!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Mathematical and Computational Sciences Division                !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module laplace_elem_mat

!----------------------------------------------------
! This module contains the elemental matrix for the Laplacian operator over
! an isoceles right triangle.  The data is for a degree 23 element.  Since
! the basis functions are p-hierarchical, elements of any degree up to 23
! (including mixed degrees on the edges) are submatrices.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
!----------------------------------------------------

implicit none
private
public laplace_elemental_matrix

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

!----------------------------------------------------
! The following variables are defined:

real(my_real), save :: lem(300,300)
!----------------------------------------------------

contains

!          ------------------------
subroutine laplace_elemental_matrix(degree,mat)
!          ------------------------

!----------------------------------------------------
! This routine sets the elemental matrix for the Laplacian operator on an
! isoceles right triangle for an element with edge1, edge2, edge3 and interior
! degrees given in degree.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: degree(4)
real(my_real), intent(out) :: mat(:,:)
!----------------------------------------------------
! Local variables:

integer :: ind(300), nind, start, num, i, j
logical, save :: first_call = .true.
!----------------------------------------------------
! Begin executable code

! if this is the first call, set up the master elemental matrix

if (first_call) then
   first_call = .false.
   call init_lem
endif

! determine the row/column indices of the submatrix for the given degrees

! linear bases

ind(1:3) = (/1,2,3/)
nind = 3
start = 4

! edges

do i=1,3
   num = degree(i)-1
   if (num > 0) then
      ind(nind+1:nind+num) = (/ (start+j,j=0,num-1) /)
      nind = nind+num
   endif
   start = start+22
end do

! interior

num = ((degree(4)-2)*(degree(4)-1))/2
if (num > 0) then
   ind(nind+1:nind+num) = (/ (start+j,j=0,num-1) /)
   nind = nind+num
endif

! copy the submatrix

mat(1:nind,1:nind) = lem(ind(1:nind),ind(1:nind))

end subroutine laplace_elemental_matrix

!          --------
subroutine init_lem
!          --------

!----------------------------------------------------
! This routine sets the degree 23 laplacian elemental matrix
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

lem(:,  1) = (/ 5.000000000000002D-01,   0.000000000000000D+00,  -4.999999999999998D-01, &
   1.632993161855453D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.632993161855451D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00 /)
lem(:,  2) = (/ 0.000000000000000D+00,   5.000000000000002D-01,  -4.999999999999998D-01, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.632993161855452D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.632993161855453D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00 /)
lem(:,  3) = (/ -4.999999999999998D-01,  -4.999999999999998D-01,   9.999999999999997D-01, &
  -1.632993161855451D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.632993161855453D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.265986323710905D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00 /)
lem(:,  4) = (/ 1.632993161855453D+00,   0.000000000000000D+00,  -1.632993161855451D+00, &
   1.599999999999999D+01,  -2.065591117977288D+00,  -1.018350154434635D+00, &
   0.000000000000000D+00,   6.382847385042202D-02,   0.000000000000000D+00, &
  -1.197893559374901D-02,   0.000000000000000D+00,   3.495293720032163D-03, &
   0.000000000000000D+00,  -1.311020180385895D-03,   0.000000000000000D+00, &
   5.794986263727882D-04,   0.000000000000000D+00,  -2.877608462748176D-04, &
   0.000000000000000D+00,   1.558791572524330D-04,   0.000000000000000D+00, &
  -9.034263192610999D-05,   0.000000000000000D+00,   5.526802731912923D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.065591117977289D+00, &
  -1.018350154434631D+00,   0.000000000000000D+00,   6.382847385042217D-02, &
   0.000000000000000D+00,  -1.197893559374915D-02,   0.000000000000000D+00, &
   3.495293720032415D-03,   0.000000000000000D+00,  -1.311020180384732D-03, &
   0.000000000000000D+00,   5.794986263737043D-04,   0.000000000000000D+00, &
  -2.877608462752944D-04,   0.000000000000000D+00,   1.558791572552085D-04, &
   0.000000000000000D+00,  -9.034263192715200D-05,   0.000000000000000D+00, &
   5.526802731974416D-05,   0.000000000000000D+00,  -7.999999999999992D+00, &
  -4.131182235954581D+00,  -1.018350154434630D+00,   0.000000000000000D+00, &
   6.382847385042240D-02,   0.000000000000000D+00,  -1.197893559374875D-02, &
   0.000000000000000D+00,   3.495293720032749D-03,   0.000000000000000D+00, &
  -1.311020180384238D-03,   0.000000000000000D+00,   5.794986263736859D-04, &
   0.000000000000000D+00,  -2.877608462753670D-04,   0.000000000000000D+00, &
   1.558791572557670D-04,   0.000000000000000D+00,  -9.034263192740976D-05, &
   0.000000000000000D+00,   5.526802732024017D-05,   0.000000000000000D+00, &
  -1.632993161855452D-01,   5.443310539518172D-02,   0.000000000000000D+00, &
   2.332847374079219D-02,   0.000000000000000D+00,   5.832118435198053D-02, &
  -2.332847374079213D-02,   0.000000000000000D+00,  -1.555231582719479D-02, &
   0.000000000000000D+00,   5.184105275731644D-03,   0.000000000000000D+00, &
  -1.166423687039608D-02,   0.000000000000000D+00,  -1.992640465359330D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.776157913597375D-03, &
   0.000000000000000D+00,   1.458029608799521D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.178205744484535D-04, &
   0.000000000000000D+00,   6.583224597306874D-03,   0.000000000000000D+00, &
   5.103103630798297D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.069234466906664D-04,   0.000000000000000D+00,  -1.192933316290501D-03, &
   0.000000000000000D+00,   1.701034543599431D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.156809000276467D-05, &
   0.000000000000000D+00,  -1.743517923809209D-03,   0.000000000000000D+00, &
  -2.258416491951676D-03,   0.000000000000000D+00,  -1.275775907699566D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.588205062655805D-04,   0.000000000000000D+00, &
  -1.323782127336613D-03,   0.000000000000000D+00,  -1.032770972899660D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.005215545901419D-04, &
   0.000000000000000D+00,   1.104001440384711D-03,   0.000000000000000D+00, &
   1.822537010999459D-04,   0.000000000000000D+00,   4.556342527498331D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.137207875242931D-05,   0.000000000000000D+00, &
   4.401278189732801D-04,   0.000000000000000D+00,   1.093522206599625D-03, &
   0.000000000000000D+00,   4.100708274748691D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.597810098239771D-06,   0.000000000000000D+00, &
  -3.547612149752827D-04,   0.000000000000000D+00,   1.033304097670920D-04, &
   0.000000000000000D+00,   1.822537010999410D-04,   0.000000000000000D+00, &
  -1.993399855780395D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.288713432709463D-05,   0.000000000000000D+00, &
  -6.663768088139803D-04,   0.000000000000000D+00,  -3.493195937748820D-04, &
   0.000000000000000D+00,  -1.866834785572354D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.015932469596923D-05,   0.000000000000000D+00,  -1.923786875041224D-05, &
   0.000000000000000D+00,  -4.556342527498466D-04,   0.000000000000000D+00, &
  -1.234009434530832D-04,   0.000000000000000D+00,   9.966999278900705D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.722512275860079D-06,   0.000000000000000D+00,   2.233645919212021D-04, &
   0.000000000000000D+00,   1.974415095249315D-04,   0.000000000000000D+00, &
   4.366494922185700D-05,   0.000000000000000D+00,   9.524021533174731D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.787209687917779D-07,   0.000000000000000D+00, &
  -3.257475653143608D-05,   0.000000000000000D+00,   3.424581332885196D-04, &
   0.000000000000000D+00,   2.575599178738652D-04,   0.000000000000000D+00, &
   7.309132804528704D-05,   0.000000000000000D+00,  -5.481849603394979D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.849477370138747D-05, &
   0.000000000000000D+00,  -1.146844109862631D-04,   0.000000000000000D+00, &
   1.378983969496708D-04,   0.000000000000000D+00,   1.182232814900278D-05, &
   0.000000000000000D+00,  -5.300631434689353D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.126461494437288D-05, &
   0.000000000000000D+00,  -1.171131519216280D-04,   0.000000000000000D+00, &
  -2.259186503217917D-04,   0.000000000000000D+00,  -8.960231674972779D-05, &
   0.000000000000000D+00,  -4.394540591152921D-05,   0.000000000000000D+00, &
   3.239274765643018D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.533296045052201D-06,   0.000000000000000D+00,   5.307657589576818D-05, &
   0.000000000000000D+00,  -1.373231011759990D-04,   0.000000000000000D+00, &
  -1.455785955282222D-04,   0.000000000000000D+00,  -1.827283201131882D-05, &
   0.000000000000000D+00,   3.156216438318735D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.359821014401073D-08, &
   0.000000000000000D+00,   9.998316850629006D-06,   0.000000000000000D+00, &
   1.354852584631371D-04,   0.000000000000000D+00,   7.852787310642701D-06, &
   0.000000000000000D+00,   2.899490699316102D-05,   0.000000000000000D+00, &
   2.740924801699078D-05,   0.000000000000000D+00,  -2.024546728527013D-05 /)
lem(:,  5) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.065591117977288D+00,   8.888888888888888D+00,  -9.014978717104166D-01, &
  -1.565247584249854D+00,   0.000000000000000D+00,   8.958064164776151D-02, &
   0.000000000000000D+00,  -1.493907666691014D-02,   0.000000000000000D+00, &
   3.881420744680548D-03,   0.000000000000000D+00,  -1.306463328200726D-03, &
   0.000000000000000D+00,   5.226386535925193D-04,   0.000000000000000D+00, &
  -2.367405092878151D-04,   0.000000000000000D+00,   1.178093986337100D-04, &
   0.000000000000000D+00,  -6.311430997323940D-05,   0.000000000000000D+00, &
   3.588529443723085D-05,  -2.065591117977287D+00,   0.000000000000000D+00, &
  -1.502496452850696D-01,  -6.708203932499371D-01,  -3.296088216486964D-01, &
   3.839170356332628D-02,   6.185895741317401D-02,  -6.402461428675744D-03, &
  -1.804961915705417D-02,   1.663466033434609D-03,   6.770079100229649D-03, &
  -5.599128549427715D-04,  -2.992518038793543D-03,   2.239879943972302D-04, &
   1.485990620419916D-03,  -1.014602182654893D-04,  -8.049565067622842D-04, &
   5.048974227103013D-05,   4.665273452124948D-04,  -2.704898998739813D-05, &
  -2.854028658510246D-04,   1.537941190077660D-05,   2.065591117977290D+00, &
   1.777777777777778D+00,   1.464934041529428D+00,   8.944271909999155D-01, &
   2.472066162365217D-01,  -5.118893808443566D-02,  -4.639421805988072D-02, &
   8.536615238233881D-03,   1.353721436778978D-02,  -2.217954711247210D-03, &
  -5.077559325172566D-03,   7.465504732569271D-04,   2.244388529095017D-03, &
  -2.986506591962830D-04,  -1.114492965314779D-03,   1.352802910205458D-04, &
   6.037173800710975D-04,  -6.731965636222378D-05,  -3.498955089095780D-04, &
   3.606531998327588D-05,   2.140521493879518D-04,  -2.050588253491626D-05, &
   7.027283689263070D-02,  -9.035079029052503D-02,   4.015590679578893D-02, &
   4.015590679578893D-02,  -2.007795339789445D-02,  -1.254872087368405D-02, &
   2.677060453052602D-02,   0.000000000000000D+00,   2.760718592210493D-02, &
  -3.513641844631533D-02,  -3.011693009684170D-02,   6.692651132631502D-03, &
  -1.254872087368407D-02,   1.505846504842085D-02,  -8.156668567894623D-03, &
   7.605285377990327D-03,  -4.258959811674599D-03,  -1.133187521320558D-02, &
   3.194219858755948D-03,  -6.331400077176955D-03,   1.619925785511941D-02, &
   0.000000000000000D+00,   1.216845660478461D-03,   9.810818137607530D-03, &
  -6.844756840191302D-03,   3.821655902440145D-03,  -4.106854104114777D-03, &
   8.784104611578821D-03,   0.000000000000000D+00,  -1.404052685167515D-04, &
  -4.855682202870649D-04,   2.702801418947327D-03,   5.120404636220023D-03, &
  -4.545620568229600D-03,   1.728739868612421D-03,  -4.392052305789432D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.000387538181802D-03, &
   1.053039513875566D-04,  -2.013938070287046D-03,   3.264422493014302D-03, &
  -1.825268490717678D-03,  -6.274360436841974D-04,  -4.156763789407833D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.228546099521398D-04, &
  -3.861144884210415D-04,  -1.908634118899499D-03,   4.914184398086127D-04, &
  -1.942272881148305D-03,   2.470255192966473D-03,  -1.647019614671026D-03, &
   9.411540655263107D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.053039513875547D-04,   6.888633486602783D-04, &
  -1.158343465263143D-03,  -3.970836500239194D-04,   8.921584770335169D-05, &
   9.019393127960350D-04,   7.058655491447259D-04,   1.617608550123334D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.259133442163677D-06,   2.655827597495129D-04,   2.642922701491596D-04, &
   1.279176308174767D-03,  -1.472534666625372D-03,   1.148530367260575D-03, &
  -4.705770327631542D-04,   1.147031517360180D-03,  -2.941106454769790D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.047877555474345D-04,   1.300813517140498D-04, &
   3.694671725766874D-04,   3.873189453813615D-04,   4.404441005952622D-04, &
  -8.627245600657771D-04,  -2.941106454769573D-05,  -2.548958927467073D-04, &
  -7.311917436163625D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.557983665404428D-06, &
  -7.721203040336519D-05,  -4.692152266322746D-04,   4.539534890956474D-04, &
  -3.124218721996575D-04,   4.928023201037039D-04,  -8.137061191529496D-04, &
  -5.882212909539368D-05,  -6.004759011821413D-04,   1.143763621299396D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.369277386463765D-05, &
  -3.418439030788403D-05,  -2.679599774343459D-04,  -6.168786501445891D-04, &
   5.466989390567569D-04,  -4.411659682154517D-04,   3.725401509374939D-04, &
  -2.361054903967900D-04,   1.045726739473659D-04,   3.717231769222938D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.150460633783014D-07, &
   7.139124310078470D-05,  -1.558096391144162D-05,   5.913413891652684D-05, &
  -4.418517530672113D-04,   6.189404619830340D-05,   2.156811400164473D-04, &
   3.325084241920137D-04,   5.555423303454015D-05,   3.259726320702949D-04, &
  -5.146936295846311D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.456894838276560D-05,   5.095396306175414D-05, &
   2.695348841505320D-04,  -1.173360688980697D-04,   3.158590035194950D-04, &
  -4.295797912724219D-04,   4.761473101396069D-04,  -7.011122457834909D-05, &
   1.930286786855197D-04,  -4.835000762766309D-05,  -2.064623309584095D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.933022977161586D-07,  -1.620937623856301D-05,  -4.905894405109629D-05, &
   1.989903956533517D-04,   2.401784328569584D-04,  -1.185487267672309D-04, &
   6.914570983309882D-05,  -2.002329040924052D-04,  -5.651974777978774D-05, &
  -3.275323097356671D-05,  -1.877461989735025D-04,   2.573468147922215D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.111441794591643D-06,  -3.911445924501062D-05, &
  -3.727750488371213D-05,  -1.590905409677209D-04,   3.266110133003947D-04, &
  -1.790658500618580D-04,   2.346943534614288D-05,  -3.019090338040091D-04, &
   9.580877087506947D-06,  -1.321826821433429D-04,   2.456492323019380D-05, &
   1.227758762238410D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.013492061256229D-07, &
   1.652203204275829D-05,  -2.736428565805347D-05,  -1.277989735717054D-04, &
  -3.789668518997873D-05,  -1.956970692386510D-04,   2.703441286707562D-04, &
  -2.115962699403794D-04,   7.330485027418766D-05,  -1.910605140125611D-05, &
   1.871613198489092D-05,   1.141976490641087D-04,  -1.393961913458059D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.327864578552854D-06, &
   1.593082834028412D-05,   7.054142284610504D-05,  -1.221780515842493D-04, &
  -4.927759839892809D-05,  -6.030753639578801D-05,   8.810179222805745D-05, &
   7.930361052575366D-05,   1.305180261735334D-04,   2.369510299370034D-06, &
   8.840297998234821D-05,  -1.344472141383848D-05,  -7.707000963832286D-05 /)
lem(:,  6) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.018350154434635D+00,  -9.014978717104166D-01,   6.361111111111108D+00, &
  -5.039526306789690D-01,  -1.730617976402338D+00,   0.000000000000000D+00, &
   9.352375699090085D-02,   0.000000000000000D+00,  -1.468266071080898D-02, &
   0.000000000000000D+00,   3.629742233595467D-03,   0.000000000000000D+00, &
  -1.177532936345358D-03,   0.000000000000000D+00,   4.595420421538579D-04, &
   0.000000000000000D+00,  -2.052236144701433D-04,   0.000000000000000D+00, &
   1.015838116292139D-04,   0.000000000000000D+00,  -5.453134373300951D-05, &
   0.000000000000000D+00,  -1.018350154434630D+00,  -1.502496452850690D-01, &
   8.333333333333336D-02,  -4.724555912615312D-02,  -2.924988129130704D-01, &
  -3.217652068271438D-01,  -9.708172600685953D-02,   5.365975287870562D-02, &
   3.661130982435148D-02,  -1.394169684122681D-02,  -1.496598152472890D-02, &
   4.692692922018772D-03,   6.895807729312744D-03,  -1.877268697522491D-03, &
  -3.506505582541178D-03,   8.503495569321089D-04,   1.928186170854954D-03, &
  -4.231602366329061D-04,  -1.128921982095002D-03,   2.267006423275972D-04, &
   6.956327631404442D-04,  -1.288965894155696D-04,   0.000000000000000D+00, &
  -4.883113471764755D-01,  -6.666666666666663D-01,  -6.771863474748648D-01, &
  -5.849976258261407D-01,  -3.217652068271441D-01,  -4.879500364742646D-02, &
   5.365975287870620D-02,   2.288206864022038D-02,  -1.394169684122625D-02, &
  -9.870037994802818D-03,   4.692692922019044D-03,   4.655489845780055D-03, &
  -1.877268697522642D-03,  -2.397610654729044D-03,   8.503495569322695D-04, &
   1.328750055563838D-03,  -4.231602366321219D-04,  -7.820036819937914D-04, &
   2.267006423280532D-04,   4.836194827765232D-04,  -1.288965894156855D-04, &
   3.266526290040733D-02,   8.908708063747458D-03,  -1.484784677291246D-02, &
  -6.087617176894103D-02,   2.227177015936870D-02,  -2.783971269921089D-02, &
   4.107904273839111D-02,  -1.781741612749494D-02,   8.537511894424671D-03, &
   1.855980846614052D-03,   1.849231825353645D-02,   1.574771627430101D-03, &
   1.872853399765093D-02,  -1.467912124140209D-02,   2.093883846043676D-02, &
  -2.857085666908914D-02,   8.323792887844869D-03,  -1.544401031758242D-02, &
   1.440916039098550D-02,  -1.158019564599497D-02,   1.172642443997065D-02, &
   8.098825512497703D-03,  -6.679800529538722D-03,  -6.673743715587053D-03, &
   2.725566278244475D-04,  -5.583517204289276D-03,   2.920249583833310D-03, &
  -6.126359154706293D-03,   0.000000000000000D+00,   2.232368570752573D-03, &
   8.961488871040474D-03,  -8.241593269929562D-03,   7.146175203816105D-03, &
  -7.002109557680321D-03,   5.001089648402531D-03,  -1.090388747385759D-02, &
   0.000000000000000D+00,  -2.907270696794048D-04,  -6.203908004765873D-04, &
   4.711335995251068D-03,   2.502329421169158D-03,  -1.362783139122209D-03, &
   2.486430284546069D-03,  -3.488075891800928D-04,  -1.832781086031375D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.072056069442807D-03, &
  -2.465988537459237D-04,  -2.128537474438497D-03,   4.490694915583660D-03, &
  -3.231851030167527D-03,   3.690871001789317D-03,  -1.113588507968434D-03, &
   4.871949722361907D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368127386726610D-04,  -6.243608129827166D-04,  -1.590906165435777D-03, &
  -5.107573781920161D-04,  -1.521621812564434D-03,   1.128813251441446D-03, &
  -9.352838897921411D-04,   1.304986532775509D-03,   2.327225983449661D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.046083368542005D-04,   6.908330736622116D-04,  -1.572735723580817D-03, &
   4.414221277244500D-04,  -1.591695078177376D-03,   8.710638477459250D-04, &
  -1.913980248070749D-03,   8.608077832863622D-04,  -1.674732717061904D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.735875913609479D-05,   2.902489001569056D-04,   6.050250840561144D-04, &
   1.024771532837700D-03,  -1.155144386176540D-03,   1.383936740994508D-03, &
  -1.240349515520686D-03,  -3.334965583759647D-05,  -1.080238852130842D-03, &
  -1.313444778821275D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.174328092019663D-04, &
   1.610306186062914D-04,   2.682390383124669D-04,   9.922318959200761D-04, &
  -4.660273484109966D-05,   5.573958585488964D-04,  -1.928480098434906D-04, &
   4.306455558159209D-04,  -8.344663884581160D-04,   6.754513627976937D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.777696118332080D-06,  -1.388700730887959D-04, &
  -4.508284723018404D-04,   4.076997913996403D-04,  -5.121630057221776D-04, &
   9.352533763483443D-04,  -7.084580720131442D-04,   1.123738403223355D-03, &
  -2.223310389173122D-05,   5.268278965649334D-04,   7.337528444702977D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.742442849627292D-05, &
  -4.603788035002149D-05,  -4.686212004163438D-04,  -3.967794944737855D-04, &
   2.234827226018523D-05,  -3.831899304570329D-04,   7.447650414324941D-05, &
   3.033544451946825D-04,   1.016527388409428D-04,   5.557891507203426D-04, &
  -3.096871449938922D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.729586508127394D-06,   8.118639760358376D-05,   2.360121246119579D-05, &
   6.525289855156509D-05,  -6.609413383283372D-04,   3.403048329880564D-04, &
  -6.461141809150894D-04,   2.653472616643543D-04,  -5.316611800196500D-04, &
   1.902720894568217D-04,  -2.623978602834577D-04,  -4.288907443797990D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.673134726262293D-05,   8.259976678042103D-05,   2.550836220744938D-04, &
   3.532139947062909D-05,   2.919196058368769D-04,  -3.185606193617972D-04, &
   3.409222393200400D-04,  -4.505938348017801D-04,  -2.122635291758168D-04, &
  -1.208595482059155D-04,  -3.538430340100840D-04,   1.565736682790405D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.037751904855668D-06,  -3.072850652032822D-05,  -4.193256373268675D-05, &
   2.854096994562599D-04,   1.053369080134939D-04,   2.686800247277972D-04, &
  -1.594222992427175D-05,   2.938416645769784D-04,  -2.612444630954037D-04, &
   1.412911555372504D-04,  -1.866484999576725D-04,   1.388401865126172D-04, &
   2.631836121178345D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.309553049407641D-06, &
  -4.633159886185783D-05,  -9.232748978689187D-05,  -1.277366747810731D-04, &
   3.249104375337216D-04,  -2.975299651300409D-04,   4.037819635794525D-04, &
  -1.516987695777749D-04,   3.499039502434624D-04,   5.136641702326943D-05, &
   8.450852474049048D-05,   2.281613719238571D-04,  -8.542625183002900D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.165186073198546D-07,   1.917100510213443D-05, &
  -3.668975462993597D-05,  -1.228668191128645D-04,  -1.735241144988332D-04, &
  -1.115199971124875D-04,  -1.345200570145134D-05,  -1.045717209553425D-04, &
   7.581368439973742D-05,   2.228045443746431D-04,  -2.596252708891148D-05, &
   1.466084274777781D-04,  -7.792362492885448D-05,  -1.687023605847237D-04 /)
lem(:,  7) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.565247584249854D+00,  -5.039526306789690D-01, &
   5.350000000000001D+00,  -3.216121008829468D-01,  -1.817798768046428D+00, &
   0.000000000000000D+00,   9.735110504930486D-02,   0.000000000000000D+00, &
  -1.513421660621134D-02,   0.000000000000000D+00,   3.714276175212994D-03, &
   0.000000000000000D+00,  -1.197871946034043D-03,   0.000000000000000D+00, &
   4.646187436721097D-04,   0.000000000000000D+00,  -2.059311987459160D-04, &
   0.000000000000000D+00,   1.009607194824246D-04,   0.000000000000000D+00, &
  -5.355644953271371D-05,   0.000000000000000D+00,  -6.708203932499371D-01, &
  -4.724555912615332D-02,   3.000000000000005D-01,   1.356801050599954D-02, &
  -2.038853399815948D-01,  -2.256358601250308D-01,  -1.360052202894715D-01, &
  -8.778338150186325D-03,   4.372881941057175D-02,   1.440509291004489D-02, &
  -1.585870726495718D-02,  -8.894083870854171D-03,   6.582035048402992D-03, &
   5.157477039381174D-03,  -3.045896485452308D-03,  -3.052498847696861D-03, &
   1.536613662743511D-03,   1.871868416990101D-03,  -8.309485944516132D-04, &
  -1.190198870581518D-03,   4.756358903217270D-04,   0.000000000000000D+00, &
   2.236067977499785D-01,   1.574851970871775D-01,   1.899999999999996D-01, &
   3.507581975254649D-01,   3.970398725957363D-01,   2.887447405431415D-01, &
   1.145307118227122D-01,  -1.392426051408886D-02,  -3.867943101292419D-02, &
  -5.251117612529084D-03,   1.423110309829029D-02,   4.702619058152758D-03, &
  -5.945952523750675D-03,  -3.033543490327975D-03,   2.761837810446773D-03, &
   1.887102272009888D-03,  -1.396576105348634D-03,  -1.190537000845088D-03, &
   7.564144596072882D-04,   7.707975216812961D-04,  -4.334582898400869D-04, &
  -1.767766952966368D-02,   3.788072042070788D-02,  -1.094331923264895D-02, &
  -1.565736444055929D-02,   1.683587574253706D-04,   9.091372900969900D-03, &
  -4.568644462861136D-02,   1.532064692570853D-02,  -2.231518802965338D-02, &
   1.821335648510804D-02,   4.655374731837839D-02,  -1.819805114352389D-02, &
   1.584102853956875D-02,  -7.688383255758495D-03,   2.515815522890449D-03, &
   7.964664293584755D-03,   4.491529087956509D-03,   1.328150449311595D-02, &
  -1.018629349121880D-02,   1.113537183618050D-02,  -1.312212290719667D-02, &
  -2.484763339484195D-02,   7.635010782542063D-03,  -1.842095970813370D-02, &
   1.589754057003323D-02,  -1.353006912711197D-02,   9.436086475686683D-03, &
  -1.101763843937954D-02,   7.903442927248235D-03,  -7.630301446670025D-03, &
  -1.835463656126919D-03,  -3.172915043785789D-03,  -2.880789052973412D-03, &
   2.456948824489967D-03,  -7.144945518358076D-04,   3.657093638129716D-04, &
   0.000000000000000D+00,   2.859744208295248D-03,   7.385415981324030D-03, &
  -7.646784122222160D-03,   8.887399791012262D-03,  -8.368637011358026D-03, &
   7.036410043182189D-03,  -3.986379232307401D-03,   7.704386106399180D-03, &
   0.000000000000000D+00,  -4.052799047524740D-04,  -5.696218756987487D-04, &
   5.715817904807867D-03,   4.440972982271669D-05,   8.520088906728073D-04, &
   5.213814820462987D-04,   1.515817483812315D-04,  -1.155974500306397D-03, &
   5.336841080104939D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.089310940166038D-03,  -6.219324398459942D-04,  -1.751197708740120D-03, &
   4.614196899182208D-03,  -4.107474379286623D-03,   3.509602259600582D-03, &
  -3.380499961143928D-03,   9.853810086289472D-04,  -1.604669406710546D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.426032100013551D-04, &
  -7.361909396928648D-04,  -1.331033459534501D-03,  -1.332114201087168D-03, &
  -5.428186195056099D-04,   3.505427599197362D-04,  -2.177423559155897D-04, &
  -9.516375456286083D-04,  -2.630605584771129D-05,  -3.893296265461645D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.750514589040308D-04,   6.208272464397597D-04,  -1.614832303271025D-03, &
   7.751609674087481D-04,  -2.138337375643392D-03,   1.767143122204593D-03, &
  -1.318651846774303D-03,   1.530354798940750D-03,  -1.141463606658717D-03, &
  -6.044638395257460D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.457335346027104D-05,   3.154448578722533D-04, &
   8.410215034287829D-04,   6.904193479388978D-04,  -7.077137187170415D-04, &
   9.922301379769200D-04,  -6.005969683717941D-04,   5.007018916768700D-04, &
   1.016729058514139D-03,  -6.349076187412151D-05,   1.867318933065065D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.257409991569144D-04,   1.510008270154434D-04, &
   2.687683956614321D-04,   1.291711689626944D-03,  -4.881433114741441D-04, &
   8.780762419446073D-04,  -8.931618300885401D-04,   8.693553592763828D-04, &
  -2.842473960703973D-04,   1.080663236859929D-03,   6.991656405775169D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.056306207248804D-05,  -1.785115724661498D-04, &
  -4.207956656630166D-04,   2.700623695272082D-04,  -4.029766673004104D-04, &
   8.751144653712434D-04,  -7.976732953530759D-04,   2.999385670612456D-04, &
  -4.997328546820364D-04,  -3.290274775020716D-04,   3.687207250283492D-04, &
  -9.241123113207545D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.780446074840797D-05,  -6.583552619140819D-05,  -5.746463750309471D-04, &
  -2.492739430694097D-04,  -2.992332568778610D-04,   2.257628562859129D-05, &
  -2.566977299766853D-04,   2.958708862773525D-04,  -8.706183488895273D-04, &
   5.238791235586382D-06,  -6.608026424662702D-04,  -5.226903999762667D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.473566147000109D-06, &
   9.011659316873932D-05,   6.510651692554574D-05,   1.196451939107515D-05, &
  -6.982845583213128D-04,   4.178283871749863D-04,  -6.214198789673220D-04, &
   6.300712800446767D-04,  -2.591370700339375D-04,   9.322955872164833D-05, &
  -9.756407928746942D-05,  -3.568249260401622D-04,   4.873463082504794D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.843006461960731D-05,   9.970865978801879D-05,   2.537566267505169D-04, &
   1.655899917281754D-04,   1.443094869873736D-04,  -1.536004613529017D-04, &
   1.660308540453100D-04,   1.562008296473229D-04,   5.485314697470765D-05, &
   5.209961499988049D-04,  -9.477362388508538D-05,   3.859953919335351D-04, &
   3.636092914731818D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.146810421401124D-06,  -4.109904591763472D-05, &
  -3.016166202904593D-05,   3.123744182020544D-04,   7.326130005941249D-05, &
   4.165159684368001D-04,  -2.384179672821541D-04,   3.221326874396685D-04, &
  -3.676208069006813D-04,   4.419994168751149D-04,   7.659479409828247D-05, &
   1.517723537492806D-04,   2.790379950687630D-04,  -2.729767084353653D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.034800661291727D-06,  -5.452823290563587D-05, &
  -1.330999730520214D-04,  -8.066314063569990D-05,   2.653084119526107D-04, &
  -2.260868276474280D-04,   2.648180341573946D-04,  -2.807809293579484D-04, &
  -9.253396589875990D-05,   9.734526935502892D-06,  -1.843092829103558D-04, &
   1.253064121545688D-04,  -2.294222141308357D-04,  -2.513300870401680D-04 /)
lem(:,  8) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.382847385042202D-02,   0.000000000000000D+00,  -1.730617976402338D+00, &
  -3.216121008829468D-01,   4.830555555555556D+00,  -2.229978693352244D-01, &
  -1.864979948301234D+00,   0.000000000000000D+00,   9.893057979515101D-02, &
   0.000000000000000D+00,  -1.514092550541258D-02,   0.000000000000000D+00, &
   3.650903219678518D-03,   0.000000000000000D+00,  -1.156686757567135D-03, &
   0.000000000000000D+00,   4.411940769698106D-04,   0.000000000000000D+00, &
  -1.926480304397434D-04,   0.000000000000000D+00,   9.325988411759929D-05, &
   0.000000000000000D+00,   6.382847385042226D-02,  -3.296088216486962D-01, &
  -2.924988129130706D-01,   1.356801050599943D-02,   2.250000000000005D-01, &
   1.065047115522923D-01,  -1.197868911105672D-01,  -1.891363336314457D-01, &
  -1.317680026869882D-01,  -4.256639051532686D-02,   2.014606046523200D-02, &
   2.678314039687541D-02,   1.349395749116873D-04,  -1.331403249278493D-02, &
  -3.005564015725605D-03,   6.734834171467675D-03,   2.707506803805436D-03, &
  -3.579641041667425D-03,  -2.010716279485517D-03,   2.002251953027421D-03, &
   1.427486785941615D-03,  -1.173163348824907D-03,   0.000000000000000D+00, &
  -8.240220541217370D-02,  -1.462494064565351D-01,  -7.487531723681105D-02, &
  -4.166666666666643D-02,  -1.541240482332509D-01,  -2.618745332257619D-01, &
  -2.461146433446844D-01,  -1.392976028405296D-01,  -2.659050092208462D-02, &
   2.877560349940494D-02,   2.124657390656561D-02,  -4.933856640030094D-03, &
  -1.106595713738014D-02,  -1.286472476424718D-04,   5.707523475917247D-03, &
   1.024894754007463D-03,  -3.065503440751186D-03,  -9.865632099430414D-04, &
   1.725731632955080D-03,   7.795760120681173D-04,  -1.015496635595740D-03, &
  -6.514466333088092D-03,  -3.908679799852864D-03,   6.514466333088093D-03, &
   3.460281467835103D-02,  -1.101537034503987D-02,   7.385881959462212D-03, &
  -2.923331515532083D-02,   6.751356017927672D-03,  -8.206535510513618D-04, &
  -4.737793696791341D-03,  -3.181740849397368D-02,   9.545439480137950D-03, &
  -1.819429922818621D-02,   1.331137805963875D-02,  -1.129537525411192D-02, &
   4.818275448691968D-02,  -1.848043246472778D-02,   2.086169443403538D-02, &
  -1.249267689609981D-02,   7.404050009924961D-03,  -4.789043868992216D-03, &
  -5.193350783021276D-04,   8.124101371100546D-03,   7.034452207227690D-03, &
  -5.002016806804699D-03,   8.181317171811572D-03,  -6.269604399239508D-03, &
   6.597036054967279D-03,  -2.091007025795408D-02,   5.539574168556037D-03, &
  -1.899434424438432D-02,   1.680192153391626D-02,  -1.606310555942694D-02, &
   1.158538614918509D-02,  -8.151624611726926D-03,   9.896265890398149D-03, &
   7.447554435589641D-03,  -7.596548400262501D-03,   1.878464761698892D-03, &
  -7.217230132286926D-03,   8.987681926019652D-04,  -6.044505603690621D-04, &
  -2.412729862154746D-04,  -1.246040747161467D-03,   3.084316174185770D-03, &
   0.000000000000000D+00,   3.178716562858533D-03,   5.737071564174548D-03, &
  -5.754891885488829D-03,   9.122147830673610D-03,  -9.368899943608983D-03, &
   8.247672825828252D-03,  -5.959800457969717D-03,   2.916607642080042D-03, &
  -4.555037006338940D-03,   0.000000000000000D+00,  -4.836239160007518D-04, &
  -4.231462446151901D-04,   5.985152721086046D-03,  -1.947756384759439D-03, &
   3.531422469314607D-03,  -1.361046055504859D-03,   1.662161175542877D-03, &
  -1.815311828931800D-03,   1.790091758719960D-03,  -5.781952401103021D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.088851955762342D-03, &
  -9.379045961531768D-04,  -1.146520593515068D-03,   4.009202067432807D-03, &
  -4.486582054071863D-03,   4.223369579119179D-03,  -3.173171958832340D-03, &
   2.274983926961075D-03,  -3.030632496103632D-04,  -1.348698108022148D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.451899320377342D-04, &
  -7.695257808539851D-04,  -1.165877068368321D-03,  -1.981831994364005D-03, &
   4.875553877588612D-04,  -8.156709373967155D-04,   3.561398524780272D-04, &
  -9.345735332967045D-04,   1.782788368344066D-03,  -4.478695603998043D-04, &
   2.821026875946309D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.210408103810917D-04,   5.331463562182404D-04, &
  -1.453516235731943D-03,   8.298755094046218D-04,  -2.456647562485214D-03, &
   2.203855025062344D-03,  -1.900373508266228D-03,   1.126738673580804D-03, &
  -4.686899428564079D-04,   5.694787704707334D-04,   2.390777528107005D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.986026331270597D-05,   3.410018563753100D-04,   9.822991830333065D-04, &
   3.739056531334059D-04,  -1.298645644435176D-04,   5.636317185307455D-04, &
  -2.421124091742313D-04,   9.991910696933763D-06,   2.463751078112073D-04, &
  -1.255124809777704D-03,   6.254404333699613D-04,  -6.047068025069209D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.324351789833256D-04,   1.257798848636751D-04, &
   3.242214775794179D-04,   1.404625502311613D-03,  -8.019251664438530D-04, &
   1.331231506104650D-03,  -1.038855202648405D-03,   9.530745395246461D-04, &
  -7.655318638993842D-04,  -3.857939756680879D-04,  -4.483033183676518D-04, &
  -1.584879321514225D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.118319367554957D-05, &
  -2.025975776494699D-04,  -3.970075803822169D-04,   1.083158220376790D-04, &
  -2.068948886887463D-04,   6.839264855136833D-04,  -7.171611115421591D-04, &
   3.810730891905964D-04,   1.016415041515241D-04,  -2.036156291728076D-04, &
   3.726944850391351D-04,  -8.201640242019322D-04,  -5.246664086069652D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.523936656907345D-05, &
  -8.677583710206977D-05,  -6.179489264001202D-04,  -1.584186705073535D-04, &
  -5.802147166133445D-04,   2.981398148095012D-04,  -5.515803285148335D-04, &
   4.725239694563815D-04,  -3.510399526443746D-04,   7.080964831475015D-04, &
   2.447785188547742D-04,   5.235171353587857D-05,   9.543439930052776D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.042234014581550D-06, &
   9.860896154459257D-05,   9.922108380217018D-05,  -5.923479021266602D-05, &
  -6.393522331592311D-04,   4.232727791782635D-04,  -6.813157346445992D-04, &
   5.296191011565833D-04,  -3.366465556690712D-04,   1.444747692901812D-05, &
   4.788854795006237D-04,   6.694053253776195D-06,   6.268511662820535D-04, &
   2.039406232570550D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.996036342805378D-05,   1.085764765456415D-04, &
   2.632841916257333D-04,   2.659392436184982D-04,  -3.811784620575297D-06, &
   5.717752747720806D-05,   4.634631907081945D-05,   1.799131700087649D-04, &
  -3.622694644786036D-04,   1.398090906360704D-04,  -3.228972563853270D-04, &
   5.436413336289007D-05,   1.052009048265464D-04,  -5.761278987921154D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.243530376981887D-06,  -4.832397472197545D-05,  -1.922252313409928D-05, &
   3.086018478635570D-04,   8.399927115327128D-05,   5.027424501465387D-04, &
  -3.371899396320261D-04,   4.575584678572749D-04,  -3.076600152374521D-04, &
   1.347485810111982D-04,  -2.375737804721874D-04,  -4.131252691253750D-04, &
   6.843505117698049D-06,  -4.270361095941767D-04,  -2.087326408705703D-04 /)
lem(:,  9) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.958064164776151D-02,   0.000000000000000D+00, &
  -1.817798768046428D+00,  -2.229978693352244D-01,   4.533730158730155D+00, &
  -1.636833997804410D-01,  -1.894975177725838D+00,   0.000000000000000D+00, &
   1.002400936831037D-01,   0.000000000000000D+00,  -1.527613833269480D-02, &
   0.000000000000000D+00,   3.671892851487647D-03,   0.000000000000000D+00, &
  -1.161460122785426D-03,   0.000000000000000D+00,   4.428397574275795D-04, &
   0.000000000000000D+00,  -1.934187247008613D-04,   0.000000000000000D+00, &
   9.366838538858191D-05,   0.000000000000000D+00,   3.839170356332654D-02, &
  -3.217652068271440D-01,  -2.038853399815945D-01,   1.065047115522916D-01, &
   1.641156462585035D-01,   9.130692996960049D-02,  -3.249187272711887D-02, &
  -1.363931569888010D-01,  -1.344109478801964D-01,  -6.283035863241929D-02, &
   1.247843284616518D-03,   2.344111371446145D-02,   1.126075805391779D-02, &
  -6.718575955263457D-03,  -8.574383115809421D-03,   1.243351963088134D-03, &
   5.442939695039873D-03,   3.945107522582499D-04,  -3.362954801762998D-03, &
  -7.779325666575627D-04,   2.098871332575246D-03,   0.000000000000000D+00, &
  -1.279723452110892D-02,   1.135641906448745D-01,   1.030157507275422D-01, &
   1.324049849177861D-02,   2.409297052153686D-03,   6.606251054196632D-02, &
   1.528577463994489D-01,   1.923601513902547D-01,   1.467698259176837D-01, &
   5.748179794992284D-02,  -9.694699765066475D-03,  -2.522267346138642D-02, &
  -7.053559979936267D-03,   8.890030595312579D-03,   6.451878119401338D-03, &
  -2.882103181690664D-03,  -4.317750360103861D-03,   7.420984415454872D-04, &
   2.734901963398710D-03,  -2.788855207871053D-07,  -1.731479327974951D-03, &
   1.011710220950957D-03,  -9.105391988558535D-03,   1.011710220950949D-03, &
   7.081971546656649D-03,   3.035130662852849D-03,  -3.540985773328326D-03, &
   2.685866598230851D-02,  -1.097316470416030D-02,   8.380215081932863D-03, &
  -4.064528125428814D-03,  -3.733422962208509D-02,   1.116452381355095D-02, &
  -7.315948485867576D-03,  -3.835604683825017D-04,   2.634514640046224D-03, &
  -1.792542981223349D-02,   4.360494635287426D-03,  -1.273498241993219D-02, &
   1.177515477441461D-02,  -8.641101895901860D-03,   6.720267455425058D-03, &
   4.578165622219293D-02,  -1.777685698258335D-02,   2.376053504928658D-02, &
  -1.612773226345084D-02,   1.101864618262463D-02,  -5.448112601545737D-03, &
   4.212879864558637D-03,  -6.318715049238302D-03,   1.118000138857520D-02, &
   2.886835280130154D-04,   2.203919485846590D-04,   3.825616594436772D-03, &
  -4.646100236449034D-03,   2.649509779453308D-03,  -1.960422648937503D-03, &
  -1.738676660258689D-02,   3.054274500846983D-03,  -1.727814613048976D-02, &
   1.625909347478019D-02,  -1.732635501133912D-02,   1.325742737034643D-02, &
  -9.516451287118763D-03,   5.897446570769503D-03,  -6.873874971409081D-03, &
   6.908735145403053D-03,  -7.031614929324240D-03,   4.152526259457918D-03, &
  -1.059452631115616D-02,   5.328259635831445D-03,  -4.070825440558472D-03, &
   2.444698731405115D-03,  -1.944801200471323D-03,   2.798529838263545D-03, &
  -5.216217651034296D-03,   0.000000000000000D+00,   3.288118453278781D-03, &
   4.327306871899803D-03,  -3.390726358408680D-03,   7.750943546768650D-03, &
  -9.163000101160890D-03,   8.835410345662632D-03,  -6.296707758168897D-03, &
   4.217619535368518D-03,  -1.425916551410405D-03,   1.266613772713980D-03, &
   0.000000000000000D+00,  -5.329953652715237D-04,  -2.537246936373612D-04, &
   5.751136863190166D-03,  -3.155349146640957D-03,   5.913382884992159D-03, &
  -3.890226428365809D-03,   3.398700219609800D-03,  -3.150677525622295D-03, &
   2.990147212542550D-03,  -2.135788714895344D-03,   5.123758990089480D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.079638303183017D-03, &
  -1.156894234558638D-03,  -5.123522969114199D-04,   2.906780428658192D-03, &
  -4.001777433634185D-03,   4.310031237394137D-03,  -3.541960402630527D-03, &
   1.898337709982869D-03,  -9.313616563581059D-04,  -3.695113502301330D-04, &
   3.171639089475300D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.461560075141252D-04,  -7.616816074385421D-04,  -1.094547566632560D-03, &
  -2.361848989779803D-03,   1.392079214957480D-03,  -2.094380767249089D-03, &
   1.507636965098310D-03,  -1.660941391791240D-03,   1.737823315653702D-03, &
  -2.024219028732910D-03,   5.182050887585635D-04,  -1.047113492029149D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.494202112811110D-04,   4.573218065077669D-04,  -1.207149538753379D-03, &
   6.258724432571664D-04,  -2.386826415283799D-03,   2.398777672575393D-03, &
  -2.106470217276062D-03,   1.525616432823428D-03,  -4.854075984194630D-04, &
  -4.200040493196688D-04,  -1.037996962605657D-04,  -2.705962575014154D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.359776627597792D-05,   3.627806005020972D-04,   1.045253120827275D-03, &
   1.488583813451616D-04,   4.375092660923345D-04,  -7.749044546155490D-05, &
   3.359538302579648D-04,  -4.037443736877909D-04,   7.917174396983856D-04, &
  -6.781915852637902D-04,   9.367421103549182D-04,  -6.215856342410371D-04, &
  -9.088461291057721D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.378634179311442D-04, &
   1.002634706568157D-04,   4.026090159074689D-04,   1.355377635040863D-03, &
  -9.101090641088653D-04,   1.629137520444256D-03,  -1.392236719632856D-03, &
   1.090394567498388D-03,  -7.461830739188064D-04,   3.708367418388800D-04, &
   7.978247182715570D-04,  -9.266283705637097D-05,   1.206987343473593D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.169706246618103D-05,  -2.169827185614590D-04, &
  -3.862121167704638D-04,  -2.901525444177292D-05,   3.074070018700853D-05, &
   3.526426453661653D-04,  -4.442094959284898D-04,   1.640252606037143D-04, &
   5.397962926731126D-05,  -5.001060462611574D-04,   5.006099681836830D-04, &
  -9.356619910423914D-06,   6.441396741261801D-04,   1.016004861126284D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.049712245617803D-05, &
  -1.041006128514723D-04,  -6.216587001902133D-04,  -1.388413149327845D-04, &
  -7.601246855187153D-04,   5.652803031655260D-04,  -8.389428376124613D-04, &
   7.291652091528369D-04,  -5.915706465809004D-04,   2.200783402743987D-04, &
  -2.657722729105922D-04,  -3.625698543195719D-04,   4.441666296663061D-04, &
  -4.440660316289193D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.469984832119781D-06,   1.060444935852683D-04,   1.223784230601403D-04, &
  -1.235959428876170D-04,  -5.219963801956276D-04,   3.093098723660578D-04, &
  -6.177915385981686D-04,   5.184525806242690D-04,  -2.226630539081674D-04, &
   5.236780950382938D-05,   1.739101487319150D-04,  -6.217081161685184D-04, &
  -1.665348613763635D-04,  -3.298898346755977D-04,  -7.779001945958573D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.132122244227582D-05,   1.135122910187729D-04,   2.804327615848416D-04, &
   3.233308880611200D-04,  -1.219864869647827D-04,   2.794153366049593D-04, &
  -1.798589439918819D-04,   3.549504776496688D-04,  -3.544671129332269D-04, &
   4.175380839589966D-04,  -9.884643170226467D-05,  -1.155154478450638D-04, &
   4.895437008469072D-06,  -4.626635443805377D-04,   1.157005934204661D-04 /)
lem(:, 10) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.197893559374901D-02,   0.000000000000000D+00,   9.352375699090085D-02, &
   0.000000000000000D+00,  -1.864979948301234D+00,  -1.636833997804410D-01, &
   4.346749441964285D+00,  -1.252448582170303D-01,  -1.914575246825788D+00, &
   0.000000000000000D+00,   1.009957973947408D-01,   0.000000000000000D+00, &
  -1.530121851232757D-02,   0.000000000000000D+00,   3.650497922236521D-03, &
   0.000000000000000D+00,  -1.144938432196503D-03,   0.000000000000000D+00, &
   4.326006019816508D-04,   0.000000000000000D+00,  -1.872024821219416D-04, &
   0.000000000000000D+00,  -1.197893559374872D-02,   6.185895741317434D-02, &
  -9.708172600685915D-02,  -2.256358601250305D-01,  -1.197868911105678D-01, &
   9.130692996959998D-02,   1.501464843750001D-01,   7.543616217590303D-02, &
  -5.320966249090855D-03,  -7.739869242411189D-02,  -1.137934319330609D-01, &
  -8.206839292587249D-02,  -1.946936977962388D-02,   1.774015688655539D-02, &
   1.673623640560183D-02,   2.749592040633584D-04,  -8.138136684189525D-03, &
  -3.563147063055984D-03,   3.436453788969876D-03,   3.344007205000425D-03, &
  -1.261945272399276D-03,  -2.536349266486547D-03,   0.000000000000000D+00, &
   1.546473935329351D-02,  -2.541406439970112D-02,  -9.423106802156399D-02, &
  -7.008807458597009D-02,   5.446218031199479D-03,   1.834542410714328D-02, &
  -2.747848834004371D-02,  -8.759531651373290D-02,  -1.345126322944374D-01, &
  -1.341574245522470D-01,  -7.985864002222626D-02,  -1.319374652009759D-02, &
   2.076761942827159D-02,   1.573129838219195D-02,  -2.195970898484108D-03, &
  -8.559055089266302D-03,  -1.970113039318546D-03,   4.117245959014424D-03, &
   2.354360557592165D-03,  -1.885319786831606D-03,  -1.917258198014866D-03, &
   1.222594994431179D-03,  -6.792194413506526D-04,  -1.222594994431176D-03, &
  -8.965696625828624D-03,   2.445189988862346D-03,  -1.018829162026023D-04, &
   1.455470231465686D-02,  -1.746564277758852D-04,  -1.508837473286094D-03, &
   2.634401118952888D-03,   1.724045767075880D-02,  -9.003152083533470D-03, &
   8.437554315963193D-03,  -5.105782786572621D-03,   2.010465797466419D-03, &
  -4.020029135535231D-02,   1.374126748319716D-02,  -1.223792619120471D-02, &
   3.591754475398078D-03,   1.183205695160006D-04,  -1.472315009319360D-03, &
  -5.931169916422898D-03,  -7.499353774287510D-04,  -6.386167845546391D-03, &
   8.586364322557525D-03,  -7.934056262353134D-03,   5.544099003860303D-03, &
  -4.055654366366080D-03,   4.109646427107926D-02,  -1.563841994385116D-02, &
   2.422189348260192D-02,  -1.826369568554724D-02,   1.442192408415348D-02, &
  -7.830176889229119D-03,   4.026165527652359D-03,  -3.270640420092587D-03, &
  -9.889315365568542D-03,   1.302527721720498D-02,  -5.744402951577405D-03, &
   5.777445582608935D-03,  -1.294541560541361D-03,  -1.703292366930118D-03, &
   1.898270408649676D-03,  -7.336450487633387D-04,   4.886047406106087D-05, &
  -1.442573773872946D-02,   7.735725492667849D-04,  -1.422632648981967D-02, &
   1.389501324461849D-02,  -1.702353256082765D-02,   1.419150264953548D-02, &
  -1.085389076775490D-02,   6.656884899172401D-03,  -3.655329787688898D-03, &
   3.759145452413464D-03,   6.365016153621295D-03,  -6.239159631137253D-03, &
   5.207153908574381D-03,  -1.265583433244607D-02,   9.483202938376090D-03, &
  -8.079469218527306D-03,   5.504683030587929D-03,  -3.702946475797174D-03, &
   3.119568540471909D-03,  -3.082364377921164D-03,   4.946524748944270D-03, &
   0.000000000000000D+00,   3.265390272241270D-03,   3.243180399633191D-03, &
  -1.122400605201510D-03,   5.339950008546866D-03,  -7.531675941409731D-03, &
   8.540359489123305D-03,  -6.439575420592440D-03,   4.148735794363786D-03, &
  -1.958585100831794D-03,  -2.217390098752922D-04,   1.582767022413019D-03, &
   0.000000000000000D+00,  -5.608409346349514D-04,  -1.003300197734986D-04, &
   5.230247293222750D-03,  -3.587443565723267D-03,   7.545427268857771D-03, &
  -6.410782878077352D-03,   5.539228747633529D-03,  -4.482720728807337D-03, &
   3.694560921564498D-03,  -3.131363610899795D-03,   1.760315950948643D-03, &
  -3.180458300129141D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.064102730263720D-03,  -1.281712637474140D-03,   2.343314717548515D-05, &
   1.605429536690057D-03,  -2.830217925685061D-03,   3.670073738017142D-03, &
  -3.415149117934273D-03,   1.815587411874773D-03,  -3.595698826583746D-04, &
  -6.101139495962213D-04,   1.128914891143327D-03,  -4.083475241002954D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.460869714515728D-04, &
  -7.347750451642086D-04,  -1.085891221231803D-03,  -2.477139577296012D-03, &
   2.003633716140607D-03,  -3.242696095889234D-03,   2.747439246274457D-03, &
  -2.584865830888572D-03,   2.306866506168983D-03,  -1.848962443607274D-03, &
   1.629656744640472D-03,  -2.587865869560134D-04,  -8.644377409152585D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.656676467729653D-04,   4.017692561129125D-04,  -9.545564225446372D-04, &
   2.765076330204937D-04,  -1.976928881408161D-03,   2.232458433513773D-03, &
  -2.066866847891660D-03,   1.458120891590642D-03,  -4.541038825968726D-04, &
  -3.861681283352032D-04,   1.182181920070981D-03,  -3.020430585765628D-04, &
   2.125351891462577D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.616521676472339D-05,   3.787162689589288D-04, &
   1.054455302485463D-03,   3.085435794900483D-05,   8.834583397084594D-04, &
  -7.421668586544402D-04,   1.041552544853057D-03,  -9.500043392147006D-04, &
   1.168489395598972D-03,  -1.193688078883620D-03,   7.266929142138643D-04, &
  -3.948995640717555D-04,   4.127363104097850D-04,   1.898260126393861D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.420275905056559D-04,   7.972498807430811D-05, &
   4.781084461437378D-04,   1.204448085691919D-03,  -8.233870943540086D-04, &
   1.701470559793998D-03,  -1.570101988577953D-03,   1.207034098161395D-03, &
  -7.149458863640556D-04,   1.133249644485418D-04,   1.292204858753893D-04, &
  -9.245956597090673D-04,   4.067700332966224D-04,  -2.418909806570762D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.211727239560483D-05,  -2.254005426519246D-04, &
  -3.857820336056624D-04,  -1.235099061142710D-04,   2.446320982037756D-04, &
  -3.153204794305543D-05,  -5.484186756913856D-05,  -1.777367889588203D-04, &
   3.360006623284897D-04,  -5.907937486335814D-04,   7.061551817475408D-04, &
  -5.202539488618533D-04,  -3.587149872992695D-04,  -2.977902642769358D-04, &
  -1.275244370915803D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.419727226967899D-05,  -1.164811275172997D-04,  -6.053830512488282D-04, &
  -1.725469355966516D-04,  -8.234354953737063D-04,   7.335316360245919D-04, &
  -1.063417228035267D-03,   9.040159473406158D-04,  -6.968573330971123D-04, &
   4.149593368839221D-04,   9.906956403532987D-05,  -1.252912188610508D-04, &
   2.293435235886289D-04,  -5.910339245396022D-04,  -3.662950385590100D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.790432550820729D-06, &
   1.120937808858848D-04,   1.361104215237279D-04,  -1.694490741598188D-04, &
  -3.912899194612646D-04,   1.271487839254407D-04,  -4.402776785653302D-04, &
   3.860340905785964D-04,  -9.350499686725300D-05,  -1.393009605509444D-04, &
   3.167358341500645D-04,  -2.572884958742408D-04,   4.844767119868537D-04, &
   2.966949253626519D-04,  -7.547152856048195D-05,   7.014996723505375D-04 /)
lem(:, 11) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.493907666691014D-02,   0.000000000000000D+00, &
   9.735110504930486D-02,   0.000000000000000D+00,  -1.894975177725838D+00, &
  -1.252448582170303D-01,   4.222095630787036D+00,  -9.891820161326995D-02, &
  -1.928399304489888D+00,   0.000000000000000D+00,   1.016095849668132D-01, &
   0.000000000000000D+00,  -1.535822080007497D-02,   0.000000000000000D+00, &
   3.656433216322437D-03,   0.000000000000000D+00,  -1.145314641774736D-03, &
   0.000000000000000D+00,   4.325772843784573D-04,   0.000000000000000D+00, &
  -1.872721525990861D-04,   0.000000000000000D+00,  -6.402461428675592D-03, &
   5.365975287870610D-02,  -1.360052202894712D-01,  -1.891363336314458D-01, &
  -3.249187272711909D-02,   7.543616217590250D-02,   1.068070023148150D-01, &
   7.791649810852150D-02,   1.070907552189693D-02,  -4.721304949003209D-02, &
  -8.231393037393453D-02,  -8.171172847043993D-02,  -4.148329364639777D-02, &
   4.147102916058689D-03,   1.964011518217025D-02,   7.776525600195168D-03, &
  -5.695894721821834D-03,  -6.318541926995001D-03,   4.593136298366651D-04, &
   3.793363710785886D-03,   1.113644725224165D-03,   0.000000000000000D+00, &
   2.134153809558426D-03,  -1.893873631013170D-02,   4.724391862686868D-02, &
   8.071927209375478D-02,   3.464263498801546D-02,  -9.219073425887698D-03, &
  -1.936848958333378D-02,   4.149162928590171D-03,   4.882686108511246D-02, &
   9.098828515998758D-02,   1.088028856962122D-01,   8.685152543789779D-02, &
   3.572620158776750D-02,  -8.337991843941123D-03,  -1.955163245104082D-02, &
  -6.104151233632039D-03,   6.663619155147434D-03,   5.883391714001500D-03, &
  -1.395365473396638D-03,  -3.849723129055474D-03,  -3.983309978407013D-04, &
  -1.687196728832537D-04,   1.518477055949290D-03,  -1.687196728832516D-04, &
  -3.350290647253203D-03,  -9.882152268876449D-04,   7.712899331805978D-04, &
  -5.712366067618780D-03,   3.591318751372156D-03,  -1.156934899770895D-03, &
  -2.892337249427264D-04,   1.833018731824504D-02,  -3.277982216017528D-03, &
   6.809043941359936D-04,   1.934250535554467D-03,  -1.965132261394707D-03, &
   7.314528755522657D-03,  -5.963135435098353D-03,   6.594991452801749D-03, &
  -6.209705896691675D-03,   2.784833546915356D-03,  -8.861881964865241D-04, &
  -3.860283212445861D-02,   1.465479130142956D-02,  -1.503341413084412D-02, &
   7.492769088092474D-03,  -2.212091877783861D-03,  -4.246921267373440D-04, &
   1.022171151680298D-03,   3.054131454160442D-03,  -5.306688304468545D-03, &
   1.715530798596778D-04,   4.225759654467641D-03,  -5.796920073778564D-03, &
   5.631291722397820D-03,  -3.186758275384014D-03,   2.175426268669658D-03, &
   3.563824384510647D-02,  -1.252320667000728D-02,   2.223524937153348D-02, &
  -1.862537982108067D-02,   1.641184423510103D-02,  -1.055246166930368D-02, &
   5.348156919343752D-03,  -2.796665946551903D-03,   2.443862265472236D-03, &
  -1.186287347063516D-02,   1.367686969506162D-02,  -1.013497790785842D-02, &
   1.092392435378868D-02,  -6.842922961441853D-03,   2.241922654113243D-03, &
  -1.498964295696677D-04,  -4.521795271670237D-04,  -3.628360400224177D-04, &
   8.759973010754201D-04,  -1.200308765256697D-02,  -1.064515919290594D-03, &
  -1.079810498557698D-02,   1.020758847872105D-02,  -1.475995051915278D-02, &
   1.385149063037660D-02,  -1.149931964312052D-02,   7.605311353733189D-03, &
  -4.136316102324880D-03,   2.004132303952990D-03,  -1.868375311383782D-03, &
   5.850171266278082D-03,  -5.395627755885543D-03,   5.422083895255423D-03, &
  -1.332548347566591D-02,   1.245995775308801D-02,  -1.200266169669604D-02, &
   9.107633603424420D-03,  -6.036794384607775D-03,   4.359885740269615D-03, &
  -3.179137280652129D-03,   2.610725406628878D-03,  -3.667306483795488D-03, &
   0.000000000000000D+00,   3.164422582334402D-03,   2.462752265726855D-03, &
   7.409877188077632D-04,   2.579666148836184D-03,  -4.712070057766877D-03, &
   6.914004738860340D-03,  -5.944538786578850D-03,   4.017090323726856D-03, &
  -1.852686210216763D-03,   1.736126856951918D-04,   1.194714589778794D-03, &
  -2.705405125805499D-03,   0.000000000000000D+00,  -5.732523660980991D-04, &
   2.183382394513304D-05,   4.592313132258529D-03,  -3.413118769172896D-03, &
   8.217291793209982D-03,  -8.362264775795904D-03,   7.793900410095886D-03, &
  -6.223338348897925D-03,   4.588304136545910D-03,  -3.452023157606505D-03, &
   2.315307182373343D-03,  -7.959059133950110D-04,   8.148056272714032D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.043333575266167D-03, &
  -1.332793111974204D-03,   4.012988421872691D-04,   3.653583174127308D-04, &
  -1.232456524652356D-03,   2.288027957854623D-03,  -2.671097435043337D-03, &
   1.486578981975075D-03,  -1.444571239237750D-04,  -9.754027829558018D-04, &
   1.559666360751457D-03,  -1.390989900779945D-03,   3.520499092361616D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.452473077233087D-04, &
  -7.004934566313602D-04,  -1.105709098059747D-03,  -2.386538065102859D-03, &
   2.243587041138008D-03,  -4.050573589067520D-03,   3.932817863906550D-03, &
  -3.619831637067400D-03,   2.976449698440771D-03,  -2.110964980240399D-03, &
   1.272508475143008D-03,  -6.081892582960189D-04,  -3.856743889662903D-04, &
   2.458504315438751D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.736552054007622D-04,   3.648275321615918D-04, &
  -7.388121420556486D-04,  -1.050974216954416D-04,  -1.333687681766748D-03, &
   1.679652225370969D-03,  -1.711568457366441D-03,   1.253121394619028D-03, &
  -2.458539116041422D-04,  -5.463023124651737D-04,   1.052539254482994D-03, &
  -1.450506552839283D-03,   4.055489364721190D-04,  -7.607752452988592D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.786588125467149D-05,   3.885696546885108D-04,   1.032580091132900D-03, &
   4.740099396903772D-06,   1.150796706788563D-03,  -1.294456838734316D-03, &
   1.774679646361575D-03,  -1.610787885397292D-03,   1.591323086850937D-03, &
  -1.396712853915747D-03,   9.345594980078846D-04,  -2.755225349223114D-04, &
  -3.729831808206081D-04,   1.371276641389936D-05,  -2.230697891530804D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.449633874757389D-04,   6.469331080919713D-05, &
   5.354933539247871D-04,   1.011856446153872D-03,  -5.849543609272916D-04, &
   1.531051683825308D-03,  -1.549072363381057D-03,   1.217619156396308D-03, &
  -6.798283608130348D-04,  -3.239216084633777D-05,   5.041612787836761D-04, &
  -5.200312780028876D-04,   7.636545334159067D-04,  -4.748666089071653D-04, &
  -8.685397125439126D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.244875317140258D-05, &
  -2.300396740786239D-04,  -3.906563966711545D-04,  -1.749002740370315D-04, &
   3.931750016724983D-04,  -3.916947146540276D-04,   4.030979724166309D-04, &
  -6.092682866594083D-04,   6.563827575986254D-04,  -7.950134552282765D-04, &
   7.326486343104130D-04,  -4.801752601209401D-04,   2.205972319495875D-04, &
   6.842877770554068D-04,  -1.056007549734680D-04,   9.261869738232498D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.677786588025998D-05, &
  -1.243390400671266D-04,  -5.824692622257551D-04,  -2.347211557549217D-04, &
  -7.819579892236716D-04,   7.685721527978945D-04,  -1.186475375503929D-03, &
   1.037521925498522D-03,  -7.394129847378032D-04,   3.736531528128047D-04, &
   6.928620118098011D-05,  -4.461484500269600D-04,   4.037868267733688D-04, &
   8.968065581526254D-06,   4.932079182075902D-04,   9.690195975733130D-04 /)
lem(:, 12) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.495293720032163D-03,   0.000000000000000D+00,  -1.468266071080898D-02, &
   0.000000000000000D+00,   9.893057979515101D-02,   0.000000000000000D+00, &
  -1.914575246825788D+00,  -9.891820161326995D-02,   4.134544994212961D+00, &
  -8.010018789148124D-02,  -1.938357173611013D+00,   0.000000000000000D+00, &
   1.020242680797499D-01,   0.000000000000000D+00,  -1.537824321752542D-02, &
   0.000000000000000D+00,   3.647495840096974D-03,   0.000000000000000D+00, &
  -1.137408961704299D-03,   0.000000000000000D+00,   4.274150686410025D-04, &
   0.000000000000000D+00,   3.495293720032787D-03,  -1.804961915705379D-02, &
   3.661130982435137D-02,  -8.778338150186008D-03,  -1.317680026869874D-01, &
  -1.363931569888010D-01,  -5.320966249091328D-03,   7.791649810852101D-02, &
   7.579933449074101D-02,   6.042426466235785D-02,   2.751502442272652D-02, &
  -2.361291587166593D-02,  -6.108720078285131D-02,  -7.037879355848067D-02, &
  -5.089698229479402D-02,  -1.375098411627570D-02,   1.395634363957804D-02, &
   1.471258571802781D-02,   2.774925049804369D-04,  -7.464055337637868D-03, &
  -3.166204991931302D-03,   2.900361714812586D-03,   0.000000000000000D+00, &
  -4.512404789263121D-03,   1.029693088809917D-02,   9.081039465712847D-04, &
  -5.145226771587096D-02,  -6.272331177551164D-02,  -1.687880277375722D-02, &
   1.439971049546754D-02,   1.528284143518548D-02,   3.722918278770047D-03, &
  -2.287018861117004D-02,  -5.892806934942549D-02,  -8.387450232262382D-02, &
  -8.135677770062504D-02,  -4.997440428411135D-02,  -8.475705271704293D-03, &
   1.588288908591875D-02,   1.323553653614593D-02,  -1.094539600296598D-03, &
  -7.391723079014793D-03,  -2.481054238025312D-03,   3.236584335730868D-03, &
  -3.567369214681288D-04,   2.751970537039835D-04,   3.567369214681057D-04, &
   1.172135599109513D-03,  -5.402016239374212D-04,  -1.579834937930171D-04, &
  -5.228744020375482D-03,  -2.089459111456082D-04,   7.338588098772567D-04, &
  -5.159944756949398D-04,  -1.146654390433221D-03,   3.460348138240682D-03, &
  -1.908542529854405D-03,   6.000824643267142D-04,   5.621791664207313D-04, &
   1.857580112501811D-02,  -5.452978656726855D-03,   2.963464569052966D-03, &
   6.765260903555955D-04,  -1.752151611603645D-03,   1.304319369117781D-03, &
  -1.140564144891997D-03,  -2.411545472899990D-03,   3.455772013493024D-03, &
  -5.884621168420313D-03,   3.896136987197043D-03,  -1.654544139583097D-03, &
   4.192272114854914D-04,  -3.423448508066545D-02,   1.387385484819804D-02, &
  -1.564914895565768D-02,   1.035548522711471D-02,  -5.063914813599301D-03, &
   9.586012953953774D-04,   6.578048888661616D-04,  -8.165396125654839D-04, &
   9.027621454813430D-03,  -8.680443475928520D-03,   6.047513142157425D-03, &
  -8.026565562034394D-04,  -2.359188047777915D-03,   4.732066094069136D-03, &
  -3.615853080536956D-03,   1.883640018906703D-03,  -1.233562544377021D-03, &
   3.031183050562961D-02,  -9.128157839757877D-03,   1.848045868303768D-02, &
  -1.695022896495577D-02,   1.671232551432137D-02,  -1.258261263132704D-02, &
   7.280392139106114D-03,  -3.484096743060780D-03,   1.704094165443369D-03, &
  -1.466471786859341D-03,  -1.277415282408885D-02,   1.343592337727050D-02, &
  -1.264496085079228D-02,   1.475714288555852D-02,  -1.214134102047792D-02, &
   6.857225739545373D-03,  -3.087380653058470D-03,   6.718568957311242D-04, &
  -3.357278472798047D-04,   7.777985725478819D-04,  -1.207316868892324D-03, &
  -1.004198690394189D-02,  -2.425479045388229D-03,  -7.621654520771887D-03, &
   6.035287439101603D-03,  -1.083127444877372D-02,   1.180750177483020D-02, &
  -1.111323368637301D-02,   8.143022933431110D-03,  -4.644196183760307D-03, &
   2.229570312632967D-03,  -7.765614244308191D-04,   5.115717303299576D-04, &
   5.377237075898072D-03,  -4.592594481236276D-03,   5.143341189028134D-03, &
  -1.288413220561327D-02,   1.385322640442445D-02,  -1.499551822667579D-02, &
   1.276127120253350D-02,  -8.860966380679096D-03,   5.992312658619482D-03, &
  -3.921094062244018D-03,   2.676178693292080D-03,  -1.926919316434951D-03, &
   2.512470217946052D-03,   0.000000000000000D+00,   3.020737229890544D-03, &
   1.925779406480802D-03,   2.100704079632807D-03,   2.356532596893657D-05, &
  -1.296510843121546D-03,   4.005602720605601D-03,  -4.390053057136098D-03, &
   3.375115924437100D-03,  -1.579795285145763D-03,  -5.309510516820831D-05, &
   9.189504146016502D-04,  -1.543347101896757D-03,   2.763404731753351D-03, &
   0.000000000000000D+00,  -5.748537353611139D-04,   1.106089339783515D-04, &
   3.948323368488266D-03,  -2.859641900730962D-03,   7.985095153480162D-03, &
  -9.346852132872504D-03,   9.640105260380689D-03,  -8.078774791628423D-03, &
   5.779162866816959D-03,  -3.999634626194542D-03,   2.481860250061145D-03, &
  -1.263260221204686D-03,  -6.396060005753080D-05,   7.157818332728695D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.018280022249123D-03, &
  -1.333291387335572D-03,   6.179530526790729D-04,  -6.408879412192719D-04, &
   4.575076043414761D-04,   3.584133049028002D-04,  -1.216754490445343D-03, &
   6.660404009593702D-04,   3.364123682867097D-04,  -1.239085191032241D-03, &
   1.769953242902585D-03,  -1.723036295641050D-03,   1.049841972058652D-03, &
  -2.083462069179736D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.437939075172622D-04,  -6.645189355024239D-04,  -1.129687259349322D-03, &
  -2.170346655522550D-03,   2.139168721703806D-03,  -4.374637514712722D-03, &
   4.801971259075806D-03,  -4.603518018015632D-03,   3.722511135899826D-03, &
  -2.500864909347941D-03,   1.404837616180511D-03,  -3.450384149018058D-04, &
  -3.545635660012778D-04,   8.584977532540390D-04,  -2.904850012849724D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.760202093338931D-04,   3.417316182967864D-04,  -5.740356131534213D-04, &
  -4.359217383210175D-04,  -6.086419486410867D-04,   8.271593749480471D-04, &
  -9.794463641125907D-04,   7.551224680272587D-04,   9.932583264857649D-05, &
  -8.187684978544555D-04,   1.165879372673940D-03,  -1.171527899488467D-03, &
   1.048830437749382D-03,  -4.275417781630836D-05,  -8.462538956231172D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.892692545775752D-05,   3.930365952786190D-04,   9.958548812282802D-04, &
   4.041441685866115D-05,   1.238811510978399D-03,  -1.632705124482196D-03, &
   2.382905477360945D-03,  -2.285047741226910D-03,   2.092158978043325D-03, &
  -1.680618832363974D-03,   1.011752049748130D-03,  -3.403770416372419D-04, &
  -3.568413244316251D-04,   9.391274607024978D-04,  -3.071247717774360D-04, &
   1.616791555433292D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.467748546763090D-04, &
   5.403720389800469D-05,   5.701731743360650D-04,   8.243558247520502D-04, &
  -2.686849039827911D-04,   1.153574116190469D-03,  -1.279244016110396D-03, &
   1.043098715603481D-03,  -5.398016344593464D-04,  -1.628616543413172D-04, &
   6.223662091216745D-04,  -7.498215959022820D-04,   4.675234724185957D-04, &
  -2.196649722061246D-04,   1.934576055500781D-04,   1.723732600825333D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.269735923019527D-05,  -2.321674097481004D-04, &
  -3.967381614680134D-04,  -1.933359680362663D-04,   4.636829201086275D-04, &
  -6.607116687916085D-04,   8.389189396076943D-04,  -1.090327315258988D-03, &
   1.050527034138627D-03,  -1.046475468820291D-03,   8.605584230842621D-04, &
  -4.298356772943440D-04,   1.187146753410161D-05,   2.158650070463725D-04, &
  -8.117420077919048D-04,   3.629412530875298D-04,  -5.743456406661187D-05 /)
lem(:, 13) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.881420744680548D-03,   0.000000000000000D+00, &
  -1.513421660621134D-02,   0.000000000000000D+00,   1.002400936831037D-01, &
   0.000000000000000D+00,  -1.928399304489888D+00,  -8.010018789148124D-02, &
   4.070871803977270D+00,  -6.618410984629487D-02,  -1.945855434215286D+00, &
   0.000000000000000D+00,   1.023641287682402D-01,   0.000000000000000D+00, &
  -1.540774561443011D-02,   0.000000000000000D+00,   3.649172637441531D-03, &
   0.000000000000000D+00,  -1.136683098086236D-03,   0.000000000000000D+00, &
   4.269043338259104D-04,   0.000000000000000D+00,   1.663466033434668D-03, &
  -1.394169684122703D-02,   4.372881941057201D-02,  -4.256639051532647D-02, &
  -1.344109478801961D-01,  -7.739869242411175D-02,   1.070907552189652D-02, &
   6.042426466235732D-02,   6.552088713842973D-02,   4.612975272066294D-02, &
   2.684988850251264D-02,  -2.930738193723886D-03,  -4.019119343515039D-02, &
  -5.990790294561191D-02,  -5.215959155529001D-02,  -2.566329663596913D-02, &
   2.639241778986934D-03,   1.502098766956133D-02,   7.741015457038304D-03, &
  -4.341644462214122D-03,  -6.378132150719544D-03,   0.000000000000000D+00, &
  -5.544886778118465D-04,   4.920598885138541D-03,  -1.545503410108929D-02, &
   1.360109519424606D-02,   5.369173465419403D-02,   4.079459945808685D-02, &
   7.085540613616902D-03,  -1.302073488851782D-02,  -1.459347559400870D-02, &
  -5.537703886092165D-03,   9.881366611684813D-03,   3.480360800229737D-02, &
   6.100488591919310D-02,   7.079766327745386D-02,   5.538132200587017D-02, &
   2.295402673364709D-02,  -6.235762572570342D-03,  -1.532708325375160D-02, &
  -5.975076854476972D-03,   5.103969485703114D-03,   5.743986203870804D-03, &
   4.383617896650334D-05,  -3.945256106985268D-04,   4.383617896650277D-05, &
   9.775467909530274D-04,   2.805515453856240D-04,  -2.093177545650548D-04, &
  -2.410989843157748D-04,  -9.687795551597248D-04,   9.534368925214405D-05, &
   2.684965961698339D-04,  -5.469780997709242D-03,   9.746243790219218D-04, &
   4.328822672942224D-04,  -7.051536455417284D-04,   2.899580587888532D-04, &
   3.287713422487756D-03,   2.294093365913677D-03,  -1.873631349326635D-03, &
   1.528056138474032D-03,   2.054820889054700D-05,  -6.945294605005365D-04, &
   1.603089064805028D-02,  -6.429306248420487D-03,   4.479935723212846D-03, &
  -1.240929166243430D-03,  -1.346227321135889D-03,   1.361996168847451D-03, &
  -8.000102661386891D-04,  -7.073182359475494D-03,   1.001468253321158D-03, &
  -1.237201072694807D-04,  -4.102661263026800D-03,   4.259733690800692D-03, &
  -2.656667166985521D-03,   8.533375850923119D-04,  -7.573191885964770D-05, &
  -2.876805566702393D-02,   1.179691290907323D-02,  -1.425276854376007D-02, &
   1.167048179321248D-02,  -7.444577085512885D-03,   3.116328175312788D-03, &
   4.317626125682535D-05,  -5.824350579699365D-04,   4.848407142796097D-04, &
   1.255123506667630D-02,  -1.065125269008881D-02,   1.041346617180700D-02, &
  -5.741049518203379D-03,   1.848400272083743D-03,   2.475841023176310D-03, &
  -3.486456756650338D-03,   2.323888549418420D-03,  -1.147046991820749D-03, &
   7.911337708894464D-04,   2.554328464052893D-02,  -5.979389936687075D-03, &
   1.395843763109414D-02,  -1.353047770024631D-02,   1.507700553094930D-02, &
  -1.332150252200419D-02,   8.989673248140911D-03,  -4.893717428617944D-03, &
   2.047612825942346D-03,  -1.036909532588873D-03,   9.121581701220038D-04, &
  -1.301316996957973D-02,   1.264182413254625D-02,  -1.354329975220797D-02, &
   1.680131830186772D-02,  -1.626435497890227D-02,   1.157109859122307D-02, &
  -6.745704049667405D-03,   2.677981496435529D-03,  -8.432577972996552D-04, &
   5.973876485385198D-04,  -7.295599153120454D-04,   9.425477328495354D-04, &
  -8.458286027628074D-03,  -3.366281497576373D-03,  -4.989577325752909D-03, &
   2.117754630711512D-03,  -6.036229428611269D-03,   8.073264503375376D-03, &
  -9.264478479709851D-03,   7.935808592382194D-03,  -5.030777569569812D-03, &
   2.596851965252260D-03,  -9.277608362442878D-04,   4.508290222609537D-05, &
   2.827985031953057D-04,   4.949392611131761D-03,  -3.870827781093686D-03, &
   4.623028665865405D-03,  -1.173283147897967D-02,   1.374489693784208D-02, &
  -1.644910658963559D-02,   1.571319985413043D-02,  -1.188983493164902D-02, &
   8.049659113809293D-03,  -4.973158940699113D-03,   2.998196145767688D-03, &
  -1.876477320282610D-03,   1.145707951974619D-03,  -1.337469567381416D-03, &
   0.000000000000000D+00,   2.856945668463977D-03,   1.568279761547668D-03, &
   2.987292409921699D-03,  -2.005796915059435D-03,   2.061044208677530D-03, &
   2.582185589756376D-04,  -1.641553043489166D-03,   1.922185149093309D-03, &
  -9.868143637049669D-04,  -3.416213628738431D-04,   1.094876626263295D-03, &
  -1.404012762795029D-03,   1.513426266648323D-03,  -2.420632518267263D-03, &
   0.000000000000000D+00,  -5.690407763512195D-04,   1.697837758167171D-04, &
   3.358862291392439D-03,  -2.135992063036518D-03,   7.077374997472257D-03, &
  -9.237380627235368D-03,   1.060916478927425D-02,  -9.660056215148155D-03, &
   7.099022163700382D-03,  -4.695331117160342D-03,   2.716117518026922D-03, &
  -1.236589295011250D-03,   2.681392036131636D-04,   6.530842031713331D-04, &
  -1.534886990971190D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.899253085662693D-04,  -1.302671702599338D-03,   7.017961987835426D-04, &
  -1.339385966868805D-03,   1.944638524645289D-03,  -1.788567457398021D-03, &
   8.467618422676120D-04,  -7.404497546995001D-04,   1.152178483667406D-03, &
  -1.692280717114779D-03,   2.025842427991905D-03,  -1.866492468952925D-03, &
   1.422521026030059D-03,  -5.297642182952582D-04,   7.569994937402523D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.418448713027486D-04, &
  -6.293688037585722D-04,  -1.144804038671542D-03,  -1.902182073602562D-03, &
   1.785756509657203D-03,  -4.191335264534872D-03,   5.149647116045555D-03, &
  -5.312993834345387D-03,   4.413866551032861D-03,  -2.899096075811537D-03, &
   1.520373995418107D-03,  -3.452145576711752D-04,  -5.359554217753594D-04, &
   9.307887486999888D-04,  -9.177699643629581D-04,   2.421016994519559D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.745452721661506D-04,   3.277595109804113D-04,  -4.571728464734997D-04, &
  -6.733772628505854D-04,   5.858548936097138D-05,  -1.650503152130347D-04, &
   7.537610793873762D-05,  -7.479848535931274D-05,   6.541638835576578D-04, &
  -1.205227731572926D-03,   1.422409885114364D-03,  -1.295460609708184D-03, &
   8.163000053271349D-04,  -3.970370807480503D-04,  -3.847648993282499D-04, &
   1.823905793049122D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.951497677715947D-05,   3.931338172640457D-04, &
   9.541288011106494D-04,   1.072482990894837D-04,   1.184958587200663D-03, &
  -1.720181728096405D-03,   2.739850240805591D-03,  -2.837458235821611D-03, &
   2.578922599657523D-03,  -1.969037871702357D-03,   1.101856846491073D-03, &
  -3.072501215878688D-04,  -3.337913594140006D-04,   7.602604372772551D-04, &
  -1.006989257279724D-03,   2.119337326355441D-04,  -4.107992135525899D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.476059051883546D-04,   4.638164818304661D-05, &
   5.842515966189254D-04,   6.692803723871540D-04,   5.106267130805294D-05, &
   6.500296743094968D-04,  -7.776022729332082D-04,   6.481305633928539D-04, &
  -2.520801733842427D-04,  -3.902506416323501D-04,   8.037137048836808D-04, &
  -8.483461733221018D-04,   6.185723458326034D-04,  -8.451151179482500D-05, &
  -3.727729279681497D-04,   1.262686317224810D-04,  -1.789278913119071D-03 /)
lem(:, 14) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.311020180385895D-03,   0.000000000000000D+00,   3.629742233595467D-03, &
   0.000000000000000D+00,  -1.514092550541258D-02,   0.000000000000000D+00, &
   1.009957973947408D-01,   0.000000000000000D+00,  -1.938357173611013D+00, &
  -6.618410984629487D-02,   4.023033835671162D+00,  -5.560384374855215D-02, &
  -1.951592647277169D+00,   0.000000000000000D+00,   1.026149667148347D-01, &
   0.000000000000000D+00,  -1.542208233586376D-02,   0.000000000000000D+00, &
   3.644798835623836D-03,   0.000000000000000D+00,  -1.132366216735540D-03, &
   0.000000000000000D+00,  -1.311020180385226D-03,   6.770079100229386D-03, &
  -1.496598152472874D-02,   1.440509291004505D-02,   2.014606046523218D-02, &
  -6.283035863241884D-02,  -1.137934319330602D-01,  -4.721304949003205D-02, &
   2.751502442272618D-02,   4.612975272066249D-02,   4.877179078381454D-02, &
   4.214548146361585D-02,   2.429077673183455D-02,   4.322554538599188D-03, &
  -2.105730964493258D-02,  -4.544983030536849D-02,  -5.065358078694638D-02, &
  -3.326560981169679D-02,  -7.476832556716723D-03,   1.012520797748868D-02, &
   1.138624792650749D-02,   1.489628234891942D-03,   0.000000000000000D+00, &
   1.692519775057697D-03,  -4.291320867305166D-03,   3.619013489716231D-03, &
   9.288526320418791D-03,  -2.325831968766476D-02,  -4.808756573958513D-02, &
  -2.653740325297178D-02,   1.099776144800952D-03,   1.028252373491060D-02, &
   1.164045237531616D-02,   7.452316170218124D-03,  -3.398038698137209D-03, &
  -2.010418475329286D-02,  -4.129374379848779D-02,  -5.718534677938846D-02, &
  -5.482739270240478D-02,  -3.292902273124911D-02,  -4.895259339684902D-03, &
   1.201843481102297D-02,   1.081879099747674D-02,   2.420886649053246D-05, &
   1.338054368514506D-04,  -1.146903744440911D-04,  -1.338054368514218D-04, &
  -2.246019832863343D-04,   1.768143272679585D-04,   8.721247223351320D-05, &
   1.320532227974243D-03,   1.720355616660905D-05,  -2.372657121311918D-04, &
   8.402662849895194D-05,  -1.964072662354918D-03,  -9.254876048890535D-04, &
   3.997437425936471D-04,   7.048679262709647D-05,  -3.072348051514232D-04, &
  -4.104322483230904D-03,   1.874869037787299D-03,  -2.127347153751063D-04, &
  -6.203634212117731D-04,   5.808987819499426D-04,  -5.324341341378902D-05, &
   6.436041512553799D-03,   6.027616345783339D-04,  -1.055390383165645D-03, &
   1.978807189627244D-03,  -7.649210806673372D-04,  -3.416817405313327D-04, &
   6.265857722558422D-04,   1.200234768557324D-02,  -6.148359823323181D-03, &
   4.884340118867661D-03,  -3.031800949438037D-03,  -1.782189947327441D-04, &
   1.349606672329038D-03,  -1.049864935438532D-03,   5.220502747156544D-04, &
  -1.040277950243279D-02,   3.630301827467617D-03,  -3.373259996209464D-03, &
  -1.389880305339575D-03,   3.464668352171455D-03,  -3.380968174239771D-03, &
   1.690992016164229D-03,  -3.212470137050037D-04,  -1.374136593698695D-04, &
  -2.335777954626437D-02,   9.102294503850066D-03,  -1.138798700033549D-02, &
   1.122046763014864D-02,  -8.755866741736437D-03,   5.342294621869045D-03, &
  -1.336884852193465D-03,  -4.151635476300120D-04,   4.907205521045612D-04, &
  -3.396531914402941D-04,   1.431336529185573D-02,  -1.138711889115282D-02, &
   1.291126875509319D-02,  -9.709426703060454D-03,   6.170672262673326D-03, &
  -8.115215430450386D-04,  -2.326184952863507D-03,   2.572409314141775D-03, &
  -1.531861316509928D-03,   6.459814012004521D-04,  -3.784545323524682D-04, &
   2.146860031080334D-02,  -3.345435113566901D-03,   9.535631289720575D-03, &
  -9.149285945517121D-03,   1.165307498695189D-02,  -1.237246475054822D-02, &
   9.888250481288831D-03,  -6.360696114839774D-03,   2.951812927669599D-03, &
  -1.157798361909334D-03,   6.404052719089623D-04,  -6.344495199871752D-04, &
  -1.284511416531973D-02,   1.156883075522746D-02,  -1.330070577431056D-02, &
   1.712014039045022D-02,  -1.855041336613522D-02,   1.551800844989771D-02, &
  -1.069829054343620D-02,   5.427188459062213D-03,  -2.021238782636585D-03, &
   8.121888665959296D-04,  -5.624106642522879D-04,   6.035575148541428D-04, &
  -7.336145631539965D-04,  -7.176567341111109D-03,  -3.972218009903805D-03, &
  -2.966117586688743D-03,  -1.099535709370565D-03,  -1.269332667614176D-03, &
   3.226638811815479D-03,  -5.818960909694595D-03,   6.562222884373915D-03, &
  -4.938815458174047D-03,   2.888299330628826D-03,  -1.181685216334520D-03, &
   2.088833822259433D-04,   2.426974269032646D-04,  -4.484010208670804D-04, &
   4.565232873205334D-03,  -3.242910554504254D-03,   4.021436151992215D-03, &
  -1.023832425236359D-02,   1.251726790277206D-02,  -1.621532959335080D-02, &
   1.726021865111396D-02,  -1.450243598463004D-02,   1.034730677693563D-02, &
  -6.387323420462716D-03,   3.611519162375667D-03,  -2.056550821976864D-03, &
   1.101914990840860D-03,  -5.106590323317859D-04,   4.621384974757343D-04, &
   0.000000000000000D+00,   2.687007761592501D-03,   1.336032456309669D-03, &
   3.489600291386719D-03,  -3.402701172308988D-03,   4.857774385047901D-03, &
  -3.660273215968780D-03,   2.046285278877328D-03,  -5.468497526364854D-04, &
   2.122516791050293D-04,  -8.504155226279594D-04,   1.313567489025694D-03, &
  -1.449614063623566D-03,   1.346080500483719D-03,  -1.167621798125984D-03, &
   1.682399197513091D-03,   0.000000000000000D+00,  -5.582644711828798D-04, &
   2.051307548895130D-04,   2.849377396710793D-03,  -1.395052718474322D-03, &
   5.785429926762608D-03,  -8.172412861792168D-03,   1.042983486914287D-02, &
  -1.049962997457186D-02,   8.255024191009616D-03,  -5.494104869364397D-03, &
   3.076023940659708D-03,  -1.295452477990926D-03,   2.030539473871566D-04, &
   4.595067816217942D-04,  -9.734623084744893D-04,   1.845676061180520D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.592310480549429D-04, &
  -1.255122740972130D-03,   6.915570464422982D-04,  -1.736399993318016D-03, &
   3.040376511413362D-03,  -3.782577080935007D-03,   3.234051029544556D-03, &
  -2.717424373469373D-03,   2.424008685068478D-03,  -2.382530869084187D-03, &
   2.371593158263596D-03,  -2.000076477592027D-03,   1.425080913077551D-03, &
  -8.245758358942446D-04,  -4.513285105569371D-06,   3.196018834102631D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.395005741317509D-04, &
  -5.960036405441328D-04,  -1.146358496683441D-03,  -1.634654086389581D-03, &
   1.302054506902045D-03,  -3.590984488417328D-03,   4.882781646492284D-03, &
  -5.516663096962954D-03,   4.865892446930311D-03,  -3.251726727551734D-03, &
   1.636390735973023D-03,  -3.084821370582758D-04,  -6.249273897261325D-04, &
   1.074716494929592D-03,  -1.131984693842424D-03,   7.382488902763875D-04, &
  -1.608769000745357D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.704419092323246D-04,   3.192603265156830D-04, &
  -3.780490147057850D-04,  -8.102300395026363D-04,   5.743686678112534D-04, &
  -1.112532830590020D-03,   1.302870404170175D-03,  -1.213096331466039D-03, &
   1.472092802924334D-03,  -1.733156244869202D-03,   1.725377386460359D-03, &
  -1.443393670391292D-03,   8.680930855230768D-04,  -2.108235800573346D-04, &
  -2.189916296347253D-04,   6.273043918209745D-04,  -2.086067844801860D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.975180768986226D-05,   3.898794688840761D-04,   9.125820823390491D-04, &
   1.819976181676072D-04,   1.042891292881920D-03,  -1.583858092735750D-03, &
   2.775986281787053D-03,  -3.126629625690773D-03,   2.942242685024866D-03, &
  -2.223385855801588D-03,   1.180982896768484D-03,  -2.485543039035215D-04, &
  -4.455460911760186D-04,   8.153579867166565D-04,  -8.216436034492833D-04, &
   7.371644110534922D-04,   5.531312900634196D-05,  -6.624771650578121D-04 /)
lem(:, 15) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.306463328200726D-03,   0.000000000000000D+00, &
   3.714276175212994D-03,   0.000000000000000D+00,  -1.527613833269480D-02, &
   0.000000000000000D+00,   1.016095849668132D-01,   0.000000000000000D+00, &
  -1.945855434215286D+00,  -5.560384374855215D-02,   3.986235398512620D+00, &
  -4.737232977966085D-02,  -1.956111107405677D+00,   0.000000000000000D+00, &
   1.028237486389946D-01,   0.000000000000000D+00,  -1.543948877550821D-02, &
   0.000000000000000D+00,   3.645133841848287D-03,   0.000000000000000D+00, &
  -1.131461703896454D-03,   0.000000000000000D+00,  -5.599128549429831D-04, &
   4.692692922018790D-03,  -1.585870726495718D-02,   2.678314039687555D-02, &
   1.247843284616896D-03,  -8.206839292587240D-02,  -8.231393037393445D-02, &
  -2.361291587166581D-02,   2.684988850251236D-02,   4.214548146361528D-02, &
   3.614670692659545D-02,   3.285142847455098D-02,   2.532809841680255D-02, &
   8.435107919934374D-03,  -1.097792640789908D-02,  -3.056634007584821D-02, &
  -4.321180391645162D-02,  -3.775753628671186D-02,  -1.681813664434309D-02, &
   3.746779756632439D-03,   1.138813127667259D-02,   0.000000000000000D+00, &
   1.866376183136989D-04,  -1.656244560712883D-03,   5.634014423076721D-03, &
  -9.273748871269740D-03,  -1.894873135899645D-03,   3.137901076615629D-02, &
   3.723337919657097D-02,   1.605220467788405D-02,  -3.630046117195844D-03, &
  -9.840269553396747D-03,  -8.672223915637229D-03,  -6.821122800317383D-03, &
  -1.057702005918542D-03,   1.100443782245060D-02,   2.730359981607806D-02, &
   4.310732592391266D-02,   4.921333451110234D-02,   3.833999049101865D-02, &
   1.520928383015175D-02,  -5.491871809320066D-03,  -1.199104961236746D-02, &
  -1.475499927355315D-05,   1.327949934620179D-04,  -1.475499927355504D-05, &
  -3.435806973699833D-04,  -9.766404281069111D-05,   7.166713932870880D-05, &
   3.154759368489261D-04,   3.309335551355016D-04,  -4.215714078160249D-06, &
  -1.166347561624069D-04,   9.253492401559415D-04,  -4.427905020093201D-04, &
  -2.002464187125666D-04,   1.825404195842973D-04,   2.309860005324640D-05, &
  -3.233312174145538D-03,  -3.714044102858435D-04,   5.397519258069957D-04, &
  -2.270864650101815D-04,  -2.399268274732367D-04,   2.853862776160531D-04, &
  -1.696824916459088D-03,   2.148493967814866D-03,  -7.668383908171729D-04, &
  -8.823617314498920D-05,   7.872542637297725D-04,  -3.330174592536803D-04, &
  -1.464433677900595D-04,   7.764080617745871D-03,  -1.059408947841414D-03, &
   1.276850370036366D-04,   1.628377049699181D-03,  -1.511986613329936D-03, &
   1.668289003361900D-04,   4.740043516630315D-04,  -5.042521001738241D-04, &
   7.775884617164800D-03,  -4.897676092202845D-03,   4.197164936215369D-03, &
  -4.071281158863301D-03,   1.302807351983578D-03,   7.131183766871071D-04, &
  -1.287632378163588D-03,   7.239786310225474D-04,  -2.627542605011424D-04, &
  -1.165596997304475D-02,   5.161776167423350D-03,  -5.662789163209289D-03, &
   1.549900628915546D-03,   1.711613247769198D-03,  -3.210117712110478D-03, &
   2.606314557953085D-03,  -8.684610968714480D-04,   5.449901995844551D-05, &
   1.319071807784306D-04,  -1.860444111509130D-02,   6.407912217303546D-03, &
  -7.869341601664186D-03,   9.200642921415654D-03,  -8.657656818853603D-03, &
   6.925532502573767D-03,  -3.172604211445882D-03,   3.551610591558013D-04, &
   5.218196104310761D-04,  -3.993242866424144D-04,   2.900977199565568D-04, &
   1.491581593520511D-02,  -1.123450939788472D-02,   1.369471798103654D-02, &
  -1.210178721665201D-02,   9.797564935570394D-03,  -4.580129854900214D-03, &
  -1.261555252738285D-04,   2.101464047546216D-03,  -1.976082241372568D-03, &
   9.407887397753275D-04,  -3.663150569935710D-04,   2.037455377690629D-04, &
   1.807446353621500D-02,  -1.295484048433897D-03,   5.741122207721697D-03, &
  -4.696572562153424D-03,   7.089163096757416D-03,  -9.646653796585868D-03, &
   9.499264625393673D-03,  -7.389024185546348D-03,   4.148163933737434D-03, &
  -1.692636732630153D-03,   6.514780827441563D-04,  -3.593693580412753D-04, &
   3.185669915366820D-04,  -1.244313820506998D-02,   1.040452741939554D-02, &
  -1.237845163359318D-02,   1.612884784697679D-02,  -1.886311410054378D-02, &
   1.788980594339996D-02,  -1.423492817979663D-02,   8.647410906001411D-03, &
  -3.877666262795568D-03,   1.455351258575030D-03,  -5.682732269305975D-04, &
   4.530972781029818D-04,  -4.701325806442417D-04,   5.946576023280294D-04, &
  -6.134314629019729D-03,  -4.326907199297713D-03,  -1.494712501765008D-03, &
  -3.461683090063511D-03,   2.790465566286650D-03,  -1.853069140733937D-03, &
  -1.128715571725725D-03,   3.762751340339648D-03,  -4.019276819273997D-03, &
   2.907747272776858D-03,  -1.430778679810878D-03,   3.476541104169074D-04, &
   1.348947088160506D-04,  -3.484058913195843D-04,   5.084268600055564D-04, &
   4.221383955680836D-03,  -2.706984359366032D-03,   3.429277596411012D-03, &
  -8.669021880008811D-03,   1.064310024920904D-02,  -1.456223124638055D-02, &
   1.704677152800424D-02,  -1.599144920970777D-02,   1.245354292665564D-02, &
  -8.053917848055689D-03,   4.484870390465530D-03,  -2.388649723600898D-03, &
   1.189468622215609D-03,  -5.392411429818440D-04,   1.401915007271368D-04, &
  -6.419412757249218D-05,   0.000000000000000D+00,   2.519242862768596D-03, &
   1.187732347088552D-03,   3.708928277994170D-03,  -4.209958518450967D-03, &
   6.830985362156513D-03,  -7.102685948737521D-03,   6.103921938853482D-03, &
  -3.939801647342860D-03,   2.219259199448546D-03,  -1.758780187506689D-03, &
   1.699881386372377D-03,  -1.594988586623455D-03,   1.378991441827527D-03, &
  -1.041851356163160D-03,   7.425159480060628D-04,  -9.673737843276095D-04, &
   0.000000000000000D+00,  -5.442755550237446D-04,   2.224584967860555D-04, &
   2.424214121115794D-03,  -7.282833322886310D-04,   4.375373089008650D-03, &
  -6.460748611877895D-03,   9.111622574641381D-03,  -1.023995728720538D-02, &
   8.848507030166607D-03,  -6.195773850114181D-03,   3.486925939951582D-03, &
  -1.414644885469618D-03,   2.017268959540615D-04,   4.665895990675903D-04, &
  -7.681789030270269D-04,   9.617372145288152D-04,  -1.603843584499964D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.270672228641981D-04, &
  -1.200064847199536D-03,   6.235007386738999D-04,  -1.885680151005556D-03, &
   3.678944735562160D-03,  -5.323379630905996D-03,   5.553063466662941D-03, &
  -5.060668970466414D-03,   4.168410525611549D-03,  -3.398032970713841D-03, &
   2.882299188353317D-03,  -2.236627547655164D-03,   1.477457893563655D-03, &
  -7.823418817853367D-04,   1.915070416749457D-04,   4.449473867087243D-04, &
  -1.025220559052228D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368492323476940D-04,  -5.646735695988578D-04,  -1.134505561566791D-03, &
  -1.397643290130740D-03,   7.936964710822747D-04,  -2.732019704921007D-03, &
   4.049959459425862D-03,  -5.068737935425852D-03,   4.863217555920816D-03, &
  -3.408045197653572D-03,   1.690639172802014D-03,  -2.512541034060450D-04, &
  -7.195109727012477D-04,   1.126275836149585D-03,  -1.162294196455502D-03, &
   9.436532319160378D-04,  -3.879131481597461D-04,   6.491323422168717D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.645433539783505D-04,   3.137467114806685D-04,  -3.254847649028467D-04, &
  -8.613834469990378D-04,   9.011782415879827D-04,  -1.862947394395914D-03, &
   2.497357616422591D-03,  -2.533894772584211D-03,   2.541262170799822D-03, &
  -2.442774638420883D-03,   2.130111881663168D-03,  -1.632705017150638D-03, &
   9.220758440335885D-04,  -1.849532523896441D-04,  -3.809525578196853D-04, &
   6.608542543442218D-04,  -6.963333367634845D-04,   1.861743074990350D-03 /)
lem(:, 16) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.794986263727882D-04,   0.000000000000000D+00,  -1.177532936345358D-03, &
   0.000000000000000D+00,   3.650903219678518D-03,   0.000000000000000D+00, &
  -1.530121851232757D-02,   0.000000000000000D+00,   1.020242680797499D-01, &
   0.000000000000000D+00,  -1.951592647277169D+00,  -4.737232977966085D-02, &
   3.957292913080571D+00,  -4.084238227953434D-02,  -1.959714073878897D+00, &
   0.000000000000000D+00,   1.029866776864025D-01,   0.000000000000000D+00, &
  -1.544972407637291D-02,   0.000000000000000D+00,   3.642719732464851D-03, &
   0.000000000000000D+00,   5.794986263736739D-04,  -2.992518038793609D-03, &
   6.895807729312304D-03,  -8.894083870854486D-03,   1.349395749113733D-04, &
   2.344111371446133D-02,  -1.946936977962383D-02,  -8.171172847043985D-02, &
  -6.108720078285084D-02,  -2.930738193723785D-03,   2.429077673183418D-02, &
   3.285142847455040D-02,   3.193902295650046D-02,   2.535649275180880D-02, &
   2.121739451357766D-02,   1.276349710993256D-02,  -3.502017449904296D-03, &
  -2.076605731291063D-02,  -3.361275426875310D-02,  -3.628408620433350D-02, &
  -2.426851604156430D-02,  -4.383571927571142D-03,   0.000000000000000D+00, &
  -7.481295096979534D-04,   1.994429335340119D-03,  -2.500008945934959D-03, &
  -9.171781259195512D-04,   9.396072162799972D-03,  -6.398904072050778D-03, &
  -3.298916747878706D-02,  -2.826469581784991D-02,  -7.329325818088248D-03, &
   4.216251536399529D-03,   8.003289245966693D-03,   7.770634146650432D-03, &
   5.561342920177481D-03,   2.507131083725348D-03,  -4.710854468849879D-03, &
  -1.729241956165931D-02,  -3.148520082701334D-02,  -4.095310032826884D-02, &
  -3.875657151315578D-02,  -2.317916179460707D-02,  -2.684178709436266D-03, &
  -5.914483088581510D-05,   5.330336610696398D-05,   5.914483088578951D-05, &
   5.038263371754218D-05,  -7.228812663819532D-05,  -4.490626048736094D-05, &
  -4.033740053534353D-04,   1.070239796982052D-05,   9.263937268019984D-05, &
  -9.075132781409354D-06,   8.667273365610284D-04,   2.850843435817935D-04, &
  -1.282097207085207D-04,  -5.987501398315879D-05,   1.037772727116589D-04, &
  -8.407726464638566D-05,  -7.575033246734562D-04,   1.514323880736854D-05, &
   2.199140797082171D-04,  -1.073561181784322D-04,  -1.131568064985967D-04, &
  -3.540365765838039D-03,   4.157662556342471D-04,   3.605482925486255D-04, &
  -4.783932064336446D-04,  -1.787075716291392D-05,   3.109003463047108D-04, &
  -1.887432033776927D-04,   8.848358773754543D-04,   1.731211019604667D-03, &
  -9.318738931911761D-04,   5.904288973908853D-04,   5.809670994692810D-04, &
  -6.490113924627487D-04,   1.165480091337009D-04,   2.284879820956160D-04, &
   7.443048384583820D-03,  -2.204130697677382D-03,   1.218722055683489D-03, &
   6.474666284814463D-04,  -1.739811770445253D-03,   8.810631066574595D-04, &
   1.777804998751192D-04,  -5.104116759845888D-04,   4.124427816301810D-04, &
   4.196654333741091D-03,  -3.208278493160690D-03,   2.730214343265721D-03, &
  -4.076715304335722D-03,   2.515169494283659D-03,  -4.322892195359949D-04, &
  -1.086227770462394D-03,   1.081517747553055D-03,  -4.385742092267916D-04, &
   7.225732203878188D-05,  -1.153896859327453D-02,   5.664071629015317D-03, &
  -6.698902941516488D-03,   3.985544548295972D-03,  -5.298522567743681D-04, &
  -2.078733113036489D-03,   3.007976085696509D-03,  -1.700929250580698D-03, &
   3.617588819618986D-04,   8.404258792718917D-05,  -1.471166406319160D-04, &
  -1.471083885708215D-02,   4.097638603148541D-03,  -4.467505530187001D-03, &
   6.222676768116928D-03,  -7.134284410306724D-03,   7.379570450837002D-03, &
  -4.866344511617466D-03,   1.680347036520952D-03,   1.753637634302632D-04, &
  -5.329569801330951D-04,   2.877745772077453D-04,  -1.511241503599994D-04, &
   1.481045702548160D-02,  -1.054795858028854D-02,   1.321718317786741D-02, &
  -1.281039331630118D-02,   1.204111226346953D-02,  -8.075296401345890D-03, &
   2.799550201526623D-03,   7.366573170831846D-04,  -1.995760948207897D-03, &
   1.377314299090151D-03,  -5.454763171305186D-04,   2.155290483385406D-04, &
  -1.629091289837813D-04,   1.528359468299693D-02,   2.154908912770068D-04, &
   2.779822274597299D-03,  -8.431802652905667D-04,   2.287972267375935D-03, &
  -5.554513632505744D-03,   7.553945223322881D-03,  -7.499699830233309D-03, &
   5.229038520215082D-03,  -2.551270370519291D-03,   9.670630263111687D-04, &
  -3.464009077435316D-04,   2.028986489241575D-04,  -1.780426956390201D-04, &
  -1.191720624599129D-02,   9.262212963178999D-03,  -1.113464888261821D-02, &
   1.434970424120491D-02,  -1.753745917112205D-02,   1.829606133929032D-02, &
  -1.654439133177068D-02,   1.178724839448836D-02,  -6.289541315945032D-03, &
   2.639133957878132D-03,  -8.767878767188261D-04,   3.839584512359496D-04, &
  -3.371920765901156D-04,   3.172764718536478D-04,  -3.521822725564998D-04, &
  -5.281554044499223D-03,  -4.500821451562435D-03,  -4.723313127106385D-04, &
  -5.012750372208420D-03,   5.798850598345773D-03,  -6.335932216762480D-03, &
   4.041391175278755D-03,  -3.333893851102991D-04,  -1.952280911525050D-03, &
   2.366316418933484D-03,  -1.551708514291985D-03,   5.397369329048398D-04, &
   1.231417316542442D-05,  -2.584031727994496D-04,   3.575792215170947D-04, &
  -5.060088795166263D-04,   3.913803454748216D-03,  -2.254620248689380D-03, &
   2.891140915836612D-03,  -7.189865710296249D-03,   8.539167575295990D-03, &
  -1.199145891145735D-02,   1.516954133475962D-02,  -1.584566410000184D-02, &
   1.375115244847151D-02,  -9.689907063801698D-03,   5.602267681838459D-03, &
  -2.908517241598819D-03,   1.365329933243775D-03,  -5.539747209351105D-04, &
   1.590880636270083D-04,   8.375860685068175D-05,  -1.981329949027824D-04, &
   0.000000000000000D+00,   2.358363738711021D-03,   1.093843562477128D-03, &
   3.735738399293885D-03,  -4.544601451180142D-03,   7.943140058370083D-03, &
  -9.623561697913605D-03,   9.840262118955251D-03,  -7.833941822533989D-03, &
   5.060844187524603D-03,  -3.259064818080003D-03,   2.344541529933424D-03, &
  -1.830771232340610D-03,   1.455665543040401D-03,  -1.048858225679758D-03, &
   7.023814326642265D-04,  -4.149004404785281D-04,   5.139279585511036D-04, &
   0.000000000000000D+00,  -5.283151506597398D-04,   2.268017136174672D-04, &
   2.076815360972200D-03,  -1.774819711561894D-04,   3.041081479033108D-03, &
  -4.461920891339691D-03,   6.916200475288919D-03,  -8.763117973501015D-03, &
   8.494799385780710D-03,  -6.497996377555422D-03,   3.844684655494238D-03, &
  -1.567297917793418D-03,   1.996618172750542D-04,   4.997309221527915D-04, &
  -7.824602365272270D-04,   7.996280555243166D-04,  -7.702876064337351D-04, &
   1.185446029226710D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.941724385532417D-04,  -1.143284304231506D-03,   5.260326130782800D-04, &
  -1.857952732513885D-03,   3.894391434564525D-03,  -6.247140046372003D-03, &
   7.418984293315779D-03,  -7.413477639712860D-03,   6.260823656204524D-03, &
  -4.777170772303774D-03,   3.600938476913295D-03,  -2.561905126092624D-03, &
   1.591252428286601D-03,  -8.042078416309590D-04,   1.834273176060731D-04, &
   2.337110136084642D-04,  -6.387070565460220D-04,   1.206561094710638D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.339679928267094D-04, &
  -5.353375310706649D-04,  -1.111644455572382D-03,  -1.203145273399015D-03, &
   3.342972475248404D-04,  -1.786195728602802D-03,   2.817336625974726D-03, &
  -3.963374126028072D-03,   4.233455532993571D-03,  -3.188226776480588D-03, &
   1.591280280515430D-03,  -1.326685014631848D-04,  -8.584969686114031D-04, &
   1.235520443612283D-03,  -1.187907477758979D-03,   9.225501192601438D-04, &
  -5.477189700812512D-04,  -1.649431206322900D-05,   1.766099774378761D-04 /)
lem(:, 17) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.226386535925193D-04,   0.000000000000000D+00, &
  -1.197871946034043D-03,   0.000000000000000D+00,   3.671892851487647D-03, &
   0.000000000000000D+00,  -1.535822080007497D-02,   0.000000000000000D+00, &
   1.023641287682402D-01,   0.000000000000000D+00,  -1.956111107405677D+00, &
  -4.084238227953434D-02,   3.934138924734935D+00,  -3.557532511813635D-02, &
  -1.962645907345427D+00,   0.000000000000000D+00,   1.031244411259724D-01, &
   0.000000000000000D+00,  -1.546094251754158D-02,   0.000000000000000D+00, &
   3.642617823018628D-03,   0.000000000000000D+00,   2.239879943971753D-04, &
  -1.877268697522699D-03,   6.582035048402805D-03,  -1.331403249278541D-02, &
   1.126075805391726D-02,   1.774015688655550D-02,  -4.148329364639721D-02, &
  -7.037879355848013D-02,  -4.019119343515042D-02,   4.322554538599269D-03, &
   2.532809841680225D-02,   2.535649275180843D-02,   2.497212554862433D-02, &
   2.300078944702522D-02,   1.746812777915902D-02,   1.222023228331503D-02, &
   2.810197727086810D-03,  -1.227742531977547D-02,  -2.593437800455405D-02, &
  -3.185526284975400D-02,  -2.696058649741498D-02,   0.000000000000000D+00, &
  -7.466266479892237D-05,   6.625654226549366D-04,  -2.343989145962994D-03, &
   4.689950310412833D-03,  -3.622446987562875D-03,  -7.400656430635275D-03, &
   1.509614368618103D-02,   2.975850873278377D-02,   1.987618400040763D-02, &
   2.963324140591913D-03,  -5.268932942058156D-03,  -6.168931123778670D-03, &
  -6.232813794056373D-03,  -5.333284475850453D-03,  -2.783723678910357D-03, &
   1.570688108288850D-03,   9.947696299624373D-03,   2.198459648521284D-02, &
   3.266773123804163D-02,   3.570332111410057D-02,   2.740604255437955D-02, &
   5.902601923566325D-06,  -5.312341731210145D-05,   5.902601923564945D-06, &
   1.404819257808932D-04,   3.974418628536274D-05,  -2.892274942548650D-05, &
  -1.751105237324962D-04,  -1.333988034726383D-04,  -4.131821346493561D-06, &
   5.213965032484372D-05,  -1.379062449415210D-04,   2.050681449238111D-04, &
   9.034047229772065D-05,  -5.823645040694043D-05,  -3.812097075637250D-05, &
   1.057996678118228D-03,  -5.366001748617091D-07,  -2.166510429842152D-04, &
   3.717617116276968D-05,   1.069846598646592D-04,  -6.235741198798692D-05, &
  -1.287733099652493D-03,  -7.253045696990045D-04,   2.548620859127936D-04, &
   8.447364371914213D-05,  -2.410056640162463D-04,  -1.051736342745132D-05, &
   1.708065431632433D-04,  -2.808457995233465D-03,   1.090264235300431D-03, &
  -1.193041055459470D-05,  -4.377507569419969D-04,   3.363109655509828D-04, &
   2.473011339249863D-04,  -2.814803292301310D-04,   6.448592601497481D-05, &
   2.881650259085663D-03,   8.319806844631167D-04,  -6.518164878016415D-04, &
   1.020514271397839D-03,   1.009638781406188D-05,  -7.360580470312223D-04, &
   4.310215106556207D-04,   5.400433593252312D-05,  -2.580082637684503D-04, &
   6.061972175503980D-03,  -2.610130570601622D-03,   1.847460742059178D-03, &
  -5.493726349295756D-04,  -1.325724048058817D-03,   1.365807393172654D-03, &
  -4.235125479521585D-04,  -4.224116576574953D-04,   4.561143408281851D-04, &
  -2.727647685773682D-04,   1.606350952056657D-03,  -1.589486375132135D-03, &
   9.924266919881203D-04,  -3.120894331557200D-03,   3.039522502200541D-03, &
  -1.617749774635797D-03,  -2.872539407270409D-04,   1.226840085266282D-03, &
  -7.648977972595573D-04,   2.326481562182589D-04,  -1.437112287529248D-05, &
  -1.067526743348504D-02,   5.428005281483746D-03,  -6.586337070962002D-03, &
   5.393723131075625D-03,  -2.644486973782450D-03,  -2.697665233550718D-04, &
   2.568140689403448D-03,  -2.384210926400556D-03,   1.006610778895708D-03, &
  -6.424654503200008D-05,  -1.418043046830401D-04,   1.659484539025773D-04, &
  -1.165405653029248D-02,   2.319511247280091D-03,  -1.683538519305836D-03, &
   3.061758952012161D-03,  -4.579340555666249D-03,   6.489572053230907D-03, &
  -5.857038883413349D-03,   3.196117043723816D-03,  -7.155115867757468D-04, &
  -4.842133137048132D-04,   4.440687554588018D-04,  -2.013245547627626D-04, &
   9.636676895285920D-05,   1.430687046340460D-02,  -9.607301753356084D-03, &
   1.199478387744894D-02,  -1.215166423148545D-02,   1.262089599392469D-02, &
  -1.052130036696053D-02,   5.873874014342703D-03,  -1.429586144093413D-03, &
  -1.310680132926790D-03,   1.671451492058808D-03,  -9.010620571365022D-04, &
   3.171779925416300D-04,  -1.164447718184619D-04,   5.634781081661182D-05, &
   1.300038553340092D-02,   1.280484008654371D-03,   6.366837583950558D-04, &
   2.077072287945357D-03,  -1.941817263815376D-03,  -9.040633711134954D-04, &
   4.191445811044446D-03,  -6.318880854197819D-03,   5.768628762983314D-03, &
  -3.509442192555839D-03,   1.558339299438377D-03,  -4.882918950266028D-04, &
   1.694212735592846D-04,  -1.216333581917231D-04,   1.456478215468529D-04, &
  -1.133555692945714D-02,   8.201851979139103D-03,  -9.807920333831578D-03, &
   1.224316375143186D-02,  -1.515454030927085D-02,   1.687174083766183D-02, &
  -1.705728118041509D-02,   1.410225633774619D-02,  -8.902587893314273D-03, &
   4.352163595990799D-03,  -1.580004085919714D-03,   4.783578433295332D-04, &
  -2.737495529240890D-04,   2.388988871405969D-04,  -2.097773164922552D-04, &
   2.269347473775514D-04,  -4.579024767873751D-03,  -4.548910495471846D-03, &
   2.089251010094764D-04,  -5.896433610952738D-03,   7.710087216243781D-03, &
  -9.680659992580550D-03,   8.801551494821954D-03,  -5.154509451397162D-03, &
   1.325152904394976D-03,   9.624844334119561D-04,  -1.334317587708677D-03, &
   6.852007996960281D-04,  -1.280696987715623D-04,  -1.685842798405071D-04, &
   2.641434884957382D-04,  -2.877721144186711D-04,   3.526119926455642D-04, &
   3.638414066402645D-03,  -1.875092698849757D-03,   2.423860158807245D-03, &
  -5.883507101267844D-03,   6.501389135883465D-03,  -9.041376341180027D-03, &
   1.208251586594655D-02,  -1.395526927773110D-02,   1.366981083074327D-02, &
  -1.081383421342154D-02,   6.798766428212812D-03,  -3.616803872218934D-03, &
   1.672008204830899D-03,  -6.467092486781277D-04,   1.882797668312693D-04, &
   5.010687667334219D-05,  -1.990487068108462D-04,   3.307330273829422D-04, &
   0.000000000000000D+00,   2.206816285498488D-03,   1.034136323148311D-03, &
   3.641273417607956D-03,  -4.542320910708007D-03,   8.310174361911867D-03, &
  -1.104682274775786D-02,   1.266627923515766D-02,  -1.158606925477711D-02, &
   8.475142556238126D-03,  -5.442285669275054D-03,   3.396144928957942D-03, &
  -2.237708677200072D-03,   1.595828665672563D-03,  -1.078245319071810D-03, &
   6.734520193604302D-04,  -3.814970846021432D-04,   1.565234139382218D-04, &
  -1.420186143679377D-04,   0.000000000000000D+00,  -5.112570257273658D-04, &
   2.221951854390676D-04,   1.796160007563428D-03,   2.487693367786800D-04, &
   1.894790648207347D-03,  -2.489652163375317D-03,   4.249220919027500D-03, &
  -6.235607357074073D-03,   6.975907656888894D-03,  -6.052350206358142D-03, &
   3.934283753381495D-03,  -1.690513572368749D-03,   2.127766144834170D-04, &
   5.254085796731316D-04,  -8.059511163757506D-04,   7.980859942067257D-04, &
  -6.854548955859771D-04,   5.574287593167621D-04,  -8.209419472946875D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.611429582507440D-04, &
  -1.088053474838930D-03,   4.188373294671743D-04,  -1.720494249822064D-03, &
   3.779387927108123D-03,  -6.536499284947641D-03,   8.560852689277004D-03, &
  -9.361119558737483D-03,   8.414318832066781D-03,  -6.452607971329137D-03, &
   4.571392694303493D-03,  -3.018474897818764D-03,   1.758821301943478D-03, &
  -8.358804167513448D-04,   1.687023957874680D-04,   2.510539769364918D-04, &
  -4.631228631073716D-04,   6.465643671951960D-04,  -1.121955112211936D-03 /)
lem(:, 18) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.877608462748176D-04,   0.000000000000000D+00,   4.595420421538579D-04, &
   0.000000000000000D+00,  -1.156686757567135D-03,   0.000000000000000D+00, &
   3.650497922236521D-03,   0.000000000000000D+00,  -1.537824321752542D-02, &
   0.000000000000000D+00,   1.026149667148347D-01,   0.000000000000000D+00, &
  -1.959714073878897D+00,  -3.557532511813635D-02,   3.915314138680694D+00, &
  -3.126526997403421D-02,  -1.965055173895256D+00,   0.000000000000000D+00, &
   1.032361316490849D-01,   0.000000000000000D+00,  -1.546838938601112D-02, &
   0.000000000000000D+00,  -2.877608462751273D-04,   1.485990620420048D-03, &
  -3.506505582541158D-03,   5.157477039380811D-03,  -3.005564015726252D-03, &
  -6.718575955263761D-03,   1.673623640560186D-02,   4.147102916058858D-03, &
  -5.089698229479365D-02,  -5.990790294561187D-02,  -2.105730964493247D-02, &
   8.435107919934452D-03,   2.121739451357734D-02,   2.300078944702485D-02, &
   1.933677295429847D-02,   1.844734834094815D-02,   1.655097673014105D-02, &
   1.122974931910804D-02,   4.761943041090671D-03,  -5.132496600951887D-03, &
  -1.812040179615686D-02,  -2.717979613926143D-02,   0.000000000000000D+00, &
   3.714976551047495D-04,  -1.018984528259454D-03,   1.505444099569651D-03, &
  -5.897280909636681D-04,  -3.015531051855232D-03,   6.332030071500284D-03, &
   2.426799162669793D-03,  -1.985918945969803D-02,  -2.566954029963526D-02, &
  -1.225654332916797D-02,  -4.589999663121133D-04,   4.797801455312015D-03, &
   5.638921747988966D-03,   4.788831729028195D-03,   4.463385393954068D-03, &
   3.267447841330807D-03,   5.284905613864993D-05,  -5.547430000262826D-03, &
  -1.443844386307562D-02,  -2.477348566336703D-02,  -3.107143758447351D-02, &
   2.936946838859863D-05,  -2.723350705125088D-05,  -2.936946838856615D-05, &
  -1.067980668675369D-05,   3.417538139761065D-05,   2.416306262879074D-05, &
   1.432388618048290D-04,  -1.121379702109609D-05,  -4.197892196532723D-05, &
  -4.106062040559259D-06,  -3.640843188667297D-04,  -1.004549089566297D-04, &
   5.227037204529333D-05,   3.350789347970495D-05,  -3.562969598243655D-05, &
   3.343103123460456D-04,   3.129506989725202D-04,  -8.032971480389637D-08, &
  -8.900359026085655D-05,   1.850964223305204D-05,   6.625323258043529D-05, &
   7.177800985017751D-04,  -3.623043510855776D-04,  -1.598815605583182D-04, &
   1.519162047377763D-04,   5.015767165054412D-05,  -1.110426832183548D-04, &
  -2.241394088017902D-06,  -2.176059157002534D-03,  -3.421869809606867D-04, &
   3.489465019644794D-04,  -1.576285054363642D-04,  -2.374364660667426D-04, &
   1.485442238153636D-04,   1.056448414569227D-04,  -1.638887635928501D-04, &
  -1.385170927272300D-03,   1.373714407371960D-03,  -3.624692781701956D-04, &
  -1.057182612381022D-04,   5.505425565821985D-04,  -3.430802322745807D-05, &
  -3.318497286066549D-04,   2.049271344011146D-04,   2.046267648380565D-05, &
   3.923303270713493D-03,  -1.734705743263165D-04,  -6.729318453565987D-05, &
   9.861015283774766D-04,  -6.086202291050018D-04,  -4.378148942479670D-04, &
   6.812266969301980D-04,  -2.170289037357982D-04,  -1.709107689969270D-04, &
   2.682377674446836D-04,   4.270396988032999D-03,  -2.363288651112244D-03, &
   1.852027580680816D-03,  -1.503633295493982D-03,  -4.584347474103440D-04, &
   1.342700180293335D-03,  -1.043776938295220D-03,  -1.484804145932533D-05, &
   5.217360796888261D-04,  -3.579189410170321D-04,   1.487998930956602D-04, &
  -1.434145469363694D-05,  -3.531964639977373D-04,  -5.209491807830007D-04, &
  -1.594165373072990D-03,   2.720468393159238D-03,  -2.388868386274926D-03, &
   8.336960075404252D-04,   8.623321227572489D-04,  -1.080714478830115D-03, &
   4.870176600281105D-04,  -8.988037699784990D-05,  -4.363031795642163D-05, &
  -9.509821152995214D-03,   4.787699212973159D-03,  -5.691846169767552D-03, &
   5.640668023410075D-03,  -4.061292866423309D-03,   1.710697102296338D-03, &
   1.319451649211176D-03,  -2.503867332881916D-03,   1.737007107701765D-03, &
  -4.995918877411871D-04,  -1.141590562027989D-04,   1.462483087358985D-04, &
  -1.032986054214191D-04,  -9.311672367681562D-03,   1.061309429761056D-03, &
   2.930742668733007D-04,   3.538335455809758D-04,  -1.667735886510134D-03, &
   4.431039113716207D-03,  -5.747846727999339D-03,   4.411294903211747D-03, &
  -1.979297425182824D-03,   6.266554598220412D-07,   5.433981302672163D-04, &
  -3.242697523962011D-04,   1.379844148888763D-04,  -9.445250836533843D-05, &
   1.360370494342046D-02,  -8.599523474177301D-03,   1.045804926231384D-02, &
  -1.063965261728157D-02,   1.172379334639750D-02,  -1.142678960230219D-02, &
   8.358106796982621D-03,  -4.028024986890290D-03,   1.585257582948452D-04, &
   1.505688829610223D-03,  -1.263253377145389D-03,   5.862758774195564D-04, &
  -1.766690959193467D-04,   6.356703637784317D-05,  -2.467128929359818D-05, &
   1.113253359011765D-02,   2.000685045948750D-03,  -8.148453622646178D-04, &
   4.021547068503580D-03,  -5.109626741346606D-03,   3.413074951815139D-03, &
   7.345475223656749D-06,  -3.764548304263401D-03,   5.343069704768904D-03, &
  -4.240494513314303D-03,   2.340225560004911D-03,  -8.690620246932798D-04, &
   2.200041466423209D-04,  -8.214452676871188D-05,   6.437054435011117D-05, &
  -5.291881759358013D-05,  -1.073952454152942D-02,   7.249098931208244D-03, &
  -8.536965680783848D-03,   1.013651588830682D-02,  -1.231182387721301D-02, &
   1.414135898037039D-02,  -1.568124551000366D-02,   1.491740672934998D-02, &
  -1.111248077208057D-02,   6.416727241372954D-03,  -2.746306232689293D-03, &
   8.382952305789160D-04,  -2.726108041888042D-04,   1.637143282303899D-04, &
  -1.552892634687103D-04,   1.428050686430141D-04,  -1.820072730357857D-04, &
  -3.996108389831299D-03,  -4.512064934940326D-03,   6.429932828422757D-04, &
  -6.283444046781997D-03,   8.670632881674150D-03,  -1.170104379657267D-02, &
   1.243905736544586D-02,  -9.856369559401180D-03,   5.476872595844049D-03, &
  -1.483754255750894D-03,  -5.132109235314366D-04,   6.603812716468078D-04, &
  -2.713847047750129D-04,  -5.969727120885923D-05,   2.187744717426349D-04, &
  -2.330909218720559D-04,   2.182707165500841D-04,  -2.615856925499649D-04, &
   3.391394905551710D-03,  -1.557556940288941D-03,   2.029044688807138D-03, &
  -4.777185875798953D-03,   4.700191012213930D-03,  -6.151430340726283D-03, &
   8.400798687569110D-03,  -1.063610813992622D-02,   1.192520136794275D-02, &
  -1.087609130418435D-02,   7.737172734505927D-03,  -4.449362869673815D-03, &
   2.121117536093115D-03,  -8.107413644683344D-04,   2.348628216060099D-04, &
   2.369990551276532D-05,  -1.365082328366315D-04,   2.100901986023727D-04, &
  -2.848023075435894D-04,   0.000000000000000D+00,   2.065648086465403D-03, &
   9.952002532857245D-04,   3.477205449611476D-03,  -4.325536434836449D-03, &
   8.119947848128743D-03,  -1.143388732434786D-02,   1.424649731140172D-02, &
  -1.453464255144274D-02,   1.193368924451464D-02,  -8.191882508696294D-03, &
   4.968453101144872D-03,  -2.910613139730545D-03,   1.830745390220171D-03, &
  -1.157151281354344D-03,   6.872100218476441D-04,  -3.783899565135351D-04, &
   1.516221203015977D-04,   2.108318464031015D-05,  -9.390907031818552D-05, &
   0.000000000000000D+00,  -4.937114797933228D-04,   2.117167495052182D-04, &
   1.570379498293325D-03,   5.605666151994539D-04,   9.805665661694934D-04, &
  -7.604566835230654D-04,   1.532444413203296D-03,  -3.049419713825020D-03, &
   4.338928916816620D-03,  -4.591292178810030D-03,   3.474185697505299D-03, &
  -1.651346878426540D-03,   2.062074994716629D-04,   5.599110439178402D-04, &
  -8.398834494432101D-04,   7.998577772720732D-04,  -6.525717565017621D-04, &
   4.832550051986215D-04,  -3.332658375856092D-04,   4.427269701623289D-04 /)
lem(:, 19) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.367405092878151D-04,   0.000000000000000D+00, &
   4.646187436721097D-04,   0.000000000000000D+00,  -1.161460122785426D-03, &
   0.000000000000000D+00,   3.656433216322437D-03,   0.000000000000000D+00, &
  -1.540774561443011D-02,   0.000000000000000D+00,   1.028237486389946D-01, &
   0.000000000000000D+00,  -1.962645907345427D+00,  -3.126526997403421D-02, &
   3.899811373694855D+00,  -2.769364170995417D-02,  -1.967064952565583D+00, &
   0.000000000000000D+00,   1.033319358844060D-01,   0.000000000000000D+00, &
  -1.547608741217421D-02,   0.000000000000000D+00,  -1.014602182650944D-04, &
   8.503495569320452D-04,  -3.045896485452536D-03,   6.734834171467607D-03, &
  -8.574383115809273D-03,   2.749592040630680D-04,   1.964011518216980D-02, &
  -1.375098411627560D-02,  -5.215959155528961D-02,  -4.544983030536835D-02, &
  -1.097792640789897D-02,   1.276349710993269D-02,   1.746812777915893D-02, &
   1.844734834094780D-02,   1.755663523108865D-02,   1.463812150481302D-02, &
   1.367522367894640D-02,   1.150444308305112D-02,   5.861339006417085D-03, &
  -1.580971234949462D-03,  -1.117324950917764D-02,   0.000000000000000D+00, &
   3.382007275431738D-05,  -3.001233730356332D-04,   1.086172874128358D-03, &
  -2.389803270582697D-03,   2.939658387561047D-03,   3.229283313569294D-04, &
  -7.637368729126264D-03,   4.377579100522022D-03,   2.118104517046833D-02, &
   2.025333793755107D-02,   7.490935351192359D-03,  -1.629225480904561D-03, &
  -4.047336887142651D-03,  -4.610403748412990D-03,  -4.348914064853915D-03, &
  -3.554469502254927D-03,  -2.992005378764415D-03,  -1.276391566881921D-03, &
   2.783411749604467D-03,   9.305967780964928D-03,   1.785132846855999D-02, &
  -2.673711513489189D-06,   2.406340362126428D-05,  -2.673711513481203D-06, &
  -6.445633458513928D-05,  -1.818565764950110D-05,   1.316968646306631D-05, &
   9.256345066066837D-05,   6.069988039268694D-05,   3.447099141339309D-06, &
  -2.510195272153777D-05,  -5.042978373437529D-06,  -1.009013058487248D-04, &
  -4.356382023813986D-05,   2.199980901422409D-05,   2.550199667905142D-05, &
  -3.392741039494141D-04,   4.796476412050482D-05,   1.005440744206153D-04, &
  -3.012528950454481D-06,  -4.721945782912797D-05,   7.196648087140331D-06, &
   7.492137402963013D-04,   2.287606947258683D-04,  -1.197958549417336D-04, &
  -5.194862098598054D-05,   8.194144610635579D-05,   3.282266771806025D-05, &
  -7.064566653392854D-05,  -4.930530267545864D-05,  -5.950597386561874D-04, &
   6.078600121268727D-06,   1.615002397976606D-04,  -9.119348513142482D-05, &
  -1.307834952585830D-04,   6.702717562890073D-05,   6.779433082900004D-05, &
  -2.408606861379079D-03,   2.198630542071861D-04,   2.458658760337240D-04, &
  -3.194213932823359D-04,  -3.708220159154613D-05,   2.811859598489711D-04, &
  -1.028947013301582D-05,  -1.559195138637407D-04,   1.168750239152281D-04, &
   1.938267229637583D-04,   1.186679662826843D-03,  -5.048512553458406D-04, &
   3.275635745205557D-04,   4.579113057140841D-04,  -3.617548132284326D-04, &
  -1.850682041952317D-04,   3.311379173759305D-04,  -1.050927659202793D-04, &
  -8.949419182058205D-05,   4.037721067659933D-03,  -9.448523999883148D-04, &
   5.550150142411950D-04,   5.072009075954646D-04,  -9.613979831407058D-04, &
   9.073093983861335D-05,   6.145188266550080D-04,  -5.547960850394113D-04, &
   8.808951098549595D-07,   2.251503933675068D-04,  -2.165928773503563D-04, &
   2.556415442142907D-03,  -1.729162155168717D-03,   1.326691233628078D-03, &
  -1.890162276363080D-03,   4.960719348378185D-04,   8.152365743546703D-04, &
  -1.322250026845828D-03,   6.138749069219938D-04,   3.443780240757819D-04, &
  -4.870113880412887D-04,   2.505004006348703D-04,  -8.922264299499395D-05, &
  -8.647373334520423D-04,   4.087306263000699D-04,  -1.507920419139209D-03, &
  -2.561898116945866D-05,   1.720579452074627D-03,  -2.457844920947941D-03, &
   1.841923293008857D-03,   1.769115345859336D-05,  -1.062029701129524D-03, &
   8.484795855795065D-04,  -2.666753336848037D-04,   2.565528287961512D-06, &
   8.458265671008095D-05,  -8.312381234447412D-03,   4.011936904062333D-03, &
  -4.444027321041358D-03,   4.958413728649818D-03,  -4.480773161449224D-03, &
   3.278488466224246D-03,  -4.137773926973189D-04,  -1.876739424737398D-03, &
   2.177764038815586D-03,  -1.167763326391733D-03,   1.192669038508025D-04, &
   1.818317897816633D-04,  -1.292437077291569D-04,   7.915321187341069D-05, &
  -7.536259664864041D-03,   2.321601605521948D-04,   1.504811477656242D-03, &
  -1.565618241905439D-03,   9.227761539393004D-04,   1.748101527372437D-03, &
  -4.454020940772245D-03,   4.862000463210649D-03,  -3.255502744518317D-03, &
   9.333945296135104D-04,   3.731898386826366D-04,  -4.802027070012210D-04, &
   2.298027504003924D-04,  -8.952219547821786D-05,   3.594751353421542D-05, &
   1.282164568002385D-02,  -7.632472956814787D-03,   8.899429427072093D-03, &
  -8.775711785171180D-03,   9.845654399657264D-03,  -1.076533300100476D-02, &
   9.617417103590334D-03,  -6.448973273453878D-03,   2.248710582801701D-03, &
   6.864605164906502D-04,  -1.390813144190906D-03,   9.371117533176882D-04, &
  -3.452897478157301D-04,   8.495072595916488D-05,  -3.654387954575773D-05, &
   4.226364980067277D-05,   9.599588157308945D-03,   2.466281088061150D-03, &
  -1.735078340822483D-03,   5.126704453517555D-03,  -7.080606712498764D-03, &
   6.727637808893711D-03,  -4.149652282245987D-03,  -1.851764476670635D-04, &
   3.694159611092551D-03,  -4.355528877968218D-03,   3.103216357703009D-03, &
  -1.463149839845815D-03,   4.277199091813351D-04,  -9.742827904233544D-05, &
   4.318729326958226D-05,  -3.464070273547708D-05,   2.375179299593979D-05, &
  -1.015337488528637D-02,   6.409438908279238D-03,  -7.389739411970418D-03, &
   8.220984909973429D-03,  -9.479063665567828D-03,   1.077627432478414D-02, &
  -1.279910286278401D-02,   1.391562281043947D-02,  -1.223412374063357D-02, &
   8.413125523782239D-03,  -4.322516394918099D-03,   1.559002001436316D-03, &
  -4.311582314150004D-04,   1.277799709059544D-04,  -9.601010554906479D-05, &
   9.690351927347912D-05,  -8.622144939806428D-05,   8.595724253192032D-05, &
  -3.508965075278462D-03,  -4.419789446182310D-03,   9.041494799728739D-04, &
  -6.330427439382061D-03,   8.910576524251205D-03,  -1.249953393386382D-02, &
   1.459282196290696D-02,  -1.361593823809748D-02,   9.810413845054825D-03, &
  -4.850550551577378D-03,   1.142578403500102D-03,   2.584758204740255D-04, &
  -3.372044080482774D-04,   7.119242650302149D-05,   1.410511222601645D-04, &
  -1.874682072231190D-04,   1.785018429640101D-04,  -1.653481438814362D-04, &
   2.159642619828938D-04,   3.169297324278317D-03,  -1.292052873700877D-03, &
   1.700884175431589D-03,  -3.865749045438841D-03,   3.206327409185232D-03, &
  -3.606200419582606D-03,   4.708601992699185D-03,  -6.491454708249489D-03, &
   8.647431978617386D-03,  -9.478619900516579D-03,   7.949633888778378D-03, &
  -5.198259001541403D-03,   2.697716238155838D-03,  -1.082340696205420D-03, &
   3.272934966341982D-04,  -3.594554073169567D-06,  -1.295052353066428D-04, &
   1.693758191707653D-04,  -1.906112936510688D-04,   2.523642358192572D-04, &
   0.000000000000000D+00,   1.935065036032065D-03,   9.683882110346160D-04, &
   3.278716925102082D-03,  -3.990220621817281D-03,   7.568826519317641D-03, &
  -1.099768122379781D-02,   1.453818091170681D-02,  -1.620061001192015D-02, &
   1.478374971519912D-02,  -1.113210612470908D-02,   7.044795265987561D-03, &
  -3.959607601020093D-03,   2.230304381715701D-03,  -1.289230323274360D-03, &
   7.167996599123155D-04,  -3.808566022464827D-04,   1.565406779670816D-04, &
  -1.463403464006981D-05,  -9.741101158659648D-05,   1.525493620321472D-04 /)
lem(:, 20) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.558791572524330D-04,   0.000000000000000D+00,  -2.052236144701433D-04, &
   0.000000000000000D+00,   4.411940769698106D-04,   0.000000000000000D+00, &
  -1.144938432196503D-03,   0.000000000000000D+00,   3.647495840096974D-03, &
   0.000000000000000D+00,  -1.542208233586376D-02,   0.000000000000000D+00, &
   1.029866776864025D-01,   0.000000000000000D+00,  -1.965055173895256D+00, &
  -2.769364170995417D-02,   3.886886760024335D+00,  -2.470088952785241D-02, &
  -1.968754880441471D+00,   0.000000000000000D+00,   1.034117927024861D-01, &
   0.000000000000000D+00,   1.558791572557201D-04,  -8.049565067620859D-04, &
   1.928186170855228D-03,  -3.052498847696235D-03,   2.707506803805866D-03, &
   1.243351963088037D-03,  -8.138136684189931D-03,   7.776525600194754D-03, &
   1.395634363957769D-02,  -2.566329663596905D-02,  -5.065358078694578D-02, &
  -3.056634007584774D-02,  -3.502017449904066D-03,   1.222023228331496D-02, &
   1.655097673014073D-02,   1.463812150481276D-02,   1.427360786396908D-02, &
   1.347813741870264D-02,   1.120311540766375D-02,   9.920049688718706D-03, &
   7.183679978111494D-03,   1.036726152421555D-03,   0.000000000000000D+00, &
  -2.012391266888751D-04,   5.619713580869074D-04,  -9.076782563967507D-04, &
   7.048241712452007D-04,   7.991659988600480D-04,  -3.255262290729337D-03, &
   2.726743610820762D-03,   5.865381823714143D-03,  -9.618097901075547D-03, &
  -2.080388009322497D-02,  -1.454770447616111D-02,  -4.094661896473301D-03, &
   2.172244824933927D-03,   3.951239439461586D-03,   3.632503997783493D-03, &
   3.587170867229401D-03,   3.304753164243687D-03,   2.547830730927814D-03, &
   1.591117992929471D-03,  -8.200419210562402D-04,  -5.691216025127620D-03, &
  -1.590934986714575D-05,   1.501931630815663D-05,   1.590934986711572D-05, &
   7.787793641018081D-07,  -1.791192537486616D-05,  -1.373989306706685D-05, &
  -5.689134965201487D-05,   8.232810420746795D-06,   2.125460822998211D-05, &
   5.302273787576912D-06,   1.618445114900590D-04,   3.918170463143223D-05, &
  -2.486025554570149D-05,  -1.860726409285975D-05,   1.253948558941879D-05, &
  -2.191209718166122D-04,  -1.382049400824633D-04,   3.788866636674523D-07, &
   4.081647999495042D-05,  -4.582914277572696D-07,  -3.257204817883572D-05, &
  -6.319705069720213D-05,   2.124651701581432D-04,   7.455937159964132D-05, &
  -5.893885473844315D-05,  -3.263953979705877D-05,   3.988010672977766D-05, &
   2.497401393246590D-05,   8.516067379767855D-04,  -2.242355527754033D-05, &
  -1.709498504390948D-04,   4.766528156285560D-05,   1.014750997993446D-04, &
  -3.277148049245367D-05,  -6.665656226586406D-05,   4.290389932987583D-05, &
  -9.203900607446921D-04,  -5.612095047398318D-04,   1.713376841184476D-04, &
   3.803851498615493D-05,  -2.026872976560361D-04,  -3.988302423489251D-05, &
   1.374400431801716D-04,  -5.360526639603156D-06,  -9.817474942240231D-05, &
  -1.957978469061993D-03,   7.054616271289605D-04,   1.106417966592880D-05, &
  -2.794880703424837D-04,   2.311660375399540D-04,   2.360809916848063D-04, &
  -1.901217247084389D-04,  -9.745414563115886D-05,   1.623903636715851D-04, &
  -6.861962555038703D-05,   1.472465162627364D-03,   6.636789527987562D-04, &
  -3.675048371572249D-04,   6.242383126423238D-04,   1.021951956321003D-04, &
  -5.067763268043299D-04,   1.239889831190321D-04,   3.186756099841508D-04, &
  -2.692159299277324D-04,   8.773560441600743D-06,   1.406844681968756D-04, &
   3.493699588204002D-03,  -1.315255038540758D-03,   9.573058364070767D-04, &
  -1.917037175419808D-04,  -8.888062826897528D-04,   5.687865850477055D-04, &
   2.248465929617648D-04,  -7.077527074233154D-04,   3.284709076328584D-04, &
   1.563948322762783D-04,  -2.295441931929616D-04,   1.552260417949900D-04, &
   1.184594933388268D-03,  -9.919503482339201D-04,   5.352002245386083D-04, &
  -1.645917635057253D-03,   1.155936513394940D-03,   1.459523610102199D-05, &
  -1.097269976499565D-03,   1.138213501685790D-03,  -1.363854612475654D-04, &
  -4.904109355050203D-04,   3.987511607028012D-04,  -1.540577641021281D-04, &
   2.420425380257632D-05,  -1.188314002188381D-03,   7.557656392424595D-04, &
  -1.906495590383563D-03,   1.150227598439748D-03,   4.463480683409276D-04, &
  -1.809254922973495D-03,   2.344908774813515D-03,  -1.009974278325739D-03, &
  -5.707370727845071D-04,   1.052714552635098D-03,  -5.930175411448682D-04, &
   9.530057148796197D-05,   4.497656286445300D-05,  -6.309814139932560D-05, &
  -7.220140746224617D-03,   3.270184801457671D-03,  -3.188116494055515D-03, &
   3.766223246342766D-03,  -3.969819401561566D-03,   3.998339717743167D-03, &
  -2.107740978292657D-03,  -6.014113877501633D-04,   2.037911572575159D-03, &
  -1.783122512452949D-03,   6.257511026780638D-04,   9.515417668472858D-05, &
  -1.800921220752248D-04,   1.028260851983301D-04,  -7.851547422307548D-05, &
  -6.191706121813095D-03,  -2.794321699926283D-04,   2.113799427806252D-03, &
  -2.654099695051735D-03,   2.748190894708729D-03,  -8.631241581208830D-04, &
  -2.288716735134133D-03,   4.270634256177861D-03,  -4.097858684234824D-03, &
   2.109910056300178D-03,  -1.989739166686104D-04,  -4.928231540085020D-04, &
   3.916082357513482D-04,  -1.554715260369362D-04,   5.766384267729560D-05, &
  -1.916467611332851D-05,   1.203013224196182D-02,  -6.757929924458944D-03, &
   7.481931358914708D-03,  -6.931178848249555D-03,   7.558799317800794D-03, &
  -8.925275117486387D-03,   9.385787834004802D-03,  -8.022333430292603D-03, &
   4.524097918028975D-03,  -8.063511983519508D-04,  -1.051628279576236D-03, &
   1.208639821426812D-03,  -6.383538457838219D-04,   1.839386569297003D-04, &
  -3.868359903300404D-05,   1.864401757542988D-05,  -3.899997064050096D-06, &
   8.334978107915381D-03,   2.749898822471807D-03,  -2.274509130724119D-03, &
   5.600896580713692D-03,  -7.985114497759658D-03,   8.742204683810472D-03, &
  -7.495472645400192D-03,   3.706393951808622D-03,   9.130583895070671D-04, &
  -3.504739516693463D-03,   3.538520682769807D-03,  -2.167796599134706D-03, &
   8.300648647703142D-04,  -2.008212202814706D-04,   2.606084259074267D-05, &
  -1.591593922322716D-05,   2.145941395845339D-05,  -3.890161407911733D-05, &
  -9.590683274529554D-03,   5.677604561119694D-03,  -6.389790605427370D-03, &
   6.581305850204596D-03,  -6.949955459282474D-03,   7.380553405718520D-03, &
  -9.079085016547842D-03,   1.126696447166283D-02,  -1.177662604593452D-02, &
   9.741410910322695D-03,  -6.061523819545971D-03,   2.677912006478384D-03, &
  -8.394591440223850D-04,   1.743464784748592D-04,  -4.290615722202998D-05, &
   6.483030244396329D-05,  -6.682374327438416D-05,   5.178012379072432D-05, &
  -4.841936928652733D-05,  -3.098998831776758D-03,  -4.292909388269924D-03, &
   1.047734097660547D-03,  -6.162941701406177D-03,   8.665614436956988D-03, &
  -1.234458853520320D-02,   1.526650607467589D-02,  -1.588714696378768D-02, &
   1.349645731329116D-02,  -8.651719224486324D-03,   3.676847166577760D-03, &
  -7.525694646553572D-04,  -1.902799253574251D-04,   1.834325372907926D-04, &
   4.200183475246609D-05,  -1.378119286883153D-04,   1.463629600138583D-04, &
  -1.248532260501766D-04,   1.108129384919682D-04,  -1.217568984791668D-04, &
   2.969072307435329D-03,  -1.069879464703071D-03,   1.430659678476454D-03, &
  -3.128071501528734D-03,   2.024658353270220D-03,  -1.540519102552758D-03, &
   1.436345864909935D-03,  -2.203338841924259D-03,   4.340109191015350D-03, &
  -6.561864426595789D-03,   6.995791913346683D-03,  -5.511158456041849D-03, &
   3.294143397783479D-03,  -1.468279145086675D-03,   4.849307660302913D-04, &
  -5.987493322247880D-05,  -1.092526579413166D-04,   1.480669140435896D-04, &
  -1.563912575892813D-04,   1.650107033680689D-04,  -2.263979354895397D-04 /)
lem(:, 21) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.178093986337100D-04,   0.000000000000000D+00, &
  -2.059311987459160D-04,   0.000000000000000D+00,   4.428397574275795D-04, &
   0.000000000000000D+00,  -1.145314641774736D-03,   0.000000000000000D+00, &
   3.649172637441531D-03,   0.000000000000000D+00,  -1.543948877550821D-02, &
   0.000000000000000D+00,   1.031244411259724D-01,   0.000000000000000D+00, &
  -1.967064952565583D+00,  -2.470088952785241D-02,   3.876002874828104D+00, &
  -2.216834216811315D-02,  -1.970192362278671D+00,   0.000000000000000D+00, &
   1.034811625612696D-01,   0.000000000000000D+00,   5.048974227136128D-05, &
  -4.231602366322541D-04,   1.536613662744333D-03,  -3.579641041666869D-03, &
   5.442939695039919D-03,  -3.563147063056099D-03,  -5.695894721822204D-03, &
   1.471258571802722D-02,   2.639241778986717D-03,  -3.326560981169630D-02, &
  -4.321180391645125D-02,  -2.076605731291064D-02,   2.810197727086719D-03, &
   1.122974931910792D-02,   1.367522367894621D-02,   1.347813741870260D-02, &
   1.144239437696908D-02,   1.113149648458182D-02,   1.049521525625626D-02, &
   8.498279341487593D-03,   6.694810799231440D-03,   0.000000000000000D+00, &
  -1.682991409127283D-05,   1.493506717520474D-04,  -5.484261550117509D-04, &
   1.275262872858023D-03,  -1.903382984590972D-03,   1.104591881462571D-03, &
   2.393761218722895D-03,  -5.484419495491444D-03,  -1.739822939238033D-03, &
   1.307606198308565D-02,   1.829060079582972D-02,   1.030921142628746D-02, &
   1.393006340675320D-03,  -2.258996657527491D-03,  -3.352437879492823D-03, &
  -3.344289727902568D-03,  -2.851648585829372D-03,  -2.777893418037593D-03, &
  -2.482038934590728D-03,  -1.599253772401908D-03,  -1.029111282706408D-04, &
   1.330521533795894D-06,  -1.197469380393175D-05,   1.330521533782423D-06, &
   3.234190805159864D-05,   9.108955115799428D-06,  -6.575846811132431D-06, &
  -5.035512266260671D-05,  -3.029495184582321D-05,  -2.226064873802457D-06, &
   1.297258495423395D-05,   2.519259142930877D-05,   5.288590488072803D-05, &
   2.251276073928044D-05,  -9.463057067834019D-06,  -1.543139268511098D-05, &
   1.102500632508157D-04,  -4.061060911651033D-05,  -5.119734494067993D-05, &
  -2.860549725700103D-06,   2.236348860512716D-05,   4.458973810104146D-06, &
  -3.459599332242671D-04,  -6.882997599840368D-05,   6.445031279479412D-05, &
   3.057937344547918D-05,  -3.100525817430756D-05,  -2.272403202773785D-05, &
   2.517657721776176D-05,   3.385508106220509D-04,   2.745340445831284D-04, &
  -1.514443411735238D-05,  -7.433618083414769D-05,   2.642072042037829D-05, &
   6.161655779737772D-05,  -1.064909283795142D-05,  -4.901426793648641D-05, &
   5.460671818415206D-04,  -3.104173300145585D-04,  -1.157723717630489D-04, &
   1.223790711733187D-04,   2.551846721515449D-05,  -1.117282209897107D-04, &
  -3.575241110859102D-05,   6.978908412795506D-05,   2.416534781398749D-06, &
  -1.540860905030418D-03,  -2.678745864614362D-04,   2.337541423728226D-04, &
  -1.325867060048824D-04,  -1.741178255502510D-04,   1.206926264383870D-04, &
   1.347964371497138D-04,  -9.141551640871553D-05,  -5.235235477345085D-05, &
   1.003393487404457D-04,  -1.058181321145315D-03,   9.200190878118134D-04, &
  -2.155438877778548D-04,  -4.862152355132702D-05,   3.867004453308165D-04, &
   2.153924876578590D-05,  -2.772368469759070D-04,   6.858984589929216D-05, &
   1.711112437964232D-04,  -1.320695862805792D-04,   8.708266579776190D-06, &
   2.216583080238109D-03,   4.447012357096642D-05,  -2.727470732871151D-05, &
   6.237345359551439D-04,  -3.243006928979775D-04,  -3.774336741858455D-04, &
   3.875309141616602D-04,   8.727999164986420D-05,  -3.795071776936796D-04, &
   1.550160852864174D-04,   6.644900884441009D-05,  -1.412584045183279D-04, &
   2.622633396447572D-03,  -1.301286612815375D-03,   1.012452516330490D-03, &
  -8.022538349660280D-04,  -4.465345293212988D-04,   7.586161263687393D-04, &
  -2.840999908552721D-04,  -5.215957545994356D-04,   6.333909442726096D-04, &
  -8.359141839983097D-05,  -2.324087477538754D-04,   1.986832507234165D-04, &
  -1.158686995786323D-04,   2.330459640301235D-04,  -3.535239578571324D-04, &
  -2.365897569614193D-04,  -9.597439563818010D-04,   1.298500200709627D-03, &
  -7.104550219254789D-04,  -4.504040007833963D-04,   1.275606667619852D-03, &
  -7.244968233331056D-04,  -2.050256532868958D-04,   5.190993614301180D-04, &
  -2.895446182379944D-04,   7.934530948365506D-05,   2.292184907545148D-05, &
  -1.198544032400023D-03,   8.127439615122110D-04,  -1.830726352526702D-03, &
   1.734513224997521D-03,  -6.512783056825856D-04,  -7.148137999688767D-04, &
   2.142255221422325D-03,  -1.824846522756530D-03,   2.765815802783391D-04, &
   8.742876407291390D-04,  -9.216964710834900D-04,   3.378402910656513D-04, &
   1.077335104151745D-05,  -6.411876673603882D-05,   5.956353956038590D-05, &
  -6.285073432119700D-03,   2.643887941648254D-03,  -2.130165616565436D-03, &
   2.475530534627560D-03,  -2.869039894136251D-03,   3.761987138059861D-03, &
  -3.232188437926192D-03,   9.596957922306373D-04,   1.230105772288011D-03, &
  -2.028584400124235D-03,   1.259931815668552D-03,  -2.233490350556952D-04, &
  -1.882329354546688D-04,   1.579741137602082D-04,  -7.799766599276707D-05, &
   3.766588741845729D-05,  -5.166498729007805D-03,  -5.742737275129336D-04, &
   2.311152076670815D-03,  -3.059442870837203D-03,   3.678849148422719D-03, &
  -2.830237103770420D-03,   1.434396311866492D-04,   2.700146111399280D-03, &
  -4.117877947786314D-03,   3.172560902480730D-03,  -1.123326251335282D-03, &
  -2.076966966511556D-04,   5.079160379908040D-04,  -2.848956754358986D-04, &
   9.280813544700205D-05,  -3.562473495757343D-05,   3.160505072115290D-05, &
   1.126654447273055D-02,  -5.992881965116243D-03,   6.273079174315280D-03, &
  -5.319974233718400D-03,   5.332698856113390D-03,  -6.500076943536734D-03, &
   7.850959634952807D-03,  -8.292908931716711D-03,   6.366217011028160D-03, &
  -2.743079441040552D-03,  -1.107435210681133D-04,   1.192624770343383D-03, &
  -9.553645451304868D-04,   3.899892863016681D-04,  -9.473079560002069D-05, &
   2.020789021627878D-05,  -1.013401826919712D-05,  -2.565174908461146D-06, &
   7.285142882974730D-03,   2.906556492344311D-03,  -2.555588182601501D-03, &
   5.651786608465147D-03,  -8.087745487576743D-03,   9.508216911981343D-03, &
  -9.557651620277819D-03,   7.090115270113159D-03,  -2.504435499471269D-03, &
  -1.563904825153295D-03,   3.292091009041387D-03,  -2.767434019513882D-03, &
   1.396175359142639D-03,  -4.525176543753730D-04,   6.996483820687986D-05, &
   5.806150315204216D-06,   3.085376087133469D-06,  -1.041223384969494D-05, &
   4.457977816983538D-06,  -9.058411611988773D-03,   5.043374202929146D-03, &
  -5.535796497980130D-03,   5.232509360689294D-03,  -4.858931075800627D-03, &
   4.372737673453900D-03,  -5.227829989902291D-03,   7.542335466371721D-03, &
  -9.664554887353584D-03,   9.832036309854641D-03,  -7.501624250447475D-03, &
   4.090713799611313D-03,  -1.576771908953254D-03,   3.854724143546503D-04, &
  -2.766607922145348D-05,   1.644128391638637D-05,  -3.552980732570401D-05, &
   3.928766826263231D-05,  -3.447924244462461D-05,   5.175834027410020D-05, &
  -2.751646711814412D-03,  -4.145872643030815D-03,   1.113356575773613D-03, &
  -5.872921425368955D-03,   8.134848726640380D-03,  -1.155448317898824D-02, &
   1.472743921232727D-02,  -1.651504563384562D-02,   1.584894791125475D-02, &
  -1.214921996516293D-02,   6.820421345634381D-03,  -2.518044751714025D-03, &
   3.654538472259741D-04,   1.900087974063368D-04,  -7.160424447380326D-05, &
  -6.405150447347882D-05,   1.107617028746795D-04,  -1.135999201336056D-04, &
   9.274662874304234D-05,  -7.269121020787796D-05,   7.880298467483956D-05 /)
lem(:, 22) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.034263192610999D-05,   0.000000000000000D+00,   1.015838116292139D-04, &
   0.000000000000000D+00,  -1.926480304397434D-04,   0.000000000000000D+00, &
   4.326006019816508D-04,   0.000000000000000D+00,  -1.137408961704299D-03, &
   0.000000000000000D+00,   3.644798835623836D-03,   0.000000000000000D+00, &
  -1.544972407637291D-02,   0.000000000000000D+00,   1.032361316490849D-01, &
   0.000000000000000D+00,  -1.968754880441471D+00,  -2.216834216811315D-02, &
   3.866748715719008D+00,  -2.000625293121219D-02,  -1.971423167883841D+00, &
   0.000000000000000D+00,  -9.034263192756957D-05,   4.665273452126936D-04, &
  -1.128921982094394D-03,   1.871868416990642D-03,  -2.010716279484798D-03, &
   3.945107522587057D-04,   3.436453788969551D-03,  -6.318541926995351D-03, &
   2.774925049804435D-04,   1.502098766956141D-02,  -7.476832556717054D-03, &
  -3.775753628671193D-02,  -3.361275426875297D-02,  -1.227742531977541D-02, &
   4.761943041090858D-03,   1.150444308305127D-02,   1.120311540766364D-02, &
   1.113149648458168D-02,   1.061356039595211D-02,   9.072572225844020D-03, &
   8.792069966295699D-03,   8.173831096372207D-03,   0.000000000000000D+00, &
   1.166318363024254D-04,  -3.296682188796275D-04,   5.626689111338748D-04, &
  -5.664638297368640D-04,  -6.952681118953512D-05,   1.473153746702824D-03, &
  -2.393724667243649D-03,  -1.568911142912880D-04,   5.991433025000204D-03, &
  -2.468643838910380D-03,  -1.505139790806519D-02,  -1.476196459359912D-02, &
  -6.785267651711316D-03,  -1.186461381106319D-04,   2.552567912135712D-03, &
   2.746492518214684D-03,   2.795167529038011D-03,   2.640094985499672D-03, &
   2.251598432978858D-03,   2.142825449789546D-03,   1.743953669188522D-03, &
   9.220556260146275D-06,  -8.810753759719448D-06,  -9.220556260128274D-06, &
   1.536759376726381D-06,   1.014261188614073D-05,   8.221662665296847D-06, &
   2.456450757508053D-05,  -5.652364665968928D-06,  -1.170922743613323D-05, &
  -4.312474168027370D-06,  -7.679856474795462D-05,  -1.644605330580811D-05, &
   1.316293512263782D-05,   1.075716408266183D-05,  -4.290877697498020D-06, &
   1.250908906079554D-04,   6.528706040496491D-05,  -1.062841585507057D-06, &
  -2.068105770121475D-05,  -2.818887065908783D-06,   1.578433331546378D-05, &
  -5.650909745860123D-05,  -1.173211211719084D-04,  -3.636984791611088D-05, &
   2.655406950615447D-05,   1.920845888467901D-05,  -1.499393004935854D-05, &
  -1.925978925182464D-05,  -2.712189341592369D-04,   7.676638015131917D-05, &
   9.015466256072146D-05,  -1.628452607435528D-05,  -4.748071780233997D-05, &
   4.944310429770864D-06,   3.385644757581749D-05,  -3.647814371034485D-06, &
   6.467629155189640D-04,   1.747577302069817D-04,  -1.049469015529341D-04, &
  -2.312806328735908D-05,   8.377873691047919D-05,   3.243384810679374D-05, &
  -5.347463318385915D-05,  -2.239480111340871D-05,   4.830434255641058D-05, &
  -5.002067333797151D-05,  -4.760903931630449D-04,   1.058184542092296D-05, &
   1.046221071693343D-04,  -9.599159266162110D-05,  -1.155935413060195D-04, &
   5.170914607492250D-05,   7.719805196903430D-05,  -4.696455422220697D-05, &
  -3.732773343499435D-05,  -1.702726574794893D-03,   1.427166219619418D-04, &
   1.575357303291574D-04,  -2.299548428189368D-04,  -9.449096517895310D-06, &
   2.215271397867829D-04,   1.529192431523139D-05,  -1.574457647176503D-04, &
   2.549502367932634D-05,   9.000075899963973D-05,  -8.868352826178491D-05, &
  -4.213591561297840D-05,   8.189524519070851D-04,  -3.032920961131465D-04, &
   2.461821730523395D-04,   3.250418765761157D-04,  -2.239841720017485D-04, &
  -1.823910004365627D-04,   2.382523157516150D-04,   6.180603171835182D-05, &
  -1.965837535603561D-04,   8.125675322713017D-05,   4.680982023847060D-05, &
   2.413764310514994D-03,  -4.583545271902050D-04,   3.461177564460609D-04, &
   3.252235230300811D-04,  -5.969675941567590D-04,  -5.014048429749435D-05, &
   4.336523772397976D-04,  -2.356656711521380D-04,  -2.882536384478960D-04, &
   3.344738236584817D-04,  -3.664954577834870D-05,  -1.117486014974432D-04, &
   1.260470600920845D-04,   1.700106390675791D-03,  -1.030406168719017D-03, &
   7.543029789497031D-04,  -1.091228429556304D-03,   1.341775631073218D-04, &
   5.789920453074556D-04,  -6.484246337983534D-04,  -6.724860117385964D-05, &
   6.917693170473958D-04,  -4.238551626450193D-04,  -1.018557091196591D-04, &
   2.516900533158922D-04,  -1.536123896922168D-04,   6.191564317000944D-05, &
  -3.358410299366372D-04,   9.233165567848943D-05,  -7.919868173450268D-04, &
  -1.400156775074172D-04,   9.504934686976768D-04,  -1.058784519498545D-03, &
   3.408972982809803D-04,   9.333305453176801D-04,  -1.126483274558094D-03, &
   3.105919035000337D-04,   4.232213292653220D-04,  -4.636694114413330D-04, &
   1.760057014038237D-04,  -2.369128377122001D-05,  -2.542645128577182D-05, &
  -1.052207323993738D-03,   7.069290153206859D-04,  -1.469110675394522D-03, &
   1.763687110778521D-03,  -1.274009419965869D-03,   3.973951817192651D-04, &
   1.323081343007383D-03,  -2.095768053705992D-03,   1.176519399477716D-03, &
   2.757906361284238D-04,  -1.010072641211756D-03,   6.805521188327464D-04, &
  -1.412000112566798D-04,  -6.106758125021509D-05,   6.399062732003159D-05, &
  -6.258299457069353D-05,  -5.511669175338172D-03,   2.154322236638027D-03, &
  -1.348142982650836D-03,   1.370886009257905D-03,  -1.602424778155615D-03, &
   2.796661265734211D-03,  -3.480196872199560D-03,   2.296692763606372D-03, &
  -7.240045750980291D-05,  -1.686768608038608D-03,   1.751467450035328D-03, &
  -7.513649469643529D-04,  -2.315693979349946D-05,   2.161119494951279D-04, &
  -1.270578446748547D-04,   5.706275578948790D-05,  -2.479003707541539D-05, &
  -4.375180263200058D-03,  -7.306778010650692D-04,   2.264265007241312D-03, &
  -3.003375841617150D-03,   3.838143083603960D-03,  -3.879311522411955D-03, &
   2.200770771948888D-03,   5.730909959208972D-04,  -3.162803787065178D-03, &
   3.704805376797532D-03,  -2.177392847669075D-03,   4.337760615674344D-04, &
   4.321168995811320D-04,  -4.399219770023847D-04,   1.888286043416057D-04, &
  -5.369998975186937D-05,   2.172364002005732D-05,  -4.882660616947671D-06, &
   1.054901935414486D-02,  -5.335297442613297D-03,   5.281371280814997D-03, &
  -4.025125064558048D-03,   3.455347930809192D-03,  -4.058922837480583D-03, &
   5.530273304107125D-03,  -7.215363071888555D-03,   7.201927184204227D-03, &
  -4.643466707286407D-03,   1.384861246288000D-03,   7.023287969285580D-04, &
  -1.140096224338246D-03,   6.749126510488932D-04,  -2.308086927107570D-04, &
   3.863179085343456D-05,  -4.532775455091117D-06,   5.669337551676962D-06, &
  -1.660323057798937D-05,   6.407621853397221D-03,   2.976310857129599D-03, &
  -2.669486351791023D-03,   5.450044271771400D-03,  -7.676962820105453D-03, &
   9.300681308796308D-03,  -1.025343022114836D-02,   9.333839248384176D-03, &
  -5.805447009086020D-03,   1.224920685113235D-03,   2.106524753669267D-03, &
  -2.955509921155468D-03,   2.007721675227284D-03,  -8.668481991151166D-04, &
   2.146959447957576D-04,   2.727929596686715D-06,  -1.134733391093444D-05, &
  -4.776857171450607D-06,   5.340797738312944D-06,   1.972138598204497D-06, &
  -8.559490324614054D-03,   4.494901201427747D-03,  -4.814385165702922D-03, &
   4.151244818022622D-03,  -3.226507285941499D-03,   1.963507127428543D-03, &
  -1.799917090191172D-03,   3.487258920712218D-03,  -6.278109389345597D-03, &
   8.396984277343334D-03,  -8.085500129180944D-03,   5.493398179029520D-03, &
  -2.641671072324951D-03,   8.436757540301659D-04,  -1.076264580997800D-04, &
  -1.976576926866615D-05,  -4.059025176667268D-07,   1.891535984050937D-05, &
  -2.003718464303283D-05,   1.910598438380383D-05,  -1.483766951530606D-05 /)
lem(:, 23) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.311430997323940D-05,   0.000000000000000D+00, &
   1.009607194824246D-04,   0.000000000000000D+00,  -1.934187247008613D-04, &
   0.000000000000000D+00,   4.325772843784573D-04,   0.000000000000000D+00, &
  -1.136683098086236D-03,   0.000000000000000D+00,   3.645133841848287D-03, &
   0.000000000000000D+00,  -1.546094251754158D-02,   0.000000000000000D+00, &
   1.033319358844060D-01,   0.000000000000000D+00,  -1.970192362278671D+00, &
  -2.000625293121219D-02,   3.858816579339788D+00,  -1.814573364684018D-02, &
  -1.972486703396304D+00,   0.000000000000000D+00,  -2.704898998778175D-05, &
   2.267006423276554D-04,  -8.309485944511030D-04,   2.002251953027937D-03, &
  -3.362954801762823D-03,   3.344007205000563D-03,   4.593136298371216D-04, &
  -7.464055337637471D-03,   7.741015457038217D-03,   1.012520797748796D-02, &
  -1.681813664434351D-02,  -3.628408620433343D-02,  -2.593437800455373D-02, &
  -5.132496600951482D-03,   5.861339006417272D-03,   9.920049688718484D-03, &
   1.049521525625603D-02,   9.072572225843935D-03,   8.884712106858815D-03, &
   8.489488519303512D-03,   7.304715677828425D-03,   0.000000000000000D+00, &
   9.016329996486229D-06,  -8.001199140984403D-05,   2.967407690595914D-04, &
  -7.150642651889848D-04,   1.187304841018221D-03,  -1.124646728993856D-03, &
  -3.462536577836929D-04,   2.909125701542961D-03,  -2.704374469823146D-03, &
  -4.369741229300788D-03,   6.340254154166489D-03,   1.487976143445225D-02, &
   1.160375735590308D-02,   3.871201784645290D-03,  -6.230145875093863D-04, &
  -2.316097528184288D-03,  -2.588265375157844D-03,  -2.261908182880610D-03, &
  -2.233185360700296D-03,  -2.107874182889752D-03,  -1.785353227944057D-03, &
  -7.128034731079539D-07,   6.415231257651353D-06,  -7.128034730868725D-07, &
  -1.742530336471992D-05,  -4.901894653243857D-06,   3.531118743515833D-06, &
   2.856696995910957D-05,   1.626288539323927D-05,   1.381742117019877D-06, &
  -7.128034730672842D-06,  -2.208003657696992D-05,  -2.934087087275944D-05, &
  -1.236469997535114D-05,   4.504942051339840D-06,   9.298605661624330D-06, &
  -3.437062427241056D-05,   2.824641928486772D-05,   2.794587290078571D-05, &
   3.119705208927043D-06,  -1.139483836052885D-05,  -5.642851754630012D-06, &
   1.554949143208730D-04,   1.717515113993941D-05,  -3.711796046497907D-05, &
  -1.822172827648507D-05,   1.280561189542264D-05,   1.385052109484500D-05, &
  -7.895142018644320D-06,  -2.404914562692894D-04,  -1.228982849643100D-04, &
   1.624356032866305D-05,   3.913802129652999D-05,  -7.217107776693434D-06, &
  -2.974822411705105D-05,  -2.297525869035488D-06,   2.573420881949157D-05, &
  -1.425185168915729D-06,   2.080449795969750D-04,   5.236153086635578D-05, &
  -5.676828099377270D-05,  -1.687912995965691D-05,   4.749299769430606D-05, &
   2.819479711025473D-05,  -2.661095181915013D-05,  -2.201216504623490D-05, &
   6.828383116815706D-04,  -4.626347606537214D-05,  -1.327218977993729D-04, &
   5.484476332068984D-05,   7.520064658568084D-05,  -4.549176277864095D-05, &
  -7.385800752899445D-05,   1.733699376282897D-05,   4.467917050390488D-05, &
  -2.790696277696196D-05,  -6.877104409150884D-04,  -4.294797585342414D-04, &
   1.267152871314766D-04,  -5.845474376818794D-07,  -1.645485601847485D-04, &
  -1.996851868353787D-05,   1.256998922469744D-04,   2.154601534568428D-05, &
  -8.634729062661154D-05,   9.321111194192974D-06,   6.135929546471565D-05, &
  -1.407079538941147D-03,   4.867390265356838D-04,  -5.002096865904870D-06, &
  -1.788580361372466D-04,   1.880782810963430D-04,   1.793659818347621D-04, &
  -1.363501838889237D-04,  -1.113463572915681D-04,   1.329059235653188D-04, &
   4.687594010697491D-05,  -1.020037313580914D-04,   5.335102677483578D-05, &
   8.096880866684057D-04,   4.943574073918115D-04,  -2.079039252495125D-04, &
   4.382452576005384D-04,   7.450560902733381D-05,  -3.458607711786224D-04, &
   3.642625809616299D-05,   2.645950600432238D-04,  -1.388347845955964D-04, &
  -1.734336712086792D-04,   1.679921203589823D-04,  -2.348841350073897D-05, &
  -7.076788683819891D-05,   2.193400465601318D-03,  -7.344635601438445D-04, &
   5.911639520397358D-04,  -1.226760083783123D-04,  -5.839791596251299D-04, &
   2.881582420402634D-04,   2.319525722835897D-04,  -4.448277344005555D-04, &
  -2.917936266950720D-06,   3.902765864457964D-04,  -2.223619684831852D-04, &
  -5.504396562105501D-05,   1.256071978597542D-04,  -1.108536259863636D-04, &
   8.998961077369817D-04,  -6.551230504876865D-04,   3.222480061673951D-04, &
  -9.971988230527060D-04,   5.809307087716277D-04,   1.431537007594930D-04, &
  -6.824860541909559D-04,   4.300330429641356D-04,   4.283436573343751D-04, &
  -6.567418328001136D-04,   1.947901015513949D-04,   2.233897761347818D-04, &
  -2.287252710999787D-04,   1.066989791206361D-04,  -1.950969753814870D-05, &
  -6.108999919379277D-04,   3.377683226678028D-04,  -1.059406903400075D-03, &
   5.387287708398877D-04,   3.360652616394729D-04,  -9.245685249378209D-04, &
   9.323211162653016D-04,   2.448078730748949D-04,  -1.126986187085010D-03, &
   8.305659734699799D-04,   4.956156187507279D-05,  -5.257232449735384D-04, &
   3.500759245641213D-04,  -8.442565545170811D-05,  -1.025570895737010D-05, &
   3.644233190382202D-05,  -8.499523443907315D-04,   5.366039376297406D-04, &
  -1.001890973928294D-03,   1.417159833331168D-03,  -1.359103283397701D-03, &
   1.149713928101647D-03,   2.288973056667493D-04,  -1.709740303740492D-03, &
   1.767360040155996D-03,  -5.732347633808498D-04,  -7.080690444965380D-04, &
   9.392006425669407D-04,  -4.288620481385368D-04,   1.140787772711300D-05, &
   8.239978609166422D-05,  -5.837136771134544D-05,   3.543175363533393D-05, &
  -4.882226992860420D-03,   1.789782466937172D-03,  -8.330203500415842D-04, &
   5.818942423018463D-04,  -5.108148287710710D-04,   1.517325994226151D-03, &
  -2.888545133346831D-03,   2.970083015258819D-03,  -1.465590656764934D-03, &
  -7.459318193934739D-04,   1.828346396323810D-03,  -1.316022768776292D-03, &
   3.601134892962729D-04,   1.712597394822222D-04,  -1.975312190991627D-04, &
   8.954946098923487D-05,  -3.835654431233665D-05,   3.011149921514943D-05, &
  -3.754824554489107D-03,  -8.037309384465645D-04,   2.096571600306272D-03, &
  -2.697344346033448D-03,   3.478053208496349D-03,  -4.048301049093238D-03, &
   3.451240938055428D-03,  -1.496779890000800D-03,  -1.431761816898866D-03, &
   3.395660789193124D-03,  -3.005684841038439D-03,   1.340255860116522D-03, &
   5.414659909056549D-05,  -5.018214902940955D-04,   3.316275620395610D-04, &
  -1.195477946817101D-04,   3.346857581045707D-05,  -1.381769313259437D-05, &
   2.318470794505790D-07,   9.884598123959538D-03,  -4.774344544202585D-03, &
   4.485062141333799D-03,  -3.043937424423276D-03,   2.039049604346664D-03, &
  -1.998408959599460D-03,   3.038554424784628D-03,  -5.151785906717265D-03, &
   6.754593454705534D-03,  -5.920439882057077D-03,   3.138578484505892D-03, &
  -3.422715023539647D-04,  -1.004129281050174D-03,   9.377822094127944D-04, &
  -4.557410072378389D-04,   1.221830668625199D-04,  -8.290762553844317D-06, &
  -2.615881909903460D-06,  -2.543433763853505D-06,  -2.837073521006182D-06, &
   5.668977329566651D-03,   2.987561756475944D-03,  -2.679962233128472D-03, &
   5.119118099148623D-03,  -6.998454141848748D-03,   8.474523327029118D-03, &
  -9.813571821961273D-03,   1.017532113239370D-02,  -8.258282347732709D-03, &
   4.275722873409716D-03,  -4.823059732867782D-06,  -2.421240523310541D-03, &
   2.437930981184262D-03,  -1.398056521599469D-03,   5.000269956941321D-04, &
  -6.585507048529777D-05,  -2.394934499070037D-05,   1.436820127552847D-05, &
   2.585887514580404D-07,  -3.835638011862515D-06,   1.524180589357915D-05 /)
lem(:, 24) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.526802731912923D-05,   0.000000000000000D+00,  -5.453134373300951D-05, &
   0.000000000000000D+00,   9.325988411759929D-05,   0.000000000000000D+00, &
  -1.872024821219416D-04,   0.000000000000000D+00,   4.274150686410025D-04, &
   0.000000000000000D+00,  -1.132366216735540D-03,   0.000000000000000D+00, &
   3.642719732464851D-03,   0.000000000000000D+00,  -1.546838938601112D-02, &
   0.000000000000000D+00,   1.034117927024861D-01,   0.000000000000000D+00, &
  -1.971423167883841D+00,  -1.814573364684018D-02,   3.851964547237532D+00, &
  -1.653319610837274D-02,   5.526802731980438D-05,  -2.854028658510547D-04, &
   6.956327631404784D-04,  -1.190198870581767D-03,   1.427486785941138D-03, &
  -7.779325666573073D-04,  -1.261945272397863D-03,   3.793363710787051D-03, &
  -3.166204991931598D-03,  -4.341644462214635D-03,   1.138624792650713D-02, &
   3.746779756632130D-03,  -2.426851604156428D-02,  -3.185526284975346D-02, &
  -1.812040179615651D-02,  -1.580971234949695D-03,   7.183679978111252D-03, &
   8.498279341487586D-03,   8.792069966295748D-03,   8.489488519303427D-03, &
   7.312886480402513D-03,   7.184168697105012D-03,   0.000000000000000D+00, &
  -7.135071646201642D-05,   2.034181473762045D-04,  -3.602758295359098D-04, &
   4.172631580302345D-04,  -1.482090616992491D-04,  -6.144098491481891D-04, &
   1.499100994631358D-03,  -1.113198376663252D-03,  -1.863137555046363D-03, &
   4.397752317607543D-03,   1.814619254869061D-03,  -9.404499525380045D-03, &
  -1.339796158958497D-02,  -8.514180082936140D-03,  -2.140316744364620D-03, &
   1.275408040639676D-03,   2.017228405380016D-03,   2.195980699706450D-03, &
   2.113284129077728D-03,   1.825058537446202D-03,   1.802780200096149D-03, &
  -5.640769417673069D-06,   5.436578669540217D-06,   5.640769417677649D-06, &
  -1.812192889871129D-06,  -6.100198601017410D-06,  -5.143054469065706D-06, &
  -1.128518509870386D-05,   3.851183360664664D-06,   6.890343871094666D-06, &
   3.186834176510262D-06,   3.866862293117645D-05,   7.252417822734766D-06, &
  -7.543025012499099D-06,  -6.508580097329326D-06,   1.234715930234094D-06, &
  -7.051923824666478D-05,  -3.273194521773676D-05,   1.226350560629491D-06, &
   1.134313263796687D-05,   2.855662306847031D-06,  -7.764304114143226D-06, &
   6.068902442120392D-05,   6.519300111601608D-05,   1.853174085188403D-05, &
  -1.340674941055798D-05,  -1.137724623651838D-05,   5.774524174587076D-06, &
   1.217337628692406D-05,   6.938205284926151D-05,  -6.480444968275841D-05, &
  -4.950032251078008D-05,   6.309281844149448D-06,   2.414893513709071D-05, &
   1.539025452605386D-06,  -1.689079612315534D-05,  -5.485304054941130D-06, &
  -3.132362648171754D-04,  -3.615452434454551D-05,   6.659210041536874D-05, &
   1.515517922608320D-05,  -3.823890728816861D-05,  -2.078415773085962D-05, &
   2.116092287279602D-05,   1.857756325070859D-05,  -1.736766941168921D-05, &
   3.315725940011644D-04,   2.335918598380348D-04,  -2.587708889353168D-05, &
  -5.090549540347070D-05,   4.207023261260506D-05,   5.815522585999791D-05, &
  -1.242086220338126D-05,  -4.260156047118520D-05,   6.526947059810062D-06, &
   3.474992530114886D-05,   4.118793567437304D-04,  -2.710654059289056D-04, &
  -7.892450307533664D-05,   9.571345419567090D-05,  -4.481818014133064D-06, &
  -1.044595255281281D-04,  -2.798041528898461D-05,   6.877604070794489D-05, &
   2.142888891068259D-05,  -4.830887180413257D-05,   3.798568207687645D-06, &
  -1.127314149084266D-03,  -1.981174103570703D-04,   1.595108467732910D-04, &
  -1.228857230578551D-04,  -1.192171153543644D-04,   1.086445240650203D-04, &
   1.094589768138221D-04,  -7.674452036497761D-05,  -7.744317596779292D-05, &
   6.667412185804857D-05,   2.825265455490739D-05,  -7.364908038458144D-05, &
  -8.115910111135584D-04,   6.393573304690777D-04,  -1.509496513893849D-04, &
   1.858995264913516D-06,   2.903760764912432D-04,   1.577828438696427D-05, &
  -2.036634866108903D-04,   3.276464388077639D-05,   1.710648462513204D-04, &
  -6.367465468208970D-05,  -1.013160016621955D-04,   9.059117673796147D-05, &
  -1.150829433951322D-05,   1.345052953375396D-03,   9.787798433155947D-05, &
   1.384650279584463D-05,   4.212054569233328D-04,  -2.249756128604771D-04, &
  -2.706170335256700D-04,   2.361595592316608D-04,   1.132654568254134D-04, &
  -2.849512423884104D-04,  -1.378451132868308D-05,   2.220481129310305D-04, &
  -1.094376480767497D-04,  -2.572605143332105D-05,   8.330852620506435D-05, &
   1.734427694204389D-03,  -7.796589812818813D-04,   6.287652392639870D-04, &
  -5.169367424742722D-04,  -3.143365463721048D-04,   4.512972691060259D-04, &
  -1.000847912006628D-04,  -4.076930620372315D-04,   3.145665554844923D-04, &
   2.314894472404713D-04,  -3.777308159146003D-04,   9.530359705919983D-05, &
   1.155191770853362D-04,  -1.198729915594019D-04,   7.520826298270766D-05, &
   2.985702643312487D-04,  -2.958068828728758D-04,  -1.231563735552900D-04, &
  -6.199977382441839D-04,   7.243528465020342D-04,  -3.110825882214870D-04, &
  -3.766026684686751D-04,   7.188552766121525D-04,  -4.364011461193080D-05, &
  -6.159310436201861D-04,   5.145411364494075D-04,   2.075450095625350D-05, &
  -2.763512656733791D-04,   1.771556995070461D-04,  -6.144648536525737D-05, &
   5.737903656990473D-06,  -6.887685745110614D-04,   4.231521445470512D-04, &
  -1.068658229690862D-03,   9.268688426562250D-04,  -2.646957332728238D-04, &
  -4.311009603458609D-04,   1.086233339753633D-03,  -4.879904669751816D-04, &
  -6.962402497408928D-04,   1.096139797748001D-03,  -4.799360274376915D-04, &
  -3.391183764305329D-04,   5.030442269089693D-04,  -2.270733120823215D-04, &
   2.295291429175393D-05,   2.610938111519604D-05,  -4.454167326763433D-05, &
  -6.483045315013837D-04,   3.636419664194281D-04,  -5.579131158713940D-04, &
   9.085690976479074D-04,  -1.036458093152822D-03,   1.368159847474710D-03, &
  -7.181578016770314D-04,  -8.385328818196057D-04,   1.790811770683746D-03, &
  -1.357195706148762D-03,  -3.968317547479909D-05,   9.163115523765572D-04, &
  -7.469141173625074D-04,   2.205172998197043D-04,   6.661730099037937D-05, &
  -8.423998048586385D-05,   4.877129030844694D-05,  -2.715090529832726D-05, &
  -4.371711136746921D-03,   1.525317901391646D-03,  -5.320636500511058D-04, &
   1.151227775243558D-04,   2.263190606460913D-04,   3.320112931366319D-04, &
  -1.780245255024669D-03,   2.811411541088615D-03,  -2.468888368961101D-03, &
   5.506606938262868D-04,   1.332077077523483D-03,  -1.666370274790044D-03, &
   8.877318070761592D-04,  -6.256543356944548D-05,  -2.288700952127732D-04, &
   1.580182165278479D-04,  -6.078270470332086D-05,   2.634030441180136D-05, &
  -8.740993400193724D-06,  -3.260157764906037D-03,  -8.293657864431577D-04, &
   1.887454729561612D-03,  -2.300939341037783D-03,   2.863944293230582D-03, &
  -3.582372333494362D-03,   3.785771414007467D-03,  -2.964272921639788D-03, &
   5.760381118028628D-04,   2.211431514163259D-03,  -3.232642497754558D-03, &
   2.264997231391467D-03,  -6.389255475214026D-04,  -3.484770419208628D-04, &
   4.558299195403553D-04,  -2.372789074531203D-04,   6.887721223875745D-05, &
  -1.571914095110402D-05,   7.535175041801204D-06,  -1.316127886133005D-05, &
   9.274211909771998D-03,  -4.296321117973460D-03,   3.851103132716341D-03, &
  -2.329915810931448D-03,   1.070312177940606D-03,  -5.009123050499085D-04, &
   8.781567298416907D-04,  -2.694366264536625D-03,   5.164056748820684D-03, &
  -6.123981884057040D-03,   4.649379229753042D-03,  -1.821432612483580D-03, &
  -3.945486209434046D-04,   1.027244136437427D-03,  -7.209501313162549D-04, &
   2.856437091995172D-04,  -4.833507602462124D-05,  -3.490679392752584D-06, &
  -4.084251633307572D-07,   1.667962755647525D-06,   4.964322455036438D-06 /)
lem(:, 25) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.588529443723085D-05,   0.000000000000000D+00, &
  -5.355644953271371D-05,   0.000000000000000D+00,   9.366838538858191D-05, &
   0.000000000000000D+00,  -1.872721525990861D-04,   0.000000000000000D+00, &
   4.269043338259104D-04,   0.000000000000000D+00,  -1.131461703896454D-03, &
   0.000000000000000D+00,   3.642617823018628D-03,   0.000000000000000D+00, &
  -1.547608741217421D-02,   0.000000000000000D+00,   1.034811625612696D-01, &
   0.000000000000000D+00,  -1.972486703396304D+00,  -1.653319610837274D-02, &
   3.846006258452960D+00,   0.000000000000000D+00,   1.537941190161395D-05, &
  -1.288965894157890D-04,   4.756358903207333D-04,  -1.173163348825407D-03, &
   2.098871332575807D-03,  -2.536349266485396D-03,   1.113644725225045D-03, &
   2.900361714812816D-03,  -6.378132150720041D-03,   1.489628234891485D-03, &
   1.138813127667275D-02,  -4.383571927570614D-03,  -2.696058649741484D-02, &
  -2.717979613926168D-02,  -1.117324950917787D-02,   1.036726152421652D-03, &
   6.694810799231604D-03,   8.173831096372210D-03,   7.304715677828397D-03, &
   7.184168697105037D-03,   6.890873985159666D-03,   0.000000000000000D+00, &
  -5.126470634711370D-06,   4.549291391131821D-05,  -1.699241458243655D-04, &
   4.196618260632488D-04,  -7.451497363762595D-04,   8.759469592883326D-04, &
  -3.067607546220178D-04,  -1.197319693530119D-03,   2.386557987168748D-03, &
  -3.015075503667417D-04,  -4.625470151873496D-03,   1.401291025101188D-03, &
   1.079670523963362D-02,   1.155807239709055D-02,   5.758841310606455D-03, &
   9.190922786248370D-04,  -1.371015152685889D-03,  -1.979687205479768D-03, &
  -1.815820772660579D-03,  -1.804339538531680D-03,  -1.716414733312949D-03, &
   4.052830890307906D-07,  -3.647547801102933D-06,   4.052830890181451D-07, &
   9.948167252041622D-06,   2.796112740113688D-06,  -2.011089613944019D-06, &
  -1.689587734926826D-05,  -9.260207722769799D-06,  -8.633551517828353D-07, &
   4.126054305304164D-06,   1.613072104115638D-05,   1.710158406000729D-05, &
   7.153757382009001D-06,  -2.325325913753121D-06,  -5.715257846559668D-06, &
   8.577697377928389D-06,  -1.880581647798860D-05,  -1.612583948018013D-05, &
  -2.418416147098828D-06,   6.192742628731983D-06,   4.620823219226911D-06, &
  -7.047157712382262D-05,  -8.204428247195867D-07,   2.238998694619239D-05, &
   1.114349038456737D-05,  -5.671137136650086D-06,  -8.363747575078231D-06, &
   1.653643257671352D-06,   1.407715049394029D-04,   5.494960158689149D-05, &
  -1.371092076632383D-05,  -2.242531058536785D-05,   1.260231956954843D-06, &
   1.506077772506806D-05,   4.318711955036264D-06,  -1.235398216020399D-05, &
  -9.794598417675299D-05,  -1.205216545980552D-04,  -2.225252037980808D-05, &
   3.039790526554522D-05,   1.127685968386053D-05,  -2.145903043064983D-05, &
  -1.789590049393338D-05,   9.332005851136358D-06,   1.769936339087742D-05, &
  -2.049775417368154D-04,   9.800143764140272D-05,   7.514538310501191D-05, &
  -2.605641400340330D-05,  -3.686681350061320D-05,   1.824842211034826D-05, &
   3.921517966244518D-05,   1.674397998597302D-06,  -2.509657422929422D-05, &
  -8.324387465173039D-07,   5.524149652181118D-04,   1.247420479081129D-04, &
  -9.174528698703637D-05,   3.439749702049197D-07,   7.521490087477471D-05, &
   1.257580779574452D-05,  -5.984651876634195D-05,  -2.805294038567844D-05, &
   3.490592025589681D-05,   1.647282557052139D-05,  -3.361108350456338D-05, &
  -5.259427896097610D-05,  -3.813461842779638D-04,   1.980378518237938D-05, &
   6.173886488461718D-05,  -9.826643511777228D-05,  -8.670753245351324D-05, &
   5.550157316806485D-05,   7.339266537717411D-05,  -3.606086827044169D-05, &
  -5.108445152280593D-05,   3.448739643368224D-05,   2.323473532531897D-05, &
  -1.242110816516658D-03,   1.037936802427951D-04,   9.370246299664507D-05, &
  -1.745381432296512D-04,   2.026845504567263D-05,   1.754099857327060D-04, &
   6.736190130891750D-06,  -1.292383637783930D-04,   8.509296208883682D-06, &
   1.071210230389203D-04,  -2.311043608415488D-05,  -5.658179716922749D-05, &
   6.212115342869062D-05,  -1.287674436683960D-04,   5.789130622936963D-04, &
  -1.953735812987547D-04,   2.108683204169298D-04,   2.280471085826246D-04, &
  -1.650231808578599D-04,  -1.351410995322684D-04,   1.677106917869277D-04, &
   8.023357120906199D-05,  -1.721381515174787D-04,  -2.669556219149184D-05, &
   1.213487069091737D-04,  -6.062493975194834D-05,  -1.638564363632263D-05, &
   1.540825114605735D-03,  -2.373059633772706D-04,   2.511212919137760D-04, &
   2.054952810330797D-04,  -4.108693392105832D-04,  -4.849004673451829D-05, &
   2.877956830500294D-04,  -1.222032305652915D-04,  -2.592053229407202D-04, &
   1.945541193051368D-04,   1.472340754424367D-04,  -2.096414158630211D-04, &
   4.344576837636987D-05,   5.734183296756083D-05,  -8.770175376478585D-05, &
   1.198945783519054D-03,  -6.573604289697544D-04,   4.758960232719497D-04, &
  -7.083751256056590D-04,   5.565601189369409D-05,   3.639181052624070D-04, &
  -3.685745125927411D-04,  -1.386460997967909D-04,   4.748536897178367D-04, &
  -7.601519249939850D-05,  -3.618142678337893D-04,   2.957965610373468D-04, &
   2.309743504706487D-05,  -1.417421388024587D-04,   1.011835925196092D-04, &
  -4.389106062616842D-05,  -9.683900729870828D-05,  -2.009216456808017D-05, &
  -4.634966690833982D-04,  -1.378719901413425D-04,   5.614186227576518D-04, &
  -5.615835115823264D-04,   9.882016632899706D-05,   6.556702715769701D-04, &
  -4.919838793683121D-04,  -2.877328041039578D-04,   6.624040138796491D-04, &
  -3.058337598691269D-04,  -1.882722782977093D-04,   2.655618759446344D-04, &
  -1.197928565071845D-04,   2.748040760006317D-05,   1.349911202305192D-05, &
  -6.514087843691127D-04,   4.042610263600931D-04,  -9.028849544162442D-04, &
   1.009457945747327D-03,  -6.474622125746496D-04,   1.626697594356613D-04, &
   7.868970402041475D-04,  -9.449534754372494D-04,  -1.531090504102720D-05, &
   9.504062640178180D-04,  -9.308784576905958D-04,   8.972270665341530D-05, &
   4.885389527096583D-04,  -4.107246350380445D-04,   1.205012434934586D-04, &
   1.727869150494759D-05,  -3.444461665484385D-05,   2.898981096417160D-05, &
  -4.737708224391704D-04,   2.189948874320647D-04,  -2.060679857998451D-04, &
   4.100660125615509D-04,  -5.215838768457250D-04,   1.108499464827591D-03, &
  -1.215061064646121D-03,   1.460433223021520D-04,   1.230274044278815D-03, &
  -1.744683212311534D-03,   7.876023124133118D-04,   5.184464791974446D-04, &
  -9.159365557115590D-04,   5.273602747128979D-04,  -5.602610068409430D-05, &
  -1.004996308204829D-04,   7.060840675663854D-05,  -3.623788492408603D-05, &
   2.882541558223410D-05,  -3.955370019932633D-03,   1.334390071371089D-03, &
  -3.809658194447766D-04,  -9.305401343971010D-05,   5.840312264381938D-04, &
  -4.953018261627752D-04,  -5.857904901755370D-04,   1.980422262107422D-03, &
  -2.746583724539179D-03,   1.771113812544137D-03,   3.209876192735984D-04, &
  -1.571239084087195D-03,   1.373869770692903D-03,  -4.869282056619387D-04, &
  -1.273601954732730D-04,   2.217952327940194D-04,  -1.171325795375154D-04, &
   4.241829515410917D-05,  -1.822883189842097D-05,   4.465230838690002D-06, &
  -2.858970401667879D-03,  -8.295764116363314D-04,   1.681305046729956D-03, &
  -1.913430949028056D-03,   2.204069154752255D-03,  -2.794404960415972D-03, &
   3.376674557314919D-03,  -3.553576573633010D-03,   2.271381211664083D-03, &
   4.581641724068651D-04,  -2.639126206668560D-03,   2.856156184319791D-03, &
  -1.518205136304230D-03,   9.822529127108641D-05,   4.541317233588111D-04, &
  -3.757089529857508D-04,   1.568654129392146D-04,  -3.538019237558386D-05, &
   5.742430087183135D-06,  -4.406080309384300D-06,  -1.423199452195497D-06 /)
lem(:, 26) = (/ 0.000000000000000D+00,   1.632993161855452D+00,  -1.632993161855453D+00, &
   0.000000000000000D+00,  -2.065591117977287D+00,  -1.018350154434630D+00, &
   0.000000000000000D+00,   6.382847385042226D-02,   0.000000000000000D+00, &
  -1.197893559374872D-02,   0.000000000000000D+00,   3.495293720032787D-03, &
   0.000000000000000D+00,  -1.311020180385226D-03,   0.000000000000000D+00, &
   5.794986263736739D-04,   0.000000000000000D+00,  -2.877608462751273D-04, &
   0.000000000000000D+00,   1.558791572557201D-04,   0.000000000000000D+00, &
  -9.034263192756957D-05,   0.000000000000000D+00,   5.526802731980438D-05, &
   0.000000000000000D+00,   1.600000000000002D+01,  -2.065591117977286D+00, &
  -1.018350154434632D+00,   0.000000000000000D+00,   6.382847385042138D-02, &
   0.000000000000000D+00,  -1.197893559374964D-02,   0.000000000000000D+00, &
   3.495293720032897D-03,   0.000000000000000D+00,  -1.311020180384975D-03, &
   0.000000000000000D+00,   5.794986263733818D-04,   0.000000000000000D+00, &
  -2.877608462760895D-04,   0.000000000000000D+00,   1.558791572553680D-04, &
   0.000000000000000D+00,  -9.034263192760775D-05,   0.000000000000000D+00, &
   5.526802731991526D-05,   0.000000000000000D+00,  -8.000000000000023D+00, &
   4.131182235954576D+00,  -1.018350154434631D+00,   0.000000000000000D+00, &
   6.382847385042273D-02,   0.000000000000000D+00,  -1.197893559374916D-02, &
   0.000000000000000D+00,   3.495293720032978D-03,   0.000000000000000D+00, &
  -1.311020180384994D-03,   0.000000000000000D+00,   5.794986263737957D-04, &
   0.000000000000000D+00,  -2.877608462754166D-04,   0.000000000000000D+00, &
   1.558791572555824D-04,   0.000000000000000D+00,  -9.034263192794420D-05, &
   0.000000000000000D+00,   5.526802732004671D-05,   0.000000000000000D+00, &
  -1.632993161855452D-01,   5.443310539518166D-02,   0.000000000000000D+00, &
   2.332847374079219D-02,   0.000000000000000D+00,   5.832118435198040D-02, &
  -2.332847374079220D-02,   0.000000000000000D+00,  -1.555231582719480D-02, &
   0.000000000000000D+00,   5.184105275731593D-03,   0.000000000000000D+00, &
  -1.166423687039608D-02,   0.000000000000000D+00,  -1.992640465359333D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.776157913597402D-03, &
   0.000000000000000D+00,   1.458029608799513D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.178205744484431D-04, &
   0.000000000000000D+00,   6.583224597306881D-03,   0.000000000000000D+00, &
   5.103103630798293D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.069234466906702D-04,   0.000000000000000D+00,  -1.192933316290502D-03, &
   0.000000000000000D+00,   1.701034543599433D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.156809000276386D-05, &
   0.000000000000000D+00,  -1.743517923809211D-03,   0.000000000000000D+00, &
  -2.258416491951689D-03,   0.000000000000000D+00,  -1.275775907699581D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.588205062655779D-04,   0.000000000000000D+00, &
  -1.323782127336625D-03,   0.000000000000000D+00,  -1.032770972899654D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.005215545901440D-04, &
   0.000000000000000D+00,   1.104001440384711D-03,   0.000000000000000D+00, &
   1.822537010999387D-04,   0.000000000000000D+00,   4.556342527498588D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.137207875242740D-05,   0.000000000000000D+00, &
   4.401278189732834D-04,   0.000000000000000D+00,   1.093522206599631D-03, &
   0.000000000000000D+00,   4.100708274748636D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.597810098235288D-06,   0.000000000000000D+00, &
  -3.547612149752866D-04,   0.000000000000000D+00,   1.033304097670960D-04, &
   0.000000000000000D+00,   1.822537010999347D-04,   0.000000000000000D+00, &
  -1.993399855780712D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.288713432709555D-05,   0.000000000000000D+00, &
  -6.663768088139968D-04,   0.000000000000000D+00,  -3.493195937748855D-04, &
   0.000000000000000D+00,  -1.866834785572345D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.015932469597371D-05,   0.000000000000000D+00,  -1.923786875041079D-05, &
   0.000000000000000D+00,  -4.556342527498552D-04,   0.000000000000000D+00, &
  -1.234009434530777D-04,   0.000000000000000D+00,   9.966999278903758D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.722512275872532D-06,   0.000000000000000D+00,   2.233645919212024D-04, &
   0.000000000000000D+00,   1.974415095249381D-04,   0.000000000000000D+00, &
   4.366494922185753D-05,   0.000000000000000D+00,   9.524021533174486D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.787209688043265D-07,   0.000000000000000D+00, &
  -3.257475653143651D-05,   0.000000000000000D+00,   3.424581332885247D-04, &
   0.000000000000000D+00,   2.575599178738762D-04,   0.000000000000000D+00, &
   7.309132804528349D-05,   0.000000000000000D+00,  -5.481849603397747D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.849477370137976D-05, &
   0.000000000000000D+00,  -1.146844109862630D-04,   0.000000000000000D+00, &
   1.378983969496694D-04,   0.000000000000000D+00,   1.182232814900208D-05, &
   0.000000000000000D+00,  -5.300631434689217D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.126461494436556D-05, &
   0.000000000000000D+00,  -1.171131519216341D-04,   0.000000000000000D+00, &
  -2.259186503218026D-04,   0.000000000000000D+00,  -8.960231674973573D-05, &
   0.000000000000000D+00,  -4.394540591152938D-05,   0.000000000000000D+00, &
   3.239274765644613D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.533296045057006D-06,   0.000000000000000D+00,   5.307657589577294D-05, &
   0.000000000000000D+00,  -1.373231011759932D-04,   0.000000000000000D+00, &
  -1.455785955282166D-04,   0.000000000000000D+00,  -1.827283201131717D-05, &
   0.000000000000000D+00,   3.156216438318786D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.359821016274069D-08, &
   0.000000000000000D+00,   9.998316850625555D-06,   0.000000000000000D+00, &
   1.354852584631377D-04,   0.000000000000000D+00,   7.852787310649266D-06, &
   0.000000000000000D+00,   2.899490699317341D-05,   0.000000000000000D+00, &
   2.740924801699067D-05,   0.000000000000000D+00,  -2.024546728527375D-05 /)
lem(:, 27) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.065591117977289D+00,   0.000000000000000D+00,  -1.502496452850690D-01, &
  -6.708203932499371D-01,  -3.296088216486962D-01,   3.839170356332654D-02, &
   6.185895741317434D-02,  -6.402461428675592D-03,  -1.804961915705379D-02, &
   1.663466033434668D-03,   6.770079100229386D-03,  -5.599128549429831D-04, &
  -2.992518038793609D-03,   2.239879943971753D-04,   1.485990620420048D-03, &
  -1.014602182650944D-04,  -8.049565067620859D-04,   5.048974227136128D-05, &
   4.665273452126936D-04,  -2.704898998778175D-05,  -2.854028658510547D-04, &
   1.537941190161395D-05,  -2.065591117977286D+00,   8.888888888888888D+00, &
  -9.014978717104244D-01,  -1.565247584249853D+00,   0.000000000000000D+00, &
   8.958064164776015D-02,   0.000000000000000D+00,  -1.493907666691005D-02, &
   0.000000000000000D+00,   3.881420744680984D-03,   0.000000000000000D+00, &
  -1.306463328199849D-03,   0.000000000000000D+00,   5.226386535917964D-04, &
   0.000000000000000D+00,  -2.367405092860386D-04,   0.000000000000000D+00, &
   1.178093986330211D-04,   0.000000000000000D+00,  -6.311430997136905D-05, &
   0.000000000000000D+00,   3.588529443539948D-05,   2.065591117977287D+00, &
  -1.777777777777781D+00,   1.464934041529432D+00,  -8.944271909999149D-01, &
   2.472066162365225D-01,   5.118893808443528D-02,  -4.639421805988059D-02, &
  -8.536615238234121D-03,   1.353721436779015D-02,   2.217954711246647D-03, &
  -5.077559325172140D-03,  -7.465504732571145D-04,   2.244388529095140D-03, &
   2.986506591962456D-04,  -1.114492965314821D-03,  -1.352802910206002D-04, &
   6.037173800715616D-04,   6.731965636161447D-05,  -3.498955089091397D-04, &
  -3.606531998358684D-05,   2.140521493882725D-04,   2.050588253465728D-05, &
   7.027283689263045D-02,  -9.035079029052498D-02,  -4.015590679578885D-02, &
   4.015590679578898D-02,   2.007795339789445D-02,  -1.254872087368406D-02, &
   2.677060453052595D-02,   0.000000000000000D+00,   2.760718592210487D-02, &
   3.513641844631532D-02,  -3.011693009684172D-02,  -6.692651132631494D-03, &
  -1.254872087368406D-02,  -1.505846504842083D-02,  -8.156668567894645D-03, &
   7.605285377990358D-03,   4.258959811674587D-03,  -1.133187521320559D-02, &
  -3.194219858755933D-03,  -6.331400077176952D-03,  -1.619925785511939D-02, &
   0.000000000000000D+00,  -1.216845660478463D-03,   9.810818137607511D-03, &
   6.844756840191297D-03,   3.821655902440149D-03,   4.106854104114789D-03, &
   8.784104611578832D-03,   0.000000000000000D+00,   1.404052685167485D-04, &
  -4.855682202870720D-04,  -2.702801418947335D-03,   5.120404636220027D-03, &
   4.545620568229609D-03,   1.728739868612419D-03,   4.392052305789413D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.000387538181804D-03, &
  -1.053039513875503D-04,  -2.013938070287053D-03,  -3.264422493014311D-03, &
  -1.825268490717682D-03,   6.274360436842001D-04,  -4.156763789407842D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.228546099521410D-04, &
   3.861144884210469D-04,  -1.908634118899502D-03,  -4.914184398086120D-04, &
  -1.942272881148300D-03,  -2.470255192966476D-03,  -1.647019614671036D-03, &
  -9.411540655263084D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.053039513875600D-04,   6.888633486602777D-04, &
   1.158343465263145D-03,  -3.970836500239171D-04,  -8.921584770334524D-05, &
   9.019393127960440D-04,  -7.058655491447264D-04,   1.617608550123337D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.259133442158699D-06,   2.655827597495097D-04,  -2.642922701491687D-04, &
   1.279176308174770D-03,   1.472534666625370D-03,   1.148530367260580D-03, &
   4.705770327631498D-04,   1.147031517360188D-03,   2.941106454769766D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.047877555474266D-04,  -1.300813517140397D-04, &
   3.694671725766910D-04,  -3.873189453813655D-04,   4.404441005952659D-04, &
   8.627245600657784D-04,  -2.941106454769897D-05,   2.548958927467035D-04, &
  -7.311917436163646D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.557983665396424D-06, &
   7.721203040336320D-05,  -4.692152266322730D-04,  -4.539534890956431D-04, &
  -3.124218721996547D-04,  -4.928023201037057D-04,  -8.137061191529560D-04, &
   5.882212909539410D-05,  -6.004759011821511D-04,  -1.143763621299426D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.369277386463940D-05, &
  -3.418439030789369D-05,   2.679599774343347D-04,  -6.168786501445997D-04, &
  -5.466989390567577D-04,  -4.411659682154538D-04,  -3.725401509374967D-04, &
  -2.361054903967894D-04,  -1.045726739473668D-04,   3.717231769222880D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.150460633770808D-07, &
   7.139124310079266D-05,   1.558096391144832D-05,   5.913413891653049D-05, &
   4.418517530672130D-04,   6.189404619830405D-05,  -2.156811400164449D-04, &
   3.325084241920257D-04,  -5.555423303453658D-05,   3.259726320703109D-04, &
   5.146936295847397D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.456894838276322D-05,  -5.095396306175622D-05, &
   2.695348841505373D-04,   1.173360688980753D-04,   3.158590035195030D-04, &
   4.295797912724229D-04,   4.761473101396064D-04,   7.011122457834902D-05, &
   1.930286786855165D-04,   4.835000762765718D-05,  -2.064623309584046D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.933022977153103D-07,   1.620937623856396D-05,  -4.905894405110352D-05, &
  -1.989903956533578D-04,   2.401784328569526D-04,   1.185487267672249D-04, &
   6.914570983309323D-05,   2.002329040924013D-04,  -5.651974777979886D-05, &
   3.275323097356638D-05,  -1.877461989735162D-04,  -2.573468147924220D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.111441794604824D-06,  -3.911445924501524D-05, &
   3.727750488371599D-05,  -1.590905409677194D-04,  -3.266110133003913D-04, &
  -1.790658500618585D-04,  -2.346943534613818D-05,  -3.019090338040061D-04, &
  -9.580877087505975D-06,  -1.321826821433401D-04,  -2.456492323018073D-05, &
   1.227758762238443D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.013492061461253D-07, &
   1.652203204276936D-05,   2.736428565805581D-05,  -1.277989735717073D-04, &
   3.789668518997875D-05,  -1.956970692386510D-04,  -2.703441286707538D-04, &
  -2.115962699403795D-04,  -7.330485027418624D-05,  -1.910605140124735D-05, &
  -1.871613198489513D-05,   1.141976490641130D-04,   1.393961913460180D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.327864578553271D-06, &
  -1.593082834028032D-05,   7.054142284610859D-05,   1.221780515842468D-04, &
  -4.927759839893150D-05,   6.030753639577651D-05,   8.810179222805146D-05, &
  -7.930361052576030D-05,   1.305180261735287D-04,  -2.369510299372307D-06, &
   8.840297998234794D-05,   1.344472141382615D-05,  -7.707000963833351D-05 /)
lem(:, 28) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.018350154434631D+00,  -1.502496452850696D-01,   8.333333333333336D-02, &
  -4.724555912615332D-02,  -2.924988129130706D-01,  -3.217652068271440D-01, &
  -9.708172600685915D-02,   5.365975287870610D-02,   3.661130982435137D-02, &
  -1.394169684122703D-02,  -1.496598152472874D-02,   4.692692922018790D-03, &
   6.895807729312304D-03,  -1.877268697522699D-03,  -3.506505582541158D-03, &
   8.503495569320452D-04,   1.928186170855228D-03,  -4.231602366322541D-04, &
  -1.128921982094394D-03,   2.267006423276554D-04,   6.956327631404784D-04, &
  -1.288965894157890D-04,  -1.018350154434632D+00,  -9.014978717104244D-01, &
   6.361111111111100D+00,  -5.039526306789686D-01,  -1.730617976402336D+00, &
   0.000000000000000D+00,   9.352375699090074D-02,   0.000000000000000D+00, &
  -1.468266071080842D-02,   0.000000000000000D+00,   3.629742233596071D-03, &
   0.000000000000000D+00,  -1.177532936345904D-03,   0.000000000000000D+00, &
   4.595420421556731D-04,   0.000000000000000D+00,  -2.052236144708565D-04, &
   0.000000000000000D+00,   1.015838116309831D-04,   0.000000000000000D+00, &
  -5.453134373329915D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.883113471764763D-01,  -6.666666666666662D-01,   6.771863474748642D-01, &
  -5.849976258261399D-01,   3.217652068271439D-01,  -4.879500364742648D-02, &
  -5.365975287870563D-02,   2.288206864021969D-02,   1.394169684122694D-02, &
  -9.870037994803207D-03,  -4.692692922018789D-03,   4.655489845779960D-03, &
   1.877268697522688D-03,  -2.397610654729162D-03,  -8.503495569318249D-04, &
   1.328750055563244D-03,   4.231602366326143D-04,  -7.820036819941662D-04, &
  -2.267006423276962D-04,   4.836194827762042D-04,   1.288965894160666D-04, &
   3.266526290040743D-02,   8.908708063747469D-03,   1.484784677291246D-02, &
  -6.087617176894111D-02,  -2.227177015936872D-02,  -2.783971269921085D-02, &
   4.107904273839118D-02,   1.781741612749493D-02,   8.537511894424655D-03, &
  -1.855980846614067D-03,   1.849231825353641D-02,  -1.574771627430112D-03, &
   1.872853399765098D-02,   1.467912124140213D-02,   2.093883846043681D-02, &
  -2.857085666908919D-02,  -8.323792887844871D-03,  -1.544401031758243D-02, &
  -1.440916039098552D-02,  -1.158019564599499D-02,  -1.172642443997064D-02, &
   8.098825512497715D-03,   6.679800529538715D-03,  -6.673743715587045D-03, &
  -2.725566278244380D-04,  -5.583517204289274D-03,  -2.920249583833304D-03, &
  -6.126359154706306D-03,   0.000000000000000D+00,  -2.232368570752589D-03, &
   8.961488871040469D-03,   8.241593269929562D-03,   7.146175203816080D-03, &
   7.002109557680319D-03,   5.001089648402530D-03,   1.090388747385759D-02, &
   0.000000000000000D+00,   2.907270696794094D-04,  -6.203908004765899D-04, &
  -4.711335995251080D-03,   2.502329421169161D-03,   1.362783139122212D-03, &
   2.486430284546069D-03,   3.488075891800858D-04,  -1.832781086031392D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.072056069442809D-03, &
   2.465988537459286D-04,  -2.128537474438507D-03,  -4.490694915583676D-03, &
  -3.231851030167524D-03,  -3.690871001789321D-03,  -1.113588507968436D-03, &
  -4.871949722361893D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368127386726629D-04,   6.243608129827086D-04,  -1.590906165435775D-03, &
   5.107573781920207D-04,  -1.521621812564439D-03,  -1.128813251441449D-03, &
  -9.352838897921364D-04,  -1.304986532775508D-03,   2.327225983449659D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.046083368541991D-04,   6.908330736622188D-04,   1.572735723580813D-03, &
   4.414221277244495D-04,   1.591695078177380D-03,   8.710638477459305D-04, &
   1.913980248070746D-03,   8.608077832863642D-04,   1.674732717061899D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.735875913609418D-05,   2.902489001569020D-04,  -6.050250840561179D-04, &
   1.024771532837703D-03,   1.155144386176542D-03,   1.383936740994508D-03, &
   1.240349515520685D-03,  -3.334965583760009D-05,   1.080238852130838D-03, &
  -1.313444778821282D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.174328092019649D-04, &
  -1.610306186062874D-04,   2.682390383124621D-04,  -9.922318959200824D-04, &
  -4.660273484109421D-05,  -5.573958585488899D-04,  -1.928480098434931D-04, &
  -4.306455558159154D-04,  -8.344663884581195D-04,  -6.754513627976910D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.777696118326820D-06,   1.388700730887950D-04, &
  -4.508284723018412D-04,  -4.076997913996378D-04,  -5.121630057221814D-04, &
  -9.352533763483517D-04,  -7.084580720131437D-04,  -1.123738403223356D-03, &
  -2.223310389172848D-05,  -5.268278965649260D-04,   7.337528444703047D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.742442849626683D-05, &
  -4.603788035001572D-05,   4.686212004163472D-04,  -3.967794944737838D-04, &
  -2.234827226018010D-05,  -3.831899304570326D-04,  -7.447650414325250D-05, &
   3.033544451946834D-04,  -1.016527388409495D-04,   5.557891507203408D-04, &
   3.096871449938897D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.729586508127006D-06,   8.118639760358128D-05,  -2.360121246120120D-05, &
   6.525289855157099D-05,   6.609413383283319D-04,   3.403048329880558D-04, &
   6.461141809150947D-04,   2.653472616643547D-04,   5.316611800196525D-04, &
   1.902720894568231D-04,   2.623978602834551D-04,  -4.288907443798040D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.673134726263536D-05,  -8.259976678042441D-05,   2.550836220744868D-04, &
  -3.532139947063227D-05,   2.919196058368838D-04,   3.185606193617934D-04, &
   3.409222393200429D-04,   4.505938348017861D-04,  -2.122635291758178D-04, &
   1.208595482059206D-04,  -3.538430340100843D-04,  -1.565736682790374D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.037751904877679D-06,   3.072850652033922D-05,  -4.193256373267728D-05, &
  -2.854096994562476D-04,   1.053369080134907D-04,  -2.686800247277957D-04, &
  -1.594222992427380D-05,  -2.938416645769814D-04,  -2.612444630954042D-04, &
  -1.412911555372492D-04,  -1.866484999576720D-04,  -1.388401865126165D-04, &
   2.631836121178332D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.309553049408524D-06, &
  -4.633159886185679D-05,   9.232748978688943D-05,  -1.277366747810722D-04, &
  -3.249104375337184D-04,  -2.975299651300503D-04,  -4.037819635794530D-04, &
  -1.516987695777831D-04,  -3.499039502434711D-04,   5.136641702326857D-05, &
  -8.450852474049551D-05,   2.281613719238590D-04,   8.542625183002560D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.165186073113118D-07,   1.917100510212105D-05, &
   3.668975462992290D-05,  -1.228668191128710D-04,   1.735241144988202D-04, &
  -1.115199971124821D-04,   1.345200570145727D-05,  -1.045717209553378D-04, &
  -7.581368439972723D-05,   2.228045443746465D-04,   2.596252708891188D-05, &
   1.466084274777779D-04,   7.792362492885428D-05,  -1.687023605847193D-04 /)
lem(:, 29) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.708203932499371D-01,  -4.724555912615312D-02, &
   3.000000000000005D-01,   1.356801050599943D-02,  -2.038853399815945D-01, &
  -2.256358601250305D-01,  -1.360052202894712D-01,  -8.778338150186008D-03, &
   4.372881941057201D-02,   1.440509291004505D-02,  -1.585870726495718D-02, &
  -8.894083870854486D-03,   6.582035048402805D-03,   5.157477039380811D-03, &
  -3.045896485452536D-03,  -3.052498847696235D-03,   1.536613662744333D-03, &
   1.871868416990642D-03,  -8.309485944511030D-04,  -1.190198870581767D-03, &
   4.756358903207333D-04,   0.000000000000000D+00,  -1.565247584249853D+00, &
  -5.039526306789686D-01,   5.349999999999998D+00,  -3.216121008829488D-01, &
  -1.817798768046429D+00,   0.000000000000000D+00,   9.735110504930423D-02, &
   0.000000000000000D+00,  -1.513421660621093D-02,   0.000000000000000D+00, &
   3.714276175212247D-03,   0.000000000000000D+00,  -1.197871946031670D-03, &
   0.000000000000000D+00,   4.646187436716498D-04,   0.000000000000000D+00, &
  -2.059311987442569D-04,   0.000000000000000D+00,   1.009607194818219D-04, &
   0.000000000000000D+00,  -5.355644953253603D-05,   0.000000000000000D+00, &
  -2.236067977499790D-01,   1.574851970871777D-01,  -1.899999999999999D-01, &
   3.507581975254647D-01,  -3.970398725957370D-01,   2.887447405431426D-01, &
  -1.145307118227126D-01,  -1.392426051408793D-02,   3.867943101292372D-02, &
  -5.251117612528752D-03,  -1.423110309829060D-02,   4.702619058152940D-03, &
   5.945952523750553D-03,  -3.033543490327724D-03,  -2.761837810447361D-03, &
   1.887102272010541D-03,   1.396576105348153D-03,  -1.190537000844730D-03, &
  -7.564144596076198D-04,   7.707975216817022D-04,   4.334582898395945D-04, &
  -1.767766952966370D-02,   3.788072042070793D-02,   1.094331923264897D-02, &
  -1.565736444055928D-02,  -1.683587574253635D-04,   9.091372900969908D-03, &
  -4.568644462861133D-02,  -1.532064692570851D-02,  -2.231518802965344D-02, &
  -1.821335648510802D-02,   4.655374731837835D-02,   1.819805114352393D-02, &
   1.584102853956876D-02,   7.688383255758478D-03,   2.515815522890462D-03, &
   7.964664293584738D-03,  -4.491529087956510D-03,   1.328150449311596D-02, &
   1.018629349121881D-02,   1.113537183618050D-02,   1.312212290719665D-02, &
  -2.484763339484200D-02,  -7.635010782542061D-03,  -1.842095970813373D-02, &
  -1.589754057003319D-02,  -1.353006912711200D-02,  -9.436086475686715D-03, &
  -1.101763843937954D-02,   7.903442927248251D-03,   7.630301446670042D-03, &
  -1.835463656126919D-03,   3.172915043785794D-03,  -2.880789052973415D-03, &
  -2.456948824489974D-03,  -7.144945518358102D-04,  -3.657093638129739D-04, &
   0.000000000000000D+00,  -2.859744208295250D-03,   7.385415981324034D-03, &
   7.646784122222154D-03,   8.887399791012234D-03,   8.368637011358019D-03, &
   7.036410043182195D-03,   3.986379232307399D-03,   7.704386106399192D-03, &
   0.000000000000000D+00,   4.052799047524740D-04,  -5.696218756987457D-04, &
  -5.715817904807874D-03,   4.440972982271805D-05,  -8.520088906728051D-04, &
   5.213814820462956D-04,  -1.515817483812351D-04,  -1.155974500306397D-03, &
  -5.336841080104940D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.089310940166040D-03,   6.219324398460000D-04,  -1.751197708740116D-03, &
  -4.614196899182206D-03,  -4.107474379286641D-03,  -3.509602259600580D-03, &
  -3.380499961143926D-03,  -9.853810086289450D-04,  -1.604669406710542D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.426032100013628D-04, &
   7.361909396928655D-04,  -1.331033459534500D-03,   1.332114201087170D-03, &
  -5.428186195055966D-04,  -3.505427599197387D-04,  -2.177423559155903D-04, &
   9.516375456286074D-04,  -2.630605584771407D-05,   3.893296265461637D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.750514589040358D-04,   6.208272464397576D-04,   1.614832303271015D-03, &
   7.751609674087477D-04,   2.138337375643387D-03,   1.767143122204594D-03, &
   1.318651846774301D-03,   1.530354798940751D-03,   1.141463606658714D-03, &
  -6.044638395257517D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.457335346026460D-05,   3.154448578722524D-04, &
  -8.410215034287848D-04,   6.904193479388955D-04,   7.077137187170384D-04, &
   9.922301379769152D-04,   6.005969683717948D-04,   5.007018916768694D-04, &
  -1.016729058514134D-03,  -6.349076187411710D-05,  -1.867318933065056D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.257409991569043D-04,  -1.510008270154327D-04, &
   2.687683956614397D-04,  -1.291711689626937D-03,  -4.881433114741329D-04, &
  -8.780762419446027D-04,  -8.931618300885421D-04,  -8.693553592763800D-04, &
  -2.842473960703965D-04,  -1.080663236859932D-03,   6.991656405775262D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.056306207248413D-05,   1.785115724661459D-04, &
  -4.207956656630267D-04,  -2.700623695272117D-04,  -4.029766673004134D-04, &
  -8.751144653712474D-04,  -7.976732953530748D-04,  -2.999385670612490D-04, &
  -4.997328546820411D-04,   3.290274775020681D-04,   3.687207250283441D-04, &
   9.241123113207421D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.780446074842740D-05,  -6.583552619141610D-05,   5.746463750309392D-04, &
  -2.492739430694188D-04,   2.992332568778570D-04,   2.257628562859365D-05, &
   2.566977299766828D-04,   2.958708862773566D-04,   8.706183488895245D-04, &
   5.238791235585845D-06,   6.608026424662723D-04,  -5.226903999762745D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.473566147020009D-06, &
   9.011659316875621D-05,  -6.510651692553703D-05,   1.196451939109041D-05, &
   6.982845583213226D-04,   4.178283871749903D-04,   6.214198789673259D-04, &
   6.300712800446811D-04,   2.591370700339417D-04,   9.322955872165721D-05, &
   9.756407928747419D-05,  -3.568249260401545D-04,  -4.873463082504681D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.843006461961445D-05,  -9.970865978801069D-05,   2.537566267505183D-04, &
  -1.655899917281771D-04,   1.443094869873784D-04,   1.536004613528974D-04, &
   1.660308540453010D-04,  -1.562008296473288D-04,   5.485314697469567D-05, &
  -5.209961499988107D-04,  -9.477362388508852D-05,  -3.859953919335410D-04, &
   3.636092914731845D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.146810421398908D-06,   4.109904591761991D-05, &
  -3.016166202906082D-05,  -3.123744182020612D-04,   7.326130005940267D-05, &
  -4.165159684368038D-04,  -2.384179672821507D-04,  -3.221326874396678D-04, &
  -3.676208069006799D-04,  -4.419994168751148D-04,   7.659479409827846D-05, &
  -1.517723537492810D-04,   2.790379950687579D-04,   2.729767084353517D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.034800661291832D-06,  -5.452823290562584D-05, &
   1.330999730520224D-04,  -8.066314063569689D-05,  -2.653084119526055D-04, &
  -2.260868276474246D-04,  -2.648180341573883D-04,  -2.807809293579378D-04, &
   9.253396589876593D-05,   9.734526935514074D-06,   1.843092829103625D-04, &
   1.253064121545741D-04,   2.294222141308455D-04,  -2.513300870401632D-04 /)
lem(:, 30) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.382847385042217D-02,  -3.296088216486964D-01,  -2.924988129130704D-01, &
   1.356801050599954D-02,   2.250000000000005D-01,   1.065047115522916D-01, &
  -1.197868911105678D-01,  -1.891363336314458D-01,  -1.317680026869874D-01, &
  -4.256639051532647D-02,   2.014606046523218D-02,   2.678314039687555D-02, &
   1.349395749113733D-04,  -1.331403249278541D-02,  -3.005564015726252D-03, &
   6.734834171467607D-03,   2.707506803805866D-03,  -3.579641041666869D-03, &
  -2.010716279484798D-03,   2.002251953027937D-03,   1.427486785941138D-03, &
  -1.173163348825407D-03,   6.382847385042138D-02,   0.000000000000000D+00, &
  -1.730617976402336D+00,  -3.216121008829488D-01,   4.830555555555562D+00, &
  -2.229978693352237D-01,  -1.864979948301230D+00,   0.000000000000000D+00, &
   9.893057979515127D-02,   0.000000000000000D+00,  -1.514092550541324D-02, &
   0.000000000000000D+00,   3.650903219680429D-03,   0.000000000000000D+00, &
  -1.156686757567310D-03,   0.000000000000000D+00,   4.411940769718729D-04, &
   0.000000000000000D+00,  -1.926480304400815D-04,   0.000000000000000D+00, &
   9.325988411718000D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.240220541217386D-02,  -1.462494064565349D-01,   7.487531723681089D-02, &
  -4.166666666666680D-02,   1.541240482332515D-01,  -2.618745332257625D-01, &
   2.461146433446843D-01,  -1.392976028405301D-01,   2.659050092208481D-02, &
   2.877560349940456D-02,  -2.124657390656525D-02,  -4.933856640030330D-03, &
   1.106595713738041D-02,  -1.286472476429799D-04,  -5.707523475916671D-03, &
   1.024894754006948D-03,   3.065503440751614D-03,  -9.865632099434359D-04, &
  -1.725731632954708D-03,   7.795760120675948D-04,   1.015496635596305D-03, &
  -6.514466333088112D-03,  -3.908679799852837D-03,  -6.514466333088112D-03, &
   3.460281467835111D-02,   1.101537034503987D-02,   7.385881959462211D-03, &
  -2.923331515532090D-02,  -6.751356017927659D-03,  -8.206535510513584D-04, &
   4.737793696791350D-03,  -3.181740849397376D-02,  -9.545439480137939D-03, &
  -1.819429922818616D-02,  -1.331137805963877D-02,  -1.129537525411189D-02, &
   4.818275448691975D-02,   1.848043246472775D-02,   2.086169443403537D-02, &
   1.249267689609981D-02,   7.404050009924969D-03,   4.789043868992216D-03, &
  -5.193350783021367D-04,  -8.124101371100560D-03,   7.034452207227694D-03, &
   5.002016806804705D-03,   8.181317171811557D-03,   6.269604399239507D-03, &
   6.597036054967271D-03,  -2.091007025795407D-02,  -5.539574168556034D-03, &
  -1.899434424438431D-02,  -1.680192153391634D-02,  -1.606310555942693D-02, &
  -1.158538614918509D-02,  -8.151624611726919D-03,  -9.896265890398151D-03, &
   7.447554435589678D-03,   7.596548400262497D-03,   1.878464761698895D-03, &
   7.217230132286928D-03,   8.987681926019608D-04,   6.044505603690604D-04, &
  -2.412729862154770D-04,   1.246040747161459D-03,   3.084316174185774D-03, &
   0.000000000000000D+00,  -3.178716562858533D-03,   5.737071564174537D-03, &
   5.754891885488836D-03,   9.122147830673619D-03,   9.368899943608981D-03, &
   8.247672825828259D-03,   5.959800457969715D-03,   2.916607642080050D-03, &
   4.555037006338945D-03,   0.000000000000000D+00,   4.836239160007587D-04, &
  -4.231462446151855D-04,  -5.985152721086035D-03,  -1.947756384759443D-03, &
  -3.531422469314597D-03,  -1.361046055504861D-03,  -1.662161175542875D-03, &
  -1.815311828931804D-03,  -1.790091758719964D-03,  -5.781952401103031D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.088851955762346D-03, &
   9.379045961531783D-04,  -1.146520593515071D-03,  -4.009202067432814D-03, &
  -4.486582054071864D-03,  -4.223369579119199D-03,  -3.173171958832345D-03, &
  -2.274983926961070D-03,  -3.030632496103673D-04,   1.348698108022143D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.451899320377247D-04, &
   7.695257808539789D-04,  -1.165877068368319D-03,   1.981831994364002D-03, &
   4.875553877588601D-04,   8.156709373967142D-04,   3.561398524780270D-04, &
   9.345735332967051D-04,   1.782788368344066D-03,   4.478695603998053D-04, &
   2.821026875946323D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.210408103810803D-04,   5.331463562182522D-04, &
   1.453516235731945D-03,   8.298755094046287D-04,   2.456647562485217D-03, &
   2.203855025062341D-03,   1.900373508266226D-03,   1.126738673580809D-03, &
   4.686899428564056D-04,   5.694787704707331D-04,  -2.390777528107007D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.986026331270669D-05,   3.410018563753122D-04,  -9.822991830333080D-04, &
   3.739056531334059D-04,   1.298645644435127D-04,   5.636317185307388D-04, &
   2.421124091742309D-04,   9.991910696929602D-06,  -2.463751078112082D-04, &
  -1.255124809777703D-03,  -6.254404333699610D-04,  -6.047068025069320D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.324351789833404D-04,  -1.257798848636871D-04, &
   3.242214775794095D-04,  -1.404625502311620D-03,  -8.019251664438532D-04, &
  -1.331231506104653D-03,  -1.038855202648401D-03,  -9.530745395246404D-04, &
  -7.655318638993821D-04,   3.857939756680900D-04,  -4.483033183676499D-04, &
   1.584879321514232D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.118319367556252D-05, &
   2.025975776494779D-04,  -3.970075803822090D-04,  -1.083158220376708D-04, &
  -2.068948886887445D-04,  -6.839264855136785D-04,  -7.171611115421565D-04, &
  -3.810730891905993D-04,   1.016415041515261D-04,   2.036156291728103D-04, &
   3.726944850391390D-04,   8.201640242019326D-04,  -5.246664086068805D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.523936656906596D-05, &
  -8.677583710206747D-05,   6.179489264001244D-04,  -1.584186705073502D-04, &
   5.802147166133485D-04,   2.981398148094986D-04,   5.515803285148310D-04, &
   4.725239694563767D-04,   3.510399526443641D-04,   7.080964831474928D-04, &
  -2.447785188547797D-04,   5.235171353587234D-05,  -9.543439930052871D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.042234014567793D-06, &
   9.860896154458220D-05,  -9.922108380218031D-05,  -5.923479021267262D-05, &
   6.393522331592195D-04,   4.232727791782651D-04,   6.813157346445975D-04, &
   5.296191011565884D-04,   3.366465556690807D-04,   1.444747692902409D-05, &
  -4.788854795006187D-04,   6.694053253776184D-06,  -6.268511662820483D-04, &
   2.039406232570521D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.996036342804227D-05,  -1.085764765456426D-04, &
   2.632841916257358D-04,  -2.659392436184973D-04,  -3.811784620573260D-06, &
  -5.717752747720849D-05,   4.634631907082609D-05,  -1.799131700087625D-04, &
  -3.622694644785966D-04,  -1.398090906360609D-04,  -3.228972563853192D-04, &
  -5.436413336288332D-05,   1.052009048265502D-04,   5.761278987921300D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.243530376984503D-06,   4.832397472198530D-05,  -1.922252313409374D-05, &
  -3.086018478635475D-04,   8.399927115327271D-05,  -5.027424501465328D-04, &
  -3.371899396320308D-04,  -4.575584678572794D-04,  -3.076600152374597D-04, &
  -1.347485810112114D-04,  -2.375737804721958D-04,   4.131252691253661D-04, &
   6.843505117693005D-06,   4.270361095941670D-04,  -2.087326408705785D-04 /)
lem(:, 31) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.839170356332628D-02,  -3.217652068271438D-01, &
  -2.038853399815948D-01,   1.065047115522923D-01,   1.641156462585035D-01, &
   9.130692996959998D-02,  -3.249187272711909D-02,  -1.363931569888010D-01, &
  -1.344109478801961D-01,  -6.283035863241884D-02,   1.247843284616896D-03, &
   2.344111371446133D-02,   1.126075805391726D-02,  -6.718575955263761D-03, &
  -8.574383115809273D-03,   1.243351963088037D-03,   5.442939695039919D-03, &
   3.945107522587057D-04,  -3.362954801762823D-03,  -7.779325666573073D-04, &
   2.098871332575807D-03,   0.000000000000000D+00,   8.958064164776015D-02, &
   0.000000000000000D+00,  -1.817798768046429D+00,  -2.229978693352237D-01, &
   4.533730158730153D+00,  -1.636833997804414D-01,  -1.894975177725836D+00, &
   0.000000000000000D+00,   1.002400936831017D-01,   0.000000000000000D+00, &
  -1.527613833269318D-02,   0.000000000000000D+00,   3.671892851487052D-03, &
   0.000000000000000D+00,  -1.161460122783058D-03,   0.000000000000000D+00, &
   4.428397574274393D-04,   0.000000000000000D+00,  -1.934187247010361D-04, &
   0.000000000000000D+00,   9.366838538669157D-05,   0.000000000000000D+00, &
   1.279723452110885D-02,   1.135641906448744D-01,  -1.030157507275424D-01, &
   1.324049849177903D-02,  -2.409297052154351D-03,   6.606251054196660D-02, &
  -1.528577463994499D-01,   1.923601513902552D-01,  -1.467698259176841D-01, &
   5.748179794992352D-02,   9.694699765066022D-03,  -2.522267346138606D-02, &
   7.053559979935639D-03,   8.890030595313321D-03,  -6.451878119401728D-03, &
  -2.882103181690363D-03,   4.317750360103513D-03,   7.420984415459507D-04, &
  -2.734901963399247D-03,  -2.788855202375729D-07,   1.731479327974452D-03, &
   1.011710220950939D-03,  -9.105391988558559D-03,  -1.011710220950947D-03, &
   7.081971546656650D-03,  -3.035130662852853D-03,  -3.540985773328331D-03, &
   2.685866598230849D-02,   1.097316470416029D-02,   8.380215081932870D-03, &
   4.064528125428814D-03,  -3.733422962208503D-02,  -1.116452381355095D-02, &
  -7.315948485867571D-03,   3.835604683825031D-04,   2.634514640046221D-03, &
  -1.792542981223350D-02,  -4.360494635287419D-03,  -1.273498241993218D-02, &
  -1.177515477441465D-02,  -8.641101895901854D-03,  -6.720267455425060D-03, &
   4.578165622219295D-02,   1.777685698258333D-02,   2.376053504928657D-02, &
   1.612773226345088D-02,   1.101864618262463D-02,   5.448112601545739D-03, &
   4.212879864558643D-03,  -6.318715049238282D-03,  -1.118000138857519D-02, &
   2.886835280130162D-04,  -2.203919485846618D-04,   3.825616594436781D-03, &
   4.646100236449033D-03,   2.649509779453315D-03,   1.960422648937501D-03, &
  -1.738676660258694D-02,  -3.054274500846976D-03,  -1.727814613048975D-02, &
  -1.625909347478017D-02,  -1.732635501133911D-02,  -1.325742737034641D-02, &
  -9.516451287118781D-03,  -5.897446570769506D-03,  -6.873874971409072D-03, &
   6.908735145403106D-03,   7.031614929324243D-03,   4.152526259457922D-03, &
   1.059452631115616D-02,   5.328259635831447D-03,   4.070825440558474D-03, &
   2.444698731405115D-03,   1.944801200471321D-03,   2.798529838263540D-03, &
   5.216217651034290D-03,   0.000000000000000D+00,  -3.288118453278790D-03, &
   4.327306871899804D-03,   3.390726358408671D-03,   7.750943546768636D-03, &
   9.163000101160879D-03,   8.835410345662636D-03,   6.296707758168882D-03, &
   4.217619535368524D-03,   1.425916551410405D-03,   1.266613772713980D-03, &
   0.000000000000000D+00,   5.329953652715065D-04,  -2.537246936373844D-04, &
  -5.751136863190179D-03,  -3.155349146640963D-03,  -5.913382884992163D-03, &
  -3.890226428365811D-03,  -3.398700219609797D-03,  -3.150677525622295D-03, &
  -2.990147212542552D-03,  -2.135788714895339D-03,  -5.123758990089470D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.079638303183001D-03, &
   1.156894234558655D-03,  -5.123522969114151D-04,  -2.906780428658174D-03, &
  -4.001777433634181D-03,  -4.310031237394138D-03,  -3.541960402630520D-03, &
  -1.898337709982864D-03,  -9.313616563581033D-04,   3.695113502301306D-04, &
   3.171639089475301D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.461560075141326D-04,   7.616816074385477D-04,  -1.094547566632553D-03, &
   2.361848989779810D-03,   1.392079214957475D-03,   2.094380767249086D-03, &
   1.507636965098317D-03,   1.660941391791234D-03,   1.737823315653700D-03, &
   2.024219028732910D-03,   5.182050887585670D-04,   1.047113492029144D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.494202112811342D-04,   4.573218065077530D-04,   1.207149538753359D-03, &
   6.258724432571569D-04,   2.386826415283786D-03,   2.398777672575386D-03, &
   2.106470217276055D-03,   1.525616432823431D-03,   4.854075984194618D-04, &
  -4.200040493196713D-04,   1.037996962605669D-04,  -2.705962575014158D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.359776627599899D-05,   3.627806005021087D-04,  -1.045253120827264D-03, &
   1.488583813451754D-04,  -4.375092660923264D-04,  -7.749044546154826D-05, &
  -3.359538302579545D-04,  -4.037443736877910D-04,  -7.917174396983792D-04, &
  -6.781915852637816D-04,  -9.367421103549116D-04,  -6.215856342410334D-04, &
   9.088461291057811D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.378634179311459D-04, &
  -1.002634706568026D-04,   4.026090159074714D-04,  -1.355377635040856D-03, &
  -9.101090641088577D-04,  -1.629137520444252D-03,  -1.392236719632854D-03, &
  -1.090394567498387D-03,  -7.461830739188163D-04,  -3.708367418388880D-04, &
   7.978247182715519D-04,   9.266283705636564D-05,   1.206987343473592D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.169706246617372D-05,   2.169827185614403D-04, &
  -3.862121167704790D-04,   2.901525444176552D-05,   3.074070018699106D-05, &
  -3.526426453661757D-04,  -4.442094959284938D-04,  -1.640252606037220D-04, &
   5.397962926730994D-05,   5.001060462611574D-04,   5.006099681836799D-04, &
   9.356619910420249D-06,   6.441396741261781D-04,  -1.016004861126286D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.049712245617162D-05, &
  -1.041006128514593D-04,   6.216587001902044D-04,  -1.388413149327808D-04, &
   7.601246855187180D-04,   5.652803031655240D-04,   8.389428376124643D-04, &
   7.291652091528435D-04,   5.915706465809054D-04,   2.200783402744124D-04, &
   2.657722729106009D-04,  -3.625698543195654D-04,  -4.441666296663007D-04, &
  -4.440660316289143D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.469984832130228D-06,   1.060444935852698D-04,  -1.223784230601279D-04, &
  -1.235959428876005D-04,   5.219963801956310D-04,   3.093098723660673D-04, &
   6.177915385981744D-04,   5.184525806242695D-04,   2.226630539081704D-04, &
   5.236780950382891D-05,  -1.739101487319170D-04,  -6.217081161685164D-04, &
   1.665348613763627D-04,  -3.298898346755991D-04,   7.779001945958542D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.132122244228464D-05,  -1.135122910187795D-04,   2.804327615848269D-04, &
  -3.233308880611198D-04,  -1.219864869647875D-04,  -2.794153366049608D-04, &
  -1.798589439918825D-04,  -3.549504776496712D-04,  -3.544671129332370D-04, &
  -4.175380839590031D-04,  -9.884643170227637D-05,   1.155154478450554D-04, &
   4.895437008462574D-06,   4.626635443805329D-04,   1.157005934204637D-04 /)
lem(:, 32) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.197893559374915D-02,   6.185895741317401D-02,  -9.708172600685953D-02, &
  -2.256358601250308D-01,  -1.197868911105672D-01,   9.130692996960049D-02, &
   1.501464843750001D-01,   7.543616217590250D-02,  -5.320966249091328D-03, &
  -7.739869242411175D-02,  -1.137934319330602D-01,  -8.206839292587240D-02, &
  -1.946936977962383D-02,   1.774015688655550D-02,   1.673623640560186D-02, &
   2.749592040630680D-04,  -8.138136684189931D-03,  -3.563147063056099D-03, &
   3.436453788969551D-03,   3.344007205000563D-03,  -1.261945272397863D-03, &
  -2.536349266485396D-03,  -1.197893559374964D-02,   0.000000000000000D+00, &
   9.352375699090074D-02,   0.000000000000000D+00,  -1.864979948301230D+00, &
  -1.636833997804414D-01,   4.346749441964286D+00,  -1.252448582170307D-01, &
  -1.914575246825786D+00,   0.000000000000000D+00,   1.009957973947417D-01, &
   0.000000000000000D+00,  -1.530121851232797D-02,   0.000000000000000D+00, &
   3.650497922238615D-03,   0.000000000000000D+00,  -1.144938432196222D-03, &
   0.000000000000000D+00,   4.326006019819445D-04,   0.000000000000000D+00, &
  -1.872024821236614D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.546473935329371D-02,  -2.541406439970151D-02,   9.423106802156456D-02, &
  -7.008807458597049D-02,  -5.446218031198898D-03,   1.834542410714296D-02, &
   2.747848834004410D-02,  -8.759531651373337D-02,   1.345126322944379D-01, &
  -1.341574245522475D-01,   7.985864002222666D-02,  -1.319374652009834D-02, &
  -2.076761942827072D-02,   1.573129838219129D-02,   2.195970898484389D-03, &
  -8.559055089266470D-03,   1.970113039318921D-03,   4.117245959013822D-03, &
  -2.354360557591487D-03,  -1.885319786832090D-03,   1.917258198015198D-03, &
   1.222594994431168D-03,  -6.792194413506586D-04,   1.222594994431171D-03, &
  -8.965696625828603D-03,  -2.445189988862350D-03,  -1.018829162025971D-04, &
   1.455470231465684D-02,   1.746564277758775D-04,  -1.508837473286090D-03, &
  -2.634401118952893D-03,   1.724045767075879D-02,   9.003152083533455D-03, &
   8.437554315963193D-03,   5.105782786572610D-03,   2.010465797466414D-03, &
  -4.020029135535232D-02,  -1.374126748319715D-02,  -1.223792619120470D-02, &
  -3.591754475398072D-03,   1.183205695160014D-04,   1.472315009319359D-03, &
  -5.931169916422906D-03,   7.499353774287460D-04,  -6.386167845546407D-03, &
  -8.586364322557516D-03,  -7.934056262353155D-03,  -5.544099003860319D-03, &
  -4.055654366366095D-03,   4.109646427107928D-02,   1.563841994385112D-02, &
   2.422189348260192D-02,   1.826369568554727D-02,   1.442192408415348D-02, &
   7.830176889229108D-03,   4.026165527652371D-03,   3.270640420092591D-03, &
  -9.889315365568547D-03,  -1.302527721720497D-02,  -5.744402951577405D-03, &
  -5.777445582608918D-03,  -1.294541560541360D-03,   1.703292366930126D-03, &
   1.898270408649683D-03,   7.336450487633441D-04,   4.886047406106941D-05, &
  -1.442573773872942D-02,  -7.735725492667845D-04,  -1.422632648981966D-02, &
  -1.389501324461848D-02,  -1.702353256082764D-02,  -1.419150264953548D-02, &
  -1.085389076775491D-02,  -6.656884899172408D-03,  -3.655329787688896D-03, &
  -3.759145452413473D-03,   6.365016153621223D-03,   6.239159631137228D-03, &
   5.207153908574362D-03,   1.265583433244607D-02,   9.483202938376097D-03, &
   8.079469218527284D-03,   5.504683030587927D-03,   3.702946475797167D-03, &
   3.119568540471905D-03,   3.082364377921164D-03,   4.946524748944262D-03, &
   0.000000000000000D+00,  -3.265390272241243D-03,   3.243180399633209D-03, &
   1.122400605201518D-03,   5.339950008546866D-03,   7.531675941409747D-03, &
   8.540359489123310D-03,   6.439575420592445D-03,   4.148735794363796D-03, &
   1.958585100831800D-03,  -2.217390098752940D-04,  -1.582767022413017D-03, &
   0.000000000000000D+00,   5.608409346349360D-04,  -1.003300197734918D-04, &
  -5.230247293222748D-03,  -3.587443565723264D-03,  -7.545427268857776D-03, &
  -6.410782878077352D-03,  -5.539228747633520D-03,  -4.482720728807329D-03, &
  -3.694560921564497D-03,  -3.131363610899787D-03,  -1.760315950948643D-03, &
  -3.180458300129146D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.064102730263731D-03,   1.281712637474122D-03,   2.343314717547449D-05, &
  -1.605429536690062D-03,  -2.830217925685065D-03,  -3.670073738017154D-03, &
  -3.415149117934274D-03,  -1.815587411874772D-03,  -3.595698826583743D-04, &
   6.101139495962204D-04,   1.128914891143332D-03,   4.083475241002952D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.460869714515824D-04, &
   7.347750451642251D-04,  -1.085891221231799D-03,   2.477139577296020D-03, &
   2.003633716140610D-03,   3.242696095889228D-03,   2.747439246274460D-03, &
   2.584865830888572D-03,   2.306866506168986D-03,   1.848962443607275D-03, &
   1.629656744640482D-03,   2.587865869560168D-04,  -8.644377409152472D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.656676467729644D-04,   4.017692561129163D-04,   9.545564225446407D-04, &
   2.765076330205020D-04,   1.976928881408169D-03,   2.232458433513774D-03, &
   2.066866847891658D-03,   1.458120891590637D-03,   4.541038825968673D-04, &
  -3.861681283352115D-04,  -1.182181920070989D-03,  -3.020430585765740D-04, &
  -2.125351891462590D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.616521676470734D-05,   3.787162689589178D-04, &
  -1.054455302485475D-03,   3.085435794899607D-05,  -8.834583397084706D-04, &
  -7.421668586544487D-04,  -1.041552544853065D-03,  -9.500043392147020D-04, &
  -1.168489395598971D-03,  -1.193688078883621D-03,  -7.266929142138649D-04, &
  -3.948995640717603D-04,  -4.127363104097862D-04,   1.898260126393852D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.420275905056439D-04,  -7.972498807430141D-05, &
   4.781084461437399D-04,  -1.204448085691913D-03,  -8.233870943540033D-04, &
  -1.701470559793999D-03,  -1.570101988577953D-03,  -1.207034098161391D-03, &
  -7.149458863640469D-04,  -1.133249644485313D-04,   1.292204858754039D-04, &
   9.245956597090777D-04,   4.067700332966349D-04,   2.418909806570909D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.211727239561184D-05,   2.254005426519320D-04, &
  -3.857820336056578D-04,   1.235099061142813D-04,   2.446320982037786D-04, &
   3.153204794305980D-05,  -5.484186756913788D-05,   1.777367889588176D-04, &
   3.360006623284863D-04,   5.907937486335745D-04,   7.061551817475348D-04, &
   5.202539488618471D-04,  -3.587149872992696D-04,   2.977902642769313D-04, &
  -1.275244370915803D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.419727226969298D-05,  -1.164811275173076D-04,   6.053830512488208D-04, &
  -1.725469355966599D-04,   8.234354953737072D-04,   7.335316360245921D-04, &
   1.063417228035267D-03,   9.040159473406146D-04,   6.968573330971104D-04, &
   4.149593368839133D-04,  -9.906956403533796D-05,  -1.252912188610630D-04, &
  -2.293435235886391D-04,  -5.910339245396117D-04,   3.662950385589943D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.790432550829609D-06, &
   1.120937808858833D-04,  -1.361104215237319D-04,  -1.694490741598136D-04, &
   3.912899194612588D-04,   1.271487839254423D-04,   4.402776785653287D-04, &
   3.860340905786014D-04,   9.350499686725777D-05,  -1.393009605509377D-04, &
  -3.167358341500561D-04,  -2.572884958742327D-04,  -4.844767119868429D-04, &
   2.966949253626565D-04,   7.547152856049230D-05,   7.014996723505441D-04 /)
lem(:, 33) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.402461428675744D-03,   5.365975287870562D-02, &
  -1.360052202894715D-01,  -1.891363336314457D-01,  -3.249187272711887D-02, &
   7.543616217590303D-02,   1.068070023148150D-01,   7.791649810852101D-02, &
   1.070907552189652D-02,  -4.721304949003205D-02,  -8.231393037393445D-02, &
  -8.171172847043985D-02,  -4.148329364639721D-02,   4.147102916058858D-03, &
   1.964011518216980D-02,   7.776525600194754D-03,  -5.695894721822204D-03, &
  -6.318541926995351D-03,   4.593136298371216D-04,   3.793363710787051D-03, &
   1.113644725225045D-03,   0.000000000000000D+00,  -1.493907666691005D-02, &
   0.000000000000000D+00,   9.735110504930423D-02,   0.000000000000000D+00, &
  -1.894975177725836D+00,  -1.252448582170307D-01,   4.222095630787041D+00, &
  -9.891820161327464D-02,  -1.928399304489886D+00,   0.000000000000000D+00, &
   1.016095849668119D-01,   0.000000000000000D+00,  -1.535822080007256D-02, &
   0.000000000000000D+00,   3.656433216322537D-03,   0.000000000000000D+00, &
  -1.145314641774179D-03,   0.000000000000000D+00,   4.325772843772778D-04, &
   0.000000000000000D+00,  -1.872721525972767D-04,   0.000000000000000D+00, &
  -2.134153809558525D-03,  -1.893873631013124D-02,  -4.724391862686926D-02, &
   8.071927209375543D-02,  -3.464263498801600D-02,  -9.219073425887285D-03, &
   1.936848958333333D-02,   4.149162928590765D-03,  -4.882686108511292D-02, &
   9.098828515998789D-02,  -1.088028856962125D-01,   8.685152543789873D-02, &
  -3.572620158776822D-02,  -8.337991843940649D-03,   1.955163245104049D-02, &
  -6.104151233631705D-03,  -6.663619155147972D-03,   5.883391714002136D-03, &
   1.395365473396059D-03,  -3.849723129055096D-03,   3.983309978403985D-04, &
  -1.687196728832557D-04,   1.518477055949300D-03,   1.687196728832527D-04, &
  -3.350290647253206D-03,   9.882152268876343D-04,   7.712899331805945D-04, &
  -5.712366067618791D-03,  -3.591318751372141D-03,  -1.156934899770890D-03, &
   2.892337249427263D-04,   1.833018731824507D-02,   3.277982216017528D-03, &
   6.809043941359881D-04,  -1.934250535554466D-03,  -1.965132261394704D-03, &
   7.314528755522660D-03,   5.963135435098357D-03,   6.594991452801752D-03, &
   6.209705896691677D-03,   2.784833546915354D-03,   8.861881964865174D-04, &
  -3.860283212445861D-02,  -1.465479130142956D-02,  -1.503341413084412D-02, &
  -7.492769088092472D-03,  -2.212091877783860D-03,   4.246921267373452D-04, &
   1.022171151680300D-03,   3.054131454160460D-03,   5.306688304468535D-03, &
   1.715530798596779D-04,  -4.225759654467630D-03,  -5.796920073778560D-03, &
  -5.631291722397832D-03,  -3.186758275384009D-03,  -2.175426268669650D-03, &
   3.563824384510653D-02,   1.252320667000728D-02,   2.223524937153350D-02, &
   1.862537982108068D-02,   1.641184423510102D-02,   1.055246166930368D-02, &
   5.348156919343749D-03,   2.796665946551904D-03,   2.443862265472237D-03, &
  -1.186287347063523D-02,  -1.367686969506163D-02,  -1.013497790785843D-02, &
  -1.092392435378869D-02,  -6.842922961441860D-03,  -2.241922654113247D-03, &
  -1.498964295696743D-04,   4.521795271670251D-04,  -3.628360400224153D-04, &
  -8.759973010754204D-04,  -1.200308765256691D-02,   1.064515919290618D-03, &
  -1.079810498557699D-02,  -1.020758847872104D-02,  -1.475995051915277D-02, &
  -1.385149063037660D-02,  -1.149931964312054D-02,  -7.605311353733197D-03, &
  -4.136316102324877D-03,  -2.004132303952988D-03,  -1.868375311383778D-03, &
   5.850171266278111D-03,   5.395627755885536D-03,   5.422083895255461D-03, &
   1.332548347566591D-02,   1.245995775308801D-02,   1.200266169669603D-02, &
   9.107633603424431D-03,   6.036794384607773D-03,   4.359885740269619D-03, &
   3.179137280652129D-03,   2.610725406628874D-03,   3.667306483795482D-03, &
   0.000000000000000D+00,  -3.164422582334416D-03,   2.462752265726844D-03, &
  -7.409877188077793D-04,   2.579666148836169D-03,   4.712070057766876D-03, &
   6.914004738860328D-03,   5.944538786578847D-03,   4.017090323726860D-03, &
   1.852686210216768D-03,   1.736126856951908D-04,  -1.194714589778788D-03, &
  -2.705405125805493D-03,   0.000000000000000D+00,   5.732523660981160D-04, &
   2.183382394512655D-05,  -4.592313132258523D-03,  -3.413118769172883D-03, &
  -8.217291793209954D-03,  -8.362264775795899D-03,  -7.793900410095864D-03, &
  -6.223338348897920D-03,  -4.588304136545906D-03,  -3.452023157606495D-03, &
  -2.315307182373339D-03,  -7.959059133950036D-04,  -8.148056272713910D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.043333575266160D-03, &
   1.332793111974213D-03,   4.012988421872722D-04,  -3.653583174127214D-04, &
  -1.232456524652341D-03,  -2.288027957854616D-03,  -2.671097435043329D-03, &
  -1.486578981975074D-03,  -1.444571239237791D-04,   9.754027829557918D-04, &
   1.559666360751450D-03,   1.390989900779934D-03,   3.520499092361606D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.452473077232889D-04, &
   7.004934566313366D-04,  -1.105709098059759D-03,   2.386538065102851D-03, &
   2.243587041137999D-03,   4.050573589067510D-03,   3.932817863906548D-03, &
   3.619831637067387D-03,   2.976449698440770D-03,   2.110964980240399D-03, &
   1.272508475143004D-03,   6.081892582960129D-04,  -3.856743889662912D-04, &
  -2.458504315438747D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.736552054007498D-04,   3.648275321616027D-04, &
   7.388121420556430D-04,  -1.050974216954426D-04,   1.333687681766746D-03, &
   1.679652225370963D-03,   1.711568457366444D-03,   1.253121394619028D-03, &
   2.458539116041454D-04,  -5.463023124651638D-04,  -1.052539254482983D-03, &
  -1.450506552839272D-03,  -4.055489364721107D-04,  -7.607752452988407D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.786588125467740D-05,   3.885696546885180D-04,  -1.032580091132885D-03, &
   4.740099396917618D-06,  -1.150796706788559D-03,  -1.294456838734306D-03, &
  -1.774679646361569D-03,  -1.610787885397293D-03,  -1.591323086850930D-03, &
  -1.396712853915748D-03,  -9.345594980078868D-04,  -2.755225349223104D-04, &
   3.729831808206110D-04,   1.371276641389873D-05,   2.230697891530802D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.449633874757549D-04,  -6.469331080920682D-05, &
   5.354933539247723D-04,  -1.011856446153869D-03,  -5.849543609272972D-04, &
  -1.531051683825309D-03,  -1.549072363381054D-03,  -1.217619156396309D-03, &
  -6.798283608130390D-04,   3.239216084633636D-05,   5.041612787836671D-04, &
   5.200312780028770D-04,   7.636545334158987D-04,   4.748666089071605D-04, &
  -8.685397125439277D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.244875317140710D-05, &
   2.300396740786217D-04,  -3.906563966711559D-04,   1.749002740370255D-04, &
   3.931750016724883D-04,   3.916947146540278D-04,   4.030979724166212D-04, &
   6.092682866594016D-04,   6.563827575986294D-04,   7.950134552282716D-04, &
   7.326486343104149D-04,   4.801752601209428D-04,   2.205972319495903D-04, &
  -6.842877770554030D-04,  -1.056007549734655D-04,  -9.261869738232373D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.677786588025031D-05, &
  -1.243390400671177D-04,   5.824692622257620D-04,  -2.347211557549112D-04, &
   7.819579892236659D-04,   7.685721527979014D-04,   1.186475375503933D-03, &
   1.037521925498522D-03,   7.394129847378009D-04,   3.736531528128086D-04, &
  -6.928620118097794D-05,  -4.461484500269577D-04,  -4.037868267733660D-04, &
   8.968065581531424D-06,  -4.932079182075896D-04,   9.690195975733211D-04 /)
lem(:, 34) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.495293720032415D-03,  -1.804961915705417D-02,   3.661130982435148D-02, &
  -8.778338150186325D-03,  -1.317680026869882D-01,  -1.363931569888010D-01, &
  -5.320966249090855D-03,   7.791649810852150D-02,   7.579933449074101D-02, &
   6.042426466235732D-02,   2.751502442272618D-02,  -2.361291587166581D-02, &
  -6.108720078285084D-02,  -7.037879355848013D-02,  -5.089698229479365D-02, &
  -1.375098411627560D-02,   1.395634363957769D-02,   1.471258571802722D-02, &
   2.774925049804435D-04,  -7.464055337637471D-03,  -3.166204991931598D-03, &
   2.900361714812816D-03,   3.495293720032897D-03,   0.000000000000000D+00, &
  -1.468266071080842D-02,   0.000000000000000D+00,   9.893057979515127D-02, &
   0.000000000000000D+00,  -1.914575246825786D+00,  -9.891820161327464D-02, &
   4.134544994212965D+00,  -8.010018789148188D-02,  -1.938357173611015D+00, &
   0.000000000000000D+00,   1.020242680797511D-01,   0.000000000000000D+00, &
  -1.537824321752530D-02,   0.000000000000000D+00,   3.647495840097358D-03, &
   0.000000000000000D+00,  -1.137408961705152D-03,   0.000000000000000D+00, &
   4.274150686431343D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.512404789263525D-03,   1.029693088809882D-02,  -9.081039465707726D-04, &
  -5.145226771587144D-02,   6.272331177551208D-02,  -1.687880277375759D-02, &
  -1.439971049546719D-02,   1.528284143518507D-02,  -3.722918278769644D-03, &
  -2.287018861117069D-02,   5.892806934942636D-02,  -8.387450232262460D-02, &
   8.135677770062526D-02,  -4.997440428411153D-02,   8.475705271704844D-03, &
   1.588288908591804D-02,  -1.323553653614521D-02,  -1.094539600296965D-03, &
   7.391723079015117D-03,  -2.481054238025787D-03,  -3.236584335730438D-03, &
  -3.567369214681134D-04,   2.751970537039691D-04,  -3.567369214681153D-04, &
   1.172135599109518D-03,   5.402016239374269D-04,  -1.579834937930232D-04, &
  -5.228744020375462D-03,   2.089459111456093D-04,   7.338588098772599D-04, &
   5.159944756949492D-04,  -1.146654390433216D-03,  -3.460348138240684D-03, &
  -1.908542529854398D-03,  -6.000824643267149D-04,   5.621791664207284D-04, &
   1.857580112501810D-02,   5.452978656726859D-03,   2.963464569052959D-03, &
  -6.765260903555943D-04,  -1.752151611603645D-03,  -1.304319369117782D-03, &
  -1.140564144891994D-03,   2.411545472899990D-03,   3.455772013493028D-03, &
   5.884621168420319D-03,   3.896136987197036D-03,   1.654544139583099D-03, &
   4.192272114854969D-04,  -3.423448508066541D-02,  -1.387385484819804D-02, &
  -1.564914895565768D-02,  -1.035548522711471D-02,  -5.063914813599288D-03, &
  -9.586012953953754D-04,   6.578048888661611D-04,   8.165396125654876D-04, &
   9.027621454813390D-03,   8.680443475928484D-03,   6.047513142157420D-03, &
   8.026565562034224D-04,  -2.359188047777920D-03,  -4.732066094069157D-03, &
  -3.615853080536957D-03,  -1.883640018906700D-03,  -1.233562544377032D-03, &
   3.031183050562962D-02,   9.128157839757883D-03,   1.848045868303771D-02, &
   1.695022896495581D-02,   1.671232551432142D-02,   1.258261263132705D-02, &
   7.280392139106100D-03,   3.484096743060776D-03,   1.704094165443368D-03, &
   1.466471786859342D-03,  -1.277415282408882D-02,  -1.343592337727050D-02, &
  -1.264496085079226D-02,  -1.475714288555853D-02,  -1.214134102047792D-02, &
  -6.857225739545364D-03,  -3.087380653058476D-03,  -6.718568957311249D-04, &
  -3.357278472798003D-04,  -7.777985725478834D-04,  -1.207316868892327D-03, &
  -1.004198690394194D-02,   2.425479045388221D-03,  -7.621654520771894D-03, &
  -6.035287439101616D-03,  -1.083127444877371D-02,  -1.180750177483024D-02, &
  -1.111323368637303D-02,  -8.143022933431113D-03,  -4.644196183760307D-03, &
  -2.229570312632969D-03,  -7.765614244308194D-04,  -5.115717303299585D-04, &
   5.377237075898134D-03,   4.592594481236297D-03,   5.143341189028127D-03, &
   1.288413220561329D-02,   1.385322640442444D-02,   1.499551822667578D-02, &
   1.276127120253351D-02,   8.860966380679092D-03,   5.992312658619473D-03, &
   3.921094062244020D-03,   2.676178693292086D-03,   1.926919316434956D-03, &
   2.512470217946068D-03,   0.000000000000000D+00,  -3.020737229890558D-03, &
   1.925779406480816D-03,  -2.100704079632801D-03,   2.356532596894663D-05, &
   1.296510843121564D-03,   4.005602720605609D-03,   4.390053057136100D-03, &
   3.375115924437105D-03,   1.579795285145762D-03,  -5.309510516821198D-05, &
  -9.189504146016513D-04,  -1.543347101896764D-03,  -2.763404731753358D-03, &
   0.000000000000000D+00,   5.748537353611025D-04,   1.106089339783204D-04, &
  -3.948323368488296D-03,  -2.859641900730974D-03,  -7.985095153480167D-03, &
  -9.346852132872487D-03,  -9.640105260380678D-03,  -8.078774791628426D-03, &
  -5.779162866816972D-03,  -3.999634626194541D-03,  -2.481860250061146D-03, &
  -1.263260221204697D-03,   6.396060005752094D-05,   7.157818332728563D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.018280022249109D-03, &
   1.333291387335588D-03,   6.179530526790790D-04,   6.408879412192731D-04, &
   4.575076043414760D-04,  -3.584133049028091D-04,  -1.216754490445345D-03, &
  -6.660404009593723D-04,   3.364123682867110D-04,   1.239085191032240D-03, &
   1.769953242902594D-03,   1.723036295641059D-03,   1.049841972058661D-03, &
   2.083462069179747D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.437939075172762D-04,   6.645189355024271D-04,  -1.129687259349314D-03, &
   2.170346655522562D-03,   2.139168721703805D-03,   4.374637514712730D-03, &
   4.801971259075809D-03,   4.603518018015633D-03,   3.722511135899834D-03, &
   2.500864909347944D-03,   1.404837616180509D-03,   3.450384149018061D-04, &
  -3.545635660012754D-04,  -8.584977532540375D-04,  -2.904850012849715D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.760202093339082D-04,   3.417316182967740D-04,   5.740356131534053D-04, &
  -4.359217383210260D-04,   6.086419486410871D-04,   8.271593749480388D-04, &
   9.794463641125907D-04,   7.551224680272641D-04,  -9.932583264857710D-05, &
  -8.187684978544601D-04,  -1.165879372673943D-03,  -1.171527899488478D-03, &
  -1.048830437749391D-03,  -4.275417781631871D-05,   8.462538956230953D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.892692545776392D-05,   3.930365952786124D-04,  -9.958548812282785D-04, &
   4.041441685866131D-05,  -1.238811510978407D-03,  -1.632705124482201D-03, &
  -2.382905477360938D-03,  -2.285047741226914D-03,  -2.092158978043318D-03, &
  -1.680618832363972D-03,  -1.011752049748130D-03,  -3.403770416372405D-04, &
   3.568413244316280D-04,   9.391274607024967D-04,   3.071247717774379D-04, &
   1.616791555433290D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.467748546762982D-04, &
  -5.403720389799393D-05,   5.701731743360679D-04,  -8.243558247520448D-04, &
  -2.686849039827865D-04,  -1.153574116190473D-03,  -1.279244016110389D-03, &
  -1.043098715603483D-03,  -5.398016344593489D-04,   1.628616543413163D-04, &
   6.223662091216797D-04,   7.498215959022847D-04,   4.675234724186021D-04, &
   2.196649722061325D-04,   1.934576055500843D-04,  -1.723732600825318D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.269735923019370D-05,   2.321674097480965D-04, &
  -3.967381614680074D-04,   1.933359680362719D-04,   4.636829201086244D-04, &
   6.607116687916146D-04,   8.389189396076952D-04,   1.090327315258987D-03, &
   1.050527034138630D-03,   1.046475468820290D-03,   8.605584230842587D-04, &
   4.298356772943412D-04,   1.187146753410262D-05,  -2.158650070463749D-04, &
  -8.117420077919028D-04,  -3.629412530875335D-04,  -5.743456406660963D-05 /)
lem(:, 35) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.663466033434609D-03,  -1.394169684122681D-02, &
   4.372881941057175D-02,  -4.256639051532686D-02,  -1.344109478801964D-01, &
  -7.739869242411189D-02,   1.070907552189693D-02,   6.042426466235785D-02, &
   6.552088713842973D-02,   4.612975272066249D-02,   2.684988850251236D-02, &
  -2.930738193723785D-03,  -4.019119343515042D-02,  -5.990790294561187D-02, &
  -5.215959155528961D-02,  -2.566329663596905D-02,   2.639241778986717D-03, &
   1.502098766956141D-02,   7.741015457038217D-03,  -4.341644462214635D-03, &
  -6.378132150720041D-03,   0.000000000000000D+00,   3.881420744680984D-03, &
   0.000000000000000D+00,  -1.513421660621093D-02,   0.000000000000000D+00, &
   1.002400936831017D-01,   0.000000000000000D+00,  -1.928399304489886D+00, &
  -8.010018789148188D-02,   4.070871803977276D+00,  -6.618410984629623D-02, &
  -1.945855434215285D+00,   0.000000000000000D+00,   1.023641287682405D-01, &
   0.000000000000000D+00,  -1.540774561442924D-02,   0.000000000000000D+00, &
   3.649172637440202D-03,   0.000000000000000D+00,  -1.136683098083811D-03, &
   0.000000000000000D+00,   4.269043338244498D-04,   0.000000000000000D+00, &
   5.544886778114630D-04,   4.920598885138869D-03,   1.545503410108890D-02, &
   1.360109519424642D-02,  -5.369173465419434D-02,   4.079459945808712D-02, &
  -7.085540613617367D-03,  -1.302073488851735D-02,   1.459347559400819D-02, &
  -5.537703886091174D-03,  -9.881366611685725D-03,   3.480360800229771D-02, &
  -6.100488591919320D-02,   7.079766327745432D-02,  -5.538132200587095D-02, &
   2.295402673364797D-02,   6.235762572569737D-03,  -1.532708325375128D-02, &
   5.975076854476663D-03,   5.103969485703548D-03,  -5.743986203871195D-03, &
   4.383617896649866D-05,  -3.945256106985230D-04,  -4.383617896650387D-05, &
   9.775467909530257D-04,  -2.805515453856219D-04,  -2.093177545650530D-04, &
  -2.410989843157627D-04,   9.687795551597224D-04,   9.534368925214117D-05, &
  -2.684965961698344D-04,  -5.469780997709252D-03,  -9.746243790219267D-04, &
   4.328822672942200D-04,   7.051536455417243D-04,   2.899580587888509D-04, &
   3.287713422487751D-03,  -2.294093365913677D-03,  -1.873631349326633D-03, &
  -1.528056138474031D-03,   2.054820889055106D-05,   6.945294605005376D-04, &
   1.603089064805027D-02,   6.429306248420483D-03,   4.479935723212847D-03, &
   1.240929166243432D-03,  -1.346227321135889D-03,  -1.361996168847451D-03, &
  -8.000102661386873D-04,  -7.073182359475558D-03,  -1.001468253321166D-03, &
  -1.237201072694748D-04,   4.102661263026791D-03,   4.259733690800688D-03, &
   2.656667166985513D-03,   8.533375850923093D-04,   7.573191885964549D-05, &
  -2.876805566702396D-02,  -1.179691290907321D-02,  -1.425276854376004D-02, &
  -1.167048179321250D-02,  -7.444577085512870D-03,  -3.116328175312787D-03, &
   4.317626125682999D-05,   5.824350579699413D-04,   4.848407142796138D-04, &
   1.255123506667634D-02,   1.065125269008881D-02,   1.041346617180703D-02, &
   5.741049518203389D-03,   1.848400272083748D-03,  -2.475841023176311D-03, &
  -3.486456756650333D-03,  -2.323888549418414D-03,  -1.147046991820754D-03, &
  -7.911337708894554D-04,   2.554328464052883D-02,   5.979389936687048D-03, &
   1.395843763109415D-02,   1.353047770024630D-02,   1.507700553094931D-02, &
   1.332150252200418D-02,   8.989673248140912D-03,   4.893717428617935D-03, &
   2.047612825942337D-03,   1.036909532588871D-03,   9.121581701220058D-04, &
  -1.301316996957969D-02,  -1.264182413254625D-02,  -1.354329975220800D-02, &
  -1.680131830186774D-02,  -1.626435497890227D-02,  -1.157109859122307D-02, &
  -6.745704049667414D-03,  -2.677981496435532D-03,  -8.432577972996540D-04, &
  -5.973876485385106D-04,  -7.295599153120374D-04,  -9.425477328495190D-04, &
  -8.458286027628052D-03,   3.366281497576363D-03,  -4.989577325752899D-03, &
  -2.117754630711496D-03,  -6.036229428611263D-03,  -8.073264503375371D-03, &
  -9.264478479709818D-03,  -7.935808592382212D-03,  -5.030777569569811D-03, &
  -2.596851965252263D-03,  -9.277608362442931D-04,  -4.508290222610061D-05, &
   2.827985031952997D-04,   4.949392611131737D-03,   3.870827781093666D-03, &
   4.623028665865394D-03,   1.173283147897962D-02,   1.374489693784207D-02, &
   1.644910658963558D-02,   1.571319985413049D-02,   1.188983493164902D-02, &
   8.049659113809296D-03,   4.973158940699112D-03,   2.998196145767681D-03, &
   1.876477320282600D-03,   1.145707951974613D-03,   1.337469567381392D-03, &
   0.000000000000000D+00,  -2.856945668463972D-03,   1.568279761547674D-03, &
  -2.987292409921701D-03,  -2.005796915059433D-03,  -2.061044208677544D-03, &
   2.582185589756300D-04,   1.641553043489164D-03,   1.922185149093305D-03, &
   9.868143637049673D-04,  -3.416213628738362D-04,  -1.094876626263286D-03, &
  -1.404012762795018D-03,  -1.513426266648315D-03,  -2.420632518267240D-03, &
   0.000000000000000D+00,   5.690407763512234D-04,   1.697837758167329D-04, &
  -3.358862291392430D-03,  -2.135992063036508D-03,  -7.077374997472235D-03, &
  -9.237380627235340D-03,  -1.060916478927424D-02,  -9.660056215148146D-03, &
  -7.099022163700377D-03,  -4.695331117160345D-03,  -2.716117518026920D-03, &
  -1.236589295011250D-03,  -2.681392036131576D-04,   6.530842031713369D-04, &
   1.534886990971200D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.899253085662845D-04,   1.302671702599329D-03,   7.017961987835350D-04, &
   1.339385966868803D-03,   1.944638524645285D-03,   1.788567457398019D-03, &
   8.467618422676174D-04,   7.404497546994994D-04,   1.152178483667412D-03, &
   1.692280717114781D-03,   2.025842427991901D-03,   1.866492468952921D-03, &
   1.422521026030051D-03,   5.297642182952497D-04,   7.569994937402311D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.418448713027518D-04, &
   6.293688037585678D-04,  -1.144804038671547D-03,   1.902182073602559D-03, &
   1.785756509657197D-03,   4.191335264534865D-03,   5.149647116045554D-03, &
   5.312993834345379D-03,   4.413866551032862D-03,   2.899096075811535D-03, &
   1.520373995418109D-03,   3.452145576711781D-04,  -5.359554217753582D-04, &
  -9.307887486999897D-04,  -9.177699643629543D-04,  -2.421016994519554D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.745452721661460D-04,   3.277595109804252D-04,   4.571728464735002D-04, &
  -6.733772628505771D-04,  -5.858548936097283D-05,  -1.650503152130351D-04, &
  -7.537610793874113D-05,  -7.479848535931208D-05,  -6.541638835576551D-04, &
  -1.205227731572930D-03,  -1.422409885114363D-03,  -1.295460609708184D-03, &
  -8.163000053271325D-04,  -3.970370807480458D-04,   3.847648993282521D-04, &
   1.823905793049139D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.951497677716107D-05,   3.931338172640420D-04, &
  -9.541288011106453D-04,   1.072482990894911D-04,  -1.184958587200656D-03, &
  -1.720181728096395D-03,  -2.739850240805592D-03,  -2.837458235821599D-03, &
  -2.578922599657520D-03,  -1.969037871702358D-03,  -1.101856846491073D-03, &
  -3.072501215878683D-04,   3.337913594140014D-04,   7.602604372772559D-04, &
   1.006989257279726D-03,   2.119337326355446D-04,   4.107992135525824D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.476059051883478D-04,  -4.638164818304852D-05, &
   5.842515966189160D-04,  -6.692803723871550D-04,   5.106267130804670D-05, &
  -6.500296743094954D-04,  -7.776022729332106D-04,  -6.481305633928534D-04, &
  -2.520801733842446D-04,   3.902506416323489D-04,   8.037137048836799D-04, &
   8.483461733221012D-04,   6.185723458326018D-04,   8.451151179482393D-05, &
  -3.727729279681546D-04,  -1.262686317224814D-04,  -1.789278913119076D-03 /)
lem(:, 36) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.311020180384732D-03,   6.770079100229649D-03,  -1.496598152472890D-02, &
   1.440509291004489D-02,   2.014606046523200D-02,  -6.283035863241929D-02, &
  -1.137934319330609D-01,  -4.721304949003209D-02,   2.751502442272652D-02, &
   4.612975272066294D-02,   4.877179078381454D-02,   4.214548146361528D-02, &
   2.429077673183418D-02,   4.322554538599269D-03,  -2.105730964493247D-02, &
  -4.544983030536835D-02,  -5.065358078694578D-02,  -3.326560981169630D-02, &
  -7.476832556717054D-03,   1.012520797748796D-02,   1.138624792650713D-02, &
   1.489628234891485D-03,  -1.311020180384975D-03,   0.000000000000000D+00, &
   3.629742233596071D-03,   0.000000000000000D+00,  -1.514092550541324D-02, &
   0.000000000000000D+00,   1.009957973947417D-01,   0.000000000000000D+00, &
  -1.938357173611015D+00,  -6.618410984629623D-02,   4.023033835671169D+00, &
  -5.560384374855298D-02,  -1.951592647277167D+00,   0.000000000000000D+00, &
   1.026149667148340D-01,   0.000000000000000D+00,  -1.542208233586503D-02, &
   0.000000000000000D+00,   3.644798835626137D-03,   0.000000000000000D+00, &
  -1.132366216736717D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.692519775057449D-03,  -4.291320867305680D-03,  -3.619013489715791D-03, &
   9.288526320418522D-03,   2.325831968766499D-02,  -4.808756573958553D-02, &
   2.653740325297222D-02,   1.099776144800341D-03,  -1.028252373490959D-02, &
   1.164045237531528D-02,  -7.452316170217832D-03,  -3.398038698137298D-03, &
   2.010418475329335D-02,  -4.129374379848841D-02,   5.718534677938909D-02, &
  -5.482739270240528D-02,   3.292902273124947D-02,  -4.895259339685391D-03, &
  -1.201843481102258D-02,   1.081879099747661D-02,  -2.420886649029630D-05, &
   1.338054368514318D-04,  -1.146903744440847D-04,   1.338054368514309D-04, &
  -2.246019832863291D-04,  -1.768143272679642D-04,   8.721247223351845D-05, &
   1.320532227974226D-03,  -1.720355616660997D-05,  -2.372657121311947D-04, &
  -8.402662849896427D-05,  -1.964072662354906D-03,   9.254876048890567D-04, &
   3.997437425936483D-04,  -7.048679262708960D-05,  -3.072348051514209D-04, &
  -4.104322483230867D-03,  -1.874869037787297D-03,  -2.127347153750994D-04, &
   6.203634212117758D-04,   5.808987819499397D-04,   5.324341341379493D-05, &
   6.436041512553754D-03,  -6.027616345783464D-04,  -1.055390383165653D-03, &
  -1.978807189627246D-03,  -7.649210806673383D-04,   3.416817405313320D-04, &
   6.265857722558368D-04,   1.200234768557322D-02,   6.148359823323195D-03, &
   4.884340118867658D-03,   3.031800949438042D-03,  -1.782189947327415D-04, &
  -1.349606672329040D-03,  -1.049864935438527D-03,  -5.220502747156508D-04, &
  -1.040277950243272D-02,  -3.630301827467621D-03,  -3.373259996209444D-03, &
   1.389880305339592D-03,   3.464668352171468D-03,   3.380968174239779D-03, &
   1.690992016164234D-03,   3.212470137050021D-04,  -1.374136593698702D-04, &
  -2.335777954626444D-02,  -9.102294503850085D-03,  -1.138798700033549D-02, &
  -1.122046763014867D-02,  -8.755866741736479D-03,  -5.342294621869042D-03, &
  -1.336884852193470D-03,   4.151635476300104D-04,   4.907205521045579D-04, &
   3.396531914402889D-04,   1.431336529185575D-02,   1.138711889115284D-02, &
   1.291126875509319D-02,   9.709426703060445D-03,   6.170672262673334D-03, &
   8.115215430450277D-04,  -2.326184952863509D-03,  -2.572409314141776D-03, &
  -1.531861316509933D-03,  -6.459814012004461D-04,  -3.784545323524553D-04, &
   2.146860031080338D-02,   3.345435113566891D-03,   9.535631289720577D-03, &
   9.149285945517139D-03,   1.165307498695190D-02,   1.237246475054824D-02, &
   9.888250481288850D-03,   6.360696114839781D-03,   2.951812927669599D-03, &
   1.157798361909334D-03,   6.404052719089628D-04,   6.344495199871790D-04, &
  -1.284511416531980D-02,  -1.156883075522747D-02,  -1.330070577431057D-02, &
  -1.712014039045024D-02,  -1.855041336613521D-02,  -1.551800844989768D-02, &
  -1.069829054343619D-02,  -5.427188459062197D-03,  -2.021238782636583D-03, &
  -8.121888665959352D-04,  -5.624106642522930D-04,  -6.035575148541503D-04, &
  -7.336145631540188D-04,  -7.176567341111074D-03,   3.972218009903806D-03, &
  -2.966117586688746D-03,   1.099535709370563D-03,  -1.269332667614177D-03, &
  -3.226638811815494D-03,  -5.818960909694589D-03,  -6.562222884373919D-03, &
  -4.938815458174060D-03,  -2.888299330628832D-03,  -1.181685216334519D-03, &
  -2.088833822259409D-04,   2.426974269032703D-04,   4.484010208670881D-04, &
   4.565232873205385D-03,   3.242910554504258D-03,   4.021436151992238D-03, &
   1.023832425236361D-02,   1.251726790277202D-02,   1.621532959335081D-02, &
   1.726021865111396D-02,   1.450243598463007D-02,   1.034730677693565D-02, &
   6.387323420462722D-03,   3.611519162375667D-03,   2.056550821976869D-03, &
   1.101914990840867D-03,   5.106590323317923D-04,   4.621384974757493D-04, &
   0.000000000000000D+00,  -2.687007761592519D-03,   1.336032456309650D-03, &
  -3.489600291386737D-03,  -3.402701172308985D-03,  -4.857774385047893D-03, &
  -3.660273215968782D-03,  -2.046285278877316D-03,  -5.468497526364811D-04, &
  -2.122516791050281D-04,  -8.504155226279530D-04,  -1.313567489025691D-03, &
  -1.449614063623570D-03,  -1.346080500483724D-03,  -1.167621798125990D-03, &
  -1.682399197513103D-03,   0.000000000000000D+00,   5.582644711828965D-04, &
   2.051307548894988D-04,  -2.849377396710789D-03,  -1.395052718474340D-03, &
  -5.785429926762623D-03,  -8.172412861792166D-03,  -1.042983486914290D-02, &
  -1.049962997457183D-02,  -8.255024191009588D-03,  -5.494104869364407D-03, &
  -3.076023940659712D-03,  -1.295452477990929D-03,  -2.030539473871565D-04, &
   4.595067816217896D-04,   9.734623084744882D-04,   1.845676061180513D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.592310480549422D-04, &
   1.255122740972144D-03,   6.915570464423050D-04,   1.736399993318031D-03, &
   3.040376511413367D-03,   3.782577080935002D-03,   3.234051029544560D-03, &
   2.717424373469367D-03,   2.424008685068470D-03,   2.382530869084181D-03, &
   2.371593158263598D-03,   2.000076477592027D-03,   1.425080913077553D-03, &
   8.245758358942496D-04,  -4.513285105569931D-06,  -3.196018834102524D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.395005741317587D-04, &
   5.960036405441262D-04,  -1.146358496683440D-03,   1.634654086389577D-03, &
   1.302054506902050D-03,   3.590984488417340D-03,   4.882781646492280D-03, &
   5.516663096962950D-03,   4.865892446930310D-03,   3.251726727551736D-03, &
   1.636390735973027D-03,   3.084821370582791D-04,  -6.249273897261298D-04, &
  -1.074716494929590D-03,  -1.131984693842419D-03,  -7.382488902763814D-04, &
  -1.608769000745342D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.704419092323232D-04,   3.192603265156905D-04, &
   3.780490147057759D-04,  -8.102300395026330D-04,  -5.743686678112614D-04, &
  -1.112532830590018D-03,  -1.302870404170170D-03,  -1.213096331466045D-03, &
  -1.472092802924333D-03,  -1.733156244869201D-03,  -1.725377386460360D-03, &
  -1.443393670391297D-03,  -8.680930855230760D-04,  -2.108235800573405D-04, &
   2.189916296347164D-04,   6.273043918209706D-04,   2.086067844801846D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.975180768986013D-05,   3.898794688840691D-04,  -9.125820823390559D-04, &
   1.819976181675991D-04,  -1.042891292881929D-03,  -1.583858092735754D-03, &
  -2.775986281787061D-03,  -3.126629625690776D-03,  -2.942242685024869D-03, &
  -2.223385855801595D-03,  -1.180982896768484D-03,  -2.485543039035188D-04, &
   4.455460911760178D-04,   8.153579867166554D-04,   8.216436034492840D-04, &
   7.371644110534881D-04,  -5.531312900634303D-05,  -6.624771650578302D-04 /)
lem(:, 37) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.599128549427715D-04,   4.692692922018772D-03, &
  -1.585870726495718D-02,   2.678314039687541D-02,   1.247843284616518D-03, &
  -8.206839292587249D-02,  -8.231393037393453D-02,  -2.361291587166593D-02, &
   2.684988850251264D-02,   4.214548146361585D-02,   3.614670692659545D-02, &
   3.285142847455040D-02,   2.532809841680225D-02,   8.435107919934452D-03, &
  -1.097792640789897D-02,  -3.056634007584774D-02,  -4.321180391645125D-02, &
  -3.775753628671193D-02,  -1.681813664434351D-02,   3.746779756632130D-03, &
   1.138813127667275D-02,   0.000000000000000D+00,  -1.306463328199849D-03, &
   0.000000000000000D+00,   3.714276175212247D-03,   0.000000000000000D+00, &
  -1.527613833269318D-02,   0.000000000000000D+00,   1.016095849668119D-01, &
   0.000000000000000D+00,  -1.945855434215285D+00,  -5.560384374855298D-02, &
   3.986235398512618D+00,  -4.737232977966290D-02,  -1.956111107405673D+00, &
   0.000000000000000D+00,   1.028237486389933D-01,   0.000000000000000D+00, &
  -1.543948877550609D-02,   0.000000000000000D+00,   3.645133841846771D-03, &
   0.000000000000000D+00,  -1.131461703894743D-03,   0.000000000000000D+00, &
  -1.866376183141407D-04,  -1.656244560712505D-03,  -5.634014423076887D-03, &
  -9.273748871269572D-03,   1.894873135899230D-03,   3.137901076615685D-02, &
  -3.723337919657167D-02,   1.605220467788504D-02,   3.630046117194874D-03, &
  -9.840269553396461D-03,   8.672223915637222D-03,  -6.821122800316859D-03, &
   1.057702005917614D-03,   1.100443782245128D-02,  -2.730359981607838D-02, &
   4.310732592391285D-02,  -4.921333451110285D-02,   3.833999049101904D-02, &
  -1.520928383015204D-02,  -5.491871809319917D-03,   1.199104961236702D-02, &
  -1.475499927354803D-05,   1.327949934620068D-04,   1.475499927355379D-05, &
  -3.435806973699744D-04,   9.766404281069141D-05,   7.166713932870597D-05, &
   3.154759368489192D-04,  -3.309335551355009D-04,  -4.215714078158037D-06, &
   1.166347561624114D-04,   9.253492401559560D-04,   4.427905020093267D-04, &
  -2.002464187125667D-04,  -1.825404195842980D-04,   2.309860005324596D-05, &
  -3.233312174145557D-03,   3.714044102858303D-04,   5.397519258069874D-04, &
   2.270864650101764D-04,  -2.399268274732388D-04,  -2.853862776160571D-04, &
  -1.696824916459085D-03,  -2.148493967814856D-03,  -7.668383908171723D-04, &
   8.823617314499047D-05,   7.872542637297722D-04,   3.330174592536838D-04, &
  -1.464433677900565D-04,   7.764080617745918D-03,   1.059408947841422D-03, &
   1.276850370036451D-04,  -1.628377049699172D-03,  -1.511986613329932D-03, &
  -1.668289003361828D-04,   4.740043516630305D-04,   5.042521001738231D-04, &
   7.775884617164774D-03,   4.897676092202819D-03,   4.197164936215350D-03, &
   4.071281158863298D-03,   1.302807351983570D-03,  -7.131183766871149D-04, &
  -1.287632378163592D-03,  -7.239786310225535D-04,  -2.627542605011486D-04, &
  -1.165596997304472D-02,  -5.161776167423327D-03,  -5.662789163209290D-03, &
  -1.549900628915558D-03,   1.711613247769195D-03,   3.210117712110471D-03, &
   2.606314557953084D-03,   8.684610968714468D-04,   5.449901995844934D-05, &
  -1.319071807784159D-04,  -1.860444111509121D-02,  -6.407912217303550D-03, &
  -7.869341601664158D-03,  -9.200642921415630D-03,  -8.657656818853597D-03, &
  -6.925532502573758D-03,  -3.172604211445879D-03,  -3.551610591557957D-04, &
   5.218196104310762D-04,   3.993242866424164D-04,   2.900977199565574D-04, &
   1.491581593520506D-02,   1.123450939788473D-02,   1.369471798103654D-02, &
   1.210178721665200D-02,   9.797564935570373D-03,   4.580129854900201D-03, &
  -1.261555252738283D-04,  -2.101464047546214D-03,  -1.976082241372570D-03, &
  -9.407887397753344D-04,  -3.663150569935772D-04,  -2.037455377690866D-04, &
   1.807446353621498D-02,   1.295484048433907D-03,   5.741122207721674D-03, &
   4.696572562153414D-03,   7.089163096757414D-03,   9.646653796585837D-03, &
   9.499264625393663D-03,   7.389024185546339D-03,   4.148163933737431D-03, &
   1.692636732630153D-03,   6.514780827441634D-04,   3.593693580412821D-04, &
   3.185669915366885D-04,  -1.244313820506994D-02,  -1.040452741939553D-02, &
  -1.237845163359316D-02,  -1.612884784697680D-02,  -1.886311410054376D-02, &
  -1.788980594339995D-02,  -1.423492817979664D-02,  -8.647410906001395D-03, &
  -3.877666262795568D-03,  -1.455351258575029D-03,  -5.682732269305956D-04, &
  -4.530972781029759D-04,  -4.701325806442377D-04,  -5.946576023280084D-04, &
  -6.134314629019749D-03,   4.326907199297699D-03,  -1.494712501765017D-03, &
   3.461683090063494D-03,   2.790465566286665D-03,   1.853069140733936D-03, &
  -1.128715571725716D-03,  -3.762751340339648D-03,  -4.019276819274000D-03, &
  -2.907747272776849D-03,  -1.430778679810876D-03,  -3.476541104169100D-04, &
   1.348947088160435D-04,   3.484058913195785D-04,   5.084268600055427D-04, &
   4.221383955680867D-03,   2.706984359366033D-03,   3.429277596411001D-03, &
   8.669021880008846D-03,   1.064310024920904D-02,   1.456223124638054D-02, &
   1.704677152800428D-02,   1.599144920970777D-02,   1.245354292665562D-02, &
   8.053917848055705D-03,   4.484870390465524D-03,   2.388649723600899D-03, &
   1.189468622215614D-03,   5.392411429818395D-04,   1.401915007271392D-04, &
   6.419412757248002D-05,   0.000000000000000D+00,  -2.519242862768614D-03, &
   1.187732347088548D-03,  -3.708928277994166D-03,  -4.209958518450969D-03, &
  -6.830985362156505D-03,  -7.102685948737530D-03,  -6.103921938853475D-03, &
  -3.939801647342867D-03,  -2.219259199448540D-03,  -1.758780187506691D-03, &
  -1.699881386372376D-03,  -1.594988586623452D-03,  -1.378991441827527D-03, &
  -1.041851356163155D-03,  -7.425159480060609D-04,  -9.673737843275970D-04, &
   0.000000000000000D+00,   5.442755550237505D-04,   2.224584967860540D-04, &
  -2.424214121115791D-03,  -7.282833322886256D-04,  -4.375373089008650D-03, &
  -6.460748611877889D-03,  -9.111622574641390D-03,  -1.023995728720536D-02, &
  -8.848507030166607D-03,  -6.195773850114171D-03,  -3.486925939951593D-03, &
  -1.414644885469621D-03,  -2.017268959540642D-04,   4.665895990675901D-04, &
   7.681789030270280D-04,   9.617372145288141D-04,   1.603843584499971D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.270672228641742D-04, &
   1.200064847199542D-03,   6.235007386739007D-04,   1.885680151005564D-03, &
   3.678944735562164D-03,   5.323379630905985D-03,   5.553063466662951D-03, &
   5.060668970466429D-03,   4.168410525611554D-03,   3.398032970713844D-03, &
   2.882299188353319D-03,   2.236627547655161D-03,   1.477457893563651D-03, &
   7.823418817853351D-04,   1.915070416749406D-04,  -4.449473867087267D-04, &
  -1.025220559052241D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368492323476916D-04,   5.646735695988468D-04,  -1.134505561566793D-03, &
   1.397643290130730D-03,   7.936964710822570D-04,   2.732019704920998D-03, &
   4.049959459425843D-03,   5.068737935425839D-03,   4.863217555920804D-03, &
   3.408045197653567D-03,   1.690639172802008D-03,   2.512541034060481D-04, &
  -7.195109727012460D-04,  -1.126275836149584D-03,  -1.162294196455503D-03, &
  -9.436532319160382D-04,  -3.879131481597422D-04,  -6.491323422168711D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.645433539783590D-04,   3.137467114806618D-04,   3.254847649028502D-04, &
  -8.613834469990212D-04,  -9.011782415879767D-04,  -1.862947394395911D-03, &
  -2.497357616422580D-03,  -2.533894772584212D-03,  -2.541262170799832D-03, &
  -2.442774638420883D-03,  -2.130111881663167D-03,  -1.632705017150634D-03, &
  -9.220758440335884D-04,  -1.849532523896402D-04,   3.809525578196847D-04, &
   6.608542543442243D-04,   6.963333367634823D-04,   1.861743074990365D-03 /)
lem(:, 38) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.794986263737043D-04,  -2.992518038793543D-03,   6.895807729312744D-03, &
  -8.894083870854171D-03,   1.349395749116873D-04,   2.344111371446145D-02, &
  -1.946936977962388D-02,  -8.171172847043993D-02,  -6.108720078285131D-02, &
  -2.930738193723886D-03,   2.429077673183455D-02,   3.285142847455098D-02, &
   3.193902295650046D-02,   2.535649275180843D-02,   2.121739451357734D-02, &
   1.276349710993269D-02,  -3.502017449904066D-03,  -2.076605731291064D-02, &
  -3.361275426875297D-02,  -3.628408620433343D-02,  -2.426851604156428D-02, &
  -4.383571927570614D-03,   5.794986263733818D-04,   0.000000000000000D+00, &
  -1.177532936345904D-03,   0.000000000000000D+00,   3.650903219680429D-03, &
   0.000000000000000D+00,  -1.530121851232797D-02,   0.000000000000000D+00, &
   1.020242680797511D-01,   0.000000000000000D+00,  -1.951592647277167D+00, &
  -4.737232977966290D-02,   3.957292913080569D+00,  -4.084238227953411D-02, &
  -1.959714073878895D+00,   0.000000000000000D+00,   1.029866776864045D-01, &
   0.000000000000000D+00,  -1.544972407637459D-02,   0.000000000000000D+00, &
   3.642719732466058D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.481295096983064D-04,   1.994429335339969D-03,   2.500008945935042D-03, &
  -9.171781259199029D-04,  -9.396072162799394D-03,  -6.398904072051552D-03, &
   3.298916747878809D-02,  -2.826469581785070D-02,   7.329325818088557D-03, &
   4.216251536399397D-03,  -8.003289245966071D-03,   7.770634146649537D-03, &
  -5.561342920176789D-03,   2.507131083724909D-03,   4.710854468850286D-03, &
  -1.729241956165965D-02,   3.148520082701364D-02,  -4.095310032826909D-02, &
   3.875657151315604D-02,  -2.317916179460770D-02,   2.684178709436921D-03, &
  -5.914483088580278D-05,   5.330336610695715D-05,  -5.914483088579736D-05, &
   5.038263371753049D-05,   7.228812663819695D-05,  -4.490626048736113D-05, &
  -4.033740053534011D-04,  -1.070239796981093D-05,   9.263937268020427D-05, &
   9.075132781418429D-06,   8.667273365609959D-04,  -2.850843435818095D-04, &
  -1.282097207085266D-04,   5.987501398315103D-05,   1.037772727116531D-04, &
  -8.407726464642167D-05,   7.575033246734654D-04,   1.514323880736196D-05, &
  -2.199140797082220D-04,  -1.073561181784322D-04,   1.131568064985953D-04, &
  -3.540365765837991D-03,  -4.157662556342361D-04,   3.605482925486360D-04, &
   4.783932064336599D-04,  -1.787075716290637D-05,  -3.109003463047056D-04, &
  -1.887432033776864D-04,   8.848358773754506D-04,  -1.731211019604685D-03, &
  -9.318738931911709D-04,  -5.904288973908896D-04,   5.809670994692757D-04, &
   6.490113924627470D-04,   1.165480091336960D-04,  -2.284879820956243D-04, &
   7.443048384583776D-03,   2.204130697677385D-03,   1.218722055683472D-03, &
  -6.474666284814517D-04,  -1.739811770445259D-03,  -8.810631066574666D-04, &
   1.777804998751157D-04,   5.104116759845882D-04,   4.124427816301846D-04, &
   4.196654333741132D-03,   3.208278493160704D-03,   2.730214343265721D-03, &
   4.076715304335740D-03,   2.515169494283667D-03,   4.322892195359988D-04, &
  -1.086227770462388D-03,  -1.081517747553053D-03,  -4.385742092267886D-04, &
  -7.225732203877065D-05,  -1.153896859327457D-02,  -5.664071629015330D-03, &
  -6.698902941516478D-03,  -3.985544548295969D-03,  -5.298522567743782D-04, &
   2.078733113036491D-03,   3.007976085696507D-03,   1.700929250580701D-03, &
   3.617588819618980D-04,  -8.404258792719169D-05,  -1.471166406319311D-04, &
  -1.471083885708216D-02,  -4.097638603148540D-03,  -4.467505530186996D-03, &
  -6.222676768116964D-03,  -7.134284410306724D-03,  -7.379570450837009D-03, &
  -4.866344511617473D-03,  -1.680347036520954D-03,   1.753637634302609D-04, &
   5.329569801330972D-04,   2.877745772077466D-04,   1.511241503599939D-04, &
   1.481045702548163D-02,   1.054795858028855D-02,   1.321718317786743D-02, &
   1.281039331630123D-02,   1.204111226346955D-02,   8.075296401345898D-03, &
   2.799550201526625D-03,  -7.366573170831840D-04,  -1.995760948207893D-03, &
  -1.377314299090150D-03,  -5.454763171305164D-04,  -2.155290483385380D-04, &
  -1.629091289837623D-04,   1.528359468299689D-02,  -2.154908912770131D-04, &
   2.779822274597288D-03,   8.431802652905663D-04,   2.287972267375934D-03, &
   5.554513632505770D-03,   7.553945223322874D-03,   7.499699830233316D-03, &
   5.229038520215089D-03,   2.551270370519290D-03,   9.670630263111721D-04, &
   3.464009077435297D-04,   2.028986489241527D-04,   1.780426956390190D-04, &
  -1.191720624599129D-02,  -9.262212963178990D-03,  -1.113464888261823D-02, &
  -1.434970424120495D-02,  -1.753745917112205D-02,  -1.829606133929036D-02, &
  -1.654439133177071D-02,  -1.178724839448836D-02,  -6.289541315945038D-03, &
  -2.639133957878128D-03,  -8.767878767188283D-04,  -3.839584512359462D-04, &
  -3.371920765901172D-04,  -3.172764718536450D-04,  -3.521822725565110D-04, &
  -5.281554044499243D-03,   4.500821451562429D-03,  -4.723313127106277D-04, &
   5.012750372208421D-03,   5.798850598345780D-03,   6.335932216762468D-03, &
   4.041391175278768D-03,   3.333893851102927D-04,  -1.952280911525060D-03, &
  -2.366316418933490D-03,  -1.551708514291987D-03,  -5.397369329048396D-04, &
   1.231417316542254D-05,   2.584031727994488D-04,   3.575792215170954D-04, &
   5.060088795166284D-04,   3.913803454748202D-03,   2.254620248689380D-03, &
   2.891140915836627D-03,   7.189865710296247D-03,   8.539167575295986D-03, &
   1.199145891145737D-02,   1.516954133475964D-02,   1.584566410000186D-02, &
   1.375115244847151D-02,   9.689907063801691D-03,   5.602267681838474D-03, &
   2.908517241598826D-03,   1.365329933243779D-03,   5.539747209351097D-04, &
   1.590880636270118D-04,  -8.375860685068143D-05,  -1.981329949027740D-04, &
   0.000000000000000D+00,  -2.358363738711016D-03,   1.093843562477136D-03, &
  -3.735738399293887D-03,  -4.544601451180157D-03,  -7.943140058370084D-03, &
  -9.623561697913634D-03,  -9.840262118955233D-03,  -7.833941822533993D-03, &
  -5.060844187524587D-03,  -3.259064818080002D-03,  -2.344541529933420D-03, &
  -1.830771232340608D-03,  -1.455665543040399D-03,  -1.048858225679757D-03, &
  -7.023814326642292D-04,  -4.149004404785261D-04,  -5.139279585511054D-04, &
   0.000000000000000D+00,   5.283151506597303D-04,   2.268017136174644D-04, &
  -2.076815360972194D-03,  -1.774819711561964D-04,  -3.041081479033127D-03, &
  -4.461920891339699D-03,  -6.916200475288942D-03,  -8.763117973501025D-03, &
  -8.494799385780698D-03,  -6.497996377555428D-03,  -3.844684655494242D-03, &
  -1.567297917793418D-03,  -1.996618172750537D-04,   4.997309221527874D-04, &
   7.824602365272260D-04,   7.996280555243133D-04,   7.702876064337327D-04, &
   1.185446029226703D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.941724385532574D-04,   1.143284304231506D-03,   5.260326130782756D-04, &
   1.857952732513897D-03,   3.894391434564536D-03,   6.247140046372006D-03, &
   7.418984293315790D-03,   7.413477639712854D-03,   6.260823656204525D-03, &
   4.777170772303768D-03,   3.600938476913298D-03,   2.561905126092625D-03, &
   1.591252428286603D-03,   8.042078416309625D-04,   1.834273176060765D-04, &
  -2.337110136084611D-04,  -6.387070565460181D-04,  -1.206561094710630D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.339679928267160D-04, &
   5.353375310706730D-04,  -1.111644455572373D-03,   1.203145273399012D-03, &
   3.342972475248436D-04,   1.786195728602808D-03,   2.817336625974739D-03, &
   3.963374126028077D-03,   4.233455532993575D-03,   3.188226776480590D-03, &
   1.591280280515429D-03,   1.326685014631819D-04,  -8.584969686114029D-04, &
  -1.235520443612285D-03,  -1.187907477758981D-03,  -9.225501192601451D-04, &
  -5.477189700812501D-04,   1.649431206322582D-05,   1.766099774378851D-04 /)
lem(:, 39) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.239879943972302D-04,  -1.877268697522491D-03, &
   6.582035048402992D-03,  -1.331403249278493D-02,   1.126075805391779D-02, &
   1.774015688655539D-02,  -4.148329364639777D-02,  -7.037879355848067D-02, &
  -4.019119343515039D-02,   4.322554538599188D-03,   2.532809841680255D-02, &
   2.535649275180880D-02,   2.497212554862433D-02,   2.300078944702485D-02, &
   1.746812777915893D-02,   1.222023228331496D-02,   2.810197727086719D-03, &
  -1.227742531977541D-02,  -2.593437800455373D-02,  -3.185526284975346D-02, &
  -2.696058649741484D-02,   0.000000000000000D+00,   5.226386535917964D-04, &
   0.000000000000000D+00,  -1.197871946031670D-03,   0.000000000000000D+00, &
   3.671892851487052D-03,   0.000000000000000D+00,  -1.535822080007256D-02, &
   0.000000000000000D+00,   1.023641287682405D-01,   0.000000000000000D+00, &
  -1.956111107405673D+00,  -4.084238227953411D-02,   3.934138924734929D+00, &
  -3.557532511813766D-02,  -1.962645907345424D+00,   0.000000000000000D+00, &
   1.031244411259704D-01,   0.000000000000000D+00,  -1.546094251753998D-02, &
   0.000000000000000D+00,   3.642617823019971D-03,   0.000000000000000D+00, &
   7.466266479902135D-05,   6.625654226549657D-04,   2.343989145962603D-03, &
   4.689950310413466D-03,   3.622446987562095D-03,  -7.400656430634391D-03, &
  -1.509614368618169D-02,   2.975850873278404D-02,  -1.987618400040785D-02, &
   2.963324140592571D-03,   5.268932942057199D-03,  -6.168931123778016D-03, &
   6.232813794056047D-03,  -5.333284475850087D-03,   2.783723678909953D-03, &
   1.570688108289172D-03,  -9.947696299624432D-03,   2.198459648521290D-02, &
  -3.266773123804208D-02,   3.570332111410141D-02,  -2.740604255438027D-02, &
   5.902601923555283D-06,  -5.312341731209926D-05,  -5.902601923562388D-06, &
   1.404819257809145D-04,  -3.974418628535683D-05,  -2.892274942547787D-05, &
  -1.751105237325160D-04,   1.333988034726249D-04,  -4.131821346500286D-06, &
  -5.213965032485385D-05,  -1.379062449415310D-04,  -2.050681449238063D-04, &
   9.034047229771777D-05,   5.823645040693959D-05,  -3.812097075637463D-05, &
   1.057996678118261D-03,   5.366001748694054D-07,  -2.166510429842093D-04, &
  -3.717617116275953D-05,   1.069846598646648D-04,   6.235741198799798D-05, &
  -1.287733099652510D-03,   7.253045696989904D-04,   2.548620859127991D-04, &
  -8.447364371914844D-05,  -2.410056640162478D-04,   1.051736342744766D-05, &
   1.708065431632339D-04,  -2.808457995233475D-03,  -1.090264235300427D-03, &
  -1.193041055460304D-05,   4.377507569419865D-04,   3.363109655509770D-04, &
  -2.473011339249918D-04,  -2.814803292301319D-04,  -6.448592601497934D-05, &
   2.881650259085715D-03,  -8.319806844631081D-04,  -6.518164878016390D-04, &
  -1.020514271397820D-03,   1.009638781406516D-05,   7.360580470312272D-04, &
   4.310215106556233D-04,  -5.400433593252059D-05,  -2.580082637684388D-04, &
   6.061972175503950D-03,   2.610130570601610D-03,   1.847460742059177D-03, &
   5.493726349295685D-04,  -1.325724048058822D-03,  -1.365807393172649D-03, &
  -4.235125479521551D-04,   4.224116576574961D-04,   4.561143408281804D-04, &
   2.727647685773582D-04,   1.606350952056606D-03,   1.589486375132140D-03, &
   9.924266919880854D-04,   3.120894331557196D-03,   3.039522502200533D-03, &
   1.617749774635789D-03,  -2.872539407270452D-04,  -1.226840085266285D-03, &
  -7.648977972595574D-04,  -2.326481562182556D-04,  -1.437112287529861D-05, &
  -1.067526743348497D-02,  -5.428005281483759D-03,  -6.586337070961975D-03, &
  -5.393723131075618D-03,  -2.644486973782445D-03,   2.697665233550752D-04, &
   2.568140689403450D-03,   2.384210926400557D-03,   1.006610778895704D-03, &
   6.424654503200116D-05,  -1.418043046830383D-04,  -1.659484539025620D-04, &
  -1.165405653029246D-02,  -2.319511247280100D-03,  -1.683538519305825D-03, &
  -3.061758952012154D-03,  -4.579340555666241D-03,  -6.489572053230897D-03, &
  -5.857038883413334D-03,  -3.196117043723812D-03,  -7.155115867757473D-04, &
   4.842133137048143D-04,   4.440687554588005D-04,   2.013245547627558D-04, &
   9.636676895285809D-05,   1.430687046340461D-02,   9.607301753356117D-03, &
   1.199478387744891D-02,   1.215166423148543D-02,   1.262089599392468D-02, &
   1.052130036696052D-02,   5.873874014342704D-03,   1.429586144093410D-03, &
  -1.310680132926790D-03,  -1.671451492058808D-03,  -9.010620571365013D-04, &
  -3.171779925416302D-04,  -1.164447718184613D-04,  -5.634781081662315D-05, &
   1.300038553340088D-02,  -1.280484008654381D-03,   6.366837583950438D-04, &
  -2.077072287945344D-03,  -1.941817263815380D-03,   9.040633711135071D-04, &
   4.191445811044450D-03,   6.318880854197813D-03,   5.768628762983313D-03, &
   3.509442192555845D-03,   1.558339299438373D-03,   4.882918950266036D-04, &
   1.694212735592851D-04,   1.216333581917264D-04,   1.456478215468556D-04, &
  -1.133555692945718D-02,  -8.201851979139080D-03,  -9.807920333831578D-03, &
  -1.224316375143187D-02,  -1.515454030927084D-02,  -1.687174083766183D-02, &
  -1.705728118041508D-02,  -1.410225633774617D-02,  -8.902587893314272D-03, &
  -4.352163595990802D-03,  -1.580004085919715D-03,  -4.783578433295312D-04, &
  -2.737495529240890D-04,  -2.388988871405942D-04,  -2.097773164922545D-04, &
  -2.269347473775409D-04,  -4.579024767873661D-03,   4.548910495471861D-03, &
   2.089251010094904D-04,   5.896433610952742D-03,   7.710087216243786D-03, &
   9.680659992580555D-03,   8.801551494821962D-03,   5.154509451397165D-03, &
   1.325152904394976D-03,  -9.624844334119541D-04,  -1.334317587708678D-03, &
  -6.852007996960327D-04,  -1.280696987715649D-04,   1.685842798405040D-04, &
   2.641434884957357D-04,   2.877721144186666D-04,   3.526119926455609D-04, &
   3.638414066402676D-03,   1.875092698849738D-03,   2.423860158807253D-03, &
   5.883507101267832D-03,   6.501389135883456D-03,   9.041376341180025D-03, &
   1.208251586594655D-02,   1.395526927773107D-02,   1.366981083074325D-02, &
   1.081383421342154D-02,   6.798766428212804D-03,   3.616803872218943D-03, &
   1.672008204830904D-03,   6.467092486781268D-04,   1.882797668312723D-04, &
  -5.010687667334289D-05,  -1.990487068108451D-04,  -3.307330273829513D-04, &
   0.000000000000000D+00,  -2.206816285498501D-03,   1.034136323148282D-03, &
  -3.641273417607968D-03,  -4.542320910708019D-03,  -8.310174361911862D-03, &
  -1.104682274775787D-02,  -1.266627923515766D-02,  -1.158606925477710D-02, &
  -8.475142556238151D-03,  -5.442285669275061D-03,  -3.396144928957937D-03, &
  -2.237708677200081D-03,  -1.595828665672560D-03,  -1.078245319071807D-03, &
  -6.734520193604268D-04,  -3.814970846021380D-04,  -1.565234139382164D-04, &
  -1.420186143679291D-04,   0.000000000000000D+00,   5.112570257274021D-04, &
   2.221951854390654D-04,  -1.796160007563404D-03,   2.487693367786834D-04, &
  -1.894790648207339D-03,  -2.489652163375304D-03,  -4.249220919027501D-03, &
  -6.235607357074058D-03,  -6.975907656888879D-03,  -6.052350206358140D-03, &
  -3.934283753381494D-03,  -1.690513572368750D-03,  -2.127766144834227D-04, &
   5.254085796731269D-04,   8.059511163757474D-04,   7.980859942067204D-04, &
   6.854548955859736D-04,   5.574287593167540D-04,   8.209419472946853D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.611429582507224D-04, &
   1.088053474838933D-03,   4.188373294671692D-04,   1.720494249822065D-03, &
   3.779387927108114D-03,   6.536499284947632D-03,   8.560852689277001D-03, &
   9.361119558737472D-03,   8.414318832066791D-03,   6.452607971329136D-03, &
   4.571392694303482D-03,   3.018474897818763D-03,   1.758821301943480D-03, &
   8.358804167513446D-04,   1.687023957874688D-04,  -2.510539769364913D-04, &
  -4.631228631073705D-04,  -6.465643671951959D-04,  -1.121955112211948D-03 /)
lem(:, 40) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.877608462752944D-04,   1.485990620419916D-03,  -3.506505582541178D-03, &
   5.157477039381174D-03,  -3.005564015725605D-03,  -6.718575955263457D-03, &
   1.673623640560183D-02,   4.147102916058689D-03,  -5.089698229479402D-02, &
  -5.990790294561191D-02,  -2.105730964493258D-02,   8.435107919934374D-03, &
   2.121739451357766D-02,   2.300078944702522D-02,   1.933677295429847D-02, &
   1.844734834094780D-02,   1.655097673014073D-02,   1.122974931910792D-02, &
   4.761943041090858D-03,  -5.132496600951482D-03,  -1.812040179615651D-02, &
  -2.717979613926168D-02,  -2.877608462760895D-04,   0.000000000000000D+00, &
   4.595420421556731D-04,   0.000000000000000D+00,  -1.156686757567310D-03, &
   0.000000000000000D+00,   3.650497922238615D-03,   0.000000000000000D+00, &
  -1.537824321752530D-02,   0.000000000000000D+00,   1.026149667148340D-01, &
   0.000000000000000D+00,  -1.959714073878895D+00,  -3.557532511813766D-02, &
   3.915314138680692D+00,  -3.126526997403580D-02,  -1.965055173895255D+00, &
   0.000000000000000D+00,   1.032361316490857D-01,   0.000000000000000D+00, &
  -1.546838938601005D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.714976551050955D-04,  -1.018984528259897D-03,  -1.505444099568734D-03, &
  -5.897280909646747D-04,   3.015531051856036D-03,   6.332030071499792D-03, &
  -2.426799162669581D-03,  -1.985918945969818D-02,   2.566954029963589D-02, &
  -1.225654332916893D-02,   4.589999663127708D-04,   4.797801455311636D-03, &
  -5.638921747988606D-03,   4.788831729027851D-03,  -4.463385393953832D-03, &
   3.267447841330626D-03,  -5.284905613844111D-05,  -5.547430000263161D-03, &
   1.443844386307628D-02,  -2.477348566336770D-02,   3.107143758447395D-02, &
   2.936946838858748D-05,  -2.723350705122865D-05,   2.936946838858088D-05, &
  -1.067980668675766D-05,  -3.417538139762003D-05,   2.416306262878450D-05, &
   1.432388618048011D-04,   1.121379702109388D-05,  -4.197892196533205D-05, &
   4.106062040548885D-06,  -3.640843188667003D-04,   1.004549089566425D-04, &
   5.227037204529690D-05,  -3.350789347969776D-05,  -3.562969598242379D-05, &
   3.343103123460477D-04,  -3.129506989725300D-04,  -8.032971479631898D-08, &
   8.900359026085979D-05,   1.850964223305161D-05,  -6.625323258043441D-05, &
   7.177800985017357D-04,   3.623043510855746D-04,  -1.598815605583279D-04, &
  -1.519162047377886D-04,   5.015767165053665D-05,   1.110426832183518D-04, &
  -2.241394088027177D-06,  -2.176059157002472D-03,   3.421869809606997D-04, &
   3.489465019644810D-04,   1.576285054363748D-04,  -2.374364660667372D-04, &
  -1.485442238153623D-04,   1.056448414569247D-04,   1.638887635928587D-04, &
  -1.385170927272314D-03,  -1.373714407371980D-03,  -3.624692781701925D-04, &
   1.057182612381059D-04,   5.505425565821971D-04,   3.430802322746377D-05, &
  -3.318497286066503D-04,  -2.049271344011166D-04,   2.046267648380512D-05, &
   3.923303270713475D-03,   1.734705743263185D-04,  -6.729318453567516D-05, &
  -9.861015283774879D-04,  -6.086202291050056D-04,   4.378148942479644D-04, &
   6.812266969301934D-04,   2.170289037357975D-04,  -1.709107689969250D-04, &
  -2.682377674446908D-04,   4.270396988033039D-03,   2.363288651112253D-03, &
   1.852027580680830D-03,   1.503633295493987D-03,  -4.584347474103375D-04, &
  -1.342700180293339D-03,  -1.043776938295221D-03,   1.484804145932323D-05, &
   5.217360796888235D-04,   3.579189410170352D-04,   1.487998930956686D-04, &
  -1.434145469365158D-05,   3.531964639977313D-04,  -5.209491807829831D-04, &
   1.594165373072991D-03,   2.720468393159245D-03,   2.388868386274933D-03, &
   8.336960075404276D-04,  -8.623321227572467D-04,  -1.080714478830113D-03, &
  -4.870176600281106D-04,  -8.988037699785438D-05,   4.363031795642402D-05, &
  -9.509821152995203D-03,  -4.787699212973159D-03,  -5.691846169767575D-03, &
  -5.640668023410072D-03,  -4.061292866423314D-03,  -1.710697102296337D-03, &
   1.319451649211171D-03,   2.503867332881919D-03,   1.737007107701768D-03, &
   4.995918877411900D-04,  -1.141590562027957D-04,  -1.462483087358975D-04, &
  -1.032986054214248D-04,  -9.311672367681538D-03,  -1.061309429761058D-03, &
   2.930742668733040D-04,  -3.538335455809817D-04,  -1.667735886510136D-03, &
  -4.431039113716212D-03,  -5.747846727999329D-03,  -4.411294903211760D-03, &
  -1.979297425182829D-03,  -6.266554598222795D-07,   5.433981302672129D-04, &
   3.242697523961979D-04,   1.379844148888786D-04,   9.445250836533321D-05, &
   1.360370494342043D-02,   8.599523474177317D-03,   1.045804926231384D-02, &
   1.063965261728157D-02,   1.172379334639752D-02,   1.142678960230219D-02, &
   8.358106796982619D-03,   4.028024986890293D-03,   1.585257582948445D-04, &
  -1.505688829610226D-03,  -1.263253377145387D-03,  -5.862758774195564D-04, &
  -1.766690959193464D-04,  -6.356703637784328D-05,  -2.467128929359116D-05, &
   1.113253359011770D-02,  -2.000685045948749D-03,  -8.148453622646082D-04, &
  -4.021547068503570D-03,  -5.109626741346607D-03,  -3.413074951815131D-03, &
   7.345475223669850D-06,   3.764548304263407D-03,   5.343069704768907D-03, &
   4.240494513314302D-03,   2.340225560004920D-03,   8.690620246932792D-04, &
   2.200041466423213D-04,   8.214452676871250D-05,   6.437054435010930D-05, &
   5.291881759358337D-05,  -1.073952454152942D-02,  -7.249098931208258D-03, &
  -8.536965680783859D-03,  -1.013651588830681D-02,  -1.231182387721301D-02, &
  -1.414135898037042D-02,  -1.568124551000366D-02,  -1.491740672934996D-02, &
  -1.111248077208061D-02,  -6.416727241372964D-03,  -2.746306232689291D-03, &
  -8.382952305789139D-04,  -2.726108041888038D-04,  -1.637143282303911D-04, &
  -1.552892634687119D-04,  -1.428050686430138D-04,  -1.820072730357905D-04, &
  -3.996108389831359D-03,   4.512064934940327D-03,   6.429932828422565D-04, &
   6.283444046782004D-03,   8.670632881674167D-03,   1.170104379657268D-02, &
   1.243905736544586D-02,   9.856369559401185D-03,   5.476872595844048D-03, &
   1.483754255750891D-03,  -5.132109235314404D-04,  -6.603812716468116D-04, &
  -2.713847047750154D-04,   5.969727120886102D-05,   2.187744717426373D-04, &
   2.330909218720556D-04,   2.182707165500875D-04,   2.615856925499689D-04, &
   3.391394905551785D-03,   1.557556940288962D-03,   2.029044688807142D-03, &
   4.777185875798972D-03,   4.700191012213927D-03,   6.151430340726302D-03, &
   8.400798687569121D-03,   1.063610813992622D-02,   1.192520136794274D-02, &
   1.087609130418435D-02,   7.737172734505941D-03,   4.449362869673823D-03, &
   2.121117536093120D-03,   8.107413644683321D-04,   2.348628216060073D-04, &
  -2.369990551276887D-05,  -1.365082328366340D-04,  -2.100901986023788D-04, &
  -2.848023075435950D-04,   0.000000000000000D+00,  -2.065648086465433D-03, &
   9.952002532857248D-04,  -3.477205449611489D-03,  -4.325536434836455D-03, &
  -8.119947848128740D-03,  -1.143388732434788D-02,  -1.424649731140173D-02, &
  -1.453464255144273D-02,  -1.193368924451464D-02,  -8.191882508696305D-03, &
  -4.968453101144864D-03,  -2.910613139730544D-03,  -1.830745390220174D-03, &
  -1.157151281354342D-03,  -6.872100218476441D-04,  -3.783899565135364D-04, &
  -1.516221203015951D-04,   2.108318464031293D-05,   9.390907031818367D-05, &
   0.000000000000000D+00,   4.937114797933231D-04,   2.117167495052041D-04, &
  -1.570379498293336D-03,   5.605666151994541D-04,  -9.805665661694973D-04, &
  -7.604566835230608D-04,  -1.532444413203303D-03,  -3.049419713825027D-03, &
  -4.338928916816623D-03,  -4.591292178810024D-03,  -3.474185697505300D-03, &
  -1.651346878426541D-03,  -2.062074994716599D-04,   5.599110439178402D-04, &
   8.398834494432097D-04,   7.998577772720770D-04,   6.525717565017661D-04, &
   4.832550051986260D-04,   3.332658375856162D-04,   4.427269701623386D-04 /)
lem(:, 41) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.014602182654893D-04,   8.503495569321089D-04, &
  -3.045896485452308D-03,   6.734834171467675D-03,  -8.574383115809421D-03, &
   2.749592040633584D-04,   1.964011518217025D-02,  -1.375098411627570D-02, &
  -5.215959155529001D-02,  -4.544983030536849D-02,  -1.097792640789908D-02, &
   1.276349710993256D-02,   1.746812777915902D-02,   1.844734834094815D-02, &
   1.755663523108865D-02,   1.463812150481276D-02,   1.367522367894621D-02, &
   1.150444308305127D-02,   5.861339006417272D-03,  -1.580971234949695D-03, &
  -1.117324950917787D-02,   0.000000000000000D+00,  -2.367405092860386D-04, &
   0.000000000000000D+00,   4.646187436716498D-04,   0.000000000000000D+00, &
  -1.161460122783058D-03,   0.000000000000000D+00,   3.656433216322537D-03, &
   0.000000000000000D+00,  -1.540774561442924D-02,   0.000000000000000D+00, &
   1.028237486389933D-01,   0.000000000000000D+00,  -1.962645907345424D+00, &
  -3.126526997403580D-02,   3.899811373694857D+00,  -2.769364170995378D-02, &
  -1.967064952565581D+00,   0.000000000000000D+00,   1.033319358844071D-01, &
   0.000000000000000D+00,  -1.547608741217479D-02,   0.000000000000000D+00, &
  -3.382007275521336D-05,  -3.001233730344878D-04,  -1.086172874129206D-03, &
  -2.389803270582092D-03,  -2.939658387561494D-03,   3.229283313571923D-04, &
   7.637368729126025D-03,   4.377579100522633D-03,  -2.118104517046914D-02, &
   2.025333793755170D-02,  -7.490935351192785D-03,  -1.629225480904097D-03, &
   4.047336887142314D-03,  -4.610403748412797D-03,   4.348914064853866D-03, &
  -3.554469502254738D-03,   2.992005378763941D-03,  -1.276391566881151D-03, &
  -2.783411749605161D-03,   9.305967780965239D-03,  -1.785132846856016D-02, &
  -2.673711513455805D-06,   2.406340362126528D-05,   2.673711513474436D-06, &
  -6.445633458516745D-05,   1.818565764949693D-05,   1.316968646305174D-05, &
   9.256345066069363D-05,  -6.069988039267312D-05,   3.447099141341606D-06, &
   2.510195272154836D-05,  -5.042978373441552D-06,   1.009013058487149D-04, &
  -4.356382023813309D-05,  -2.199980901422122D-05,   2.550199667905886D-05, &
  -3.392741039494496D-04,  -4.796476412051003D-05,   1.005440744206090D-04, &
   3.012528950440584D-06,  -4.721945782913318D-05,  -7.196648087152777D-06, &
   7.492137402963546D-04,  -2.287606947258487D-04,  -1.197958549417297D-04, &
   5.194862098599241D-05,   8.194144610635821D-05,  -3.282266771805744D-05, &
  -7.064566653392559D-05,  -4.930530267548316D-05,   5.950597386561692D-04, &
   6.078600121277154D-06,  -1.615002397976573D-04,  -9.119348513142535D-05, &
   1.307834952585896D-04,   6.702717562890035D-05,  -6.779433082899302D-05, &
  -2.408606861379136D-03,  -2.198630542071789D-04,   2.458658760337128D-04, &
   3.194213932823171D-04,  -3.708220159154934D-05,  -2.811859598489764D-04, &
  -1.028947013301732D-05,   1.559195138637400D-04,   1.168750239152212D-04, &
   1.938267229638313D-04,  -1.186679662826831D-03,  -5.048512553458395D-04, &
  -3.275635745205413D-04,   4.579113057140843D-04,   3.617548132284311D-04, &
  -1.850682041952321D-04,  -3.311379173759345D-04,  -1.050927659202762D-04, &
   8.949419182058145D-05,   4.037721067659937D-03,   9.448523999883052D-04, &
   5.550150142412170D-04,  -5.072009075954634D-04,  -9.613979831407007D-04, &
  -9.073093983860273D-05,   6.145188266550103D-04,   5.547960850394152D-04, &
   8.808951098541732D-07,  -2.251503933675107D-04,  -2.165928773503510D-04, &
   2.556415442142882D-03,   1.729162155168727D-03,   1.326691233628061D-03, &
   1.890162276363072D-03,   4.960719348378188D-04,  -8.152365743546736D-04, &
  -1.322250026845830D-03,  -6.138749069219924D-04,   3.443780240757821D-04, &
   4.870113880412934D-04,   2.505004006348694D-04,   8.922264299499174D-05, &
  -8.647373334520265D-04,  -4.087306263000727D-04,  -1.507920419139215D-03, &
   2.561898116946430D-05,   1.720579452074610D-03,   2.457844920947942D-03, &
   1.841923293008852D-03,  -1.769115345859743D-05,  -1.062029701129523D-03, &
  -8.484795855795098D-04,  -2.666753336848060D-04,  -2.565528287958425D-06, &
   8.458265671007676D-05,  -8.312381234447441D-03,  -4.011936904062336D-03, &
  -4.444027321041325D-03,  -4.958413728649800D-03,  -4.480773161449231D-03, &
  -3.278488466224234D-03,  -4.137773926973170D-04,   1.876739424737398D-03, &
   2.177764038815587D-03,   1.167763326391735D-03,   1.192669038508064D-04, &
  -1.818317897816651D-04,  -1.292437077291542D-04,  -7.915321187340522D-05, &
  -7.536259664863986D-03,  -2.321601605521783D-04,   1.504811477656250D-03, &
   1.565618241905446D-03,   9.227761539393095D-04,  -1.748101527372432D-03, &
  -4.454020940772231D-03,  -4.862000463210638D-03,  -3.255502744518314D-03, &
  -9.333945296135076D-04,   3.731898386826346D-04,   4.802027070012168D-04, &
   2.298027504003928D-04,   8.952219547821359D-05,   3.594751353421653D-05, &
   1.282164568002387D-02,   7.632472956814789D-03,   8.899429427072076D-03, &
   8.775711785171159D-03,   9.845654399657237D-03,   1.076533300100477D-02, &
   9.617417103590299D-03,   6.448973273453857D-03,   2.248710582801695D-03, &
  -6.864605164906506D-04,  -1.390813144190908D-03,  -9.371117533176873D-04, &
  -3.452897478157317D-04,  -8.495072595916559D-05,  -3.654387954575934D-05, &
  -4.226364980068049D-05,   9.599588157308849D-03,  -2.466281088061166D-03, &
  -1.735078340822496D-03,  -5.126704453517550D-03,  -7.080606712498771D-03, &
  -6.727637808893704D-03,  -4.149652282246001D-03,   1.851764476670627D-04, &
   3.694159611092552D-03,   4.355528877968203D-03,   3.103216357703008D-03, &
   1.463149839845819D-03,   4.277199091813396D-04,   9.742827904234221D-05, &
   4.318729326958497D-05,   3.464070273548485D-05,   2.375179299594668D-05, &
  -1.015337488528637D-02,  -6.409438908279204D-03,  -7.389739411970421D-03, &
  -8.220984909973413D-03,  -9.479063665567793D-03,  -1.077627432478413D-02, &
  -1.279910286278397D-02,  -1.391562281043946D-02,  -1.223412374063358D-02, &
  -8.413125523782242D-03,  -4.322516394918101D-03,  -1.559002001436314D-03, &
  -4.311582314150013D-04,  -1.277799709059586D-04,  -9.601010554906711D-05, &
  -9.690351927348278D-05,  -8.622144939806916D-05,  -8.595724253192151D-05, &
  -3.508965075278416D-03,   4.419789446182302D-03,   9.041494799728905D-04, &
   6.330427439382063D-03,   8.910576524251207D-03,   1.249953393386382D-02, &
   1.459282196290695D-02,   1.361593823809747D-02,   9.810413845054825D-03, &
   4.850550551577381D-03,   1.142578403500097D-03,  -2.584758204740260D-04, &
  -3.372044080482796D-04,  -7.119242650302614D-05,   1.410511222601627D-04, &
   1.874682072231129D-04,   1.785018429640095D-04,   1.653481438814329D-04, &
   2.159642619828883D-04,   3.169297324278273D-03,   1.292052873700856D-03, &
   1.700884175431602D-03,   3.865749045438817D-03,   3.206327409185230D-03, &
   3.606200419582603D-03,   4.708601992699174D-03,   6.491454708249487D-03, &
   8.647431978617377D-03,   9.478619900516606D-03,   7.949633888778385D-03, &
   5.198259001541404D-03,   2.697716238155833D-03,   1.082340696205429D-03, &
   3.272934966342009D-04,   3.594554073176340D-06,  -1.295052353066369D-04, &
  -1.693758191707603D-04,  -1.906112936510604D-04,  -2.523642358192469D-04, &
   0.000000000000000D+00,  -1.935065036032034D-03,   9.683882110346116D-04, &
  -3.278716925102079D-03,  -3.990220621817287D-03,  -7.568826519317635D-03, &
  -1.099768122379779D-02,  -1.453818091170680D-02,  -1.620061001192017D-02, &
  -1.478374971519910D-02,  -1.113210612470909D-02,  -7.044795265987552D-03, &
  -3.959607601020092D-03,  -2.230304381715704D-03,  -1.289230323274361D-03, &
  -7.167996599123144D-04,  -3.808566022464855D-04,  -1.565406779670827D-04, &
  -1.463403464007418D-05,   9.741101158659161D-05,   1.525493620321422D-04 /)
lem(:, 42) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.558791572552085D-04,  -8.049565067622842D-04,   1.928186170854954D-03, &
  -3.052498847696861D-03,   2.707506803805436D-03,   1.243351963088134D-03, &
  -8.138136684189525D-03,   7.776525600195168D-03,   1.395634363957804D-02, &
  -2.566329663596913D-02,  -5.065358078694638D-02,  -3.056634007584821D-02, &
  -3.502017449904296D-03,   1.222023228331503D-02,   1.655097673014105D-02, &
   1.463812150481302D-02,   1.427360786396908D-02,   1.347813741870260D-02, &
   1.120311540766364D-02,   9.920049688718484D-03,   7.183679978111252D-03, &
   1.036726152421652D-03,   1.558791572553680D-04,   0.000000000000000D+00, &
  -2.052236144708565D-04,   0.000000000000000D+00,   4.411940769718729D-04, &
   0.000000000000000D+00,  -1.144938432196222D-03,   0.000000000000000D+00, &
   3.647495840097358D-03,   0.000000000000000D+00,  -1.542208233586503D-02, &
   0.000000000000000D+00,   1.029866776864045D-01,   0.000000000000000D+00, &
  -1.965055173895255D+00,  -2.769364170995378D-02,   3.886886760024336D+00, &
  -2.470088952785544D-02,  -1.968754880441467D+00,   0.000000000000000D+00, &
   1.034117927024845D-01,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.012391266908333D-04,   5.619713580858680D-04,   9.076782563970733D-04, &
   7.048241712449214D-04,  -7.991659988597428D-04,  -3.255262290729701D-03, &
  -2.726743610820203D-03,   5.865381823713393D-03,   9.618097901076126D-03, &
  -2.080388009322525D-02,   1.454770447616140D-02,  -4.094661896473640D-03, &
  -2.172244824933721D-03,   3.951239439461449D-03,  -3.632503997783307D-03, &
   3.587170867229069D-03,  -3.304753164242944D-03,   2.547830730926940D-03, &
  -1.591117992928992D-03,  -8.200419210562890D-04,   5.691216025127758D-03, &
  -1.590934986710848D-05,   1.501931630812646D-05,  -1.590934986713603D-05, &
   7.787793641179761D-07,   1.791192537488112D-05,  -1.373989306707092D-05, &
  -5.689134965200542D-05,  -8.232810420751161D-06,   2.125460822998726D-05, &
  -5.302273787558203D-06,   1.618445114900218D-04,  -3.918170463143883D-05, &
  -2.486025554570639D-05,   1.860726409284712D-05,   1.253948558940484D-05, &
  -2.191209718165745D-04,   1.382049400824766D-04,   3.788866636656613D-07, &
  -4.081647999494714D-05,  -4.582914277502375D-07,   3.257204817883314D-05, &
  -6.319705069719759D-05,  -2.124651701581500D-04,   7.455937159965300D-05, &
   5.893885473844987D-05,  -3.263953979705772D-05,  -3.988010672977826D-05, &
   2.497401393247647D-05,   8.516067379767322D-04,   2.242355527754029D-05, &
  -1.709498504391083D-04,  -4.766528156286709D-05,   1.014750997993422D-04, &
   3.277148049245638D-05,  -6.665656226586713D-05,  -4.290389932988044D-05, &
  -9.203900607446720D-04,   5.612095047398327D-04,   1.713376841184420D-04, &
  -3.803851498614810D-05,  -2.026872976560370D-04,   3.988302423488603D-05, &
   1.374400431801672D-04,   5.360526639607387D-06,  -9.817474942240450D-05, &
  -1.957978469061967D-03,  -7.054616271289631D-04,   1.106417966595217D-05, &
   2.794880703424880D-04,   2.311660375399565D-04,  -2.360809916847998D-04, &
  -1.901217247084334D-04,   9.745414563115693D-05,   1.623903636715822D-04, &
   6.861962555038948D-05,   1.472465162627354D-03,  -6.636789527987508D-04, &
  -3.675048371572445D-04,  -6.242383126423259D-04,   1.021951956321061D-04, &
   5.067763268043326D-04,   1.239889831190305D-04,  -3.186756099841512D-04, &
  -2.692159299277247D-04,  -8.773560441603005D-06,   1.406844681968747D-04, &
   3.493699588203981D-03,   1.315255038540746D-03,   9.573058364070630D-04, &
   1.917037175419789D-04,  -8.888062826897664D-04,  -5.687865850477135D-04, &
   2.248465929617633D-04,   7.077527074233171D-04,   3.284709076328541D-04, &
  -1.563948322762803D-04,  -2.295441931929588D-04,  -1.552260417949935D-04, &
   1.184594933388224D-03,   9.919503482339255D-04,   5.352002245386151D-04, &
   1.645917635057249D-03,   1.155936513394947D-03,  -1.459523610101826D-05, &
  -1.097269976499565D-03,  -1.138213501685797D-03,  -1.363854612475654D-04, &
   4.904109355050256D-04,   3.987511607027992D-04,   1.540577641021299D-04, &
   2.420425380258019D-05,  -1.188314002188339D-03,  -7.557656392424470D-04, &
  -1.906495590383550D-03,  -1.150227598439739D-03,   4.463480683409455D-04, &
   1.809254922973492D-03,   2.344908774813516D-03,   1.009974278325743D-03, &
  -5.707370727845038D-04,  -1.052714552635101D-03,  -5.930175411448693D-04, &
  -9.530057148796162D-05,   4.497656286444877D-05,   6.309814139932507D-05, &
  -7.220140746224549D-03,  -3.270184801457689D-03,  -3.188116494055502D-03, &
  -3.766223246342780D-03,  -3.969819401561585D-03,  -3.998339717743166D-03, &
  -2.107740978292659D-03,   6.014113877501609D-04,   2.037911572575154D-03, &
   1.783122512452950D-03,   6.257511026780650D-04,  -9.515417668473225D-05, &
  -1.800921220752238D-04,  -1.028260851983314D-04,  -7.851547422307702D-05, &
  -6.191706121813146D-03,   2.794321699926312D-04,   2.113799427806244D-03, &
   2.654099695051732D-03,   2.748190894708736D-03,   8.631241581208790D-04, &
  -2.288716735134139D-03,  -4.270634256177853D-03,  -4.097858684234835D-03, &
  -2.109910056300182D-03,  -1.989739166686129D-04,   4.928231540085030D-04, &
   3.916082357513521D-04,   1.554715260369384D-04,   5.766384267730075D-05, &
   1.916467611333390D-05,   1.203013224196182D-02,   6.757929924458957D-03, &
   7.481931358914693D-03,   6.931178848249569D-03,   7.558799317800791D-03, &
   8.925275117486399D-03,   9.385787834004839D-03,   8.022333430292627D-03, &
   4.524097918028976D-03,   8.063511983519532D-04,  -1.051628279576232D-03, &
  -1.208639821426814D-03,  -6.383538457838257D-04,  -1.839386569297006D-04, &
  -3.868359903300872D-05,  -1.864401757543383D-05,  -3.899997064059823D-06, &
   8.334978107915381D-03,  -2.749898822471829D-03,  -2.274509130724086D-03, &
  -5.600896580713687D-03,  -7.985114497759661D-03,  -8.742204683810468D-03, &
  -7.495472645400205D-03,  -3.706393951808637D-03,   9.130583895070671D-04, &
   3.504739516693461D-03,   3.538520682769804D-03,   2.167796599134704D-03, &
   8.300648647703141D-04,   2.008212202814712D-04,   2.606084259074048D-05, &
   1.591593922322769D-05,   2.145941395845214D-05,   3.890161407911257D-05, &
  -9.590683274529653D-03,  -5.677604561119697D-03,  -6.389790605427384D-03, &
  -6.581305850204610D-03,  -6.949955459282444D-03,  -7.380553405718518D-03, &
  -9.079085016547833D-03,  -1.126696447166285D-02,  -1.177662604593453D-02, &
  -9.741410910322724D-03,  -6.061523819545972D-03,  -2.677912006478387D-03, &
  -8.394591440223849D-04,  -1.743464784748564D-04,  -4.290615722202402D-05, &
  -6.483030244395969D-05,  -6.682374327437733D-05,  -5.178012379071706D-05, &
  -4.841936928650651D-05,  -3.098998831776646D-03,   4.292909388269948D-03, &
   1.047734097660558D-03,   6.162941701406186D-03,   8.665614436957004D-03, &
   1.234458853520319D-02,   1.526650607467589D-02,   1.588714696378767D-02, &
   1.349645731329118D-02,   8.651719224486317D-03,   3.676847166577759D-03, &
   7.525694646553582D-04,  -1.902799253574262D-04,  -1.834325372907937D-04, &
   4.200183475246545D-05,   1.378119286883117D-04,   1.463629600138556D-04, &
   1.248532260501715D-04,   1.108129384919603D-04,   1.217568984791569D-04, &
   2.969072307435254D-03,   1.069879464703033D-03,   1.430659678476472D-03, &
   3.128071501528731D-03,   2.024658353270215D-03,   1.540519102552765D-03, &
   1.436345864909927D-03,   2.203338841924260D-03,   4.340109191015351D-03, &
   6.561864426595785D-03,   6.995791913346682D-03,   5.511158456041857D-03, &
   3.294143397783477D-03,   1.468279145086674D-03,   4.849307660302905D-04, &
   5.987493322247768D-05,  -1.092526579413204D-04,  -1.480669140435918D-04, &
  -1.563912575892870D-04,  -1.650107033680717D-04,  -2.263979354895589D-04 /)
lem(:, 43) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.048974227103013D-05,  -4.231602366329061D-04, &
   1.536613662743511D-03,  -3.579641041667425D-03,   5.442939695039873D-03, &
  -3.563147063055984D-03,  -5.695894721821834D-03,   1.471258571802781D-02, &
   2.639241778986934D-03,  -3.326560981169679D-02,  -4.321180391645162D-02, &
  -2.076605731291063D-02,   2.810197727086810D-03,   1.122974931910804D-02, &
   1.367522367894640D-02,   1.347813741870264D-02,   1.144239437696908D-02, &
   1.113149648458168D-02,   1.049521525625603D-02,   8.498279341487586D-03, &
   6.694810799231604D-03,   0.000000000000000D+00,   1.178093986330211D-04, &
   0.000000000000000D+00,  -2.059311987442569D-04,   0.000000000000000D+00, &
   4.428397574274393D-04,   0.000000000000000D+00,  -1.145314641774179D-03, &
   0.000000000000000D+00,   3.649172637440202D-03,   0.000000000000000D+00, &
  -1.543948877550609D-02,   0.000000000000000D+00,   1.031244411259704D-01, &
   0.000000000000000D+00,  -1.967064952565581D+00,  -2.470088952785544D-02, &
   3.876002874828102D+00,  -2.216834216811275D-02,  -1.970192362278672D+00, &
   0.000000000000000D+00,   1.034811625612694D-01,   0.000000000000000D+00, &
   1.682991409035108D-05,   1.493506717525642D-04,   5.484261550112639D-04, &
   1.275262872858414D-03,   1.903382984590620D-03,   1.104591881463065D-03, &
  -2.393761218723612D-03,  -5.484419495490818D-03,   1.739822939237639D-03, &
   1.307606198308586D-02,  -1.829060079582991D-02,   1.030921142628755D-02, &
  -1.393006340675440D-03,  -2.258996657527253D-03,   3.352437879492321D-03, &
  -3.344289727901832D-03,   2.851648585828668D-03,  -2.777893418037196D-03, &
   2.482038934590601D-03,  -1.599253772401667D-03,   1.029111282702509D-04, &
   1.330521533762330D-06,  -1.197469380391648D-05,  -1.330521533771061D-06, &
   3.234190805160946D-05,  -9.108955115803460D-06,  -6.575846811118765D-06, &
  -5.035512266263836D-05,   3.029495184581591D-05,  -2.226064873807666D-06, &
  -1.297258495424837D-05,   2.519259142933970D-05,  -5.288590488071635D-05, &
   2.251276073927657D-05,   9.463057067840016D-06,  -1.543139268511229D-05, &
   1.102500632508360D-04,   4.061060911650704D-05,  -5.119734494067043D-05, &
   2.860549725708180D-06,   2.236348860512512D-05,  -4.458973810096925D-06, &
  -3.459599332243430D-04,   6.882997599839539D-05,   6.445031279478527D-05, &
  -3.057937344549070D-05,  -3.100525817430407D-05,   2.272403202773719D-05, &
   2.517657721776076D-05,   3.385508106220747D-04,  -2.745340445831147D-04, &
  -1.514443411735636D-05,   7.433618083414940D-05,   2.642072042037715D-05, &
  -6.161655779738661D-05,  -1.064909283794887D-05,   4.901426793648414D-05, &
   5.460671818415555D-04,   3.104173300145545D-04,  -1.157723717630237D-04, &
  -1.223790711733099D-04,   2.551846721515613D-05,   1.117282209897194D-04, &
  -3.575241110859205D-05,  -6.978908412795645D-05,   2.416534781397690D-06, &
  -1.540860905030437D-03,   2.678745864614356D-04,   2.337541423728189D-04, &
   1.325867060048724D-04,  -1.741178255502457D-04,  -1.206926264383857D-04, &
   1.347964371497104D-04,   9.141551640872361D-05,  -5.235235477345186D-05, &
  -1.003393487404435D-04,  -1.058181321145319D-03,  -9.200190878118139D-04, &
  -2.155438877778764D-04,   4.862152355133082D-05,   3.867004453308069D-04, &
  -2.153924876579339D-05,  -2.772368469759044D-04,  -6.858984589929746D-05, &
   1.711112437964217D-04,   1.320695862805791D-04,   8.708266579773379D-06, &
   2.216583080238104D-03,  -4.447012357096799D-05,  -2.727470732870178D-05, &
  -6.237345359551415D-04,  -3.243006928979785D-04,   3.774336741858517D-04, &
   3.875309141616558D-04,  -8.727999164986726D-05,  -3.795071776936738D-04, &
  -1.550160852864217D-04,   6.644900884441264D-05,   1.412584045183283D-04, &
   2.622633396447589D-03,   1.301286612815384D-03,   1.012452516330499D-03, &
   8.022538349660270D-04,  -4.465345293212789D-04,  -7.586161263687353D-04, &
  -2.840999908552629D-04,   5.215957545994443D-04,   6.333909442726046D-04, &
   8.359141839983302D-05,  -2.324087477538747D-04,  -1.986832507234183D-04, &
  -1.158686995786310D-04,   2.330459640301672D-04,   3.535239578571215D-04, &
  -2.365897569614289D-04,   9.597439563817990D-04,   1.298500200709622D-03, &
   7.104550219254686D-04,  -4.504040007833957D-04,  -1.275606667619859D-03, &
  -7.244968233331075D-04,   2.050256532868987D-04,   5.190993614301110D-04, &
   2.895446182379963D-04,   7.934530948365139D-05,  -2.292184907545435D-05, &
  -1.198544032400042D-03,  -8.127439615122200D-04,  -1.830726352526698D-03, &
  -1.734513224997527D-03,  -6.512783056825901D-04,   7.148137999688679D-04, &
   2.142255221422312D-03,   1.824846522756526D-03,   2.765815802783389D-04, &
  -8.742876407291445D-04,  -9.216964710834839D-04,  -3.378402910656486D-04, &
   1.077335104151933D-05,   6.411876673604354D-05,   5.956353956039266D-05, &
  -6.285073432119702D-03,  -2.643887941648231D-03,  -2.130165616565434D-03, &
  -2.475530534627553D-03,  -2.869039894136241D-03,  -3.761987138059842D-03, &
  -3.232188437926183D-03,  -9.596957922306245D-04,   1.230105772288012D-03, &
   2.028584400124242D-03,   1.259931815668557D-03,   2.233490350556902D-04, &
  -1.882329354546658D-04,  -1.579741137602114D-04,  -7.799766599276897D-05, &
  -3.766588741845866D-05,  -5.166498729007772D-03,   5.742737275129377D-04, &
   2.311152076670810D-03,   3.059442870837208D-03,   3.678849148422716D-03, &
   2.830237103770412D-03,   1.434396311866385D-04,  -2.700146111399289D-03, &
  -4.117877947786312D-03,  -3.172560902480728D-03,  -1.123326251335287D-03, &
   2.076966966511542D-04,   5.079160379908019D-04,   2.848956754358926D-04, &
   9.280813544699953D-05,   3.562473495757053D-05,   3.160505072114414D-05, &
   1.126654447273051D-02,   5.992881965116217D-03,   6.273079174315266D-03, &
   5.319974233718391D-03,   5.332698856113392D-03,   6.500076943536717D-03, &
   7.850959634952807D-03,   8.292908931716683D-03,   6.366217011028167D-03, &
   2.743079441040541D-03,  -1.107435210681132D-04,  -1.192624770343385D-03, &
  -9.553645451304908D-04,  -3.899892863016621D-04,  -9.473079560001887D-05, &
  -2.020789021627347D-05,  -1.013401826919156D-05,   2.565174908470707D-06, &
   7.285142882974773D-03,  -2.906556492344292D-03,  -2.555588182601495D-03, &
  -5.651786608465137D-03,  -8.087745487576722D-03,  -9.508216911981332D-03, &
  -9.557651620277829D-03,  -7.090115270113152D-03,  -2.504435499471255D-03, &
   1.563904825153305D-03,   3.292091009041397D-03,   2.767434019513887D-03, &
   1.396175359142649D-03,   4.525176543753768D-04,   6.996483820688096D-05, &
  -5.806150315202604D-06,   3.085376087132875D-06,   1.041223384969276D-05, &
   4.457977816983598D-06,  -9.058411611988823D-03,  -5.043374202929164D-03, &
  -5.535796497980133D-03,  -5.232509360689269D-03,  -4.858931075800628D-03, &
  -4.372737673453914D-03,  -5.227829989902296D-03,  -7.542335466371715D-03, &
  -9.664554887353584D-03,  -9.832036309854644D-03,  -7.501624250447482D-03, &
  -4.090713799611330D-03,  -1.576771908953259D-03,  -3.854724143546570D-04, &
  -2.766607922145329D-05,  -1.644128391639253D-05,  -3.552980732570719D-05, &
  -3.928766826263646D-05,  -3.447924244462805D-05,  -5.175834027411042D-05, &
  -2.751646711814439D-03,   4.145872643030809D-03,   1.113356575773599D-03, &
   5.872921425368931D-03,   8.134848726640359D-03,   1.155448317898822D-02, &
   1.472743921232727D-02,   1.651504563384563D-02,   1.584894791125475D-02, &
   1.214921996516293D-02,   6.820421345634379D-03,   2.518044751714024D-03, &
   3.654538472259654D-04,  -1.900087974063395D-04,  -7.160424447380839D-05, &
   6.405150447347675D-05,   1.107617028746805D-04,   1.135999201336066D-04, &
   9.274662874304510D-05,   7.269121020788289D-05,   7.880298467484695D-05 /)
lem(:, 44) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.034263192715200D-05,   4.665273452124948D-04,  -1.128921982095002D-03, &
   1.871868416990101D-03,  -2.010716279485517D-03,   3.945107522582499D-04, &
   3.436453788969876D-03,  -6.318541926995001D-03,   2.774925049804369D-04, &
   1.502098766956133D-02,  -7.476832556716723D-03,  -3.775753628671186D-02, &
  -3.361275426875310D-02,  -1.227742531977547D-02,   4.761943041090671D-03, &
   1.150444308305112D-02,   1.120311540766375D-02,   1.113149648458182D-02, &
   1.061356039595211D-02,   9.072572225843935D-03,   8.792069966295748D-03, &
   8.173831096372210D-03,  -9.034263192760775D-05,   0.000000000000000D+00, &
   1.015838116309831D-04,   0.000000000000000D+00,  -1.926480304400815D-04, &
   0.000000000000000D+00,   4.326006019819445D-04,   0.000000000000000D+00, &
  -1.137408961705152D-03,   0.000000000000000D+00,   3.644798835626137D-03, &
   0.000000000000000D+00,  -1.544972407637459D-02,   0.000000000000000D+00, &
   1.032361316490857D-01,   0.000000000000000D+00,  -1.968754880441467D+00, &
  -2.216834216811275D-02,   3.866748715719011D+00,  -2.000625293121497D-02, &
  -1.971423167883838D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.166318363034772D-04,  -3.296682188797968D-04,  -5.626689111331123D-04, &
  -5.664638297375536D-04,   6.952681119004634D-05,   1.473153746702198D-03, &
   2.393724667244259D-03,  -1.568911142916656D-04,  -5.991433024999983D-03, &
  -2.468643838910531D-03,   1.505139790806519D-02,  -1.476196459359903D-02, &
   6.785267651711477D-03,  -1.186461381111319D-04,  -2.552567912134924D-03, &
   2.746492518213866D-03,  -2.795167529037633D-03,   2.640094985499741D-03, &
  -2.251598432978756D-03,   2.142825449789002D-03,  -1.743953669188067D-03, &
   9.220556260098442D-06,  -8.810753759690933D-06,   9.220556260152449D-06, &
   1.536759376707836D-06,  -1.014261188615765D-05,   8.221662665303013D-06, &
   2.456450757509500D-05,   5.652364665975912D-06,  -1.170922743613434D-05, &
   4.312474168010029D-06,  -7.679856474791898D-05,   1.644605330580974D-05, &
   1.316293512264182D-05,  -1.075716408265138D-05,  -4.290877697489836D-06, &
   1.250908906078992D-04,  -6.528706040497243D-05,  -1.062841585513603D-06, &
   2.068105770121383D-05,  -2.818887065913755D-06,  -1.578433331545730D-05, &
  -5.650909745859704D-05,   1.173211211719215D-04,  -3.636984791612133D-05, &
  -2.655406950615686D-05,   1.920845888467720D-05,   1.499393004935660D-05, &
  -1.925978925183420D-05,  -2.712189341591949D-04,  -7.676638015133010D-05, &
   9.015466256073650D-05,   1.628452607436014D-05,  -4.748071780233323D-05, &
  -4.944310429773221D-06,   3.385644757582132D-05,   3.647814371036309D-06, &
   6.467629155189375D-04,  -1.747577302069741D-04,  -1.049469015529278D-04, &
   2.312806328735412D-05,   8.377873691048350D-05,  -3.243384810679192D-05, &
  -5.347463318385490D-05,   2.239480111340659D-05,   4.830434255641510D-05, &
  -5.002067333797776D-05,   4.760903931630427D-04,   1.058184542090169D-05, &
  -1.046221071693325D-04,  -9.599159266162939D-05,   1.155935413060196D-04, &
   5.170914607491633D-05,  -7.719805196903337D-05,  -4.696455422220720D-05, &
   3.732773343498884D-05,  -1.702726574794919D-03,  -1.427166219619534D-04, &
   1.575357303291606D-04,   2.299548428189346D-04,  -9.449096517903354D-06, &
  -2.215271397867857D-04,   1.529192431522812D-05,   1.574457647176529D-04, &
   2.549502367932251D-05,  -9.000075899963603D-05,  -8.868352826178134D-05, &
  -4.213591561296909D-05,  -8.189524519070593D-04,  -3.032920961131350D-04, &
  -2.461821730523362D-04,   3.250418765761363D-04,   2.239841720017555D-04, &
  -1.823910004365530D-04,  -2.382523157516181D-04,   6.180603171835617D-05, &
   1.965837535603552D-04,   8.125675322712712D-05,  -4.680982023846941D-05, &
   2.413764310515057D-03,   4.583545271901873D-04,   3.461177564460488D-04, &
  -3.252235230300777D-04,  -5.969675941567661D-04,   5.014048429749244D-05, &
   4.336523772397966D-04,   2.356656711521393D-04,  -2.882536384478936D-04, &
  -3.344738236584880D-04,  -3.664954577834738D-05,   1.117486014974397D-04, &
   1.260470600920772D-04,   1.700106390675727D-03,   1.030406168719006D-03, &
   7.543029789497064D-04,   1.091228429556293D-03,   1.341775631073076D-04, &
  -5.789920453074650D-04,  -6.484246337983626D-04,   6.724860117385904D-05, &
   6.917693170473909D-04,   4.238551626450236D-04,  -1.018557091196579D-04, &
  -2.516900533158896D-04,  -1.536123896922105D-04,  -6.191564317000157D-05, &
  -3.358410299366815D-04,  -9.233165567845929D-05,  -7.919868173450481D-04, &
   1.400156775074338D-04,   9.504934686976981D-04,   1.058784519498547D-03, &
   3.408972982809892D-04,  -9.333305453176766D-04,  -1.126483274558088D-03, &
  -3.105919035000299D-04,   4.232213292653185D-04,   4.636694114413361D-04, &
   1.760057014038218D-04,   2.369128377121778D-05,  -2.542645128577078D-05, &
  -1.052207323993668D-03,  -7.069290153206947D-04,  -1.469110675394510D-03, &
  -1.763687110778506D-03,  -1.274009419965878D-03,  -3.973951817192666D-04, &
   1.323081343007381D-03,   2.095768053705997D-03,   1.176519399477718D-03, &
  -2.757906361284286D-04,  -1.010072641211755D-03,  -6.805521188327514D-04, &
  -1.412000112566837D-04,   6.106758125021078D-05,   6.399062732002712D-05, &
   6.258299457068256D-05,  -5.511669175338237D-03,  -2.154322236638054D-03, &
  -1.348142982650811D-03,  -1.370886009257933D-03,  -1.602424778155606D-03, &
  -2.796661265734213D-03,  -3.480196872199572D-03,  -2.296692763606377D-03, &
  -7.240045750980933D-05,   1.686768608038605D-03,   1.751467450035326D-03, &
   7.513649469643494D-04,  -2.315693979349490D-05,  -2.161119494951262D-04, &
  -1.270578446748493D-04,  -5.706275578948079D-05,  -2.479003707540056D-05, &
  -4.375180263200012D-03,   7.306778010651058D-04,   2.264265007241293D-03, &
   3.003375841617157D-03,   3.838143083603974D-03,   3.879311522411955D-03, &
   2.200770771948900D-03,  -5.730909959208939D-04,  -3.162803787065163D-03, &
  -3.704805376797519D-03,  -2.177392847669079D-03,  -4.337760615674296D-04, &
   4.321168995811331D-04,   4.399219770023868D-04,   1.888286043416062D-04, &
   5.369998975186991D-05,   2.172364002005389D-05,   4.882660616947334D-06, &
   1.054901935414488D-02,   5.335297442613277D-03,   5.281371280815002D-03, &
   4.025125064558079D-03,   3.455347930809169D-03,   4.058922837480584D-03, &
   5.530273304107110D-03,   7.215363071888544D-03,   7.201927184204228D-03, &
   4.643466707286413D-03,   1.384861246287996D-03,  -7.023287969285625D-04, &
  -1.140096224338253D-03,  -6.749126510488947D-04,  -2.308086927107618D-04, &
  -3.863179085343737D-05,  -4.532775455097334D-06,  -5.669337551681890D-06, &
  -1.660323057801058D-05,   6.407621853397094D-03,  -2.976310857129620D-03, &
  -2.669486351791019D-03,  -5.450044271771422D-03,  -7.676962820105454D-03, &
  -9.300681308796294D-03,  -1.025343022114837D-02,  -9.333839248384211D-03, &
  -5.805447009086025D-03,  -1.224920685113236D-03,   2.106524753669260D-03, &
   2.955509921155465D-03,   2.007721675227286D-03,   8.668481991151156D-04, &
   2.146959447957561D-04,  -2.727929596686739D-06,  -1.134733391093175D-05, &
   4.776857171453648D-06,   5.340797738320515D-06,  -1.972138598191549D-06, &
  -8.559490324613965D-03,  -4.494901201427705D-03,  -4.814385165702909D-03, &
  -4.151244818022616D-03,  -3.226507285941476D-03,  -1.963507127428547D-03, &
  -1.799917090191148D-03,  -3.487258920712217D-03,  -6.278109389345592D-03, &
  -8.396984277343353D-03,  -8.085500129180934D-03,  -5.493398179029506D-03, &
  -2.641671072324942D-03,  -8.436757540301640D-04,  -1.076264580997729D-04, &
   1.976576926866889D-05,  -4.059025176614194D-07,  -1.891535984050726D-05, &
  -2.003718464302683D-05,  -1.910598438380267D-05,  -1.483766951528733D-05 /)
lem(:, 45) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.704898998739813D-05,   2.267006423275972D-04, &
  -8.309485944516132D-04,   2.002251953027421D-03,  -3.362954801762998D-03, &
   3.344007205000425D-03,   4.593136298366651D-04,  -7.464055337637868D-03, &
   7.741015457038304D-03,   1.012520797748868D-02,  -1.681813664434309D-02, &
  -3.628408620433350D-02,  -2.593437800455405D-02,  -5.132496600951887D-03, &
   5.861339006417085D-03,   9.920049688718706D-03,   1.049521525625626D-02, &
   9.072572225844020D-03,   8.884712106858815D-03,   8.489488519303427D-03, &
   7.304715677828397D-03,   0.000000000000000D+00,  -6.311430997136905D-05, &
   0.000000000000000D+00,   1.009607194818219D-04,   0.000000000000000D+00, &
  -1.934187247010361D-04,   0.000000000000000D+00,   4.325772843772778D-04, &
   0.000000000000000D+00,  -1.136683098083811D-03,   0.000000000000000D+00, &
   3.645133841846771D-03,   0.000000000000000D+00,  -1.546094251753998D-02, &
   0.000000000000000D+00,   1.033319358844071D-01,   0.000000000000000D+00, &
  -1.970192362278672D+00,  -2.000625293121497D-02,   3.858816579339791D+00, &
  -1.814573364683924D-02,  -1.972486703396304D+00,   0.000000000000000D+00, &
  -9.016329995719977D-06,  -8.001199141002478D-05,  -2.967407690598735D-04, &
  -7.150642651883736D-04,  -1.187304841019022D-03,  -1.124646728993183D-03, &
   3.462536577833881D-04,   2.909125701543134D-03,   2.704374469823016D-03, &
  -4.369741229300784D-03,  -6.340254154166453D-03,   1.487976143445232D-02, &
  -1.160375735590347D-02,   3.871201784646026D-03,   6.230145875085923D-04, &
  -2.316097528183894D-03,   2.588265375157775D-03,  -2.261908182880474D-03, &
   2.233185360699975D-03,  -2.107874182889425D-03,   1.785353227943824D-03, &
  -7.128034730831018D-07,   6.415231257598411D-06,   7.128034730672786D-07, &
  -1.742530336469492D-05,   4.901894653261747D-06,   3.531118743516019D-06, &
   2.856696995915070D-05,  -1.626288539324521D-05,   1.381742117029426D-06, &
   7.128034730687558D-06,  -2.208003657701742D-05,   2.934087087275129D-05, &
  -1.236469997535275D-05,  -4.504942051351060D-06,   9.298605661617140D-06, &
  -3.437062427240913D-05,  -2.824641928485600D-05,   2.794587290076841D-05, &
  -3.119705208927157D-06,  -1.139483836052565D-05,   5.642851754627806D-06, &
   1.554949143209001D-04,  -1.717515113994812D-05,  -3.711796046496832D-05, &
   1.822172827649552D-05,   1.280561189541768D-05,  -1.385052109484385D-05, &
  -7.895142018640850D-06,  -2.404914562692998D-04,   1.228982849643127D-04, &
   1.624356032867672D-05,  -3.913802129653783D-05,  -7.217107776687271D-06, &
   2.974822411705768D-05,  -2.297525869039285D-06,  -2.573420881949501D-05, &
  -1.425185168942083D-06,  -2.080449795969801D-04,   5.236153086633528D-05, &
   5.676828099376979D-05,  -1.687912995965377D-05,  -4.749299769431686D-05, &
   2.819479711025696D-05,   2.661095181915274D-05,  -2.201216504623398D-05, &
   6.828383116815419D-04,   4.626347606537292D-05,  -1.327218977993672D-04, &
  -5.484476332068431D-05,   7.520064658566661D-05,   4.549176277864180D-05, &
  -7.385800752899321D-05,  -1.733699376283536D-05,   4.467917050390626D-05, &
   2.790696277696554D-05,  -6.877104409150756D-04,   4.294797585342562D-04, &
   1.267152871314901D-04,   5.845474376848962D-07,  -1.645485601847315D-04, &
   1.996851868355012D-05,   1.256998922469676D-04,  -2.154601534567710D-05, &
  -8.634729062661104D-05,  -9.321111194195462D-06,   6.135929546471368D-05, &
  -1.407079538941112D-03,  -4.867390265357009D-04,  -5.002096865906474D-06, &
   1.788580361372430D-04,   1.880782810963399D-04,  -1.793659818347758D-04, &
  -1.363501838889181D-04,   1.113463572915713D-04,   1.329059235653127D-04, &
  -4.687594010697291D-05,  -1.020037313580932D-04,  -5.335102677483971D-05, &
   8.096880866683945D-04,  -4.943574073918135D-04,  -2.079039252495129D-04, &
  -4.382452576005520D-04,   7.450560902731750D-05,   3.458607711786212D-04, &
   3.642625809615716D-05,  -2.645950600432369D-04,  -1.388347845955901D-04, &
   1.734336712086787D-04,   1.679921203589836D-04,   2.348841350074246D-05, &
  -7.076788683819059D-05,   2.193400465601329D-03,   7.344635601438593D-04, &
   5.911639520397265D-04,   1.226760083783320D-04,  -5.839791596251179D-04, &
  -2.881582420402500D-04,   2.319525722835897D-04,   4.448277344005672D-04, &
  -2.917936266948315D-06,  -3.902765864458013D-04,  -2.223619684831807D-04, &
   5.504396562105302D-05,   1.256071978597534D-04,   1.108536259863615D-04, &
   8.998961077370099D-04,   6.551230504876782D-04,   3.222480061673972D-04, &
   9.971988230527019D-04,   5.809307087716143D-04,  -1.431537007594948D-04, &
  -6.824860541909514D-04,  -4.300330429641372D-04,   4.283436573343690D-04, &
   6.567418328001189D-04,   1.947901015513894D-04,  -2.233897761347847D-04, &
  -2.287252710999814D-04,  -1.066989791206387D-04,  -1.950969753816006D-05, &
  -6.108999919379978D-04,  -3.377683226678148D-04,  -1.059406903400070D-03, &
  -5.387287708399101D-04,   3.360652616394702D-04,   9.245685249378081D-04, &
   9.323211162652873D-04,  -2.448078730749058D-04,  -1.126986187085008D-03, &
  -8.305659734699839D-04,   4.956156187506767D-05,   5.257232449735436D-04, &
   3.500759245641203D-04,   8.442565545171179D-05,  -1.025570895736521D-05, &
  -3.644233190380972D-05,  -8.499523443906553D-04,  -5.366039376297156D-04, &
  -1.001890973928291D-03,  -1.417159833331161D-03,  -1.359103283397683D-03, &
  -1.149713928101644D-03,   2.288973056667601D-04,   1.709740303740499D-03, &
   1.767360040155996D-03,   5.732347633808513D-04,  -7.080690444965260D-04, &
  -9.392006425669384D-04,  -4.288620481385338D-04,  -1.140787772711052D-05, &
   8.239978609166544D-05,   5.837136771134320D-05,   3.543175363533983D-05, &
  -4.882226992860420D-03,  -1.789782466937189D-03,  -8.330203500415736D-04, &
  -5.818942423018426D-04,  -5.108148287710804D-04,  -1.517325994226143D-03, &
  -2.888545133346831D-03,  -2.970083015258812D-03,  -1.465590656764932D-03, &
   7.459318193934752D-04,   1.828346396323803D-03,   1.316022768776287D-03, &
   3.601134892962754D-04,  -1.712597394822283D-04,  -1.975312190991635D-04, &
  -8.954946098923839D-05,  -3.835654431233893D-05,  -3.011149921516229D-05, &
  -3.754824554489242D-03,   8.037309384465608D-04,   2.096571600306241D-03, &
   2.697344346033436D-03,   3.478053208496343D-03,   4.048301049093228D-03, &
   3.451240938055411D-03,   1.496779890000801D-03,  -1.431761816898876D-03, &
  -3.395660789193123D-03,  -3.005684841038438D-03,  -1.340255860116522D-03, &
   5.414659909055774D-05,   5.018214902940952D-04,   3.316275620395579D-04, &
   1.195477946817083D-04,   3.346857581045655D-05,   1.381769313259658D-05, &
   2.318470794541390D-07,   9.884598123959693D-03,   4.774344544202614D-03, &
   4.485062141333792D-03,   3.043937424423283D-03,   2.039049604346671D-03, &
   1.998408959599454D-03,   3.038554424784627D-03,   5.151785906717264D-03, &
   6.754593454705532D-03,   5.920439882057091D-03,   3.138578484505901D-03, &
   3.422715023539697D-04,  -1.004129281050168D-03,  -9.377822094127848D-04, &
  -4.557410072378375D-04,  -1.221830668625128D-04,  -8.290762553842003D-06, &
   2.615881909907928D-06,  -2.543433763852632D-06,   2.837073521014060D-06, &
   5.668977329566658D-03,  -2.987561756475967D-03,  -2.679962233128420D-03, &
  -5.119118099148611D-03,  -6.998454141848738D-03,  -8.474523327029112D-03, &
  -9.813571821961235D-03,  -1.017532113239372D-02,  -8.258282347732722D-03, &
  -4.275722873409720D-03,  -4.823059732861458D-06,   2.421240523310541D-03, &
   2.437930981184259D-03,   1.398056521599468D-03,   5.000269956941354D-04, &
   6.585507048529742D-05,  -2.394934499069879D-05,  -1.436820127552741D-05, &
   2.585887514586145D-07,   3.835638011863093D-06,   1.524180589357600D-05 /)
lem(:, 46) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.526802731974416D-05,  -2.854028658510246D-04,   6.956327631404442D-04, &
  -1.190198870581518D-03,   1.427486785941615D-03,  -7.779325666575627D-04, &
  -1.261945272399276D-03,   3.793363710785886D-03,  -3.166204991931302D-03, &
  -4.341644462214122D-03,   1.138624792650749D-02,   3.746779756632439D-03, &
  -2.426851604156430D-02,  -3.185526284975400D-02,  -1.812040179615686D-02, &
  -1.580971234949462D-03,   7.183679978111494D-03,   8.498279341487593D-03, &
   8.792069966295699D-03,   8.489488519303512D-03,   7.312886480402513D-03, &
   7.184168697105037D-03,   5.526802731991526D-05,   0.000000000000000D+00, &
  -5.453134373329915D-05,   0.000000000000000D+00,   9.325988411718000D-05, &
   0.000000000000000D+00,  -1.872024821236614D-04,   0.000000000000000D+00, &
   4.274150686431343D-04,   0.000000000000000D+00,  -1.132366216736717D-03, &
   0.000000000000000D+00,   3.642719732466058D-03,   0.000000000000000D+00, &
  -1.546838938601005D-02,   0.000000000000000D+00,   1.034117927024845D-01, &
   0.000000000000000D+00,  -1.971423167883838D+00,  -1.814573364683924D-02, &
   3.851964547237531D+00,  -1.653319610837655D-02,   0.000000000000000D+00, &
   7.135071646284959D-05,   2.034181473763018D-04,   3.602758295356620D-04, &
   4.172631580300124D-04,   1.482090616998441D-04,  -6.144098491487001D-04, &
  -1.499100994631133D-03,  -1.113198376663255D-03,   1.863137555046271D-03, &
   4.397752317607617D-03,  -1.814619254868875D-03,  -9.404499525380553D-03, &
   1.339796158958559D-02,  -8.514180082936662D-03,   2.140316744364909D-03, &
   1.275408040639561D-03,  -2.017228405379874D-03,   2.195980699705980D-03, &
  -2.113284129077396D-03,   1.825058537446123D-03,  -1.802780200095800D-03, &
  -5.640769417664116D-06,   5.436578669526820D-06,  -5.640769417688615D-06, &
  -1.812192889858988D-06,   6.100198601020795D-06,  -5.143054469065110D-06, &
  -1.128518509871262D-05,  -3.851183360659434D-06,   6.890343871099949D-06, &
  -3.186834176498014D-06,   3.866862293115864D-05,  -7.252417822736484D-06, &
  -7.543025012508755D-06,   6.508580097324824D-06,   1.234715930227571D-06, &
  -7.051923824663529D-05,   3.273194521773126D-05,   1.226350560635762D-06, &
  -1.134313263797034D-05,   2.855662306844438D-06,   7.764304114137495D-06, &
   6.068902442121736D-05,  -6.519300111601038D-05,   1.853174085189352D-05, &
   1.340674941055932D-05,  -1.137724623650823D-05,  -5.774524174583468D-06, &
   1.217337628693256D-05,   6.938205284923403D-05,   6.480444968275613D-05, &
  -4.950032251080092D-05,  -6.309281844144173D-06,   2.414893513708375D-05, &
  -1.539025452605638D-06,  -1.689079612315554D-05,   5.485304054938965D-06, &
  -3.132362648171763D-04,   3.615452434454017D-05,   6.659210041536667D-05, &
  -1.515517922608683D-05,  -3.823890728817821D-05,   2.078415773086215D-05, &
   2.116092287278936D-05,  -1.857756325070992D-05,  -1.736766941169275D-05, &
   3.315725940011682D-04,  -2.335918598380220D-04,  -2.587708889351637D-05, &
   5.090549540347290D-05,   4.207023261262708D-05,  -5.815522586000178D-05, &
  -1.242086220337603D-05,   4.260156047118858D-05,   6.526947059811607D-06, &
  -3.474992530114360D-05,   4.118793567437627D-04,   2.710654059289038D-04, &
  -7.892450307533812D-05,  -9.571345419566798D-05,  -4.481818014136527D-06, &
   1.044595255281262D-04,  -2.798041528897786D-05,  -6.877604070795133D-05, &
   2.142888891068263D-05,   4.830887180412988D-05,   3.798568207681859D-06, &
  -1.127314149084275D-03,   1.981174103570525D-04,   1.595108467732898D-04, &
   1.228857230578417D-04,  -1.192171153543769D-04,  -1.086445240650178D-04, &
   1.094589768138047D-04,   7.674452036498189D-05,  -7.744317596779329D-05, &
  -6.667412185804891D-05,   2.825265455490961D-05,   7.364908038458286D-05, &
  -8.115910111135826D-04,  -6.393573304690622D-04,  -1.509496513893819D-04, &
  -1.858995264904180D-06,   2.903760764912581D-04,  -1.577828438696383D-05, &
  -2.036634866108824D-04,  -3.276464388076779D-05,   1.710648462513162D-04, &
   6.367465468209491D-05,  -1.013160016621941D-04,  -9.059117673796060D-05, &
  -1.150829433950628D-05,   1.345052953375449D-03,  -9.787798433155165D-05, &
   1.384650279583836D-05,  -4.212054569233223D-04,  -2.249756128604840D-04, &
   2.706170335256770D-04,   2.361595592316693D-04,  -1.132654568254238D-04, &
  -2.849512423884019D-04,   1.378451132867826D-05,   2.220481129310280D-04, &
   1.094376480767462D-04,  -2.572605143332579D-05,  -8.330852620507315D-05, &
   1.734427694204379D-03,   7.796589812818536D-04,   6.287652392640012D-04, &
   5.169367424742561D-04,  -3.143365463721144D-04,  -4.512972691060368D-04, &
  -1.000847912006759D-04,   4.076930620372294D-04,   3.145665554844876D-04, &
  -2.314894472404800D-04,  -3.777308159145990D-04,  -9.530359705920068D-05, &
   1.155191770853366D-04,   1.198729915594055D-04,   7.520826298270875D-05, &
   2.985702643312196D-04,   2.958068828728993D-04,  -1.231563735552979D-04, &
   6.199977382441826D-04,   7.243528465020499D-04,   3.110825882214831D-04, &
  -3.766026684686636D-04,  -7.188552766121495D-04,  -4.364011461193319D-05, &
   6.159310436201987D-04,   5.145411364494098D-04,  -2.075450095624752D-05, &
  -2.763512656733751D-04,  -1.771556995070404D-04,  -6.144648536525528D-05, &
  -5.737903656983557D-06,  -6.887685745110302D-04,  -4.231521445470482D-04, &
  -1.068658229690875D-03,  -9.268688426562108D-04,  -2.646957332728284D-04, &
   4.311009603458675D-04,   1.086233339753631D-03,   4.879904669751826D-04, &
  -6.962402497408895D-04,  -1.096139797748002D-03,  -4.799360274376977D-04, &
   3.391183764305354D-04,   5.030442269089660D-04,   2.270733120823176D-04, &
   2.295291429175169D-05,  -2.610938111520022D-05,  -4.454167326764391D-05, &
  -6.483045315014486D-04,  -3.636419664194561D-04,  -5.579131158713758D-04, &
  -9.085690976479209D-04,  -1.036458093152832D-03,  -1.368159847474707D-03, &
  -7.181578016770404D-04,   8.385328818196045D-04,   1.790811770683738D-03, &
   1.357195706148754D-03,  -3.968317547479658D-05,  -9.163115523765660D-04, &
  -7.469141173625129D-04,  -2.205172998197107D-04,   6.661730099037707D-05, &
   8.423998048586046D-05,   4.877129030844796D-05,   2.715090529832597D-05, &
  -4.371711136746870D-03,  -1.525317901391620D-03,  -5.320636500510985D-04, &
  -1.151227775243699D-04,   2.263190606461071D-04,  -3.320112931366401D-04, &
  -1.780245255024656D-03,  -2.811411541088613D-03,  -2.468888368961090D-03, &
  -5.506606938262785D-04,   1.332077077523486D-03,   1.666370274790047D-03, &
   8.877318070761708D-04,   6.256543356944979D-05,  -2.288700952127678D-04, &
  -1.580182165278422D-04,  -6.078270470331619D-05,  -2.634030441179739D-05, &
  -8.740993400178296D-06,  -3.260157764905942D-03,   8.293657864431540D-04, &
   1.887454729561632D-03,   2.300939341037821D-03,   2.863944293230579D-03, &
   3.582372333494362D-03,   3.785771414007481D-03,   2.964272921639797D-03, &
   5.760381118028642D-04,  -2.211431514163267D-03,  -3.232642497754558D-03, &
  -2.264997231391460D-03,  -6.389255475214063D-04,   3.484770419208645D-04, &
   4.558299195403555D-04,   2.372789074531219D-04,   6.887721223875698D-05, &
   1.571914095110428D-05,   7.535175041797971D-06,   1.316127886132660D-05, &
   9.274211909771941D-03,   4.296321117973434D-03,   3.851103132716310D-03, &
   2.329915810931415D-03,   1.070312177940583D-03,   5.009123050499075D-04, &
   8.781567298416714D-04,   2.694366264536618D-03,   5.164056748820685D-03, &
   6.123981884057045D-03,   4.649379229753045D-03,   1.821432612483573D-03, &
  -3.945486209434132D-04,  -1.027244136437428D-03,  -7.209501313162622D-04, &
  -2.856437091995214D-04,  -4.833507602462703D-05,   3.490679392747690D-06, &
  -4.084251633376299D-07,  -1.667962755651127D-06,   4.964322455016449D-06 /)
lem(:, 47) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.537941190077660D-05,  -1.288965894155696D-04, &
   4.756358903217270D-04,  -1.173163348824907D-03,   2.098871332575246D-03, &
  -2.536349266486547D-03,   1.113644725224165D-03,   2.900361714812586D-03, &
  -6.378132150719544D-03,   1.489628234891942D-03,   1.138813127667259D-02, &
  -4.383571927571142D-03,  -2.696058649741498D-02,  -2.717979613926143D-02, &
  -1.117324950917764D-02,   1.036726152421555D-03,   6.694810799231440D-03, &
   8.173831096372207D-03,   7.304715677828425D-03,   7.184168697105012D-03, &
   6.890873985159666D-03,   0.000000000000000D+00,   3.588529443539948D-05, &
   0.000000000000000D+00,  -5.355644953253603D-05,   0.000000000000000D+00, &
   9.366838538669157D-05,   0.000000000000000D+00,  -1.872721525972767D-04, &
   0.000000000000000D+00,   4.269043338244498D-04,   0.000000000000000D+00, &
  -1.131461703894743D-03,   0.000000000000000D+00,   3.642617823019971D-03, &
   0.000000000000000D+00,  -1.547608741217479D-02,   0.000000000000000D+00, &
   1.034811625612694D-01,   0.000000000000000D+00,  -1.972486703396304D+00, &
  -1.653319610837655D-02,   3.846006258452961D+00,   0.000000000000000D+00, &
   5.126470633516430D-06,   4.549291391176628D-05,   1.699241458241624D-04, &
   4.196618260631814D-04,   7.451497363763514D-04,   8.759469592883509D-04, &
   3.067607546219216D-04,  -1.197319693530119D-03,  -2.386557987168722D-03, &
  -3.015075503665575D-04,   4.625470151873006D-03,   1.401291025101851D-03, &
  -1.079670523963422D-02,   1.155807239709076D-02,  -5.758841310606438D-03, &
   9.190922786250326D-04,   1.371015152685546D-03,  -1.979687205479406D-03, &
   1.815820772660331D-03,  -1.804339538531316D-03,   1.716414733312382D-03, &
   4.052830890223062D-07,  -3.647547801049115D-06,  -4.052830890043910D-07, &
   9.948167252005689D-06,  -2.796112740128660D-06,  -2.011089613949935D-06, &
  -1.689587734930921D-05,   9.260207722778332D-06,  -8.633551517951512D-07, &
  -4.126054305309634D-06,   1.613072104120169D-05,  -1.710158406000181D-05, &
   7.153757382017463D-06,   2.325325913763351D-06,  -5.715257846548995D-06, &
   8.577697377947374D-06,   1.880581647797951D-05,  -1.612583948016284D-05, &
   2.418416147094651D-06,   6.192742628730854D-06,  -4.620823219230381D-06, &
  -7.047157712384749D-05,   8.204428247205206D-07,   2.238998694617856D-05, &
  -1.114349038457589D-05,  -5.671137136649834D-06,   8.363747575077615D-06, &
   1.653643257666515D-06,   1.407715049393947D-04,  -5.494960158689059D-05, &
  -1.371092076634007D-05,   2.242531058537557D-05,   1.260231956947367D-06, &
  -1.506077772506847D-05,   4.318711955038814D-06,   1.235398216021137D-05, &
  -9.794598417674281D-05,   1.205216545980680D-04,  -2.225252037978869D-05, &
  -3.039790526553594D-05,   1.127685968386139D-05,   2.145903043065733D-05, &
  -1.789590049393366D-05,  -9.332005851138413D-06,   1.769936339087733D-05, &
  -2.049775417367912D-04,  -9.800143764140661D-05,   7.514538310501531D-05, &
   2.605641400339634D-05,  -3.686681350060012D-05,  -1.824842211035062D-05, &
   3.921517966244459D-05,  -1.674397998597579D-06,  -2.509657422929760D-05, &
   8.324387465090582D-07,   5.524149652180875D-04,  -1.247420479081310D-04, &
  -9.174528698704300D-05,  -3.439749702175987D-07,   7.521490087476399D-05, &
  -1.257580779576059D-05,  -5.984651876633989D-05,   2.805294038567473D-05, &
   3.490592025589908D-05,  -1.647282557051684D-05,  -3.361108350455752D-05, &
  -5.259427896099713D-05,   3.813461842779894D-04,   1.980378518238770D-05, &
  -6.173886488460211D-05,  -9.826643511776881D-05,   8.670753245352895D-05, &
   5.550157316806307D-05,  -7.339266537717272D-05,  -3.606086827043856D-05, &
   5.108445152280649D-05,   3.448739643368267D-05,  -2.323473532531601D-05, &
  -1.242110816516610D-03,  -1.037936802427985D-04,   9.370246299663423D-05, &
   1.745381432296607D-04,   2.026845504567323D-05,  -1.754099857326988D-04, &
   6.736190130888279D-06,   1.292383637784042D-04,   8.509296208879428D-06, &
  -1.071210230389202D-04,  -2.311043608415844D-05,   5.658179716922303D-05, &
   6.212115342867953D-05,  -1.287674436684458D-04,  -5.789130622937241D-04, &
  -1.953735812987446D-04,  -2.108683204169516D-04,   2.280471085826095D-04, &
   1.650231808578439D-04,  -1.351410995322701D-04,  -1.677106917869397D-04, &
   8.023357120905768D-05,   1.721381515174775D-04,  -2.669556219149484D-05, &
  -1.213487069091723D-04,  -6.062493975194547D-05,   1.638564363633009D-05, &
   1.540825114605756D-03,   2.373059633773025D-04,   2.511212919137742D-04, &
  -2.054952810330786D-04,  -4.108693392105621D-04,   4.849004673452320D-05, &
   2.877956830500372D-04,   1.222032305652881D-04,  -2.592053229407076D-04, &
  -1.945541193051382D-04,   1.472340754424435D-04,   2.096414158630238D-04, &
   4.344576837637352D-05,  -5.734183296755904D-05,  -8.770175376477692D-05, &
   1.198945783519128D-03,   6.573604289697478D-04,   4.758960232719397D-04, &
   7.083751256056750D-04,   5.565601189368674D-05,  -3.639181052624045D-04, &
  -3.685745125927357D-04,   1.386460997967992D-04,   4.748536897178341D-04, &
   7.601519249939804D-05,  -3.618142678337897D-04,  -2.957965610373484D-04, &
   2.309743504706444D-05,   1.417421388024553D-04,   1.011835925196051D-04, &
   4.389106062615365D-05,  -9.683900729886415D-05,   2.009216456805269D-05, &
  -4.634966690833998D-04,   1.378719901413147D-04,   5.614186227576353D-04, &
   5.615835115823255D-04,   9.882016632898659D-05,  -6.556702715769770D-04, &
  -4.919838793683231D-04,   2.877328041039595D-04,   6.624040138796404D-04, &
   3.058337598691267D-04,  -1.882722782977145D-04,  -2.655618759446369D-04, &
  -1.197928565071870D-04,  -2.748040760006267D-05,   1.349911202304798D-05, &
  -6.514087843690619D-04,  -4.042610263600565D-04,  -9.028849544162566D-04, &
  -1.009457945747331D-03,  -6.474622125746304D-04,  -1.626697594356651D-04, &
   7.868970402041469D-04,   9.449534754372515D-04,  -1.531090504101956D-05, &
  -9.504062640178167D-04,  -9.308784576905897D-04,  -8.972270665341020D-05, &
   4.885389527096610D-04,   4.107246350380491D-04,   1.205012434934604D-04, &
  -1.727869150494328D-05,  -3.444461665484023D-05,  -2.898981096415690D-05, &
  -4.737708224390195D-04,  -2.189948874320758D-04,  -2.060679857997844D-04, &
  -4.100660125615199D-04,  -5.215838768457169D-04,  -1.108499464827593D-03, &
  -1.215061064646112D-03,  -1.460433223021540D-04,   1.230274044278818D-03, &
   1.744683212311528D-03,   7.876023124133205D-04,  -5.184464791974482D-04, &
  -9.159365557115541D-04,  -5.273602747128994D-04,  -5.602610068409103D-05, &
   1.004996308204846D-04,   7.060840675663923D-05,   3.623788492408553D-05, &
   2.882541558223362D-05,  -3.955370019932776D-03,  -1.334390071371113D-03, &
  -3.809658194447933D-04,   9.305401343967011D-05,   5.840312264381830D-04, &
   4.953018261627783D-04,  -5.857904901755480D-04,  -1.980422262107425D-03, &
  -2.746583724539169D-03,  -1.771113812544134D-03,   3.209876192735972D-04, &
   1.571239084087192D-03,   1.373869770692899D-03,   4.869282056619345D-04, &
  -1.273601954732743D-04,  -2.217952327940238D-04,  -1.171325795375171D-04, &
  -4.241829515411326D-05,  -1.822883189842262D-05,  -4.465230838701345D-06, &
  -2.858970401667891D-03,   8.295764116363570D-04,   1.681305046729889D-03, &
   1.913430949028060D-03,   2.204069154752237D-03,   2.794404960415950D-03, &
   3.376674557314911D-03,   3.553576573633006D-03,   2.271381211664069D-03, &
  -4.581641724068643D-04,  -2.639126206668565D-03,  -2.856156184319782D-03, &
  -1.518205136304232D-03,  -9.822529127108276D-05,   4.541317233588057D-04, &
   3.757089529857533D-04,   1.568654129392125D-04,   3.538019237558161D-05, &
   5.742430087180668D-06,   4.406080309382820D-06,  -1.423199452194552D-06 /)
lem(:, 48) = (/ -1.632993161855451D+00,  -1.632993161855453D+00,   3.265986323710905D+00, &
  -7.999999999999992D+00,   2.065591117977290D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.000000000000023D+00,   2.065591117977287D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.600000000000000D+01, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00 /)
lem(:, 49) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.131182235954581D+00,   1.777777777777778D+00,  -4.883113471764755D-01, &
   2.236067977499785D-01,  -8.240220541217370D-02,  -1.279723452110892D-02, &
   1.546473935329351D-02,   2.134153809558426D-03,  -4.512404789263121D-03, &
  -5.544886778118465D-04,   1.692519775057697D-03,   1.866376183136989D-04, &
  -7.481295096979534D-04,  -7.466266479892237D-05,   3.714976551047495D-04, &
   3.382007275431738D-05,  -2.012391266888751D-04,  -1.682991409127283D-05, &
   1.166318363024254D-04,   9.016329996486229D-06,  -7.135071646201642D-05, &
  -5.126470634711370D-06,   4.131182235954576D+00,  -1.777777777777781D+00, &
   4.883113471764763D-01,  -2.236067977499790D-01,   8.240220541217386D-02, &
   1.279723452110885D-02,  -1.546473935329371D-02,  -2.134153809558525D-03, &
   4.512404789263525D-03,   5.544886778114630D-04,  -1.692519775057449D-03, &
  -1.866376183141407D-04,   7.481295096983064D-04,   7.466266479902135D-05, &
  -3.714976551050955D-04,  -3.382007275521336D-05,   2.012391266908333D-04, &
   1.682991409035108D-05,  -1.166318363034772D-04,  -9.016329995719977D-06, &
   7.135071646284959D-05,   5.126470633516430D-06,   0.000000000000000D+00, &
   7.111111111111113D+00,   0.000000000000000D+00,  -4.472135954999579D-01, &
   0.000000000000000D+00,   2.559446904221702D-02,   0.000000000000000D+00, &
  -4.268307619117014D-03,   0.000000000000000D+00,   1.108977355622888D-03, &
   0.000000000000000D+00,  -3.732752366284262D-04,   0.000000000000000D+00, &
   1.493253295978426D-04,   0.000000000000000D+00,  -6.764014551072315D-05, &
   0.000000000000000D+00,   3.365982818001015D-05,   0.000000000000000D+00, &
  -1.803265999247915D-05,   0.000000000000000D+00,   1.025294126712685D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.007795339789447D-02, &
   0.000000000000000D+00,   1.003897669894724D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.756820922315767D-02,   0.000000000000000D+00,  -3.346325566315736D-03, &
   0.000000000000000D+00,  -7.529232524210436D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.129479905837284D-03,   0.000000000000000D+00, &
  -1.597109929377965D-03,   0.000000000000000D+00,  -8.099628927559725D-03, &
   0.000000000000000D+00,  -6.084228302392187D-04,   0.000000000000000D+00, &
   3.422378420095644D-03,   0.000000000000000D+00,   2.053427052057391D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.020263425836853D-05, &
   0.000000000000000D+00,  -1.351400709473664D-03,   0.000000000000000D+00, &
   2.272810284114805D-03,   0.000000000000000D+00,   2.196026152894709D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.265197569378296D-05,   0.000000000000000D+00,  -1.632211246507157D-03, &
   0.000000000000000D+00,   3.137180218421004D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.930572442105254D-04,   0.000000000000000D+00,  -2.457092199043029D-04, &
   0.000000000000000D+00,  -1.235127596483242D-03,   0.000000000000000D+00, &
  -4.705770327631500D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.265197569378173D-05,   0.000000000000000D+00, &
   5.791717326315721D-04,   0.000000000000000D+00,  -4.460792385167036D-05, &
   0.000000000000000D+00,  -3.529327745723648D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.129566721083062D-06,   0.000000000000000D+00,  -1.321461350745828D-04, &
   0.000000000000000D+00,   7.362673333126852D-04,   0.000000000000000D+00, &
   2.352885163815785D-04,   0.000000000000000D+00,   1.470553227384795D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.504067585701929D-05, &
   0.000000000000000D+00,  -1.936594726906823D-04,   0.000000000000000D+00, &
   4.313622800328893D-04,   0.000000000000000D+00,   1.274479463733589D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.860601520167883D-05,   0.000000000000000D+00,  -2.269767445478215D-04, &
   0.000000000000000D+00,  -2.464011600518544D-04,   0.000000000000000D+00, &
   2.941106454769522D-05,   0.000000000000000D+00,  -5.718818106496238D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.846386932314541D-06, &
   0.000000000000000D+00,   1.339799887171674D-04,   0.000000000000000D+00, &
  -2.733494695283782D-04,   0.000000000000000D+00,  -1.862700754687486D-04, &
   0.000000000000000D+00,  -5.228633697368785D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.075230316820916D-07, &
   0.000000000000000D+00,   7.790481955726327D-06,   0.000000000000000D+00, &
   2.209258765336081D-04,   0.000000000000000D+00,  -1.078405700082216D-04, &
   0.000000000000000D+00,  -2.777711651726771D-05,   0.000000000000000D+00, &
   2.573468147923535D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.547698153088149D-05, &
   0.000000000000000D+00,   5.866803444903738D-05,   0.000000000000000D+00, &
   2.147898956362085D-04,   0.000000000000000D+00,   3.505561228917395D-05, &
   0.000000000000000D+00,   2.417500381382550D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.104688119285616D-06,   0.000000000000000D+00, &
  -9.949519782668157D-05,   0.000000000000000D+00,   5.927436338361391D-05, &
   0.000000000000000D+00,   1.001164520462002D-04,   0.000000000000000D+00, &
   1.637661548678674D-05,   0.000000000000000D+00,  -1.286734073961736D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.055720897301632D-06,   0.000000000000000D+00, &
   1.863875244186248D-05,   0.000000000000000D+00,  -1.633055066501957D-04, &
   0.000000000000000D+00,  -1.173471767306745D-05,   0.000000000000000D+00, &
  -4.790438543754831D-06,   0.000000000000000D+00,  -1.228246161508855D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.067460307226333D-08, &
   0.000000000000000D+00,   1.368214282902316D-05,   0.000000000000000D+00, &
   1.894834259498961D-05,   0.000000000000000D+00,  -1.351720643353774D-04, &
   0.000000000000000D+00,  -3.665242513708752D-05,   0.000000000000000D+00, &
  -9.358065992450859D-06,   0.000000000000000D+00,   6.969809567288024D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.965414170140399D-06,   0.000000000000000D+00,   6.108902579212175D-05, &
   0.000000000000000D+00,   3.015376819788663D-05,   0.000000000000000D+00, &
  -3.965180526288310D-05,   0.000000000000000D+00,  -1.184755149683300D-06, &
   0.000000000000000D+00,   6.722360706919343D-06,   0.000000000000000D+00 /)
lem(:, 50) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.018350154434630D+00,   1.464934041529428D+00,  -6.666666666666663D-01, &
   1.574851970871775D-01,  -1.462494064565351D-01,   1.135641906448745D-01, &
  -2.541406439970112D-02,  -1.893873631013170D-02,   1.029693088809917D-02, &
   4.920598885138541D-03,  -4.291320867305166D-03,  -1.656244560712883D-03, &
   1.994429335340119D-03,   6.625654226549366D-04,  -1.018984528259454D-03, &
  -3.001233730356332D-04,   5.619713580869074D-04,   1.493506717520474D-04, &
  -3.296682188796275D-04,  -8.001199140984403D-05,   2.034181473762045D-04, &
   4.549291391131821D-05,  -1.018350154434631D+00,   1.464934041529432D+00, &
  -6.666666666666662D-01,   1.574851970871777D-01,  -1.462494064565349D-01, &
   1.135641906448744D-01,  -2.541406439970151D-02,  -1.893873631013124D-02, &
   1.029693088809882D-02,   4.920598885138869D-03,  -4.291320867305680D-03, &
  -1.656244560712505D-03,   1.994429335339969D-03,   6.625654226549657D-04, &
  -1.018984528259897D-03,  -3.001233730344878D-04,   5.619713580858680D-04, &
   1.493506717525642D-04,  -3.296682188797968D-04,  -8.001199141002478D-05, &
   2.034181473763018D-04,   4.549291391176628D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.333333333333332D+00,   0.000000000000000D+00, &
  -5.849976258261413D-01,   0.000000000000000D+00,   3.049687727964110D-02, &
   0.000000000000000D+00,  -4.576413728044295D-03,   0.000000000000000D+00, &
   1.072830216826533D-03,   0.000000000000000D+00,  -3.278513975904689D-04, &
   0.000000000000000D+00,   1.198805327358500D-04,   0.000000000000000D+00, &
  -4.995300960777740D-05,   0.000000000000000D+00,   2.300010829297768D-05, &
   0.000000000000000D+00,  -1.146017731756322D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.484784677291246D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.908708063747469D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.889725952916128D-03,   0.000000000000000D+00,   1.822235740311983D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.889725952916130D-03, &
   0.000000000000000D+00,  -9.673597139927824D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.180453022595537D-03, &
   0.000000000000000D+00,   3.634088370992590D-04,   0.000000000000000D+00, &
  -9.902890810954745D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.142142059454799D-03,   0.000000000000000D+00,   3.478341726521442D-03, &
   0.000000000000000D+00,   3.698982806188855D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.530257274678427D-04, &
   0.000000000000000D+00,  -2.543861859694793D-03,   0.000000000000000D+00, &
   1.933854168849611D-03,   0.000000000000000D+00,   2.783971269921083D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.229865778844100D-05, &
   0.000000000000000D+00,   6.749021260414733D-04,   0.000000000000000D+00, &
  -2.660671843048128D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.886169238305224D-06,   0.000000000000000D+00,   1.435312213753388D-04, &
   0.000000000000000D+00,   6.138250105626076D-04,   0.000000000000000D+00, &
  -1.606709869065935D-03,   0.000000000000000D+00,  -5.219946131102079D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.720338752654750D-04,   0.000000000000000D+00, &
   6.581059192847883D-04,   0.000000000000000D+00,   5.306074407226530D-04, &
   0.000000000000000D+00,  -3.673295425590302D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.754107195114377D-05,   0.000000000000000D+00, &
  -5.203609509548188D-04,   0.000000000000000D+00,   8.100151528820601D-04, &
   0.000000000000000D+00,   4.059958101968212D-04,   0.000000000000000D+00, &
   1.449985036417282D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.000838070572848D-06, &
   0.000000000000000D+00,   7.931988549689675D-05,   0.000000000000000D+00, &
  -6.837824171736012D-04,   0.000000000000000D+00,   4.059958101968255D-04, &
   0.000000000000000D+00,   1.217987430590399D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.357641708696746D-07,   0.000000000000000D+00, &
   7.212725150300269D-05,   0.000000000000000D+00,  -4.791740803193886D-05, &
   0.000000000000000D+00,  -5.336893404024163D-04,   0.000000000000000D+00, &
   5.799940145672556D-06,   0.000000000000000D+00,  -5.074947627460875D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.568728967070233D-05,   0.000000000000000D+00,   2.914824971603270D-04, &
   0.000000000000000D+00,  -1.209901995833740D-04,   0.000000000000000D+00, &
  -2.367430114004874D-04,   0.000000000000000D+00,  -4.547680341490214D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.171401771412183D-05,   0.000000000000000D+00, &
  -1.262991238774594D-04,   0.000000000000000D+00,   4.258067823253165D-04, &
   0.000000000000000D+00,  -2.372702786864754D-05,   0.000000000000000D+00, &
  -1.911343911640875D-05,   0.000000000000000D+00,   2.076114938506552D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.441322090103769D-06,   0.000000000000000D+00,  -1.395199783221439D-05, &
   0.000000000000000D+00,  -1.391274182362171D-04,   0.000000000000000D+00, &
   3.137240351520953D-04,   0.000000000000000D+00,   4.679497162982350D-05, &
   0.000000000000000D+00,   1.922328646765324D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.918346032741719D-08,   0.000000000000000D+00,   3.173791242410757D-05, &
   0.000000000000000D+00,  -1.274723273014850D-04,   0.000000000000000D+00, &
  -1.007506077468731D-04,   0.000000000000000D+00,   1.008398684417483D-04, &
   0.000000000000000D+00,   1.153397188059042D-05,   0.000000000000000D+00, &
  -9.515526801481516D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.274128727652546D-05,   0.000000000000000D+00,   1.149292218123464D-04, &
   0.000000000000000D+00,  -2.083792820333847D-04,   0.000000000000000D+00, &
  -7.457790169824690D-05,   0.000000000000000D+00,  -8.999033005732695D-06, &
   0.000000000000000D+00,  -8.983189637771483D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.506619261590031D-06, &
   0.000000000000000D+00,  -1.861140512784935D-05,   0.000000000000000D+00, &
   1.676676895409998D-04,   0.000000000000000D+00,  -1.423114684343750D-04, &
   0.000000000000000D+00,  -3.720445288662916D-05,   0.000000000000000D+00, &
  -6.321503819169671D-06,   0.000000000000000D+00,   4.757763400729286D-06 /)
lem(:, 51) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.944271909999155D-01,  -6.771863474748648D-01, &
   1.899999999999996D-01,  -7.487531723681105D-02,   1.030157507275422D-01, &
  -9.423106802156399D-02,   4.724391862686868D-02,   9.081039465712847D-04, &
  -1.545503410108929D-02,   3.619013489716231D-03,   5.634014423076721D-03, &
  -2.500008945934959D-03,  -2.343989145962994D-03,   1.505444099569651D-03, &
   1.086172874128358D-03,  -9.076782563967507D-04,  -5.484261550117509D-04, &
   5.626689111338748D-04,   2.967407690595914D-04,  -3.602758295359098D-04, &
  -1.699241458243655D-04,   0.000000000000000D+00,  -8.944271909999149D-01, &
   6.771863474748642D-01,  -1.899999999999999D-01,   7.487531723681089D-02, &
  -1.030157507275424D-01,   9.423106802156456D-02,  -4.724391862686926D-02, &
  -9.081039465707726D-04,   1.545503410108890D-02,  -3.619013489715791D-03, &
  -5.634014423076887D-03,   2.500008945935042D-03,   2.343989145962603D-03, &
  -1.505444099568734D-03,  -1.086172874129206D-03,   9.076782563970733D-04, &
   5.484261550112639D-04,  -5.626689111331123D-04,  -2.967407690598735D-04, &
   3.602758295356620D-04,   1.699241458241624D-04,   0.000000000000000D+00, &
  -4.472135954999579D-01,   0.000000000000000D+00,   3.219999999999995D+00, &
   0.000000000000000D+00,  -6.567254108880844D-01,   0.000000000000000D+00, &
   3.435921354681309D-02,   0.000000000000000D+00,  -5.216771438454159D-03, &
   0.000000000000000D+00,   1.252003205127755D-03,   0.000000000000000D+00, &
  -3.956733814771676D-04,   0.000000000000000D+00,   1.507479701014281D-04, &
   0.000000000000000D+00,  -6.578508510270316D-05,   0.000000000000000D+00, &
   3.182356319092472D-05,   0.000000000000000D+00,  -1.668897860879074D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.417937871268419D-03, &
   0.000000000000000D+00,  -3.030457633656631D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.499923475244190D-03,   0.000000000000000D+00, &
  -1.928473039599677D-02,   0.000000000000000D+00,   1.867251673263178D-03, &
   0.000000000000000D+00,   9.826029297007882D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.120668888033479D-04,   0.000000000000000D+00, &
  -5.933763198768250D-04,   0.000000000000000D+00,   2.109929637419148D-02, &
   0.000000000000000D+00,  -2.943334920023923D-04,   0.000000000000000D+00, &
  -2.465337329012040D-03,   0.000000000000000D+00,  -1.062691072874635D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.095654463057024D-04, &
   0.000000000000000D+00,   1.730680932974064D-03,   0.000000000000000D+00, &
   2.781451499422600D-04,   0.000000000000000D+00,  -1.169239796979503D-02, &
   0.000000000000000D+00,  -4.944802665640184D-05,   0.000000000000000D+00, &
  -8.123604379265994D-04,   0.000000000000000D+00,   3.280346768366664D-03, &
   0.000000000000000D+00,   4.381889862185608D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.878276600502713D-06,   0.000000000000000D+00, &
   4.325663508345734D-04,   0.000000000000000D+00,  -2.392411877934738D-03, &
   0.000000000000000D+00,   1.913340835363790D-03,   0.000000000000000D+00, &
   3.373751662469281D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.073492882482717D-04,   0.000000000000000D+00, &
   1.006101130603470D-03,   0.000000000000000D+00,  -2.744443391237008D-03, &
   0.000000000000000D+00,  -5.918862565735108D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.121407634288473D-05,   0.000000000000000D+00,  -2.438229487773365D-04, &
   0.000000000000000D+00,   9.387803177384330D-04,   0.000000000000000D+00, &
  -1.792431592713806D-03,   0.000000000000000D+00,  -6.313453403451229D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.185391948556935D-06,   0.000000000000000D+00,  -7.214359926263081D-05, &
   0.000000000000000D+00,   2.708050376262268D-04,   0.000000000000000D+00, &
   6.888672342603349D-04,   0.000000000000000D+00,  -4.458876466187537D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.467495207471076D-07,   0.000000000000000D+00, &
   1.139562688617955D-04,   0.000000000000000D+00,  -4.856935461410069D-04, &
   0.000000000000000D+00,   6.611265304514115D-04,   0.000000000000000D+00, &
   4.814008220131648D-04,   0.000000000000000D+00,   1.783879412173093D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.425122119617099D-05, &
   0.000000000000000D+00,   2.898182522102199D-04,   0.000000000000000D+00, &
  -7.183660794969154D-04,   0.000000000000000D+00,   4.222121963558061D-04, &
   0.000000000000000D+00,   1.508562623130547D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.296450086424739D-05, &
   0.000000000000000D+00,  -5.658580203023234D-05,   0.000000000000000D+00, &
   1.979234565121732D-04,   0.000000000000000D+00,  -5.815388547485241D-04, &
   0.000000000000000D+00,   1.319786779178017D-05,   0.000000000000000D+00, &
  -6.434521046841710D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.560019013003318D-06,   0.000000000000000D+00,  -4.788901794142275D-05, &
   0.000000000000000D+00,   1.493416786990422D-04,   0.000000000000000D+00, &
  -9.164405978498039D-06,   0.000000000000000D+00,  -2.739267531369639D-04, &
   0.000000000000000D+00,  -5.806762895929701D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.488091262487475D-08, &
   0.000000000000000D+00,   4.198881803171351D-05,   0.000000000000000D+00, &
  -2.019486398555056D-04,   0.000000000000000D+00,   3.546214174957490D-04, &
   0.000000000000000D+00,  -1.442349084831599D-05,   0.000000000000000D+00, &
  -2.667972141373388D-05,   0.000000000000000D+00,   2.736633189129712D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.492644020429135D-05,   0.000000000000000D+00, &
   9.339137393785246D-05,   0.000000000000000D+00,  -2.458295403187362D-04, &
   0.000000000000000D+00,   3.136548761160612D-04,   0.000000000000000D+00, &
   5.196250848065800D-05,   0.000000000000000D+00,   2.551022005085561D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.822264238104903D-06, &
   0.000000000000000D+00,   5.015888274868128D-06,   0.000000000000000D+00, &
   1.351188330699623D-05,   0.000000000000000D+00,  -1.515398853019700D-04, &
   0.000000000000000D+00,   1.179460987385351D-04,   0.000000000000000D+00, &
   1.622966084868147D-05,   0.000000000000000D+00,  -1.311350560437832D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.776140501356878D-07,   0.000000000000000D+00, &
  -2.806858637647980D-05,   0.000000000000000D+00,   1.103666341601158D-04, &
   0.000000000000000D+00,  -1.252789650757792D-04,   0.000000000000000D+00, &
  -8.245361866777655D-05,   0.000000000000000D+00,  -8.637063982230535D-06, &
   0.000000000000000D+00,  -1.245707580714603D-05,   0.000000000000000D+00 /)
lem(:, 52) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.382847385042240D-02,   2.472066162365217D-01,  -5.849976258261407D-01, &
   3.507581975254649D-01,  -4.166666666666643D-02,   1.324049849177861D-02, &
  -7.008807458597009D-02,   8.071927209375478D-02,  -5.145226771587096D-02, &
   1.360109519424606D-02,   9.288526320418791D-03,  -9.273748871269740D-03, &
  -9.171781259195512D-04,   4.689950310412833D-03,  -5.897280909636681D-04, &
  -2.389803270582697D-03,   7.048241712452007D-04,   1.275262872858023D-03, &
  -5.664638297368640D-04,  -7.150642651889848D-04,   4.172631580302345D-04, &
   4.196618260632488D-04,   6.382847385042273D-02,   2.472066162365225D-01, &
  -5.849976258261399D-01,   3.507581975254647D-01,  -4.166666666666680D-02, &
   1.324049849177903D-02,  -7.008807458597049D-02,   8.071927209375543D-02, &
  -5.145226771587144D-02,   1.360109519424642D-02,   9.288526320418522D-03, &
  -9.273748871269572D-03,  -9.171781259199029D-04,   4.689950310413466D-03, &
  -5.897280909646747D-04,  -2.389803270582092D-03,   7.048241712449214D-04, &
   1.275262872858414D-03,  -5.664638297375536D-04,  -7.150642651883736D-04, &
   4.172631580300124D-04,   4.196618260631814D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.849976258261413D-01,   0.000000000000000D+00, &
   2.653333333333335D+00,   0.000000000000000D+00,  -6.957834313443615D-01, &
   0.000000000000000D+00,   3.614208073700244D-02,   0.000000000000000D+00, &
  -5.397386915919424D-03,   0.000000000000000D+00,   1.265829108811616D-03, &
   0.000000000000000D+00,  -3.889017680958921D-04,   0.000000000000000D+00, &
   1.434553790231373D-04,   0.000000000000000D+00,  -6.042811620135665D-05, &
   0.000000000000000D+00,   2.815754139096722D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.001808023903555D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.264014327112208D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.195576372727134D-04,   0.000000000000000D+00,  -2.122896021831506D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.191880380603467D-03, &
   0.000000000000000D+00,   1.132818621091777D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.202670707647032D-03, &
   0.000000000000000D+00,  -8.200027552138862D-04,   0.000000000000000D+00, &
   2.311952212616930D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.644456689839423D-05,   0.000000000000000D+00,  -3.293026937604969D-03, &
   0.000000000000000D+00,  -1.189003995060134D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.065631215570883D-04, &
   0.000000000000000D+00,   2.326020887338736D-03,   0.000000000000000D+00, &
   3.443475622385108D-04,   0.000000000000000D+00,  -1.296830337749413D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.865104305976683D-04, &
   0.000000000000000D+00,  -6.860153769109627D-04,   0.000000000000000D+00, &
   3.773084573010297D-03,   0.000000000000000D+00,   5.067737615862291D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.438477165642956D-05,   0.000000000000000D+00,   1.478092341699590D-04, &
   0.000000000000000D+00,  -2.505916572628603D-03,   0.000000000000000D+00, &
   1.985957395222276D-03,   0.000000000000000D+00,   3.766175848816563D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.123866305527966D-06, &
   0.000000000000000D+00,  -1.530747030716882D-04,   0.000000000000000D+00, &
   8.142201420450211D-04,   0.000000000000000D+00,  -3.027920820545880D-03, &
   0.000000000000000D+00,  -1.696475607575023D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.835634705699754D-07, &
   0.000000000000000D+00,   1.294741192442975D-04,   0.000000000000000D+00, &
  -2.640961746023380D-04,   0.000000000000000D+00,   9.744937283807961D-04, &
   0.000000000000000D+00,  -1.976258736473739D-03,   0.000000000000000D+00, &
  -6.819831942451591D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.556950215931722D-05, &
   0.000000000000000D+00,   1.303203553177886D-04,   0.000000000000000D+00, &
   2.552635855852642D-04,   0.000000000000000D+00,   8.352283933174217D-04, &
   0.000000000000000D+00,  -4.747047200105409D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.287458539691790D-05,   0.000000000000000D+00, &
   5.595582732330444D-06,   0.000000000000000D+00,  -3.145445687843006D-04, &
   0.000000000000000D+00,   6.886193988682411D-04,   0.000000000000000D+00, &
   5.616876511625690D-04,   0.000000000000000D+00,   1.862267541951678D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.533036365022320D-06,   0.000000000000000D+00, &
  -6.764369657022983D-05,   0.000000000000000D+00,   2.023309059737866D-04, &
   0.000000000000000D+00,  -7.203156283567363D-04,   0.000000000000000D+00, &
   4.321386047659272D-04,   0.000000000000000D+00,   1.561528684245177D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.358574552454562D-08, &
   0.000000000000000D+00,   4.653746511115256D-05,   0.000000000000000D+00, &
  -1.497078997080426D-04,   0.000000000000000D+00,   1.755903634512060D-04, &
   0.000000000000000D+00,  -6.601452253661014D-04,   0.000000000000000D+00, &
   5.783439571278515D-06,   0.000000000000000D+00,  -6.477452319831951D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.564432949777588D-05,   0.000000000000000D+00,   5.510882473162914D-05, &
   0.000000000000000D+00,   5.365673907432917D-05,   0.000000000000000D+00, &
   7.732162100189771D-06,   0.000000000000000D+00,  -3.047427774096723D-04, &
   0.000000000000000D+00,  -5.807166502853050D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.899891588073856D-06,   0.000000000000000D+00,   2.254161340764737D-05, &
   0.000000000000000D+00,  -9.183154191549803D-05,   0.000000000000000D+00, &
   3.479364329460513D-04,   0.000000000000000D+00,   8.274767386595761D-06, &
   0.000000000000000D+00,  -2.455737417958404D-05,   0.000000000000000D+00, &
   2.650928503488332D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.833051202523317D-07,   0.000000000000000D+00, &
  -3.390049305064106D-05,   0.000000000000000D+00,   1.091121705527848D-04, &
   0.000000000000000D+00,  -2.066330333479213D-04,   0.000000000000000D+00, &
   3.454937823888373D-04,   0.000000000000000D+00,   5.895295105843810D-05, &
   0.000000000000000D+00,   2.457405717834903D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.139733241904403D-08,   0.000000000000000D+00,   1.639213582607971D-05, &
   0.000000000000000D+00,  -7.054034932530094D-05,   0.000000000000000D+00, &
   4.523640774280160D-05,   0.000000000000000D+00,  -1.711813815096109D-04, &
   0.000000000000000D+00,   1.250409293974893D-04,   0.000000000000000D+00, &
   1.489791789562906D-05,   0.000000000000000D+00,  -1.220361359534430D-05 /)
lem(:, 53) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.118893808443566D-02,  -3.217652068271441D-01, &
   3.970398725957363D-01,  -1.541240482332509D-01,   2.409297052153686D-03, &
   5.446218031199479D-03,   3.464263498801546D-02,  -6.272331177551164D-02, &
   5.369173465419403D-02,  -2.325831968766476D-02,  -1.894873135899645D-03, &
   9.396072162799972D-03,  -3.622446987562875D-03,  -3.015531051855232D-03, &
   2.939658387561047D-03,   7.991659988600480D-04,  -1.903382984590972D-03, &
  -6.952681118953512D-05,   1.187304841018221D-03,  -1.482090616992491D-04, &
  -7.451497363762595D-04,   0.000000000000000D+00,   5.118893808443528D-02, &
   3.217652068271439D-01,  -3.970398725957370D-01,   1.541240482332515D-01, &
  -2.409297052154351D-03,  -5.446218031198898D-03,  -3.464263498801600D-02, &
   6.272331177551208D-02,  -5.369173465419434D-02,   2.325831968766499D-02, &
   1.894873135899230D-03,  -9.396072162799394D-03,   3.622446987562095D-03, &
   3.015531051856036D-03,  -2.939658387561494D-03,  -7.991659988597428D-04, &
   1.903382984590620D-03,   6.952681119004634D-05,  -1.187304841019022D-03, &
   1.482090616998441D-04,   7.451497363763514D-04,   0.000000000000000D+00, &
   2.559446904221702D-02,   0.000000000000000D+00,  -6.567254108880844D-01, &
   0.000000000000000D+00,   2.329365079365080D+00,   0.000000000000000D+00, &
  -7.205053574004202D-01,   0.000000000000000D+00,   3.749573698487107D-02, &
   0.000000000000000D+00,  -5.612607779731865D-03,   0.000000000000000D+00, &
   1.324289225126035D-03,   0.000000000000000D+00,  -4.112526546317832D-04, &
   0.000000000000000D+00,   1.540530374277926D-04,   0.000000000000000D+00, &
  -6.617550630029674D-05,   0.000000000000000D+00,   3.155737882759176D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.023420441901896D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.061234497500999D-03,   0.000000000000000D+00, &
   1.087057870273520D-02,   0.000000000000000D+00,  -3.099478532383893D-04, &
   0.000000000000000D+00,  -4.677012035415097D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.433601900670822D-04,   0.000000000000000D+00, &
  -9.121563180901421D-04,   0.000000000000000D+00,  -2.341790791152196D-02, &
   0.000000000000000D+00,   3.207286481336349D-04,   0.000000000000000D+00, &
   2.130049241412716D-03,   0.000000000000000D+00,   1.229705473979278D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.323407823536783D-06, &
   0.000000000000000D+00,  -1.072173833497646D-03,   0.000000000000000D+00, &
  -8.994482579309889D-04,   0.000000000000000D+00,   2.511763882636132D-02, &
   0.000000000000000D+00,  -8.323407823537142D-05,   0.000000000000000D+00, &
   2.302611321468543D-04,   0.000000000000000D+00,  -3.458710373315478D-03, &
   0.000000000000000D+00,  -1.287988836731126D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.482689063006167D-05,   0.000000000000000D+00, &
   7.611537417576769D-05,   0.000000000000000D+00,   2.673032304934878D-03, &
   0.000000000000000D+00,   4.549296813090796D-04,   0.000000000000000D+00, &
  -1.414447395107834D-02,   0.000000000000000D+00,  -5.782578066876689D-06, &
   0.000000000000000D+00,  -1.600065635551040D-04,   0.000000000000000D+00, &
  -1.016386661925488D-03,   0.000000000000000D+00,   4.028384502711314D-03, &
   0.000000000000000D+00,   5.583293711429131D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.442010754076853D-07,   0.000000000000000D+00, &
   1.235447178168473D-04,   0.000000000000000D+00,   1.600202533706024D-04, &
   0.000000000000000D+00,  -2.823066978716745D-03,   0.000000000000000D+00, &
   2.028206914768516D-03,   0.000000000000000D+00,   4.131808734391480D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.156914384804979D-05,   0.000000000000000D+00,   2.045258435586357D-05, &
   0.000000000000000D+00,   9.104537287744188D-04,   0.000000000000000D+00, &
  -3.248160277554215D-03,   0.000000000000000D+00,  -2.272396004089013D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.182493951458721D-05,   0.000000000000000D+00, &
   4.210630121943965D-05,   0.000000000000000D+00,  -1.277823746956599D-04, &
   0.000000000000000D+00,   1.129627625316454D-03,   0.000000000000000D+00, &
  -2.110299409333564D-03,   0.000000000000000D+00,  -7.423160280024226D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.402993513898767D-06,   0.000000000000000D+00,  -7.604811551700834D-05, &
   0.000000000000000D+00,   5.978148985472547D-05,   0.000000000000000D+00, &
   2.777593999700535D-04,   0.000000000000000D+00,   9.490189961524150D-04, &
   0.000000000000000D+00,  -5.161542438800443D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.734368866868577D-08,   0.000000000000000D+00,   4.718241678262883D-05, &
   0.000000000000000D+00,  -8.614140816132017D-05,   0.000000000000000D+00, &
  -3.741333723280348D-04,   0.000000000000000D+00,   6.754496357801892D-04, &
   0.000000000000000D+00,   6.146546767318575D-04,   0.000000000000000D+00, &
   2.027641889024152D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.540818494907709D-05,   0.000000000000000D+00,   2.584181831088450D-05, &
   0.000000000000000D+00,   1.321012362013364D-04,   0.000000000000000D+00, &
  -7.854454522154327D-04,   0.000000000000000D+00,   4.395095301716382D-04, &
   0.000000000000000D+00,   1.702569733376820D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.825458877169845D-06, &
   0.000000000000000D+00,   3.341032209104755D-05,   0.000000000000000D+00, &
  -5.888441261331548D-05,   0.000000000000000D+00,   1.958016608062649D-04, &
   0.000000000000000D+00,  -7.079899062352527D-04,   0.000000000000000D+00, &
   7.720895514046114D-06,   0.000000000000000D+00,  -7.104907960870594D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.749867287250219D-07, &
   0.000000000000000D+00,  -3.665016573128330D-05,   0.000000000000000D+00, &
   7.876896634748522D-05,   0.000000000000000D+00,   9.495866677884910D-05, &
   0.000000000000000D+00,   4.964483781987226D-05,   0.000000000000000D+00, &
  -3.269992951156264D-04,   0.000000000000000D+00,  -6.383618644809655D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.106268448978790D-08,   0.000000000000000D+00,   1.681113756065406D-05, &
   0.000000000000000D+00,  -5.104123664180167D-05,   0.000000000000000D+00, &
  -6.964537360709226D-05,   0.000000000000000D+00,   3.610235199445334D-04, &
   0.000000000000000D+00,   2.019178267543493D-05,   0.000000000000000D+00, &
  -2.777172064513909D-05,   0.000000000000000D+00,   2.946776068221608D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.337757607265784D-06,   0.000000000000000D+00, &
  -3.145313103900539D-06,   0.000000000000000D+00,   3.067037682854872D-05, &
   0.000000000000000D+00,  -2.292640591419017D-04,   0.000000000000000D+00, &
   3.558002834313628D-04,   0.000000000000000D+00,   6.206573501590571D-05, &
   0.000000000000000D+00,   2.738532116226510D-05,   0.000000000000000D+00 /)
lem(:, 54) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.197893559374875D-02,  -4.639421805988072D-02,  -4.879500364742646D-02, &
   2.887447405431415D-01,  -2.618745332257619D-01,   6.606251054196632D-02, &
   1.834542410714328D-02,  -9.219073425887698D-03,  -1.687880277375722D-02, &
   4.079459945808685D-02,  -4.808756573958513D-02,   3.137901076615629D-02, &
  -6.398904072050778D-03,  -7.400656430635275D-03,   6.332030071500284D-03, &
   3.229283313569294D-04,  -3.255262290729337D-03,   1.104591881462571D-03, &
   1.473153746702824D-03,  -1.124646728993856D-03,  -6.144098491481891D-04, &
   8.759469592883326D-04,  -1.197893559374916D-02,  -4.639421805988059D-02, &
  -4.879500364742648D-02,   2.887447405431426D-01,  -2.618745332257625D-01, &
   6.606251054196660D-02,   1.834542410714296D-02,  -9.219073425887285D-03, &
  -1.687880277375759D-02,   4.079459945808712D-02,  -4.808756573958553D-02, &
   3.137901076615685D-02,  -6.398904072051552D-03,  -7.400656430634391D-03, &
   6.332030071499792D-03,   3.229283313571923D-04,  -3.255262290729701D-03, &
   1.104591881463065D-03,   1.473153746702198D-03,  -1.124646728993183D-03, &
  -6.144098491487001D-04,   8.759469592883509D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.049687727964110D-02,   0.000000000000000D+00, &
  -6.957834313443615D-01,   0.000000000000000D+00,   2.125837053571425D+00, &
   0.000000000000000D+00,  -7.367008615299606D-01,   0.000000000000000D+00, &
   3.832321433573528D-02,   0.000000000000000D+00,  -5.712310933678395D-03, &
   0.000000000000000D+00,   1.338089373599307D-03,   0.000000000000000D+00, &
  -4.113209170138435D-04,   0.000000000000000D+00,   1.520695147793388D-04, &
   0.000000000000000D+00,  -6.429395032336220D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.445189988862347D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.493128555517588D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.189617906669296D-03,   0.000000000000000D+00,   1.209770571412664D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.987783480099794D-04, &
   0.000000000000000D+00,  -5.333331477538582D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.517466569011496D-04, &
   0.000000000000000D+00,  -1.078041834132912D-03,   0.000000000000000D+00, &
  -2.527096561170268D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.595328093453609D-04,   0.000000000000000D+00,   2.554906061048085D-03, &
   0.000000000000000D+00,   1.331898054736475D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.326130639559476D-04, &
   0.000000000000000D+00,  -1.087394354531685D-03,   0.000000000000000D+00, &
  -9.193318002600075D-04,   0.000000000000000D+00,   2.689153957572794D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.198589640526537D-05, &
   0.000000000000000D+00,   8.398781795664702D-06,   0.000000000000000D+00, &
  -3.783135308292042D-03,   0.000000000000000D+00,  -1.385175161431465D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.284708323646279D-05,   0.000000000000000D+00,   8.959891004675491D-05, &
   0.000000000000000D+00,   2.793887903593416D-03,   0.000000000000000D+00, &
   5.159100608347577D-04,   0.000000000000000D+00,  -1.516759243651619D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.234347308307773D-05, &
   0.000000000000000D+00,  -1.394054927364525D-05,   0.000000000000000D+00, &
  -1.081152460989937D-03,   0.000000000000000D+00,   4.296156896140436D-03, &
   0.000000000000000D+00,   6.069081335753952D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.856151375363110D-06, &
   0.000000000000000D+00,   5.337200503321725D-05,   0.000000000000000D+00, &
   3.212905800349868D-04,   0.000000000000000D+00,  -2.992480239538585D-03, &
   0.000000000000000D+00,   2.098510974068149D-03,   0.000000000000000D+00, &
   4.436682616510004D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.175504292474783D-06,   0.000000000000000D+00,  -7.518624098384640D-05, &
   0.000000000000000D+00,  -4.957905136642281D-05,   0.000000000000000D+00, &
   1.027915094354835D-03,   0.000000000000000D+00,  -3.453257809402685D-03, &
   0.000000000000000D+00,  -2.913214635168035D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.642420604135345D-08, &
   0.000000000000000D+00,   4.479844118517605D-05,   0.000000000000000D+00, &
  -4.138752093323168D-05,   0.000000000000000D+00,  -1.956522476339603D-04, &
   0.000000000000000D+00,   1.223882911031168D-03,   0.000000000000000D+00, &
  -2.242779849639380D-03,   0.000000000000000D+00,  -7.856127991559639D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.444242658415786D-05,   0.000000000000000D+00, &
   6.915906878269983D-06,   0.000000000000000D+00,  -6.011674862214575D-06, &
   0.000000000000000D+00,   2.422992288306665D-04,   0.000000000000000D+00, &
   1.043036371456839D-03,   0.000000000000000D+00,  -5.433514330719546D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.635072043887836D-06, &
   0.000000000000000D+00,   3.893513126828883D-05,   0.000000000000000D+00, &
  -3.720662207880085D-06,   0.000000000000000D+00,  -3.909471482138662D-04, &
   0.000000000000000D+00,   6.846333556825368D-04,   0.000000000000000D+00, &
   6.679553973481400D-04,   0.000000000000000D+00,   2.114721751435253D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.559987125808790D-07,   0.000000000000000D+00, &
  -3.715149668173642D-05,   0.000000000000000D+00,   5.114015600314457D-05, &
   0.000000000000000D+00,   1.859498187051659D-04,   0.000000000000000D+00, &
  -8.133407655562778D-04,   0.000000000000000D+00,   4.505413508624836D-04, &
   0.000000000000000D+00,   1.770113092719973D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.029879877826760D-08,   0.000000000000000D+00, &
   1.653645340511471D-05,   0.000000000000000D+00,  -3.582488630134230D-05, &
   0.000000000000000D+00,  -4.215819594414485D-05,   0.000000000000000D+00, &
   2.279923989263479D-04,   0.000000000000000D+00,  -7.549236522066601D-04, &
   0.000000000000000D+00,   4.706979914314547D-06,   0.000000000000000D+00, &
  -7.292794754848671D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.220133684125492D-06,   0.000000000000000D+00, &
  -9.810726864994649D-06,   0.000000000000000D+00,   7.803240105774279D-06, &
   0.000000000000000D+00,   7.974674293061727D-05,   0.000000000000000D+00, &
   6.775922131324473D-05,   0.000000000000000D+00,  -3.488250217520470D-04, &
   0.000000000000000D+00,  -6.535459617609575D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.322587182111900D-07,   0.000000000000000D+00,   2.607503749809489D-05, &
   0.000000000000000D+00,  -3.754789259802773D-05,   0.000000000000000D+00, &
  -1.014479161496219D-04,   0.000000000000000D+00,   3.589524138331944D-04, &
   0.000000000000000D+00,   3.162490947920653D-05,   0.000000000000000D+00, &
  -2.748783931410407D-05,   0.000000000000000D+00,   2.971361829867939D-05 /)
lem(:, 55) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.536615238233881D-03,   5.365975287870620D-02, &
   1.145307118227122D-01,  -2.461146433446844D-01,   1.528577463994489D-01, &
  -2.747848834004371D-02,  -1.936848958333378D-02,   1.439971049546754D-02, &
   7.085540613616902D-03,  -2.653740325297178D-02,   3.723337919657097D-02, &
  -3.298916747878706D-02,   1.509614368618103D-02,   2.426799162669793D-03, &
  -7.637368729126264D-03,   2.726743610820762D-03,   2.393761218722895D-03, &
  -2.393724667243649D-03,  -3.462536577836929D-04,   1.499100994631358D-03, &
  -3.067607546220178D-04,   0.000000000000000D+00,  -8.536615238234121D-03, &
  -5.365975287870563D-02,  -1.145307118227126D-01,   2.461146433446843D-01, &
  -1.528577463994499D-01,   2.747848834004410D-02,   1.936848958333333D-02, &
  -1.439971049546719D-02,  -7.085540613617367D-03,   2.653740325297222D-02, &
  -3.723337919657167D-02,   3.298916747878809D-02,  -1.509614368618169D-02, &
  -2.426799162669581D-03,   7.637368729126025D-03,  -2.726743610820203D-03, &
  -2.393761218723612D-03,   2.393724667244259D-03,   3.462536577833881D-04, &
  -1.499100994631133D-03,   3.067607546219216D-04,   0.000000000000000D+00, &
  -4.268307619117014D-03,   0.000000000000000D+00,   3.435921354681309D-02, &
   0.000000000000000D+00,  -7.205053574004202D-01,   0.000000000000000D+00, &
   1.990125868055552D+00,   0.000000000000000D+00,  -7.481007591048005D-01, &
   0.000000000000000D+00,   3.896556444148536D-02,   0.000000000000000D+00, &
  -5.812303926672576D-03,   0.000000000000000D+00,   1.364149781378617D-03, &
   0.000000000000000D+00,  -4.210132796760198D-04,   0.000000000000000D+00, &
   1.566613964112046D-04,   0.000000000000000D+00,  -6.683516282784280D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.374393457665111D-04, &
   0.000000000000000D+00,   2.410281041189376D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.410281041189384D-04,   0.000000000000000D+00, &
  -3.036954111898590D-03,   0.000000000000000D+00,  -2.410281041189349D-04, &
   0.000000000000000D+00,   3.615421561784013D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.699197493432303D-04,   0.000000000000000D+00, &
   1.363479427331803D-03,   0.000000000000000D+00,   1.331361143537609D-02, &
   0.000000000000000D+00,   6.515988071861866D-05,   0.000000000000000D+00, &
  -8.095906969079273D-04,   0.000000000000000D+00,  -5.794248450439813D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.688329100145935D-04, &
   0.000000000000000D+00,  -1.896481908069914D-04,   0.000000000000000D+00, &
  -1.182245551174143D-03,   0.000000000000000D+00,  -2.705466564547382D-02, &
   0.000000000000000D+00,   4.741348273451812D-05,   0.000000000000000D+00, &
   4.389050337465382D-04,   0.000000000000000D+00,   2.785916849892301D-03, &
   0.000000000000000D+00,   1.421100108998870D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.921726720565467D-05,   0.000000000000000D+00, &
  -2.718239073720461D-04,   0.000000000000000D+00,  -1.273597780475902D-03, &
   0.000000000000000D+00,  -9.717809706857909D-04,   0.000000000000000D+00, &
   2.857850654000552D-02,   0.000000000000000D+00,  -2.343121507534138D-05, &
   0.000000000000000D+00,   8.624286035146159D-05,   0.000000000000000D+00, &
   9.204756790777368D-06,   0.000000000000000D+00,  -4.057298744084387D-03, &
   0.000000000000000D+00,  -1.472695766034419D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.456399612286325D-06,   0.000000000000000D+00, &
   3.355594414931165D-05,   0.000000000000000D+00,   2.433578911658384D-04, &
   0.000000000000000D+00,   3.018481153419588D-03,   0.000000000000000D+00, &
   5.903145247527504D-04,   0.000000000000000D+00,  -1.612308619711108D-02, &
   0.000000000000000D+00,  -8.111054789297258D-07,   0.000000000000000D+00, &
  -6.518817298525215D-05,   0.000000000000000D+00,  -8.285497769919556D-05, &
   0.000000000000000D+00,  -1.111628907297414D-03,   0.000000000000000D+00, &
   4.544663432601220D-03,   0.000000000000000D+00,   6.493633789501396D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.893306298844778D-08, &
   0.000000000000000D+00,   3.956946882596771D-05,   0.000000000000000D+00, &
  -2.081144931162051D-05,   0.000000000000000D+00,   2.935654087958578D-04, &
   0.000000000000000D+00,  -3.190504447397142D-03,   0.000000000000000D+00, &
   2.157607464809634D-03,   0.000000000000000D+00,   4.721844126453455D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.279924445751990D-05,   0.000000000000000D+00,  -2.291784202785407D-06, &
   0.000000000000000D+00,  -1.203605171659938D-04,   0.000000000000000D+00, &
   1.079553523290319D-03,   0.000000000000000D+00,  -3.646526569754395D-03, &
   0.000000000000000D+00,  -3.377688763776098D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.337065253291493D-06,   0.000000000000000D+00,   4.000597686719507D-05, &
   0.000000000000000D+00,   3.690672071114317D-05,   0.000000000000000D+00, &
  -2.221393255742278D-04,   0.000000000000000D+00,   1.323743664689196D-03, &
   0.000000000000000D+00,  -2.356561582350663D-03,   0.000000000000000D+00, &
  -8.304171399722752D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.271095340989752D-07,   0.000000000000000D+00, &
  -3.584602673221236D-05,   0.000000000000000D+00,   3.117973691192439D-05, &
   0.000000000000000D+00,   4.360015806510472D-05,   0.000000000000000D+00, &
   2.423881049176467D-04,   0.000000000000000D+00,   1.128806209751430D-03, &
   0.000000000000000D+00,  -5.735758461388542D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.136590449043141D-09,   0.000000000000000D+00,   1.570755064547652D-05, &
   0.000000000000000D+00,  -2.505277198528167D-05,   0.000000000000000D+00, &
   1.352858446659848D-05,   0.000000000000000D+00,  -3.924315610979108D-04, &
   0.000000000000000D+00,   6.908476520894726D-04,   0.000000000000000D+00, &
   7.106306230282202D-04,   0.000000000000000D+00,   2.228965649144419D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.984695511015491D-06, &
   0.000000000000000D+00,  -1.382069798658583D-05,   0.000000000000000D+00, &
  -1.544856553112459D-05,   0.000000000000000D+00,   1.847619564641879D-04, &
   0.000000000000000D+00,  -8.551345260875043D-04,   0.000000000000000D+00, &
   4.591618962009505D-04,   0.000000000000000D+00,   1.866074795940537D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.954625462923718D-07,   0.000000000000000D+00,   2.662389604718496D-05, &
   0.000000000000000D+00,  -2.477576934666115D-05,   0.000000000000000D+00, &
  -7.627842615174438D-05,   0.000000000000000D+00,   2.412329148326283D-04, &
   0.000000000000000D+00,  -7.950858665979933D-04,   0.000000000000000D+00, &
   5.162052992195188D-06,   0.000000000000000D+00,  -7.693668414207122D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.874076277391457D-08, &
   0.000000000000000D+00,  -1.620463030064949D-05,   0.000000000000000D+00, &
   3.386511927828273D-05,   0.000000000000000D+00,   1.772236349708432D-05, &
   0.000000000000000D+00,   7.468465195590954D-05,   0.000000000000000D+00, &
   8.986614003352484D-05,   0.000000000000000D+00,  -3.664331426723476D-04, &
   0.000000000000000D+00,  -6.899731059758047D-05,   0.000000000000000D+00 /)
lem(:, 56) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.495293720032749D-03,   1.353721436778978D-02,   2.288206864022038D-02, &
  -1.392426051408886D-02,  -1.392976028405296D-01,   1.923601513902547D-01, &
  -8.759531651373290D-02,   4.149162928590171D-03,   1.528284143518548D-02, &
  -1.302073488851782D-02,   1.099776144800952D-03,   1.605220467788405D-02, &
  -2.826469581784991D-02,   2.975850873278377D-02,  -1.985918945969803D-02, &
   4.377579100522022D-03,   5.865381823714143D-03,  -5.484419495491444D-03, &
  -1.568911142912880D-04,   2.909125701542961D-03,  -1.113198376663252D-03, &
  -1.197319693530119D-03,   3.495293720032978D-03,   1.353721436779015D-02, &
   2.288206864021969D-02,  -1.392426051408793D-02,  -1.392976028405301D-01, &
   1.923601513902552D-01,  -8.759531651373337D-02,   4.149162928590765D-03, &
   1.528284143518507D-02,  -1.302073488851735D-02,   1.099776144800341D-03, &
   1.605220467788504D-02,  -2.826469581785070D-02,   2.975850873278404D-02, &
  -1.985918945969818D-02,   4.377579100522633D-03,   5.865381823713393D-03, &
  -5.484419495490818D-03,  -1.568911142916656D-04,   2.909125701543134D-03, &
  -1.113198376663255D-03,  -1.197319693530119D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.576413728044295D-03,   0.000000000000000D+00, &
   3.614208073700244D-02,   0.000000000000000D+00,  -7.367008615299606D-01, &
   0.000000000000000D+00,   1.894918981481482D+00,   0.000000000000000D+00, &
  -7.563229283801729D-01,   0.000000000000000D+00,   3.941387782893296D-02, &
   0.000000000000000D+00,  -5.871362444253836D-03,   0.000000000000000D+00, &
   1.374181153528935D-03,   0.000000000000000D+00,  -4.222870447382378D-04, &
   0.000000000000000D+00,   1.561988267437408D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.669294049386305D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.446196032924193D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.669294049386292D-04,   0.000000000000000D+00,  -3.414481962623362D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.630797355282806D-04, &
   0.000000000000000D+00,   4.586617561732892D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.286814362532195D-04, &
   0.000000000000000D+00,   1.555100142368273D-03,   0.000000000000000D+00, &
   1.435857613150127D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.188629508417011D-04,   0.000000000000000D+00,  -9.365846117365619D-04, &
   0.000000000000000D+00,  -6.260987844536127D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.986842479230275D-04, &
   0.000000000000000D+00,  -2.911982089880038D-04,   0.000000000000000D+00, &
  -1.309064781518296D-03,   0.000000000000000D+00,  -2.869848738384408D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.812946428783066D-04, &
   0.000000000000000D+00,   5.008316146074569D-04,   0.000000000000000D+00, &
   3.019609399410917D-03,   0.000000000000000D+00,   1.507573622359452D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.930941880052487D-05,   0.000000000000000D+00,  -1.776568749548330D-04, &
   0.000000000000000D+00,  -1.342140618307534D-03,   0.000000000000000D+00, &
  -1.002249217117879D-03,   0.000000000000000D+00,   3.015142101915397D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.007296502986493D-05, &
   0.000000000000000D+00,   3.672120340401209D-05,   0.000000000000000D+00, &
   3.797092188117408D-05,   0.000000000000000D+00,  -4.317060359612220D-03, &
   0.000000000000000D+00,  -1.556483190758877D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.060626211607137D-05, &
   0.000000000000000D+00,  -3.410176565239234D-05,   0.000000000000000D+00, &
   2.260219488061348D-04,   0.000000000000000D+00,   3.192136010017871D-03, &
   0.000000000000000D+00,   6.460477016156145D-04,   0.000000000000000D+00, &
  -1.700368416531085D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.033928596497591D-05,   0.000000000000000D+00,  -1.260643994483622D-06, &
   0.000000000000000D+00,  -1.561364999503369D-04,   0.000000000000000D+00, &
  -1.193649881968796D-03,   0.000000000000000D+00,   4.777631338094803D-03, &
   0.000000000000000D+00,   6.894038675619621D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.911633728484183D-06, &
   0.000000000000000D+00,   3.685593945811778D-05,   0.000000000000000D+00, &
   5.337445832348575D-05,   0.000000000000000D+00,   3.002402268494826D-04, &
   0.000000000000000D+00,  -3.362282419071952D-03,   0.000000000000000D+00, &
   2.224819114640352D-03,   0.000000000000000D+00,   4.978988360959207D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.865900864163237D-07, &
   0.000000000000000D+00,  -3.287964280131317D-05,   0.000000000000000D+00, &
   2.010447497147710D-05,   0.000000000000000D+00,  -8.272241673645235D-05, &
   0.000000000000000D+00,   1.156153221233867D-03,   0.000000000000000D+00, &
  -3.825288678076212D-03,   0.000000000000000D+00,  -3.840097776294531D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.506497731089851D-09,   0.000000000000000D+00,   1.436966021749580D-05, &
   0.000000000000000D+00,  -1.845490935817142D-05,   0.000000000000000D+00, &
   5.693067997052100D-05,   0.000000000000000D+00,  -2.317178303546575D-04, &
   0.000000000000000D+00,   1.410034728914574D-03,   0.000000000000000D+00, &
  -2.467192271526806D-03,   0.000000000000000D+00,  -8.682522784504177D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.640572548997478D-06,   0.000000000000000D+00, &
  -1.558006678183403D-05,   0.000000000000000D+00,  -3.218870186208761D-05, &
   0.000000000000000D+00,   4.159429198037780D-05,   0.000000000000000D+00, &
   2.294807870647616D-04,   0.000000000000000D+00,   1.203803620697113D-03, &
   0.000000000000000D+00,  -5.981514664817229D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.437047465909832D-07,   0.000000000000000D+00, &
   2.606803784106977D-05,   0.000000000000000D+00,  -1.535016781475988D-05, &
   0.000000000000000D+00,  -1.787555563632152D-05,   0.000000000000000D+00, &
  -4.093951311803401D-04,   0.000000000000000D+00,   7.008050586432084D-04, &
   0.000000000000000D+00,   7.523525933618484D-04,   0.000000000000000D+00, &
   2.311785145556011D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.449415992853366D-08, &
   0.000000000000000D+00,  -1.554686359535579D-05,   0.000000000000000D+00, &
   2.751930487717295D-05,   0.000000000000000D+00,  -7.093488401575535D-06, &
   0.000000000000000D+00,   1.911895770008764D-04,   0.000000000000000D+00, &
  -8.871701475232480D-04,   0.000000000000000D+00,   4.698924763915544D-04, &
   0.000000000000000D+00,   1.932597598617521D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.540923287498840D-09,   0.000000000000000D+00, &
   5.231721784841231D-06,   0.000000000000000D+00,  -1.949869340844512D-06, &
   0.000000000000000D+00,   1.862252028449445D-05,   0.000000000000000D+00, &
  -6.778704844548786D-05,   0.000000000000000D+00,   2.631367871959921D-04, &
   0.000000000000000D+00,  -8.323116106009148D-04,   0.000000000000000D+00, &
   3.665986392492393D-06,   0.000000000000000D+00,  -7.911943288481939D-05 /)
lem(:, 57) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.217954711247210D-03,  -1.394169684122625D-02, &
  -3.867943101292419D-02,  -2.659050092208462D-02,   1.467698259176837D-01, &
  -1.345126322944374D-01,   4.882686108511246D-02,   3.722918278770047D-03, &
  -1.459347559400870D-02,   1.028252373491060D-02,  -3.630046117195844D-03, &
  -7.329325818088248D-03,   1.987618400040763D-02,  -2.566954029963526D-02, &
   2.118104517046833D-02,  -9.618097901075547D-03,  -1.739822939238033D-03, &
   5.991433025000204D-03,  -2.704374469823146D-03,  -1.863137555046363D-03, &
   2.386557987168748D-03,   0.000000000000000D+00,   2.217954711246647D-03, &
   1.394169684122694D-02,   3.867943101292372D-02,   2.659050092208481D-02, &
  -1.467698259176841D-01,   1.345126322944379D-01,  -4.882686108511292D-02, &
  -3.722918278769644D-03,   1.459347559400819D-02,  -1.028252373490959D-02, &
   3.630046117194874D-03,   7.329325818088557D-03,  -1.987618400040785D-02, &
   2.566954029963589D-02,  -2.118104517046914D-02,   9.618097901076126D-03, &
   1.739822939237639D-03,  -5.991433024999983D-03,   2.704374469823016D-03, &
   1.863137555046271D-03,  -2.386557987168722D-03,   0.000000000000000D+00, &
   1.108977355622888D-03,   0.000000000000000D+00,  -5.216771438454159D-03, &
   0.000000000000000D+00,   3.749573698487107D-02,   0.000000000000000D+00, &
  -7.481007591048005D-01,   0.000000000000000D+00,   1.825671164772727D+00, &
   0.000000000000000D+00,  -7.625075596945748D-01,   0.000000000000000D+00, &
   3.977179890414836D-02,   0.000000000000000D+00,  -5.926717240130590D-03, &
   0.000000000000000D+00,   1.388151574579140D-03,   0.000000000000000D+00, &
  -4.272976869277262D-04,   0.000000000000000D+00,   1.585234341426620D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.767235793300508D-05, &
   0.000000000000000D+00,  -7.452150424305770D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.767235793301448D-06,   0.000000000000000D+00, &
   4.767184462607178D-04,   0.000000000000000D+00,   1.300473309339610D-04, &
   0.000000000000000D+00,   3.254836288262922D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.191808948325089D-05,   0.000000000000000D+00, &
  -4.306904583458973D-04,   0.000000000000000D+00,  -3.789089719417131D-03, &
   0.000000000000000D+00,  -1.461205965550106D-04,   0.000000000000000D+00, &
  -2.005505187717540D-04,   0.000000000000000D+00,   4.931570133731607D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.358546590929714D-05, &
   0.000000000000000D+00,   4.351816048549205D-04,   0.000000000000000D+00, &
   1.699868809115091D-03,   0.000000000000000D+00,   1.534718409245538D-02, &
   0.000000000000000D+00,   7.602794878339047D-05,   0.000000000000000D+00, &
  -1.392994973588934D-04,   0.000000000000000D+00,  -1.026402363786424D-03, &
   0.000000000000000D+00,  -6.664329550772115D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.918455864280479D-05,   0.000000000000000D+00, &
  -1.155774058875463D-04,   0.000000000000000D+00,  -2.968090976790894D-04, &
   0.000000000000000D+00,  -1.409713539359907D-03,   0.000000000000000D+00, &
  -3.026661055074404D-02,   0.000000000000000D+00,  -3.850196093098431D-06, &
   0.000000000000000D+00,   1.575199646853735D-04,   0.000000000000000D+00, &
   5.841320855663433D-04,   0.000000000000000D+00,   3.232814928777646D-03, &
   0.000000000000000D+00,   1.587895258898031D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.438350650810820D-05,   0.000000000000000D+00, &
  -9.763773889247974D-05,   0.000000000000000D+00,  -2.315122229933219D-04, &
   0.000000000000000D+00,  -1.438463139685571D-03,   0.000000000000000D+00, &
  -1.042942611309612D-03,   0.000000000000000D+00,   3.165175319029987D-02, &
   0.000000000000000D+00,  -6.229351747872020D-06,   0.000000000000000D+00, &
   1.732366966257554D-05,   0.000000000000000D+00,  -2.118862313858111D-05, &
   0.000000000000000D+00,   1.535888529744379D-05,   0.000000000000000D+00, &
  -4.564485053141020D-03,   0.000000000000000D+00,  -1.634972443918362D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.262811868017068D-06, &
   0.000000000000000D+00,   2.841606985509119D-05,   0.000000000000000D+00, &
   3.238067680026591D-05,   0.000000000000000D+00,   2.388437336563566D-04, &
   0.000000000000000D+00,   3.370015289492683D-03,   0.000000000000000D+00, &
   7.040849100791468D-04,   0.000000000000000D+00,  -1.783747043711587D-02, &
   0.000000000000000D+00,  -1.270613883084912D-07,   0.000000000000000D+00, &
  -2.805376787789942D-05,   0.000000000000000D+00,   1.955651105732899D-05, &
   0.000000000000000D+00,  -1.285853106929746D-04,   0.000000000000000D+00, &
  -1.244498295941365D-03,   0.000000000000000D+00,   5.002429920320719D-03, &
   0.000000000000000D+00,   7.262863450907201D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.111665043078483D-09,   0.000000000000000D+00, &
   1.247485407388296D-05,   0.000000000000000D+00,  -1.608593963480873D-05, &
   0.000000000000000D+00,   7.551122950472073D-05,   0.000000000000000D+00, &
   3.207950111988033D-04,   0.000000000000000D+00,  -3.530194919262532D-03, &
   0.000000000000000D+00,   2.287188400391179D-03,   0.000000000000000D+00, &
   5.222230969803728D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.178062442196349D-06,   0.000000000000000D+00, &
  -1.524892824349720D-05,   0.000000000000000D+00,  -3.997133195127447D-05, &
   0.000000000000000D+00,  -9.230654216940686D-05,   0.000000000000000D+00, &
   1.214048040208130D-03,   0.000000000000000D+00,  -3.997106649960503D-03, &
   0.000000000000000D+00,  -4.224262255827252D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.756517147310603D-07,   0.000000000000000D+00,   2.450653989354149D-05, &
   0.000000000000000D+00,  -9.681413403987852D-06,   0.000000000000000D+00, &
   3.038183322306608D-05,   0.000000000000000D+00,  -2.544258394786315D-04, &
   0.000000000000000D+00,   1.491893559204493D-03,   0.000000000000000D+00, &
  -2.568543072199442D-03,   0.000000000000000D+00,  -9.057859171700323D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.895583443415034D-08,   0.000000000000000D+00,  -1.450749692093803D-05, &
   0.000000000000000D+00,   2.308515595220694D-05,   0.000000000000000D+00, &
  -2.602073794718939D-05,   0.000000000000000D+00,   4.631669789208862D-05, &
   0.000000000000000D+00,   2.258349546255620D-04,   0.000000000000000D+00, &
   1.274479039875717D-03,   0.000000000000000D+00,  -6.233349333473479D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.349119807501207D-09,   0.000000000000000D+00, &
   4.868336415493039D-06,   0.000000000000000D+00,  -8.834647157009698D-08, &
   0.000000000000000D+00,   2.629920722979211D-05,   0.000000000000000D+00, &
  -8.982434785297268D-06,   0.000000000000000D+00,  -4.161419880951692D-04, &
   0.000000000000000D+00,   7.115086666386720D-04,   0.000000000000000D+00, &
   7.889199857660623D-04,   0.000000000000000D+00,   2.404067698732010D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.003625313340645D-06, &
   0.000000000000000D+00,  -1.655601276157301D-05,   0.000000000000000D+00, &
   2.341630189506742D-06,   0.000000000000000D+00,   8.127568605456295D-06, &
   0.000000000000000D+00,   2.001347751720831D-04,   0.000000000000000D+00, &
  -9.213442833341482D-04,   0.000000000000000D+00,   4.793101724606559D-04, &
   0.000000000000000D+00,   2.009505273521665D-04,   0.000000000000000D+00 /)
lem(:, 58) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.311020180384238D-03,  -5.077559325172566D-03,  -9.870037994802818D-03, &
  -5.251117612529084D-03,   2.877560349940494D-02,   5.748179794992284D-02, &
  -1.341574245522470D-01,   9.098828515998758D-02,  -2.287018861117004D-02, &
  -5.537703886092165D-03,   1.164045237531616D-02,  -9.840269553396747D-03, &
   4.216251536399529D-03,   2.963324140591913D-03,  -1.225654332916797D-02, &
   2.025333793755107D-02,  -2.080388009322497D-02,   1.307606198308565D-02, &
  -2.468643838910380D-03,  -4.369741229300788D-03,   4.397752317607543D-03, &
  -3.015075503667417D-04,  -1.311020180384994D-03,  -5.077559325172140D-03, &
  -9.870037994803207D-03,  -5.251117612528752D-03,   2.877560349940456D-02, &
   5.748179794992352D-02,  -1.341574245522475D-01,   9.098828515998789D-02, &
  -2.287018861117069D-02,  -5.537703886091174D-03,   1.164045237531528D-02, &
  -9.840269553396461D-03,   4.216251536399397D-03,   2.963324140592571D-03, &
  -1.225654332916893D-02,   2.025333793755170D-02,  -2.080388009322525D-02, &
   1.307606198308586D-02,  -2.468643838910531D-03,  -4.369741229300784D-03, &
   4.397752317607617D-03,  -3.015075503665575D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.072830216826533D-03,   0.000000000000000D+00, &
  -5.397386915919424D-03,   0.000000000000000D+00,   3.832321433573528D-02, &
   0.000000000000000D+00,  -7.563229283801729D-01,   0.000000000000000D+00, &
   1.773676091974433D+00,   0.000000000000000D+00,  -7.672431019386425D-01, &
   0.000000000000000D+00,   4.004127397659218D-02,   0.000000000000000D+00, &
  -5.964184546676235D-03,   0.000000000000000D+00,   1.395220904585443D-03, &
   0.000000000000000D+00,  -4.285988121049305D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.601778083306487D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.072573090718224D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.734518722202029D-06,   0.000000000000000D+00,   5.272571380693293D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.586550179809822D-04, &
   0.000000000000000D+00,   3.552215764031996D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.778765601836635D-05, &
   0.000000000000000D+00,  -5.049562319274190D-04,   0.000000000000000D+00, &
  -4.099464071535681D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.729044281391865D-04,   0.000000000000000D+00,  -2.182302958172142D-04, &
   0.000000000000000D+00,   5.435845872089291D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.456570843079487D-04, &
   0.000000000000000D+00,   5.148036581862231D-04,   0.000000000000000D+00, &
   1.847948947696499D-03,   0.000000000000000D+00,   1.625475021935456D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.970987611237475D-05, &
   0.000000000000000D+00,  -1.301237969410205D-04,   0.000000000000000D+00, &
  -1.108671207407664D-03,   0.000000000000000D+00,  -7.052687324510194D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.102109812357167D-04,   0.000000000000000D+00,  -1.623496789979534D-04, &
   0.000000000000000D+00,  -3.337942685810891D-04,   0.000000000000000D+00, &
  -1.514978130093472D-03,   0.000000000000000D+00,  -3.174863383762668D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.948174096775458D-05, &
   0.000000000000000D+00,   1.221866060005751D-04,   0.000000000000000D+00, &
   6.164983166388479D-04,   0.000000000000000D+00,   3.429746011827520D-03, &
   0.000000000000000D+00,   1.665072604855691D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.007865616076672D-05, &
   0.000000000000000D+00,  -4.698645641555539D-05,   0.000000000000000D+00, &
  -2.554246997890453D-04,   0.000000000000000D+00,  -1.515333851910379D-03, &
   0.000000000000000D+00,  -1.075609106549040D-03,   0.000000000000000D+00, &
   3.307708555506086D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.009924754273798D-05,   0.000000000000000D+00,   3.853281356185133D-05, &
   0.000000000000000D+00,   1.218617391381551D-05,   0.000000000000000D+00, &
   1.870578144589227D-05,   0.000000000000000D+00,  -4.796600595944610D-03, &
   0.000000000000000D+00,  -1.710210803733801D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.783981770713371D-06, &
   0.000000000000000D+00,  -1.928743836846898D-05,   0.000000000000000D+00, &
   5.301457632164633D-05,   0.000000000000000D+00,   2.642028975495887D-04, &
   0.000000000000000D+00,   3.534356667148132D-03,   0.000000000000000D+00, &
   7.537851755149934D-04,   0.000000000000000D+00,  -1.862469589076506D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.555361916024350D-06, &
   0.000000000000000D+00,  -1.261982844040061D-05,   0.000000000000000D+00, &
  -3.622118690731094D-05,   0.000000000000000D+00,  -1.441808149228616D-04, &
   0.000000000000000D+00,  -1.306505152360607D-03,   0.000000000000000D+00, &
   5.215271387192906D-03,   0.000000000000000D+00,   7.613642719313233D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.859004796419856D-07,   0.000000000000000D+00,   2.189431481635558D-05, &
   0.000000000000000D+00,  -8.259737992709898D-06,   0.000000000000000D+00, &
   5.350054432300477D-05,   0.000000000000000D+00,   3.271664643079432D-04, &
   0.000000000000000D+00,  -3.688635381472540D-03,   0.000000000000000D+00, &
   2.351823510667721D-03,   0.000000000000000D+00,   5.449217082247894D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.170898295341131D-08, &
   0.000000000000000D+00,  -1.307227488128896D-05,   0.000000000000000D+00, &
   2.055627004941015D-05,   0.000000000000000D+00,  -3.578392758044761D-05, &
   0.000000000000000D+00,  -9.511283040478254D-05,   0.000000000000000D+00, &
   1.275719912654162D-03,   0.000000000000000D+00,  -4.159199570352192D-03, &
   0.000000000000000D+00,  -4.595501807711615D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.098146592279610D-09, &
   0.000000000000000D+00,   4.396901390873822D-06,   0.000000000000000D+00, &
   6.959278129233227D-07,   0.000000000000000D+00,   3.010806899556804D-05, &
   0.000000000000000D+00,   4.155982701950546D-05,   0.000000000000000D+00, &
  -2.662140366101959D-04,   0.000000000000000D+00,   1.568308745250299D-03, &
   0.000000000000000D+00,  -2.666740100537530D-03,   0.000000000000000D+00, &
  -9.397485431899716D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.072292345637440D-07, &
   0.000000000000000D+00,  -1.589548347401999D-05,   0.000000000000000D+00, &
  -4.128850622155995D-07,   0.000000000000000D+00,  -1.276516776166843D-05, &
   0.000000000000000D+00,   5.326460298599542D-05,   0.000000000000000D+00, &
   2.199549424168102D-04,   0.000000000000000D+00,   1.338942741334868D-03, &
   0.000000000000000D+00,  -6.457305561118393D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.141157337548749D-07,   0.000000000000000D+00, &
   1.332344696351763D-05,   0.000000000000000D+00,  -1.972038129011333D-05, &
   0.000000000000000D+00,   1.009501760942771D-05,   0.000000000000000D+00, &
  -1.468492364891919D-05,   0.000000000000000D+00,  -4.289849456110488D-04, &
   0.000000000000000D+00,   7.229040795111082D-04,   0.000000000000000D+00, &
   8.243593792787864D-04,   0.000000000000000D+00,   2.481212306238817D-04 /)
lem(:, 59) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.465504732569271D-04,   4.692692922019044D-03, &
   1.423110309829029D-02,   2.124657390656561D-02,  -9.694699765066475D-03, &
  -7.985864002222626D-02,   1.088028856962122D-01,  -5.892806934942549D-02, &
   9.881366611684813D-03,   7.452316170218124D-03,  -8.672223915637229D-03, &
   8.003289245966693D-03,  -5.268932942058156D-03,  -4.589999663121133D-04, &
   7.490935351192359D-03,  -1.454770447616111D-02,   1.829060079582972D-02, &
  -1.505139790806519D-02,   6.340254154166489D-03,   1.814619254869061D-03, &
  -4.625470151873496D-03,   0.000000000000000D+00,  -7.465504732571145D-04, &
  -4.692692922018789D-03,  -1.423110309829060D-02,  -2.124657390656525D-02, &
   9.694699765066022D-03,   7.985864002222666D-02,  -1.088028856962125D-01, &
   5.892806934942636D-02,  -9.881366611685725D-03,  -7.452316170217832D-03, &
   8.672223915637222D-03,  -8.003289245966071D-03,   5.268932942057199D-03, &
   4.589999663127708D-04,  -7.490935351192785D-03,   1.454770447616140D-02, &
  -1.829060079582991D-02,   1.505139790806519D-02,  -6.340254154166453D-03, &
  -1.814619254868875D-03,   4.625470151873006D-03,   0.000000000000000D+00, &
  -3.732752366284262D-04,   0.000000000000000D+00,   1.252003205127755D-03, &
   0.000000000000000D+00,  -5.612607779731865D-03,   0.000000000000000D+00, &
   3.896556444148536D-02,   0.000000000000000D+00,  -7.625075596945748D-01, &
   0.000000000000000D+00,   1.733677790715147D+00,   0.000000000000000D+00, &
  -7.709702881307903D-01,   0.000000000000000D+00,   4.026190376641968D-02, &
   0.000000000000000D+00,  -5.998335557301818D-03,   0.000000000000000D+00, &
   1.403657172633524D-03,   0.000000000000000D+00,  -4.315246740108446D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.950999854711189D-05, &
   0.000000000000000D+00,   2.669952249501014D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.264714223447715D-05,   0.000000000000000D+00, &
  -1.180399941884553D-04,   0.000000000000000D+00,  -2.276485602205906D-05, &
   0.000000000000000D+00,  -9.864770942892905D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.424218780564872D-05,   0.000000000000000D+00, &
  -1.686285631263337D-06,   0.000000000000000D+00,   5.924132208333243D-04, &
   0.000000000000000D+00,   2.920340115961131D-05,   0.000000000000000D+00, &
   1.947659904109582D-04,   0.000000000000000D+00,   4.012305873888092D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.492199680365299D-05, &
   0.000000000000000D+00,  -4.300028359722536D-05,   0.000000000000000D+00, &
  -5.546825798338058D-04,   0.000000000000000D+00,  -4.394776533629071D-03, &
   0.000000000000000D+00,  -4.918333091185713D-05,   0.000000000000000D+00, &
  -1.743006147951648D-04,   0.000000000000000D+00,  -2.374500954523224D-04, &
   0.000000000000000D+00,   5.756908883232985D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.534687733216428D-05,   0.000000000000000D+00, &
   1.433009342827700D-04,   0.000000000000000D+00,   5.587567959572777D-04, &
   0.000000000000000D+00,   1.975193042591151D-03,   0.000000000000000D+00, &
   1.711616106842150D-02,   0.000000000000000D+00,   3.731707655437570D-05, &
   0.000000000000000D+00,  -5.254837377576721D-06,   0.000000000000000D+00, &
  -1.596594502224848D-04,   0.000000000000000D+00,  -1.185561609654426D-03, &
   0.000000000000000D+00,  -7.409981526841409D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.569334457161581D-06,   0.000000000000000D+00, &
  -7.098642833400902D-05,   0.000000000000000D+00,  -1.836987081654245D-04, &
   0.000000000000000D+00,  -3.573330529752329D-04,   0.000000000000000D+00, &
  -1.608503080212376D-03,   0.000000000000000D+00,  -3.316865404760185D-02, &
   0.000000000000000D+00,  -7.363447256517740D-06,   0.000000000000000D+00, &
   7.355129700120012D-05,   0.000000000000000D+00,   1.676587587868729D-04, &
   0.000000000000000D+00,   6.694857398902498D-04,   0.000000000000000D+00, &
   3.617806701637277D-03,   0.000000000000000D+00,   1.738329686942848D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.354230763732841D-06, &
   0.000000000000000D+00,  -3.350710546178353D-05,   0.000000000000000D+00, &
  -3.549333382432435D-05,   0.000000000000000D+00,  -2.651219596154639D-04, &
   0.000000000000000D+00,  -1.591895757207888D-03,   0.000000000000000D+00, &
  -1.111385023003458D-03,   0.000000000000000D+00,   3.444506158312022D-02, &
   0.000000000000000D+00,  -1.601198855031614D-06,   0.000000000000000D+00, &
  -6.219173452792473D-06,   0.000000000000000D+00,  -1.032021580701383D-05, &
   0.000000000000000D+00,   5.273284008844743D-06,   0.000000000000000D+00, &
   8.987438799018844D-06,   0.000000000000000D+00,  -5.020126116721040D-03, &
   0.000000000000000D+00,  -1.781881309897951D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.558855736413937D-07,   0.000000000000000D+00, &
   1.790566659627029D-05,   0.000000000000000D+00,  -1.287514659539121D-05, &
   0.000000000000000D+00,   3.250330898284701D-05,   0.000000000000000D+00, &
   2.757359058423789D-04,   0.000000000000000D+00,   3.692860020447049D-03, &
   0.000000000000000D+00,   8.030372572687001D-04,   0.000000000000000D+00, &
  -1.937749520631668D-02,   0.000000000000000D+00,  -2.144590180108382D-08, &
   0.000000000000000D+00,  -1.115148835168313D-05,   0.000000000000000D+00, &
   2.024324216812099D-05,   0.000000000000000D+00,  -3.310102483005611D-05, &
   0.000000000000000D+00,  -1.530622547050845D-04,   0.000000000000000D+00, &
  -1.359379367990366D-03,   0.000000000000000D+00,   5.421393377320578D-03, &
   0.000000000000000D+00,   7.944545619605030D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.427152147572568D-10,   0.000000000000000D+00, &
   3.795125392296143D-06,   0.000000000000000D+00,   3.623481949602663D-07, &
   0.000000000000000D+00,   2.925325759990340D-05,   0.000000000000000D+00, &
   6.668973198277617D-05,   0.000000000000000D+00,   3.424776264198547D-04, &
   0.000000000000000D+00,  -3.840742928368895D-03,   0.000000000000000D+00, &
   2.413720959165769D-03,   0.000000000000000D+00,   5.665993478030948D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.867049442568774D-07,   0.000000000000000D+00,  -1.461759960320827D-05, &
   0.000000000000000D+00,  -1.235233905097663D-06,   0.000000000000000D+00, &
  -2.445723832161944D-05,   0.000000000000000D+00,  -9.532425352321505D-05, &
   0.000000000000000D+00,   1.331163571199924D-03,   0.000000000000000D+00, &
  -4.315779903975792D-03,   0.000000000000000D+00,  -4.924758512898433D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.915594089767809D-08,   0.000000000000000D+00, &
   1.228492887453045D-05,   0.000000000000000D+00,  -1.783117378593785D-05, &
   0.000000000000000D+00,   1.555118412413053D-05,   0.000000000000000D+00, &
   3.872947466117457D-05,   0.000000000000000D+00,  -2.829545573219867D-04, &
   0.000000000000000D+00,   1.640482202957864D-03,   0.000000000000000D+00, &
  -2.759343153030523D-03,   0.000000000000000D+00,  -9.730043060993357D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.002743446785481D-09,   0.000000000000000D+00,  -5.361025885081817D-06, &
   0.000000000000000D+00,   7.697898241169488D-06,   0.000000000000000D+00, &
  -2.103148612403700D-05,   0.000000000000000D+00,  -2.156628139862103D-05, &
   0.000000000000000D+00,   5.557238740596404D-05,   0.000000000000000D+00, &
   2.168886015444206D-04,   0.000000000000000D+00,   1.400429975248260D-03, &
   0.000000000000000D+00,  -6.680269871112318D-04,   0.000000000000000D+00 /)
lem(:, 60) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.794986263736859D-04,   2.244388529095017D-03,   4.655489845780055D-03, &
   4.702619058152758D-03,  -4.933856640030094D-03,  -2.522267346138642D-02, &
  -1.319374652009759D-02,   8.685152543789779D-02,  -8.387450232262382D-02, &
   3.480360800229737D-02,  -3.398038698137209D-03,  -6.821122800317383D-03, &
   7.770634146650432D-03,  -6.168931123778670D-03,   4.797801455312015D-03, &
  -1.629225480904561D-03,  -4.094661896473301D-03,   1.030921142628746D-02, &
  -1.476196459359912D-02,   1.487976143445225D-02,  -9.404499525380045D-03, &
   1.401291025101188D-03,   5.794986263737957D-04,   2.244388529095140D-03, &
   4.655489845779960D-03,   4.702619058152940D-03,  -4.933856640030330D-03, &
  -2.522267346138606D-02,  -1.319374652009834D-02,   8.685152543789873D-02, &
  -8.387450232262460D-02,   3.480360800229771D-02,  -3.398038698137298D-03, &
  -6.821122800316859D-03,   7.770634146649537D-03,  -6.168931123778016D-03, &
   4.797801455311636D-03,  -1.629225480904097D-03,  -4.094661896473640D-03, &
   1.030921142628755D-02,  -1.476196459359903D-02,   1.487976143445232D-02, &
  -9.404499525380553D-03,   1.401291025101851D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.278513975904689D-04,   0.000000000000000D+00, &
   1.265829108811616D-03,   0.000000000000000D+00,  -5.712310933678395D-03, &
   0.000000000000000D+00,   3.941387782893296D-02,   0.000000000000000D+00, &
  -7.672431019386425D-01,   0.000000000000000D+00,   1.702229594136334D+00, &
   0.000000000000000D+00,  -7.739436436854455D-01,   0.000000000000000D+00, &
   4.043632944455110D-02,   0.000000000000000D+00,  -6.023478150621724D-03, &
   0.000000000000000D+00,   1.408711822217659D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.628659150480343D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.328240961853364D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.261500187223733D-06,   0.000000000000000D+00,  -1.270518589398611D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.860800478054713D-05, &
   0.000000000000000D+00,  -1.052032634034047D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.291456731485588D-05, &
   0.000000000000000D+00,   4.039714278877122D-06,   0.000000000000000D+00, &
   6.394213383542336D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.725270813565580D-05,   0.000000000000000D+00,   2.236636173276267D-04, &
   0.000000000000000D+00,   4.316377809594816D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.922761681807772D-05, &
   0.000000000000000D+00,  -6.048191941471902D-05,   0.000000000000000D+00, &
  -6.072800060142753D-04,   0.000000000000000D+00,  -4.660612673800838D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.037125519371164D-04, &
   0.000000000000000D+00,  -2.095530660436700D-04,   0.000000000000000D+00, &
  -2.577878235072896D-04,   0.000000000000000D+00,   6.111632524865731D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.195258701347978D-05,   0.000000000000000D+00,   1.582563420351143D-04, &
   0.000000000000000D+00,   6.090566725976731D-04,   0.000000000000000D+00, &
   2.099735705693194D-03,   0.000000000000000D+00,   1.792760025237640D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.540149078467009D-05, &
   0.000000000000000D+00,   2.274127865577465D-05,   0.000000000000000D+00, &
  -1.650732438304251D-04,   0.000000000000000D+00,  -1.254186302101828D-03, &
   0.000000000000000D+00,  -7.752584394490895D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.640802795093244D-05, &
   0.000000000000000D+00,  -6.478224442378990D-05,   0.000000000000000D+00, &
  -1.925155755865902D-04,   0.000000000000000D+00,  -3.840436676643128D-04, &
   0.000000000000000D+00,  -1.701387213182977D-03,   0.000000000000000D+00, &
  -3.452695806100701D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.121339628069625D-06,   0.000000000000000D+00,   3.490019952688066D-05, &
   0.000000000000000D+00,   1.755046058189668D-04,   0.000000000000000D+00, &
   7.073960988090158D-04,   0.000000000000000D+00,   3.794101064465602D-03, &
   0.000000000000000D+00,   1.808877776436639D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.121829299422285D-05, &
   0.000000000000000D+00,  -2.998042851333533D-05,   0.000000000000000D+00, &
  -6.134252986526779D-05,   0.000000000000000D+00,  -2.856975527836732D-04, &
   0.000000000000000D+00,  -1.662871156002590D-03,   0.000000000000000D+00, &
  -1.143517298251463D-03,   0.000000000000000D+00,   3.575770939474055D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.419625219200676D-06, &
   0.000000000000000D+00,   2.384088982630042D-05,   0.000000000000000D+00, &
  -5.952947009646554D-06,   0.000000000000000D+00,   4.029657653432801D-06, &
   0.000000000000000D+00,   5.713594942773833D-06,   0.000000000000000D+00, &
  -5.232536339396063D-03,   0.000000000000000D+00,  -1.850898906768039D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.998124792770684D-06,   0.000000000000000D+00,  -1.432450028665164D-06, &
   0.000000000000000D+00,   2.155903445905824D-05,   0.000000000000000D+00, &
   4.486391322066664D-05,   0.000000000000000D+00,   2.953532946867656D-04, &
   0.000000000000000D+00,   3.844449023820919D-03,   0.000000000000000D+00, &
   8.477767339889263D-04,   0.000000000000000D+00,  -2.009739456094397D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.314027406320748D-07, &
   0.000000000000000D+00,  -1.262064405468898D-05,   0.000000000000000D+00, &
   3.541134883600588D-07,   0.000000000000000D+00,  -2.298878149684640D-05, &
   0.000000000000000D+00,  -1.590841447488417D-04,   0.000000000000000D+00, &
  -1.413383826833486D-03,   0.000000000000000D+00,   5.618751461711570D-03, &
   0.000000000000000D+00,   8.261669692848944D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.009936535904505D-08, &
   0.000000000000000D+00,   1.092451148540841D-05,   0.000000000000000D+00, &
  -1.696198933648919D-05,   0.000000000000000D+00,   1.605666152778508D-05, &
   0.000000000000000D+00,   6.649183423115579D-05,   0.000000000000000D+00, &
   3.518231963464998D-04,   0.000000000000000D+00,  -3.987017595672918D-03, &
   0.000000000000000D+00,   2.475909232219228D-03,   0.000000000000000D+00, &
   5.871917148705073D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.670751532053100D-09,   0.000000000000000D+00,  -4.801139984800442D-06, &
   0.000000000000000D+00,   7.155563597939448D-06,   0.000000000000000D+00, &
  -2.094753471290551D-05,   0.000000000000000D+00,  -3.395927144438671D-05, &
   0.000000000000000D+00,  -9.961951261173978D-05,   0.000000000000000D+00, &
   1.385906682700681D-03,   0.000000000000000D+00,  -4.465479683023489D-03, &
   0.000000000000000D+00,  -5.240881732240409D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.724214983561654D-10, &
   0.000000000000000D+00,   1.298280232945372D-06,   0.000000000000000D+00, &
   7.600661446436790D-06,   0.000000000000000D+00,   7.314431318084812D-06, &
   0.000000000000000D+00,   1.326956598654126D-05,   0.000000000000000D+00, &
   3.988170151282608D-05,   0.000000000000000D+00,  -2.955026225118109D-04, &
   0.000000000000000D+00,   1.709375366491624D-03,   0.000000000000000D+00, &
  -2.849161814878216D-03,   0.000000000000000D+00,  -1.004114624384731D-03 /)
lem(:, 61) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.986506591962830D-04,  -1.877268697522642D-03, &
  -5.945952523750675D-03,  -1.106595713738014D-02,  -7.053559979936267D-03, &
   2.076761942827159D-02,   3.572620158776750D-02,  -8.135677770062504D-02, &
   6.100488591919310D-02,  -2.010418475329286D-02,  -1.057702005918542D-03, &
   5.561342920177481D-03,  -6.232813794056373D-03,   5.638921747988966D-03, &
  -4.047336887142651D-03,   2.172244824933927D-03,   1.393006340675320D-03, &
  -6.785267651711316D-03,   1.160375735590308D-02,  -1.339796158958497D-02, &
   1.079670523963362D-02,   0.000000000000000D+00,   2.986506591962456D-04, &
   1.877268697522688D-03,   5.945952523750553D-03,   1.106595713738041D-02, &
   7.053559979935639D-03,  -2.076761942827072D-02,  -3.572620158776822D-02, &
   8.135677770062526D-02,  -6.100488591919320D-02,   2.010418475329335D-02, &
   1.057702005917614D-03,  -5.561342920176789D-03,   6.232813794056047D-03, &
  -5.638921747988606D-03,   4.047336887142314D-03,  -2.172244824933721D-03, &
  -1.393006340675440D-03,   6.785267651711477D-03,  -1.160375735590347D-02, &
   1.339796158958559D-02,  -1.079670523963422D-02,   0.000000000000000D+00, &
   1.493253295978426D-04,   0.000000000000000D+00,  -3.956733814771676D-04, &
   0.000000000000000D+00,   1.324289225126035D-03,   0.000000000000000D+00, &
  -5.812303926672576D-03,   0.000000000000000D+00,   3.977179890414836D-02, &
   0.000000000000000D+00,  -7.709702881307903D-01,   0.000000000000000D+00, &
   1.677070225306918D+00,   0.000000000000000D+00,  -7.763621496707679D-01, &
   0.000000000000000D+00,   4.058219933266805D-02,   0.000000000000000D+00, &
  -6.046152238629331D-03,   0.000000000000000D+00,   1.414238975120059D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.180520384713449D-05, &
   0.000000000000000D+00,  -1.101819025732697D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.083122308279874D-06,   0.000000000000000D+00, &
   3.836691250319017D-05,   0.000000000000000D+00,   3.495566853436879D-06, &
   0.000000000000000D+00,   3.442673407628750D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.860213939548547D-05,   0.000000000000000D+00, &
   1.472584194178494D-05,   0.000000000000000D+00,  -1.451503473022778D-04, &
   0.000000000000000D+00,   1.609800524609504D-05,   0.000000000000000D+00, &
  -3.410477397136869D-05,   0.000000000000000D+00,  -1.207350393457018D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.176673035229230D-05, &
   0.000000000000000D+00,  -7.048626582753609D-05,   0.000000000000000D+00, &
   1.341500437174475D-06,   0.000000000000000D+00,   6.889812095284361D-04, &
   0.000000000000000D+00,  -2.067699340498438D-05,   0.000000000000000D+00, &
   7.048626582753627D-05,   0.000000000000000D+00,   2.472193662792945D-04, &
   0.000000000000000D+00,   4.658718001538245D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.013769308026331D-05,   0.000000000000000D+00, &
   5.060139649022129D-05,   0.000000000000000D+00,  -6.600182150898460D-05, &
   0.000000000000000D+00,  -6.496752467192191D-04,   0.000000000000000D+00, &
  -4.913362732439718D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.481725089948845D-05,   0.000000000000000D+00, &
  -2.229573726583948D-04,   0.000000000000000D+00,  -2.749941746163928D-04, &
   0.000000000000000D+00,   6.393782726493113D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.369238476484505D-05,   0.000000000000000D+00, &
   4.023612809115141D-05,   0.000000000000000D+00,   1.813217608163999D-04, &
   0.000000000000000D+00,   6.516371319387563D-04,   0.000000000000000D+00, &
   2.213337770532503D-03,   0.000000000000000D+00,   1.870345290122032D-02, &
   0.000000000000000D+00,   1.148780256428942D-05,   0.000000000000000D+00, &
   2.064740017743872D-05,   0.000000000000000D+00,   1.861194920680989D-05, &
   0.000000000000000D+00,  -1.803285880559766D-04,   0.000000000000000D+00, &
  -1.320158335068008D-03,   0.000000000000000D+00,  -8.075994453153702D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.202359159139984D-06, &
   0.000000000000000D+00,  -4.604878024320526D-05,   0.000000000000000D+00, &
  -9.226382292670143D-05,   0.000000000000000D+00,  -2.117919777385765D-04, &
   0.000000000000000D+00,  -4.082686318717155D-04,   0.000000000000000D+00, &
  -1.787690110826312D-03,   0.000000000000000D+00,  -3.583532416807945D-02, &
   0.000000000000000D+00,  -4.070468136539857D-06,   0.000000000000000D+00, &
   3.393776636274582D-05,   0.000000000000000D+00,   4.332427771607425D-05, &
   0.000000000000000D+00,   1.876907662384850D-04,   0.000000000000000D+00, &
   7.489019664007036D-04,   0.000000000000000D+00,   3.963718351107791D-03, &
   0.000000000000000D+00,   1.876568130201107D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.763757222909296D-06,   0.000000000000000D+00, &
  -6.193067599583796D-06,   0.000000000000000D+00,  -3.675160666351840D-07, &
   0.000000000000000D+00,  -5.631224214862313D-05,   0.000000000000000D+00, &
  -2.969089339902595D-04,   0.000000000000000D+00,  -1.730781211032026D-03, &
   0.000000000000000D+00,  -1.176478519744455D-03,   0.000000000000000D+00, &
   3.702432415953643D-02,   0.000000000000000D+00,  -4.038926047396397D-07, &
   0.000000000000000D+00,  -9.501425170169350D-06,   0.000000000000000D+00, &
   6.360430292384956D-06,   0.000000000000000D+00,   5.007376663089358D-06, &
   0.000000000000000D+00,   5.075362680775138D-06,   0.000000000000000D+00, &
  -5.392827692268255D-07,   0.000000000000000D+00,  -5.437795132584633D-03, &
   0.000000000000000D+00,  -1.917253253487636D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.321566929954437D-08,   0.000000000000000D+00, &
   9.128646742357214D-06,   0.000000000000000D+00,  -1.752278987178328D-05, &
   0.000000000000000D+00,   9.123805872321174D-06,   0.000000000000000D+00, &
   4.467444185591678D-05,   0.000000000000000D+00,   3.089564007954139D-04, &
   0.000000000000000D+00,   3.990164163591345D-03,   0.000000000000000D+00, &
   8.914345467957424D-04,   0.000000000000000D+00,  -2.079067972989783D-02, &
   0.000000000000000D+00,  -3.820067984358133D-09,   0.000000000000000D+00, &
  -4.100703709364830D-06,   0.000000000000000D+00,   7.280159442666784D-06, &
   0.000000000000000D+00,  -1.813862050676312D-05,   0.000000000000000D+00, &
  -3.250090382939770D-05,   0.000000000000000D+00,  -1.686897643731036D-04, &
   0.000000000000000D+00,  -1.463826639838616D-03,   0.000000000000000D+00, &
   5.810264628507311D-03,   0.000000000000000D+00,   8.564738507059971D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.161507142971809D-10, &
   0.000000000000000D+00,   1.117792545375140D-06,   0.000000000000000D+00, &
   6.738206901131966D-06,   0.000000000000000D+00,   6.449096606499967D-06, &
   0.000000000000000D+00,   1.422861129014091D-05,   0.000000000000000D+00, &
   6.992249267280667D-05,   0.000000000000000D+00,   3.641194118646957D-04, &
   0.000000000000000D+00,  -4.127807568499465D-03,   0.000000000000000D+00, &
   2.536202881383451D-03,   0.000000000000000D+00,   6.070006607814336D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.941405440826680D-07,   0.000000000000000D+00,  -9.469824202974018D-06, &
   0.000000000000000D+00,   1.135642144882996D-05,   0.000000000000000D+00, &
  -6.563888227537639D-06,   0.000000000000000D+00,  -3.091432465664181D-05, &
   0.000000000000000D+00,  -1.005391762554571D-04,   0.000000000000000D+00, &
   1.437706416749725D-03,   0.000000000000000D+00,  -4.610585399863202D-03, &
   0.000000000000000D+00,  -5.531286477696872D-04,   0.000000000000000D+00 /)
lem(:, 62) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.877608462753670D-04,  -1.114492965314779D-03,  -2.397610654729044D-03, &
  -3.033543490327975D-03,  -1.286472476424718D-04,   8.890030595312579D-03, &
   1.573129838219195D-02,  -8.337991843941123D-03,  -4.997440428411135D-02, &
   7.079766327745386D-02,  -4.129374379848779D-02,   1.100443782245060D-02, &
   2.507131083725348D-03,  -5.333284475850453D-03,   4.788831729028195D-03, &
  -4.610403748412990D-03,   3.951239439461586D-03,  -2.258996657527491D-03, &
  -1.186461381106319D-04,   3.871201784645290D-03,  -8.514180082936140D-03, &
   1.155807239709055D-02,  -2.877608462754166D-04,  -1.114492965314821D-03, &
  -2.397610654729162D-03,  -3.033543490327724D-03,  -1.286472476429799D-04, &
   8.890030595313321D-03,   1.573129838219129D-02,  -8.337991843940649D-03, &
  -4.997440428411153D-02,   7.079766327745432D-02,  -4.129374379848841D-02, &
   1.100443782245128D-02,   2.507131083724909D-03,  -5.333284475850087D-03, &
   4.788831729027851D-03,  -4.610403748412797D-03,   3.951239439461449D-03, &
  -2.258996657527253D-03,  -1.186461381111319D-04,   3.871201784646026D-03, &
  -8.514180082936662D-03,   1.155807239709076D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.198805327358500D-04,   0.000000000000000D+00, &
  -3.889017680958921D-04,   0.000000000000000D+00,   1.338089373599307D-03, &
   0.000000000000000D+00,  -5.871362444253836D-03,   0.000000000000000D+00, &
   4.004127397659218D-02,   0.000000000000000D+00,  -7.739436436854455D-01, &
   0.000000000000000D+00,   1.656619573235510D+00,   0.000000000000000D+00, &
  -7.783501782477856D-01,   0.000000000000000D+00,   4.070152188984712D-02, &
   0.000000000000000D+00,  -6.063802199148954D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.611826018083128D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.835112804498136D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.951546736585639D-06,   0.000000000000000D+00,   3.988745982250904D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.312163585747956D-06, &
   0.000000000000000D+00,   3.551844799610629D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.854402797428015D-05, &
   0.000000000000000D+00,   1.367338886410414D-05,   0.000000000000000D+00, &
  -1.554640041560773D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.133848655874142D-06,   0.000000000000000D+00,  -4.017908291551758D-05, &
   0.000000000000000D+00,  -1.289437289500162D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.684924989232429D-05, &
   0.000000000000000D+00,  -7.773978165750436D-05,   0.000000000000000D+00, &
   3.657647080307086D-06,   0.000000000000000D+00,   7.308992701248897D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.143429733637880D-06, &
   0.000000000000000D+00,   8.861874557851224D-05,   0.000000000000000D+00, &
   2.703167364164190D-04,   0.000000000000000D+00,   4.937451472242433D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.912729338394921D-05,   0.000000000000000D+00,   6.109770527033176D-05, &
   0.000000000000000D+00,  -7.387177474851710D-05,   0.000000000000000D+00, &
  -6.920254672790410D-04,   0.000000000000000D+00,  -5.148918362862719D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.873857152741592D-05, &
   0.000000000000000D+00,  -1.123557984359079D-04,   0.000000000000000D+00, &
  -2.451594435322668D-04,   0.000000000000000D+00,  -2.932691186618105D-04, &
   0.000000000000000D+00,   6.680481309962455D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.944901369802237D-05, &
   0.000000000000000D+00,   2.714432186339494D-05,   0.000000000000000D+00, &
   1.922986324448405D-04,   0.000000000000000D+00,   6.925592258541681D-04, &
   0.000000000000000D+00,   2.323279068837901D-03,   0.000000000000000D+00, &
   1.944411433312049D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.518437701413017D-05,   0.000000000000000D+00,   2.710535997174008D-05, &
   0.000000000000000D+00,   1.952777447807919D-05,   0.000000000000000D+00, &
  -1.888884244605555D-04,   0.000000000000000D+00,  -1.380809003683616D-03, &
   0.000000000000000D+00,  -8.387180043636727D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.355176956248313D-05, &
   0.000000000000000D+00,  -2.197532194991138D-05,   0.000000000000000D+00, &
  -9.136428958771743D-05,   0.000000000000000D+00,  -2.225335785644296D-04, &
   0.000000000000000D+00,  -4.318624944293020D-04,   0.000000000000000D+00, &
  -1.872065612449377D-03,   0.000000000000000D+00,  -3.709603648719308D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.178433417394286D-06, &
   0.000000000000000D+00,   1.991211018670905D-05,   0.000000000000000D+00, &
   5.769684328504801D-05,   0.000000000000000D+00,   2.013208529492142D-04, &
   0.000000000000000D+00,   7.854353487107504D-04,   0.000000000000000D+00, &
   4.125008938156947D-03,   0.000000000000000D+00,   1.942003665550443D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.506341123620217D-06,   0.000000000000000D+00,  -2.122776463286134D-05, &
   0.000000000000000D+00,  -1.346194890789629D-05,   0.000000000000000D+00, &
  -6.260495782843133D-05,   0.000000000000000D+00,  -3.123330083361087D-04, &
   0.000000000000000D+00,  -1.795835060413011D-03,   0.000000000000000D+00, &
  -1.207424349958248D-03,   0.000000000000000D+00,   3.824751055368562D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.167871064239224D-06, &
   0.000000000000000D+00,   8.498801540419126D-06,   0.000000000000000D+00, &
  -1.033737084275428D-05,   0.000000000000000D+00,  -2.525442529182631D-06, &
   0.000000000000000D+00,   2.350867060881136D-06,   0.000000000000000D+00, &
  -5.063169041524536D-06,   0.000000000000000D+00,  -5.634747913246654D-03, &
   0.000000000000000D+00,  -1.981432388877080D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.866617718059080D-07, &
   0.000000000000000D+00,   5.353116104992933D-06,   0.000000000000000D+00, &
   3.928739799470039D-06,   0.000000000000000D+00,   7.202873144067887D-06, &
   0.000000000000000D+00,   4.792405250441660D-05,   0.000000000000000D+00, &
   3.250914132433966D-04,   0.000000000000000D+00,   4.130920697096041D-03, &
   0.000000000000000D+00,   9.322666447999501D-04,   0.000000000000000D+00, &
  -2.145889700615451D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.556730546428769D-07,   0.000000000000000D+00,  -8.289894367128633D-06, &
   0.000000000000000D+00,   1.133277248053872D-05,   0.000000000000000D+00, &
  -4.663946109595923D-06,   0.000000000000000D+00,  -3.010358009880306D-05, &
   0.000000000000000D+00,  -1.747896230829385D-04,   0.000000000000000D+00, &
  -1.513336244440110D-03,   0.000000000000000D+00,   5.995031260027777D-03, &
   0.000000000000000D+00,   8.856941025781889D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.697817966690733D-08, &
   0.000000000000000D+00,   4.805904539436045D-06,   0.000000000000000D+00, &
  -1.001761981832768D-05,   0.000000000000000D+00,   1.208603807927189D-05, &
   0.000000000000000D+00,   1.952183386650109D-05,   0.000000000000000D+00, &
   7.321619138387159D-05,   0.000000000000000D+00,   3.741509227072016D-04, &
   0.000000000000000D+00,  -4.264130294994804D-03,   0.000000000000000D+00, &
   2.596060415746171D-03,   0.000000000000000D+00,   6.260171143938123D-03 /)
lem(:, 63) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.352802910205458D-04,   8.503495569322695D-04, &
   2.761837810446773D-03,   5.707523475917247D-03,   6.451878119401338D-03, &
  -2.195970898484108D-03,  -1.955163245104082D-02,  -8.475705271704293D-03, &
   5.538132200587017D-02,  -5.718534677938846D-02,   2.730359981607806D-02, &
  -4.710854468849879D-03,  -2.783723678910357D-03,   4.463385393954068D-03, &
  -4.348914064853915D-03,   3.632503997783493D-03,  -3.352437879492823D-03, &
   2.552567912135712D-03,  -6.230145875093863D-04,  -2.140316744364620D-03, &
   5.758841310606455D-03,   0.000000000000000D+00,  -1.352802910206002D-04, &
  -8.503495569318249D-04,  -2.761837810447361D-03,  -5.707523475916671D-03, &
  -6.451878119401728D-03,   2.195970898484389D-03,   1.955163245104049D-02, &
   8.475705271704844D-03,  -5.538132200587095D-02,   5.718534677938909D-02, &
  -2.730359981607838D-02,   4.710854468850286D-03,   2.783723678909953D-03, &
  -4.463385393953832D-03,   4.348914064853866D-03,  -3.632503997783307D-03, &
   3.352437879492321D-03,  -2.552567912134924D-03,   6.230145875085923D-04, &
   2.140316744364909D-03,  -5.758841310606438D-03,   0.000000000000000D+00, &
  -6.764014551072315D-05,   0.000000000000000D+00,   1.507479701014281D-04, &
   0.000000000000000D+00,  -4.112526546317832D-04,   0.000000000000000D+00, &
   1.364149781378617D-03,   0.000000000000000D+00,  -5.926717240130590D-03, &
   0.000000000000000D+00,   4.026190376641968D-02,   0.000000000000000D+00, &
  -7.763621496707679D-01,   0.000000000000000D+00,   1.639777501278064D+00, &
   0.000000000000000D+00,  -7.800081015143782D-01,   0.000000000000000D+00, &
   4.080308673857361D-02,   0.000000000000000D+00,  -6.079677281380242D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.347423026950253D-06, &
   0.000000000000000D+00,   5.082261554536547D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.756454192486483D-06,   0.000000000000000D+00, &
  -1.498162319120391D-05,   0.000000000000000D+00,   7.365596455650198D-08, &
   0.000000000000000D+00,  -1.392097730155896D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.142907444178037D-06,   0.000000000000000D+00, &
  -8.617747853347513D-06,   0.000000000000000D+00,   4.656898359212771D-05, &
   0.000000000000000D+00,  -9.961685914370442D-06,   0.000000000000000D+00, &
   5.514484467441736D-06,   0.000000000000000D+00,   4.164819857527016D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.889580148069186D-07, &
   0.000000000000000D+00,   2.469061248870142D-05,   0.000000000000000D+00, &
   1.704427349101257D-05,   0.000000000000000D+00,  -1.685985028744578D-04, &
   0.000000000000000D+00,   2.211864333506458D-05,   0.000000000000000D+00, &
   1.659007519967288D-05,   0.000000000000000D+00,  -4.372796781629507D-05, &
   0.000000000000000D+00,  -1.400568570783334D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.103689454008210D-07,   0.000000000000000D+00, &
  -5.178661833427089D-05,   0.000000000000000D+00,  -8.715908590729678D-05, &
   0.000000000000000D+00,   2.651371902246631D-06,   0.000000000000000D+00, &
   7.726364356723449D-04,   0.000000000000000D+00,  -2.603001787497926D-05, &
   0.000000000000000D+00,   2.258243308966610D-07,   0.000000000000000D+00, &
   9.577683376064203D-05,   0.000000000000000D+00,   2.897483594943366D-04, &
   0.000000000000000D+00,   5.223163112989417D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.062539900159998D-05,   0.000000000000000D+00, &
   5.132040036080490D-05,   0.000000000000000D+00,   6.204523491455513D-05, &
   0.000000000000000D+00,  -7.959882725939580D-05,   0.000000000000000D+00, &
  -7.292843105680581D-04,   0.000000000000000D+00,  -5.374150376091564D-03, &
   0.000000000000000D+00,   1.215323415215448D-05,   0.000000000000000D+00, &
  -4.185471856820876D-05,   0.000000000000000D+00,  -1.242598380772827D-04, &
   0.000000000000000D+00,  -2.609219038902108D-04,   0.000000000000000D+00, &
  -3.093825203688300D-04,   0.000000000000000D+00,   6.931616088110412D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.130849126010338D-05, &
   0.000000000000000D+00,   4.663600072884483D-06,   0.000000000000000D+00, &
   3.840077175108654D-05,   0.000000000000000D+00,   2.092063700989327D-04, &
   0.000000000000000D+00,   7.307034491272945D-04,   0.000000000000000D+00, &
   2.426442728009091D-03,   0.000000000000000D+00,   2.015681920963911D-02, &
   0.000000000000000D+00,   1.263083333122156D-06,   0.000000000000000D+00, &
   2.379130758068204D-05,   0.000000000000000D+00,   3.934282147989523D-05, &
   0.000000000000000D+00,   2.267690813302584D-05,   0.000000000000000D+00, &
  -1.993721500550431D-04,   0.000000000000000D+00,  -1.439353504880320D-03, &
   0.000000000000000D+00,  -8.684853434878915D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.607998563812215D-06,   0.000000000000000D+00, &
  -2.775593988228994D-05,   0.000000000000000D+00,  -3.613013476486561D-05, &
   0.000000000000000D+00,  -1.008036976352387D-04,   0.000000000000000D+00, &
  -2.374052351565895D-04,   0.000000000000000D+00,  -4.546504832646129D-04, &
   0.000000000000000D+00,  -1.952104693491101D-03,   0.000000000000000D+00, &
  -3.831589825911436D-02,   0.000000000000000D+00,  -1.724701286371560D-06, &
   0.000000000000000D+00,   1.203125060493881D-05,   0.000000000000000D+00, &
   6.061332389970149D-06,   0.000000000000000D+00,   5.464671961476784D-05, &
   0.000000000000000D+00,   2.107621811320952D-04,   0.000000000000000D+00, &
   8.217263041463989D-04,   0.000000000000000D+00,   4.280811080318264D-03, &
   0.000000000000000D+00,   2.005205406980712D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.425647640954340D-07,   0.000000000000000D+00, &
   2.990063882836013D-06,   0.000000000000000D+00,  -2.216951682783282D-06, &
   0.000000000000000D+00,  -1.682156397995440D-05,   0.000000000000000D+00, &
  -6.533783068958914D-05,   0.000000000000000D+00,  -3.240512208221560D-04, &
   0.000000000000000D+00,  -1.858120704770775D-03,   0.000000000000000D+00, &
  -1.238397336028898D-03,   0.000000000000000D+00,   3.943289824754460D-02, &
   0.000000000000000D+00,  -1.007827159868525D-07,   0.000000000000000D+00, &
  -6.709920164131742D-06,   0.000000000000000D+00,   1.244813719572357D-05, &
   0.000000000000000D+00,   2.638631440678765D-06,   0.000000000000000D+00, &
   1.078909898069225D-06,   0.000000000000000D+00,   2.660918767523557D-06, &
   0.000000000000000D+00,  -1.018214779668051D-05,   0.000000000000000D+00, &
  -5.825644392669009D-03,   0.000000000000000D+00,  -2.043503986830637D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.129216920495516D-08, &
   0.000000000000000D+00,   4.055410514042104D-06,   0.000000000000000D+00, &
  -1.003167084884444D-05,   0.000000000000000D+00,   9.143815590925439D-06, &
   0.000000000000000D+00,   1.205305730631416D-05,   0.000000000000000D+00, &
   5.106300780427987D-05,   0.000000000000000D+00,   3.387986146014003D-04, &
   0.000000000000000D+00,   4.266772584033841D-03,   0.000000000000000D+00, &
   9.719524403269025D-04,   0.000000000000000D+00,  -2.210574414873325D-02, &
   0.000000000000000D+00,  -7.087773263730010D-10,   0.000000000000000D+00, &
  -1.418238153580233D-06,   0.000000000000000D+00,  -1.009110409834493D-06, &
   0.000000000000000D+00,  -7.892542956971879D-06,   0.000000000000000D+00, &
  -7.739445604191605D-06,   0.000000000000000D+00,  -3.342028590137117D-05, &
   0.000000000000000D+00,  -1.829247302476908D-04,   0.000000000000000D+00, &
  -1.560886015184944D-03,   0.000000000000000D+00,   6.174728334599008D-03, &
   0.000000000000000D+00,   9.138457487147996D-03,   0.000000000000000D+00 /)
lem(:, 64) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.558791572557670D-04,   6.037173800710975D-04,   1.328750055563838D-03, &
   1.887102272009888D-03,   1.024894754007463D-03,  -2.882103181690664D-03, &
  -8.559055089266302D-03,  -6.104151233632039D-03,   1.588288908591875D-02, &
   2.295402673364709D-02,  -5.482739270240478D-02,   4.310732592391266D-02, &
  -1.729241956165931D-02,   1.570688108288850D-03,   3.267447841330807D-03, &
  -3.554469502254927D-03,   3.587170867229401D-03,  -3.344289727902568D-03, &
   2.746492518214684D-03,  -2.316097528184288D-03,   1.275408040639676D-03, &
   9.190922786248370D-04,   1.558791572555824D-04,   6.037173800715616D-04, &
   1.328750055563244D-03,   1.887102272010541D-03,   1.024894754006948D-03, &
  -2.882103181690363D-03,  -8.559055089266470D-03,  -6.104151233631705D-03, &
   1.588288908591804D-02,   2.295402673364797D-02,  -5.482739270240528D-02, &
   4.310732592391285D-02,  -1.729241956165965D-02,   1.570688108289172D-03, &
   3.267447841330626D-03,  -3.554469502254738D-03,   3.587170867229069D-03, &
  -3.344289727901832D-03,   2.746492518213866D-03,  -2.316097528183894D-03, &
   1.275408040639561D-03,   9.190922786250326D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.995300960777740D-05,   0.000000000000000D+00, &
   1.434553790231373D-04,   0.000000000000000D+00,  -4.113209170138435D-04, &
   0.000000000000000D+00,   1.374181153528935D-03,   0.000000000000000D+00, &
  -5.964184546676235D-03,   0.000000000000000D+00,   4.043632944455110D-02, &
   0.000000000000000D+00,  -7.783501782477856D-01,   0.000000000000000D+00, &
   1.625738424932772D+00,   0.000000000000000D+00,  -7.814024611950848D-01, &
   0.000000000000000D+00,   4.088828723502302D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.005151015500775D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.762414590320056D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.548732464409102D-06,   0.000000000000000D+00,  -1.496874621954752D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.811133682126108D-07, &
   0.000000000000000D+00,  -1.381730420266073D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.077746645900831D-06, &
   0.000000000000000D+00,  -8.060094118218340D-06,   0.000000000000000D+00, &
   4.925915628330066D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.807132532268631D-06,   0.000000000000000D+00,   7.210961202288758D-06, &
   0.000000000000000D+00,   4.393095837890043D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.482909032544240D-06, &
   0.000000000000000D+00,   2.735452791472742D-05,   0.000000000000000D+00, &
   1.728996885687670D-05,   0.000000000000000D+00,  -1.784437818513287D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.476578395165703D-05, &
   0.000000000000000D+00,   1.566672458499832D-05,   0.000000000000000D+00, &
  -4.842713939735434D-05,   0.000000000000000D+00,  -1.480830566708976D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.828392116932849D-05,   0.000000000000000D+00,  -6.350713981177748D-05, &
   0.000000000000000D+00,  -9.463019518467251D-05,   0.000000000000000D+00, &
   3.720569231308135D-06,   0.000000000000000D+00,   8.103566819901950D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.643026229576385D-05, &
   0.000000000000000D+00,   1.391511008827175D-06,   0.000000000000000D+00, &
   1.065063859872411D-04,   0.000000000000000D+00,   3.090031358273693D-04, &
   0.000000000000000D+00,   5.476741044684654D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.604635913950592D-05, &
   0.000000000000000D+00,   6.464635609296491D-05,   0.000000000000000D+00, &
   6.943428762245690D-05,   0.000000000000000D+00,  -8.495116358783548D-05, &
   0.000000000000000D+00,  -7.657387831401805D-04,   0.000000000000000D+00, &
  -5.587830469117038D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.335050338499289D-06,   0.000000000000000D+00,  -3.872112782468670D-05, &
   0.000000000000000D+00,  -1.332016133185535D-04,   0.000000000000000D+00, &
  -2.782498811736517D-04,   0.000000000000000D+00,  -3.258581498241561D-04, &
   0.000000000000000D+00,   7.179773324468064D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.247182353870431D-05, &
   0.000000000000000D+00,  -5.145121043346034D-06,   0.000000000000000D+00, &
   3.931410346108217D-05,   0.000000000000000D+00,   2.206049652289355D-04, &
   0.000000000000000D+00,   7.667030273580902D-04,   0.000000000000000D+00, &
   2.526177923954616D-03,   0.000000000000000D+00,   2.084278758027835D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.539703541341532D-05, &
   0.000000000000000D+00,   1.213126993322035D-05,   0.000000000000000D+00, &
   3.575930372282924D-05,   0.000000000000000D+00,   2.259346893872765D-05, &
   0.000000000000000D+00,  -2.077489549059392D-04,   0.000000000000000D+00, &
  -1.494373841637266D-03,   0.000000000000000D+00,  -8.972508613725930D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.465602096885839D-07,   0.000000000000000D+00,  -1.181034913372882D-05, &
   0.000000000000000D+00,  -4.135378511905947D-05,   0.000000000000000D+00, &
  -1.067921179017668D-04,   0.000000000000000D+00,  -2.488654932657664D-04, &
   0.000000000000000D+00,  -4.764293969421697D-04,   0.000000000000000D+00, &
  -2.030035319326738D-03,   0.000000000000000D+00,  -3.949730718822409D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.321381509608080D-06, &
   0.000000000000000D+00,   1.610412838677471D-05,   0.000000000000000D+00, &
   1.936368430588518D-05,   0.000000000000000D+00,   6.181272427128369D-05, &
   0.000000000000000D+00,   2.225386679690506D-04,   0.000000000000000D+00, &
   8.557736824749128D-04,   0.000000000000000D+00,   4.430387182831808D-03, &
   0.000000000000000D+00,   2.066520881746686D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.035852021564385D-06, &
   0.000000000000000D+00,  -1.098716665289833D-05,   0.000000000000000D+00, &
   2.166198383647698D-06,   0.000000000000000D+00,  -1.371537425726934D-05, &
   0.000000000000000D+00,  -6.788260790325924D-05,   0.000000000000000D+00, &
  -3.369581874910330D-04,   0.000000000000000D+00,  -1.918382877745401D-03, &
   0.000000000000000D+00,  -1.268068366414194D-03,   0.000000000000000D+00, &
   4.058284812047366D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.108601070008453D-06,   0.000000000000000D+00,  -1.203266963062804D-07, &
   0.000000000000000D+00,  -5.170475083981780D-06,   0.000000000000000D+00, &
   1.425625803807910D-07,   0.000000000000000D+00,  -8.073145769940777D-07, &
   0.000000000000000D+00,   7.540211221348963D-07,   0.000000000000000D+00, &
  -1.475495898503477D-05,   0.000000000000000D+00,  -6.010031080823063D-03, &
   0.000000000000000D+00,  -2.103762874746390D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.550997431553349D-07, &
   0.000000000000000D+00,   5.509218724131010D-06,   0.000000000000000D+00, &
  -5.910391891397722D-06,   0.000000000000000D+00,  -6.982375842867377D-07, &
   0.000000000000000D+00,   9.473261591523329D-06,   0.000000000000000D+00, &
   5.314799212660586D-05,   0.000000000000000D+00,   3.530983091672206D-04, &
   0.000000000000000D+00,   4.398621608632327D-03,   0.000000000000000D+00, &
   1.009675942996165D-03,   0.000000000000000D+00,  -2.273254435447944D-02 /)
lem(:, 65) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.731965636222378D-05,  -4.231602366321219D-04, &
  -1.396576105348634D-03,  -3.065503440751186D-03,  -4.317750360103861D-03, &
  -1.970113039318546D-03,   6.663619155147434D-03,   1.323553653614593D-02, &
  -6.235762572570342D-03,  -3.292902273124911D-02,   4.921333451110234D-02, &
  -3.148520082701334D-02,   9.947696299624373D-03,   5.284905613864993D-05, &
  -2.992005378764415D-03,   3.304753164243687D-03,  -2.851648585829372D-03, &
   2.795167529038011D-03,  -2.588265375157844D-03,   2.017228405380016D-03, &
  -1.371015152685889D-03,   0.000000000000000D+00,   6.731965636161447D-05, &
   4.231602366326143D-04,   1.396576105348153D-03,   3.065503440751614D-03, &
   4.317750360103513D-03,   1.970113039318921D-03,  -6.663619155147972D-03, &
  -1.323553653614521D-02,   6.235762572569737D-03,   3.292902273124947D-02, &
  -4.921333451110285D-02,   3.148520082701364D-02,  -9.947696299624432D-03, &
  -5.284905613844111D-05,   2.992005378763941D-03,  -3.304753164242944D-03, &
   2.851648585828668D-03,  -2.795167529037633D-03,   2.588265375157775D-03, &
  -2.017228405379874D-03,   1.371015152685546D-03,   0.000000000000000D+00, &
   3.365982818001015D-05,   0.000000000000000D+00,  -6.578508510270316D-05, &
   0.000000000000000D+00,   1.540530374277926D-04,   0.000000000000000D+00, &
  -4.210132796760198D-04,   0.000000000000000D+00,   1.388151574579140D-03, &
   0.000000000000000D+00,  -5.998335557301818D-03,   0.000000000000000D+00, &
   4.058219933266805D-02,   0.000000000000000D+00,  -7.800081015143782D-01, &
   0.000000000000000D+00,   1.613915870151324D+00,   0.000000000000000D+00, &
  -7.825882917195222D-01,   0.000000000000000D+00,   4.096188552912975D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.661043067535928D-06, &
   0.000000000000000D+00,  -2.558695257248841D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.046956205798104D-06,   0.000000000000000D+00, &
   6.678194621417613D-06,   0.000000000000000D+00,  -5.976253258188272D-07, &
   0.000000000000000D+00,   6.331428788651490D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.066280617531475D-06,   0.000000000000000D+00, &
   4.597599624823781D-06,   0.000000000000000D+00,  -1.794352147797413D-05, &
   0.000000000000000D+00,   4.725176528209905D-06,   0.000000000000000D+00, &
  -1.938831521104293D-07,   0.000000000000000D+00,  -1.663241382080069D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.388929242816626D-06, &
   0.000000000000000D+00,  -8.088876677930138D-06,   0.000000000000000D+00, &
  -1.007687553496874D-05,   0.000000000000000D+00,   5.389620928164262D-05, &
   0.000000000000000D+00,  -5.506946116099904D-06,   0.000000000000000D+00, &
  -1.205677143100458D-05,   0.000000000000000D+00,   7.348005315930059D-06, &
   0.000000000000000D+00,   4.814673029565764D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.185435077423712D-05,   0.000000000000000D+00, &
   3.322341818790479D-06,   0.000000000000000D+00,   3.062919259690281D-05, &
   0.000000000000000D+00,   1.939933536578754D-05,   0.000000000000000D+00, &
  -1.892120874494825D-04,   0.000000000000000D+00,   8.301981996208950D-06, &
   0.000000000000000D+00,   2.945324079560668D-05,   0.000000000000000D+00, &
   1.883884115583418D-05,   0.000000000000000D+00,  -5.154900577984603D-05, &
   0.000000000000000D+00,  -1.571214952016887D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.355377430553441D-05,   0.000000000000000D+00, &
  -1.659388002359830D-05,   0.000000000000000D+00,  -6.824469682452056D-05, &
   0.000000000000000D+00,  -1.017397766851266D-04,   0.000000000000000D+00, &
   3.331466037251363D-06,   0.000000000000000D+00,   8.472569410515666D-04, &
   0.000000000000000D+00,  -1.304934581196278D-05,   0.000000000000000D+00, &
  -2.193132855477287D-05,   0.000000000000000D+00,   3.951056186630679D-06, &
   0.000000000000000D+00,   1.139366341253233D-04,   0.000000000000000D+00, &
   3.260582332476007D-04,   0.000000000000000D+00,   5.728421434536498D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.764937307093773D-06, &
   0.000000000000000D+00,   3.234964037244822D-05,   0.000000000000000D+00, &
   6.851865103450766D-05,   0.000000000000000D+00,   7.247816324885902D-05, &
   0.000000000000000D+00,  -8.986673713906255D-05,   0.000000000000000D+00, &
  -7.992318445360026D-04,   0.000000000000000D+00,  -5.793340155855630D-03, &
   0.000000000000000D+00,   9.065091768534407D-06,   0.000000000000000D+00, &
  -1.486503533105199D-05,   0.000000000000000D+00,  -4.913768472345652D-05, &
   0.000000000000000D+00,  -1.444214719607031D-04,   0.000000000000000D+00, &
  -2.933594995579017D-04,   0.000000000000000D+00,  -3.409598420867472D-04, &
   0.000000000000000D+00,   7.407570551517811D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.243067135072303D-06,   0.000000000000000D+00, &
  -8.269656094680297D-06,   0.000000000000000D+00,  -7.342860718890911D-06, &
   0.000000000000000D+00,   4.271338537347143D-05,   0.000000000000000D+00, &
   2.338427943394171D-04,   0.000000000000000D+00,   8.011261519163142D-04, &
   0.000000000000000D+00,   2.621203253097821D-03,   0.000000000000000D+00, &
   2.150619327436531D-02,   0.000000000000000D+00,  -1.173973501160155D-06, &
   0.000000000000000D+00,   1.956272092770086D-05,   0.000000000000000D+00, &
   2.496674647712782D-05,   0.000000000000000D+00,   4.177912867892635D-05, &
   0.000000000000000D+00,   2.528902105679230D-05,   0.000000000000000D+00, &
  -2.162547507299201D-04,   0.000000000000000D+00,  -1.547616454622814D-03, &
   0.000000000000000D+00,  -9.249881417200206D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.493840237366546D-06,   0.000000000000000D+00, &
  -1.355455480811528D-05,   0.000000000000000D+00,  -8.765258707094589D-06, &
   0.000000000000000D+00,  -4.044168330856233D-05,   0.000000000000000D+00, &
  -1.124757155199695D-04,   0.000000000000000D+00,  -2.615300388924301D-04, &
   0.000000000000000D+00,  -4.976935463551129D-04,   0.000000000000000D+00, &
  -2.104797194221069D-03,   0.000000000000000D+00,  -4.064464963563603D-02, &
   0.000000000000000D+00,  -6.428551712400412D-07,   0.000000000000000D+00, &
   1.637163257806285D-06,   0.000000000000000D+00,   4.986468141569819D-07, &
   0.000000000000000D+00,   1.827758680973248D-05,   0.000000000000000D+00, &
   6.362029949156743D-05,   0.000000000000000D+00,   2.317164512394880D-04, &
   0.000000000000000D+00,   8.888490960049011D-04,   0.000000000000000D+00, &
   4.575366663407179D-03,   0.000000000000000D+00,   2.126015578448134D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.598266002258679D-07, &
   0.000000000000000D+00,   4.186843670077697D-06,   0.000000000000000D+00, &
  -7.434405522630568D-06,   0.000000000000000D+00,  -7.475050367727129D-06, &
   0.000000000000000D+00,  -1.755000432141532D-05,   0.000000000000000D+00, &
  -7.118600235728083D-05,   0.000000000000000D+00,  -3.482902836319618D-04, &
   0.000000000000000D+00,  -1.976415192650958D-03,   0.000000000000000D+00, &
  -1.297471180312458D-03,   0.000000000000000D+00,   4.170110626946237D-02, &
   0.000000000000000D+00,  -2.498771740148417D-08,   0.000000000000000D+00, &
  -3.668812445943319D-06,   0.000000000000000D+00,   9.775319625484858D-06, &
   0.000000000000000D+00,  -2.673731161777537D-06,   0.000000000000000D+00, &
  -9.427681571708486D-07,   0.000000000000000D+00,  -5.537807086355188D-07, &
   0.000000000000000D+00,   3.694583461297502D-07,   0.000000000000000D+00, &
  -1.933132163947558D-05,   0.000000000000000D+00,  -6.189253803310088D-03, &
   0.000000000000000D+00,  -2.162290796105441D-02,   0.000000000000000D+00 /)
lem(:, 66) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.034263192740976D-05,  -3.498955089095780D-04,  -7.820036819937914D-04, &
  -1.190537000845088D-03,  -9.865632099430414D-04,   7.420984415454872D-04, &
   4.117245959014424D-03,   5.883391714001500D-03,  -1.094539600296598D-03, &
  -1.532708325375160D-02,  -4.895259339684902D-03,   3.833999049101865D-02, &
  -4.095310032826884D-02,   2.198459648521284D-02,  -5.547430000262826D-03, &
  -1.276391566881921D-03,   2.547830730927814D-03,  -2.777893418037593D-03, &
   2.640094985499672D-03,  -2.261908182880610D-03,   2.195980699706450D-03, &
  -1.979687205479768D-03,  -9.034263192794420D-05,  -3.498955089091397D-04, &
  -7.820036819941662D-04,  -1.190537000844730D-03,  -9.865632099434359D-04, &
   7.420984415459507D-04,   4.117245959013822D-03,   5.883391714002136D-03, &
  -1.094539600296965D-03,  -1.532708325375128D-02,  -4.895259339685391D-03, &
   3.833999049101904D-02,  -4.095310032826909D-02,   2.198459648521290D-02, &
  -5.547430000263161D-03,  -1.276391566881151D-03,   2.547830730926940D-03, &
  -2.777893418037196D-03,   2.640094985499741D-03,  -2.261908182880474D-03, &
   2.195980699705980D-03,  -1.979687205479406D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.300010829297768D-05,   0.000000000000000D+00, &
  -6.042811620135665D-05,   0.000000000000000D+00,   1.520695147793388D-04, &
   0.000000000000000D+00,  -4.222870447382378D-04,   0.000000000000000D+00, &
   1.395220904585443D-03,   0.000000000000000D+00,  -6.023478150621724D-03, &
   0.000000000000000D+00,   4.070152188984712D-02,   0.000000000000000D+00, &
  -7.814024611950848D-01,   0.000000000000000D+00,   1.603864665676498D+00, &
   0.000000000000000D+00,  -7.836037806651482D-01,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.844111252024118D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.756816044834179D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.320340008847312D-06,   0.000000000000000D+00,   6.392555276997201D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.593970011782093D-08, &
   0.000000000000000D+00,   6.025518610379915D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.064333252001734D-06, &
   0.000000000000000D+00,   4.190335277268439D-06,   0.000000000000000D+00, &
  -1.866617047632839D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.810237395932537D-06,   0.000000000000000D+00,  -8.601585373621863D-07, &
   0.000000000000000D+00,  -1.725456729632748D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.278676387231042D-06, &
   0.000000000000000D+00,  -9.016371447851659D-06,   0.000000000000000D+00, &
  -1.019655139628258D-05,   0.000000000000000D+00,   5.676575446100118D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.718681899959866D-06, &
   0.000000000000000D+00,  -1.229958607458120D-05,   0.000000000000000D+00, &
   8.539058449411944D-06,   0.000000000000000D+00,   5.065250812354608D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.367866922285164D-06,   0.000000000000000D+00,   6.223916808553403D-06, &
   0.000000000000000D+00,   3.336840695258702D-05,   0.000000000000000D+00, &
   2.008627643626479D-05,   0.000000000000000D+00,  -1.983310805544601D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.880666120041382D-05, &
   0.000000000000000D+00,   3.504684719941087D-05,   0.000000000000000D+00, &
   1.964395630209555D-05,   0.000000000000000D+00,  -5.532211606773223D-05, &
   0.000000000000000D+00,  -1.646270684355061D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.329681943248792D-06, &
   0.000000000000000D+00,  -2.194607441417344D-05,   0.000000000000000D+00, &
  -7.606488083597397D-05,   0.000000000000000D+00,  -1.084146172416721D-04, &
   0.000000000000000D+00,   3.892992159244525D-06,   0.000000000000000D+00, &
   8.816906184473624D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.636860852382162D-05,   0.000000000000000D+00,  -2.969999144511627D-05, &
   0.000000000000000D+00,   3.435538328311300D-06,   0.000000000000000D+00, &
   1.219362841455837D-04,   0.000000000000000D+00,   3.428255108447940D-04, &
   0.000000000000000D+00,   5.961091403280394D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.127744957970897D-05, &
   0.000000000000000D+00,   3.472517605916126D-05,   0.000000000000000D+00, &
   7.452794032367416D-05,   0.000000000000000D+00,   7.785787537576455D-05, &
   0.000000000000000D+00,  -9.420087654867651D-05,   0.000000000000000D+00, &
  -8.317982502685373D-04,   0.000000000000000D+00,  -5.990377113578255D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.127744957971038D-05, &
   0.000000000000000D+00,  -6.424381654624102D-06,   0.000000000000000D+00, &
  -5.019230063081975D-05,   0.000000000000000D+00,  -1.524243082322084D-04, &
   0.000000000000000D+00,  -3.083692509212223D-04,   0.000000000000000D+00, &
  -3.560613227227485D-04,   0.000000000000000D+00,   7.630611179105693D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.387027195935052D-05,   0.000000000000000D+00,  -5.262997753916541D-06, &
   0.000000000000000D+00,  -4.133408731843569D-06,   0.000000000000000D+00, &
   4.582723851614364D-05,   0.000000000000000D+00,   2.447546006549075D-04, &
   0.000000000000000D+00,   8.338003447535647D-04,   0.000000000000000D+00, &
   2.713246882354356D-03,   0.000000000000000D+00,   2.214828305363794D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.221966845851634D-06, &
   0.000000000000000D+00,   6.197574336805414D-06,   0.000000000000000D+00, &
   2.463659493133539D-05,   0.000000000000000D+00,   4.287565311643191D-05, &
   0.000000000000000D+00,   2.599598904187262D-05,   0.000000000000000D+00, &
  -2.239398471700363D-04,   0.000000000000000D+00,  -1.598362518858990D-03, &
   0.000000000000000D+00,  -9.518950425607105D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.134567330030225D-06, &
   0.000000000000000D+00,  -1.080129150151648D-05,   0.000000000000000D+00, &
  -1.859626300136525D-05,   0.000000000000000D+00,  -4.595071762393019D-05, &
   0.000000000000000D+00,  -1.187698644684998D-04,   0.000000000000000D+00, &
  -2.727339288221702D-04,   0.000000000000000D+00,  -5.180690714849957D-04, &
   0.000000000000000D+00,  -2.177577776561889D-03,   0.000000000000000D+00, &
  -4.176002821276301D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.618342844913530D-06,   0.000000000000000D+00,   1.068672731480960D-05, &
   0.000000000000000D+00,   3.228937965601645D-06,   0.000000000000000D+00, &
   1.842504713885620D-05,   0.000000000000000D+00,   6.729569509706442D-05, &
   0.000000000000000D+00,   2.419213116264741D-04,   0.000000000000000D+00, &
   9.205750964586542D-04,   0.000000000000000D+00,   4.715456337784409D-03, &
   0.000000000000000D+00,   2.183911723260118D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.255491717231636D-06, &
   0.000000000000000D+00,  -3.229315414398279D-06,   0.000000000000000D+00, &
   4.339061090294847D-06,   0.000000000000000D+00,  -3.570665105505552D-06, &
   0.000000000000000D+00,  -1.648078088249850D-05,   0.000000000000000D+00, &
  -7.329176417530845D-05,   0.000000000000000D+00,  -3.598103423642412D-04, &
   0.000000000000000D+00,  -2.032819754853376D-03,   0.000000000000000D+00, &
  -1.325933067517771D-03,   0.000000000000000D+00,   4.278965703332629D-02 /)
lem(:, 67) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.606531998327588D-05,   2.267006423280532D-04, &
   7.564144596072882D-04,   1.725731632955080D-03,   2.734901963398710D-03, &
   2.354360557592165D-03,  -1.395365473396638D-03,  -7.391723079014793D-03, &
  -5.975076854476972D-03,   1.201843481102297D-02,   1.520928383015175D-02, &
  -3.875657151315578D-02,   3.266773123804163D-02,  -1.443844386307562D-02, &
   2.783411749604467D-03,   1.591117992929471D-03,  -2.482038934590728D-03, &
   2.251598432978858D-03,  -2.233185360700296D-03,   2.113284129077728D-03, &
  -1.815820772660579D-03,   0.000000000000000D+00,  -3.606531998358684D-05, &
  -2.267006423276962D-04,  -7.564144596076198D-04,  -1.725731632954708D-03, &
  -2.734901963399247D-03,  -2.354360557591487D-03,   1.395365473396059D-03, &
   7.391723079015117D-03,   5.975076854476663D-03,  -1.201843481102258D-02, &
  -1.520928383015204D-02,   3.875657151315604D-02,  -3.266773123804208D-02, &
   1.443844386307628D-02,  -2.783411749605161D-03,  -1.591117992928992D-03, &
   2.482038934590601D-03,  -2.251598432978756D-03,   2.233185360699975D-03, &
  -2.113284129077396D-03,   1.815820772660331D-03,   0.000000000000000D+00, &
  -1.803265999247915D-05,   0.000000000000000D+00,   3.182356319092472D-05, &
   0.000000000000000D+00,  -6.617550630029674D-05,   0.000000000000000D+00, &
   1.566613964112046D-04,   0.000000000000000D+00,  -4.272976869277262D-04, &
   0.000000000000000D+00,   1.403657172633524D-03,   0.000000000000000D+00, &
  -6.046152238629331D-03,   0.000000000000000D+00,   4.080308673857361D-02, &
   0.000000000000000D+00,  -7.825882917195222D-01,   0.000000000000000D+00, &
   1.595249255593745D+00,   0.000000000000000D+00,  -7.844811332740665D-01, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.425606946135812D-06, &
   0.000000000000000D+00,   1.381742117026409D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.162417971464778D-06,   0.000000000000000D+00, &
  -3.289862183394183D-06,   0.000000000000000D+00,   5.333674001363997D-07, &
   0.000000000000000D+00,  -3.158990742692055D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.832790688585429D-07,   0.000000000000000D+00, &
  -2.504633539179840D-06,   0.000000000000000D+00,   7.898832569171176D-06, &
   0.000000000000000D+00,  -2.172536029268544D-06,   0.000000000000000D+00, &
  -6.572493900466436D-07,   0.000000000000000D+00,   7.474946480151689D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.532350326804549D-06, &
   0.000000000000000D+00,   2.693348710580975D-06,   0.000000000000000D+00, &
   5.355516035081859D-06,   0.000000000000000D+00,  -2.066265428633214D-05, &
   0.000000000000000D+00,   3.357346535855266D-07,   0.000000000000000D+00, &
   5.838437238452558D-06,   0.000000000000000D+00,  -4.898640943390558D-07, &
   0.000000000000000D+00,  -1.913387009190644D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.578425286868491D-06,   0.000000000000000D+00, &
   3.507499220139913D-06,   0.000000000000000D+00,  -1.002933466581737D-05, &
   0.000000000000000D+00,  -1.148994911980160D-05,   0.000000000000000D+00, &
   6.044157935174118D-05,   0.000000000000000D+00,   4.297909698561121D-06, &
   0.000000000000000D+00,  -8.129830019080735D-06,   0.000000000000000D+00, &
  -1.437696670898045D-05,   0.000000000000000D+00,   8.813459535171990D-06, &
   0.000000000000000D+00,   5.397190776653149D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.313443181082666D-06,   0.000000000000000D+00, &
  -1.306951449198597D-05,   0.000000000000000D+00,   5.576483871213825D-06, &
   0.000000000000000D+00,   3.582275666278567D-05,   0.000000000000000D+00, &
   2.162354984045446D-05,   0.000000000000000D+00,  -2.077162438088464D-04, &
   0.000000000000000D+00,  -4.616773264030043D-06,   0.000000000000000D+00, &
   1.794853393458622D-05,   0.000000000000000D+00,   3.761077256929778D-05, &
   0.000000000000000D+00,   2.149094633463822D-05,   0.000000000000000D+00, &
  -5.819443392782223D-05,   0.000000000000000D+00,  -1.724721203612428D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.132078131330093D-05, &
   0.000000000000000D+00,   4.831759129786107D-06,   0.000000000000000D+00, &
  -2.449142368303905D-05,   0.000000000000000D+00,  -8.099128050135821D-05, &
   0.000000000000000D+00,  -1.145059127306763D-04,   0.000000000000000D+00, &
   3.748496876840082D-06,   0.000000000000000D+00,   9.152121814316771D-04, &
   0.000000000000000D+00,  -1.509681201934909D-06,   0.000000000000000D+00, &
  -2.095401195277082D-05,   0.000000000000000D+00,  -2.927834683283984D-05, &
   0.000000000000000D+00,   5.138462043292760D-06,   0.000000000000000D+00, &
   1.287689759826506D-04,   0.000000000000000D+00,   3.582072880108011D-04, &
   0.000000000000000D+00,   6.189383200403241D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.049691015019519D-06,   0.000000000000000D+00, &
   1.668900087604266D-05,   0.000000000000000D+00,   4.163730496727843D-05, &
   0.000000000000000D+00,   8.016223709192293D-05,   0.000000000000000D+00, &
   8.153960579027803D-05,   0.000000000000000D+00,  -9.842159028282533D-05, &
   0.000000000000000D+00,  -8.624309869164701D-04,   0.000000000000000D+00, &
  -6.180815324432303D-03,   0.000000000000000D+00,   4.002665656464316D-06, &
   0.000000000000000D+00,  -1.272080044244692D-06,   0.000000000000000D+00, &
  -9.502516331454614D-06,   0.000000000000000D+00,  -5.536901011101991D-05, &
   0.000000000000000D+00,  -1.617846802349981D-04,   0.000000000000000D+00, &
  -3.223853516703373D-04,   0.000000000000000D+00,  -3.702417415946628D-04, &
   0.000000000000000D+00,   7.840567529058775D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.376360369805256D-08,   0.000000000000000D+00, &
  -1.167994397953411D-05,   0.000000000000000D+00,  -1.399430128425244D-05, &
   0.000000000000000D+00,  -7.088640588986829D-06,   0.000000000000000D+00, &
   4.787998778669208D-05,   0.000000000000000D+00,   2.560317753143088D-04, &
   0.000000000000000D+00,   8.653313879379439D-04,   0.000000000000000D+00, &
   2.801757046579681D-03,   0.000000000000000D+00,   2.277171294829318D-02, &
   0.000000000000000D+00,  -1.111372324901234D-06,   0.000000000000000D+00, &
   1.241753975194047D-05,   0.000000000000000D+00,   9.245981688240721D-06, &
   0.000000000000000D+00,   2.541391117956528D-05,   0.000000000000000D+00, &
   4.580553861293856D-05,   0.000000000000000D+00,   2.809161284726717D-05, &
   0.000000000000000D+00,  -2.313772649202183D-04,   0.000000000000000D+00, &
  -1.647599463259163D-03,   0.000000000000000D+00,  -9.779795904545897D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.813280348903170D-07, &
   0.000000000000000D+00,  -4.517099011779443D-06,   0.000000000000000D+00, &
  -4.429807165872065D-07,   0.000000000000000D+00,  -1.580863745887231D-05, &
   0.000000000000000D+00,  -4.675037911154501D-05,   0.000000000000000D+00, &
  -1.238162236200652D-04,   0.000000000000000D+00,  -2.841491664550914D-04, &
   0.000000000000000D+00,  -5.379960281468562D-04,   0.000000000000000D+00, &
  -2.247885918476053D-03,   0.000000000000000D+00,  -4.284649204702546D-02, &
   0.000000000000000D+00,  -2.217436211385796D-07,   0.000000000000000D+00, &
  -1.817106859081354D-06,   0.000000000000000D+00,   3.564260233828023D-06, &
   0.000000000000000D+00,   8.618714838289651D-06,   0.000000000000000D+00, &
   2.115716614846337D-05,   0.000000000000000D+00,   7.023752254839831D-05, &
   0.000000000000000D+00,   2.507695887441527D-04,   0.000000000000000D+00, &
   9.512961046519366D-04,   0.000000000000000D+00,   4.851648968525008D-03, &
   0.000000000000000D+00,   2.240281357861489D-02,   0.000000000000000D+00 /)
lem(:, 68) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.526802732024017D-05,   2.140521493879518D-04,   4.836194827765232D-04, &
   7.707975216812961D-04,   7.795760120681173D-04,  -2.788855207871053D-07, &
  -1.885319786831606D-03,  -3.849723129055474D-03,  -2.481054238025312D-03, &
   5.103969485703114D-03,   1.081879099747674D-02,  -5.491871809320066D-03, &
  -2.317916179460707D-02,   3.570332111410057D-02,  -2.477348566336703D-02, &
   9.305967780964928D-03,  -8.200419210562402D-04,  -1.599253772401908D-03, &
   2.142825449789546D-03,  -2.107874182889752D-03,   1.825058537446202D-03, &
  -1.804339538531680D-03,   5.526802732004671D-05,   2.140521493882725D-04, &
   4.836194827762042D-04,   7.707975216817022D-04,   7.795760120675948D-04, &
  -2.788855202375729D-07,  -1.885319786832090D-03,  -3.849723129055096D-03, &
  -2.481054238025787D-03,   5.103969485703548D-03,   1.081879099747661D-02, &
  -5.491871809319917D-03,  -2.317916179460770D-02,   3.570332111410141D-02, &
  -2.477348566336770D-02,   9.305967780965239D-03,  -8.200419210562890D-04, &
  -1.599253772401667D-03,   2.142825449789002D-03,  -2.107874182889425D-03, &
   1.825058537446123D-03,  -1.804339538531316D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.146017731756322D-05,   0.000000000000000D+00, &
   2.815754139096722D-05,   0.000000000000000D+00,  -6.429395032336220D-05, &
   0.000000000000000D+00,   1.561988267437408D-04,   0.000000000000000D+00, &
  -4.285988121049305D-04,   0.000000000000000D+00,   1.408711822217659D-03, &
   0.000000000000000D+00,  -6.063802199148954D-03,   0.000000000000000D+00, &
   4.088828723502302D-02,   0.000000000000000D+00,  -7.836037806651482D-01, &
   0.000000000000000D+00,   1.587807570793765D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.188583666756751D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.838542384292053D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.088335971534997D-07,   0.000000000000000D+00,  -3.011813535223896D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.106979258392655D-07, &
   0.000000000000000D+00,  -2.877182272719257D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.236430360302114D-07, &
   0.000000000000000D+00,  -2.204025960126812D-06,   0.000000000000000D+00, &
   8.052379955697771D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.737355246759321D-06,   0.000000000000000D+00,  -3.107738309768980D-07, &
   0.000000000000000D+00,   7.599724407759373D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.512582234380177D-06, &
   0.000000000000000D+00,   3.067629412816832D-06,   0.000000000000000D+00, &
   5.336446668031503D-06,   0.000000000000000D+00,  -2.160178591419347D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.590470379488270D-06, &
   0.000000000000000D+00,   5.979255450154043D-06,   0.000000000000000D+00, &
  -8.971169026884724D-07,   0.000000000000000D+00,  -1.997886791552477D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.589711831286089D-06,   0.000000000000000D+00,   2.669840229079386D-06, &
   0.000000000000000D+00,  -1.100284387248639D-05,   0.000000000000000D+00, &
  -1.186427792221932D-05,   0.000000000000000D+00,   6.322634909188315D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.969312796727259D-07, &
   0.000000000000000D+00,  -1.053942249823400D-05,   0.000000000000000D+00, &
  -1.527002156038713D-05,   0.000000000000000D+00,   9.679920748096870D-06, &
   0.000000000000000D+00,   5.642998006249055D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.367348759942164D-05, &
   0.000000000000000D+00,  -1.450843241203623D-05,   0.000000000000000D+00, &
   7.001964077018367D-06,   0.000000000000000D+00,   3.826320526125638D-05, &
   0.000000000000000D+00,   2.244813491558906D-05,   0.000000000000000D+00, &
  -2.161559471234447D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.965676030168019D-06,   0.000000000000000D+00,   2.359110337356395D-05, &
   0.000000000000000D+00,   4.201507143958421D-05,   0.000000000000000D+00, &
   2.265737133960794D-05,   0.000000000000000D+00,  -6.139049080821160D-05, &
   0.000000000000000D+00,  -1.794501077588843D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.577956931606573D-05, &
   0.000000000000000D+00,   7.844074924565336D-06,   0.000000000000000D+00, &
  -2.669529007173258D-05,   0.000000000000000D+00,  -8.691300789326968D-05, &
   0.000000000000000D+00,  -1.204513954964840D-04,   0.000000000000000D+00, &
   4.079089063875029D-06,   0.000000000000000D+00,   9.470453889701647D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.411979023466381D-05, &
   0.000000000000000D+00,  -2.544712349799557D-05,   0.000000000000000D+00, &
  -3.281849204432392D-05,   0.000000000000000D+00,   4.940829752852388D-06, &
   0.000000000000000D+00,   1.354708080550339D-04,   0.000000000000000D+00, &
   3.732521445395117D-04,   0.000000000000000D+00,   6.405056733994684D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.807622787080599D-06,   0.000000000000000D+00,   1.139457299951627D-05, &
   0.000000000000000D+00,   4.260267268490523D-05,   0.000000000000000D+00, &
   8.469732522255384D-05,   0.000000000000000D+00,   8.598214348237853D-05, &
   0.000000000000000D+00,  -1.021873994958485D-04,   0.000000000000000D+00, &
  -8.921929044540271D-04,   0.000000000000000D+00,  -6.364670102153277D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.058145627026917D-05, &
   0.000000000000000D+00,   6.380960879713946D-07,   0.000000000000000D+00, &
  -1.135228255335206D-05,   0.000000000000000D+00,  -5.846162982276075D-05, &
   0.000000000000000D+00,  -1.693738249588746D-04,   0.000000000000000D+00, &
  -3.359648099526909D-04,   0.000000000000000D+00,  -3.842671721461833D-04, &
   0.000000000000000D+00,   8.045657846967595D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.841430228493608D-06, &
   0.000000000000000D+00,  -2.630040243142562D-06,   0.000000000000000D+00, &
  -1.120005141598454D-05,   0.000000000000000D+00,  -5.785691745177876D-06, &
   0.000000000000000D+00,   5.070818090418334D-05,   0.000000000000000D+00, &
   2.662024093236509D-04,   0.000000000000000D+00,   8.955008014834865D-04, &
   0.000000000000000D+00,   2.887712591578258D-03,   0.000000000000000D+00, &
   2.337757503152102D-02,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.814558794691070D-07,   0.000000000000000D+00,   6.354801842958855D-06, &
   0.000000000000000D+00,   1.476452148981905D-05,   0.000000000000000D+00, &
   2.860658072958324D-05,   0.000000000000000D+00,   4.809892252695678D-05, &
   0.000000000000000D+00,   2.924048974484226D-05,   0.000000000000000D+00, &
  -2.384088474186373D-04,   0.000000000000000D+00,  -1.694959340249962D-03, &
   0.000000000000000D+00,  -1.003364703566952D-02,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.972222146917436D-06, &
   0.000000000000000D+00,  -8.902616227951762D-06,   0.000000000000000D+00, &
  -6.166440221865851D-06,   0.000000000000000D+00,  -1.780988222460352D-05, &
   0.000000000000000D+00,  -5.004441143457031D-05,   0.000000000000000D+00, &
  -1.295266002951160D-04,   0.000000000000000D+00,  -2.948289746267931D-04, &
   0.000000000000000D+00,  -5.571935189834851D-04,   0.000000000000000D+00, &
  -2.316412006419817D-03,   0.000000000000000D+00,  -4.390580171783246D-02 /)
lem(:, 69) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.050588253491626D-05,  -1.288965894156855D-04, &
  -4.334582898400869D-04,  -1.015496635595740D-03,  -1.731479327974951D-03, &
  -1.917258198014866D-03,  -3.983309978407013D-04,   3.236584335730868D-03, &
   5.743986203870804D-03,   2.420886649053246D-05,  -1.199104961236746D-02, &
  -2.684178709436266D-03,   2.740604255437955D-02,  -3.107143758447351D-02, &
   1.785132846855999D-02,  -5.691216025127620D-03,  -1.029111282706408D-04, &
   1.743953669188522D-03,  -1.785353227944057D-03,   1.802780200096149D-03, &
  -1.716414733312949D-03,   0.000000000000000D+00,   2.050588253465728D-05, &
   1.288965894160666D-04,   4.334582898395945D-04,   1.015496635596305D-03, &
   1.731479327974452D-03,   1.917258198015198D-03,   3.983309978403985D-04, &
  -3.236584335730438D-03,  -5.743986203871195D-03,  -2.420886649029630D-05, &
   1.199104961236702D-02,   2.684178709436921D-03,  -2.740604255438027D-02, &
   3.107143758447395D-02,  -1.785132846856016D-02,   5.691216025127758D-03, &
   1.029111282702509D-04,  -1.743953669188067D-03,   1.785353227943824D-03, &
  -1.802780200095800D-03,   1.716414733312382D-03,   0.000000000000000D+00, &
   1.025294126712685D-05,   0.000000000000000D+00,  -1.668897860879074D-05, &
   0.000000000000000D+00,   3.155737882759176D-05,   0.000000000000000D+00, &
  -6.683516282784280D-05,   0.000000000000000D+00,   1.585234341426620D-04, &
   0.000000000000000D+00,  -4.315246740108446D-04,   0.000000000000000D+00, &
   1.414238975120059D-03,   0.000000000000000D+00,  -6.079677281380242D-03, &
   0.000000000000000D+00,   4.096188552912975D-02,   0.000000000000000D+00, &
  -7.844811332740665D-01,   0.000000000000000D+00,   1.581336489021815D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.105661780166077D-07, &
   0.000000000000000D+00,  -7.901317365509126D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.879595292381923D-07,   0.000000000000000D+00, &
   1.752253355415624D-06,   0.000000000000000D+00,  -3.923412760824809D-07, &
   0.000000000000000D+00,   1.697080363463892D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.988952302362055D-07,   0.000000000000000D+00, &
   1.421215403719570D-06,   0.000000000000000D+00,  -3.847805327396360D-06, &
   0.000000000000000D+00,   1.009618596261854D-06,   0.000000000000000D+00, &
   6.315814292052919D-07,   0.000000000000000D+00,  -3.689625066828665D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.519903276782281D-06, &
   0.000000000000000D+00,  -8.836717022464886D-07,   0.000000000000000D+00, &
  -2.898723763967568D-06,   0.000000000000000D+00,   9.043400427442309D-06, &
   0.000000000000000D+00,   7.289054651111262D-07,   0.000000000000000D+00, &
  -2.686221438663571D-06,   0.000000000000000D+00,  -6.694307997821057D-07, &
   0.000000000000000D+00,   8.550687382475144D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.836855949819978D-06,   0.000000000000000D+00, &
  -2.945529961698765D-06,   0.000000000000000D+00,   3.346294088016439D-06, &
   0.000000000000000D+00,   6.087122157577419D-06,   0.000000000000000D+00, &
  -2.313125292795250D-05,   0.000000000000000D+00,  -3.764558351167265D-06, &
   0.000000000000000D+00,   8.844221979463883D-07,   0.000000000000000D+00, &
   6.988228669480202D-06,   0.000000000000000D+00,  -7.029888188600504D-07, &
   0.000000000000000D+00,  -2.141108850094855D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.607509395065169D-07,   0.000000000000000D+00, &
   6.997561812716067D-06,   0.000000000000000D+00,   3.739282608847600D-06, &
   0.000000000000000D+00,  -1.174967771492115D-05,   0.000000000000000D+00, &
  -1.281026636587886D-05,   0.000000000000000D+00,   6.636335971787123D-05, &
   0.000000000000000D+00,   7.271482469774047D-06,   0.000000000000000D+00, &
   3.269133236787923D-06,   0.000000000000000D+00,  -1.080460043345066D-05, &
   0.000000000000000D+00,  -1.661276592988756D-05,   0.000000000000000D+00, &
   1.002975723462510D-05,   0.000000000000000D+00,   5.925096744578024D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.583448276120241D-07, &
   0.000000000000000D+00,  -1.377163155355674D-05,   0.000000000000000D+00, &
  -1.576771354478957D-05,   0.000000000000000D+00,   7.042164258159239D-06, &
   0.000000000000000D+00,   4.037214110850346D-05,   0.000000000000000D+00, &
   2.368900608532046D-05,   0.000000000000000D+00,  -2.246152929738210D-04, &
   0.000000000000000D+00,  -7.941959199092474D-06,   0.000000000000000D+00, &
   3.604274866202425D-06,   0.000000000000000D+00,   2.516308153449473D-05, &
   0.000000000000000D+00,   4.449100551211303D-05,   0.000000000000000D+00, &
   2.404062658852954D-05,   0.000000000000000D+00,  -6.405440947759436D-05, &
   0.000000000000000D+00,  -1.865062596374604D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.170442453156060D-06,   0.000000000000000D+00, &
   1.122480898239233D-05,   0.000000000000000D+00,   5.841848923485515D-06, &
   0.000000000000000D+00,  -2.938162146519615D-05,   0.000000000000000D+00, &
  -9.161327869510441D-05,   0.000000000000000D+00,  -1.259222958032246D-04, &
   0.000000000000000D+00,   4.038907044360951D-06,   0.000000000000000D+00, &
   9.780229877983672D-04,   0.000000000000000D+00,   3.080378525803535D-06, &
   0.000000000000000D+00,  -1.438158961433138D-05,   0.000000000000000D+00, &
  -2.873368902811457D-05,   0.000000000000000D+00,  -3.456637497316476D-05, &
   0.000000000000000D+00,   5.895457075906551D-06,   0.000000000000000D+00, &
   1.416607735831439D-04,   0.000000000000000D+00,   3.873619851756580D-04, &
   0.000000000000000D+00,   6.615890890712495D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.846443031299473D-06,   0.000000000000000D+00, &
   6.194871216383805D-06,   0.000000000000000D+00,   1.625242533574456D-05, &
   0.000000000000000D+00,   4.729540258459305D-05,   0.000000000000000D+00, &
   8.987412061534371D-05,   0.000000000000000D+00,   8.971850052716267D-05, &
   0.000000000000000D+00,  -1.058988895726153D-04,   0.000000000000000D+00, &
  -9.205804045644003D-04,   0.000000000000000D+00,  -6.543078801379863D-03, &
   0.000000000000000D+00,   9.899351641856822D-07,   0.000000000000000D+00, &
   5.193924158745646D-06,   0.000000000000000D+00,   4.929904163231219D-06, &
   0.000000000000000D+00,  -1.076452619536673D-05,   0.000000000000000D+00, &
  -6.184314672217862D-05,   0.000000000000000D+00,  -1.774692877464355D-04, &
   0.000000000000000D+00,  -3.489667908422632D-04,   0.000000000000000D+00, &
  -3.976330045186157D-04,   0.000000000000000D+00,   8.241580955341009D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.388450583970957D-07, &
   0.000000000000000D+00,  -9.848013072054867D-06,   0.000000000000000D+00, &
  -8.066646871470008D-06,   0.000000000000000D+00,  -1.287704526330208D-05, &
   0.000000000000000D+00,  -6.967872365134377D-06,   0.000000000000000D+00, &
   5.254701507478027D-05,   0.000000000000000D+00,   2.762766165101593D-04, &
   0.000000000000000D+00,   9.247526976815786D-04,   0.000000000000000D+00, &
   2.970876026388674D-03,   0.000000000000000D+00,   2.396768538640084D-02, &
   0.000000000000000D+00,  -6.449725163172609D-07,   0.000000000000000D+00, &
   5.887378801107770D-06,   0.000000000000000D+00,   1.164051211534770D-06, &
   0.000000000000000D+00,   1.208130274647383D-05,   0.000000000000000D+00, &
   2.884850590281819D-05,   0.000000000000000D+00,   5.025840400531404D-05, &
   0.000000000000000D+00,   3.101095751452163D-05,   0.000000000000000D+00, &
  -2.451509370343702D-04,   0.000000000000000D+00,  -1.741031287438332D-03, &
   0.000000000000000D+00,  -1.028069393391832D-02,   0.000000000000000D+00 /)
lem(:, 70) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.632993161855452D-01,   7.027283689263070D-02,   3.266526290040733D-02, &
  -1.767766952966368D-02,  -6.514466333088092D-03,   1.011710220950957D-03, &
   1.222594994431179D-03,  -1.687196728832537D-04,  -3.567369214681288D-04, &
   4.383617896650334D-05,   1.338054368514506D-04,  -1.475499927355315D-05, &
  -5.914483088581510D-05,   5.902601923566325D-06,   2.936946838859863D-05, &
  -2.673711513489189D-06,  -1.590934986714575D-05,   1.330521533795894D-06, &
   9.220556260146275D-06,  -7.128034731079539D-07,  -5.640769417673069D-06, &
   4.052830890307906D-07,  -1.632993161855452D-01,   7.027283689263045D-02, &
   3.266526290040743D-02,  -1.767766952966370D-02,  -6.514466333088112D-03, &
   1.011710220950939D-03,   1.222594994431168D-03,  -1.687196728832557D-04, &
  -3.567369214681134D-04,   4.383617896649866D-05,   1.338054368514318D-04, &
  -1.475499927354803D-05,  -5.914483088580278D-05,   5.902601923555283D-06, &
   2.936946838858748D-05,  -2.673711513455805D-06,  -1.590934986710848D-05, &
   1.330521533762330D-06,   9.220556260098442D-06,  -7.128034730831018D-07, &
  -5.640769417664116D-06,   4.052830890223062D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.111111111111112D-02,  -4.761904761904751D-03,   0.000000000000000D+00, &
  -7.936507936507948D-04,   0.000000000000000D+00,  -3.769841269841268D-03, &
   1.851851851851852D-03,   0.000000000000000D+00,   1.388888888888889D-03, &
   0.000000000000000D+00,  -7.936507936507927D-04,   0.000000000000000D+00, &
   5.158730158730160D-04,   0.000000000000000D+00,   1.091269841269841D-03, &
   1.202501202501212D-04,   0.000000000000000D+00,  -6.553631553631544D-04, &
   0.000000000000000D+00,  -1.893939393939399D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.623376623376628D-04, &
   0.000000000000000D+00,  -3.427128427128424D-04,   0.000000000000000D+00, &
  -1.736111111111113D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.561253561253551D-05,   0.000000000000000D+00,   1.540126540126542D-04, &
   0.000000000000000D+00,  -7.683288933288950D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.081252081252053D-05, &
   0.000000000000000D+00,   7.492507492507487D-05,   0.000000000000000D+00, &
   1.094391719391717D-04,   0.000000000000000D+00,   1.240079365079356D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.942501942501346D-06, &
   0.000000000000000D+00,  -5.827505827505824D-05,   0.000000000000000D+00, &
   5.035473785473783D-05,   0.000000000000000D+00,   3.720238095238121D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.387501387501484D-06, &
   0.000000000000000D+00,  -7.388444888444893D-05,   0.000000000000000D+00, &
   3.100198412698465D-06,   0.000000000000000D+00,  -2.325148809523984D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.569861511038325D-06,   0.000000000000000D+00, &
  -6.726661138425972D-06,   0.000000000000000D+00,  -4.880442610222033D-05, &
   0.000000000000000D+00,  -6.975446428571433D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.999634352575946D-06,   0.000000000000000D+00, &
   2.756638786050553D-05,   0.000000000000000D+00,   6.363632282749649D-07, &
   0.000000000000000D+00,  -1.162574404761897D-05,   0.000000000000000D+00, &
   6.458746693122813D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.353136028062282D-07, &
   0.000000000000000D+00,  -5.405026689856607D-06,   0.000000000000000D+00, &
   3.586725541543655D-05,   0.000000000000000D+00,   8.525545634920556D-06, &
   0.000000000000000D+00,   1.937624007936668D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.612121452678448D-06,   0.000000000000000D+00,  -8.575633817506873D-06, &
   0.000000000000000D+00,   2.092633928571447D-05,   0.000000000000000D+00, &
   3.823578042327896D-06,   0.000000000000000D+00,  -2.260561342591345D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.724174518292069D-06,   0.000000000000000D+00,  -1.274323908805324D-05, &
   0.000000000000000D+00,  -1.111214301157224D-05,   0.000000000000000D+00, &
   2.583498677248823D-06,   0.000000000000000D+00,  -6.781684027779110D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.657945769400609D-07,   0.000000000000000D+00, &
   6.256509013861874D-06,   0.000000000000000D+00,  -1.597702276760915D-05, &
   0.000000000000000D+00,  -8.243709415584547D-06,   0.000000000000000D+00, &
  -1.423860066438076D-06,   0.000000000000000D+00,   9.247750946951839D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.412434949912706D-08,   0.000000000000000D+00,   9.692513835903736D-07, &
   0.000000000000000D+00,   1.082051167258837D-05,   0.000000000000000D+00, &
  -7.351228054353110D-06,   0.000000000000000D+00,  -1.617622467231901D-06, &
   0.000000000000000D+00,   2.774325284091350D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.308267872366115D-06, &
   0.000000000000000D+00,   4.656244291194703D-06,   0.000000000000000D+00, &
   1.100335572991829D-05,   0.000000000000000D+00,   7.133979301948402D-07, &
   0.000000000000000D+00,   6.011038115532376D-07,   0.000000000000000D+00, &
  -4.238552517356801D-08,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.551310875225913D-07,   0.000000000000000D+00,  -5.270942401560007D-06, &
   0.000000000000000D+00,   5.315045843375522D-06,   0.000000000000000D+00, &
   5.569201125841725D-06,   0.000000000000000D+00,   8.168846669821257D-07, &
   0.000000000000000D+00,  -1.271565755206647D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.131372228519565D-08, &
   0.000000000000000D+00,   5.030884636460702D-07,   0.000000000000000D+00, &
  -9.075891015413231D-06,   0.000000000000000D+00,   9.121285976756568D-07, &
   0.000000000000000D+00,   2.881030102710867D-07,   0.000000000000000D+00, &
  -2.806929534227890D-07,   0.000000000000000D+00,   2.119276258673093D-08 /)
lem(:, 71) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.443310539518172D-02,  -9.035079029052503D-02,   8.908708063747458D-03, &
   3.788072042070788D-02,  -3.908679799852864D-03,  -9.105391988558535D-03, &
  -6.792194413506526D-04,   1.518477055949290D-03,   2.751970537039835D-04, &
  -3.945256106985268D-04,  -1.146903744440911D-04,   1.327949934620179D-04, &
   5.330336610696398D-05,  -5.312341731210145D-05,  -2.723350705125088D-05, &
   2.406340362126428D-05,   1.501931630815663D-05,  -1.197469380393175D-05, &
  -8.810753759719448D-06,   6.415231257651353D-06,   5.436578669540217D-06, &
  -3.647547801102933D-06,   5.443310539518166D-02,  -9.035079029052498D-02, &
   8.908708063747469D-03,   3.788072042070793D-02,  -3.908679799852837D-03, &
  -9.105391988558559D-03,  -6.792194413506586D-04,   1.518477055949300D-03, &
   2.751970537039691D-04,  -3.945256106985230D-04,  -1.146903744440847D-04, &
   1.327949934620068D-04,   5.330336610695715D-05,  -5.312341731209926D-05, &
  -2.723350705122865D-05,   2.406340362126528D-05,   1.501931630812646D-05, &
  -1.197469380391648D-05,  -8.810753759690933D-06,   6.415231257598411D-06, &
   5.436578669526820D-06,  -3.647547801049115D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.761904761904751D-03,   4.761904761904761D-03,   0.000000000000000D+00, &
  -2.380952380952385D-03,   0.000000000000000D+00,   1.388888888888889D-03, &
  -7.936507936507941D-04,   0.000000000000000D+00,  -1.547619047619046D-03, &
   0.000000000000000D+00,   1.659451659451659D-03,   0.000000000000000D+00, &
   7.395382395382401D-04,   0.000000000000000D+00,  -1.893939393939399D-04, &
  -7.936507936507943D-04,   0.000000000000000D+00,   4.004329004329001D-04, &
   0.000000000000000D+00,   4.178691678691672D-04,   0.000000000000000D+00, &
   1.332001332001340D-04,   0.000000000000000D+00,  -6.007881007881003D-04, &
   0.000000000000000D+00,  -1.734376734376734D-04,   0.000000000000000D+00, &
  -7.683288933288939D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.956376956376959D-04,   0.000000000000000D+00,  -2.317127317127319D-04, &
   0.000000000000000D+00,  -8.897352647352645D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.192252192252168D-05, &
   0.000000000000000D+00,   1.859251859251860D-04,   0.000000000000000D+00, &
   2.306721056721060D-05,   0.000000000000000D+00,   3.720238095238115D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.081252081252021D-05, &
   0.000000000000000D+00,   3.052503052503065D-05,   0.000000000000000D+00, &
   9.833916083916093D-05,   0.000000000000000D+00,   3.720238095238092D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.089413854118940D-06,   0.000000000000000D+00,  -5.737726325961615D-05, &
   0.000000000000000D+00,  -1.271196124137298D-05,   0.000000000000000D+00, &
   1.648933174668498D-06,   0.000000000000000D+00,  -6.975446428571483D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.706622353680874D-06,   0.000000000000000D+00, &
  -6.262120232708472D-05,   0.000000000000000D+00,  -3.656601772410588D-05, &
   0.000000000000000D+00,  -1.679274140211644D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.420769804671125D-06,   0.000000000000000D+00, &
   1.322958946797957D-05,   0.000000000000000D+00,  -2.117208180965923D-05, &
   0.000000000000000D+00,  -1.007564484126985D-05,   0.000000000000000D+00, &
   1.937624007936714D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.999634352575841D-06, &
   0.000000000000000D+00,   2.131189244192360D-05,   0.000000000000000D+00, &
   2.560816645085228D-05,   0.000000000000000D+00,   1.317584325396843D-05, &
   0.000000000000000D+00,   5.192832341269794D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.431889976783689D-07,   0.000000000000000D+00, &
  -7.623203250138448D-06,   0.000000000000000D+00,   1.914083401119011D-05, &
   0.000000000000000D+00,   1.824428630615737D-05,   0.000000000000000D+00, &
   7.750496031746133D-06,   0.000000000000000D+00,  -6.781684027779279D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.430812221833787D-06,   0.000000000000000D+00,  -1.547447972564076D-05, &
   0.000000000000000D+00,  -2.167048020134302D-06,   0.000000000000000D+00, &
  -7.609577922079429D-07,   0.000000000000000D+00,  -1.893587098665129D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.578308508918299D-06,   0.000000000000000D+00, &
  -6.706550477162236D-06,   0.000000000000000D+00,  -1.701575048398938D-05, &
   0.000000000000000D+00,  -1.148482593795079D-05,   0.000000000000000D+00, &
  -3.426071541305959D-06,   0.000000000000000D+00,   2.774325284091319D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.657945769399360D-07,   0.000000000000000D+00,   6.339971079084783D-06, &
   0.000000000000000D+00,  -2.467296476889505D-06,   0.000000000000000D+00, &
  -5.707183441558477D-06,   0.000000000000000D+00,  -2.457259537337573D-06, &
   0.000000000000000D+00,   7.911964699074549D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.479204602059287D-08,   0.000000000000000D+00,   9.062054235509180D-08, &
   0.000000000000000D+00,   1.104283535386522D-05,   0.000000000000000D+00, &
   7.481927902403087D-06,   0.000000000000000D+00,   3.972129216269722D-06, &
   0.000000000000000D+00,   1.556704742739801D-06,   0.000000000000000D+00, &
  -1.271565755206994D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.067897124524747D-06,   0.000000000000000D+00,   3.090391620034274D-07, &
   0.000000000000000D+00,   8.350981880209791D-06,   0.000000000000000D+00, &
   6.908374828296874D-06,   0.000000000000000D+00,   1.695929125128421D-06, &
   0.000000000000000D+00,  -3.672424174134842D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.329858195504313D-07, &
   0.000000000000000D+00,  -4.270167324599732D-06,   0.000000000000000D+00, &
  -3.337024109319626D-06,   0.000000000000000D+00,  -6.781684027778140D-07, &
   0.000000000000000D+00,  -2.580675798448881D-07,   0.000000000000000D+00, &
  -7.567150053719656D-07,   0.000000000000000D+00,   6.357828776031219D-08 /)
lem(:, 72) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.015590679578893D-02,  -1.484784677291246D-02, &
  -1.094331923264895D-02,   6.514466333088093D-03,   1.011710220950949D-03, &
  -1.222594994431176D-03,  -1.687196728832516D-04,   3.567369214681057D-04, &
   4.383617896650277D-05,  -1.338054368514218D-04,  -1.475499927355504D-05, &
   5.914483088578951D-05,   5.902601923564945D-06,  -2.936946838856615D-05, &
  -2.673711513481203D-06,   1.590934986711572D-05,   1.330521533782423D-06, &
  -9.220556260128274D-06,  -7.128034730868725D-07,   5.640769417677649D-06, &
   4.052830890181451D-07,   0.000000000000000D+00,  -4.015590679578885D-02, &
   1.484784677291246D-02,   1.094331923264897D-02,  -6.514466333088112D-03, &
  -1.011710220950947D-03,   1.222594994431171D-03,   1.687196728832527D-04, &
  -3.567369214681153D-04,  -4.383617896650387D-05,   1.338054368514309D-04, &
   1.475499927355379D-05,  -5.914483088579736D-05,  -5.902601923562388D-06, &
   2.936946838858088D-05,   2.673711513474436D-06,  -1.590934986713603D-05, &
  -1.330521533771061D-06,   9.220556260152449D-06,   7.128034730672786D-07, &
  -5.640769417688615D-06,  -4.052830890043910D-07,   0.000000000000000D+00, &
  -2.007795339789447D-02,   0.000000000000000D+00,   8.417937871268419D-03, &
   0.000000000000000D+00,  -2.023420441901896D-03,   0.000000000000000D+00, &
   3.374393457665111D-04,   0.000000000000000D+00,  -8.767235793300508D-05, &
   0.000000000000000D+00,   2.950999854711189D-05,   0.000000000000000D+00, &
  -1.180520384713449D-05,   0.000000000000000D+00,   5.347423026950253D-06, &
   0.000000000000000D+00,  -2.661043067535928D-06,   0.000000000000000D+00, &
   1.425606946135812D-06,   0.000000000000000D+00,  -8.105661780166077D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.587301587301586D-03, &
   0.000000000000000D+00,  -7.936507936507941D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.388888888888886D-03,   0.000000000000000D+00,   2.645502645502644D-04, &
   0.000000000000000D+00,   5.952380952380964D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.683501683501688D-04,   0.000000000000000D+00, &
   1.262626262626263D-04,   0.000000000000000D+00,   6.403318903318902D-04, &
   0.000000000000000D+00,   4.810004810004818D-05,   0.000000000000000D+00, &
  -2.705627705627707D-04,   0.000000000000000D+00,  -1.623376623376626D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.550005550005634D-06, &
   0.000000000000000D+00,   1.068376068376067D-04,   0.000000000000000D+00, &
  -1.796814296814298D-04,   0.000000000000000D+00,  -1.736111111111112D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.162504162504113D-06,   0.000000000000000D+00,   1.290376290376290D-04, &
   0.000000000000000D+00,  -2.480158730158738D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.526251526251514D-05,   0.000000000000000D+00,   1.942501942501944D-05, &
   0.000000000000000D+00,   9.764541014541012D-05,   0.000000000000000D+00, &
   3.720238095238083D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.162504162504108D-06,   0.000000000000000D+00, &
  -4.578754578754577D-05,   0.000000000000000D+00,   3.526566026566070D-06, &
   0.000000000000000D+00,   2.790178571428573D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.264709147061684D-07,   0.000000000000000D+00,   1.044706927059864D-05, &
   0.000000000000000D+00,  -5.820704350116126D-05,   0.000000000000000D+00, &
  -1.860119047619037D-05,   0.000000000000000D+00,  -1.162574404761905D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.141916906622838D-06, &
   0.000000000000000D+00,   1.531012560424333D-05,   0.000000000000000D+00, &
  -3.410218253968257D-05,   0.000000000000000D+00,  -1.007564484126980D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.052073485510072D-06,   0.000000000000000D+00,   1.794408721653306D-05, &
   0.000000000000000D+00,   1.947972209678869D-05,   0.000000000000000D+00, &
  -2.325148809523829D-06,   0.000000000000000D+00,   4.521122685185281D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.412544112235568D-07, &
   0.000000000000000D+00,  -1.059204813074785D-05,   0.000000000000000D+00, &
   2.161017302271172D-05,   0.000000000000000D+00,   1.472594246031749D-05, &
   0.000000000000000D+00,   4.133597883597872D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.221752447775484D-08, &
   0.000000000000000D+00,  -6.158916762631309D-07,   0.000000000000000D+00, &
  -1.746572409788354D-05,   0.000000000000000D+00,   8.525545634920661D-06, &
   0.000000000000000D+00,   2.195973875661310D-06,   0.000000000000000D+00, &
  -2.034505208333256D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.014132238590368D-06, &
   0.000000000000000D+00,  -4.638115367604468D-06,   0.000000000000000D+00, &
  -1.698063221500720D-05,   0.000000000000000D+00,  -2.771389490139397D-06, &
   0.000000000000000D+00,  -1.911201862373843D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.407318545559246D-07,   0.000000000000000D+00, &
   7.865786034533449D-06,   0.000000000000000D+00,  -4.686049878717594D-06, &
   0.000000000000000D+00,  -7.914900493025535D-06,   0.000000000000000D+00, &
  -1.294685132575634D-06,   0.000000000000000D+00,   1.017252604166524D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.346206522256908D-08,   0.000000000000000D+00, &
  -1.473522761507311D-06,   0.000000000000000D+00,   1.291043388655999D-05, &
   0.000000000000000D+00,   9.277108886483404D-07,   0.000000000000000D+00, &
   3.787174197328887D-07,   0.000000000000000D+00,   9.710138494319489D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.006179130539050D-09, &
   0.000000000000000D+00,  -1.081668365286650D-06,   0.000000000000000D+00, &
  -1.497998012133864D-06,   0.000000000000000D+00,   1.068628998316513D-05, &
   0.000000000000000D+00,   2.897628630050546D-06,   0.000000000000000D+00, &
   7.398200757574787D-07,   0.000000000000000D+00,  -5.510118272568938D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.297212821056065D-07,   0.000000000000000D+00,  -4.829511538597134D-06, &
   0.000000000000000D+00,  -2.383864688552264D-06,   0.000000000000000D+00, &
   3.134750449203660D-06,   0.000000000000000D+00,   9.366311856554352D-08, &
   0.000000000000000D+00,  -5.314492771768276D-07,   0.000000000000000D+00 /)
lem(:, 73) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.332847374079219D-02,   4.015590679578893D-02,  -6.087617176894103D-02, &
  -1.565736444055929D-02,   3.460281467835103D-02,   7.081971546656649D-03, &
  -8.965696625828624D-03,  -3.350290647253203D-03,   1.172135599109513D-03, &
   9.775467909530274D-04,  -2.246019832863343D-04,  -3.435806973699833D-04, &
   5.038263371754218D-05,   1.404819257808932D-04,  -1.067980668675369D-05, &
  -6.445633458513928D-05,   7.787793641018081D-07,   3.234190805159864D-05, &
   1.536759376726381D-06,  -1.742530336471992D-05,  -1.812192889871129D-06, &
   9.948167252041622D-06,   2.332847374079219D-02,   4.015590679578898D-02, &
  -6.087617176894111D-02,  -1.565736444055928D-02,   3.460281467835111D-02, &
   7.081971546656650D-03,  -8.965696625828603D-03,  -3.350290647253206D-03, &
   1.172135599109518D-03,   9.775467909530257D-04,  -2.246019832863291D-04, &
  -3.435806973699744D-04,   5.038263371753049D-05,   1.404819257809145D-04, &
  -1.067980668675766D-05,  -6.445633458516745D-05,   7.787793641179761D-07, &
   3.234190805160946D-05,   1.536759376707836D-06,  -1.742530336469492D-05, &
  -1.812192889858988D-06,   9.948167252005689D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.936507936507948D-04,  -2.380952380952385D-03,   0.000000000000000D+00, &
   4.444444444444449D-03,   0.000000000000000D+00,   5.158730158730161D-04, &
  -2.405002405002406D-03,   0.000000000000000D+00,   7.395382395382400D-04, &
   0.000000000000000D+00,  -1.010101010101008D-03,   0.000000000000000D+00, &
  -1.562049062049062D-03,   0.000000000000000D+00,  -3.427128427128425D-04, &
   1.957301957301961D-03,   0.000000000000000D+00,   7.654382654382653D-04, &
   0.000000000000000D+00,  -1.734376734376734D-04,   0.000000000000000D+00, &
  -9.490509490509491D-04,   0.000000000000000D+00,   5.019980019980018D-04, &
   0.000000000000000D+00,   5.143467643467643D-04,   0.000000000000000D+00, &
   1.094391719391717D-04,   1.631701631701644D-04,   0.000000000000000D+00, &
  -7.214082214082214D-04,   0.000000000000000D+00,  -1.685814185814186D-04, &
   0.000000000000000D+00,   2.306721056721055D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.455877455877459D-04, &
   0.000000000000000D+00,  -2.955377955377956D-04,   0.000000000000000D+00, &
  -1.163766788766788D-04,   0.000000000000000D+00,   3.100198412698425D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.074722662957926D-05, &
   0.000000000000000D+00,   2.328553799142033D-04,   0.000000000000000D+00, &
  -9.409843968667558D-06,   0.000000000000000D+00,   1.648933174668492D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.448531860296514D-05,   0.000000000000000D+00,   2.677061500590888D-05, &
   0.000000000000000D+00,   1.219368866427690D-04,   0.000000000000000D+00, &
   1.358680127062464D-05,   0.000000000000000D+00,  -1.162574404761895D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.505234703376748D-06, &
   0.000000000000000D+00,  -6.810870587960345D-05,   0.000000000000000D+00, &
  -1.391224301441014D-05,   0.000000000000000D+00,   1.619436241146775D-05, &
   0.000000000000000D+00,  -1.007564484126985D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.749388554341918D-06,   0.000000000000000D+00, &
  -7.462115627750311D-05,   0.000000000000000D+00,  -2.661543392967551D-05, &
   0.000000000000000D+00,  -4.650297619047532D-07,   0.000000000000000D+00, &
   3.823578042327859D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.247686266262302D-06, &
   0.000000000000000D+00,   1.958932879985516D-05,   0.000000000000000D+00, &
  -3.008007071474559D-05,   0.000000000000000D+00,   4.522367198933768D-07, &
   0.000000000000000D+00,   7.750496031746102D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.330400937212237D-06,   0.000000000000000D+00, &
   2.371478280921001D-05,   0.000000000000000D+00,   2.866319320943947D-05, &
   0.000000000000000D+00,   1.233108969418975D-06,   0.000000000000000D+00, &
   4.126551978114391D-06,   0.000000000000000D+00,  -1.423860066438058D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.694921939909953D-07,   0.000000000000000D+00, &
  -9.762913796094608D-06,   0.000000000000000D+00,   2.217485809553218D-05, &
   0.000000000000000D+00,   1.410237723508557D-05,   0.000000000000000D+00, &
  -4.004422949735328D-06,   0.000000000000000D+00,  -3.426071541305978D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.398454621162823D-06,   0.000000000000000D+00, &
  -1.957309455269297D-05,   0.000000000000000D+00,   3.204547151568527D-06, &
   0.000000000000000D+00,  -2.811316287878914D-06,   0.000000000000000D+00, &
  -4.354956747534771D-06,   0.000000000000000D+00,   6.011038115532074D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.777767669879424D-06,   0.000000000000000D+00,  -6.591982742247055D-06, &
   0.000000000000000D+00,  -1.918436423495766D-05,   0.000000000000000D+00, &
  -1.823163382671448D-06,   0.000000000000000D+00,   3.552310681215217D-07, &
   0.000000000000000D+00,   1.556704742739812D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.078594578119279D-07,   0.000000000000000D+00,   7.557425804329910D-06, &
   0.000000000000000D+00,  -2.948697108882384D-06,   0.000000000000000D+00, &
  -5.409188363289174D-06,   0.000000000000000D+00,   4.550706455003570D-06, &
   0.000000000000000D+00,   2.514168773934400D-06,   0.000000000000000D+00, &
  -2.806929534227790D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.736010956604256D-08,   0.000000000000000D+00, &
  -2.187733334253266D-07,   0.000000000000000D+00,   1.331259905791526D-05, &
   0.000000000000000D+00,   3.696212897747694D-06,   0.000000000000000D+00, &
   1.493415800056313D-06,   0.000000000000000D+00,   1.503352333430595D-06, &
   0.000000000000000D+00,  -7.567150053719636D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.147937245073747D-06, &
   0.000000000000000D+00,  -5.135907735203558D-07,   0.000000000000000D+00, &
   1.009218044679802D-05,   0.000000000000000D+00,   3.972355046572592D-07, &
   0.000000000000000D+00,  -2.243511154644221D-06,   0.000000000000000D+00, &
  -1.368065866967344D-06,   0.000000000000000D+00,   1.420613755821542D-07 /)
lem(:, 74) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.007795339789445D-02,   2.227177015936870D-02, &
   1.683587574253706D-04,  -1.101537034503987D-02,   3.035130662852849D-03, &
   2.445189988862346D-03,  -9.882152268876449D-04,  -5.402016239374212D-04, &
   2.805515453856240D-04,   1.768143272679585D-04,  -9.766404281069111D-05, &
  -7.228812663819532D-05,   3.974418628536274D-05,   3.417538139761065D-05, &
  -1.818565764950110D-05,  -1.791192537486616D-05,   9.108955115799428D-06, &
   1.014261188614073D-05,  -4.901894653243857D-06,  -6.100198601017410D-06, &
   2.796112740113688D-06,   0.000000000000000D+00,   2.007795339789445D-02, &
  -2.227177015936872D-02,  -1.683587574253635D-04,   1.101537034503987D-02, &
  -3.035130662852853D-03,  -2.445189988862350D-03,   9.882152268876343D-04, &
   5.402016239374269D-04,  -2.805515453856219D-04,  -1.768143272679642D-04, &
   9.766404281069141D-05,   7.228812663819695D-05,  -3.974418628535683D-05, &
  -3.417538139762003D-05,   1.818565764949693D-05,   1.791192537488112D-05, &
  -9.108955115803460D-06,  -1.014261188615765D-05,   4.901894653261747D-06, &
   6.100198601020795D-06,  -2.796112740128660D-06,   0.000000000000000D+00, &
   1.003897669894724D-02,   0.000000000000000D+00,  -3.030457633656631D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.410281041189376D-04,   0.000000000000000D+00,  -7.452150424305770D-05, &
   0.000000000000000D+00,   2.669952249501014D-05,   0.000000000000000D+00, &
  -1.101819025732697D-05,   0.000000000000000D+00,   5.082261554536547D-06, &
   0.000000000000000D+00,  -2.558695257248841D-06,   0.000000000000000D+00, &
   1.381742117026409D-06,   0.000000000000000D+00,  -7.901317365509126D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.936507936507941D-04, &
   0.000000000000000D+00,   6.349206349206349D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.030303030303030D-04,   0.000000000000000D+00, &
   5.952380952380964D-04,   0.000000000000000D+00,  -3.848003848003852D-05, &
   0.000000000000000D+00,  -4.894179894179889D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.813001813001813D-04,   0.000000000000000D+00, &
   2.145077145077145D-04,   0.000000000000000D+00,  -1.623376623376626D-04, &
   0.000000000000000D+00,  -1.350501350501354D-04,   0.000000000000000D+00, &
   7.932543646829361D-05,   0.000000000000000D+00,   1.653439153439154D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.106005106005116D-05, &
   0.000000000000000D+00,  -1.734376734376737D-04,   0.000000000000000D+00, &
  -6.105006105006105D-05,   0.000000000000000D+00,  -2.480158730158736D-05, &
   0.000000000000000D+00,  -9.990009990010031D-06,   0.000000000000000D+00, &
   9.573759573759579D-05,   0.000000000000000D+00,  -7.284382284382269D-05, &
   0.000000000000000D+00,  -2.480158730158728D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.835301952949592D-07,   0.000000000000000D+00, &
  -1.398927869516112D-05,   0.000000000000000D+00,   8.292361233537711D-05, &
   0.000000000000000D+00,   1.176315514550812D-05,   0.000000000000000D+00, &
   2.790178571428573D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.384861326037725D-06,   0.000000000000000D+00, &
  -1.077354018530485D-05,   0.000000000000000D+00,   3.188872674166790D-05, &
   0.000000000000000D+00,   1.033399470899466D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.357527577341728D-06,   0.000000000000000D+00,  -2.381948976376220D-05, &
   0.000000000000000D+00,  -1.490991235573272D-05,   0.000000000000000D+00, &
  -8.224395680858551D-06,   0.000000000000000D+00,  -1.007564484126981D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.560379077406098D-07,   0.000000000000000D+00,   1.182454742826256D-05, &
   0.000000000000000D+00,  -2.123958199809596D-05,   0.000000000000000D+00, &
  -1.169965977435014D-05,   0.000000000000000D+00,  -8.112185846560857D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.500225641308921D-08,   0.000000000000000D+00, &
   9.450473846759476D-07,   0.000000000000000D+00,   1.538279402056491D-05, &
   0.000000000000000D+00,  -1.486749261517071D-06,   0.000000000000000D+00, &
   1.705109126984146D-06,   0.000000000000000D+00,   4.133597883597877D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.257374548396319D-06, &
   0.000000000000000D+00,   4.988346706613032D-06,   0.000000000000000D+00, &
   1.197639488564411D-05,   0.000000000000000D+00,   8.046424062049031D-06, &
   0.000000000000000D+00,   3.689705838143341D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.025599299122997D-07, &
   0.000000000000000D+00,  -7.366863816251422D-06,   0.000000000000000D+00, &
   2.629171784677265D-06,   0.000000000000000D+00,   2.516650961775233D-06, &
   0.000000000000000D+00,   1.470245610870626D-06,   0.000000000000000D+00, &
  -1.911201862373840D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.104952569758486D-08,   0.000000000000000D+00,   1.268366585031084D-06, &
   0.000000000000000D+00,  -9.651939404640038D-06,   0.000000000000000D+00, &
  -4.252499468933845D-06,   0.000000000000000D+00,  -3.638035864598410D-06, &
   0.000000000000000D+00,  -1.776155340608358D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.370377233420050D-09, &
   0.000000000000000D+00,   1.055730923478826D-06,   0.000000000000000D+00, &
   1.058704624334300D-06,   0.000000000000000D+00,  -5.176496144903330D-06, &
   0.000000000000000D+00,  -4.004422949735453D-06,   0.000000000000000D+00, &
  -1.100922731782231D-06,   0.000000000000000D+00,   9.710138494319293D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.975883869946131D-07,   0.000000000000000D+00, &
   3.944545605603609D-06,   0.000000000000000D+00,   2.415861062875754D-06, &
   0.000000000000000D+00,   1.553712491212549D-06,   0.000000000000000D+00, &
   6.673285481880150D-07,   0.000000000000000D+00,   9.216134704414782D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.299782562400809D-07, &
   0.000000000000000D+00,  -1.793196738330070D-06,   0.000000000000000D+00, &
   4.720582629211378D-06,   0.000000000000000D+00,   3.573093457725190D-06, &
   0.000000000000000D+00,   2.813168096371207D-06,   0.000000000000000D+00, &
   6.746115754905967D-07,   0.000000000000000D+00,  -5.314492771768268D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.335393043563064D-08,   0.000000000000000D+00, &
  -2.948992971204457D-07,   0.000000000000000D+00,  -2.634612237025542D-06, &
   0.000000000000000D+00,   1.027892754729740D-06,   0.000000000000000D+00, &
   1.049433422480228D-06,   0.000000000000000D+00,  -4.452083135183610D-09, &
   0.000000000000000D+00,  -5.108704907288865D-07,   0.000000000000000D+00 /)
lem(:, 75) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.832118435198053D-02,  -1.254872087368405D-02,  -2.783971269921089D-02, &
   9.091372900969900D-03,   7.385881959462212D-03,  -3.540985773328326D-03, &
  -1.018829162026023D-04,   7.712899331805978D-04,  -1.579834937930171D-04, &
  -2.093177545650548D-04,   8.721247223351320D-05,   7.166713932870880D-05, &
  -4.490626048736094D-05,  -2.892274942548650D-05,   2.416306262879074D-05, &
   1.316968646306631D-05,  -1.373989306706685D-05,  -6.575846811132431D-06, &
   8.221662665296847D-06,   3.531118743515833D-06,  -5.143054469065706D-06, &
  -2.011089613944019D-06,   5.832118435198040D-02,  -1.254872087368406D-02, &
  -2.783971269921085D-02,   9.091372900969908D-03,   7.385881959462211D-03, &
  -3.540985773328331D-03,  -1.018829162025971D-04,   7.712899331805945D-04, &
  -1.579834937930232D-04,  -2.093177545650530D-04,   8.721247223351845D-05, &
   7.166713932870597D-05,  -4.490626048736113D-05,  -2.892274942547787D-05, &
   2.416306262878450D-05,   1.316968646305174D-05,  -1.373989306707092D-05, &
  -6.575846811118765D-06,   8.221662665303013D-06,   3.531118743516019D-06, &
  -5.143054469065110D-06,  -2.011089613949935D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.484784677291246D-02,   0.000000000000000D+00, &
   9.001808023903555D-03,   0.000000000000000D+00,  -2.445189988862347D-03, &
   0.000000000000000D+00,   3.669294049386305D-04,   0.000000000000000D+00, &
  -8.601778083306487D-05,   0.000000000000000D+00,   2.628659150480343D-05, &
   0.000000000000000D+00,  -9.611826018083128D-06,   0.000000000000000D+00, &
   4.005151015500775D-06,   0.000000000000000D+00,  -1.844111252024118D-06, &
   0.000000000000000D+00,   9.188583666756751D-07,   0.000000000000000D+00, &
  -3.769841269841268D-03,   1.388888888888889D-03,   0.000000000000000D+00, &
   5.158730158730161D-04,   0.000000000000000D+00,   1.944444444444443D-03, &
  -6.553631553631546D-04,   0.000000000000000D+00,  -7.611832611832602D-04, &
   0.000000000000000D+00,   1.623376623376627D-04,   0.000000000000000D+00, &
  -2.092352092352091D-04,   0.000000000000000D+00,  -1.215277777777777D-03, &
   3.561253561253547D-05,   0.000000000000000D+00,   3.206053206053204D-04, &
   0.000000000000000D+00,   4.472957597957596D-04,   0.000000000000000D+00, &
  -2.081252081252057D-05,   0.000000000000000D+00,  -1.021201021201020D-04, &
   0.000000000000000D+00,   1.650086025086025D-04,   0.000000000000000D+00, &
   5.532661782661787D-04,   1.942501942501363D-06,   0.000000000000000D+00, &
   2.590002590002636D-06,   0.000000000000000D+00,  -2.117673992673995D-04, &
   0.000000000000000D+00,  -1.307720057720058D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.665001665001540D-06, &
   0.000000000000000D+00,   5.723443223443225D-05,   0.000000000000000D+00, &
  -1.514978077478074D-04,   0.000000000000000D+00,  -1.666356646825398D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.508831038245610D-07, &
   0.000000000000000D+00,   8.141368435486064D-06,   0.000000000000000D+00, &
   9.963654265124858D-05,   0.000000000000000D+00,  -1.308784914023899D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.897063720624601D-08,   0.000000000000000D+00,  -8.304603892839130D-06, &
   0.000000000000000D+00,   1.318636428930544D-05,   0.000000000000000D+00, &
   8.362342059492802D-05,   0.000000000000000D+00,   3.952752976190474D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.319661762357329D-08, &
   0.000000000000000D+00,   3.305303227903689D-06,   0.000000000000000D+00, &
  -3.078966169136445D-05,   0.000000000000000D+00,  -4.401310969373979D-06, &
   0.000000000000000D+00,   2.247643849206354D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.031497742023863D-06,   0.000000000000000D+00, &
   9.858025531400142D-06,   0.000000000000000D+00,  -4.212919554323382D-05, &
   0.000000000000000D+00,  -2.046130952380960D-05,   0.000000000000000D+00, &
  -1.142552290013235D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.571118252216448D-08, &
   0.000000000000000D+00,   8.446361000541363D-07,   0.000000000000000D+00, &
   1.590066275994106D-05,   0.000000000000000D+00,  -2.433335927917974D-05, &
   0.000000000000000D+00,  -9.164961557539688D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.953273077124572D-08,   0.000000000000000D+00, &
  -2.271872434411251D-06,   0.000000000000000D+00,   8.517222774430556D-06, &
   0.000000000000000D+00,   1.895944263639696D-05,   0.000000000000000D+00, &
  -1.097986937830274D-07,   0.000000000000000D+00,   4.063138828764006D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.852095318040439D-09,   0.000000000000000D+00, &
   1.113722467538667D-06,   0.000000000000000D+00,  -8.565773766707235D-06, &
   0.000000000000000D+00,   1.158931621214146D-05,   0.000000000000000D+00, &
   1.286523625390813D-05,   0.000000000000000D+00,   3.628641323953797D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.983829088756160D-07,   0.000000000000000D+00, &
   2.197203798140615D-06,   0.000000000000000D+00,  -1.424084401136962D-05, &
   0.000000000000000D+00,   3.605742131132795D-06,   0.000000000000000D+00, &
   1.586503051346842D-06,   0.000000000000000D+00,  -1.676154859138262D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.168468913117993D-08,   0.000000000000000D+00,   8.978409779430284D-07, &
   0.000000000000000D+00,   9.277775367067025D-07,   0.000000000000000D+00, &
  -1.336217774861021D-05,   0.000000000000000D+00,  -3.035610945767253D-06, &
   0.000000000000000D+00,  -1.575970890546045D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.164897291451128D-09,   0.000000000000000D+00,  -9.462431392641745D-07, &
   0.000000000000000D+00,   4.864112984790838D-06,   0.000000000000000D+00, &
   3.638242756650482D-07,   0.000000000000000D+00,  -5.795257260101034D-06, &
   0.000000000000000D+00,  -1.016109338252965D-06,   0.000000000000000D+00, &
   7.721279236170333D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.564137681634481D-10,   0.000000000000000D+00, &
   2.946291154348404D-07,   0.000000000000000D+00,  -2.892764684865404D-06, &
   0.000000000000000D+00,   8.448498771177069D-06,   0.000000000000000D+00, &
   2.468325222231499D-06,   0.000000000000000D+00,   4.844483451661665D-07, &
   0.000000000000000D+00,   7.478229371540181D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.268081622433245D-08, &
   0.000000000000000D+00,   1.393955592662060D-07,   0.000000000000000D+00, &
  -3.957449977790220D-06,   0.000000000000000D+00,   6.908600658600529D-06, &
   0.000000000000000D+00,   2.446800771385682D-06,   0.000000000000000D+00, &
   5.715482705229559D-07,   0.000000000000000D+00,  -3.873501359613433D-07 /)
lem(:, 76) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.332847374079213D-02,   2.677060453052602D-02,   4.107904273839111D-02, &
  -4.568644462861136D-02,  -2.923331515532083D-02,   2.685866598230851D-02, &
   1.455470231465686D-02,  -5.712366067618780D-03,  -5.228744020375482D-03, &
  -2.410989843157748D-04,   1.320532227974243D-03,   3.154759368489261D-04, &
  -4.033740053534353D-04,  -1.751105237324962D-04,   1.432388618048290D-04, &
   9.256345066066837D-05,  -5.689134965201487D-05,  -5.035512266260671D-05, &
   2.456450757508053D-05,   2.856696995910957D-05,  -1.128518509870386D-05, &
  -1.689587734926826D-05,  -2.332847374079220D-02,   2.677060453052595D-02, &
   4.107904273839118D-02,  -4.568644462861133D-02,  -2.923331515532090D-02, &
   2.685866598230849D-02,   1.455470231465684D-02,  -5.712366067618791D-03, &
  -5.228744020375462D-03,  -2.410989843157627D-04,   1.320532227974226D-03, &
   3.154759368489192D-04,  -4.033740053534011D-04,  -1.751105237325160D-04, &
   1.432388618048011D-04,   9.256345066069363D-05,  -5.689134965200542D-05, &
  -5.035512266263836D-05,   2.456450757509500D-05,   2.856696995915070D-05, &
  -1.128518509871262D-05,  -1.689587734930921D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.851851851851852D-03,  -7.936507936507941D-04,   0.000000000000000D+00, &
  -2.405002405002406D-03,   0.000000000000000D+00,  -6.553631553631546D-04, &
   4.713804713804713D-03,   0.000000000000000D+00,   4.004329004329002D-04, &
   0.000000000000000D+00,  -2.586302586302589D-03,   0.000000000000000D+00, &
   7.654382654382653D-04,   0.000000000000000D+00,   1.540126540126542D-04, &
  -1.246901246901248D-03,   0.000000000000000D+00,  -1.680554894840610D-03, &
   0.000000000000000D+00,  -2.317127317127318D-04,   0.000000000000000D+00, &
   2.318052318052323D-03,   0.000000000000000D+00,   8.422133422133435D-04, &
   0.000000000000000D+00,  -1.685814185814187D-04,   0.000000000000000D+00, &
   5.035473785473781D-05,  -1.126651126651127D-03,   0.000000000000000D+00, &
   6.026381026381032D-04,   0.000000000000000D+00,   5.749211106353964D-04, &
   0.000000000000000D+00,   9.833916083916094D-05,   0.000000000000000D+00, &
   1.950119597178423D-04,   0.000000000000000D+00,  -8.597665950607133D-04, &
   0.000000000000000D+00,  -2.078803549391783D-04,   0.000000000000000D+00, &
  -9.409843968667602D-06,   0.000000000000000D+00,  -4.880442610222037D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.984760337701512D-04, &
   0.000000000000000D+00,  -3.477731418907888D-04,   0.000000000000000D+00, &
  -1.525333326803918D-04,   0.000000000000000D+00,  -3.656601772410588D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.137410806141464D-05,   0.000000000000000D+00,   2.861431654001316D-04, &
   0.000000000000000D+00,   1.018932907477779D-05,   0.000000000000000D+00, &
   1.619436241146775D-05,   0.000000000000000D+00,   8.525545634920552D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.860916173609655D-05, &
   0.000000000000000D+00,   2.409011696937392D-05,   0.000000000000000D+00, &
   1.490855206025486D-04,   0.000000000000000D+00,   4.485337181738122D-05, &
   0.000000000000000D+00,   1.317584325396841D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.945602237950619D-06, &
   0.000000000000000D+00,  -8.005410482190680D-05,   0.000000000000000D+00, &
  -2.669508280963380D-05,   0.000000000000000D+00,   1.259758902947763D-05, &
   0.000000000000000D+00,   4.522367198933839D-07,   0.000000000000000D+00, &
   2.583498677248813D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.036115587199170D-05, &
   0.000000000000000D+00,  -8.703564237619985D-05,   0.000000000000000D+00, &
  -4.422768064408935D-05,   0.000000000000000D+00,  -1.766357661408940D-05, &
   0.000000000000000D+00,  -7.609577922079499D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.330742052893008D-06,   0.000000000000000D+00, &
   2.643570060954652D-05,   0.000000000000000D+00,  -3.132347747306860D-05, &
   0.000000000000000D+00,  -1.077079483782730D-05,   0.000000000000000D+00, &
  -4.004422949735373D-06,   0.000000000000000D+00,  -1.617622467231855D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.696466722580747D-06,   0.000000000000000D+00, &
   2.662239104728006D-05,   0.000000000000000D+00,   3.814144883455196D-05, &
   0.000000000000000D+00,   1.133008114069452D-05,   0.000000000000000D+00, &
   3.149519751082174D-06,   0.000000000000000D+00,  -2.457259537337562D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.972272802800671D-07, &
   0.000000000000000D+00,  -1.182852403575287D-05,   0.000000000000000D+00, &
   2.349347312393303D-05,   0.000000000000000D+00,   2.114624169162339D-05, &
   0.000000000000000D+00,   8.463193124996781D-06,   0.000000000000000D+00, &
   3.552310681215345D-07,   0.000000000000000D+00,   8.168846669821569D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.520040757017048D-06,   0.000000000000000D+00,  -2.401948646711561D-05, &
   0.000000000000000D+00,  -1.046918234050215D-06,   0.000000000000000D+00, &
  -1.094766583712921D-06,   0.000000000000000D+00,   2.209297863595004D-06, &
   0.000000000000000D+00,   1.695929125128415D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.017482134814934D-06,   0.000000000000000D+00,  -6.626518333174421D-06, &
   0.000000000000000D+00,  -2.348966241918460D-05,   0.000000000000000D+00, &
  -1.027975280035413D-05,   0.000000000000000D+00,  -2.667281720016274D-06, &
   0.000000000000000D+00,   1.503352333430585D-06,   0.000000000000000D+00, &
  -4.176308039830182D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.538791565445274D-07,   0.000000000000000D+00, &
   8.799598141198264D-06,   0.000000000000000D+00,  -1.326485608154377D-06, &
   0.000000000000000D+00,  -6.464885629571788D-06,   0.000000000000000D+00, &
  -3.253988849692027D-06,   0.000000000000000D+00,  -1.115053256515652D-06, &
   0.000000000000000D+00,  -9.812185562919915D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.005647789597624D-08,   0.000000000000000D+00,  -4.411590270409586D-07, &
   0.000000000000000D+00,   1.551118364136671D-05,   0.000000000000000D+00, &
   7.897803724690038D-06,   0.000000000000000D+00,   3.965815530072582D-06, &
   0.000000000000000D+00,  -1.375645296543531D-06,   0.000000000000000D+00, &
  -1.313951280381987D-06,   0.000000000000000D+00,   2.231062259136308D-07 /)
lem(:, 77) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.781741612749494D-02, &
   1.532064692570853D-02,   6.751356017927672D-03,  -1.097316470416030D-02, &
  -1.746564277758852D-04,   3.591318751372156D-03,  -2.089459111456082D-04, &
  -9.687795551597248D-04,   1.720355616660905D-05,   3.309335551355016D-04, &
   1.070239796982052D-05,  -1.333988034726383D-04,  -1.121379702109609D-05, &
   6.069988039268694D-05,   8.232810420746795D-06,  -3.029495184582321D-05, &
  -5.652364665968928D-06,   1.626288539323927D-05,   3.851183360664664D-06, &
  -9.260207722769799D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.781741612749493D-02,  -1.532064692570851D-02,  -6.751356017927659D-03, &
   1.097316470416029D-02,   1.746564277758775D-04,  -3.591318751372141D-03, &
   2.089459111456093D-04,   9.687795551597224D-04,  -1.720355616660997D-05, &
  -3.309335551355009D-04,  -1.070239796981093D-05,   1.333988034726249D-04, &
   1.121379702109388D-05,  -6.069988039267312D-05,  -8.232810420751161D-06, &
   3.029495184581591D-05,   5.652364665975912D-06,  -1.626288539324521D-05, &
  -3.851183360659434D-06,   9.260207722778332D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.499923475244190D-03, &
   0.000000000000000D+00,   1.061234497500999D-03,   0.000000000000000D+00, &
  -2.410281041189384D-04,   0.000000000000000D+00,  -8.767235793301448D-06, &
   0.000000000000000D+00,   1.264714223447715D-05,   0.000000000000000D+00, &
  -7.083122308279874D-06,   0.000000000000000D+00,   3.756454192486483D-06, &
   0.000000000000000D+00,  -2.046956205798104D-06,   0.000000000000000D+00, &
   1.162417971464778D-06,   0.000000000000000D+00,  -6.879595292381923D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.030303030303030D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.906204906204907D-04,   0.000000000000000D+00, &
   1.262626262626263D-04,   0.000000000000000D+00,  -3.104303104303103D-04, &
   0.000000000000000D+00,   2.145077145077145D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.405002405002410D-05,   0.000000000000000D+00, &
  -4.288965003250718D-04,   0.000000000000000D+00,  -1.796814296814298D-04, &
   0.000000000000000D+00,   1.924001924001924D-04,   0.000000000000000D+00, &
   2.493934636791780D-04,   0.000000000000000D+00,  -6.105006105006109D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.542901542901546D-04, &
   0.000000000000000D+00,   7.115899973042827D-05,   0.000000000000000D+00, &
   2.074314574314576D-04,   0.000000000000000D+00,   9.764541014541015D-05, &
   0.000000000000000D+00,   6.281300398947468D-05,   0.000000000000000D+00, &
  -1.946582828935771D-04,   0.000000000000000D+00,  -8.961626608685443D-05, &
   0.000000000000000D+00,   1.176315514550812D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.358119005177840D-05,   0.000000000000000D+00, &
   1.177417353887942D-04,   0.000000000000000D+00,  -8.684126331185155D-05, &
   0.000000000000000D+00,  -6.644703335879809D-05,   0.000000000000000D+00, &
  -1.860119047619039D-05,   0.000000000000000D+00,   1.237152939939377D-06, &
   0.000000000000000D+00,  -2.328510842442741D-05,   0.000000000000000D+00, &
   1.056605932766924D-04,   0.000000000000000D+00,   3.277417170606018D-06, &
   0.000000000000000D+00,  -8.224395680858549D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.687530907345025D-06,   0.000000000000000D+00,  -2.153848903074917D-05, &
   0.000000000000000D+00,   5.368012333956610D-05,   0.000000000000000D+00, &
   1.291439468684051D-05,   0.000000000000000D+00,  -2.325148809523831D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.243171983552297D-06,   0.000000000000000D+00,  -2.445524891345321D-05, &
   0.000000000000000D+00,  -2.143503497992655D-05,   0.000000000000000D+00, &
   1.235589047605301D-05,   0.000000000000000D+00,   1.705109126984147D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.055507468599061D-06,   0.000000000000000D+00, &
   1.515727134922170D-05,   0.000000000000000D+00,  -2.562904072191997D-05, &
   0.000000000000000D+00,  -1.609392875356496D-05,   0.000000000000000D+00, &
   1.564191017315999D-06,   0.000000000000000D+00,   2.195973875661308D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.043831438593263D-08,   0.000000000000000D+00,  -2.930393965525743D-07, &
   0.000000000000000D+00,   2.071927675979365D-05,   0.000000000000000D+00, &
  -1.072314193822806D-05,   0.000000000000000D+00,  -5.114710462050646D-06, &
   0.000000000000000D+00,   1.470245610870641D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.397637510034892D-06, &
   0.000000000000000D+00,   3.310817560716653D-06,   0.000000000000000D+00, &
   1.761781182484671D-05,   0.000000000000000D+00,  -1.056923759813258D-06, &
   0.000000000000000D+00,  -1.312887055074556D-06,   0.000000000000000D+00, &
  -1.294685132575639D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.874107003060498D-07,   0.000000000000000D+00,  -8.369435824342350D-06, &
   0.000000000000000D+00,   3.056733353543195D-06,   0.000000000000000D+00, &
   8.790093825565050D-06,   0.000000000000000D+00,  -9.253622534872865D-07, &
   0.000000000000000D+00,  -1.100922731782234D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.290942198196240D-07, &
   0.000000000000000D+00,   2.197267853816607D-06,   0.000000000000000D+00, &
  -1.215311559486416D-05,   0.000000000000000D+00,  -6.765498916717544D-07, &
   0.000000000000000D+00,   2.431740712990790D-06,   0.000000000000000D+00, &
  -2.296694191224728D-07,   0.000000000000000D+00,   7.398200757574823D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.798364585259131D-09, &
   0.000000000000000D+00,   9.208562260789209D-07,   0.000000000000000D+00, &
   2.667917443417206D-06,   0.000000000000000D+00,  -9.443115440042624D-06, &
   0.000000000000000D+00,  -9.937678059176388D-07,   0.000000000000000D+00, &
   1.297395096223227D-06,   0.000000000000000D+00,   6.746115754905952D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.890285696315752D-07, &
   0.000000000000000D+00,   3.991262961598747D-06,   0.000000000000000D+00, &
   3.212908999932379D-06,   0.000000000000000D+00,  -3.973543926241989D-06, &
   0.000000000000000D+00,   2.770937829531348D-07,   0.000000000000000D+00, &
   3.659902690314799D-07,   0.000000000000000D+00,  -4.336365267761846D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.695795671872977D-07,   0.000000000000000D+00, &
  -2.357989969918697D-06,   0.000000000000000D+00,   5.827266103000359D-06, &
   0.000000000000000D+00,   3.069317448398303D-06,   0.000000000000000D+00, &
  -1.409858587592951D-06,   0.000000000000000D+00,  -3.498514676667250D-07, &
   0.000000000000000D+00,  -4.079765584892631D-07,   0.000000000000000D+00 /)
lem(:, 78) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.555231582719479D-02,   2.760718592210493D-02,   8.537511894424671D-03, &
  -2.231518802965338D-02,  -8.206535510513618D-04,   8.380215081932863D-03, &
  -1.508837473286094D-03,  -1.156934899770895D-03,   7.338588098772567D-04, &
   9.534368925214405D-05,  -2.372657121311918D-04,  -4.215714078160249D-06, &
   9.263937268019984D-05,  -4.131821346493561D-06,  -4.197892196532723D-05, &
   3.447099141339309D-06,   2.125460822998211D-05,  -2.226064873802457D-06, &
  -1.170922743613323D-05,   1.381742117019877D-06,   6.890343871094666D-06, &
  -8.633551517828353D-07,  -1.555231582719480D-02,   2.760718592210487D-02, &
   8.537511894424655D-03,  -2.231518802965344D-02,  -8.206535510513584D-04, &
   8.380215081932870D-03,  -1.508837473286090D-03,  -1.156934899770890D-03, &
   7.338588098772599D-04,   9.534368925214117D-05,  -2.372657121311947D-04, &
  -4.215714078158037D-06,   9.263937268020427D-05,  -4.131821346500286D-06, &
  -4.197892196533205D-05,   3.447099141341606D-06,   2.125460822998726D-05, &
  -2.226064873807666D-06,  -1.170922743613434D-05,   1.381742117029426D-06, &
   6.890343871099949D-06,  -8.633551517951512D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.908708063747469D-03,   0.000000000000000D+00, &
  -4.264014327112208D-03,   0.000000000000000D+00,   3.493128555517588D-04, &
   0.000000000000000D+00,   2.446196032924193D-04,   0.000000000000000D+00, &
  -7.072573090718224D-05,   0.000000000000000D+00,   2.328240961853364D-05, &
   0.000000000000000D+00,  -8.835112804498136D-06,   0.000000000000000D+00, &
   3.762414590320056D-06,   0.000000000000000D+00,  -1.756816044834179D-06, &
   0.000000000000000D+00,   8.838542384292053D-07,   0.000000000000000D+00, &
   1.388888888888889D-03,  -1.547619047619046D-03,   0.000000000000000D+00, &
   7.395382395382400D-04,   0.000000000000000D+00,  -7.611832611832602D-04, &
   4.004329004329002D-04,   0.000000000000000D+00,   7.900432900432899D-04, &
   0.000000000000000D+00,  -6.007881007881003D-04,   0.000000000000000D+00, &
  -3.624153624153628D-04,   0.000000000000000D+00,   4.472957597957595D-04, &
   1.956376956376959D-04,   0.000000000000000D+00,  -1.889380460809033D-04, &
   0.000000000000000D+00,  -4.503945128945133D-04,   0.000000000000000D+00, &
   2.192252192252166D-05,   0.000000000000000D+00,   2.808302808302809D-04, &
   0.000000000000000D+00,   1.781204906204906D-04,   0.000000000000000D+00, &
  -1.307720057720058D-04,  -2.081252081252020D-05,   0.000000000000000D+00, &
  -9.546009546009546D-05,   0.000000000000000D+00,   1.375757179328605D-04, &
   0.000000000000000D+00,   1.592157842157840D-04,   0.000000000000000D+00, &
   2.089413854118975D-06,   0.000000000000000D+00,   3.591180061769338D-07, &
   0.000000000000000D+00,  -1.590117398940929D-04,   0.000000000000000D+00, &
  -3.942646405881690D-05,   0.000000000000000D+00,  -1.308784914023898D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.814714697064608D-07, &
   0.000000000000000D+00,   3.878882555353142D-05,   0.000000000000000D+00, &
  -9.020289351171701D-05,   0.000000000000000D+00,  -2.969896984143306D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.382378543369577D-06,   0.000000000000000D+00,   6.437705741111250D-06, &
   0.000000000000000D+00,   5.799181253902614D-05,   0.000000000000000D+00, &
  -1.234015087294909D-06,   0.000000000000000D+00,   2.247643849206352D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.644510288475618D-07, &
   0.000000000000000D+00,  -1.213884727816445D-06,   0.000000000000000D+00, &
   1.220050804029137D-05,   0.000000000000000D+00,   3.554252402331929D-05, &
   0.000000000000000D+00,   8.938905423280377D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.945602237966173D-08, &
   0.000000000000000D+00,   6.883811063357790D-08,   0.000000000000000D+00, &
  -1.583115456954464D-05,   0.000000000000000D+00,   2.678249521806108D-07, &
   0.000000000000000D+00,  -1.076199557236670D-06,   0.000000000000000D+00, &
  -9.164961557539676D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.010019392372205D-06, &
   0.000000000000000D+00,   1.670478644162790D-06,   0.000000000000000D+00, &
  -2.268310048761871D-05,   0.000000000000000D+00,  -8.371576071846962D-06, &
   0.000000000000000D+00,  -6.469902710137092D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.953428215008890D-07,   0.000000000000000D+00, &
  -2.145500361939718D-08,   0.000000000000000D+00,   4.049330623451054D-06, &
   0.000000000000000D+00,  -9.254783999975966D-06,   0.000000000000000D+00, &
   2.624599792568476D-07,   0.000000000000000D+00,   3.628641323953772D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.890993752207801D-08,   0.000000000000000D+00, &
   9.167986856768554D-08,   0.000000000000000D+00,   5.869689335404353D-06, &
   0.000000000000000D+00,   6.989365770725115D-06,   0.000000000000000D+00, &
   3.817119295634970D-06,   0.000000000000000D+00,   3.184749278499335D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.792975275183929D-09, &
   0.000000000000000D+00,   4.999669532244341D-07,   0.000000000000000D+00, &
  -2.185049270110810D-06,   0.000000000000000D+00,   7.871332497323119D-06, &
   0.000000000000000D+00,   3.896197085439298D-06,   0.000000000000000D+00, &
   1.356336805555515D-06,   0.000000000000000D+00,  -1.575970890546049D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.076423397125972D-07,   0.000000000000000D+00,   4.753113544989233D-07, &
   0.000000000000000D+00,  -5.128889188349949D-06,   0.000000000000000D+00, &
   1.278278387418141D-06,   0.000000000000000D+00,  -1.881527760434014D-06, &
   0.000000000000000D+00,  -1.517734900913655D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.574631776550527D-08,   0.000000000000000D+00,  -2.310476557830246D-07, &
   0.000000000000000D+00,  -1.648726533100699D-06,   0.000000000000000D+00, &
  -5.259692908004478D-06,   0.000000000000000D+00,  -2.456582046425760D-06, &
   0.000000000000000D+00,  -1.049729824754196D-06,   0.000000000000000D+00, &
   7.478229371540282D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.778142977017632D-09,   0.000000000000000D+00, &
  -2.243604766761965D-07,   0.000000000000000D+00,   1.584379109586482D-06, &
   0.000000000000000D+00,  -1.959494720490262D-06,   0.000000000000000D+00, &
  -1.096180295399148D-06,   0.000000000000000D+00,   2.159562771832410D-07, &
   0.000000000000000D+00,   7.572231235561210D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.653006419842522D-10,   0.000000000000000D+00,   2.346587435700158D-07, &
   0.000000000000000D+00,  -5.668679145348312D-07,   0.000000000000000D+00, &
   3.790236721662274D-06,   0.000000000000000D+00,   8.330933553020353D-07, &
   0.000000000000000D+00,   1.652414448434933D-06,   0.000000000000000D+00, &
   6.399748526212923D-07,   0.000000000000000D+00,  -3.823430546911281D-07 /)
lem(:, 79) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.513641844631533D-02,   1.855980846614052D-03, &
   1.821335648510804D-02,  -4.737793696791341D-03,  -4.064528125428814D-03, &
   2.634401118952888D-03,  -2.892337249427264D-04,  -5.159944756949398D-04, &
   2.684965961698339D-04,   8.402662849895194D-05,  -1.166347561624069D-04, &
  -9.075132781409354D-06,   5.213965032484372D-05,  -4.106062040559259D-06, &
  -2.510195272153777D-05,   5.302273787576912D-06,   1.297258495423395D-05, &
  -4.312474168027370D-06,  -7.128034730672842D-06,   3.186834176510262D-06, &
   4.126054305304164D-06,   0.000000000000000D+00,   3.513641844631532D-02, &
  -1.855980846614067D-03,  -1.821335648510802D-02,   4.737793696791350D-03, &
   4.064528125428814D-03,  -2.634401118952893D-03,   2.892337249427263D-04, &
   5.159944756949492D-04,  -2.684965961698344D-04,  -8.402662849896427D-05, &
   1.166347561624114D-04,   9.075132781418429D-06,  -5.213965032485385D-05, &
   4.106062040548885D-06,   2.510195272154836D-05,  -5.302273787558203D-06, &
  -1.297258495424837D-05,   4.312474168010029D-06,   7.128034730687558D-06, &
  -3.186834176498014D-06,  -4.126054305309634D-06,   0.000000000000000D+00, &
   1.756820922315767D-02,   0.000000000000000D+00,  -1.928473039599677D-02, &
   0.000000000000000D+00,   1.087057870273520D-02,   0.000000000000000D+00, &
  -3.036954111898590D-03,   0.000000000000000D+00,   4.767184462607178D-04, &
   0.000000000000000D+00,  -1.180399941884553D-04,   0.000000000000000D+00, &
   3.836691250319017D-05,   0.000000000000000D+00,  -1.498162319120391D-05, &
   0.000000000000000D+00,   6.678194621417613D-06,   0.000000000000000D+00, &
  -3.289862183394183D-06,   0.000000000000000D+00,   1.752253355415624D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.388888888888886D-03, &
   0.000000000000000D+00,   5.952380952380964D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.262626262626263D-04,   0.000000000000000D+00, &
   1.617364117364118D-03,   0.000000000000000D+00,  -2.705627705627707D-04, &
   0.000000000000000D+00,  -6.660006660006667D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.068376068376068D-04,   0.000000000000000D+00, &
  -1.693412407698123D-04,   0.000000000000000D+00,  -1.194927757427758D-03, &
   0.000000000000000D+00,   4.162504162504113D-06,   0.000000000000000D+00, &
   3.036645893788751D-04,   0.000000000000000D+00,   4.230144855144851D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.526251526251515D-05, &
   0.000000000000000D+00,  -1.022786737072452D-04,   0.000000000000000D+00, &
   1.946548821548823D-04,   0.000000000000000D+00,   5.877802752802750D-04, &
   0.000000000000000D+00,   4.162504162504114D-06,   0.000000000000000D+00, &
  -1.093677564265796D-05,   0.000000000000000D+00,  -2.216635488694311D-04, &
   0.000000000000000D+00,  -1.254719545160720D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.264709147061776D-07,   0.000000000000000D+00, &
   1.229707112060041D-05,   0.000000000000000D+00,   3.799305269893504D-05, &
   0.000000000000000D+00,  -1.721521942110177D-04,   0.000000000000000D+00, &
  -1.966042493386241D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.381583328951042D-07,   0.000000000000000D+00, &
   2.556890134289517D-05,   0.000000000000000D+00,   9.577652524633944D-05, &
   0.000000000000000D+00,  -1.449152640982166D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.457823795285642D-07,   0.000000000000000D+00,  -8.215826714278623D-06, &
   0.000000000000000D+00,   2.877535046652703D-05,   0.000000000000000D+00, &
   9.309324732965395D-05,   0.000000000000000D+00,   5.296172288359784D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.477447086458110D-08,   0.000000000000000D+00,  -7.055637860592325D-07, &
   0.000000000000000D+00,  -3.019802265158303D-05,   0.000000000000000D+00, &
   9.898890328940613D-07,   0.000000000000000D+00,   2.635168650793645D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.602503496780862D-08,   0.000000000000000D+00, &
  -7.601545914238700D-07,   0.000000000000000D+00,   1.404415587852072D-06, &
   0.000000000000000D+00,  -4.529537276126862D-05,   0.000000000000000D+00, &
  -2.496599176286684D-05,   0.000000000000000D+00,  -1.624961952110389D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.891549610053248D-09,   0.000000000000000D+00,   6.913304884093713D-07, &
   0.000000000000000D+00,   1.703606663897508D-06,   0.000000000000000D+00, &
   8.414982230908416D-06,   0.000000000000000D+00,  -2.812993432787899D-05, &
   0.000000000000000D+00,  -1.235675674152240D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.097228656370468D-08, &
   0.000000000000000D+00,   5.480247605701108D-07,   0.000000000000000D+00, &
   1.076491941971605D-05,   0.000000000000000D+00,   1.817308552842437D-05, &
   0.000000000000000D+00,   7.662422213204214D-07,   0.000000000000000D+00, &
   6.141680946368469D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.158694921798788D-08,   0.000000000000000D+00,   4.992665722576519D-07, &
   0.000000000000000D+00,  -3.429486172105923D-06,   0.000000000000000D+00, &
   1.489316182079550D-05,   0.000000000000000D+00,   1.579163566468239D-05, &
   0.000000000000000D+00,   5.266814348845667D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.218662882414605D-08, &
   0.000000000000000D+00,  -6.837819379885558D-07,   0.000000000000000D+00, &
   2.670112456823198D-07,   0.000000000000000D+00,  -1.069840312062509D-05, &
   0.000000000000000D+00,   4.768858528624199D-06,   0.000000000000000D+00, &
   2.017116274928641D-06,   0.000000000000000D+00,  -2.700125914847626D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.093291171838460D-10, &
   0.000000000000000D+00,   1.557530540234983D-07,   0.000000000000000D+00, &
  -8.942059682998175D-07,   0.000000000000000D+00,  -2.315614581786451D-06, &
   0.000000000000000D+00,  -1.398597929891694D-05,   0.000000000000000D+00, &
  -4.412272478678561D-06,   0.000000000000000D+00,  -2.445022357742130D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.856542683559307D-08, &
   0.000000000000000D+00,  -1.622962659310291D-08,   0.000000000000000D+00, &
   2.192384496162285D-06,   0.000000000000000D+00,  -2.241016804371679D-06, &
   0.000000000000000D+00,  -7.274897411616240D-06,   0.000000000000000D+00, &
  -1.450179378558912D-06,   0.000000000000000D+00,   1.327331725891293D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.703777331440426D-08,   0.000000000000000D+00, &
   5.339037625905814D-07,   0.000000000000000D+00,  -5.133735642948517D-07, &
   0.000000000000000D+00,   7.364105820281045D-06,   0.000000000000000D+00, &
   2.469680204055095D-06,   0.000000000000000D+00,   8.377437249680638D-07, &
   0.000000000000000D+00,   1.238284014162916D-06,   0.000000000000000D+00 /)
lem(:, 80) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.184105275731644D-03,  -3.011693009684170D-02,   1.849231825353645D-02, &
   4.655374731837839D-02,  -3.181740849397368D-02,  -3.733422962208509D-02, &
   1.724045767075880D-02,   1.833018731824504D-02,  -1.146654390433221D-03, &
  -5.469780997709242D-03,  -1.964072662354918D-03,   9.253492401559415D-04, &
   8.667273365610284D-04,  -1.379062449415210D-04,  -3.640843188667297D-04, &
  -5.042978373437529D-06,   1.618445114900590D-04,   2.519259142930877D-05, &
  -7.679856474795462D-05,  -2.208003657696992D-05,   3.866862293117645D-05, &
   1.613072104115638D-05,   5.184105275731593D-03,  -3.011693009684172D-02, &
   1.849231825353641D-02,   4.655374731837835D-02,  -3.181740849397376D-02, &
  -3.733422962208503D-02,   1.724045767075879D-02,   1.833018731824507D-02, &
  -1.146654390433216D-03,  -5.469780997709252D-03,  -1.964072662354906D-03, &
   9.253492401559560D-04,   8.667273365609959D-04,  -1.379062449415310D-04, &
  -3.640843188667003D-04,  -5.042978373441552D-06,   1.618445114900218D-04, &
   2.519259142933970D-05,  -7.679856474791898D-05,  -2.208003657701742D-05, &
   3.866862293115864D-05,   1.613072104120169D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.936507936507927D-04,   1.659451659451659D-03,   0.000000000000000D+00, &
  -1.010101010101008D-03,   0.000000000000000D+00,   1.623376623376627D-04, &
  -2.586302586302589D-03,   0.000000000000000D+00,  -6.007881007881003D-04, &
   0.000000000000000D+00,   5.283605283605283D-03,   0.000000000000000D+00, &
   5.019980019980018D-04,   0.000000000000000D+00,   7.492507492507493D-05, &
  -2.886002886002885D-03,   0.000000000000000D+00,   8.422133422133437D-04, &
   0.000000000000000D+00,   1.859251859251859D-04,   0.000000000000000D+00, &
  -1.492951492951495D-03,   0.000000000000000D+00,  -1.913364413364413D-03, &
   0.000000000000000D+00,  -2.955377955377956D-04,   0.000000000000000D+00, &
  -7.388444888444898D-05,   2.703505644682109D-03,   0.000000000000000D+00, &
   9.538663950428665D-04,   0.000000000000000D+00,  -2.078803549391782D-04, &
   0.000000000000000D+00,  -1.271196124137305D-05,   0.000000000000000D+00, &
  -1.311760135289545D-03,   0.000000000000000D+00,   7.117882117882126D-04, &
   0.000000000000000D+00,   6.807780092093814D-04,   0.000000000000000D+00, &
   1.219368866427690D-04,   0.000000000000000D+00,   6.363632282749176D-07, &
   2.273268527138494D-04,   0.000000000000000D+00,  -1.006337144108041D-03, &
   0.000000000000000D+00,  -2.454481363150097D-04,   0.000000000000000D+00, &
   1.018932907477784D-05,   0.000000000000000D+00,  -2.117208180965921D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.520473334714812D-04,   0.000000000000000D+00,  -4.083077225492086D-04, &
   0.000000000000000D+00,  -1.874914945735377D-04,   0.000000000000000D+00, &
  -2.661543392967548D-05,   0.000000000000000D+00,   2.092633928571443D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.270691125180314D-05, &
   0.000000000000000D+00,   3.404361376497595D-04,   0.000000000000000D+00, &
   1.301641684768630D-05,   0.000000000000000D+00,   1.259758902947765D-05, &
   0.000000000000000D+00,   1.824428630615738D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.286187496713834D-05, &
   0.000000000000000D+00,   2.308493020567347D-05,   0.000000000000000D+00, &
   1.779228898036949D-04,   0.000000000000000D+00,   4.663526940036241D-05, &
   0.000000000000000D+00,   1.233108969418981D-06,   0.000000000000000D+00, &
  -8.243709415584545D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.389843879815788D-06,   0.000000000000000D+00,  -9.246112019028978D-05, &
   0.000000000000000D+00,  -3.613879408905619D-05,   0.000000000000000D+00, &
   1.671766761491147D-05,   0.000000000000000D+00,  -1.077079483782731D-05, &
   0.000000000000000D+00,  -1.148482593795078D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.276561042343781D-05,   0.000000000000000D+00, &
  -1.005601389233102D-04,   0.000000000000000D+00,  -5.285026041415675D-05, &
   0.000000000000000D+00,  -1.340403758687481D-05,   0.000000000000000D+00, &
  -2.811316287878874D-06,   0.000000000000000D+00,   7.133979301948694D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.049824377324716D-05,   0.000000000000000D+00, &
   3.274533157424525D-05,   0.000000000000000D+00,  -3.587181761923502D-05, &
   0.000000000000000D+00,  -8.529282459204400D-06,   0.000000000000000D+00, &
   8.463193124996803D-06,   0.000000000000000D+00,   3.972129216269742D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.070376774778129D-06, &
   0.000000000000000D+00,   2.986942724794427D-05,   0.000000000000000D+00, &
   4.627535821274050D-05,   0.000000000000000D+00,   1.116118514600174D-05, &
   0.000000000000000D+00,   7.226982781108779D-06,   0.000000000000000D+00, &
   4.550706455003589D-06,   0.000000000000000D+00,   2.881030102710757D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.250384573408557D-07,   0.000000000000000D+00, &
  -1.384340334737458D-05,   0.000000000000000D+00,   2.579566038119520D-05, &
   0.000000000000000D+00,   2.467299438917215D-05,   0.000000000000000D+00, &
   2.327514856933929D-06,   0.000000000000000D+00,  -2.667281720016261D-06, &
   0.000000000000000D+00,  -2.580675798448778D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.699805205189676D-06,   0.000000000000000D+00,  -2.836821317395789D-05, &
   0.000000000000000D+00,  -1.899148506563503D-06,   0.000000000000000D+00, &
  -2.375890389187158D-06,   0.000000000000000D+00,  -6.603503917957010D-06, &
   0.000000000000000D+00,  -2.243511154644223D-06,   0.000000000000000D+00, &
  -3.061835489815674D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.268287879256125D-06,   0.000000000000000D+00, &
  -6.896161986548835D-06,   0.000000000000000D+00,  -2.765599544166694D-05, &
   0.000000000000000D+00,  -9.687344821097315D-06,   0.000000000000000D+00, &
  -1.659719774266934D-06,   0.000000000000000D+00,  -1.375645296543523D-06, &
   0.000000000000000D+00,  -4.205948267227394D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.005027929794071D-07,   0.000000000000000D+00,   1.005805259130412D-05, &
   0.000000000000000D+00,  -4.509401944155013D-07,   0.000000000000000D+00, &
  -7.611565212819532D-06,   0.000000000000000D+00,   1.956160101549133D-06, &
   0.000000000000000D+00,   3.498563068875601D-06,   0.000000000000000D+00, &
   2.557951624114927D-07,   0.000000000000000D+00,   2.129015190540212D-07 /)
lem(:, 81) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.692651132631502D-03,   1.574771627430101D-03, &
  -1.819805114352389D-02,   9.545439480137950D-03,   1.116452381355095D-02, &
  -9.003152083533470D-03,  -3.277982216017528D-03,   3.460348138240682D-03, &
   9.746243790219218D-04,  -9.254876048890535D-04,  -4.427905020093201D-04, &
   2.850843435817935D-04,   2.050681449238111D-04,  -1.004549089566297D-04, &
  -1.009013058487248D-04,   3.918170463143223D-05,   5.288590488072803D-05, &
  -1.644605330580811D-05,  -2.934087087275944D-05,   7.252417822734766D-06, &
   1.710158406000729D-05,   0.000000000000000D+00,  -6.692651132631494D-03, &
  -1.574771627430112D-03,   1.819805114352393D-02,  -9.545439480137939D-03, &
  -1.116452381355095D-02,   9.003152083533455D-03,   3.277982216017528D-03, &
  -3.460348138240684D-03,  -9.746243790219267D-04,   9.254876048890567D-04, &
   4.427905020093267D-04,  -2.850843435818095D-04,  -2.050681449238063D-04, &
   1.004549089566425D-04,   1.009013058487149D-04,  -3.918170463143883D-05, &
  -5.288590488071635D-05,   1.644605330580974D-05,   2.934087087275129D-05, &
  -7.252417822736484D-06,  -1.710158406000181D-05,   0.000000000000000D+00, &
  -3.346325566315736D-03,   0.000000000000000D+00,   1.867251673263178D-03, &
   0.000000000000000D+00,  -3.099478532383893D-04,   0.000000000000000D+00, &
  -2.410281041189349D-04,   0.000000000000000D+00,   1.300473309339610D-04, &
   0.000000000000000D+00,  -2.276485602205906D-05,   0.000000000000000D+00, &
   3.495566853436879D-06,   0.000000000000000D+00,   7.365596455650198D-08, &
   0.000000000000000D+00,  -5.976253258188272D-07,   0.000000000000000D+00, &
   5.333674001363997D-07,   0.000000000000000D+00,  -3.923412760824809D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.645502645502644D-04, &
   0.000000000000000D+00,  -3.848003848003852D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.104303104303103D-04,   0.000000000000000D+00, &
  -2.705627705627707D-04,   0.000000000000000D+00,   4.932633504062075D-04, &
   0.000000000000000D+00,   7.932543646829360D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.145003145003144D-04,   0.000000000000000D+00, &
   2.493934636791780D-04,   0.000000000000000D+00,   1.290376290376290D-04, &
   0.000000000000000D+00,  -3.145003145003146D-05,   0.000000000000000D+00, &
  -4.328607900036463D-04,   0.000000000000000D+00,  -7.284382284382271D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.160366866249219D-04, &
   0.000000000000000D+00,   2.528867024665342D-04,   0.000000000000000D+00, &
  -8.961626608685440D-05,   0.000000000000000D+00,   3.526566026566076D-06, &
   0.000000000000000D+00,  -1.780137074254723D-04,   0.000000000000000D+00, &
   7.955396610858806D-05,   0.000000000000000D+00,   2.006338666002531D-04, &
   0.000000000000000D+00,   3.188872674166789D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.503561683128255D-05,   0.000000000000000D+00, &
  -2.194425801236947D-04,   0.000000000000000D+00,  -9.145910848697217D-05, &
   0.000000000000000D+00,   3.277417170606041D-06,   0.000000000000000D+00, &
  -3.410218253968255D-05,   0.000000000000000D+00,  -1.690775684583737D-05, &
   0.000000000000000D+00,   1.394950079160605D-04,   0.000000000000000D+00, &
  -9.114552458205715D-05,   0.000000000000000D+00,  -5.432357889788221D-05, &
   0.000000000000000D+00,  -1.169965977435013D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.620081230872849D-06,   0.000000000000000D+00, &
  -3.125725814801450D-05,   0.000000000000000D+00,   1.207512817419939D-04, &
   0.000000000000000D+00,   1.023228577408135D-05,   0.000000000000000D+00, &
   1.235589047605301D-05,   0.000000000000000D+00,   1.472594246031751D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.636473801887350D-06,   0.000000000000000D+00,  -3.094600617820435D-05, &
   0.000000000000000D+00,   5.254033891804797D-05,   0.000000000000000D+00, &
   1.368909191084117D-05,   0.000000000000000D+00,   8.046424062049038D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.164419755943458D-06,   0.000000000000000D+00, &
  -2.526974528589816D-05,   0.000000000000000D+00,  -2.787452435790026D-05, &
   0.000000000000000D+00,   6.908324397219724D-07,   0.000000000000000D+00, &
  -5.114710462050652D-06,   0.000000000000000D+00,  -2.771389490139412D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.320718394733558D-06,   0.000000000000000D+00,   1.819281584391291D-05, &
   0.000000000000000D+00,  -2.482585939309588D-05,   0.000000000000000D+00, &
  -1.663772788230153D-05,   0.000000000000000D+00,  -8.665576303538664D-06, &
   0.000000000000000D+00,  -3.638035864598408D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.156980870163133D-08,   0.000000000000000D+00,  -1.182376478032563D-06, &
   0.000000000000000D+00,   2.456358228501179D-05,   0.000000000000000D+00, &
  -6.006163673750178D-06,   0.000000000000000D+00,  -2.177083166175167D-07, &
   0.000000000000000D+00,  -9.253622534872890D-07,   0.000000000000000D+00, &
   3.787174197328991D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.626428625890030D-06,   0.000000000000000D+00,   1.757465960252415D-06, &
   0.000000000000000D+00,   1.884707964240037D-05,   0.000000000000000D+00, &
   5.427511399409951D-06,   0.000000000000000D+00,   4.607480193417823D-06, &
   0.000000000000000D+00,   6.673285481880158D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.062786094418747D-06, &
   0.000000000000000D+00,  -9.289440698446000D-06,   0.000000000000000D+00, &
   8.217540702145458D-07,   0.000000000000000D+00,   5.712148150407415D-06, &
   0.000000000000000D+00,   2.829978181929700D-06,   0.000000000000000D+00, &
   1.297395096223219D-06,   0.000000000000000D+00,   9.366311856555393D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.618658234622322D-07, &
   0.000000000000000D+00,   2.942059199630860D-06,   0.000000000000000D+00, &
  -1.304671610651562D-05,   0.000000000000000D+00,  -3.460404898445454D-06, &
   0.000000000000000D+00,  -3.045845639587246D-06,   0.000000000000000D+00, &
  -4.766632487057150D-07,   0.000000000000000D+00,  -4.452083135184844D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.790988688095131D-09,   0.000000000000000D+00,   8.748603385109318D-07, &
   0.000000000000000D+00,   4.031542816922381D-06,   0.000000000000000D+00, &
  -8.443207133725538D-06,   0.000000000000000D+00,  -3.251459586304583D-06, &
   0.000000000000000D+00,  -2.846139320748676D-06,   0.000000000000000D+00, &
  -3.498514676667224D-07,   0.000000000000000D+00,  -1.554418210912397D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.878563134604132D-07,   0.000000000000000D+00, &
   4.018356769951131D-06,   0.000000000000000D+00,   4.602803640183311D-06, &
   0.000000000000000D+00,  -4.290252386290322D-07,   0.000000000000000D+00, &
  -4.810185474243258D-08,   0.000000000000000D+00,  -5.788675919814441D-07, &
   0.000000000000000D+00,  -1.179257618466572D-07,   0.000000000000000D+00 /)
lem(:, 82) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.166423687039608D-02,  -1.254872087368407D-02,   1.872853399765093D-02, &
   1.584102853956875D-02,  -1.819429922818621D-02,  -7.315948485867576D-03, &
   8.437554315963193D-03,   6.809043941359936D-04,  -1.908542529854405D-03, &
   4.328822672942224D-04,   3.997437425936471D-04,  -2.002464187125666D-04, &
  -1.282097207085207D-04,   9.034047229772065D-05,   5.227037204529333D-05, &
  -4.356382023813986D-05,  -2.486025554570149D-05,   2.251276073928044D-05, &
   1.316293512263782D-05,  -1.236469997535114D-05,  -7.543025012499099D-06, &
   7.153757382009001D-06,  -1.166423687039608D-02,  -1.254872087368406D-02, &
   1.872853399765098D-02,   1.584102853956876D-02,  -1.819429922818616D-02, &
  -7.315948485867571D-03,   8.437554315963193D-03,   6.809043941359881D-04, &
  -1.908542529854398D-03,   4.328822672942200D-04,   3.997437425936483D-04, &
  -2.002464187125667D-04,  -1.282097207085266D-04,   9.034047229771777D-05, &
   5.227037204529690D-05,  -4.356382023813309D-05,  -2.486025554570639D-05, &
   2.251276073927657D-05,   1.316293512264182D-05,  -1.236469997535275D-05, &
  -7.543025012508755D-06,   7.153757382017463D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.889725952916128D-03,   0.000000000000000D+00, &
  -6.195576372727134D-04,   0.000000000000000D+00,   1.189617906669296D-03, &
   0.000000000000000D+00,  -3.669294049386292D-04,   0.000000000000000D+00, &
   5.734518722202029D-06,   0.000000000000000D+00,   8.261500187223733D-06, &
   0.000000000000000D+00,  -4.951546736585639D-06,   0.000000000000000D+00, &
   2.548732464409102D-06,   0.000000000000000D+00,  -1.320340008847312D-06, &
   0.000000000000000D+00,   7.088335971534997D-07,   0.000000000000000D+00, &
   5.158730158730160D-04,   7.395382395382401D-04,   0.000000000000000D+00, &
  -1.562049062049062D-03,   0.000000000000000D+00,  -2.092352092352091D-04, &
   7.654382654382653D-04,   0.000000000000000D+00,  -3.624153624153628D-04, &
   0.000000000000000D+00,   5.019980019980018D-04,   0.000000000000000D+00, &
   7.339882339882335D-04,   0.000000000000000D+00,   1.650086025086026D-04, &
  -7.214082214082212D-04,   0.000000000000000D+00,  -3.620850049421480D-04, &
   0.000000000000000D+00,   1.781204906204906D-04,   0.000000000000000D+00, &
   2.455877455877460D-04,   0.000000000000000D+00,  -2.203352203352203D-04, &
   0.000000000000000D+00,  -4.050463425463431D-04,   0.000000000000000D+00, &
  -1.514978077478074D-04,   2.074722662957931D-05,   0.000000000000000D+00, &
   3.204747322394381D-04,   0.000000000000000D+00,   1.896256754450034D-04, &
   0.000000000000000D+00,  -3.942646405881693D-05,   0.000000000000000D+00, &
  -2.448531860296517D-05,   0.000000000000000D+00,  -1.122407004759948D-04, &
   0.000000000000000D+00,   1.382277852866087D-04,   0.000000000000000D+00, &
   1.712033881151527D-04,   0.000000000000000D+00,   8.362342059492802D-05, &
   2.505234703376726D-06,   0.000000000000000D+00,   1.890094769351842D-06, &
   0.000000000000000D+00,  -1.729900646309314D-04,   0.000000000000000D+00, &
  -6.250002863779938D-05,   0.000000000000000D+00,  -1.234015087294892D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.061921566588233D-08,   0.000000000000000D+00,   4.785054127159400D-05, &
   0.000000000000000D+00,  -8.895688075254630D-05,   0.000000000000000D+00, &
  -5.639273262749467D-05,   0.000000000000000D+00,  -2.046130952380960D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.525885878827213D-06, &
   0.000000000000000D+00,   2.803139413046423D-06,   0.000000000000000D+00, &
   7.026355710566226D-05,   0.000000000000000D+00,   5.920512674817978D-06, &
   0.000000000000000D+00,  -1.076199557236679D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.671116587968312D-07, &
   0.000000000000000D+00,   1.359042574212969D-06,   0.000000000000000D+00, &
   4.641202880366938D-06,   0.000000000000000D+00,   4.554515925573397D-05, &
   0.000000000000000D+00,   1.621337858155773D-05,   0.000000000000000D+00, &
  -1.097986937830219D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.547017257884914D-08,   0.000000000000000D+00,  -7.744392503445399D-07, &
   0.000000000000000D+00,  -1.399423038859705D-05,   0.000000000000000D+00, &
  -1.004234707642597D-05,   0.000000000000000D+00,   6.308437155613394D-06, &
   0.000000000000000D+00,   2.624599792568412D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.275410416469557D-06,   0.000000000000000D+00, &
   1.657579961355663D-06,   0.000000000000000D+00,  -2.235424827762496D-05, &
   0.000000000000000D+00,  -1.660312599836974D-05,   0.000000000000000D+00, &
  -2.956344509078878D-06,   0.000000000000000D+00,   1.586503051346846D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.956459971131639D-07,   0.000000000000000D+00, &
  -1.068559568189365D-06,   0.000000000000000D+00,   7.382336836850024D-06, &
   0.000000000000000D+00,  -9.217905255340864D-06,   0.000000000000000D+00, &
  -4.585009132781483D-06,   0.000000000000000D+00,   1.356336805555518D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.816395611379936D-08, &
   0.000000000000000D+00,   8.742715533670003D-07,   0.000000000000000D+00, &
   4.018570766860834D-06,   0.000000000000000D+00,   1.286202457551652D-05, &
   0.000000000000000D+00,   3.331286468060782D-06,   0.000000000000000D+00, &
   5.764995999371311D-07,   0.000000000000000D+00,  -1.016109338252954D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.329910777581229D-09,   0.000000000000000D+00, &
   4.421271583743143D-07,   0.000000000000000D+00,  -1.507048393353319D-06, &
   0.000000000000000D+00,   5.215304633652444D-06,   0.000000000000000D+00, &
   8.394927555132406D-06,   0.000000000000000D+00,   1.071338961964005D-06, &
   0.000000000000000D+00,  -1.049729824754199D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.175437864360622D-07,   0.000000000000000D+00,   7.328361831722516D-07, &
   0.000000000000000D+00,  -6.234692438568276D-06,   0.000000000000000D+00, &
  -1.406506253004953D-06,   0.000000000000000D+00,   3.647611069485093D-07, &
   0.000000000000000D+00,  -7.701801372455798D-07,   0.000000000000000D+00, &
   5.715482705229628D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.078830883163391D-07,   0.000000000000000D+00, &
  -7.555583329428351D-07,   0.000000000000000D+00,  -3.548592652417539D-07, &
   0.000000000000000D+00,  -7.699902588065971D-06,   0.000000000000000D+00, &
  -3.603131161999475D-06,   0.000000000000000D+00,  -6.809388566855455D-08, &
   0.000000000000000D+00,   6.399748526212970D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.182106216077176D-08,   0.000000000000000D+00,  -5.269395198612485D-08, &
   0.000000000000000D+00,   8.735921714181566D-07,   0.000000000000000D+00, &
   7.459020539686772D-07,   0.000000000000000D+00,  -2.800691740477826D-06, &
   0.000000000000000D+00,  -1.990129553519770D-09,   0.000000000000000D+00, &
   6.057192183901321D-07,   0.000000000000000D+00,  -3.213963956915607D-07 /)
lem(:, 83) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.505846504842085D-02,  -1.467912124140209D-02, &
  -7.688383255758495D-03,   1.331137805963875D-02,  -3.835604683825017D-04, &
  -5.105782786572621D-03,   1.934250535554467D-03,   6.000824643267142D-04, &
  -7.051536455417284D-04,   7.048679262709647D-05,   1.825404195842973D-04, &
  -5.987501398315879D-05,  -5.823645040694043D-05,   3.350789347970495D-05, &
   2.199980901422409D-05,  -1.860726409285975D-05,  -9.463057067834019D-06, &
   1.075716408266183D-05,   4.504942051339840D-06,  -6.508580097329326D-06, &
  -2.325325913753121D-06,   0.000000000000000D+00,  -1.505846504842083D-02, &
   1.467912124140213D-02,   7.688383255758478D-03,  -1.331137805963877D-02, &
   3.835604683825031D-04,   5.105782786572610D-03,  -1.934250535554466D-03, &
  -6.000824643267149D-04,   7.051536455417243D-04,  -7.048679262708960D-05, &
  -1.825404195842980D-04,   5.987501398315103D-05,   5.823645040693959D-05, &
  -3.350789347969776D-05,  -2.199980901422122D-05,   1.860726409284712D-05, &
   9.463057067840016D-06,  -1.075716408265138D-05,  -4.504942051351060D-06, &
   6.508580097324824D-06,   2.325325913763351D-06,   0.000000000000000D+00, &
  -7.529232524210436D-03,   0.000000000000000D+00,   9.826029297007882D-03, &
   0.000000000000000D+00,  -4.677012035415097D-03,   0.000000000000000D+00, &
   3.615421561784013D-04,   0.000000000000000D+00,   3.254836288262922D-04, &
   0.000000000000000D+00,  -9.864770942892905D-05,   0.000000000000000D+00, &
   3.442673407628750D-05,   0.000000000000000D+00,  -1.392097730155896D-05, &
   0.000000000000000D+00,   6.331428788651490D-06,   0.000000000000000D+00, &
  -3.158990742692055D-06,   0.000000000000000D+00,   1.697080363463892D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.952380952380964D-04, &
   0.000000000000000D+00,  -4.894179894179889D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.145077145077145D-04,   0.000000000000000D+00, &
  -6.660006660006667D-04,   0.000000000000000D+00,   7.932543646829360D-05, &
   0.000000000000000D+00,   5.530580530580537D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.734376734376737D-04,   0.000000000000000D+00, &
  -2.352409495266635D-04,   0.000000000000000D+00,   4.230144855144850D-04, &
   0.000000000000000D+00,   9.573759573759576D-05,   0.000000000000000D+00, &
  -1.038247466818895D-04,   0.000000000000000D+00,  -3.650863025863031D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.398927869516111D-05, &
   0.000000000000000D+00,   1.975615420993575D-04,   0.000000000000000D+00, &
   1.382502301619947D-04,   0.000000000000000D+00,  -1.254719545160719D-04, &
   0.000000000000000D+00,  -8.384861326037715D-06,   0.000000000000000D+00, &
  -9.598244892362541D-05,   0.000000000000000D+00,   1.058042680941840D-04, &
   0.000000000000000D+00,   1.348800981153924D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.357527577341718D-06,   0.000000000000000D+00, &
   6.788876757917070D-06,   0.000000000000000D+00,  -1.428943863154390D-04, &
   0.000000000000000D+00,  -3.634163609008806D-05,   0.000000000000000D+00, &
  -1.449152640982167D-05,   0.000000000000000D+00,  -7.560379077405990D-07, &
   0.000000000000000D+00,   9.271773977656256D-06,   0.000000000000000D+00, &
   4.351835814684108D-05,   0.000000000000000D+00,  -7.725440194480440D-05, &
   0.000000000000000D+00,  -2.572023980813648D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.500225641309620D-08,   0.000000000000000D+00, &
  -1.819645782493979D-06,   0.000000000000000D+00,   1.385353552536213D-05, &
   0.000000000000000D+00,   6.053726545212605D-05,   0.000000000000000D+00, &
   3.299997404341492D-06,   0.000000000000000D+00,   2.635168650793645D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.787435916583486D-07,   0.000000000000000D+00,  -9.086415820161890D-06, &
   0.000000000000000D+00,   1.026691961289485D-05,   0.000000000000000D+00, &
   3.353718448411753D-05,   0.000000000000000D+00,   8.046424062049028D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.239095214559140D-07,   0.000000000000000D+00, &
  -8.653206846100751D-07,   0.000000000000000D+00,  -2.280512800939500D-05, &
   0.000000000000000D+00,  -5.492046836303644D-06,   0.000000000000000D+00, &
  -3.953292780229418D-06,   0.000000000000000D+00,  -1.235675674152239D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.802068106049834D-08,   0.000000000000000D+00,   8.979584376678043D-07, &
   0.000000000000000D+00,   2.292093989327709D-06,   0.000000000000000D+00, &
  -2.422156607428225D-05,   0.000000000000000D+00,  -9.846272701037374D-06, &
   0.000000000000000D+00,  -6.794014362373691D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.120609546881054D-09,   0.000000000000000D+00,   2.573032915031177D-07, &
   0.000000000000000D+00,   3.798994767664921D-06,   0.000000000000000D+00, &
   8.087552183378067D-06,   0.000000000000000D+00,  -6.849411414450066D-06, &
   0.000000000000000D+00,   1.995752728174508D-06,   0.000000000000000D+00, &
   5.266814348845657D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.073828043002058D-07,   0.000000000000000D+00,  -4.356136153821630D-07, &
   0.000000000000000D+00,   8.589458316404736D-06,   0.000000000000000D+00, &
   1.054680964861937D-05,   0.000000000000000D+00,   5.186734923063125D-06, &
   0.000000000000000D+00,   3.929447288822205D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.979409013737290D-08, &
   0.000000000000000D+00,  -4.573993544580735D-07,   0.000000000000000D+00, &
  -4.172056337503886D-06,   0.000000000000000D+00,   6.933294466998926D-06, &
   0.000000000000000D+00,   3.793706293527953D-06,   0.000000000000000D+00, &
   6.165167297980949D-07,   0.000000000000000D+00,  -2.445022357742118D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.521371481029844D-09, &
   0.000000000000000D+00,  -7.445049678204855D-08,   0.000000000000000D+00, &
  -1.156257235947032D-06,   0.000000000000000D+00,  -8.299363992635330D-06, &
   0.000000000000000D+00,  -9.791655862415653D-07,   0.000000000000000D+00, &
  -3.040966350118237D-06,   0.000000000000000D+00,  -2.052501060679718D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.756833010527194D-12,   0.000000000000000D+00,   1.476525443645141D-07, &
   0.000000000000000D+00,   5.745661216318299D-07,   0.000000000000000D+00, &
  -1.949887918668203D-06,   0.000000000000000D+00,  -6.795364323794171D-06, &
   0.000000000000000D+00,  -2.802102411477386D-06,   0.000000000000000D+00, &
  -9.319189944016516D-07,   0.000000000000000D+00,   1.238284014162913D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.386518197141286D-08,   0.000000000000000D+00, &
   2.176235530782622D-07,   0.000000000000000D+00,   3.332330578890793D-06, &
   0.000000000000000D+00,  -8.591396041273769D-08,   0.000000000000000D+00, &
  -3.563602196413962D-07,   0.000000000000000D+00,   9.633013412834228D-07, &
   0.000000000000000D+00,   1.102828174971711D-06,   0.000000000000000D+00 /)
lem(:, 84) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.992640465359330D-02,  -8.156668567894623D-03,   2.093883846043676D-02, &
   2.515815522890449D-03,  -1.129537525411192D-02,   2.634514640046224D-03, &
   2.010465797466419D-03,  -1.965132261394707D-03,   5.621791664207313D-04, &
   2.899580587888532D-04,  -3.072348051514232D-04,   2.309860005324640D-05, &
   1.037772727116589D-04,  -3.812097075637250D-05,  -3.562969598243655D-05, &
   2.550199667905142D-05,   1.253948558941879D-05,  -1.543139268511098D-05, &
  -4.290877697498020D-06,   9.298605661624330D-06,   1.234715930234094D-06, &
  -5.715257846559668D-06,  -1.992640465359333D-02,  -8.156668567894645D-03, &
   2.093883846043681D-02,   2.515815522890462D-03,  -1.129537525411189D-02, &
   2.634514640046221D-03,   2.010465797466414D-03,  -1.965132261394704D-03, &
   5.621791664207284D-04,   2.899580587888509D-04,  -3.072348051514209D-04, &
   2.309860005324596D-05,   1.037772727116531D-04,  -3.812097075637463D-05, &
  -3.562969598242379D-05,   2.550199667905886D-05,   1.253948558940484D-05, &
  -1.543139268511229D-05,  -4.290877697489836D-06,   9.298605661617140D-06, &
   1.234715930227571D-06,  -5.715257846548995D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.822235740311983D-02,   0.000000000000000D+00, &
  -2.122896021831506D-02,   0.000000000000000D+00,   1.209770571412664D-02, &
   0.000000000000000D+00,  -3.414481962623362D-03,   0.000000000000000D+00, &
   5.272571380693293D-04,   0.000000000000000D+00,  -1.270518589398611D-04, &
   0.000000000000000D+00,   3.988745982250904D-05,   0.000000000000000D+00, &
  -1.496874621954752D-05,   0.000000000000000D+00,   6.392555276997201D-06, &
   0.000000000000000D+00,  -3.011813535223896D-06,   0.000000000000000D+00, &
   1.091269841269841D-03,  -1.893939393939399D-04,   0.000000000000000D+00, &
  -3.427128427128425D-04,   0.000000000000000D+00,  -1.215277777777777D-03, &
   1.540126540126542D-04,   0.000000000000000D+00,   4.472957597957595D-04, &
   0.000000000000000D+00,   7.492507492507493D-05,   0.000000000000000D+00, &
   1.650086025086026D-04,   0.000000000000000D+00,   1.466993654493653D-03, &
  -5.827505827505832D-05,   0.000000000000000D+00,  -2.117673992673995D-04, &
   0.000000000000000D+00,  -6.000365375365380D-04,   0.000000000000000D+00, &
  -1.387501387501460D-06,   0.000000000000000D+00,   5.723443223443230D-05, &
   0.000000000000000D+00,  -1.546052327302328D-04,   0.000000000000000D+00, &
  -1.167896773365523D-03,   8.569861511038334D-06,   0.000000000000000D+00, &
   8.141368435486024D-06,   0.000000000000000D+00,   2.791096770876178D-04, &
   0.000000000000000D+00,   4.224795068085509D-04,   0.000000000000000D+00, &
  -1.999634352575931D-06,   0.000000000000000D+00,  -8.304603892839108D-06, &
   0.000000000000000D+00,  -1.025705299602360D-04,   0.000000000000000D+00, &
   1.910154541128803D-04,   0.000000000000000D+00,   6.098640878627081D-04, &
   1.353136028062101D-07,   0.000000000000000D+00,   3.305303227903682D-06, &
   0.000000000000000D+00,  -7.683745348218991D-06,   0.000000000000000D+00, &
  -2.350902357553866D-04,   0.000000000000000D+00,  -1.393293913822891D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.031497742023849D-06,   0.000000000000000D+00,   1.830644487378229D-05, &
   0.000000000000000D+00,   4.744725169696161D-05,   0.000000000000000D+00, &
  -1.801404958453140D-04,   0.000000000000000D+00,  -2.156898458167991D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.571118252218851D-08, &
   0.000000000000000D+00,  -5.264432992761080D-06,   0.000000000000000D+00, &
   3.520774702808368D-05,   0.000000000000000D+00,   1.121915217672766D-04, &
   0.000000000000000D+00,  -7.921115370957040D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.953273077124022D-08, &
   0.000000000000000D+00,  -5.434917393123006D-07,   0.000000000000000D+00, &
  -1.884043915535782D-05,   0.000000000000000D+00,   3.070898923380540D-05, &
   0.000000000000000D+00,   1.024226181027955D-04,   0.000000000000000D+00, &
   5.987258184523810D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.852095318036311D-09,   0.000000000000000D+00,   5.952732387986838D-07, &
   0.000000000000000D+00,  -1.078751570384922D-06,   0.000000000000000D+00, &
  -4.289905576118177D-05,   0.000000000000000D+00,  -5.785369273098551D-06, &
   0.000000000000000D+00,   2.608746505230873D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.842367804502671D-09,   0.000000000000000D+00, &
   2.487299405941783D-06,   0.000000000000000D+00,   2.506423089373921D-06, &
   0.000000000000000D+00,  -5.277043537501004D-05,   0.000000000000000D+00, &
  -3.047530269209955D-05,   0.000000000000000D+00,  -1.782834271847963D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.333583993769739D-08,   0.000000000000000D+00, &
   2.911877643713502D-07,   0.000000000000000D+00,   8.369628064326481D-06, &
   0.000000000000000D+00,   1.353712780233445D-05,   0.000000000000000D+00, &
  -2.811573121591960D-05,   0.000000000000000D+00,  -1.297107162585666D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.368661139503161D-08, &
   0.000000000000000D+00,  -6.191644981563931D-07,   0.000000000000000D+00, &
  -9.977798701939426D-07,   0.000000000000000D+00,   1.689569715498388D-05, &
   0.000000000000000D+00,   2.383137572122601D-05,   0.000000000000000D+00, &
   2.543131510416589D-06,   0.000000000000000D+00,   6.480193514800479D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.852037221665739D-10,   0.000000000000000D+00, &
   9.405970472822168D-08,   0.000000000000000D+00,  -1.700646459966986D-06, &
   0.000000000000000D+00,  -7.165085488414927D-06,   0.000000000000000D+00, &
   1.598899915446957D-05,   0.000000000000000D+00,   1.750585704443136D-05, &
   0.000000000000000D+00,   5.418776971816487D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.393565733521165D-08,   0.000000000000000D+00,   1.356017803911420D-07, &
   0.000000000000000D+00,  -3.293939115057447D-06,   0.000000000000000D+00, &
  -1.565918857161386D-05,   0.000000000000000D+00,   2.640351456269439D-06, &
   0.000000000000000D+00,   1.643098948396848D-06,   0.000000000000000D+00, &
  -2.727975378503061D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.229661801830731D-08,   0.000000000000000D+00, &
   5.219110411247925D-07,   0.000000000000000D+00,   1.468884742573553D-06, &
   0.000000000000000D+00,  -4.184713718941853D-06,   0.000000000000000D+00, &
  -1.676539968300051D-05,   0.000000000000000D+00,  -5.396624430437912D-06, &
   0.000000000000000D+00,  -2.430757998308745D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.100678448611911D-09,   0.000000000000000D+00,  -1.905163693947422D-07, &
   0.000000000000000D+00,   1.117143760986505D-06,   0.000000000000000D+00, &
   6.059751250466208D-06,   0.000000000000000D+00,  -8.369671088131758D-07, &
   0.000000000000000D+00,  -7.202688171926449D-06,   0.000000000000000D+00, &
  -1.309817527239991D-06,   0.000000000000000D+00,   1.281172364622298D-06 /)
lem(:, 85) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.605285377990327D-03,  -2.857085666908914D-02, &
   7.964664293584755D-03,   4.818275448691968D-02,  -1.792542981223349D-02, &
  -4.020029135535231D-02,   7.314528755522657D-03,   1.857580112501811D-02, &
   3.287713422487756D-03,  -4.104322483230904D-03,  -3.233312174145538D-03, &
  -8.407726464638566D-05,   1.057996678118228D-03,   3.343103123460456D-04, &
  -3.392741039494141D-04,  -2.191209718166122D-04,   1.102500632508157D-04, &
   1.250908906079554D-04,  -3.437062427241056D-05,  -7.051923824666478D-05, &
   8.577697377928389D-06,   0.000000000000000D+00,   7.605285377990358D-03, &
  -2.857085666908919D-02,   7.964664293584738D-03,   4.818275448691975D-02, &
  -1.792542981223350D-02,  -4.020029135535232D-02,   7.314528755522660D-03, &
   1.857580112501810D-02,   3.287713422487751D-03,  -4.104322483230867D-03, &
  -3.233312174145557D-03,  -8.407726464642167D-05,   1.057996678118261D-03, &
   3.343103123460477D-04,  -3.392741039494496D-04,  -2.191209718165745D-04, &
   1.102500632508360D-04,   1.250908906078992D-04,  -3.437062427240913D-05, &
  -7.051923824663529D-05,   8.577697377947374D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.202501202501212D-04,  -7.936507936507943D-04,   0.000000000000000D+00, &
   1.957301957301961D-03,   0.000000000000000D+00,   3.561253561253547D-05, &
  -1.246901246901248D-03,   0.000000000000000D+00,   1.956376956376959D-04, &
   0.000000000000000D+00,  -2.886002886002885D-03,   0.000000000000000D+00, &
  -7.214082214082212D-04,   0.000000000000000D+00,  -5.827505827505832D-05, &
   5.979205979205984D-03,   0.000000000000000D+00,   6.026381026381033D-04, &
   0.000000000000000D+00,   3.052503052503058D-05,   0.000000000000000D+00, &
  -3.239135592076768D-03,   0.000000000000000D+00,   9.538663950428670D-04, &
   0.000000000000000D+00,   2.328553799142033D-04,   0.000000000000000D+00, &
  -6.726661138425948D-06,  -1.743681155445858D-03,   0.000000000000000D+00, &
  -2.184770567123509D-03,   0.000000000000000D+00,  -3.477731418907889D-04, &
   0.000000000000000D+00,  -6.262120232708472D-05,   0.000000000000000D+00, &
   3.101187311713625D-03,   0.000000000000000D+00,   1.080439741430454D-03, &
   0.000000000000000D+00,  -2.454481363150097D-04,   0.000000000000000D+00, &
  -1.391224301441015D-05,   0.000000000000000D+00,   3.586725541543657D-05, &
  -1.500357227911406D-03,   0.000000000000000D+00,   8.246990367733398D-04, &
   0.000000000000000D+00,   7.929656342969038D-04,   0.000000000000000D+00, &
   1.490855206025485D-04,   0.000000000000000D+00,   2.560816645085227D-05, &
   0.000000000000000D+00,   2.598021173872556D-04,   0.000000000000000D+00, &
  -1.156895506740708D-03,   0.000000000000000D+00,  -2.853055097637141D-04, &
   0.000000000000000D+00,   1.301641684768624D-05,   0.000000000000000D+00, &
  -3.008007071474558D-05,   0.000000000000000D+00,  -1.111214301157224D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.058634863588417D-04, &
   0.000000000000000D+00,  -4.711963303294564D-04,   0.000000000000000D+00, &
  -2.237082877562755D-04,   0.000000000000000D+00,  -4.422768064408939D-05, &
   0.000000000000000D+00,  -2.167048020134312D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.439454922965556D-05, &
   0.000000000000000D+00,   3.952042483473364D-04,   0.000000000000000D+00, &
   1.787639150507634D-05,   0.000000000000000D+00,   1.671766761491148D-05, &
   0.000000000000000D+00,   1.410237723508562D-05,   0.000000000000000D+00, &
  -7.351228054353059D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.716501562779691D-05,   0.000000000000000D+00,   2.289069121751883D-05, &
   0.000000000000000D+00,   2.070901150942204D-04,   0.000000000000000D+00, &
   5.804823406241823D-05,   0.000000000000000D+00,   1.133008114069450D-05, &
   0.000000000000000D+00,  -5.707183441558479D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.834098328646732D-06, &
   0.000000000000000D+00,  -1.050655869608626D-04,   0.000000000000000D+00, &
  -4.507004439700642D-05,   0.000000000000000D+00,   1.846376892122433D-05, &
   0.000000000000000D+00,  -8.529282459204393D-06,   0.000000000000000D+00, &
  -1.823163382671476D-06,   0.000000000000000D+00,   5.569201125841703D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.505583750838868D-05,   0.000000000000000D+00, &
  -1.144288026775572D-04,   0.000000000000000D+00,  -6.275532820050083D-05, &
   0.000000000000000D+00,  -1.932128906417812D-05,   0.000000000000000D+00, &
  -1.094766583712924D-06,   0.000000000000000D+00,   6.908374828296848D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.170317918870853D-05, &
   0.000000000000000D+00,   3.878572520045139D-05,   0.000000000000000D+00, &
  -4.020263889210732D-05,   0.000000000000000D+00,  -1.057120114535514D-05, &
   0.000000000000000D+00,   2.327514856933918D-06,   0.000000000000000D+00, &
   1.493415800056319D-06,   0.000000000000000D+00,  -1.573303270080554D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.446204314423369D-06,   0.000000000000000D+00, &
   3.327070134513918D-05,   0.000000000000000D+00,   5.424644068913959D-05, &
   0.000000000000000D+00,   1.416440472446691D-05,   0.000000000000000D+00, &
   7.727166852432858D-06,   0.000000000000000D+00,  -3.253988849692021D-06, &
   0.000000000000000D+00,  -3.934020352477601D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.527116214857326D-07,   0.000000000000000D+00, &
  -1.582276853164971D-05,   0.000000000000000D+00,   2.832326921745558D-05, &
   0.000000000000000D+00,   2.854539917865655D-05,   0.000000000000000D+00, &
   5.583043543280505D-06,   0.000000000000000D+00,  -1.659719774266918D-06, &
   0.000000000000000D+00,  -3.249585158764701D-06,   0.000000000000000D+00, &
   3.666496128653669D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.907956496468300D-06,   0.000000000000000D+00, &
  -3.263502547638058D-05,   0.000000000000000D+00,  -3.109218989151876D-06, &
   0.000000000000000D+00,  -3.193023752324904D-06,   0.000000000000000D+00, &
  -2.028715653976686D-06,   0.000000000000000D+00,   2.049071262938549D-06, &
   0.000000000000000D+00,   1.316745930393565D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.523044264604196D-06,   0.000000000000000D+00,  -7.277720524527080D-06, &
   0.000000000000000D+00,  -3.172112313191211D-05,   0.000000000000000D+00, &
  -1.188465224420958D-05,   0.000000000000000D+00,  -3.169545135413804D-06, &
   0.000000000000000D+00,   3.444363795926181D-06,   0.000000000000000D+00, &
   2.442820512278462D-06,   0.000000000000000D+00,  -3.793949106443673D-08 /)
lem(:, 86) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.258959811674599D-03,   8.323792887844869D-03, &
   4.491529087956509D-03,  -1.848043246472778D-02,   4.360494635287426D-03, &
   1.374126748319716D-02,  -5.963135435098353D-03,  -5.452978656726855D-03, &
   2.294093365913677D-03,   1.874869037787299D-03,  -3.714044102858435D-04, &
  -7.575033246734562D-04,  -5.366001748617091D-07,   3.129506989725202D-04, &
   4.796476412050482D-05,  -1.382049400824633D-04,  -4.061060911651033D-05, &
   6.528706040496491D-05,   2.824641928486772D-05,  -3.273194521773676D-05, &
  -1.880581647798860D-05,   0.000000000000000D+00,   4.258959811674587D-03, &
  -8.323792887844871D-03,  -4.491529087956510D-03,   1.848043246472775D-02, &
  -4.360494635287419D-03,  -1.374126748319715D-02,   5.963135435098357D-03, &
   5.452978656726859D-03,  -2.294093365913677D-03,  -1.874869037787297D-03, &
   3.714044102858303D-04,   7.575033246734654D-04,   5.366001748694054D-07, &
  -3.129506989725300D-04,  -4.796476412051003D-05,   1.382049400824766D-04, &
   4.061060911650704D-05,  -6.528706040497243D-05,  -2.824641928485600D-05, &
   3.273194521773126D-05,   1.880581647797951D-05,   0.000000000000000D+00, &
   2.129479905837284D-03,   0.000000000000000D+00,  -4.120668888033479D-04, &
   0.000000000000000D+00,  -4.433601900670822D-04,   0.000000000000000D+00, &
   2.699197493432303D-04,   0.000000000000000D+00,   2.191808948325089D-05, &
   0.000000000000000D+00,  -5.424218780564872D-05,   0.000000000000000D+00, &
   1.860213939548547D-05,   0.000000000000000D+00,  -6.142907444178037D-06, &
   0.000000000000000D+00,   2.066280617531475D-06,   0.000000000000000D+00, &
  -6.832790688585429D-07,   0.000000000000000D+00,   1.988952302362055D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.683501683501688D-04, &
   0.000000000000000D+00,   1.813001813001813D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.405002405002410D-05,   0.000000000000000D+00, &
   1.068376068376068D-04,   0.000000000000000D+00,  -3.145003145003144D-04, &
   0.000000000000000D+00,  -1.734376734376737D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.180005180005184D-04,   0.000000000000000D+00, &
   7.115899973042827D-05,   0.000000000000000D+00,   1.942501942501943D-05, &
   0.000000000000000D+00,  -3.370268076150426D-04,   0.000000000000000D+00, &
   2.528867024665342D-04,   0.000000000000000D+00,   8.292361233537711D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.743533155297875D-05, &
   0.000000000000000D+00,  -4.622219258073598D-04,   0.000000000000000D+00, &
  -8.684126331185156D-05,   0.000000000000000D+00,  -5.820704350116126D-05, &
   0.000000000000000D+00,   2.438966835251665D-04,   0.000000000000000D+00, &
   2.771609195757808D-04,   0.000000000000000D+00,  -9.145910848697214D-05, &
   0.000000000000000D+00,  -1.490991235573272D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.037980366153741D-04,   0.000000000000000D+00, &
   8.793665914408964D-05,   0.000000000000000D+00,   2.249573134506053D-04, &
   0.000000000000000D+00,   5.368012333956610D-05,   0.000000000000000D+00, &
   1.947972209678870D-05,   0.000000000000000D+00,   8.742547442237832D-05, &
   0.000000000000000D+00,  -2.486634452578732D-04,   0.000000000000000D+00, &
  -1.070223206446116D-04,   0.000000000000000D+00,   1.023228577408135D-05, &
   0.000000000000000D+00,  -1.486749261517078D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.010373527401386D-05,   0.000000000000000D+00, &
   1.622818186285679D-04,   0.000000000000000D+00,  -1.001556922609555D-04, &
   0.000000000000000D+00,  -6.997270445412862D-05,   0.000000000000000D+00, &
  -1.609392875356496D-05,   0.000000000000000D+00,   8.525545634920637D-06, &
   0.000000000000000D+00,   1.972272802801814D-06,   0.000000000000000D+00, &
  -3.867538384777764D-05,   0.000000000000000D+00,   1.393722638170072D-04, &
   0.000000000000000D+00,   1.440830728924680D-05,   0.000000000000000D+00, &
   6.908324397219673D-07,   0.000000000000000D+00,   2.516650961775237D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.874416562377052D-06,   0.000000000000000D+00, &
  -3.975063538934834D-05,   0.000000000000000D+00,   6.041935242534246D-05, &
   0.000000000000000D+00,   1.788508595657937D-05,   0.000000000000000D+00, &
  -1.056923759813241D-06,   0.000000000000000D+00,  -7.914900493025543D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.092551272694835D-06,   0.000000000000000D+00,  -2.691834869786152D-05, &
   0.000000000000000D+00,  -3.503967845568326D-05,   0.000000000000000D+00, &
   4.470814782389362D-06,   0.000000000000000D+00,  -2.177083166175257D-07, &
   0.000000000000000D+00,  -4.004422949735456D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.569600438896397D-06,   0.000000000000000D+00,   2.122845372273308D-05, &
   0.000000000000000D+00,  -2.602178324011682D-05,   0.000000000000000D+00, &
  -2.100464114825061D-05,   0.000000000000000D+00,  -3.388703572053885D-06, &
   0.000000000000000D+00,   2.431740712990794D-06,   0.000000000000000D+00, &
   2.897628630050546D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.107881636139945D-07,   0.000000000000000D+00, &
  -1.915121715902582D-06,   0.000000000000000D+00,   2.879424907961721D-05, &
   0.000000000000000D+00,  -5.951346217322461D-06,   0.000000000000000D+00, &
  -3.782024974574861D-06,   0.000000000000000D+00,   2.829978181929694D-06, &
   0.000000000000000D+00,   2.813168096371214D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.891016991165004D-06, &
   0.000000000000000D+00,   6.045221585671061D-07,   0.000000000000000D+00, &
   2.211429848792713D-05,   0.000000000000000D+00,   4.496458501928970D-06, &
   0.000000000000000D+00,  -3.088968241631945D-07,   0.000000000000000D+00, &
   2.770937829531303D-07,   0.000000000000000D+00,  -1.027923086028591D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.232322964903595D-06, &
   0.000000000000000D+00,  -1.033710516527148D-05,   0.000000000000000D+00, &
  -6.350605800541119D-07,   0.000000000000000D+00,   7.937917155005299D-06, &
   0.000000000000000D+00,   6.956684512202077D-07,   0.000000000000000D+00, &
  -2.846139320748673D-06,   0.000000000000000D+00,  -1.095615719639163D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.921319416598022D-07,   0.000000000000000D+00,   3.619707607206806D-06, &
   0.000000000000000D+00,  -1.449342693393904D-05,   0.000000000000000D+00, &
  -4.598103956431375D-06,   0.000000000000000D+00,   7.002895869787150D-08, &
   0.000000000000000D+00,  -1.113117568195664D-06,   0.000000000000000D+00, &
  -9.288642362721600D-07,   0.000000000000000D+00,   3.648711992217899D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.058699712978415D-08, &
   0.000000000000000D+00,   8.671781816125761D-07,   0.000000000000000D+00, &
   5.280208048269495D-06,   0.000000000000000D+00,  -9.475991755032815D-06, &
   0.000000000000000D+00,  -3.382592243942504D-06,   0.000000000000000D+00, &
   1.210500604308408D-06,   0.000000000000000D+00,   7.773240369501953D-07, &
   0.000000000000000D+00,   4.047584766579141D-07,   0.000000000000000D+00 /)
lem(:, 87) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.776157913597375D-03,  -1.133187521320558D-02,  -1.544401031758242D-02, &
   1.328150449311595D-02,   2.086169443403538D-02,  -1.273498241993219D-02, &
  -1.223792619120471D-02,   6.594991452801749D-03,   2.963464569052966D-03, &
  -1.873631349326635D-03,  -2.127347153751063D-04,   5.397519258069957D-04, &
   1.514323880736854D-05,  -2.166510429842152D-04,  -8.032971480389637D-08, &
   1.005440744206153D-04,   3.788866636674523D-07,  -5.119734494067993D-05, &
  -1.062841585507057D-06,   2.794587290078571D-05,   1.226350560629491D-06, &
  -1.612583948018013D-05,   7.776157913597402D-03,  -1.133187521320559D-02, &
  -1.544401031758243D-02,   1.328150449311596D-02,   2.086169443403537D-02, &
  -1.273498241993218D-02,  -1.223792619120470D-02,   6.594991452801752D-03, &
   2.963464569052959D-03,  -1.873631349326633D-03,  -2.127347153750994D-04, &
   5.397519258069874D-04,   1.514323880736196D-05,  -2.166510429842093D-04, &
  -8.032971479631898D-08,   1.005440744206090D-04,   3.788866636656613D-07, &
  -5.119734494067043D-05,  -1.062841585513603D-06,   2.794587290076841D-05, &
   1.226350560635762D-06,  -1.612583948016284D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.889725952916130D-03,   0.000000000000000D+00, &
   2.191880380603467D-03,   0.000000000000000D+00,  -7.987783480099794D-04, &
   0.000000000000000D+00,  -1.630797355282806D-04,   0.000000000000000D+00, &
   1.586550179809822D-04,   0.000000000000000D+00,  -2.860800478054713D-05, &
   0.000000000000000D+00,   5.312163585747956D-06,   0.000000000000000D+00, &
  -7.811133682126108D-07,   0.000000000000000D+00,  -9.593970011782093D-08, &
   0.000000000000000D+00,   2.106979258392655D-07,   0.000000000000000D+00, &
  -6.553631553631544D-04,   4.004329004329001D-04,   0.000000000000000D+00, &
   7.654382654382653D-04,   0.000000000000000D+00,   3.206053206053204D-04, &
  -1.680554894840610D-03,   0.000000000000000D+00,  -1.889380460809033D-04, &
   0.000000000000000D+00,   8.422133422133437D-04,   0.000000000000000D+00, &
  -3.620850049421480D-04,   0.000000000000000D+00,  -2.117673992673995D-04, &
   6.026381026381033D-04,   0.000000000000000D+00,   7.851539994397135D-04, &
   0.000000000000000D+00,   1.375757179328605D-04,   0.000000000000000D+00, &
  -8.597665950607130D-04,   0.000000000000000D+00,  -3.956407737920341D-04, &
   0.000000000000000D+00,   1.896256754450035D-04,   0.000000000000000D+00, &
   9.963654265124855D-05,   2.984760337701513D-04,   0.000000000000000D+00, &
  -2.636781208209780D-04,   0.000000000000000D+00,  -4.387358090089181D-04, &
   0.000000000000000D+00,  -9.020289351171695D-05,   0.000000000000000D+00, &
   2.137410806141460D-05,   0.000000000000000D+00,   3.774988914307806D-04, &
   0.000000000000000D+00,   2.057793632318176D-04,   0.000000000000000D+00, &
  -6.250002863779944D-05,   0.000000000000000D+00,  -4.401310969373971D-06, &
  -2.860916173609655D-05,   0.000000000000000D+00,  -1.328942815010928D-04, &
   0.000000000000000D+00,   1.637828484356570D-04,   0.000000000000000D+00, &
   1.879902525358076D-04,   0.000000000000000D+00,   3.554252402331928D-05, &
   0.000000000000000D+00,   2.945602237950619D-06,   0.000000000000000D+00, &
   2.739737370431762D-06,   0.000000000000000D+00,  -2.005985500567543D-04, &
   0.000000000000000D+00,  -6.319574818026830D-05,   0.000000000000000D+00, &
   5.920512674817986D-06,   0.000000000000000D+00,  -2.433335927917974D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.036979826896768D-07, &
   0.000000000000000D+00,   5.576316528328909D-05,   0.000000000000000D+00, &
  -1.047048067171906D-04,   0.000000000000000D+00,  -5.505986679191980D-05, &
   0.000000000000000D+00,  -8.371576071846946D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.505890945134957D-06, &
   0.000000000000000D+00,   1.807877381161654D-06,   0.000000000000000D+00, &
   8.001775115871900D-05,   0.000000000000000D+00,   6.195621102678440D-07, &
   0.000000000000000D+00,   6.308437155613387D-06,   0.000000000000000D+00, &
   1.286523625390813D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.122610635523363D-06,   0.000000000000000D+00,   3.074002080069076D-06, &
   0.000000000000000D+00,   4.565094815296771D-06,   0.000000000000000D+00, &
   5.075309874150349D-05,   0.000000000000000D+00,   1.209557740156786D-05, &
   0.000000000000000D+00,   3.817119295634963D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.607337953653208D-08, &
   0.000000000000000D+00,  -1.456755724538464D-06,   0.000000000000000D+00, &
  -1.409430704675697D-05,   0.000000000000000D+00,  -9.066167757771200D-06, &
   0.000000000000000D+00,   6.377105132037294D-06,   0.000000000000000D+00, &
  -4.585009132781490D-06,   0.000000000000000D+00,  -3.035610945767237D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.513913481732565D-06,   0.000000000000000D+00, &
   9.820391727445550D-07,   0.000000000000000D+00,  -2.482180131768104D-05, &
   0.000000000000000D+00,  -1.432694855561531D-05,   0.000000000000000D+00, &
  -4.499366998045788D-06,   0.000000000000000D+00,  -1.881527760434013D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.633027482843837D-07, &
   0.000000000000000D+00,  -1.593878088984075D-06,   0.000000000000000D+00, &
   7.247575562256651D-06,   0.000000000000000D+00,  -1.172663827410010D-05, &
   0.000000000000000D+00,  -1.654701207111207D-06,   0.000000000000000D+00, &
   1.071338961964004D-06,   0.000000000000000D+00,   4.844483451661788D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.193086007103854D-07,   0.000000000000000D+00, &
   1.462068614444328D-06,   0.000000000000000D+00,   4.181829570246137D-06, &
   0.000000000000000D+00,   1.267338046398573D-05,   0.000000000000000D+00, &
   2.017819551046823D-06,   0.000000000000000D+00,   2.819678461419452D-06, &
   0.000000000000000D+00,   2.159562771832318D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.251369733854208D-09,   0.000000000000000D+00, &
   4.134373893187016D-07,   0.000000000000000D+00,  -8.636552923819757D-07, &
   0.000000000000000D+00,   6.605595525840017D-06,   0.000000000000000D+00, &
   7.288153159952722D-06,   0.000000000000000D+00,   7.205169113168067D-07, &
   0.000000000000000D+00,  -6.809388566856068D-08,   0.000000000000000D+00, &
   4.005665016396929D-08,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.119497738706249D-07,   0.000000000000000D+00, &
   6.987764554929510D-07,   0.000000000000000D+00,  -6.246714947469227D-06, &
   0.000000000000000D+00,   5.362920449022579D-07,   0.000000000000000D+00, &
  -9.888181133304323D-07,   0.000000000000000D+00,  -9.379559223482426D-07, &
   0.000000000000000D+00,   1.844106065995726D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.417220222879505D-07,   0.000000000000000D+00,  -1.114906195645419D-06, &
   0.000000000000000D+00,  -7.093031526431901D-07,   0.000000000000000D+00, &
  -8.022978270748111D-06,   0.000000000000000D+00,  -2.018805218740980D-06, &
   0.000000000000000D+00,  -9.910905667219099D-07,   0.000000000000000D+00, &
   2.515971383682502D-07,   0.000000000000000D+00,  -1.157927240537006D-07 /)
lem(:, 88) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.194219858755948D-03,   1.440916039098550D-02, &
  -1.018629349121880D-02,  -1.249267689609981D-02,   1.177515477441461D-02, &
   3.591754475398078D-03,  -6.209705896691675D-03,   6.765260903555955D-04, &
   1.528056138474032D-03,  -6.203634212117731D-04,  -2.270864650101815D-04, &
   2.199140797082171D-04,   3.717617116276968D-05,  -8.900359026085655D-05, &
  -3.012528950454481D-06,   4.081647999495042D-05,  -2.860549725700103D-06, &
  -2.068105770121475D-05,   3.119705208927043D-06,   1.134313263796687D-05, &
  -2.418416147098828D-06,   0.000000000000000D+00,  -3.194219858755933D-03, &
  -1.440916039098552D-02,   1.018629349121881D-02,   1.249267689609981D-02, &
  -1.177515477441465D-02,  -3.591754475398072D-03,   6.209705896691677D-03, &
  -6.765260903555943D-04,  -1.528056138474031D-03,   6.203634212117758D-04, &
   2.270864650101764D-04,  -2.199140797082220D-04,  -3.717617116275953D-05, &
   8.900359026085979D-05,   3.012528950440584D-06,  -4.081647999494714D-05, &
   2.860549725708180D-06,   2.068105770121383D-05,  -3.119705208927157D-06, &
  -1.134313263797034D-05,   2.418416147094651D-06,   0.000000000000000D+00, &
  -1.597109929377965D-03,   0.000000000000000D+00,  -5.933763198768250D-04, &
   0.000000000000000D+00,  -9.121563180901421D-04,   0.000000000000000D+00, &
   1.363479427331803D-03,   0.000000000000000D+00,  -4.306904583458973D-04, &
   0.000000000000000D+00,  -1.686285631263337D-06,   0.000000000000000D+00, &
   1.472584194178494D-05,   0.000000000000000D+00,  -8.617747853347513D-06, &
   0.000000000000000D+00,   4.597599624823781D-06,   0.000000000000000D+00, &
  -2.504633539179840D-06,   0.000000000000000D+00,   1.421215403719570D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.262626262626263D-04, &
   0.000000000000000D+00,   2.145077145077145D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.288965003250718D-04,   0.000000000000000D+00, &
  -1.693412407698123D-04,   0.000000000000000D+00,   2.493934636791780D-04, &
   0.000000000000000D+00,  -2.352409495266635D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.115899973042827D-05,   0.000000000000000D+00, &
   4.834186977044116D-04,   0.000000000000000D+00,   1.946548821548823D-04, &
   0.000000000000000D+00,  -1.946582828935771D-04,   0.000000000000000D+00, &
  -2.661157703174508D-04,   0.000000000000000D+00,   1.382502301619947D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.177417353887943D-04, &
   0.000000000000000D+00,  -9.896732585808214D-05,   0.000000000000000D+00, &
  -3.368106177980126D-04,   0.000000000000000D+00,  -1.721521942110177D-04, &
   0.000000000000000D+00,  -2.328510842442739D-05,   0.000000000000000D+00, &
   2.178628781901669D-04,   0.000000000000000D+00,   1.614506832054379D-04, &
   0.000000000000000D+00,  -3.634163609008808D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.687530907345019D-06,   0.000000000000000D+00, &
  -1.147487171370409D-04,   0.000000000000000D+00,   1.038404565604920D-04, &
   0.000000000000000D+00,   1.559655729547372D-04,   0.000000000000000D+00, &
   9.309324732965399D-05,   0.000000000000000D+00,   5.243171983552293D-06, &
   0.000000000000000D+00,   1.342384579890121D-05,   0.000000000000000D+00, &
  -1.534703256839480D-04,   0.000000000000000D+00,  -5.120465404908122D-05, &
   0.000000000000000D+00,   3.299997404341478D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.055507468599067D-06,   0.000000000000000D+00, &
   9.363619491881158D-06,   0.000000000000000D+00,   5.246695455673783D-05, &
   0.000000000000000D+00,  -7.904220607393977D-05,   0.000000000000000D+00, &
  -5.263554282154334D-05,   0.000000000000000D+00,  -2.496599176286680D-05, &
   0.000000000000000D+00,   7.043831438590440D-08,   0.000000000000000D+00, &
  -2.188319208491046D-06,   0.000000000000000D+00,   1.100760750444415D-05, &
   0.000000000000000D+00,   6.554566640042454D-05,   0.000000000000000D+00, &
   1.483170266138680D-06,   0.000000000000000D+00,  -3.953292780229418D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.073784076687740D-06,   0.000000000000000D+00, &
  -8.807158364635684D-06,   0.000000000000000D+00,   5.729057944194548D-06, &
   0.000000000000000D+00,   3.959205235848074D-05,   0.000000000000000D+00, &
   1.403366334479835D-05,   0.000000000000000D+00,   7.662422213204429D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.287854065808096D-07,   0.000000000000000D+00,  -1.566671067709574D-06, &
   0.000000000000000D+00,  -2.094282435681002D-05,   0.000000000000000D+00, &
  -8.159635489766988D-06,   0.000000000000000D+00,   7.083548820323574D-06, &
   0.000000000000000D+00,   1.995752728174501D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.780231998045811D-08,   0.000000000000000D+00,   1.470930101249553D-06, &
   0.000000000000000D+00,   1.977152123537849D-06,   0.000000000000000D+00, &
  -2.201196139883419D-05,   0.000000000000000D+00,  -1.233199435293841D-05, &
   0.000000000000000D+00,  -1.417446485805805D-06,   0.000000000000000D+00, &
   2.017116274928637D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.699790250342613D-09,   0.000000000000000D+00, &
   2.458495057479905D-07,   0.000000000000000D+00,   3.971358523619988D-06, &
   0.000000000000000D+00,   8.232842713057483D-06,   0.000000000000000D+00, &
  -8.233410410741274D-06,   0.000000000000000D+00,  -4.130853490632537D-06, &
   0.000000000000000D+00,   6.165167297980917D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.259946556488213D-07, &
   0.000000000000000D+00,  -1.623856617800082D-07,   0.000000000000000D+00, &
   7.245102341967292D-06,   0.000000000000000D+00,   1.086222571123381D-05, &
   0.000000000000000D+00,   1.428478203438647D-06,   0.000000000000000D+00, &
  -6.672962867159931D-07,   0.000000000000000D+00,  -1.450179378558906D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.355797179667149D-08, &
   0.000000000000000D+00,  -8.956731831265110D-07,   0.000000000000000D+00, &
  -3.418924351093847D-06,   0.000000000000000D+00,   5.273919698912036D-06, &
   0.000000000000000D+00,   6.009853796535539D-06,   0.000000000000000D+00, &
   4.919874479026264D-07,   0.000000000000000D+00,  -9.319189944016490D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.768447691517327D-09,   0.000000000000000D+00,   5.864075159727163D-08, &
   0.000000000000000D+00,  -1.515813000354034D-06,   0.000000000000000D+00, &
  -7.138865134835070D-06,   0.000000000000000D+00,  -5.919895899732180D-07, &
   0.000000000000000D+00,   8.805446165937527D-07,   0.000000000000000D+00, &
   1.346714821538766D-08,   0.000000000000000D+00,   8.794255467838961D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.272080156959073D-10, &
   0.000000000000000D+00,   1.959274905849259D-07,   0.000000000000000D+00, &
   5.724354303179536D-07,   0.000000000000000D+00,  -1.522744833439025D-06, &
   0.000000000000000D+00,  -6.302111622812775D-06,   0.000000000000000D+00, &
  -1.881038594202418D-06,   0.000000000000000D+00,   5.990491590554325D-07, &
   0.000000000000000D+00,   6.846892528045315D-07,   0.000000000000000D+00 /)
lem(:, 89) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.458029608799521D-03,  -6.331400077176955D-03,  -1.158019564599497D-02, &
   1.113537183618050D-02,   7.404050009924961D-03,  -8.641101895901860D-03, &
   1.183205695160006D-04,   2.784833546915356D-03,  -1.752151611603645D-03, &
   2.054820889054700D-05,   5.808987819499426D-04,  -2.399268274732367D-04, &
  -1.073561181784322D-04,   1.069846598646592D-04,   1.850964223305204D-05, &
  -4.721945782912797D-05,  -4.582914277572696D-07,   2.236348860512716D-05, &
  -2.818887065908783D-06,  -1.139483836052885D-05,   2.855662306847031D-06, &
   6.192742628731983D-06,   1.458029608799513D-03,  -6.331400077176952D-03, &
  -1.158019564599499D-02,   1.113537183618050D-02,   7.404050009924969D-03, &
  -8.641101895901854D-03,   1.183205695160014D-04,   2.784833546915354D-03, &
  -1.752151611603645D-03,   2.054820889055106D-05,   5.808987819499397D-04, &
  -2.399268274732388D-04,  -1.073561181784322D-04,   1.069846598646648D-04, &
   1.850964223305161D-05,  -4.721945782913318D-05,  -4.582914277502375D-07, &
   2.236348860512512D-05,  -2.818887065913755D-06,  -1.139483836052565D-05, &
   2.855662306844438D-06,   6.192742628730854D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.673597139927824D-03,   0.000000000000000D+00, &
   1.132818621091777D-02,   0.000000000000000D+00,  -5.333331477538582D-03, &
   0.000000000000000D+00,   4.586617561732892D-04,   0.000000000000000D+00, &
   3.552215764031996D-04,   0.000000000000000D+00,  -1.052032634034047D-04, &
   0.000000000000000D+00,   3.551844799610629D-05,   0.000000000000000D+00, &
  -1.381730420266073D-05,   0.000000000000000D+00,   6.025518610379915D-06, &
   0.000000000000000D+00,  -2.877182272719257D-06,   0.000000000000000D+00, &
  -1.893939393939399D-04,   4.178691678691672D-04,   0.000000000000000D+00, &
  -1.734376734376734D-04,   0.000000000000000D+00,   4.472957597957596D-04, &
  -2.317127317127318D-04,   0.000000000000000D+00,  -4.503945128945133D-04, &
   0.000000000000000D+00,   1.859251859251859D-04,   0.000000000000000D+00, &
   1.781204906204906D-04,   0.000000000000000D+00,  -6.000365375365380D-04, &
   3.052503052503058D-05,   0.000000000000000D+00,   1.375757179328605D-04, &
   0.000000000000000D+00,   5.071895696895697D-04,   0.000000000000000D+00, &
  -5.737726325961620D-05,   0.000000000000000D+00,  -1.590117398940928D-04, &
   0.000000000000000D+00,  -1.855128463584345D-04,   0.000000000000000D+00, &
   4.224795068085509D-04,   4.706622353680867D-06,   0.000000000000000D+00, &
   3.878882555353144D-05,   0.000000000000000D+00,  -1.286818874791564D-04, &
   0.000000000000000D+00,  -3.461575784645632D-04,   0.000000000000000D+00, &
   7.420769804671122D-06,   0.000000000000000D+00,   6.437705741111252D-06, &
   0.000000000000000D+00,   1.683012603581487D-04,   0.000000000000000D+00, &
   1.044290448436152D-04,   0.000000000000000D+00,  -1.393293913822891D-04, &
  -1.999634352575813D-06,   0.000000000000000D+00,  -1.213884727816445D-06, &
   0.000000000000000D+00,  -6.344914795959669D-05,   0.000000000000000D+00, &
   1.153568935471993D-04,   0.000000000000000D+00,   1.292507988612624D-04, &
   0.000000000000000D+00,   1.431889976783249D-07,   0.000000000000000D+00, &
   6.883811063357794D-08,   0.000000000000000D+00,   4.698299493191188D-06, &
   0.000000000000000D+00,  -1.226097966253733D-04,   0.000000000000000D+00, &
  -2.144213849660830D-05,   0.000000000000000D+00,  -7.921115370957040D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.010019392372223D-06, &
   0.000000000000000D+00,   4.147916783365585D-06,   0.000000000000000D+00, &
   3.211627419493129D-05,   0.000000000000000D+00,  -7.580377434527954D-05, &
   0.000000000000000D+00,  -2.275739865761986D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.953428215008983D-07, &
   0.000000000000000D+00,  -3.132123139021652D-06,   0.000000000000000D+00, &
   9.969629779349288D-06,   0.000000000000000D+00,   4.977517868432065D-05, &
   0.000000000000000D+00,  -7.687089363351003D-07,   0.000000000000000D+00, &
   2.608746505230873D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.890993752207888D-08,   0.000000000000000D+00,   1.004739298082887D-06, &
   0.000000000000000D+00,  -8.824929072543942D-06,   0.000000000000000D+00, &
   9.707245911805324D-06,   0.000000000000000D+00,   2.910055470699220D-05, &
   0.000000000000000D+00,   5.774119543650835D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.792975275168655D-09, &
   0.000000000000000D+00,   2.024754127396994D-07,   0.000000000000000D+00, &
   2.251880935186992D-06,   0.000000000000000D+00,  -1.873769509142181D-05, &
   0.000000000000000D+00,  -2.633719123619130D-06,   0.000000000000000D+00, &
  -3.897114183246141D-06,   0.000000000000000D+00,  -1.297107162585668D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.791905248796745D-07,   0.000000000000000D+00, &
   6.156830211002045D-07,   0.000000000000000D+00,   4.374027847677353D-06, &
   0.000000000000000D+00,  -1.985055850401033D-05,   0.000000000000000D+00, &
  -6.860634089251167D-06,   0.000000000000000D+00,  -5.935667251585299D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.227653805902903D-08, &
   0.000000000000000D+00,  -5.620455120513823D-07,   0.000000000000000D+00, &
   2.329579083921264D-06,   0.000000000000000D+00,   7.112146232439420D-06, &
   0.000000000000000D+00,  -5.880095697544164D-06,   0.000000000000000D+00, &
   2.924897639253172D-06,   0.000000000000000D+00,   5.418776971816492D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.566129662055289D-09,   0.000000000000000D+00, &
  -3.845529246414030D-08,   0.000000000000000D+00,  -1.807326883806144D-06, &
   0.000000000000000D+00,   5.685716904755840D-06,   0.000000000000000D+00, &
   7.781491300240842D-06,   0.000000000000000D+00,   4.066131080291206D-06, &
   0.000000000000000D+00,   3.755325067275796D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.382971578302438D-11,   0.000000000000000D+00, &
   1.093366840653054D-07,   0.000000000000000D+00,   1.013013772579662D-07, &
   0.000000000000000D+00,  -4.455443986007150D-06,   0.000000000000000D+00, &
   5.036816162872035D-06,   0.000000000000000D+00,   2.161936436059707D-06, &
   0.000000000000000D+00,   3.227961907158944D-08,   0.000000000000000D+00, &
  -2.430757998308766D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.781959825528572D-08,   0.000000000000000D+00, &
   3.027589686411978D-07,   0.000000000000000D+00,   2.212144155338389D-07, &
   0.000000000000000D+00,  -6.542123825412111D-06,   0.000000000000000D+00, &
  -4.482330965934612D-07,   0.000000000000000D+00,  -3.011553969369938D-06, &
   0.000000000000000D+00,  -1.963130482188662D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.606829946258047D-09,   0.000000000000000D+00,  -1.540381659026662D-08, &
   0.000000000000000D+00,   9.925132034820833D-07,   0.000000000000000D+00, &
  -2.717976446222407D-07,   0.000000000000000D+00,  -4.718098349448366D-06, &
   0.000000000000000D+00,  -1.720601085835431D-06,   0.000000000000000D+00, &
  -7.204321913031743D-07,   0.000000000000000D+00,   1.182883253424776D-06 /)
lem(:, 90) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.619925785511941D-02,   1.172642443997065D-02, &
  -1.312212290719667D-02,  -4.789043868992216D-03,   6.720267455425058D-03, &
  -1.472315009319360D-03,  -8.861881964865241D-04,   1.304319369117781D-03, &
  -6.945294605005365D-04,  -5.324341341378902D-05,   2.853862776160531D-04, &
  -1.131568064985967D-04,  -6.235741198798692D-05,   6.625323258043529D-05, &
   7.196648087140331D-06,  -3.257204817883572D-05,   4.458973810104146D-06, &
   1.578433331546378D-05,  -5.642851754630012D-06,  -7.764304114143226D-06, &
   4.620823219226911D-06,   0.000000000000000D+00,  -1.619925785511939D-02, &
  -1.172642443997064D-02,   1.312212290719665D-02,   4.789043868992216D-03, &
  -6.720267455425060D-03,   1.472315009319359D-03,   8.861881964865174D-04, &
  -1.304319369117782D-03,   6.945294605005376D-04,   5.324341341379493D-05, &
  -2.853862776160571D-04,   1.131568064985953D-04,   6.235741198799798D-05, &
  -6.625323258043441D-05,  -7.196648087152777D-06,   3.257204817883314D-05, &
  -4.458973810096925D-06,  -1.578433331545730D-05,   5.642851754627806D-06, &
   7.764304114137495D-06,  -4.620823219230381D-06,   0.000000000000000D+00, &
  -8.099628927559725D-03,   0.000000000000000D+00,   2.109929637419148D-02, &
   0.000000000000000D+00,  -2.341790791152196D-02,   0.000000000000000D+00, &
   1.331361143537609D-02,   0.000000000000000D+00,  -3.789089719417131D-03, &
   0.000000000000000D+00,   5.924132208333243D-04,   0.000000000000000D+00, &
  -1.451503473022778D-04,   0.000000000000000D+00,   4.656898359212771D-05, &
   0.000000000000000D+00,  -1.794352147797413D-05,   0.000000000000000D+00, &
   7.898832569171176D-06,   0.000000000000000D+00,  -3.847805327396360D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.403318903318902D-04, &
   0.000000000000000D+00,  -1.623376623376626D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.796814296814298D-04,   0.000000000000000D+00, &
  -1.194927757427758D-03,   0.000000000000000D+00,   1.290376290376290D-04, &
   0.000000000000000D+00,   4.230144855144850D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.942501942501943D-05,   0.000000000000000D+00, &
   1.946548821548823D-04,   0.000000000000000D+00,   1.485724923224923D-03, &
   0.000000000000000D+00,  -4.578754578754577D-05,   0.000000000000000D+00, &
  -2.216635488694311D-04,   0.000000000000000D+00,  -5.749229324596972D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.044706927059866D-05, &
   0.000000000000000D+00,   3.799305269893505D-05,   0.000000000000000D+00, &
  -1.876680645062998D-04,   0.000000000000000D+00,  -1.215523268602312D-03, &
   0.000000000000000D+00,   5.141916906622822D-06,   0.000000000000000D+00, &
   2.556890134289517D-05,   0.000000000000000D+00,   2.755257459688573D-04, &
   0.000000000000000D+00,   4.213388722023595D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.052073485510063D-06,   0.000000000000000D+00, &
  -8.215826714278645D-06,   0.000000000000000D+00,  -8.381644048246214D-05, &
   0.000000000000000D+00,   2.097021284206850D-04,   0.000000000000000D+00, &
   6.576452999301537D-04,   0.000000000000000D+00,   5.412544112235542D-07, &
   0.000000000000000D+00,  -7.055637860592257D-07,   0.000000000000000D+00, &
  -9.685124816703764D-06,   0.000000000000000D+00,  -2.360973018746822D-04, &
   0.000000000000000D+00,  -1.454936693423535D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.221752447774844D-08,   0.000000000000000D+00, &
  -7.601545914238587D-07,   0.000000000000000D+00,   1.310855001837966D-05, &
   0.000000000000000D+00,   4.584872552413197D-05,   0.000000000000000D+00, &
  -1.927149313257334D-04,   0.000000000000000D+00,  -2.426756638182418D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.913304884093740D-07,   0.000000000000000D+00,  -6.899768076406819D-06, &
   0.000000000000000D+00,   3.055745670467118D-05,   0.000000000000000D+00, &
   1.168814701901885D-04,   0.000000000000000D+00,  -7.389386262477370D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.097228656370923D-08,   0.000000000000000D+00, &
   2.798796272850553D-06,   0.000000000000000D+00,  -2.092336631442248D-05, &
   0.000000000000000D+00,   2.848155451261740D-05,   0.000000000000000D+00, &
   1.111548720605540D-04,   0.000000000000000D+00,   6.991299715909078D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.158694921799383D-08,   0.000000000000000D+00,   2.649210462342702D-07, &
   0.000000000000000D+00,   5.483033429362776D-06,   0.000000000000000D+00, &
  -4.332258637007666D-05,   0.000000000000000D+00,  -7.421562080241037D-06, &
   0.000000000000000D+00,   2.902032320977638D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.218662882414851D-08,   0.000000000000000D+00,  -6.031400734592363D-07, &
   0.000000000000000D+00,   1.581432499856176D-06,   0.000000000000000D+00, &
   8.779144147348675D-06,   0.000000000000000D+00,  -5.374344522660166D-05, &
   0.000000000000000D+00,  -3.442195825008331D-05,   0.000000000000000D+00, &
  -2.114375741084733D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.093291171869555D-10,   0.000000000000000D+00, &
   7.561274008654276D-08,   0.000000000000000D+00,  -2.144005340410584D-06, &
   0.000000000000000D+00,   6.171240914177157D-06,   0.000000000000000D+00, &
   1.652449272553068D-05,   0.000000000000000D+00,  -2.980268212733850D-05, &
   0.000000000000000D+00,  -1.520069421680747D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.174908794704933D-08, &
   0.000000000000000D+00,   1.355788720971713D-07,   0.000000000000000D+00, &
  -5.172987143847838D-06,   0.000000000000000D+00,   1.415698337952303D-05, &
   0.000000000000000D+00,   2.551148259738402D-05,   0.000000000000000D+00, &
   3.151744179576224D-06,   0.000000000000000D+00,   7.840722295373466D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.393158828683223D-08, &
   0.000000000000000D+00,   5.287319114864217D-07,   0.000000000000000D+00, &
   4.310093926392329D-07,   0.000000000000000D+00,  -1.032063198836697D-05, &
   0.000000000000000D+00,   1.491703446999449D-05,   0.000000000000000D+00, &
   1.946877404641066D-05,   0.000000000000000D+00,   6.515133697140422D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.272107424389454D-09,   0.000000000000000D+00,  -1.698978507557429D-07, &
   0.000000000000000D+00,   1.534436845571682D-06,   0.000000000000000D+00, &
   1.771773797070784D-07,   0.000000000000000D+00,  -1.649247605911069D-05, &
   0.000000000000000D+00,   2.517048110309865D-06,   0.000000000000000D+00, &
   1.974404707253768D-06,   0.000000000000000D+00,  -3.384210013009699D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.603319280005432D-10, &
   0.000000000000000D+00,  -3.414749895737554D-08,   0.000000000000000D+00, &
  -4.879897957258918D-07,   0.000000000000000D+00,   3.296692078200505D-06, &
   0.000000000000000D+00,  -1.221544117626674D-06,   0.000000000000000D+00, &
  -1.732099898508381D-05,   0.000000000000000D+00,  -6.235950983875629D-06, &
   0.000000000000000D+00,  -3.001184174508703D-06,   0.000000000000000D+00 /)
lem(:, 91) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.098825512497703D-03, &
  -2.484763339484195D-02,  -5.193350783021276D-04,   4.578165622219293D-02, &
  -5.931169916422898D-03,  -3.860283212445861D-02,  -1.140564144891997D-03, &
   1.603089064805028D-02,   6.436041512553799D-03,  -1.696824916459088D-03, &
  -3.540365765838039D-03,  -1.287733099652493D-03,   7.177800985017751D-04, &
   7.492137402963013D-04,  -6.319705069720213D-05,  -3.459599332242671D-04, &
  -5.650909745860123D-05,   1.554949143208730D-04,   6.068902442120392D-05, &
  -7.047157712382262D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.098825512497715D-03,  -2.484763339484200D-02,  -5.193350783021367D-04, &
   4.578165622219295D-02,  -5.931169916422906D-03,  -3.860283212445861D-02, &
  -1.140564144891994D-03,   1.603089064805027D-02,   6.436041512553754D-03, &
  -1.696824916459085D-03,  -3.540365765837991D-03,  -1.287733099652510D-03, &
   7.177800985017357D-04,   7.492137402963546D-04,  -6.319705069719759D-05, &
  -3.459599332243430D-04,  -5.650909745859704D-05,   1.554949143209001D-04, &
   6.068902442121736D-05,  -7.047157712384749D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.332001332001340D-04,   0.000000000000000D+00, &
  -9.490509490509491D-04,   0.000000000000000D+00,  -2.081252081252057D-05, &
   2.318052318052323D-03,   0.000000000000000D+00,   2.192252192252166D-05, &
   0.000000000000000D+00,  -1.492951492951495D-03,   0.000000000000000D+00, &
   2.455877455877460D-04,   0.000000000000000D+00,  -1.387501387501460D-06, &
  -3.239135592076768D-03,   0.000000000000000D+00,  -8.597665950607130D-04, &
   0.000000000000000D+00,  -5.737726325961620D-05,   0.000000000000000D+00, &
   6.741624388683208D-03,   0.000000000000000D+00,   7.117882117882126D-04, &
   0.000000000000000D+00,   2.677061500590887D-05,   0.000000000000000D+00, &
   2.756638786050553D-05,  -3.622074519907338D-03,   0.000000000000000D+00, &
   1.080439741430454D-03,   0.000000000000000D+00,   2.861431654001316D-04, &
   0.000000000000000D+00,   1.322958946797956D-05,   0.000000000000000D+00, &
  -1.997108498656485D-03,   0.000000000000000D+00,  -2.477482098225128D-03, &
   0.000000000000000D+00,  -4.083077225492087D-04,   0.000000000000000D+00, &
  -7.462115627750315D-05,   0.000000000000000D+00,  -8.575633817506886D-06, &
   3.505885239631364D-03,   0.000000000000000D+00,   1.215507059160311D-03, &
   0.000000000000000D+00,  -2.853055097637141D-04,   0.000000000000000D+00, &
  -2.669508280963383D-05,   0.000000000000000D+00,   1.914083401119018D-05, &
   0.000000000000000D+00,  -1.690878780662061D-03,   0.000000000000000D+00, &
   9.397534010537102D-04,   0.000000000000000D+00,   9.106712860582840D-04, &
   0.000000000000000D+00,   1.779228898036948D-04,   0.000000000000000D+00, &
   2.866319320943947D-05,   0.000000000000000D+00,  -1.597702276760916D-05, &
   2.923401361952980D-04,   0.000000000000000D+00,  -1.309735940911064D-03, &
   0.000000000000000D+00,  -3.264149776129855D-04,   0.000000000000000D+00, &
   1.787639150507634D-05,   0.000000000000000D+00,  -3.132347747306854D-05, &
   0.000000000000000D+00,  -1.701575048398939D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.597872177627184D-04, &
   0.000000000000000D+00,  -5.355366877779045D-04,   0.000000000000000D+00, &
  -2.609766329227228D-04,   0.000000000000000D+00,  -5.285026041415675D-05, &
   0.000000000000000D+00,   3.204547151568502D-06,   0.000000000000000D+00, &
   1.100335572991830D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.628891725674636D-05,   0.000000000000000D+00,   4.502508305173536D-04, &
   0.000000000000000D+00,   2.236436425930257D-05,   0.000000000000000D+00, &
   1.846376892122424D-05,   0.000000000000000D+00,   2.114624169162343D-05, &
   0.000000000000000D+00,   7.481927902403082D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.149168908894250D-05, &
   0.000000000000000D+00,   2.317428014022498D-05,   0.000000000000000D+00, &
   2.366974831303813D-04,   0.000000000000000D+00,   6.813259353978192D-05, &
   0.000000000000000D+00,   1.116118514600175D-05,   0.000000000000000D+00, &
  -5.409188363289167D-06,   0.000000000000000D+00,   9.121285976756479D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.277530997142283D-06, &
   0.000000000000000D+00,  -1.177734486389767D-04,   0.000000000000000D+00, &
  -5.364877740029180D-05,   0.000000000000000D+00,   2.091462234987620D-05, &
   0.000000000000000D+00,  -1.057120114535510D-05,   0.000000000000000D+00, &
  -1.027975280035415D-05,   0.000000000000000D+00,  -6.781684027778102D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.727546619187504D-05, &
   0.000000000000000D+00,  -1.285123993429288D-04,   0.000000000000000D+00, &
  -7.262678645519565D-05,   0.000000000000000D+00,  -2.198321241382021D-05, &
   0.000000000000000D+00,  -2.375890389187176D-06,   0.000000000000000D+00, &
   3.972355046572864D-07,   0.000000000000000D+00,  -2.959280303030723D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.292783260932360D-05,   0.000000000000000D+00, &
   4.466287327952552D-05,   0.000000000000000D+00,  -4.476880933558708D-05, &
   0.000000000000000D+00,  -1.225562914440491D-05,   0.000000000000000D+00, &
   5.583043543280519D-06,   0.000000000000000D+00,   3.965815530072586D-06, &
   0.000000000000000D+00,  -2.925405757436886D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.822263274967502D-06,   0.000000000000000D+00, &
   3.675817237447758D-05,   0.000000000000000D+00,   6.212772740227632D-05, &
   0.000000000000000D+00,   1.654545274525839D-05,   0.000000000000000D+00, &
   8.814051915120619D-06,   0.000000000000000D+00,   1.956160101549159D-06, &
   0.000000000000000D+00,   9.095315491810670D-08,   0.000000000000000D+00, &
   1.402279158016768D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.802245802759748D-07, &
   0.000000000000000D+00,  -1.777743177580640D-05,   0.000000000000000D+00, &
   3.099949764968391D-05,   0.000000000000000D+00,   3.253273958765912D-05, &
   0.000000000000000D+00,   5.911837300413260D-06,   0.000000000000000D+00, &
  -3.169545135413808D-06,   0.000000000000000D+00,   6.644492118123145D-07, &
   0.000000000000000D+00,   2.737359686212096D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.132292129842922D-06,   0.000000000000000D+00,  -3.685110288856343D-05, &
   0.000000000000000D+00,  -4.191015896549489D-06,   0.000000000000000D+00, &
  -3.594569363423295D-06,   0.000000000000000D+00,  -4.458654564487705D-06, &
   0.000000000000000D+00,  -1.304508750797863D-06,   0.000000000000000D+00, &
   1.598158717909563D-06,   0.000000000000000D+00,  -5.657578403857180D-07 /)
lem(:, 92) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.216845660478461D-03,  -6.679800529538722D-03, &
   7.635010782542063D-03,   8.124101371100546D-03,  -1.777685698258335D-02, &
  -7.499353774287510D-04,   1.465479130142956D-02,  -2.411545472899990D-03, &
  -6.429306248420487D-03,   6.027616345783339D-04,   2.148493967814866D-03, &
   4.157662556342471D-04,  -7.253045696990045D-04,  -3.623043510855776D-04, &
   2.287606947258683D-04,   2.124651701581432D-04,  -6.882997599840368D-05, &
  -1.173211211719084D-04,   1.717515113993941D-05,   6.519300111601608D-05, &
  -8.204428247195867D-07,   0.000000000000000D+00,  -1.216845660478463D-03, &
   6.679800529538715D-03,  -7.635010782542061D-03,  -8.124101371100560D-03, &
   1.777685698258333D-02,   7.499353774287460D-04,  -1.465479130142956D-02, &
   2.411545472899990D-03,   6.429306248420483D-03,  -6.027616345783464D-04, &
  -2.148493967814856D-03,  -4.157662556342361D-04,   7.253045696989904D-04, &
   3.623043510855746D-04,  -2.287606947258487D-04,  -2.124651701581500D-04, &
   6.882997599839539D-05,   1.173211211719215D-04,  -1.717515113994812D-05, &
  -6.519300111601038D-05,   8.204428247205206D-07,   0.000000000000000D+00, &
  -6.084228302392187D-04,   0.000000000000000D+00,  -2.943334920023923D-04, &
   0.000000000000000D+00,   3.207286481336349D-04,   0.000000000000000D+00, &
   6.515988071861866D-05,   0.000000000000000D+00,  -1.461205965550106D-04, &
   0.000000000000000D+00,   2.920340115961131D-05,   0.000000000000000D+00, &
   1.609800524609504D-05,   0.000000000000000D+00,  -9.961685914370442D-06, &
   0.000000000000000D+00,   4.725176528209905D-06,   0.000000000000000D+00, &
  -2.172536029268544D-06,   0.000000000000000D+00,   1.009618596261854D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.810004810004818D-05, &
   0.000000000000000D+00,  -1.350501350501354D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.924001924001924D-04,   0.000000000000000D+00, &
   4.162504162504113D-06,   0.000000000000000D+00,  -3.145003145003146D-05, &
   0.000000000000000D+00,   9.573759573759576D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.370268076150426D-04,   0.000000000000000D+00, &
  -1.946582828935771D-04,   0.000000000000000D+00,  -4.578754578754577D-05, &
   0.000000000000000D+00,   5.630535042299745D-04,   0.000000000000000D+00, &
   7.955396610858808D-05,   0.000000000000000D+00,  -1.077354018530483D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.685283871042389D-04, &
   0.000000000000000D+00,   2.771609195757808D-04,   0.000000000000000D+00, &
   1.056605932766924D-04,   0.000000000000000D+00,   1.531012560424333D-05, &
   0.000000000000000D+00,  -4.348363481490405D-05,   0.000000000000000D+00, &
  -5.088521729388606D-04,   0.000000000000000D+00,  -9.114552458205715D-05, &
   0.000000000000000D+00,  -2.123958199809594D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.740522864361873D-04,   0.000000000000000D+00, &
   3.072134264084724D-04,   0.000000000000000D+00,  -1.070223206446116D-04, &
   0.000000000000000D+00,  -2.143503497992655D-05,   0.000000000000000D+00, &
   2.161017302271172D-05,   0.000000000000000D+00,  -2.306259272203550D-04, &
   0.000000000000000D+00,   9.783388266360392D-05,   0.000000000000000D+00, &
   2.530741219877164D-04,   0.000000000000000D+00,   5.254033891804797D-05, &
   0.000000000000000D+00,   1.197639488564411D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.987768770915890D-05,   0.000000000000000D+00, &
  -2.801131654274743D-04,   0.000000000000000D+00,  -1.220467063783795D-04, &
   0.000000000000000D+00,   1.440830728924681D-05,   0.000000000000000D+00, &
  -1.072314193822806D-05,   0.000000000000000D+00,  -1.698063221500720D-05, &
   0.000000000000000D+00,  -2.322351225299150D-05,   0.000000000000000D+00, &
   1.855068250276217D-04,   0.000000000000000D+00,  -1.110138384670622D-04, &
   0.000000000000000D+00,  -7.922086158558081D-05,   0.000000000000000D+00, &
  -1.663772788230152D-05,   0.000000000000000D+00,  -4.252499468933840D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.307559179278378D-06, &
   0.000000000000000D+00,  -4.578207614570900D-05,   0.000000000000000D+00, &
   1.587500308727931D-04,   0.000000000000000D+00,   1.760212231843672D-05, &
   0.000000000000000D+00,   4.470814782389368D-06,   0.000000000000000D+00, &
   8.790093825565050D-06,   0.000000000000000D+00,   9.277108886483135D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.261357702736005D-06,   0.000000000000000D+00,  -4.807302895874677D-05, &
   0.000000000000000D+00,   6.787405802246298D-05,   0.000000000000000D+00, &
   1.916850489589083D-05,   0.000000000000000D+00,   5.427511399409958D-06, &
   0.000000000000000D+00,   1.553712491212553D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.020576064727137D-06,   0.000000000000000D+00,  -2.896790620575802D-05, &
   0.000000000000000D+00,  -4.165442649760865D-05,   0.000000000000000D+00, &
   4.000345540887668D-06,   0.000000000000000D+00,  -3.782024974574862D-06, &
   0.000000000000000D+00,  -9.937678059176303D-07,   0.000000000000000D+00, &
   3.134750449203680D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.809012443014073D-06,   0.000000000000000D+00, &
   2.424483579416956D-05,   0.000000000000000D+00,  -2.770100108234454D-05, &
   0.000000000000000D+00,  -2.373941870972095D-05,   0.000000000000000D+00, &
  -5.894715114653533D-06,   0.000000000000000D+00,  -3.045845639587245D-06, &
   0.000000000000000D+00,   1.049433422480229D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.289345007576916D-07,   0.000000000000000D+00,  -2.563074900744991D-06, &
   0.000000000000000D+00,   3.299682319112247D-05,   0.000000000000000D+00, &
  -5.690525519257561D-06,   0.000000000000000D+00,  -2.519131021512728D-06, &
   0.000000000000000D+00,   6.956684512202091D-07,   0.000000000000000D+00, &
  -1.409858587592948D-06,   0.000000000000000D+00,  -1.862083771263552D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.172706888573538D-06, &
   0.000000000000000D+00,  -3.287558203986110D-07,   0.000000000000000D+00, &
   2.507854144964910D-05,   0.000000000000000D+00,   5.382310492317756D-06, &
   0.000000000000000D+00,   3.231547061743286D-06,   0.000000000000000D+00, &
  -4.810185474243505D-08,   0.000000000000000D+00,  -1.545864888150005D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.397933568495762D-06,   0.000000000000000D+00,  -1.143842076240589D-05, &
   0.000000000000000D+00,  -1.812285818153171D-06,   0.000000000000000D+00, &
   8.410618463298417D-06,   0.000000000000000D+00,   1.167009984495272D-06, &
   0.000000000000000D+00,   1.210500604308407D-06,   0.000000000000000D+00, &
   2.222170190920347D-07,   0.000000000000000D+00,   9.250714969709468D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.209373963500383D-07, &
   0.000000000000000D+00,   4.257678236419093D-06,   0.000000000000000D+00, &
  -1.602049654544037D-05,   0.000000000000000D+00,  -5.591929834278602D-06, &
   0.000000000000000D+00,  -1.428889074893432D-06,   0.000000000000000D+00, &
  -1.150063642902420D-06,   0.000000000000000D+00,   1.495343423007971D-06, &
   0.000000000000000D+00,   8.840832968030072D-07,   0.000000000000000D+00 /)
lem(:, 93) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.178205744484535D-04,   9.810818137607530D-03,  -6.673743715587053D-03, &
  -1.842095970813370D-02,   7.034452207227690D-03,   2.376053504928658D-02, &
  -6.386167845546391D-03,  -1.503341413084412D-02,   3.455772013493024D-03, &
   4.479935723212846D-03,  -1.055390383165645D-03,  -7.668383908171729D-04, &
   3.605482925486255D-04,   2.548620859127936D-04,  -1.598815605583182D-04, &
  -1.197958549417336D-04,   7.455937159964132D-05,   6.445031279479412D-05, &
  -3.636984791611088D-05,  -3.711796046497907D-05,   1.853174085188403D-05, &
   2.238998694619239D-05,  -1.178205744484431D-04,   9.810818137607511D-03, &
  -6.673743715587045D-03,  -1.842095970813373D-02,   7.034452207227694D-03, &
   2.376053504928657D-02,  -6.386167845546407D-03,  -1.503341413084412D-02, &
   3.455772013493028D-03,   4.479935723212847D-03,  -1.055390383165653D-03, &
  -7.668383908171723D-04,   3.605482925486360D-04,   2.548620859127991D-04, &
  -1.598815605583279D-04,  -1.197958549417297D-04,   7.455937159965300D-05, &
   6.445031279478527D-05,  -3.636984791612133D-05,  -3.711796046496832D-05, &
   1.853174085189352D-05,   2.238998694617856D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.180453022595537D-03,   0.000000000000000D+00, &
  -1.202670707647032D-03,   0.000000000000000D+00,  -2.517466569011496D-04, &
   0.000000000000000D+00,   4.286814362532195D-04,   0.000000000000000D+00, &
  -4.778765601836635D-05,   0.000000000000000D+00,  -5.291456731485588D-05, &
   0.000000000000000D+00,   1.854402797428015D-05,   0.000000000000000D+00, &
  -6.077746645900831D-06,   0.000000000000000D+00,   2.064333252001734D-06, &
   0.000000000000000D+00,  -7.236430360302114D-07,   0.000000000000000D+00, &
   1.623376623376628D-04,  -6.007881007881003D-04,   0.000000000000000D+00, &
   5.019980019980018D-04,   0.000000000000000D+00,  -1.021201021201020D-04, &
   8.422133422133435D-04,   0.000000000000000D+00,   2.808302808302809D-04, &
   0.000000000000000D+00,  -1.913364413364413D-03,   0.000000000000000D+00, &
  -2.203352203352203D-04,   0.000000000000000D+00,   5.723443223443230D-05, &
   9.538663950428670D-04,   0.000000000000000D+00,  -3.956407737920341D-04, &
   0.000000000000000D+00,  -1.590117398940928D-04,   0.000000000000000D+00, &
   7.117882117882126D-04,   0.000000000000000D+00,   8.809654397889690D-04, &
   0.000000000000000D+00,   1.382277852866087D-04,   0.000000000000000D+00, &
   1.318636428930542D-05,  -1.006337144108041D-03,   0.000000000000000D+00, &
  -4.402434510793638D-04,   0.000000000000000D+00,   2.057793632318176D-04, &
   0.000000000000000D+00,   5.799181253902612D-05,   0.000000000000000D+00, &
   3.520473334714811D-04,   0.000000000000000D+00,  -3.115237016165814D-04, &
   0.000000000000000D+00,  -4.834067124443798D-04,   0.000000000000000D+00, &
  -8.895688075254627D-05,   0.000000000000000D+00,  -4.212919554323379D-05, &
   2.270691125180318D-05,   0.000000000000000D+00,   4.384796490059653D-04, &
   0.000000000000000D+00,   2.256524363722507D-04,   0.000000000000000D+00, &
  -6.319574818026826D-05,   0.000000000000000D+00,   2.678249521806305D-07, &
   0.000000000000000D+00,  -3.286187496713834D-05,   0.000000000000000D+00, &
  -1.542764081463768D-04,   0.000000000000000D+00,   1.897134298450087D-04, &
   0.000000000000000D+00,   2.051424023173247D-04,   0.000000000000000D+00, &
   4.554515925573399D-05,   0.000000000000000D+00,   1.895944263639695D-05, &
   3.389843879815845D-06,   0.000000000000000D+00,   3.436685358882063D-06, &
   0.000000000000000D+00,  -2.295908622046057D-04,   0.000000000000000D+00, &
  -7.165726075572628D-05,   0.000000000000000D+00,   6.195621102678271D-07, &
   0.000000000000000D+00,  -9.254783999975964D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.788780590407501D-07, &
   0.000000000000000D+00,   6.403648549293961D-05,   0.000000000000000D+00, &
  -1.170001696864328D-04,   0.000000000000000D+00,  -6.323156464152705D-05, &
   0.000000000000000D+00,  -1.660312599836974D-05,   0.000000000000000D+00, &
   3.605742131132799D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.418819525697924D-06,   0.000000000000000D+00,   1.067193825303932D-06, &
   0.000000000000000D+00,   9.081530662917790D-05,   0.000000000000000D+00, &
   5.491442757719762D-06,   0.000000000000000D+00,   6.377105132037276D-06, &
   0.000000000000000D+00,   3.896197085439297D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.357170222427859D-06, &
   0.000000000000000D+00,   4.539315792993519D-06,   0.000000000000000D+00, &
   3.818348628383754D-06,   0.000000000000000D+00,   5.617414518685082D-05, &
   0.000000000000000D+00,   1.877562254468843D-05,   0.000000000000000D+00, &
   3.331286468060779D-06,   0.000000000000000D+00,  -5.795257260101029D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.436777507836212D-08, &
   0.000000000000000D+00,  -2.043976249791351D-06,   0.000000000000000D+00, &
  -1.443265330985754D-05,   0.000000000000000D+00,  -1.188133786760145D-05, &
   0.000000000000000D+00,   4.789186520776777D-06,   0.000000000000000D+00, &
  -1.654701207111199D-06,   0.000000000000000D+00,  -2.456582046425758D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.746877237634243D-06, &
   0.000000000000000D+00,   4.859943529346152D-07,   0.000000000000000D+00, &
  -2.653712623063101D-05,   0.000000000000000D+00,  -1.837615632057181D-05, &
   0.000000000000000D+00,  -7.163766486231442D-06,   0.000000000000000D+00, &
   3.647611069485121D-07,   0.000000000000000D+00,   2.446800771385686D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.168625220185844D-07,   0.000000000000000D+00, &
  -2.070657544487873D-06,   0.000000000000000D+00,   7.758461733163283D-06, &
   0.000000000000000D+00,  -1.125153266690854D-05,   0.000000000000000D+00, &
  -3.827105495344455D-06,   0.000000000000000D+00,   7.205169113168067D-07, &
   0.000000000000000D+00,   1.652414448434929D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.469478877640186D-07,   0.000000000000000D+00, &
   1.971398748822764D-06,   0.000000000000000D+00,   4.179761786007957D-06, &
   0.000000000000000D+00,   1.433700925147396D-05,   0.000000000000000D+00, &
   4.946200842053047D-06,   0.000000000000000D+00,   2.098639956999245D-06, &
   0.000000000000000D+00,  -1.990129553524768D-09,   0.000000000000000D+00, &
  -8.561791398704661D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.912677542033784D-09, &
   0.000000000000000D+00,   4.028930475335314D-07,   0.000000000000000D+00, &
  -3.463483869684371D-07,   0.000000000000000D+00,   6.635893527969846D-06, &
   0.000000000000000D+00,   8.563572731788455D-06,   0.000000000000000D+00, &
   2.791896892528648D-06,   0.000000000000000D+00,  -9.910905667219181D-07, &
   0.000000000000000D+00,  -6.158392388860976D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.002643469503280D-07,   0.000000000000000D+00,   7.013762913260534D-07, &
   0.000000000000000D+00,  -6.416487766246295D-06,   0.000000000000000D+00, &
  -8.626109610814151D-07,   0.000000000000000D+00,  -9.668185752763974D-07, &
   0.000000000000000D+00,  -6.550490254104087D-07,   0.000000000000000D+00, &
  -2.872773182222276D-07,   0.000000000000000D+00,   2.828974453348401D-07 /)
lem(:, 94) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.844756840191302D-03,   2.725566278244475D-04, &
   1.589754057003323D-02,  -5.002016806804699D-03,  -1.612773226345084D-02, &
   8.586364322557525D-03,   7.492769088092474D-03,  -5.884621168420313D-03, &
  -1.240929166243430D-03,   1.978807189627244D-03,  -8.823617314498920D-05, &
  -4.783932064336446D-04,   8.447364371914213D-05,   1.519162047377763D-04, &
  -5.194862098598054D-05,  -5.893885473844315D-05,   3.057937344547918D-05, &
   2.655406950615447D-05,  -1.822172827648507D-05,  -1.340674941055798D-05, &
   1.114349038456737D-05,   0.000000000000000D+00,   6.844756840191297D-03, &
  -2.725566278244380D-04,  -1.589754057003319D-02,   5.002016806804705D-03, &
   1.612773226345088D-02,  -8.586364322557516D-03,  -7.492769088092472D-03, &
   5.884621168420319D-03,   1.240929166243432D-03,  -1.978807189627246D-03, &
   8.823617314499047D-05,   4.783932064336599D-04,  -8.447364371914844D-05, &
  -1.519162047377886D-04,   5.194862098599241D-05,   5.893885473844987D-05, &
  -3.057937344549070D-05,  -2.655406950615686D-05,   1.822172827649552D-05, &
   1.340674941055932D-05,  -1.114349038457589D-05,   0.000000000000000D+00, &
   3.422378420095644D-03,   0.000000000000000D+00,  -2.465337329012040D-03, &
   0.000000000000000D+00,   2.130049241412716D-03,   0.000000000000000D+00, &
  -8.095906969079273D-04,   0.000000000000000D+00,  -2.005505187717540D-04, &
   0.000000000000000D+00,   1.947659904109582D-04,   0.000000000000000D+00, &
  -3.410477397136869D-05,   0.000000000000000D+00,   5.514484467441736D-06, &
   0.000000000000000D+00,  -1.938831521104293D-07,   0.000000000000000D+00, &
  -6.572493900466436D-07,   0.000000000000000D+00,   6.315814292052919D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.705627705627707D-04, &
   0.000000000000000D+00,   7.932543646829361D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.493934636791780D-04,   0.000000000000000D+00, &
   3.036645893788751D-04,   0.000000000000000D+00,  -4.328607900036463D-04, &
   0.000000000000000D+00,  -1.038247466818895D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.528867024665342D-04,   0.000000000000000D+00, &
  -2.661157703174508D-04,   0.000000000000000D+00,  -2.216635488694311D-04, &
   0.000000000000000D+00,   7.955396610858808D-05,   0.000000000000000D+00, &
   4.804519580429949D-04,   0.000000000000000D+00,   1.058042680941840D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.194425801236947D-04, &
   0.000000000000000D+00,  -2.710678186131526D-04,   0.000000000000000D+00, &
   1.614506832054379D-04,   0.000000000000000D+00,   9.577652524633946D-05, &
   0.000000000000000D+00,   1.394950079160604D-04,   0.000000000000000D+00, &
  -1.058275789220800D-04,   0.000000000000000D+00,  -3.313245529539941D-04, &
   0.000000000000000D+00,  -7.725440194480441D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.125725814801451D-05,   0.000000000000000D+00, &
   2.437588947982577D-04,   0.000000000000000D+00,   1.689965643968299D-04, &
   0.000000000000000D+00,  -5.120465404908124D-05,   0.000000000000000D+00, &
   9.898890328940482D-07,   0.000000000000000D+00,  -7.636473801887346D-06, &
   0.000000000000000D+00,  -1.362753419365540D-04,   0.000000000000000D+00, &
   1.070099791167166D-04,   0.000000000000000D+00,   1.517048053425758D-04, &
   0.000000000000000D+00,   3.353718448411752D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.164419755943461D-06,   0.000000000000000D+00, &
   1.975400904308852D-05,   0.000000000000000D+00,  -1.721543420459828D-04, &
   0.000000000000000D+00,  -5.800408087896307D-05,   0.000000000000000D+00, &
   1.483170266138665D-06,   0.000000000000000D+00,  -2.812993432787898D-05, &
   0.000000000000000D+00,  -1.320718394733533D-06,   0.000000000000000D+00, &
   9.989620444786474D-06,   0.000000000000000D+00,   6.528914019828005D-05, &
   0.000000000000000D+00,  -8.243643900781488D-05,   0.000000000000000D+00, &
  -4.802252930234292D-05,   0.000000000000000D+00,  -9.846272701037369D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.156980870164805D-08, &
   0.000000000000000D+00,  -2.652642884937274D-06,   0.000000000000000D+00, &
   9.560650443147575D-06,   0.000000000000000D+00,   7.705397625810862D-05, &
   0.000000000000000D+00,   4.965837483725566D-06,   0.000000000000000D+00, &
   7.083548820323574D-06,   0.000000000000000D+00,   1.579163566468241D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.318450494459361D-06,   0.000000000000000D+00,  -9.896346375732134D-06, &
   0.000000000000000D+00,   1.849017425642819D-06,   0.000000000000000D+00, &
   4.367905545275281D-05,   0.000000000000000D+00,   1.306154866763008D-05, &
   0.000000000000000D+00,   5.186734923063118D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.953431649013828D-07,   0.000000000000000D+00,  -1.921003804308579D-06, &
   0.000000000000000D+00,  -2.316595261052789D-05,   0.000000000000000D+00, &
  -1.284173575484446D-05,   0.000000000000000D+00,   4.374296691829702D-06, &
   0.000000000000000D+00,  -4.130853490632541D-06,   0.000000000000000D+00, &
  -4.412272478678575D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.086813386103679D-07,   0.000000000000000D+00, &
   1.965632524205920D-06,   0.000000000000000D+00,   2.805438219020053D-06, &
   0.000000000000000D+00,  -2.414513991952286D-05,   0.000000000000000D+00, &
  -1.475819223150208D-05,   0.000000000000000D+00,  -5.369053536595672D-06, &
   0.000000000000000D+00,  -3.040966350118232D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.612588897946475D-09,   0.000000000000000D+00,   2.337916783596095D-07, &
   0.000000000000000D+00,   4.630130420805188D-06,   0.000000000000000D+00, &
   1.093830001948445D-05,   0.000000000000000D+00,  -8.433308288596034D-06, &
   0.000000000000000D+00,  -2.552650301334413D-06,   0.000000000000000D+00, &
   4.919874479026158D-07,   0.000000000000000D+00,   8.377437249680484D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.204556001800255D-07, &
   0.000000000000000D+00,  -2.287039036659231D-07,   0.000000000000000D+00, &
   7.675240731376844D-06,   0.000000000000000D+00,   1.394705947286475D-05, &
   0.000000000000000D+00,   3.136155241346879D-06,   0.000000000000000D+00, &
   2.919017985982324D-06,   0.000000000000000D+00,   9.633013412834183D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.170459967160283D-07,   0.000000000000000D+00,  -1.213769810284104D-06, &
   0.000000000000000D+00,  -4.132350453082401D-06,   0.000000000000000D+00, &
   5.126558491789411D-06,   0.000000000000000D+00,   7.159750631222012D-06, &
   0.000000000000000D+00,   1.744529490526621D-06,   0.000000000000000D+00, &
   5.990491590554280D-07,   0.000000000000000D+00,  -7.452400030444693D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.337008308517458D-08, &
   0.000000000000000D+00,   1.633112302464342D-07,   0.000000000000000D+00, &
  -1.807898631347235D-06,   0.000000000000000D+00,  -8.837997515699185D-06, &
   0.000000000000000D+00,  -1.425960470867529D-06,   0.000000000000000D+00, &
  -5.730970160811700D-07,   0.000000000000000D+00,  -8.232179972030957D-07, &
   0.000000000000000D+00,  -1.506207473702158D-07,   0.000000000000000D+00 /)
lem(:, 95) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.583224597306874D-03,   3.821655902440145D-03,  -5.583517204289276D-03, &
  -1.353006912711197D-02,   8.181317171811572D-03,   1.101864618262463D-02, &
  -7.934056262353134D-03,  -2.212091877783861D-03,   3.896136987197043D-03, &
  -1.346227321135889D-03,  -7.649210806673372D-04,   7.872542637297725D-04, &
  -1.787075716291392D-05,  -2.410056640162463D-04,   5.015767165054412D-05, &
   8.194144610635579D-05,  -3.263953979705877D-05,  -3.100525817430756D-05, &
   1.920845888467901D-05,   1.280561189542264D-05,  -1.137724623651838D-05, &
  -5.671137136650086D-06,   6.583224597306881D-03,   3.821655902440149D-03, &
  -5.583517204289274D-03,  -1.353006912711200D-02,   8.181317171811557D-03, &
   1.101864618262463D-02,  -7.934056262353155D-03,  -2.212091877783860D-03, &
   3.896136987197036D-03,  -1.346227321135889D-03,  -7.649210806673383D-04, &
   7.872542637297722D-04,  -1.787075716290637D-05,  -2.410056640162478D-04, &
   5.015767165053665D-05,   8.194144610635821D-05,  -3.263953979705772D-05, &
  -3.100525817430407D-05,   1.920845888467720D-05,   1.280561189541768D-05, &
  -1.137724623650823D-05,  -5.671137136649834D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.634088370992590D-04,   0.000000000000000D+00, &
  -8.200027552138862D-04,   0.000000000000000D+00,  -1.078041834132912D-03, &
   0.000000000000000D+00,   1.555100142368273D-03,   0.000000000000000D+00, &
  -5.049562319274190D-04,   0.000000000000000D+00,   4.039714278877122D-06, &
   0.000000000000000D+00,   1.367338886410414D-05,   0.000000000000000D+00, &
  -8.060094118218340D-06,   0.000000000000000D+00,   4.190335277268439D-06, &
   0.000000000000000D+00,  -2.204025960126812D-06,   0.000000000000000D+00, &
  -3.427128427128424D-04,  -1.734376734376734D-04,   0.000000000000000D+00, &
   5.143467643467643D-04,   0.000000000000000D+00,   1.650086025086025D-04, &
  -1.685814185814187D-04,   0.000000000000000D+00,   1.781204906204906D-04, &
   0.000000000000000D+00,  -2.955377955377956D-04,   0.000000000000000D+00, &
  -4.050463425463431D-04,   0.000000000000000D+00,  -1.546052327302328D-04, &
   2.328553799142033D-04,   0.000000000000000D+00,   1.896256754450035D-04, &
   0.000000000000000D+00,  -1.855128463584345D-04,   0.000000000000000D+00, &
   2.677061500590887D-05,   0.000000000000000D+00,   1.382277852866087D-04, &
   0.000000000000000D+00,   4.101736362765770D-04,   0.000000000000000D+00, &
   1.910154541128803D-04,  -6.810870587960346D-05,   0.000000000000000D+00, &
  -1.729900646309315D-04,   0.000000000000000D+00,  -2.147929444341429D-04, &
   0.000000000000000D+00,   1.044290448436152D-04,   0.000000000000000D+00, &
   7.749388554341911D-06,   0.000000000000000D+00,   4.785054127159398D-05, &
   0.000000000000000D+00,  -1.045934772714959D-04,   0.000000000000000D+00, &
  -2.985096943712463D-04,   0.000000000000000D+00,  -1.801404958453140D-04, &
   8.247686266262275D-06,   0.000000000000000D+00,   2.803139413046432D-06, &
   0.000000000000000D+00,   1.747958210806508D-04,   0.000000000000000D+00, &
   1.453882576294534D-04,   0.000000000000000D+00,  -2.144213849660826D-05, &
   0.000000000000000D+00,  -2.330400937212223D-06,   0.000000000000000D+00, &
   1.359042574212963D-06,   0.000000000000000D+00,  -7.684680551360116D-05, &
   0.000000000000000D+00,   9.673791631270351D-05,   0.000000000000000D+00, &
   1.498204409297917D-04,   0.000000000000000D+00,   1.024226181027955D-04, &
   1.694921939910239D-07,   0.000000000000000D+00,  -7.744392503445229D-07, &
   0.000000000000000D+00,   1.280479946224002D-05,   0.000000000000000D+00, &
  -1.321253125179253D-04,   0.000000000000000D+00,  -5.242333188754619D-05, &
   0.000000000000000D+00,  -7.687089363351010D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.275410416469574D-06, &
   0.000000000000000D+00,   1.153642237149368D-06,   0.000000000000000D+00, &
   4.296648275007987D-05,   0.000000000000000D+00,  -7.675893384494710D-05, &
   0.000000000000000D+00,  -5.703837911974025D-05,   0.000000000000000D+00, &
  -3.047530269209951D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.956459971131624D-07,   0.000000000000000D+00,  -3.235711429529540D-06, &
   0.000000000000000D+00,   5.753533632559193D-06,   0.000000000000000D+00, &
   6.084306382007427D-05,   0.000000000000000D+00,   3.181466710853451D-06, &
   0.000000000000000D+00,  -3.897114183246142D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.816395611379123D-08, &
   0.000000000000000D+00,   1.534128922212335D-06,   0.000000000000000D+00, &
  -8.329611870738654D-06,   0.000000000000000D+00,   5.969407534694348D-06, &
   0.000000000000000D+00,   4.206322096328613D-05,   0.000000000000000D+00, &
   1.766238954130684D-05,   0.000000000000000D+00,   2.543131510416598D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.329910777583593D-09, &
   0.000000000000000D+00,   2.028225460046511D-07,   0.000000000000000D+00, &
   2.799596360830179D-06,   0.000000000000000D+00,  -1.896314196111657D-05, &
   0.000000000000000D+00,  -7.407803242058981D-06,   0.000000000000000D+00, &
   8.259504684384209D-06,   0.000000000000000D+00,   2.924897639253166D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.005269878895945D-07, &
   0.000000000000000D+00,   6.638633951150272D-07,   0.000000000000000D+00, &
   4.579680978495966D-06,   0.000000000000000D+00,  -2.254240913236837D-05, &
   0.000000000000000D+00,  -1.467737521093410D-05,   0.000000000000000D+00, &
  -2.759183362210760D-06,   0.000000000000000D+00,   1.643098948396831D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.659005777015582D-08,   0.000000000000000D+00, &
  -9.510943391005755D-07,   0.000000000000000D+00,   2.372841784801264D-06, &
   0.000000000000000D+00,   8.130047213091291D-06,   0.000000000000000D+00, &
  -1.009534323885802D-05,   0.000000000000000D+00,  -5.996607400458939D-06, &
   0.000000000000000D+00,   3.227961907158869D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.883870108046609D-09,   0.000000000000000D+00, &
   7.988573820183067D-08,   0.000000000000000D+00,  -2.109224615305689D-06, &
   0.000000000000000D+00,   6.183114204489696D-06,   0.000000000000000D+00, &
   1.157793492771812D-05,   0.000000000000000D+00,   1.888563392227241D-06, &
   0.000000000000000D+00,  -5.588947163043682D-07,   0.000000000000000D+00, &
  -1.309817527239995D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.836560226648907D-10, &
   0.000000000000000D+00,   1.628152721626416D-07,   0.000000000000000D+00, &
   1.733220991355446D-07,   0.000000000000000D+00,  -4.392305283795383D-06, &
   0.000000000000000D+00,   6.110276050676163D-06,   0.000000000000000D+00, &
   8.045959725861847D-06,   0.000000000000000D+00,   1.681138248390282D-06, &
   0.000000000000000D+00,  -7.204321913031795D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.822153441327561D-08,   0.000000000000000D+00,   5.386354731886648D-07, &
   0.000000000000000D+00,  -2.876198707155626D-08,   0.000000000000000D+00, &
  -7.302165532880538D-06,   0.000000000000000D+00,  -1.038278771383879D-08, &
   0.000000000000000D+00,   1.502572513162066D-06,   0.000000000000000D+00, &
   1.459251909386494D-07,   0.000000000000000D+00,   7.943812505905118D-07 /)
lem(:, 96) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.106854104114777D-03,   2.920249583833310D-03, &
   9.436086475686683D-03,  -6.269604399239508D-03,  -5.448112601545737D-03, &
   5.544099003860303D-03,  -4.246921267373440D-04,  -1.654544139583097D-03, &
   1.361996168847451D-03,  -3.416817405313327D-04,  -3.330174592536803D-04, &
   3.109003463047108D-04,  -1.051736342745132D-05,  -1.110426832183548D-04, &
   3.282266771806025D-05,   3.988010672977766D-05,  -2.272403202773785D-05, &
  -1.499393004935854D-05,   1.385052109484500D-05,   5.774524174587076D-06, &
  -8.363747575078231D-06,   0.000000000000000D+00,   4.106854104114789D-03, &
  -2.920249583833304D-03,  -9.436086475686715D-03,   6.269604399239507D-03, &
   5.448112601545739D-03,  -5.544099003860319D-03,   4.246921267373452D-04, &
   1.654544139583099D-03,  -1.361996168847451D-03,   3.416817405313320D-04, &
   3.330174592536838D-04,  -3.109003463047056D-04,   1.051736342744766D-05, &
   1.110426832183518D-04,  -3.282266771805744D-05,  -3.988010672977826D-05, &
   2.272403202773719D-05,   1.499393004935660D-05,  -1.385052109484385D-05, &
  -5.774524174583468D-06,   8.363747575077615D-06,   0.000000000000000D+00, &
   2.053427052057391D-03,   0.000000000000000D+00,  -1.062691072874635D-02, &
   0.000000000000000D+00,   1.229705473979278D-02,   0.000000000000000D+00, &
  -5.794248450439813D-03,   0.000000000000000D+00,   4.931570133731607D-04, &
   0.000000000000000D+00,   4.012305873888092D-04,   0.000000000000000D+00, &
  -1.207350393457018D-04,   0.000000000000000D+00,   4.164819857527016D-05, &
   0.000000000000000D+00,  -1.663241382080069D-05,   0.000000000000000D+00, &
   7.474946480151689D-06,   0.000000000000000D+00,  -3.689625066828665D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.623376623376626D-04, &
   0.000000000000000D+00,   1.653439153439154D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.105006105006109D-05,   0.000000000000000D+00, &
   4.230144855144851D-04,   0.000000000000000D+00,  -7.284382284382271D-05, &
   0.000000000000000D+00,  -3.650863025863031D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.292361233537711D-05,   0.000000000000000D+00, &
   1.382502301619947D-04,   0.000000000000000D+00,  -5.749229324596972D-04, &
   0.000000000000000D+00,  -1.077354018530483D-05,   0.000000000000000D+00, &
   1.058042680941840D-04,   0.000000000000000D+00,   4.639100686894807D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.381948976376220D-05, &
   0.000000000000000D+00,  -1.428943863154390D-04,   0.000000000000000D+00, &
  -1.636122681498065D-04,   0.000000000000000D+00,   4.213388722023595D-04, &
   0.000000000000000D+00,   1.182454742826257D-05,   0.000000000000000D+00, &
   4.351835814684109D-05,   0.000000000000000D+00,  -1.257141229083953D-04, &
   0.000000000000000D+00,  -3.348659267897853D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.450473846759368D-07,   0.000000000000000D+00, &
   1.385353552536212D-05,   0.000000000000000D+00,   1.666403127316440D-04, &
   0.000000000000000D+00,   9.432066229908720D-05,   0.000000000000000D+00, &
  -1.454936693423536D-04,   0.000000000000000D+00,  -2.257374548396325D-06, &
   0.000000000000000D+00,  -9.086415820161883D-06,   0.000000000000000D+00, &
  -5.702178436927237D-05,   0.000000000000000D+00,   1.218627662005178D-04, &
   0.000000000000000D+00,   1.296707453435151D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.025599299123007D-07,   0.000000000000000D+00, &
  -8.653206846100692D-07,   0.000000000000000D+00,  -6.661953719094564D-06, &
   0.000000000000000D+00,  -1.246132538907266D-04,   0.000000000000000D+00, &
  -1.746717374057235D-05,   0.000000000000000D+00,  -7.389386262477352D-06, &
   0.000000000000000D+00,  -9.104952569758801D-08,   0.000000000000000D+00, &
   8.979584376677988D-07,   0.000000000000000D+00,   6.794179809137965D-06, &
   0.000000000000000D+00,   2.439065773790754D-05,   0.000000000000000D+00, &
  -8.115307150273049D-05,   0.000000000000000D+00,  -2.313694945166745D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.370377233418303D-09, &
   0.000000000000000D+00,   2.573032915031231D-07,   0.000000000000000D+00, &
  -6.228102729144720D-07,   0.000000000000000D+00,   1.678935318273128D-05, &
   0.000000000000000D+00,   4.942665780695475D-05,   0.000000000000000D+00, &
  -1.878864762437370D-06,   0.000000000000000D+00,   2.902032320977639D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.073828043001873D-07,   0.000000000000000D+00,   7.833340886240680D-07, &
   0.000000000000000D+00,  -7.551089010256540D-06,   0.000000000000000D+00, &
   1.573540840581331D-05,   0.000000000000000D+00,   3.080868454304038D-05, &
   0.000000000000000D+00,   5.606914786602238D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.979409013735545D-08,   0.000000000000000D+00,  -5.737952782473283D-07, &
   0.000000000000000D+00,   7.669385787444099D-07,   0.000000000000000D+00, &
  -1.997572496669679D-05,   0.000000000000000D+00,  -6.782767429584940D-07, &
   0.000000000000000D+00,  -4.276074565866015D-06,   0.000000000000000D+00, &
  -1.520069421680747D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.521371481015584D-09,   0.000000000000000D+00, &
  -2.952806156478050D-08,   0.000000000000000D+00,  -4.598826913614632D-07, &
   0.000000000000000D+00,   1.366235924481011D-06,   0.000000000000000D+00, &
  -2.242114376126916D-05,   0.000000000000000D+00,  -6.939978698669339D-06, &
   0.000000000000000D+00,  -6.382811253416725D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.756833011390984D-12,   0.000000000000000D+00,   9.793115779792266D-08, &
   0.000000000000000D+00,  -1.403544908314541D-07,   0.000000000000000D+00, &
   2.432511908578289D-06,   0.000000000000000D+00,   5.005675668220130D-06, &
   0.000000000000000D+00,  -7.109985276162786D-06,   0.000000000000000D+00, &
   3.617970842091880D-06,   0.000000000000000D+00,   6.515133697140412D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.292334829133762D-08, &
   0.000000000000000D+00,   3.026779850793803D-07,   0.000000000000000D+00, &
  -6.915492643794380D-07,   0.000000000000000D+00,   7.585027615089838D-06, &
   0.000000000000000D+00,   7.785756300830923D-06,   0.000000000000000D+00, &
   4.237536280993383D-06,   0.000000000000000D+00,   4.348225593037649D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.856356943952926D-10,   0.000000000000000D+00,  -1.159456701954639D-08, &
   0.000000000000000D+00,   4.278174142633099D-07,   0.000000000000000D+00, &
  -3.006448165904639D-06,   0.000000000000000D+00,   7.064629331992151D-06, &
   0.000000000000000D+00,   2.405150373985149D-06,   0.000000000000000D+00, &
  -2.246729236474696D-07,   0.000000000000000D+00,  -3.001184174508699D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.326887501488398D-10, &
   0.000000000000000D+00,  -6.334478606887480D-08,   0.000000000000000D+00, &
  -6.270922487464588D-08,   0.000000000000000D+00,  -4.502467716077008D-07, &
   0.000000000000000D+00,  -6.170391584454801D-06,   0.000000000000000D+00, &
   1.179800188833643D-07,   0.000000000000000D+00,  -3.456920364094678D-06, &
   0.000000000000000D+00,  -2.358885739775644D-06,   0.000000000000000D+00 /)
lem(:, 97) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.103103630798297D-03,   8.784104611578821D-03,  -6.126359154706293D-03, &
  -1.101763843937954D-02,   6.597036054967279D-03,   4.212879864558637D-03, &
  -4.055654366366080D-03,   1.022171151680298D-03,   4.192272114854914D-04, &
  -8.000102661386891D-04,   6.265857722558422D-04,  -1.464433677900595D-04, &
  -1.887432033776927D-04,   1.708065431632433D-04,  -2.241394088017902D-06, &
  -7.064566653392854D-05,   2.497401393246590D-05,   2.517657721776176D-05, &
  -1.925978925182464D-05,  -7.895142018644320D-06,   1.217337628692406D-05, &
   1.653643257671352D-06,   5.103103630798293D-03,   8.784104611578832D-03, &
  -6.126359154706306D-03,  -1.101763843937954D-02,   6.597036054967271D-03, &
   4.212879864558643D-03,  -4.055654366366095D-03,   1.022171151680300D-03, &
   4.192272114854969D-04,  -8.000102661386873D-04,   6.265857722558368D-04, &
  -1.464433677900565D-04,  -1.887432033776864D-04,   1.708065431632339D-04, &
  -2.241394088027177D-06,  -7.064566653392559D-05,   2.497401393247647D-05, &
   2.517657721776076D-05,  -1.925978925183420D-05,  -7.895142018640850D-06, &
   1.217337628693256D-05,   1.653643257666515D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.902890810954745D-03,   0.000000000000000D+00, &
   2.311952212616930D-02,   0.000000000000000D+00,  -2.527096561170268D-02, &
   0.000000000000000D+00,   1.435857613150127D-02,   0.000000000000000D+00, &
  -4.099464071535681D-03,   0.000000000000000D+00,   6.394213383542336D-04, &
   0.000000000000000D+00,  -1.554640041560773D-04,   0.000000000000000D+00, &
   4.925915628330066D-05,   0.000000000000000D+00,  -1.866617047632839D-05, &
   0.000000000000000D+00,   8.052379955697771D-06,   0.000000000000000D+00, &
  -1.736111111111113D-04,  -7.683288933288939D-05,   0.000000000000000D+00, &
   1.094391719391717D-04,   0.000000000000000D+00,   5.532661782661787D-04, &
   5.035473785473781D-05,   0.000000000000000D+00,  -1.307720057720058D-04, &
   0.000000000000000D+00,  -7.388444888444898D-05,   0.000000000000000D+00, &
  -1.514978077478074D-04,   0.000000000000000D+00,  -1.167896773365523D-03, &
  -6.726661138425948D-06,   0.000000000000000D+00,   9.963654265124855D-05, &
   0.000000000000000D+00,   4.224795068085509D-04,   0.000000000000000D+00, &
   2.756638786050553D-05,   0.000000000000000D+00,   1.318636428930542D-05, &
   0.000000000000000D+00,   1.910154541128803D-04,   0.000000000000000D+00, &
   1.513424152440697D-03,  -5.405026689856635D-06,   0.000000000000000D+00, &
  -3.078966169136447D-05,   0.000000000000000D+00,  -2.350902357553866D-04, &
   0.000000000000000D+00,  -5.864368380694809D-04,   0.000000000000000D+00, &
  -3.612121452678474D-06,   0.000000000000000D+00,   9.858025531400134D-06, &
   0.000000000000000D+00,   4.744725169696157D-05,   0.000000000000000D+00, &
  -2.052115357552665D-04,   0.000000000000000D+00,  -1.270407472278513D-03, &
   1.724174518292107D-06,   0.000000000000000D+00,   8.446361000541473D-07, &
   0.000000000000000D+00,   3.520774702808369D-05,   0.000000000000000D+00, &
   2.992308187741626D-04,   0.000000000000000D+00,   4.399621597892515D-04, &
   0.000000000000000D+00,  -2.657945769400758D-07,   0.000000000000000D+00, &
  -2.271872434411242D-06,   0.000000000000000D+00,  -1.884043915535782D-05, &
   0.000000000000000D+00,  -8.153067424062014D-05,   0.000000000000000D+00, &
   2.293990876637392D-04,   0.000000000000000D+00,   7.049287484477597D-04, &
   1.412434949915106D-08,   0.000000000000000D+00,   1.113722467538646D-06, &
   0.000000000000000D+00,  -1.078751570384911D-06,   0.000000000000000D+00, &
  -2.590170584159511D-05,   0.000000000000000D+00,  -2.549593734695844D-04, &
   0.000000000000000D+00,  -1.588151601014252D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.983829088755882D-07, &
   0.000000000000000D+00,   2.487299405941784D-06,   0.000000000000000D+00, &
   1.630466609087127D-05,   0.000000000000000D+00,   4.029200023824774D-05, &
   0.000000000000000D+00,  -2.088502572758082D-04,   0.000000000000000D+00, &
  -2.667220181258268D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.168468913117834D-08,   0.000000000000000D+00,   2.911877643713577D-07, &
   0.000000000000000D+00,  -1.253315689877574D-06,   0.000000000000000D+00, &
   3.886825371013048D-05,   0.000000000000000D+00,   1.275351424224953D-04, &
   0.000000000000000D+00,  -4.227771985403897D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.164897291448852D-09, &
   0.000000000000000D+00,  -6.191644981564046D-07,   0.000000000000000D+00, &
   2.049868301950496D-06,   0.000000000000000D+00,  -1.716762869855227D-05, &
   0.000000000000000D+00,   3.411266800879886D-05,   0.000000000000000D+00, &
   1.220068177383314D-04,   0.000000000000000D+00,   7.780954893991974D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.564137681514488D-10, &
   0.000000000000000D+00,   9.405970472821142D-08,   0.000000000000000D+00, &
  -2.239330691910922D-06,   0.000000000000000D+00,   3.071999269301056D-06, &
   0.000000000000000D+00,  -4.589912767437336D-05,   0.000000000000000D+00, &
  -1.009387104759381D-05,   0.000000000000000D+00,   3.039071795175311D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.393565733521179D-08, &
   0.000000000000000D+00,   1.310920565277620D-07,   0.000000000000000D+00, &
  -2.477696909299154D-06,   0.000000000000000D+00,   5.555474358055264D-06, &
   0.000000000000000D+00,  -5.930753807795969D-05,   0.000000000000000D+00, &
  -3.936978729459198D-05,   0.000000000000000D+00,  -2.324092876422877D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.229661801831149D-08,   0.000000000000000D+00, &
   5.320248995214150D-07,   0.000000000000000D+00,  -1.467989393376351D-07, &
   0.000000000000000D+00,   4.115078843951780D-06,   0.000000000000000D+00, &
   1.674707890967960D-05,   0.000000000000000D+00,  -3.154761986423964D-05, &
   0.000000000000000D+00,  -1.637613565599714D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.100678448615469D-09,   0.000000000000000D+00, &
  -1.636293210174190D-07,   0.000000000000000D+00,   1.671580795124329D-06, &
   0.000000000000000D+00,  -1.935863218990556D-06,   0.000000000000000D+00, &
   1.576780341270431D-05,   0.000000000000000D+00,   2.906822882765880D-05, &
   0.000000000000000D+00,   4.397905789888834D-06,   0.000000000000000D+00, &
   8.470594770490771D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.530947225665813D-10, &
   0.000000000000000D+00,  -3.897077157907449D-08,   0.000000000000000D+00, &
  -4.841519236615756D-07,   0.000000000000000D+00,   2.389274746397436D-06, &
   0.000000000000000D+00,  -8.116074532411982D-06,   0.000000000000000D+00, &
   1.652925180149349D-05,   0.000000000000000D+00,   2.154180272886025D-05, &
   0.000000000000000D+00,   6.948611436984311D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.153383858567032D-08,   0.000000000000000D+00,  -3.727577869998249D-07, &
   0.000000000000000D+00,  -8.564397893859128D-07,   0.000000000000000D+00, &
   4.656927441083872D-07,   0.000000000000000D+00,  -1.750590336714500D-05, &
   0.000000000000000D+00,   1.698922258805442D-06,   0.000000000000000D+00, &
   1.831034309843729D-06,   0.000000000000000D+00,  -3.580799979022173D-06 /)
lem(:, 98) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.903442927248235D-03,  -2.091007025795408D-02,  -6.318715049238302D-03, &
   4.109646427107926D-02,   3.054131454160442D-03,  -3.423448508066545D-02, &
  -7.073182359475494D-03,   1.200234768557324D-02,   7.764080617745871D-03, &
   8.848358773754543D-04,  -2.808457995233465D-03,  -2.176059157002534D-03, &
  -4.930530267545864D-05,   8.516067379767855D-04,   3.385508106220509D-04, &
  -2.712189341592369D-04,  -2.404914562692894D-04,   6.938205284926151D-05, &
   1.407715049394029D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.903442927248251D-03,  -2.091007025795407D-02, &
  -6.318715049238282D-03,   4.109646427107928D-02,   3.054131454160460D-03, &
  -3.423448508066541D-02,  -7.073182359475558D-03,   1.200234768557322D-02, &
   7.764080617745918D-03,   8.848358773754506D-04,  -2.808457995233475D-03, &
  -2.176059157002472D-03,  -4.930530267548316D-05,   8.516067379767322D-04, &
   3.385508106220747D-04,  -2.712189341591949D-04,  -2.404914562692998D-04, &
   6.938205284923403D-05,   1.407715049393947D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.631701631701644D-04,   0.000000000000000D+00,   1.942501942501363D-06, &
  -1.126651126651127D-03,   0.000000000000000D+00,  -2.081252081252020D-05, &
   0.000000000000000D+00,   2.703505644682109D-03,   0.000000000000000D+00, &
   2.074722662957931D-05,   0.000000000000000D+00,   8.569861511038334D-06, &
  -1.743681155445858D-03,   0.000000000000000D+00,   2.984760337701513D-04, &
   0.000000000000000D+00,   4.706622353680867D-06,   0.000000000000000D+00, &
  -3.622074519907338D-03,   0.000000000000000D+00,  -1.006337144108041D-03, &
   0.000000000000000D+00,  -6.810870587960346D-05,   0.000000000000000D+00, &
  -5.405026689856635D-06,   7.543931816377632D-03,   0.000000000000000D+00, &
   8.246990367733396D-04,   0.000000000000000D+00,   2.409011696937379D-05, &
   0.000000000000000D+00,   2.131189244192360D-05,   0.000000000000000D+00, &
  -4.023427552839316D-03,   0.000000000000000D+00,   1.215507059160311D-03, &
   0.000000000000000D+00,   3.404361376497595D-04,   0.000000000000000D+00, &
   1.958932879985517D-05,   0.000000000000000D+00,  -1.274323908805326D-05, &
  -2.252236927159525D-03,   0.000000000000000D+00,  -2.782983270599366D-03, &
   0.000000000000000D+00,  -4.711963303294563D-04,   0.000000000000000D+00, &
  -8.703564237619982D-05,   0.000000000000000D+00,  -1.547447972564078D-05, &
   0.000000000000000D+00,   3.915010820381667D-03,   0.000000000000000D+00, &
   1.355841553041712D-03,   0.000000000000000D+00,  -3.264149776129854D-04, &
   0.000000000000000D+00,  -3.613879408905631D-05,   0.000000000000000D+00, &
   2.217485809553217D-05,   0.000000000000000D+00,   1.082051167258842D-05, &
  -1.882583697094399D-03,   0.000000000000000D+00,   1.056158137191922D-03, &
   0.000000000000000D+00,   1.031603284766565D-03,   0.000000000000000D+00, &
   2.070901150942204D-04,   0.000000000000000D+00,   3.814144883455198D-05, &
   0.000000000000000D+00,  -2.467296476889479D-06,   0.000000000000000D+00, &
   3.249043324423612D-04,   0.000000000000000D+00,  -1.464014918988669D-03, &
   0.000000000000000D+00,  -3.682812304430285D-04,   0.000000000000000D+00, &
   2.236436425930256D-05,   0.000000000000000D+00,  -3.587181761923500D-05, &
   0.000000000000000D+00,  -1.918436423495765D-05,   0.000000000000000D+00, &
   5.315045843375562D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.137647384248552D-04,   0.000000000000000D+00,  -6.008009877978910D-04, &
   0.000000000000000D+00,  -2.986215309122160D-04,   0.000000000000000D+00, &
  -6.275532820050081D-05,   0.000000000000000D+00,  -1.046918234050199D-06, &
   0.000000000000000D+00,   8.350981880209757D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.831559191769092D-05, &
   0.000000000000000D+00,   5.054700695657304D-04,   0.000000000000000D+00, &
   2.667317800579311D-05,   0.000000000000000D+00,   2.091462234987625D-05, &
   0.000000000000000D+00,   2.467299438917217D-05,   0.000000000000000D+00, &
   3.696212897747689D-06,   0.000000000000000D+00,  -8.076369160353803D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.583068925510649D-05, &
   0.000000000000000D+00,   2.376616120054049D-05,   0.000000000000000D+00, &
   2.665624746322010D-04,   0.000000000000000D+00,   7.827377842689436D-05, &
   0.000000000000000D+00,   1.416440472446688D-05,   0.000000000000000D+00, &
  -6.464885629571742D-06,   0.000000000000000D+00,  -6.690222754675954D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.720034203744618D-06,   0.000000000000000D+00,  -1.305422922571003D-04, &
   0.000000000000000D+00,  -6.199447471071650D-05,   0.000000000000000D+00, &
   2.340479427376214D-05,   0.000000000000000D+00,  -1.225562914440495D-05, &
   0.000000000000000D+00,  -9.687344821097332D-06,   0.000000000000000D+00, &
   1.323652162822481D-06,   0.000000000000000D+00,   1.731666770729618D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.944798891526730D-05,   0.000000000000000D+00, &
  -1.427330426917325D-04,   0.000000000000000D+00,  -8.246666567926761D-05, &
   0.000000000000000D+00,  -2.526503892269882D-05,   0.000000000000000D+00, &
  -3.193023752324895D-06,   0.000000000000000D+00,   5.810301211193466D-06, &
   0.000000000000000D+00,   3.103811697561687D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.416411344468132D-05,   0.000000000000000D+00, &
   5.043168749189936D-05,   0.000000000000000D+00,  -4.946722120287545D-05, &
   0.000000000000000D+00,  -1.386292400708617D-05,   0.000000000000000D+00, &
   5.911837300413248D-06,   0.000000000000000D+00,   3.518816999173640D-06, &
   0.000000000000000D+00,   4.693318291949331D-07,   0.000000000000000D+00, &
   1.051423552033578D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.198031093089650D-06, &
   0.000000000000000D+00,   4.029923206278637D-05,   0.000000000000000D+00, &
   6.994323507771172D-05,   0.000000000000000D+00,   1.892391684200440D-05, &
   0.000000000000000D+00,   1.030027891395081D-05,   0.000000000000000D+00, &
   3.290622929385451D-07,   0.000000000000000D+00,  -3.757195224480349D-06, &
   0.000000000000000D+00,   2.884840989088797D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.075942785105441D-07,   0.000000000000000D+00, &
  -1.971433991549795D-05,   0.000000000000000D+00,   3.377292092407631D-05, &
   0.000000000000000D+00,   3.652069281774567D-05,   0.000000000000000D+00, &
   6.701809076796206D-06,   0.000000000000000D+00,  -2.892847338709675D-06, &
   0.000000000000000D+00,  -2.549670610034272D-06,   0.000000000000000D+00, &
  -1.043801779275257D-06,   0.000000000000000D+00,  -8.827600723656708D-07 /)
lem(:, 99) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.404052685167515D-04,   2.232368570752573D-03, &
  -7.630301446670025D-03,   5.539574168556037D-03,   1.118000138857520D-02, &
  -1.563841994385116D-02,  -5.306688304468545D-03,   1.387385484819804D-02, &
   1.001468253321158D-03,  -6.148359823323181D-03,  -1.059408947841414D-03, &
   1.731211019604667D-03,   1.090264235300431D-03,  -3.421869809606867D-04, &
  -5.950597386561874D-04,  -2.242355527754033D-05,   2.745340445831284D-04, &
   7.676638015131917D-05,  -1.228982849643100D-04,  -6.480444968275841D-05, &
   5.494960158689149D-05,   0.000000000000000D+00,   1.404052685167485D-04, &
  -2.232368570752589D-03,   7.630301446670042D-03,  -5.539574168556034D-03, &
  -1.118000138857519D-02,   1.563841994385112D-02,   5.306688304468535D-03, &
  -1.387385484819804D-02,  -1.001468253321166D-03,   6.148359823323195D-03, &
   1.059408947841422D-03,  -1.731211019604685D-03,  -1.090264235300427D-03, &
   3.421869809606997D-04,   5.950597386561692D-04,   2.242355527754029D-05, &
  -2.745340445831147D-04,  -7.676638015133010D-05,   1.228982849643127D-04, &
   6.480444968275613D-05,  -5.494960158689059D-05,   0.000000000000000D+00, &
   7.020263425836853D-05,   0.000000000000000D+00,   2.095654463057024D-04, &
   0.000000000000000D+00,   8.323407823536783D-06,   0.000000000000000D+00, &
  -1.688329100145935D-04,   0.000000000000000D+00,   3.358546590929714D-05, &
   0.000000000000000D+00,   6.492199680365299D-05,   0.000000000000000D+00, &
  -3.176673035229230D-05,   0.000000000000000D+00,   2.889580148069186D-07, &
   0.000000000000000D+00,   3.388929242816626D-06,   0.000000000000000D+00, &
  -2.532350326804549D-06,   0.000000000000000D+00,   1.519903276782281D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.550005550005634D-06, &
   0.000000000000000D+00,   5.106005106005116D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.542901542901546D-04,   0.000000000000000D+00, &
  -1.526251526251515D-05,   0.000000000000000D+00,   2.160366866249219D-04, &
   0.000000000000000D+00,  -1.398927869516111D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.743533155297875D-05,   0.000000000000000D+00, &
   1.177417353887943D-04,   0.000000000000000D+00,   1.044706927059866D-05, &
   0.000000000000000D+00,  -3.685283871042389D-04,   0.000000000000000D+00, &
  -2.194425801236947D-04,   0.000000000000000D+00,  -2.381948976376220D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.179991319310209D-04, &
   0.000000000000000D+00,   8.793665914408967D-05,   0.000000000000000D+00, &
  -2.153848903074916D-05,   0.000000000000000D+00,   1.794408721653307D-05, &
   0.000000000000000D+00,  -4.048342438435312D-04,   0.000000000000000D+00, &
   3.072134264084723D-04,   0.000000000000000D+00,   1.207512817419939D-04, &
   0.000000000000000D+00,   1.538279402056490D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.961670596345511D-05,   0.000000000000000D+00, &
  -5.631253851068088D-04,   0.000000000000000D+00,  -1.001556922609555D-04, &
   0.000000000000000D+00,  -2.562904072191998D-05,   0.000000000000000D+00, &
  -1.746572409788353D-05,   0.000000000000000D+00,   3.055588464122591D-04, &
   0.000000000000000D+00,   3.405875693532173D-04,   0.000000000000000D+00, &
  -1.220467063783794D-04,   0.000000000000000D+00,  -2.787452435790028D-05, &
   0.000000000000000D+00,   2.629171784677269D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.580584489320532D-04,   0.000000000000000D+00, &
   1.085225927061972D-04,   0.000000000000000D+00,   2.840021474763704D-04, &
   0.000000000000000D+00,   6.041935242534246D-05,   0.000000000000000D+00, &
   1.761781182484670D-05,   0.000000000000000D+00,  -4.686049878717588D-06, &
   0.000000000000000D+00,   1.123564370300139D-04,   0.000000000000000D+00, &
  -3.128506527348902D-04,   0.000000000000000D+00,  -1.379835297590044D-04, &
   0.000000000000000D+00,   1.760212231843670D-05,   0.000000000000000D+00, &
  -6.006163673750182D-06,   0.000000000000000D+00,  -5.176496144903323D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.629434100695414D-05, &
   0.000000000000000D+00,   2.089762737225385D-04,   0.000000000000000D+00, &
  -1.227833203873044D-04,   0.000000000000000D+00,  -9.034977507349289D-05, &
   0.000000000000000D+00,  -2.100464114825062D-05,   0.000000000000000D+00, &
  -6.765498916717639D-07,   0.000000000000000D+00,   1.068628998316513D-05, &
   0.000000000000000D+00,   2.632326767473076D-06,   0.000000000000000D+00, &
  -5.269732137412867D-05,   0.000000000000000D+00,   1.786150763132858D-04, &
   0.000000000000000D+00,   2.109267786848883D-05,   0.000000000000000D+00, &
   4.000345540887668D-06,   0.000000000000000D+00,   5.712148150407415D-06, &
   0.000000000000000D+00,   3.573093457725186D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.736182380010297D-06,   0.000000000000000D+00,  -5.612079848181596D-05, &
   0.000000000000000D+00,   7.578841119694538D-05,   0.000000000000000D+00, &
   2.234889401104558D-05,   0.000000000000000D+00,   4.496458501928976D-06, &
   0.000000000000000D+00,  -3.973543926241999D-06,   0.000000000000000D+00, &
  -3.571054596445402D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.946924640780297D-06,   0.000000000000000D+00, &
  -3.125147532433046D-05,   0.000000000000000D+00,  -4.819874600058017D-05, &
   0.000000000000000D+00,   4.143645459053561D-06,   0.000000000000000D+00, &
  -2.519131021512736D-06,   0.000000000000000D+00,  -3.251459586304588D-06, &
   0.000000000000000D+00,  -2.483004191988487D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.042322492004714D-06,   0.000000000000000D+00,   2.724540117028070D-05, &
   0.000000000000000D+00,  -2.966676833884113D-05,   0.000000000000000D+00, &
  -2.696631946073357D-05,   0.000000000000000D+00,  -6.556554652297497D-06, &
   0.000000000000000D+00,   7.002895869787296D-08,   0.000000000000000D+00, &
   1.201191386738294D-06,   0.000000000000000D+00,  -4.466358836474937D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.464030460218923D-07,   0.000000000000000D+00,  -3.159506541616414D-06, &
   0.000000000000000D+00,   3.719281005473127D-05,   0.000000000000000D+00, &
  -5.537787544270006D-06,   0.000000000000000D+00,  -3.154373621802900D-06, &
   0.000000000000000D+00,   1.167009984495274D-06,   0.000000000000000D+00, &
   2.855845713422518D-06,   0.000000000000000D+00,   4.942296202061538D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.463893422127647D-06,   0.000000000000000D+00,  -1.133125141114453D-06, &
   0.000000000000000D+00,   2.811488472489454D-05,   0.000000000000000D+00, &
   6.296355646194702D-06,   0.000000000000000D+00,   2.180552495272101D-06, &
   0.000000000000000D+00,  -1.447762632438351D-08,   0.000000000000000D+00, &
   8.267082852043329D-07,   0.000000000000000D+00,   8.082890010379816D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.560788051842864D-06, &
   0.000000000000000D+00,  -1.256924046105279D-05,   0.000000000000000D+00, &
  -2.862080897808897D-06,   0.000000000000000D+00,   9.275081226669544D-06, &
   0.000000000000000D+00,   1.559715892822081D-06,   0.000000000000000D+00, &
  -1.443473205087069D-06,   0.000000000000000D+00,  -1.016575113254742D-06, &
   0.000000000000000D+00,   5.239968771462760D-07,   0.000000000000000D+00 /)
lem(:,100) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.069234466906664D-04,  -4.855682202870649D-04,   8.961488871040474D-03, &
  -1.835463656126919D-03,  -1.899434424438432D-02,   2.886835280130154D-04, &
   2.422189348260192D-02,   1.715530798596778D-04,  -1.564914895565768D-02, &
  -1.237201072694807D-04,   4.884340118867661D-03,   1.276850370036366D-04, &
  -9.318738931911761D-04,  -1.193041055459470D-05,   3.489465019644794D-04, &
   6.078600121268727D-06,  -1.709498504390948D-04,  -1.514443411735238D-05, &
   9.015466256072146D-05,   1.624356032866305D-05,  -4.950032251078008D-05, &
  -1.371092076632383D-05,  -7.069234466906702D-04,  -4.855682202870720D-04, &
   8.961488871040469D-03,  -1.835463656126919D-03,  -1.899434424438431D-02, &
   2.886835280130162D-04,   2.422189348260192D-02,   1.715530798596779D-04, &
  -1.564914895565768D-02,  -1.237201072694748D-04,   4.884340118867658D-03, &
   1.276850370036451D-04,  -9.318738931911709D-04,  -1.193041055460304D-05, &
   3.489465019644810D-04,   6.078600121277154D-06,  -1.709498504391083D-04, &
  -1.514443411735636D-05,   9.015466256073650D-05,   1.624356032867672D-05, &
  -4.950032251080092D-05,  -1.371092076634007D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.142142059454799D-03,   0.000000000000000D+00, &
   3.644456689839423D-05,   0.000000000000000D+00,   5.595328093453609D-04, &
   0.000000000000000D+00,  -1.188629508417011D-04,   0.000000000000000D+00, &
  -1.729044281391865D-04,   0.000000000000000D+00,   6.725270813565580D-05, &
   0.000000000000000D+00,   9.133848655874142D-06,   0.000000000000000D+00, &
  -7.807132532268631D-06,   0.000000000000000D+00,   3.810237395932537D-06, &
   0.000000000000000D+00,  -1.737355246759321D-06,   0.000000000000000D+00, &
   3.561253561253551D-05,   1.956376956376959D-04,   0.000000000000000D+00, &
  -7.214082214082214D-04,   0.000000000000000D+00,   2.590002590002636D-06, &
   6.026381026381032D-04,   0.000000000000000D+00,  -9.546009546009546D-05, &
   0.000000000000000D+00,   9.538663950428665D-04,   0.000000000000000D+00, &
   3.204747322394381D-04,   0.000000000000000D+00,   8.141368435486024D-06, &
  -2.184770567123509D-03,   0.000000000000000D+00,  -2.636781208209780D-04, &
   0.000000000000000D+00,   3.878882555353144D-05,   0.000000000000000D+00, &
   1.080439741430454D-03,   0.000000000000000D+00,  -4.402434510793638D-04, &
   0.000000000000000D+00,  -1.729900646309315D-04,   0.000000000000000D+00, &
  -3.078966169136447D-05,   8.246990367733396D-04,   0.000000000000000D+00, &
   9.963551527019023D-04,   0.000000000000000D+00,   1.637828484356570D-04, &
   0.000000000000000D+00,   1.220050804029137D-05,   0.000000000000000D+00, &
  -1.156895506740708D-03,   0.000000000000000D+00,  -4.936165772079079D-04, &
   0.000000000000000D+00,   2.256524363722507D-04,   0.000000000000000D+00, &
   7.026355710566233D-05,   0.000000000000000D+00,   1.590066275994108D-05, &
   4.058634863588417D-04,   0.000000000000000D+00,  -3.607623886261663D-04, &
   0.000000000000000D+00,  -5.431386970755107D-04,   0.000000000000000D+00, &
  -1.047048067171906D-04,   0.000000000000000D+00,  -2.268310048761871D-05, &
   0.000000000000000D+00,   2.439454922965536D-05,   0.000000000000000D+00, &
   5.016419594424716D-04,   0.000000000000000D+00,   2.518990885948082D-04, &
   0.000000000000000D+00,  -7.165726075572628D-05,   0.000000000000000D+00, &
  -1.004234707642597D-05,   0.000000000000000D+00,   1.158931621214145D-05, &
  -3.716501562779689D-05,   0.000000000000000D+00,  -1.761046868059920D-04, &
   0.000000000000000D+00,   2.170597393288872D-04,   0.000000000000000D+00, &
   2.332674097922998D-04,   0.000000000000000D+00,   5.075309874150351D-05, &
   0.000000000000000D+00,   6.989365770725118D-06,   0.000000000000000D+00, &
   3.834098328646736D-06,   0.000000000000000D+00,   4.082386983373262D-06, &
   0.000000000000000D+00,  -2.603366481411977D-04,   0.000000000000000D+00, &
  -8.150452765727382D-05,   0.000000000000000D+00,   5.491442757719766D-06, &
   0.000000000000000D+00,  -9.217905255340852D-06,   0.000000000000000D+00, &
  -1.336217774861020D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.204072546110388D-06,   0.000000000000000D+00,   7.261886047725315D-05, &
   0.000000000000000D+00,  -1.319057334545623D-04,   0.000000000000000D+00, &
  -7.368794314247724D-05,   0.000000000000000D+00,  -1.432694855561529D-05, &
   0.000000000000000D+00,   1.278278387418128D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.294755184242296D-06, &
   0.000000000000000D+00,   4.887471836494981D-07,   0.000000000000000D+00, &
   1.026789508794151D-04,   0.000000000000000D+00,   6.356794570372768D-06, &
   0.000000000000000D+00,   4.789186520776790D-06,   0.000000000000000D+00, &
   8.394927555132394D-06,   0.000000000000000D+00,   2.468325222231512D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.580037049144741D-06, &
   0.000000000000000D+00,   5.858518696720253D-06,   0.000000000000000D+00, &
   3.329295945226669D-06,   0.000000000000000D+00,   6.367846110246040D-05, &
   0.000000000000000D+00,   1.985815004396117D-05,   0.000000000000000D+00, &
   2.017819551046838D-06,   0.000000000000000D+00,  -1.096180295399136D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.113994086551253D-07,   0.000000000000000D+00,  -2.576624927117547D-06, &
   0.000000000000000D+00,  -1.510144025235158D-05,   0.000000000000000D+00, &
  -1.412964325396883D-05,   0.000000000000000D+00,   6.623072633241856D-06, &
   0.000000000000000D+00,  -3.827105495344458D-06,   0.000000000000000D+00, &
  -3.603131161999463D-06,   0.000000000000000D+00,   1.774814473178722D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.976627526565409D-06,   0.000000000000000D+00, &
   8.941707451979671D-08,   0.000000000000000D+00,  -2.898404862351587D-05, &
   0.000000000000000D+00,  -2.089228083703509D-05,   0.000000000000000D+00, &
  -5.791558564189709D-06,   0.000000000000000D+00,  -9.888181133304385D-07, &
   0.000000000000000D+00,   1.107739984082847D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.062502366981426D-06,   0.000000000000000D+00, &
  -2.511222297283430D-06,   0.000000000000000D+00,   8.412664397637088D-06, &
   0.000000000000000D+00,  -1.254475633139206D-05,   0.000000000000000D+00, &
  -3.930314540584695D-06,   0.000000000000000D+00,   2.791896892528657D-06, &
   0.000000000000000D+00,   7.697567054255956D-07,   0.000000000000000D+00, &
  -1.344819460053798D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.726507641286646D-07, &
   0.000000000000000D+00,   2.435075816722526D-06,   0.000000000000000D+00, &
   4.293690818018080D-06,   0.000000000000000D+00,   1.611803705885071D-05, &
   0.000000000000000D+00,   4.546357268700671D-06,   0.000000000000000D+00, &
   2.638005796431304D-06,   0.000000000000000D+00,   6.122037742607822D-07, &
   0.000000000000000D+00,  -1.075686195315176D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.432051988252860D-09,   0.000000000000000D+00, &
   4.023621121322074D-07,   0.000000000000000D+00,   6.842699758503759D-08, &
   0.000000000000000D+00,   7.075280154905624D-06,   0.000000000000000D+00, &
   9.825117296179248D-06,   0.000000000000000D+00,   1.265520868577612D-06, &
   0.000000000000000D+00,  -1.066603657054818D-06,   0.000000000000000D+00, &
  -3.748853617627662D-07,   0.000000000000000D+00,   6.819448853216357D-07 /)
lem(:,101) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.702801418947327D-03,  -8.241593269929562D-03, &
  -3.172915043785789D-03,   1.680192153391626D-02,   2.203919485846590D-04, &
  -1.826369568554724D-02,   4.225759654467641D-03,   1.035548522711471D-02, &
  -4.102661263026800D-03,  -3.031800949438037D-03,   1.628377049699181D-03, &
   5.904288973908853D-04,  -4.377507569419969D-04,  -1.576285054363642D-04, &
   1.615002397976606D-04,   4.766528156285560D-05,  -7.433618083414769D-05, &
  -1.628452607435528D-05,   3.913802129652999D-05,   6.309281844149448D-06, &
  -2.242531058536785D-05,   0.000000000000000D+00,  -2.702801418947335D-03, &
   8.241593269929562D-03,   3.172915043785794D-03,  -1.680192153391634D-02, &
  -2.203919485846618D-04,   1.826369568554727D-02,  -4.225759654467630D-03, &
  -1.035548522711471D-02,   4.102661263026791D-03,   3.031800949438042D-03, &
  -1.628377049699172D-03,  -5.904288973908896D-04,   4.377507569419865D-04, &
   1.576285054363748D-04,  -1.615002397976573D-04,  -4.766528156286709D-05, &
   7.433618083414940D-05,   1.628452607436014D-05,  -3.913802129653783D-05, &
  -6.309281844144173D-06,   2.242531058537557D-05,   0.000000000000000D+00, &
  -1.351400709473664D-03,   0.000000000000000D+00,   1.730680932974064D-03, &
   0.000000000000000D+00,  -1.072173833497646D-03,   0.000000000000000D+00, &
  -1.896481908069914D-04,   0.000000000000000D+00,   4.351816048549205D-04, &
   0.000000000000000D+00,  -4.300028359722536D-05,   0.000000000000000D+00, &
  -7.048626582753609D-05,   0.000000000000000D+00,   2.469061248870142D-05, &
   0.000000000000000D+00,  -8.088876677930138D-06,   0.000000000000000D+00, &
   2.693348710580975D-06,   0.000000000000000D+00,  -8.836717022464886D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.068376068376067D-04, &
   0.000000000000000D+00,  -1.734376734376737D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.115899973042827D-05,   0.000000000000000D+00, &
  -1.022786737072452D-04,   0.000000000000000D+00,   2.528867024665342D-04, &
   0.000000000000000D+00,   1.975615420993575D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.622219258073598D-04,   0.000000000000000D+00, &
  -9.896732585808214D-05,   0.000000000000000D+00,   3.799305269893505D-05, &
   0.000000000000000D+00,   2.771609195757808D-04,   0.000000000000000D+00, &
  -2.710678186131526D-04,   0.000000000000000D+00,  -1.428943863154390D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.793665914408967D-05, &
   0.000000000000000D+00,   5.145302303639325D-04,   0.000000000000000D+00, &
   1.038404565604920D-04,   0.000000000000000D+00,   2.877535046652703D-05, &
   0.000000000000000D+00,  -2.486634452578732D-04,   0.000000000000000D+00, &
  -2.968786582231954D-04,   0.000000000000000D+00,   1.689965643968299D-04, &
   0.000000000000000D+00,   6.053726545212605D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.622818186285679D-04,   0.000000000000000D+00, &
  -1.171030306368653D-04,   0.000000000000000D+00,  -3.574693080668916D-04, &
   0.000000000000000D+00,  -7.904220607393978D-05,   0.000000000000000D+00, &
  -4.529537276126860D-05,   0.000000000000000D+00,  -3.867538384777767D-05, &
   0.000000000000000D+00,   2.758703775952783D-04,   0.000000000000000D+00, &
   1.848015859636476D-04,   0.000000000000000D+00,  -5.800408087896305D-05, &
   0.000000000000000D+00,  -5.492046836303626D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.874416562377038D-06,   0.000000000000000D+00, &
  -1.586619829798107D-04,   0.000000000000000D+00,   1.182811505571514D-04, &
   0.000000000000000D+00,   1.643013956806752D-04,   0.000000000000000D+00, &
   3.959205235848073D-05,   0.000000000000000D+00,   1.817308552842437D-05, &
   0.000000000000000D+00,   7.092551272694835D-06,   0.000000000000000D+00, &
   2.531729114810852D-05,   0.000000000000000D+00,  -1.945342454243384D-04, &
   0.000000000000000D+00,  -6.256175165601050D-05,   0.000000000000000D+00, &
   4.965837483725577D-06,   0.000000000000000D+00,  -6.849411414450057D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.569600438896404D-06, &
   0.000000000000000D+00,   1.088592131416597D-05,   0.000000000000000D+00, &
   7.706660646154523D-05,   0.000000000000000D+00,  -9.038747891554463D-05, &
   0.000000000000000D+00,  -5.072711168937744D-05,   0.000000000000000D+00, &
  -1.233199435293842D-05,   0.000000000000000D+00,   4.768858528624214D-06, &
   0.000000000000000D+00,   1.107881636140089D-07,   0.000000000000000D+00, &
  -3.116930986477132D-06,   0.000000000000000D+00,   9.088481388156501D-06, &
   0.000000000000000D+00,   8.692609111995987D-05,   0.000000000000000D+00, &
   5.075232613374902D-06,   0.000000000000000D+00,   4.374296691829713D-06, &
   0.000000000000000D+00,   3.793706293527953D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.549224911338652D-06,   0.000000000000000D+00,  -1.109929737845447D-05, &
   0.000000000000000D+00,  -9.103785268140318D-08,   0.000000000000000D+00, &
   4.634350746951065D-05,   0.000000000000000D+00,   1.279651692229301D-05, &
   0.000000000000000D+00,   1.428478203438640D-06,   0.000000000000000D+00, &
  -7.274897411616231D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.457741775958603D-07,   0.000000000000000D+00, &
  -2.284239812893494D-06,   0.000000000000000D+00,  -2.541947224598155D-05, &
   0.000000000000000D+00,  -1.472246244238813D-05,   0.000000000000000D+00, &
   3.678377981877853D-06,   0.000000000000000D+00,  -2.552650301334421D-06, &
   0.000000000000000D+00,  -2.802102411477392D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.357719364043831D-07,   0.000000000000000D+00,   2.422596028559995D-06, &
   0.000000000000000D+00,   3.268611557104688D-06,   0.000000000000000D+00, &
  -2.567991567585530D-05,   0.000000000000000D+00,  -1.456160442621457D-05, &
   0.000000000000000D+00,  -4.317457981452423D-06,   0.000000000000000D+00, &
   8.805446165937523D-07,   0.000000000000000D+00,   3.498411843225668D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.247051606809718D-09,   0.000000000000000D+00,   2.296407458046004D-07, &
   0.000000000000000D+00,   5.346171628261917D-06,   0.000000000000000D+00, &
   1.231947143753683D-05,   0.000000000000000D+00,  -8.142608059482856D-06, &
   0.000000000000000D+00,  -1.536498703459090D-06,   0.000000000000000D+00, &
   1.744529490526621D-06,   0.000000000000000D+00,   2.191737541120207D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.057986959904506D-07,   0.000000000000000D+00,  -2.782061759905625D-07, &
   0.000000000000000D+00,   8.308536150604466D-06,   0.000000000000000D+00, &
   1.456643704368719D-05,   0.000000000000000D+00,   3.332965081997457D-06, &
   0.000000000000000D+00,   1.848477371560479D-06,   0.000000000000000D+00, &
  -8.287770520961663D-08,   0.000000000000000D+00,  -1.345369921419750D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.462352389883015D-07, &
   0.000000000000000D+00,  -1.505387763952033D-06,   0.000000000000000D+00, &
  -4.612272670857870D-06,   0.000000000000000D+00,   5.197845992934126D-06, &
   0.000000000000000D+00,   6.172096971217271D-06,   0.000000000000000D+00, &
   7.562379746401870D-07,   0.000000000000000D+00,  -1.459394061497187D-06, &
   0.000000000000000D+00,  -1.111868444326970D-06,   0.000000000000000D+00 /)
lem(:,102) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.192933316290501D-03,   5.120404636220023D-03,   7.146175203816105D-03, &
  -2.880789052973412D-03,  -1.606310555942694D-02,   3.825616594436772D-03, &
   1.442192408415348D-02,  -5.796920073778564D-03,  -5.063914813599301D-03, &
   4.259733690800692D-03,  -1.782189947327441D-04,  -1.511986613329936D-03, &
   5.809670994692810D-04,   3.363109655509828D-04,  -2.374364660667426D-04, &
  -9.119348513142482D-05,   1.014750997993446D-04,   2.642072042037829D-05, &
  -4.748071780233997D-05,  -7.217107776693434D-06,   2.414893513709071D-05, &
   1.260231956954843D-06,  -1.192933316290502D-03,   5.120404636220027D-03, &
   7.146175203816080D-03,  -2.880789052973415D-03,  -1.606310555942693D-02, &
   3.825616594436781D-03,   1.442192408415348D-02,  -5.796920073778560D-03, &
  -5.063914813599288D-03,   4.259733690800688D-03,  -1.782189947327415D-04, &
  -1.511986613329932D-03,   5.809670994692757D-04,   3.363109655509770D-04, &
  -2.374364660667372D-04,  -9.119348513142535D-05,   1.014750997993422D-04, &
   2.642072042037715D-05,  -4.748071780233323D-05,  -7.217107776687271D-06, &
   2.414893513708375D-05,   1.260231956947367D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.478341726521442D-03,   0.000000000000000D+00, &
  -3.293026937604969D-03,   0.000000000000000D+00,   2.554906061048085D-03, &
   0.000000000000000D+00,  -9.365846117365619D-04,   0.000000000000000D+00, &
  -2.182302958172142D-04,   0.000000000000000D+00,   2.236636173276267D-04, &
   0.000000000000000D+00,  -4.017908291551758D-05,   0.000000000000000D+00, &
   7.210961202288758D-06,   0.000000000000000D+00,  -8.601585373621863D-07, &
   0.000000000000000D+00,  -3.107738309768980D-07,   0.000000000000000D+00, &
   1.540126540126542D-04,  -2.317127317127319D-04,   0.000000000000000D+00, &
  -1.685814185814186D-04,   0.000000000000000D+00,  -2.117673992673995D-04, &
   5.749211106353964D-04,   0.000000000000000D+00,   1.375757179328605D-04, &
   0.000000000000000D+00,  -2.078803549391782D-04,   0.000000000000000D+00, &
   1.896256754450034D-04,   0.000000000000000D+00,   2.791096770876178D-04, &
  -3.477731418907889D-04,   0.000000000000000D+00,  -4.387358090089181D-04, &
   0.000000000000000D+00,  -1.286818874791564D-04,   0.000000000000000D+00, &
   2.861431654001316D-04,   0.000000000000000D+00,   2.057793632318176D-04, &
   0.000000000000000D+00,  -2.147929444341429D-04,   0.000000000000000D+00, &
  -2.350902357553866D-04,   2.409011696937379D-05,   0.000000000000000D+00, &
   1.637828484356570D-04,   0.000000000000000D+00,   4.299617045866866D-04, &
   0.000000000000000D+00,   1.153568935471993D-04,   0.000000000000000D+00, &
  -8.005410482190677D-05,   0.000000000000000D+00,  -2.005985500567543D-04, &
   0.000000000000000D+00,  -2.174731143292251D-04,   0.000000000000000D+00, &
   1.453882576294534D-04,   0.000000000000000D+00,   1.121915217672766D-04, &
   1.036115587199174D-05,   0.000000000000000D+00,   5.576316528328915D-05, &
   0.000000000000000D+00,  -1.211062308407945D-04,   0.000000000000000D+00, &
  -3.015309694375236D-04,   0.000000000000000D+00,  -7.580377434527954D-05, &
   0.000000000000000D+00,   9.330742052892969D-06,   0.000000000000000D+00, &
   1.807877381161667D-06,   0.000000000000000D+00,   1.923050826103895D-04, &
   0.000000000000000D+00,   1.392281142273782D-04,   0.000000000000000D+00, &
  -5.242333188754617D-05,   0.000000000000000D+00,  -5.785369273098583D-06, &
  -2.696466722580728D-06,   0.000000000000000D+00,   3.074002080069072D-06, &
   0.000000000000000D+00,  -8.662496426874210D-05,   0.000000000000000D+00, &
   1.040729849554241D-04,   0.000000000000000D+00,   1.391816890615674D-04, &
   0.000000000000000D+00,   2.910055470699221D-05,   0.000000000000000D+00, &
   1.972272802800910D-07,   0.000000000000000D+00,  -1.456755724538441D-06, &
   0.000000000000000D+00,   1.844034131684597D-05,   0.000000000000000D+00, &
  -1.398318844491571D-04,   0.000000000000000D+00,  -4.604858324563727D-05, &
   0.000000000000000D+00,   3.181466710853463D-06,   0.000000000000000D+00, &
  -2.811573121591963D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.513913481732575D-06,   0.000000000000000D+00,  -1.433293795933906D-06, &
   0.000000000000000D+00,   4.998216387650955D-05,   0.000000000000000D+00, &
  -7.625232494555319D-05,   0.000000000000000D+00,  -4.208674142199139D-05, &
   0.000000000000000D+00,  -6.860634089251168D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.633027482843968D-07, &
   0.000000000000000D+00,  -3.306155765709354D-06,   0.000000000000000D+00, &
   1.256692109788950D-06,   0.000000000000000D+00,   6.318555253025535D-05, &
   0.000000000000000D+00,   2.671281247959417D-06,   0.000000000000000D+00, &
   8.259504684384202D-06,   0.000000000000000D+00,   1.750585704443134D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.193086007103830D-07, &
   0.000000000000000D+00,   2.006367274156256D-06,   0.000000000000000D+00, &
  -7.532500510294518D-06,   0.000000000000000D+00,   1.612959127692354D-06, &
   0.000000000000000D+00,   3.824567232560691D-05,   0.000000000000000D+00, &
   1.032028784005301D-05,   0.000000000000000D+00,   4.066131080291212D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.251369733832454D-09,   0.000000000000000D+00,   1.953027617628895D-07, &
   0.000000000000000D+00,   3.625491979785719D-06,   0.000000000000000D+00, &
  -1.731144347911928D-05,   0.000000000000000D+00,  -9.253317431204150D-06, &
   0.000000000000000D+00,   2.800393570388285D-06,   0.000000000000000D+00, &
  -5.996607400458928D-06,   0.000000000000000D+00,  -5.396624430437901D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.951924949469125D-07,   0.000000000000000D+00, &
   5.014451830626345D-07,   0.000000000000000D+00,   5.690169752225577D-06, &
   0.000000000000000D+00,  -2.008486570530813D-05,   0.000000000000000D+00, &
  -1.217319718386050D-05,   0.000000000000000D+00,  -4.916734406731237D-06, &
   0.000000000000000D+00,  -3.011553969369944D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.100321868602302D-07,   0.000000000000000D+00, &
  -1.243094989786310D-06,   0.000000000000000D+00,   1.525998207762389D-06, &
   0.000000000000000D+00,   9.052484606345401D-06,   0.000000000000000D+00, &
  -7.339942190403786D-06,   0.000000000000000D+00,  -6.534802057321898D-07, &
   0.000000000000000D+00,   1.681138248390284D-06,   0.000000000000000D+00, &
   1.173281936908050D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.248174866920080D-08, &
   0.000000000000000D+00,   1.774762363875203D-07,   0.000000000000000D+00, &
  -2.399464139179371D-06,   0.000000000000000D+00,   4.519211538229660D-06, &
   0.000000000000000D+00,   1.110944446468065D-05,   0.000000000000000D+00, &
   3.216347435852935D-06,   0.000000000000000D+00,   3.569751031342451D-06, &
   0.000000000000000D+00,   1.136481174987799D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.141315622231244D-10,   0.000000000000000D+00, &
   2.025124152393751D-07,   0.000000000000000D+00,   4.205426914246842D-07, &
   0.000000000000000D+00,  -4.377687974023206D-06,   0.000000000000000D+00, &
   4.167339492080376D-06,   0.000000000000000D+00,   5.439194917684107D-06, &
   0.000000000000000D+00,   8.066980040168238D-07,   0.000000000000000D+00, &
   1.677047090472612D-07,   0.000000000000000D+00,  -1.431034148263159D-07 /)
lem(:,103) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.545620568229600D-03,  -7.002109557680321D-03, &
   2.456948824489967D-03,   1.158538614918509D-02,  -4.646100236449034D-03, &
  -7.830176889229119D-03,   5.631291722397820D-03,   9.586012953953774D-04, &
  -2.656667166985521D-03,   1.349606672329038D-03,   1.668289003361900D-04, &
  -6.490113924627487D-04,   2.473011339249863D-04,   1.485442238153636D-04, &
  -1.307834952585830D-04,  -3.277148049245367D-05,   6.161655779737772D-05, &
   4.944310429770864D-06,  -2.974822411705105D-05,   1.539025452605386D-06, &
   1.506077772506806D-05,   0.000000000000000D+00,   4.545620568229609D-03, &
   7.002109557680319D-03,  -2.456948824489974D-03,  -1.158538614918509D-02, &
   4.646100236449033D-03,   7.830176889229108D-03,  -5.631291722397832D-03, &
  -9.586012953953754D-04,   2.656667166985513D-03,  -1.349606672329040D-03, &
  -1.668289003361828D-04,   6.490113924627470D-04,  -2.473011339249918D-04, &
  -1.485442238153623D-04,   1.307834952585896D-04,   3.277148049245638D-05, &
  -6.161655779738661D-05,  -4.944310429773221D-06,   2.974822411705768D-05, &
  -1.539025452605638D-06,  -1.506077772506847D-05,   0.000000000000000D+00, &
   2.272810284114805D-03,   0.000000000000000D+00,   2.781451499422600D-04, &
   0.000000000000000D+00,  -8.994482579309889D-04,   0.000000000000000D+00, &
  -1.182245551174143D-03,   0.000000000000000D+00,   1.699868809115091D-03, &
   0.000000000000000D+00,  -5.546825798338058D-04,   0.000000000000000D+00, &
   1.341500437174475D-06,   0.000000000000000D+00,   1.704427349101257D-05, &
   0.000000000000000D+00,  -1.007687553496874D-05,   0.000000000000000D+00, &
   5.355516035081859D-06,   0.000000000000000D+00,  -2.898723763967568D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.796814296814298D-04, &
   0.000000000000000D+00,  -6.105006105006105D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.074314574314576D-04,   0.000000000000000D+00, &
   1.946548821548823D-04,   0.000000000000000D+00,  -8.961626608685440D-05, &
   0.000000000000000D+00,   1.382502301619947D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.684126331185156D-05,   0.000000000000000D+00, &
  -3.368106177980126D-04,   0.000000000000000D+00,  -1.876680645062998D-04, &
   0.000000000000000D+00,   1.056605932766924D-04,   0.000000000000000D+00, &
   1.614506832054379D-04,   0.000000000000000D+00,  -1.636122681498065D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.153848903074916D-05, &
   0.000000000000000D+00,   1.038404565604920D-04,   0.000000000000000D+00, &
   3.723569745241573D-04,   0.000000000000000D+00,   2.097021284206850D-04, &
   0.000000000000000D+00,  -2.445524891345322D-05,   0.000000000000000D+00, &
  -1.534703256839480D-04,   0.000000000000000D+00,  -1.848924800259938D-04, &
   0.000000000000000D+00,   9.432066229908717D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.515727134922171D-05,   0.000000000000000D+00, &
   5.246695455673779D-05,   0.000000000000000D+00,  -9.984595249248373D-05, &
   0.000000000000000D+00,  -2.777482920308964D-04,   0.000000000000000D+00, &
  -1.927149313257334D-04,   0.000000000000000D+00,  -2.930393965525722D-07, &
   0.000000000000000D+00,   1.100760750444415D-05,   0.000000000000000D+00, &
   1.611532873483124D-04,   0.000000000000000D+00,   1.284225384220966D-04, &
   0.000000000000000D+00,  -1.746717374057234D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.397637510034901D-06,   0.000000000000000D+00, &
  -8.807158364635679D-06,   0.000000000000000D+00,  -6.508182725521271D-05, &
   0.000000000000000D+00,   9.191571981556753D-05,   0.000000000000000D+00, &
   1.433900283918737D-04,   0.000000000000000D+00,   1.111548720605539D-04, &
   0.000000000000000D+00,   8.874107003060553D-07,   0.000000000000000D+00, &
  -1.566671067709579D-06,   0.000000000000000D+00,   1.149627731267327D-06, &
   0.000000000000000D+00,  -1.216421610206734D-04,   0.000000000000000D+00, &
  -4.956779539364475D-05,   0.000000000000000D+00,  -1.878864762437368D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.290942198196294D-07, &
   0.000000000000000D+00,   1.470930101249545D-06,   0.000000000000000D+00, &
   3.159746942557431D-06,   0.000000000000000D+00,   3.585085609237170D-05, &
   0.000000000000000D+00,  -7.106580507957288D-05,   0.000000000000000D+00, &
  -5.718533436743029D-05,   0.000000000000000D+00,  -3.442195825008331D-05, &
   0.000000000000000D+00,   6.798364585271844D-09,   0.000000000000000D+00, &
   2.458495057479964D-07,   0.000000000000000D+00,   6.619926970076009D-07, &
   0.000000000000000D+00,   8.977789008433193D-06,   0.000000000000000D+00, &
   5.778352738626142D-05,   0.000000000000000D+00,   5.034911784465148D-06, &
   0.000000000000000D+00,  -4.276074565866016D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.259946556488173D-07,   0.000000000000000D+00,   7.908071535409735D-07, &
   0.000000000000000D+00,  -6.068321025435817D-06,   0.000000000000000D+00, &
   5.778377507307184D-06,   0.000000000000000D+00,   3.975083355703616D-05, &
   0.000000000000000D+00,   1.860010236869765D-05,   0.000000000000000D+00, &
   3.151744179576239D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.355797179666274D-08,   0.000000000000000D+00, &
  -9.515002145670005D-07,   0.000000000000000D+00,   1.298369132562246D-06, &
   0.000000000000000D+00,  -1.851578303289465D-05,   0.000000000000000D+00, &
  -9.289457057230361D-06,   0.000000000000000D+00,   7.276130071426000D-06, &
   0.000000000000000D+00,   3.617970842091877D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.768447691510303D-09,   0.000000000000000D+00,   8.455254228754003D-08, &
   0.000000000000000D+00,  -1.046013707125784D-06,   0.000000000000000D+00, &
   3.614146374641646D-06,   0.000000000000000D+00,  -2.123922242831889D-05, &
   0.000000000000000D+00,  -1.506720250677384D-05,   0.000000000000000D+00, &
  -2.880499401495464D-06,   0.000000000000000D+00,   1.974404707253754D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.272080156935302D-10,   0.000000000000000D+00,   1.530317350727926D-07, &
   0.000000000000000D+00,  -2.383984463434631D-08,   0.000000000000000D+00, &
   1.544917230930554D-06,   0.000000000000000D+00,   8.964762870593276D-06, &
   0.000000000000000D+00,  -8.123001764921341D-06,   0.000000000000000D+00, &
  -6.025162857614552D-06,   0.000000000000000D+00,  -2.246729236474618D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.379868260078590D-08,   0.000000000000000D+00,   5.296248362789797D-07, &
   0.000000000000000D+00,  -8.885581088009923D-07,   0.000000000000000D+00, &
   5.777238076705587D-06,   0.000000000000000D+00,   1.227822649087481D-05, &
   0.000000000000000D+00,   2.552430117615667D-06,   0.000000000000000D+00, &
  -9.101047952104763D-07,   0.000000000000000D+00,  -1.624469712619628D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.057035083498611D-08, &
   0.000000000000000D+00,  -1.571600267417627D-07,   0.000000000000000D+00, &
   7.975200413803352D-07,   0.000000000000000D+00,  -3.957907812096804D-06, &
   0.000000000000000D+00,   4.660356190529319D-06,   0.000000000000000D+00, &
   7.585758815987931D-06,   0.000000000000000D+00,   1.734167034808732D-06, &
   0.000000000000000D+00,  -7.626377579533910D-07,   0.000000000000000D+00 /)
lem(:,104) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.701034543599431D-03,   1.728739868612421D-03,   5.001089648402531D-03, &
  -7.144945518358076D-04,  -8.151624611726926D-03,   2.649509779453308D-03, &
   4.026165527652359D-03,  -3.186758275384014D-03,   6.578048888661616D-04, &
   8.533375850923119D-04,  -1.049864935438532D-03,   4.740043516630315D-04, &
   1.165480091337009D-04,  -2.814803292301310D-04,   1.056448414569227D-04, &
   6.702717562890073D-05,  -6.665656226586406D-05,  -1.064909283795142D-05, &
   3.385644757581749D-05,  -2.297525869035488D-06,  -1.689079612315534D-05, &
   4.318711955036264D-06,   1.701034543599433D-03,   1.728739868612419D-03, &
   5.001089648402530D-03,  -7.144945518358102D-04,  -8.151624611726919D-03, &
   2.649509779453315D-03,   4.026165527652371D-03,  -3.186758275384009D-03, &
   6.578048888661611D-04,   8.533375850923093D-04,  -1.049864935438527D-03, &
   4.740043516630305D-04,   1.165480091336960D-04,  -2.814803292301319D-04, &
   1.056448414569247D-04,   6.702717562890035D-05,  -6.665656226586713D-05, &
  -1.064909283794887D-05,   3.385644757582132D-05,  -2.297525869039285D-06, &
  -1.689079612315554D-05,   4.318711955038814D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.698982806188855D-03,   0.000000000000000D+00, &
  -1.189003995060134D-02,   0.000000000000000D+00,   1.331898054736475D-02, &
   0.000000000000000D+00,  -6.260987844536127D-03,   0.000000000000000D+00, &
   5.435845872089291D-04,   0.000000000000000D+00,   4.316377809594816D-04, &
   0.000000000000000D+00,  -1.289437289500162D-04,   0.000000000000000D+00, &
   4.393095837890043D-05,   0.000000000000000D+00,  -1.725456729632748D-05, &
   0.000000000000000D+00,   7.599724407759373D-06,   0.000000000000000D+00, &
  -7.683288933288950D-05,  -8.897352647352645D-05,   0.000000000000000D+00, &
   2.306721056721055D-05,   0.000000000000000D+00,  -1.307720057720058D-04, &
   9.833916083916094D-05,   0.000000000000000D+00,   1.592157842157840D-04, &
   0.000000000000000D+00,  -1.271196124137305D-05,   0.000000000000000D+00, &
  -3.942646405881693D-05,   0.000000000000000D+00,   4.224795068085509D-04, &
  -6.262120232708472D-05,   0.000000000000000D+00,  -9.020289351171695D-05, &
   0.000000000000000D+00,  -3.461575784645632D-04,   0.000000000000000D+00, &
   1.322958946797956D-05,   0.000000000000000D+00,   5.799181253902612D-05, &
   0.000000000000000D+00,   1.044290448436152D-04,   0.000000000000000D+00, &
  -5.864368380694809D-04,   2.131189244192360D-05,   0.000000000000000D+00, &
   1.220050804029137D-05,   0.000000000000000D+00,   1.153568935471993D-04, &
   0.000000000000000D+00,   4.485863199945051D-04,   0.000000000000000D+00, &
  -7.623203250138440D-06,   0.000000000000000D+00,  -1.583115456954465D-05, &
   0.000000000000000D+00,  -1.226097966253734D-04,   0.000000000000000D+00, &
  -1.359343808066716D-04,   0.000000000000000D+00,   4.399621597892515D-04, &
  -2.430812221833798D-06,   0.000000000000000D+00,   1.670478644162797D-06, &
   0.000000000000000D+00,   3.211627419493130D-05,   0.000000000000000D+00, &
  -1.319282658775695D-04,   0.000000000000000D+00,  -3.310842725003490D-04, &
   0.000000000000000D+00,   1.578308508918303D-06,   0.000000000000000D+00, &
  -2.145500361939928D-08,   0.000000000000000D+00,   9.969629779349290D-06, &
   0.000000000000000D+00,   1.557580354829735D-04,   0.000000000000000D+00, &
   8.018293955422595D-05,   0.000000000000000D+00,  -1.588151601014252D-04, &
  -2.657945769399394D-07,   0.000000000000000D+00,   9.167986856769574D-08, &
   0.000000000000000D+00,  -8.824929072543932D-06,   0.000000000000000D+00, &
  -5.281186786415456D-05,   0.000000000000000D+00,   1.253916578627554D-04, &
   0.000000000000000D+00,   1.299837712539419D-04,   0.000000000000000D+00, &
   1.479204602059686D-08,   0.000000000000000D+00,   4.999669532244282D-07, &
   0.000000000000000D+00,   2.251880935186991D-06,   0.000000000000000D+00, &
  -6.854209053255855D-06,   0.000000000000000D+00,  -1.222217490003373D-04, &
   0.000000000000000D+00,  -1.225248889300615D-05,   0.000000000000000D+00, &
  -4.227771985403860D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.076423397125949D-07,   0.000000000000000D+00,   6.156830211002079D-07, &
   0.000000000000000D+00,   1.049667941399244D-05,   0.000000000000000D+00, &
   2.460123542847947D-05,   0.000000000000000D+00,  -8.243564104772870D-05, &
   0.000000000000000D+00,  -2.206166187161378D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.574631776550523D-08, &
   0.000000000000000D+00,  -5.620455120513854D-07,   0.000000000000000D+00, &
  -2.514270376596175D-06,   0.000000000000000D+00,   1.902765061208607D-05, &
   0.000000000000000D+00,   5.001349589328034D-05,   0.000000000000000D+00, &
  -3.091439268671212D-06,   0.000000000000000D+00,   3.039071795175309D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.778142977026847D-09, &
   0.000000000000000D+00,  -3.845529246413880D-08,   0.000000000000000D+00, &
  -2.171014656923762D-07,   0.000000000000000D+00,  -1.123764359488984D-05, &
   0.000000000000000D+00,   1.553618420077929D-05,   0.000000000000000D+00, &
   3.034573042123182D-05,   0.000000000000000D+00,   4.382180590676685D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.653006419853997D-10,   0.000000000000000D+00,   1.093366840653112D-07, &
   0.000000000000000D+00,  -1.925590600027391D-07,   0.000000000000000D+00, &
   6.941770019847493D-07,   0.000000000000000D+00,  -2.296812070071748D-05, &
   0.000000000000000D+00,  -1.117123432387653D-06,   0.000000000000000D+00, &
  -4.893088972127556D-06,   0.000000000000000D+00,  -1.637613565599716D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.781959825527839D-08,   0.000000000000000D+00, &
   3.001792854134021D-07,   0.000000000000000D+00,   6.964842996584105D-07, &
   0.000000000000000D+00,   2.802080502769115D-06,   0.000000000000000D+00, &
  -2.301775229903317D-05,   0.000000000000000D+00,  -6.099700345909401D-06, &
   0.000000000000000D+00,  -6.112945098769632D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.606829946251353D-09,   0.000000000000000D+00, &
  -9.814871217590054D-09,   0.000000000000000D+00,   1.299496269251311D-07, &
   0.000000000000000D+00,   4.242411299612529D-06,   0.000000000000000D+00, &
   6.816150018845667D-06,   0.000000000000000D+00,  -6.621686742311859D-06, &
   0.000000000000000D+00,   4.553235401547791D-06,   0.000000000000000D+00, &
   6.948611436984310D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.265801893819882D-10, &
   0.000000000000000D+00,  -5.939506622669672D-08,   0.000000000000000D+00, &
   1.154461264971188D-08,   0.000000000000000D+00,  -1.393099902989085D-06, &
   0.000000000000000D+00,   8.521349055656016D-06,   0.000000000000000D+00, &
   8.183160187927377D-06,   0.000000000000000D+00,   4.053215351204230D-06, &
   0.000000000000000D+00,   4.428688224617128D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.375287806226161D-11,   0.000000000000000D+00, &
   1.578134333401319D-08,   0.000000000000000D+00,  -1.608962070417857D-07, &
   0.000000000000000D+00,  -2.058546011764565D-07,   0.000000000000000D+00, &
  -4.776052597330332D-06,   0.000000000000000D+00,   6.664549104456058D-06, &
   0.000000000000000D+00,   1.685543023738079D-06,   0.000000000000000D+00, &
  -6.939092160382779D-07,   0.000000000000000D+00,  -3.145183577682122D-06 /)
lem(:,105) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.392052305789432D-03,  -1.090388747385759D-02, &
   3.657093638129716D-04,   9.896265890398149D-03,  -1.960422648937503D-03, &
  -3.270640420092587D-03,   2.175426268669658D-03,  -8.165396125654839D-04, &
  -7.573191885964770D-05,   5.220502747156544D-04,  -5.042521001738241D-04, &
   2.284879820956160D-04,   6.448592601497481D-05,  -1.638887635928501D-04, &
   6.779433082900004D-05,   4.290389932987583D-05,  -4.901426793648641D-05, &
  -3.647814371034485D-06,   2.573420881949157D-05,  -5.485304054941130D-06, &
  -1.235398216020399D-05,   0.000000000000000D+00,   4.392052305789413D-03, &
   1.090388747385759D-02,  -3.657093638129739D-04,  -9.896265890398151D-03, &
   1.960422648937501D-03,   3.270640420092591D-03,  -2.175426268669650D-03, &
   8.165396125654876D-04,   7.573191885964549D-05,  -5.220502747156508D-04, &
   5.042521001738231D-04,  -2.284879820956243D-04,  -6.448592601497934D-05, &
   1.638887635928587D-04,  -6.779433082899302D-05,  -4.290389932988044D-05, &
   4.901426793648414D-05,   3.647814371036309D-06,  -2.573420881949501D-05, &
   5.485304054938965D-06,   1.235398216021137D-05,   0.000000000000000D+00, &
   2.196026152894709D-03,   0.000000000000000D+00,  -1.169239796979503D-02, &
   0.000000000000000D+00,   2.511763882636132D-02,   0.000000000000000D+00, &
  -2.705466564547382D-02,   0.000000000000000D+00,   1.534718409245538D-02, &
   0.000000000000000D+00,  -4.394776533629071D-03,   0.000000000000000D+00, &
   6.889812095284361D-04,   0.000000000000000D+00,  -1.685985028744578D-04, &
   0.000000000000000D+00,   5.389620928164262D-05,   0.000000000000000D+00, &
  -2.066265428633214D-05,   0.000000000000000D+00,   9.043400427442309D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.736111111111112D-04, &
   0.000000000000000D+00,  -2.480158730158736D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.764541014541015D-05,   0.000000000000000D+00, &
   5.877802752802750D-04,   0.000000000000000D+00,   3.526566026566076D-06, &
   0.000000000000000D+00,  -1.254719545160719D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.820704350116126D-05,   0.000000000000000D+00, &
  -1.721521942110177D-04,   0.000000000000000D+00,  -1.215523268602312D-03, &
   0.000000000000000D+00,   1.531012560424333D-05,   0.000000000000000D+00, &
   9.577652524633946D-05,   0.000000000000000D+00,   4.213388722023595D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.794408721653307D-05, &
   0.000000000000000D+00,   2.877535046652703D-05,   0.000000000000000D+00, &
   2.097021284206850D-04,   0.000000000000000D+00,   1.584321150824153D-03, &
   0.000000000000000D+00,  -1.059204813074784D-05,   0.000000000000000D+00, &
  -3.019802265158306D-05,   0.000000000000000D+00,  -2.360973018746821D-04, &
   0.000000000000000D+00,  -5.991254214696556D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.158916762631271D-07,   0.000000000000000D+00, &
   1.404415587852082D-06,   0.000000000000000D+00,   4.584872552413202D-05, &
   0.000000000000000D+00,  -2.257194734337729D-04,   0.000000000000000D+00, &
  -1.346869373461046D-03,   0.000000000000000D+00,   2.014132238590375D-06, &
   0.000000000000000D+00,   1.703606663897504D-06,   0.000000000000000D+00, &
   3.055745670467116D-05,   0.000000000000000D+00,   3.133671605465000D-04, &
   0.000000000000000D+00,   4.584096863165207D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.407318545559256D-07,   0.000000000000000D+00, &
   5.480247605700996D-07,   0.000000000000000D+00,  -2.092336631442246D-05, &
   0.000000000000000D+00,  -8.684112595936196D-05,   0.000000000000000D+00, &
   2.487188490948895D-04,   0.000000000000000D+00,   7.601992500733128D-04, &
   0.000000000000000D+00,   8.346206522256466D-08,   0.000000000000000D+00, &
   4.992665722576605D-07,   0.000000000000000D+00,   5.483033429362782D-06, &
   0.000000000000000D+00,  -2.869331097232317D-05,   0.000000000000000D+00, &
  -2.713203703747845D-04,   0.000000000000000D+00,  -1.698710977227943D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.006179130541460D-09, &
   0.000000000000000D+00,  -6.837819379885452D-07,   0.000000000000000D+00, &
   1.581432499856188D-06,   0.000000000000000D+00,   2.564284453039676D-05, &
   0.000000000000000D+00,   4.348741617296036D-05,   0.000000000000000D+00, &
  -2.244847682523567D-04,   0.000000000000000D+00,  -2.932410448889940D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.557530540235015D-07,   0.000000000000000D+00,  -2.144005340410579D-06, &
   0.000000000000000D+00,  -4.616956916959449D-06,   0.000000000000000D+00, &
   4.550834841765431D-05,   0.000000000000000D+00,   1.376116182365999D-04, &
   0.000000000000000D+00,  -3.206779015530599D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.856542683557664D-08,   0.000000000000000D+00,   1.355788720971706D-07, &
   0.000000000000000D+00,  -2.020049699014824D-06,   0.000000000000000D+00, &
  -2.328430570393959D-05,   0.000000000000000D+00,   3.552991880743293D-05, &
   0.000000000000000D+00,   1.320340878681072D-04,   0.000000000000000D+00, &
   8.679217511004627D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.703777331439395D-08,   0.000000000000000D+00, &
   5.287319114864235D-07,   0.000000000000000D+00,  -2.784725350879196D-07, &
   0.000000000000000D+00,   5.069017900853474D-07,   0.000000000000000D+00, &
  -5.167700436190973D-05,   0.000000000000000D+00,  -1.243132798342323D-05, &
   0.000000000000000D+00,   3.298229006236808D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.551828660969299D-09,   0.000000000000000D+00,  -1.698978507557404D-07, &
   0.000000000000000D+00,   1.710667693483836D-06,   0.000000000000000D+00, &
   1.097465535354115D-06,   0.000000000000000D+00,   6.616743135223169D-06, &
   0.000000000000000D+00,  -6.316777652593629D-05,   0.000000000000000D+00, &
  -4.345457195798992D-05,   0.000000000000000D+00,  -2.602760132018892D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.299742953297385D-10,   0.000000000000000D+00,  -3.414749895737850D-08, &
   0.000000000000000D+00,  -4.818925994589027D-07,   0.000000000000000D+00, &
   1.878427934055580D-06,   0.000000000000000D+00,   7.797471344084744D-06, &
   0.000000000000000D+00,   1.963522084423208D-05,   0.000000000000000D+00, &
  -3.332913058303437D-05,   0.000000000000000D+00,  -1.822814598426464D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.000172492324074D-08,   0.000000000000000D+00,  -3.792127906236556D-07, &
   0.000000000000000D+00,  -7.230960023145379D-07,   0.000000000000000D+00, &
  -1.784406451212471D-06,   0.000000000000000D+00,   1.741908898247055D-05, &
   0.000000000000000D+00,   3.176167531047001D-05,   0.000000000000000D+00, &
   5.048919510055488D-06,   0.000000000000000D+00,   9.550909604887881D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.927945776196511D-09, &
   0.000000000000000D+00,   2.275576111678755D-07,   0.000000000000000D+00, &
  -1.105575422703793D-06,   0.000000000000000D+00,  -7.822488541629099D-07, &
   0.000000000000000D+00,  -1.016510603502517D-05,   0.000000000000000D+00, &
   1.664158145063624D-05,   0.000000000000000D+00,   2.342674229118308D-05, &
   0.000000000000000D+00,   7.814022713753335D-06,   0.000000000000000D+00 /)
lem(:,106) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.447554435589641D-03,  -1.738676660258689D-02, &
  -9.889315365568542D-03,   3.563824384510647D-02,   9.027621454813430D-03, &
  -2.876805566702393D-02,  -1.040277950243279D-02,   7.775884617164800D-03, &
   7.443048384583820D-03,   2.881650259085663D-03,  -1.385170927272300D-03, &
  -2.408606861379079D-03,  -9.203900607446921D-04,   5.460671818415206D-04, &
   6.467629155189640D-04,  -1.425185168915729D-06,  -3.132362648171754D-04, &
  -9.794598417675299D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.447554435589678D-03, &
  -1.738676660258694D-02,  -9.889315365568547D-03,   3.563824384510653D-02, &
   9.027621454813390D-03,  -2.876805566702396D-02,  -1.040277950243272D-02, &
   7.775884617164774D-03,   7.443048384583776D-03,   2.881650259085715D-03, &
  -1.385170927272314D-03,  -2.408606861379136D-03,  -9.203900607446720D-04, &
   5.460671818415555D-04,   6.467629155189375D-04,  -1.425185168942083D-06, &
  -3.132362648171763D-04,  -9.794598417674281D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.950119597178423D-04,   0.000000000000000D+00,   2.089413854118975D-06, &
   0.000000000000000D+00,  -1.311760135289545D-03,   0.000000000000000D+00, &
  -2.448531860296517D-05,   0.000000000000000D+00,  -1.999634352575931D-06, &
   3.101187311713625D-03,   0.000000000000000D+00,   2.137410806141460D-05, &
   0.000000000000000D+00,   7.420769804671122D-06,   0.000000000000000D+00, &
  -1.997108498656485D-03,   0.000000000000000D+00,   3.520473334714811D-04, &
   0.000000000000000D+00,   7.749388554341911D-06,   0.000000000000000D+00, &
  -3.612121452678474D-06,  -4.023427552839316D-03,   0.000000000000000D+00, &
  -1.156895506740708D-03,   0.000000000000000D+00,  -8.005410482190677D-05, &
   0.000000000000000D+00,  -7.623203250138440D-06,   0.000000000000000D+00, &
   8.371985771366570D-03,   0.000000000000000D+00,   9.397534010537102D-04, &
   0.000000000000000D+00,   2.308493020567340D-05,   0.000000000000000D+00, &
   2.371478280921003D-05,   0.000000000000000D+00,   6.256509013861891D-06, &
  -4.436972817643169D-03,   0.000000000000000D+00,   1.355841553041712D-03, &
   0.000000000000000D+00,   3.952042483473364D-04,   0.000000000000000D+00, &
   2.643570060954649D-05,   0.000000000000000D+00,  -6.706550477162232D-06, &
   0.000000000000000D+00,  -2.508509124473590D-03,   0.000000000000000D+00, &
  -3.096756707942600D-03,   0.000000000000000D+00,  -5.355366877779045D-04, &
   0.000000000000000D+00,  -1.005601389233101D-04,   0.000000000000000D+00, &
  -1.957309455269297D-05,   0.000000000000000D+00,   4.656244291194726D-06, &
   4.327122153209114D-03,   0.000000000000000D+00,   1.499668613344747D-03, &
   0.000000000000000D+00,  -3.682812304430286D-04,   0.000000000000000D+00, &
  -4.507004439700641D-05,   0.000000000000000D+00,   2.349347312393303D-05, &
   0.000000000000000D+00,   1.104283535386522D-05,   0.000000000000000D+00, &
  -2.075072591965852D-03,   0.000000000000000D+00,   1.173471500904535D-03, &
   0.000000000000000D+00,   1.154622941707341D-03,   0.000000000000000D+00, &
   2.366974831303813D-04,   0.000000000000000D+00,   4.627535821274048D-05, &
   0.000000000000000D+00,  -2.948697108882375D-06,   0.000000000000000D+00, &
  -9.075891015413227D-06,   3.574793761898334D-04,   0.000000000000000D+00, &
  -1.619263336385427D-03,   0.000000000000000D+00,  -4.106476486675716D-04, &
   0.000000000000000D+00,   2.667317800579324D-05,   0.000000000000000D+00, &
  -4.020263889210732D-05,   0.000000000000000D+00,  -2.348966241918463D-05, &
   0.000000000000000D+00,  -3.337024109319609D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.677717311180752D-04, &
   0.000000000000000D+00,  -6.666893303721948D-04,   0.000000000000000D+00, &
  -3.365347121656859D-04,   0.000000000000000D+00,  -7.262678645519566D-05, &
   0.000000000000000D+00,  -1.899148506563489D-06,   0.000000000000000D+00, &
   1.009218044679799D-05,   0.000000000000000D+00,   1.262391399110275D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.043263124445509D-05, &
   0.000000000000000D+00,   5.608037706500861D-04,   0.000000000000000D+00, &
   3.086671473405033D-05,   0.000000000000000D+00,   2.340479427376219D-05, &
   0.000000000000000D+00,   2.854539917865650D-05,   0.000000000000000D+00, &
   7.897803724690032D-06,   0.000000000000000D+00,  -2.510724364993570D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.017670289136765D-05,   0.000000000000000D+00,   2.456958844090387D-05, &
   0.000000000000000D+00,   2.966000363011029D-04,   0.000000000000000D+00, &
   8.846043623256441D-05,   0.000000000000000D+00,   1.654545274525837D-05, &
   0.000000000000000D+00,  -7.611565212819528D-06,   0.000000000000000D+00, &
  -4.296295700654839D-06,   0.000000000000000D+00,   4.865739729020847D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.161687749810156D-06, &
   0.000000000000000D+00,  -1.433501939336112D-04,   0.000000000000000D+00, &
  -7.017950960470539D-05,   0.000000000000000D+00,   2.594351023523844D-05, &
   0.000000000000000D+00,  -1.386292400708614D-05,   0.000000000000000D+00, &
  -1.188465224420958D-05,   0.000000000000000D+00,  -3.860671290809350D-07, &
   0.000000000000000D+00,   4.215348453629796D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.158732521142971D-05,   0.000000000000000D+00, &
  -1.570464753676091D-04,   0.000000000000000D+00,  -9.229082030473599D-05, &
   0.000000000000000D+00,  -2.853740542093732D-05,   0.000000000000000D+00, &
  -3.594569363423258D-06,   0.000000000000000D+00,   4.955058656778154D-06, &
   0.000000000000000D+00,   1.862253143990891D-07,   0.000000000000000D+00, &
  -2.365734749462589D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.540782138261812D-05, &
   0.000000000000000D+00,   5.612450043020509D-05,   0.000000000000000D+00, &
  -5.425479219807682D-05,   0.000000000000000D+00,  -1.549562017766971D-05, &
   0.000000000000000D+00,   6.701809076796192D-06,   0.000000000000000D+00, &
   4.923138005825862D-06,   0.000000000000000D+00,  -2.136343036278069D-06, &
   0.000000000000000D+00,  -3.322902378383442D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.573362197581300D-06,   0.000000000000000D+00, &
   4.387610906754082D-05,   0.000000000000000D+00,   7.771190489024742D-05, &
   0.000000000000000D+00,   2.128841777068580D-05,   0.000000000000000D+00, &
   1.161991785079578D-05,   0.000000000000000D+00,   7.067717652252458D-07, &
   0.000000000000000D+00,  -2.191617874973684D-06,   0.000000000000000D+00, &
  -6.466214868062944D-07,   0.000000000000000D+00,   9.473041582912668D-08 /)
lem(:,107) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.907270696794048D-04, &
   2.859744208295248D-03,  -7.596548400262501D-03,   3.054274500846983D-03, &
   1.302527721720498D-02,  -1.252320667000728D-02,  -8.680443475928520D-03, &
   1.179691290907323D-02,   3.630301827467617D-03,  -4.897676092202845D-03, &
  -2.204130697677382D-03,   8.319806844631167D-04,   1.373714407371960D-03, &
   2.198630542071861D-04,  -5.612095047398318D-04,  -3.104173300145585D-04, &
   1.747577302069817D-04,   2.080449795969750D-04,  -3.615452434454551D-05, &
  -1.205216545980552D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.907270696794094D-04,  -2.859744208295250D-03,   7.596548400262497D-03, &
  -3.054274500846976D-03,  -1.302527721720497D-02,   1.252320667000728D-02, &
   8.680443475928484D-03,  -1.179691290907321D-02,  -3.630301827467621D-03, &
   4.897676092202819D-03,   2.204130697677385D-03,  -8.319806844631081D-04, &
  -1.373714407371980D-03,  -2.198630542071789D-04,   5.612095047398327D-04, &
   3.104173300145545D-04,  -1.747577302069741D-04,  -2.080449795969801D-04, &
   3.615452434454017D-05,   1.205216545980680D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.944802665640184D-05, &
   0.000000000000000D+00,  -8.323407823537142D-05,   0.000000000000000D+00, &
   4.741348273451812D-05,   0.000000000000000D+00,   7.602794878339047D-05, &
   0.000000000000000D+00,  -4.918333091185713D-05,   0.000000000000000D+00, &
  -2.067699340498438D-05,   0.000000000000000D+00,   2.211864333506458D-05, &
   0.000000000000000D+00,  -5.506946116099904D-06,   0.000000000000000D+00, &
   3.357346535855266D-07,   0.000000000000000D+00,   7.289054651111262D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.990009990010031D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.281300398947468D-05,   0.000000000000000D+00, &
   4.162504162504114D-06,   0.000000000000000D+00,  -1.780137074254723D-04, &
   0.000000000000000D+00,  -8.384861326037715D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.438966835251665D-04,   0.000000000000000D+00, &
  -2.328510842442739D-05,   0.000000000000000D+00,   5.141916906622822D-06, &
   0.000000000000000D+00,  -4.348363481490405D-05,   0.000000000000000D+00, &
   1.394950079160604D-04,   0.000000000000000D+00,   1.182454742826257D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.048342438435312D-04, &
   0.000000000000000D+00,  -2.486634452578732D-04,   0.000000000000000D+00, &
  -2.445524891345322D-05,   0.000000000000000D+00,  -1.059204813074784D-05, &
   0.000000000000000D+00,   6.787914527852615D-04,   0.000000000000000D+00, &
   9.783388266360391D-05,   0.000000000000000D+00,  -3.094600617820437D-05, &
   0.000000000000000D+00,   4.988346706613026D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.440906904226327D-04,   0.000000000000000D+00, &
   3.405875693532173D-04,   0.000000000000000D+00,   1.393722638170072D-04, &
   0.000000000000000D+00,   2.071927675979365D-05,   0.000000000000000D+00, &
  -4.638115367604467D-06,   0.000000000000000D+00,  -5.581083788110315D-05, &
   0.000000000000000D+00,  -6.220701943544861D-04,   0.000000000000000D+00, &
  -1.110138384670623D-04,   0.000000000000000D+00,  -2.482585939309587D-05, &
   0.000000000000000D+00,  -9.651939404640043D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.379292220320627D-04,   0.000000000000000D+00, &
   3.760397332883543D-04,   0.000000000000000D+00,  -1.379835297590045D-04, &
   0.000000000000000D+00,  -3.503967845568327D-05,   0.000000000000000D+00, &
   3.056733353543190D-06,   0.000000000000000D+00,   1.291043388656000D-05, &
   0.000000000000000D+00,  -2.858750259477144D-04,   0.000000000000000D+00, &
   1.197083677902092D-04,   0.000000000000000D+00,   3.167912201766022D-04, &
   0.000000000000000D+00,   6.787405802246298D-05,   0.000000000000000D+00, &
   1.884707964240038D-05,   0.000000000000000D+00,   2.415861062875761D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.248474981144253D-04, &
   0.000000000000000D+00,  -3.464094887346114D-04,   0.000000000000000D+00, &
  -1.544380000430616D-04,   0.000000000000000D+00,   2.109267786848882D-05, &
   0.000000000000000D+00,  -5.951346217322473D-06,   0.000000000000000D+00, &
  -9.443115440042621D-06,   0.000000000000000D+00,  -2.383864688552264D-06, &
   0.000000000000000D+00,  -2.933164112326874D-05,   0.000000000000000D+00, &
   2.325979659402728D-04,   0.000000000000000D+00,  -1.351503673485099D-04, &
   0.000000000000000D+00,  -1.018603868499546D-04,   0.000000000000000D+00, &
  -2.373941870972094D-05,   0.000000000000000D+00,  -3.460404898445452D-06, &
   0.000000000000000D+00,   1.027892754729737D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.950021377340298D-06,   0.000000000000000D+00, &
  -5.948574645340845D-05,   0.000000000000000D+00,   1.987958233674384D-04, &
   0.000000000000000D+00,   2.449385906457687D-05,   0.000000000000000D+00, &
   4.143645459053564D-06,   0.000000000000000D+00,   7.937917155005295D-06, &
   0.000000000000000D+00,   3.069317448398294D-06,   0.000000000000000D+00, &
  -5.285106603465882D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.267713678665996D-06,   0.000000000000000D+00, &
  -6.399025627027735D-05,   0.000000000000000D+00,   8.399386293450081D-05, &
   0.000000000000000D+00,   2.520416295388321D-05,   0.000000000000000D+00, &
   5.382310492317744D-06,   0.000000000000000D+00,  -4.290252386290326D-07, &
   0.000000000000000D+00,  -1.757185595076252D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.871399667363579D-06,   0.000000000000000D+00,  -3.368580990236548D-05, &
   0.000000000000000D+00,  -5.468067433090107D-05,   0.000000000000000D+00, &
   4.393427082728745D-06,   0.000000000000000D+00,  -3.154373621802899D-06, &
   0.000000000000000D+00,  -3.382592243942506D-06,   0.000000000000000D+00, &
   1.455668292912018D-06,   0.000000000000000D+00,   3.426749032217800D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.271425830095020D-06,   0.000000000000000D+00,   3.023383172723505D-05, &
   0.000000000000000D+00,  -3.182276834896121D-05,   0.000000000000000D+00, &
  -3.022446717816984D-05,   0.000000000000000D+00,  -7.243712977303180D-06, &
   0.000000000000000D+00,  -1.428889074893431D-06,   0.000000000000000D+00, &
   9.307657524632789D-07,   0.000000000000000D+00,   2.041675320483899D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.634094604587889D-07,   0.000000000000000D+00, &
  -3.721954628055633D-06,   0.000000000000000D+00,   4.138409301572558D-05, &
   0.000000000000000D+00,  -5.498817141160832D-06,   0.000000000000000D+00, &
  -3.464737136987737D-06,   0.000000000000000D+00,   1.559715892822082D-06, &
   0.000000000000000D+00,   4.513128253316377D-07,   0.000000000000000D+00, &
  -1.024027513285316D-06,   0.000000000000000D+00,  -7.516338235333200D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.760920449098265D-06, &
   0.000000000000000D+00,  -1.853303474099336D-06,   0.000000000000000D+00, &
   3.118834215924511D-05,   0.000000000000000D+00,   7.108556574079050D-06, &
   0.000000000000000D+00,   2.691719428455621D-06,   0.000000000000000D+00, &
   6.166788865473878D-07,   0.000000000000000D+00,  -1.824373431536798D-06, &
   0.000000000000000D+00,  -1.030680374408774D-06,   0.000000000000000D+00 /)
lem(:,108) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.156809000276467D-05,  -1.000387538181802D-03,  -6.203908004765873D-04, &
   7.385415981324030D-03,   1.878464761698892D-03,  -1.727814613048976D-02, &
  -5.744402951577405D-03,   2.223524937153348D-02,   6.047513142157425D-03, &
  -1.425276854376007D-02,  -3.373259996209464D-03,   4.197164936215369D-03, &
   1.218722055683489D-03,  -6.518164878016415D-04,  -3.624692781701956D-04, &
   2.458658760337240D-04,   1.713376841184476D-04,  -1.157723717630489D-04, &
  -1.049469015529341D-04,   5.236153086635578D-05,   6.659210041536874D-05, &
  -2.225252037980808D-05,   8.156809000276386D-05,  -1.000387538181804D-03, &
  -6.203908004765899D-04,   7.385415981324034D-03,   1.878464761698895D-03, &
  -1.727814613048975D-02,  -5.744402951577405D-03,   2.223524937153350D-02, &
   6.047513142157420D-03,  -1.425276854376004D-02,  -3.373259996209444D-03, &
   4.197164936215350D-03,   1.218722055683472D-03,  -6.518164878016390D-04, &
  -3.624692781701925D-04,   2.458658760337128D-04,   1.713376841184420D-04, &
  -1.157723717630237D-04,  -1.049469015529278D-04,   5.236153086633528D-05, &
   6.659210041536667D-05,  -2.225252037978869D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.530257274678427D-04,   0.000000000000000D+00, &
   3.065631215570883D-04,   0.000000000000000D+00,  -2.326130639559476D-04, &
   0.000000000000000D+00,  -1.986842479230275D-04,   0.000000000000000D+00, &
   1.456570843079487D-04,   0.000000000000000D+00,   4.922761681807772D-05, &
   0.000000000000000D+00,  -4.684924989232429D-05,   0.000000000000000D+00, &
   5.482909032544240D-06,   0.000000000000000D+00,   1.278676387231042D-06, &
   0.000000000000000D+00,  -1.512582234380177D-06,   0.000000000000000D+00, &
  -2.081252081252053D-05,   2.192252192252168D-05,   0.000000000000000D+00, &
   2.455877455877459D-04,   0.000000000000000D+00,   1.665001665001540D-06, &
  -8.597665950607133D-04,   0.000000000000000D+00,   3.591180061769338D-07, &
   0.000000000000000D+00,   7.117882117882126D-04,   0.000000000000000D+00, &
  -1.122407004759948D-04,   0.000000000000000D+00,  -8.304603892839108D-06, &
   1.080439741430454D-03,   0.000000000000000D+00,   3.774988914307806D-04, &
   0.000000000000000D+00,   6.437705741111252D-06,   0.000000000000000D+00, &
  -2.477482098225128D-03,   0.000000000000000D+00,  -3.115237016165814D-04, &
   0.000000000000000D+00,   4.785054127159398D-05,   0.000000000000000D+00, &
   9.858025531400134D-06,   1.215507059160311D-03,   0.000000000000000D+00, &
  -4.936165772079079D-04,   0.000000000000000D+00,  -2.005985500567543D-04, &
   0.000000000000000D+00,  -1.583115456954465D-05,   0.000000000000000D+00, &
   9.397534010537102D-04,   0.000000000000000D+00,   1.123143282586005D-03, &
   0.000000000000000D+00,   1.897134298450087D-04,   0.000000000000000D+00, &
   4.641202880366936D-06,   0.000000000000000D+00,   8.517222774430581D-06, &
  -1.309735940911064D-03,   0.000000000000000D+00,  -5.516485503697782D-04, &
   0.000000000000000D+00,   2.518990885948083D-04,   0.000000000000000D+00, &
   8.001775115871901D-05,   0.000000000000000D+00,   4.049330623451037D-06, &
   0.000000000000000D+00,   4.597872177627181D-04,   0.000000000000000D+00, &
  -4.109381293389907D-04,   0.000000000000000D+00,  -6.096092729976829D-04, &
   0.000000000000000D+00,  -1.170001696864328D-04,   0.000000000000000D+00, &
  -2.235424827762496D-05,   0.000000000000000D+00,  -1.424084401136961D-05, &
   2.628891725674624D-05,   0.000000000000000D+00,   5.660965319061957D-04, &
   0.000000000000000D+00,   2.805524957311871D-04,   0.000000000000000D+00, &
  -8.150452765727383D-05,   0.000000000000000D+00,  -9.066167757771171D-06, &
   0.000000000000000D+00,   7.871332497323124D-06,   0.000000000000000D+00, &
  -4.149168908894246D-05,   0.000000000000000D+00,  -1.981937836023175D-04, &
   0.000000000000000D+00,   2.453567692832732D-04,   0.000000000000000D+00, &
   2.627613742320139D-04,   0.000000000000000D+00,   5.617414518685088D-05, &
   0.000000000000000D+00,   1.286202457551650D-05,   0.000000000000000D+00, &
   3.638242756650611D-07,   4.277530997142278D-06,   0.000000000000000D+00, &
   4.702884498280682D-06,   0.000000000000000D+00,  -2.920587698148610D-04, &
   0.000000000000000D+00,  -9.163289635089346D-05,   0.000000000000000D+00, &
   6.356794570372770D-06,   0.000000000000000D+00,  -1.172663827410011D-05, &
   0.000000000000000D+00,  -5.259692908004476D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.500899602932385D-06, &
   0.000000000000000D+00,   8.136676194288234D-05,   0.000000000000000D+00, &
  -1.476084171433479D-04,   0.000000000000000D+00,  -8.340033320657193D-05, &
   0.000000000000000D+00,  -1.837615632057179D-05,   0.000000000000000D+00, &
  -1.406506253004968D-06,   0.000000000000000D+00,   6.908600658600523D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.147539307156656D-06, &
   0.000000000000000D+00,   7.565531305516386D-09,   0.000000000000000D+00, &
   1.149341691386843D-04,   0.000000000000000D+00,   7.706689692845051D-06, &
   0.000000000000000D+00,   6.623072633241841D-06,   0.000000000000000D+00, &
   7.288153159952761D-06,   0.000000000000000D+00,   8.330933553020293D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.795541857554435D-06,   0.000000000000000D+00,   7.088423579191362D-06, &
   0.000000000000000D+00,   2.987799202163225D-06,   0.000000000000000D+00, &
   7.119660789184331D-05,   0.000000000000000D+00,   2.280171919988297D-05, &
   0.000000000000000D+00,   4.946200842053047D-06,   0.000000000000000D+00, &
  -2.800691740477831D-06,   0.000000000000000D+00,  -3.312041352202362D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.276451557501405D-07, &
   0.000000000000000D+00,  -3.075364405869295D-06,   0.000000000000000D+00, &
  -1.595194632310964D-05,   0.000000000000000D+00,  -1.630391291650340D-05, &
   0.000000000000000D+00,   6.955396264001499D-06,   0.000000000000000D+00, &
  -3.930314540584710D-06,   0.000000000000000D+00,  -2.018805218740978D-06, &
   0.000000000000000D+00,  -1.429929255857445D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.204155613709174D-06,   0.000000000000000D+00, &
  -2.499155536815297D-07,   0.000000000000000D+00,  -3.162406105775583D-05, &
   0.000000000000000D+00,  -2.351144805455855D-05,   0.000000000000000D+00, &
  -7.801652196999713D-06,   0.000000000000000D+00,  -9.668185752763948D-07, &
   0.000000000000000D+00,   2.227209029577145D-06,   0.000000000000000D+00, &
   1.002898265470456D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.203115333859035D-06, &
   0.000000000000000D+00,  -2.929074916436743D-06,   0.000000000000000D+00, &
   9.116389086806821D-06,   0.000000000000000D+00,  -1.364401620739650D-05, &
   0.000000000000000D+00,  -4.186765650264330D-06,   0.000000000000000D+00, &
   1.265520868577601D-06,   0.000000000000000D+00,   1.062423905028842D-06, &
   0.000000000000000D+00,  -2.504387498894055D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.971387090522940D-07,   0.000000000000000D+00, &
   2.870340808410676D-06,   0.000000000000000D+00,   4.468392063962734D-06, &
   0.000000000000000D+00,   1.787194398963523D-05,   0.000000000000000D+00, &
   5.435617065507250D-06,   0.000000000000000D+00,   3.441168657089653D-06, &
   0.000000000000000D+00,  -5.988110882373960D-07,   0.000000000000000D+00, &
  -7.889580930710724D-07,   0.000000000000000D+00,   4.798106770239020D-07 /)
lem(:,109) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.053039513875566D-04,   4.711335995251068D-03, &
  -7.646784122222160D-03,  -7.217230132286926D-03,   1.625909347478019D-02, &
   5.777445582608935D-03,  -1.862537982108067D-02,  -8.026565562034394D-04, &
   1.167048179321248D-02,  -1.389880305339575D-03,  -4.071281158863301D-03, &
   6.474666284814463D-04,   1.020514271397839D-03,  -1.057182612381022D-04, &
  -3.194213932823359D-04,   3.803851498615493D-05,   1.223790711733187D-04, &
  -2.312806328735908D-05,  -5.676828099377270D-05,   1.515517922608320D-05, &
   3.039790526554522D-05,   0.000000000000000D+00,  -1.053039513875503D-04, &
  -4.711335995251080D-03,   7.646784122222154D-03,   7.217230132286928D-03, &
  -1.625909347478017D-02,  -5.777445582608918D-03,   1.862537982108068D-02, &
   8.026565562034224D-04,  -1.167048179321250D-02,   1.389880305339592D-03, &
   4.071281158863298D-03,  -6.474666284814517D-04,  -1.020514271397820D-03, &
   1.057182612381059D-04,   3.194213932823171D-04,  -3.803851498614810D-05, &
  -1.223790711733099D-04,   2.312806328735412D-05,   5.676828099376979D-05, &
  -1.515517922608683D-05,  -3.039790526553594D-05,   0.000000000000000D+00, &
  -5.265197569378296D-05,   0.000000000000000D+00,  -8.123604379265994D-04, &
   0.000000000000000D+00,   2.302611321468543D-04,   0.000000000000000D+00, &
   4.389050337465382D-04,   0.000000000000000D+00,  -1.392994973588934D-04, &
   0.000000000000000D+00,  -1.743006147951648D-04,   0.000000000000000D+00, &
   7.048626582753627D-05,   0.000000000000000D+00,   1.659007519967288D-05, &
   0.000000000000000D+00,  -1.205677143100458D-05,   0.000000000000000D+00, &
   5.838437238452558D-06,   0.000000000000000D+00,  -2.686221438663571D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.162504162504113D-06, &
   0.000000000000000D+00,   9.573759573759579D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.946582828935771D-04,   0.000000000000000D+00, &
  -1.093677564265796D-05,   0.000000000000000D+00,   7.955396610858806D-05, &
   0.000000000000000D+00,  -9.598244892362541D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.771609195757808D-04,   0.000000000000000D+00, &
   2.178628781901669D-04,   0.000000000000000D+00,   2.556890134289517D-05, &
   0.000000000000000D+00,  -5.088521729388606D-04,   0.000000000000000D+00, &
  -1.058275789220800D-04,   0.000000000000000D+00,   4.351835814684109D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.072134264084723D-04, &
   0.000000000000000D+00,  -2.968786582231954D-04,   0.000000000000000D+00, &
  -1.534703256839480D-04,   0.000000000000000D+00,  -3.019802265158306D-05, &
   0.000000000000000D+00,   9.783388266360391D-05,   0.000000000000000D+00, &
   5.643393848197308D-04,   0.000000000000000D+00,   1.070099791167166D-04, &
   0.000000000000000D+00,   1.026691961289485D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.801131654274743D-04,   0.000000000000000D+00, &
  -3.279362984572301D-04,   0.000000000000000D+00,   1.848015859636476D-04, &
   0.000000000000000D+00,   6.554566640042456D-05,   0.000000000000000D+00, &
   8.414982230908443D-06,   0.000000000000000D+00,   1.855068250276217D-04, &
   0.000000000000000D+00,  -1.302086660142925D-04,   0.000000000000000D+00, &
  -3.908361899172376D-04,   0.000000000000000D+00,  -8.243643900781488D-05, &
   0.000000000000000D+00,  -2.422156607428224D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.578207614570896D-05,   0.000000000000000D+00, &
   3.103505728354293D-04,   0.000000000000000D+00,   2.044670228285724D-04, &
   0.000000000000000D+00,  -6.256175165601050D-05,   0.000000000000000D+00, &
  -8.159635489766977D-06,   0.000000000000000D+00,   1.489316182079547D-05, &
   0.000000000000000D+00,  -8.261357702736037D-06,   0.000000000000000D+00, &
  -1.814028860545404D-04,   0.000000000000000D+00,   1.304502741018421D-04, &
   0.000000000000000D+00,   1.804397236648307D-04,   0.000000000000000D+00, &
   4.367905545275280D-05,   0.000000000000000D+00,   1.054680964861936D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.020576064727131D-06, &
   0.000000000000000D+00,   3.053255461770343D-05,   0.000000000000000D+00, &
  -2.184264113864330D-04,   0.000000000000000D+00,  -7.127379050333015D-05, &
   0.000000000000000D+00,   5.075232613374880D-06,   0.000000000000000D+00, &
  -8.233410410741267D-06,   0.000000000000000D+00,  -1.398597929891695D-05, &
   0.000000000000000D+00,  -1.809012443014059D-06,   0.000000000000000D+00, &
   1.189733012897058D-05,   0.000000000000000D+00,   8.888833680191860D-05, &
   0.000000000000000D+00,  -9.883020727347025D-05,   0.000000000000000D+00, &
  -5.794685756527661D-05,   0.000000000000000D+00,  -1.475819223150209D-05, &
   0.000000000000000D+00,  -9.791655862415719D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.289345007577024D-07,   0.000000000000000D+00, &
  -3.574439519521421D-06,   0.000000000000000D+00,   8.917739410907439D-06, &
   0.000000000000000D+00,   9.804147880233662D-05,   0.000000000000000D+00, &
   7.826033066620664D-06,   0.000000000000000D+00,   3.678377981877865D-06, &
   0.000000000000000D+00,   6.009853796535530D-06,   0.000000000000000D+00, &
   2.469680204055102D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.772670774857296D-06,   0.000000000000000D+00, &
  -1.235728445556195D-05,   0.000000000000000D+00,  -1.991974573069791D-06, &
   0.000000000000000D+00,   5.124665001387524D-05,   0.000000000000000D+00, &
   1.618024696935191D-05,   0.000000000000000D+00,   3.136155241346859D-06, &
   0.000000000000000D+00,  -3.563602196414057D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.872990997864227D-07,   0.000000000000000D+00,  -2.638226001634109D-06, &
   0.000000000000000D+00,  -2.789304030294761D-05,   0.000000000000000D+00, &
  -1.791493401000041D-05,   0.000000000000000D+00,   4.100113818477098D-06, &
   0.000000000000000D+00,  -1.536498703459089D-06,   0.000000000000000D+00, &
  -1.881038594202415D-06,   0.000000000000000D+00,   2.319234878439028D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.607900189070785D-07,   0.000000000000000D+00,   2.853980085612513D-06, &
   0.000000000000000D+00,   3.742300310804671D-06,   0.000000000000000D+00, &
  -2.764866945635656D-05,   0.000000000000000D+00,  -1.737798595869266D-05, &
   0.000000000000000D+00,  -5.034695349928813D-06,   0.000000000000000D+00, &
  -5.730970160811662D-07,   0.000000000000000D+00,   8.979718604962199D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.730835295146190D-09,   0.000000000000000D+00, &
   2.304772968589854D-07,   0.000000000000000D+00,   6.055723353769038D-06, &
   0.000000000000000D+00,   1.401462033320932D-05,   0.000000000000000D+00, &
  -8.176396201858162D-06,   0.000000000000000D+00,  -3.185140417885023D-06, &
   0.000000000000000D+00,   7.562379746401799D-07,   0.000000000000000D+00, &
  -5.290276504355945D-08,   0.000000000000000D+00,  -1.960722213713046D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.861465742692736D-07, &
   0.000000000000000D+00,  -3.243252853655003D-07,   0.000000000000000D+00, &
   8.958780874747969D-06,   0.000000000000000D+00,   1.646225698333916D-05, &
   0.000000000000000D+00,   4.219805007630948D-06,   0.000000000000000D+00, &
   1.381271965672737D-06,   0.000000000000000D+00,  -2.021075658965018D-07, &
   0.000000000000000D+00,  -1.173110384751532D-06,   0.000000000000000D+00 /)
lem(:,110) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.743517923809209D-03,  -2.013938070287046D-03,   2.502329421169158D-03, &
   8.887399791012262D-03,   8.987681926019652D-04,  -1.732635501133912D-02, &
  -1.294541560541361D-03,   1.641184423510103D-02,  -2.359188047777915D-03, &
  -7.444577085512885D-03,   3.464668352171455D-03,   1.302807351983578D-03, &
  -1.739811770445253D-03,   1.009638781406188D-05,   5.505425565821985D-04, &
  -3.708220159154613D-05,  -2.026872976560361D-04,   2.551846721515449D-05, &
   8.377873691047919D-05,  -1.687912995965691D-05,  -3.823890728816861D-05, &
   1.127685968386053D-05,  -1.743517923809211D-03,  -2.013938070287053D-03, &
   2.502329421169161D-03,   8.887399791012234D-03,   8.987681926019608D-04, &
  -1.732635501133911D-02,  -1.294541560541360D-03,   1.641184423510102D-02, &
  -2.359188047777920D-03,  -7.444577085512870D-03,   3.464668352171468D-03, &
   1.302807351983570D-03,  -1.739811770445259D-03,   1.009638781406516D-05, &
   5.505425565821971D-04,  -3.708220159154934D-05,  -2.026872976560370D-04, &
   2.551846721515613D-05,   8.377873691048350D-05,  -1.687912995965377D-05, &
  -3.823890728817821D-05,   1.127685968386139D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.543861859694793D-03,   0.000000000000000D+00, &
   2.326020887338736D-03,   0.000000000000000D+00,  -1.087394354531685D-03, &
   0.000000000000000D+00,  -2.911982089880038D-04,   0.000000000000000D+00, &
   5.148036581862231D-04,   0.000000000000000D+00,  -6.048191941471902D-05, &
   0.000000000000000D+00,  -7.773978165750436D-05,   0.000000000000000D+00, &
   2.735452791472742D-05,   0.000000000000000D+00,  -9.016371447851659D-06, &
   0.000000000000000D+00,   3.067629412816832D-06,   0.000000000000000D+00, &
   7.492507492507487D-05,   1.859251859251860D-04,   0.000000000000000D+00, &
  -2.955377955377956D-04,   0.000000000000000D+00,   5.723443223443225D-05, &
  -2.078803549391783D-04,   0.000000000000000D+00,  -1.590117398940929D-04, &
   0.000000000000000D+00,   6.807780092093814D-04,   0.000000000000000D+00, &
   1.382277852866087D-04,   0.000000000000000D+00,  -1.025705299602360D-04, &
  -2.454481363150097D-04,   0.000000000000000D+00,   2.057793632318176D-04, &
   0.000000000000000D+00,   1.683012603581487D-04,   0.000000000000000D+00, &
  -4.083077225492087D-04,   0.000000000000000D+00,  -4.834067124443798D-04, &
   0.000000000000000D+00,  -1.045934772714959D-04,   0.000000000000000D+00, &
   4.744725169696157D-05,   3.404361376497595D-04,   0.000000000000000D+00, &
   2.256524363722507D-04,   0.000000000000000D+00,  -2.174731143292251D-04, &
   0.000000000000000D+00,  -1.226097966253734D-04,   0.000000000000000D+00, &
   2.308493020567340D-05,   0.000000000000000D+00,   1.897134298450087D-04, &
   0.000000000000000D+00,   4.528231507046357D-04,   0.000000000000000D+00, &
   9.673791631270352D-05,   0.000000000000000D+00,   3.070898923380538D-05, &
  -9.246112019028978D-05,   0.000000000000000D+00,  -2.295908622046057D-04, &
   0.000000000000000D+00,  -2.335526368754307D-04,   0.000000000000000D+00, &
   1.392281142273782D-04,   0.000000000000000D+00,   4.977517868432065D-05, &
   0.000000000000000D+00,   1.276561042343777D-05,   0.000000000000000D+00, &
   6.403648549293972D-05,   0.000000000000000D+00,  -1.346351860336906D-04, &
   0.000000000000000D+00,  -3.148143824669320D-04,   0.000000000000000D+00, &
  -7.675893384494712D-05,   0.000000000000000D+00,  -5.277043537501003D-05, &
   1.049824377324712D-05,   0.000000000000000D+00,   1.067193825303938D-06, &
   0.000000000000000D+00,   2.138040779366158D-04,   0.000000000000000D+00, &
   1.521563814979839D-04,   0.000000000000000D+00,  -4.604858324563726D-05, &
   0.000000000000000D+00,  -2.633719123619118D-06,   0.000000000000000D+00, &
  -3.070376774778082D-06,   0.000000000000000D+00,   4.539315792993540D-06, &
   0.000000000000000D+00,  -9.845383071296687D-05,   0.000000000000000D+00, &
   1.116643150146526D-04,   0.000000000000000D+00,   1.494753124822028D-04, &
   0.000000000000000D+00,   4.206322096328613D-05,   0.000000000000000D+00, &
   2.383137572122599D-05,   2.250384573408714D-07,   0.000000000000000D+00, &
  -2.043976249791347D-06,   0.000000000000000D+00,   2.370818560960239D-05, &
   0.000000000000000D+00,  -1.543531369240792D-04,   0.000000000000000D+00, &
  -5.342157758403950D-05,   0.000000000000000D+00,   2.671281247959399D-06, &
   0.000000000000000D+00,  -5.880095697544152D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.746877237634281D-06, &
   0.000000000000000D+00,  -3.459100996142170D-06,   0.000000000000000D+00, &
   5.845249249738825D-05,   0.000000000000000D+00,  -8.236452885116487D-05, &
   0.000000000000000D+00,  -4.992961478501809D-05,   0.000000000000000D+00, &
  -1.467737521093411D-05,   0.000000000000000D+00,   2.640351456269433D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.168625220185833D-07, &
   0.000000000000000D+00,  -3.503464143741838D-06,   0.000000000000000D+00, &
  -1.806570464814545D-06,   0.000000000000000D+00,   7.091814565215422D-05, &
   0.000000000000000D+00,   3.804308282881093D-06,   0.000000000000000D+00, &
   2.800393570388281D-06,   0.000000000000000D+00,   2.161936436059725D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.469478877639954D-07,   0.000000000000000D+00,   2.447964727628545D-06, &
   0.000000000000000D+00,  -7.460907856328605D-06,   0.000000000000000D+00, &
  -7.891749841256093D-07,   0.000000000000000D+00,   4.311852743007091D-05, &
   0.000000000000000D+00,   1.411595782217473D-05,   0.000000000000000D+00, &
   1.888563392227240D-06,   0.000000000000000D+00,  -7.202688171926459D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.912677542022607D-09, &
   0.000000000000000D+00,   1.932396235083975D-07,   0.000000000000000D+00, &
   4.427144890258774D-06,   0.000000000000000D+00,  -1.780083350546279D-05, &
   0.000000000000000D+00,  -1.106130309276023D-05,   0.000000000000000D+00, &
   5.245448267263976D-06,   0.000000000000000D+00,  -6.534802057322068D-07, &
   0.000000000000000D+00,  -1.720601085835444D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.798582723292204D-07,   0.000000000000000D+00, &
   4.057592546768423D-07,   0.000000000000000D+00,   6.764071832690060D-06, &
   0.000000000000000D+00,  -2.155506095350630D-05,   0.000000000000000D+00, &
  -1.423293910335406D-05,   0.000000000000000D+00,  -4.039009294632970D-06, &
   0.000000000000000D+00,   1.502572513162069D-06,   0.000000000000000D+00, &
   3.771689824965868D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.389635756286641D-07, &
   0.000000000000000D+00,  -1.516501586004570D-06,   0.000000000000000D+00, &
   1.140559954965399D-06,   0.000000000000000D+00,   1.023904644287294D-05, &
   0.000000000000000D+00,  -8.845153365873787D-06,   0.000000000000000D+00, &
  -3.310040802136249D-06,   0.000000000000000D+00,   8.066980040168380D-07, &
   0.000000000000000D+00,   1.810718845058727D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.636384362950219D-08,   0.000000000000000D+00, &
   2.608444597796202D-07,   0.000000000000000D+00,  -2.774799133052859D-06, &
   0.000000000000000D+00,   4.276392920537134D-06,   0.000000000000000D+00, &
   1.245937352806299D-05,   0.000000000000000D+00,   2.622318048324802D-06, &
   0.000000000000000D+00,   7.492921958744871D-07,   0.000000000000000D+00, &
  -8.103041940191536D-07,   0.000000000000000D+00,  -1.480566136192586D-06 /)
lem(:,111) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.264422493014302D-03,  -1.362783139122209D-03, &
  -8.368637011358026D-03,  -6.044505603690621D-04,   1.325742737034643D-02, &
  -1.703292366930118D-03,  -1.055246166930368D-02,   4.732066094069136D-03, &
   3.116328175312788D-03,  -3.380968174239771D-03,   7.131183766871071D-04, &
   8.810631066574595D-04,  -7.360580470312223D-04,  -3.430802322745807D-05, &
   2.811859598489711D-04,  -3.988302423489251D-05,  -1.117282209897107D-04, &
   3.243384810679374D-05,   4.749299769430606D-05,  -2.078415773085962D-05, &
  -2.145903043064983D-05,   0.000000000000000D+00,  -3.264422493014311D-03, &
   1.362783139122212D-03,   8.368637011358019D-03,   6.044505603690604D-04, &
  -1.325742737034641D-02,   1.703292366930126D-03,   1.055246166930368D-02, &
  -4.732066094069157D-03,  -3.116328175312787D-03,   3.380968174239779D-03, &
  -7.131183766871149D-04,  -8.810631066574666D-04,   7.360580470312272D-04, &
   3.430802322746377D-05,  -2.811859598489764D-04,   3.988302423488603D-05, &
   1.117282209897194D-04,  -3.243384810679192D-05,  -4.749299769431686D-05, &
   2.078415773086215D-05,   2.145903043065733D-05,   0.000000000000000D+00, &
  -1.632211246507157D-03,   0.000000000000000D+00,   3.280346768366664D-03, &
   0.000000000000000D+00,  -3.458710373315478D-03,   0.000000000000000D+00, &
   2.785916849892301D-03,   0.000000000000000D+00,  -1.026402363786424D-03, &
   0.000000000000000D+00,  -2.374500954523224D-04,   0.000000000000000D+00, &
   2.472193662792945D-04,   0.000000000000000D+00,  -4.372796781629507D-05, &
   0.000000000000000D+00,   7.348005315930059D-06,   0.000000000000000D+00, &
  -4.898640943390558D-07,   0.000000000000000D+00,  -6.694307997821057D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.290376290376290D-04, &
   0.000000000000000D+00,  -7.284382284382269D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.961626608685443D-05,   0.000000000000000D+00, &
  -2.216635488694311D-04,   0.000000000000000D+00,   2.006338666002531D-04, &
   0.000000000000000D+00,   1.058042680941840D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.145910848697214D-05,   0.000000000000000D+00, &
   1.614506832054379D-04,   0.000000000000000D+00,   2.755257459688573D-04, &
   0.000000000000000D+00,  -9.114552458205715D-05,   0.000000000000000D+00, &
  -3.313245529539941D-04,   0.000000000000000D+00,  -1.257141229083953D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.207512817419939D-04, &
   0.000000000000000D+00,   1.689965643968299D-04,   0.000000000000000D+00, &
  -1.848924800259938D-04,   0.000000000000000D+00,  -2.360973018746821D-04, &
   0.000000000000000D+00,  -3.094600617820437D-05,   0.000000000000000D+00, &
   1.070099791167166D-04,   0.000000000000000D+00,   3.717035709722145D-04, &
   0.000000000000000D+00,   1.218627662005178D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.526974528589816D-05,   0.000000000000000D+00, &
  -1.721543420459829D-04,   0.000000000000000D+00,  -1.933844639488372D-04, &
   0.000000000000000D+00,   1.284225384220966D-04,   0.000000000000000D+00, &
   1.168814701901885D-04,   0.000000000000000D+00,   1.819281584391291D-05, &
   0.000000000000000D+00,   6.528914019828009D-05,   0.000000000000000D+00, &
  -1.033258814003015D-04,   0.000000000000000D+00,  -2.777807301753789D-04, &
   0.000000000000000D+00,  -8.115307150273049D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.182376478032571D-06,   0.000000000000000D+00, &
   9.560650443147563D-06,   0.000000000000000D+00,   1.776748696222745D-04, &
   0.000000000000000D+00,   1.293332700576483D-04,   0.000000000000000D+00, &
  -4.956779539364476D-05,   0.000000000000000D+00,  -7.421562080241045D-06, &
   0.000000000000000D+00,  -2.626428625890034D-06,   0.000000000000000D+00, &
  -9.896346375732133D-06,   0.000000000000000D+00,  -7.613515485097531D-05, &
   0.000000000000000D+00,   9.736063086688590D-05,   0.000000000000000D+00, &
   1.352832686495792D-04,   0.000000000000000D+00,   3.080868454304038D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.062786094418755D-06, &
   0.000000000000000D+00,  -1.921003804308576D-06,   0.000000000000000D+00, &
   3.864852757284466D-06,   0.000000000000000D+00,  -1.316539017309434D-04, &
   0.000000000000000D+00,  -4.313455266001349D-05,   0.000000000000000D+00, &
   5.034911784465149D-06,   0.000000000000000D+00,  -2.980268212733850D-05, &
   0.000000000000000D+00,  -1.618658234622364D-07,   0.000000000000000D+00, &
   1.965632524205918D-06,   0.000000000000000D+00,   1.995249967722634D-06, &
   0.000000000000000D+00,   4.096678536321741D-05,   0.000000000000000D+00, &
  -7.473558948090767D-05,   0.000000000000000D+00,  -4.236937943162210D-05, &
   0.000000000000000D+00,  -6.939978698669341D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.790988688104114D-09,   0.000000000000000D+00, &
   2.337916783596043D-07,   0.000000000000000D+00,   1.840866012816570D-06, &
   0.000000000000000D+00,   7.436333483346491D-06,   0.000000000000000D+00, &
   6.003874345141958D-05,   0.000000000000000D+00,   4.782266032711278D-07, &
   0.000000000000000D+00,   7.276130071426015D-06,   0.000000000000000D+00, &
   1.946877404641066D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.204556001800075D-07,   0.000000000000000D+00, &
   6.094034113887326D-07,   0.000000000000000D+00,  -4.615683456645783D-06, &
   0.000000000000000D+00,   5.385401295663196D-06,   0.000000000000000D+00, &
   3.809950946690496D-05,   0.000000000000000D+00,   9.633046145875052D-06, &
   0.000000000000000D+00,   4.237536280993395D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.170459967160040D-07,   0.000000000000000D+00,  -1.236934145788302D-06, &
   0.000000000000000D+00,   6.773893908721617D-07,   0.000000000000000D+00, &
  -1.770114199142554D-05,   0.000000000000000D+00,  -7.525210330508650D-06, &
   0.000000000000000D+00,   4.387275408524590D-06,   0.000000000000000D+00, &
  -6.025162857614545D-06,   0.000000000000000D+00,  -6.235950983875625D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.337008308516611D-08,   0.000000000000000D+00,   1.800463789950010D-07, &
   0.000000000000000D+00,  -1.463627857221856D-06,   0.000000000000000D+00, &
   2.663068377727483D-06,   0.000000000000000D+00,  -2.117896831576186D-05, &
   0.000000000000000D+00,  -1.113751839126321D-05,   0.000000000000000D+00, &
  -4.154162290471595D-06,   0.000000000000000D+00,  -3.456920364094685D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.574655080898666D-10,   0.000000000000000D+00, &
   1.931347112646564D-07,   0.000000000000000D+00,   2.438832718963131D-07, &
   0.000000000000000D+00,   1.234763675448716D-06,   0.000000000000000D+00, &
   7.782337883029635D-06,   0.000000000000000D+00,  -8.543498330741399D-06, &
   0.000000000000000D+00,  -9.357932502578219D-07,   0.000000000000000D+00, &
   1.734167034808735D-06,   0.000000000000000D+00,   1.340077023670622D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.366265577293958D-08, &
   0.000000000000000D+00,   6.539326981029621D-07,   0.000000000000000D+00, &
  -2.610452430597083D-07,   0.000000000000000D+00,   5.872183069518668D-06, &
   0.000000000000000D+00,   1.066276328840265D-05,   0.000000000000000D+00, &
   1.883472012827763D-06,   0.000000000000000D+00,   3.358733748304811D-06, &
   0.000000000000000D+00,   1.449768593354781D-06,   0.000000000000000D+00 /)
lem(:,112) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.258416491951676D-03,  -1.825268490717678D-03,   2.486430284546069D-03, &
   7.036410043182189D-03,  -2.412729862154746D-04,  -9.516451287118763D-03, &
   1.898270408649676D-03,   5.348156919343752D-03,  -3.615853080536956D-03, &
   4.317626125682535D-05,   1.690992016164229D-03,  -1.287632378163588D-03, &
   1.777804998751192D-04,   4.310215106556207D-04,  -3.318497286066549D-04, &
  -1.028947013301582D-05,   1.374400431801716D-04,  -3.575241110859102D-05, &
  -5.347463318385915D-05,   2.819479711025473D-05,   2.116092287279602D-05, &
  -1.789590049393338D-05,  -2.258416491951689D-03,  -1.825268490717682D-03, &
   2.486430284546069D-03,   7.036410043182195D-03,  -2.412729862154770D-04, &
  -9.516451287118781D-03,   1.898270408649683D-03,   5.348156919343749D-03, &
  -3.615853080536957D-03,   4.317626125682999D-05,   1.690992016164234D-03, &
  -1.287632378163592D-03,   1.777804998751157D-04,   4.310215106556233D-04, &
  -3.318497286066503D-04,  -1.028947013301732D-05,   1.374400431801672D-04, &
  -3.575241110859205D-05,  -5.347463318385490D-05,   2.819479711025696D-05, &
   2.116092287278936D-05,  -1.789590049393366D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.933854168849611D-03,   0.000000000000000D+00, &
   3.443475622385108D-04,   0.000000000000000D+00,  -9.193318002600075D-04, &
   0.000000000000000D+00,  -1.309064781518296D-03,   0.000000000000000D+00, &
   1.847948947696499D-03,   0.000000000000000D+00,  -6.072800060142753D-04, &
   0.000000000000000D+00,   3.657647080307086D-06,   0.000000000000000D+00, &
   1.728996885687670D-05,   0.000000000000000D+00,  -1.019655139628258D-05, &
   0.000000000000000D+00,   5.336446668031503D-06,   0.000000000000000D+00, &
   1.094391719391717D-04,   2.306721056721060D-05,   0.000000000000000D+00, &
  -1.163766788766788D-04,   0.000000000000000D+00,  -1.514978077478074D-04, &
  -9.409843968667602D-06,   0.000000000000000D+00,  -3.942646405881690D-05, &
   0.000000000000000D+00,   1.219368866427690D-04,   0.000000000000000D+00, &
   1.712033881151527D-04,   0.000000000000000D+00,   1.910154541128803D-04, &
  -1.391224301441015D-05,   0.000000000000000D+00,  -6.250002863779944D-05, &
   0.000000000000000D+00,   1.044290448436152D-04,   0.000000000000000D+00, &
  -7.462115627750315D-05,   0.000000000000000D+00,  -8.895688075254627D-05, &
   0.000000000000000D+00,  -2.985096943712463D-04,   0.000000000000000D+00, &
  -2.052115357552665D-04,   1.958932879985517D-05,   0.000000000000000D+00, &
   7.026355710566233D-05,   0.000000000000000D+00,   1.453882576294534D-04, &
   0.000000000000000D+00,  -1.359343808066716D-04,   0.000000000000000D+00, &
   2.371478280921003D-05,   0.000000000000000D+00,   4.641202880366936D-06, &
   0.000000000000000D+00,   9.673791631270352D-05,   0.000000000000000D+00, &
   3.509078197666628D-04,   0.000000000000000D+00,   2.293990876637393D-04, &
  -9.762913796094605D-06,   0.000000000000000D+00,  -1.399423038859705D-05, &
   0.000000000000000D+00,  -1.321253125179253D-04,   0.000000000000000D+00, &
  -1.742473145163403D-04,   0.000000000000000D+00,   8.018293955422596D-05, &
   0.000000000000000D+00,  -2.398454621162817D-06,   0.000000000000000D+00, &
   1.657579961355655D-06,   0.000000000000000D+00,   4.296648275007987D-05, &
   0.000000000000000D+00,  -1.009112005055900D-04,   0.000000000000000D+00, &
  -2.719388945715290D-04,   0.000000000000000D+00,  -2.088502572758082D-04, &
   1.777767669879400D-06,   0.000000000000000D+00,  -1.068559568189383D-06, &
   0.000000000000000D+00,   5.753533632559210D-06,   0.000000000000000D+00, &
   1.521591505496428D-04,   0.000000000000000D+00,   1.224504929255935D-04, &
   0.000000000000000D+00,  -1.225248889300615D-05,   0.000000000000000D+00, &
  -3.078594578119357D-07,   0.000000000000000D+00,   8.742715533670011D-07, &
   0.000000000000000D+00,  -8.329611870738647D-06,   0.000000000000000D+00, &
  -5.795106246179280D-05,   0.000000000000000D+00,   9.582865648033241D-05, &
   0.000000000000000D+00,   1.455861440703622D-04,   0.000000000000000D+00, &
   1.220068177383314D-04,   1.736010956602455D-08,   0.000000000000000D+00, &
   4.421271583743218D-07,   0.000000000000000D+00,   2.799596360830171D-06, &
   0.000000000000000D+00,  -1.428314804315916D-06,   0.000000000000000D+00, &
  -1.166212066834641D-04,   0.000000000000000D+00,  -4.775605331308692D-05, &
   0.000000000000000D+00,  -3.091439268671211D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.175437864360708D-07, &
   0.000000000000000D+00,   6.638633951150335D-07,   0.000000000000000D+00, &
   7.019092098468469D-06,   0.000000000000000D+00,   2.988698618910062D-05, &
   0.000000000000000D+00,  -7.298786530921057D-05,   0.000000000000000D+00, &
  -6.073575499734774D-05,   0.000000000000000D+00,  -3.936978729459200D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.078830883163222D-07, &
   0.000000000000000D+00,  -9.510943391005848D-07,   0.000000000000000D+00, &
  -1.084668712820032D-06,   0.000000000000000D+00,   1.142531864432104D-05, &
   0.000000000000000D+00,   5.405069434227579D-05,   0.000000000000000D+00, &
   4.547085538392118D-06,   0.000000000000000D+00,  -4.893088972127564D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.182106216076087D-08,   0.000000000000000D+00,   7.988573820182934D-08, &
   0.000000000000000D+00,  -8.547173696035187D-07,   0.000000000000000D+00, &
  -7.209397371890980D-06,   0.000000000000000D+00,   8.188755649439833D-06, &
   0.000000000000000D+00,   4.031971149941456D-05,   0.000000000000000D+00, &
   2.070389340826287D-05,   0.000000000000000D+00,   4.397905789888870D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.769056826194756D-10, &
   0.000000000000000D+00,   1.628152721626476D-07,   0.000000000000000D+00, &
  -6.701651510968322D-08,   0.000000000000000D+00,   3.705154089221315D-07, &
   0.000000000000000D+00,  -1.771644645551902D-05,   0.000000000000000D+00, &
  -7.337893629578542D-06,   0.000000000000000D+00,   8.201613837450140D-06, &
   0.000000000000000D+00,   4.553235401547781D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.822153441328371D-08,   0.000000000000000D+00, &
   5.248191518024793D-07,   0.000000000000000D+00,   1.381539638182048D-07, &
   0.000000000000000D+00,   2.919795739225533D-06,   0.000000000000000D+00, &
  -2.071085374288215D-05,   0.000000000000000D+00,  -1.535913025669427D-05, &
   0.000000000000000D+00,  -3.417037623428010D-06,   0.000000000000000D+00, &
   1.831034309843713D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.149964618159621D-08, &
   0.000000000000000D+00,  -1.538894144332519D-07,   0.000000000000000D+00, &
   5.630359244537759D-07,   0.000000000000000D+00,   2.055373944010564D-06, &
   0.000000000000000D+00,   7.451442911610366D-06,   0.000000000000000D+00, &
  -8.852266046091325D-06,   0.000000000000000D+00,  -7.100054019893400D-06, &
   0.000000000000000D+00,  -6.939092160382761D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.773719377248601D-10,   0.000000000000000D+00, &
  -6.498666583649367D-08,   0.000000000000000D+00,  -1.650086617860391D-07, &
   0.000000000000000D+00,  -6.051198765448907D-07,   0.000000000000000D+00, &
   5.476989025035600D-06,   0.000000000000000D+00,   1.121198044591221D-05, &
   0.000000000000000D+00,   2.434950622542176D-06,   0.000000000000000D+00, &
  -1.093272283971703D-06,   0.000000000000000D+00,  -1.607243383390752D-06 /)
lem(:,113) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.274360436841974D-04,  -3.488075891800928D-04, &
  -3.986379232307401D-03,  -1.246040747161467D-03,   5.897446570769503D-03, &
  -7.336450487633387D-04,  -2.796665946551903D-03,   1.883640018906703D-03, &
  -5.824350579699365D-04,  -3.212470137050037D-04,   7.239786310225474D-04, &
  -5.104116759845888D-04,   5.400433593252312D-05,   2.049271344011146D-04, &
  -1.559195138637407D-04,  -5.360526639603156D-06,   6.978908412795506D-05, &
  -2.239480111340871D-05,  -2.661095181915013D-05,   1.857756325070859D-05, &
   9.332005851136358D-06,   0.000000000000000D+00,   6.274360436842001D-04, &
   3.488075891800858D-04,   3.986379232307399D-03,   1.246040747161459D-03, &
  -5.897446570769506D-03,   7.336450487633441D-04,   2.796665946551904D-03, &
  -1.883640018906700D-03,   5.824350579699413D-04,   3.212470137050021D-04, &
  -7.239786310225535D-04,   5.104116759845882D-04,  -5.400433593252059D-05, &
  -2.049271344011166D-04,   1.559195138637400D-04,   5.360526639607387D-06, &
  -6.978908412795645D-05,   2.239480111340659D-05,   2.661095181915274D-05, &
  -1.857756325070992D-05,  -9.332005851138413D-06,   0.000000000000000D+00, &
   3.137180218421004D-04,   0.000000000000000D+00,   4.381889862185608D-03, &
   0.000000000000000D+00,  -1.287988836731126D-02,   0.000000000000000D+00, &
   1.421100108998870D-02,   0.000000000000000D+00,  -6.664329550772115D-03, &
   0.000000000000000D+00,   5.756908883232985D-04,   0.000000000000000D+00, &
   4.658718001538245D-04,   0.000000000000000D+00,  -1.400568570783334D-04, &
   0.000000000000000D+00,   4.814673029565764D-05,   0.000000000000000D+00, &
  -1.913387009190644D-05,   0.000000000000000D+00,   8.550687382475144D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.480158730158738D-05, &
   0.000000000000000D+00,  -2.480158730158728D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.176315514550812D-05,   0.000000000000000D+00, &
  -1.254719545160720D-04,   0.000000000000000D+00,   3.188872674166789D-05, &
   0.000000000000000D+00,   1.348800981153924D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.490991235573272D-05,   0.000000000000000D+00, &
  -3.634163609008808D-05,   0.000000000000000D+00,   4.213388722023595D-04, &
   0.000000000000000D+00,  -2.123958199809594D-05,   0.000000000000000D+00, &
  -7.725440194480441D-05,   0.000000000000000D+00,  -3.348659267897853D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.538279402056490D-05, &
   0.000000000000000D+00,   6.053726545212605D-05,   0.000000000000000D+00, &
   9.432066229908717D-05,   0.000000000000000D+00,  -5.991254214696556D-04, &
   0.000000000000000D+00,   4.988346706613026D-06,   0.000000000000000D+00, &
   1.026691961289485D-05,   0.000000000000000D+00,   1.218627662005178D-04, &
   0.000000000000000D+00,   4.473516236408073D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.366863816251419D-06,   0.000000000000000D+00, &
  -2.280512800939501D-05,   0.000000000000000D+00,  -1.246132538907266D-04, &
   0.000000000000000D+00,  -1.229211121714320D-04,   0.000000000000000D+00, &
   4.584096863165207D-04,   0.000000000000000D+00,   1.268366585031082D-06, &
   0.000000000000000D+00,   2.292093989327713D-06,   0.000000000000000D+00, &
   2.439065773790753D-05,   0.000000000000000D+00,  -1.450484150802373D-04, &
   0.000000000000000D+00,  -3.363096225328414D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.055730923478827D-06,   0.000000000000000D+00, &
   3.798994767664929D-06,   0.000000000000000D+00,   1.678935318273128D-05, &
   0.000000000000000D+00,   1.567470065790626D-04,   0.000000000000000D+00, &
   7.138808518086386D-05,   0.000000000000000D+00,  -1.698710977227943D-04, &
   0.000000000000000D+00,  -5.975883869946132D-07,   0.000000000000000D+00, &
  -4.356136153821678D-07,   0.000000000000000D+00,  -7.551089010256532D-06, &
   0.000000000000000D+00,  -4.416973046141666D-05,   0.000000000000000D+00, &
   1.355853610355409D-04,   0.000000000000000D+00,   1.341267792908430D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.299782562400873D-07, &
   0.000000000000000D+00,  -4.573993544580747D-07,   0.000000000000000D+00, &
   7.669385787443968D-07,   0.000000000000000D+00,  -1.105576027177604D-05, &
   0.000000000000000D+00,  -1.221984580501520D-04,   0.000000000000000D+00, &
  -8.034731019415616D-06,   0.000000000000000D+00,  -3.206779015530637D-06, &
   0.000000000000000D+00,  -1.335393043563179D-08,   0.000000000000000D+00, &
  -7.445049678205289D-08,   0.000000000000000D+00,  -4.598826913614629D-07, &
   0.000000000000000D+00,   9.384047147793865D-06,   0.000000000000000D+00, &
   1.961610303276639D-05,   0.000000000000000D+00,  -8.725984125455901D-05, &
   0.000000000000000D+00,  -2.266499544963368D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.372270865766900D-10,   0.000000000000000D+00, &
   1.476525443645217D-07,   0.000000000000000D+00,  -1.403544908314441D-07, &
   0.000000000000000D+00,  -3.046510866662228D-06,   0.000000000000000D+00, &
   1.970136616865214D-05,   0.000000000000000D+00,   4.904904608788696D-05, &
   0.000000000000000D+00,  -4.480679696369301D-06,   0.000000000000000D+00, &
   3.298229006236808D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.386518197142380D-08,   0.000000000000000D+00, &
   3.026779850793826D-07,   0.000000000000000D+00,   9.323627749212825D-07, &
   0.000000000000000D+00,  -1.076643368055745D-05,   0.000000000000000D+00, &
   1.747694842112098D-05,   0.000000000000000D+00,   3.135105585545839D-05, &
   0.000000000000000D+00,   4.164322096451036D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.927025587127951D-09,   0.000000000000000D+00,  -1.159456701954315D-08, &
   0.000000000000000D+00,   6.246348471354790D-08,   0.000000000000000D+00, &
   2.257630499169069D-06,   0.000000000000000D+00,  -2.237205335608971D-05, &
   0.000000000000000D+00,   2.033943703056921D-07,   0.000000000000000D+00, &
  -5.162006559558050D-06,   0.000000000000000D+00,  -1.822814598426460D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.169768654350691D-10,   0.000000000000000D+00,  -6.334478606887594D-08, &
   0.000000000000000D+00,   3.249124703203885D-08,   0.000000000000000D+00, &
   1.703768536307899D-07,   0.000000000000000D+00,   3.367862484588420D-06, &
   0.000000000000000D+00,  -2.347164263254870D-05,   0.000000000000000D+00, &
  -5.813038998031571D-06,   0.000000000000000D+00,  -6.472092558365839D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.452676137235116D-12,   0.000000000000000D+00, &
   1.877454379634362D-08,   0.000000000000000D+00,  -1.597658790041490D-07, &
   0.000000000000000D+00,  -4.784348997478507D-07,   0.000000000000000D+00, &
   3.010642327825725D-06,   0.000000000000000D+00,   6.274026100430684D-06, &
   0.000000000000000D+00,  -7.266396344600874D-06,   0.000000000000000D+00, &
   5.147541015231997D-06,   0.000000000000000D+00,   7.814022713753345D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.948485060975339D-10, &
   0.000000000000000D+00,   9.954014262140822D-09,   0.000000000000000D+00, &
   3.454574347483774D-08,   0.000000000000000D+00,  -2.048256948683540D-06, &
   0.000000000000000D+00,   7.780649070644781D-06,   0.000000000000000D+00, &
   7.817729678676351D-06,   0.000000000000000D+00,   4.050972657024009D-06, &
   0.000000000000000D+00,   4.897988919119686D-06,   0.000000000000000D+00 /)
lem(:,114) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.275775907699566D-03,  -4.156763789407833D-03,  -1.832781086031375D-03, &
   7.704386106399180D-03,   3.084316174185770D-03,  -6.873874971409081D-03, &
   4.886047406106087D-05,   2.443862265472236D-03,  -1.233562544377021D-03, &
   4.848407142796097D-04,  -1.374136593698695D-04,  -2.627542605011424D-04, &
   4.124427816301810D-04,  -2.580082637684503D-04,   2.046267648380565D-05, &
   1.168750239152281D-04,  -9.817474942240231D-05,   2.416534781398749D-06, &
   4.830434255641058D-05,  -2.201216504623490D-05,  -1.736766941168921D-05, &
   1.769936339087742D-05,  -1.275775907699581D-03,  -4.156763789407842D-03, &
  -1.832781086031392D-03,   7.704386106399192D-03,   3.084316174185774D-03, &
  -6.873874971409072D-03,   4.886047406106941D-05,   2.443862265472237D-03, &
  -1.233562544377032D-03,   4.848407142796138D-04,  -1.374136593698702D-04, &
  -2.627542605011486D-04,   4.124427816301846D-04,  -2.580082637684388D-04, &
   2.046267648380512D-05,   1.168750239152212D-04,  -9.817474942240450D-05, &
   2.416534781397690D-06,   4.830434255641510D-05,  -2.201216504623398D-05, &
  -1.736766941169275D-05,   1.769936339087733D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.783971269921083D-03,   0.000000000000000D+00, &
  -1.296830337749413D-02,   0.000000000000000D+00,   2.689153957572794D-02, &
   0.000000000000000D+00,  -2.869848738384408D-02,   0.000000000000000D+00, &
   1.625475021935456D-02,   0.000000000000000D+00,  -4.660612673800838D-03, &
   0.000000000000000D+00,   7.308992701248897D-04,   0.000000000000000D+00, &
  -1.784437818513287D-04,   0.000000000000000D+00,   5.676575446100118D-05, &
   0.000000000000000D+00,  -2.160178591419347D-05,   0.000000000000000D+00, &
   1.240079365079356D-05,   3.720238095238115D-05,   0.000000000000000D+00, &
   3.100198412698425D-06,   0.000000000000000D+00,  -1.666356646825398D-04, &
  -4.880442610222037D-05,   0.000000000000000D+00,  -1.308784914023898D-05, &
   0.000000000000000D+00,   6.363632282749176D-07,   0.000000000000000D+00, &
   8.362342059492802D-05,   0.000000000000000D+00,   6.098640878627081D-04, &
   3.586725541543657D-05,   0.000000000000000D+00,  -4.401310969373971D-06, &
   0.000000000000000D+00,  -1.393293913822891D-04,   0.000000000000000D+00, &
  -8.575633817506886D-06,   0.000000000000000D+00,  -4.212919554323379D-05, &
   0.000000000000000D+00,  -1.801404958453140D-04,   0.000000000000000D+00, &
  -1.270407472278513D-03,  -1.274323908805326D-05,   0.000000000000000D+00, &
   1.590066275994108D-05,   0.000000000000000D+00,   1.121915217672766D-04, &
   0.000000000000000D+00,   4.399621597892515D-04,   0.000000000000000D+00, &
   6.256509013861891D-06,   0.000000000000000D+00,   8.517222774430581D-06, &
   0.000000000000000D+00,   3.070898923380538D-05,   0.000000000000000D+00, &
   2.293990876637393D-04,   0.000000000000000D+00,   1.665945657346314D-03, &
   9.692513835903404D-07,   0.000000000000000D+00,  -8.565773766707232D-06, &
   0.000000000000000D+00,  -4.289905576118179D-05,   0.000000000000000D+00, &
  -2.549593734695845D-04,   0.000000000000000D+00,  -6.256630386782874D-04, &
   0.000000000000000D+00,  -1.308267872366110D-06,   0.000000000000000D+00, &
   2.197203798140605D-06,   0.000000000000000D+00,   2.506423089373916D-06, &
   0.000000000000000D+00,   4.029200023824772D-05,   0.000000000000000D+00, &
  -2.492478641572056D-04,   0.000000000000000D+00,  -1.429162029944675D-03, &
   3.551310875225679D-07,   0.000000000000000D+00,   8.978409779430486D-07, &
   0.000000000000000D+00,   8.369628064326477D-06,   0.000000000000000D+00, &
   3.886825371013048D-05,   0.000000000000000D+00,   3.354590570300163D-04, &
   0.000000000000000D+00,   4.839992905628560D-04,   0.000000000000000D+00, &
  -4.131372228519892D-08,   0.000000000000000D+00,  -9.462431392641717D-07, &
   0.000000000000000D+00,  -9.977798701939367D-07,   0.000000000000000D+00, &
  -1.716762869855228D-05,   0.000000000000000D+00,  -8.549806635544010D-05, &
   0.000000000000000D+00,   2.707870433208345D-04,   0.000000000000000D+00, &
   8.162399105495763D-04,   1.808344746580495D-09,   0.000000000000000D+00, &
   2.946291154348472D-07,   0.000000000000000D+00,  -1.700646459966990D-06, &
   0.000000000000000D+00,   3.071999269301062D-06,   0.000000000000000D+00, &
  -3.283829676652103D-05,   0.000000000000000D+00,  -2.906113580058820D-04, &
   0.000000000000000D+00,  -1.834779565238407D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.268081622432681D-08, &
   0.000000000000000D+00,   1.356017803911444D-07,   0.000000000000000D+00, &
  -2.477696909299165D-06,   0.000000000000000D+00,   2.504424129138065D-05, &
   0.000000000000000D+00,   4.279332052106455D-05,   0.000000000000000D+00, &
  -2.421479161376912D-04,   0.000000000000000D+00,  -3.187737824998257D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.694952731854858D-09, &
   0.000000000000000D+00,   5.219110411247988D-07,   0.000000000000000D+00, &
  -1.467989393376345D-07,   0.000000000000000D+00,  -8.230297663467074D-06, &
   0.000000000000000D+00,   4.902111769061729D-05,   0.000000000000000D+00, &
   1.483235811412106D-04,   0.000000000000000D+00,  -1.250133986722514D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.485768661144444D-09,   0.000000000000000D+00,  -1.905163693947497D-07, &
   0.000000000000000D+00,   1.671580795124341D-06,   0.000000000000000D+00, &
   1.615425633481753D-06,   0.000000000000000D+00,  -2.516624274203225D-05, &
   0.000000000000000D+00,   3.905095629555748D-05,   0.000000000000000D+00, &
   1.431251620019389D-04,   0.000000000000000D+00,   9.491262616930138D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.529563136546220D-11, &
   0.000000000000000D+00,  -1.882724084196798D-08,   0.000000000000000D+00, &
  -4.841519236615819D-07,   0.000000000000000D+00,   1.775120614083177D-06, &
   0.000000000000000D+00,   3.861592932462969D-06,   0.000000000000000D+00, &
  -5.483676842919187D-05,   0.000000000000000D+00,  -1.451057270756055D-05, &
   0.000000000000000D+00,   3.493732505745005D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.517615635445244D-08,   0.000000000000000D+00, &
  -3.727577869998259D-07,   0.000000000000000D+00,  -6.860997656179171D-07, &
   0.000000000000000D+00,   1.643245902081270D-06,   0.000000000000000D+00, &
   7.725085972709376D-06,   0.000000000000000D+00,  -6.807301281213363D-05, &
   0.000000000000000D+00,  -4.803752585565729D-05,   0.000000000000000D+00, &
  -2.828233902522370D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.119315357500470D-09, &
   0.000000000000000D+00,   2.283109657046516D-07,   0.000000000000000D+00, &
  -1.126947700896971D-06,   0.000000000000000D+00,  -1.397829164400969D-06, &
   0.000000000000000D+00,   7.162619704812788D-06,   0.000000000000000D+00, &
   2.085064401235581D-05,   0.000000000000000D+00,  -3.542614448503280D-05, &
   0.000000000000000D+00,  -1.960676206853289D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.487637400714108D-10,   0.000000000000000D+00, &
  -2.979999711157653D-09,   0.000000000000000D+00,   6.950544403084417D-07, &
   0.000000000000000D+00,  -1.079822134194387D-06,   0.000000000000000D+00, &
  -3.974778176041934D-06,   0.000000000000000D+00,   1.802266743838744D-05, &
   0.000000000000000D+00,   3.480851342643084D-05,   0.000000000000000D+00, &
   6.029054891363028D-06,   0.000000000000000D+00,   1.028783732850875D-05 /)
lem(:,115) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.908735145403053D-03, &
  -1.442573773872946D-02,  -1.186287347063516D-02,   3.031183050562961D-02, &
   1.255123506667630D-02,  -2.335777954626437D-02,  -1.165596997304475D-02, &
   4.196654333741091D-03,   6.061972175503980D-03,   3.923303270713493D-03, &
   1.938267229637583D-04,  -1.957978469061993D-03,  -1.540860905030418D-03, &
  -5.002067333797151D-05,   6.828383116815706D-04,   3.315725940011644D-04, &
  -2.049775417368154D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.908735145403106D-03,  -1.442573773872942D-02,  -1.186287347063523D-02, &
   3.031183050562962D-02,   1.255123506667634D-02,  -2.335777954626444D-02, &
  -1.165596997304472D-02,   4.196654333741132D-03,   6.061972175503950D-03, &
   3.923303270713475D-03,   1.938267229638313D-04,  -1.957978469061967D-03, &
  -1.540860905030437D-03,  -5.002067333797776D-05,   6.828383116815419D-04, &
   3.315725940011682D-04,  -2.049775417367912D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.273268527138494D-04,   0.000000000000000D+00, &
   2.505234703376726D-06,   0.000000000000000D+00,   1.353136028062101D-07, &
  -1.500357227911406D-03,   0.000000000000000D+00,  -2.860916173609655D-05, &
   0.000000000000000D+00,  -1.999634352575813D-06,   0.000000000000000D+00, &
   3.505885239631364D-03,   0.000000000000000D+00,   2.270691125180318D-05, &
   0.000000000000000D+00,   8.247686266262275D-06,   0.000000000000000D+00, &
   1.724174518292107D-06,  -2.252236927159525D-03,   0.000000000000000D+00, &
   4.058634863588417D-04,   0.000000000000000D+00,   1.036115587199174D-05, &
   0.000000000000000D+00,  -2.430812221833798D-06,   0.000000000000000D+00, &
  -4.436972817643169D-03,   0.000000000000000D+00,  -1.309735940911064D-03, &
   0.000000000000000D+00,  -9.246112019028978D-05,   0.000000000000000D+00, &
  -9.762913796094605D-06,   0.000000000000000D+00,   9.692513835903404D-07, &
   9.217630606781255D-03,   0.000000000000000D+00,   1.056158137191922D-03, &
   0.000000000000000D+00,   2.289069121751889D-05,   0.000000000000000D+00, &
   2.662239104728005D-05,   0.000000000000000D+00,   6.339971079084788D-06, &
   0.000000000000000D+00,  -4.859014544032712D-03,   0.000000000000000D+00, &
   1.499668613344747D-03,   0.000000000000000D+00,   4.502508305173536D-04, &
   0.000000000000000D+00,   3.274533157424527D-05,   0.000000000000000D+00, &
  -6.591982742247004D-06,   0.000000000000000D+00,  -5.270942401559983D-06, &
  -2.765588310076128D-03,   0.000000000000000D+00,  -3.416188102888865D-03, &
   0.000000000000000D+00,  -6.008009877978910D-04,   0.000000000000000D+00, &
  -1.144288026775572D-04,   0.000000000000000D+00,  -2.401948646711563D-05, &
   0.000000000000000D+00,   3.090391620034156D-07,   0.000000000000000D+00, &
   4.741348018146150D-03,   0.000000000000000D+00,   1.645931368303152D-03, &
   0.000000000000000D+00,  -4.106476486675716D-04,   0.000000000000000D+00, &
  -5.364877740029171D-05,   0.000000000000000D+00,   2.579566038119519D-05, &
   0.000000000000000D+00,   1.331259905791523D-05,   0.000000000000000D+00, &
  -2.435712862038821D-07,  -2.268109888247187D-03,   0.000000000000000D+00, &
   1.291426562997436D-03,   0.000000000000000D+00,   1.279072359466758D-03, &
   0.000000000000000D+00,   2.665624746322010D-04,   0.000000000000000D+00, &
   5.424644068913961D-05,   0.000000000000000D+00,  -1.326485608154397D-06, &
   0.000000000000000D+00,  -6.298718615815296D-06,   0.000000000000000D+00, &
   3.900583821149850D-04,   0.000000000000000D+00,  -1.775198011261867D-03, &
   0.000000000000000D+00,  -4.533626030810866D-04,   0.000000000000000D+00, &
   3.086671473405024D-05,   0.000000000000000D+00,  -4.476880933558715D-05, &
   0.000000000000000D+00,  -2.765599544166694D-05,   0.000000000000000D+00, &
  -3.597385276910607D-06,   0.000000000000000D+00,   6.281839142617738D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.217960237042021D-04, &
   0.000000000000000D+00,  -7.330192691033938D-04,   0.000000000000000D+00, &
  -3.746336570999131D-04,   0.000000000000000D+00,  -8.246666567926778D-05, &
   0.000000000000000D+00,  -3.109218989151846D-06,   0.000000000000000D+00, &
   1.115649160959139D-05,   0.000000000000000D+00,   4.419832952222165D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.261440844744009D-05,   0.000000000000000D+00,   6.162174190933194D-04, &
   0.000000000000000D+00,   3.497979077060419D-05,   0.000000000000000D+00, &
   2.594351023523854D-05,   0.000000000000000D+00,   3.253273958765910D-05, &
   0.000000000000000D+00,   9.430169373391873D-06,   0.000000000000000D+00, &
  -4.078931515352120D-06,   0.000000000000000D+00,  -2.483681682900434D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.452696483030158D-05, &
   0.000000000000000D+00,   2.552493535055889D-05,   0.000000000000000D+00, &
   3.267595053219864D-04,   0.000000000000000D+00,   9.867212238364244D-05, &
   0.000000000000000D+00,   1.892391684200441D-05,   0.000000000000000D+00, &
  -8.375093779761063D-06,   0.000000000000000D+00,  -7.059422236891090D-06, &
   0.000000000000000D+00,  -6.930732028390599D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.602610072870281D-06, &
   0.000000000000000D+00,  -1.561846767760491D-04,   0.000000000000000D+00, &
  -7.824909266943306D-05,   0.000000000000000D+00,   2.851792966013797D-05, &
   0.000000000000000D+00,  -1.549562017766975D-05,   0.000000000000000D+00, &
  -1.364442534505104D-05,   0.000000000000000D+00,   2.694878867329592D-09, &
   0.000000000000000D+00,   3.134835150226061D-06,   0.000000000000000D+00, &
  -2.244336846692952D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.370231051393786D-05, &
   0.000000000000000D+00,  -1.714258057538221D-04,   0.000000000000000D+00, &
  -1.021041191578757D-04,   0.000000000000000D+00,  -3.181603125337392D-05, &
   0.000000000000000D+00,  -4.156116494190726D-06,   0.000000000000000D+00, &
   5.952127976632868D-06,   0.000000000000000D+00,   2.769159860520600D-06, &
   0.000000000000000D+00,  -1.789020610736581D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.665656489610244D-05,   0.000000000000000D+00, &
   6.176176998501419D-05,   0.000000000000000D+00,  -5.910602841857933D-05, &
   0.000000000000000D+00,  -1.713847620745555D-05,   0.000000000000000D+00, &
   7.500113686605931D-06,   0.000000000000000D+00,   5.453505514422518D-06, &
   0.000000000000000D+00,  -1.518832238846140D-06,   0.000000000000000D+00, &
  -4.984427668146395D-07,   0.000000000000000D+00,   2.024491045792465D-06 /)
lem(:,116) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.052799047524740D-04,   3.178716562858533D-03,  -7.031614929324240D-03, &
   7.735725492667849D-04,   1.367686969506162D-02,  -9.128157839757877D-03, &
  -1.065125269008881D-02,   9.102294503850066D-03,   5.161776167423350D-03, &
  -3.208278493160690D-03,  -2.610130570601622D-03,  -1.734705743263165D-04, &
   1.186679662826843D-03,   7.054616271289605D-04,  -2.678745864614362D-04, &
  -4.760903931630449D-04,  -4.626347606537214D-05,   2.335918598380348D-04, &
   9.800143764140272D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.052799047524740D-04,  -3.178716562858533D-03, &
   7.031614929324243D-03,  -7.735725492667845D-04,  -1.367686969506163D-02, &
   9.128157839757883D-03,   1.065125269008881D-02,  -9.102294503850085D-03, &
  -5.161776167423327D-03,   3.208278493160704D-03,   2.610130570601610D-03, &
   1.734705743263185D-04,  -1.186679662826831D-03,  -7.054616271289631D-04, &
   2.678745864614356D-04,   4.760903931630427D-04,   4.626347606537292D-05, &
  -2.335918598380220D-04,  -9.800143764140661D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.878276600502713D-06, &
   0.000000000000000D+00,   3.482689063006167D-05,   0.000000000000000D+00, &
   2.921726720565467D-05,   0.000000000000000D+00,  -4.918455864280479D-05, &
   0.000000000000000D+00,  -2.534687733216428D-05,   0.000000000000000D+00, &
   4.013769308026331D-05,   0.000000000000000D+00,  -4.103689454008210D-07, &
   0.000000000000000D+00,  -1.185435077423712D-05,   0.000000000000000D+00, &
   5.578425286868491D-06,   0.000000000000000D+00,  -1.836855949819978D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.835301952949592D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.358119005177840D-05,   0.000000000000000D+00, &
  -3.264709147061776D-07,   0.000000000000000D+00,   7.503561683128255D-05, &
   0.000000000000000D+00,   4.357527577341718D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.037980366153741D-04,   0.000000000000000D+00, &
  -7.687530907345019D-06,   0.000000000000000D+00,  -3.052073485510063D-06, &
   0.000000000000000D+00,   2.740522864361873D-04,   0.000000000000000D+00, &
  -3.125725814801451D-05,   0.000000000000000D+00,   9.450473846759368D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.961670596345511D-05, &
   0.000000000000000D+00,   1.622818186285679D-04,   0.000000000000000D+00, &
   1.515727134922171D-05,   0.000000000000000D+00,  -6.158916762631271D-07, &
   0.000000000000000D+00,  -4.440906904226327D-04,   0.000000000000000D+00, &
  -2.801131654274743D-04,   0.000000000000000D+00,  -2.526974528589816D-05, &
   0.000000000000000D+00,  -7.366863816251419D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.433475275715140D-04,   0.000000000000000D+00, &
   1.085225927061972D-04,   0.000000000000000D+00,  -3.975063538934836D-05, &
   0.000000000000000D+00,   3.310817560716655D-06,   0.000000000000000D+00, &
   7.865786034533440D-06,   0.000000000000000D+00,  -4.852978941012324D-04, &
   0.000000000000000D+00,   3.760397332883542D-04,   0.000000000000000D+00, &
   1.587500308727931D-04,   0.000000000000000D+00,   2.456358228501178D-05, &
   0.000000000000000D+00,   1.058704624334308D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.204770237614524D-05,   0.000000000000000D+00, &
  -6.840402758426728D-04,   0.000000000000000D+00,  -1.227833203873044D-04, &
   0.000000000000000D+00,  -2.602178324011682D-05,   0.000000000000000D+00, &
  -1.215311559486417D-05,   0.000000000000000D+00,  -1.497998012133863D-06, &
   0.000000000000000D+00,   3.708868867032818D-04,   0.000000000000000D+00, &
   4.128723387708450D-04,   0.000000000000000D+00,  -1.544380000430616D-04, &
   0.000000000000000D+00,  -4.165442649760864D-05,   0.000000000000000D+00, &
   8.217540702145527D-07,   0.000000000000000D+00,   4.720582629211378D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.139519728562671D-04, &
   0.000000000000000D+00,   1.312288404859991D-04,   0.000000000000000D+00, &
   3.507680642894001D-04,   0.000000000000000D+00,   7.578841119694536D-05, &
   0.000000000000000D+00,   2.211429848792713D-05,   0.000000000000000D+00, &
   3.212908999932380D-06,   0.000000000000000D+00,  -7.712809431711934D-06, &
   0.000000000000000D+00,   1.373445666964543D-04,   0.000000000000000D+00, &
  -3.805266222285269D-04,   0.000000000000000D+00,  -1.712325585799065D-04, &
   0.000000000000000D+00,   2.449385906457688D-05,   0.000000000000000D+00, &
  -5.690525519257561D-06,   0.000000000000000D+00,  -8.443207133725536D-06, &
   0.000000000000000D+00,  -2.908465296741336D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.234487724440900D-05,   0.000000000000000D+00, &
   2.563215409725594D-04,   0.000000000000000D+00,  -1.479226127679767D-04, &
   0.000000000000000D+00,  -1.136223075538826D-04,   0.000000000000000D+00, &
  -2.696631946073358D-05,   0.000000000000000D+00,  -4.598103956431379D-06, &
   0.000000000000000D+00,   3.996554039424922D-06,   0.000000000000000D+00, &
   4.518186891233797D-06,   0.000000000000000D+00,   3.262696454201266D-06, &
   0.000000000000000D+00,  -6.618544896497200D-05,   0.000000000000000D+00, &
   2.191919177592675D-04,   0.000000000000000D+00,   2.785309057952044D-05, &
   0.000000000000000D+00,   4.393427082728741D-06,   0.000000000000000D+00, &
   8.410618463298417D-06,   0.000000000000000D+00,   3.819121013800688D-06, &
   0.000000000000000D+00,   8.219981304088249D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.838219334165535D-06,   0.000000000000000D+00,  -7.173652426428016D-05, &
   0.000000000000000D+00,   9.239158225615444D-05,   0.000000000000000D+00, &
   2.813976843765110D-05,   0.000000000000000D+00,   6.296355646194695D-06, &
   0.000000000000000D+00,  -1.609180519410166D-06,   0.000000000000000D+00, &
  -2.850137078208944D-06,   0.000000000000000D+00,   1.634108079420494D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.079416467784902D-05,   0.000000000000000D+00,  -3.622366004979266D-05, &
   0.000000000000000D+00,  -6.111836504641266D-05,   0.000000000000000D+00, &
   4.670010634386026D-06,   0.000000000000000D+00,  -3.464737136987723D-06, &
   0.000000000000000D+00,  -3.521124279803820D-06,   0.000000000000000D+00, &
  -1.279506513696729D-06,   0.000000000000000D+00,   2.787028238688755D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.497487026968932D-06,   0.000000000000000D+00, &
   3.321294727099874D-05,   0.000000000000000D+00,  -3.411008151322505D-05, &
   0.000000000000000D+00,  -3.351444101242739D-05,   0.000000000000000D+00, &
  -8.059035515038112D-06,   0.000000000000000D+00,  -1.397768736765282D-06, &
   0.000000000000000D+00,   1.664391322579362D-06,   0.000000000000000D+00, &
  -2.319708956025234D-07,   0.000000000000000D+00,  -2.339820722083717D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.800838951993302D-07,   0.000000000000000D+00,  -4.260642468079286D-06, &
   0.000000000000000D+00,   4.557153833568637D-05,   0.000000000000000D+00, &
  -5.532425822184630D-06,   0.000000000000000D+00,  -3.791925726578216D-06, &
   0.000000000000000D+00,   1.576853778886021D-06,   0.000000000000000D+00, &
   1.497863339640506D-06,   0.000000000000000D+00,   4.384050857441865D-07, &
   0.000000000000000D+00,  -1.196711727161186D-06,   0.000000000000000D+00 /)
lem(:,117) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.228546099521398D-04,  -1.072056069442807D-03, &
  -5.696218756987487D-04,   5.737071564174548D-03,   4.152526259457918D-03, &
  -1.422632648981967D-02,  -1.013497790785842D-02,   1.848045868303768D-02, &
   1.041346617180700D-02,  -1.138798700033549D-02,  -5.662789163209289D-03, &
   2.730214343265721D-03,   1.847460742059178D-03,  -6.729318453565987D-05, &
  -5.048512553458406D-04,   1.106417966592880D-05,   2.337541423728226D-04, &
   1.058184542092296D-05,  -1.327218977993729D-04,  -2.587708889353168D-05, &
   7.514538310501191D-05,   0.000000000000000D+00,   1.228546099521410D-04, &
  -1.072056069442809D-03,  -5.696218756987457D-04,   5.737071564174537D-03, &
   4.152526259457922D-03,  -1.422632648981966D-02,  -1.013497790785843D-02, &
   1.848045868303771D-02,   1.041346617180703D-02,  -1.138798700033549D-02, &
  -5.662789163209290D-03,   2.730214343265721D-03,   1.847460742059177D-03, &
  -6.729318453567516D-05,  -5.048512553458395D-04,   1.106417966595217D-05, &
   2.337541423728189D-04,   1.058184542090169D-05,  -1.327218977993672D-04, &
  -2.587708889351637D-05,   7.514538310501531D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.229865778844100D-05,   0.000000000000000D+00, &
  -1.865104305976683D-04,   0.000000000000000D+00,  -5.198589640526537D-05, &
   0.000000000000000D+00,   1.812946428783066D-04,   0.000000000000000D+00, &
   3.970987611237475D-05,   0.000000000000000D+00,  -1.037125519371164D-04, &
   0.000000000000000D+00,   2.143429733637880D-06,   0.000000000000000D+00, &
   2.476578395165703D-05,   0.000000000000000D+00,  -7.718681899959866D-06, &
   0.000000000000000D+00,   1.590470379488270D-06,   0.000000000000000D+00, &
   1.942501942501346D-06,  -2.081252081252021D-05,   0.000000000000000D+00, &
   2.074722662957926D-05,   0.000000000000000D+00,   7.508831038245610D-07, &
   2.984760337701512D-04,   0.000000000000000D+00,   8.814714697064608D-07, &
   0.000000000000000D+00,  -1.006337144108041D-03,   0.000000000000000D+00, &
   1.890094769351842D-06,   0.000000000000000D+00,   3.305303227903682D-06, &
   8.246990367733398D-04,   0.000000000000000D+00,  -1.328942815010928D-04, &
   0.000000000000000D+00,  -1.213884727816445D-06,   0.000000000000000D+00, &
   1.215507059160311D-03,   0.000000000000000D+00,   4.384796490059653D-04, &
   0.000000000000000D+00,   2.803139413046432D-06,   0.000000000000000D+00, &
   8.446361000541473D-07,  -2.782983270599366D-03,   0.000000000000000D+00, &
  -3.607623886261663D-04,   0.000000000000000D+00,   5.576316528328915D-05, &
   0.000000000000000D+00,   1.670478644162797D-06,   0.000000000000000D+00, &
   1.355841553041712D-03,   0.000000000000000D+00,  -5.516485503697782D-04, &
   0.000000000000000D+00,  -2.295908622046057D-04,   0.000000000000000D+00, &
  -1.399423038859705D-05,   0.000000000000000D+00,  -8.565773766707232D-06, &
   1.056158137191922D-03,   0.000000000000000D+00,   1.256611764419008D-03, &
   0.000000000000000D+00,   2.170597393288872D-04,   0.000000000000000D+00, &
   4.565094815296772D-06,   0.000000000000000D+00,   5.869689335404335D-06, &
   0.000000000000000D+00,  -1.464014918988669D-03,   0.000000000000000D+00, &
  -6.125271208054881D-04,   0.000000000000000D+00,   2.805524957311871D-04, &
   0.000000000000000D+00,   9.081530662917788D-05,   0.000000000000000D+00, &
   7.382336836850029D-06,   0.000000000000000D+00,   9.277775367067000D-07, &
   5.137647384248548D-04,   0.000000000000000D+00,  -4.617175797684616D-04, &
   0.000000000000000D+00,  -6.799905376850459D-04,   0.000000000000000D+00, &
  -1.319057334545623D-04,   0.000000000000000D+00,  -2.482180131768104D-05, &
   0.000000000000000D+00,  -5.128889188349966D-06,   0.000000000000000D+00, &
   2.831559191769099D-05,   0.000000000000000D+00,   6.313747573674878D-04, &
   0.000000000000000D+00,   3.107534132907667D-04,   0.000000000000000D+00, &
  -9.163289635089349D-05,   0.000000000000000D+00,  -1.188133786760143D-05, &
   0.000000000000000D+00,   5.215304633652438D-06,   0.000000000000000D+00, &
   8.448498771177061D-06,  -4.583068925510663D-05,   0.000000000000000D+00, &
  -2.204478341239682D-04,   0.000000000000000D+00,   2.742364212615925D-04, &
   0.000000000000000D+00,   2.938411926835674D-04,   0.000000000000000D+00, &
   6.367846110246041D-05,   0.000000000000000D+00,   1.267338046398571D-05, &
   0.000000000000000D+00,  -1.959494720490260D-06,   0.000000000000000D+00, &
   4.720034203744557D-06,   0.000000000000000D+00,   5.308700960875072D-06, &
   0.000000000000000D+00,  -3.244001365745442D-04,   0.000000000000000D+00, &
  -1.021877582774766D-04,   0.000000000000000D+00,   7.706689692845068D-06, &
   0.000000000000000D+00,  -1.125153266690853D-05,   0.000000000000000D+00, &
  -7.699902588065969D-06,   0.000000000000000D+00,  -3.940598375707331D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.779738171604469D-06, &
   0.000000000000000D+00,   9.022009602214853D-05,   0.000000000000000D+00, &
  -1.638232714809364D-04,   0.000000000000000D+00,  -9.389315494433712D-05, &
   0.000000000000000D+00,  -2.089228083703511D-05,   0.000000000000000D+00, &
   5.362920449022367D-07,   0.000000000000000D+00,   3.038089836106778D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.984678341325602D-06,   0.000000000000000D+00,  -4.112297368041471D-07, &
   0.000000000000000D+00,   1.274574100468043D-04,   0.000000000000000D+00, &
   9.063996074708432D-06,   0.000000000000000D+00,   6.955396264001509D-06, &
   0.000000000000000D+00,   8.563572731788466D-06,   0.000000000000000D+00, &
   3.574744532148957D-06,   0.000000000000000D+00,  -2.525389717162316D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.006048523012538D-06, &
   0.000000000000000D+00,   8.259482023470951D-06,   0.000000000000000D+00, &
   2.736254705700860D-06,   0.000000000000000D+00,   7.898784475170752D-05, &
   0.000000000000000D+00,   2.582140127132252D-05,   0.000000000000000D+00, &
   4.546357268700668D-06,   0.000000000000000D+00,  -2.726642024120173D-06, &
   0.000000000000000D+00,  -1.088967840066522D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.433609048058881D-07, &
   0.000000000000000D+00,  -3.551459560194294D-06,   0.000000000000000D+00, &
  -1.691894950413843D-05,   0.000000000000000D+00,  -1.846881704856498D-05, &
   0.000000000000000D+00,   7.651438223117667D-06,   0.000000000000000D+00, &
  -4.186765650264331D-06,   0.000000000000000D+00,  -4.264550698415521D-06, &
   0.000000000000000D+00,   1.386961410166760D-08,   0.000000000000000D+00, &
   2.494923797719321D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.430056025065901D-06, &
   0.000000000000000D+00,  -5.516119394493206D-07,   0.000000000000000D+00, &
  -3.439842876744588D-05,   0.000000000000000D+00,  -2.628111377632027D-05, &
   0.000000000000000D+00,  -8.559586546576961D-06,   0.000000000000000D+00, &
  -7.093200227700261D-07,   0.000000000000000D+00,   1.003575168875261D-06, &
   0.000000000000000D+00,   5.437873331765982D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.340289769804136D-06,   0.000000000000000D+00, &
  -3.331858289556837D-06,   0.000000000000000D+00,   9.857453178522480D-06, &
   0.000000000000000D+00,  -1.482209317175120D-05,   0.000000000000000D+00, &
  -4.785258581947926D-06,   0.000000000000000D+00,   2.224032166834955D-06, &
   0.000000000000000D+00,   1.882637255182735D-06,   0.000000000000000D+00, &
  -9.476404129725543D-07,   0.000000000000000D+00,  -7.384753372477850D-07 /)
lem(:,118) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.861144884210415D-04,  -2.465988537459237D-04, &
   5.715817904807867D-03,  -5.754891885488829D-03,  -1.059452631115616D-02, &
   1.389501324461849D-02,   1.092392435378868D-02,  -1.695022896495577D-02, &
  -5.741049518203379D-03,   1.122046763014864D-02,   1.549900628915546D-03, &
  -4.076715304335722D-03,  -5.493726349295756D-04,   9.861015283774766D-04, &
   3.275635745205557D-04,  -2.794880703424837D-04,  -1.325867060048824D-04, &
   1.046221071693343D-04,   5.484476332068984D-05,  -5.090549540347070D-05, &
  -2.605641400340330D-05,   0.000000000000000D+00,   3.861144884210469D-04, &
   2.465988537459286D-04,  -5.715817904807874D-03,   5.754891885488836D-03, &
   1.059452631115616D-02,  -1.389501324461848D-02,  -1.092392435378869D-02, &
   1.695022896495581D-02,   5.741049518203389D-03,  -1.122046763014867D-02, &
  -1.549900628915558D-03,   4.076715304335740D-03,   5.493726349295685D-04, &
  -9.861015283774879D-04,  -3.275635745205413D-04,   2.794880703424880D-04, &
   1.325867060048724D-04,  -1.046221071693325D-04,  -5.484476332068431D-05, &
   5.090549540347290D-05,   2.605641400339634D-05,   0.000000000000000D+00, &
   1.930572442105254D-04,   0.000000000000000D+00,   4.325663508345734D-04, &
   0.000000000000000D+00,   7.611537417576769D-05,   0.000000000000000D+00, &
  -2.718239073720461D-04,   0.000000000000000D+00,  -1.155774058875463D-04, &
   0.000000000000000D+00,   1.433009342827700D-04,   0.000000000000000D+00, &
   5.060139649022129D-05,   0.000000000000000D+00,  -5.178661833427089D-05, &
   0.000000000000000D+00,   3.322341818790479D-06,   0.000000000000000D+00, &
   3.507499220139913D-06,   0.000000000000000D+00,  -2.945529961698765D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.526251526251514D-05, &
   0.000000000000000D+00,  -1.398927869516112D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.177417353887942D-04,   0.000000000000000D+00, &
   1.229707112060041D-05,   0.000000000000000D+00,  -2.194425801236947D-04, &
   0.000000000000000D+00,   6.788876757917070D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.793665914408964D-05,   0.000000000000000D+00, &
  -1.147487171370409D-04,   0.000000000000000D+00,  -8.215826714278645D-06, &
   0.000000000000000D+00,   3.072134264084724D-04,   0.000000000000000D+00, &
   2.437588947982577D-04,   0.000000000000000D+00,   1.385353552536212D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.631253851068088D-04, &
   0.000000000000000D+00,  -1.171030306368653D-04,   0.000000000000000D+00, &
   5.246695455673779D-05,   0.000000000000000D+00,   1.404415587852082D-06, &
   0.000000000000000D+00,   3.405875693532173D-04,   0.000000000000000D+00, &
  -3.279362984572301D-04,   0.000000000000000D+00,  -1.721543420459829D-04, &
   0.000000000000000D+00,  -2.280512800939501D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.085225927061972D-04,   0.000000000000000D+00, &
   6.230816565316494D-04,   0.000000000000000D+00,   1.182811505571514D-04, &
   0.000000000000000D+00,   5.729057944194545D-06,   0.000000000000000D+00, &
   1.076491941971605D-05,   0.000000000000000D+00,  -3.128506527348902D-04, &
   0.000000000000000D+00,  -3.628614397223898D-04,   0.000000000000000D+00, &
   2.044670228285724D-04,   0.000000000000000D+00,   7.705397625810862D-05, &
   0.000000000000000D+00,   8.087552183378065D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.089762737225386D-04,   0.000000000000000D+00, &
  -1.443179246382906D-04,   0.000000000000000D+00,  -4.315345200623858D-04, &
   0.000000000000000D+00,  -9.038747891554467D-05,   0.000000000000000D+00, &
  -2.201196139883418D-05,   0.000000000000000D+00,  -1.069840312062507D-05, &
   0.000000000000000D+00,  -5.269732137412866D-05,   0.000000000000000D+00, &
   3.462910586858564D-04,   0.000000000000000D+00,   2.266910888949523D-04, &
   0.000000000000000D+00,  -7.127379050333017D-05,   0.000000000000000D+00, &
  -1.284173575484446D-05,   0.000000000000000D+00,   6.933294466998921D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.736182380010295D-06, &
   0.000000000000000D+00,  -2.043691063555303D-04,   0.000000000000000D+00, &
   1.438383597826319D-04,   0.000000000000000D+00,   2.004065070323087D-04, &
   0.000000000000000D+00,   4.634350746951066D-05,   0.000000000000000D+00, &
   1.086222571123381D-05,   0.000000000000000D+00,  -2.241016804371701D-06, &
   0.000000000000000D+00,   8.946924640780297D-06,   0.000000000000000D+00, &
   3.554527683197570D-05,   0.000000000000000D+00,  -2.434332279326576D-04, &
   0.000000000000000D+00,  -8.002596214609605D-05,   0.000000000000000D+00, &
   7.826033066620664D-06,   0.000000000000000D+00,  -8.433308288596022D-06, &
   0.000000000000000D+00,  -6.795364323794181D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.042322492004711D-06,   0.000000000000000D+00, &
   1.297244566399429D-05,   0.000000000000000D+00,   1.007272391955438D-04, &
   0.000000000000000D+00,  -1.086098761948820D-04,   0.000000000000000D+00, &
  -6.447046829908127D-05,   0.000000000000000D+00,  -1.456160442621458D-05, &
   0.000000000000000D+00,  -5.919895899732100D-07,   0.000000000000000D+00, &
   7.918062117281035D-06,   0.000000000000000D+00,   1.464030460218951D-07, &
   0.000000000000000D+00,  -4.026074114868480D-06,   0.000000000000000D+00, &
   8.938521253820577D-06,   0.000000000000000D+00,   1.096208912474865D-04, &
   0.000000000000000D+00,   9.110223392120265D-06,   0.000000000000000D+00, &
   4.100113818477097D-06,   0.000000000000000D+00,   7.159750631222013D-06, &
   0.000000000000000D+00,   2.441247380080189D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.991513728280008D-06,   0.000000000000000D+00,  -1.364928895622291D-05, &
   0.000000000000000D+00,  -3.635688015320572D-06,   0.000000000000000D+00, &
   5.657290722128429D-05,   0.000000000000000D+00,   1.693420485700574D-05, &
   0.000000000000000D+00,   3.332965081997447D-06,   0.000000000000000D+00, &
  -1.954136373323255D-06,   0.000000000000000D+00,  -3.799439605714163D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.123234182376185D-06,   0.000000000000000D+00,  -2.984975006588870D-06, &
   0.000000000000000D+00,  -3.052836081705153D-05,   0.000000000000000D+00, &
  -2.075527253173002D-05,   0.000000000000000D+00,   4.503230546167890D-06, &
   0.000000000000000D+00,  -3.185140417885023D-06,   0.000000000000000D+00, &
  -2.930379375338365D-06,   0.000000000000000D+00,  -9.702093289739447D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.845162356885817D-07,   0.000000000000000D+00, &
   3.268645211920744D-06,   0.000000000000000D+00,   4.213619880241559D-06, &
   0.000000000000000D+00,  -2.994639012669679D-05,   0.000000000000000D+00, &
  -1.913834648394661D-05,   0.000000000000000D+00,  -5.150205752212546D-06, &
   0.000000000000000D+00,  -1.291127516255197D-06,   0.000000000000000D+00, &
   1.111468556265136D-06,   0.000000000000000D+00,   1.002860903841435D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.012126528306924D-08,   0.000000000000000D+00,   2.344655069519129D-07, &
   0.000000000000000D+00,   6.763632317583372D-06,   0.000000000000000D+00, &
   1.571734630122343D-05,   0.000000000000000D+00,  -8.639093250299704D-06, &
   0.000000000000000D+00,  -2.733278390431324D-06,   0.000000000000000D+00, &
   1.651865704418265D-06,   0.000000000000000D+00,   1.133128106788217D-06, &
   0.000000000000000D+00,   1.804379977310323D-07,   0.000000000000000D+00 /)
lem(:,119) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.588205062655805D-04,  -1.908634118899499D-03,  -2.128537474438497D-03, &
   4.440972982271669D-05,   9.122147830673610D-03,   5.328259635831445D-03, &
  -1.702353256082765D-02,  -6.842922961441853D-03,   1.671232551432137D-02, &
   1.848400272083743D-03,  -8.755866741736437D-03,   1.711613247769198D-03, &
   2.515169494283659D-03,  -1.325724048058817D-03,  -6.086202291050018D-04, &
   4.579113057140841D-04,   2.311660375399540D-04,  -1.741178255502510D-04, &
  -9.599159266162110D-05,   7.520064658568084D-05,   4.207023261260506D-05, &
  -3.686681350061320D-05,   4.588205062655779D-04,  -1.908634118899502D-03, &
  -2.128537474438507D-03,   4.440972982271805D-05,   9.122147830673619D-03, &
   5.328259635831447D-03,  -1.702353256082764D-02,  -6.842922961441860D-03, &
   1.671232551432142D-02,   1.848400272083748D-03,  -8.755866741736479D-03, &
   1.711613247769195D-03,   2.515169494283667D-03,  -1.325724048058822D-03, &
  -6.086202291050056D-04,   4.579113057140843D-04,   2.311660375399565D-04, &
  -1.741178255502457D-04,  -9.599159266162939D-05,   7.520064658566661D-05, &
   4.207023261262708D-05,  -3.686681350060012D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.749021260414733D-04,   0.000000000000000D+00, &
  -6.860153769109627D-04,   0.000000000000000D+00,   8.398781795664702D-06, &
   0.000000000000000D+00,   5.008316146074569D-04,   0.000000000000000D+00, &
  -1.301237969410205D-04,   0.000000000000000D+00,  -2.095530660436700D-04, &
   0.000000000000000D+00,   8.861874557851224D-05,   0.000000000000000D+00, &
   1.566672458499832D-05,   0.000000000000000D+00,  -1.229958607458120D-05, &
   0.000000000000000D+00,   5.979255450154043D-06,   0.000000000000000D+00, &
  -5.827505827505824D-05,   3.052503052503065D-05,   0.000000000000000D+00, &
   2.328553799142033D-04,   0.000000000000000D+00,   8.141368435486064D-06, &
  -3.477731418907888D-04,   0.000000000000000D+00,   3.878882555353142D-05, &
   0.000000000000000D+00,  -2.454481363150097D-04,   0.000000000000000D+00, &
  -1.729900646309314D-04,   0.000000000000000D+00,  -7.683745348218991D-06, &
   7.929656342969038D-04,   0.000000000000000D+00,   1.637828484356570D-04, &
   0.000000000000000D+00,  -6.344914795959669D-05,   0.000000000000000D+00, &
  -2.853055097637141D-04,   0.000000000000000D+00,   2.256524363722507D-04, &
   0.000000000000000D+00,   1.747958210806508D-04,   0.000000000000000D+00, &
   3.520774702808369D-05,  -4.711963303294563D-04,   0.000000000000000D+00, &
  -5.431386970755107D-04,   0.000000000000000D+00,  -1.211062308407945D-04, &
   0.000000000000000D+00,   3.211627419493130D-05,   0.000000000000000D+00, &
   3.952042483473364D-04,   0.000000000000000D+00,   2.518990885948083D-04, &
   0.000000000000000D+00,  -2.335526368754307D-04,   0.000000000000000D+00, &
  -1.321253125179253D-04,   0.000000000000000D+00,  -4.289905576118179D-05, &
   2.289069121751889D-05,   0.000000000000000D+00,   2.170597393288872D-04, &
   0.000000000000000D+00,   4.971101988937763D-04,   0.000000000000000D+00, &
   1.040729849554241D-04,   0.000000000000000D+00,   9.707245911805313D-06, &
   0.000000000000000D+00,  -1.050655869608626D-04,   0.000000000000000D+00, &
  -2.603366481411977D-04,   0.000000000000000D+00,  -2.560942186604900D-04, &
   0.000000000000000D+00,   1.521563814979838D-04,   0.000000000000000D+00, &
   6.084306382007429D-05,   0.000000000000000D+00,   1.353712780233446D-05, &
   1.505583750838867D-05,   0.000000000000000D+00,   7.261886047725312D-05, &
   0.000000000000000D+00,  -1.509462692325230D-04,   0.000000000000000D+00, &
  -3.428611415971338D-04,   0.000000000000000D+00,  -7.625232494555323D-05, &
   0.000000000000000D+00,  -1.985055850401032D-05,   0.000000000000000D+00, &
   1.170317918870848D-05,   0.000000000000000D+00,   4.887471836495331D-07, &
   0.000000000000000D+00,   2.380398274550414D-04,   0.000000000000000D+00, &
   1.658586449109706D-04,   0.000000000000000D+00,  -5.342157758403950D-05, &
   0.000000000000000D+00,  -7.407803242058977D-06,   0.000000000000000D+00, &
   1.598899915446957D-05,  -3.446204314423375D-06,   0.000000000000000D+00, &
   5.858518696720213D-06,   0.000000000000000D+00,  -1.108265429779933D-04, &
   0.000000000000000D+00,   1.225940035855684D-04,   0.000000000000000D+00, &
   1.614591326736208D-04,   0.000000000000000D+00,   3.824567232560690D-05, &
   0.000000000000000D+00,   7.781491300240835D-06,   0.000000000000000D+00, &
   2.527116214857028D-07,   0.000000000000000D+00,  -2.576624927117518D-06, &
   0.000000000000000D+00,   2.861175016959822D-05,   0.000000000000000D+00, &
  -1.704283164992861D-04,   0.000000000000000D+00,  -5.820220652865165D-05, &
   0.000000000000000D+00,   3.804308282881087D-06,   0.000000000000000D+00, &
  -1.009534323885802D-05,   0.000000000000000D+00,  -1.676539968300050D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.976627526565405D-06, &
   0.000000000000000D+00,  -5.200725570755341D-06,   0.000000000000000D+00, &
   6.670994478461571D-05,   0.000000000000000D+00,  -8.903295867909655D-05, &
   0.000000000000000D+00,  -5.240420937581813D-05,   0.000000000000000D+00, &
  -1.217319718386052D-05,   0.000000000000000D+00,  -4.482330965934682D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.062502366981418D-06,   0.000000000000000D+00,  -3.750658260266595D-06, &
   0.000000000000000D+00,  -4.337834968357184D-06,   0.000000000000000D+00, &
   7.817888578916733D-05,   0.000000000000000D+00,   5.061225751834691D-06, &
   0.000000000000000D+00,   5.245448267263974D-06,   0.000000000000000D+00, &
   8.045959725861842D-06,   0.000000000000000D+00,   4.347685011747555D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.726507641286418D-07, &
   0.000000000000000D+00,   2.867448474989186D-06,   0.000000000000000D+00, &
  -7.567107847206264D-06,   0.000000000000000D+00,  -2.800233706103405D-06, &
   0.000000000000000D+00,   4.580970062357324D-05,   0.000000000000000D+00, &
   1.451984806030163D-05,   0.000000000000000D+00,   3.216347435852936D-06, &
   0.000000000000000D+00,   1.717157173699538D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.432051988265465D-09, &
   0.000000000000000D+00,   1.949647826667213D-07,   0.000000000000000D+00, &
   5.183617731247332D-06,   0.000000000000000D+00,  -1.841302069917897D-05, &
   0.000000000000000D+00,  -1.279568776237766D-05,   0.000000000000000D+00, &
   3.942967856716197D-06,   0.000000000000000D+00,  -3.310040802136255D-06, &
   0.000000000000000D+00,  -2.925690387467039D-06,   0.000000000000000D+00, &
   1.691395758206347D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.591010272632454D-07, &
   0.000000000000000D+00,   3.377453351561931D-07,   0.000000000000000D+00, &
   7.738798795285721D-06,   0.000000000000000D+00,  -2.238674832294910D-05, &
   0.000000000000000D+00,  -1.517840789503505D-05,   0.000000000000000D+00, &
  -5.412767873998047D-06,   0.000000000000000D+00,  -1.661834343878844D-06, &
   0.000000000000000D+00,   2.240871244055363D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.654406602196844D-07,   0.000000000000000D+00, &
  -1.774834162934798D-06,   0.000000000000000D+00,   8.465744871828291D-07, &
   0.000000000000000D+00,   1.122130008654053D-05,   0.000000000000000D+00, &
  -8.495743304961798D-06,   0.000000000000000D+00,  -2.315369927837792D-06, &
   0.000000000000000D+00,   1.334206623338427D-06,   0.000000000000000D+00, &
   7.373893903028928D-08,   0.000000000000000D+00,  -1.852281752734435D-06 /)
lem(:,120) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.914184398086127D-04,   4.490694915583660D-03, &
   8.520088906728073D-04,  -9.368899943608983D-03,  -4.070825440558472D-03, &
   1.419150264953548D-02,   2.241922654113243D-03,  -1.258261263132704D-02, &
   2.475841023176310D-03,   5.342294621869045D-03,  -3.210117712110478D-03, &
  -4.322892195359949D-04,   1.365807393172654D-03,  -4.378148942479670D-04, &
  -3.617548132284326D-04,   2.360809916848063D-04,   1.206926264383870D-04, &
  -1.155935413060195D-04,  -4.549176277864095D-05,   5.815522585999791D-05, &
   1.824842211034826D-05,   0.000000000000000D+00,  -4.914184398086120D-04, &
  -4.490694915583676D-03,  -8.520088906728051D-04,   9.368899943608981D-03, &
   4.070825440558474D-03,  -1.419150264953548D-02,  -2.241922654113247D-03, &
   1.258261263132705D-02,  -2.475841023176311D-03,  -5.342294621869042D-03, &
   3.210117712110471D-03,   4.322892195359988D-04,  -1.365807393172649D-03, &
   4.378148942479644D-04,   3.617548132284311D-04,  -2.360809916847998D-04, &
  -1.206926264383857D-04,   1.155935413060196D-04,   4.549176277864180D-05, &
  -5.815522586000178D-05,  -1.824842211035062D-05,   0.000000000000000D+00, &
  -2.457092199043029D-04,   0.000000000000000D+00,  -2.392411877934738D-03, &
   0.000000000000000D+00,   2.673032304934878D-03,   0.000000000000000D+00, &
  -1.273597780475902D-03,   0.000000000000000D+00,  -2.968090976790894D-04, &
   0.000000000000000D+00,   5.587567959572777D-04,   0.000000000000000D+00, &
  -6.600182150898460D-05,   0.000000000000000D+00,  -8.715908590729678D-05, &
   0.000000000000000D+00,   3.062919259690281D-05,   0.000000000000000D+00, &
  -1.002933466581737D-05,   0.000000000000000D+00,   3.346294088016439D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.942501942501944D-05, &
   0.000000000000000D+00,   8.292361233537711D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.684126331185155D-05,   0.000000000000000D+00, &
   3.799305269893504D-05,   0.000000000000000D+00,  -9.145910848697217D-05, &
   0.000000000000000D+00,  -1.428943863154390D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.249573134506053D-04,   0.000000000000000D+00, &
   1.038404565604920D-04,   0.000000000000000D+00,  -8.381644048246214D-05, &
   0.000000000000000D+00,  -1.070223206446116D-04,   0.000000000000000D+00, &
   1.689965643968299D-04,   0.000000000000000D+00,   1.666403127316440D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.001556922609555D-04, &
   0.000000000000000D+00,  -3.574693080668916D-04,   0.000000000000000D+00, &
  -9.984595249248373D-05,   0.000000000000000D+00,   4.584872552413202D-05, &
   0.000000000000000D+00,   1.393722638170072D-04,   0.000000000000000D+00, &
   1.848015859636476D-04,   0.000000000000000D+00,  -1.933844639488372D-04, &
   0.000000000000000D+00,  -1.246132538907266D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.975063538934836D-05,   0.000000000000000D+00, &
   1.182811505571514D-04,   0.000000000000000D+00,   3.863684136268387D-04, &
   0.000000000000000D+00,   9.191571981556753D-05,   0.000000000000000D+00, &
   2.848155451261740D-05,   0.000000000000000D+00,  -2.691834869786151D-05, &
   0.000000000000000D+00,  -1.945342454243384D-04,   0.000000000000000D+00, &
  -2.028420981297989D-04,   0.000000000000000D+00,   1.293332700576483D-04, &
   0.000000000000000D+00,   4.942665780695474D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.122845372273306D-05,   0.000000000000000D+00, &
   7.706660646154525D-05,   0.000000000000000D+00,  -1.121232274375682D-04, &
   0.000000000000000D+00,  -2.800867440713502D-04,   0.000000000000000D+00, &
  -7.106580507957288D-05,   0.000000000000000D+00,  -5.374344522660166D-05, &
   0.000000000000000D+00,  -1.915121715902586D-06,   0.000000000000000D+00, &
   9.088481388156503D-06,   0.000000000000000D+00,   1.954031612917932D-04, &
   0.000000000000000D+00,   1.345404445894484D-04,   0.000000000000000D+00, &
  -4.313455266001349D-05,   0.000000000000000D+00,  -6.782767429584996D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.891016991165012D-06, &
   0.000000000000000D+00,  -1.109929737845447D-05,   0.000000000000000D+00, &
  -8.660737713669861D-05,   0.000000000000000D+00,   1.009407923401885D-04, &
   0.000000000000000D+00,   1.344205174876414D-04,   0.000000000000000D+00, &
   3.975083355703614D-05,   0.000000000000000D+00,   2.551148259738401D-05, &
   0.000000000000000D+00,   1.232322964903600D-06,   0.000000000000000D+00, &
  -2.284239812893502D-06,   0.000000000000000D+00,   6.379963301899012D-06, &
   0.000000000000000D+00,  -1.419315535832828D-04,   0.000000000000000D+00, &
  -4.828987666212565D-05,   0.000000000000000D+00,   4.782266032711421D-07, &
   0.000000000000000D+00,  -7.109985276162802D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.921319416597831D-07,   0.000000000000000D+00, &
   2.422596028560011D-06,   0.000000000000000D+00,   9.279404185521454D-07, &
   0.000000000000000D+00,   4.753860941041026D-05,   0.000000000000000D+00, &
  -7.516414684106749D-05,   0.000000000000000D+00,  -4.507503528330680D-05, &
   0.000000000000000D+00,  -1.506720250677385D-05,   0.000000000000000D+00, &
   2.517048110309856D-06,   0.000000000000000D+00,   1.058699712978010D-08, &
   0.000000000000000D+00,   2.296407458045937D-07,   0.000000000000000D+00, &
   2.846933045531750D-06,   0.000000000000000D+00,   5.429009664625525D-06, &
   0.000000000000000D+00,   6.558511972732672D-05,   0.000000000000000D+00, &
   5.075437303649706D-06,   0.000000000000000D+00,   4.387275408524585D-06, &
   0.000000000000000D+00,   2.405150373985143D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.057986959904422D-07,   0.000000000000000D+00,   5.035275226475004D-07, &
   0.000000000000000D+00,  -3.646464585023818D-06,   0.000000000000000D+00, &
   2.474189932679552D-06,   0.000000000000000D+00,   3.931043974709304D-05, &
   0.000000000000000D+00,   1.385683850462090D-05,   0.000000000000000D+00, &
   2.552430117615665D-06,   0.000000000000000D+00,  -7.981520432692344D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.462352389882939D-07,   0.000000000000000D+00,  -1.505347895064731D-06, &
   0.000000000000000D+00,   4.307044469776215D-07,   0.000000000000000D+00, &
  -1.762300970295542D-05,   0.000000000000000D+00,  -1.100532852145090D-05, &
   0.000000000000000D+00,   3.201064223635787D-06,   0.000000000000000D+00, &
  -9.357932502578143D-07,   0.000000000000000D+00,  -1.774401627154302D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.728253246469771D-08,   0.000000000000000D+00, &
   2.619663906282169D-07,   0.000000000000000D+00,  -1.902699573028107D-06, &
   0.000000000000000D+00,   2.906157890866794D-06,   0.000000000000000D+00, &
  -2.106837605139940D-05,   0.000000000000000D+00,  -1.383792632320036D-05, &
   0.000000000000000D+00,  -4.963125672685721D-06,   0.000000000000000D+00, &
   1.351692470141221D-06,   0.000000000000000D+00,   4.364328424936825D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.491631336111697D-10,   0.000000000000000D+00,   2.291084539223320D-07, &
   0.000000000000000D+00,   4.580370314599601D-07,   0.000000000000000D+00, &
   7.913947403377642D-07,   0.000000000000000D+00,   9.084673666498279D-06, &
   0.000000000000000D+00,  -7.178533905221292D-06,   0.000000000000000D+00, &
  -2.770402371971401D-06,   0.000000000000000D+00,   5.198577646265371D-07, &
   0.000000000000000D+00,   2.036331413730174D-06,   0.000000000000000D+00 /)
lem(:,121) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.323782127336613D-03,  -1.942272881148305D-03,  -3.231851030167527D-03, &
   5.213814820462987D-04,   8.247672825828252D-03,   2.444698731405115D-03, &
  -1.085389076775490D-02,  -1.498964295696677D-04,   7.280392139106114D-03, &
  -3.486456756650338D-03,  -1.336884852193465D-03,   2.606314557953085D-03, &
  -1.086227770462394D-03,  -4.235125479521585D-04,   6.812266969301980D-04, &
  -1.850682041952317D-04,  -1.901217247084389D-04,   1.347964371497138D-04, &
   5.170914607492250D-05,  -7.385800752899445D-05,  -1.242086220338126D-05, &
   3.921517966244518D-05,  -1.323782127336625D-03,  -1.942272881148300D-03, &
  -3.231851030167524D-03,   5.213814820462956D-04,   8.247672825828259D-03, &
   2.444698731405115D-03,  -1.085389076775491D-02,  -1.498964295696743D-04, &
   7.280392139106100D-03,  -3.486456756650333D-03,  -1.336884852193470D-03, &
   2.606314557953084D-03,  -1.086227770462388D-03,  -4.235125479521551D-04, &
   6.812266969301934D-04,  -1.850682041952321D-04,  -1.901217247084334D-04, &
   1.347964371497104D-04,   5.170914607491633D-05,  -7.385800752899321D-05, &
  -1.242086220337603D-05,   3.921517966244459D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.660671843048128D-03,   0.000000000000000D+00, &
   3.773084573010297D-03,   0.000000000000000D+00,  -3.783135308292042D-03, &
   0.000000000000000D+00,   3.019609399410917D-03,   0.000000000000000D+00, &
  -1.108671207407664D-03,   0.000000000000000D+00,  -2.577878235072896D-04, &
   0.000000000000000D+00,   2.703167364164190D-04,   0.000000000000000D+00, &
  -4.842713939735434D-05,   0.000000000000000D+00,   8.539058449411944D-06, &
   0.000000000000000D+00,  -8.971169026884724D-07,   0.000000000000000D+00, &
   5.035473785473783D-05,   9.833916083916093D-05,   0.000000000000000D+00, &
  -9.409843968667558D-06,   0.000000000000000D+00,   9.963654265124858D-05, &
  -1.525333326803918D-04,   0.000000000000000D+00,  -9.020289351171701D-05, &
   0.000000000000000D+00,   1.018932907477784D-05,   0.000000000000000D+00, &
  -6.250002863779938D-05,   0.000000000000000D+00,  -2.350902357553866D-04, &
   1.490855206025485D-04,   0.000000000000000D+00,   1.879902525358076D-04, &
   0.000000000000000D+00,   1.153568935471993D-04,   0.000000000000000D+00, &
  -2.669508280963383D-05,   0.000000000000000D+00,  -6.319574818026826D-05, &
   0.000000000000000D+00,   1.453882576294534D-04,   0.000000000000000D+00, &
   2.992308187741626D-04,  -8.703564237619982D-05,   0.000000000000000D+00, &
  -1.047048067171906D-04,   0.000000000000000D+00,  -3.015309694375236D-04, &
   0.000000000000000D+00,  -1.319282658775695D-04,   0.000000000000000D+00, &
   2.643570060954649D-05,   0.000000000000000D+00,   8.001775115871901D-05, &
   0.000000000000000D+00,   1.392281142273782D-04,   0.000000000000000D+00, &
  -1.742473145163403D-04,   0.000000000000000D+00,  -2.549593734695845D-04, &
   2.662239104728005D-05,   0.000000000000000D+00,   4.565094815296772D-06, &
   0.000000000000000D+00,   1.040729849554241D-04,   0.000000000000000D+00, &
   3.432014190396673D-04,   0.000000000000000D+00,   1.253916578627554D-04, &
   0.000000000000000D+00,  -1.182852403575285D-05,   0.000000000000000D+00, &
  -1.409430704675697D-05,   0.000000000000000D+00,  -1.398318844491572D-04, &
   0.000000000000000D+00,  -1.691915715111344D-04,   0.000000000000000D+00, &
   1.224504929255934D-04,   0.000000000000000D+00,   1.275351424224953D-04, &
  -2.520040757017024D-06,   0.000000000000000D+00,   9.820391727445471D-07, &
   0.000000000000000D+00,   4.998216387650955D-05,   0.000000000000000D+00, &
  -9.914914271275765D-05,   0.000000000000000D+00,  -2.610332438104029D-04, &
   0.000000000000000D+00,  -8.243564104772869D-05,   0.000000000000000D+00, &
   2.017482134814933D-06,   0.000000000000000D+00,  -1.593878088984073D-06, &
   0.000000000000000D+00,   1.256692109788954D-06,   0.000000000000000D+00, &
   1.572457109653952D-04,   0.000000000000000D+00,   1.183437083977499D-04, &
   0.000000000000000D+00,  -4.775605331308692D-05,   0.000000000000000D+00, &
  -1.009387104759381D-05,  -3.538791565444828D-07,   0.000000000000000D+00, &
   1.462068614444325D-06,   0.000000000000000D+00,  -7.532500510294511D-06, &
   0.000000000000000D+00,  -6.776606861877222D-05,   0.000000000000000D+00, &
   9.221269538355264D-05,   0.000000000000000D+00,   1.302226594071422D-04, &
   0.000000000000000D+00,   3.034573042123180D-05,   0.000000000000000D+00, &
   2.005647789598198D-08,   0.000000000000000D+00,   4.134373893186849D-07, &
   0.000000000000000D+00,   3.625491979785707D-06,   0.000000000000000D+00, &
   4.544995340002747D-06,   0.000000000000000D+00,  -1.213814532000456D-04, &
   0.000000000000000D+00,  -4.202678460051083D-05,   0.000000000000000D+00, &
   4.547085538392113D-06,   0.000000000000000D+00,  -3.154761986423965D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.119497738706097D-07, &
   0.000000000000000D+00,   5.014451830626365D-07,   0.000000000000000D+00, &
   6.085105001216345D-06,   0.000000000000000D+00,   3.804465924695407D-05, &
   0.000000000000000D+00,  -7.138852836490812D-05,   0.000000000000000D+00, &
  -4.193110513455487D-05,   0.000000000000000D+00,  -6.099700345909412D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.417220222879507D-07,   0.000000000000000D+00,  -1.243094989786318D-06, &
   0.000000000000000D+00,  -1.268393265360576D-06,   0.000000000000000D+00, &
   8.266603341730587D-06,   0.000000000000000D+00,   5.768666766168369D-05, &
   0.000000000000000D+00,   1.566447577307091D-06,   0.000000000000000D+00, &
   8.201613837450138D-06,   0.000000000000000D+00,   2.154180272886029D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.646174437777956D-08, &
   0.000000000000000D+00,   1.774762363875138D-07,   0.000000000000000D+00, &
  -1.296509036924955D-06,   0.000000000000000D+00,  -7.498784082909457D-06, &
   0.000000000000000D+00,   4.862572647822637D-06,   0.000000000000000D+00, &
   3.782468359638852D-05,   0.000000000000000D+00,   9.680146759398490D-06, &
   0.000000000000000D+00,   4.053215351204212D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.077388190019304D-10, &
   0.000000000000000D+00,   2.025124152393867D-07,   0.000000000000000D+00, &
   2.051432758603117D-07,   0.000000000000000D+00,   3.085207256339404D-07, &
   0.000000000000000D+00,  -1.829365795301037D-05,   0.000000000000000D+00, &
  -8.247078084733821D-06,   0.000000000000000D+00,   3.922552855701597D-06, &
   0.000000000000000D+00,  -7.100054019893413D-06,   0.000000000000000D+00, &
  -7.196185405967047D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.803572900012858D-08, &
   0.000000000000000D+00,   6.474517958416172D-07,   0.000000000000000D+00, &
   6.022764563864215D-07,   0.000000000000000D+00,   3.738494858852573D-06, &
   0.000000000000000D+00,  -2.093774125638238D-05,   0.000000000000000D+00, &
  -1.194613355474646D-05,   0.000000000000000D+00,  -4.385394181883556D-06, &
   0.000000000000000D+00,  -3.757098286614837D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.859872670786714D-08,   0.000000000000000D+00, &
  -2.627924419534741D-07,   0.000000000000000D+00,   6.399524663972971D-07, &
   0.000000000000000D+00,   2.506393378461507D-06,   0.000000000000000D+00, &
   8.712077792817492D-06,   0.000000000000000D+00,  -8.334489943865835D-06, &
   0.000000000000000D+00,  -7.071086856528381D-07,   0.000000000000000D+00, &
   2.330232824619859D-06,   0.000000000000000D+00,   1.645191864597767D-06 /)
lem(:,122) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.470255192966473D-03,   3.690871001789317D-03, &
   1.515817483812315D-04,  -5.959800457969717D-03,  -1.944801200471323D-03, &
   6.656884899172401D-03,  -4.521795271670237D-04,  -3.484096743060780D-03, &
   2.323888549418420D-03,  -4.151635476300120D-04,  -8.684610968714480D-04, &
   1.081517747553055D-03,  -4.224116576574953D-04,  -2.170289037357982D-04, &
   3.311379173759305D-04,  -9.745414563115886D-05,  -9.141551640871553D-05, &
   7.719805196903430D-05,   1.733699376282897D-05,  -4.260156047118520D-05, &
   1.674397998597302D-06,   0.000000000000000D+00,  -2.470255192966476D-03, &
  -3.690871001789321D-03,  -1.515817483812351D-04,   5.959800457969715D-03, &
   1.944801200471321D-03,  -6.656884899172408D-03,   4.521795271670251D-04, &
   3.484096743060776D-03,  -2.323888549418414D-03,   4.151635476300104D-04, &
   8.684610968714468D-04,  -1.081517747553053D-03,   4.224116576574961D-04, &
   2.170289037357975D-04,  -3.311379173759345D-04,   9.745414563115693D-05, &
   9.141551640872361D-05,  -7.719805196903337D-05,  -1.733699376283536D-05, &
   4.260156047118858D-05,  -1.674397998597579D-06,   0.000000000000000D+00, &
  -1.235127596483242D-03,   0.000000000000000D+00,   1.913340835363790D-03, &
   0.000000000000000D+00,   4.549296813090796D-04,   0.000000000000000D+00, &
  -9.717809706857909D-04,   0.000000000000000D+00,  -1.409713539359907D-03, &
   0.000000000000000D+00,   1.975193042591151D-03,   0.000000000000000D+00, &
  -6.496752467192191D-04,   0.000000000000000D+00,   2.651371902246631D-06, &
   0.000000000000000D+00,   1.939933536578754D-05,   0.000000000000000D+00, &
  -1.148994911980160D-05,   0.000000000000000D+00,   6.087122157577419D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.764541014541012D-05, &
   0.000000000000000D+00,   1.176315514550812D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.644703335879809D-05,   0.000000000000000D+00, &
  -1.721521942110177D-04,   0.000000000000000D+00,   3.277417170606041D-06, &
   0.000000000000000D+00,  -3.634163609008806D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.368012333956610D-05,   0.000000000000000D+00, &
   1.559655729547372D-04,   0.000000000000000D+00,   2.097021284206850D-04, &
   0.000000000000000D+00,  -2.143503497992655D-05,   0.000000000000000D+00, &
  -5.120465404908124D-05,   0.000000000000000D+00,   9.432066229908720D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.562904072191998D-05, &
   0.000000000000000D+00,  -7.904220607393978D-05,   0.000000000000000D+00, &
  -2.777482920308964D-04,   0.000000000000000D+00,  -2.257194734337729D-04, &
   0.000000000000000D+00,   2.071927675979365D-05,   0.000000000000000D+00, &
   6.554566640042456D-05,   0.000000000000000D+00,   1.284225384220966D-04, &
   0.000000000000000D+00,  -1.229211121714320D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.310817560716655D-06,   0.000000000000000D+00, &
   5.729057944194545D-06,   0.000000000000000D+00,   9.191571981556753D-05, &
   0.000000000000000D+00,   3.347898049116459D-04,   0.000000000000000D+00, &
   2.487188490948895D-04,   0.000000000000000D+00,  -8.369435824342340D-06, &
   0.000000000000000D+00,  -2.094282435681002D-05,   0.000000000000000D+00, &
  -1.216421610206734D-04,   0.000000000000000D+00,  -1.634131432262070D-04, &
   0.000000000000000D+00,   7.138808518086386D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.197267853816606D-06,   0.000000000000000D+00, &
   1.977152123537832D-06,   0.000000000000000D+00,   3.585085609237170D-05, &
   0.000000000000000D+00,  -9.811623521935298D-05,   0.000000000000000D+00, &
  -2.666747651258245D-04,   0.000000000000000D+00,  -2.244847682523566D-04, &
   0.000000000000000D+00,   9.208562260789180D-07,   0.000000000000000D+00, &
   3.971358523619991D-06,   0.000000000000000D+00,   8.977789008433184D-06, &
   0.000000000000000D+00,   1.466177464950226D-04,   0.000000000000000D+00, &
   1.197311500340002D-04,   0.000000000000000D+00,  -8.034731019415634D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.890285696315789D-07, &
   0.000000000000000D+00,  -1.623856617800016D-07,   0.000000000000000D+00, &
  -6.068321025435816D-06,   0.000000000000000D+00,  -5.393576476599952D-05, &
   0.000000000000000D+00,   9.540971463134722D-05,   0.000000000000000D+00, &
   1.472383760393849D-04,   0.000000000000000D+00,   1.320340878681072D-04, &
   0.000000000000000D+00,   1.695795671872987D-07,   0.000000000000000D+00, &
  -8.956731831265194D-07,   0.000000000000000D+00,   1.298369132562248D-06, &
   0.000000000000000D+00,  -4.097577548752158D-06,   0.000000000000000D+00, &
  -1.164829188568652D-04,   0.000000000000000D+00,  -4.908697531600557D-05, &
   0.000000000000000D+00,  -4.480679696369292D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.912528427914396D-08,   0.000000000000000D+00, &
   5.864075159727675D-08,   0.000000000000000D+00,  -1.046013707125796D-06, &
   0.000000000000000D+00,   7.657075724525855D-06,   0.000000000000000D+00, &
   2.783060220980730D-05,   0.000000000000000D+00,  -7.387003104790293D-05, &
   0.000000000000000D+00,  -6.367067066289998D-05,   0.000000000000000D+00, &
  -4.345457195798993D-05,   0.000000000000000D+00,   8.318354887923945D-10, &
   0.000000000000000D+00,   1.959274905849223D-07,   0.000000000000000D+00, &
  -2.383984463434444D-08,   0.000000000000000D+00,  -2.324271577590999D-06, &
   0.000000000000000D+00,   1.454049514363367D-05,   0.000000000000000D+00, &
   5.533644669235109D-05,   0.000000000000000D+00,   5.854727113496472D-06, &
   0.000000000000000D+00,  -5.162006559558046D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.281968750309408D-08,   0.000000000000000D+00,   5.296248362789739D-07, &
   0.000000000000000D+00,   3.059032487806984D-07,   0.000000000000000D+00, &
  -8.422652731974667D-06,   0.000000000000000D+00,   1.011526826590353D-05, &
   0.000000000000000D+00,   4.165561136999599D-05,   0.000000000000000D+00, &
   2.243636424728915D-05,   0.000000000000000D+00,   5.048919510055489D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.454403273260091D-08,   0.000000000000000D+00,  -1.571600267417641D-07, &
   0.000000000000000D+00,   5.096025555234926D-07,   0.000000000000000D+00, &
   6.756316228345471D-07,   0.000000000000000D+00,  -1.974522369611858D-05, &
   0.000000000000000D+00,  -8.025590429428634D-06,   0.000000000000000D+00, &
   8.149628622007446D-06,   0.000000000000000D+00,   5.147541015232004D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.191037177083429D-09,   0.000000000000000D+00, &
  -6.805415721235018D-08,   0.000000000000000D+00,  -1.457953553816794D-07, &
   0.000000000000000D+00,   5.490190387853140D-07,   0.000000000000000D+00, &
   2.365677391793167D-06,   0.000000000000000D+00,  -2.236973260237828D-05, &
   0.000000000000000D+00,  -1.654922238768622D-05,   0.000000000000000D+00, &
  -3.817408786007081D-06,   0.000000000000000D+00,   2.057722597713006D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.673039821697282D-11,   0.000000000000000D+00,   5.003683017731743D-08, &
   0.000000000000000D+00,  -2.646681462283288D-07,   0.000000000000000D+00, &
  -1.839027589836634D-07,   0.000000000000000D+00,   2.922165303959604D-06, &
   0.000000000000000D+00,   7.698739881631031D-06,   0.000000000000000D+00, &
  -8.919066991481029D-06,   0.000000000000000D+00,  -7.603024508401014D-06, &
   0.000000000000000D+00,  -8.977770351569670D-07,   0.000000000000000D+00 /)
lem(:,123) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.032770972899660D-03,  -1.647019614671026D-03,  -1.113588507968434D-03, &
  -1.155974500306397D-03,   2.916607642080042D-03,   2.798529838263545D-03, &
  -3.655329787688898D-03,  -3.628360400224177D-04,   1.704094165443369D-03, &
  -1.147046991820749D-03,   4.907205521045612D-04,   5.449901995844551D-05, &
  -4.385742092267916D-04,   4.561143408281851D-04,  -1.709107689969270D-04, &
  -1.050927659202793D-04,   1.623903636715851D-04,  -5.235235477345085D-05, &
  -4.696455422220697D-05,   4.467917050390488D-05,   6.526947059810062D-06, &
  -2.509657422929422D-05,  -1.032770972899654D-03,  -1.647019614671036D-03, &
  -1.113588507968436D-03,  -1.155974500306397D-03,   2.916607642080050D-03, &
   2.798529838263540D-03,  -3.655329787688896D-03,  -3.628360400224153D-04, &
   1.704094165443368D-03,  -1.147046991820754D-03,   4.907205521045579D-04, &
   5.449901995844934D-05,  -4.385742092267886D-04,   4.561143408281804D-04, &
  -1.709107689969250D-04,  -1.050927659202762D-04,   1.623903636715822D-04, &
  -5.235235477345186D-05,  -4.696455422220720D-05,   4.467917050390626D-05, &
   6.526947059811607D-06,  -2.509657422929760D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.067737615862291D-03,   0.000000000000000D+00,  -1.385175161431465D-02, &
   0.000000000000000D+00,   1.507573622359452D-02,   0.000000000000000D+00, &
  -7.052687324510194D-03,   0.000000000000000D+00,   6.111632524865731D-04, &
   0.000000000000000D+00,   4.937451472242433D-04,   0.000000000000000D+00, &
  -1.480830566708976D-04,   0.000000000000000D+00,   5.065250812354608D-05, &
   0.000000000000000D+00,  -1.997886791552477D-05,   0.000000000000000D+00, &
   3.720238095238121D-05,   3.720238095238092D-05,   0.000000000000000D+00, &
   1.648933174668492D-06,   0.000000000000000D+00,  -1.308784914023899D-05, &
  -3.656601772410588D-05,   0.000000000000000D+00,  -2.969896984143306D-05, &
   0.000000000000000D+00,  -2.117208180965921D-05,   0.000000000000000D+00, &
  -1.234015087294892D-06,   0.000000000000000D+00,  -1.393293913822891D-04, &
   2.560816645085227D-05,   0.000000000000000D+00,   3.554252402331928D-05, &
   0.000000000000000D+00,   1.292507988612624D-04,   0.000000000000000D+00, &
   1.914083401119018D-05,   0.000000000000000D+00,   2.678249521806305D-07, &
   0.000000000000000D+00,  -2.144213849660826D-05,   0.000000000000000D+00, &
   4.399621597892515D-04,  -1.547447972564078D-05,   0.000000000000000D+00, &
  -2.268310048761871D-05,   0.000000000000000D+00,  -7.580377434527954D-05, &
   0.000000000000000D+00,  -3.310842725003490D-04,   0.000000000000000D+00, &
  -6.706550477162232D-06,   0.000000000000000D+00,   4.049330623451037D-06, &
   0.000000000000000D+00,   4.977517868432065D-05,   0.000000000000000D+00, &
   8.018293955422596D-05,   0.000000000000000D+00,  -6.256630386782874D-04, &
   6.339971079084788D-06,   0.000000000000000D+00,   5.869689335404335D-06, &
   0.000000000000000D+00,   9.707245911805313D-06,   0.000000000000000D+00, &
   1.253916578627554D-04,   0.000000000000000D+00,   4.508759028566394D-04, &
   0.000000000000000D+00,   9.062054235507243D-08,   0.000000000000000D+00, &
  -2.185049270110804D-06,   0.000000000000000D+00,  -1.873769509142181D-05, &
   0.000000000000000D+00,  -1.222217490003373D-04,   0.000000000000000D+00, &
  -1.114780210871813D-04,   0.000000000000000D+00,   4.839992905628560D-04, &
  -1.067897124524741D-06,   0.000000000000000D+00,   4.753113544989164D-07, &
   0.000000000000000D+00,   4.374027847677349D-06,   0.000000000000000D+00, &
   2.460123542847945D-05,   0.000000000000000D+00,  -1.539547809588272D-04, &
   0.000000000000000D+00,  -3.434519703454365D-04,   0.000000000000000D+00, &
   3.329858195504413D-07,   0.000000000000000D+00,  -2.310476557830185D-07, &
   0.000000000000000D+00,   2.329579083921259D-06,   0.000000000000000D+00, &
   1.902765061208607D-05,   0.000000000000000D+00,   1.603230249219410D-04, &
   0.000000000000000D+00,   6.444754768898557D-05,   0.000000000000000D+00, &
  -1.834779565238407D-04,  -4.131372228555897D-08,   0.000000000000000D+00, &
  -2.243604766761929D-07,   0.000000000000000D+00,  -1.807326883806141D-06, &
   0.000000000000000D+00,  -1.123764359488984D-05,   0.000000000000000D+00, &
  -4.292484635097513D-05,   0.000000000000000D+00,   1.439193330836288D-04, &
   0.000000000000000D+00,   1.380567011588266D-04,   0.000000000000000D+00, &
   1.880294802952159D-09,   0.000000000000000D+00,   2.346587435700196D-07, &
   0.000000000000000D+00,   1.013013772579727D-07,   0.000000000000000D+00, &
   6.941770019847597D-07,   0.000000000000000D+00,  -1.653314795884399D-05, &
   0.000000000000000D+00,  -1.261146873915508D-04,   0.000000000000000D+00, &
  -4.704655662233200D-06,   0.000000000000000D+00,  -1.250133986722482D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.695418263696588D-08, &
   0.000000000000000D+00,   3.027589686412020D-07,   0.000000000000000D+00, &
   6.964842996584112D-07,   0.000000000000000D+00,   1.240510467275273D-05, &
   0.000000000000000D+00,   1.731295007768328D-05,   0.000000000000000D+00, &
  -9.154660920704312D-05,   0.000000000000000D+00,  -2.258390053495511D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.161888964419345D-08,   0.000000000000000D+00,  -1.540381659026695D-08, &
   0.000000000000000D+00,   1.299496269251333D-07,   0.000000000000000D+00, &
  -1.967625314998764D-06,   0.000000000000000D+00,   2.412636819682806D-05, &
   0.000000000000000D+00,   5.053722015778618D-05,   0.000000000000000D+00, &
  -5.400913022095290D-06,   0.000000000000000D+00,   3.493732505745005D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.017789329885840D-10, &
   0.000000000000000D+00,  -7.630004117257147D-08,   0.000000000000000D+00, &
   1.154461264971164D-08,   0.000000000000000D+00,   4.409509387617953D-07, &
   0.000000000000000D+00,  -1.117854527884966D-05,   0.000000000000000D+00, &
   1.983420605506433D-05,   0.000000000000000D+00,   3.220106730517380D-05, &
   0.000000000000000D+00,   3.446832756296213D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.890535205595603D-11, &
   0.000000000000000D+00,   2.832973532589152D-08,   0.000000000000000D+00, &
  -1.608962070417818D-07,   0.000000000000000D+00,  -5.349185375123171D-07, &
   0.000000000000000D+00,   8.409130232809211D-07,   0.000000000000000D+00, &
  -2.430707338057382D-05,   0.000000000000000D+00,   3.587237946032855D-07, &
   0.000000000000000D+00,  -5.755918757993512D-06,   0.000000000000000D+00, &
  -1.960676206853288D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.696634303061219D-09, &
   0.000000000000000D+00,   1.340496449440399D-08,   0.000000000000000D+00, &
   5.358416857404380D-08,   0.000000000000000D+00,  -3.189720773104326D-07, &
   0.000000000000000D+00,   2.195025812472940D-06,   0.000000000000000D+00, &
  -2.498202531190891D-05,   0.000000000000000D+00,  -5.507022150707782D-06, &
   0.000000000000000D+00,  -6.474766364581160D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.918675765687707D-10,   0.000000000000000D+00, &
   4.597379744062087D-08,   0.000000000000000D+00,  -1.663791157252839D-08, &
   0.000000000000000D+00,   2.013143602082153D-07,   0.000000000000000D+00, &
   3.557151239793045D-06,   0.000000000000000D+00,   6.425467514154587D-06, &
   0.000000000000000D+00,  -7.360654648076421D-06,   0.000000000000000D+00, &
   5.959008684499495D-06,   0.000000000000000D+00,   8.360165568857964D-06 /)
lem(:,124) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.411540655263107D-04,   4.871949722361907D-03, &
   5.336841080104939D-03,  -4.555037006338940D-03,  -5.216217651034296D-03, &
   3.759145452413464D-03,   8.759973010754201D-04,  -1.466471786859341D-03, &
   7.911337708894464D-04,  -3.396531914402941D-04,   1.319071807784306D-04, &
   7.225732203878188D-05,  -2.727647685773682D-04,   2.682377674446836D-04, &
  -8.949419182058205D-05,  -6.861962555038703D-05,   1.003393487404457D-04, &
  -3.732773343499435D-05,  -2.790696277696196D-05,   3.474992530114886D-05, &
  -8.324387465173039D-07,   0.000000000000000D+00,  -9.411540655263084D-04, &
  -4.871949722361893D-03,  -5.336841080104940D-03,   4.555037006338945D-03, &
   5.216217651034290D-03,  -3.759145452413473D-03,  -8.759973010754204D-04, &
   1.466471786859342D-03,  -7.911337708894554D-04,   3.396531914402889D-04, &
  -1.319071807784159D-04,  -7.225732203877065D-05,   2.727647685773582D-04, &
  -2.682377674446908D-04,   8.949419182058145D-05,   6.861962555038948D-05, &
  -1.003393487404435D-04,   3.732773343498884D-05,   2.790696277696554D-05, &
  -3.474992530114360D-05,   8.324387465090582D-07,   0.000000000000000D+00, &
  -4.705770327631500D-04,   0.000000000000000D+00,   3.373751662469281D-03, &
   0.000000000000000D+00,  -1.414447395107834D-02,   0.000000000000000D+00, &
   2.857850654000552D-02,   0.000000000000000D+00,  -3.026661055074404D-02, &
   0.000000000000000D+00,   1.711616106842150D-02,   0.000000000000000D+00, &
  -4.913362732439718D-03,   0.000000000000000D+00,   7.726364356723449D-04, &
   0.000000000000000D+00,  -1.892120874494825D-04,   0.000000000000000D+00, &
   6.044157935174118D-05,   0.000000000000000D+00,  -2.313125292795250D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.720238095238083D-05, &
   0.000000000000000D+00,   2.790178571428573D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.860119047619039D-05,   0.000000000000000D+00, &
  -1.966042493386241D-04,   0.000000000000000D+00,  -3.410218253968255D-05, &
   0.000000000000000D+00,  -1.449152640982167D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.947972209678870D-05,   0.000000000000000D+00, &
   9.309324732965399D-05,   0.000000000000000D+00,   6.576452999301537D-04, &
   0.000000000000000D+00,   2.161017302271172D-05,   0.000000000000000D+00, &
   9.898890328940482D-07,   0.000000000000000D+00,  -1.454936693423536D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.746572409788353D-05, &
   0.000000000000000D+00,  -4.529537276126860D-05,   0.000000000000000D+00, &
  -1.927149313257334D-04,   0.000000000000000D+00,  -1.346869373461046D-03, &
   0.000000000000000D+00,  -4.638115367604467D-06,   0.000000000000000D+00, &
   8.414982230908443D-06,   0.000000000000000D+00,   1.168814701901885D-04, &
   0.000000000000000D+00,   4.584096863165207D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.865786034533440D-06,   0.000000000000000D+00, &
   1.076491941971605D-05,   0.000000000000000D+00,   2.848155451261740D-05, &
   0.000000000000000D+00,   2.487188490948895D-04,   0.000000000000000D+00, &
   1.763857592179344D-03,   0.000000000000000D+00,  -1.473522761507321D-06, &
   0.000000000000000D+00,  -3.429486172105928D-06,   0.000000000000000D+00, &
  -4.332258637007663D-05,   0.000000000000000D+00,  -2.713203703747845D-04, &
   0.000000000000000D+00,  -6.543186757528847D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.081668365286646D-06,   0.000000000000000D+00, &
   2.670112456823269D-07,   0.000000000000000D+00,   8.779144147348663D-06, &
   0.000000000000000D+00,   4.348741617296045D-05,   0.000000000000000D+00, &
  -2.717844326807615D-04,   0.000000000000000D+00,  -1.520476460562043D-03, &
   0.000000000000000D+00,   6.297212821056079D-07,   0.000000000000000D+00, &
  -8.942059682998260D-07,   0.000000000000000D+00,   6.171240914177161D-06, &
   0.000000000000000D+00,   4.550834841765431D-05,   0.000000000000000D+00, &
   3.583265002909326D-04,   0.000000000000000D+00,   5.103710412375814D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.378515110596924D-07, &
   0.000000000000000D+00,  -1.622962659310455D-08,   0.000000000000000D+00, &
  -5.172987143847849D-06,   0.000000000000000D+00,  -2.328430570393958D-05, &
   0.000000000000000D+00,  -8.956892642838677D-05,   0.000000000000000D+00, &
   2.923289399147370D-04,   0.000000000000000D+00,   8.756479037429644D-04, &
   0.000000000000000D+00,   1.418855108797657D-08,   0.000000000000000D+00, &
   5.339037625905866D-07,   0.000000000000000D+00,   4.310093926392462D-07, &
   0.000000000000000D+00,   5.069017900853515D-07,   0.000000000000000D+00, &
  -3.997295700094324D-05,   0.000000000000000D+00,  -3.109238640906486D-04, &
   0.000000000000000D+00,  -1.963695414911587D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.708037794144892D-10,   0.000000000000000D+00, &
  -2.333192426085970D-07,   0.000000000000000D+00,   1.534436845571678D-06, &
   0.000000000000000D+00,   1.097465535354131D-06,   0.000000000000000D+00, &
   2.870332567018804D-05,   0.000000000000000D+00,   4.452394761033676D-05, &
   0.000000000000000D+00,  -2.595394910318938D-04,   0.000000000000000D+00, &
  -3.452891395205541D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.026808536275895D-08,   0.000000000000000D+00, &
  -4.879897957258858D-07,   0.000000000000000D+00,   1.878427934055564D-06, &
   0.000000000000000D+00,  -5.865335201792831D-06,   0.000000000000000D+00, &
   5.534760220403144D-05,   0.000000000000000D+00,   1.595295769868729D-04, &
   0.000000000000000D+00,  -1.878438176187450D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.620355238472149D-08,   0.000000000000000D+00,  -3.537516516022247D-07, &
   0.000000000000000D+00,  -7.230960023145388D-07,   0.000000000000000D+00, &
   2.238731086576394D-06,   0.000000000000000D+00,  -2.701078374656346D-05, &
   0.000000000000000D+00,   4.176004316552767D-05,   0.000000000000000D+00, &
   1.538131326630015D-04,   0.000000000000000D+00,   1.034756719847738D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.629308582534725D-09,   0.000000000000000D+00,   2.310416455269243D-07, &
   0.000000000000000D+00,  -1.105575422703780D-06,   0.000000000000000D+00, &
  -1.528656601662424D-06,   0.000000000000000D+00,   2.837164422956440D-06, &
   0.000000000000000D+00,  -5.981664679365698D-05,   0.000000000000000D+00, &
  -1.671857842976333D-05,   0.000000000000000D+00,   3.743146980678164D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.189125266498008D-10,   0.000000000000000D+00, &
  -1.123060948951494D-08,   0.000000000000000D+00,   6.933573184650414D-07, &
   0.000000000000000D+00,  -1.058910771277853D-06,   0.000000000000000D+00, &
  -2.436311156580268D-07,   0.000000000000000D+00,   7.701488606968997D-06, &
   0.000000000000000D+00,  -7.273294412638321D-05,   0.000000000000000D+00, &
  -5.218610460266790D-05,   0.000000000000000D+00,  -3.084752967942762D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.168113773032231D-11,   0.000000000000000D+00,  -2.741463701113055D-08, &
   0.000000000000000D+00,   1.717028529964347D-07,   0.000000000000000D+00, &
   1.098826846704146D-06,   0.000000000000000D+00,  -5.164054023605020D-07, &
   0.000000000000000D+00,   8.006472115757420D-06,   0.000000000000000D+00, &
   2.304273560588360D-05,   0.000000000000000D+00,  -3.741243449238283D-05, &
   0.000000000000000D+00,  -2.130062861756966D-05,   0.000000000000000D+00 /)
lem(:,125) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.365016153621295D-03,  -1.200308765256697D-02,  -1.277415282408885D-02, &
   2.554328464052893D-02,   1.431336529185573D-02,  -1.860444111509130D-02, &
  -1.153896859327453D-02,   1.606350952056657D-03,   4.270396988032999D-03, &
   4.037721067659933D-03,   1.472465162627364D-03,  -1.058181321145315D-03, &
  -1.702726574794893D-03,  -6.877104409150884D-04,   4.118793567437304D-04, &
   5.524149652181118D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.365016153621223D-03,  -1.200308765256691D-02, &
  -1.277415282408882D-02,   2.554328464052883D-02,   1.431336529185575D-02, &
  -1.860444111509121D-02,  -1.153896859327457D-02,   1.606350952056606D-03, &
   4.270396988033039D-03,   4.037721067659937D-03,   1.472465162627354D-03, &
  -1.058181321145319D-03,  -1.702726574794919D-03,  -6.877104409150756D-04, &
   4.118793567437627D-04,   5.524149652180875D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.598021173872556D-04,   0.000000000000000D+00,   2.945602237950619D-06, &
   0.000000000000000D+00,   1.431889976783249D-07,   0.000000000000000D+00, &
  -1.690878780662061D-03,   0.000000000000000D+00,  -3.286187496713834D-05, &
   0.000000000000000D+00,  -2.330400937212223D-06,   0.000000000000000D+00, &
  -2.657945769400758D-07,   3.915010820381667D-03,   0.000000000000000D+00, &
   2.439454922965536D-05,   0.000000000000000D+00,   9.330742052892969D-06, &
   0.000000000000000D+00,   1.578308508918303D-06,   0.000000000000000D+00, &
  -2.508509124473590D-03,   0.000000000000000D+00,   4.597872177627181D-04, &
   0.000000000000000D+00,   1.276561042343777D-05,   0.000000000000000D+00, &
  -2.398454621162817D-06,   0.000000000000000D+00,  -1.308267872366110D-06, &
  -4.859014544032712D-03,   0.000000000000000D+00,  -1.464014918988669D-03, &
   0.000000000000000D+00,  -1.050655869608626D-04,   0.000000000000000D+00, &
  -1.182852403575285D-05,   0.000000000000000D+00,   9.062054235507243D-08, &
   0.000000000000000D+00,   1.007582895858572D-02,   0.000000000000000D+00, &
   1.173471500904535D-03,   0.000000000000000D+00,   2.317428014022501D-05, &
   0.000000000000000D+00,   2.986942724794428D-05,   0.000000000000000D+00, &
   7.557425804329895D-06,   0.000000000000000D+00,   5.030884636460577D-07, &
  -5.287216335809688D-03,   0.000000000000000D+00,   1.645931368303152D-03, &
   0.000000000000000D+00,   5.054700695657303D-04,   0.000000000000000D+00, &
   3.878572520045143D-05,   0.000000000000000D+00,  -6.626518333174415D-06, &
   0.000000000000000D+00,  -4.270167324599759D-06,   0.000000000000000D+00, &
  -3.023258629665951D-03,   0.000000000000000D+00,  -3.739659860893250D-03, &
   0.000000000000000D+00,  -6.666893303721947D-04,   0.000000000000000D+00, &
  -1.285123993429288D-04,   0.000000000000000D+00,  -2.836821317395786D-05, &
   0.000000000000000D+00,  -5.135907735203452D-07,   0.000000000000000D+00, &
   3.729342415519637D-06,   5.157129037820276D-03,   0.000000000000000D+00, &
   1.793960813587999D-03,   0.000000000000000D+00,  -4.533626030810867D-04, &
   0.000000000000000D+00,  -6.199447471071659D-05,   0.000000000000000D+00, &
   2.832326921745565D-05,   0.000000000000000D+00,   1.551118364136671D-05, &
   0.000000000000000D+00,   2.201375046465415D-06,   0.000000000000000D+00, &
  -2.461547004275646D-03,   0.000000000000000D+00,   1.409852091498109D-03, &
   0.000000000000000D+00,   1.404543867120350D-03,   0.000000000000000D+00, &
   2.966000363011027D-04,   0.000000000000000D+00,   6.212772740227629D-05, &
   0.000000000000000D+00,  -4.509401944154796D-07,   0.000000000000000D+00, &
  -7.435255086679334D-06,   0.000000000000000D+00,  -1.921481982842994D-06, &
   4.226381432939439D-04,   0.000000000000000D+00,  -1.931637235903234D-03, &
   0.000000000000000D+00,  -4.963302949880304D-04,   0.000000000000000D+00, &
   3.497979077060417D-05,   0.000000000000000D+00,  -4.946722120287540D-05, &
   0.000000000000000D+00,  -3.172112313191212D-05,   0.000000000000000D+00, &
  -5.518304872905855D-06,   0.000000000000000D+00,   2.664808639891046D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.758310350847696D-04,   0.000000000000000D+00,  -7.996733717146503D-04, &
   0.000000000000000D+00,  -4.128658874285371D-04,   0.000000000000000D+00, &
  -9.229082030473591D-05,   0.000000000000000D+00,  -4.191015896549517D-06, &
   0.000000000000000D+00,   1.280244359106400D-05,   0.000000000000000D+00, &
   4.911725386832484D-06,   0.000000000000000D+00,  -3.576897955560075D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.484429531804844D-05, &
   0.000000000000000D+00,   6.716892231970755D-04,   0.000000000000000D+00, &
   3.903431014888693D-05,   0.000000000000000D+00,   2.851792966013809D-05, &
   0.000000000000000D+00,   3.652069281774566D-05,   0.000000000000000D+00, &
   1.125044454999954D-05,   0.000000000000000D+00,  -3.510947529577148D-06, &
   0.000000000000000D+00,  -3.561832802367134D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.887993023457575D-05, &
   0.000000000000000D+00,   2.659329335774389D-05,   0.000000000000000D+00, &
   3.570084220156669D-04,   0.000000000000000D+00,   1.089012834350040D-04, &
   0.000000000000000D+00,   2.128841777068574D-05,   0.000000000000000D+00, &
  -9.330510552078064D-06,   0.000000000000000D+00,  -8.222631001698584D-06, &
   0.000000000000000D+00,   6.870236248981195D-07,   0.000000000000000D+00, &
   3.045469849974317D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.042915196633671D-06,   0.000000000000000D+00,  -1.690381096231454D-04, &
   0.000000000000000D+00,  -8.623287160514115D-05,   0.000000000000000D+00, &
   3.111860017823468D-05,   0.000000000000000D+00,  -1.713847620745557D-05, &
   0.000000000000000D+00,  -1.539463375881575D-05,   0.000000000000000D+00, &
  -4.123260121201170D-07,   0.000000000000000D+00,   4.373970947021826D-06, &
   0.000000000000000D+00,   1.982105520618386D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.579885217513817D-05,   0.000000000000000D+00, &
  -1.858537219598393D-04,   0.000000000000000D+00,  -1.119094176384784D-04, &
   0.000000000000000D+00,  -3.510123880716526D-05,   0.000000000000000D+00, &
  -4.699259338320084D-06,   0.000000000000000D+00,   6.763103289680305D-06, &
   0.000000000000000D+00,   2.523823554261967D-06,   0.000000000000000D+00, &
  -1.787242197092850D-06,   0.000000000000000D+00,   4.448151268220333D-07 /)
lem(:,126) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.836239160007518D-04,   3.288118453278781D-03, &
  -6.239159631137253D-03,  -1.064515919290594D-03,   1.343592337727050D-02, &
  -5.979389936687075D-03,  -1.138711889115282D-02,   6.407912217303546D-03, &
   5.664071629015317D-03,  -1.589486375132135D-03,  -2.363288651112244D-03, &
  -9.448523999883148D-04,   6.636789527987562D-04,   9.200190878118134D-04, &
   1.427166219619418D-04,  -4.294797585342414D-04,  -2.710654059289056D-04, &
   1.247420479081129D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.836239160007587D-04, &
  -3.288118453278790D-03,   6.239159631137228D-03,   1.064515919290618D-03, &
  -1.343592337727050D-02,   5.979389936687048D-03,   1.138711889115284D-02, &
  -6.407912217303550D-03,  -5.664071629015330D-03,   1.589486375132140D-03, &
   2.363288651112253D-03,   9.448523999883052D-04,  -6.636789527987508D-04, &
  -9.200190878118139D-04,  -1.427166219619534D-04,   4.294797585342562D-04, &
   2.710654059289038D-04,  -1.247420479081310D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.782578066876689D-06,   0.000000000000000D+00, &
  -2.343121507534138D-05,   0.000000000000000D+00,  -3.850196093098431D-06, &
   0.000000000000000D+00,   3.731707655437570D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.603001787497926D-05, &
   0.000000000000000D+00,   8.301981996208950D-06,   0.000000000000000D+00, &
   4.297909698561121D-06,   0.000000000000000D+00,  -3.764558351167265D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.237152939939377D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.690775684583737D-05, &
   0.000000000000000D+00,  -7.560379077405990D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.742547442237832D-05,   0.000000000000000D+00, &
   5.243171983552293D-06,   0.000000000000000D+00,   5.412544112235542D-07, &
   0.000000000000000D+00,  -2.306259272203550D-04,   0.000000000000000D+00, &
  -7.636473801887346D-06,   0.000000000000000D+00,  -2.257374548396325D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.055588464122591D-04, &
   0.000000000000000D+00,  -3.867538384777767D-05,   0.000000000000000D+00, &
  -2.930393965525722D-07,   0.000000000000000D+00,   2.014132238590375D-06, &
   0.000000000000000D+00,  -5.581083788110315D-05,   0.000000000000000D+00, &
   1.855068250276217D-04,   0.000000000000000D+00,   1.819281584391291D-05, &
   0.000000000000000D+00,   1.268366585031082D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.852978941012324D-04,   0.000000000000000D+00, &
  -3.128506527348902D-04,   0.000000000000000D+00,  -2.691834869786151D-05, &
   0.000000000000000D+00,  -8.369435824342340D-06,   0.000000000000000D+00, &
  -1.473522761507321D-06,   0.000000000000000D+00,   8.104726370980238D-04, &
   0.000000000000000D+00,   1.197083677902092D-04,   0.000000000000000D+00, &
  -4.807302895874674D-05,   0.000000000000000D+00,   1.757465960252409D-06, &
   0.000000000000000D+00,   3.944545605603610D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.278640236238836D-04,   0.000000000000000D+00, &
   4.128723387708451D-04,   0.000000000000000D+00,   1.786150763132858D-04, &
   0.000000000000000D+00,   2.879424907961722D-05,   0.000000000000000D+00, &
   2.667917443417206D-06,   0.000000000000000D+00,  -4.829511538597127D-06, &
   0.000000000000000D+00,  -6.831514706060413D-05,   0.000000000000000D+00, &
  -7.480802370039545D-04,   0.000000000000000D+00,  -1.351503673485099D-04, &
   0.000000000000000D+00,  -2.770100108234454D-05,   0.000000000000000D+00, &
  -1.304671610651563D-05,   0.000000000000000D+00,  -2.634612237025543D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.042627890765739D-04, &
   0.000000000000000D+00,   4.506685853534540D-04,   0.000000000000000D+00, &
  -1.712325585799065D-04,   0.000000000000000D+00,  -4.819874600058017D-05, &
   0.000000000000000D+00,  -6.350605800541018D-07,   0.000000000000000D+00, &
   5.827266103000358D-06,   0.000000000000000D+00,   3.811478361261452D-06, &
   0.000000000000000D+00,  -3.422141862055034D-04,   0.000000000000000D+00, &
   1.429858710357118D-04,   0.000000000000000D+00,   3.855576560226554D-04, &
   0.000000000000000D+00,   8.399386293450081D-05,   0.000000000000000D+00, &
   2.507854144964911D-05,   0.000000000000000D+00,   4.602803640183309D-06, &
   0.000000000000000D+00,  -1.591165031619978D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.498446214358508D-04,   0.000000000000000D+00, &
  -4.150414591679771D-04,   0.000000000000000D+00,  -1.882643597836635D-04, &
   0.000000000000000D+00,   2.785309057952043D-05,   0.000000000000000D+00, &
  -5.537787544269999D-06,   0.000000000000000D+00,  -9.475991755032815D-06, &
   0.000000000000000D+00,  -4.137735110765244D-06,   0.000000000000000D+00, &
   3.514145359848483D-06,   0.000000000000000D+00,  -3.534021385027985D-05, &
   0.000000000000000D+00,   2.801169348222646D-04,   0.000000000000000D+00, &
  -1.609820624297830D-04,   0.000000000000000D+00,  -1.255618362424210D-04, &
   0.000000000000000D+00,  -3.022446717816985D-05,   0.000000000000000D+00, &
  -5.591929834278601D-06,   0.000000000000000D+00,   2.554337781859674D-06, &
   0.000000000000000D+00,   2.022134247931225D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.571663921455250D-06,   0.000000000000000D+00, &
  -7.282046511520464D-05,   0.000000000000000D+00,   2.397416291655131D-04, &
   0.000000000000000D+00,   3.118418848785516D-05,   0.000000000000000D+00, &
   4.670010634386017D-06,   0.000000000000000D+00,   9.275081226669540D-06, &
   0.000000000000000D+00,   4.837314277104507D-06,   0.000000000000000D+00, &
  -9.586086806553373D-07,   0.000000000000000D+00,  -4.291312122122657D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.043679230739189D-05,   0.000000000000000D+00,  -7.939384358809687D-05, &
   0.000000000000000D+00,   1.009252567074195D-04,   0.000000000000000D+00, &
   3.111933391883783D-05,   0.000000000000000D+00,   7.108556574079045D-06, &
   0.000000000000000D+00,  -1.366481872088513D-06,   0.000000000000000D+00, &
  -2.311288870766782D-06,   0.000000000000000D+00,  -1.279949705242674D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.171545784880626D-05,   0.000000000000000D+00, &
  -3.883596126012327D-05,   0.000000000000000D+00,  -6.752405717711948D-05, &
   0.000000000000000D+00,   4.971333903035614D-06,   0.000000000000000D+00, &
  -3.791925726578202D-06,   0.000000000000000D+00,  -4.070627563263875D-06, &
   0.000000000000000D+00,  -6.869111274043034D-07,   0.000000000000000D+00, &
   1.994019807478267D-06,   0.000000000000000D+00,   4.189646142159611D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.721267749679910D-06,   0.000000000000000D+00,   3.618481855394543D-05, &
   0.000000000000000D+00,  -3.649195230708370D-05,   0.000000000000000D+00, &
  -3.682852130450033D-05,   0.000000000000000D+00,  -8.880857847336853D-06, &
   0.000000000000000D+00,  -1.576499000476968D-06,   0.000000000000000D+00, &
   1.312933681587883D-06,   0.000000000000000D+00,   1.362021322055115D-06, &
   0.000000000000000D+00,   5.668643673620718D-07,   0.000000000000000D+00 /)
lem(:,127) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.368127386726610D-04, &
  -1.089310940166038D-03,  -4.231462446151901D-04,   4.327306871899803D-03, &
   5.207153908574381D-03,  -1.079810498557698D-02,  -1.264496085079228D-02, &
   1.395843763109414D-02,   1.291126875509319D-02,  -7.869341601664186D-03, &
  -6.698902941516488D-03,   9.924266919881203D-04,   1.852027580680816D-03, &
   5.550150142411950D-04,  -3.675048371572249D-04,  -2.155438877778548D-04, &
   1.575357303291574D-04,   1.267152871314766D-04,  -7.892450307533664D-05, &
  -9.174528698703637D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368127386726629D-04,  -1.089310940166040D-03,  -4.231462446151855D-04, &
   4.327306871899804D-03,   5.207153908574362D-03,  -1.079810498557699D-02, &
  -1.264496085079226D-02,   1.395843763109415D-02,   1.291126875509319D-02, &
  -7.869341601664158D-03,  -6.698902941516478D-03,   9.924266919880854D-04, &
   1.852027580680830D-03,   5.550150142412170D-04,  -3.675048371572445D-04, &
  -2.155438877778764D-04,   1.575357303291606D-04,   1.267152871314901D-04, &
  -7.892450307533812D-05,  -9.174528698704300D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.886169238305224D-06,   0.000000000000000D+00, &
   5.438477165642956D-05,   0.000000000000000D+00,   9.284708323646279D-05, &
   0.000000000000000D+00,  -2.930941880052487D-05,   0.000000000000000D+00, &
  -1.102109812357167D-04,   0.000000000000000D+00,   2.195258701347978D-05, &
   0.000000000000000D+00,   5.912729338394921D-05,   0.000000000000000D+00, &
  -1.828392116932849D-05,   0.000000000000000D+00,  -9.367866922285164D-06, &
   0.000000000000000D+00,   5.589711831286089D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.089413854118940D-06,   0.000000000000000D+00, &
  -2.448531860296514D-05,   0.000000000000000D+00,  -4.897063720624601D-08, &
   2.137410806141464D-05,   0.000000000000000D+00,   2.382378543369577D-06, &
   0.000000000000000D+00,   3.520473334714812D-04,   0.000000000000000D+00, &
  -2.061921566588233D-08,   0.000000000000000D+00,  -1.031497742023849D-06, &
  -1.156895506740708D-03,   0.000000000000000D+00,   2.739737370431762D-06, &
   0.000000000000000D+00,   6.883811063357794D-08,   0.000000000000000D+00, &
   9.397534010537102D-04,   0.000000000000000D+00,  -1.542764081463768D-04, &
   0.000000000000000D+00,   1.359042574212963D-06,   0.000000000000000D+00, &
  -2.271872434411242D-06,   1.355841553041712D-03,   0.000000000000000D+00, &
   5.016419594424716D-04,   0.000000000000000D+00,   1.807877381161667D-06, &
   0.000000000000000D+00,  -2.145500361939928D-08,   0.000000000000000D+00, &
  -3.096756707942600D-03,   0.000000000000000D+00,  -4.109381293389907D-04, &
   0.000000000000000D+00,   6.403648549293972D-05,   0.000000000000000D+00, &
   1.657579961355655D-06,   0.000000000000000D+00,   2.197203798140605D-06, &
   1.499668613344747D-03,   0.000000000000000D+00,  -6.125271208054881D-04, &
   0.000000000000000D+00,  -2.603366481411977D-04,   0.000000000000000D+00, &
  -1.409430704675697D-05,   0.000000000000000D+00,  -2.185049270110804D-06, &
   0.000000000000000D+00,   1.173471500904535D-03,   0.000000000000000D+00, &
   1.394385702839062D-03,   0.000000000000000D+00,   2.453567692832733D-04, &
   0.000000000000000D+00,   3.818348628383754D-06,   0.000000000000000D+00, &
   4.018570766860828D-06,   0.000000000000000D+00,   4.864112984790845D-06, &
  -1.619263336385427D-03,   0.000000000000000D+00,  -6.752872693779955D-04, &
   0.000000000000000D+00,   3.107534132907667D-04,   0.000000000000000D+00, &
   1.026789508794151D-04,   0.000000000000000D+00,   7.247575562256650D-06, &
   0.000000000000000D+00,  -1.648726533100699D-06,   0.000000000000000D+00, &
   5.677717311180754D-04,   0.000000000000000D+00,  -5.129058648470990D-04, &
   0.000000000000000D+00,  -7.529230474361045D-04,   0.000000000000000D+00, &
  -1.476084171433479D-04,   0.000000000000000D+00,  -2.653712623063100D-05, &
   0.000000000000000D+00,  -6.234692438568264D-06,   0.000000000000000D+00, &
  -3.957449977790228D-06,   3.043263124445528D-05,   0.000000000000000D+00, &
   6.972111365471952D-04,   0.000000000000000D+00,   3.419858062123808D-04, &
   0.000000000000000D+00,  -1.021877582774766D-04,   0.000000000000000D+00, &
  -1.412964325396884D-05,   0.000000000000000D+00,   6.605595525840014D-06, &
   0.000000000000000D+00,   3.790236721662269D-06,   0.000000000000000D+00, &
  -5.017670289136769D-05,   0.000000000000000D+00,  -2.428130676607057D-04, &
   0.000000000000000D+00,   3.035076064489735D-04,   0.000000000000000D+00, &
   3.259432718368113D-04,   0.000000000000000D+00,   7.119660789184332D-05, &
   0.000000000000000D+00,   1.433700925147398D-05,   0.000000000000000D+00, &
   7.459020539686860D-07,   0.000000000000000D+00,  -3.655155137703740D-06, &
   5.161687749810187D-06,   0.000000000000000D+00,   5.905062287686504D-06, &
   0.000000000000000D+00,  -3.571618643978381D-04,   0.000000000000000D+00, &
  -1.130084385252220D-04,   0.000000000000000D+00,   9.063996074708437D-06, &
   0.000000000000000D+00,  -1.254475633139205D-05,   0.000000000000000D+00, &
  -8.022978270748101D-06,   0.000000000000000D+00,  -3.062523956393866D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.046322432369110D-06,   0.000000000000000D+00,   9.914542786249221D-05, &
   0.000000000000000D+00,  -1.803890975786334D-04,   0.000000000000000D+00, &
  -1.046217432297973D-04,   0.000000000000000D+00,  -2.351144805455855D-05, &
   0.000000000000000D+00,  -8.626109610814325D-07,   0.000000000000000D+00, &
   3.377347744514201D-06,   0.000000000000000D+00,   4.040111194957385D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.810670779946010D-06, &
   0.000000000000000D+00,  -7.874501604801202D-07,   0.000000000000000D+00, &
   1.401646999228513D-04,   0.000000000000000D+00,   1.038692012267270D-05, &
   0.000000000000000D+00,   7.651438223117655D-06,   0.000000000000000D+00, &
   9.825117296179254D-06,   0.000000000000000D+00,   2.151862612583055D-06, &
   0.000000000000000D+00,  -1.448852723118351D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.212979484206307D-06, &
   0.000000000000000D+00,   9.389499858767548D-06,   0.000000000000000D+00, &
   2.545637416870728D-06,   0.000000000000000D+00,   8.695313494149864D-05, &
   0.000000000000000D+00,   2.880974012223605D-05,   0.000000000000000D+00, &
   5.435617065507261D-06,   0.000000000000000D+00,  -2.508963049122620D-06, &
   0.000000000000000D+00,  -3.047977859300876D-06,   0.000000000000000D+00, &
   1.164860936589200D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.586989326444745D-07,   0.000000000000000D+00,  -4.011597539617705D-06, &
   0.000000000000000D+00,  -1.796596422933300D-05,   0.000000000000000D+00, &
  -2.062186227086454D-05,   0.000000000000000D+00,   8.375026813910205D-06, &
   0.000000000000000D+00,  -4.785258581947925D-06,   0.000000000000000D+00, &
  -4.047426605654686D-06,   0.000000000000000D+00,   5.670651344628880D-07, &
   0.000000000000000D+00,   8.044482253538260D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.654723965015260D-06,   0.000000000000000D+00, &
  -8.269890654143825D-07,   0.000000000000000D+00,  -3.726641646135911D-05, &
   0.000000000000000D+00,  -2.908011272539710D-05,   0.000000000000000D+00, &
  -9.532215347638338D-06,   0.000000000000000D+00,  -1.243919532234205D-06, &
   0.000000000000000D+00,   2.288473705492514D-06,   0.000000000000000D+00, &
   1.041164048114889D-06,   0.000000000000000D+00,  -1.394129652217666D-06 /)
lem(:,128) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.053039513875547D-04,  -6.243608129827166D-04, &
  -6.219324398459942D-04,   5.985152721086046D-03,  -3.390726358408680D-03, &
  -1.265583433244607D-02,   1.020758847872105D-02,   1.475714288555852D-02, &
  -1.353047770024631D-02,  -9.709426703060454D-03,   9.200642921415654D-03, &
   3.985544548295972D-03,  -3.120894331557200D-03,  -1.503633295493982D-03, &
   5.072009075954646D-04,   6.242383126423238D-04,  -4.862152355132702D-05, &
  -2.299548428189368D-04,  -5.845474376818794D-07,   9.571345419567090D-05, &
   3.439749702049197D-07,   0.000000000000000D+00,  -1.053039513875600D-04, &
   6.243608129827086D-04,   6.219324398460000D-04,  -5.985152721086035D-03, &
   3.390726358408671D-03,   1.265583433244607D-02,  -1.020758847872104D-02, &
  -1.475714288555853D-02,   1.353047770024630D-02,   9.709426703060445D-03, &
  -9.200642921415630D-03,  -3.985544548295969D-03,   3.120894331557196D-03, &
   1.503633295493987D-03,  -5.072009075954634D-04,  -6.242383126423259D-04, &
   4.862152355133082D-05,   2.299548428189346D-04,   5.845474376848962D-07, &
  -9.571345419566798D-05,  -3.439749702175987D-07,   0.000000000000000D+00, &
  -5.265197569378173D-05,   0.000000000000000D+00,  -2.073492882482717D-04, &
   0.000000000000000D+00,  -1.600065635551040D-04,   0.000000000000000D+00, &
   8.624286035146159D-05,   0.000000000000000D+00,   1.575199646853735D-04, &
   0.000000000000000D+00,  -5.254837377576721D-06,   0.000000000000000D+00, &
  -9.481725089948845D-05,   0.000000000000000D+00,   2.258243308966610D-07, &
   0.000000000000000D+00,   2.945324079560668D-05,   0.000000000000000D+00, &
  -8.129830019080735D-06,   0.000000000000000D+00,   8.844221979463883D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.162504162504108D-06, &
   0.000000000000000D+00,  -8.384861326037725D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.328510842442741D-05,   0.000000000000000D+00, &
  -4.381583328951042D-07,   0.000000000000000D+00,   1.394950079160605D-04, &
   0.000000000000000D+00,   9.271773977656256D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.486634452578732D-04,   0.000000000000000D+00, &
   1.342384579890121D-05,   0.000000000000000D+00,  -7.055637860592257D-07, &
   0.000000000000000D+00,   9.783388266360392D-05,   0.000000000000000D+00, &
  -1.362753419365540D-04,   0.000000000000000D+00,  -9.086415820161883D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.405875693532173D-04, &
   0.000000000000000D+00,   2.758703775952783D-04,   0.000000000000000D+00, &
   1.100760750444415D-05,   0.000000000000000D+00,   1.703606663897504D-06, &
   0.000000000000000D+00,  -6.220701943544861D-04,   0.000000000000000D+00, &
  -1.302086660142925D-04,   0.000000000000000D+00,   6.528914019828009D-05, &
   0.000000000000000D+00,   2.292093989327713D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.760397332883542D-04,   0.000000000000000D+00, &
  -3.628614397223898D-04,   0.000000000000000D+00,  -1.945342454243384D-04, &
   0.000000000000000D+00,  -2.094282435681002D-05,   0.000000000000000D+00, &
  -3.429486172105928D-06,   0.000000000000000D+00,   1.197083677902092D-04, &
   0.000000000000000D+00,   6.872354324905131D-04,   0.000000000000000D+00, &
   1.304502741018420D-04,   0.000000000000000D+00,   1.849017425642802D-06, &
   0.000000000000000D+00,   8.589458316404741D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.464094887346114D-04,   0.000000000000000D+00, &
  -4.001037774089332D-04,   0.000000000000000D+00,   2.266910888949524D-04, &
   0.000000000000000D+00,   8.692609111995986D-05,   0.000000000000000D+00, &
   8.232842713057487D-06,   0.000000000000000D+00,  -2.315614581786456D-06, &
   0.000000000000000D+00,   2.325979659402729D-04,   0.000000000000000D+00, &
  -1.590931508436267D-04,   0.000000000000000D+00,  -4.759081526227174D-04, &
   0.000000000000000D+00,  -9.883020727347027D-05,   0.000000000000000D+00, &
  -2.414513991952286D-05,   0.000000000000000D+00,  -8.299363992635323D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.948574645340843D-05, &
   0.000000000000000D+00,   3.831738044021551D-04,   0.000000000000000D+00, &
   2.502661803177953D-04,   0.000000000000000D+00,  -8.002596214609608D-05, &
   0.000000000000000D+00,  -1.472246244238813D-05,   0.000000000000000D+00, &
   5.273919698912038D-06,   0.000000000000000D+00,   7.364105820281047D-06, &
   0.000000000000000D+00,  -9.267713678666001D-06,   0.000000000000000D+00, &
  -2.274794824574780D-04,   0.000000000000000D+00,   1.579911173779101D-04, &
   0.000000000000000D+00,   2.216066669864345D-04,   0.000000000000000D+00, &
   5.124665001387523D-05,   0.000000000000000D+00,   1.394705947286474D-05, &
   0.000000000000000D+00,  -8.591396041273849D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.871399667363590D-06,   0.000000000000000D+00, &
   4.042675998354024D-05,   0.000000000000000D+00,  -2.691372861901551D-04, &
   0.000000000000000D+00,  -8.910337100729531D-05,   0.000000000000000D+00, &
   9.110223392120289D-06,   0.000000000000000D+00,  -8.142608059482852D-06, &
   0.000000000000000D+00,  -6.302111622812761D-06,   0.000000000000000D+00, &
  -2.279929680223774D-06,   0.000000000000000D+00,  -2.271425830094999D-06, &
   0.000000000000000D+00,   1.408723133771413D-05,   0.000000000000000D+00, &
   1.125713758475013D-04,   0.000000000000000D+00,  -1.189842100736601D-04, &
   0.000000000000000D+00,  -7.150676800602072D-05,   0.000000000000000D+00, &
  -1.737798595869266D-05,   0.000000000000000D+00,  -1.425960470867532D-06, &
   0.000000000000000D+00,   3.188602144484650D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.634094604587638D-07,   0.000000000000000D+00, &
  -4.473152128576444D-06,   0.000000000000000D+00,   9.083445199000418D-06, &
   0.000000000000000D+00,   1.214632515765755D-04,   0.000000000000000D+00, &
   1.066529053966639D-05,   0.000000000000000D+00,   4.503230546167889D-06, &
   0.000000000000000D+00,   6.172096971217263D-06,   0.000000000000000D+00, &
   2.247300467522461D-06,   0.000000000000000D+00,  -3.447031416391785D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.207207950311490D-06,   0.000000000000000D+00,  -1.496246308663882D-05, &
   0.000000000000000D+00,  -5.123634076511546D-06,   0.000000000000000D+00, &
   6.210723508397070D-05,   0.000000000000000D+00,   1.900057579026964D-05, &
   0.000000000000000D+00,   4.219805007630945D-06,   0.000000000000000D+00, &
  -2.321049370874294D-06,   0.000000000000000D+00,  -1.950892368552321D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.255375047983942D-06,   0.000000000000000D+00, &
  -3.326536907328953D-06,   0.000000000000000D+00,  -3.326226984985678D-05, &
   0.000000000000000D+00,  -2.356837253144726D-05,   0.000000000000000D+00, &
   4.673266208813699D-06,   0.000000000000000D+00,  -2.733278390431320D-06, &
   0.000000000000000D+00,  -2.404793014946834D-06,   0.000000000000000D+00, &
   1.369551031529902D-07,   0.000000000000000D+00,   3.100705285494694D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.073693338660516D-07,   0.000000000000000D+00,   3.671805441246071D-06, &
   0.000000000000000D+00,   4.682030758260896D-06,   0.000000000000000D+00, &
  -3.239099120009515D-05,   0.000000000000000D+00,  -2.116350932137317D-05, &
   0.000000000000000D+00,  -6.047842023183511D-06,   0.000000000000000D+00, &
  -5.582148735288890D-07,   0.000000000000000D+00,   1.686728061438089D-06, &
   0.000000000000000D+00,   1.180414168515859D-06,   0.000000000000000D+00 /)
lem(:,129) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.005215545901419D-04,   6.888633486602783D-04,  -1.590906165435777D-03, &
  -1.751197708740120D-03,  -1.947756384759439D-03,   7.750943546768650D-03, &
   9.483202938376090D-03,  -1.475995051915278D-02,  -1.214134102047792D-02, &
   1.507700553094930D-02,   6.170672262673326D-03,  -8.657656818853603D-03, &
  -5.298522567743681D-04,   3.039522502200541D-03,  -4.584347474103440D-04, &
  -9.613979831407058D-04,   1.021951956321003D-04,   3.867004453308165D-04, &
  -9.449096517895310D-06,  -1.645485601847485D-04,  -4.481818014133064D-06, &
   7.521490087477471D-05,   2.005215545901440D-04,   6.888633486602777D-04, &
  -1.590906165435775D-03,  -1.751197708740116D-03,  -1.947756384759443D-03, &
   7.750943546768636D-03,   9.483202938376097D-03,  -1.475995051915277D-02, &
  -1.214134102047792D-02,   1.507700553094931D-02,   6.170672262673334D-03, &
  -8.657656818853597D-03,  -5.298522567743782D-04,   3.039522502200533D-03, &
  -4.584347474103375D-04,  -9.613979831407007D-04,   1.021951956321061D-04, &
   3.867004453308069D-04,  -9.449096517903354D-06,  -1.645485601847315D-04, &
  -4.481818014136527D-06,   7.521490087476399D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.435312213753388D-04,   0.000000000000000D+00, &
   1.478092341699590D-04,   0.000000000000000D+00,   8.959891004675491D-05, &
   0.000000000000000D+00,  -1.776568749548330D-04,   0.000000000000000D+00, &
  -1.623496789979534D-04,   0.000000000000000D+00,   1.582563420351143D-04, &
   0.000000000000000D+00,   6.109770527033176D-05,   0.000000000000000D+00, &
  -6.350713981177748D-05,   0.000000000000000D+00,   6.223916808553403D-06, &
   0.000000000000000D+00,   2.669840229079386D-06,   0.000000000000000D+00, &
  -1.387501387501484D-06,  -5.737726325961615D-05,   0.000000000000000D+00, &
   2.677061500590888D-05,   0.000000000000000D+00,  -8.304603892839130D-06, &
   2.861431654001316D-04,   0.000000000000000D+00,   6.437705741111250D-06, &
   0.000000000000000D+00,  -4.083077225492086D-04,   0.000000000000000D+00, &
   4.785054127159400D-05,   0.000000000000000D+00,   1.830644487378229D-05, &
  -2.853055097637141D-04,   0.000000000000000D+00,  -2.005985500567543D-04, &
   0.000000000000000D+00,   4.698299493191188D-06,   0.000000000000000D+00, &
   9.106712860582840D-04,   0.000000000000000D+00,   1.897134298450087D-04, &
   0.000000000000000D+00,  -7.684680551360116D-05,   0.000000000000000D+00, &
  -1.884043915535782D-05,  -3.264149776129854D-04,   0.000000000000000D+00, &
   2.518990885948082D-04,   0.000000000000000D+00,   1.923050826103895D-04, &
   0.000000000000000D+00,   9.969629779349290D-06,   0.000000000000000D+00, &
  -5.355366877779045D-04,   0.000000000000000D+00,  -6.096092729976829D-04, &
   0.000000000000000D+00,  -1.346351860336906D-04,   0.000000000000000D+00, &
   4.296648275007987D-05,   0.000000000000000D+00,   2.506423089373916D-06, &
   4.502508305173536D-04,   0.000000000000000D+00,   2.805524957311871D-04, &
   0.000000000000000D+00,  -2.560942186604900D-04,   0.000000000000000D+00, &
  -1.398318844491572D-04,   0.000000000000000D+00,  -1.873769509142181D-05, &
   0.000000000000000D+00,   2.317428014022501D-05,   0.000000000000000D+00, &
   2.453567692832733D-04,   0.000000000000000D+00,   5.483700924667954D-04, &
   0.000000000000000D+00,   1.116643150146526D-04,   0.000000000000000D+00, &
   5.969407534694357D-06,   0.000000000000000D+00,   1.689569715498388D-05, &
  -1.177734486389766D-04,   0.000000000000000D+00,  -2.920587698148610D-04, &
   0.000000000000000D+00,  -2.815906424231907D-04,   0.000000000000000D+00, &
   1.658586449109706D-04,   0.000000000000000D+00,   6.318555253025537D-05, &
   0.000000000000000D+00,   7.112146232439419D-06,   0.000000000000000D+00, &
   1.727546619187492D-05,   0.000000000000000D+00,   8.136676194288230D-05, &
   0.000000000000000D+00,  -1.682626285113163D-04,   0.000000000000000D+00, &
  -3.756236851152707D-04,   0.000000000000000D+00,  -8.236452885116485D-05, &
   0.000000000000000D+00,  -2.254240913236835D-05,   0.000000000000000D+00, &
  -1.565918857161386D-05,   1.292783260932361D-05,   0.000000000000000D+00, &
   7.565531305476565D-09,   0.000000000000000D+00,   2.636923924127651D-04, &
   0.000000000000000D+00,   1.821998255254736D-04,   0.000000000000000D+00, &
  -5.820220652865165D-05,   0.000000000000000D+00,  -9.253317431204156D-06, &
   0.000000000000000D+00,   5.036816162872038D-06,   0.000000000000000D+00, &
  -3.822263274967433D-06,   0.000000000000000D+00,   7.088423579191361D-06, &
   0.000000000000000D+00,  -1.235404786675542D-04,   0.000000000000000D+00, &
   1.344140876422262D-04,   0.000000000000000D+00,   1.774159725901818D-04, &
   0.000000000000000D+00,   4.311852743007090D-05,   0.000000000000000D+00, &
   1.157793492771813D-05,   0.000000000000000D+00,  -8.369671088131815D-07, &
   2.802245802759323D-07,   0.000000000000000D+00,  -3.075364405869293D-06, &
   0.000000000000000D+00,   3.331808284799745D-05,   0.000000000000000D+00, &
  -1.876920174504279D-04,   0.000000000000000D+00,  -6.496643606240453D-05, &
   0.000000000000000D+00,   5.061225751834713D-06,   0.000000000000000D+00, &
  -7.339942190403773D-06,   0.000000000000000D+00,  -4.718098349448357D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.204155613709190D-06,   0.000000000000000D+00,  -6.777530891676200D-06, &
   0.000000000000000D+00,   7.507521522235254D-05,   0.000000000000000D+00, &
  -9.678779039761526D-05,   0.000000000000000D+00,  -5.879932830718967D-05, &
   0.000000000000000D+00,  -1.423293910335406D-05,   0.000000000000000D+00, &
  -1.038278771384379D-08,   0.000000000000000D+00,   8.642628061725914D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.203115333859049D-06, &
   0.000000000000000D+00,  -4.025302375402282D-06,   0.000000000000000D+00, &
  -6.585192858776349D-06,   0.000000000000000D+00,   8.627051996030386D-05, &
   0.000000000000000D+00,   6.308988330746158D-06,   0.000000000000000D+00, &
   3.942967856716201D-06,   0.000000000000000D+00,   5.439194917684101D-06, &
   0.000000000000000D+00,   1.421146787839603D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.971387090522987D-07, &
   0.000000000000000D+00,   3.272232464089589D-06,   0.000000000000000D+00, &
  -7.787831228089671D-06,   0.000000000000000D+00,  -4.645094181567548D-06, &
   0.000000000000000D+00,   5.015390938027009D-05,   0.000000000000000D+00, &
   1.648178254002429D-05,   0.000000000000000D+00,   2.622318048324806D-06, &
   0.000000000000000D+00,  -3.241389536715152D-06,   0.000000000000000D+00, &
  -4.890327885513695D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.086301526900967D-08,   0.000000000000000D+00,   1.991470014977699D-07, &
   0.000000000000000D+00,   5.916269507328889D-06,   0.000000000000000D+00, &
  -1.927648248867731D-05,   0.000000000000000D+00,  -1.479781010773592D-05, &
   0.000000000000000D+00,   4.850261067695398D-06,   0.000000000000000D+00, &
  -2.315369927837800D-06,   0.000000000000000D+00,  -2.416917396700609D-06, &
   0.000000000000000D+00,  -8.248827025013797D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.349531220327396D-07,   0.000000000000000D+00, &
   2.851098739653239D-07,   0.000000000000000D+00,   8.713167423094192D-06, &
   0.000000000000000D+00,  -2.368585808469806D-05,   0.000000000000000D+00, &
  -1.706254574857974D-05,   0.000000000000000D+00,  -5.085682206033279D-06, &
   0.000000000000000D+00,   3.733424272237198D-08,   0.000000000000000D+00, &
   2.135883659408610D-06,   0.000000000000000D+00,   7.076486756884091D-07 /)
lem(:,130) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.158343465263143D-03,  -5.107573781920161D-04, &
   4.614196899182208D-03,   3.531422469314607D-03,  -9.163000101160890D-03, &
  -8.079469218527306D-03,   1.385149063037660D-02,   6.857225739545373D-03, &
  -1.332150252200419D-02,  -8.115215430450386D-04,   6.925532502573767D-03, &
  -2.078733113036489D-03,  -1.617749774635797D-03,   1.342700180293335D-03, &
   9.073093983861335D-05,  -5.067763268043299D-04,   2.153924876578590D-05, &
   2.215271397867829D-04,  -1.996851868353787D-05,  -1.044595255281281D-04, &
   1.257580779574452D-05,   0.000000000000000D+00,   1.158343465263145D-03, &
   5.107573781920207D-04,  -4.614196899182206D-03,  -3.531422469314597D-03, &
   9.163000101160879D-03,   8.079469218527284D-03,  -1.385149063037660D-02, &
  -6.857225739545364D-03,   1.332150252200418D-02,   8.115215430450277D-04, &
  -6.925532502573758D-03,   2.078733113036491D-03,   1.617749774635789D-03, &
  -1.342700180293339D-03,  -9.073093983860273D-05,   5.067763268043326D-04, &
  -2.153924876579339D-05,  -2.215271397867857D-04,   1.996851868355012D-05, &
   1.044595255281262D-04,  -1.257580779576059D-05,   0.000000000000000D+00, &
   5.791717326315721D-04,   0.000000000000000D+00,   1.006101130603470D-03, &
   0.000000000000000D+00,  -1.016386661925488D-03,   0.000000000000000D+00, &
   9.204756790777368D-06,   0.000000000000000D+00,   5.841320855663433D-04, &
   0.000000000000000D+00,  -1.596594502224848D-04,   0.000000000000000D+00, &
  -2.229573726583948D-04,   0.000000000000000D+00,   9.577683376064203D-05, &
   0.000000000000000D+00,   1.883884115583418D-05,   0.000000000000000D+00, &
  -1.437696670898045D-05,   0.000000000000000D+00,   6.988228669480202D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.578754578754577D-05, &
   0.000000000000000D+00,  -1.077354018530485D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.056605932766924D-04,   0.000000000000000D+00, &
   2.556890134289517D-05,   0.000000000000000D+00,  -9.114552458205715D-05, &
   0.000000000000000D+00,   4.351835814684108D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.070223206446116D-04,   0.000000000000000D+00, &
  -1.534703256839480D-04,   0.000000000000000D+00,  -9.685124816703764D-06, &
   0.000000000000000D+00,   2.530741219877164D-04,   0.000000000000000D+00, &
   1.070099791167166D-04,   0.000000000000000D+00,  -5.702178436927237D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.220467063783794D-04, &
   0.000000000000000D+00,   1.848015859636476D-04,   0.000000000000000D+00, &
   1.611532873483124D-04,   0.000000000000000D+00,   3.055745670467116D-05, &
   0.000000000000000D+00,  -1.110138384670623D-04,   0.000000000000000D+00, &
  -3.908361899172376D-04,   0.000000000000000D+00,  -1.033258814003015D-04, &
   0.000000000000000D+00,   2.439065773790753D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.587500308727931D-04,   0.000000000000000D+00, &
   2.044670228285724D-04,   0.000000000000000D+00,  -2.028420981297989D-04, &
   0.000000000000000D+00,  -1.216421610206734D-04,   0.000000000000000D+00, &
  -4.332258637007663D-05,   0.000000000000000D+00,  -4.807302895874674D-05, &
   0.000000000000000D+00,   1.304502741018420D-04,   0.000000000000000D+00, &
   4.146522878054532D-04,   0.000000000000000D+00,   9.736063086688592D-05, &
   0.000000000000000D+00,   1.573540840581331D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.896790620575804D-05,   0.000000000000000D+00, &
  -2.184264113864330D-04,   0.000000000000000D+00,  -2.213460670179171D-04, &
   0.000000000000000D+00,   1.345404445894484D-04,   0.000000000000000D+00, &
   5.778352738626142D-05,   0.000000000000000D+00,   1.652449272553068D-05, &
   0.000000000000000D+00,   2.424483579416958D-05,   0.000000000000000D+00, &
   8.888833680191857D-05,   0.000000000000000D+00,  -1.217940181496522D-04, &
   0.000000000000000D+00,  -3.001202127559596D-04,   0.000000000000000D+00, &
  -7.473558948090766D-05,   0.000000000000000D+00,  -2.242114376126916D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.563074900744986D-06, &
   0.000000000000000D+00,   8.917739410907402D-06,   0.000000000000000D+00, &
   2.161165022863000D-04,   0.000000000000000D+00,   1.471681246976150D-04, &
   0.000000000000000D+00,  -4.828987666212564D-05,   0.000000000000000D+00, &
  -9.289457057230374D-06,   0.000000000000000D+00,   1.491703446999451D-05, &
   0.000000000000000D+00,  -3.172706888573543D-06,   0.000000000000000D+00, &
  -1.235728445556195D-05,   0.000000000000000D+00,  -9.784587981998258D-05, &
   0.000000000000000D+00,   1.082781125951815D-04,   0.000000000000000D+00, &
   1.458345543697444D-04,   0.000000000000000D+00,   3.809950946690497D-05, &
   0.000000000000000D+00,   7.785756300830926D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.397933568495767D-06,   0.000000000000000D+00, &
  -2.638226001634100D-06,   0.000000000000000D+00,   8.616975666913328D-06, &
   0.000000000000000D+00,  -1.557504773501397D-04,   0.000000000000000D+00, &
  -5.301361444670677D-05,   0.000000000000000D+00,   5.075437303649700D-06, &
   0.000000000000000D+00,  -8.123001764921347D-06,   0.000000000000000D+00, &
  -1.732099898508383D-05,   0.000000000000000D+00,  -2.209373963500312D-07, &
   0.000000000000000D+00,   2.853980085612503D-06,   0.000000000000000D+00, &
   9.794166784057983D-08,   0.000000000000000D+00,   5.420776223777919D-05, &
   0.000000000000000D+00,  -8.090361171637975D-05,   0.000000000000000D+00, &
  -4.860688186348680D-05,   0.000000000000000D+00,  -1.113751839126320D-05, &
   0.000000000000000D+00,   1.179800188833615D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.227129212752654D-08,   0.000000000000000D+00, &
   2.304772968589756D-07,   0.000000000000000D+00,   3.733198920217930D-06, &
   0.000000000000000D+00,   4.168433784289642D-06,   0.000000000000000D+00, &
   7.213335708563619D-05,   0.000000000000000D+00,   4.458250584561109D-06, &
   0.000000000000000D+00,   3.201064223635796D-06,   0.000000000000000D+00, &
   7.585758815987931D-06,   0.000000000000000D+00,   4.711169471758985D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.861465742692682D-07,   0.000000000000000D+00,   4.295083089298563D-07, &
   0.000000000000000D+00,  -3.006744345120459D-06,   0.000000000000000D+00, &
   1.113143233626295D-06,   0.000000000000000D+00,   4.244061300994342D-05, &
   0.000000000000000D+00,   1.274982242527812D-05,   0.000000000000000D+00, &
   1.883472012827750D-06,   0.000000000000000D+00,  -3.657330075137049D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.730723722079702D-07,   0.000000000000000D+00, &
  -1.759407916300222D-06,   0.000000000000000D+00,   2.315669846884759D-07, &
   0.000000000000000D+00,  -1.825470033957734D-05,   0.000000000000000D+00, &
  -1.215900843371170D-05,   0.000000000000000D+00,   4.371059773374262D-06, &
   0.000000000000000D+00,  -2.770402371971401D-06,   0.000000000000000D+00, &
  -3.119425045727307D-06,   0.000000000000000D+00,   1.927518153725799D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.082074180776383D-08,   0.000000000000000D+00,   3.352981115815769D-07, &
   0.000000000000000D+00,  -2.299635590798435D-06,   0.000000000000000D+00, &
   3.028124471899634D-06,   0.000000000000000D+00,  -2.228094418100337D-05, &
   0.000000000000000D+00,  -1.397871707361348D-05,   0.000000000000000D+00, &
  -3.645595914459904D-06,   0.000000000000000D+00,  -1.009148949458107D-06, &
   0.000000000000000D+00,   2.319368031140716D-07,   0.000000000000000D+00 /)
lem(:,131) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.104001440384711D-03,  -3.970836500239194D-04,  -1.521621812564434D-03, &
  -4.107474379286623D-03,  -1.361046055504859D-03,   8.835410345662632D-03, &
   5.504683030587929D-03,  -1.149931964312052D-02,  -3.087380653058470D-03, &
   8.989673248140911D-03,  -2.326184952863507D-03,  -3.172604211445882D-03, &
   3.007976085696509D-03,  -2.872539407270409D-04,  -1.043776938295220D-03, &
   6.145188266550080D-04,   1.239889831190321D-04,  -2.772368469759070D-04, &
   1.529192431523139D-05,   1.256998922469744D-04,  -2.798041528898461D-05, &
  -5.984651876634195D-05,   1.104001440384711D-03,  -3.970836500239171D-04, &
  -1.521621812564439D-03,  -4.107474379286641D-03,  -1.361046055504861D-03, &
   8.835410345662636D-03,   5.504683030587927D-03,  -1.149931964312054D-02, &
  -3.087380653058476D-03,   8.989673248140912D-03,  -2.326184952863509D-03, &
  -3.172604211445879D-03,   3.007976085696507D-03,  -2.872539407270452D-04, &
  -1.043776938295221D-03,   6.145188266550103D-04,   1.239889831190305D-04, &
  -2.772368469759044D-04,   1.529192431522812D-05,   1.256998922469676D-04, &
  -2.798041528897786D-05,  -5.984651876633989D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.138250105626076D-04,   0.000000000000000D+00, &
  -2.505916572628603D-03,   0.000000000000000D+00,   2.793887903593416D-03, &
   0.000000000000000D+00,  -1.342140618307534D-03,   0.000000000000000D+00, &
  -3.337942685810891D-04,   0.000000000000000D+00,   6.090566725976731D-04, &
   0.000000000000000D+00,  -7.387177474851710D-05,   0.000000000000000D+00, &
  -9.463019518467251D-05,   0.000000000000000D+00,   3.336840695258702D-05, &
   0.000000000000000D+00,  -1.100284387248639D-05,   0.000000000000000D+00, &
  -7.388444888444893D-05,  -1.271196124137298D-05,   0.000000000000000D+00, &
   1.219368866427690D-04,   0.000000000000000D+00,   1.318636428930544D-05, &
   1.018932907477779D-05,   0.000000000000000D+00,   5.799181253902614D-05, &
   0.000000000000000D+00,  -1.874914945735377D-04,   0.000000000000000D+00, &
  -8.895688075254630D-05,   0.000000000000000D+00,   4.744725169696161D-05, &
   1.301641684768624D-05,   0.000000000000000D+00,  -6.319574818026830D-05, &
   0.000000000000000D+00,  -1.226097966253733D-04,   0.000000000000000D+00, &
   1.779228898036948D-04,   0.000000000000000D+00,   2.051424023173247D-04, &
   0.000000000000000D+00,   9.673791631270351D-05,   0.000000000000000D+00, &
  -8.153067424062014D-05,  -3.613879408905631D-05,   0.000000000000000D+00, &
  -7.165726075572628D-05,   0.000000000000000D+00,   1.392281142273782D-04, &
   0.000000000000000D+00,   1.557580354829735D-04,   0.000000000000000D+00, &
  -1.005601389233101D-04,   0.000000000000000D+00,  -1.170001696864328D-04, &
   0.000000000000000D+00,  -3.148143824669320D-04,   0.000000000000000D+00, &
  -1.009112005055900D-04,   0.000000000000000D+00,   4.029200023824772D-05, &
   3.274533157424527D-05,   0.000000000000000D+00,   9.081530662917788D-05, &
   0.000000000000000D+00,   1.521563814979838D-04,   0.000000000000000D+00, &
  -1.691915715111344D-04,   0.000000000000000D+00,  -1.222217490003373D-04, &
   0.000000000000000D+00,   2.986942724794428D-05,   0.000000000000000D+00, &
   3.818348628383754D-06,   0.000000000000000D+00,   1.116643150146526D-04, &
   0.000000000000000D+00,   3.523237704412915D-04,   0.000000000000000D+00, &
   9.582865648033238D-05,   0.000000000000000D+00,   3.411266800879887D-05, &
  -1.384340334737457D-05,   0.000000000000000D+00,  -1.443265330985753D-05, &
   0.000000000000000D+00,  -1.543531369240793D-04,   0.000000000000000D+00, &
  -1.792195817357684D-04,   0.000000000000000D+00,   1.183437083977499D-04, &
   0.000000000000000D+00,   5.001349589328033D-05,   0.000000000000000D+00, &
  -2.699805205189688D-06,   0.000000000000000D+00,   4.859943529345893D-07, &
   0.000000000000000D+00,   5.845249249738822D-05,   0.000000000000000D+00, &
  -1.054862833677366D-04,   0.000000000000000D+00,  -2.637057042944167D-04, &
   0.000000000000000D+00,  -7.298786530921059D-05,   0.000000000000000D+00, &
  -5.930753807795972D-05,   2.268287879256140D-06,   0.000000000000000D+00, &
  -2.070657544487864D-06,   0.000000000000000D+00,  -1.806570464814566D-06, &
   0.000000000000000D+00,   1.706392818149322D-04,   0.000000000000000D+00, &
   1.218654176663793D-04,   0.000000000000000D+00,  -4.202678460051083D-05, &
   0.000000000000000D+00,  -1.117123432387641D-06,   0.000000000000000D+00, &
  -4.005027929794282D-07,   0.000000000000000D+00,   1.971398748822779D-06, &
   0.000000000000000D+00,  -7.460907856328600D-06,   0.000000000000000D+00, &
  -7.688818814744920D-05,   0.000000000000000D+00,   9.681345961855054D-05, &
   0.000000000000000D+00,   1.294585324598228D-04,   0.000000000000000D+00, &
   4.031971149941455D-05,   0.000000000000000D+00,   2.906822882765879D-05, &
   2.274550164557623D-08,   0.000000000000000D+00,   4.028930475335123D-07, &
   0.000000000000000D+00,   4.427144890258750D-06,   0.000000000000000D+00, &
   8.111840601581885D-06,   0.000000000000000D+00,  -1.289374588206117D-04, &
   0.000000000000000D+00,  -4.271901324738573D-05,   0.000000000000000D+00, &
   1.566447577307092D-06,   0.000000000000000D+00,  -6.621686742311856D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.002643469503041D-07,   0.000000000000000D+00,   4.057592546768529D-07, &
   0.000000000000000D+00,   5.632980174291352D-06,   0.000000000000000D+00, &
   4.308874643146214D-05,   0.000000000000000D+00,  -7.259342262414246D-05, &
   0.000000000000000D+00,  -4.325227575227905D-05,   0.000000000000000D+00, &
  -1.535913025669427D-05,   0.000000000000000D+00,   1.698922258805447D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.719625063055478D-07, &
   0.000000000000000D+00,  -1.516501586004559D-06,   0.000000000000000D+00, &
  -1.250985247087169D-06,   0.000000000000000D+00,   6.247854483515517D-06, &
   0.000000000000000D+00,   5.976969387622577D-05,   0.000000000000000D+00, &
   2.846423405964955D-06,   0.000000000000000D+00,   3.922552855701593D-06, &
   0.000000000000000D+00,   1.685543023738087D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.049958218260547D-08, &
   0.000000000000000D+00,   2.608444597796019D-07,   0.000000000000000D+00, &
  -1.747969351482100D-06,   0.000000000000000D+00,  -7.095055076554729D-06, &
   0.000000000000000D+00,   3.035825373699615D-06,   0.000000000000000D+00, &
   3.715043158846705D-05,   0.000000000000000D+00,   1.268990140051789D-05, &
   0.000000000000000D+00,   2.434950622542183D-06,   0.000000000000000D+00, &
  -8.444380371164896D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.044271928314429D-10,   0.000000000000000D+00,   2.384869169906713D-07, &
   0.000000000000000D+00,   4.211441139051364D-07,   0.000000000000000D+00, &
   1.772748936531506D-08,   0.000000000000000D+00,  -1.748036678273068D-05, &
   0.000000000000000D+00,  -8.802071815477001D-06,   0.000000000000000D+00, &
   3.525118946399669D-06,   0.000000000000000D+00,  -7.071086856528417D-07, &
   0.000000000000000D+00,  -1.320123966708196D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.236631231625098D-07,   0.000000000000000D+00, &
   7.687196453576160D-07,   0.000000000000000D+00,   8.613822402368922D-07, &
   0.000000000000000D+00,   3.777299959487109D-06,   0.000000000000000D+00, &
  -1.979698309666543D-05,   0.000000000000000D+00,  -1.208174871206214D-05, &
   0.000000000000000D+00,  -4.399283439126640D-06,   0.000000000000000D+00, &
   1.845635624488169D-06,   0.000000000000000D+00,   4.804607840005181D-06 /)
lem(:,132) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.921584770335169D-05,   1.128813251441446D-03, &
   3.509602259600582D-03,   1.662161175542877D-03,  -6.296707758168897D-03, &
  -3.702946475797174D-03,   7.605311353733189D-03,   6.718568957311242D-04, &
  -4.893717428617944D-03,   2.572409314141775D-03,   3.551610591558013D-04, &
  -1.700929250580698D-03,   1.226840085266282D-03,  -1.484804145932533D-05, &
  -5.547960850394113D-04,   3.186756099841508D-04,   6.858984589929216D-05, &
  -1.574457647176503D-04,   2.154601534568428D-05,   6.877604070794489D-05, &
  -2.805294038567844D-05,   0.000000000000000D+00,  -8.921584770334524D-05, &
  -1.128813251441449D-03,  -3.509602259600580D-03,  -1.662161175542875D-03, &
   6.296707758168882D-03,   3.702946475797167D-03,  -7.605311353733197D-03, &
  -6.718568957311249D-04,   4.893717428617935D-03,  -2.572409314141776D-03, &
  -3.551610591557957D-04,   1.700929250580701D-03,  -1.226840085266285D-03, &
   1.484804145932323D-05,   5.547960850394152D-04,  -3.186756099841512D-04, &
  -6.858984589929746D-05,   1.574457647176529D-04,  -2.154601534567710D-05, &
  -6.877604070795133D-05,   2.805294038567473D-05,   0.000000000000000D+00, &
  -4.460792385167036D-05,   0.000000000000000D+00,  -2.744443391237008D-03, &
   0.000000000000000D+00,   4.028384502711314D-03,   0.000000000000000D+00, &
  -4.057298744084387D-03,   0.000000000000000D+00,   3.232814928777646D-03, &
   0.000000000000000D+00,  -1.185561609654426D-03,   0.000000000000000D+00, &
  -2.749941746163928D-04,   0.000000000000000D+00,   2.897483594943366D-04, &
   0.000000000000000D+00,  -5.154900577984603D-05,   0.000000000000000D+00, &
   8.813459535171990D-06,   0.000000000000000D+00,  -7.029888188600504D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.526566026566070D-06, &
   0.000000000000000D+00,   3.188872674166790D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.277417170606018D-06,   0.000000000000000D+00, &
   9.577652524633944D-05,   0.000000000000000D+00,  -5.432357889788221D-05, &
   0.000000000000000D+00,  -7.725440194480440D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.023228577408135D-05,   0.000000000000000D+00, &
  -5.120465404908122D-05,   0.000000000000000D+00,  -2.360973018746822D-04, &
   0.000000000000000D+00,   5.254033891804797D-05,   0.000000000000000D+00, &
   1.517048053425758D-04,   0.000000000000000D+00,   1.218627662005178D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.787452435790028D-05, &
   0.000000000000000D+00,  -5.800408087896305D-05,   0.000000000000000D+00, &
   1.284225384220966D-04,   0.000000000000000D+00,   3.133671605465000D-04, &
   0.000000000000000D+00,  -2.482585939309587D-05,   0.000000000000000D+00, &
  -8.243643900781488D-05,   0.000000000000000D+00,  -2.777807301753789D-04, &
   0.000000000000000D+00,  -1.450484150802373D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.456358228501178D-05,   0.000000000000000D+00, &
   7.705397625810862D-05,   0.000000000000000D+00,   1.293332700576483D-04, &
   0.000000000000000D+00,  -1.634131432262070D-04,   0.000000000000000D+00, &
  -2.713203703747845D-04,   0.000000000000000D+00,   1.757465960252409D-06, &
   0.000000000000000D+00,   1.849017425642802D-06,   0.000000000000000D+00, &
   9.736063086688592D-05,   0.000000000000000D+00,   3.293618351031652D-04, &
   0.000000000000000D+00,   1.355853610355408D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.289440698446018D-06,   0.000000000000000D+00, &
  -2.316595261052787D-05,   0.000000000000000D+00,  -1.316539017309434D-04, &
   0.000000000000000D+00,  -1.576946665792951D-04,   0.000000000000000D+00, &
   1.197311500340001D-04,   0.000000000000000D+00,   1.376116182365999D-04, &
   0.000000000000000D+00,   2.942059199630862D-06,   0.000000000000000D+00, &
   2.805438219020050D-06,   0.000000000000000D+00,   4.096678536321741D-05, &
   0.000000000000000D+00,  -9.945245424553712D-05,   0.000000000000000D+00, &
  -2.554493971097046D-04,   0.000000000000000D+00,  -8.725984125455901D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.748603385109303D-07, &
   0.000000000000000D+00,   4.630130420805188D-06,   0.000000000000000D+00, &
   7.436333483346484D-06,   0.000000000000000D+00,   1.494588786492193D-04, &
   0.000000000000000D+00,   1.102159814161934D-04,   0.000000000000000D+00, &
  -4.908697531600557D-05,   0.000000000000000D+00,  -1.243132798342320D-05, &
   0.000000000000000D+00,  -7.878563134604115D-07,   0.000000000000000D+00, &
  -2.287039036659137D-07,   0.000000000000000D+00,  -4.615683456645781D-06, &
   0.000000000000000D+00,  -5.834681952897042D-05,   0.000000000000000D+00, &
   9.282429868468195D-05,   0.000000000000000D+00,   1.291700834343296D-04, &
   0.000000000000000D+00,   3.135105585545837D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.053478330170487D-07,   0.000000000000000D+00, &
  -1.213769810284105D-06,   0.000000000000000D+00,   6.773893908721552D-07, &
   0.000000000000000D+00,   7.049783355227438D-07,   0.000000000000000D+00, &
  -1.152818782428697D-04,   0.000000000000000D+00,  -3.865010931643694D-05, &
   0.000000000000000D+00,   5.854727113496462D-06,   0.000000000000000D+00, &
  -3.332913058303436D-05,   0.000000000000000D+00,  -2.400191983188210D-08, &
   0.000000000000000D+00,   1.633112302464384D-07,   0.000000000000000D+00, &
  -1.463627857221856D-06,   0.000000000000000D+00,   4.563545220264987D-06, &
   0.000000000000000D+00,   3.242836598611326D-05,   0.000000000000000D+00, &
  -7.027994851623639D-05,   0.000000000000000D+00,  -4.160850671271564D-05, &
   0.000000000000000D+00,  -5.813038998031579D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.071303280922049D-09,   0.000000000000000D+00, &
   2.344149733278557D-07,   0.000000000000000D+00,   2.438832718963117D-07, &
   0.000000000000000D+00,  -1.876906404051227D-06,   0.000000000000000D+00, &
   9.199860760341528D-06,   0.000000000000000D+00,   5.435353421845499D-05, &
   0.000000000000000D+00,   6.127497863439598D-07,   0.000000000000000D+00, &
   8.149628622007439D-06,   0.000000000000000D+00,   2.342674229118307D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.125432272915156D-07,   0.000000000000000D+00,   6.539326981029514D-07, &
   0.000000000000000D+00,   7.390052404078627D-07,   0.000000000000000D+00, &
  -6.667123032080213D-06,   0.000000000000000D+00,   5.910208229451751D-06, &
   0.000000000000000D+00,   3.651262591746097D-05,   0.000000000000000D+00, &
   9.002825544332542D-06,   0.000000000000000D+00,   4.050972657024017D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.163333198808223D-08,   0.000000000000000D+00, &
  -2.669824271933597D-07,   0.000000000000000D+00,   5.932512228391311D-07, &
   0.000000000000000D+00,   1.412765026244330D-06,   0.000000000000000D+00, &
  -1.723081853058295D-05,   0.000000000000000D+00,  -7.603648051882227D-06, &
   0.000000000000000D+00,   4.050832712304216D-06,   0.000000000000000D+00, &
  -7.603024508401008D-06,   0.000000000000000D+00,  -7.952968596315961D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.958954570913974D-09,   0.000000000000000D+00,  -6.938194734377430D-08, &
   0.000000000000000D+00,  -3.841334256853866D-07,   0.000000000000000D+00, &
   1.368069074859399D-07,   0.000000000000000D+00,   3.441140760426020D-06, &
   0.000000000000000D+00,  -1.994339794718193D-05,   0.000000000000000D+00, &
  -1.115288743586457D-05,   0.000000000000000D+00,  -3.974873565077538D-06, &
   0.000000000000000D+00,  -4.070896680885770D-06,   0.000000000000000D+00 /)
lem(:,133) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.822537010999459D-04,   9.019393127960350D-04,  -9.352838897921411D-04, &
  -3.380499961143928D-03,  -1.815311828931800D-03,   4.217619535368518D-03, &
   3.119568540471909D-03,  -4.136316102324880D-03,  -3.357278472798047D-04, &
   2.047612825942346D-03,  -1.531861316509928D-03,   5.218196104310761D-04, &
   3.617588819618986D-04,  -7.648977972595573D-04,   5.217360796888261D-04, &
   8.808951098549595D-07,  -2.692159299277324D-04,   1.711112437964232D-04, &
   2.549502367932634D-05,  -8.634729062661154D-05,   2.142888891068259D-05, &
   3.490592025589681D-05,   1.822537010999387D-04,   9.019393127960440D-04, &
  -9.352838897921364D-04,  -3.380499961143926D-03,  -1.815311828931804D-03, &
   4.217619535368524D-03,   3.119568540471905D-03,  -4.136316102324877D-03, &
  -3.357278472798003D-04,   2.047612825942337D-03,  -1.531861316509933D-03, &
   5.218196104310762D-04,   3.617588819618980D-04,  -7.648977972595574D-04, &
   5.217360796888235D-04,   8.808951098541732D-07,  -2.692159299277247D-04, &
   1.711112437964217D-04,   2.549502367932251D-05,  -8.634729062661104D-05, &
   2.142888891068263D-05,   3.490592025589908D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.606709869065935D-03,   0.000000000000000D+00, &
   1.985957395222276D-03,   0.000000000000000D+00,   5.159100608347577D-04, &
   0.000000000000000D+00,  -1.002249217117879D-03,   0.000000000000000D+00, &
  -1.514978130093472D-03,   0.000000000000000D+00,   2.099735705693194D-03, &
   0.000000000000000D+00,  -6.920254672790410D-04,   0.000000000000000D+00, &
   3.720569231308135D-06,   0.000000000000000D+00,   2.008627643626479D-05, &
   0.000000000000000D+00,  -1.186427792221932D-05,   0.000000000000000D+00, &
   3.100198412698465D-06,   1.648933174668498D-06,   0.000000000000000D+00, &
   1.358680127062464D-05,   0.000000000000000D+00,   8.362342059492802D-05, &
   1.619436241146775D-05,   0.000000000000000D+00,  -1.234015087294909D-06, &
   0.000000000000000D+00,  -2.661543392967548D-05,   0.000000000000000D+00, &
  -5.639273262749467D-05,   0.000000000000000D+00,  -1.801404958453140D-04, &
  -3.008007071474558D-05,   0.000000000000000D+00,   5.920512674817986D-06, &
   0.000000000000000D+00,  -2.144213849660830D-05,   0.000000000000000D+00, &
   2.866319320943947D-05,   0.000000000000000D+00,   4.554515925573399D-05, &
   0.000000000000000D+00,   1.498204409297917D-04,   0.000000000000000D+00, &
   2.293990876637392D-04,   2.217485809553217D-05,   0.000000000000000D+00, &
  -1.004234707642597D-05,   0.000000000000000D+00,  -5.242333188754617D-05, &
   0.000000000000000D+00,   8.018293955422595D-05,   0.000000000000000D+00, &
  -1.957309455269297D-05,   0.000000000000000D+00,  -2.235424827762496D-05, &
   0.000000000000000D+00,  -7.675893384494712D-05,   0.000000000000000D+00, &
  -2.719388945715290D-04,   0.000000000000000D+00,  -2.492478641572056D-04, &
  -6.591982742247004D-06,   0.000000000000000D+00,   7.382336836850029D-06, &
   0.000000000000000D+00,   6.084306382007429D-05,   0.000000000000000D+00, &
   1.224504929255934D-04,   0.000000000000000D+00,  -1.114780210871813D-04, &
   0.000000000000000D+00,   7.557425804329895D-06,   0.000000000000000D+00, &
   4.018570766860828D-06,   0.000000000000000D+00,   5.969407534694357D-06, &
   0.000000000000000D+00,   9.582865648033238D-05,   0.000000000000000D+00, &
   3.311778137954876D-04,   0.000000000000000D+00,   2.707870433208345D-04, &
  -2.187733334253281D-07,   0.000000000000000D+00,  -1.507048393353321D-06, &
   0.000000000000000D+00,  -1.896314196111656D-05,   0.000000000000000D+00, &
  -1.166212066834641D-04,   0.000000000000000D+00,  -1.562943738642568D-04, &
   0.000000000000000D+00,   6.444754768898556D-05,   0.000000000000000D+00, &
  -1.147937245073750D-06,   0.000000000000000D+00,   7.328361831722457D-07, &
   0.000000000000000D+00,   4.579680978495968D-06,   0.000000000000000D+00, &
   2.988698618910063D-05,   0.000000000000000D+00,  -1.027977146684439D-04, &
   0.000000000000000D+00,  -2.680201660626502D-04,   0.000000000000000D+00, &
  -2.421479161376912D-04,   3.763717034551942D-07,   0.000000000000000D+00, &
  -7.555583329428392D-07,   0.000000000000000D+00,   2.372841784801263D-06, &
   0.000000000000000D+00,   1.142531864432103D-05,   0.000000000000000D+00, &
   1.420032733962687D-04,   0.000000000000000D+00,   1.167159037084532D-04, &
   0.000000000000000D+00,  -4.704655662233207D-06,   0.000000000000000D+00, &
  -4.764532723320626D-08,   0.000000000000000D+00,  -5.269395198611388D-08, &
   0.000000000000000D+00,  -2.109224615305695D-06,   0.000000000000000D+00, &
  -7.209397371890993D-06,   0.000000000000000D+00,  -4.866206603809990D-05, &
   0.000000000000000D+00,   9.818567839302858D-05,   0.000000000000000D+00, &
   1.513813116493542D-04,   0.000000000000000D+00,   1.431251620019389D-04, &
   2.193316229750130D-09,   0.000000000000000D+00,   2.723886389106113D-07, &
   0.000000000000000D+00,   1.733220991355529D-07,   0.000000000000000D+00, &
   3.705154089221359D-07,   0.000000000000000D+00,  -5.703524293570526D-06, &
   0.000000000000000D+00,  -1.141619893612672D-04,   0.000000000000000D+00, &
  -4.916368955588300D-05,   0.000000000000000D+00,  -5.400913022095283D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.133991606659040D-07,   0.000000000000000D+00,   5.386354731886684D-07, &
   0.000000000000000D+00,   1.381539638182079D-07,   0.000000000000000D+00, &
   8.194907276697311D-06,   0.000000000000000D+00,   2.605799741837081D-05, &
   0.000000000000000D+00,  -7.517676487908392D-05,   0.000000000000000D+00, &
  -6.729508500318408D-05,   0.000000000000000D+00,  -4.803752585565731D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.078355443093395D-08, &
   0.000000000000000D+00,  -1.660636380429425D-07,   0.000000000000000D+00, &
   5.630359244537728D-07,   0.000000000000000D+00,  -2.368181098803722D-06, &
   0.000000000000000D+00,   1.496834962015458D-05,   0.000000000000000D+00, &
   5.482594733957814D-05,   0.000000000000000D+00,   6.121271204251234D-06, &
   0.000000000000000D+00,  -5.755918757993510D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.827320431096988D-09, &
   0.000000000000000D+00,  -7.819566411270720D-08,   0.000000000000000D+00, &
  -1.650086617860320D-07,   0.000000000000000D+00,   7.439730579878235D-07, &
   0.000000000000000D+00,  -9.566556391863500D-06,   0.000000000000000D+00, &
   1.008520226971179D-05,   0.000000000000000D+00,   4.251096049027573D-05, &
   0.000000000000000D+00,   2.428868511917115D-05,   0.000000000000000D+00, &
   6.029054891363018D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.284341571017151D-11,   0.000000000000000D+00,   5.845532571458012D-08, &
   0.000000000000000D+00,  -2.659568519534415D-07,   0.000000000000000D+00, &
  -2.209764668434328D-07,   0.000000000000000D+00,   1.430968461028530D-06, &
   0.000000000000000D+00,  -2.026258996734342D-05,   0.000000000000000D+00, &
  -8.064143530421678D-06,   0.000000000000000D+00,   8.697012826207119D-06, &
   0.000000000000000D+00,   5.959008684499491D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.480850315581006D-08,   0.000000000000000D+00, &
   1.471699098429258D-07,   0.000000000000000D+00,  -2.847073080621138D-07, &
   0.000000000000000D+00,   3.631130943761542D-07,   0.000000000000000D+00, &
   3.553417873926694D-06,   0.000000000000000D+00,  -2.223952061956892D-05, &
   0.000000000000000D+00,  -1.714145405379967D-05,   0.000000000000000D+00, &
  -4.203012677273291D-06,   0.000000000000000D+00,   2.032222140240467D-06 /)
lem(:,134) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.058655491447259D-04,   1.304986532775509D-03, &
   9.853810086289472D-04,   1.790091758719960D-03,  -1.425916551410405D-03, &
  -3.082364377921164D-03,   2.004132303952990D-03,   7.777985725478819D-04, &
  -1.036909532588873D-03,   6.459814012004521D-04,  -3.993242866424144D-04, &
   8.404258792718917D-05,   2.326481562182589D-04,  -3.579189410170321D-04, &
   2.251503933675068D-04,   8.773560441600743D-06,  -1.320695862805792D-04, &
   9.000075899963973D-05,   9.321111194192974D-06,  -4.830887180413257D-05, &
   1.647282557052139D-05,   0.000000000000000D+00,  -7.058655491447264D-04, &
  -1.304986532775508D-03,  -9.853810086289450D-04,  -1.790091758719964D-03, &
   1.425916551410405D-03,   3.082364377921164D-03,  -2.004132303952988D-03, &
  -7.777985725478834D-04,   1.036909532588871D-03,  -6.459814012004461D-04, &
   3.993242866424164D-04,  -8.404258792719169D-05,  -2.326481562182556D-04, &
   3.579189410170352D-04,  -2.251503933675107D-04,  -8.773560441603005D-06, &
   1.320695862805791D-04,  -9.000075899963603D-05,  -9.321111194195462D-06, &
   4.830887180412988D-05,  -1.647282557051684D-05,   0.000000000000000D+00, &
  -3.529327745723648D-04,   0.000000000000000D+00,  -5.918862565735108D-05, &
   0.000000000000000D+00,   5.583293711429131D-03,   0.000000000000000D+00, &
  -1.472695766034419D-02,   0.000000000000000D+00,   1.587895258898031D-02, &
   0.000000000000000D+00,  -7.409981526841409D-03,   0.000000000000000D+00, &
   6.393782726493113D-04,   0.000000000000000D+00,   5.223163112989417D-04, &
   0.000000000000000D+00,  -1.571214952016887D-04,   0.000000000000000D+00, &
   5.397190776653149D-05,   0.000000000000000D+00,  -2.141108850094855D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.790178571428573D-05, &
   0.000000000000000D+00,   1.033399470899466D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.224395680858549D-06,   0.000000000000000D+00, &
  -1.449152640982166D-05,   0.000000000000000D+00,  -1.169965977435013D-05, &
   0.000000000000000D+00,  -2.572023980813648D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.486749261517078D-06,   0.000000000000000D+00, &
   3.299997404341478D-06,   0.000000000000000D+00,  -1.454936693423535D-04, &
   0.000000000000000D+00,   1.197639488564411D-05,   0.000000000000000D+00, &
   3.353718448411752D-05,   0.000000000000000D+00,   1.296707453435151D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.629171784677269D-06, &
   0.000000000000000D+00,  -5.492046836303626D-06,   0.000000000000000D+00, &
  -1.746717374057234D-05,   0.000000000000000D+00,   4.584096863165207D-04, &
   0.000000000000000D+00,  -9.651939404640043D-06,   0.000000000000000D+00, &
  -2.422156607428224D-05,   0.000000000000000D+00,  -8.115307150273049D-05, &
   0.000000000000000D+00,  -3.363096225328414D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.058704624334308D-06,   0.000000000000000D+00, &
   8.087552183378065D-06,   0.000000000000000D+00,   4.942665780695474D-05, &
   0.000000000000000D+00,   7.138808518086386D-05,   0.000000000000000D+00, &
  -6.543186757528847D-04,   0.000000000000000D+00,   3.944545605603610D-06, &
   0.000000000000000D+00,   8.589458316404741D-06,   0.000000000000000D+00, &
   1.573540840581331D-05,   0.000000000000000D+00,   1.355853610355408D-04, &
   0.000000000000000D+00,   4.619387744137319D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.793196738330082D-06,   0.000000000000000D+00, &
  -4.172056337503888D-06,   0.000000000000000D+00,  -1.997572496669678D-05, &
   0.000000000000000D+00,  -1.221984580501521D-04,   0.000000000000000D+00, &
  -1.028620299321905D-04,   0.000000000000000D+00,   5.103710412375815D-04, &
   0.000000000000000D+00,  -2.948992971204410D-07,   0.000000000000000D+00, &
  -1.156257235947030D-06,   0.000000000000000D+00,   1.366235924481004D-06, &
   0.000000000000000D+00,   1.961610303276636D-05,   0.000000000000000D+00, &
  -1.664077421815216D-04,   0.000000000000000D+00,  -3.549269918010076D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.532488995319171D-07, &
   0.000000000000000D+00,   5.745661216318399D-07,   0.000000000000000D+00, &
   2.432511908578300D-06,   0.000000000000000D+00,   1.970136616865215D-05, &
   0.000000000000000D+00,   1.632793757795850D-04,   0.000000000000000D+00, &
   5.856310473658402D-05,   0.000000000000000D+00,  -1.963695414911587D-04, &
   0.000000000000000D+00,  -1.541250148879276D-07,   0.000000000000000D+00, &
   2.176235530782584D-07,   0.000000000000000D+00,  -6.915492643794377D-07, &
   0.000000000000000D+00,  -1.076643368055746D-05,   0.000000000000000D+00, &
  -4.054834773685129D-05,   0.000000000000000D+00,   1.539662266089800D-04, &
   0.000000000000000D+00,   1.437852606202818D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.576307153077543D-08,   0.000000000000000D+00, &
  -9.217261040598894D-09,   0.000000000000000D+00,   4.278174142633074D-07, &
   0.000000000000000D+00,   2.257630499169065D-06,   0.000000000000000D+00, &
  -1.804848870318966D-05,   0.000000000000000D+00,  -1.291823804800371D-04, &
   0.000000000000000D+00,  -1.359363261259290D-06,   0.000000000000000D+00, &
  -1.878438176187453D-07,   0.000000000000000D+00,  -2.191149991795691D-09, &
   0.000000000000000D+00,  -1.030593650802877D-07,   0.000000000000000D+00, &
  -6.270922487464291D-08,   0.000000000000000D+00,   1.703768536307951D-07, &
   0.000000000000000D+00,   1.458312918828840D-05,   0.000000000000000D+00, &
   1.590335713566152D-05,   0.000000000000000D+00,  -9.663354155777738D-05, &
   0.000000000000000D+00,  -2.328508066919901D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.588398235567626D-11,   0.000000000000000D+00, &
   4.667472941464160D-08,   0.000000000000000D+00,  -1.625450736784428D-07, &
   0.000000000000000D+00,  -4.784348997478632D-07,   0.000000000000000D+00, &
  -3.851088959231865D-06,   0.000000000000000D+00,   2.615487253521705D-05, &
   0.000000000000000D+00,   5.131266344095265D-05,   0.000000000000000D+00, &
  -6.601623108193964D-06,   0.000000000000000D+00,   3.743146980678167D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.342158892148238D-09,   0.000000000000000D+00,   2.363041532780011D-08, &
   0.000000000000000D+00,   3.454574347483732D-08,   0.000000000000000D+00, &
  -1.747128032576280D-08,   0.000000000000000D+00,  -1.341094035385173D-05, &
   0.000000000000000D+00,   2.124187972113179D-05,   0.000000000000000D+00, &
   3.333571943556245D-05,   0.000000000000000D+00,   3.215808134812599D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.395902555751229D-10,   0.000000000000000D+00, &
   4.821773391518108D-08,   0.000000000000000D+00,  -5.794041246490118D-09, &
   0.000000000000000D+00,   1.311222112806028D-07,   0.000000000000000D+00, &
   1.159495793382125D-06,   0.000000000000000D+00,  -2.573302629502150D-05, &
   0.000000000000000D+00,   1.059174169504446D-06,   0.000000000000000D+00, &
  -6.089093667652741D-06,   0.000000000000000D+00,  -2.130062861756963D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.530433256443777D-11,   0.000000000000000D+00,  -2.030494870522078D-08, &
   0.000000000000000D+00,   9.926224986618239D-08,   0.000000000000000D+00, &
   2.588814768376132D-07,   0.000000000000000D+00,   8.996343594848017D-07, &
   0.000000000000000D+00,   2.958272690541997D-06,   0.000000000000000D+00, &
  -2.583565225278134D-05,   0.000000000000000D+00,  -5.264011783803267D-06, &
   0.000000000000000D+00,  -6.793103851962748D-06,   0.000000000000000D+00 /)
lem(:,135) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.556342527498331D-04,   1.617608550123334D-03,   2.327225983449661D-03, &
  -1.604669406710546D-03,  -5.781952401103021D-03,   1.266613772713980D-03, &
   4.946524748944270D-03,  -1.868375311383782D-03,  -1.207316868892324D-03, &
   9.121581701220038D-04,  -3.784545323524682D-04,   2.900977199565568D-04, &
  -1.471166406319160D-04,  -1.437112287529248D-05,   1.487998930956602D-04, &
  -2.165928773503563D-04,   1.406844681968756D-04,   8.708266579776190D-06, &
  -8.868352826178491D-05,   6.135929546471565D-05,   3.798568207687645D-06, &
  -3.361108350456338D-05,   4.556342527498588D-04,   1.617608550123337D-03, &
   2.327225983449659D-03,  -1.604669406710542D-03,  -5.781952401103031D-03, &
   1.266613772713980D-03,   4.946524748944262D-03,  -1.868375311383778D-03, &
  -1.207316868892327D-03,   9.121581701220058D-04,  -3.784545323524553D-04, &
   2.900977199565574D-04,  -1.471166406319311D-04,  -1.437112287529861D-05, &
   1.487998930956686D-04,  -2.165928773503510D-04,   1.406844681968747D-04, &
   8.708266579773379D-06,  -8.868352826178134D-05,   6.135929546471368D-05, &
   3.798568207681859D-06,  -3.361108350455752D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.219946131102079D-04,   0.000000000000000D+00, &
   3.766175848816563D-03,   0.000000000000000D+00,  -1.516759243651619D-02, &
   0.000000000000000D+00,   3.015142101915397D-02,   0.000000000000000D+00, &
  -3.174863383762668D-02,   0.000000000000000D+00,   1.792760025237640D-02, &
   0.000000000000000D+00,  -5.148918362862719D-03,   0.000000000000000D+00, &
   8.103566819901950D-04,   0.000000000000000D+00,  -1.983310805544601D-04, &
   0.000000000000000D+00,   6.322634909188315D-05,   0.000000000000000D+00, &
  -2.325148809523984D-06,  -6.975446428571483D-06,   0.000000000000000D+00, &
  -1.162574404761895D-05,   0.000000000000000D+00,   3.952752976190474D-05, &
   8.525545634920552D-06,   0.000000000000000D+00,   2.247643849206352D-05, &
   0.000000000000000D+00,   2.092633928571443D-05,   0.000000000000000D+00, &
  -2.046130952380960D-05,   0.000000000000000D+00,  -2.156898458167991D-04, &
  -1.111214301157224D-05,   0.000000000000000D+00,  -2.433335927917974D-05, &
   0.000000000000000D+00,  -7.921115370957040D-06,   0.000000000000000D+00, &
  -1.597702276760916D-05,   0.000000000000000D+00,   1.895944263639695D-05, &
   0.000000000000000D+00,   1.024226181027955D-04,   0.000000000000000D+00, &
   7.049287484477597D-04,   1.082051167258842D-05,   0.000000000000000D+00, &
   1.158931621214145D-05,   0.000000000000000D+00,  -5.785369273098583D-06, &
   0.000000000000000D+00,  -1.588151601014252D-04,   0.000000000000000D+00, &
   4.656244291194726D-06,   0.000000000000000D+00,  -1.424084401136961D-05, &
   0.000000000000000D+00,  -5.277043537501003D-05,   0.000000000000000D+00, &
  -2.088502572758082D-04,   0.000000000000000D+00,  -1.429162029944675D-03, &
  -5.270942401559983D-06,   0.000000000000000D+00,   9.277775367067000D-07, &
   0.000000000000000D+00,   1.353712780233446D-05,   0.000000000000000D+00, &
   1.275351424224953D-04,   0.000000000000000D+00,   4.839992905628560D-04, &
   0.000000000000000D+00,   5.030884636460577D-07,   0.000000000000000D+00, &
   4.864112984790845D-06,   0.000000000000000D+00,   1.689569715498388D-05, &
   0.000000000000000D+00,   3.411266800879887D-05,   0.000000000000000D+00, &
   2.707870433208345D-04,   0.000000000000000D+00,   1.868483143387340D-03, &
   8.383973106789340D-07,   0.000000000000000D+00,  -2.892764684865401D-06, &
   0.000000000000000D+00,  -7.165085488414918D-06,   0.000000000000000D+00, &
  -4.589912767437334D-05,   0.000000000000000D+00,  -2.906113580058820D-04, &
   0.000000000000000D+00,  -6.888253605772994D-04,   0.000000000000000D+00, &
  -3.879386343276659D-07,   0.000000000000000D+00,   1.393955592662113D-07, &
   0.000000000000000D+00,  -3.293939115057444D-06,   0.000000000000000D+00, &
   5.555474358055262D-06,   0.000000000000000D+00,   4.279332052106455D-05, &
   0.000000000000000D+00,  -2.961215179321848D-04,   0.000000000000000D+00, &
  -1.615521262838142D-03,   7.518708925628384D-08,   0.000000000000000D+00, &
   6.445847247207605D-07,   0.000000000000000D+00,   1.468884742573550D-06, &
   0.000000000000000D+00,   4.115078843951765D-06,   0.000000000000000D+00, &
   4.902111769061728D-05,   0.000000000000000D+00,   3.828119375292877D-04, &
   0.000000000000000D+00,   5.399372683476860D-04,   0.000000000000000D+00, &
  -7.049426675387876D-09,   0.000000000000000D+00,  -3.349708736348047D-07, &
   0.000000000000000D+00,   1.117143760986499D-06,   0.000000000000000D+00, &
  -1.935863218990548D-06,   0.000000000000000D+00,  -2.516624274203225D-05, &
   0.000000000000000D+00,  -9.227197118358979D-05,   0.000000000000000D+00, &
   3.152711735637539D-04,   0.000000000000000D+00,   9.357902668049758D-04, &
   2.623856077945198D-10,   0.000000000000000D+00,   6.326734006484835D-08, &
   0.000000000000000D+00,  -4.702812642185089D-07,   0.000000000000000D+00, &
   2.389274746397434D-06,   0.000000000000000D+00,   3.861592932462988D-06, &
   0.000000000000000D+00,  -4.414750080745376D-05,   0.000000000000000D+00, &
  -3.319393842351644D-04,   0.000000000000000D+00,  -2.103585322601944D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.613785895243098D-10,   0.000000000000000D+00,  -3.249894737223463D-07, &
   0.000000000000000D+00,  -8.564397893859244D-07,   0.000000000000000D+00, &
   1.643245902081265D-06,   0.000000000000000D+00,   3.242392302065262D-05, &
   0.000000000000000D+00,   4.564921275036527D-05,   0.000000000000000D+00, &
  -2.779494461735579D-04,   0.000000000000000D+00,  -3.714167097688231D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.115907675279073D-09, &
   0.000000000000000D+00,   2.376974360881995D-07,   0.000000000000000D+00, &
  -1.036054626223595D-06,   0.000000000000000D+00,  -1.397829164400967D-06, &
   0.000000000000000D+00,  -7.900418679193386D-06,   0.000000000000000D+00, &
   5.992797053510557D-05,   0.000000000000000D+00,   1.707660223332044D-04, &
   0.000000000000000D+00,   1.252019608432203D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.021410032510299D-10, &
   0.000000000000000D+00,  -2.644250990202216D-08,   0.000000000000000D+00, &
   6.873918627410565D-07,   0.000000000000000D+00,  -1.079822134194396D-06, &
   0.000000000000000D+00,   3.992245413785356D-07,   0.000000000000000D+00, &
  -3.004665106305447D-05,   0.000000000000000D+00,   4.489962558753013D-05, &
   0.000000000000000D+00,   1.650302938793897D-04,   0.000000000000000D+00, &
   1.116360909554543D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.390302955118615D-11,   0.000000000000000D+00,  -2.021705010106777D-08, &
   0.000000000000000D+00,   1.567501362212697D-07,   0.000000000000000D+00, &
   1.108508143079305D-06,   0.000000000000000D+00,  -6.642303920364193D-07, &
   0.000000000000000D+00,   3.293559447706148D-06,   0.000000000000000D+00, &
  -6.387773181459507D-05,   0.000000000000000D+00,  -1.866258663531472D-05, &
   0.000000000000000D+00,   3.963639284228228D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.792547312964153D-09,   0.000000000000000D+00, &
  -2.427385990480124D-07,   0.000000000000000D+00,   5.297198706320437D-07, &
   0.000000000000000D+00,   1.360911026960773D-06,   0.000000000000000D+00, &
   1.147990403813613D-06,   0.000000000000000D+00,   8.750023798300623D-06, &
   0.000000000000000D+00,  -7.762755356542751D-05,   0.000000000000000D+00, &
  -5.658128997543573D-05,   0.000000000000000D+00,  -3.315102847995812D-05 /)
lem(:,136) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.850171266278082D-03,  -1.004198690394189D-02, &
  -1.301316996957973D-02,   2.146860031080334D-02,   1.491581593520511D-02, &
  -1.471083885708215D-02,  -1.067526743348504D-02,  -1.434145469363694D-05, &
   2.556415442142907D-03,   3.493699588204002D-03,   2.216583080238109D-03, &
  -4.213591561297840D-05,  -1.407079538941147D-03,  -1.127314149084266D-03, &
  -5.259427896097610D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.850171266278111D-03, &
  -1.004198690394194D-02,  -1.301316996957969D-02,   2.146860031080338D-02, &
   1.491581593520506D-02,  -1.471083885708216D-02,  -1.067526743348497D-02, &
  -1.434145469365158D-05,   2.556415442142882D-03,   3.493699588203981D-03, &
   2.216583080238104D-03,  -4.213591561296909D-05,  -1.407079538941112D-03, &
  -1.127314149084275D-03,  -5.259427896099713D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.923401361952980D-04,   0.000000000000000D+00,   3.389843879815845D-06, &
   0.000000000000000D+00,   1.694921939910239D-07,   0.000000000000000D+00, &
   1.412434949915106D-08,  -1.882583697094399D-03,   0.000000000000000D+00, &
  -3.716501562779689D-05,   0.000000000000000D+00,  -2.696466722580728D-06, &
   0.000000000000000D+00,  -2.657945769399394D-07,   0.000000000000000D+00, &
   4.327122153209114D-03,   0.000000000000000D+00,   2.628891725674624D-05, &
   0.000000000000000D+00,   1.049824377324712D-05,   0.000000000000000D+00, &
   1.777767669879400D-06,   0.000000000000000D+00,   3.551310875225679D-07, &
  -2.765588310076128D-03,   0.000000000000000D+00,   5.137647384248548D-04, &
   0.000000000000000D+00,   1.505583750838867D-05,   0.000000000000000D+00, &
  -2.520040757017024D-06,   0.000000000000000D+00,  -1.067897124524741D-06, &
   0.000000000000000D+00,  -5.287216335809688D-03,   0.000000000000000D+00, &
  -1.619263336385427D-03,   0.000000000000000D+00,  -1.177734486389766D-04, &
   0.000000000000000D+00,  -1.384340334737457D-05,   0.000000000000000D+00, &
  -2.187733334253281D-07,   0.000000000000000D+00,   8.383973106789340D-07, &
   1.094330027968472D-02,   0.000000000000000D+00,   1.291426562997435D-03, &
   0.000000000000000D+00,   2.376616120054044D-05,   0.000000000000000D+00, &
   3.327070134513919D-05,   0.000000000000000D+00,   8.799598141198272D-06, &
   0.000000000000000D+00,   9.760401665748062D-07,   0.000000000000000D+00, &
  -5.720027761112477D-03,   0.000000000000000D+00,   1.793960813587999D-03, &
   0.000000000000000D+00,   5.608037706500862D-04,   0.000000000000000D+00, &
   4.466287327952552D-05,   0.000000000000000D+00,  -6.896161986548859D-06, &
   0.000000000000000D+00,  -4.818880198288938D-06,   0.000000000000000D+00, &
  -1.234484128984861D-06,  -3.281375185423159D-03,   0.000000000000000D+00, &
  -4.066117427796495D-03,   0.000000000000000D+00,  -7.330192691033936D-04, &
   0.000000000000000D+00,  -1.427330426917326D-04,   0.000000000000000D+00, &
  -3.263502547638064D-05,   0.000000000000000D+00,  -1.325031995633394D-06, &
   0.000000000000000D+00,   2.340899732099606D-06,   0.000000000000000D+00, &
   5.574088702363269D-03,   0.000000000000000D+00,   1.943312663602548D-03, &
   0.000000000000000D+00,  -4.963302949880305D-04,   0.000000000000000D+00, &
  -7.017950960470541D-05,   0.000000000000000D+00,   3.099949764968392D-05, &
   0.000000000000000D+00,   1.782397318981120D-05,   0.000000000000000D+00, &
   3.085847232563002D-06,   0.000000000000000D+00,  -2.165064330400595D-06, &
  -2.655285260595511D-03,   0.000000000000000D+00,   1.528633027653042D-03, &
   0.000000000000000D+00,   1.530771306131741D-03,   0.000000000000000D+00, &
   3.267595053219865D-04,   0.000000000000000D+00,   6.994323507771175D-05, &
   0.000000000000000D+00,   2.783506729007653D-07,   0.000000000000000D+00, &
  -8.159534448925587D-06,   0.000000000000000D+00,  -2.738475799682732D-06, &
   0.000000000000000D+00,   4.552171617026273D-04,   0.000000000000000D+00, &
  -2.088458760948072D-03,   0.000000000000000D+00,  -5.394871433627489D-04, &
   0.000000000000000D+00,   3.903431014888694D-05,   0.000000000000000D+00, &
  -5.425479219807692D-05,   0.000000000000000D+00,  -3.578807803438344D-05, &
   0.000000000000000D+00,  -6.986789569097378D-06,   0.000000000000000D+00, &
   3.148636966590079D-06,   0.000000000000000D+00,   2.667373811368493D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.298729841893741D-04, &
   0.000000000000000D+00,  -8.665725691264919D-04,   0.000000000000000D+00, &
  -4.511968830521604D-04,   0.000000000000000D+00,  -1.021041191578757D-04, &
   0.000000000000000D+00,  -5.198560764020832D-06,   0.000000000000000D+00, &
   1.434068144405011D-05,   0.000000000000000D+00,   6.294871131542212D-06, &
   0.000000000000000D+00,  -3.294853144634423D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.711098879643724D-05, &
   0.000000000000000D+00,   7.272047200762057D-04,   0.000000000000000D+00, &
   4.304486273903045D-05,   0.000000000000000D+00,   3.111860017823468D-05, &
   0.000000000000000D+00,   4.051603870480506D-05,   0.000000000000000D+00, &
   1.302009508975328D-05,   0.000000000000000D+00,  -3.854813862573230D-06, &
   0.000000000000000D+00,  -4.147136647323197D-06,   0.000000000000000D+00, &
   1.413860018291211D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.323468624576151D-05,   0.000000000000000D+00,   2.774812012383504D-05, &
   0.000000000000000D+00,   3.873249594157237D-04,   0.000000000000000D+00, &
   1.191433287144560D-04,   0.000000000000000D+00,   2.364090823537926D-05, &
   0.000000000000000D+00,  -1.029406809095440D-05,   0.000000000000000D+00, &
  -9.498536891936983D-06,   0.000000000000000D+00,  -4.644317965280365D-07, &
   0.000000000000000D+00,   2.128824646219715D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.482701651245815D-06, &
   0.000000000000000D+00,  -1.819055558908901D-04,   0.000000000000000D+00, &
  -9.415110646724764D-05,   0.000000000000000D+00,   3.373904569911279D-05, &
   0.000000000000000D+00,  -1.878818167575000D-05,   0.000000000000000D+00, &
  -1.714447384238347D-05,   0.000000000000000D+00,  -6.173233574717278D-07, &
   0.000000000000000D+00,   5.088959815145169D-06,   0.000000000000000D+00, &
   9.025379484950011D-07,   0.000000000000000D+00,  -2.646067785838592D-06 /)
lem(:,137) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.329953652715237D-04, &
   3.265390272241270D-03,  -5.395627755885543D-03,  -2.425479045388229D-03, &
   1.264182413254625D-02,  -3.345435113566901D-03,  -1.123450939788472D-02, &
   4.097638603148541D-03,   5.428005281483746D-03,  -3.531964639977373D-04, &
  -1.729162155168717D-03,  -1.315255038540758D-03,   4.447012357096642D-05, &
   8.189524519070851D-04,   4.867390265356838D-04,  -1.981174103570703D-04, &
  -3.813461842779638D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.329953652715065D-04,  -3.265390272241243D-03,   5.395627755885536D-03, &
   2.425479045388221D-03,  -1.264182413254625D-02,   3.345435113566891D-03, &
   1.123450939788473D-02,  -4.097638603148540D-03,  -5.428005281483759D-03, &
   3.531964639977313D-04,   1.729162155168727D-03,   1.315255038540746D-03, &
  -4.447012357096799D-05,  -8.189524519070593D-04,  -4.867390265357009D-04, &
   1.981174103570525D-04,   3.813461842779894D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.442010754076853D-07,   0.000000000000000D+00, &
   6.456399612286325D-06,   0.000000000000000D+00,   1.438350650810820D-05, &
   0.000000000000000D+00,  -7.569334457161581D-06,   0.000000000000000D+00, &
  -2.369238476484505D-05,   0.000000000000000D+00,   1.062539900159998D-05, &
   0.000000000000000D+00,   1.355377430553441D-05,   0.000000000000000D+00, &
  -9.313443181082666D-06,   0.000000000000000D+00,   1.607509395065169D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.620081230872849D-06, &
   0.000000000000000D+00,   4.500225641309620D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.010373527401386D-05,   0.000000000000000D+00, &
  -1.055507468599067D-06,   0.000000000000000D+00,  -3.221752447774844D-08, &
   0.000000000000000D+00,   9.987768770915890D-05,   0.000000000000000D+00, &
   6.164419755943461D-06,   0.000000000000000D+00,   7.025599299123007D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.580584489320532D-04, &
   0.000000000000000D+00,  -7.874416562377038D-06,   0.000000000000000D+00, &
  -2.397637510034901D-06,   0.000000000000000D+00,  -6.407318545559256D-07, &
   0.000000000000000D+00,   3.379292220320627D-04,   0.000000000000000D+00, &
  -4.578207614570896D-05,   0.000000000000000D+00,  -1.182376478032571D-06, &
   0.000000000000000D+00,   1.055730923478827D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.204770237614524D-05,   0.000000000000000D+00, &
   2.089762737225386D-04,   0.000000000000000D+00,   2.122845372273306D-05, &
   0.000000000000000D+00,   2.197267853816606D-06,   0.000000000000000D+00, &
  -1.081668365286646D-06,   0.000000000000000D+00,  -5.278640236238836D-04, &
   0.000000000000000D+00,  -3.464094887346114D-04,   0.000000000000000D+00, &
  -2.896790620575804D-05,   0.000000000000000D+00,  -9.289440698446018D-06, &
   0.000000000000000D+00,  -1.793196738330082D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.794303381154137D-04,   0.000000000000000D+00, &
   1.312288404859991D-04,   0.000000000000000D+00,  -5.612079848181596D-05, &
   0.000000000000000D+00,   6.045221585671074D-07,   0.000000000000000D+00, &
   3.991262961598745D-06,   0.000000000000000D+00,   2.234397723938240D-06, &
   0.000000000000000D+00,  -5.714153587121135D-04,   0.000000000000000D+00, &
   4.506685853534541D-04,   0.000000000000000D+00,   1.987958233674384D-04, &
   0.000000000000000D+00,   3.299682319112247D-05,   0.000000000000000D+00, &
   4.031542816922392D-06,   0.000000000000000D+00,  -1.643737967469061D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.460510411821128D-05, &
   0.000000000000000D+00,  -8.135985844627100D-04,   0.000000000000000D+00, &
  -1.479226127679767D-04,   0.000000000000000D+00,  -2.966676833884111D-05, &
   0.000000000000000D+00,  -1.449342693393904D-05,   0.000000000000000D+00, &
  -3.829120355159180D-06,   0.000000000000000D+00,   2.251432899056321D-06, &
   0.000000000000000D+00,   4.379475156260942D-04,   0.000000000000000D+00, &
   4.891642540860370D-04,   0.000000000000000D+00,  -1.882643597836635D-04, &
   0.000000000000000D+00,  -5.468067433090104D-05,   0.000000000000000D+00, &
  -1.812285818153167D-06,   0.000000000000000D+00,   5.555321689925915D-06, &
   0.000000000000000D+00,   2.498694802790881D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.706132399256132D-04,   0.000000000000000D+00, &
   1.549164360572818D-04,   0.000000000000000D+00,   4.209282519159283D-04, &
   0.000000000000000D+00,   9.239158225615447D-05,   0.000000000000000D+00, &
   2.811488472489452D-05,   0.000000000000000D+00,   5.997193091161098D-06, &
   0.000000000000000D+00,  -1.946664349402569D-06,   0.000000000000000D+00, &
  -4.068749592877568D-06,   0.000000000000000D+00,   1.623461317553152D-04, &
   0.000000000000000D+00,  -4.498500667629974D-04,   0.000000000000000D+00, &
  -2.054684307792750D-04,   0.000000000000000D+00,   3.118418848785516D-05, &
   0.000000000000000D+00,  -5.498817141160859D-06,   0.000000000000000D+00, &
  -1.030913676424800D-05,   0.000000000000000D+00,  -4.588333918818385D-06, &
   0.000000000000000D+00,  -7.529491256350214D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.832185289843566D-05,   0.000000000000000D+00, &
   3.039650007388793D-04,   0.000000000000000D+00,  -1.742525498007689D-04, &
   0.000000000000000D+00,  -1.376289495134076D-04,   0.000000000000000D+00, &
  -3.351444101242739D-05,   0.000000000000000D+00,  -6.641613684849831D-06, &
   0.000000000000000D+00,   2.631142949946538D-06,   0.000000000000000D+00, &
   3.325083513738286D-06,   0.000000000000000D+00,  -6.539481026785161D-07, &
   0.000000000000000D+00,   3.877806543294190D-06,   0.000000000000000D+00, &
  -7.940676875286580D-05,   0.000000000000000D+00,   2.604048957556262D-04, &
   0.000000000000000D+00,   3.449496598549944D-05,   0.000000000000000D+00, &
   4.971333903035599D-06,   0.000000000000000D+00,   1.017471747436565D-05, &
   0.000000000000000D+00,   5.433926047774003D-06,   0.000000000000000D+00, &
   1.627106140893852D-07,   0.000000000000000D+00,  -9.578190656877378D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.105631744350697D-05,   0.000000000000000D+00, &
  -8.698471584875190D-05,   0.000000000000000D+00,   1.095588377038783D-04, &
   0.000000000000000D+00,   3.412877781885074D-05,   0.000000000000000D+00, &
   7.931721427297478D-06,   0.000000000000000D+00,  -1.349800828656869D-06, &
   0.000000000000000D+00,  -3.068576234213195D-06,   0.000000000000000D+00, &
  -5.054917265531495D-07,   0.000000000000000D+00,   3.111609900245198D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.263550809175753D-05,   0.000000000000000D+00,  -4.150375417016624D-05, &
   0.000000000000000D+00,  -7.390590125185929D-05,   0.000000000000000D+00, &
   5.291159682054547D-06,   0.000000000000000D+00,  -4.131887812385644D-06, &
   0.000000000000000D+00,  -4.500308540576926D-06,   0.000000000000000D+00, &
  -9.700610184167437D-07,   0.000000000000000D+00,   8.781263860408546D-07, &
   0.000000000000000D+00,   1.044966216780128D-06,   0.000000000000000D+00 /)
lem(:,138) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.426032100013551D-04,  -1.088851955762342D-03,  -2.537246936373612D-04, &
   3.243180399633191D-03,   5.422083895255423D-03,  -7.621654520771887D-03, &
  -1.354329975220797D-02,   9.535631289720575D-03,   1.369471798103654D-02, &
  -4.467505530187001D-03,  -6.586337070962002D-03,  -5.209491807830007D-04, &
   1.326691233628078D-03,   9.573058364070767D-04,  -2.727470732871151D-05, &
  -3.032920961131465D-04,  -5.002096865904870D-06,   1.595108467732910D-04, &
   1.980378518237938D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.426032100013628D-04,  -1.088851955762346D-03, &
  -2.537246936373844D-04,   3.243180399633209D-03,   5.422083895255461D-03, &
  -7.621654520771894D-03,  -1.354329975220800D-02,   9.535631289720577D-03, &
   1.369471798103654D-02,  -4.467505530186996D-03,  -6.586337070961975D-03, &
  -5.209491807829831D-04,   1.326691233628061D-03,   9.573058364070630D-04, &
  -2.727470732870178D-05,  -3.032920961131350D-04,  -5.002096865906474D-06, &
   1.595108467732898D-04,   1.980378518238770D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.123866305527966D-06,   0.000000000000000D+00,  -4.234347308307773D-05, &
   0.000000000000000D+00,  -4.007296502986493D-05,   0.000000000000000D+00, &
   4.948174096775458D-05,   0.000000000000000D+00,   5.540149078467009D-05, &
   0.000000000000000D+00,  -3.873857152741592D-05,   0.000000000000000D+00, &
  -2.643026229576385D-05,   0.000000000000000D+00,   1.880666120041382D-05, &
   0.000000000000000D+00,   7.969312796727259D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.505234703376748D-06,   0.000000000000000D+00,  -2.319661762357329D-08, &
  -2.860916173609655D-05,   0.000000000000000D+00,  -5.644510288475618D-07, &
   0.000000000000000D+00,   2.270691125180314D-05,   0.000000000000000D+00, &
   3.525885878827213D-06,   0.000000000000000D+00,   7.571118252218851D-08, &
   4.058634863588417D-04,   0.000000000000000D+00,  -5.036979826896768D-07, &
   0.000000000000000D+00,  -1.010019392372223D-06,   0.000000000000000D+00, &
  -1.309735940911064D-03,   0.000000000000000D+00,   3.436685358882063D-06, &
   0.000000000000000D+00,  -7.744392503445229D-07,   0.000000000000000D+00, &
   1.113722467538646D-06,   1.056158137191922D-03,   0.000000000000000D+00, &
  -1.761046868059920D-04,   0.000000000000000D+00,   3.074002080069072D-06, &
   0.000000000000000D+00,   9.167986856769574D-08,   0.000000000000000D+00, &
   1.499668613344747D-03,   0.000000000000000D+00,   5.660965319061957D-04, &
   0.000000000000000D+00,   1.067193825303938D-06,   0.000000000000000D+00, &
  -1.068559568189383D-06,   0.000000000000000D+00,   8.978409779430486D-07, &
  -3.416188102888865D-03,   0.000000000000000D+00,  -4.617175797684616D-04, &
   0.000000000000000D+00,   7.261886047725312D-05,   0.000000000000000D+00, &
   9.820391727445471D-07,   0.000000000000000D+00,   4.753113544989164D-07, &
   0.000000000000000D+00,   1.645931368303152D-03,   0.000000000000000D+00, &
  -6.752872693779955D-04,   0.000000000000000D+00,  -2.920587698148610D-04, &
   0.000000000000000D+00,  -1.443265330985753D-05,   0.000000000000000D+00, &
  -1.507048393353321D-06,   0.000000000000000D+00,  -2.892764684865401D-06, &
   1.291426562997435D-03,   0.000000000000000D+00,   1.535101390734719D-03, &
   0.000000000000000D+00,   2.742364212615925D-04,   0.000000000000000D+00, &
   3.329295945226705D-06,   0.000000000000000D+00,   4.181829570246128D-06, &
   0.000000000000000D+00,   1.584379109586487D-06,   0.000000000000000D+00, &
  -1.775198011261867D-03,   0.000000000000000D+00,  -7.393583452477930D-04, &
   0.000000000000000D+00,   3.419858062123808D-04,   0.000000000000000D+00, &
   1.149341691386843D-04,   0.000000000000000D+00,   7.758461733163296D-06, &
   0.000000000000000D+00,  -3.548592652417728D-07,   0.000000000000000D+00, &
  -1.771815345194584D-06,   6.217960237042018D-04,   0.000000000000000D+00, &
  -5.643835393001061D-04,   0.000000000000000D+00,  -8.276030235856244D-04, &
   0.000000000000000D+00,  -1.638232714809364D-04,   0.000000000000000D+00, &
  -2.898404862351588D-05,   0.000000000000000D+00,  -6.246714947469236D-06, &
   0.000000000000000D+00,  -1.265569040638521D-07,   0.000000000000000D+00, &
   3.261440844743994D-05,   0.000000000000000D+00,   7.634439491787550D-04, &
   0.000000000000000D+00,   3.739386631665689D-04,   0.000000000000000D+00, &
  -1.130084385252220D-04,   0.000000000000000D+00,  -1.630391291650346D-05, &
   0.000000000000000D+00,   6.635893527969852D-06,   0.000000000000000D+00, &
   3.698360536483293D-06,   0.000000000000000D+00,   4.082394122738607D-06, &
  -5.452696483030152D-05,   0.000000000000000D+00,  -2.652569951504122D-04, &
   0.000000000000000D+00,   3.330553561877189D-04,   0.000000000000000D+00, &
   3.587411580736529D-04,   0.000000000000000D+00,   7.898784475170752D-05, &
   0.000000000000000D+00,   1.611803705885068D-05,   0.000000000000000D+00, &
  -2.055722055727776D-07,   0.000000000000000D+00,  -2.334667401906459D-06, &
   0.000000000000000D+00,   5.602610072870282D-06,   0.000000000000000D+00, &
   6.494911551271577D-06,   0.000000000000000D+00,  -3.902228128640278D-04, &
   0.000000000000000D+00,  -1.240136533556765D-04,   0.000000000000000D+00, &
   1.038692012267267D-05,   0.000000000000000D+00,  -1.364401620739649D-05, &
   0.000000000000000D+00,  -8.551898450182302D-06,   0.000000000000000D+00, &
  -2.346524894800168D-06,   0.000000000000000D+00,   6.022597782226881D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.304124887046316D-06, &
   0.000000000000000D+00,   1.081223155499791D-04,   0.000000000000000D+00, &
  -1.972038333752597D-04,   0.000000000000000D+00,  -1.155234426687471D-04, &
   0.000000000000000D+00,  -2.628111377632024D-05,   0.000000000000000D+00, &
  -9.595585048613920D-07,   0.000000000000000D+00,   4.065607926138019D-06, &
   0.000000000000000D+00,   1.009640133325007D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.628404676611331D-06, &
   0.000000000000000D+00,  -1.133203679694523D-06,   0.000000000000000D+00, &
   1.530035523553764D-04,   0.000000000000000D+00,   1.169142254492289D-05, &
   0.000000000000000D+00,   8.375026813910210D-06,   0.000000000000000D+00, &
   1.080936328802085D-05,   0.000000000000000D+00,   3.277458658322216D-06, &
   0.000000000000000D+00,  -7.208674248084534D-07,   0.000000000000000D+00, &
  -2.925278727891019D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.417251118270301D-06,   0.000000000000000D+00,   1.048964154376533D-05, &
   0.000000000000000D+00,   2.398487498643718D-06,   0.000000000000000D+00, &
   9.504031585117140D-05,   0.000000000000000D+00,   3.185030433950663D-05, &
   0.000000000000000D+00,   6.140135095960162D-06,   0.000000000000000D+00, &
  -3.142007926016599D-06,   0.000000000000000D+00,  -2.214854063933862D-06, &
   0.000000000000000D+00,   4.825802636075879D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.737566232055574D-07, &
   0.000000000000000D+00,  -4.460021540421971D-06,   0.000000000000000D+00, &
  -1.907057388144435D-05,   0.000000000000000D+00,  -2.276597316614149D-05, &
   0.000000000000000D+00,   9.117611101902720D-06,   0.000000000000000D+00, &
  -5.278562409326530D-06,   0.000000000000000D+00,  -4.635105502362163D-06, &
   0.000000000000000D+00,  -2.582170067781213D-07,   0.000000000000000D+00, &
   1.868953804576533D-06,   0.000000000000000D+00,   9.103401171478346D-07 /)
lem(:,139) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.259133442163677D-06,   2.046083368542005D-04, &
  -7.361909396928648D-04,  -9.379045961531768D-04,   5.751136863190166D-03, &
  -1.122400605201510D-03,  -1.332548347566591D-02,   6.035287439101603D-03, &
   1.680131830186772D-02,  -9.149285945517121D-03,  -1.210178721665201D-02, &
   6.222676768116928D-03,   5.393723131075625D-03,  -1.594165373072990D-03, &
  -1.890162276363080D-03,  -1.917037175419808D-04,   6.237345359551439D-04, &
   2.461821730523395D-04,  -1.788580361372466D-04,  -1.228857230578551D-04, &
   6.173886488461718D-05,   0.000000000000000D+00,   8.259133442158699D-06, &
  -2.046083368541991D-04,   7.361909396928655D-04,   9.379045961531783D-04, &
  -5.751136863190179D-03,   1.122400605201518D-03,   1.332548347566591D-02, &
  -6.035287439101616D-03,  -1.680131830186774D-02,   9.149285945517139D-03, &
   1.210178721665200D-02,  -6.222676768116964D-03,  -5.393723131075618D-03, &
   1.594165373072991D-03,   1.890162276363072D-03,   1.917037175419789D-04, &
  -6.237345359551415D-04,  -2.461821730523362D-04,   1.788580361372430D-04, &
   1.228857230578417D-04,  -6.173886488460211D-05,   0.000000000000000D+00, &
   4.129566721083062D-06,   0.000000000000000D+00,   6.121407634288473D-05, &
   0.000000000000000D+00,   1.235447178168473D-04,   0.000000000000000D+00, &
   3.355594414931165D-05,   0.000000000000000D+00,  -9.763773889247974D-05, &
   0.000000000000000D+00,  -7.098642833400902D-05,   0.000000000000000D+00, &
   4.023612809115141D-05,   0.000000000000000D+00,   5.132040036080490D-05, &
   0.000000000000000D+00,  -1.659388002359830D-05,   0.000000000000000D+00, &
  -1.306951449198597D-05,   0.000000000000000D+00,   6.997561812716067D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.264709147061684D-07, &
   0.000000000000000D+00,   4.357527577341728D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.687530907345025D-06,   0.000000000000000D+00, &
  -6.457823795285642D-07,   0.000000000000000D+00,  -3.125725814801450D-05, &
   0.000000000000000D+00,  -1.819645782493979D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.622818186285679D-04,   0.000000000000000D+00, &
   9.363619491881158D-06,   0.000000000000000D+00,  -7.601545914238587D-07, &
   0.000000000000000D+00,  -2.801131654274743D-04,   0.000000000000000D+00, &
   1.975400904308852D-05,   0.000000000000000D+00,  -8.653206846100692D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.085225927061972D-04, &
   0.000000000000000D+00,  -1.586619829798107D-04,   0.000000000000000D+00, &
  -8.807158364635679D-06,   0.000000000000000D+00,   5.480247605700996D-07, &
   0.000000000000000D+00,   3.760397332883543D-04,   0.000000000000000D+00, &
   3.103505728354293D-04,   0.000000000000000D+00,   9.560650443147563D-06, &
   0.000000000000000D+00,   3.798994767664929D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.840402758426728D-04,   0.000000000000000D+00, &
  -1.443179246382906D-04,   0.000000000000000D+00,   7.706660646154525D-05, &
   0.000000000000000D+00,   1.977152123537832D-06,   0.000000000000000D+00, &
   2.670112456823269D-07,   0.000000000000000D+00,   4.128723387708451D-04, &
   0.000000000000000D+00,  -4.001037774089332D-04,   0.000000000000000D+00, &
  -2.184264113864330D-04,   0.000000000000000D+00,  -2.316595261052787D-05, &
   0.000000000000000D+00,  -4.172056337503888D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.312288404859991D-04,   0.000000000000000D+00, &
   7.548586156070345D-04,   0.000000000000000D+00,   1.438383597826319D-04, &
   0.000000000000000D+00,  -9.103785268140945D-08,   0.000000000000000D+00, &
   7.245102341967292D-06,   0.000000000000000D+00,   2.192384496162285D-06, &
   0.000000000000000D+00,  -3.805266222285269D-04,   0.000000000000000D+00, &
  -4.388748008755785D-04,   0.000000000000000D+00,   2.502661803177953D-04, &
   0.000000000000000D+00,   9.804147880233662D-05,   0.000000000000000D+00, &
   1.093830001948445D-05,   0.000000000000000D+00,  -1.949887918668196D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.563215409725594D-04, &
   0.000000000000000D+00,  -1.743213725098726D-04,   0.000000000000000D+00, &
  -5.226724612992198D-04,   0.000000000000000D+00,  -1.086098761948820D-04, &
   0.000000000000000D+00,  -2.567991567585529D-05,   0.000000000000000D+00, &
  -7.138865134835072D-06,   0.000000000000000D+00,  -8.365944667990493D-07, &
   0.000000000000000D+00,  -6.618544896497197D-05,   0.000000000000000D+00, &
   4.206992517729646D-04,   0.000000000000000D+00,   2.747574463569521D-04, &
   0.000000000000000D+00,  -8.910337100729531D-05,   0.000000000000000D+00, &
  -1.791493401000041D-05,   0.000000000000000D+00,   5.126558491789412D-06, &
   0.000000000000000D+00,   5.556913143504281D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.838219334165560D-06,   0.000000000000000D+00, &
  -2.506873645212392D-04,   0.000000000000000D+00,   1.726552771828699D-04, &
   0.000000000000000D+00,   2.438699622730050D-04,   0.000000000000000D+00, &
   5.657290722128428D-05,   0.000000000000000D+00,   1.456643704368719D-05, &
   0.000000000000000D+00,   4.838600631681223D-07,   0.000000000000000D+00, &
  -4.002907849821682D-06,   0.000000000000000D+00,   1.079416467784902D-05, &
   0.000000000000000D+00,   4.521754504483098D-05,   0.000000000000000D+00, &
  -2.953152150003184D-04,   0.000000000000000D+00,  -9.843343070200176D-05, &
   0.000000000000000D+00,   1.066529053966638D-05,   0.000000000000000D+00, &
  -8.176396201858152D-06,   0.000000000000000D+00,  -8.056392784023287D-06, &
   0.000000000000000D+00,  -2.128197431963528D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.497487026968925D-06,   0.000000000000000D+00, &
   1.522865356104047D-05,   0.000000000000000D+00,   1.244180772147488D-04, &
   0.000000000000000D+00,  -1.297720878952872D-04,   0.000000000000000D+00, &
  -7.893491167628769D-05,   0.000000000000000D+00,  -1.913834648394661D-05, &
   0.000000000000000D+00,  -1.583013552534162D-06,   0.000000000000000D+00, &
   2.678440485121714D-06,   0.000000000000000D+00,   3.268366020495210D-06, &
   0.000000000000000D+00,   1.800838951992935D-07,   0.000000000000000D+00, &
  -4.916726902533147D-06,   0.000000000000000D+00,   9.313510509613111D-06, &
   0.000000000000000D+00,   1.334906468598345D-04,   0.000000000000000D+00, &
   1.219054792504620D-05,   0.000000000000000D+00,   4.673266208813706D-06, &
   0.000000000000000D+00,   7.398483850310372D-06,   0.000000000000000D+00, &
   3.304440293305896D-06,   0.000000000000000D+00,  -8.639287903579482D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.420629697587887D-06,   0.000000000000000D+00, &
  -1.628966668315446D-05,   0.000000000000000D+00,  -6.506155678297237D-06, &
   0.000000000000000D+00,   6.785185663490725D-05,   0.000000000000000D+00, &
   2.096050348725599D-05,   0.000000000000000D+00,   4.389575952151072D-06, &
   0.000000000000000D+00,  -1.464331179458972D-06,   0.000000000000000D+00, &
  -1.841022404198240D-06,   0.000000000000000D+00,   7.388710903256771D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.384798490247918D-06,   0.000000000000000D+00,  -3.664308476192686D-06, &
   0.000000000000000D+00,  -3.606307347411875D-05,   0.000000000000000D+00, &
  -2.636271651988766D-05,   0.000000000000000D+00,   4.992709015375541D-06, &
   0.000000000000000D+00,  -3.055354216780081D-06,   0.000000000000000D+00, &
  -3.385528643403104D-06,   0.000000000000000D+00,   1.877536816355626D-08, &
   0.000000000000000D+00,   1.051007592540323D-06,   0.000000000000000D+00 /)
lem(:,140) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.137207875242931D-05,   2.655827597495129D-04,   6.908330736622116D-04, &
  -1.331033459534501D-03,  -1.146520593515068D-03,  -3.155349146640957D-03, &
   5.339950008546866D-03,   1.245995775308801D-02,  -1.083127444877372D-02, &
  -1.626435497890227D-02,   1.165307498695189D-02,   9.797564935570394D-03, &
  -7.134284410306724D-03,  -2.644486973782450D-03,   2.720468393159238D-03, &
   4.960719348378185D-04,  -8.888062826897528D-04,  -3.243006928979775D-04, &
   3.250418765761157D-04,   1.880782810963430D-04,  -1.192171153543644D-04, &
  -9.826643511777228D-05,  -7.137207875242740D-05,   2.655827597495097D-04, &
   6.908330736622188D-04,  -1.331033459534500D-03,  -1.146520593515071D-03, &
  -3.155349146640963D-03,   5.339950008546866D-03,   1.245995775308801D-02, &
  -1.083127444877371D-02,  -1.626435497890227D-02,   1.165307498695190D-02, &
   9.797564935570373D-03,  -7.134284410306724D-03,  -2.644486973782445D-03, &
   2.720468393159245D-03,   4.960719348378188D-04,  -8.888062826897664D-04, &
  -3.243006928979785D-04,   3.250418765761363D-04,   1.880782810963399D-04, &
  -1.192171153543769D-04,  -9.826643511776881D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.720338752654750D-04,   0.000000000000000D+00, &
  -1.530747030716882D-04,   0.000000000000000D+00,  -1.394054927364525D-05, &
   0.000000000000000D+00,   3.672120340401209D-05,   0.000000000000000D+00, &
   1.221866060005751D-04,   0.000000000000000D+00,   2.274127865577465D-05, &
   0.000000000000000D+00,  -1.123557984359079D-04,   0.000000000000000D+00, &
   1.391511008827175D-06,   0.000000000000000D+00,   3.504684719941087D-05, &
   0.000000000000000D+00,  -1.053942249823400D-05,   0.000000000000000D+00, &
   8.569861511038325D-06,   4.706622353680874D-06,   0.000000000000000D+00, &
  -6.810870587960345D-05,   0.000000000000000D+00,   3.305303227903689D-06, &
   2.409011696937392D-05,   0.000000000000000D+00,  -1.213884727816445D-06, &
   0.000000000000000D+00,   3.404361376497595D-04,   0.000000000000000D+00, &
   2.803139413046423D-06,   0.000000000000000D+00,  -5.264432992761080D-06, &
  -4.711963303294564D-04,   0.000000000000000D+00,   5.576316528328909D-05, &
   0.000000000000000D+00,   4.147916783365585D-06,   0.000000000000000D+00, &
  -3.264149776129855D-04,   0.000000000000000D+00,  -2.295908622046057D-04, &
   0.000000000000000D+00,   1.280479946224002D-05,   0.000000000000000D+00, &
  -1.078751570384911D-06,   1.031603284766565D-03,   0.000000000000000D+00, &
   2.170597393288872D-04,   0.000000000000000D+00,  -8.662496426874210D-05, &
   0.000000000000000D+00,  -8.824929072543932D-06,   0.000000000000000D+00, &
  -3.682812304430286D-04,   0.000000000000000D+00,   2.805524957311871D-04, &
   0.000000000000000D+00,   2.138040779366158D-04,   0.000000000000000D+00, &
   5.753533632559210D-06,   0.000000000000000D+00,   8.369628064326477D-06, &
  -6.008009877978910D-04,   0.000000000000000D+00,  -6.799905376850459D-04, &
   0.000000000000000D+00,  -1.509462692325230D-04,   0.000000000000000D+00, &
   4.998216387650955D-05,   0.000000000000000D+00,   4.374027847677349D-06, &
   0.000000000000000D+00,   5.054700695657303D-04,   0.000000000000000D+00, &
   3.107534132907667D-04,   0.000000000000000D+00,  -2.815906424231907D-04, &
   0.000000000000000D+00,  -1.543531369240793D-04,   0.000000000000000D+00, &
  -1.896314196111656D-05,   0.000000000000000D+00,  -7.165085488414918D-06, &
   2.376616120054044D-05,   0.000000000000000D+00,   2.742364212615925D-04, &
   0.000000000000000D+00,   6.044525593758336D-04,   0.000000000000000D+00, &
   1.225940035855683D-04,   0.000000000000000D+00,   1.612959127692367D-06, &
   0.000000000000000D+00,   5.685716904755846D-06,   0.000000000000000D+00, &
  -1.305422922571002D-04,   0.000000000000000D+00,  -3.244001365745442D-04, &
   0.000000000000000D+00,  -3.091252981956817D-04,   0.000000000000000D+00, &
   1.821998255254736D-04,   0.000000000000000D+00,   7.091814565215421D-05, &
   0.000000000000000D+00,   8.130047213091287D-06,   0.000000000000000D+00, &
  -4.184713718941859D-06,   1.944798891526731D-05,   0.000000000000000D+00, &
   9.022009602214859D-05,   0.000000000000000D+00,  -1.862384536300092D-04, &
   0.000000000000000D+00,  -4.122995505836101D-04,   0.000000000000000D+00, &
  -8.903295867909654D-05,   0.000000000000000D+00,  -2.008486570530812D-05, &
   0.000000000000000D+00,  -6.542123825412103D-06,   0.000000000000000D+00, &
   1.416411344468129D-05,   0.000000000000000D+00,  -4.112297368041701D-07, &
   0.000000000000000D+00,   2.902763817019073D-04,   0.000000000000000D+00, &
   1.998851705939908D-04,   0.000000000000000D+00,  -6.496643606240449D-05, &
   0.000000000000000D+00,  -1.106130309276021D-05,   0.000000000000000D+00, &
   6.110276050676158D-06,   0.000000000000000D+00,   1.064110289567965D-05, &
  -4.198031093089579D-06,   0.000000000000000D+00,   8.259482023470975D-06, &
   0.000000000000000D+00,  -1.364821502300053D-04,   0.000000000000000D+00, &
   1.470507813089826D-04,   0.000000000000000D+00,   1.949633221475310D-04, &
   0.000000000000000D+00,   4.580970062357324D-05,   0.000000000000000D+00, &
   1.110944446468067D-05,   0.000000000000000D+00,   2.587483332135405D-07, &
   0.000000000000000D+00,   3.075942785105414D-07,   0.000000000000000D+00, &
  -3.551459560194285D-06,   0.000000000000000D+00,   3.789851521003255D-05, &
   0.000000000000000D+00,  -2.057537419804512D-04,   0.000000000000000D+00, &
  -7.180460344408148D-05,   0.000000000000000D+00,   6.308988330746172D-06, &
   0.000000000000000D+00,  -8.845153365873782D-06,   0.000000000000000D+00, &
  -7.487063015896519D-06,   0.000000000000000D+00,  -4.209611311653803D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.430056025065957D-06, &
   0.000000000000000D+00,  -8.247401466742629D-06,   0.000000000000000D+00, &
   8.350034277978467D-05,   0.000000000000000D+00,  -1.052900452272084D-04, &
   0.000000000000000D+00,  -6.455958639164686D-05,   0.000000000000000D+00, &
  -1.517840789503506D-05,   0.000000000000000D+00,  -1.164264103784871D-06, &
   0.000000000000000D+00,   1.979962626142478D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.340289769804167D-06, &
   0.000000000000000D+00,  -4.317143959061764D-06,   0.000000000000000D+00, &
  -8.643341367856004D-06,   0.000000000000000D+00,   9.469929493674874D-05, &
   0.000000000000000D+00,   7.256034637345693D-06,   0.000000000000000D+00, &
   4.850261067695400D-06,   0.000000000000000D+00,   7.013030765891129D-06, &
   0.000000000000000D+00,   2.537919130178118D-06,   0.000000000000000D+00, &
  -3.070402667514558D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.208017279427432D-07,   0.000000000000000D+00,   3.666812678299063D-06, &
   0.000000000000000D+00,  -8.084060452667733D-06,   0.000000000000000D+00, &
  -6.318256940566707D-06,   0.000000000000000D+00,   5.464018909126179D-05, &
   0.000000000000000D+00,   1.779464788804446D-05,   0.000000000000000D+00, &
   3.802826111912216D-06,   0.000000000000000D+00,  -1.314691877672810D-06, &
   0.000000000000000D+00,  -9.938949945094412D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.223420074268873D-08, &
   0.000000000000000D+00,   2.049950208667283D-07,   0.000000000000000D+00, &
   6.633090982621242D-06,   0.000000000000000D+00,  -2.030290672609923D-05, &
   0.000000000000000D+00,  -1.668350523440921D-05,   0.000000000000000D+00, &
   5.095682958064287D-06,   0.000000000000000D+00,  -3.270226895393352D-06, &
   0.000000000000000D+00,  -2.607053412970906D-06,   0.000000000000000D+00, &
   6.775517765606700D-07,   0.000000000000000D+00,   3.477724852601323D-06 /)
lem(:,141) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.642922701491596D-04,  -1.572735723580817D-03, &
  -1.332114201087168D-03,   4.009202067432807D-03,   5.913382884992159D-03, &
  -7.531675941409731D-03,  -1.200266169669604D-02,   1.180750177483020D-02, &
   1.157109859122307D-02,  -1.237246475054822D-02,  -4.580129854900214D-03, &
   7.379570450837002D-03,  -2.697665233550718D-04,  -2.388868386274926D-03, &
   8.152365743546703D-04,   5.687865850477055D-04,  -3.774336741858455D-04, &
  -2.239841720017485D-04,   1.793659818347621D-04,   1.086445240650203D-04, &
  -8.670753245351324D-05,   0.000000000000000D+00,  -2.642922701491687D-04, &
   1.572735723580813D-03,   1.332114201087170D-03,  -4.009202067432814D-03, &
  -5.913382884992163D-03,   7.531675941409747D-03,   1.200266169669603D-02, &
  -1.180750177483024D-02,  -1.157109859122307D-02,   1.237246475054824D-02, &
   4.580129854900201D-03,  -7.379570450837009D-03,   2.697665233550752D-04, &
   2.388868386274933D-03,  -8.152365743546736D-04,  -5.687865850477135D-04, &
   3.774336741858517D-04,   2.239841720017555D-04,  -1.793659818347758D-04, &
  -1.086445240650178D-04,   8.670753245352895D-05,   0.000000000000000D+00, &
  -1.321461350745828D-04,   0.000000000000000D+00,  -2.438229487773365D-04, &
   0.000000000000000D+00,   1.600202533706024D-04,   0.000000000000000D+00, &
   2.433578911658384D-04,   0.000000000000000D+00,  -2.315122229933219D-04, &
   0.000000000000000D+00,  -1.836987081654245D-04,   0.000000000000000D+00, &
   1.813217608163999D-04,   0.000000000000000D+00,   6.204523491455513D-05, &
   0.000000000000000D+00,  -6.824469682452056D-05,   0.000000000000000D+00, &
   5.576483871213825D-06,   0.000000000000000D+00,   3.739282608847600D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.044706927059864D-05, &
   0.000000000000000D+00,  -2.381948976376220D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.153848903074917D-05,   0.000000000000000D+00, &
  -8.215826714278623D-06,   0.000000000000000D+00,   1.207512817419939D-04, &
   0.000000000000000D+00,   1.385353552536213D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.001556922609555D-04,   0.000000000000000D+00, &
   5.246695455673783D-05,   0.000000000000000D+00,   1.310855001837966D-05, &
   0.000000000000000D+00,  -1.220467063783795D-04,   0.000000000000000D+00, &
  -1.721543420459828D-04,   0.000000000000000D+00,  -6.661953719094564D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.840021474763704D-04, &
   0.000000000000000D+00,   1.182811505571514D-04,   0.000000000000000D+00, &
  -6.508182725521271D-05,   0.000000000000000D+00,  -2.092336631442246D-05, &
   0.000000000000000D+00,  -1.379835297590045D-04,   0.000000000000000D+00, &
   2.044670228285724D-04,   0.000000000000000D+00,   1.776748696222745D-04, &
   0.000000000000000D+00,   1.678935318273128D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.227833203873044D-04,   0.000000000000000D+00, &
  -4.315345200623858D-04,   0.000000000000000D+00,  -1.121232274375682D-04, &
   0.000000000000000D+00,   3.585085609237170D-05,   0.000000000000000D+00, &
   8.779144147348663D-06,   0.000000000000000D+00,   1.786150763132858D-04, &
   0.000000000000000D+00,   2.266910888949524D-04,   0.000000000000000D+00, &
  -2.213460670179171D-04,   0.000000000000000D+00,  -1.316539017309434D-04, &
   0.000000000000000D+00,  -1.997572496669678D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.612079848181596D-05,   0.000000000000000D+00, &
   1.438383597826319D-04,   0.000000000000000D+00,   4.521116444481940D-04, &
   0.000000000000000D+00,   1.009407923401885D-04,   0.000000000000000D+00, &
   5.778377507307189D-06,   0.000000000000000D+00,   1.415698337952303D-05, &
   0.000000000000000D+00,  -3.125147532433048D-05,   0.000000000000000D+00, &
  -2.434332279326577D-04,   0.000000000000000D+00,  -2.423656126500469D-04, &
   0.000000000000000D+00,   1.471681246976150D-04,   0.000000000000000D+00, &
   6.003874345141958D-05,   0.000000000000000D+00,   5.005675668220126D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.724540117028070D-05, &
   0.000000000000000D+00,   1.007272391955438D-04,   0.000000000000000D+00, &
  -1.331228542898876D-04,   0.000000000000000D+00,  -3.248222434878284D-04, &
   0.000000000000000D+00,  -7.516414684106749D-05,   0.000000000000000D+00, &
  -2.123922242831889D-05,   0.000000000000000D+00,  -1.649247605911071D-05, &
   0.000000000000000D+00,  -3.159506541616414D-06,   0.000000000000000D+00, &
   8.938521253820582D-06,   0.000000000000000D+00,   2.383564601360798D-04, &
   0.000000000000000D+00,   1.604430329581977D-04,   0.000000000000000D+00, &
  -5.301361444670677D-05,   0.000000000000000D+00,  -7.525210330508672D-06, &
   0.000000000000000D+00,   7.064629331992153D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.463893422127666D-06,   0.000000000000000D+00, &
  -1.364928895622291D-05,   0.000000000000000D+00,  -1.093236536591280D-04, &
   0.000000000000000D+00,   1.169975171533444D-04,   0.000000000000000D+00, &
   1.567994736842224D-04,   0.000000000000000D+00,   3.931043974709304D-05, &
   0.000000000000000D+00,   1.227822649087481D-05,   0.000000000000000D+00, &
   1.752515866393965D-07,   0.000000000000000D+00,   1.560788051842870D-06, &
   0.000000000000000D+00,  -2.984975006588866D-06,   0.000000000000000D+00, &
   1.066640620973215D-05,   0.000000000000000D+00,  -1.705560269291402D-04, &
   0.000000000000000D+00,  -5.833797489541550D-05,   0.000000000000000D+00, &
   4.458250584561136D-06,   0.000000000000000D+00,  -8.543498330741411D-06, &
   0.000000000000000D+00,  -5.374406747432938D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.487944325473709D-07,   0.000000000000000D+00, &
   3.268645211920764D-06,   0.000000000000000D+00,  -6.031496386672049D-07, &
   0.000000000000000D+00,   6.087726793778554D-05,   0.000000000000000D+00, &
  -8.665206406552048D-05,   0.000000000000000D+00,  -5.217238138723947D-05, &
   0.000000000000000D+00,  -1.383792632320035D-05,   0.000000000000000D+00, &
  -1.406061264618724D-06,   0.000000000000000D+00,   8.643971212366635D-06, &
   0.000000000000000D+00,   1.388409052145502D-08,   0.000000000000000D+00, &
   2.344655069519024D-07,   0.000000000000000D+00,   4.551871467890827D-06, &
   0.000000000000000D+00,   3.170509307191117D-06,   0.000000000000000D+00, &
   7.898178639000583D-05,   0.000000000000000D+00,   6.137757375244304D-06, &
   0.000000000000000D+00,   4.371059773374256D-06,   0.000000000000000D+00, &
   5.463854821436069D-06,   0.000000000000000D+00,   1.327483711689442D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.633482152720837D-07,   0.000000000000000D+00, &
   3.732832362176235D-07,   0.000000000000000D+00,  -2.504913487141664D-06, &
   0.000000000000000D+00,  -1.779603948386414D-07,   0.000000000000000D+00, &
   4.534185147470755D-05,   0.000000000000000D+00,   1.474985565684116D-05, &
   0.000000000000000D+00,   3.714492120229656D-06,   0.000000000000000D+00, &
  -2.523030078699364D-06,   0.000000000000000D+00,  -5.161986173825774D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.984397940456065D-07,   0.000000000000000D+00,  -2.003784631307719D-06, &
   0.000000000000000D+00,   6.150061325869984D-08,   0.000000000000000D+00, &
  -1.903565104710569D-05,   0.000000000000000D+00,  -1.401480983130202D-05, &
   0.000000000000000D+00,   3.748073937748421D-06,   0.000000000000000D+00, &
  -2.176624548447287D-06,   0.000000000000000D+00,  -1.673549960569969D-06, &
   0.000000000000000D+00,  -7.250739069554818D-07,   0.000000000000000D+00 /)
lem(:,142) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.401278189732801D-04,   1.279176308174767D-03,   4.414221277244500D-04, &
  -5.428186195056099D-04,  -4.486582054071863D-03,  -3.890226428365809D-03, &
   8.540359489123305D-03,   9.107633603424420D-03,  -1.111323368637301D-02, &
  -6.745704049667405D-03,   9.888250481288831D-03,  -1.261555252738285D-04, &
  -4.866344511617466D-03,   2.568140689403448D-03,   8.336960075404252D-04, &
  -1.322250026845828D-03,   2.248465929617648D-04,   3.875309141616602D-04, &
  -1.823910004365627D-04,  -1.363501838889237D-04,   1.094589768138221D-04, &
   5.550157316806485D-05,   4.401278189732834D-04,   1.279176308174770D-03, &
   4.414221277244495D-04,  -5.428186195055966D-04,  -4.486582054071864D-03, &
  -3.890226428365811D-03,   8.540359489123310D-03,   9.107633603424431D-03, &
  -1.111323368637303D-02,  -6.745704049667414D-03,   9.888250481288850D-03, &
  -1.261555252738283D-04,  -4.866344511617473D-03,   2.568140689403450D-03, &
   8.336960075404276D-04,  -1.322250026845830D-03,   2.248465929617633D-04, &
   3.875309141616558D-04,  -1.823910004365530D-04,  -1.363501838889181D-04, &
   1.094589768138047D-04,   5.550157316806307D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.581059192847883D-04,   0.000000000000000D+00, &
   8.142201420450211D-04,   0.000000000000000D+00,  -1.081152460989937D-03, &
   0.000000000000000D+00,   3.797092188117408D-05,   0.000000000000000D+00, &
   6.164983166388479D-04,   0.000000000000000D+00,  -1.650732438304251D-04, &
   0.000000000000000D+00,  -2.451594435322668D-04,   0.000000000000000D+00, &
   1.065063859872411D-04,   0.000000000000000D+00,   1.964395630209555D-05, &
   0.000000000000000D+00,  -1.527002156038713D-05,   0.000000000000000D+00, &
  -6.726661138425972D-06,  -6.262120232708472D-05,   0.000000000000000D+00, &
  -1.391224301441014D-05,   0.000000000000000D+00,  -3.078966169136445D-05, &
   1.490855206025486D-04,   0.000000000000000D+00,   1.220050804029137D-05, &
   0.000000000000000D+00,   1.301641684768630D-05,   0.000000000000000D+00, &
   7.026355710566226D-05,   0.000000000000000D+00,   3.520774702808368D-05, &
  -2.237082877562755D-04,   0.000000000000000D+00,  -1.047048067171906D-04, &
   0.000000000000000D+00,   3.211627419493129D-05,   0.000000000000000D+00, &
   1.787639150507634D-05,   0.000000000000000D+00,  -7.165726075572628D-05, &
   0.000000000000000D+00,  -1.321253125179253D-04,   0.000000000000000D+00, &
  -2.590170584159511D-05,   2.070901150942204D-04,   0.000000000000000D+00, &
   2.332674097922998D-04,   0.000000000000000D+00,   1.040729849554241D-04, &
   0.000000000000000D+00,  -5.281186786415456D-05,   0.000000000000000D+00, &
  -4.507004439700641D-05,   0.000000000000000D+00,  -8.150452765727383D-05, &
   0.000000000000000D+00,   1.521563814979839D-04,   0.000000000000000D+00, &
   1.521591505496428D-04,   0.000000000000000D+00,   3.886825371013048D-05, &
  -1.144288026775572D-04,   0.000000000000000D+00,  -1.319057334545623D-04, &
   0.000000000000000D+00,  -3.428611415971338D-04,   0.000000000000000D+00, &
  -9.914914271275765D-05,   0.000000000000000D+00,   2.460123542847945D-05, &
   0.000000000000000D+00,   3.878572520045143D-05,   0.000000000000000D+00, &
   1.026789508794151D-04,   0.000000000000000D+00,   1.658586449109706D-04, &
   0.000000000000000D+00,  -1.792195817357684D-04,   0.000000000000000D+00, &
  -1.166212066834641D-04,   0.000000000000000D+00,  -4.589912767437334D-05, &
   3.327070134513919D-05,   0.000000000000000D+00,   3.329295945226705D-06, &
   0.000000000000000D+00,   1.225940035855683D-04,   0.000000000000000D+00, &
   3.726520972010389D-04,   0.000000000000000D+00,   9.221269538355262D-05, &
   0.000000000000000D+00,   1.553618420077930D-05,   0.000000000000000D+00, &
  -1.582276853164967D-05,   0.000000000000000D+00,  -1.510144025235159D-05, &
   0.000000000000000D+00,  -1.704283164992861D-04,   0.000000000000000D+00, &
  -1.915374350166570D-04,   0.000000000000000D+00,   1.218654176663793D-04, &
   0.000000000000000D+00,   5.405069434227583D-05,   0.000000000000000D+00, &
   1.674707890967958D-05,  -2.907956496468297D-06,   0.000000000000000D+00, &
   8.941707451976625D-08,   0.000000000000000D+00,   6.670994478461568D-05, &
   0.000000000000000D+00,  -1.128249204281203D-04,   0.000000000000000D+00, &
  -2.747919166630959D-04,   0.000000000000000D+00,  -7.138852836490810D-05, &
   0.000000000000000D+00,  -2.301775229903315D-05,   0.000000000000000D+00, &
   2.523044264604194D-06,   0.000000000000000D+00,  -2.511222297283428D-06, &
   0.000000000000000D+00,  -4.337834968357165D-06,   0.000000000000000D+00, &
   1.854125571525330D-04,   0.000000000000000D+00,   1.308091110388809D-04, &
   0.000000000000000D+00,  -4.271901324738573D-05,   0.000000000000000D+00, &
  -7.337893629578547D-06,   0.000000000000000D+00,   1.652925180149349D-05, &
  -4.470855803495616D-07,   0.000000000000000D+00,   2.435075816722542D-06, &
   0.000000000000000D+00,  -7.567107847206247D-06,   0.000000000000000D+00, &
  -8.645476052742768D-05,   0.000000000000000D+00,   1.008841124751131D-04, &
   0.000000000000000D+00,   1.356422180908787D-04,   0.000000000000000D+00, &
   3.782468359638852D-05,   0.000000000000000D+00,   8.183160187927377D-06, &
   0.000000000000000D+00,   2.540879311114882D-08,   0.000000000000000D+00, &
   4.023621121322034D-07,   0.000000000000000D+00,   5.183617731247308D-06, &
   0.000000000000000D+00,   1.146834041315696D-05,   0.000000000000000D+00, &
  -1.388955009740331D-04,   0.000000000000000D+00,  -4.818482794943219D-05, &
   0.000000000000000D+00,   2.846423405964968D-06,   0.000000000000000D+00, &
  -8.852266046091312D-06,   0.000000000000000D+00,  -1.945159699794049D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.850480642114199D-07, &
   0.000000000000000D+00,   3.377453351561830D-07,   0.000000000000000D+00, &
   5.339783832154179D-06,   0.000000000000000D+00,   4.935664578438932D-05, &
   0.000000000000000D+00,  -7.542394732804667D-05,   0.000000000000000D+00, &
  -4.683564140979155D-05,   0.000000000000000D+00,  -1.194613355474645D-05, &
   0.000000000000000D+00,  -1.377244507827445D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.001869326545069D-07, &
   0.000000000000000D+00,  -1.774834162934793D-06,   0.000000000000000D+00, &
  -1.273974061940714D-06,   0.000000000000000D+00,   4.509963320888090D-06, &
   0.000000000000000D+00,   6.526974204569154D-05,   0.000000000000000D+00, &
   4.576568785312908D-06,   0.000000000000000D+00,   3.525118946399681D-06, &
   0.000000000000000D+00,   7.998816676818831D-06,   0.000000000000000D+00, &
   5.879344047740801D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.420444947476824D-08,   0.000000000000000D+00,   3.353179038795566D-07, &
   0.000000000000000D+00,  -2.151414840409400D-06,   0.000000000000000D+00, &
  -7.080958064506110D-06,   0.000000000000000D+00,   9.019414063792612D-07, &
   0.000000000000000D+00,   3.999486513400433D-05,   0.000000000000000D+00, &
   1.358860401206082D-05,   0.000000000000000D+00,   2.386182054109009D-06, &
   0.000000000000000D+00,   4.024746691159046D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.082439489598731D-09, &
   0.000000000000000D+00,   2.725725368784435D-07,   0.000000000000000D+00, &
   6.059952546997325D-07,   0.000000000000000D+00,  -1.348542351445838D-07, &
   0.000000000000000D+00,  -1.784216594962090D-05,   0.000000000000000D+00, &
  -1.097487800702617D-05,   0.000000000000000D+00,   4.231725108692807D-06, &
   0.000000000000000D+00,  -2.319143429171435D-06,   0.000000000000000D+00, &
  -3.309378693699374D-06,   0.000000000000000D+00,   1.719138296895295D-06 /)
lem(:,143) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.472534666625372D-03,  -1.591695078177376D-03, &
   3.505427599197362D-04,   4.223369579119179D-03,   3.398700219609800D-03, &
  -6.439575420592440D-03,  -6.036794384607775D-03,   8.143022933431110D-03, &
   2.677981496435529D-03,  -6.360696114839774D-03,   2.101464047546216D-03, &
   1.680347036520952D-03,  -2.384210926400556D-03,   8.623321227572489D-04, &
   6.138749069219938D-04,  -7.077527074233154D-04,   8.727999164986420D-05, &
   2.382523157516150D-04,  -1.113463572915681D-04,  -7.674452036497761D-05, &
   7.339266537717411D-05,   0.000000000000000D+00,   1.472534666625370D-03, &
   1.591695078177380D-03,  -3.505427599197387D-04,  -4.223369579119199D-03, &
  -3.398700219609797D-03,   6.439575420592445D-03,   6.036794384607773D-03, &
  -8.143022933431113D-03,  -2.677981496435532D-03,   6.360696114839781D-03, &
  -2.101464047546214D-03,  -1.680347036520954D-03,   2.384210926400557D-03, &
  -8.623321227572467D-04,  -6.138749069219924D-04,   7.077527074233171D-04, &
  -8.727999164986726D-05,  -2.382523157516181D-04,   1.113463572915713D-04, &
   7.674452036498189D-05,  -7.339266537717272D-05,   0.000000000000000D+00, &
   7.362673333126852D-04,   0.000000000000000D+00,   9.387803177384330D-04, &
   0.000000000000000D+00,  -2.823066978716745D-03,   0.000000000000000D+00, &
   3.018481153419588D-03,   0.000000000000000D+00,  -1.438463139685571D-03, &
   0.000000000000000D+00,  -3.573330529752329D-04,   0.000000000000000D+00, &
   6.516371319387563D-04,   0.000000000000000D+00,  -7.959882725939580D-05, &
   0.000000000000000D+00,  -1.017397766851266D-04,   0.000000000000000D+00, &
   3.582275666278567D-05,   0.000000000000000D+00,  -1.174967771492115D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.820704350116126D-05, &
   0.000000000000000D+00,  -1.490991235573272D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.368012333956610D-05,   0.000000000000000D+00, &
   2.877535046652703D-05,   0.000000000000000D+00,   1.023228577408135D-05, &
   0.000000000000000D+00,   6.053726545212605D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.997270445412862D-05,   0.000000000000000D+00, &
  -7.904220607393977D-05,   0.000000000000000D+00,   4.584872552413197D-05, &
   0.000000000000000D+00,   1.440830728924681D-05,   0.000000000000000D+00, &
  -5.800408087896307D-05,   0.000000000000000D+00,  -1.246132538907266D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.041935242534246D-05, &
   0.000000000000000D+00,   1.643013956806752D-04,   0.000000000000000D+00, &
   9.191571981556753D-05,   0.000000000000000D+00,  -8.684112595936196D-05, &
   0.000000000000000D+00,  -3.503967845568327D-05,   0.000000000000000D+00, &
  -6.256175165601050D-05,   0.000000000000000D+00,   1.293332700576483D-04, &
   0.000000000000000D+00,   1.567470065790626D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.602178324011682D-05,   0.000000000000000D+00, &
  -9.038747891554467D-05,   0.000000000000000D+00,  -2.800867440713502D-04, &
   0.000000000000000D+00,  -9.811623521935298D-05,   0.000000000000000D+00, &
   4.348741617296045D-05,   0.000000000000000D+00,   2.879424907961722D-05, &
   0.000000000000000D+00,   8.692609111995986D-05,   0.000000000000000D+00, &
   1.345404445894484D-04,   0.000000000000000D+00,  -1.576946665792951D-04, &
   0.000000000000000D+00,  -1.221984580501521D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.045221585671074D-07,   0.000000000000000D+00, &
  -9.103785268140945D-08,   0.000000000000000D+00,   1.009407923401885D-04, &
   0.000000000000000D+00,   3.242417759192182D-04,   0.000000000000000D+00, &
   9.540971463134720D-05,   0.000000000000000D+00,   3.552991880743294D-05, &
   0.000000000000000D+00,  -1.033710516527149D-05,   0.000000000000000D+00, &
  -2.541947224598156D-05,   0.000000000000000D+00,  -1.419315535832828D-04, &
   0.000000000000000D+00,  -1.636674450103504D-04,   0.000000000000000D+00, &
   1.102159814161934D-04,   0.000000000000000D+00,   4.904904608788694D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.619707607206799D-06, &
   0.000000000000000D+00,   3.268611557104689D-06,   0.000000000000000D+00, &
   4.753860941041026D-05,   0.000000000000000D+00,  -9.945966655710289D-05, &
   0.000000000000000D+00,  -2.487658159938892D-04,   0.000000000000000D+00, &
  -7.387003104790295D-05,   0.000000000000000D+00,  -6.316777652593627D-05, &
   0.000000000000000D+00,   8.671781816125759D-07,   0.000000000000000D+00, &
   5.346171628261913D-06,   0.000000000000000D+00,   5.429009664625538D-06, &
   0.000000000000000D+00,   1.587723109789904D-04,   0.000000000000000D+00, &
   1.150462125877129D-04,   0.000000000000000D+00,  -3.865010931643696D-05, &
   0.000000000000000D+00,   2.033943703056784D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.885938192799528D-07,   0.000000000000000D+00, &
  -2.782061759905692D-07,   0.000000000000000D+00,  -3.646464585023811D-06, &
   0.000000000000000D+00,  -6.658119396569137D-05,   0.000000000000000D+00, &
   9.273605064097729D-05,   0.000000000000000D+00,   1.255347380262549D-04, &
   0.000000000000000D+00,   4.165561136999599D-05,   0.000000000000000D+00, &
   3.176167531047004D-05,   0.000000000000000D+00,   2.390778693334483D-07, &
   0.000000000000000D+00,  -1.505387763952037D-06,   0.000000000000000D+00, &
   4.307044469776154D-07,   0.000000000000000D+00,   4.396566450625547D-06, &
   0.000000000000000D+00,  -1.223099796146540D-04,   0.000000000000000D+00, &
  -4.187999851068141D-05,   0.000000000000000D+00,   6.127497863439564D-07, &
   0.000000000000000D+00,  -7.266396344600874D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.845255478596222D-08,   0.000000000000000D+00, &
   2.514063402354939D-07,   0.000000000000000D+00,  -1.902699573028097D-06, &
   0.000000000000000D+00,   3.281362303974192D-06,   0.000000000000000D+00, &
   3.782454156390418D-05,   0.000000000000000D+00,  -7.096979305170024D-05, &
   0.000000000000000D+00,  -4.345089029750039D-05,   0.000000000000000D+00, &
  -1.654922238768624D-05,   0.000000000000000D+00,   1.324795151347240D-06, &
   0.000000000000000D+00,   1.285563937211140D-09,   0.000000000000000D+00, &
   2.705187725111171D-07,   0.000000000000000D+00,   4.580370314599603D-07, &
   0.000000000000000D+00,  -1.857176962740605D-06,   0.000000000000000D+00, &
   7.379833397070349D-06,   0.000000000000000D+00,   5.788436421529511D-05, &
   0.000000000000000D+00,   2.934689306473595D-06,   0.000000000000000D+00, &
   4.050832712304216D-06,   0.000000000000000D+00,   1.608829854354249D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.385547799834611D-07,   0.000000000000000D+00, &
   7.765553358529259D-07,   0.000000000000000D+00,   9.788004229152853D-07, &
   0.000000000000000D+00,  -6.257182036659439D-06,   0.000000000000000D+00, &
   4.676926113452024D-06,   0.000000000000000D+00,   3.738595956100997D-05, &
   0.000000000000000D+00,   1.328676636586768D-05,   0.000000000000000D+00, &
   2.976599818105447D-06,   0.000000000000000D+00,  -9.065843158235920D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.766773562181054D-08,   0.000000000000000D+00,  -3.623903988832829D-07, &
   0.000000000000000D+00,   7.582985867820806D-07,   0.000000000000000D+00, &
   1.483757380289089D-06,   0.000000000000000D+00,  -1.734087725101092D-05, &
   0.000000000000000D+00,  -8.456431114092630D-06,   0.000000000000000D+00, &
   3.917343298557625D-06,   0.000000000000000D+00,  -3.726844778658303D-07, &
   0.000000000000000D+00,  -1.216474317641375D-06,   0.000000000000000D+00 /)
lem(:,144) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.093522206599625D-03,   1.148530367260575D-03,   8.710638477459250D-04, &
  -2.177423559155897D-04,  -3.173171958832340D-03,  -3.150677525622295D-03, &
   4.148735794363786D-03,   4.359885740269615D-03,  -4.644196183760307D-03, &
  -8.432577972996552D-04,   2.951812927669599D-03,  -1.976082241372568D-03, &
   1.753637634302632D-04,   1.006610778895708D-03,  -1.080714478830115D-03, &
   3.443780240757819D-04,   3.284709076328584D-04,  -3.795071776936796D-04, &
   6.180603171835182D-05,   1.329059235653188D-04,  -7.744317596779292D-05, &
  -3.606086827044169D-05,   1.093522206599631D-03,   1.148530367260580D-03, &
   8.710638477459305D-04,  -2.177423559155903D-04,  -3.173171958832345D-03, &
  -3.150677525622295D-03,   4.148735794363796D-03,   4.359885740269619D-03, &
  -4.644196183760307D-03,  -8.432577972996540D-04,   2.951812927669599D-03, &
  -1.976082241372570D-03,   1.753637634302609D-04,   1.006610778895704D-03, &
  -1.080714478830113D-03,   3.443780240757821D-04,   3.284709076328541D-04, &
  -3.795071776936738D-04,   6.180603171835617D-05,   1.329059235653127D-04, &
  -7.744317596779329D-05,  -3.606086827043856D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.306074407226530D-04,   0.000000000000000D+00, &
  -3.027920820545880D-03,   0.000000000000000D+00,   4.296156896140436D-03, &
   0.000000000000000D+00,  -4.317060359612220D-03,   0.000000000000000D+00, &
   3.429746011827520D-03,   0.000000000000000D+00,  -1.254186302101828D-03, &
   0.000000000000000D+00,  -2.932691186618105D-04,   0.000000000000000D+00, &
   3.090031358273693D-04,   0.000000000000000D+00,  -5.532211606773223D-05, &
   0.000000000000000D+00,   9.679920748096870D-06,   0.000000000000000D+00, &
  -4.880442610222033D-05,  -3.656601772410588D-05,   0.000000000000000D+00, &
   1.619436241146775D-05,   0.000000000000000D+00,  -4.401310969373979D-06, &
   4.485337181738122D-05,   0.000000000000000D+00,   3.554252402331929D-05, &
   0.000000000000000D+00,   1.259758902947765D-05,   0.000000000000000D+00, &
   5.920512674817978D-06,   0.000000000000000D+00,   1.121915217672766D-04, &
  -4.422768064408939D-05,   0.000000000000000D+00,  -5.505986679191980D-05, &
   0.000000000000000D+00,  -7.580377434527954D-05,   0.000000000000000D+00, &
  -3.132347747306854D-05,   0.000000000000000D+00,   6.195621102678271D-07, &
   0.000000000000000D+00,  -5.242333188754619D-05,   0.000000000000000D+00, &
  -2.549593734695844D-04,   3.814144883455198D-05,   0.000000000000000D+00, &
   5.075309874150351D-05,   0.000000000000000D+00,   1.391816890615674D-04, &
   0.000000000000000D+00,   1.253916578627554D-04,   0.000000000000000D+00, &
   2.349347312393303D-05,   0.000000000000000D+00,  -9.066167757771171D-06, &
   0.000000000000000D+00,  -4.604858324563726D-05,   0.000000000000000D+00, &
   1.224504929255935D-04,   0.000000000000000D+00,   3.354590570300163D-04, &
  -2.401948646711563D-05,   0.000000000000000D+00,  -2.482180131768104D-05, &
   0.000000000000000D+00,  -7.625232494555323D-05,   0.000000000000000D+00, &
  -2.610332438104029D-04,   0.000000000000000D+00,  -1.539547809588272D-04, &
   0.000000000000000D+00,  -6.626518333174415D-06,   0.000000000000000D+00, &
   7.247575562256650D-06,   0.000000000000000D+00,   6.318555253025537D-05, &
   0.000000000000000D+00,   1.183437083977499D-04,   0.000000000000000D+00, &
  -1.562943738642568D-04,   0.000000000000000D+00,  -2.906113580058820D-04, &
   8.799598141198272D-06,   0.000000000000000D+00,   4.181829570246128D-06, &
   0.000000000000000D+00,   1.612959127692367D-06,   0.000000000000000D+00, &
   9.221269538355262D-05,   0.000000000000000D+00,   3.182209290346883D-04, &
   0.000000000000000D+00,   1.439193330836288D-04,   0.000000000000000D+00, &
  -4.411590270409389D-07,   0.000000000000000D+00,  -8.636552923819741D-07, &
   0.000000000000000D+00,  -1.731144347911928D-05,   0.000000000000000D+00, &
  -1.213814532000455D-04,   0.000000000000000D+00,  -1.508997096443502D-04, &
   0.000000000000000D+00,   1.167159037084532D-04,   0.000000000000000D+00, &
   1.483235811412107D-04,  -1.267539343871703D-06,   0.000000000000000D+00, &
   6.987764554929718D-07,   0.000000000000000D+00,   5.690169752225577D-06, &
   0.000000000000000D+00,   3.804465924695407D-05,   0.000000000000000D+00, &
  -9.786381729560714D-05,   0.000000000000000D+00,  -2.524670988256515D-04, &
   0.000000000000000D+00,  -9.154660920704311D-05,   0.000000000000000D+00, &
   4.266604809112198D-07,   0.000000000000000D+00,  -1.114906195645414D-06, &
   0.000000000000000D+00,   1.525998207762396D-06,   0.000000000000000D+00, &
   8.266603341730594D-06,   0.000000000000000D+00,   1.444190262209794D-04, &
   0.000000000000000D+00,   1.075899444781548D-04,   0.000000000000000D+00, &
  -4.916368955588300D-05,   0.000000000000000D+00,  -1.451057270756054D-05, &
  -5.450797001506366D-08,   0.000000000000000D+00,   6.814912043362147D-08, &
   0.000000000000000D+00,  -2.399464139179381D-06,   0.000000000000000D+00, &
  -7.498784082909457D-06,   0.000000000000000D+00,  -5.477862310154228D-05, &
   0.000000000000000D+00,   9.384322032602456D-05,   0.000000000000000D+00, &
   1.302244697995353D-04,   0.000000000000000D+00,   3.220106730517380D-05, &
   0.000000000000000D+00,   2.520713601997269D-09,   0.000000000000000D+00, &
   3.089108344246352D-07,   0.000000000000000D+00,   4.205426914246940D-07, &
   0.000000000000000D+00,   3.085207256339321D-07,   0.000000000000000D+00, &
  -2.573183221190986D-06,   0.000000000000000D+00,  -1.135979174364575D-04, &
   0.000000000000000D+00,  -3.805929385089781D-05,   0.000000000000000D+00, &
   6.121271204251233D-06,   0.000000000000000D+00,  -3.542614448503281D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.432668992090305D-07, &
   0.000000000000000D+00,   6.688596963686461D-07,   0.000000000000000D+00, &
   6.022764563864340D-07,   0.000000000000000D+00,   6.730948398231338D-06, &
   0.000000000000000D+00,   2.929139309095697D-05,   0.000000000000000D+00, &
  -7.169694232669713D-05,   0.000000000000000D+00,  -4.259180063636747D-05, &
   0.000000000000000D+00,  -5.507022150707788D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.792649648646232D-08, &
   0.000000000000000D+00,  -2.789602772858333D-07,   0.000000000000000D+00, &
   6.399524663972951D-07,   0.000000000000000D+00,  -1.045570812210382D-06, &
   0.000000000000000D+00,   1.197162020149809D-05,   0.000000000000000D+00, &
   5.342709978647242D-05,   0.000000000000000D+00,   2.659663060761089D-07, &
   0.000000000000000D+00,   8.697012826207119D-06,   0.000000000000000D+00, &
   2.549188966758989D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.598388716822263D-09,   0.000000000000000D+00,  -7.852930267474912D-08, &
   0.000000000000000D+00,  -4.023105438265990D-07,   0.000000000000000D+00, &
   2.970638127888545D-07,   0.000000000000000D+00,  -7.275755709281037D-06, &
   0.000000000000000D+00,   8.003938897478706D-06,   0.000000000000000D+00, &
   3.738766289003608D-05,   0.000000000000000D+00,   9.146896579687648D-06, &
   0.000000000000000D+00,   4.084222610521048D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.492822508738392D-11, &
   0.000000000000000D+00,   8.028668979115336D-08,   0.000000000000000D+00, &
  -2.776344762325870D-07,   0.000000000000000D+00,  -5.135163771924727D-07, &
   0.000000000000000D+00,   4.739344556224156D-07,   0.000000000000000D+00, &
  -1.793948318034836D-05,   0.000000000000000D+00,  -6.717416062683983D-06, &
   0.000000000000000D+00,   4.347000539569268D-06,   0.000000000000000D+00, &
  -8.388835624126092D-06,   0.000000000000000D+00,  -8.836786411486896D-06 /)
lem(:,145) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.705770327631542D-04,  -1.913980248070749D-03, &
  -9.516375456286083D-04,   2.274983926961075D-03,   2.990147212542550D-03, &
  -1.958585100831794D-03,  -3.179137280652129D-03,   2.229570312632967D-03, &
   5.973876485385198D-04,  -1.157798361909334D-03,   9.407887397753275D-04, &
  -5.329569801330951D-04,  -6.424654503200008D-05,   4.870176600281105D-04, &
  -4.870113880412887D-04,   1.563948322762783D-04,   1.550160852864174D-04, &
  -1.965837535603561D-04,   4.687594010697491D-05,   6.667412185804857D-05, &
  -5.108445152280593D-05,   0.000000000000000D+00,   4.705770327631498D-04, &
   1.913980248070746D-03,   9.516375456286074D-04,  -2.274983926961070D-03, &
  -2.990147212542552D-03,   1.958585100831800D-03,   3.179137280652129D-03, &
  -2.229570312632969D-03,  -5.973876485385106D-04,   1.157798361909334D-03, &
  -9.407887397753344D-04,   5.329569801330972D-04,   6.424654503200116D-05, &
  -4.870176600281106D-04,   4.870113880412934D-04,  -1.563948322762803D-04, &
  -1.550160852864217D-04,   1.965837535603552D-04,  -4.687594010697291D-05, &
  -6.667412185804891D-05,   5.108445152280649D-05,   0.000000000000000D+00, &
   2.352885163815785D-04,   0.000000000000000D+00,  -1.792431592713806D-03, &
   0.000000000000000D+00,   2.028206914768516D-03,   0.000000000000000D+00, &
   5.903145247527504D-04,   0.000000000000000D+00,  -1.042942611309612D-03, &
   0.000000000000000D+00,  -1.608503080212376D-03,   0.000000000000000D+00, &
   2.213337770532503D-03,   0.000000000000000D+00,  -7.292843105680581D-04, &
   0.000000000000000D+00,   3.331466037251363D-06,   0.000000000000000D+00, &
   2.162354984045446D-05,   0.000000000000000D+00,  -1.281026636587886D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.860119047619037D-05, &
   0.000000000000000D+00,  -8.224395680858551D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.291439468684051D-05,   0.000000000000000D+00, &
   9.309324732965395D-05,   0.000000000000000D+00,   1.235589047605301D-05, &
   0.000000000000000D+00,   3.299997404341492D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.609392875356496D-05,   0.000000000000000D+00, &
  -5.263554282154334D-05,   0.000000000000000D+00,  -1.927149313257334D-04, &
   0.000000000000000D+00,  -1.072314193822806D-05,   0.000000000000000D+00, &
   1.483170266138665D-06,   0.000000000000000D+00,  -1.746717374057235D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.761781182484670D-05, &
   0.000000000000000D+00,   3.959205235848073D-05,   0.000000000000000D+00, &
   1.433900283918737D-04,   0.000000000000000D+00,   2.487188490948895D-04, &
   0.000000000000000D+00,   3.056733353543190D-06,   0.000000000000000D+00, &
  -8.159635489766977D-06,   0.000000000000000D+00,  -4.956779539364476D-05, &
   0.000000000000000D+00,   7.138808518086386D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.215311559486417D-05,   0.000000000000000D+00, &
  -2.201196139883418D-05,   0.000000000000000D+00,  -7.106580507957288D-05, &
   0.000000000000000D+00,  -2.666747651258245D-04,   0.000000000000000D+00, &
  -2.717844326807615D-04,   0.000000000000000D+00,   2.667917443417206D-06, &
   0.000000000000000D+00,   8.232842713057487D-06,   0.000000000000000D+00, &
   5.778352738626142D-05,   0.000000000000000D+00,   1.197311500340001D-04, &
   0.000000000000000D+00,  -1.028620299321905D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.991262961598745D-06,   0.000000000000000D+00, &
   7.245102341967292D-06,   0.000000000000000D+00,   5.778377507307189D-06, &
   0.000000000000000D+00,   9.540971463134720D-05,   0.000000000000000D+00, &
   3.298756148153949D-04,   0.000000000000000D+00,   2.923289399147370D-04, &
   0.000000000000000D+00,  -2.357989969918700D-06,   0.000000000000000D+00, &
  -3.418924351093852D-06,   0.000000000000000D+00,  -1.851578303289464D-05, &
   0.000000000000000D+00,  -1.164829188568651D-04,   0.000000000000000D+00, &
  -1.542332891845476D-04,   0.000000000000000D+00,   5.856310473658402D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.409224923397804D-08, &
   0.000000000000000D+00,  -1.515813000354033D-06,   0.000000000000000D+00, &
   3.614146374641651D-06,   0.000000000000000D+00,   2.783060220980732D-05, &
   0.000000000000000D+00,  -1.064365566932751D-04,   0.000000000000000D+00, &
  -2.711639771261896D-04,   0.000000000000000D+00,  -2.595394910318938D-04, &
   0.000000000000000D+00,   4.725695785011938D-07,   0.000000000000000D+00, &
   5.724354303179598D-07,   0.000000000000000D+00,   1.544917230930557D-06, &
   0.000000000000000D+00,   1.454049514363368D-05,   0.000000000000000D+00, &
   1.435256210696913D-04,   0.000000000000000D+00,   1.169589671398000D-04, &
   0.000000000000000D+00,  -1.359363261259285D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.866874956747439D-07,   0.000000000000000D+00, &
   4.851201033414626D-07,   0.000000000000000D+00,  -8.885581088009923D-07, &
   0.000000000000000D+00,  -8.422652731974674D-06,   0.000000000000000D+00, &
  -4.494813008399947D-05,   0.000000000000000D+00,   1.020742912710588D-04, &
   0.000000000000000D+00,   1.562058946814602D-04,   0.000000000000000D+00, &
   1.538131326630015D-04,   0.000000000000000D+00,   3.436697600608361D-08, &
   0.000000000000000D+00,  -1.717723172415013D-07,   0.000000000000000D+00, &
   7.975200413803356D-07,   0.000000000000000D+00,   6.756316228345414D-07, &
   0.000000000000000D+00,  -9.748187904112846D-06,   0.000000000000000D+00, &
  -1.156082781559342D-04,   0.000000000000000D+00,  -5.066258155920596D-05, &
   0.000000000000000D+00,  -6.601623108193971D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.160344678852327D-09,   0.000000000000000D+00, &
  -9.953845895327885D-08,   0.000000000000000D+00,  -2.320728066692198D-07, &
   0.000000000000000D+00,   5.490190387853214D-07,   0.000000000000000D+00, &
   8.834115799657551D-06,   0.000000000000000D+00,   2.289199204628304D-05, &
   0.000000000000000D+00,  -7.783641413582491D-05,   0.000000000000000D+00, &
  -7.105916802771200D-05,   0.000000000000000D+00,  -5.218610460266790D-05, &
   0.000000000000000D+00,   1.170781443084380D-10,   0.000000000000000D+00, &
   7.478862895396795D-08,   0.000000000000000D+00,  -2.685479130163423D-07, &
   0.000000000000000D+00,  -1.839027589836648D-07,   0.000000000000000D+00, &
  -1.933999716830372D-06,   0.000000000000000D+00,   1.704998491137098D-05, &
   0.000000000000000D+00,   5.553292911719418D-05,   0.000000000000000D+00, &
   6.930303954483806D-06,   0.000000000000000D+00,  -6.089093667652733D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.993549677152591D-08,   0.000000000000000D+00, &
   1.586486569222069D-07,   0.000000000000000D+00,  -2.990240302005754D-07, &
   0.000000000000000D+00,   5.845617468058177D-07,   0.000000000000000D+00, &
  -9.029978252539761D-06,   0.000000000000000D+00,   1.199038455677183D-05, &
   0.000000000000000D+00,   4.427793442232718D-05,   0.000000000000000D+00, &
   2.616923825910540D-05,   0.000000000000000D+00,   6.702579068254104D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.683793958774408D-09,   0.000000000000000D+00,   1.859775639873156D-08, &
   0.000000000000000D+00,   1.864155457270626D-07,   0.000000000000000D+00, &
  -4.425025217886840D-07,   0.000000000000000D+00,   1.381952853415268D-06, &
   0.000000000000000D+00,  -2.053736329845053D-05,   0.000000000000000D+00, &
  -8.097793172604237D-06,   0.000000000000000D+00,   8.994252686387387D-06, &
   0.000000000000000D+00,   6.550627105671959D-06,   0.000000000000000D+00 /)
lem(:,146) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.100708274748691D-04,   1.147031517360180D-03,   8.608077832863622D-04, &
  -2.630605584771129D-05,  -3.030632496103632D-04,  -2.135788714895344D-03, &
  -2.217390098752922D-04,   2.610725406628878D-03,  -7.765614244308191D-04, &
  -7.295599153120454D-04,   6.404052719089623D-04,  -3.663150569935710D-04, &
   2.877745772077453D-04,  -1.418043046830401D-04,  -8.988037699784990D-05, &
   2.505004006348703D-04,  -2.295441931929616D-04,   6.644900884441009D-05, &
   8.125675322713017D-05,  -1.020037313580914D-04,   2.825265455490739D-05, &
   3.448739643368224D-05,   4.100708274748636D-04,   1.147031517360188D-03, &
   8.608077832863642D-04,  -2.630605584771407D-05,  -3.030632496103673D-04, &
  -2.135788714895339D-03,  -2.217390098752940D-04,   2.610725406628874D-03, &
  -7.765614244308194D-04,  -7.295599153120374D-04,   6.404052719089628D-04, &
  -3.663150569935772D-04,   2.877745772077466D-04,  -1.418043046830383D-04, &
  -8.988037699785438D-05,   2.505004006348694D-04,  -2.295441931929588D-04, &
   6.644900884441264D-05,   8.125675322712712D-05,  -1.020037313580932D-04, &
   2.825265455490961D-05,   3.448739643368267D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.673295425590302D-04,   0.000000000000000D+00, &
  -1.696475607575023D-04,   0.000000000000000D+00,   6.069081335753952D-03, &
   0.000000000000000D+00,  -1.556483190758877D-02,   0.000000000000000D+00, &
   1.665072604855691D-02,   0.000000000000000D+00,  -7.752584394490895D-03, &
   0.000000000000000D+00,   6.680481309962455D-04,   0.000000000000000D+00, &
   5.476741044684654D-04,   0.000000000000000D+00,  -1.646270684355061D-04, &
   0.000000000000000D+00,   5.642998006249055D-05,   0.000000000000000D+00, &
  -6.975446428571433D-06,  -1.679274140211644D-05,   0.000000000000000D+00, &
  -1.007564484126985D-05,   0.000000000000000D+00,   2.247643849206354D-05, &
   1.317584325396841D-05,   0.000000000000000D+00,   8.938905423280377D-06, &
   0.000000000000000D+00,   1.824428630615738D-05,   0.000000000000000D+00, &
  -1.076199557236679D-06,   0.000000000000000D+00,  -7.921115370957040D-06, &
  -2.167048020134312D-06,   0.000000000000000D+00,  -8.371576071846946D-06, &
   0.000000000000000D+00,  -2.275739865761986D-05,   0.000000000000000D+00, &
  -1.701575048398939D-05,   0.000000000000000D+00,  -9.254783999975964D-06, &
   0.000000000000000D+00,  -7.687089363351010D-07,   0.000000000000000D+00, &
  -1.588151601014252D-04,  -2.467296476889479D-06,   0.000000000000000D+00, &
   6.989365770725118D-06,   0.000000000000000D+00,   2.910055470699221D-05, &
   0.000000000000000D+00,   1.299837712539419D-04,   0.000000000000000D+00, &
   1.104283535386522D-05,   0.000000000000000D+00,   7.871332497323124D-06, &
   0.000000000000000D+00,  -2.633719123619118D-06,   0.000000000000000D+00, &
  -1.225248889300615D-05,   0.000000000000000D+00,   4.839992905628560D-04, &
   3.090391620034156D-07,   0.000000000000000D+00,  -5.128889188349966D-06, &
   0.000000000000000D+00,  -1.985055850401032D-05,   0.000000000000000D+00, &
  -8.243564104772869D-05,   0.000000000000000D+00,  -3.434519703454365D-04, &
   0.000000000000000D+00,  -4.270167324599759D-06,   0.000000000000000D+00, &
  -1.648726533100699D-06,   0.000000000000000D+00,   7.112146232439419D-06, &
   0.000000000000000D+00,   5.001349589328033D-05,   0.000000000000000D+00, &
   6.444754768898556D-05,   0.000000000000000D+00,  -6.888253605772994D-04, &
   9.760401665748062D-07,   0.000000000000000D+00,   1.584379109586487D-06, &
   0.000000000000000D+00,   5.685716904755846D-06,   0.000000000000000D+00, &
   1.553618420077930D-05,   0.000000000000000D+00,   1.439193330836288D-04, &
   0.000000000000000D+00,   4.753820179614827D-04,   0.000000000000000D+00, &
   5.689525044503138D-07,   0.000000000000000D+00,  -5.668679145348388D-07, &
   0.000000000000000D+00,  -4.455443986007152D-06,   0.000000000000000D+00, &
  -2.296812070071748D-05,   0.000000000000000D+00,  -1.261146873915508D-04, &
   0.000000000000000D+00,  -9.654694889499048D-05,   0.000000000000000D+00, &
   5.399372683476860D-04,  -3.373273658874328D-07,   0.000000000000000D+00, &
   2.952927200278581D-07,   0.000000000000000D+00,   2.212144155338433D-07, &
   0.000000000000000D+00,   2.802080502769118D-06,   0.000000000000000D+00, &
   1.731295007768326D-05,   0.000000000000000D+00,  -1.783501688924865D-04, &
   0.000000000000000D+00,  -3.675688748242198D-04,   0.000000000000000D+00, &
   7.146187122647823D-08,   0.000000000000000D+00,   5.564614253927258D-08, &
   0.000000000000000D+00,   9.925132034820809D-07,   0.000000000000000D+00, &
   4.242411299612531D-06,   0.000000000000000D+00,   2.412636819682805D-05, &
   0.000000000000000D+00,   1.694745790484060D-04,   0.000000000000000D+00, &
   5.414492912571831D-05,   0.000000000000000D+00,  -2.103585322601944D-04, &
  -7.049426675364037D-09,   0.000000000000000D+00,  -1.662297809408989D-07, &
   0.000000000000000D+00,  -2.917319714108803D-07,   0.000000000000000D+00, &
  -1.393099902989086D-06,   0.000000000000000D+00,  -1.117854527884966D-05, &
   0.000000000000000D+00,  -3.835299724333374D-05,   0.000000000000000D+00, &
   1.640778451980881D-04,   0.000000000000000D+00,   1.495190230181749D-04, &
   0.000000000000000D+00,   2.713968311393437D-10,   0.000000000000000D+00, &
   8.070497353113320D-08,   0.000000000000000D+00,  -1.514591555793368D-07, &
   0.000000000000000D+00,  -2.058546011764625D-07,   0.000000000000000D+00, &
   8.409130232809264D-07,   0.000000000000000D+00,  -2.180318481995559D-05, &
   0.000000000000000D+00,  -1.341586635520609D-04,   0.000000000000000D+00, &
   1.339666859539218D-06,   0.000000000000000D+00,   1.252019608432232D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.831585458652096D-08, &
   0.000000000000000D+00,   3.929300334437422D-08,   0.000000000000000D+00, &
  -3.450840380594280D-08,   0.000000000000000D+00,  -3.189720773104371D-07, &
   0.000000000000000D+00,   1.494575943574779D-05,   0.000000000000000D+00, &
   1.384329840576914D-05,   0.000000000000000D+00,  -1.019245144327372D-04, &
   0.000000000000000D+00,  -2.368148368226603D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.156991040424754D-09, &
   0.000000000000000D+00,   5.314856053423058D-08,   0.000000000000000D+00, &
   2.986689645436527D-08,   0.000000000000000D+00,   2.013143602082236D-07, &
   0.000000000000000D+00,  -4.014377619822075D-06,   0.000000000000000D+00, &
   2.875726274368142D-05,   0.000000000000000D+00,   5.298219717500372D-05, &
   0.000000000000000D+00,  -7.484177833051179D-06,   0.000000000000000D+00, &
   3.963639284228230D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.231281382730736D-10,   0.000000000000000D+00,  -2.989706912166565D-08, &
   0.000000000000000D+00,   9.635763707633788D-08,   0.000000000000000D+00, &
   2.451145337146138D-07,   0.000000000000000D+00,   1.227559323867717D-06, &
   0.000000000000000D+00,  -1.374470033904040D-05,   0.000000000000000D+00, &
   2.342520746211112D-05,   0.000000000000000D+00,   3.461794709674706D-05, &
   0.000000000000000D+00,   2.744704381287080D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.475579260931331D-12, &
   0.000000000000000D+00,   4.931918549626985D-09,   0.000000000000000D+00, &
  -1.829599565305326D-08,   0.000000000000000D+00,  -8.985184625886952D-08, &
   0.000000000000000D+00,   4.440094689684855D-08,   0.000000000000000D+00, &
   1.863489238603884D-06,   0.000000000000000D+00,  -2.704078068997871D-05, &
   0.000000000000000D+00,   1.413504089516081D-06,   0.000000000000000D+00, &
  -6.628138316261674D-06,   0.000000000000000D+00,  -2.275960160571696D-05 /)
lem(:,147) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.941106454769790D-04,  -1.674732717061904D-03, &
  -3.893296265461645D-03,  -1.348698108022148D-03,   5.123758990089480D-03, &
   1.582767022413019D-03,  -3.667306483795488D-03,   5.115717303299576D-04, &
   9.425477328495354D-04,  -6.344495199871752D-04,   2.037455377690629D-04, &
  -1.511241503599994D-04,   1.659484539025773D-04,  -4.363031795642163D-05, &
  -8.922264299499395D-05,   1.552260417949900D-04,  -1.412584045183279D-04, &
   4.680982023847060D-05,   5.335102677483578D-05,  -7.364908038458144D-05, &
   2.323473532531897D-05,   0.000000000000000D+00,   2.941106454769766D-04, &
   1.674732717061899D-03,   3.893296265461637D-03,   1.348698108022143D-03, &
  -5.123758990089470D-03,  -1.582767022413017D-03,   3.667306483795482D-03, &
  -5.115717303299585D-04,  -9.425477328495190D-04,   6.344495199871790D-04, &
  -2.037455377690866D-04,   1.511241503599939D-04,  -1.659484539025620D-04, &
   4.363031795642402D-05,   8.922264299499174D-05,  -1.552260417949935D-04, &
   1.412584045183283D-04,  -4.680982023846941D-05,  -5.335102677483971D-05, &
   7.364908038458286D-05,  -2.323473532531601D-05,   0.000000000000000D+00, &
   1.470553227384795D-04,   0.000000000000000D+00,  -6.313453403451229D-04, &
   0.000000000000000D+00,   4.131808734391480D-03,   0.000000000000000D+00, &
  -1.612308619711108D-02,   0.000000000000000D+00,   3.165175319029987D-02, &
   0.000000000000000D+00,  -3.316865404760185D-02,   0.000000000000000D+00, &
   1.870345290122032D-02,   0.000000000000000D+00,  -5.374150376091564D-03, &
   0.000000000000000D+00,   8.472569410515666D-04,   0.000000000000000D+00, &
  -2.077162438088464D-04,   0.000000000000000D+00,   6.636335971787123D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.162574404761905D-05, &
   0.000000000000000D+00,  -1.007564484126981D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.325148809523831D-06,   0.000000000000000D+00, &
   5.296172288359784D-05,   0.000000000000000D+00,   1.472594246031751D-05, &
   0.000000000000000D+00,   2.635168650793645D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.525545634920637D-06,   0.000000000000000D+00, &
  -2.496599176286680D-05,   0.000000000000000D+00,  -2.426756638182418D-04, &
   0.000000000000000D+00,  -1.698063221500720D-05,   0.000000000000000D+00, &
  -2.812993432787898D-05,   0.000000000000000D+00,  -7.389386262477352D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.686049878717588D-06, &
   0.000000000000000D+00,   1.817308552842437D-05,   0.000000000000000D+00, &
   1.111548720605539D-04,   0.000000000000000D+00,   7.601992500733128D-04, &
   0.000000000000000D+00,   1.291043388656000D-05,   0.000000000000000D+00, &
   1.489316182079547D-05,   0.000000000000000D+00,  -7.421562080241045D-06, &
   0.000000000000000D+00,  -1.698710977227943D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.497998012133863D-06,   0.000000000000000D+00, &
  -1.069840312062507D-05,   0.000000000000000D+00,  -5.374344522660166D-05, &
   0.000000000000000D+00,  -2.244847682523566D-04,   0.000000000000000D+00, &
  -1.520476460562043D-03,   0.000000000000000D+00,  -4.829511538597127D-06, &
   0.000000000000000D+00,  -2.315614581786456D-06,   0.000000000000000D+00, &
   1.652449272553068D-05,   0.000000000000000D+00,   1.376116182365999D-04, &
   0.000000000000000D+00,   5.103710412375815D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.234397723938240D-06,   0.000000000000000D+00, &
   2.192384496162285D-06,   0.000000000000000D+00,   1.415698337952303D-05, &
   0.000000000000000D+00,   3.552991880743294D-05,   0.000000000000000D+00, &
   2.923289399147370D-04,   0.000000000000000D+00,   1.981008157698581D-03, &
   0.000000000000000D+00,   3.082462723838932D-07,   0.000000000000000D+00, &
  -5.133735642948467D-07,   0.000000000000000D+00,  -1.032063198836698D-05, &
   0.000000000000000D+00,  -5.167700436190971D-05,   0.000000000000000D+00, &
  -3.109238640906486D-04,   0.000000000000000D+00,  -7.250191595630848D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.254860364825532D-07, &
   0.000000000000000D+00,   9.163542372660185D-07,   0.000000000000000D+00, &
   1.771773797070893D-07,   0.000000000000000D+00,   6.616743135223183D-06, &
   0.000000000000000D+00,   4.452394761033676D-05,   0.000000000000000D+00, &
  -3.199865619889988D-04,   0.000000000000000D+00,  -1.715198057434663D-03, &
   0.000000000000000D+00,   1.802639107033177D-07,   0.000000000000000D+00, &
  -3.507787538348013D-07,   0.000000000000000D+00,   3.296692078200491D-06, &
   0.000000000000000D+00,   7.797471344084743D-06,   0.000000000000000D+00, &
   5.534760220403140D-05,   0.000000000000000D+00,   4.085464058582044D-04, &
   0.000000000000000D+00,   5.702509441131757D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.020932969894109D-08,   0.000000000000000D+00, &
  -3.127179941669431D-07,   0.000000000000000D+00,  -1.289604004618827D-06, &
   0.000000000000000D+00,  -1.784406451212481D-06,   0.000000000000000D+00, &
  -2.701078374656345D-05,   0.000000000000000D+00,  -9.637034759287811D-05, &
   0.000000000000000D+00,   3.380238945787757D-04,   0.000000000000000D+00, &
   9.976476255897542D-04,   0.000000000000000D+00,   2.571378961093124D-09, &
   0.000000000000000D+00,   2.553502255714325D-07,   0.000000000000000D+00, &
  -8.854662313060814D-07,   0.000000000000000D+00,  -7.822488541629141D-07, &
   0.000000000000000D+00,   2.837164422956429D-06,   0.000000000000000D+00, &
  -4.998245383112602D-05,   0.000000000000000D+00,  -3.538347777356960D-04, &
   0.000000000000000D+00,  -2.241046312973308D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.905208525554365D-11,   0.000000000000000D+00, &
  -5.202444022096961D-08,   0.000000000000000D+00,   6.725056907645831D-07, &
   0.000000000000000D+00,  -1.165662061381185D-06,   0.000000000000000D+00, &
  -2.436311156580258D-07,   0.000000000000000D+00,   3.495369216324462D-05, &
   0.000000000000000D+00,   4.737019835575434D-05,   0.000000000000000D+00, &
  -2.962790880174827D-04,   0.000000000000000D+00,  -3.980035937967754D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.621508443358333D-09,   0.000000000000000D+00,   1.319230529847389D-07, &
   0.000000000000000D+00,   1.142939515835250D-06,   0.000000000000000D+00, &
  -5.164054023605024D-07,   0.000000000000000D+00,  -8.454295522488836D-06, &
   0.000000000000000D+00,   6.551257975941347D-05,   0.000000000000000D+00, &
   1.823987935868071D-04,   0.000000000000000D+00,   2.262369880455776D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.014196184717017D-09,   0.000000000000000D+00, &
  -2.359309949231650D-07,   0.000000000000000D+00,   4.929758145816674D-07, &
   0.000000000000000D+00,   1.323234051913271D-06,   0.000000000000000D+00, &
   1.854929093627056D-06,   0.000000000000000D+00,  -3.201898158050235D-05, &
   0.000000000000000D+00,   4.794092383244523D-05,   0.000000000000000D+00, &
   1.760574211548985D-04,   0.000000000000000D+00,   1.199988736992787D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.174680342119233D-09,   0.000000000000000D+00,   5.754689294316654D-08, &
   0.000000000000000D+00,  -7.315827886631377D-07,   0.000000000000000D+00, &
   3.331767548181195D-07,   0.000000000000000D+00,   1.257238537609422D-07, &
   0.000000000000000D+00,   3.706088495917181D-06,   0.000000000000000D+00, &
  -6.855352139018906D-05,   0.000000000000000D+00,  -2.072559251874392D-05, &
   0.000000000000000D+00,   4.210869653175476D-05,   0.000000000000000D+00 /)
lem(:,148) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.377237075898072D-03, &
  -8.458286027628074D-03,  -1.284511416531973D-02,   1.807446353621500D-02, &
   1.481045702548160D-02,  -1.165405653029248D-02,  -9.509821152995214D-03, &
  -8.647373334520423D-04,   1.184594933388268D-03,   2.622633396447572D-03, &
   2.413764310514994D-03,   8.096880866684057D-04,  -8.115910111135584D-04, &
  -1.242110816516658D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.377237075898134D-03,  -8.458286027628052D-03,  -1.284511416531980D-02, &
   1.807446353621498D-02,   1.481045702548163D-02,  -1.165405653029246D-02, &
  -9.509821152995203D-03,  -8.647373334520265D-04,   1.184594933388224D-03, &
   2.622633396447589D-03,   2.413764310515057D-03,   8.096880866683945D-04, &
  -8.115910111135826D-04,  -1.242110816516610D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.249043324423612D-04,   0.000000000000000D+00, &
   3.834098328646736D-06,   0.000000000000000D+00,   1.972272802800910D-07, &
   0.000000000000000D+00,   1.479204602059686D-08,   0.000000000000000D+00, &
  -2.075072591965852D-03,   0.000000000000000D+00,  -4.149168908894246D-05, &
   0.000000000000000D+00,  -3.070376774778082D-06,   0.000000000000000D+00, &
  -3.078594578119357D-07,   0.000000000000000D+00,  -4.131372228519892D-08, &
   4.741348018146150D-03,   0.000000000000000D+00,   2.831559191769099D-05, &
   0.000000000000000D+00,   1.170317918870848D-05,   0.000000000000000D+00, &
   2.017482134814933D-06,   0.000000000000000D+00,   3.329858195504413D-07, &
   0.000000000000000D+00,  -3.023258629665951D-03,   0.000000000000000D+00, &
   5.677717311180754D-04,   0.000000000000000D+00,   1.727546619187492D-05, &
   0.000000000000000D+00,  -2.699805205189688D-06,   0.000000000000000D+00, &
  -1.147937245073750D-06,   0.000000000000000D+00,  -3.879386343276659D-07, &
  -5.720027761112477D-03,   0.000000000000000D+00,  -1.775198011261867D-03, &
   0.000000000000000D+00,  -1.305422922571002D-04,   0.000000000000000D+00, &
  -1.582276853164967D-05,   0.000000000000000D+00,  -4.411590270409389D-07, &
   0.000000000000000D+00,   5.689525044503138D-07,   0.000000000000000D+00, &
   1.181781595974501D-02,   0.000000000000000D+00,   1.409852091498109D-03, &
   0.000000000000000D+00,   2.456958844090382D-05,   0.000000000000000D+00, &
   3.675817237447759D-05,   0.000000000000000D+00,   1.005805259130413D-05, &
   0.000000000000000D+00,   1.324439319037129D-06,   0.000000000000000D+00, &
  -3.796051709592077D-07,  -6.156379055735834D-03,   0.000000000000000D+00, &
   1.943312663602548D-03,   0.000000000000000D+00,   6.162174190933195D-04, &
   0.000000000000000D+00,   5.043168749189940D-05,   0.000000000000000D+00, &
  -7.277720524527077D-06,   0.000000000000000D+00,  -5.403497073265632D-06, &
   0.000000000000000D+00,  -1.315537520370208D-06,   0.000000000000000D+00, &
  -3.539837011004651D-03,   0.000000000000000D+00,  -4.394843906726759D-03, &
   0.000000000000000D+00,  -7.996733717146501D-04,   0.000000000000000D+00, &
  -1.570464753676090D-04,   0.000000000000000D+00,  -3.685110288856340D-05, &
   0.000000000000000D+00,  -1.994009599771187D-06,   0.000000000000000D+00, &
   2.460419077243910D-06,   0.000000000000000D+00,   1.457062788083543D-06, &
   5.991963831269982D-03,   0.000000000000000D+00,   2.093680223759923D-03, &
   0.000000000000000D+00,  -5.394871433627488D-04,   0.000000000000000D+00, &
  -7.824909266943304D-05,   0.000000000000000D+00,   3.377292092407632D-05, &
   0.000000000000000D+00,   2.013266240761629D-05,   0.000000000000000D+00, &
   3.962066486161473D-06,   0.000000000000000D+00,  -8.457304079234184D-07, &
   0.000000000000000D+00,  -2.849256375396529D-03,   0.000000000000000D+00, &
   1.647689123056277D-03,   0.000000000000000D+00,   1.657573528987656D-03, &
   0.000000000000000D+00,   3.570084220156669D-04,   0.000000000000000D+00, &
   7.771190489024729D-05,   0.000000000000000D+00,   9.117473031361262D-07, &
   0.000000000000000D+00,  -9.090830329679717D-06,   0.000000000000000D+00, &
  -3.463824567043381D-06,   0.000000000000000D+00,   8.330890902985170D-07, &
   4.877947688477916D-04,   0.000000000000000D+00,  -2.245577164093711D-03, &
   0.000000000000000D+00,  -5.827892683430648D-04,   0.000000000000000D+00, &
   4.304486273903047D-05,   0.000000000000000D+00,  -5.910602841857929D-05, &
   0.000000000000000D+00,  -3.984850035864523D-05,   0.000000000000000D+00, &
  -8.355286063250623D-06,   0.000000000000000D+00,   3.060480717875587D-06, &
   0.000000000000000D+00,   2.317537947928233D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.839195888719511D-04, &
   0.000000000000000D+00,  -9.336616385702186D-04,   0.000000000000000D+00, &
  -4.896030683354715D-04,   0.000000000000000D+00,  -1.119094176384783D-04, &
   0.000000000000000D+00,  -6.156309815272161D-06,   0.000000000000000D+00, &
   1.591234219066087D-05,   0.000000000000000D+00,   7.524671901603998D-06, &
   0.000000000000000D+00,  -4.450754073874750D-07,   0.000000000000000D+00, &
  -2.550195764612346D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.940651516863459D-05,   0.000000000000000D+00,   7.827539166771909D-04, &
   0.000000000000000D+00,   4.702155960827984D-05,   0.000000000000000D+00, &
   3.373904569911270D-05,   0.000000000000000D+00,   4.451711953030083D-05, &
   0.000000000000000D+00,   1.474887336643993D-05,   0.000000000000000D+00, &
  -4.086817401973734D-06,   0.000000000000000D+00,  -4.839400902034631D-06, &
   0.000000000000000D+00,  -8.902290381917169D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.759066825807605D-05, &
   0.000000000000000D+00,   2.897065188500171D-05,   0.000000000000000D+00, &
   4.176938797670679D-04,   0.000000000000000D+00,   1.293951763618676D-04, &
   0.000000000000000D+00,   2.598429638269277D-05,   0.000000000000000D+00, &
  -1.126976565928369D-05,   0.000000000000000D+00,  -1.078480984074275D-05, &
   0.000000000000000D+00,  -6.871069817905259D-07,   0.000000000000000D+00, &
   2.753110591513201D-06,   0.000000000000000D+00,   7.165172112691862D-08 /)
lem(:,149) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.608409346349514D-04,   3.164422582334402D-03,  -4.592594481236276D-03, &
  -3.366281497576373D-03,   1.156883075522746D-02,  -1.295484048433897D-03, &
  -1.054795858028854D-02,   2.319511247280091D-03,   4.787699212973159D-03, &
   4.087306263000699D-04,  -9.919503482339201D-04,  -1.301286612815375D-03, &
  -4.583545271902050D-04,   4.943574073918115D-04,   6.393573304690777D-04, &
   1.037936802427951D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.608409346349360D-04,  -3.164422582334416D-03, &
   4.592594481236297D-03,   3.366281497576363D-03,  -1.156883075522747D-02, &
   1.295484048433907D-03,   1.054795858028855D-02,  -2.319511247280100D-03, &
  -4.787699212973159D-03,  -4.087306263000727D-04,   9.919503482339255D-04, &
   1.301286612815384D-03,   4.583545271901873D-04,  -4.943574073918135D-04, &
  -6.393573304690622D-04,  -1.037936802427985D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.111054789297258D-07,   0.000000000000000D+00,  -6.229351747872020D-06, &
   0.000000000000000D+00,  -7.363447256517740D-06,   0.000000000000000D+00, &
   1.148780256428942D-05,   0.000000000000000D+00,   1.215323415215448D-05, &
   0.000000000000000D+00,  -1.304934581196278D-05,   0.000000000000000D+00, &
  -4.616773264030043D-06,   0.000000000000000D+00,   7.271482469774047D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.972272802801814D-06,   0.000000000000000D+00, &
   7.043831438590440D-08,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.322351225299150D-05,   0.000000000000000D+00, &
  -1.320718394733533D-06,   0.000000000000000D+00,  -9.104952569758801D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.123564370300139D-04, &
   0.000000000000000D+00,   7.092551272694835D-06,   0.000000000000000D+00, &
   8.874107003060553D-07,   0.000000000000000D+00,   8.346206522256466D-08, &
   0.000000000000000D+00,  -2.858750259477144D-04,   0.000000000000000D+00, &
  -8.261357702736037D-06,   0.000000000000000D+00,  -2.626428625890034D-06, &
   0.000000000000000D+00,  -5.975883869946132D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.708868867032818D-04,   0.000000000000000D+00, &
  -5.269732137412866D-05,   0.000000000000000D+00,  -1.915121715902586D-06, &
   0.000000000000000D+00,   9.208562260789180D-07,   0.000000000000000D+00, &
   6.297212821056079D-07,   0.000000000000000D+00,  -6.831514706060413D-05, &
   0.000000000000000D+00,   2.325979659402729D-04,   0.000000000000000D+00, &
   2.424483579416958D-05,   0.000000000000000D+00,   2.942059199630862D-06, &
   0.000000000000000D+00,  -2.948992971204410D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.714153587121135D-04,   0.000000000000000D+00, &
  -3.805266222285269D-04,   0.000000000000000D+00,  -3.125147532433048D-05, &
   0.000000000000000D+00,  -1.033710516527149D-05,   0.000000000000000D+00, &
  -2.357989969918700D-06,   0.000000000000000D+00,   3.082462723838932D-07, &
   0.000000000000000D+00,   9.497414543993428D-04,   0.000000000000000D+00, &
   1.429858710357118D-04,   0.000000000000000D+00,  -6.399025627027738D-05, &
   0.000000000000000D+00,  -3.287558203986230D-07,   0.000000000000000D+00, &
   4.018356769951121D-06,   0.000000000000000D+00,   1.637087160093982D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.157039855285107D-04, &
   0.000000000000000D+00,   4.891642540860371D-04,   0.000000000000000D+00, &
   2.191919177592676D-04,   0.000000000000000D+00,   3.719281005473126D-05, &
   0.000000000000000D+00,   5.280208048269486D-06,   0.000000000000000D+00, &
  -1.319287190587269D-06,   0.000000000000000D+00,  -2.218606925042864D-06, &
   0.000000000000000D+00,  -8.091207762509627D-05,   0.000000000000000D+00, &
  -8.802096032423212D-04,   0.000000000000000D+00,  -1.609820624297830D-04, &
   0.000000000000000D+00,  -3.182276834896121D-05,   0.000000000000000D+00, &
  -1.602049654544037D-05,   0.000000000000000D+00,  -4.627446780972173D-06, &
   0.000000000000000D+00,   2.414685476157380D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.718669940727841D-04,   0.000000000000000D+00, &
   5.281837129021248D-04,   0.000000000000000D+00,  -2.054684307792749D-04, &
   0.000000000000000D+00,  -6.111836504641270D-05,   0.000000000000000D+00, &
  -2.862080897808906D-06,   0.000000000000000D+00,   5.735275404238465D-06, &
   0.000000000000000D+00,   3.299174552981335D-06,   0.000000000000000D+00, &
  -3.634368656564556D-07,   0.000000000000000D+00,  -3.991164125613035D-04, &
   0.000000000000000D+00,   1.669781490858917D-04,   0.000000000000000D+00, &
   4.567286012941521D-04,   0.000000000000000D+00,   1.009252567074195D-04, &
   0.000000000000000D+00,   3.118834215924510D-05,   0.000000000000000D+00, &
   7.210032253175086D-06,   0.000000000000000D+00,  -1.381584321888951D-06, &
   0.000000000000000D+00,  -1.781337740339019D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.748482924669026D-04,   0.000000000000000D+00, &
  -4.848820621580653D-04,   0.000000000000000D+00,  -2.228015812720763D-04, &
   0.000000000000000D+00,   3.449496598549945D-05,   0.000000000000000D+00, &
  -5.532425822184609D-06,   0.000000000000000D+00,  -1.120897621350297D-05, &
   0.000000000000000D+00,  -5.480653014386482D-06,   0.000000000000000D+00, &
  -1.607020876565753D-07,   0.000000000000000D+00,   3.306070963541612D-06, &
   0.000000000000000D+00,  -4.129276422162263D-05,   0.000000000000000D+00, &
   3.278529234587710D-04,   0.000000000000000D+00,  -1.876827093026544D-04, &
   0.000000000000000D+00,  -1.497905928130990D-04,   0.000000000000000D+00, &
  -3.682852130450034D-05,   0.000000000000000D+00,  -7.651443967479317D-06, &
   0.000000000000000D+00,   2.616521748276330D-06,   0.000000000000000D+00, &
   3.159057485302791D-06,   0.000000000000000D+00,   7.504327375475041D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.181742731822966D-06, &
   0.000000000000000D+00,  -8.595541998716722D-05,   0.000000000000000D+00, &
   2.811545506297201D-04,   0.000000000000000D+00,   3.779052035309391D-05, &
   0.000000000000000D+00,   5.291159682054524D-06,   0.000000000000000D+00, &
   1.109715660968724D-05,   0.000000000000000D+00,   6.152360836215132D-06, &
   0.000000000000000D+00,   4.369381427226944D-07,   0.000000000000000D+00, &
  -2.116841218752819D-06,   0.000000000000000D+00,  -9.764972914997265D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.169193607997234D-05,   0.000000000000000D+00,  -9.452457939664434D-05, &
   0.000000000000000D+00,   1.182680425585440D-04,   0.000000000000000D+00, &
   3.716016798612354D-05,   0.000000000000000D+00,   8.757039499723927D-06, &
   0.000000000000000D+00,  -1.370623384514412D-06,   0.000000000000000D+00, &
  -3.268760429481094D-06,   0.000000000000000D+00,  -1.077871372354266D-06, &
   0.000000000000000D+00,   1.455546880962050D-07,   0.000000000000000D+00 /)
lem(:,150) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.451899320377342D-04,  -1.079638303183017D-03, &
  -1.003300197734986D-04,   2.462752265726855D-03,   5.143341189028134D-03, &
  -4.989577325752909D-03,  -1.330070577431056D-02,   5.741122207721697D-03, &
   1.321718317786741D-02,  -1.683538519305836D-03,  -5.691846169767552D-03, &
  -1.507920419139209D-03,   5.352002245386083D-04,   1.012452516330490D-03, &
   3.461177564460609D-04,  -2.079039252495125D-04,  -1.509496513893849D-04, &
   9.370246299664507D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.451899320377247D-04, &
  -1.079638303183001D-03,  -1.003300197734918D-04,   2.462752265726844D-03, &
   5.143341189028127D-03,  -4.989577325752899D-03,  -1.330070577431057D-02, &
   5.741122207721674D-03,   1.321718317786743D-02,  -1.683538519305825D-03, &
  -5.691846169767575D-03,  -1.507920419139215D-03,   5.352002245386151D-04, &
   1.012452516330499D-03,   3.461177564460488D-04,  -2.079039252495129D-04, &
  -1.509496513893819D-04,   9.370246299663423D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.835634705699754D-07,   0.000000000000000D+00,   9.856151375363110D-06, &
   0.000000000000000D+00,   3.060626211607137D-05,   0.000000000000000D+00, &
   1.007865616076672D-05,   0.000000000000000D+00,  -4.640802795093244D-05, &
   0.000000000000000D+00,  -1.944901369802237D-05,   0.000000000000000D+00, &
   3.604635913950592D-05,   0.000000000000000D+00,   6.329681943248792D-06, &
   0.000000000000000D+00,  -1.367348759942164D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.945602237950619D-06,   0.000000000000000D+00,   2.945602237966173D-08, &
   0.000000000000000D+00,  -3.286187496713834D-05,   0.000000000000000D+00, &
  -8.671116587968312D-07,   0.000000000000000D+00,   2.953273077124022D-08, &
   2.439454922965556D-05,   0.000000000000000D+00,   4.505890945134957D-06, &
   0.000000000000000D+00,   3.953428215008983D-07,   0.000000000000000D+00, &
   4.597872177627184D-04,   0.000000000000000D+00,  -8.788780590407501D-07, &
   0.000000000000000D+00,  -1.275410416469574D-06,   0.000000000000000D+00, &
  -1.983829088755882D-07,  -1.464014918988669D-03,   0.000000000000000D+00, &
   4.082386983373262D-06,   0.000000000000000D+00,  -1.456755724538441D-06, &
   0.000000000000000D+00,   4.999669532244282D-07,   0.000000000000000D+00, &
   1.173471500904535D-03,   0.000000000000000D+00,  -1.981937836023175D-04, &
   0.000000000000000D+00,   4.539315792993540D-06,   0.000000000000000D+00, &
   8.742715533670011D-07,   0.000000000000000D+00,  -9.462431392641717D-07, &
   1.645931368303152D-03,   0.000000000000000D+00,   6.313747573674878D-04, &
   0.000000000000000D+00,   4.887471836495331D-07,   0.000000000000000D+00, &
  -1.593878088984073D-06,   0.000000000000000D+00,  -2.310476557830185D-07, &
   0.000000000000000D+00,  -3.739659860893250D-03,   0.000000000000000D+00, &
  -5.129058648470990D-04,   0.000000000000000D+00,   8.136676194288230D-05, &
   0.000000000000000D+00,   4.859943529345893D-07,   0.000000000000000D+00, &
   7.328361831722457D-07,   0.000000000000000D+00,   1.393955592662113D-07, &
   1.793960813587999D-03,   0.000000000000000D+00,  -7.393583452477930D-04, &
   0.000000000000000D+00,  -3.244001365745442D-04,   0.000000000000000D+00, &
  -1.510144025235159D-05,   0.000000000000000D+00,  -8.636552923819741D-07, &
   0.000000000000000D+00,  -5.668679145348388D-07,   0.000000000000000D+00, &
   1.409852091498109D-03,   0.000000000000000D+00,   1.677917005016500D-03, &
   0.000000000000000D+00,   3.035076064489734D-04,   0.000000000000000D+00, &
   2.987799202163210D-06,   0.000000000000000D+00,   4.179761786007961D-06, &
   0.000000000000000D+00,   8.735921714181594D-07,   0.000000000000000D+00, &
   2.498989133128155D-06,  -1.931637235903234D-03,   0.000000000000000D+00, &
  -8.043799752967338D-04,   0.000000000000000D+00,   3.739386631665690D-04, &
   0.000000000000000D+00,   1.274574100468044D-04,   0.000000000000000D+00, &
   8.412664397637067D-06,   0.000000000000000D+00,  -7.093031526431956D-07, &
   0.000000000000000D+00,  -9.275862756346661D-07,   0.000000000000000D+00, &
   6.758310350847694D-04,   0.000000000000000D+00,  -6.160735649810223D-04, &
   0.000000000000000D+00,  -9.035317594273299D-04,   0.000000000000000D+00, &
  -1.803890975786334D-04,   0.000000000000000D+00,  -3.162406105775584D-05, &
   0.000000000000000D+00,  -6.416487766246305D-06,   0.000000000000000D+00, &
  -9.122287415389517D-07,   0.000000000000000D+00,  -2.332586879562018D-07, &
   3.484429531804840D-05,   0.000000000000000D+00,   8.299688744826711D-04, &
   0.000000000000000D+00,   4.064145781598479D-04,   0.000000000000000D+00, &
  -1.240136533556764D-04,   0.000000000000000D+00,  -1.846881704856494D-05, &
   0.000000000000000D+00,   7.075280154905629D-06,   0.000000000000000D+00, &
   3.960380018957528D-06,   0.000000000000000D+00,   7.842532601617625D-07, &
   0.000000000000000D+00,  -5.887993023457569D-05,   0.000000000000000D+00, &
  -2.877587868230466D-04,   0.000000000000000D+00,   3.628059152056788D-04, &
   0.000000000000000D+00,   3.920363440738272D-04,   0.000000000000000D+00, &
   8.695313494149865D-05,   0.000000000000000D+00,   1.787194398963524D-05, &
   0.000000000000000D+00,   5.923051482652621D-08,   0.000000000000000D+00, &
  -1.642436646690619D-06,   0.000000000000000D+00,  -3.062967812413605D-06, &
   6.042915196633645D-06,   0.000000000000000D+00,   7.080054248163696D-06, &
   0.000000000000000D+00,  -4.235046645220767D-04,   0.000000000000000D+00, &
  -1.351527930107809D-04,   0.000000000000000D+00,   1.169142254492287D-05, &
   0.000000000000000D+00,  -1.482209317175121D-05,   0.000000000000000D+00, &
  -9.503669450444258D-06,   0.000000000000000D+00,  -1.798895990439537D-06, &
   0.000000000000000D+00,   1.205038345501960D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.555390472879539D-06, &
   0.000000000000000D+00,   1.171374753498166D-04,   0.000000000000000D+00, &
  -2.142014354299925D-04,   0.000000000000000D+00,  -1.265506469155292D-04, &
   0.000000000000000D+00,  -2.908011272539711D-05,   0.000000000000000D+00, &
  -1.215937843774778D-06,   0.000000000000000D+00,   3.976536078772789D-06, &
   0.000000000000000D+00,   2.224521303017490D-06,   0.000000000000000D+00, &
   9.700822994280427D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.043983114359501D-05,   0.000000000000000D+00,  -1.456337706856368D-06, &
   0.000000000000000D+00,   1.659397683825674D-04,   0.000000000000000D+00, &
   1.298295145204358D-05,   0.000000000000000D+00,   9.117611101902769D-06, &
   0.000000000000000D+00,   1.194285479743490D-05,   0.000000000000000D+00, &
   3.641989750277392D-06,   0.000000000000000D+00,  -1.545254230723247D-06, &
   0.000000000000000D+00,  -1.041191994769106D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.619484521390580D-06, &
   0.000000000000000D+00,   1.156729562664224D-05,   0.000000000000000D+00, &
   2.283533000081281D-06,   0.000000000000000D+00,   1.032169398662661D-04, &
   0.000000000000000D+00,   3.492285034800830D-05,   0.000000000000000D+00, &
   6.832823925825707D-06,   0.000000000000000D+00,  -3.310367072637556D-06, &
   0.000000000000000D+00,  -2.826989183095552D-06,   0.000000000000000D+00, &
  -6.058820523629014D-07,   0.000000000000000D+00,   1.625969159417252D-06 /)
lem(:,151) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.735875913609479D-05, &
   2.750514589040308D-04,  -7.695257808539851D-04,  -1.156894234558638D-03, &
   5.230247293222750D-03,   7.409877188077632D-04,  -1.288413220561327D-02, &
   2.117754630711512D-03,   1.712014039045022D-02,  -4.696572562153424D-03, &
  -1.281039331630118D-02,   3.061758952012161D-03,   5.640668023410075D-03, &
  -2.561898116945866D-05,  -1.645917635057253D-03,  -8.022538349660280D-04, &
   3.252235230300811D-04,   4.382452576005384D-04,   1.858995264913516D-06, &
  -1.745381432296512D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.735875913609418D-05,  -2.750514589040358D-04,   7.695257808539789D-04, &
   1.156894234558655D-03,  -5.230247293222748D-03,  -7.409877188077793D-04, &
   1.288413220561329D-02,  -2.117754630711496D-03,  -1.712014039045024D-02, &
   4.696572562153414D-03,   1.281039331630123D-02,  -3.061758952012154D-03, &
  -5.640668023410072D-03,   2.561898116946430D-05,   1.645917635057249D-03, &
   8.022538349660270D-04,  -3.252235230300777D-04,  -4.382452576005520D-04, &
  -1.858995264904180D-06,   1.745381432296607D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.185391948556935D-06, &
   0.000000000000000D+00,  -5.156914384804979D-05,   0.000000000000000D+00, &
  -6.518817298525215D-05,   0.000000000000000D+00,   1.732366966257554D-05, &
   0.000000000000000D+00,   7.355129700120012D-05,   0.000000000000000D+00, &
   2.064740017743872D-05,   0.000000000000000D+00,  -4.185471856820876D-05, &
   0.000000000000000D+00,  -2.193132855477287D-05,   0.000000000000000D+00, &
   1.794853393458622D-05,   0.000000000000000D+00,   3.269133236787923D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.560379077406098D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.243171983552297D-06,   0.000000000000000D+00, &
  -3.477447086458110D-08,   0.000000000000000D+00,  -7.636473801887350D-06, &
   0.000000000000000D+00,  -7.787435916583486D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.867538384777764D-05,   0.000000000000000D+00, &
  -2.188319208491046D-06,   0.000000000000000D+00,   6.913304884093740D-07, &
   0.000000000000000D+00,   1.855068250276217D-04,   0.000000000000000D+00, &
   9.989620444786474D-06,   0.000000000000000D+00,   8.979584376677988D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.128506527348902D-04, &
   0.000000000000000D+00,   2.531729114810852D-05,   0.000000000000000D+00, &
  -1.566671067709579D-06,   0.000000000000000D+00,   4.992665722576605D-07, &
   0.000000000000000D+00,   1.197083677902092D-04,   0.000000000000000D+00, &
  -1.814028860545404D-04,   0.000000000000000D+00,  -9.896346375732133D-06, &
   0.000000000000000D+00,  -4.356136153821678D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.128723387708450D-04,   0.000000000000000D+00, &
   3.462910586858564D-04,   0.000000000000000D+00,   9.088481388156503D-06, &
   0.000000000000000D+00,   3.971358523619991D-06,   0.000000000000000D+00, &
  -8.942059682998260D-07,   0.000000000000000D+00,  -7.480802370039545D-04, &
   0.000000000000000D+00,  -1.590931508436267D-04,   0.000000000000000D+00, &
   8.888833680191857D-05,   0.000000000000000D+00,   2.805438219020050D-06, &
   0.000000000000000D+00,  -1.156257235947030D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.506685853534541D-04,   0.000000000000000D+00, &
  -4.388748008755785D-04,   0.000000000000000D+00,  -2.434332279326577D-04, &
   0.000000000000000D+00,  -2.541947224598156D-05,   0.000000000000000D+00, &
  -3.418924351093852D-06,   0.000000000000000D+00,  -5.133735642948467D-07, &
   0.000000000000000D+00,   1.429858710357118D-04,   0.000000000000000D+00, &
   8.248517448390082D-04,   0.000000000000000D+00,   1.579911173779101D-04, &
   0.000000000000000D+00,  -1.991974573069760D-06,   0.000000000000000D+00, &
   7.675240731376841D-06,   0.000000000000000D+00,   3.332330578890787D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.150414591679770D-04, &
   0.000000000000000D+00,  -4.787111743796252D-04,   0.000000000000000D+00, &
   2.747574463569521D-04,   0.000000000000000D+00,   1.096208912474864D-04, &
   0.000000000000000D+00,   1.231947143753685D-05,   0.000000000000000D+00, &
  -1.522744833439030D-06,   0.000000000000000D+00,  -9.800848958672680D-07, &
   0.000000000000000D+00,   2.801169348222646D-04,   0.000000000000000D+00, &
  -1.898708552895799D-04,   0.000000000000000D+00,  -5.710790946099737D-04, &
   0.000000000000000D+00,  -1.189842100736601D-04,   0.000000000000000D+00, &
  -2.764866945635655D-05,   0.000000000000000D+00,  -8.837997515699179D-06, &
   0.000000000000000D+00,  -8.956119758389770D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.282046511520466D-05,   0.000000000000000D+00, &
   4.586832032050648D-04,   0.000000000000000D+00,   2.998900871970221D-04, &
   0.000000000000000D+00,  -9.843343070200173D-05,   0.000000000000000D+00, &
  -2.075527253173003D-05,   0.000000000000000D+00,   5.197845992934111D-06, &
   0.000000000000000D+00,   4.314589734191615D-06,   0.000000000000000D+00, &
   1.706140349783170D-06,   0.000000000000000D+00,  -1.043679230739189D-05, &
   0.000000000000000D+00,  -2.739643499769489D-04,   0.000000000000000D+00, &
   1.876809488927795D-04,   0.000000000000000D+00,   2.668506216283533D-04, &
   0.000000000000000D+00,   6.210723508397070D-05,   0.000000000000000D+00, &
   1.646225698333917D-05,   0.000000000000000D+00,   1.377551440485464D-06, &
   0.000000000000000D+00,  -2.914261159206306D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.171545784880627D-05,   0.000000000000000D+00, &
   4.994271592788823D-05,   0.000000000000000D+00,  -3.218310548684547D-04, &
   0.000000000000000D+00,  -1.079344468353159D-04,   0.000000000000000D+00, &
   1.219054792504617D-05,   0.000000000000000D+00,  -8.639093250299687D-06, &
   0.000000000000000D+00,  -7.989333712251705D-06,   0.000000000000000D+00, &
  -1.967414457913507D-06,   0.000000000000000D+00,   1.590176217809779D-06, &
   0.000000000000000D+00,  -2.721267749679925D-06,   0.000000000000000D+00, &
   1.638894300501731D-05,   0.000000000000000D+00,   1.362662019472518D-04, &
   0.000000000000000D+00,  -1.408544965544367D-04,   0.000000000000000D+00, &
  -8.656163671874140D-05,   0.000000000000000D+00,  -2.116350932137319D-05, &
   0.000000000000000D+00,  -2.428771586521507D-06,   0.000000000000000D+00, &
   3.462658531620759D-06,   0.000000000000000D+00,   2.331150066465516D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.965104667209384D-07, &
   0.000000000000000D+00,  -5.357578844664797D-06,   0.000000000000000D+00, &
   9.604706399547836D-06,   0.000000000000000D+00,   1.456506571517006D-04, &
   0.000000000000000D+00,   1.369513142753734D-05,   0.000000000000000D+00, &
   4.992709015375541D-06,   0.000000000000000D+00,   7.956137871810849D-06, &
   0.000000000000000D+00,   3.029285569845636D-06,   0.000000000000000D+00, &
  -6.021713594752867D-07,   0.000000000000000D+00,  -2.819273817162588D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.632347844049752D-06,   0.000000000000000D+00,  -1.762666737027512D-05, &
   0.000000000000000D+00,  -7.813143621033870D-06,   0.000000000000000D+00, &
   7.374027779329453D-05,   0.000000000000000D+00,   2.296366483139619D-05, &
   0.000000000000000D+00,   5.021635266031639D-06,   0.000000000000000D+00, &
  -1.797994312435369D-06,   0.000000000000000D+00,  -2.768033651409225D-06, &
   0.000000000000000D+00,  -3.388651442462359D-07,   0.000000000000000D+00 /)
lem(:,152) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.597810098239771D-06,  -1.047877555474345D-04,   2.902489001569056D-04, &
   6.208272464397597D-04,  -1.165877068368321D-03,  -5.123522969114199D-04, &
  -3.587443565723267D-03,   2.579666148836184D-03,   1.385322640442445D-02, &
  -6.036229428611269D-03,  -1.855041336613522D-02,   7.089163096757416D-03, &
   1.204111226346953D-02,  -4.579340555666249D-03,  -4.061292866423309D-03, &
   1.720579452074627D-03,   1.155936513394940D-03,  -4.465345293212988D-04, &
  -5.969675941567590D-04,   7.450560902733381D-05,   2.903760764912432D-04, &
   2.026845504567263D-05,   5.597810098235288D-06,  -1.047877555474266D-04, &
   2.902489001569020D-04,   6.208272464397576D-04,  -1.165877068368319D-03, &
  -5.123522969114151D-04,  -3.587443565723264D-03,   2.579666148836169D-03, &
   1.385322640442444D-02,  -6.036229428611263D-03,  -1.855041336613521D-02, &
   7.089163096757414D-03,   1.204111226346955D-02,  -4.579340555666241D-03, &
  -4.061292866423314D-03,   1.720579452074610D-03,   1.155936513394947D-03, &
  -4.465345293212789D-04,  -5.969675941567661D-04,   7.450560902731750D-05, &
   2.903760764912581D-04,   2.026845504567323D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.754107195114377D-05,   0.000000000000000D+00, &
   1.294741192442975D-04,   0.000000000000000D+00,   5.337200503321725D-05, &
   0.000000000000000D+00,  -3.410176565239234D-05,   0.000000000000000D+00, &
  -4.698645641555539D-05,   0.000000000000000D+00,  -6.478224442378990D-05, &
   0.000000000000000D+00,   2.714432186339494D-05,   0.000000000000000D+00, &
   6.464635609296491D-05,   0.000000000000000D+00,  -2.194607441417344D-05, &
   0.000000000000000D+00,  -1.450843241203623D-05,   0.000000000000000D+00, &
  -1.999634352575946D-06,   7.420769804671125D-06,   0.000000000000000D+00, &
   7.749388554341918D-06,   0.000000000000000D+00,  -1.031497742023863D-06, &
  -8.005410482190680D-05,   0.000000000000000D+00,   6.883811063357790D-08, &
   0.000000000000000D+00,   2.308493020567347D-05,   0.000000000000000D+00, &
   1.359042574212969D-06,   0.000000000000000D+00,  -5.434917393123006D-07, &
   3.952042483473364D-04,   0.000000000000000D+00,   1.807877381161654D-06, &
   0.000000000000000D+00,  -3.132123139021652D-06,   0.000000000000000D+00, &
  -5.355366877779045D-04,   0.000000000000000D+00,   6.403648549293961D-05, &
   0.000000000000000D+00,   1.153642237149368D-06,   0.000000000000000D+00, &
   2.487299405941784D-06,  -3.682812304430285D-04,   0.000000000000000D+00, &
  -2.603366481411977D-04,   0.000000000000000D+00,   1.844034131684597D-05, &
   0.000000000000000D+00,   2.251880935186991D-06,   0.000000000000000D+00, &
   1.154622941707341D-03,   0.000000000000000D+00,   2.453567692832732D-04, &
   0.000000000000000D+00,  -9.845383071296687D-05,   0.000000000000000D+00, &
  -8.329611870738647D-06,   0.000000000000000D+00,  -9.977798701939367D-07, &
  -4.106476486675716D-04,   0.000000000000000D+00,   3.107534132907667D-04, &
   0.000000000000000D+00,   2.380398274550414D-04,   0.000000000000000D+00, &
   1.256692109788954D-06,   0.000000000000000D+00,   2.329579083921259D-06, &
   0.000000000000000D+00,  -6.666893303721947D-04,   0.000000000000000D+00, &
  -7.529230474361045D-04,   0.000000000000000D+00,  -1.682626285113163D-04, &
   0.000000000000000D+00,   5.845249249738822D-05,   0.000000000000000D+00, &
   4.579680978495968D-06,   0.000000000000000D+00,  -3.293939115057444D-06, &
   5.608037706500862D-04,   0.000000000000000D+00,   3.419858062123808D-04, &
   0.000000000000000D+00,  -3.091252981956817D-04,   0.000000000000000D+00, &
  -1.704283164992861D-04,   0.000000000000000D+00,  -1.731144347911928D-05, &
   0.000000000000000D+00,  -4.455443986007152D-06,   0.000000000000000D+00, &
   2.456958844090382D-05,   0.000000000000000D+00,   3.035076064489734D-04, &
   0.000000000000000D+00,   6.636326022782173D-04,   0.000000000000000D+00, &
   1.344140876422262D-04,   0.000000000000000D+00,  -7.891749841256007D-07, &
   0.000000000000000D+00,   6.183114204489701D-06,   0.000000000000000D+00, &
   6.059751250466204D-06,  -1.433501939336112D-04,   0.000000000000000D+00, &
  -3.571618643978382D-04,   0.000000000000000D+00,  -3.379891614875666D-04, &
   0.000000000000000D+00,   1.998851705939908D-04,   0.000000000000000D+00, &
   7.817888578916733D-05,   0.000000000000000D+00,   9.052484606345399D-06, &
   0.000000000000000D+00,  -2.717976446222309D-07,   0.000000000000000D+00, &
   2.158732521142974D-05,   0.000000000000000D+00,   9.914542786249229D-05, &
   0.000000000000000D+00,  -2.046669110088177D-04,   0.000000000000000D+00, &
  -4.512538479130094D-04,   0.000000000000000D+00,  -9.678779039761526D-05, &
   0.000000000000000D+00,  -2.155506095350631D-05,   0.000000000000000D+00, &
  -7.302165532880555D-06,   0.000000000000000D+00,  -1.293591142406640D-06, &
   1.540782138261809D-05,   0.000000000000000D+00,  -7.874501604801626D-07, &
   0.000000000000000D+00,   3.174978954985336D-04,   0.000000000000000D+00, &
   2.184300020156041D-04,   0.000000000000000D+00,  -7.180460344408148D-05, &
   0.000000000000000D+00,  -1.279568776237767D-05,   0.000000000000000D+00, &
   4.167339492080359D-06,   0.000000000000000D+00,   3.799777795108989D-06, &
   0.000000000000000D+00,  -4.573362197581298D-06,   0.000000000000000D+00, &
   9.389499858767541D-06,   0.000000000000000D+00,  -1.495812611427084D-04, &
   0.000000000000000D+00,   1.602272022883594D-04,   0.000000000000000D+00, &
   2.134902067134996D-04,   0.000000000000000D+00,   5.015390938027009D-05, &
   0.000000000000000D+00,   1.245937352806301D-05,   0.000000000000000D+00, &
   6.310928977388946D-08,   0.000000000000000D+00,  -5.581349601808763D-06, &
   3.348434926362646D-07,   0.000000000000000D+00,  -4.011597539617730D-06, &
   0.000000000000000D+00,   4.239249424197367D-05,   0.000000000000000D+00, &
  -2.243547912690518D-04,   0.000000000000000D+00,  -7.890020747787720D-05, &
   0.000000000000000D+00,   7.256034637345701D-06,   0.000000000000000D+00, &
  -8.495743304961809D-06,   0.000000000000000D+00,  -6.277504188111470D-06, &
   0.000000000000000D+00,  -1.616441002770354D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.654723965015250D-06, &
   0.000000000000000D+00,  -9.643258955346020D-06,   0.000000000000000D+00, &
   9.196547693987149D-05,   0.000000000000000D+00,  -1.142301393943694D-04, &
   0.000000000000000D+00,  -7.083932515537062D-05,   0.000000000000000D+00, &
  -1.706254574857973D-05,   0.000000000000000D+00,  -6.997406811012245D-07, &
   0.000000000000000D+00,   3.734441122482583D-06,   0.000000000000000D+00, &
   4.847269072611760D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.474987785873834D-06,   0.000000000000000D+00,  -4.620611482872691D-06, &
   0.000000000000000D+00,  -1.057066802714808D-05,   0.000000000000000D+00, &
   1.033592144533433D-04,   0.000000000000000D+00,   8.297182942004424D-06, &
   0.000000000000000D+00,   5.095682958064302D-06,   0.000000000000000D+00, &
   6.785301345751179D-06,   0.000000000000000D+00,   2.599813353919404D-06, &
   0.000000000000000D+00,  -4.182778647721666D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.438734022647428D-07, &
   0.000000000000000D+00,   4.054009720035296D-06,   0.000000000000000D+00, &
  -8.432469217346658D-06,   0.000000000000000D+00,  -7.878795778994631D-06, &
   0.000000000000000D+00,   5.933613208453422D-05,   0.000000000000000D+00, &
   1.963669257507024D-05,   0.000000000000000D+00,   3.923597365381508D-06, &
   0.000000000000000D+00,  -2.471481808718031D-06,   0.000000000000000D+00, &
  -2.527031541791730D-06,   0.000000000000000D+00,  -3.922200689119510D-08 /)
lem(:,153) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.300813517140498D-04,   6.050250840561144D-04, &
  -1.614832303271025D-03,  -1.981831994364005D-03,   2.906780428658192D-03, &
   7.545427268857771D-03,  -4.712070057766877D-03,  -1.499551822667579D-02, &
   8.073264503375376D-03,   1.551800844989771D-02,  -9.646653796585868D-03, &
  -8.075296401345890D-03,   6.489572053230907D-03,   1.710697102296338D-03, &
  -2.457844920947941D-03,   1.459523610102199D-05,   7.586161263687393D-04, &
  -5.014048429749435D-05,  -3.458607711786224D-04,   1.577828438696427D-05, &
   1.754099857327060D-04,   0.000000000000000D+00,  -1.300813517140397D-04, &
  -6.050250840561179D-04,   1.614832303271015D-03,   1.981831994364002D-03, &
  -2.906780428658174D-03,  -7.545427268857776D-03,   4.712070057766876D-03, &
   1.499551822667578D-02,  -8.073264503375371D-03,  -1.551800844989768D-02, &
   9.646653796585837D-03,   8.075296401345898D-03,  -6.489572053230897D-03, &
  -1.710697102296337D-03,   2.457844920947942D-03,  -1.459523610101826D-05, &
  -7.586161263687353D-04,   5.014048429749244D-05,   3.458607711786212D-04, &
  -1.577828438696383D-05,  -1.754099857326988D-04,   0.000000000000000D+00, &
  -6.504067585701929D-05,   0.000000000000000D+00,  -7.214359926263081D-05, &
   0.000000000000000D+00,   2.045258435586357D-05,   0.000000000000000D+00, &
  -8.285497769919556D-05,   0.000000000000000D+00,  -2.118862313858111D-05, &
   0.000000000000000D+00,   1.676587587868729D-04,   0.000000000000000D+00, &
   1.861194920680989D-05,   0.000000000000000D+00,  -1.242598380772827D-04, &
   0.000000000000000D+00,   3.951056186630679D-06,   0.000000000000000D+00, &
   3.761077256929778D-05,   0.000000000000000D+00,  -1.080460043345066D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.141916906622838D-06, &
   0.000000000000000D+00,   1.182454742826256D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.445524891345321D-05,   0.000000000000000D+00, &
  -7.055637860592325D-07,   0.000000000000000D+00,  -3.094600617820435D-05, &
   0.000000000000000D+00,  -9.086415820161890D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.393722638170072D-04,   0.000000000000000D+00, &
   1.100760750444415D-05,   0.000000000000000D+00,  -6.899768076406819D-06, &
   0.000000000000000D+00,  -1.110138384670622D-04,   0.000000000000000D+00, &
   6.528914019828005D-05,   0.000000000000000D+00,   6.794179809137965D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.379835297590044D-04, &
   0.000000000000000D+00,  -1.945342454243384D-04,   0.000000000000000D+00, &
   1.149627731267327D-06,   0.000000000000000D+00,   5.483033429362782D-06, &
   0.000000000000000D+00,   3.167912201766022D-04,   0.000000000000000D+00, &
   1.304502741018421D-04,   0.000000000000000D+00,  -7.613515485097531D-05, &
   0.000000000000000D+00,  -7.551089010256532D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.544380000430616D-04,   0.000000000000000D+00, &
   2.266910888949523D-04,   0.000000000000000D+00,   1.954031612917932D-04, &
   0.000000000000000D+00,   8.977789008433184D-06,   0.000000000000000D+00, &
   6.171240914177161D-06,   0.000000000000000D+00,  -1.351503673485099D-04, &
   0.000000000000000D+00,  -4.759081526227174D-04,   0.000000000000000D+00, &
  -1.217940181496522D-04,   0.000000000000000D+00,   4.096678536321741D-05, &
   0.000000000000000D+00,   1.366235924481004D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.987958233674384D-04,   0.000000000000000D+00, &
   2.502661803177953D-04,   0.000000000000000D+00,  -2.423656126500469D-04, &
   0.000000000000000D+00,  -1.419315535832828D-04,   0.000000000000000D+00, &
  -1.851578303289464D-05,   0.000000000000000D+00,  -1.032063198836698D-05, &
   0.000000000000000D+00,  -6.399025627027738D-05,   0.000000000000000D+00, &
   1.579911173779101D-04,   0.000000000000000D+00,   4.939051676173226D-04, &
   0.000000000000000D+00,   1.082781125951814D-04,   0.000000000000000D+00, &
   5.385401295663206D-06,   0.000000000000000D+00,   7.585027615089845D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.368580990236550D-05, &
   0.000000000000000D+00,  -2.691372861901551D-04,   0.000000000000000D+00, &
  -2.652078533134717D-04,   0.000000000000000D+00,   1.604430329581977D-04, &
   0.000000000000000D+00,   6.558511972732672D-05,   0.000000000000000D+00, &
   8.964762870593283D-06,   0.000000000000000D+00,  -1.221544117626681D-06, &
   0.000000000000000D+00,   3.023383172723505D-05,   0.000000000000000D+00, &
   1.125713758475014D-04,   0.000000000000000D+00,  -1.452683681129303D-04, &
   0.000000000000000D+00,  -3.531848630842476D-04,   0.000000000000000D+00, &
  -8.090361171637975D-05,   0.000000000000000D+00,  -2.117896831576187D-05, &
   0.000000000000000D+00,  -6.170391584454799D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.721954628055658D-06,   0.000000000000000D+00, &
   9.083445199000416D-06,   0.000000000000000D+00,   2.615501525849153D-04, &
   0.000000000000000D+00,   1.752828703536348D-04,   0.000000000000000D+00, &
  -5.833797489541550D-05,   0.000000000000000D+00,  -1.100532852145089D-05, &
   0.000000000000000D+00,   4.660356190529302D-06,   0.000000000000000D+00, &
   1.010579086174458D-05,   0.000000000000000D+00,  -3.760920449098271D-06, &
   0.000000000000000D+00,  -1.496246308663884D-05,   0.000000000000000D+00, &
  -1.209708333383862D-04,   0.000000000000000D+00,   1.265406834269395D-04, &
   0.000000000000000D+00,   1.706589221904144D-04,   0.000000000000000D+00, &
   4.244061300994341D-05,   0.000000000000000D+00,   1.066276328840265D-05, &
   0.000000000000000D+00,  -9.669677219617993D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.721633946568100D-06,   0.000000000000000D+00, &
  -3.326536907328955D-06,   0.000000000000000D+00,   1.260248130981536D-05, &
   0.000000000000000D+00,  -1.861947056843112D-04,   0.000000000000000D+00, &
  -6.431671072738878D-05,   0.000000000000000D+00,   6.137757375244322D-06, &
   0.000000000000000D+00,  -7.178533905221297D-06,   0.000000000000000D+00, &
  -7.266819591200746D-06,   0.000000000000000D+00,  -4.658879021500317D-06, &
   0.000000000000000D+00,  -2.759922121125328D-07,   0.000000000000000D+00, &
   3.671805441246062D-06,   0.000000000000000D+00,  -1.222828952000394D-06, &
   0.000000000000000D+00,   6.761301436470501D-05,   0.000000000000000D+00, &
  -9.331730369472067D-05,   0.000000000000000D+00,  -5.711199831171848D-05, &
   0.000000000000000D+00,  -1.397871707361348D-05,   0.000000000000000D+00, &
  -9.642195223544744D-08,   0.000000000000000D+00,   2.725832568037650D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.544761422901139D-08, &
   0.000000000000000D+00,   2.405645934874660D-07,   0.000000000000000D+00, &
   5.326458117678828D-06,   0.000000000000000D+00,   2.341906484866978D-06, &
   0.000000000000000D+00,   8.630914696820814D-05,   0.000000000000000D+00, &
   7.076994903198330D-06,   0.000000000000000D+00,   3.748073937748418D-06, &
   0.000000000000000D+00,   6.323206314942813D-06,   0.000000000000000D+00, &
   3.247923057703529D-06,   0.000000000000000D+00,  -2.948161014574018D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -7.384022485977567D-07,   0.000000000000000D+00,   3.285255868268361D-07, &
   0.000000000000000D+00,  -2.097688995402538D-06,   0.000000000000000D+00, &
  -1.319300027889034D-06,   0.000000000000000D+00,   4.903998561187983D-05, &
   0.000000000000000D+00,   1.568974396179137D-05,   0.000000000000000D+00, &
   2.798465931190664D-06,   0.000000000000000D+00,  -1.835746577303632D-06, &
   0.000000000000000D+00,  -1.162616876306239D-06,   0.000000000000000D+00 /)
lem(:,154) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.547612149752827D-04,   3.694671725766874D-04,   1.024771532837700D-03, &
   7.751609674087481D-04,   4.875553877588612D-04,  -4.001777433634185D-03, &
  -6.410782878077352D-03,   6.914004738860340D-03,   1.276127120253350D-02, &
  -9.264478479709851D-03,  -1.069829054343620D-02,   9.499264625393673D-03, &
   2.799550201526623D-03,  -5.857038883413349D-03,   1.319451649211176D-03, &
   1.841923293008857D-03,  -1.097269976499565D-03,  -2.840999908552721D-04, &
   4.336523772397976D-04,   3.642625809616299D-05,  -2.036634866108903D-04, &
   6.736190130891750D-06,  -3.547612149752866D-04,   3.694671725766910D-04, &
   1.024771532837703D-03,   7.751609674087477D-04,   4.875553877588601D-04, &
  -4.001777433634181D-03,  -6.410782878077352D-03,   6.914004738860328D-03, &
   1.276127120253351D-02,  -9.264478479709818D-03,  -1.069829054343619D-02, &
   9.499264625393663D-03,   2.799550201526625D-03,  -5.857038883413334D-03, &
   1.319451649211171D-03,   1.841923293008852D-03,  -1.097269976499565D-03, &
  -2.840999908552629D-04,   4.336523772397966D-04,   3.642625809615716D-05, &
  -2.036634866108824D-04,   6.736190130888279D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.203609509548188D-04,   0.000000000000000D+00, &
  -2.640961746023380D-04,   0.000000000000000D+00,   3.212905800349868D-04, &
   0.000000000000000D+00,   2.260219488061348D-04,   0.000000000000000D+00, &
  -2.554246997890453D-04,   0.000000000000000D+00,  -1.925155755865902D-04, &
   0.000000000000000D+00,   1.922986324448405D-04,   0.000000000000000D+00, &
   6.943428762245690D-05,   0.000000000000000D+00,  -7.606488083597397D-05, &
   0.000000000000000D+00,   7.001964077018367D-06,   0.000000000000000D+00, &
   2.756638786050553D-05,   1.322958946797957D-05,   0.000000000000000D+00, &
  -7.462115627750311D-05,   0.000000000000000D+00,   9.858025531400142D-06, &
  -2.669508280963380D-05,   0.000000000000000D+00,  -1.583115456954464D-05, &
   0.000000000000000D+00,   1.779228898036949D-04,   0.000000000000000D+00, &
   4.641202880366938D-06,   0.000000000000000D+00,  -1.884043915535782D-05, &
   1.787639150507634D-05,   0.000000000000000D+00,   8.001775115871900D-05, &
   0.000000000000000D+00,   9.969629779349288D-06,   0.000000000000000D+00, &
  -2.609766329227228D-04,   0.000000000000000D+00,  -1.170001696864328D-04, &
   0.000000000000000D+00,   4.296648275007987D-05,   0.000000000000000D+00, &
   1.630466609087127D-05,   2.236436425930256D-05,   0.000000000000000D+00, &
  -8.150452765727382D-05,   0.000000000000000D+00,  -1.398318844491571D-04, &
   0.000000000000000D+00,  -6.854209053255855D-06,   0.000000000000000D+00, &
   2.366974831303813D-04,   0.000000000000000D+00,   2.627613742320139D-04, &
   0.000000000000000D+00,   1.116643150146526D-04,   0.000000000000000D+00, &
  -5.795106246179280D-05,   0.000000000000000D+00,  -1.716762869855228D-05, &
  -5.364877740029171D-05,   0.000000000000000D+00,  -9.163289635089349D-05, &
   0.000000000000000D+00,   1.658586449109706D-04,   0.000000000000000D+00, &
   1.572457109653952D-04,   0.000000000000000D+00,   1.902765061208607D-05, &
   0.000000000000000D+00,  -1.285123993429288D-04,   0.000000000000000D+00, &
  -1.476084171433479D-04,   0.000000000000000D+00,  -3.756236851152707D-04, &
   0.000000000000000D+00,  -1.054862833677366D-04,   0.000000000000000D+00, &
   2.988698618910063D-05,   0.000000000000000D+00,   5.555474358055262D-06, &
   4.466287327952552D-05,   0.000000000000000D+00,   1.149341691386843D-04, &
   0.000000000000000D+00,   1.821998255254736D-04,   0.000000000000000D+00, &
  -1.915374350166570D-04,   0.000000000000000D+00,  -1.213814532000455D-04, &
   0.000000000000000D+00,  -2.296812070071748D-05,   0.000000000000000D+00, &
   3.675817237447759D-05,   0.000000000000000D+00,   2.987799202163210D-06, &
   0.000000000000000D+00,   1.344140876422262D-04,   0.000000000000000D+00, &
   4.023161113506701D-04,   0.000000000000000D+00,   9.681345961855054D-05, &
   0.000000000000000D+00,   8.188755649439822D-06,   0.000000000000000D+00, &
   1.576780341270433D-05,  -1.777743177580647D-05,   0.000000000000000D+00, &
  -1.595194632310965D-05,   0.000000000000000D+00,  -1.876920174504279D-04, &
   0.000000000000000D+00,  -2.082301758175900D-04,   0.000000000000000D+00, &
   1.308091110388809D-04,   0.000000000000000D+00,   5.768666766168368D-05, &
   0.000000000000000D+00,   6.816150018845674D-06,   0.000000000000000D+00, &
  -3.132292129842920D-06,   0.000000000000000D+00,  -2.499155536815333D-07, &
   0.000000000000000D+00,   7.507521522235252D-05,   0.000000000000000D+00, &
  -1.216804704551154D-04,   0.000000000000000D+00,  -2.952979883279126D-04, &
   0.000000000000000D+00,  -7.259342262414246D-05,   0.000000000000000D+00, &
  -2.071085374288215D-05,   0.000000000000000D+00,  -1.750590336714499D-05, &
   2.779378991774040D-06,   0.000000000000000D+00,  -2.929074916436719D-06, &
   0.000000000000000D+00,  -6.585192858776305D-06,   0.000000000000000D+00, &
   2.021548173792724D-04,   0.000000000000000D+00,   1.417768156137390D-04, &
   0.000000000000000D+00,  -4.818482794943218D-05,   0.000000000000000D+00, &
  -8.247078084733807D-06,   0.000000000000000D+00,   6.664549104456052D-06, &
   0.000000000000000D+00,  -4.934864037071268D-07,   0.000000000000000D+00, &
   2.870340808410668D-06,   0.000000000000000D+00,  -7.787831228089667D-06, &
   0.000000000000000D+00,  -9.636811770735413D-05,   0.000000000000000D+00, &
   1.076899368698198D-04,   0.000000000000000D+00,   1.454222719463021D-04, &
   0.000000000000000D+00,   3.715043158846705D-05,   0.000000000000000D+00, &
   1.121198044591219D-05,   0.000000000000000D+00,   1.670767416618773D-07, &
   2.804671989572912D-08,   0.000000000000000D+00,   4.078945432415283D-07, &
   0.000000000000000D+00,   5.916269507328871D-06,   0.000000000000000D+00, &
   1.453076605546590D-05,   0.000000000000000D+00,  -1.505536507286103D-04, &
   0.000000000000000D+00,  -5.195962116696768D-05,   0.000000000000000D+00, &
   4.576568785312897D-06,   0.000000000000000D+00,  -8.334489943865829D-06, &
   0.000000000000000D+00,  -5.359679215840410D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.675202083265739D-07, &
   0.000000000000000D+00,   2.851098739653252D-07,   0.000000000000000D+00, &
   5.195291344452029D-06,   0.000000000000000D+00,   5.550571810417385D-05, &
   0.000000000000000D+00,  -8.029276775973195D-05,   0.000000000000000D+00, &
  -4.902652579407962D-05,   0.000000000000000D+00,  -1.208174871206213D-05, &
   0.000000000000000D+00,  -7.890006509359904D-07,   0.000000000000000D+00, &
   9.574273700939097D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.271313294494896D-07,   0.000000000000000D+00,  -2.023069296913810D-06, &
   0.000000000000000D+00,  -1.319483390770139D-06,   0.000000000000000D+00, &
   3.207454500370191D-06,   0.000000000000000D+00,   7.070687944454378D-05, &
   0.000000000000000D+00,   4.688992400504394D-06,   0.000000000000000D+00, &
   4.231725108692808D-06,   0.000000000000000D+00,   5.940032160547950D-06, &
   0.000000000000000D+00,   1.323287252660904D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.769992721955712D-08, &
   0.000000000000000D+00,   4.041036745748266D-07,   0.000000000000000D+00, &
  -2.529740725476107D-06,   0.000000000000000D+00,  -7.205413842529359D-06, &
   0.000000000000000D+00,  -5.173454620708117D-07,   0.000000000000000D+00, &
   4.220800440264300D-05,   0.000000000000000D+00,   1.338917824819949D-05, &
   0.000000000000000D+00,   2.944163428456075D-06,   0.000000000000000D+00, &
  -2.927622901813304D-06,   0.000000000000000D+00,  -6.043534688275249D-06 /)
lem(:,155) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.873189453813615D-04,  -1.155144386176540D-03, &
  -2.138337375643392D-03,  -8.156709373967155D-04,   4.310031237394137D-03, &
   5.539228747633529D-03,  -5.944538786578850D-03,  -8.860966380679096D-03, &
   7.935808592382194D-03,   5.427188459062213D-03,  -7.389024185546348D-03, &
   7.366573170831846D-04,   3.196117043723816D-03,  -2.503867332881916D-03, &
   1.769115345859336D-05,   1.138213501685790D-03,  -5.215957545994356D-04, &
  -2.356656711521380D-04,   2.645950600432238D-04,   3.276464388077639D-05, &
  -1.292383637783930D-04,   0.000000000000000D+00,  -3.873189453813655D-04, &
   1.155144386176542D-03,   2.138337375643387D-03,   8.156709373967142D-04, &
  -4.310031237394138D-03,  -5.539228747633520D-03,   5.944538786578847D-03, &
   8.860966380679092D-03,  -7.935808592382212D-03,  -5.427188459062197D-03, &
   7.389024185546339D-03,  -7.366573170831840D-04,  -3.196117043723812D-03, &
   2.503867332881919D-03,  -1.769115345859743D-05,  -1.138213501685797D-03, &
   5.215957545994443D-04,   2.356656711521393D-04,  -2.645950600432369D-04, &
  -3.276464388076779D-05,   1.292383637784042D-04,   0.000000000000000D+00, &
  -1.936594726906823D-04,   0.000000000000000D+00,   2.708050376262268D-04, &
   0.000000000000000D+00,   9.104537287744188D-04,   0.000000000000000D+00, &
  -1.111628907297414D-03,   0.000000000000000D+00,   1.535888529744379D-05, &
   0.000000000000000D+00,   6.694857398902498D-04,   0.000000000000000D+00, &
  -1.803285880559766D-04,   0.000000000000000D+00,  -2.609219038902108D-04, &
   0.000000000000000D+00,   1.139366341253233D-04,   0.000000000000000D+00, &
   2.149094633463822D-05,   0.000000000000000D+00,  -1.661276592988756D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.531012560424333D-05, &
   0.000000000000000D+00,  -2.123958199809596D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.143503497992655D-05,   0.000000000000000D+00, &
  -3.019802265158303D-05,   0.000000000000000D+00,   5.254033891804797D-05, &
   0.000000000000000D+00,   1.026691961289485D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.440830728924680D-05,   0.000000000000000D+00, &
   6.554566640042454D-05,   0.000000000000000D+00,   3.055745670467118D-05, &
   0.000000000000000D+00,  -7.922086158558081D-05,   0.000000000000000D+00, &
  -8.243643900781488D-05,   0.000000000000000D+00,   2.439065773790754D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.760212231843670D-05, &
   0.000000000000000D+00,  -6.256175165601050D-05,   0.000000000000000D+00, &
  -1.216421610206734D-04,   0.000000000000000D+00,  -2.869331097232317D-05, &
   0.000000000000000D+00,   6.787405802246298D-05,   0.000000000000000D+00, &
   1.804397236648307D-04,   0.000000000000000D+00,   9.736063086688590D-05, &
   0.000000000000000D+00,  -4.416973046141666D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.165442649760864D-05,   0.000000000000000D+00, &
  -7.127379050333017D-05,   0.000000000000000D+00,   1.345404445894484D-04, &
   0.000000000000000D+00,   1.466177464950226D-04,   0.000000000000000D+00, &
   4.550834841765431D-05,   0.000000000000000D+00,  -2.770100108234454D-05, &
   0.000000000000000D+00,  -9.883020727347027D-05,   0.000000000000000D+00, &
  -3.001202127559596D-04,   0.000000000000000D+00,  -9.945245424553712D-05, &
   0.000000000000000D+00,   1.961610303276636D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.299682319112247D-05,   0.000000000000000D+00, &
   9.804147880233662D-05,   0.000000000000000D+00,   1.471681246976150D-04, &
   0.000000000000000D+00,  -1.636674450103504D-04,   0.000000000000000D+00, &
  -1.164829188568651D-04,   0.000000000000000D+00,  -5.167700436190971D-05, &
   0.000000000000000D+00,  -3.287558203986230D-07,   0.000000000000000D+00, &
  -1.991974573069760D-06,   0.000000000000000D+00,   1.082781125951814D-04, &
   0.000000000000000D+00,   3.408073568607780D-04,   0.000000000000000D+00, &
   9.282429868468193D-05,   0.000000000000000D+00,   1.747694842112098D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.143842076240588D-05, &
   0.000000000000000D+00,  -2.789304030294763D-05,   0.000000000000000D+00, &
  -1.557504773501397D-04,   0.000000000000000D+00,  -1.740140859392810D-04, &
   0.000000000000000D+00,   1.150462125877129D-04,   0.000000000000000D+00, &
   5.533644669235110D-05,   0.000000000000000D+00,   1.963522084423207D-05, &
   0.000000000000000D+00,   4.257678236419092D-06,   0.000000000000000D+00, &
   3.742300310804665D-06,   0.000000000000000D+00,   5.420776223777913D-05, &
   0.000000000000000D+00,  -1.053450275603329D-04,   0.000000000000000D+00, &
  -2.572443068824512D-04,   0.000000000000000D+00,  -7.027994851623638D-05, &
   0.000000000000000D+00,  -2.347164263254871D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.792995281093221D-07,   0.000000000000000D+00, &
   6.055723353769041D-06,   0.000000000000000D+00,   4.168433784289652D-06, &
   0.000000000000000D+00,   1.710326377937255D-04,   0.000000000000000D+00, &
   1.202246033850720D-04,   0.000000000000000D+00,  -4.187999851068140D-05, &
   0.000000000000000D+00,  -8.025590429428634D-06,   0.000000000000000D+00, &
   1.664158145063625D-05,   0.000000000000000D+00,  -9.896471270016868D-07, &
   0.000000000000000D+00,  -3.243252853655058D-07,   0.000000000000000D+00, &
  -3.006744345120450D-06,   0.000000000000000D+00,  -7.413242801174466D-05, &
   0.000000000000000D+00,   9.626497960193815D-05,   0.000000000000000D+00, &
   1.281272878793174D-04,   0.000000000000000D+00,   3.651262591746097D-05, &
   0.000000000000000D+00,   7.817729678676349D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.715312692402316D-07,   0.000000000000000D+00, &
  -1.778196285545568D-06,   0.000000000000000D+00,   2.315669846884841D-07, &
   0.000000000000000D+00,   7.264983871413899D-06,   0.000000000000000D+00, &
  -1.296374993924570D-04,   0.000000000000000D+00,  -4.408537227763391D-05, &
   0.000000000000000D+00,   2.934689306473592D-06,   0.000000000000000D+00, &
  -8.919066991481026D-06,   0.000000000000000D+00,  -2.069641460048096D-05, &
   0.000000000000000D+00,  -3.265332400426220D-08,   0.000000000000000D+00, &
   3.294861711958345D-07,   0.000000000000000D+00,  -2.299635590798432D-06, &
   0.000000000000000D+00,   2.172498850711521D-06,   0.000000000000000D+00, &
   4.255676658233859D-05,   0.000000000000000D+00,  -7.200405318426007D-05, &
   0.000000000000000D+00,  -4.376007965930402D-05,   0.000000000000000D+00, &
  -1.115288743586457D-05,   0.000000000000000D+00,   2.963457273658171D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.485347521953355D-09, &
   0.000000000000000D+00,   3.053789390070406D-07,   0.000000000000000D+00, &
   6.422325375468891D-07,   0.000000000000000D+00,  -1.800055428157534D-06, &
   0.000000000000000D+00,   5.489763718957549D-06,   0.000000000000000D+00, &
   6.103964919596442D-05,   0.000000000000000D+00,   4.145797725124651D-06, &
   0.000000000000000D+00,   3.917343298557616D-06,   0.000000000000000D+00, &
   8.492046327553492D-06,   0.000000000000000D+00,   6.502993502112490D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.625303050353998D-07,   0.000000000000000D+00,   8.941497882027376D-07, &
   0.000000000000000D+00,   1.189308500340803D-06,   0.000000000000000D+00, &
  -5.700815069918568D-06,   0.000000000000000D+00,   2.711479773285586D-06, &
   0.000000000000000D+00,   3.760643035360820D-05,   0.000000000000000D+00, &
   1.247377053742862D-05,   0.000000000000000D+00,   1.995914755419505D-06, &
   0.000000000000000D+00,   3.429847679729593D-07,   0.000000000000000D+00 /)
lem(:,156) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.033304097670920D-04,   4.404441005952622D-04,   1.383936740994508D-03, &
   1.767143122204593D-03,   3.561398524780272D-04,  -3.541960402630527D-03, &
  -4.482720728807337D-03,   4.017090323726856D-03,   5.992312658619482D-03, &
  -5.030777569569812D-03,  -2.021238782636585D-03,   4.148163933737434D-03, &
  -1.995760948207897D-03,  -7.155115867757468D-04,   1.737007107701765D-03, &
  -1.062029701129524D-03,  -1.363854612475654D-04,   6.333909442726096D-04, &
  -2.882536384478960D-04,  -1.388347845955964D-04,   1.710648462513204D-04, &
   8.509296208883682D-06,   1.033304097670960D-04,   4.404441005952659D-04, &
   1.383936740994508D-03,   1.767143122204594D-03,   3.561398524780270D-04, &
  -3.541960402630520D-03,  -4.482720728807329D-03,   4.017090323726860D-03, &
   5.992312658619473D-03,  -5.030777569569811D-03,  -2.021238782636583D-03, &
   4.148163933737431D-03,  -1.995760948207893D-03,  -7.155115867757473D-04, &
   1.737007107701768D-03,  -1.062029701129523D-03,  -1.363854612475654D-04, &
   6.333909442726046D-04,  -2.882536384478936D-04,  -1.388347845955901D-04, &
   1.710648462513162D-04,   8.509296208879428D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.100151528820601D-04,   0.000000000000000D+00, &
   9.744937283807961D-04,   0.000000000000000D+00,  -2.992480239538585D-03, &
   0.000000000000000D+00,   3.192136010017871D-03,   0.000000000000000D+00, &
  -1.515333851910379D-03,   0.000000000000000D+00,  -3.840436676643128D-04, &
   0.000000000000000D+00,   6.925592258541681D-04,   0.000000000000000D+00, &
  -8.495116358783548D-05,   0.000000000000000D+00,  -1.084146172416721D-04, &
   0.000000000000000D+00,   3.826320526125638D-05,   0.000000000000000D+00, &
   6.363632282749649D-07,  -2.117208180965923D-05,   0.000000000000000D+00, &
  -2.661543392967551D-05,   0.000000000000000D+00,  -4.212919554323382D-05, &
   1.259758902947763D-05,   0.000000000000000D+00,   2.678249521806108D-07, &
   0.000000000000000D+00,   4.663526940036241D-05,   0.000000000000000D+00, &
   4.554515925573397D-05,   0.000000000000000D+00,   3.070898923380540D-05, &
   1.671766761491148D-05,   0.000000000000000D+00,   6.195621102678440D-07, &
   0.000000000000000D+00,   4.977517868432065D-05,   0.000000000000000D+00, &
  -5.285026041415675D-05,   0.000000000000000D+00,  -6.323156464152705D-05, &
   0.000000000000000D+00,  -7.675893384494710D-05,   0.000000000000000D+00, &
   4.029200023824774D-05,  -3.587181761923500D-05,   0.000000000000000D+00, &
   5.491442757719766D-06,   0.000000000000000D+00,  -4.604858324563727D-05, &
   0.000000000000000D+00,  -1.222217490003373D-04,   0.000000000000000D+00, &
   4.627535821274048D-05,   0.000000000000000D+00,   5.617414518685088D-05, &
   0.000000000000000D+00,   1.494753124822028D-04,   0.000000000000000D+00, &
   9.582865648033241D-05,   0.000000000000000D+00,  -8.549806635544010D-05, &
   2.579566038119519D-05,   0.000000000000000D+00,  -1.188133786760143D-05, &
   0.000000000000000D+00,  -5.342157758403950D-05,   0.000000000000000D+00, &
   1.183437083977499D-04,   0.000000000000000D+00,   1.603230249219410D-04, &
   0.000000000000000D+00,  -2.836821317395786D-05,   0.000000000000000D+00, &
  -2.653712623063100D-05,   0.000000000000000D+00,  -8.236452885116485D-05, &
   0.000000000000000D+00,  -2.637057042944167D-04,   0.000000000000000D+00, &
  -1.027977146684439D-04,   0.000000000000000D+00,   4.279332052106455D-05, &
  -6.896161986548859D-06,   0.000000000000000D+00,   7.758461733163296D-06, &
   0.000000000000000D+00,   7.091814565215421D-05,   0.000000000000000D+00, &
   1.218654176663793D-04,   0.000000000000000D+00,  -1.508997096443502D-04, &
   0.000000000000000D+00,  -1.261146873915508D-04,   0.000000000000000D+00, &
   1.005805259130413D-05,   0.000000000000000D+00,   4.179761786007961D-06, &
   0.000000000000000D+00,  -7.891749841256007D-07,   0.000000000000000D+00, &
   9.681345961855054D-05,   0.000000000000000D+00,   3.102921718049837D-04, &
   0.000000000000000D+00,   9.818567839302858D-05,   0.000000000000000D+00, &
   3.905095629555749D-05,  -6.267391456894909D-07,   0.000000000000000D+00, &
  -3.463483869684455D-07,   0.000000000000000D+00,  -1.780083350546280D-05, &
   0.000000000000000D+00,  -1.289374588206117D-04,   0.000000000000000D+00, &
  -1.506923099251123D-04,   0.000000000000000D+00,   1.075899444781548D-04, &
   0.000000000000000D+00,   5.053722015778617D-05,   0.000000000000000D+00, &
  -1.399880563743002D-06,   0.000000000000000D+00,   7.013762913260404D-07, &
   0.000000000000000D+00,   6.764071832690063D-06,   0.000000000000000D+00, &
   4.308874643146215D-05,   0.000000000000000D+00,  -9.799621754831182D-05, &
   0.000000000000000D+00,  -2.403373764556966D-04,   0.000000000000000D+00, &
  -7.517676487908392D-05,   0.000000000000000D+00,  -6.807301281213363D-05, &
   4.784703355404652D-07,   0.000000000000000D+00,  -1.433220858866506D-06, &
   0.000000000000000D+00,   1.140559954965409D-06,   0.000000000000000D+00, &
   6.247854483515524D-06,   0.000000000000000D+00,   1.486029744436217D-04, &
   0.000000000000000D+00,   1.071673989955909D-04,   0.000000000000000D+00, &
  -3.805929385089781D-05,   0.000000000000000D+00,   3.587237946032867D-07, &
   0.000000000000000D+00,  -6.141285444363537D-08,   0.000000000000000D+00, &
   1.650824938521712D-07,   0.000000000000000D+00,  -2.774799133052872D-06, &
   0.000000000000000D+00,  -7.095055076554724D-06,   0.000000000000000D+00, &
  -6.077405222581878D-05,   0.000000000000000D+00,   9.079881150681305D-05, &
   0.000000000000000D+00,   1.222227725848170D-04,   0.000000000000000D+00, &
   4.251096049027574D-05,   0.000000000000000D+00,   3.480851342643082D-05, &
   2.845649339891154D-09,   0.000000000000000D+00,   3.459513258842746D-07, &
   0.000000000000000D+00,   6.257589999600995D-07,   0.000000000000000D+00, &
   1.772748936531770D-08,   0.000000000000000D+00,   1.532396437671412D-06, &
   0.000000000000000D+00,  -1.157844890036444D-04,   0.000000000000000D+00, &
  -3.963893556576631D-05,   0.000000000000000D+00,   2.659663060761089D-07, &
   0.000000000000000D+00,  -7.360654648076430D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.703463789555830D-07, &
   0.000000000000000D+00,   7.960596493383984D-07,   0.000000000000000D+00, &
   8.613822402369028D-07,   0.000000000000000D+00,   5.188819823863466D-06, &
   0.000000000000000D+00,   3.474549507112610D-05,   0.000000000000000D+00, &
  -6.868408465981141D-05,   0.000000000000000D+00,  -4.257670354636689D-05, &
   0.000000000000000D+00,  -1.714145405379966D-05,   0.000000000000000D+00, &
   8.720856938024546D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.419263868760322D-08,   0.000000000000000D+00,  -3.769290907215429D-07, &
   0.000000000000000D+00,   8.013474764512626D-07,   0.000000000000000D+00, &
  -6.181665623878282D-07,   0.000000000000000D+00,   8.714616585722216D-06, &
   0.000000000000000D+00,   5.542887308077155D-05,   0.000000000000000D+00, &
   3.141361068189115D-06,   0.000000000000000D+00,   4.347000539569270D-06, &
   0.000000000000000D+00,   1.267635777395108D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.260191230558117D-09, &
   0.000000000000000D+00,  -8.029859567928667D-08,   0.000000000000000D+00, &
  -6.108834377726716D-07,   0.000000000000000D+00,   3.229479921033057D-07, &
   0.000000000000000D+00,  -6.786673892517790D-06,   0.000000000000000D+00, &
   4.644124048983329D-06,   0.000000000000000D+00,   3.621592290411081D-05, &
   0.000000000000000D+00,   1.317149383568732D-05,   0.000000000000000D+00, &
   3.057605538381726D-06,   0.000000000000000D+00,  -9.659653955060580D-06 /)
lem(:,157) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.627245600657771D-04,  -1.240349515520686D-03, &
  -1.318651846774303D-03,  -9.345735332967045D-04,   1.898337709982869D-03, &
   3.694560921564498D-03,  -1.852686210216763D-03,  -3.921094062244018D-03, &
   2.596851965252260D-03,   8.121888665959296D-04,  -1.692636732630153D-03, &
   1.377314299090151D-03,  -4.842133137048132D-04,  -4.995918877411871D-04, &
   8.484795855795065D-04,  -4.904109355050203D-04,  -8.359141839983097D-05, &
   3.344738236584817D-04,  -1.734336712086792D-04,  -6.367465468208970D-05, &
   1.071210230389203D-04,   0.000000000000000D+00,   8.627245600657784D-04, &
   1.240349515520685D-03,   1.318651846774301D-03,   9.345735332967051D-04, &
  -1.898337709982864D-03,  -3.694560921564497D-03,   1.852686210216768D-03, &
   3.921094062244020D-03,  -2.596851965252263D-03,  -8.121888665959352D-04, &
   1.692636732630153D-03,  -1.377314299090150D-03,   4.842133137048143D-04, &
   4.995918877411900D-04,  -8.484795855795098D-04,   4.904109355050256D-04, &
   8.359141839983302D-05,  -3.344738236584880D-04,   1.734336712086787D-04, &
   6.367465468209491D-05,  -1.071210230389202D-04,   0.000000000000000D+00, &
   4.313622800328893D-04,   0.000000000000000D+00,   6.888672342603349D-04, &
   0.000000000000000D+00,  -3.248160277554215D-03,   0.000000000000000D+00, &
   4.544663432601220D-03,   0.000000000000000D+00,  -4.564485053141020D-03, &
   0.000000000000000D+00,   3.617806701637277D-03,   0.000000000000000D+00, &
  -1.320158335068008D-03,   0.000000000000000D+00,  -3.093825203688300D-04, &
   0.000000000000000D+00,   3.260582332476007D-04,   0.000000000000000D+00, &
  -5.819443392782223D-05,   0.000000000000000D+00,   1.002975723462510D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.410218253968257D-05, &
   0.000000000000000D+00,  -1.169965977435014D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.235589047605301D-05,   0.000000000000000D+00, &
   9.898890328940613D-07,   0.000000000000000D+00,   1.368909191084117D-05, &
   0.000000000000000D+00,   3.353718448411753D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.908324397219673D-07,   0.000000000000000D+00, &
   1.483170266138680D-06,   0.000000000000000D+00,   1.168814701901885D-04, &
   0.000000000000000D+00,  -1.663772788230152D-05,   0.000000000000000D+00, &
  -4.802252930234292D-05,   0.000000000000000D+00,  -8.115307150273049D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.006163673750182D-06, &
   0.000000000000000D+00,   4.965837483725577D-06,   0.000000000000000D+00, &
  -4.956779539364475D-05,   0.000000000000000D+00,  -2.713203703747845D-04, &
   0.000000000000000D+00,   1.884707964240038D-05,   0.000000000000000D+00, &
   4.367905545275280D-05,   0.000000000000000D+00,   1.352832686495792D-04, &
   0.000000000000000D+00,   1.355853610355409D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.217540702145527D-07,   0.000000000000000D+00, &
  -1.284173575484446D-05,   0.000000000000000D+00,  -4.313455266001349D-05, &
   0.000000000000000D+00,   1.197311500340002D-04,   0.000000000000000D+00, &
   3.583265002909326D-04,   0.000000000000000D+00,  -1.304671610651563D-05, &
   0.000000000000000D+00,  -2.414513991952286D-05,   0.000000000000000D+00, &
  -7.473558948090766D-05,   0.000000000000000D+00,  -2.554493971097046D-04, &
   0.000000000000000D+00,  -1.664077421815216D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.031542816922392D-06,   0.000000000000000D+00, &
   1.093830001948445D-05,   0.000000000000000D+00,   6.003874345141958D-05, &
   0.000000000000000D+00,   1.102159814161934D-04,   0.000000000000000D+00, &
  -1.542332891845476D-04,   0.000000000000000D+00,  -3.109238640906486D-04, &
   0.000000000000000D+00,   4.018356769951121D-06,   0.000000000000000D+00, &
   7.675240731376841D-06,   0.000000000000000D+00,   5.385401295663206D-06, &
   0.000000000000000D+00,   9.282429868468193D-05,   0.000000000000000D+00, &
   3.139557709357772D-04,   0.000000000000000D+00,   1.539662266089801D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.824458347905232D-06, &
   0.000000000000000D+00,  -4.132350453082394D-06,   0.000000000000000D+00, &
  -1.770114199142554D-05,   0.000000000000000D+00,  -1.152818782428697D-04, &
   0.000000000000000D+00,  -1.437255972181094D-04,   0.000000000000000D+00, &
   1.169589671398000D-04,   0.000000000000000D+00,   1.595295769868729D-04, &
   0.000000000000000D+00,   7.238531538387497D-08,   0.000000000000000D+00, &
  -1.807898631347241D-06,   0.000000000000000D+00,   2.663068377727478D-06, &
   0.000000000000000D+00,   3.242836598611328D-05,   0.000000000000000D+00, &
  -9.849532117357512D-05,   0.000000000000000D+00,  -2.520181297924083D-04, &
   0.000000000000000D+00,  -9.663354155777742D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.095254286555481D-07,   0.000000000000000D+00, &
   7.914705890549162D-07,   0.000000000000000D+00,   1.234763675448716D-06, &
   0.000000000000000D+00,   9.199860760341536D-06,   0.000000000000000D+00, &
   1.396442114964970D-04,   0.000000000000000D+00,   1.044138018219134D-04, &
   0.000000000000000D+00,  -5.066258155920595D-05,   0.000000000000000D+00, &
  -1.671857842976333D-05,   0.000000000000000D+00,  -2.182576611669044D-07, &
   0.000000000000000D+00,   6.316546321622480D-07,   0.000000000000000D+00, &
  -2.610452430597069D-07,   0.000000000000000D+00,  -6.667123032080223D-06, &
   0.000000000000000D+00,  -5.093820055768604D-05,   0.000000000000000D+00, &
   9.415281496123268D-05,   0.000000000000000D+00,   1.315525984352056D-04, &
   0.000000000000000D+00,   3.333571943556243D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.190326261886884D-08,   0.000000000000000D+00, &
  -2.913365068703193D-07,   0.000000000000000D+00,   8.447071007038026D-07, &
   0.000000000000000D+00,   1.412765026244339D-06,   0.000000000000000D+00, &
  -3.704578360347210D-06,   0.000000000000000D+00,  -1.118897334456108D-04, &
   0.000000000000000D+00,  -3.738631819508166D-05,   0.000000000000000D+00, &
   6.930303954483818D-06,   0.000000000000000D+00,  -3.741243449238278D-05, &
   0.000000000000000D+00,  -3.967648222743598D-09,   0.000000000000000D+00, &
  -9.802559473087801D-08,   0.000000000000000D+00,  -4.655087638180170D-07, &
   0.000000000000000D+00,   1.368069074859417D-07,   0.000000000000000D+00, &
   7.501542171233522D-06,   0.000000000000000D+00,   2.792076819872264D-05, &
   0.000000000000000D+00,  -7.195711342923363D-05,   0.000000000000000D+00, &
  -4.317140282013910D-05,   0.000000000000000D+00,  -5.264011783803277D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.503030230718491D-10, &
   0.000000000000000D+00,   9.629764731165878D-08,   0.000000000000000D+00, &
  -2.816124483818635D-07,   0.000000000000000D+00,  -4.843276329354191D-07, &
   0.000000000000000D+00,  -2.232394889233928D-06,   0.000000000000000D+00, &
   1.314455698077465D-05,   0.000000000000000D+00,   5.349361651941627D-05, &
   0.000000000000000D+00,   1.018701268717415D-07,   0.000000000000000D+00, &
   8.994252686387397D-06,   0.000000000000000D+00,   2.739503165316098D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.850672797699199D-08,   0.000000000000000D+00,   2.483882407801735D-07, &
   0.000000000000000D+00,  -1.992309504049048D-07,   0.000000000000000D+00, &
   1.161717145466750D-07,   0.000000000000000D+00,  -8.523667781050933D-06, &
   0.000000000000000D+00,   8.486109857450336D-06,   0.000000000000000D+00, &
   3.762022095529544D-05,   0.000000000000000D+00,   8.896561812510026D-06, &
   0.000000000000000D+00,   4.099732225133031D-06,   0.000000000000000D+00 /)
lem(:,158) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.822537010999410D-04,  -2.941106454769573D-05,  -3.334965583759647D-05, &
   1.530354798940750D-03,   1.782788368344066D-03,  -9.313616563581059D-04, &
  -3.131363610899795D-03,   1.736126856951918D-04,   2.676178693292080D-03, &
  -9.277608362442878D-04,  -5.624106642522879D-04,   6.514780827441563D-04, &
  -5.454763171305186D-04,   4.440687554588018D-04,  -1.141590562027989D-04, &
  -2.666753336848037D-04,   3.987511607028012D-04,  -2.324087477538754D-04, &
  -3.664954577834870D-05,   1.679921203589823D-04,  -1.013160016621955D-04, &
  -2.311043608415488D-05,   1.822537010999347D-04,  -2.941106454769897D-05, &
  -3.334965583760009D-05,   1.530354798940751D-03,   1.782788368344066D-03, &
  -9.313616563581033D-04,  -3.131363610899787D-03,   1.736126856951908D-04, &
   2.676178693292086D-03,  -9.277608362442931D-04,  -5.624106642522930D-04, &
   6.514780827441634D-04,  -5.454763171305164D-04,   4.440687554588005D-04, &
  -1.141590562027957D-04,  -2.666753336848060D-04,   3.987511607027992D-04, &
  -2.324087477538747D-04,  -3.664954577834738D-05,   1.679921203589836D-04, &
  -1.013160016621941D-04,  -2.311043608415844D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.059958101968212D-04,   0.000000000000000D+00, &
  -1.976258736473739D-03,   0.000000000000000D+00,   2.098510974068149D-03, &
   0.000000000000000D+00,   6.460477016156145D-04,   0.000000000000000D+00, &
  -1.075609106549040D-03,   0.000000000000000D+00,  -1.701387213182977D-03, &
   0.000000000000000D+00,   2.323279068837901D-03,   0.000000000000000D+00, &
  -7.657387831401805D-04,   0.000000000000000D+00,   3.892992159244525D-06, &
   0.000000000000000D+00,   2.244813491558906D-05,   0.000000000000000D+00, &
  -1.162574404761897D-05,  -1.007564484126985D-05,   0.000000000000000D+00, &
  -4.650297619047532D-07,   0.000000000000000D+00,  -2.046130952380960D-05, &
   4.522367198933839D-07,   0.000000000000000D+00,  -1.076199557236670D-06, &
   0.000000000000000D+00,   1.233108969418981D-06,   0.000000000000000D+00, &
   1.621337858155773D-05,   0.000000000000000D+00,   1.024226181027955D-04, &
   1.410237723508562D-05,   0.000000000000000D+00,   6.308437155613387D-06, &
   0.000000000000000D+00,  -7.687089363351003D-07,   0.000000000000000D+00, &
   3.204547151568502D-06,   0.000000000000000D+00,  -1.660312599836974D-05, &
   0.000000000000000D+00,  -5.703837911974025D-05,   0.000000000000000D+00, &
  -2.088502572758082D-04,  -1.918436423495765D-05,   0.000000000000000D+00, &
  -9.217905255340852D-06,   0.000000000000000D+00,   3.181466710853463D-06, &
   0.000000000000000D+00,  -1.225248889300615D-05,   0.000000000000000D+00, &
  -2.948697108882375D-06,   0.000000000000000D+00,   1.286202457551650D-05, &
   0.000000000000000D+00,   4.206322096328613D-05,   0.000000000000000D+00, &
   1.455861440703622D-04,   0.000000000000000D+00,   2.707870433208345D-04, &
   1.331259905791523D-05,   0.000000000000000D+00,   5.215304633652438D-06, &
   0.000000000000000D+00,  -7.407803242058977D-06,   0.000000000000000D+00, &
  -4.775605331308692D-05,   0.000000000000000D+00,   6.444754768898557D-05, &
   0.000000000000000D+00,  -5.135907735203452D-07,   0.000000000000000D+00, &
  -6.234692438568264D-06,   0.000000000000000D+00,  -2.254240913236835D-05, &
   0.000000000000000D+00,  -7.298786530921059D-05,   0.000000000000000D+00, &
  -2.680201660626502D-04,   0.000000000000000D+00,  -2.961215179321848D-04, &
  -4.818880198288938D-06,   0.000000000000000D+00,  -3.548592652417728D-07, &
   0.000000000000000D+00,   8.130047213091287D-06,   0.000000000000000D+00, &
   5.405069434227583D-05,   0.000000000000000D+00,   1.167159037084532D-04, &
   0.000000000000000D+00,  -9.654694889499048D-05,   0.000000000000000D+00, &
   1.324439319037129D-06,   0.000000000000000D+00,   8.735921714181594D-07, &
   0.000000000000000D+00,   6.183114204489701D-06,   0.000000000000000D+00, &
   8.188755649439822D-06,   0.000000000000000D+00,   9.818567839302858D-05, &
   0.000000000000000D+00,   3.331229644107724D-04,   0.000000000000000D+00, &
   3.152711735637539D-04,   5.584377696523460D-07,   0.000000000000000D+00, &
  -4.713360004408447D-07,   0.000000000000000D+00,  -4.392305283795385D-06, &
   0.000000000000000D+00,  -1.771644645551904D-05,   0.000000000000000D+00, &
  -1.141619893612672D-04,   0.000000000000000D+00,  -1.523069060714184D-04, &
   0.000000000000000D+00,   5.414492912571831D-05,   0.000000000000000D+00, &
  -3.702242519679226D-07,   0.000000000000000D+00,   5.857609102691576D-07, &
   0.000000000000000D+00,  -2.876198707155615D-08,   0.000000000000000D+00, &
   2.919795739225541D-06,   0.000000000000000D+00,   2.605799741837081D-05, &
   0.000000000000000D+00,  -1.104875484610037D-04,   0.000000000000000D+00, &
  -2.767350254446306D-04,   0.000000000000000D+00,  -2.779494461735579D-04, &
   8.085820534789738D-08,   0.000000000000000D+00,  -1.333811696671857D-07, &
   0.000000000000000D+00,   1.253760260489356D-06,   0.000000000000000D+00, &
   2.055373944010563D-06,   0.000000000000000D+00,   1.496834962015458D-05, &
   0.000000000000000D+00,   1.441051210714556D-04,   0.000000000000000D+00, &
   1.172457630190536D-04,   0.000000000000000D+00,   1.339666859539219D-06, &
   0.000000000000000D+00,  -8.103739756374472D-09,   0.000000000000000D+00, &
  -1.520541826449241D-07,   0.000000000000000D+00,  -4.300717602197040D-07, &
   0.000000000000000D+00,  -6.051198765448920D-07,   0.000000000000000D+00, &
  -9.566556391863495D-06,   0.000000000000000D+00,  -4.393485030416834D-05, &
   0.000000000000000D+00,   1.057144386116536D-04,   0.000000000000000D+00, &
   1.618822761100893D-04,   0.000000000000000D+00,   1.650302938793897D-04, &
   3.151353384654175D-10,   0.000000000000000D+00,   1.056277228423076D-07, &
   0.000000000000000D+00,  -2.621003502405827D-07,   0.000000000000000D+00, &
   5.341204836205378D-09,   0.000000000000000D+00,   1.430968461028531D-06, &
   0.000000000000000D+00,  -1.194646067913975D-05,   0.000000000000000D+00, &
  -1.172932108680935D-04,   0.000000000000000D+00,  -5.190884977364252D-05, &
   0.000000000000000D+00,  -7.484177833051193D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.916556225999307D-08, &
   0.000000000000000D+00,   1.772905533249782D-07,   0.000000000000000D+00, &
  -3.513329389762476D-07,   0.000000000000000D+00,   3.631130943761550D-07, &
   0.000000000000000D+00,   1.117269538626056D-05,   0.000000000000000D+00, &
   2.206557633069135D-05,   0.000000000000000D+00,  -8.021696174304484D-05, &
   0.000000000000000D+00,  -7.496843227807915D-05,   0.000000000000000D+00, &
  -5.658128997543570D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.184329135053695D-09,   0.000000000000000D+00,   2.095572728773918D-08, &
   0.000000000000000D+00,   2.200433210789462D-07,   0.000000000000000D+00, &
  -3.910214088409301D-07,   0.000000000000000D+00,  -2.318005435852889D-06, &
   0.000000000000000D+00,   1.946541153186401D-05,   0.000000000000000D+00, &
   5.668744131689977D-05,   0.000000000000000D+00,   7.456437637109811D-06, &
   0.000000000000000D+00,  -6.628138316261676D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.086737820454430D-10, &
   0.000000000000000D+00,  -4.637308584985502D-08,   0.000000000000000D+00, &
   1.158196746397801D-07,   0.000000000000000D+00,   1.763574818556499D-07, &
   0.000000000000000D+00,   1.097038488508894D-08,   0.000000000000000D+00, &
  -1.037889507616426D-05,   0.000000000000000D+00,   1.285755257664804D-05, &
   0.000000000000000D+00,   4.576248102849166D-05,   0.000000000000000D+00, &
   2.800190375634929D-05,   0.000000000000000D+00,   7.547535571339693D-06 /)
lem(:,159) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.548958927467073D-04,  -1.080238852130842D-03, &
  -1.141463606658717D-03,  -4.478695603998043D-04,  -3.695113502301330D-04, &
   1.760315950948643D-03,   1.194714589778794D-03,  -1.926919316434951D-03, &
   4.508290222609537D-05,   6.035575148541428D-04,  -3.593693580412753D-04, &
   2.155290483385406D-04,  -2.013245547627626D-04,   1.462483087358985D-04, &
   2.565528287961512D-06,  -1.540577641021281D-04,   1.986832507234165D-04, &
  -1.117486014974432D-04,  -2.348841350073897D-05,   9.059117673796147D-05, &
  -5.658179716922749D-05,   0.000000000000000D+00,   2.548958927467035D-04, &
   1.080238852130838D-03,   1.141463606658714D-03,   4.478695603998053D-04, &
   3.695113502301306D-04,  -1.760315950948643D-03,  -1.194714589778788D-03, &
   1.926919316434956D-03,  -4.508290222610061D-05,  -6.035575148541503D-04, &
   3.593693580412821D-04,  -2.155290483385380D-04,   2.013245547627558D-04, &
  -1.462483087358975D-04,  -2.565528287958425D-06,   1.540577641021299D-04, &
  -1.986832507234183D-04,   1.117486014974397D-04,   2.348841350074246D-05, &
  -9.059117673796060D-05,   5.658179716922303D-05,   0.000000000000000D+00, &
   1.274479463733589D-04,   0.000000000000000D+00,  -4.458876466187537D-04, &
   0.000000000000000D+00,  -2.272396004089013D-04,   0.000000000000000D+00, &
   6.493633789501396D-03,   0.000000000000000D+00,  -1.634972443918362D-02, &
   0.000000000000000D+00,   1.738329686942848D-02,   0.000000000000000D+00, &
  -8.075994453153702D-03,   0.000000000000000D+00,   6.931616088110412D-04, &
   0.000000000000000D+00,   5.728421434536498D-04,   0.000000000000000D+00, &
  -1.724721203612428D-04,   0.000000000000000D+00,   5.925096744578024D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.007564484126980D-05, &
   0.000000000000000D+00,  -8.112185846560857D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.705109126984147D-06,   0.000000000000000D+00, &
   2.635168650793645D-05,   0.000000000000000D+00,   8.046424062049038D-06, &
   0.000000000000000D+00,   8.046424062049028D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.516650961775237D-06,   0.000000000000000D+00, &
  -3.953292780229418D-06,   0.000000000000000D+00,  -7.389386262477370D-06, &
   0.000000000000000D+00,  -4.252499468933840D-06,   0.000000000000000D+00, &
  -9.846272701037369D-06,   0.000000000000000D+00,  -2.313694945166745D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.176496144903323D-06, &
   0.000000000000000D+00,  -6.849411414450057D-06,   0.000000000000000D+00, &
  -1.878864762437368D-06,   0.000000000000000D+00,  -1.698710977227943D-04, &
   0.000000000000000D+00,   2.415861062875761D-06,   0.000000000000000D+00, &
   1.054680964861936D-05,   0.000000000000000D+00,   3.080868454304038D-05, &
   0.000000000000000D+00,   1.341267792908430D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.720582629211378D-06,   0.000000000000000D+00, &
   6.933294466998921D-06,   0.000000000000000D+00,  -6.782767429584996D-07, &
   0.000000000000000D+00,  -8.034731019415634D-06,   0.000000000000000D+00, &
   5.103710412375814D-04,   0.000000000000000D+00,  -2.634612237025543D-06, &
   0.000000000000000D+00,  -8.299363992635323D-06,   0.000000000000000D+00, &
  -2.242114376126916D-05,   0.000000000000000D+00,  -8.725984125455901D-05, &
   0.000000000000000D+00,  -3.549269918010076D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.643737967469061D-06,   0.000000000000000D+00, &
  -1.949887918668196D-06,   0.000000000000000D+00,   5.005675668220126D-06, &
   0.000000000000000D+00,   4.904904608788694D-05,   0.000000000000000D+00, &
   5.856310473658402D-05,   0.000000000000000D+00,  -7.250191595630848D-04, &
   0.000000000000000D+00,   1.637087160093982D-06,   0.000000000000000D+00, &
   3.332330578890787D-06,   0.000000000000000D+00,   7.585027615089845D-06, &
   0.000000000000000D+00,   1.747694842112098D-05,   0.000000000000000D+00, &
   1.539662266089801D-04,   0.000000000000000D+00,   4.926272797193466D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.725976832738493D-07, &
   0.000000000000000D+00,  -1.079835787238049D-07,   0.000000000000000D+00, &
  -3.006448165904648D-06,   0.000000000000000D+00,  -2.237205335608970D-05, &
   0.000000000000000D+00,  -1.291823804800371D-04,   0.000000000000000D+00, &
  -9.123320992467805D-05,   0.000000000000000D+00,   5.702509441131757D-04, &
   0.000000000000000D+00,  -2.894430840073267D-07,   0.000000000000000D+00, &
  -4.265499094061503D-07,   0.000000000000000D+00,  -4.502467716076958D-07, &
   0.000000000000000D+00,   3.367862484588417D-06,   0.000000000000000D+00, &
   1.590335713566153D-05,   0.000000000000000D+00,  -1.911123779486561D-04, &
   0.000000000000000D+00,  -3.824601932749331D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.586867147995444D-07,   0.000000000000000D+00, &
  -7.772795281268802D-08,   0.000000000000000D+00,   3.053750743019015D-07, &
   0.000000000000000D+00,   3.010642327825726D-06,   0.000000000000000D+00, &
   2.615487253521705D-05,   0.000000000000000D+00,   1.755169364483818D-04, &
   0.000000000000000D+00,   5.015606072697769D-05,   0.000000000000000D+00, &
  -2.241046312973308D-04,   0.000000000000000D+00,  -3.921631203386745D-08, &
   0.000000000000000D+00,   4.581158854771778D-08,   0.000000000000000D+00, &
  -2.620471598309825D-07,   0.000000000000000D+00,  -2.048256948683538D-06, &
   0.000000000000000D+00,  -1.341094035385173D-05,   0.000000000000000D+00, &
  -3.761798973589609D-05,   0.000000000000000D+00,   1.746281240868026D-04, &
   0.000000000000000D+00,   1.561774697574059D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.329555272372994D-09,   0.000000000000000D+00, &
   6.485394852323057D-08,   0.000000000000000D+00,   1.088977436891489D-07, &
   0.000000000000000D+00,   5.258000894271972D-07,   0.000000000000000D+00, &
   1.159495793382121D-06,   0.000000000000000D+00,  -2.487842926762715D-05, &
   0.000000000000000D+00,  -1.391325209535254D-04,   0.000000000000000D+00, &
   4.094693794982148D-06,   0.000000000000000D+00,   2.262369880455808D-06, &
   0.000000000000000D+00,  -3.867404844288634D-10,   0.000000000000000D+00, &
  -4.613651275099898D-08,   0.000000000000000D+00,   8.839011227334916D-08, &
   0.000000000000000D+00,   1.914405821499281D-07,   0.000000000000000D+00, &
   8.996343594847989D-07,   0.000000000000000D+00,   1.721798391309308D-05, &
   0.000000000000000D+00,   1.280909088306866D-05,   0.000000000000000D+00, &
  -1.074301567017706D-04,   0.000000000000000D+00,  -2.447340411143772D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.175900125595566D-11, &
   0.000000000000000D+00,   1.223155635045261D-08,   0.000000000000000D+00, &
  -3.169632047829754D-08,   0.000000000000000D+00,  -7.056977565545099D-08, &
   0.000000000000000D+00,   1.195755250189854D-07,   0.000000000000000D+00, &
  -3.786047330192701D-06,   0.000000000000000D+00,   3.155134099353443D-05, &
   0.000000000000000D+00,   5.456775852592886D-05,   0.000000000000000D+00, &
  -8.542696347850615D-06,   0.000000000000000D+00,   4.210869653175477D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.440789552333344D-09,   0.000000000000000D+00,  -3.626082694628764D-08, &
   0.000000000000000D+00,   7.779247111660590D-09,   0.000000000000000D+00, &
  -3.147281226316350D-07,   0.000000000000000D+00,   7.016408814252885D-07, &
   0.000000000000000D+00,  -1.506058590288591D-05,   0.000000000000000D+00, &
   2.507480111740242D-05,   0.000000000000000D+00,   3.594910023527132D-05, &
   0.000000000000000D+00,   2.527331852262579D-06,   0.000000000000000D+00 /)
lem(:,160) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.993399855780395D-04,  -7.311917436163625D-04,  -1.313444778821275D-03, &
  -6.044638395257460D-04,   2.821026875946309D-03,   3.171639089475300D-03, &
  -3.180458300129141D-03,  -2.705405125805499D-03,   2.512470217946052D-03, &
   2.827985031953057D-04,  -7.336145631539965D-04,   3.185669915366820D-04, &
  -1.629091289837813D-04,   9.636676895285920D-05,  -1.032986054214191D-04, &
   8.458265671008095D-05,   2.420425380257632D-05,  -1.158686995786323D-04, &
   1.260470600920845D-04,  -7.076788683819891D-05,  -1.150829433951322D-05, &
   6.212115342869062D-05,  -1.993399855780712D-04,  -7.311917436163646D-04, &
  -1.313444778821282D-03,  -6.044638395257517D-04,   2.821026875946323D-03, &
   3.171639089475301D-03,  -3.180458300129146D-03,  -2.705405125805493D-03, &
   2.512470217946068D-03,   2.827985031952997D-04,  -7.336145631540188D-04, &
   3.185669915366885D-04,  -1.629091289837623D-04,   9.636676895285809D-05, &
  -1.032986054214248D-04,   8.458265671007676D-05,   2.420425380258019D-05, &
  -1.158686995786310D-04,   1.260470600920772D-04,  -7.076788683819059D-05, &
  -1.150829433950628D-05,   6.212115342867953D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.449985036417282D-04,   0.000000000000000D+00, &
  -6.819831942451591D-04,   0.000000000000000D+00,   4.436682616510004D-03, &
   0.000000000000000D+00,  -1.700368416531085D-02,   0.000000000000000D+00, &
   3.307708555506086D-02,   0.000000000000000D+00,  -3.452695806100701D-02, &
   0.000000000000000D+00,   1.944411433312049D-02,   0.000000000000000D+00, &
  -5.587830469117038D-03,   0.000000000000000D+00,   8.816906184473624D-04, &
   0.000000000000000D+00,  -2.161559471234447D-04,   0.000000000000000D+00, &
   6.458746693122813D-07,   1.937624007936714D-06,   0.000000000000000D+00, &
   3.823578042327859D-06,   0.000000000000000D+00,  -1.142552290013235D-05, &
   2.583498677248813D-06,   0.000000000000000D+00,  -9.164961557539676D-06, &
   0.000000000000000D+00,  -8.243709415584545D-06,   0.000000000000000D+00, &
  -1.097986937830219D-07,   0.000000000000000D+00,   5.987258184523810D-05, &
  -7.351228054353059D-06,   0.000000000000000D+00,   1.286523625390813D-05, &
   0.000000000000000D+00,   2.608746505230873D-05,   0.000000000000000D+00, &
   1.100335572991830D-05,   0.000000000000000D+00,   3.605742131132799D-06, &
   0.000000000000000D+00,  -3.047530269209951D-05,   0.000000000000000D+00, &
  -2.667220181258268D-04,   5.315045843375562D-06,   0.000000000000000D+00, &
  -1.336217774861020D-05,   0.000000000000000D+00,  -2.811573121591963D-05, &
   0.000000000000000D+00,  -4.227771985403860D-06,   0.000000000000000D+00, &
  -9.075891015413227D-06,   0.000000000000000D+00,   3.638242756650611D-07, &
   0.000000000000000D+00,   2.383137572122599D-05,   0.000000000000000D+00, &
   1.220068177383314D-04,   0.000000000000000D+00,   8.162399105495763D-04, &
  -2.435712862038821D-07,   0.000000000000000D+00,   8.448498771177061D-06, &
   0.000000000000000D+00,   1.598899915446957D-05,   0.000000000000000D+00, &
  -1.009387104759381D-05,   0.000000000000000D+00,  -1.834779565238407D-04, &
   0.000000000000000D+00,   3.729342415519637D-06,   0.000000000000000D+00, &
  -3.957449977790228D-06,   0.000000000000000D+00,  -1.565918857161386D-05, &
   0.000000000000000D+00,  -5.930753807795972D-05,   0.000000000000000D+00, &
  -2.421479161376912D-04,   0.000000000000000D+00,  -1.615521262838142D-03, &
  -1.234484128984861D-06,   0.000000000000000D+00,  -1.771815345194584D-06, &
   0.000000000000000D+00,  -4.184713718941859D-06,   0.000000000000000D+00, &
   1.674707890967958D-05,   0.000000000000000D+00,   1.483235811412107D-04, &
   0.000000000000000D+00,   5.399372683476860D-04,   0.000000000000000D+00, &
  -3.796051709592077D-07,   0.000000000000000D+00,   2.498989133128155D-06, &
   0.000000000000000D+00,   6.059751250466204D-06,   0.000000000000000D+00, &
   1.576780341270433D-05,   0.000000000000000D+00,   3.905095629555749D-05, &
   0.000000000000000D+00,   3.152711735637539D-04,   0.000000000000000D+00, &
   2.097722833241045D-03,   3.613124793208068D-07,   0.000000000000000D+00, &
  -7.592538917535712D-07,   0.000000000000000D+00,   7.191270038379347D-07, &
   0.000000000000000D+00,  -8.116074532411968D-06,   0.000000000000000D+00, &
  -5.483676842919187D-05,   0.000000000000000D+00,  -3.319393842351641D-04, &
   0.000000000000000D+00,  -7.641794939209049D-04,   0.000000000000000D+00, &
  -1.071611978136851D-07,   0.000000000000000D+00,  -3.031154029191968D-07, &
   0.000000000000000D+00,  -1.456827488351373D-06,   0.000000000000000D+00, &
   4.656927441083732D-07,   0.000000000000000D+00,   7.725085972709373D-06, &
   0.000000000000000D+00,   4.564921275036526D-05,   0.000000000000000D+00, &
  -3.447519967537107D-04,   0.000000000000000D+00,  -1.817290325451267D-03, &
   1.627612382820021D-08,   0.000000000000000D+00,   3.154242515673320D-07, &
   0.000000000000000D+00,  -5.516305511397086D-07,   0.000000000000000D+00, &
   2.967252813490544D-07,   0.000000000000000D+00,   7.162619704812772D-06, &
   0.000000000000000D+00,   5.992797053510559D-05,   0.000000000000000D+00, &
   4.348799021008724D-04,   0.000000000000000D+00,   6.022452792847012D-04, &
   0.000000000000000D+00,  -1.280123725576727D-09,   0.000000000000000D+00, &
  -9.951031176410340D-08,   0.000000000000000D+00,   6.165623783546453D-07, &
   0.000000000000000D+00,  -1.535831332286322D-06,   0.000000000000000D+00, &
  -3.974778176041928D-06,   0.000000000000000D+00,  -3.004665106305449D-05, &
   0.000000000000000D+00,  -1.004698852520413D-04,   0.000000000000000D+00, &
   3.615067367513969D-04,   0.000000000000000D+00,   1.060067700284547D-03, &
   4.145162566281613D-11,   0.000000000000000D+00,   1.051171079090721D-08, &
   0.000000000000000D+00,   1.005552388056796D-07,   0.000000000000000D+00, &
   1.236559917456650D-06,   0.000000000000000D+00,   1.725876560712009D-07, &
   0.000000000000000D+00,   3.293559447706156D-06,   0.000000000000000D+00, &
  -5.473796917941583D-05,   0.000000000000000D+00,  -3.760303487933162D-04, &
   0.000000000000000D+00,  -2.383978466058864D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.587750973340133D-09, &
   0.000000000000000D+00,  -2.278085102316019D-07,   0.000000000000000D+00, &
   4.233838006338194D-07,   0.000000000000000D+00,   1.184406076746002D-06, &
   0.000000000000000D+00,   1.147990403813625D-06,   0.000000000000000D+00, &
   3.855927712565527D-05,   0.000000000000000D+00,   4.911876503336971D-05, &
   0.000000000000000D+00,  -3.151586807952425D-04,   0.000000000000000D+00, &
  -4.244206537290533D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.768350254088788D-10,   0.000000000000000D+00,   6.795690127786417D-08, &
   0.000000000000000D+00,  -7.105008719092168D-07,   0.000000000000000D+00, &
   3.539287952408322D-07,   0.000000000000000D+00,   2.805040964020591D-07, &
   0.000000000000000D+00,  -9.050921062375390D-06,   0.000000000000000D+00, &
   7.053464895918323D-05,   0.000000000000000D+00,   1.940089969106714D-04, &
   0.000000000000000D+00,   3.440303400815905D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.183145868357614D-11, &
   0.000000000000000D+00,   7.846269848622232D-09,   0.000000000000000D+00, &
   5.586818676649600D-08,   0.000000000000000D+00,  -1.129140776448501D-06, &
   0.000000000000000D+00,  -1.433764192730908D-07,   0.000000000000000D+00, &
   1.415648243323581D-06,   0.000000000000000D+00,  -3.478381353946024D-05, &
   0.000000000000000D+00,   5.105429334853264D-05,   0.000000000000000D+00, &
   1.873675878922278D-04,   0.000000000000000D+00,   1.281514507595194D-04 /)
lem(:,161) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.949392611131761D-03,  -7.176567341111109D-03,  -1.244313820506998D-02, &
   1.528359468299693D-02,   1.430687046340460D-02,  -9.311672367681562D-03, &
  -8.312381234447412D-03,  -1.188314002188381D-03,   2.330459640301235D-04, &
   1.700106390675791D-03,   2.193400465601318D-03,   1.345052953375396D-03, &
  -1.287674436683960D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.949392611131737D-03,  -7.176567341111074D-03, &
  -1.244313820506994D-02,   1.528359468299689D-02,   1.430687046340461D-02, &
  -9.311672367681538D-03,  -8.312381234447441D-03,  -1.188314002188339D-03, &
   2.330459640301672D-04,   1.700106390675727D-03,   2.193400465601329D-03, &
   1.345052953375449D-03,  -1.287674436684458D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.574793761898334D-04,   0.000000000000000D+00,   4.277530997142278D-06, &
   0.000000000000000D+00,   2.250384573408714D-07,   0.000000000000000D+00, &
   1.736010956602455D-08,   0.000000000000000D+00,   1.808344746580495D-09, &
  -2.268109888247187D-03,   0.000000000000000D+00,  -4.583068925510663D-05, &
   0.000000000000000D+00,  -3.446204314423375D-06,   0.000000000000000D+00, &
  -3.538791565444828D-07,   0.000000000000000D+00,  -4.131372228555897D-08, &
   0.000000000000000D+00,   5.157129037820276D-03,   0.000000000000000D+00, &
   3.043263124445528D-05,   0.000000000000000D+00,   1.292783260932361D-05, &
   0.000000000000000D+00,   2.268287879256140D-06,   0.000000000000000D+00, &
   3.763717034551942D-07,   0.000000000000000D+00,   7.518708925628384D-08, &
  -3.281375185423159D-03,   0.000000000000000D+00,   6.217960237042018D-04, &
   0.000000000000000D+00,   1.944798891526731D-05,   0.000000000000000D+00, &
  -2.907956496468297D-06,   0.000000000000000D+00,  -1.267539343871703D-06, &
   0.000000000000000D+00,  -3.373273658874328D-07,   0.000000000000000D+00, &
  -6.156379055735834D-03,   0.000000000000000D+00,  -1.931637235903234D-03, &
   0.000000000000000D+00,  -1.433501939336112D-04,   0.000000000000000D+00, &
  -1.777743177580647D-05,   0.000000000000000D+00,  -6.267391456894909D-07, &
   0.000000000000000D+00,   5.584377696523460D-07,   0.000000000000000D+00, &
   3.613124793208068D-07,   1.269780880169235D-02,   0.000000000000000D+00, &
   1.528633027653043D-03,   0.000000000000000D+00,   2.552493535055898D-05, &
   0.000000000000000D+00,   4.029923206278635D-05,   0.000000000000000D+00, &
   1.131657952231370D-05,   0.000000000000000D+00,   1.629021230760643D-06, &
   0.000000000000000D+00,  -1.434954405797416D-07,   0.000000000000000D+00, &
  -6.595508045911873D-03,   0.000000000000000D+00,   2.093680223759923D-03, &
   0.000000000000000D+00,   6.716892231970756D-04,   0.000000000000000D+00, &
   5.612450043020509D-05,   0.000000000000000D+00,  -7.724438243501075D-06, &
   0.000000000000000D+00,  -6.032900400827948D-06,   0.000000000000000D+00, &
  -1.599273372053388D-06,   0.000000000000000D+00,  -1.598942116937906D-08, &
  -3.798571551630510D-03,   0.000000000000000D+00,  -4.725334919004202D-03, &
   0.000000000000000D+00,  -8.665725691264918D-04,   0.000000000000000D+00, &
  -1.714258057538222D-04,   0.000000000000000D+00,  -4.103177322459683D-05, &
   0.000000000000000D+00,  -2.588069150894415D-06,   0.000000000000000D+00, &
   2.603252525729074D-06,   0.000000000000000D+00,   1.228895973634681D-06, &
   0.000000000000000D+00,   6.410564675303267D-03,   0.000000000000000D+00, &
   2.244844912180604D-03,   0.000000000000000D+00,  -5.827892683430647D-04, &
   0.000000000000000D+00,  -8.623287160514105D-05,   0.000000000000000D+00, &
   3.661328548813057D-05,   0.000000000000000D+00,   2.243947602784934D-05, &
   0.000000000000000D+00,   4.776649320619277D-06,   0.000000000000000D+00, &
  -7.209176631955364D-07,   0.000000000000000D+00,  -1.336607143562493D-06, &
  -3.043411497249097D-03,   0.000000000000000D+00,   1.766962664457464D-03, &
   0.000000000000000D+00,   1.784822972777737D-03,   0.000000000000000D+00, &
   3.873249594157238D-04,   0.000000000000000D+00,   8.544619064097900D-05, &
   0.000000000000000D+00,   1.480362207488238D-06,   0.000000000000000D+00, &
  -1.005049985553653D-05,   0.000000000000000D+00,  -4.115152487700175D-06, &
   0.000000000000000D+00,  -1.494080458749326D-07,   0.000000000000000D+00, &
   5.203707094697056D-04,   0.000000000000000D+00,  -2.402930874275847D-03, &
   0.000000000000000D+00,  -6.262054004842212D-04,   0.000000000000000D+00, &
   4.702155960827982D-05,   0.000000000000000D+00,  -6.400443560714663D-05, &
   0.000000000000000D+00,  -4.390310486350425D-05,   0.000000000000000D+00, &
  -9.668457623716237D-06,   0.000000000000000D+00,   3.190891550191664D-06, &
   0.000000000000000D+00,   2.832078829671312D-06,   0.000000000000000D+00, &
   1.512033149951538D-07,   0.000000000000000D+00,   0.000000000000000D+00, &
   8.379694141651865D-04,   0.000000000000000D+00,  -1.000900825706165D-03, &
   0.000000000000000D+00,  -5.280678074221608D-04,   0.000000000000000D+00, &
  -1.217086214794309D-04,   0.000000000000000D+00,  -7.078161213520282D-06, &
   0.000000000000000D+00,   1.749972000869274D-05,   0.000000000000000D+00, &
   8.677939223551047D-06,   0.000000000000000D+00,  -1.476050477243915D-08, &
   0.000000000000000D+00,  -1.529071964448533D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.172507620783621D-05, &
   0.000000000000000D+00,   8.383296766524495D-04,   0.000000000000000D+00, &
   5.097163229124507D-05,   0.000000000000000D+00,   3.637473975176332D-05, &
   0.000000000000000D+00,   4.852248110037242D-05,   0.000000000000000D+00, &
   1.645114485548197D-05,   0.000000000000000D+00,  -4.355610217782858D-06, &
   0.000000000000000D+00,  -5.633897612306407D-06,   0.000000000000000D+00, &
  -9.850992537798034D-07,   0.000000000000000D+00,   1.970926920573156D-06 /)
lem(:,162) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.732523660980991D-04,   3.020737229890544D-03, &
  -3.870827781093686D-03,  -3.972218009903805D-03,   1.040452741939554D-02, &
   2.154908912770068D-04,  -9.607301753356084D-03,   1.061309429761056D-03, &
   4.011936904062333D-03,   7.557656392424595D-04,  -3.535239578571324D-04, &
  -1.030406168719017D-03,  -7.344635601438445D-04,   9.787798433155947D-05, &
   5.789130622936963D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.732523660981160D-04, &
  -3.020737229890558D-03,   3.870827781093666D-03,   3.972218009903806D-03, &
  -1.040452741939553D-02,  -2.154908912770131D-04,   9.607301753356117D-03, &
  -1.061309429761058D-03,  -4.011936904062336D-03,  -7.557656392424470D-04, &
   3.535239578571215D-04,   1.030406168719006D-03,   7.344635601438593D-04, &
  -9.787798433155165D-05,  -5.789130622937241D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.893306298844778D-08,   0.000000000000000D+00,   1.262811868017068D-06, &
   0.000000000000000D+00,   5.354230763732841D-06,   0.000000000000000D+00, &
   2.202359159139984D-06,   0.000000000000000D+00,  -1.130849126010338D-05, &
   0.000000000000000D+00,  -3.764937307093773D-06,   0.000000000000000D+00, &
   1.132078131330093D-05,   0.000000000000000D+00,  -7.583448276120241D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.307559179278378D-06,   0.000000000000000D+00, &
   9.156980870164805D-08,   0.000000000000000D+00,   4.370377233418303D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.629434100695414D-05, &
   0.000000000000000D+00,  -1.569600438896404D-06,   0.000000000000000D+00, &
  -1.290942198196294D-07,   0.000000000000000D+00,  -4.006179130541460D-09, &
   0.000000000000000D+00,   1.248474981144253D-04,   0.000000000000000D+00, &
   8.020576064727131D-06,   0.000000000000000D+00,   1.062786094418755D-06, &
   0.000000000000000D+00,   1.299782562400873D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.139519728562671D-04,   0.000000000000000D+00, &
  -8.736182380010295D-06,   0.000000000000000D+00,  -2.891016991165012D-06, &
   0.000000000000000D+00,  -6.890285696315789D-07,   0.000000000000000D+00, &
  -1.378515110596924D-07,   0.000000000000000D+00,   4.042627890765739D-04, &
   0.000000000000000D+00,  -5.948574645340843D-05,   0.000000000000000D+00, &
  -2.563074900744986D-06,   0.000000000000000D+00,   8.748603385109303D-07, &
   0.000000000000000D+00,   4.532488995319171D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.460510411821128D-05,   0.000000000000000D+00, &
   2.563215409725594D-04,   0.000000000000000D+00,   2.724540117028070D-05, &
   0.000000000000000D+00,   3.619707607206799D-06,   0.000000000000000D+00, &
  -7.409224923397804D-08,   0.000000000000000D+00,  -5.254860364825532D-07, &
   0.000000000000000D+00,  -6.157039855285107D-04,   0.000000000000000D+00, &
  -4.150414591679770D-04,   0.000000000000000D+00,  -3.368580990236550D-05, &
   0.000000000000000D+00,  -1.143842076240588D-05,   0.000000000000000D+00, &
  -2.824458347905232D-06,   0.000000000000000D+00,  -1.725976832738493D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.021080606021583D-03, &
   0.000000000000000D+00,   1.549164360572818D-04,   0.000000000000000D+00, &
  -7.173652426428013D-05,   0.000000000000000D+00,  -1.133125141114456D-06, &
   0.000000000000000D+00,   4.177357229357691D-06,   0.000000000000000D+00, &
   1.909782004518252D-06,   0.000000000000000D+00,   2.657174890111337D-07, &
   0.000000000000000D+00,  -6.605588585345746D-04,   0.000000000000000D+00, &
   5.281837129021248D-04,   0.000000000000000D+00,   2.397416291655131D-04, &
   0.000000000000000D+00,   4.138409301572558D-05,   0.000000000000000D+00, &
   6.425884068897881D-06,   0.000000000000000D+00,  -1.023989020078925D-06, &
   0.000000000000000D+00,  -1.213685963065879D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.723221439443970D-05,   0.000000000000000D+00, &
  -9.476510316143676D-04,   0.000000000000000D+00,  -1.742525498007689D-04, &
   0.000000000000000D+00,  -3.411008151322504D-05,   0.000000000000000D+00, &
  -1.760108717510404D-05,   0.000000000000000D+00,  -5.475972917715155D-06, &
   0.000000000000000D+00,  -1.888138450875643D-07,   0.000000000000000D+00, &
   1.771560184858361D-06,   0.000000000000000D+00,   5.059692220950299D-04, &
   0.000000000000000D+00,   5.676056372656788D-04,   0.000000000000000D+00, &
  -2.228015812720763D-04,   0.000000000000000D+00,  -6.752405717711948D-05, &
   0.000000000000000D+00,  -3.823264703106282D-06,   0.000000000000000D+00, &
   6.012843331671461D-06,   0.000000000000000D+00,   3.615633128941117D-06, &
   0.000000000000000D+00,   4.932452733447065D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.277007487660230D-04,   0.000000000000000D+00, &
   1.791414277255398D-04,   0.000000000000000D+00,   4.928557563638327D-04, &
   0.000000000000000D+00,   1.095588377038783D-04,   0.000000000000000D+00, &
   3.428570776313759D-05,   0.000000000000000D+00,   8.380305875676559D-06, &
   0.000000000000000D+00,  -1.077609480805548D-06,   0.000000000000000D+00, &
  -2.263763838452371D-06,   0.000000000000000D+00,  -8.176315963165973D-07, &
   0.000000000000000D+00,   1.873506705606788D-04,   0.000000000000000D+00, &
  -5.200880142081024D-04,   0.000000000000000D+00,  -2.402339580671914D-04, &
   0.000000000000000D+00,   3.779052035309390D-05,   0.000000000000000D+00, &
  -5.617674056874094D-06,   0.000000000000000D+00,  -1.215437395834039D-05, &
   0.000000000000000D+00,  -6.243515460891264D-06,   0.000000000000000D+00, &
  -6.344285081967431D-07,   0.000000000000000D+00,   9.902216649735651D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.425510988358650D-05, &
   0.000000000000000D+00,   3.517717798319429D-04,   0.000000000000000D+00, &
  -2.012366284912053D-04,   0.000000000000000D+00,  -1.620241832851591D-04, &
   0.000000000000000D+00,  -4.016048972262819D-05,   0.000000000000000D+00, &
  -8.638679172999461D-06,   0.000000000000000D+00,   2.631314928200424D-06, &
   0.000000000000000D+00,   3.604306114473789D-06,   0.000000000000000D+00, &
   1.096997150231724D-06,   0.000000000000000D+00,  -2.162670963024975D-06, &
   0.000000000000000D+00,   4.483919945544193D-06,   0.000000000000000D+00, &
  -9.247433517326246D-05,   0.000000000000000D+00,   3.019714986298608D-04, &
   0.000000000000000D+00,   4.107436271826624D-05,   0.000000000000000D+00, &
   5.624959288410944D-06,   0.000000000000000D+00,   1.203789572992409D-05, &
   0.000000000000000D+00,   6.862325049889300D-06,   0.000000000000000D+00, &
   6.835687851405842D-07,   0.000000000000000D+00,  -1.671384990659081D-06, &
   0.000000000000000D+00,  -8.334506727445517D-07,   0.000000000000000D+00 /)
lem(:,163) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.461560075141252D-04, &
  -1.064102730263720D-03,   2.183382394513304D-05,   1.925779406480802D-03, &
   4.623028665865405D-03,  -2.966117586688743D-03,  -1.237845163359318D-02, &
   2.779822274597299D-03,   1.199478387744894D-02,   2.930742668733007D-04, &
  -4.444027321041358D-03,  -1.906495590383563D-03,  -2.365897569614193D-04, &
   7.543029789497031D-04,   5.911639520397358D-04,   1.384650279584463D-05, &
  -1.953735812987547D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.461560075141326D-04,  -1.064102730263731D-03,   2.183382394512655D-05, &
   1.925779406480816D-03,   4.623028665865394D-03,  -2.966117586688746D-03, &
  -1.237845163359316D-02,   2.779822274597288D-03,   1.199478387744891D-02, &
   2.930742668733040D-04,  -4.444027321041325D-03,  -1.906495590383550D-03, &
  -2.365897569614289D-04,   7.543029789497064D-04,   5.911639520397265D-04, &
   1.384650279583836D-05,  -1.953735812987446D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.175504292474783D-06, &
   0.000000000000000D+00,  -1.033928596497591D-05,   0.000000000000000D+00, &
  -2.009924754273798D-05,   0.000000000000000D+00,   6.121339628069625D-06, &
   0.000000000000000D+00,   3.518437701413017D-05,   0.000000000000000D+00, &
  -1.335050338499289D-06,   0.000000000000000D+00,  -2.636860852382162D-05, &
   0.000000000000000D+00,   3.965676030168019D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.389843879815788D-06,   0.000000000000000D+00, &
   5.547017257884914D-08,   0.000000000000000D+00,  -3.852095318036311D-09, &
  -3.716501562779691D-05,   0.000000000000000D+00,  -1.122610635523363D-06, &
   0.000000000000000D+00,  -4.890993752207888D-08,   0.000000000000000D+00, &
   2.628891725674636D-05,   0.000000000000000D+00,   5.418819525697924D-06, &
   0.000000000000000D+00,   5.956459971131624D-07,   0.000000000000000D+00, &
  -1.168468913117834D-08,   5.137647384248552D-04,   0.000000000000000D+00, &
  -1.204072546110388D-06,   0.000000000000000D+00,  -1.513913481732575D-06, &
   0.000000000000000D+00,  -3.076423397125949D-07,   0.000000000000000D+00, &
  -1.619263336385427D-03,   0.000000000000000D+00,   4.702884498280682D-06, &
   0.000000000000000D+00,  -2.043976249791347D-06,   0.000000000000000D+00, &
   4.421271583743218D-07,   0.000000000000000D+00,   2.946291154348472D-07, &
   1.291426562997436D-03,   0.000000000000000D+00,  -2.204478341239682D-04, &
   0.000000000000000D+00,   5.858518696720213D-06,   0.000000000000000D+00, &
   1.462068614444325D-06,   0.000000000000000D+00,  -2.243604766761929D-07, &
   0.000000000000000D+00,   1.793960813587999D-03,   0.000000000000000D+00, &
   6.972111365471952D-04,   0.000000000000000D+00,   7.565531305476565D-09, &
   0.000000000000000D+00,  -2.070657544487864D-06,   0.000000000000000D+00, &
  -7.555583329428392D-07,   0.000000000000000D+00,   6.445847247207605D-07, &
  -4.066117427796495D-03,   0.000000000000000D+00,  -5.643835393001061D-04, &
   0.000000000000000D+00,   9.022009602214859D-05,   0.000000000000000D+00, &
   8.941707451976625D-08,   0.000000000000000D+00,   6.987764554929718D-07, &
   0.000000000000000D+00,   2.952927200278581D-07,   0.000000000000000D+00, &
   1.943312663602548D-03,   0.000000000000000D+00,  -8.043799752967338D-04, &
   0.000000000000000D+00,  -3.571618643978382D-04,   0.000000000000000D+00, &
  -1.595194632310965D-05,   0.000000000000000D+00,  -3.463483869684455D-07, &
   0.000000000000000D+00,  -4.713360004408447D-07,   0.000000000000000D+00, &
  -7.592538917535712D-07,   1.528633027653043D-03,   0.000000000000000D+00, &
   1.822284226526554D-03,   0.000000000000000D+00,   3.330553561877190D-04, &
   0.000000000000000D+00,   2.736254705700825D-06,   0.000000000000000D+00, &
   4.293690818018065D-06,   0.000000000000000D+00,   5.355518786284881D-07, &
   0.000000000000000D+00,   5.063638911213269D-07,   0.000000000000000D+00, &
  -2.088458760948072D-03,   0.000000000000000D+00,  -8.701130517366598D-04, &
   0.000000000000000D+00,   4.064145781598478D-04,   0.000000000000000D+00, &
   1.401646999228513D-04,   0.000000000000000D+00,   9.116389086806819D-06, &
   0.000000000000000D+00,  -7.988571072349726D-07,   0.000000000000000D+00, &
  -3.248208440660569D-07,   0.000000000000000D+00,  -1.686808677862445D-06, &
   7.298729841893743D-04,   0.000000000000000D+00,  -6.679240571640717D-04, &
   0.000000000000000D+00,  -9.803837701518551D-04,   0.000000000000000D+00, &
  -1.972038333752597D-04,   0.000000000000000D+00,  -3.439842876744587D-05, &
   0.000000000000000D+00,  -6.784953762064741D-06,   0.000000000000000D+00, &
  -5.815353751745146D-07,   0.000000000000000D+00,   4.117643514432205D-07, &
   0.000000000000000D+00,   3.711098879643723D-05,   0.000000000000000D+00, &
   8.967154831373431D-04,   0.000000000000000D+00,   4.392822744378018D-04, &
   0.000000000000000D+00,  -1.351527930107809D-04,   0.000000000000000D+00, &
  -2.062186227086453D-05,   0.000000000000000D+00,   7.578573998236979D-06, &
   0.000000000000000D+00,   3.989182485669879D-06,   0.000000000000000D+00, &
   1.146428434869874D-06,   0.000000000000000D+00,   1.265086845963140D-06, &
  -6.323468624576155D-05,   0.000000000000000D+00,  -3.103044425133791D-04, &
   0.000000000000000D+00,   3.927097034197233D-04,   0.000000000000000D+00, &
   4.256994224203359D-04,   0.000000000000000D+00,   9.504031585117143D-05, &
   0.000000000000000D+00,   1.967709233667479D-05,   0.000000000000000D+00, &
   1.417775552407557D-07,   0.000000000000000D+00,  -2.052951920121643D-06, &
   0.000000000000000D+00,  -9.107662323656564D-07,   0.000000000000000D+00, &
   6.482701651245800D-06,   0.000000000000000D+00,   7.661669075443939D-06, &
   0.000000000000000D+00,  -4.569544212920669D-04,   0.000000000000000D+00, &
  -1.463922844156403D-04,   0.000000000000000D+00,   1.298295145204354D-05, &
   0.000000000000000D+00,  -1.605487528084676D-05,   0.000000000000000D+00, &
  -1.039161279303662D-05,   0.000000000000000D+00,  -2.017993613179798D-06, &
   0.000000000000000D+00,   3.078525092577053D-07,   0.000000000000000D+00, &
   1.781498019477693D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.801643769178326D-06,   0.000000000000000D+00,   1.261818710356560D-04, &
   0.000000000000000D+00,  -2.313373604426494D-04,   0.000000000000000D+00, &
  -1.376709495352444D-04,   0.000000000000000D+00,  -3.190518858708481D-05, &
   0.000000000000000D+00,  -1.472757720157571D-06,   0.000000000000000D+00, &
   4.418763384734989D-06,   0.000000000000000D+00,   2.051876604498146D-06, &
   0.000000000000000D+00,  -4.368857506596637D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.124632149442350D-05, &
   0.000000000000000D+00,  -1.762179849762337D-06,   0.000000000000000D+00, &
   1.789501125204395D-04,   0.000000000000000D+00,   1.426480301681308D-05, &
   0.000000000000000D+00,   9.875973415858750D-06,   0.000000000000000D+00, &
   1.309271861642871D-05,   0.000000000000000D+00,   4.049733345692801D-06, &
   0.000000000000000D+00,  -1.200737762375810D-06,   0.000000000000000D+00, &
  -1.520803040436786D-06,   0.000000000000000D+00,  -1.507073240547787D-06 /)
lem(:,164) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.457335346027104D-05,   3.210408103810917D-04,  -7.616816074385421D-04, &
  -1.281712637474140D-03,   4.592313132258529D-03,   2.100704079632807D-03, &
  -1.173283147897967D-02,  -1.099535709370565D-03,   1.612884784697679D-02, &
  -8.431802652905667D-04,  -1.215166423148545D-02,   3.538335455809758D-04, &
   4.958413728649818D-03,   1.150227598439748D-03,  -9.597439563818010D-04, &
  -1.091228429556304D-03,  -1.226760083783123D-04,   4.212054569233328D-04, &
   2.108683204169298D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.457335346026460D-05,  -3.210408103810803D-04, &
   7.616816074385477D-04,   1.281712637474122D-03,  -4.592313132258523D-03, &
  -2.100704079632801D-03,   1.173283147897962D-02,   1.099535709370563D-03, &
  -1.612884784697680D-02,   8.431802652905663D-04,   1.215166423148543D-02, &
  -3.538335455809817D-04,  -4.958413728649800D-03,  -1.150227598439739D-03, &
   9.597439563817990D-04,   1.091228429556293D-03,   1.226760083783320D-04, &
  -4.212054569233223D-04,  -2.108683204169516D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.467495207471076D-07, &
   0.000000000000000D+00,   1.182493951458721D-05,   0.000000000000000D+00, &
   3.956946882596771D-05,   0.000000000000000D+00,   2.841606985509119D-05, &
   0.000000000000000D+00,  -3.350710546178353D-05,   0.000000000000000D+00, &
  -4.604878024320526D-05,   0.000000000000000D+00,   4.663600072884483D-06, &
   0.000000000000000D+00,   3.234964037244822D-05,   0.000000000000000D+00, &
   4.831759129786107D-06,   0.000000000000000D+00,  -1.377163155355674D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.500225641308921D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.055507468599061D-06,   0.000000000000000D+00, &
   4.602503496780862D-08,   0.000000000000000D+00,   6.164419755943458D-06, &
   0.000000000000000D+00,   3.239095214559140D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.874416562377052D-06,   0.000000000000000D+00, &
  -1.073784076687740D-06,   0.000000000000000D+00,  -4.097228656370923D-08, &
   0.000000000000000D+00,  -4.578207614570900D-05,   0.000000000000000D+00, &
  -2.652642884937274D-06,   0.000000000000000D+00,   2.573032915031231D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.089762737225385D-04, &
   0.000000000000000D+00,   1.088592131416597D-05,   0.000000000000000D+00, &
   1.470930101249545D-06,   0.000000000000000D+00,  -6.837819379885452D-07, &
   0.000000000000000D+00,  -3.464094887346114D-04,   0.000000000000000D+00, &
   3.053255461770343D-05,   0.000000000000000D+00,  -1.921003804308576D-06, &
   0.000000000000000D+00,  -4.573993544580747D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.312288404859991D-04,   0.000000000000000D+00, &
  -2.043691063555303D-04,   0.000000000000000D+00,  -1.109929737845447D-05, &
   0.000000000000000D+00,  -1.623856617800016D-07,   0.000000000000000D+00, &
  -1.622962659310455D-08,   0.000000000000000D+00,   4.506685853534540D-04, &
   0.000000000000000D+00,   3.831738044021551D-04,   0.000000000000000D+00, &
   8.917739410907402D-06,   0.000000000000000D+00,   4.630130420805188D-06, &
   0.000000000000000D+00,   5.745661216318399D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.135985844627100D-04,   0.000000000000000D+00, &
  -1.743213725098726D-04,   0.000000000000000D+00,   1.007272391955438D-04, &
   0.000000000000000D+00,   3.268611557104689D-06,   0.000000000000000D+00, &
  -1.515813000354033D-06,   0.000000000000000D+00,   9.163542372660185D-07, &
   0.000000000000000D+00,   4.891642540860371D-04,   0.000000000000000D+00, &
  -4.787111743796252D-04,   0.000000000000000D+00,  -2.691372861901551D-04, &
   0.000000000000000D+00,  -2.789304030294763D-05,   0.000000000000000D+00, &
  -4.132350453082394D-06,   0.000000000000000D+00,  -1.079835787238049D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.549164360572818D-04, &
   0.000000000000000D+00,   8.965368922365395D-04,   0.000000000000000D+00, &
   1.726552771828700D-04,   0.000000000000000D+00,  -3.635688015320592D-06, &
   0.000000000000000D+00,   8.308536150604469D-06,   0.000000000000000D+00, &
   2.688601926423918D-06,   0.000000000000000D+00,   2.916767081698878D-07, &
   0.000000000000000D+00,  -4.498500667629974D-04,   0.000000000000000D+00, &
  -5.193203918849652D-04,   0.000000000000000D+00,   2.998900871970221D-04, &
   0.000000000000000D+00,   1.214632515765755D-04,   0.000000000000000D+00, &
   1.401462033320931D-05,   0.000000000000000D+00,  -1.262582920310530D-06, &
   0.000000000000000D+00,  -2.054243141362819D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.039650007388793D-04,   0.000000000000000D+00, &
  -2.056563471647828D-04,   0.000000000000000D+00,  -6.206600512632393D-04, &
   0.000000000000000D+00,  -1.297720878952873D-04,   0.000000000000000D+00, &
  -2.994639012669677D-05,   0.000000000000000D+00,  -9.557266042594554D-06, &
   0.000000000000000D+00,  -7.034179377209223D-07,   0.000000000000000D+00, &
   2.075969138771546D-07,   0.000000000000000D+00,  -7.940676875286581D-05, &
   0.000000000000000D+00,   4.970060905655846D-04,   0.000000000000000D+00, &
   3.254887454185666D-04,   0.000000000000000D+00,  -1.079344468353159D-04, &
   0.000000000000000D+00,  -2.356837253144727D-05,   0.000000000000000D+00, &
   5.122778952825531D-06,   0.000000000000000D+00,   5.225537945958704D-06, &
   0.000000000000000D+00,   1.829455634431061D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.105631744350698D-05,   0.000000000000000D+00, &
  -2.972921725007927D-04,   0.000000000000000D+00,   2.029719495467116D-04, &
   0.000000000000000D+00,   2.903429485859620D-04,   0.000000000000000D+00, &
   6.785185663490729D-05,   0.000000000000000D+00,   1.824834975540136D-05, &
   0.000000000000000D+00,   1.473037030042468D-06,   0.000000000000000D+00, &
  -2.030969816697011D-06,   0.000000000000000D+00,  -1.631628447667663D-06, &
   0.000000000000000D+00,   1.263550809175754D-05,   0.000000000000000D+00, &
   5.461874001598685D-05,   0.000000000000000D+00,  -3.485967174928456D-04, &
   0.000000000000000D+00,  -1.175592949741629D-04,   0.000000000000000D+00, &
   1.369513142753737D-05,   0.000000000000000D+00,  -9.080130308931080D-06, &
   0.000000000000000D+00,  -8.739018904391984D-06,   0.000000000000000D+00, &
  -3.007207885278287D-06,   0.000000000000000D+00,   1.002467826584194D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.943290101554190D-06, &
   0.000000000000000D+00,   1.756314406530065D-05,   0.000000000000000D+00, &
   1.481151634912577D-04,   0.000000000000000D+00,  -1.521534766833258D-04, &
   0.000000000000000D+00,  -9.433599401151117D-05,   0.000000000000000D+00, &
  -2.325476609176459D-05,   0.000000000000000D+00,  -2.799643423149977D-06, &
   0.000000000000000D+00,   3.263910305953127D-06,   0.000000000000000D+00, &
   1.849464110973787D-06,   0.000000000000000D+00,  -1.082964241067019D-07, &
   0.000000000000000D+00,   2.127463742275859D-07,   0.000000000000000D+00, &
  -5.796282324523070D-06,   0.000000000000000D+00,   9.941397323907679D-06, &
   0.000000000000000D+00,   1.579088791112338D-04,   0.000000000000000D+00, &
   1.518657769008570D-05,   0.000000000000000D+00,   5.331490575482450D-06, &
   0.000000000000000D+00,   8.608745967199175D-06,   0.000000000000000D+00, &
   3.828890410589409D-06,   0.000000000000000D+00,  -7.227221575688694D-07, &
   0.000000000000000D+00,  -1.755795271741424D-06,   0.000000000000000D+00 /)
lem(:,165) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.557983665404428D-06,  -1.174328092019663D-04, &
   3.154448578722533D-04,   5.331463562182404D-04,  -1.094547566632560D-03, &
   2.343314717548515D-05,  -3.413118769172896D-03,   2.356532596893657D-05, &
   1.374489693784208D-02,  -1.269332667614176D-03,  -1.886311410054378D-02, &
   2.287972267375935D-03,   1.262089599392469D-02,  -1.667735886510134D-03, &
  -4.480773161449224D-03,   4.463480683409276D-04,   1.298500200709627D-03, &
   1.341775631073218D-04,  -5.839791596251299D-04,  -2.249756128604771D-04, &
   2.280471085826246D-04,   0.000000000000000D+00,   8.557983665396424D-06, &
  -1.174328092019649D-04,   3.154448578722524D-04,   5.331463562182522D-04, &
  -1.094547566632553D-03,   2.343314717547449D-05,  -3.413118769172883D-03, &
   2.356532596894663D-05,   1.374489693784207D-02,  -1.269332667614177D-03, &
  -1.886311410054376D-02,   2.287972267375934D-03,   1.262089599392468D-02, &
  -1.667735886510136D-03,  -4.480773161449231D-03,   4.463480683409455D-04, &
   1.298500200709622D-03,   1.341775631073076D-04,  -5.839791596251179D-04, &
  -2.249756128604840D-04,   2.280471085826095D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.000838070572848D-06,   0.000000000000000D+00, &
  -5.556950215931722D-05,   0.000000000000000D+00,  -7.518624098384640D-05, &
   0.000000000000000D+00,  -1.260643994483622D-06,   0.000000000000000D+00, &
   3.853281356185133D-05,   0.000000000000000D+00,   3.490019952688066D-05, &
   0.000000000000000D+00,   2.710535997174008D-05,   0.000000000000000D+00, &
  -3.872112782468670D-05,   0.000000000000000D+00,  -2.969999144511627D-05, &
   0.000000000000000D+00,   2.359110337356395D-05,   0.000000000000000D+00, &
   1.353136028062282D-07,  -1.999634352575841D-06,   0.000000000000000D+00, &
   8.247686266262302D-06,   0.000000000000000D+00,   7.571118252216448D-08, &
   1.036115587199170D-05,   0.000000000000000D+00,  -1.010019392372205D-06, &
   0.000000000000000D+00,  -9.246112019028978D-05,   0.000000000000000D+00, &
  -7.744392503445399D-07,   0.000000000000000D+00,   5.952732387986838D-07, &
   2.289069121751883D-05,   0.000000000000000D+00,   3.074002080069076D-06, &
   0.000000000000000D+00,   1.004739298082887D-06,   0.000000000000000D+00, &
   4.502508305173536D-04,   0.000000000000000D+00,   1.067193825303932D-06, &
   0.000000000000000D+00,  -3.235711429529540D-06,   0.000000000000000D+00, &
   2.911877643713577D-07,  -6.008009877978910D-04,   0.000000000000000D+00, &
   7.261886047725315D-05,   0.000000000000000D+00,  -1.433293795933906D-06, &
   0.000000000000000D+00,   6.156830211002079D-07,   0.000000000000000D+00, &
  -4.106476486675716D-04,   0.000000000000000D+00,  -2.920587698148610D-04, &
   0.000000000000000D+00,   2.370818560960239D-05,   0.000000000000000D+00, &
   2.799596360830171D-06,   0.000000000000000D+00,  -1.700646459966990D-06, &
   1.279072359466758D-03,   0.000000000000000D+00,   2.742364212615925D-04, &
   0.000000000000000D+00,  -1.108265429779933D-04,   0.000000000000000D+00, &
  -7.532500510294511D-06,   0.000000000000000D+00,  -1.807326883806141D-06, &
   0.000000000000000D+00,  -4.533626030810867D-04,   0.000000000000000D+00, &
   3.419858062123808D-04,   0.000000000000000D+00,   2.636923924127651D-04, &
   0.000000000000000D+00,  -1.806570464814566D-06,   0.000000000000000D+00, &
   2.372841784801263D-06,   0.000000000000000D+00,   1.468884742573550D-06, &
  -7.330192691033936D-04,   0.000000000000000D+00,  -8.276030235856244D-04, &
   0.000000000000000D+00,  -1.862384536300092D-04,   0.000000000000000D+00, &
   6.670994478461568D-05,   0.000000000000000D+00,   5.690169752225577D-06, &
   0.000000000000000D+00,   2.212144155338433D-07,   0.000000000000000D+00, &
   6.162174190933195D-04,   0.000000000000000D+00,   3.739386631665690D-04, &
   0.000000000000000D+00,  -3.379891614875666D-04,   0.000000000000000D+00, &
  -1.876920174504279D-04,   0.000000000000000D+00,  -1.780083350546280D-05, &
   0.000000000000000D+00,  -4.392305283795385D-06,   0.000000000000000D+00, &
   7.191270038379347D-07,   2.552493535055898D-05,   0.000000000000000D+00, &
   3.330553561877190D-04,   0.000000000000000D+00,   7.249247905022039D-04, &
   0.000000000000000D+00,   1.470507813089826D-04,   0.000000000000000D+00, &
  -2.800233706103397D-06,   0.000000000000000D+00,   4.519211538229654D-06, &
   0.000000000000000D+00,   2.706687717554071D-06,   0.000000000000000D+00, &
  -1.561846767760491D-04,   0.000000000000000D+00,  -3.902228128640278D-04, &
   0.000000000000000D+00,  -3.677785130024936D-04,   0.000000000000000D+00, &
   2.184300020156041D-04,   0.000000000000000D+00,   8.627051996030387D-05, &
   0.000000000000000D+00,   1.023904644287294D-05,   0.000000000000000D+00, &
  -8.934546094787676D-07,   0.000000000000000D+00,  -3.966845013332533D-06, &
   2.370231051393785D-05,   0.000000000000000D+00,   1.081223155499791D-04, &
   0.000000000000000D+00,  -2.234165496947115D-04,   0.000000000000000D+00, &
  -4.917811543978935D-04,   0.000000000000000D+00,  -1.052900452272084D-04, &
   0.000000000000000D+00,  -2.238674832294909D-05,   0.000000000000000D+00, &
  -6.850942187518968D-06,   0.000000000000000D+00,  -1.510664431251173D-06, &
   0.000000000000000D+00,   1.665656489610241D-05,   0.000000000000000D+00, &
  -1.133203679694510D-06,   0.000000000000000D+00,   3.451749155042537D-04, &
   0.000000000000000D+00,   2.375815448387929D-04,   0.000000000000000D+00, &
  -7.890020747787717D-05,   0.000000000000000D+00,  -1.479781010773593D-05, &
   0.000000000000000D+00,   4.522064506163154D-06,   0.000000000000000D+00, &
   4.615963692712065D-06,   0.000000000000000D+00,   3.149347768350607D-06, &
  -4.948242724515414D-06,   0.000000000000000D+00,   1.048964154376532D-05, &
   0.000000000000000D+00,  -1.627937305506170D-04,   0.000000000000000D+00, &
   1.737839892459658D-04,   0.000000000000000D+00,   2.327577392235544D-04, &
   0.000000000000000D+00,   5.464018909126182D-05,   0.000000000000000D+00, &
   1.317836339365469D-05,   0.000000000000000D+00,   1.059102638665090D-06, &
   0.000000000000000D+00,  -1.591718413441436D-06,   0.000000000000000D+00, &
   3.619929650116304D-07,   0.000000000000000D+00,  -4.460021540421970D-06, &
   0.000000000000000D+00,   4.682416926131373D-05,   0.000000000000000D+00, &
  -2.433407397838047D-04,   0.000000000000000D+00,  -8.617532767602917D-05, &
   0.000000000000000D+00,   8.297182942004439D-06,   0.000000000000000D+00, &
  -8.988955705252805D-06,   0.000000000000000D+00,  -7.277003678068085D-06, &
   0.000000000000000D+00,  -1.989883080139868D-06,   0.000000000000000D+00, &
   1.903276050083151D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.878435371578486D-06,   0.000000000000000D+00,  -1.098548073433846D-05, &
   0.000000000000000D+00,   1.004593628685650D-04,   0.000000000000000D+00, &
  -1.234859093334541D-04,   0.000000000000000D+00,  -7.735945574087309D-05, &
   0.000000000000000D+00,  -1.860269656548919D-05,   0.000000000000000D+00, &
  -1.250426371822419D-06,   0.000000000000000D+00,   2.542055740721383D-06, &
   0.000000000000000D+00,   1.500348856849344D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.607833162663034D-06, &
   0.000000000000000D+00,  -4.932364754042550D-06,   0.000000000000000D+00, &
  -1.240358294024620D-05,   0.000000000000000D+00,   1.121864033993638D-04, &
   0.000000000000000D+00,   9.320310309623323D-06,   0.000000000000000D+00, &
   5.431444128102040D-06,   0.000000000000000D+00,   7.720402279393862D-06, &
   0.000000000000000D+00,   2.656345011872785D-06,   0.000000000000000D+00, &
  -1.280674206227470D-06,   0.000000000000000D+00,  -3.762471067791737D-06 /)
lem(:,166) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.721203040336519D-05,   1.610306186062914D-04, &
   8.410215034287829D-04,  -1.453516235731943D-03,  -2.361848989779803D-03, &
   1.605429536690057D-03,   8.217291793209982D-03,  -1.296510843121546D-03, &
  -1.644910658963559D-02,   3.226638811815479D-03,   1.788980594339996D-02, &
  -5.554513632505744D-03,  -1.052130036696053D-02,   4.431039113716207D-03, &
   3.278488466224246D-03,  -1.809254922973495D-03,  -7.104550219254789D-04, &
   5.789920453074556D-04,   2.881582420402634D-04,  -2.706170335256700D-04, &
  -1.650231808578599D-04,   0.000000000000000D+00,   7.721203040336320D-05, &
  -1.610306186062874D-04,  -8.410215034287848D-04,   1.453516235731945D-03, &
   2.361848989779810D-03,  -1.605429536690062D-03,  -8.217291793209954D-03, &
   1.296510843121564D-03,   1.644910658963558D-02,  -3.226638811815494D-03, &
  -1.788980594339995D-02,   5.554513632505770D-03,   1.052130036696052D-02, &
  -4.431039113716212D-03,  -3.278488466224234D-03,   1.809254922973492D-03, &
   7.104550219254686D-04,  -5.789920453074650D-04,  -2.881582420402500D-04, &
   2.706170335256770D-04,   1.650231808578439D-04,   0.000000000000000D+00, &
   3.860601520167883D-05,   0.000000000000000D+00,   1.139562688617955D-04, &
   0.000000000000000D+00,   4.210630121943965D-05,   0.000000000000000D+00, &
  -2.081144931162051D-05,   0.000000000000000D+00,   3.238067680026591D-05, &
   0.000000000000000D+00,  -3.549333382432435D-05,   0.000000000000000D+00, &
  -9.226382292670143D-05,   0.000000000000000D+00,   3.840077175108654D-05, &
   0.000000000000000D+00,   6.851865103450766D-05,   0.000000000000000D+00, &
  -2.449142368303905D-05,   0.000000000000000D+00,  -1.576771354478957D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.052073485510072D-06, &
   0.000000000000000D+00,   9.450473846759476D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.515727134922170D-05,   0.000000000000000D+00, &
  -7.601545914238700D-07,   0.000000000000000D+00,  -2.526974528589816D-05, &
   0.000000000000000D+00,  -8.653206846100751D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.975063538934834D-05,   0.000000000000000D+00, &
  -8.807158364635684D-06,   0.000000000000000D+00,   2.798796272850553D-06, &
   0.000000000000000D+00,   1.587500308727931D-04,   0.000000000000000D+00, &
   9.560650443147575D-06,   0.000000000000000D+00,  -6.228102729144720D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.227833203873044D-04, &
   0.000000000000000D+00,   7.706660646154523D-05,   0.000000000000000D+00, &
   3.159746942557431D-06,   0.000000000000000D+00,   1.581432499856188D-06, &
   0.000000000000000D+00,  -1.544380000430616D-04,   0.000000000000000D+00, &
  -2.184264113864330D-04,   0.000000000000000D+00,   3.864852757284466D-06, &
   0.000000000000000D+00,   7.669385787443968D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.507680642894001D-04,   0.000000000000000D+00, &
   1.438383597826319D-04,   0.000000000000000D+00,  -8.660737713669861D-05, &
   0.000000000000000D+00,  -6.068321025435816D-06,   0.000000000000000D+00, &
  -5.172987143847849D-06,   0.000000000000000D+00,  -1.712325585799065D-04, &
   0.000000000000000D+00,   2.502661803177953D-04,   0.000000000000000D+00, &
   2.161165022863000D-04,   0.000000000000000D+00,   7.436333483346484D-06, &
   0.000000000000000D+00,   2.432511908578300D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.479226127679767D-04,   0.000000000000000D+00, &
  -5.226724612992198D-04,   0.000000000000000D+00,  -1.331228542898876D-04, &
   0.000000000000000D+00,   4.753860941041026D-05,   0.000000000000000D+00, &
   3.614146374641651D-06,   0.000000000000000D+00,   1.771773797070893D-07, &
   0.000000000000000D+00,   2.191919177592676D-04,   0.000000000000000D+00, &
   2.747574463569521D-04,   0.000000000000000D+00,  -2.652078533134717D-04, &
   0.000000000000000D+00,  -1.557504773501397D-04,   0.000000000000000D+00, &
  -1.770114199142554D-05,   0.000000000000000D+00,  -3.006448165904648D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.173652426428013D-05, &
   0.000000000000000D+00,   1.726552771828700D-04,   0.000000000000000D+00, &
   5.387472724498874D-04,   0.000000000000000D+00,   1.169975171533444D-04, &
   0.000000000000000D+00,   2.474189932679549D-06,   0.000000000000000D+00, &
   5.777238076705588D-06,   0.000000000000000D+00,   6.634715238406958D-06, &
   0.000000000000000D+00,  -3.622366004979268D-05,   0.000000000000000D+00, &
  -2.953152150003184D-04,   0.000000000000000D+00,  -2.892886446615076D-04, &
   0.000000000000000D+00,   1.752828703536348D-04,   0.000000000000000D+00, &
   7.213335708563620D-05,   0.000000000000000D+00,   7.782337883029643D-06, &
   0.000000000000000D+00,  -1.976180407849846D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.321294727099873D-05,   0.000000000000000D+00, &
   1.244180772147486D-04,   0.000000000000000D+00,  -1.579803430386101D-04, &
   0.000000000000000D+00,  -3.839724566945816D-04,   0.000000000000000D+00, &
  -8.665206406552048D-05,   0.000000000000000D+00,  -2.106837605139942D-05, &
   0.000000000000000D+00,  -7.719402444627299D-06,   0.000000000000000D+00, &
  -3.319503162110522D-06,   0.000000000000000D+00,  -4.260642468079284D-06, &
   0.000000000000000D+00,   9.313510509613099D-06,   0.000000000000000D+00, &
   2.854018854748742D-04,   0.000000000000000D+00,   1.909570075821798D-04, &
   0.000000000000000D+00,  -6.431671072738877D-05,   0.000000000000000D+00, &
  -1.215900843371171D-05,   0.000000000000000D+00,   5.390332984603721D-06, &
   0.000000000000000D+00,   4.259318061527114D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.061810379639237D-06,   0.000000000000000D+00, &
  -1.628966668315447D-05,   0.000000000000000D+00,  -1.327370033742221D-04, &
   0.000000000000000D+00,   1.367072759655286D-04,   0.000000000000000D+00, &
   1.854672819814874D-04,   0.000000000000000D+00,   4.534185147470755D-05, &
   0.000000000000000D+00,   1.231272551480222D-05,   0.000000000000000D+00, &
   1.293744915365217D-06,   0.000000000000000D+00,  -4.596632214483711D-06, &
   0.000000000000000D+00,   1.880969648359458D-06,   0.000000000000000D+00, &
  -3.664308476192714D-06,   0.000000000000000D+00,   1.446207077537008D-05, &
   0.000000000000000D+00,  -2.024019943449647D-04,   0.000000000000000D+00, &
  -7.044139784614878D-05,   0.000000000000000D+00,   7.076994903198347D-06, &
   0.000000000000000D+00,  -7.980289262914372D-06,   0.000000000000000D+00, &
  -6.529947514682244D-06,   0.000000000000000D+00,  -1.092037719436197D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.027093577745238D-07, &
   0.000000000000000D+00,   4.066685882480669D-06,   0.000000000000000D+00, &
  -1.786797417654831D-06,   0.000000000000000D+00,   7.438994133078912D-05, &
   0.000000000000000D+00,  -1.004890209114343D-04,   0.000000000000000D+00, &
  -6.199423243983808D-05,   0.000000000000000D+00,  -1.550506022079550D-05, &
   0.000000000000000D+00,  -1.673586988758205D-06,   0.000000000000000D+00, &
   3.053973783879094D-06,   0.000000000000000D+00,   4.831392444086751D-06, &
   0.000000000000000D+00,   1.697540025155638D-08,   0.000000000000000D+00, &
   2.481506551958836D-07,   0.000000000000000D+00,   6.070301974402063D-06, &
   0.000000000000000D+00,   1.631171598803688D-06,   0.000000000000000D+00, &
   9.388342773355987D-05,   0.000000000000000D+00,   8.055490605335467D-06, &
   0.000000000000000D+00,   4.294278989785562D-06,   0.000000000000000D+00, &
   6.367398863172071D-06,   0.000000000000000D+00,   1.955595596492756D-06, &
   0.000000000000000D+00,  -9.961282604063157D-07,   0.000000000000000D+00 /)
lem(:,167) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.288713432709463D-05,  -4.692152266322746D-04,   2.682390383124669D-04, &
   6.904193479388978D-04,   8.298755094046218D-04,   1.392079214957480D-03, &
  -2.830217925685061D-03,  -8.362264775795904D-03,   4.005602720605601D-03, &
   1.571319985413043D-02,  -5.818960909694595D-03,  -1.423492817979663D-02, &
   7.553945223322881D-03,   5.873874014342703D-03,  -5.747846727999339D-03, &
  -4.137773926973189D-04,   2.344908774813515D-03,  -4.504040007833963D-04, &
  -6.484246337983534D-04,   2.319525722835897D-04,   2.361595592316608D-04, &
  -1.351410995322684D-04,  -3.288713432709555D-05,  -4.692152266322730D-04, &
   2.682390383124621D-04,   6.904193479388955D-04,   8.298755094046287D-04, &
   1.392079214957475D-03,  -2.830217925685065D-03,  -8.362264775795899D-03, &
   4.005602720605609D-03,   1.571319985413049D-02,  -5.818960909694589D-03, &
  -1.423492817979664D-02,   7.553945223322874D-03,   5.873874014342704D-03, &
  -5.747846727999329D-03,  -4.137773926973170D-04,   2.344908774813516D-03, &
  -4.504040007833957D-04,  -6.484246337983626D-04,   2.319525722835897D-04, &
   2.361595592316693D-04,  -1.351410995322701D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.931988549689675D-05,   0.000000000000000D+00, &
   1.303203553177886D-04,   0.000000000000000D+00,  -4.957905136642281D-05, &
   0.000000000000000D+00,  -1.561364999503369D-04,   0.000000000000000D+00, &
   1.218617391381551D-05,   0.000000000000000D+00,   1.755046058189668D-04, &
   0.000000000000000D+00,   1.952777447807919D-05,   0.000000000000000D+00, &
  -1.332016133185535D-04,   0.000000000000000D+00,   3.435538328311300D-06, &
   0.000000000000000D+00,   4.201507143958421D-05,   0.000000000000000D+00, &
  -5.405026689856607D-06,   2.131189244192360D-05,   0.000000000000000D+00, &
   1.958932879985516D-05,   0.000000000000000D+00,   8.446361000541363D-07, &
  -8.703564237619985D-05,   0.000000000000000D+00,   1.670478644162790D-06, &
   0.000000000000000D+00,  -3.613879408905619D-05,   0.000000000000000D+00, &
  -1.399423038859705D-05,   0.000000000000000D+00,  -1.078751570384922D-06, &
   2.070901150942204D-04,   0.000000000000000D+00,   4.565094815296771D-06, &
   0.000000000000000D+00,  -8.824929072543942D-06,   0.000000000000000D+00, &
   2.236436425930257D-05,   0.000000000000000D+00,   9.081530662917790D-05, &
   0.000000000000000D+00,   5.753533632559193D-06,   0.000000000000000D+00, &
  -1.253315689877574D-06,  -2.986215309122160D-04,   0.000000000000000D+00, &
  -1.319057334545623D-04,   0.000000000000000D+00,   4.998216387650955D-05, &
   0.000000000000000D+00,   1.049667941399244D-05,   0.000000000000000D+00, &
   2.667317800579324D-05,   0.000000000000000D+00,  -9.163289635089346D-05, &
   0.000000000000000D+00,  -1.543531369240792D-04,   0.000000000000000D+00, &
  -1.428314804315916D-06,   0.000000000000000D+00,   3.071999269301062D-06, &
   2.665624746322010D-04,   0.000000000000000D+00,   2.938411926835674D-04, &
   0.000000000000000D+00,   1.225940035855684D-04,   0.000000000000000D+00, &
  -6.776606861877222D-05,   0.000000000000000D+00,  -1.123764359488984D-05, &
   0.000000000000000D+00,  -6.199447471071659D-05,   0.000000000000000D+00, &
  -1.021877582774766D-04,   0.000000000000000D+00,   1.821998255254736D-04, &
   0.000000000000000D+00,   1.706392818149322D-04,   0.000000000000000D+00, &
   1.142531864432103D-05,   0.000000000000000D+00,   4.115078843951765D-06, &
  -1.427330426917326D-04,   0.000000000000000D+00,  -1.638232714809364D-04, &
   0.000000000000000D+00,  -4.122995505836101D-04,   0.000000000000000D+00, &
  -1.128249204281203D-04,   0.000000000000000D+00,   3.804465924695407D-05, &
   0.000000000000000D+00,   2.802080502769118D-06,   0.000000000000000D+00, &
   5.043168749189940D-05,   0.000000000000000D+00,   1.274574100468044D-04, &
   0.000000000000000D+00,   1.998851705939908D-04,   0.000000000000000D+00, &
  -2.082301758175900D-04,   0.000000000000000D+00,  -1.289374588206117D-04, &
   0.000000000000000D+00,  -1.771644645551904D-05,   0.000000000000000D+00, &
  -8.116074532411968D-06,   4.029923206278635D-05,   0.000000000000000D+00, &
   2.736254705700825D-06,   0.000000000000000D+00,   1.470507813089826D-04, &
   0.000000000000000D+00,   4.366121482631863D-04,   0.000000000000000D+00, &
   1.008841124751131D-04,   0.000000000000000D+00,   4.862572647822641D-06, &
   0.000000000000000D+00,   8.521349055656026D-06,   0.000000000000000D+00, &
  -1.971433991549793D-05,   0.000000000000000D+00,  -1.691894950413842D-05, &
   0.000000000000000D+00,  -2.057537419804512D-04,   0.000000000000000D+00, &
  -2.264396392698717D-04,   0.000000000000000D+00,   1.417768156137391D-04, &
   0.000000000000000D+00,   5.976969387622576D-05,   0.000000000000000D+00, &
   7.451442911610372D-06,   0.000000000000000D+00,  -2.392558799625183D-06, &
  -3.366828145899136D-06,   0.000000000000000D+00,  -5.516119394493237D-07, &
   0.000000000000000D+00,   8.350034277978468D-05,   0.000000000000000D+00, &
  -1.315695643407012D-04,   0.000000000000000D+00,  -3.183634911835009D-04, &
   0.000000000000000D+00,  -7.542394732804663D-05,   0.000000000000000D+00, &
  -2.093774125638237D-05,   0.000000000000000D+00,  -7.553613584037194D-06, &
   0.000000000000000D+00,   3.036359959391394D-06,   0.000000000000000D+00, &
  -3.331858289556859D-06,   0.000000000000000D+00,  -8.643341367855953D-06, &
   0.000000000000000D+00,   2.199740209250965D-04,   0.000000000000000D+00, &
   1.537473234836114D-04,   0.000000000000000D+00,  -5.195962116696766D-05, &
   0.000000000000000D+00,  -8.802071815476981D-06,   0.000000000000000D+00, &
   4.904135573562567D-06,   0.000000000000000D+00,   1.106441872426132D-05, &
  -5.396889234255248D-07,   0.000000000000000D+00,   3.286579611373572D-06, &
   0.000000000000000D+00,  -8.084060452667724D-06,   0.000000000000000D+00, &
  -1.064475855371683D-04,   0.000000000000000D+00,   1.153214701531801D-04, &
   0.000000000000000D+00,   1.562845070225248D-04,   0.000000000000000D+00, &
   3.999486513400433D-05,   0.000000000000000D+00,   1.142559991709096D-05, &
   0.000000000000000D+00,  -2.912059330990458D-07,   0.000000000000000D+00, &
   3.066293350707587D-08,   0.000000000000000D+00,   4.174027578159681D-07, &
   0.000000000000000D+00,   6.633090982621226D-06,   0.000000000000000D+00, &
   1.740573531054120D-05,   0.000000000000000D+00,  -1.630246777852102D-04, &
   0.000000000000000D+00,  -5.685990454145788D-05,   0.000000000000000D+00, &
   4.688992400504394D-06,   0.000000000000000D+00,  -7.342009783728861D-06, &
   0.000000000000000D+00,  -6.779060749722217D-06,   0.000000000000000D+00, &
  -5.212809560292718D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.483717340239765D-07,   0.000000000000000D+00,   2.423935843760142D-07, &
   0.000000000000000D+00,   5.140953852883357D-06,   0.000000000000000D+00, &
   6.168660606223377D-05,   0.000000000000000D+00,  -8.553375478635304D-05, &
   0.000000000000000D+00,  -5.311985243432893D-05,   0.000000000000000D+00, &
  -1.392265423135514D-05,   0.000000000000000D+00,  -5.657476653077574D-07, &
   0.000000000000000D+00,   2.473163224616074D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.531981316568740D-07, &
   0.000000000000000D+00,  -2.264282385873360D-06,   0.000000000000000D+00, &
  -1.377337133702468D-06,   0.000000000000000D+00,   2.110120852615431D-06, &
   0.000000000000000D+00,   7.661800304396016D-05,   0.000000000000000D+00, &
   5.879238946655946D-06,   0.000000000000000D+00,   3.998879626170358D-06, &
   0.000000000000000D+00,   5.490655058269340D-06,   0.000000000000000D+00, &
   2.822123384024537D-06,   0.000000000000000D+00,  -3.128121039980161D-06 /)
lem(:,168) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.539534890956474D-04,   9.922318959200761D-04, &
  -7.077137187170415D-04,  -2.456647562485214D-03,  -2.094380767249089D-03, &
   3.670073738017142D-03,   7.793900410095886D-03,  -4.390053057136098D-03, &
  -1.188983493164902D-02,   6.562222884373915D-03,   8.647410906001411D-03, &
  -7.499699830233309D-03,  -1.429586144093413D-03,   4.411294903211747D-03, &
  -1.876739424737398D-03,  -1.009974278325739D-03,   1.275606667619852D-03, &
  -6.724860117385964D-05,  -4.448277344005555D-04,   1.132654568254134D-04, &
   1.677106917869277D-04,   0.000000000000000D+00,  -4.539534890956431D-04, &
  -9.922318959200824D-04,   7.077137187170384D-04,   2.456647562485217D-03, &
   2.094380767249086D-03,  -3.670073738017154D-03,  -7.793900410095864D-03, &
   4.390053057136100D-03,   1.188983493164902D-02,  -6.562222884373919D-03, &
  -8.647410906001395D-03,   7.499699830233316D-03,   1.429586144093410D-03, &
  -4.411294903211760D-03,   1.876739424737398D-03,   1.009974278325743D-03, &
  -1.275606667619859D-03,   6.724860117385904D-05,   4.448277344005672D-04, &
  -1.132654568254238D-04,  -1.677106917869397D-04,   0.000000000000000D+00, &
  -2.269767445478215D-04,   0.000000000000000D+00,  -4.856935461410069D-04, &
   0.000000000000000D+00,  -1.277823746956599D-04,   0.000000000000000D+00, &
   2.935654087958578D-04,   0.000000000000000D+00,   2.388437336563566D-04, &
   0.000000000000000D+00,  -2.651219596154639D-04,   0.000000000000000D+00, &
  -2.117919777385765D-04,   0.000000000000000D+00,   2.092063700989327D-04, &
   0.000000000000000D+00,   7.247816324885902D-05,   0.000000000000000D+00, &
  -8.099128050135821D-05,   0.000000000000000D+00,   7.042164258159239D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.794408721653306D-05, &
   0.000000000000000D+00,   1.538279402056491D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.562904072191997D-05,   0.000000000000000D+00, &
   1.404415587852072D-06,   0.000000000000000D+00,  -2.787452435790026D-05, &
   0.000000000000000D+00,  -2.280512800939500D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.041935242534246D-05,   0.000000000000000D+00, &
   5.729057944194548D-06,   0.000000000000000D+00,  -2.092336631442248D-05, &
   0.000000000000000D+00,   1.760212231843672D-05,   0.000000000000000D+00, &
   7.705397625810862D-05,   0.000000000000000D+00,   1.678935318273128D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.034977507349289D-05, &
   0.000000000000000D+00,  -9.038747891554463D-05,   0.000000000000000D+00, &
   3.585085609237170D-05,   0.000000000000000D+00,   2.564284453039676D-05, &
   0.000000000000000D+00,   2.109267786848882D-05,   0.000000000000000D+00, &
  -7.127379050333015D-05,   0.000000000000000D+00,  -1.316539017309434D-04, &
   0.000000000000000D+00,  -1.105576027177604D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.578841119694536D-05,   0.000000000000000D+00, &
   2.004065070323087D-04,   0.000000000000000D+00,   1.009407923401885D-04, &
   0.000000000000000D+00,  -5.393576476599952D-05,   0.000000000000000D+00, &
  -2.328430570393958D-05,   0.000000000000000D+00,  -4.819874600058017D-05, &
   0.000000000000000D+00,  -8.002596214609608D-05,   0.000000000000000D+00, &
   1.471681246976150D-04,   0.000000000000000D+00,   1.494588786492193D-04, &
   0.000000000000000D+00,   1.970136616865215D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.966676833884111D-05,   0.000000000000000D+00, &
  -1.086098761948820D-04,   0.000000000000000D+00,  -3.248222434878284D-04, &
   0.000000000000000D+00,  -9.945966655710289D-05,   0.000000000000000D+00, &
   2.783060220980732D-05,   0.000000000000000D+00,   6.616743135223183D-06, &
   0.000000000000000D+00,   3.719281005473126D-05,   0.000000000000000D+00, &
   1.096208912474864D-04,   0.000000000000000D+00,   1.604430329581977D-04, &
   0.000000000000000D+00,  -1.740140859392810D-04,   0.000000000000000D+00, &
  -1.152818782428697D-04,   0.000000000000000D+00,  -2.237205335608970D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.133125141114456D-06, &
   0.000000000000000D+00,  -3.635688015320592D-06,   0.000000000000000D+00, &
   1.169975171533444D-04,   0.000000000000000D+00,   3.623555546957561D-04, &
   0.000000000000000D+00,   9.273605064097732D-05,   0.000000000000000D+00, &
   1.011526826590354D-05,   0.000000000000000D+00,   1.741908898247055D-05, &
   0.000000000000000D+00,  -1.256924046105279D-05,   0.000000000000000D+00, &
  -3.052836081705153D-05,   0.000000000000000D+00,  -1.705560269291402D-04, &
   0.000000000000000D+00,  -1.873667122024967D-04,   0.000000000000000D+00, &
   1.202246033850720D-04,   0.000000000000000D+00,   5.435353421845503D-05, &
   0.000000000000000D+00,   6.274026100430684D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.870606478555527D-06,   0.000000000000000D+00, &
   4.213619880241541D-06,   0.000000000000000D+00,   6.087726793778555D-05, &
   0.000000000000000D+00,  -1.117548701367608D-04,   0.000000000000000D+00, &
  -2.715775311336025D-04,   0.000000000000000D+00,  -7.096979305170024D-05, &
   0.000000000000000D+00,  -2.236973260237829D-05,   0.000000000000000D+00, &
  -1.958395033260835D-05,   0.000000000000000D+00,   9.031078781426744D-07, &
   0.000000000000000D+00,   6.763632317583393D-06,   0.000000000000000D+00, &
   3.170509307191083D-06,   0.000000000000000D+00,   1.848200975469974D-04, &
   0.000000000000000D+00,   1.297938000679667D-04,   0.000000000000000D+00, &
  -4.408537227763391D-05,   0.000000000000000D+00,  -7.603648051882244D-06, &
   0.000000000000000D+00,   6.994697524979545D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.090567834685028D-06,   0.000000000000000D+00, &
  -3.691645301087503D-07,   0.000000000000000D+00,  -2.504913487141662D-06, &
   0.000000000000000D+00,  -8.215895198345082D-05,   0.000000000000000D+00, &
   1.007939855441580D-04,   0.000000000000000D+00,   1.360339689929573D-04, &
   0.000000000000000D+00,   3.738595956100998D-05,   0.000000000000000D+00, &
   1.224698674979757D-05,   0.000000000000000D+00,   1.047981303188809D-06, &
   0.000000000000000D+00,   3.031174457344950D-07,   0.000000000000000D+00, &
  -2.038745637233536D-06,   0.000000000000000D+00,   6.150061325869208D-08, &
   0.000000000000000D+00,   9.926997561993735D-06,   0.000000000000000D+00, &
  -1.392306025414869D-04,   0.000000000000000D+00,  -4.853173148260313D-05, &
   0.000000000000000D+00,   4.145797725124652D-06,   0.000000000000000D+00, &
  -7.860420211038035D-06,   0.000000000000000D+00,  -5.298858356000283D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.669068966627076D-08, &
   0.000000000000000D+00,   4.012317937685847D-07,   0.000000000000000D+00, &
  -2.674845178769629D-06,   0.000000000000000D+00,   1.255945014856946D-06, &
   0.000000000000000D+00,   4.779361312832277D-05,   0.000000000000000D+00, &
  -7.567297100655206D-05,   0.000000000000000D+00,  -4.706049035490309D-05, &
   0.000000000000000D+00,  -1.244698685761121D-05,   0.000000000000000D+00, &
  -1.231738014246962D-06,   0.000000000000000D+00,   9.861980149449433D-06, &
   0.000000000000000D+00,   1.675776691303188D-09,   0.000000000000000D+00, &
   3.394358519428652D-07,   0.000000000000000D+00,   8.094899555285661D-07, &
   0.000000000000000D+00,  -1.788099484354864D-06,   0.000000000000000D+00, &
   4.023780662664383D-06,   0.000000000000000D+00,   6.599321603347865D-05, &
   0.000000000000000D+00,   4.596504221212215D-06,   0.000000000000000D+00, &
   3.333837602713509D-06,   0.000000000000000D+00,   5.382195852507861D-06, &
   0.000000000000000D+00,   1.054255063498632D-06,   0.000000000000000D+00 /)
lem(:,169) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -6.663768088139803D-04,  -3.124218721996575D-04,  -4.660273484109966D-05, &
   9.922301379769200D-04,   2.203855025062344D-03,   1.507636965098310D-03, &
  -3.415149117934273D-03,  -6.223338348897925D-03,   3.375115924437100D-03, &
   8.049659113809293D-03,  -4.938815458174047D-03,  -3.877666262795568D-03, &
   5.229038520215082D-03,  -1.310680132926790D-03,  -1.979297425182824D-03, &
   2.177764038815586D-03,  -5.707370727845071D-04,  -7.244968233331056D-04, &
   6.917693170473958D-04,  -2.917936266950720D-06,  -2.849512423884104D-04, &
   8.023357120906199D-05,  -6.663768088139968D-04,  -3.124218721996547D-04, &
  -4.660273484109421D-05,   9.922301379769152D-04,   2.203855025062341D-03, &
   1.507636965098317D-03,  -3.415149117934274D-03,  -6.223338348897920D-03, &
   3.375115924437105D-03,   8.049659113809296D-03,  -4.938815458174060D-03, &
  -3.877666262795568D-03,   5.229038520215089D-03,  -1.310680132926790D-03, &
  -1.979297425182829D-03,   2.177764038815587D-03,  -5.707370727845038D-04, &
  -7.244968233331075D-04,   6.917693170473909D-04,  -2.917936266948315D-06, &
  -2.849512423884019D-04,   8.023357120905768D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.837824171736012D-04,   0.000000000000000D+00, &
   2.552635855852642D-04,   0.000000000000000D+00,   1.027915094354835D-03, &
   0.000000000000000D+00,  -1.193649881968796D-03,   0.000000000000000D+00, &
   1.870578144589227D-05,   0.000000000000000D+00,   7.073960988090158D-04, &
   0.000000000000000D+00,  -1.888884244605555D-04,   0.000000000000000D+00, &
  -2.782498811736517D-04,   0.000000000000000D+00,   1.219362841455837D-04, &
   0.000000000000000D+00,   2.265737133960794D-05,   0.000000000000000D+00, &
   3.586725541543655D-05,   2.560816645085228D-05,   0.000000000000000D+00, &
  -3.008007071474559D-05,   0.000000000000000D+00,   1.590066275994106D-05, &
  -4.422768064408935D-05,   0.000000000000000D+00,  -2.268310048761871D-05, &
   0.000000000000000D+00,   1.671766761491147D-05,   0.000000000000000D+00, &
  -1.004234707642597D-05,   0.000000000000000D+00,  -4.289905576118177D-05, &
   5.804823406241823D-05,   0.000000000000000D+00,   5.075309874150349D-05, &
   0.000000000000000D+00,   9.707245911805324D-06,   0.000000000000000D+00, &
   1.846376892122424D-05,   0.000000000000000D+00,   5.491442757719762D-06, &
   0.000000000000000D+00,   6.084306382007427D-05,   0.000000000000000D+00, &
   3.886825371013048D-05,  -6.275532820050081D-05,   0.000000000000000D+00, &
  -7.368794314247724D-05,   0.000000000000000D+00,  -7.625232494555319D-05, &
   0.000000000000000D+00,   2.460123542847947D-05,   0.000000000000000D+00, &
  -4.020263889210732D-05,   0.000000000000000D+00,   6.356794570372770D-06, &
   0.000000000000000D+00,  -5.342157758403950D-05,   0.000000000000000D+00, &
  -1.166212066834641D-04,   0.000000000000000D+00,  -3.283829676652103D-05, &
   5.424644068913961D-05,   0.000000000000000D+00,   6.367846110246041D-05, &
   0.000000000000000D+00,   1.614591326736208D-04,   0.000000000000000D+00, &
   9.221269538355264D-05,   0.000000000000000D+00,  -4.292484635097513D-05, &
   0.000000000000000D+00,   2.832326921745565D-05,   0.000000000000000D+00, &
  -1.412964325396884D-05,   0.000000000000000D+00,  -5.820220652865165D-05, &
   0.000000000000000D+00,   1.218654176663793D-04,   0.000000000000000D+00, &
   1.420032733962687D-04,   0.000000000000000D+00,   4.902111769061728D-05, &
  -3.263502547638064D-05,   0.000000000000000D+00,  -2.898404862351588D-05, &
   0.000000000000000D+00,  -8.903295867909654D-05,   0.000000000000000D+00, &
  -2.747919166630959D-04,   0.000000000000000D+00,  -9.786381729560714D-05, &
   0.000000000000000D+00,   1.731295007768326D-05,   0.000000000000000D+00, &
  -7.277720524527077D-06,   0.000000000000000D+00,   8.412664397637067D-06, &
   0.000000000000000D+00,   7.817888578916733D-05,   0.000000000000000D+00, &
   1.308091110388809D-04,   0.000000000000000D+00,  -1.506923099251123D-04, &
   0.000000000000000D+00,  -1.141619893612672D-04,   0.000000000000000D+00, &
  -5.483676842919187D-05,   1.131657952231370D-05,   0.000000000000000D+00, &
   4.293690818018065D-06,   0.000000000000000D+00,  -2.800233706103397D-06, &
   0.000000000000000D+00,   1.008841124751131D-04,   0.000000000000000D+00, &
   3.188427999703754D-04,   0.000000000000000D+00,   9.384322032602457D-05, &
   0.000000000000000D+00,   1.983420605506433D-05,   0.000000000000000D+00, &
  -7.925130589511055D-07,   0.000000000000000D+00,   6.842699758503816D-08, &
   0.000000000000000D+00,  -1.841302069917898D-05,   0.000000000000000D+00, &
  -1.388955009740331D-04,   0.000000000000000D+00,  -1.602048515182998D-04, &
   0.000000000000000D+00,   1.071673989955909D-04,   0.000000000000000D+00, &
   5.482594733957815D-05,   0.000000000000000D+00,   2.085064401235581D-05, &
  -1.537618656757391D-06,   0.000000000000000D+00,   7.189912972348793D-07, &
   0.000000000000000D+00,   7.738798795285711D-06,   0.000000000000000D+00, &
   4.935664578438932D-05,   0.000000000000000D+00,  -1.004725447922547D-04, &
   0.000000000000000D+00,  -2.457323956933252D-04,   0.000000000000000D+00, &
  -7.169694232669711D-05,   0.000000000000000D+00,  -2.498202531190891D-05, &
   0.000000000000000D+00,   5.306074562546438D-07,   0.000000000000000D+00, &
  -1.725916069520274D-06,   0.000000000000000D+00,   8.465744871828305D-07, &
   0.000000000000000D+00,   4.509963320888085D-06,   0.000000000000000D+00, &
   1.583128580887407D-04,   0.000000000000000D+00,   1.130907064849989D-04, &
   0.000000000000000D+00,  -3.963893556576631D-05,   0.000000000000000D+00, &
  -8.064143530421671D-06,   0.000000000000000D+00,   1.778161998553770D-05, &
  -6.827549722587220D-08,   0.000000000000000D+00,   2.484527330795823D-07, &
   0.000000000000000D+00,  -3.139749243396192D-06,   0.000000000000000D+00, &
  -7.080958064506108D-06,   0.000000000000000D+00,  -6.798148477487473D-05, &
   0.000000000000000D+00,   9.318643739626760D-05,   0.000000000000000D+00, &
   1.250499705151469D-04,   0.000000000000000D+00,   3.738766289003607D-05, &
   0.000000000000000D+00,   8.259991923563529D-06,   0.000000000000000D+00, &
   3.166135507219896D-09,   0.000000000000000D+00,   3.831459473688856D-07, &
   0.000000000000000D+00,   8.064690275619453D-07,   0.000000000000000D+00, &
  -1.348542351445901D-07,   0.000000000000000D+00,   4.314301876188870D-06, &
   0.000000000000000D+00,  -1.223973133559024D-04,   0.000000000000000D+00, &
  -4.157274217652753D-05,   0.000000000000000D+00,   3.141361068189126D-06, &
   0.000000000000000D+00,  -9.002618010120568D-06,   0.000000000000000D+00, &
  -2.247920506748467D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.958313172042737D-07,   0.000000000000000D+00,   9.177151433022941D-07, &
   0.000000000000000D+00,   1.084860137326866D-06,   0.000000000000000D+00, &
   4.431087397893792D-06,   0.000000000000000D+00,   3.905782588543808D-05, &
   0.000000000000000D+00,  -7.055712236691487D-05,   0.000000000000000D+00, &
  -4.347328868975921D-05,   0.000000000000000D+00,  -1.141554109391776D-05, &
   0.000000000000000D+00,   2.163861465082152D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.996804457869507D-08, &
   0.000000000000000D+00,  -4.653729762904867D-07,   0.000000000000000D+00, &
   9.463565375029841D-07,   0.000000000000000D+00,  -2.493488132243861D-07, &
   0.000000000000000D+00,   7.235591562974228D-06,   0.000000000000000D+00, &
   5.802353269977634D-05,   0.000000000000000D+00,   2.988343660728294D-06, &
   0.000000000000000D+00,   3.422867280519547D-06,   0.000000000000000D+00, &
   8.762629860204927D-06,   0.000000000000000D+00,   7.382523596173003D-06 /)
lem(:,170) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.928023201037039D-04,   5.573958585488964D-04, &
  -6.005969683717941D-04,  -1.900373508266228D-03,  -1.660941391791240D-03, &
   1.815587411874773D-03,   4.588304136545910D-03,  -1.579795285145763D-03, &
  -4.973158940699113D-03,   2.888299330628826D-03,   1.455351258575030D-03, &
  -2.551270370519291D-03,   1.671451492058808D-03,   6.266554598220412D-07, &
  -1.167763326391733D-03,   1.052714552635098D-03,  -2.050256532868958D-04, &
  -4.238551626450193D-04,   3.902765864457964D-04,  -1.378451132868308D-05, &
  -1.721381515174787D-04,   0.000000000000000D+00,  -4.928023201037057D-04, &
  -5.573958585488899D-04,   6.005969683717948D-04,   1.900373508266226D-03, &
   1.660941391791234D-03,  -1.815587411874772D-03,  -4.588304136545906D-03, &
   1.579795285145762D-03,   4.973158940699112D-03,  -2.888299330628832D-03, &
  -1.455351258575029D-03,   2.551270370519290D-03,  -1.671451492058808D-03, &
  -6.266554598222795D-07,   1.167763326391735D-03,  -1.052714552635101D-03, &
   2.050256532868987D-04,   4.238551626450236D-04,  -3.902765864458013D-04, &
   1.378451132867826D-05,   1.721381515174775D-04,   0.000000000000000D+00, &
  -2.464011600518544D-04,   0.000000000000000D+00,   6.611265304514115D-04, &
   0.000000000000000D+00,   1.129627625316454D-03,   0.000000000000000D+00, &
  -3.190504447397142D-03,   0.000000000000000D+00,   3.370015289492683D-03, &
   0.000000000000000D+00,  -1.591895757207888D-03,   0.000000000000000D+00, &
  -4.082686318717155D-04,   0.000000000000000D+00,   7.307034491272945D-04, &
   0.000000000000000D+00,  -8.986673713906255D-05,   0.000000000000000D+00, &
  -1.145059127306763D-04,   0.000000000000000D+00,   4.037214110850346D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.947972209678869D-05, &
   0.000000000000000D+00,  -1.486749261517071D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.609392875356496D-05,   0.000000000000000D+00, &
  -4.529537276126862D-05,   0.000000000000000D+00,   6.908324397219724D-07, &
   0.000000000000000D+00,  -5.492046836303644D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.788508595657937D-05,   0.000000000000000D+00, &
   3.959205235848074D-05,   0.000000000000000D+00,   2.848155451261740D-05, &
   0.000000000000000D+00,   4.470814782389368D-06,   0.000000000000000D+00, &
   4.965837483725566D-06,   0.000000000000000D+00,   4.942665780695475D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.100464114825062D-05, &
   0.000000000000000D+00,  -5.072711168937744D-05,   0.000000000000000D+00, &
  -7.106580507957288D-05,   0.000000000000000D+00,   4.348741617296036D-05, &
   0.000000000000000D+00,  -5.951346217322473D-06,   0.000000000000000D+00, &
   5.075232613374880D-06,   0.000000000000000D+00,  -4.313455266001349D-05, &
   0.000000000000000D+00,  -1.221984580501520D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.211429848792713D-05,   0.000000000000000D+00, &
   4.634350746951066D-05,   0.000000000000000D+00,   1.344205174876414D-04, &
   0.000000000000000D+00,   9.540971463134722D-05,   0.000000000000000D+00, &
  -8.956892642838677D-05,   0.000000000000000D+00,  -6.350605800541018D-07, &
   0.000000000000000D+00,  -1.472246244238813D-05,   0.000000000000000D+00, &
  -4.828987666212564D-05,   0.000000000000000D+00,   1.102159814161934D-04, &
   0.000000000000000D+00,   1.632793757795850D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.449342693393904D-05,   0.000000000000000D+00, &
  -2.567991567585529D-05,   0.000000000000000D+00,  -7.516414684106749D-05, &
   0.000000000000000D+00,  -2.487658159938892D-04,   0.000000000000000D+00, &
  -1.064365566932751D-04,   0.000000000000000D+00,   4.452394761033676D-05, &
   0.000000000000000D+00,   5.280208048269486D-06,   0.000000000000000D+00, &
   1.231947143753685D-05,   0.000000000000000D+00,   6.558511972732672D-05, &
   0.000000000000000D+00,   1.150462125877129D-04,   0.000000000000000D+00, &
  -1.437255972181094D-04,   0.000000000000000D+00,  -1.291823804800371D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.177357229357691D-06, &
   0.000000000000000D+00,   8.308536150604469D-06,   0.000000000000000D+00, &
   2.474189932679549D-06,   0.000000000000000D+00,   9.273605064097732D-05, &
   0.000000000000000D+00,   3.005001594330563D-04,   0.000000000000000D+00, &
   1.020742912710588D-04,   0.000000000000000D+00,   4.176004316552768D-05, &
   0.000000000000000D+00,  -3.279110899491910D-06,   0.000000000000000D+00, &
  -4.612272670857863D-06,   0.000000000000000D+00,  -1.762300970295542D-05, &
   0.000000000000000D+00,  -1.223099796146541D-04,   0.000000000000000D+00, &
  -1.445124622294988D-04,   0.000000000000000D+00,   1.044138018219134D-04, &
   0.000000000000000D+00,   5.131266344095261D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.882445286510723D-07,   0.000000000000000D+00, &
  -2.168154485292682D-06,   0.000000000000000D+00,   2.906157890866792D-06, &
   0.000000000000000D+00,   3.782454156390419D-05,   0.000000000000000D+00, &
  -9.754637534987110D-05,   0.000000000000000D+00,  -2.366641165964755D-04, &
   0.000000000000000D+00,  -7.783641413582491D-05,   0.000000000000000D+00, &
  -7.273294412638319D-05,   0.000000000000000D+00,   5.533798389478748D-07, &
   0.000000000000000D+00,   9.893105407002325D-07,   0.000000000000000D+00, &
   7.913947403377679D-07,   0.000000000000000D+00,   7.379833397070354D-06, &
   0.000000000000000D+00,   1.437979507008840D-04,   0.000000000000000D+00, &
   1.032273287523155D-04,   0.000000000000000D+00,  -3.738631819508166D-05, &
   0.000000000000000D+00,   1.059174169504457D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.489647120027733D-07,   0.000000000000000D+00, &
   7.692964612056563D-07,   0.000000000000000D+00,   9.379648830493821D-08, &
   0.000000000000000D+00,  -6.257182036659461D-06,   0.000000000000000D+00, &
  -5.534814081138659D-05,   0.000000000000000D+00,   9.169720756063272D-05, &
   0.000000000000000D+00,   1.219068770101146D-04,   0.000000000000000D+00, &
   4.427793442232718D-05,   0.000000000000000D+00,   3.770498347521970D-05, &
   0.000000000000000D+00,   4.889206882004981D-08,   0.000000000000000D+00, &
  -3.944863662876364D-07,   0.000000000000000D+00,   9.901165458067445D-07, &
   0.000000000000000D+00,   1.483757380289102D-06,   0.000000000000000D+00, &
  -7.684273888715403D-07,   0.000000000000000D+00,  -1.126154981060877D-04, &
   0.000000000000000D+00,  -3.787868184592714D-05,   0.000000000000000D+00, &
   1.018701268717363D-07,   0.000000000000000D+00,  -7.791270366499046D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.697628694423271D-09, &
   0.000000000000000D+00,  -9.906621879252237D-08,   0.000000000000000D+00, &
  -6.734369267143954D-07,   0.000000000000000D+00,   1.832633842858908D-07, &
   0.000000000000000D+00,   5.777610368685016D-06,   0.000000000000000D+00, &
   3.083350572323679D-05,   0.000000000000000D+00,  -6.903063431367237D-05, &
   0.000000000000000D+00,  -4.292964107673270D-05,   0.000000000000000D+00, &
  -1.824399254458148D-05,   0.000000000000000D+00,   5.037493179966933D-07, &
   0.000000000000000D+00,   1.798497713320260D-10,   0.000000000000000D+00, &
   1.154001378595954D-07,   0.000000000000000D+00,  -3.035989639674472D-07, &
   0.000000000000000D+00,  -7.116650957987761D-07,   0.000000000000000D+00, &
  -1.617557944030206D-06,   0.000000000000000D+00,   1.000869396983364D-05, &
   0.000000000000000D+00,   5.334722153403152D-05,   0.000000000000000D+00, &
   2.456445217266435D-06,   0.000000000000000D+00,   4.353819508711697D-06, &
   0.000000000000000D+00,   1.112969546651137D-06,   0.000000000000000D+00 /)
lem(:,171) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.493195937748820D-04,  -8.137061191529496D-04,  -1.928480098434906D-04, &
   5.007018916768700D-04,   1.126738673580804D-03,   1.737823315653702D-03, &
  -3.595698826583746D-04,  -3.452023157606505D-03,  -5.309510516820831D-05, &
   2.998196145767688D-03,  -1.181685216334520D-03,  -5.682732269305975D-04, &
   9.670630263111687D-04,  -9.010620571365022D-04,   5.433981302672163D-04, &
   1.192669038508025D-04,  -5.930175411448682D-04,   5.190993614301180D-04, &
  -1.018557091196591D-04,  -2.223619684831852D-04,   2.220481129310305D-04, &
  -2.669556219149184D-05,  -3.493195937748855D-04,  -8.137061191529560D-04, &
  -1.928480098434931D-04,   5.007018916768694D-04,   1.126738673580809D-03, &
   1.737823315653700D-03,  -3.595698826583743D-04,  -3.452023157606495D-03, &
  -5.309510516821198D-05,   2.998196145767681D-03,  -1.181685216334519D-03, &
  -5.682732269305956D-04,   9.670630263111721D-04,  -9.010620571365013D-04, &
   5.433981302672129D-04,   1.192669038508064D-04,  -5.930175411448693D-04, &
   5.190993614301110D-04,  -1.018557091196579D-04,  -2.223619684831807D-04, &
   2.220481129310280D-04,  -2.669556219149484D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.059958101968255D-04,   0.000000000000000D+00, &
   8.352283933174217D-04,   0.000000000000000D+00,  -3.453257809402685D-03, &
   0.000000000000000D+00,   4.777631338094803D-03,   0.000000000000000D+00, &
  -4.796600595944610D-03,   0.000000000000000D+00,   3.794101064465602D-03, &
   0.000000000000000D+00,  -1.380809003683616D-03,   0.000000000000000D+00, &
  -3.258581498241561D-04,   0.000000000000000D+00,   3.428255108447940D-04, &
   0.000000000000000D+00,  -6.139049080821160D-05,   0.000000000000000D+00, &
   8.525545634920556D-06,   1.317584325396843D-05,   0.000000000000000D+00, &
   4.522367198933768D-07,   0.000000000000000D+00,  -2.433335927917974D-05, &
  -1.766357661408940D-05,   0.000000000000000D+00,  -8.371576071846962D-06, &
   0.000000000000000D+00,  -1.077079483782731D-05,   0.000000000000000D+00, &
   6.308437155613394D-06,   0.000000000000000D+00,  -5.785369273098551D-06, &
   1.133008114069450D-05,   0.000000000000000D+00,   1.209557740156786D-05, &
   0.000000000000000D+00,   2.910055470699220D-05,   0.000000000000000D+00, &
   2.114624169162343D-05,   0.000000000000000D+00,   6.377105132037276D-06, &
   0.000000000000000D+00,   3.181466710853451D-06,   0.000000000000000D+00, &
   1.275351424224953D-04,  -1.046918234050199D-06,   0.000000000000000D+00, &
  -1.432694855561529D-05,   0.000000000000000D+00,  -4.208674142199139D-05, &
   0.000000000000000D+00,  -8.243564104772870D-05,   0.000000000000000D+00, &
  -2.348966241918463D-05,   0.000000000000000D+00,  -1.172663827410011D-05, &
   0.000000000000000D+00,   2.671281247959399D-06,   0.000000000000000D+00, &
  -4.775605331308692D-05,   0.000000000000000D+00,  -2.906113580058820D-04, &
  -1.326485608154397D-06,   0.000000000000000D+00,   1.267338046398571D-05, &
   0.000000000000000D+00,   3.824567232560690D-05,   0.000000000000000D+00, &
   1.302226594071422D-04,   0.000000000000000D+00,   1.439193330836288D-04, &
   0.000000000000000D+00,   1.551118364136671D-05,   0.000000000000000D+00, &
   6.605595525840014D-06,   0.000000000000000D+00,  -9.253317431204156D-06, &
   0.000000000000000D+00,  -4.202678460051083D-05,   0.000000000000000D+00, &
   1.167159037084532D-04,   0.000000000000000D+00,   3.828119375292877D-04, &
  -1.325031995633394D-06,   0.000000000000000D+00,  -6.246714947469236D-06, &
   0.000000000000000D+00,  -2.008486570530812D-05,   0.000000000000000D+00, &
  -7.138852836490810D-05,   0.000000000000000D+00,  -2.524670988256515D-04, &
   0.000000000000000D+00,  -1.783501688924865D-04,   0.000000000000000D+00, &
  -5.403497073265632D-06,   0.000000000000000D+00,  -7.093031526431956D-07, &
   0.000000000000000D+00,   9.052484606345399D-06,   0.000000000000000D+00, &
   5.768666766168368D-05,   0.000000000000000D+00,   1.075899444781548D-04, &
   0.000000000000000D+00,  -1.523069060714184D-04,   0.000000000000000D+00, &
  -3.319393842351641D-04,   1.629021230760643D-06,   0.000000000000000D+00, &
   5.355518786284881D-07,   0.000000000000000D+00,   4.519211538229654D-06, &
   0.000000000000000D+00,   4.862572647822641D-06,   0.000000000000000D+00, &
   9.384322032602457D-05,   0.000000000000000D+00,   3.139123962715284D-04, &
   0.000000000000000D+00,   1.640778451980881D-04,   0.000000000000000D+00, &
   5.825128239387568D-07,   0.000000000000000D+00,  -2.503106463777732D-07, &
   0.000000000000000D+00,  -4.377687974023204D-06,   0.000000000000000D+00, &
  -1.829365795301036D-05,   0.000000000000000D+00,  -1.135979174364575D-04, &
   0.000000000000000D+00,  -1.407652457256996D-04,   0.000000000000000D+00, &
   1.172457630190536D-04,   0.000000000000000D+00,   1.707660223332044D-04, &
  -4.126096967157585D-07,   0.000000000000000D+00,   7.488066482597765D-07, &
   0.000000000000000D+00,   5.847329050301942D-07,   0.000000000000000D+00, &
   3.738494858852575D-06,   0.000000000000000D+00,   2.929139309095698D-05, &
   0.000000000000000D+00,  -1.017207370202181D-04,   0.000000000000000D+00, &
  -2.546830315455607D-04,   0.000000000000000D+00,  -1.019245144327372D-04, &
   0.000000000000000D+00,   9.151196973852793D-08,   0.000000000000000D+00, &
  -2.654726609514886D-07,   0.000000000000000D+00,   1.254469345520339D-06, &
   0.000000000000000D+00,   2.506393378461505D-06,   0.000000000000000D+00, &
   1.197162020149809D-05,   0.000000000000000D+00,   1.382394019477321D-04, &
   0.000000000000000D+00,   1.029569308917073D-04,   0.000000000000000D+00, &
  -5.190884977364252D-05,   0.000000000000000D+00,  -1.866258663531467D-05, &
  -9.237790800758476D-09,   0.000000000000000D+00,  -1.473286177342732D-07, &
   0.000000000000000D+00,  -6.497019707037894D-07,   0.000000000000000D+00, &
  -8.295341303561286D-07,   0.000000000000000D+00,  -7.275755709281035D-06, &
   0.000000000000000D+00,  -4.687908158671962D-05,   0.000000000000000D+00, &
   9.695934498609987D-05,   0.000000000000000D+00,   1.344520637429826D-04, &
   0.000000000000000D+00,   3.461794709674705D-05,   0.000000000000000D+00, &
   3.607272553519103D-10,   0.000000000000000D+00,   1.268977283090137D-07, &
   0.000000000000000D+00,  -2.787350035878322D-07,   0.000000000000000D+00, &
  -3.303192064079006D-07,   0.000000000000000D+00,   4.739344556224158D-07, &
   0.000000000000000D+00,  -6.199926517862620D-06,   0.000000000000000D+00, &
  -1.109231568857014D-04,   0.000000000000000D+00,  -3.691705615772343D-05, &
   0.000000000000000D+00,   7.456437637109816D-06,   0.000000000000000D+00, &
  -3.962462175303654D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.776831374016271D-08,   0.000000000000000D+00,   2.694276080095424D-07, &
   0.000000000000000D+00,  -2.443721286750149D-07,   0.000000000000000D+00, &
  -6.419097110114519D-08,   0.000000000000000D+00,   7.648170256031733D-06, &
   0.000000000000000D+00,   2.543362615728204D-05,   0.000000000000000D+00, &
  -7.362552642819425D-05,   0.000000000000000D+00,  -4.445355268698689D-05, &
   0.000000000000000D+00,  -5.124259934830275D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.730647191224161D-09, &
   0.000000000000000D+00,  -2.909440242144060D-09,   0.000000000000000D+00, &
   3.744413075370888D-07,   0.000000000000000D+00,   9.369132291168010D-08, &
   0.000000000000000D+00,  -1.983369122816320D-06,   0.000000000000000D+00, &
   1.401267507411623D-05,   0.000000000000000D+00,   5.289710649932277D-05, &
   0.000000000000000D+00,  -3.434114636003579D-07,   0.000000000000000D+00, &
   9.452709798816852D-06,   0.000000000000000D+00,   2.942325457436106D-05 /)
lem(:,172) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.882212909539368D-05,   4.306455558159209D-04, &
   1.016729058514139D-03,  -4.686899428564079D-04,  -2.024219028732910D-03, &
  -6.101139495962213D-04,   2.315307182373343D-03,   9.189504146016502D-04, &
  -1.876477320282610D-03,   2.088833822259433D-04,   4.530972781029818D-04, &
  -3.464009077435316D-04,   3.171779925416300D-04,  -3.242697523962011D-04, &
   1.818317897816633D-04,   9.530057148796197D-05,  -2.895446182379944D-04, &
   2.516900533158922D-04,  -5.504396562105501D-05,  -1.094376480767497D-04, &
   1.213487069091737D-04,   0.000000000000000D+00,   5.882212909539410D-05, &
  -4.306455558159154D-04,  -1.016729058514134D-03,   4.686899428564056D-04, &
   2.024219028732910D-03,   6.101139495962204D-04,  -2.315307182373339D-03, &
  -9.189504146016513D-04,   1.876477320282600D-03,  -2.088833822259409D-04, &
  -4.530972781029759D-04,   3.464009077435297D-04,  -3.171779925416302D-04, &
   3.242697523961979D-04,  -1.818317897816651D-04,  -9.530057148796162D-05, &
   2.895446182379963D-04,  -2.516900533158896D-04,   5.504396562105302D-05, &
   1.094376480767462D-04,  -1.213487069091723D-04,   0.000000000000000D+00, &
   2.941106454769522D-05,   0.000000000000000D+00,   4.814008220131648D-04, &
   0.000000000000000D+00,  -2.110299409333564D-03,   0.000000000000000D+00, &
   2.157607464809634D-03,   0.000000000000000D+00,   7.040849100791468D-04, &
   0.000000000000000D+00,  -1.111385023003458D-03,   0.000000000000000D+00, &
  -1.787690110826312D-03,   0.000000000000000D+00,   2.426442728009091D-03, &
   0.000000000000000D+00,  -7.992318445360026D-04,   0.000000000000000D+00, &
   3.748496876840082D-06,   0.000000000000000D+00,   2.368900608532046D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.325148809523829D-06, &
   0.000000000000000D+00,   1.705109126984146D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.564191017315999D-06,   0.000000000000000D+00, &
  -2.496599176286684D-05,   0.000000000000000D+00,  -5.114710462050652D-06, &
   0.000000000000000D+00,  -3.953292780229418D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.056923759813241D-06,   0.000000000000000D+00, &
   1.403366334479835D-05,   0.000000000000000D+00,   1.111548720605540D-04, &
   0.000000000000000D+00,   8.790093825565050D-06,   0.000000000000000D+00, &
   7.083548820323574D-06,   0.000000000000000D+00,  -1.878864762437370D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.765498916717639D-07, &
   0.000000000000000D+00,  -1.233199435293842D-05,   0.000000000000000D+00, &
  -5.718533436743029D-05,   0.000000000000000D+00,  -2.244847682523567D-04, &
   0.000000000000000D+00,  -9.443115440042621D-06,   0.000000000000000D+00, &
  -8.233410410741267D-06,   0.000000000000000D+00,   5.034911784465149D-06, &
   0.000000000000000D+00,  -8.034731019415616D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.212908999932380D-06,   0.000000000000000D+00, &
   1.086222571123381D-05,   0.000000000000000D+00,   3.975083355703614D-05, &
   0.000000000000000D+00,   1.472383760393849D-04,   0.000000000000000D+00, &
   2.923289399147370D-04,   0.000000000000000D+00,   5.827266103000358D-06, &
   0.000000000000000D+00,   5.273919698912038D-06,   0.000000000000000D+00, &
  -9.289457057230374D-06,   0.000000000000000D+00,  -4.908697531600557D-05, &
   0.000000000000000D+00,   5.856310473658402D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.829120355159180D-06,   0.000000000000000D+00, &
  -7.138865134835072D-06,   0.000000000000000D+00,  -2.123922242831889D-05, &
   0.000000000000000D+00,  -7.387003104790295D-05,   0.000000000000000D+00, &
  -2.711639771261896D-04,   0.000000000000000D+00,  -3.199865619889988D-04, &
   0.000000000000000D+00,  -1.319287190587269D-06,   0.000000000000000D+00, &
  -1.522744833439030D-06,   0.000000000000000D+00,   8.964762870593283D-06, &
   0.000000000000000D+00,   5.533644669235110D-05,   0.000000000000000D+00, &
   1.169589671398000D-04,   0.000000000000000D+00,  -9.123320992467805D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.909782004518252D-06, &
   0.000000000000000D+00,   2.688601926423918D-06,   0.000000000000000D+00, &
   5.777238076705588D-06,   0.000000000000000D+00,   1.011526826590354D-05, &
   0.000000000000000D+00,   1.020742912710588D-04,   0.000000000000000D+00, &
   3.386352778772096D-04,   0.000000000000000D+00,   3.380238945787758D-04, &
   0.000000000000000D+00,  -3.897186207645244D-07,   0.000000000000000D+00, &
   2.811041006273336D-07,   0.000000000000000D+00,  -3.957907812096800D-06, &
   0.000000000000000D+00,  -1.974522369611859D-05,   0.000000000000000D+00, &
  -1.156082781559342D-04,   0.000000000000000D+00,  -1.528406158089720D-04, &
   0.000000000000000D+00,   5.015606072697769D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.586148718088494D-07,   0.000000000000000D+00, &
  -5.569947196788028D-07,   0.000000000000000D+00,   2.436900797261277D-07, &
   0.000000000000000D+00,   2.365677391793175D-06,   0.000000000000000D+00, &
   2.289199204628302D-05,   0.000000000000000D+00,  -1.159173811272673D-04, &
   0.000000000000000D+00,  -2.838072594189467D-04,   0.000000000000000D+00, &
  -2.962790880174826D-04,   0.000000000000000D+00,   1.786363905850931D-07, &
   0.000000000000000D+00,  -2.044563466882637D-07,   0.000000000000000D+00, &
   3.716787512833762D-07,   0.000000000000000D+00,   2.922165303959606D-06, &
   0.000000000000000D+00,   1.704998491137096D-05,   0.000000000000000D+00, &
   1.465248327318462D-04,   0.000000000000000D+00,   1.188853572004684D-04, &
   0.000000000000000D+00,   4.094693794982140D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.909823110608486D-08,   0.000000000000000D+00, &
   1.920528805278518D-07,   0.000000000000000D+00,  -5.273797250266416D-07, &
   0.000000000000000D+00,  -9.189726202562671D-07,   0.000000000000000D+00, &
  -9.029978252539768D-06,   0.000000000000000D+00,  -4.154939029690186D-05, &
   0.000000000000000D+00,   1.104709893287914D-04,   0.000000000000000D+00, &
   1.681816032247189D-04,   0.000000000000000D+00,   1.760574211548985D-04, &
   0.000000000000000D+00,   7.224683193173467D-09,   0.000000000000000D+00, &
   2.799351217897311D-08,   0.000000000000000D+00,   2.923694756442284D-07, &
   0.000000000000000D+00,  -1.511823469826868D-07,   0.000000000000000D+00, &
   1.381952853415264D-06,   0.000000000000000D+00,  -1.367262189834554D-05, &
   0.000000000000000D+00,  -1.196141065231320D-04,   0.000000000000000D+00, &
  -5.365326071353536D-05,   0.000000000000000D+00,  -8.542696347850618D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.607186895976360D-10, &
   0.000000000000000D+00,  -6.037708667779733D-08,   0.000000000000000D+00, &
   1.070872790304099D-07,   0.000000000000000D+00,   1.444733015562726D-07, &
   0.000000000000000D+00,  -2.245943466066908D-07,   0.000000000000000D+00, &
   1.163516344742597D-05,   0.000000000000000D+00,   2.051424809275798D-05, &
   0.000000000000000D+00,  -8.336502073824147D-05,   0.000000000000000D+00, &
  -7.906992506986133D-05,   0.000000000000000D+00,  -6.074631115892424D-05, &
   0.000000000000000D+00,   1.809077117758828D-11,   0.000000000000000D+00, &
   2.379653455007376D-08,   0.000000000000000D+00,  -1.355443964002634D-07, &
   0.000000000000000D+00,   1.702826596923537D-07,   0.000000000000000D+00, &
  -1.040665112506709D-07,   0.000000000000000D+00,  -3.351243293532568D-06, &
   0.000000000000000D+00,   2.072914322002014D-05,   0.000000000000000D+00, &
   5.784862683384986D-05,   0.000000000000000D+00,   8.112272778580795D-06, &
   0.000000000000000D+00,  -7.010130689297564D-06,   0.000000000000000D+00 /)
lem(:,173) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.866834785572354D-04,  -6.004759011821413D-04,  -8.344663884581160D-04, &
  -6.349076187412151D-05,   5.694787704707334D-04,   5.182050887585635D-04, &
   1.128914891143327D-03,  -7.959059133950110D-04,  -1.543347101896757D-03, &
   1.145707951974619D-03,   2.426974269032646D-04,  -4.701325806442417D-04, &
   2.028986489241575D-04,  -1.164447718184619D-04,   1.379844148888763D-04, &
  -1.292437077291569D-04,   4.497656286445300D-05,   7.934530948365506D-05, &
  -1.536123896922168D-04,   1.256071978597542D-04,  -2.572605143332105D-05, &
  -6.062493975194834D-05,  -1.866834785572345D-04,  -6.004759011821511D-04, &
  -8.344663884581195D-04,  -6.349076187411710D-05,   5.694787704707331D-04, &
   5.182050887585670D-04,   1.128914891143332D-03,  -7.959059133950036D-04, &
  -1.543347101896764D-03,   1.145707951974613D-03,   2.426974269032703D-04, &
  -4.701325806442377D-04,   2.028986489241527D-04,  -1.164447718184613D-04, &
   1.379844148888786D-04,  -1.292437077291542D-04,   4.497656286444877D-05, &
   7.934530948365139D-05,  -1.536123896922105D-04,   1.256071978597534D-04, &
  -2.572605143332579D-05,  -6.062493975194547D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.217987430590399D-04,   0.000000000000000D+00, &
  -4.747047200105409D-04,   0.000000000000000D+00,  -2.913214635168035D-04, &
   0.000000000000000D+00,   6.894038675619621D-03,   0.000000000000000D+00, &
  -1.710210803733801D-02,   0.000000000000000D+00,   1.808877776436639D-02, &
   0.000000000000000D+00,  -8.387180043636727D-03,   0.000000000000000D+00, &
   7.179773324468064D-04,   0.000000000000000D+00,   5.961091403280394D-04, &
   0.000000000000000D+00,  -1.794501077588843D-04,   0.000000000000000D+00, &
   1.937624007936668D-06,   5.192832341269794D-06,   0.000000000000000D+00, &
   7.750496031746102D-06,   0.000000000000000D+00,  -9.164961557539688D-06, &
  -7.609577922079499D-07,   0.000000000000000D+00,  -6.469902710137092D-06, &
   0.000000000000000D+00,  -1.148482593795078D-05,   0.000000000000000D+00, &
   2.624599792568412D-07,   0.000000000000000D+00,   2.608746505230873D-05, &
  -5.707183441558479D-06,   0.000000000000000D+00,   3.817119295634963D-06, &
   0.000000000000000D+00,   5.774119543650835D-06,   0.000000000000000D+00, &
   7.481927902403082D-06,   0.000000000000000D+00,   3.896197085439297D-06, &
   0.000000000000000D+00,  -3.897114183246142D-06,   0.000000000000000D+00, &
  -4.227771985403897D-06,   8.350981880209757D-06,   0.000000000000000D+00, &
   1.278278387418128D-06,   0.000000000000000D+00,  -6.860634089251168D-06, &
   0.000000000000000D+00,  -2.206166187161378D-05,   0.000000000000000D+00, &
  -3.337024109319609D-06,   0.000000000000000D+00,  -5.259692908004476D-06, &
   0.000000000000000D+00,  -5.880095697544152D-06,   0.000000000000000D+00, &
  -3.091439268671211D-06,   0.000000000000000D+00,  -1.834779565238407D-04, &
  -6.298718615815296D-06,   0.000000000000000D+00,  -1.959494720490260D-06, &
   0.000000000000000D+00,   7.781491300240835D-06,   0.000000000000000D+00, &
   3.034573042123180D-05,   0.000000000000000D+00,   1.380567011588266D-04, &
   0.000000000000000D+00,   2.201375046465415D-06,   0.000000000000000D+00, &
   3.790236721662269D-06,   0.000000000000000D+00,   5.036816162872038D-06, &
   0.000000000000000D+00,  -1.117123432387641D-06,   0.000000000000000D+00, &
  -4.704655662233207D-06,   0.000000000000000D+00,   5.399372683476860D-04, &
   2.340899732099606D-06,   0.000000000000000D+00,  -1.265569040638521D-07, &
   0.000000000000000D+00,  -6.542123825412103D-06,   0.000000000000000D+00, &
  -2.301775229903315D-05,   0.000000000000000D+00,  -9.154660920704311D-05, &
   0.000000000000000D+00,  -3.675688748242198D-04,   0.000000000000000D+00, &
  -1.315537520370208D-06,   0.000000000000000D+00,  -9.275862756346661D-07, &
   0.000000000000000D+00,  -2.717976446222309D-07,   0.000000000000000D+00, &
   6.816150018845674D-06,   0.000000000000000D+00,   5.053722015778617D-05, &
   0.000000000000000D+00,   5.414492912571831D-05,   0.000000000000000D+00, &
  -7.641794939209049D-04,  -1.434954405797416D-07,   0.000000000000000D+00, &
   5.063638911213269D-07,   0.000000000000000D+00,   2.706687717554071D-06, &
   0.000000000000000D+00,   8.521349055656026D-06,   0.000000000000000D+00, &
   1.983420605506433D-05,   0.000000000000000D+00,   1.640778451980881D-04, &
   0.000000000000000D+00,   5.113884744195802D-04,   0.000000000000000D+00, &
   2.855399455394586D-07,   0.000000000000000D+00,  -3.037369499034275D-07, &
   0.000000000000000D+00,  -1.020717848847430D-06,   0.000000000000000D+00, &
  -4.776052597330330D-06,   0.000000000000000D+00,  -2.430707338057382D-05, &
   0.000000000000000D+00,  -1.341586635520608D-04,   0.000000000000000D+00, &
  -8.736946149610581D-05,   0.000000000000000D+00,   6.022452792847012D-04, &
  -9.623747535880172D-08,   0.000000000000000D+00,   4.829161894732671D-08, &
   0.000000000000000D+00,  -3.669095853464858D-07,   0.000000000000000D+00, &
  -1.045014050151753D-06,   0.000000000000000D+00,   2.195025812472930D-06, &
   0.000000000000000D+00,   1.384329840576913D-05,   0.000000000000000D+00, &
  -2.041097920266436D-04,   0.000000000000000D+00,  -3.981561688412321D-04, &
   0.000000000000000D+00,   1.560605989376217D-08,   0.000000000000000D+00, &
   1.025944069898876D-07,   0.000000000000000D+00,   2.923097261097772D-07, &
   0.000000000000000D+00,   1.124127929315009D-06,   0.000000000000000D+00, &
   3.557151239793047D-06,   0.000000000000000D+00,   2.875726274368140D-05, &
   0.000000000000000D+00,   1.828975391752859D-04,   0.000000000000000D+00, &
   4.706404452599146D-05,   0.000000000000000D+00,  -2.383978466058864D-04, &
  -1.280123725232840D-09,   0.000000000000000D+00,  -7.656201345748434D-08, &
   0.000000000000000D+00,   5.657905388140023D-08,   0.000000000000000D+00, &
  -1.782464356249972D-08,   0.000000000000000D+00,  -9.829050261331650D-07, &
   0.000000000000000D+00,  -1.374470033904040D-05,   0.000000000000000D+00, &
  -3.646398742605504D-05,   0.000000000000000D+00,   1.854649195015159D-04, &
   0.000000000000000D+00,   1.629126309325736D-04,   0.000000000000000D+00, &
   4.270374144553907D-11,   0.000000000000000D+00,   2.457472877220651D-08, &
   0.000000000000000D+00,  -4.864888179532930D-08,   0.000000000000000D+00, &
  -1.921573593869110D-08,   0.000000000000000D+00,   4.691467394563776D-07, &
   0.000000000000000D+00,   1.863489238603883D-06,   0.000000000000000D+00, &
  -2.767352572558118D-05,   0.000000000000000D+00,  -1.448812605711903D-04, &
   0.000000000000000D+00,   6.458958833614565D-06,   0.000000000000000D+00, &
   3.440303400815927D-06,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.291904358472736D-09,   0.000000000000000D+00,  -3.269079501938751D-08, &
   0.000000000000000D+00,  -2.709191845738822D-08,   0.000000000000000D+00, &
  -3.890935367977604D-07,   0.000000000000000D+00,   3.415989773524664D-07, &
   0.000000000000000D+00,   1.851658589307912D-05,   0.000000000000000D+00, &
   1.142310559360769D-05,   0.000000000000000D+00,  -1.131753706718524D-04, &
   0.000000000000000D+00,  -2.512375515334318D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.159057365864043D-10, &
   0.000000000000000D+00,   3.005994164775927D-08,   0.000000000000000D+00, &
  -6.431517924570710D-08,   0.000000000000000D+00,  -4.729201585735729D-08, &
   0.000000000000000D+00,  -6.287367462752404D-07,   0.000000000000000D+00, &
  -4.753286199609019D-06,   0.000000000000000D+00,   3.395741668124653D-05, &
   0.000000000000000D+00,   5.648448546612812D-05,   0.000000000000000D+00, &
  -9.411939897724168D-06,   0.000000000000000D+00,   4.443720810943157D-05 /)
lem(:,174) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.143763621299396D-04,   6.754513627976937D-04, &
   1.867318933065065D-03,   2.390777528107005D-03,  -1.047113492029149D-03, &
  -4.083475241002954D-03,   8.148056272714032D-04,   2.763404731753351D-03, &
  -1.337469567381416D-03,  -4.484010208670804D-04,   5.946576023280294D-04, &
  -1.780426956390201D-04,   5.634781081661182D-05,  -9.445250836533843D-05, &
   7.915321187341069D-05,  -6.309814139932560D-05,   2.292184907545148D-05, &
   6.191564317000944D-05,  -1.108536259863636D-04,   8.330852620506435D-05, &
  -1.638564363632263D-05,   0.000000000000000D+00,  -1.143763621299426D-04, &
  -6.754513627976910D-04,  -1.867318933065056D-03,  -2.390777528107007D-03, &
   1.047113492029144D-03,   4.083475241002952D-03,  -8.148056272713910D-04, &
  -2.763404731753358D-03,   1.337469567381392D-03,   4.484010208670881D-04, &
  -5.946576023280084D-04,   1.780426956390190D-04,  -5.634781081662315D-05, &
   9.445250836533321D-05,  -7.915321187340522D-05,   6.309814139932507D-05, &
  -2.292184907545435D-05,  -6.191564317000157D-05,   1.108536259863615D-04, &
  -8.330852620507315D-05,   1.638564363633009D-05,   0.000000000000000D+00, &
  -5.718818106496238D-05,   0.000000000000000D+00,   1.783879412173093D-04, &
   0.000000000000000D+00,  -7.423160280024226D-04,   0.000000000000000D+00, &
   4.721844126453455D-03,   0.000000000000000D+00,  -1.783747043711587D-02, &
   0.000000000000000D+00,   3.444506158312022D-02,   0.000000000000000D+00, &
  -3.583532416807945D-02,   0.000000000000000D+00,   2.015681920963911D-02, &
   0.000000000000000D+00,  -5.793340155855630D-03,   0.000000000000000D+00, &
   9.152121814316771D-04,   0.000000000000000D+00,  -2.246152929738210D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.521122685185281D-06, &
   0.000000000000000D+00,   4.133597883597877D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.195973875661308D-06,   0.000000000000000D+00, &
  -1.624961952110389D-05,   0.000000000000000D+00,  -2.771389490139412D-06, &
   0.000000000000000D+00,  -1.235675674152239D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.914900493025543D-06,   0.000000000000000D+00, &
   7.662422213204429D-07,   0.000000000000000D+00,   6.991299715909078D-05, &
   0.000000000000000D+00,   9.277108886483135D-07,   0.000000000000000D+00, &
   1.579163566468241D-05,   0.000000000000000D+00,   2.902032320977639D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.068628998316513D-05, &
   0.000000000000000D+00,   4.768858528624214D-06,   0.000000000000000D+00, &
  -3.442195825008331D-05,   0.000000000000000D+00,  -2.932410448889940D-04, &
   0.000000000000000D+00,  -2.383864688552264D-06,   0.000000000000000D+00, &
  -1.398597929891695D-05,   0.000000000000000D+00,  -2.980268212733850D-05, &
   0.000000000000000D+00,  -3.206779015530637D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.712809431711934D-06,   0.000000000000000D+00, &
  -2.241016804371701D-06,   0.000000000000000D+00,   2.551148259738401D-05, &
   0.000000000000000D+00,   1.320340878681072D-04,   0.000000000000000D+00, &
   8.756479037429644D-04,   0.000000000000000D+00,   3.811478361261452D-06, &
   0.000000000000000D+00,   7.364105820281047D-06,   0.000000000000000D+00, &
   1.491703446999451D-05,   0.000000000000000D+00,  -1.243132798342320D-05, &
   0.000000000000000D+00,  -1.963695414911587D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.251432899056321D-06,   0.000000000000000D+00, &
  -8.365944667990493D-07,   0.000000000000000D+00,  -1.649247605911071D-05, &
   0.000000000000000D+00,  -6.316777652593627D-05,   0.000000000000000D+00, &
  -2.595394910318938D-04,   0.000000000000000D+00,  -1.715198057434663D-03, &
   0.000000000000000D+00,  -2.218606925042864D-06,   0.000000000000000D+00, &
  -9.800848958672680D-07,   0.000000000000000D+00,  -1.221544117626681D-06, &
   0.000000000000000D+00,   1.963522084423207D-05,   0.000000000000000D+00, &
   1.595295769868729D-04,   0.000000000000000D+00,   5.702509441131757D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.657174890111337D-07, &
   0.000000000000000D+00,   2.916767081698878D-07,   0.000000000000000D+00, &
   6.634715238406958D-06,   0.000000000000000D+00,   1.741908898247055D-05, &
   0.000000000000000D+00,   4.176004316552768D-05,   0.000000000000000D+00, &
   3.380238945787758D-04,   0.000000000000000D+00,   2.218882161571891D-03, &
   0.000000000000000D+00,   3.560893397760400D-07,   0.000000000000000D+00, &
  -7.066611024246212D-07,   0.000000000000000D+00,  -2.431159770844963D-06, &
   0.000000000000000D+00,  -1.016510603502516D-05,   0.000000000000000D+00, &
  -5.981664679365698D-05,   0.000000000000000D+00,  -3.538347777356959D-04, &
   0.000000000000000D+00,  -8.045415350389992D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.990239894851026D-07,   0.000000000000000D+00, &
   5.339038393769514D-07,   0.000000000000000D+00,  -1.543585958313504D-06, &
   0.000000000000000D+00,  -1.425381210101359D-06,   0.000000000000000D+00, &
   7.701488606968995D-06,   0.000000000000000D+00,   4.737019835575439D-05, &
   0.000000000000000D+00,  -3.693360433409147D-04,   0.000000000000000D+00, &
  -1.922084345855228D-03,   0.000000000000000D+00,   4.937010500194511D-08, &
   0.000000000000000D+00,   8.677358546239187D-08,   0.000000000000000D+00, &
   1.496599289194139D-06,   0.000000000000000D+00,   1.372467856969039D-06, &
   0.000000000000000D+00,   8.006472115757415D-06,   0.000000000000000D+00, &
   6.551257975941344D-05,   0.000000000000000D+00,   4.620646364603822D-04, &
   0.000000000000000D+00,   6.348197227895361D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.717601643732351D-09,   0.000000000000000D+00, &
  -2.254788859210767D-07,   0.000000000000000D+00,   2.802224713148970D-07, &
   0.000000000000000D+00,   7.025023068997220D-07,   0.000000000000000D+00, &
  -2.938324993400566D-06,   0.000000000000000D+00,  -3.201898158050232D-05, &
   0.000000000000000D+00,  -1.050065097794215D-04,   0.000000000000000D+00, &
   3.849451302432145D-04,   0.000000000000000D+00,   1.123477656280079D-03, &
   0.000000000000000D+00,   4.876661811401078D-10,   0.000000000000000D+00, &
   8.562133967975432D-08,   0.000000000000000D+00,  -6.729845023794154D-07, &
   0.000000000000000D+00,   4.263429340986675D-07,   0.000000000000000D+00, &
   1.012301917328786D-06,   0.000000000000000D+00,   3.706088495917185D-06, &
   0.000000000000000D+00,  -6.000202963193547D-05,   0.000000000000000D+00, &
  -3.988019020829497D-04,   0.000000000000000D+00,  -2.526103244902797D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.482768802845533D-11, &
   0.000000000000000D+00,  -4.603234230457598D-09,   0.000000000000000D+00, &
   7.469656824970808D-08,   0.000000000000000D+00,  -1.141049212601456D-06, &
   0.000000000000000D+00,  -2.862561525928940D-07,   0.000000000000000D+00, &
   6.502103240738651D-07,   0.000000000000000D+00,   4.137502218310073D-05, &
   0.000000000000000D+00,   5.103901439563163D-05,   0.000000000000000D+00, &
  -3.340346888334408D-04,   0.000000000000000D+00,  -4.510791977207168D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.657608473356593D-09,   0.000000000000000D+00,   1.890533255898805D-07, &
   0.000000000000000D+00,   1.778064945470548D-08,   0.000000000000000D+00, &
  -9.611702506927345D-07,   0.000000000000000D+00,  -7.189145229462911D-07, &
   0.000000000000000D+00,  -1.002937457112427D-05,   0.000000000000000D+00, &
   7.587041797722258D-05,   0.000000000000000D+00,   2.058760307370230D-04, &
   0.000000000000000D+00,   4.388504612744742D-06,   0.000000000000000D+00 /)
lem(:,175) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.565232873205334D-03,  -6.134314629019729D-03, &
  -1.191720624599129D-02,   1.300038553340092D-02,   1.360370494342046D-02, &
  -7.536259664864041D-03,  -7.220140746224617D-03,  -1.198544032400023D-03, &
  -3.358410299366372D-04,   8.998961077369817D-04,   1.734427694204389D-03, &
   1.540825114605735D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.565232873205385D-03, &
  -6.134314629019749D-03,  -1.191720624599129D-02,   1.300038553340088D-02, &
   1.360370494342043D-02,  -7.536259664863986D-03,  -7.220140746224549D-03, &
  -1.198544032400042D-03,  -3.358410299366815D-04,   8.998961077370099D-04, &
   1.734427694204379D-03,   1.540825114605756D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.900583821149850D-04,   0.000000000000000D+00,   4.720034203744557D-06, &
   0.000000000000000D+00,   2.527116214857028D-07,   0.000000000000000D+00, &
   2.005647789598198D-08,   0.000000000000000D+00,   1.880294802952159D-09, &
   0.000000000000000D+00,  -2.461547004275646D-03,   0.000000000000000D+00, &
  -5.017670289136769D-05,   0.000000000000000D+00,  -3.822263274967433D-06, &
   0.000000000000000D+00,  -4.005027929794282D-07,   0.000000000000000D+00, &
  -4.764532723320626D-08,   0.000000000000000D+00,  -7.049426675387876D-09, &
   5.574088702363269D-03,   0.000000000000000D+00,   3.261440844743994D-05, &
   0.000000000000000D+00,   1.416411344468129D-05,   0.000000000000000D+00, &
   2.523044264604194D-06,   0.000000000000000D+00,   4.266604809112198D-07, &
   0.000000000000000D+00,   7.146187122647823D-08,   0.000000000000000D+00, &
  -3.539837011004651D-03,   0.000000000000000D+00,   6.758310350847694D-04, &
   0.000000000000000D+00,   2.158732521142974D-05,   0.000000000000000D+00, &
  -3.132292129842920D-06,   0.000000000000000D+00,  -1.399880563743002D-06, &
   0.000000000000000D+00,  -3.702242519679226D-07,   0.000000000000000D+00, &
  -1.071611978136851D-07,  -6.595508045911873D-03,   0.000000000000000D+00, &
  -2.088458760948072D-03,   0.000000000000000D+00,  -1.561846767760491D-04, &
   0.000000000000000D+00,  -1.971433991549793D-05,   0.000000000000000D+00, &
  -7.925130589511055D-07,   0.000000000000000D+00,   5.825128239387568D-07, &
   0.000000000000000D+00,   2.855399455394586D-07,   0.000000000000000D+00, &
   1.358214452022388D-02,   0.000000000000000D+00,   1.647689123056277D-03, &
   0.000000000000000D+00,   2.659329335774388D-05,   0.000000000000000D+00, &
   4.387610906754080D-05,   0.000000000000000D+00,   1.257215751748277D-05, &
   0.000000000000000D+00,   1.915117826032757D-06,   0.000000000000000D+00, &
  -7.590757339394439D-08,   0.000000000000000D+00,  -2.827991423355122D-07, &
  -7.036856907083090D-03,   0.000000000000000D+00,   2.244844912180604D-03, &
   0.000000000000000D+00,   7.272047200762058D-04,   0.000000000000000D+00, &
   6.176176998501416D-05,   0.000000000000000D+00,  -8.213077627226769D-06, &
   0.000000000000000D+00,  -6.679193610931725D-06,   0.000000000000000D+00, &
  -1.864354617314255D-06,   0.000000000000000D+00,  -1.710368254629840D-07, &
   0.000000000000000D+00,  -4.057525303354003D-03,   0.000000000000000D+00, &
  -5.057225281911330D-03,   0.000000000000000D+00,  -9.336616385702186D-04, &
   0.000000000000000D+00,  -1.858537219598393D-04,   0.000000000000000D+00, &
  -4.518661444739779D-05,   0.000000000000000D+00,  -3.137187734750344D-06, &
   0.000000000000000D+00,   2.797334482781059D-06,   0.000000000000000D+00, &
   1.410588049582722D-06,   0.000000000000000D+00,   3.112553892580085D-07, &
   6.829750841615885D-03,   0.000000000000000D+00,   2.396646714286282D-03, &
   0.000000000000000D+00,  -6.262054004842212D-04,   0.000000000000000D+00, &
  -9.415110646724745D-05,   0.000000000000000D+00,   3.950183357911352D-05, &
   0.000000000000000D+00,   2.474350600058438D-05,   0.000000000000000D+00, &
   5.546714867278306D-06,   0.000000000000000D+00,  -6.116026459833094D-07, &
   0.000000000000000D+00,  -9.177974980769405D-07,   0.000000000000000D+00, &
  -3.237714694889330D-03,   0.000000000000000D+00,   1.886411063274246D-03, &
   0.000000000000000D+00,   1.912427213546813D-03,   0.000000000000000D+00, &
   4.176938797670679D-04,   0.000000000000000D+00,   9.315450159950361D-05, &
   0.000000000000000D+00,   2.003563492483021D-06,   0.000000000000000D+00, &
  -1.102898925933108D-05,   0.000000000000000D+00,  -4.770165616118815D-06, &
   0.000000000000000D+00,  -3.874719007214490D-07,   0.000000000000000D+00, &
   1.010022829285178D-06,   5.529449355923201D-04,   0.000000000000000D+00, &
  -2.560474346162147D-03,   0.000000000000000D+00,  -6.697126504763011D-04, &
   0.000000000000000D+00,   5.097163229124517D-05,   0.000000000000000D+00, &
  -6.893880658423893D-05,   0.000000000000000D+00,  -4.795297523572742D-05, &
   0.000000000000000D+00,  -1.094226489936661D-05,   0.000000000000000D+00, &
   3.361489466822727D-06,   0.000000000000000D+00,   3.192069762457796D-06, &
   0.000000000000000D+00,   6.834413276598173D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.920215260593442D-04, &
   0.000000000000000D+00,  -1.068260772763359D-03,   0.000000000000000D+00, &
  -5.665790301456866D-04,   0.000000000000000D+00,  -1.315030703557183D-04, &
   0.000000000000000D+00,  -7.973165557962296D-06,   0.000000000000000D+00, &
   1.909701748781807D-05,   0.000000000000000D+00,   9.805610993544880D-06, &
   0.000000000000000D+00,   2.433601621259907D-07,   0.000000000000000D+00, &
  -1.846509058561484D-06,   0.000000000000000D+00,  -7.594073260275182D-07 /)
lem(:,176) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.748537353611139D-04, &
   2.856945668463977D-03,  -3.242910554504254D-03,  -4.326907199297713D-03, &
   9.262212963178999D-03,   1.280484008654371D-03,  -8.599523474177301D-03, &
   2.321601605521948D-04,   3.270184801457671D-03,   8.127439615122110D-04, &
   9.233165567848943D-05,  -6.551230504876865D-04,  -7.796589812818813D-04, &
  -2.373059633772706D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.748537353611025D-04,  -2.856945668463972D-03,   3.242910554504258D-03, &
   4.326907199297699D-03,  -9.262212963178990D-03,  -1.280484008654381D-03, &
   8.599523474177317D-03,  -2.321601605521783D-04,  -3.270184801457689D-03, &
  -8.127439615122200D-04,  -9.233165567845929D-05,   6.551230504876782D-04, &
   7.796589812818536D-04,   2.373059633773025D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.270613883084912D-07, &
   0.000000000000000D+00,  -1.601198855031614D-06,   0.000000000000000D+00, &
  -4.070468136539857D-06,   0.000000000000000D+00,   1.263083333122156D-06, &
   0.000000000000000D+00,   9.065091768534407D-06,   0.000000000000000D+00, &
  -1.509681201934909D-06,   0.000000000000000D+00,  -7.941959199092474D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.632326767473076D-06, &
   0.000000000000000D+00,   1.107881636140089D-07,   0.000000000000000D+00, &
   6.798364585271844D-09,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.933164112326874D-05,   0.000000000000000D+00, &
  -1.809012443014059D-06,   0.000000000000000D+00,  -1.618658234622364D-07, &
   0.000000000000000D+00,  -1.335393043563179D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.373445666964543D-04,   0.000000000000000D+00, &
   8.946924640780297D-06,   0.000000000000000D+00,   1.232322964903600D-06, &
   0.000000000000000D+00,   1.695795671872987D-07,   0.000000000000000D+00, &
   1.418855108797657D-08,   0.000000000000000D+00,  -3.422141862055034D-04, &
   0.000000000000000D+00,  -9.267713678666001D-06,   0.000000000000000D+00, &
  -3.172706888573543D-06,   0.000000000000000D+00,  -7.878563134604115D-07, &
   0.000000000000000D+00,  -1.541250148879276D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.379475156260942D-04,   0.000000000000000D+00, &
  -6.618544896497197D-05,   0.000000000000000D+00,  -3.159506541616414D-06, &
   0.000000000000000D+00,   8.671781816125759D-07,   0.000000000000000D+00, &
   4.725695785011938D-07,   0.000000000000000D+00,   1.802639107033177D-07, &
   0.000000000000000D+00,  -8.091207762509627D-05,   0.000000000000000D+00, &
   2.801169348222646D-04,   0.000000000000000D+00,   3.023383172723505D-05, &
   0.000000000000000D+00,   4.257678236419092D-06,   0.000000000000000D+00, &
   7.238531538387497D-08,   0.000000000000000D+00,  -2.894430840073267D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.605588585345746D-04, &
   0.000000000000000D+00,  -4.498500667629974D-04,   0.000000000000000D+00, &
  -3.622366004979268D-05,   0.000000000000000D+00,  -1.256924046105279D-05, &
   0.000000000000000D+00,  -3.279110899491910D-06,   0.000000000000000D+00, &
  -3.897186207645244D-07,   0.000000000000000D+00,   3.560893397760400D-07, &
   0.000000000000000D+00,   1.093219034635423D-03,   0.000000000000000D+00, &
   1.669781490858917D-04,   0.000000000000000D+00,  -7.939384358809687D-05, &
   0.000000000000000D+00,  -1.853303474099335D-06,   0.000000000000000D+00, &
   4.400667157793108D-06,   0.000000000000000D+00,   2.130670089408290D-06, &
   0.000000000000000D+00,   4.198784161436243D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.058581060289880D-04,   0.000000000000000D+00, &
   5.676056372656789D-04,   0.000000000000000D+00,   2.604048957556261D-04, &
   0.000000000000000D+00,   4.557153833568638D-05,   0.000000000000000D+00, &
   7.511931919573498D-06,   0.000000000000000D+00,  -8.222650657462417D-07, &
   0.000000000000000D+00,  -1.280054600234169D-06,   0.000000000000000D+00, &
  -6.231034888635977D-07,   0.000000000000000D+00,  -9.356273933662664D-05, &
   0.000000000000000D+00,  -1.015738326251009D-03,   0.000000000000000D+00, &
  -1.876827093026544D-04,   0.000000000000000D+00,  -3.649195230708372D-05, &
   0.000000000000000D+00,  -1.921710113325977D-05,   0.000000000000000D+00, &
  -6.289906059341867D-06,   0.000000000000000D+00,  -5.320773829621393D-07, &
   0.000000000000000D+00,   7.307739722207702D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.402165803255590D-04,   0.000000000000000D+00, &
   6.073435120902055D-04,   0.000000000000000D+00,  -2.402339580671915D-04, &
   0.000000000000000D+00,  -7.390590125185929D-05,   0.000000000000000D+00, &
  -4.722425517998103D-06,   0.000000000000000D+00,   6.345701570372750D-06, &
   0.000000000000000D+00,   4.058205978929093D-06,   0.000000000000000D+00, &
   9.071398746118963D-07,   0.000000000000000D+00,  -1.136103274872931D-06, &
   0.000000000000000D+00,  -4.563496413700349D-04,   0.000000000000000D+00, &
   1.913849864471329D-04,   0.000000000000000D+00,   5.292372125909348D-04, &
   0.000000000000000D+00,   1.182680425585441D-04,   0.000000000000000D+00, &
   3.740000193259881D-05,   0.000000000000000D+00,   9.517604540089325D-06, &
   0.000000000000000D+00,  -8.643524791175222D-07,   0.000000000000000D+00, &
  -2.270175244758515D-06,   0.000000000000000D+00,  -7.632386693951583D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.998530318886273D-04, &
   0.000000000000000D+00,  -5.554321500541940D-04,   0.000000000000000D+00, &
  -2.577442599836276D-04,   0.000000000000000D+00,   4.107436271826625D-05, &
   0.000000000000000D+00,  -5.741201724533210D-06,   0.000000000000000D+00, &
  -1.312951462632104D-05,   0.000000000000000D+00,  -7.003307630859236D-06, &
   0.000000000000000D+00,  -1.046533543656542D-06,   0.000000000000000D+00, &
   1.224041320874281D-06,   0.000000000000000D+00,   1.379547926483990D-06, &
   0.000000000000000D+00,  -4.721050327279449D-05,   0.000000000000000D+00, &
   3.757151533908318D-04,   0.000000000000000D+00,  -2.148884341818009D-04, &
   0.000000000000000D+00,  -1.743137893140169D-04,   0.000000000000000D+00, &
  -4.350602910272271D-05,   0.000000000000000D+00,  -9.610029986986167D-06, &
   0.000000000000000D+00,   2.680849504046547D-06,   0.000000000000000D+00, &
   3.960673922590106D-06,   0.000000000000000D+00,   1.350235294510407D-06, &
   0.000000000000000D+00,  -3.470359757059061D-07,   0.000000000000000D+00 /)
lem(:,177) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.460869714515728D-04,  -1.043333575266167D-03,   1.106089339783515D-04, &
   1.568279761547668D-03,   4.021436151992215D-03,  -1.494712501765008D-03, &
  -1.113464888261821D-02,   6.366837583950558D-04,   1.045804926231384D-02, &
   1.504811477656242D-03,  -3.188116494055515D-03,  -1.830726352526702D-03, &
  -7.919868173450268D-04,   3.222480061673951D-04,   6.287652392639870D-04, &
   2.511212919137760D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.460869714515824D-04,  -1.043333575266160D-03, &
   1.106089339783204D-04,   1.568279761547674D-03,   4.021436151992238D-03, &
  -1.494712501765017D-03,  -1.113464888261823D-02,   6.366837583950438D-04, &
   1.045804926231384D-02,   1.504811477656250D-03,  -3.188116494055502D-03, &
  -1.830726352526698D-03,  -7.919868173450481D-04,   3.222480061673972D-04, &
   6.287652392640012D-04,   2.511212919137742D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.642420604135345D-08, &
   0.000000000000000D+00,   1.911633728484183D-06,   0.000000000000000D+00, &
   9.783981770713371D-06,   0.000000000000000D+00,   1.121829299422285D-05, &
   0.000000000000000D+00,  -1.355176956248313D-05,   0.000000000000000D+00, &
  -2.247182353870431D-05,   0.000000000000000D+00,   1.127744957970897D-05, &
   0.000000000000000D+00,   1.577956931606573D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.834098328646732D-06,   0.000000000000000D+00,   7.607337953653208D-08, &
   0.000000000000000D+00,   1.792975275168655D-09,   0.000000000000000D+00, &
  -4.149168908894250D-05,   0.000000000000000D+00,  -1.357170222427859D-06, &
   0.000000000000000D+00,  -8.816395611379123D-08,   0.000000000000000D+00, &
   7.164897291448852D-09,   2.831559191769092D-05,   0.000000000000000D+00, &
   6.294755184242296D-06,   0.000000000000000D+00,   7.633027482843968D-07, &
   0.000000000000000D+00,   6.574631776550523D-08,   0.000000000000000D+00, &
   5.677717311180752D-04,   0.000000000000000D+00,  -1.500899602932385D-06, &
   0.000000000000000D+00,  -1.746877237634281D-06,   0.000000000000000D+00, &
  -4.175437864360708D-07,   0.000000000000000D+00,  -2.268081622432681D-08, &
  -1.775198011261867D-03,   0.000000000000000D+00,   5.308700960875072D-06, &
   0.000000000000000D+00,  -2.576624927117518D-06,   0.000000000000000D+00, &
   4.134373893186849D-07,   0.000000000000000D+00,   2.346587435700196D-07, &
   0.000000000000000D+00,   1.409852091498109D-03,   0.000000000000000D+00, &
  -2.428130676607057D-04,   0.000000000000000D+00,   7.088423579191361D-06, &
   0.000000000000000D+00,   1.971398748822779D-06,   0.000000000000000D+00, &
  -5.269395198611388D-08,   0.000000000000000D+00,  -3.349708736348047D-07, &
   1.943312663602548D-03,   0.000000000000000D+00,   7.634439491787550D-04, &
   0.000000000000000D+00,  -4.112297368041701D-07,   0.000000000000000D+00, &
  -2.511222297283428D-06,   0.000000000000000D+00,  -1.114906195645414D-06, &
   0.000000000000000D+00,   5.564614253927258D-08,   0.000000000000000D+00, &
  -4.394843906726759D-03,   0.000000000000000D+00,  -6.160735649810223D-04, &
   0.000000000000000D+00,   9.914542786249229D-05,   0.000000000000000D+00, &
  -2.499155536815333D-07,   0.000000000000000D+00,   7.013762913260404D-07, &
   0.000000000000000D+00,   5.857609102691576D-07,   0.000000000000000D+00, &
  -3.031154029191968D-07,   2.093680223759923D-03,   0.000000000000000D+00, &
  -8.701130517366598D-04,   0.000000000000000D+00,  -3.902228128640278D-04, &
   0.000000000000000D+00,  -1.691894950413842D-05,   0.000000000000000D+00, &
   6.842699758503816D-08,   0.000000000000000D+00,  -2.503106463777732D-07, &
   0.000000000000000D+00,  -3.037369499034275D-07,   0.000000000000000D+00, &
   1.647689123056277D-03,   0.000000000000000D+00,   1.967830445072523D-03, &
   0.000000000000000D+00,   3.628059152056788D-04,   0.000000000000000D+00, &
   2.545637416870732D-06,   0.000000000000000D+00,   4.468392063962749D-06, &
   0.000000000000000D+00,   2.166008859358890D-07,   0.000000000000000D+00, &
   2.703051808219430D-07,   0.000000000000000D+00,   1.012306380826977D-06, &
  -2.245577164093712D-03,   0.000000000000000D+00,  -9.363925748615912D-04, &
   0.000000000000000D+00,   4.392822744378019D-04,   0.000000000000000D+00, &
   1.530035523553764D-04,   0.000000000000000D+00,   9.857453178522461D-06, &
   0.000000000000000D+00,  -8.937390645205385D-07,   0.000000000000000D+00, &
  -2.439880924282419D-07,   0.000000000000000D+00,  -3.881064275027709D-07, &
   0.000000000000000D+00,   7.839195888719506D-04,   0.000000000000000D+00, &
  -7.198987917058426D-04,   0.000000000000000D+00,  -1.057937631327633D-03, &
   0.000000000000000D+00,  -2.142014354299925D-04,   0.000000000000000D+00, &
  -3.726641646135915D-05,   0.000000000000000D+00,  -7.209450732629709D-06, &
   0.000000000000000D+00,  -4.498794304493893D-07,   0.000000000000000D+00, &
  -5.217432822383241D-08,   0.000000000000000D+00,   8.217241648930325D-07, &
   3.940651516863447D-05,   0.000000000000000D+00,   9.636344417739803D-04, &
   0.000000000000000D+00,   4.724510448396389D-04,   0.000000000000000D+00, &
  -1.463922844156403D-04,   0.000000000000000D+00,  -2.276597316614146D-05, &
   0.000000000000000D+00,   8.122055211943368D-06,   0.000000000000000D+00, &
   4.194456982920293D-06,   0.000000000000000D+00,   9.530956334072193D-07, &
   0.000000000000000D+00,  -5.534359681200354D-08,   0.000000000000000D+00, &
  -6.759066825807615D-05,   0.000000000000000D+00,  -3.328841862478262D-04, &
   0.000000000000000D+00,   4.227321301394397D-04,   0.000000000000000D+00, &
   4.596423102619445D-04,   0.000000000000000D+00,   1.032169398662661D-04, &
   0.000000000000000D+00,   2.151250595388388D-05,   0.000000000000000D+00, &
   2.003709226524959D-07,   0.000000000000000D+00,  -2.030813792563988D-06, &
   0.000000000000000D+00,  -9.390969521505893D-07,   0.000000000000000D+00, &
  -1.574587510882383D-06,   6.922051415933808D-06,   0.000000000000000D+00, &
   8.240562848051663D-06,   0.000000000000000D+00,  -4.905348829299326D-04, &
   0.000000000000000D+00,  -1.577089286189178D-04,   0.000000000000000D+00, &
   1.426480301681312D-05,   0.000000000000000D+00,  -1.732501795717659D-05, &
   0.000000000000000D+00,  -1.131778054296302D-05,   0.000000000000000D+00, &
  -2.225895486357716D-06,   0.000000000000000D+00,   7.338672930018135D-07, &
   0.000000000000000D+00,   7.908989636079195D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.043960871498729D-06, &
   0.000000000000000D+00,   1.352491284223208D-04,   0.000000000000000D+00, &
  -2.485804735978500D-04,   0.000000000000000D+00,  -1.488621318338421D-04, &
   0.000000000000000D+00,  -3.474996131851343D-05,   0.000000000000000D+00, &
  -1.715730144682914D-06,   0.000000000000000D+00,   4.781265751600553D-06, &
   0.000000000000000D+00,   2.146354103447444D-06,   0.000000000000000D+00, &
   4.054007708041892D-07,   0.000000000000000D+00,  -6.603078721838386D-07 /)
lem(:,178) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.986026331270597D-05,   3.494202112811110D-04, &
  -7.347750451642086D-04,  -1.332793111974204D-03,   3.948323368488266D-03, &
   2.987292409921699D-03,  -1.023832425236359D-02,  -3.461683090063511D-03, &
   1.434970424120491D-02,   2.077072287945357D-03,  -1.063965261728157D-02, &
  -1.565618241905439D-03,   3.766223246342766D-03,   1.734513224997521D-03, &
  -1.400156775074172D-04,  -9.971988230527060D-04,  -5.169367424742722D-04, &
   2.054952810330797D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.986026331270669D-05, &
  -3.494202112811342D-04,   7.347750451642251D-04,   1.332793111974213D-03, &
  -3.948323368488296D-03,  -2.987292409921701D-03,   1.023832425236361D-02, &
   3.461683090063494D-03,  -1.434970424120495D-02,  -2.077072287945344D-03, &
   1.063965261728157D-02,   1.565618241905446D-03,  -3.766223246342780D-03, &
  -1.734513224997527D-03,   1.400156775074338D-04,   9.971988230527019D-04, &
   5.169367424742561D-04,  -2.054952810330786D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.402993513898767D-06,   0.000000000000000D+00, &
  -1.279924445751990D-05,   0.000000000000000D+00,  -2.805376787789942D-05, &
   0.000000000000000D+00,  -6.219173452792473D-06,   0.000000000000000D+00, &
   3.393776636274582D-05,   0.000000000000000D+00,   2.379130758068204D-05, &
   0.000000000000000D+00,  -1.486503533105199D-05,   0.000000000000000D+00, &
  -2.095401195277082D-05,   0.000000000000000D+00,   3.604274866202425D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.043831438593263D-08,   0.000000000000000D+00, &
  -4.891549610053248D-09,   0.000000000000000D+00,  -1.320718394733558D-06, &
   0.000000000000000D+00,  -3.802068106049834D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.092551272694835D-06,   0.000000000000000D+00, &
   5.287854065808096D-07,   0.000000000000000D+00,  -5.158694921799383D-08, &
   0.000000000000000D+00,  -8.261357702736005D-06,   0.000000000000000D+00, &
  -1.318450494459361D-06,   0.000000000000000D+00,  -2.073828043001873D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.269732137412867D-05, &
   0.000000000000000D+00,  -3.116930986477132D-06,   0.000000000000000D+00, &
   2.458495057479964D-07,   0.000000000000000D+00,   1.557530540235015D-07, &
   0.000000000000000D+00,   2.325979659402728D-04,   0.000000000000000D+00, &
   1.189733012897058D-05,   0.000000000000000D+00,   1.965632524205918D-06, &
   0.000000000000000D+00,  -7.445049678205289D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.805266222285269D-04,   0.000000000000000D+00, &
   3.554527683197570D-05,   0.000000000000000D+00,  -2.284239812893502D-06, &
   0.000000000000000D+00,  -8.956731831265194D-07,   0.000000000000000D+00, &
   5.339037625905866D-07,   0.000000000000000D+00,   1.429858710357118D-04, &
   0.000000000000000D+00,  -2.274794824574780D-04,   0.000000000000000D+00, &
  -1.235728445556195D-05,   0.000000000000000D+00,  -2.287039036659137D-07, &
   0.000000000000000D+00,   2.176235530782584D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.891642540860370D-04,   0.000000000000000D+00, &
   4.206992517729646D-04,   0.000000000000000D+00,   8.938521253820582D-06, &
   0.000000000000000D+00,   5.346171628261913D-06,   0.000000000000000D+00, &
   5.724354303179598D-07,   0.000000000000000D+00,  -3.507787538348013D-07, &
   0.000000000000000D+00,  -8.802096032423212D-04,   0.000000000000000D+00, &
  -1.898708552895799D-04,   0.000000000000000D+00,   1.125713758475014D-04, &
   0.000000000000000D+00,   3.742300310804665D-06,   0.000000000000000D+00, &
  -1.807898631347241D-06,   0.000000000000000D+00,  -4.265499094061503D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.281837129021248D-04, &
   0.000000000000000D+00,  -5.193203918849652D-04,   0.000000000000000D+00, &
  -2.953152150003184D-04,   0.000000000000000D+00,  -3.052836081705153D-05, &
   0.000000000000000D+00,  -4.612272670857863D-06,   0.000000000000000D+00, &
   2.811041006273336D-07,   0.000000000000000D+00,  -7.066611024246212D-07, &
   0.000000000000000D+00,   1.669781490858917D-04,   0.000000000000000D+00, &
   9.694724901262647D-04,   0.000000000000000D+00,   1.876809488927794D-04, &
   0.000000000000000D+00,  -5.123634076511561D-06,   0.000000000000000D+00, &
   8.958780874747964D-06,   0.000000000000000D+00,   3.211313965712540D-06, &
   0.000000000000000D+00,   6.373321817528780D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.848820621580653D-04,   0.000000000000000D+00, &
  -5.605082929407533D-04,   0.000000000000000D+00,   3.254887454185666D-04, &
   0.000000000000000D+00,   1.334906468598344D-04,   0.000000000000000D+00, &
   1.571734630122341D-05,   0.000000000000000D+00,  -1.369316108065286D-06, &
   0.000000000000000D+00,  -1.653088349774700D-06,   0.000000000000000D+00, &
  -3.628978836308478D-08,   0.000000000000000D+00,   3.278529234587712D-04, &
   0.000000000000000D+00,  -2.216203237384162D-04,   0.000000000000000D+00, &
  -6.711095231939786D-04,   0.000000000000000D+00,  -1.408544965544367D-04, &
   0.000000000000000D+00,  -3.239099120009514D-05,   0.000000000000000D+00, &
  -1.051928810151977D-05,   0.000000000000000D+00,  -1.233405466918691D-06, &
   0.000000000000000D+00,   9.270090967052188D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.595541998716730D-05,   0.000000000000000D+00, &
   5.355868198764417D-04,   0.000000000000000D+00,   3.514366357340658D-04, &
   0.000000000000000D+00,  -1.175592949741628D-04,   0.000000000000000D+00, &
  -2.636271651988766D-05,   0.000000000000000D+00,   5.246594270622900D-06, &
   0.000000000000000D+00,   5.563309496557967D-06,   0.000000000000000D+00, &
   1.330093689132353D-06,   0.000000000000000D+00,   2.022294077524740D-07, &
   0.000000000000000D+00,  -1.169193607997233D-05,   0.000000000000000D+00, &
  -3.206585378080420D-04,   0.000000000000000D+00,   2.184635519489865D-04, &
   0.000000000000000D+00,   3.142137506440675D-04,   0.000000000000000D+00, &
   7.374027779329453D-05,   0.000000000000000D+00,   2.006709486919097D-05, &
   0.000000000000000D+00,   1.975547493503456D-06,   0.000000000000000D+00, &
  -2.341610144373988D-06,   0.000000000000000D+00,  -1.809479694353513D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.355451328160797D-05, &
   0.000000000000000D+00,   5.925692396929402D-05,   0.000000000000000D+00, &
  -3.755524753305420D-04,   0.000000000000000D+00,  -1.272769320472504D-04, &
   0.000000000000000D+00,   1.518657769008571D-05,   0.000000000000000D+00, &
  -9.589302209008892D-06,   0.000000000000000D+00,  -9.504020766117847D-06, &
   0.000000000000000D+00,  -3.018230524485836D-06,   0.000000000000000D+00, &
   5.672157543931072D-07,   0.000000000000000D+00,   1.212178915478443D-06, &
   0.000000000000000D+00,  -3.163925396357155D-06,   0.000000000000000D+00, &
   1.874792966113659D-05,   0.000000000000000D+00,   1.599646342080253D-04, &
   0.000000000000000D+00,  -1.636163454536218D-04,   0.000000000000000D+00, &
  -1.022202003553103D-04,   0.000000000000000D+00,  -2.537949001718391D-05, &
   0.000000000000000D+00,  -3.259847779207474D-06,   0.000000000000000D+00, &
   3.376776893047392D-06,   0.000000000000000D+00,   2.740278956348551D-06, &
   0.000000000000000D+00,   1.548338689854494D-07,   0.000000000000000D+00 /)
lem(:,179) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.777696118332080D-06, &
  -1.257409991569144D-04,   3.410018563753100D-04,   4.573218065077669D-04, &
  -1.085891221231803D-03,   4.012988421872691D-04,  -2.859641900730962D-03, &
  -2.005796915059435D-03,   1.251726790277206D-02,   2.790465566286650D-03, &
  -1.753745917112205D-02,  -1.941817263815376D-03,   1.172379334639750D-02, &
   9.227761539393004D-04,  -3.969819401561566D-03,  -6.512783056825856D-04, &
   9.504934686976768D-04,   5.809307087716277D-04,  -3.143365463721048D-04, &
  -4.108693392105832D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.777696118326820D-06,  -1.257409991569043D-04,   3.410018563753122D-04, &
   4.573218065077530D-04,  -1.085891221231799D-03,   4.012988421872722D-04, &
  -2.859641900730974D-03,  -2.005796915059433D-03,   1.251726790277202D-02, &
   2.790465566286665D-03,  -1.753745917112205D-02,  -1.941817263815380D-03, &
   1.172379334639752D-02,   9.227761539393095D-04,  -3.969819401561585D-03, &
  -6.512783056825901D-04,   9.504934686976981D-04,   5.809307087716143D-04, &
  -3.143365463721144D-04,  -4.108693392105621D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.357641708696746D-07,   0.000000000000000D+00, &
   1.287458539691790D-05,   0.000000000000000D+00,   4.479844118517605D-05, &
   0.000000000000000D+00,   3.685593945811778D-05,   0.000000000000000D+00, &
  -1.928743836846898D-05,   0.000000000000000D+00,  -2.998042851333533D-05, &
   0.000000000000000D+00,  -2.197532194991138D-05,   0.000000000000000D+00, &
  -5.145121043346034D-06,   0.000000000000000D+00,   3.472517605916126D-05, &
   0.000000000000000D+00,   7.844074924565336D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.431889976783689D-07,   0.000000000000000D+00, &
  -2.330400937212237D-06,   0.000000000000000D+00,   2.953273077124572D-08, &
   9.330742052893008D-06,   0.000000000000000D+00,   3.953428215008890D-07, &
   0.000000000000000D+00,   1.276561042343781D-05,   0.000000000000000D+00, &
  -1.275410416469557D-06,   0.000000000000000D+00,   3.842367804502671D-09, &
  -1.050655869608626D-04,   0.000000000000000D+00,  -1.456755724538464D-06, &
   0.000000000000000D+00,   2.024754127396994D-07,   0.000000000000000D+00, &
   2.317428014022498D-05,   0.000000000000000D+00,   4.539315792993519D-06, &
   0.000000000000000D+00,   1.534128922212335D-06,   0.000000000000000D+00, &
  -6.191644981564046D-07,   5.054700695657304D-04,   0.000000000000000D+00, &
   4.887471836494981D-07,   0.000000000000000D+00,  -3.306155765709354D-06, &
   0.000000000000000D+00,  -5.620455120513854D-07,   0.000000000000000D+00, &
  -6.666893303721948D-04,   0.000000000000000D+00,   8.136676194288234D-05, &
   0.000000000000000D+00,  -3.459100996142170D-06,   0.000000000000000D+00, &
   6.638633951150335D-07,   0.000000000000000D+00,   1.356017803911444D-07, &
  -4.533626030810866D-04,   0.000000000000000D+00,  -3.244001365745442D-04, &
   0.000000000000000D+00,   2.861175016959822D-05,   0.000000000000000D+00, &
   3.625491979785707D-06,   0.000000000000000D+00,   1.013013772579727D-07, &
   0.000000000000000D+00,   1.404543867120350D-03,   0.000000000000000D+00, &
   3.035076064489735D-04,   0.000000000000000D+00,  -1.235404786675542D-04, &
   0.000000000000000D+00,  -7.460907856328600D-06,   0.000000000000000D+00, &
  -2.109224615305695D-06,   0.000000000000000D+00,   1.117143760986499D-06, &
  -4.963302949880305D-04,   0.000000000000000D+00,   3.739386631665689D-04, &
   0.000000000000000D+00,   2.902763817019073D-04,   0.000000000000000D+00, &
  -4.337834968357165D-06,   0.000000000000000D+00,   1.525998207762396D-06, &
   0.000000000000000D+00,   9.925132034820809D-07,   0.000000000000000D+00, &
  -7.996733717146501D-04,   0.000000000000000D+00,  -9.035317594273299D-04, &
   0.000000000000000D+00,  -2.046669110088177D-04,   0.000000000000000D+00, &
   7.507521522235252D-05,   0.000000000000000D+00,   6.764071832690063D-06, &
   0.000000000000000D+00,  -2.876198707155615D-08,   0.000000000000000D+00, &
  -1.456827488351373D-06,   6.716892231970756D-04,   0.000000000000000D+00, &
   4.064145781598478D-04,   0.000000000000000D+00,  -3.677785130024936D-04, &
   0.000000000000000D+00,  -2.057537419804512D-04,   0.000000000000000D+00, &
  -1.841302069917898D-05,   0.000000000000000D+00,  -4.377687974023204D-06, &
   0.000000000000000D+00,  -1.020717848847430D-06,   0.000000000000000D+00, &
   2.659329335774388D-05,   0.000000000000000D+00,   3.628059152056788D-04, &
   0.000000000000000D+00,   7.877250051955240D-04,   0.000000000000000D+00, &
   1.602272022883593D-04,   0.000000000000000D+00,  -4.645094181567547D-06, &
   0.000000000000000D+00,   4.276392920537145D-06,   0.000000000000000D+00, &
   2.895598152075911D-06,   0.000000000000000D+00,   5.961183358581864D-07, &
  -1.690381096231455D-04,   0.000000000000000D+00,  -4.235046645220766D-04, &
   0.000000000000000D+00,  -3.982393130350830D-04,   0.000000000000000D+00, &
   2.375815448387930D-04,   0.000000000000000D+00,   9.469929493674874D-05, &
   0.000000000000000D+00,   1.122130008654056D-05,   0.000000000000000D+00, &
   1.413504040371420D-07,   0.000000000000000D+00,  -1.159916088792988D-06, &
   0.000000000000000D+00,   2.579885217513818D-05,   0.000000000000000D+00, &
   1.171374753498165D-04,   0.000000000000000D+00,  -2.424020885477146D-04, &
   0.000000000000000D+00,  -5.334347479941962D-04,   0.000000000000000D+00, &
  -1.142301393943694D-04,   0.000000000000000D+00,  -2.368585808469808D-05, &
   0.000000000000000D+00,  -7.465139723148578D-06,   0.000000000000000D+00, &
  -1.102787635713732D-06,   0.000000000000000D+00,   2.013274794746220D-06, &
   1.790888370397503D-05,   0.000000000000000D+00,  -1.456337706856421D-06, &
   0.000000000000000D+00,   3.731889742624513D-04,   0.000000000000000D+00, &
   2.571744418177518D-04,   0.000000000000000D+00,  -8.617532767602914D-05, &
   0.000000000000000D+00,  -1.668350523440920D-05,   0.000000000000000D+00, &
   4.348165751845187D-06,   0.000000000000000D+00,   3.732167139882350D-06, &
   0.000000000000000D+00,   1.711530758311665D-06,   0.000000000000000D+00, &
  -5.322704336100407D-06,   0.000000000000000D+00,   1.156729562664227D-05, &
   0.000000000000000D+00,  -1.760906098022989D-04,   0.000000000000000D+00, &
   1.876198103959929D-04,   0.000000000000000D+00,   2.525497092025625D-04, &
   0.000000000000000D+00,   5.933613208453420D-05,   0.000000000000000D+00, &
   1.443557398837055D-05,   0.000000000000000D+00,   1.059433156608613D-06, &
   0.000000000000000D+00,  -2.328767603356499D-06,   0.000000000000000D+00, &
  -3.207143222034068D-06,   3.890599318688151D-07,   0.000000000000000D+00, &
  -4.899566823779754D-06,   0.000000000000000D+00,   5.120929819892323D-05, &
   0.000000000000000D+00,  -2.626115330995250D-04,   0.000000000000000D+00, &
  -9.357907684213191D-05,   0.000000000000000D+00,   9.320310309623323D-06, &
   0.000000000000000D+00,  -9.519218156999232D-06,   0.000000000000000D+00, &
  -7.318818808289693D-06,   0.000000000000000D+00,  -2.223353064387005D-06, &
   0.000000000000000D+00,   2.486628964156903D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.101389343676275D-06, &
   0.000000000000000D+00,  -1.228745231311116D-05,   0.000000000000000D+00, &
   1.089746952637811D-04,   0.000000000000000D+00,  -1.329741775603131D-04, &
   0.000000000000000D+00,  -8.403874042497026D-05,   0.000000000000000D+00, &
  -2.031179966916721D-05,   0.000000000000000D+00,  -1.538190058105250D-06, &
   0.000000000000000D+00,   3.283649657050082D-06,   0.000000000000000D+00, &
   2.161217760526612D-06,   0.000000000000000D+00,   3.260155309080622D-07 /)
lem(:,180) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.369277386463765D-05,  -1.388700730887959D-04, &
   1.510008270154434D-04,   9.822991830333065D-04,  -1.207149538753379D-03, &
  -2.477139577296012D-03,   3.653583174127308D-04,   7.985095153480162D-03, &
   2.061044208677530D-03,  -1.621532959335080D-02,  -1.853069140733937D-03, &
   1.829606133929032D-02,  -9.040633711134954D-04,  -1.142678960230219D-02, &
   1.748101527372437D-03,   3.998339717743167D-03,  -7.148137999688767D-04, &
  -1.058784519498545D-03,   1.431537007594930D-04,   4.512972691060259D-04, &
  -4.849004673451829D-05,   0.000000000000000D+00,  -1.369277386463940D-05, &
   1.388700730887950D-04,  -1.510008270154327D-04,  -9.822991830333080D-04, &
   1.207149538753359D-03,   2.477139577296020D-03,  -3.653583174127214D-04, &
  -7.985095153480167D-03,  -2.061044208677544D-03,   1.621532959335081D-02, &
   1.853069140733936D-03,  -1.829606133929036D-02,   9.040633711135071D-04, &
   1.142678960230219D-02,  -1.748101527372432D-03,  -3.998339717743166D-03, &
   7.148137999688679D-04,   1.058784519498547D-03,  -1.431537007594948D-04, &
  -4.512972691060368D-04,   4.849004673452320D-05,   0.000000000000000D+00, &
  -6.846386932314541D-06,   0.000000000000000D+00,  -5.425122119617099D-05, &
   0.000000000000000D+00,  -7.604811551700834D-05,   0.000000000000000D+00, &
  -2.291784202785407D-06,   0.000000000000000D+00,   1.955651105732899D-05, &
   0.000000000000000D+00,  -1.032021580701383D-05,   0.000000000000000D+00, &
   4.332427771607425D-05,   0.000000000000000D+00,   3.934282147989523D-05, &
   0.000000000000000D+00,  -4.913768472345652D-05,   0.000000000000000D+00, &
  -2.927834683283984D-05,   0.000000000000000D+00,   2.516308153449473D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.412544112235568D-07, &
   0.000000000000000D+00,  -2.257374548396319D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.930393965525743D-07,   0.000000000000000D+00, &
   6.913304884093713D-07,   0.000000000000000D+00,   1.819281584391291D-05, &
   0.000000000000000D+00,   8.979584376678043D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.691834869786152D-05,   0.000000000000000D+00, &
  -1.566671067709574D-06,   0.000000000000000D+00,   2.649210462342702D-07, &
   0.000000000000000D+00,  -4.807302895874677D-05,   0.000000000000000D+00, &
  -9.896346375732134D-06,   0.000000000000000D+00,   7.833340886240680D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.786150763132858D-04, &
   0.000000000000000D+00,   9.088481388156501D-06,   0.000000000000000D+00, &
   6.619926970076009D-07,   0.000000000000000D+00,  -2.144005340410579D-06, &
   0.000000000000000D+00,  -1.351503673485099D-04,   0.000000000000000D+00, &
   8.888833680191860D-05,   0.000000000000000D+00,   1.995249967722634D-06, &
   0.000000000000000D+00,  -4.598826913614629D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.712325585799065D-04,   0.000000000000000D+00, &
  -2.434332279326576D-04,   0.000000000000000D+00,   6.379963301899012D-06, &
   0.000000000000000D+00,   1.298369132562248D-06,   0.000000000000000D+00, &
   4.310093926392462D-07,   0.000000000000000D+00,   3.855576560226554D-04, &
   0.000000000000000D+00,   1.579911173779101D-04,   0.000000000000000D+00, &
  -9.784587981998258D-05,   0.000000000000000D+00,  -4.615683456645781D-06, &
   0.000000000000000D+00,  -6.915492643794377D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.882643597836635D-04,   0.000000000000000D+00, &
   2.747574463569521D-04,   0.000000000000000D+00,   2.383564601360798D-04, &
   0.000000000000000D+00,   5.429009664625538D-06,   0.000000000000000D+00, &
   1.544917230930557D-06,   0.000000000000000D+00,   3.296692078200491D-06, &
   0.000000000000000D+00,  -1.609820624297830D-04,   0.000000000000000D+00, &
  -5.710790946099737D-04,   0.000000000000000D+00,  -1.452683681129303D-04, &
   0.000000000000000D+00,   5.420776223777913D-05,   0.000000000000000D+00, &
   2.663068377727478D-06,   0.000000000000000D+00,  -4.502467716076958D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.397416291655131D-04, &
   0.000000000000000D+00,   2.998900871970221D-04,   0.000000000000000D+00, &
  -2.892886446615076D-04,   0.000000000000000D+00,  -1.705560269291402D-04, &
   0.000000000000000D+00,  -1.762300970295542D-05,   0.000000000000000D+00, &
  -3.957907812096800D-06,   0.000000000000000D+00,  -2.431159770844963D-06, &
   0.000000000000000D+00,  -7.939384358809687D-05,   0.000000000000000D+00, &
   1.876809488927794D-04,   0.000000000000000D+00,   5.856615073581735D-04, &
   0.000000000000000D+00,   1.265406834269395D-04,   0.000000000000000D+00, &
   1.113143233626287D-06,   0.000000000000000D+00,   5.872183069518663D-06, &
   0.000000000000000D+00,   2.244368133763348D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.883596126012328D-05,   0.000000000000000D+00, &
  -3.218310548684547D-04,   0.000000000000000D+00,  -3.142273045499182D-04, &
   0.000000000000000D+00,   1.909570075821797D-04,   0.000000000000000D+00, &
   7.898178639000583D-05,   0.000000000000000D+00,   9.084673666498285D-06, &
   0.000000000000000D+00,  -3.674237260125896D-07,   0.000000000000000D+00, &
  -3.078971829889313D-06,   0.000000000000000D+00,   3.618481855394546D-05, &
   0.000000000000000D+00,   1.362662019472518D-04,   0.000000000000000D+00, &
  -1.710955035722903D-04,   0.000000000000000D+00,  -4.163540085443533D-04, &
   0.000000000000000D+00,  -9.331730369472067D-05,   0.000000000000000D+00, &
  -2.228094418100337D-05,   0.000000000000000D+00,  -6.809991668197071D-06, &
   0.000000000000000D+00,  -2.784542346713600D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.781982611104446D-06,   0.000000000000000D+00, &
   9.604706399547819D-06,   0.000000000000000D+00,   3.097244396655799D-04, &
   0.000000000000000D+00,   2.072156258011585D-04,   0.000000000000000D+00, &
  -7.044139784614882D-05,   0.000000000000000D+00,  -1.401480983130200D-05, &
   0.000000000000000D+00,   4.433491271351753D-06,   0.000000000000000D+00, &
   4.526814058202821D-06,   0.000000000000000D+00,   4.070249897720250D-06, &
   0.000000000000000D+00,  -4.365397485900213D-06,   0.000000000000000D+00, &
  -1.762666737027510D-05,   0.000000000000000D+00,  -1.445892333935517D-04, &
   0.000000000000000D+00,   1.473096123214649D-04,   0.000000000000000D+00, &
   2.010431488649749D-04,   0.000000000000000D+00,   4.903998561187980D-05, &
   0.000000000000000D+00,   1.296356315851790D-05,   0.000000000000000D+00, &
   2.572961290598971D-07,   0.000000000000000D+00,  -2.357736202758533D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.039140173808006D-06, &
   0.000000000000000D+00,  -3.999233505975092D-06,   0.000000000000000D+00, &
   1.626712402970330D-05,   0.000000000000000D+00,  -2.190138425404549D-04, &
   0.000000000000000D+00,  -7.675153363896602D-05,   0.000000000000000D+00, &
   8.055490605335464D-06,   0.000000000000000D+00,  -7.824664634111138D-06, &
   0.000000000000000D+00,  -7.002125323612567D-06,   0.000000000000000D+00, &
  -2.773364182296379D-06,   0.000000000000000D+00,   1.028270984330977D-06, &
   0.000000000000000D+00,  -3.290634184863721D-07,   0.000000000000000D+00, &
   4.455390625466726D-06,   0.000000000000000D+00,  -2.310684650970229D-06, &
   0.000000000000000D+00,   8.119637107568813D-05,   0.000000000000000D+00, &
  -1.080070911692979D-04,   0.000000000000000D+00,  -6.723916750081515D-05, &
   0.000000000000000D+00,  -1.687837254385695D-05,   0.000000000000000D+00, &
  -1.238273132275045D-06,   0.000000000000000D+00,   3.070456253104014D-06, &
   0.000000000000000D+00,   1.477966531305627D-06,   0.000000000000000D+00 /)
lem(:,181) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.015932469596923D-05,  -3.418439030788403D-05,  -4.508284723018404D-04, &
   2.687683956614321D-04,   3.739056531334059D-04,   6.258724432571664D-04, &
   2.003633716140607D-03,  -1.232456524652356D-03,  -9.346852132872504D-03, &
   2.582185589756376D-04,   1.726021865111396D-02,  -1.128715571725725D-03, &
  -1.654439133177068D-02,   4.191445811044446D-03,   8.358106796982621D-03, &
  -4.454020940772245D-03,  -2.107740978292657D-03,   2.142255221422325D-03, &
   3.408972982809803D-04,  -6.824860541909559D-04,  -1.000847912006628D-04, &
   2.877956830500294D-04,   5.015932469597371D-05,  -3.418439030789369D-05, &
  -4.508284723018412D-04,   2.687683956614397D-04,   3.739056531334059D-04, &
   6.258724432571569D-04,   2.003633716140610D-03,  -1.232456524652341D-03, &
  -9.346852132872487D-03,   2.582185589756300D-04,   1.726021865111396D-02, &
  -1.128715571725716D-03,  -1.654439133177071D-02,   4.191445811044450D-03, &
   8.358106796982619D-03,  -4.454020940772231D-03,  -2.107740978292659D-03, &
   2.142255221422312D-03,   3.408972982809892D-04,  -6.824860541909514D-04, &
  -1.000847912006759D-04,   2.877956830500372D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.212725150300269D-05,   0.000000000000000D+00, &
   5.595582732330444D-06,   0.000000000000000D+00,  -4.138752093323168D-05, &
   0.000000000000000D+00,   5.337445832348575D-05,   0.000000000000000D+00, &
   5.301457632164633D-05,   0.000000000000000D+00,  -6.134252986526779D-05, &
   0.000000000000000D+00,  -9.136428958771743D-05,   0.000000000000000D+00, &
   3.931410346108217D-05,   0.000000000000000D+00,   7.452794032367416D-05, &
   0.000000000000000D+00,  -2.669529007173258D-05,   0.000000000000000D+00, &
  -3.612121452678448D-06,  -7.623203250138448D-06,   0.000000000000000D+00, &
   2.371478280921001D-05,   0.000000000000000D+00,  -2.271872434411251D-06, &
   2.643570060954652D-05,   0.000000000000000D+00,  -2.145500361939718D-08, &
   0.000000000000000D+00,  -1.005601389233102D-04,   0.000000000000000D+00, &
   1.657579961355663D-06,   0.000000000000000D+00,   2.487299405941783D-06, &
  -4.507004439700642D-05,   0.000000000000000D+00,  -1.409430704675697D-05, &
   0.000000000000000D+00,   2.251880935186992D-06,   0.000000000000000D+00, &
   2.366974831303813D-04,   0.000000000000000D+00,   3.818348628383754D-06, &
   0.000000000000000D+00,  -8.329611870738654D-06,   0.000000000000000D+00, &
   2.049868301950496D-06,   2.667317800579311D-05,   0.000000000000000D+00, &
   1.026789508794151D-04,   0.000000000000000D+00,   1.256692109788950D-06, &
   0.000000000000000D+00,  -2.514270376596175D-06,   0.000000000000000D+00, &
  -3.365347121656859D-04,   0.000000000000000D+00,  -1.476084171433479D-04, &
   0.000000000000000D+00,   5.845249249738825D-05,   0.000000000000000D+00, &
   7.019092098468469D-06,   0.000000000000000D+00,  -2.477696909299165D-06, &
   3.086671473405024D-05,   0.000000000000000D+00,  -1.021877582774766D-04, &
   0.000000000000000D+00,  -1.704283164992861D-04,   0.000000000000000D+00, &
   4.544995340002747D-06,   0.000000000000000D+00,   6.941770019847597D-07, &
   0.000000000000000D+00,   2.966000363011027D-04,   0.000000000000000D+00, &
   3.259432718368113D-04,   0.000000000000000D+00,   1.344140876422262D-04, &
   0.000000000000000D+00,  -7.688818814744920D-05,   0.000000000000000D+00, &
  -7.209397371890993D-06,   0.000000000000000D+00,  -1.935863218990548D-06, &
  -7.017950960470541D-05,   0.000000000000000D+00,  -1.130084385252220D-04, &
   0.000000000000000D+00,   1.998851705939908D-04,   0.000000000000000D+00, &
   1.854125571525330D-04,   0.000000000000000D+00,   8.266603341730594D-06, &
   0.000000000000000D+00,   4.242411299612531D-06,   0.000000000000000D+00, &
  -1.570464753676090D-04,   0.000000000000000D+00,  -1.803890975786334D-04, &
   0.000000000000000D+00,  -4.512538479130094D-04,   0.000000000000000D+00, &
  -1.216804704551154D-04,   0.000000000000000D+00,   4.308874643146215D-05, &
   0.000000000000000D+00,   2.919795739225541D-06,   0.000000000000000D+00, &
   4.656927441083732D-07,   5.612450043020509D-05,   0.000000000000000D+00, &
   1.401646999228513D-04,   0.000000000000000D+00,   2.184300020156041D-04, &
   0.000000000000000D+00,  -2.264396392698717D-04,   0.000000000000000D+00, &
  -1.388955009740331D-04,   0.000000000000000D+00,  -1.829365795301036D-05, &
   0.000000000000000D+00,  -4.776052597330330D-06,   0.000000000000000D+00, &
   4.387610906754080D-05,   0.000000000000000D+00,   2.545637416870732D-06, &
   0.000000000000000D+00,   1.602272022883593D-04,   0.000000000000000D+00, &
   4.738416993960265D-04,   0.000000000000000D+00,   1.076899368698198D-04, &
   0.000000000000000D+00,   3.035825373699619D-06,   0.000000000000000D+00, &
   5.476989025035603D-06,   0.000000000000000D+00,   5.303681662892724D-06, &
  -2.163807856446797D-05,   0.000000000000000D+00,  -1.796596422933300D-05, &
   0.000000000000000D+00,  -2.243547912690518D-04,   0.000000000000000D+00, &
  -2.458569693853017D-04,   0.000000000000000D+00,   1.537473234836114D-04, &
   0.000000000000000D+00,   6.526974204569158D-05,   0.000000000000000D+00, &
   8.712077792817478D-06,   0.000000000000000D+00,  -1.733889189062117D-06, &
   0.000000000000000D+00,  -3.608257255857885D-06,   0.000000000000000D+00, &
  -8.269890654143804D-07,   0.000000000000000D+00,   9.196547693987144D-05, &
   0.000000000000000D+00,  -1.420935563817551D-04,   0.000000000000000D+00, &
  -3.440212554461401D-04,   0.000000000000000D+00,  -8.029276775973197D-05, &
   0.000000000000000D+00,  -1.979698309666543D-05,   0.000000000000000D+00, &
  -6.590440893456512D-06,   0.000000000000000D+00,  -2.477030279803564D-06, &
   3.293582936805048D-06,   0.000000000000000D+00,  -3.724085687538793D-06, &
   0.000000000000000D+00,  -1.057066802714807D-05,   0.000000000000000D+00, &
   2.385317400267561D-04,   0.000000000000000D+00,   1.666737861777551D-04, &
   0.000000000000000D+00,  -5.685990454145791D-05,   0.000000000000000D+00, &
  -1.097487800702617D-05,   0.000000000000000D+00,   4.981744513938477D-06, &
   0.000000000000000D+00,   4.940306325503907D-06,   0.000000000000000D+00, &
  -5.857099408164771D-07,   0.000000000000000D+00,   3.689366112200712D-06, &
   0.000000000000000D+00,  -8.432469217346670D-06,   0.000000000000000D+00, &
  -1.166483141777014D-04,   0.000000000000000D+00,   1.236249262799109D-04, &
   0.000000000000000D+00,   1.687830405024037D-04,   0.000000000000000D+00, &
   4.220800440264299D-05,   0.000000000000000D+00,   1.076480609259697D-05, &
   0.000000000000000D+00,   7.083835317861039D-07,   0.000000000000000D+00, &
  -5.330612484135430D-06,   3.326117486812377D-08,   0.000000000000000D+00, &
   4.296764951876112D-07,   0.000000000000000D+00,   7.338633395447437D-06, &
   0.000000000000000D+00,   2.015434880733202D-05,   0.000000000000000D+00, &
  -1.761323965525997D-04,   0.000000000000000D+00,  -6.197434178259917D-05, &
   0.000000000000000D+00,   5.879238946655952D-06,   0.000000000000000D+00, &
  -7.660543862135279D-06,   0.000000000000000D+00,  -7.020317243879304D-06, &
   0.000000000000000D+00,  -1.656597098103416D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.280332549827649D-07, &
   0.000000000000000D+00,   2.065510025886052D-07,   0.000000000000000D+00, &
   5.147693781273942D-06,   0.000000000000000D+00,   6.790267717088057D-05, &
   0.000000000000000D+00,  -9.142154656976810D-05,   0.000000000000000D+00, &
  -5.735464865051421D-05,   0.000000000000000D+00,  -1.415758490152761D-05, &
   0.000000000000000D+00,  -9.089899000337996D-07,   0.000000000000000D+00, &
   2.988859720824322D-06,   0.000000000000000D+00,   5.488669231007688D-06 /)
lem(:,182) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.679599774343459D-04,   4.076997913996403D-04, &
   1.291711689626944D-03,  -1.298645644435176D-04,  -2.386826415283799D-03, &
  -3.242696095889234D-03,   2.288027957854623D-03,   9.640105260380689D-03, &
  -1.641553043489166D-03,  -1.450243598463004D-02,   3.762751340339648D-03, &
   1.178724839448836D-02,  -6.318880854197819D-03,  -4.028024986890290D-03, &
   4.862000463210649D-03,  -6.014113877501633D-04,  -1.824846522756530D-03, &
   9.333305453176801D-04,   4.300330429641356D-04,  -4.076930620372315D-04, &
  -1.222032305652915D-04,   0.000000000000000D+00,   2.679599774343347D-04, &
  -4.076997913996378D-04,  -1.291711689626937D-03,   1.298645644435127D-04, &
   2.386826415283786D-03,   3.242696095889228D-03,  -2.288027957854616D-03, &
  -9.640105260380678D-03,   1.641553043489164D-03,   1.450243598463007D-02, &
  -3.762751340339648D-03,  -1.178724839448836D-02,   6.318880854197813D-03, &
   4.028024986890293D-03,  -4.862000463210638D-03,   6.014113877501609D-04, &
   1.824846522756526D-03,  -9.333305453176766D-04,  -4.300330429641372D-04, &
   4.076930620372294D-04,   1.222032305652881D-04,   0.000000000000000D+00, &
   1.339799887171674D-04,   0.000000000000000D+00,   2.898182522102199D-04, &
   0.000000000000000D+00,   5.978148985472547D-05,   0.000000000000000D+00, &
  -1.203605171659938D-04,   0.000000000000000D+00,  -1.285853106929746D-04, &
   0.000000000000000D+00,   5.273284008844743D-06,   0.000000000000000D+00, &
   1.876907662384850D-04,   0.000000000000000D+00,   2.267690813302584D-05, &
   0.000000000000000D+00,  -1.444214719607031D-04,   0.000000000000000D+00, &
   5.138462043292760D-06,   0.000000000000000D+00,   4.449100551211303D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.059204813074785D-05, &
   0.000000000000000D+00,   4.988346706613032D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.071927675979365D-05,   0.000000000000000D+00, &
   1.703606663897508D-06,   0.000000000000000D+00,  -2.482585939309588D-05, &
   0.000000000000000D+00,   2.292093989327709D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.503967845568326D-05,   0.000000000000000D+00, &
  -2.094282435681002D-05,   0.000000000000000D+00,   5.483033429362776D-06, &
   0.000000000000000D+00,   6.787405802246298D-05,   0.000000000000000D+00, &
   1.849017425642819D-06,   0.000000000000000D+00,  -7.551089010256540D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.109267786848883D-05, &
   0.000000000000000D+00,   8.692609111995987D-05,   0.000000000000000D+00, &
   8.977789008433193D-06,   0.000000000000000D+00,  -4.616956916959449D-06, &
   0.000000000000000D+00,  -1.018603868499546D-04,   0.000000000000000D+00, &
  -9.883020727347025D-05,   0.000000000000000D+00,   4.096678536321741D-05, &
   0.000000000000000D+00,   9.384047147793865D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.449385906457688D-05,   0.000000000000000D+00, &
  -8.002596214609605D-05,   0.000000000000000D+00,  -1.419315535832828D-04, &
   0.000000000000000D+00,  -4.097577548752158D-06,   0.000000000000000D+00, &
   5.069017900853515D-07,   0.000000000000000D+00,   8.399386293450081D-05, &
   0.000000000000000D+00,   2.216066669864345D-04,   0.000000000000000D+00, &
   1.082781125951815D-04,   0.000000000000000D+00,  -5.834681952897042D-05, &
   0.000000000000000D+00,  -1.076643368055746D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.468067433090104D-05,   0.000000000000000D+00, &
  -8.910337100729531D-05,   0.000000000000000D+00,   1.604430329581977D-04, &
   0.000000000000000D+00,   1.587723109789904D-04,   0.000000000000000D+00, &
   1.454049514363368D-05,   0.000000000000000D+00,   7.797471344084743D-06, &
   0.000000000000000D+00,  -3.182276834896121D-05,   0.000000000000000D+00, &
  -1.189842100736601D-04,   0.000000000000000D+00,  -3.531848630842476D-04, &
   0.000000000000000D+00,  -1.053450275603329D-04,   0.000000000000000D+00, &
   3.242836598611328D-05,   0.000000000000000D+00,   3.367862484588417D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.138409301572558D-05, &
   0.000000000000000D+00,   1.214632515765755D-04,   0.000000000000000D+00, &
   1.752828703536348D-04,   0.000000000000000D+00,  -1.873667122024967D-04, &
   0.000000000000000D+00,  -1.223099796146541D-04,   0.000000000000000D+00, &
  -1.974522369611859D-05,   0.000000000000000D+00,  -1.016510603502516D-05, &
   0.000000000000000D+00,  -1.853303474099335D-06,   0.000000000000000D+00, &
  -5.123634076511561D-06,   0.000000000000000D+00,   1.265406834269395D-04, &
   0.000000000000000D+00,   3.898658041016049D-04,   0.000000000000000D+00, &
   9.626497960193818D-05,   0.000000000000000D+00,   5.910208229451754D-06, &
   0.000000000000000D+00,   7.780649070644779D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.371829641345200D-05,   0.000000000000000D+00, &
  -3.326226984985677D-05,   0.000000000000000D+00,  -1.861947056843112D-04, &
   0.000000000000000D+00,  -2.027933376166766D-04,   0.000000000000000D+00, &
   1.297938000679667D-04,   0.000000000000000D+00,   5.788436421529511D-05, &
   0.000000000000000D+00,   7.698739881631029D-06,   0.000000000000000D+00, &
  -2.141273604675251D-06,   0.000000000000000D+00,   5.466574519337734D-06, &
   0.000000000000000D+00,   4.682030758260895D-06,   0.000000000000000D+00, &
   6.761301436470499D-05,   0.000000000000000D+00,  -1.194299789561205D-04, &
   0.000000000000000D+00,  -2.905492879264727D-04,   0.000000000000000D+00, &
  -7.200405318426007D-05,   0.000000000000000D+00,  -1.994339794718193D-05, &
   0.000000000000000D+00,  -7.094090614889026D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.344699976263584D-07,   0.000000000000000D+00, &
   7.469871221745377D-06,   0.000000000000000D+00,   2.341906484866961D-06, &
   0.000000000000000D+00,   1.998825280730075D-04,   0.000000000000000D+00, &
   1.399311820285290D-04,   0.000000000000000D+00,  -4.853173148260311D-05, &
   0.000000000000000D+00,  -8.456431114092623D-06,   0.000000000000000D+00, &
   5.565143593915274D-06,   0.000000000000000D+00,   1.195110697020075D-05, &
   0.000000000000000D+00,  -1.191236221004669D-06,   0.000000000000000D+00, &
  -4.131366408568506D-07,   0.000000000000000D+00,  -2.097688995402541D-06, &
   0.000000000000000D+00,  -9.039441275682023D-05,   0.000000000000000D+00, &
   1.068732058561280D-04,   0.000000000000000D+00,   1.446656268055321D-04, &
   0.000000000000000D+00,   3.760643035360819D-05,   0.000000000000000D+00, &
   1.081016202728983D-05,   0.000000000000000D+00,  -6.193035056775716D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.340818332301639D-07, &
   0.000000000000000D+00,  -2.290882283573376D-06,   0.000000000000000D+00, &
  -8.759580553551124D-08,   0.000000000000000D+00,   1.238312325300816D-05, &
   0.000000000000000D+00,  -1.497684398357018D-04,   0.000000000000000D+00, &
  -5.231710116300109D-05,   0.000000000000000D+00,   4.596504221212211D-06, &
   0.000000000000000D+00,  -7.925630517438175D-06,   0.000000000000000D+00, &
  -7.505843494125838D-06,   0.000000000000000D+00,  -6.038621560156226D-06, &
   0.000000000000000D+00,  -4.061364510152360D-08,   0.000000000000000D+00, &
   4.686935834051351D-07,   0.000000000000000D+00,  -3.035962484862383D-06, &
   0.000000000000000D+00,   4.748933576281288D-07,   0.000000000000000D+00, &
   5.298053971740661D-05,   0.000000000000000D+00,  -7.994497350209886D-05, &
   0.000000000000000D+00,  -4.931806657883116D-05,   0.000000000000000D+00, &
  -1.292071576528403D-05,   0.000000000000000D+00,  -6.273306131061563D-07, &
   0.000000000000000D+00,   2.582966090654734D-06,   0.000000000000000D+00 /)
lem(:,183) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.923786875041224D-05,  -6.168786501445891D-04,  -5.121630057221776D-04, &
  -4.881433114741441D-04,   5.636317185307455D-04,   2.398777672575393D-03, &
   2.747439246274457D-03,  -2.671097435043337D-03,  -8.078774791628423D-03, &
   1.922185149093309D-03,   1.034730677693563D-02,  -4.019276819273997D-03, &
  -6.289541315945032D-03,   5.768628762983314D-03,   1.585257582948452D-04, &
  -3.255502744518317D-03,   2.037911572575159D-03,   2.765815802783391D-04, &
  -1.126483274558094D-03,   4.283436573343751D-04,   3.145665554844923D-04, &
  -2.592053229407202D-04,  -1.923786875041079D-05,  -6.168786501445997D-04, &
  -5.121630057221814D-04,  -4.881433114741329D-04,   5.636317185307388D-04, &
   2.398777672575386D-03,   2.747439246274460D-03,  -2.671097435043329D-03, &
  -8.078774791628426D-03,   1.922185149093305D-03,   1.034730677693565D-02, &
  -4.019276819274000D-03,  -6.289541315945038D-03,   5.768628762983313D-03, &
   1.585257582948445D-04,  -3.255502744518314D-03,   2.037911572575154D-03, &
   2.765815802783389D-04,  -1.126483274558088D-03,   4.283436573343690D-04, &
   3.145665554844876D-04,  -2.592053229407076D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.791740803193886D-05,   0.000000000000000D+00, &
  -3.145445687843006D-04,   0.000000000000000D+00,  -1.956522476339603D-04, &
   0.000000000000000D+00,   3.002402268494826D-04,   0.000000000000000D+00, &
   2.642028975495887D-04,   0.000000000000000D+00,  -2.856975527836732D-04, &
   0.000000000000000D+00,  -2.225335785644296D-04,   0.000000000000000D+00, &
   2.206049652289355D-04,   0.000000000000000D+00,   7.785787537576455D-05, &
   0.000000000000000D+00,  -8.691300789326968D-05,   0.000000000000000D+00, &
  -8.575633817506873D-06,   1.914083401119011D-05,   0.000000000000000D+00, &
   2.866319320943947D-05,   0.000000000000000D+00,   8.517222774430556D-06, &
  -3.132347747306860D-05,   0.000000000000000D+00,   4.049330623451054D-06, &
   0.000000000000000D+00,  -5.285026041415675D-05,   0.000000000000000D+00, &
  -2.235424827762496D-05,   0.000000000000000D+00,   2.506423089373921D-06, &
   1.846376892122433D-05,   0.000000000000000D+00,  -9.066167757771200D-06, &
   0.000000000000000D+00,  -1.873769509142181D-05,   0.000000000000000D+00, &
   6.813259353978192D-05,   0.000000000000000D+00,   5.617414518685082D-05, &
   0.000000000000000D+00,   5.969407534694348D-06,   0.000000000000000D+00, &
  -1.716762869855227D-05,   2.091462234987625D-05,   0.000000000000000D+00, &
   6.356794570372768D-06,   0.000000000000000D+00,   6.318555253025535D-05, &
   0.000000000000000D+00,   1.902765061208607D-05,   0.000000000000000D+00, &
  -7.262678645519566D-05,   0.000000000000000D+00,  -8.340033320657193D-05, &
   0.000000000000000D+00,  -8.236452885116487D-05,   0.000000000000000D+00, &
   2.988698618910062D-05,   0.000000000000000D+00,   2.504424129138065D-05, &
  -4.476880933558715D-05,   0.000000000000000D+00,   7.706689692845068D-06, &
   0.000000000000000D+00,  -5.820220652865165D-05,   0.000000000000000D+00, &
  -1.213814532000456D-04,   0.000000000000000D+00,  -1.653314795884399D-05, &
   0.000000000000000D+00,   6.212772740227629D-05,   0.000000000000000D+00, &
   7.119660789184332D-05,   0.000000000000000D+00,   1.774159725901818D-04, &
   0.000000000000000D+00,   9.681345961855054D-05,   0.000000000000000D+00, &
  -4.866206603809990D-05,   0.000000000000000D+00,  -2.516624274203225D-05, &
   3.099949764968392D-05,   0.000000000000000D+00,  -1.630391291650346D-05, &
   0.000000000000000D+00,  -6.496643606240449D-05,   0.000000000000000D+00, &
   1.308091110388809D-04,   0.000000000000000D+00,   1.444190262209794D-04, &
   0.000000000000000D+00,   2.412636819682805D-05,   0.000000000000000D+00, &
  -3.685110288856340D-05,   0.000000000000000D+00,  -3.162406105775584D-05, &
   0.000000000000000D+00,  -9.678779039761526D-05,   0.000000000000000D+00, &
  -2.952979883279126D-04,   0.000000000000000D+00,  -9.799621754831182D-05, &
   0.000000000000000D+00,   2.605799741837081D-05,   0.000000000000000D+00, &
   7.725085972709373D-06,  -7.724438243501075D-06,   0.000000000000000D+00, &
   9.116389086806819D-06,   0.000000000000000D+00,   8.627051996030387D-05, &
   0.000000000000000D+00,   1.417768156137391D-04,   0.000000000000000D+00, &
  -1.602048515182998D-04,   0.000000000000000D+00,  -1.135979174364575D-04, &
   0.000000000000000D+00,  -2.430707338057382D-05,   0.000000000000000D+00, &
   1.257215751748277D-05,   0.000000000000000D+00,   4.468392063962749D-06, &
   0.000000000000000D+00,  -4.645094181567547D-06,   0.000000000000000D+00, &
   1.076899368698198D-04,   0.000000000000000D+00,   3.365230399009810D-04, &
   0.000000000000000D+00,   9.079881150681306D-05,   0.000000000000000D+00, &
   1.008520226971179D-05,   0.000000000000000D+00,   1.802266743838743D-05, &
  -9.461811626143145D-07,   0.000000000000000D+00,   4.249066421423012D-07, &
   0.000000000000000D+00,  -1.927648248867730D-05,   0.000000000000000D+00, &
  -1.505536507286103D-04,   0.000000000000000D+00,  -1.704378482022359D-04, &
   0.000000000000000D+00,   1.130907064849989D-04,   0.000000000000000D+00, &
   5.342709978647242D-05,   0.000000000000000D+00,   6.425467514154582D-06, &
   0.000000000000000D+00,  -1.678046134561173D-06,   0.000000000000000D+00, &
   7.449192701361720D-07,   0.000000000000000D+00,   8.713167423094190D-06, &
   0.000000000000000D+00,   5.550571810417385D-05,   0.000000000000000D+00, &
  -1.055770284147542D-04,   0.000000000000000D+00,  -2.557661449492186D-04, &
   0.000000000000000D+00,  -6.868408465981140D-05,   0.000000000000000D+00, &
  -2.223952061956891D-05,   0.000000000000000D+00,  -2.068712852946639D-05, &
   5.827317280356852D-07,   0.000000000000000D+00,  -2.002444115022358D-06, &
   0.000000000000000D+00,   6.067438892652721D-07,   0.000000000000000D+00, &
   3.207454500370189D-06,   0.000000000000000D+00,   1.692736013537761D-04, &
   0.000000000000000D+00,   1.195751246510810D-04,   0.000000000000000D+00, &
  -4.157274217652752D-05,   0.000000000000000D+00,  -6.717416062683984D-06, &
   0.000000000000000D+00,   7.497784505885667D-06,   0.000000000000000D+00, &
  -7.508334003490698D-08,   0.000000000000000D+00,   3.234391412712453D-07, &
   0.000000000000000D+00,  -3.501412278623731D-06,   0.000000000000000D+00, &
  -7.205413842529361D-06,   0.000000000000000D+00,  -7.505414444311641D-05, &
   0.000000000000000D+00,   9.625939889622011D-05,   0.000000000000000D+00, &
   1.290757108828395D-04,   0.000000000000000D+00,   3.621592290411080D-05, &
   0.000000000000000D+00,   1.232101673452541D-05,   0.000000000000000D+00, &
   1.178787562274071D-06,   3.482473436081629D-09,   0.000000000000000D+00, &
   4.203340100053728D-07,   0.000000000000000D+00,   9.733649352844797D-07, &
   0.000000000000000D+00,  -2.733004005921906D-07,   0.000000000000000D+00, &
   6.793863186018509D-06,   0.000000000000000D+00,  -1.296204395388156D-04, &
   0.000000000000000D+00,  -4.495738013380897D-05,   0.000000000000000D+00, &
   2.988343660728292D-06,   0.000000000000000D+00,  -8.563350476682885D-06, &
   0.000000000000000D+00,  -5.598637559575195D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.202849866786760D-07, &
   0.000000000000000D+00,   1.035315185060441D-06,   0.000000000000000D+00, &
   1.297063423142190D-06,   0.000000000000000D+00,   3.778567309572550D-06, &
   0.000000000000000D+00,   4.370817558500269D-05,   0.000000000000000D+00, &
  -7.218069777993009D-05,   0.000000000000000D+00,  -4.497333341779854D-05, &
   0.000000000000000D+00,  -1.234215335577404D-05,   0.000000000000000D+00, &
  -1.294104001636407D-06,   0.000000000000000D+00,   1.063996492366813D-05 /)
lem(:,184) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.466989390567569D-04,   9.352533763483443D-04, &
   8.780762419446073D-04,  -2.421124091742313D-04,  -2.106470217276062D-03, &
  -2.584865830888572D-03,   1.486578981975075D-03,   5.779162866816959D-03, &
  -9.868143637049669D-04,  -6.387323420462716D-03,   2.907747272776858D-03, &
   2.639133957878132D-03,  -3.509442192555839D-03,   1.505688829610223D-03, &
   9.333945296135104D-04,  -1.783122512452949D-03,   8.742876407291390D-04, &
   3.105919035000337D-04,  -6.567418328001136D-04,   2.314894472404713D-04, &
   1.945541193051368D-04,   0.000000000000000D+00,  -5.466989390567577D-04, &
  -9.352533763483517D-04,  -8.780762419446027D-04,   2.421124091742309D-04, &
   2.106470217276055D-03,   2.584865830888572D-03,  -1.486578981975074D-03, &
  -5.779162866816972D-03,   9.868143637049673D-04,   6.387323420462722D-03, &
  -2.907747272776849D-03,  -2.639133957878128D-03,   3.509442192555845D-03, &
  -1.505688829610226D-03,  -9.333945296135076D-04,   1.783122512452950D-03, &
  -8.742876407291445D-04,  -3.105919035000299D-04,   6.567418328001189D-04, &
  -2.314894472404800D-04,  -1.945541193051382D-04,   0.000000000000000D+00, &
  -2.733494695283782D-04,   0.000000000000000D+00,  -7.183660794969154D-04, &
   0.000000000000000D+00,   2.777593999700535D-04,   0.000000000000000D+00, &
   1.079553523290319D-03,   0.000000000000000D+00,  -1.244498295941365D-03, &
   0.000000000000000D+00,   8.987438799018844D-06,   0.000000000000000D+00, &
   7.489019664007036D-04,   0.000000000000000D+00,  -1.993721500550431D-04, &
   0.000000000000000D+00,  -2.933594995579017D-04,   0.000000000000000D+00, &
   1.287689759826506D-04,   0.000000000000000D+00,   2.404062658852954D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.161017302271172D-05, &
   0.000000000000000D+00,   1.197639488564411D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.072314193822806D-05,   0.000000000000000D+00, &
   8.414982230908416D-06,   0.000000000000000D+00,  -1.663772788230153D-05, &
   0.000000000000000D+00,  -2.422156607428225D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.470814782389362D-06,   0.000000000000000D+00, &
  -8.159635489766988D-06,   0.000000000000000D+00,  -4.332258637007666D-05, &
   0.000000000000000D+00,   1.916850489589083D-05,   0.000000000000000D+00, &
   4.367905545275281D-05,   0.000000000000000D+00,   1.573540840581331D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.000345540887668D-06, &
   0.000000000000000D+00,   5.075232613374902D-06,   0.000000000000000D+00, &
   5.778352738626142D-05,   0.000000000000000D+00,   4.550834841765431D-05, &
   0.000000000000000D+00,  -2.373941870972094D-05,   0.000000000000000D+00, &
  -5.794685756527661D-05,   0.000000000000000D+00,  -7.473558948090767D-05, &
   0.000000000000000D+00,   1.961610303276639D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.690525519257561D-06,   0.000000000000000D+00, &
   7.826033066620664D-06,   0.000000000000000D+00,  -4.828987666212565D-05, &
   0.000000000000000D+00,  -1.164829188568652D-04,   0.000000000000000D+00, &
  -3.997295700094324D-05,   0.000000000000000D+00,   2.507854144964911D-05, &
   0.000000000000000D+00,   5.124665001387523D-05,   0.000000000000000D+00, &
   1.458345543697444D-04,   0.000000000000000D+00,   9.282429868468195D-05, &
   0.000000000000000D+00,  -4.054834773685129D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.812285818153167D-06,   0.000000000000000D+00, &
  -1.791493401000041D-05,   0.000000000000000D+00,  -5.301361444670677D-05, &
   0.000000000000000D+00,   1.150462125877129D-04,   0.000000000000000D+00, &
   1.435256210696913D-04,   0.000000000000000D+00,   5.534760220403140D-05, &
   0.000000000000000D+00,  -1.602049654544037D-05,   0.000000000000000D+00, &
  -2.764866945635655D-05,   0.000000000000000D+00,  -8.090361171637975D-05, &
   0.000000000000000D+00,  -2.572443068824512D-04,   0.000000000000000D+00, &
  -9.849532117357512D-05,   0.000000000000000D+00,   1.590335713566153D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.425884068897881D-06, &
   0.000000000000000D+00,   1.401462033320931D-05,   0.000000000000000D+00, &
   7.213335708563620D-05,   0.000000000000000D+00,   1.202246033850720D-04, &
   0.000000000000000D+00,  -1.445124622294988D-04,   0.000000000000000D+00, &
  -1.156082781559342D-04,   0.000000000000000D+00,  -5.981664679365698D-05, &
   0.000000000000000D+00,   4.400667157793108D-06,   0.000000000000000D+00, &
   8.958780874747964D-06,   0.000000000000000D+00,   1.113143233626287D-06, &
   0.000000000000000D+00,   9.626497960193818D-05,   0.000000000000000D+00, &
   3.029218958369850D-04,   0.000000000000000D+00,   9.415281496123268D-05, &
   0.000000000000000D+00,   2.124187972113180D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.722910785926876D-06,   0.000000000000000D+00, &
  -5.115922324654523D-06,   0.000000000000000D+00,  -1.825470033957733D-05, &
   0.000000000000000D+00,  -1.296374993924570D-04,   0.000000000000000D+00, &
  -1.493911885255945D-04,   0.000000000000000D+00,   1.032273287523155D-04, &
   0.000000000000000D+00,   5.553292911719418D-05,   0.000000000000000D+00, &
   2.304273560588361D-05,   0.000000000000000D+00,   2.873597752166643D-07, &
   0.000000000000000D+00,  -2.508254206591940D-06,   0.000000000000000D+00, &
   3.028124471899635D-06,   0.000000000000000D+00,   4.255676658233858D-05, &
   0.000000000000000D+00,  -9.773265426183524D-05,   0.000000000000000D+00, &
  -2.359248073035503D-04,   0.000000000000000D+00,  -7.195711342923363D-05, &
   0.000000000000000D+00,  -2.583565225278135D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.004344058651668D-07,   0.000000000000000D+00, &
   1.172287811390652D-06,   0.000000000000000D+00,   5.340587432010526D-07, &
   0.000000000000000D+00,   5.489763718957564D-06,   0.000000000000000D+00, &
   1.498478936289847D-04,   0.000000000000000D+00,   1.072748116923373D-04, &
   0.000000000000000D+00,  -3.787868184592714D-05,   0.000000000000000D+00, &
  -8.097793172604240D-06,   0.000000000000000D+00,   1.844279183611540D-05, &
   0.000000000000000D+00,  -2.789856516696358D-07,   0.000000000000000D+00, &
   8.987449689629845D-07,   0.000000000000000D+00,   3.776471997643987D-07, &
   0.000000000000000D+00,  -5.700815069918569D-06,   0.000000000000000D+00, &
  -6.146827829037027D-05,   0.000000000000000D+00,   9.064795466190599D-05, &
   0.000000000000000D+00,   1.214483748404275D-04,   0.000000000000000D+00, &
   3.762022095529544D-05,   0.000000000000000D+00,   8.237430035394004D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.554392913984698D-08, &
   0.000000000000000D+00,  -4.874469206110132D-07,   0.000000000000000D+00, &
   1.126377979965907D-06,   0.000000000000000D+00,   1.610349158811429D-06, &
   0.000000000000000D+00,   2.544758240418010D-06,   0.000000000000000D+00, &
  -1.171643056128585D-04,   0.000000000000000D+00,  -4.051566482866815D-05, &
   0.000000000000000D+00,   2.456445217266440D-06,   0.000000000000000D+00, &
  -9.401794779658960D-06,   0.000000000000000D+00,  -2.397721816722097D-05, &
   0.000000000000000D+00,  -5.382004402739814D-09,   0.000000000000000D+00, &
  -1.016698758991308D-07,   0.000000000000000D+00,  -8.563638813271654D-07, &
   0.000000000000000D+00,   1.427913560211364D-07,   0.000000000000000D+00, &
   4.720127850363343D-06,   0.000000000000000D+00,   3.560670187369493D-05, &
   0.000000000000000D+00,  -6.877454177865064D-05,   0.000000000000000D+00, &
  -4.291374440092752D-05,   0.000000000000000D+00,  -1.145482303683267D-05, &
   0.000000000000000D+00,   4.543236987993985D-07,   0.000000000000000D+00 /)
lem(:,185) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.556342527498466D-04,  -4.411659682154517D-04,  -7.084580720131442D-04, &
  -8.931618300885401D-04,   9.991910696933763D-06,   1.525616432823428D-03, &
   2.306866506168983D-03,  -1.444571239237750D-04,  -3.999634626194542D-03, &
  -3.416213628738431D-04,   3.611519162375667D-03,  -1.430778679810878D-03, &
  -8.767878767188261D-04,   1.558339299438377D-03,  -1.263253377145389D-03, &
   3.731898386826366D-04,   6.257511026780638D-04,  -9.216964710834900D-04, &
   4.232213292653220D-04,   1.947901015513949D-04,  -3.777308159146003D-04, &
   1.472340754424367D-04,  -4.556342527498552D-04,  -4.411659682154538D-04, &
  -7.084580720131437D-04,  -8.931618300885421D-04,   9.991910696929602D-06, &
   1.525616432823431D-03,   2.306866506168986D-03,  -1.444571239237791D-04, &
  -3.999634626194541D-03,  -3.416213628738362D-04,   3.611519162375667D-03, &
  -1.430778679810876D-03,  -8.767878767188283D-04,   1.558339299438373D-03, &
  -1.263253377145387D-03,   3.731898386826346D-04,   6.257511026780650D-04, &
  -9.216964710834839D-04,   4.232213292653185D-04,   1.947901015513894D-04, &
  -3.777308159145990D-04,   1.472340754424435D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.336893404024163D-04,   0.000000000000000D+00, &
   6.886193988682411D-04,   0.000000000000000D+00,   1.223882911031168D-03, &
   0.000000000000000D+00,  -3.362282419071952D-03,   0.000000000000000D+00, &
   3.534356667148132D-03,   0.000000000000000D+00,  -1.662871156002590D-03, &
   0.000000000000000D+00,  -4.318624944293020D-04,   0.000000000000000D+00, &
   7.667030273580902D-04,   0.000000000000000D+00,  -9.420087654867651D-05, &
   0.000000000000000D+00,  -1.204513954964840D-04,   0.000000000000000D+00, &
   2.092633928571447D-05,   1.824428630615737D-05,   0.000000000000000D+00, &
   1.233108969418975D-06,   0.000000000000000D+00,   1.895944263639696D-05, &
  -1.077079483782730D-05,   0.000000000000000D+00,  -9.254783999975966D-06, &
   0.000000000000000D+00,  -1.340403758687481D-05,   0.000000000000000D+00, &
  -1.660312599836974D-05,   0.000000000000000D+00,  -5.277043537501004D-05, &
  -8.529282459204393D-06,   0.000000000000000D+00,   6.377105132037294D-06, &
   0.000000000000000D+00,  -2.633719123619130D-06,   0.000000000000000D+00, &
   1.116118514600175D-05,   0.000000000000000D+00,   1.877562254468843D-05, &
   0.000000000000000D+00,   4.206322096328613D-05,   0.000000000000000D+00, &
   3.411266800879886D-05,   2.467299438917217D-05,   0.000000000000000D+00, &
   4.789186520776790D-06,   0.000000000000000D+00,   2.671281247959417D-06, &
   0.000000000000000D+00,   5.001349589328034D-05,   0.000000000000000D+00, &
  -1.899148506563489D-06,   0.000000000000000D+00,  -1.837615632057179D-05, &
   0.000000000000000D+00,  -4.992961478501809D-05,   0.000000000000000D+00, &
  -7.298786530921057D-05,   0.000000000000000D+00,   4.279332052106455D-05, &
  -2.765599544166694D-05,   0.000000000000000D+00,  -1.125153266690853D-05, &
   0.000000000000000D+00,   3.804308282881087D-06,   0.000000000000000D+00, &
  -4.202678460051083D-05,   0.000000000000000D+00,  -1.261146873915508D-04, &
   0.000000000000000D+00,  -4.509401944154796D-07,   0.000000000000000D+00, &
   1.433700925147398D-05,   0.000000000000000D+00,   4.311852743007090D-05, &
   0.000000000000000D+00,   1.294585324598228D-04,   0.000000000000000D+00, &
   9.818567839302858D-05,   0.000000000000000D+00,  -9.227197118358979D-05, &
   1.782397318981120D-05,   0.000000000000000D+00,   6.635893527969852D-06, &
   0.000000000000000D+00,  -1.106130309276021D-05,   0.000000000000000D+00, &
  -4.271901324738573D-05,   0.000000000000000D+00,   1.075899444781548D-04, &
   0.000000000000000D+00,   1.694745790484060D-04,   0.000000000000000D+00, &
  -1.994009599771187D-06,   0.000000000000000D+00,  -6.416487766246305D-06, &
   0.000000000000000D+00,  -2.155506095350631D-05,   0.000000000000000D+00, &
  -7.259342262414246D-05,   0.000000000000000D+00,  -2.403373764556966D-04, &
   0.000000000000000D+00,  -1.104875484610037D-04,   0.000000000000000D+00, &
   4.564921275036526D-05,  -6.032900400827948D-06,   0.000000000000000D+00, &
  -7.988571072349726D-07,   0.000000000000000D+00,   1.023904644287294D-05, &
   0.000000000000000D+00,   5.976969387622576D-05,   0.000000000000000D+00, &
   1.071673989955909D-04,   0.000000000000000D+00,  -1.407652457256996D-04, &
   0.000000000000000D+00,  -1.341586635520608D-04,   0.000000000000000D+00, &
   1.915117826032757D-06,   0.000000000000000D+00,   2.166008859358890D-07, &
   0.000000000000000D+00,   4.276392920537145D-06,   0.000000000000000D+00, &
   3.035825373699619D-06,   0.000000000000000D+00,   9.079881150681306D-05, &
   0.000000000000000D+00,   2.935357521658873D-04,   0.000000000000000D+00, &
   1.057144386116536D-04,   0.000000000000000D+00,   4.489962558753014D-05, &
   6.190935871265753D-07,   0.000000000000000D+00,  -8.229816017984205D-08, &
   0.000000000000000D+00,  -4.767449865042469D-06,   0.000000000000000D+00, &
  -1.748036678273068D-05,   0.000000000000000D+00,  -1.157844890036444D-04, &
   0.000000000000000D+00,  -1.384961963169884D-04,   0.000000000000000D+00, &
   1.029569308917073D-04,   0.000000000000000D+00,   5.298219717500372D-05, &
   0.000000000000000D+00,  -4.575785167887719D-07,   0.000000000000000D+00, &
   9.013891748671817D-07,   0.000000000000000D+00,   9.573003802246408D-07, &
   0.000000000000000D+00,   3.777299959487106D-06,   0.000000000000000D+00, &
   3.474549507112611D-05,   0.000000000000000D+00,  -9.650494599702847D-05, &
   0.000000000000000D+00,  -2.338814845151818D-04,   0.000000000000000D+00, &
  -8.021696174304483D-05,   0.000000000000000D+00,  -7.762755356542751D-05, &
   1.023717186182376D-07,   0.000000000000000D+00,  -3.768654507789690D-07, &
   0.000000000000000D+00,   1.378348032605796D-06,   0.000000000000000D+00, &
   2.409936541014415D-06,   0.000000000000000D+00,   8.714616585722223D-06, &
   0.000000000000000D+00,   1.395466931358522D-04,   0.000000000000000D+00, &
   1.006891004333143D-04,   0.000000000000000D+00,  -3.691705615772343D-05, &
   0.000000000000000D+00,   1.413504089516071D-06,   0.000000000000000D+00, &
  -1.037216111654230D-08,   0.000000000000000D+00,  -1.475662332877719D-07, &
   0.000000000000000D+00,  -8.536264966165166D-07,   0.000000000000000D+00, &
  -6.743730012261097D-07,   0.000000000000000D+00,  -6.786673892517795D-06, &
   0.000000000000000D+00,  -5.256487073727487D-05,   0.000000000000000D+00, &
   9.154909674363392D-05,   0.000000000000000D+00,   1.217705597131962D-04, &
   0.000000000000000D+00,   4.576248102849166D-05,   0.000000000000000D+00, &
   4.063167544690201D-05,   4.057860464630538D-10,   0.000000000000000D+00, &
   1.467633274873211D-07,   0.000000000000000D+00,  -3.029217103923405D-07, &
   0.000000000000000D+00,  -5.803274369709333D-07,   0.000000000000000D+00, &
   6.843406033226455D-07,   0.000000000000000D+00,  -2.837859104857746D-06, &
   0.000000000000000D+00,  -1.111284759532496D-04,   0.000000000000000D+00, &
  -3.748305755490186D-05,   0.000000000000000D+00,  -3.434114636003633D-07, &
   0.000000000000000D+00,  -8.088484829916405D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.540121429227469D-08, &
   0.000000000000000D+00,   3.506737758352459D-07,   0.000000000000000D+00, &
  -2.446107872754382D-07,   0.000000000000000D+00,  -2.187864208907325D-07, &
   0.000000000000000D+00,   6.812160455592401D-06,   0.000000000000000D+00, &
   2.912154447631317D-05,   0.000000000000000D+00,  -6.934025407019054D-05, &
   0.000000000000000D+00,  -4.331161516657076D-05,   0.000000000000000D+00, &
  -1.909087219745931D-05,   0.000000000000000D+00,   1.750372395555585D-07 /)
lem(:,186) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.725401509374939D-04,   1.123738403223355D-03, &
   8.693553592763828D-04,   2.463751078112073D-04,  -4.854075984194630D-04, &
  -1.848962443607274D-03,  -9.754027829558018D-04,   2.481860250061145D-03, &
   1.094876626263295D-03,  -2.056550821976864D-03,   3.476541104169074D-04, &
   3.839584512359496D-04,  -4.882918950266028D-04,   5.862758774195564D-04, &
  -4.802027070012210D-04,   9.515417668472858D-05,   3.378402910656513D-04, &
  -4.636694114413330D-04,   2.233897761347818D-04,   9.530359705919983D-05, &
  -2.096414158630211D-04,   0.000000000000000D+00,  -3.725401509374967D-04, &
  -1.123738403223356D-03,  -8.693553592763800D-04,  -2.463751078112082D-04, &
   4.854075984194618D-04,   1.848962443607275D-03,   9.754027829557918D-04, &
  -2.481860250061146D-03,  -1.094876626263286D-03,   2.056550821976869D-03, &
  -3.476541104169100D-04,  -3.839584512359462D-04,   4.882918950266036D-04, &
  -5.862758774195564D-04,   4.802027070012168D-04,  -9.515417668473225D-05, &
  -3.378402910656486D-04,   4.636694114413361D-04,  -2.233897761347847D-04, &
  -9.530359705920068D-05,   2.096414158630238D-04,   0.000000000000000D+00, &
  -1.862700754687486D-04,   0.000000000000000D+00,   4.222121963558061D-04, &
   0.000000000000000D+00,   9.490189961524150D-04,   0.000000000000000D+00, &
  -3.646526569754395D-03,   0.000000000000000D+00,   5.002429920320719D-03, &
   0.000000000000000D+00,  -5.020126116721040D-03,   0.000000000000000D+00, &
   3.963718351107791D-03,   0.000000000000000D+00,  -1.439353504880320D-03, &
   0.000000000000000D+00,  -3.409598420867472D-04,   0.000000000000000D+00, &
   3.582072880108011D-04,   0.000000000000000D+00,  -6.405440947759436D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.472594246031749D-05, &
   0.000000000000000D+00,   8.046424062049031D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.114710462050646D-06,   0.000000000000000D+00, &
  -2.812993432787899D-05,   0.000000000000000D+00,  -8.665576303538664D-06, &
   0.000000000000000D+00,  -9.846272701037374D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.177083166175257D-07,   0.000000000000000D+00, &
   7.083548820323574D-06,   0.000000000000000D+00,  -7.421562080241037D-06, &
   0.000000000000000D+00,   5.427511399409958D-06,   0.000000000000000D+00, &
   1.306154866763008D-05,   0.000000000000000D+00,   3.080868454304038D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.712148150407415D-06, &
   0.000000000000000D+00,   4.374296691829713D-06,   0.000000000000000D+00, &
   5.034911784465148D-06,   0.000000000000000D+00,   1.376116182365999D-04, &
   0.000000000000000D+00,  -3.460404898445452D-06,   0.000000000000000D+00, &
  -1.475819223150209D-05,   0.000000000000000D+00,  -4.236937943162210D-05, &
   0.000000000000000D+00,  -8.725984125455901D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.443207133725536D-06,   0.000000000000000D+00, &
  -8.433308288596022D-06,   0.000000000000000D+00,   4.782266032711421D-07, &
   0.000000000000000D+00,  -4.908697531600557D-05,   0.000000000000000D+00, &
  -3.109238640906486D-04,   0.000000000000000D+00,   4.602803640183309D-06, &
   0.000000000000000D+00,   1.394705947286474D-05,   0.000000000000000D+00, &
   3.809950946690497D-05,   0.000000000000000D+00,   1.291700834343296D-04, &
   0.000000000000000D+00,   1.539662266089800D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.555321689925915D-06,   0.000000000000000D+00, &
   5.126558491789412D-06,   0.000000000000000D+00,  -7.525210330508672D-06, &
   0.000000000000000D+00,  -3.865010931643696D-05,   0.000000000000000D+00, &
   1.169589671398000D-04,   0.000000000000000D+00,   4.085464058582044D-04, &
   0.000000000000000D+00,  -4.627446780972173D-06,   0.000000000000000D+00, &
  -8.837997515699179D-06,   0.000000000000000D+00,  -2.117896831576187D-05, &
   0.000000000000000D+00,  -7.027994851623638D-05,   0.000000000000000D+00, &
  -2.520181297924083D-04,   0.000000000000000D+00,  -1.911123779486561D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.023989020078925D-06, &
   0.000000000000000D+00,  -1.262582920310530D-06,   0.000000000000000D+00, &
   7.782337883029643D-06,   0.000000000000000D+00,   5.435353421845503D-05, &
   0.000000000000000D+00,   1.044138018219134D-04,   0.000000000000000D+00, &
  -1.528406158089720D-04,   0.000000000000000D+00,  -3.538347777356959D-04, &
   0.000000000000000D+00,   2.130670089408290D-06,   0.000000000000000D+00, &
   3.211313965712540D-06,   0.000000000000000D+00,   5.872183069518663D-06, &
   0.000000000000000D+00,   5.910208229451754D-06,   0.000000000000000D+00, &
   9.415281496123268D-05,   0.000000000000000D+00,   3.157404405291208D-04, &
   0.000000000000000D+00,   1.746281240868026D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.495676781021201D-07,   0.000000000000000D+00, &
   3.071709562168194D-07,   0.000000000000000D+00,  -3.040392469955626D-06, &
   0.000000000000000D+00,  -1.723081853058296D-05,   0.000000000000000D+00, &
  -1.118897334456108D-04,   0.000000000000000D+00,  -1.387881635936623D-04, &
   0.000000000000000D+00,   1.188853572004685D-04,   0.000000000000000D+00, &
   1.823987935868073D-04,   0.000000000000000D+00,  -2.504244234434647D-07, &
   0.000000000000000D+00,  -7.735351343886224D-07,   0.000000000000000D+00, &
  -2.597023307919505D-08,   0.000000000000000D+00,   3.441140760426020D-06, &
   0.000000000000000D+00,   2.792076819872266D-05,   0.000000000000000D+00, &
  -1.038343605486562D-04,   0.000000000000000D+00,  -2.583263069655135D-04, &
   0.000000000000000D+00,  -1.074301567017706D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.009280247257791D-07,   0.000000000000000D+00, &
  -2.290290181167915D-07,   0.000000000000000D+00,  -2.140355396054670D-08, &
   0.000000000000000D+00,   1.647495806521627D-06,   0.000000000000000D+00, &
   1.314455698077464D-05,   0.000000000000000D+00,   1.385635740867772D-04, &
   0.000000000000000D+00,   1.025305506862585D-04,   0.000000000000000D+00, &
  -5.365326071353536D-05,   0.000000000000000D+00,  -2.072559251874392D-05, &
   0.000000000000000D+00,  -5.817889139345375D-08,   0.000000000000000D+00, &
   2.893514424057034D-07,   0.000000000000000D+00,  -3.981316035468768D-07, &
   0.000000000000000D+00,  -1.125816074037707D-06,   0.000000000000000D+00, &
  -8.523667781050933D-06,   0.000000000000000D+00,  -4.508953332719159D-05, &
   0.000000000000000D+00,   9.936167698424896D-05,   0.000000000000000D+00, &
   1.375394327857082D-04,   0.000000000000000D+00,   3.594910023527131D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.849678772857640D-09, &
   0.000000000000000D+00,   2.076474070465781D-09,   0.000000000000000D+00, &
   4.431782844375087D-07,   0.000000000000000D+00,   2.931516228751032D-07, &
   0.000000000000000D+00,   9.871582589486363D-07,   0.000000000000000D+00, &
  -8.838763922786746D-06,   0.000000000000000D+00,  -1.119313801397589D-04, &
   0.000000000000000D+00,  -3.702618111530187D-05,   0.000000000000000D+00, &
   8.112272778580795D-06,   0.000000000000000D+00,  -4.175160283602444D-05, &
   0.000000000000000D+00,  -7.040325086601305D-10,   0.000000000000000D+00, &
  -7.149630139944640D-08,   0.000000000000000D+00,   5.817837258365896D-08, &
   0.000000000000000D+00,   2.557914099705574D-07,   0.000000000000000D+00, &
   5.248997018587881D-07,   0.000000000000000D+00,   9.072628159972067D-06, &
   0.000000000000000D+00,   2.372468647778980D-05,   0.000000000000000D+00, &
  -7.534943546337624D-05,   0.000000000000000D+00,  -4.561006860508925D-05, &
   0.000000000000000D+00,  -4.963033396294079D-06,   0.000000000000000D+00 /)
lem(:,187) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.234009434530832D-04,  -2.361054903967900D-04,  -2.223310389173122D-05, &
  -2.842473960703973D-04,  -1.255124809777704D-03,  -4.200040493196688D-04, &
   1.629656744640472D-03,   1.559666360751457D-03,  -1.263260221204686D-03, &
  -1.404012762795029D-03,   1.101914990840860D-03,   1.348947088160506D-04, &
  -3.371920765901156D-04,   1.694212735592846D-04,  -1.766690959193467D-04, &
   2.298027504003924D-04,  -1.800921220752248D-04,   1.077335104151745D-05, &
   1.760057014038237D-04,  -2.287252710999787D-04,   1.155191770853362D-04, &
   4.344576837636987D-05,  -1.234009434530777D-04,  -2.361054903967894D-04, &
  -2.223310389172848D-05,  -2.842473960703965D-04,  -1.255124809777703D-03, &
  -4.200040493196713D-04,   1.629656744640482D-03,   1.559666360751450D-03, &
  -1.263260221204697D-03,  -1.404012762795018D-03,   1.101914990840867D-03, &
   1.348947088160435D-04,  -3.371920765901172D-04,   1.694212735592851D-04, &
  -1.766690959193464D-04,   2.298027504003928D-04,  -1.800921220752238D-04, &
   1.077335104151933D-05,   1.760057014038218D-04,  -2.287252710999814D-04, &
   1.155191770853366D-04,   4.344576837637352D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.799940145672556D-06,   0.000000000000000D+00, &
   5.616876511625690D-04,   0.000000000000000D+00,  -2.242779849639380D-03, &
   0.000000000000000D+00,   2.224819114640352D-03,   0.000000000000000D+00, &
   7.537851755149934D-04,   0.000000000000000D+00,  -1.143517298251463D-03, &
   0.000000000000000D+00,  -1.872065612449377D-03,   0.000000000000000D+00, &
   2.526177923954616D-03,   0.000000000000000D+00,  -8.317982502685373D-04, &
   0.000000000000000D+00,   4.079089063875029D-06,   0.000000000000000D+00, &
   3.823578042327896D-06,   7.750496031746133D-06,   0.000000000000000D+00, &
   4.126551978114391D-06,   0.000000000000000D+00,  -1.097986937830274D-07, &
  -4.004422949735373D-06,   0.000000000000000D+00,   2.624599792568476D-07, &
   0.000000000000000D+00,  -2.811316287878874D-06,   0.000000000000000D+00, &
  -2.956344509078878D-06,   0.000000000000000D+00,  -3.047530269209955D-05, &
  -1.823163382671476D-06,   0.000000000000000D+00,  -4.585009132781490D-06, &
   0.000000000000000D+00,  -3.897114183246141D-06,   0.000000000000000D+00, &
  -5.409188363289167D-06,   0.000000000000000D+00,   3.331286468060779D-06, &
   0.000000000000000D+00,   1.766238954130684D-05,   0.000000000000000D+00, &
   1.220068177383314D-04,   3.696212897747689D-06,   0.000000000000000D+00, &
   8.394927555132394D-06,   0.000000000000000D+00,   8.259504684384202D-06, &
   0.000000000000000D+00,  -3.091439268671212D-06,   0.000000000000000D+00, &
   1.009218044679799D-05,   0.000000000000000D+00,  -1.406506253004968D-06, &
   0.000000000000000D+00,  -1.467737521093411D-05,   0.000000000000000D+00, &
  -6.073575499734774D-05,   0.000000000000000D+00,  -2.421479161376912D-04, &
  -3.597385276910607D-06,   0.000000000000000D+00,  -7.699902588065969D-06, &
   0.000000000000000D+00,  -1.009534323885802D-05,   0.000000000000000D+00, &
   4.547085538392113D-06,   0.000000000000000D+00,  -4.704655662233200D-06, &
   0.000000000000000D+00,  -7.435255086679334D-06,   0.000000000000000D+00, &
   7.459020539686860D-07,   0.000000000000000D+00,   1.157793492771813D-05, &
   0.000000000000000D+00,   4.031971149941455D-05,   0.000000000000000D+00, &
   1.513813116493542D-04,   0.000000000000000D+00,   3.152711735637539D-04, &
   3.085847232563002D-06,   0.000000000000000D+00,   3.698360536483293D-06, &
   0.000000000000000D+00,   6.110276050676158D-06,   0.000000000000000D+00, &
  -7.337893629578547D-06,   0.000000000000000D+00,  -4.916368955588300D-05, &
   0.000000000000000D+00,   5.414492912571831D-05,   0.000000000000000D+00, &
   2.460419077243910D-06,   0.000000000000000D+00,  -9.122287415389517D-07, &
   0.000000000000000D+00,  -7.302165532880555D-06,   0.000000000000000D+00, &
  -2.071085374288215D-05,   0.000000000000000D+00,  -7.517676487908392D-05, &
   0.000000000000000D+00,  -2.767350254446306D-04,   0.000000000000000D+00, &
  -3.447519967537107D-04,  -1.599273372053388D-06,   0.000000000000000D+00, &
  -3.248208440660569D-07,   0.000000000000000D+00,  -8.934546094787676D-07, &
   0.000000000000000D+00,   7.451442911610372D-06,   0.000000000000000D+00, &
   5.482594733957815D-05,   0.000000000000000D+00,   1.172457630190536D-04, &
   0.000000000000000D+00,  -8.736946149610581D-05,   0.000000000000000D+00, &
  -7.590757339394439D-08,   0.000000000000000D+00,   2.703051808219430D-07, &
   0.000000000000000D+00,   2.895598152075911D-06,   0.000000000000000D+00, &
   5.476989025035603D-06,   0.000000000000000D+00,   1.008520226971179D-05, &
   0.000000000000000D+00,   1.057144386116536D-04,   0.000000000000000D+00, &
   3.462046316325060D-04,   0.000000000000000D+00,   3.615067367513968D-04, &
   3.008768879110485D-07,   0.000000000000000D+00,  -4.201227690083197D-07, &
   0.000000000000000D+00,  -8.091981299116294D-07,   0.000000000000000D+00, &
  -3.410015806162686D-06,   0.000000000000000D+00,  -2.026258996734342D-05, &
   0.000000000000000D+00,  -1.172932108680935D-04,   0.000000000000000D+00, &
  -1.539237630364898D-04,   0.000000000000000D+00,   4.706404452599148D-05, &
   0.000000000000000D+00,  -1.065635417057295D-07,   0.000000000000000D+00, &
   2.091166630667318D-07,   0.000000000000000D+00,  -6.021055482860772D-07, &
   0.000000000000000D+00,  -6.574297802491540D-08,   0.000000000000000D+00, &
   3.553417873926685D-06,   0.000000000000000D+00,   2.206557633069135D-05, &
   0.000000000000000D+00,  -1.210865768925074D-04,   0.000000000000000D+00, &
  -2.920656254110370D-04,   0.000000000000000D+00,  -3.151586807952425D-04, &
   1.766230635155172D-08,   0.000000000000000D+00,   5.629185957273016D-08, &
   0.000000000000000D+00,   4.554380557661598D-07,   0.000000000000000D+00, &
   2.985582621793514D-07,   0.000000000000000D+00,   3.021461403823303D-06, &
   0.000000000000000D+00,   1.946541153186402D-05,   0.000000000000000D+00, &
   1.498909484190325D-04,   0.000000000000000D+00,   1.208800273984850D-04, &
   0.000000000000000D+00,   6.458958833614568D-06,   0.000000000000000D+00, &
  -1.467941105894157D-09,   0.000000000000000D+00,  -8.718216120794930D-08, &
   0.000000000000000D+00,   7.782798550787888D-08,   0.000000000000000D+00, &
  -7.210138907325513D-10,   0.000000000000000D+00,  -1.598546479444323D-06, &
   0.000000000000000D+00,  -1.037889507616424D-05,   0.000000000000000D+00, &
  -4.062084568560877D-05,   0.000000000000000D+00,   1.151034386818255D-04, &
   0.000000000000000D+00,   1.748797330349027D-04,   0.000000000000000D+00, &
   1.873675878922278D-04,   4.941041901778782D-11,   0.000000000000000D+00, &
   3.509490665961318D-08,   0.000000000000000D+00,  -1.534141447225670D-07, &
   0.000000000000000D+00,   2.043087929028076D-07,   0.000000000000000D+00, &
   1.566503423195811D-07,   0.000000000000000D+00,   6.843129546973277D-07, &
   0.000000000000000D+00,  -1.644479835576124D-05,   0.000000000000000D+00, &
  -1.227722763904277D-04,   0.000000000000000D+00,  -5.544068160334509D-05, &
   0.000000000000000D+00,  -9.411939897724173D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.344217807247117D-09, &
   0.000000000000000D+00,   1.739135177812572D-08,   0.000000000000000D+00, &
  -2.104720292062854D-07,   0.000000000000000D+00,   2.288956986742698D-07, &
   0.000000000000000D+00,   2.702448487615787D-07,   0.000000000000000D+00, &
   1.265593256213949D-05,   0.000000000000000D+00,   1.953540246714422D-05, &
   0.000000000000000D+00,  -8.645864300312188D-05,   0.000000000000000D+00, &
  -8.320432199401562D-05,   0.000000000000000D+00,  -6.503969559800011D-05 /)
lem(:,188) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.045726739473659D-04,   5.268278965649334D-04, &
   1.080663236859929D-03,   6.254404333699613D-04,  -1.037996962605657D-04, &
  -2.587865869560134D-04,  -1.390989900779945D-03,  -6.396060005753080D-05, &
   1.513426266648323D-03,  -5.106590323317859D-04,  -3.484058913195843D-04, &
   3.172764718536478D-04,  -1.216333581917231D-04,   6.356703637784317D-05, &
  -8.952219547821786D-05,   1.028260851983301D-04,  -6.411876673603882D-05, &
  -2.369128377122001D-05,   1.066989791206361D-04,  -1.198729915594019D-04, &
   5.734183296756083D-05,   0.000000000000000D+00,  -1.045726739473668D-04, &
  -5.268278965649260D-04,  -1.080663236859932D-03,  -6.254404333699610D-04, &
   1.037996962605669D-04,   2.587865869560168D-04,   1.390989900779934D-03, &
   6.396060005752094D-05,  -1.513426266648315D-03,   5.106590323317923D-04, &
   3.484058913195785D-04,  -3.172764718536450D-04,   1.216333581917264D-04, &
  -6.356703637784328D-05,   8.952219547821359D-05,  -1.028260851983314D-04, &
   6.411876673604354D-05,   2.369128377121778D-05,  -1.066989791206387D-04, &
   1.198729915594055D-04,  -5.734183296755904D-05,   0.000000000000000D+00, &
  -5.228633697368785D-05,   0.000000000000000D+00,   1.508562623130547D-04, &
   0.000000000000000D+00,  -5.161542438800443D-04,   0.000000000000000D+00, &
  -3.377688763776098D-04,   0.000000000000000D+00,   7.262863450907201D-03, &
   0.000000000000000D+00,  -1.781881309897951D-02,   0.000000000000000D+00, &
   1.876568130201107D-02,   0.000000000000000D+00,  -8.684853434878915D-03, &
   0.000000000000000D+00,   7.407570551517811D-04,   0.000000000000000D+00, &
   6.189383200403241D-04,   0.000000000000000D+00,  -1.865062596374604D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.133597883597872D-06, &
   0.000000000000000D+00,   3.689705838143341D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.470245610870641D-06,   0.000000000000000D+00, &
  -1.235675674152240D-05,   0.000000000000000D+00,  -3.638035864598408D-06, &
   0.000000000000000D+00,  -6.794014362373691D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.004422949735456D-06,   0.000000000000000D+00, &
   1.995752728174501D-06,   0.000000000000000D+00,   2.902032320977638D-05, &
   0.000000000000000D+00,   1.553712491212553D-06,   0.000000000000000D+00, &
   5.186734923063118D-06,   0.000000000000000D+00,   5.606914786602238D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.573093457725186D-06, &
   0.000000000000000D+00,   3.793706293527953D-06,   0.000000000000000D+00, &
  -4.276074565866016D-06,   0.000000000000000D+00,  -3.206779015530599D-06, &
   0.000000000000000D+00,   1.027892754729737D-06,   0.000000000000000D+00, &
  -9.791655862415719D-07,   0.000000000000000D+00,  -6.939978698669341D-06, &
   0.000000000000000D+00,  -2.266499544963368D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.908465296741336D-06,   0.000000000000000D+00, &
  -6.795364323794181D-06,   0.000000000000000D+00,  -7.109985276162802D-06, &
   0.000000000000000D+00,  -4.480679696369292D-06,   0.000000000000000D+00, &
  -1.963695414911587D-04,   0.000000000000000D+00,  -1.591165031619978D-06, &
   0.000000000000000D+00,  -8.591396041273849D-08,   0.000000000000000D+00, &
   7.785756300830926D-06,   0.000000000000000D+00,   3.135105585545837D-05, &
   0.000000000000000D+00,   1.437852606202818D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.498694802790881D-06,   0.000000000000000D+00, &
   5.556913143504281D-06,   0.000000000000000D+00,   7.064629331992153D-06, &
   0.000000000000000D+00,   2.033943703056784D-07,   0.000000000000000D+00, &
  -1.359363261259285D-06,   0.000000000000000D+00,   5.702509441131757D-04, &
   0.000000000000000D+00,   2.414685476157380D-07,   0.000000000000000D+00, &
  -8.956119758389770D-07,   0.000000000000000D+00,  -6.170391584454799D-06, &
   0.000000000000000D+00,  -2.347164263254871D-05,   0.000000000000000D+00, &
  -9.663354155777742D-05,   0.000000000000000D+00,  -3.824601932749331D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.213685963065879D-06, &
   0.000000000000000D+00,  -2.054243141362819D-06,   0.000000000000000D+00, &
  -1.976180407849846D-06,   0.000000000000000D+00,   6.274026100430684D-06, &
   0.000000000000000D+00,   5.131266344095261D-05,   0.000000000000000D+00, &
   5.015606072697769D-05,   0.000000000000000D+00,  -8.045415350389992D-04, &
   0.000000000000000D+00,   4.198784161436243D-07,   0.000000000000000D+00, &
   6.373321817528780D-07,   0.000000000000000D+00,   2.244368133763348D-06, &
   0.000000000000000D+00,   7.780649070644779D-06,   0.000000000000000D+00, &
   2.124187972113180D-05,   0.000000000000000D+00,   1.746281240868026D-04, &
   0.000000000000000D+00,   5.322753255196544D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.276866433155341D-07,   0.000000000000000D+00, &
   2.143989598001517D-07,   0.000000000000000D+00,  -2.221714384589376D-07, &
   0.000000000000000D+00,  -4.733171583772289D-06,   0.000000000000000D+00, &
  -2.573302629502149D-05,   0.000000000000000D+00,  -1.391325209535254D-04, &
   0.000000000000000D+00,  -8.406613432069192D-05,   0.000000000000000D+00, &
   6.348197227895361D-04,   0.000000000000000D+00,  -1.426094934909911D-07, &
   0.000000000000000D+00,   2.343023472915809D-09,   0.000000000000000D+00, &
  -6.838393646044647D-08,   0.000000000000000D+00,   1.119461766874832D-07, &
   0.000000000000000D+00,   2.958272690541992D-06,   0.000000000000000D+00, &
   1.280909088306866D-05,   0.000000000000000D+00,  -2.173425304243358D-04, &
   0.000000000000000D+00,  -4.151462748079044D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.070399951087409D-08,   0.000000000000000D+00, &
  -5.487136205416378D-08,   0.000000000000000D+00,   1.249360507599833D-07, &
   0.000000000000000D+00,   1.107786710957110D-06,   0.000000000000000D+00, &
   4.467189464488960D-06,   0.000000000000000D+00,   3.155134099353444D-05, &
   0.000000000000000D+00,   1.905405625380020D-04,   0.000000000000000D+00, &
   4.423867966421799D-05,   0.000000000000000D+00,  -2.526103244902797D-04, &
   0.000000000000000D+00,  -9.907333872694177D-09,   0.000000000000000D+00, &
  -3.259653910561398D-08,   0.000000000000000D+00,  -1.013896679272653D-07, &
   0.000000000000000D+00,  -6.441238512775862D-07,   0.000000000000000D+00, &
  -1.712970740331702D-06,   0.000000000000000D+00,  -1.506058590288592D-05, &
   0.000000000000000D+00,  -3.613728631376689D-05,   0.000000000000000D+00, &
   1.964533892832111D-04,   0.000000000000000D+00,   1.701722244959680D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.135390283089498D-09, &
   0.000000000000000D+00,   4.156500639699954D-08,   0.000000000000000D+00, &
  -4.519972924474051D-08,   0.000000000000000D+00,   7.422027263696349D-10, &
   0.000000000000000D+00,  -2.488173656619604D-07,   0.000000000000000D+00, &
   1.376465021237273D-06,   0.000000000000000D+00,  -3.079304450028801D-05, &
   0.000000000000000D+00,  -1.508433630852731D-04,   0.000000000000000D+00, &
   8.847461879758397D-06,   0.000000000000000D+00,   4.388504612744750D-06, &
   0.000000000000000D+00,  -7.185725685466161D-11,   0.000000000000000D+00, &
  -1.577497866672470D-08,   0.000000000000000D+00,   3.745958351127317D-08, &
   0.000000000000000D+00,   6.762173464247981D-08,   0.000000000000000D+00, &
  -9.005056477830635D-09,   0.000000000000000D+00,   2.468066400742059D-07, &
   0.000000000000000D+00,   2.008650906732161D-05,   0.000000000000000D+00, &
   1.052767734417702D-05,   0.000000000000000D+00,  -1.189989977198700D-04, &
   0.000000000000000D+00,  -2.599187542950752D-05,   0.000000000000000D+00 /)
lem(:,189) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.966999278900705D-05,   3.717231769222938D-04,   7.337528444702977D-04, &
   6.991656405775169D-04,  -6.047068025069209D-04,  -2.705962575014154D-03, &
  -8.644377409152585D-04,   3.520499092361616D-03,   7.157818332728695D-04, &
  -2.420632518267263D-03,   4.621384974757343D-04,   5.084268600055564D-04, &
  -3.521822725564998D-04,   1.456478215468529D-04,  -2.467128929359818D-05, &
   3.594751353421542D-05,  -7.851547422307548D-05,   5.956353956038590D-05, &
  -2.542645128577182D-05,  -1.950969753814870D-05,   7.520826298270766D-05, &
  -8.770175376478585D-05,   9.966999278903758D-05,   3.717231769222880D-04, &
   7.337528444703047D-04,   6.991656405775262D-04,  -6.047068025069320D-04, &
  -2.705962575014158D-03,  -8.644377409152472D-04,   3.520499092361606D-03, &
   7.157818332728563D-04,  -2.420632518267240D-03,   4.621384974757493D-04, &
   5.084268600055427D-04,  -3.521822725565110D-04,   1.456478215468556D-04, &
  -2.467128929359116D-05,   3.594751353421653D-05,  -7.851547422307702D-05, &
   5.956353956039266D-05,  -2.542645128577078D-05,  -1.950969753816006D-05, &
   7.520826298270875D-05,  -8.770175376477692D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.074947627460875D-05,   0.000000000000000D+00, &
   1.862267541951678D-04,   0.000000000000000D+00,  -7.856127991559639D-04, &
   0.000000000000000D+00,   4.978988360959207D-03,   0.000000000000000D+00, &
  -1.862469589076506D-02,   0.000000000000000D+00,   3.575770939474055D-02, &
   0.000000000000000D+00,  -3.709603648719308D-02,   0.000000000000000D+00, &
   2.084278758027835D-02,   0.000000000000000D+00,  -5.990377113578255D-03, &
   0.000000000000000D+00,   9.470453889701647D-04,   0.000000000000000D+00, &
  -2.260561342591345D-07,  -6.781684027779279D-07,   0.000000000000000D+00, &
  -1.423860066438058D-06,   0.000000000000000D+00,   4.063138828764006D-06, &
  -1.617622467231855D-06,   0.000000000000000D+00,   3.628641323953772D-06, &
   0.000000000000000D+00,   7.133979301948694D-07,   0.000000000000000D+00, &
   1.586503051346846D-06,   0.000000000000000D+00,  -1.782834271847963D-05, &
   5.569201125841703D-06,   0.000000000000000D+00,  -3.035610945767237D-06, &
   0.000000000000000D+00,  -1.297107162585668D-05,   0.000000000000000D+00, &
   9.121285976756479D-07,   0.000000000000000D+00,  -5.795257260101029D-06, &
   0.000000000000000D+00,   2.543131510416598D-06,   0.000000000000000D+00, &
   7.780954893991974D-05,  -8.076369160353803D-06,   0.000000000000000D+00, &
   2.468325222231512D-06,   0.000000000000000D+00,   1.750585704443134D-05, &
   0.000000000000000D+00,   3.039071795175309D-05,   0.000000000000000D+00, &
   1.262391399110275D-07,   0.000000000000000D+00,   6.908600658600523D-06, &
   0.000000000000000D+00,   2.640351456269433D-06,   0.000000000000000D+00, &
  -3.936978729459200D-05,   0.000000000000000D+00,  -3.187737824998257D-04, &
   6.281839142617738D-06,   0.000000000000000D+00,  -3.940598375707331D-06, &
   0.000000000000000D+00,  -1.676539968300050D-05,   0.000000000000000D+00, &
  -3.154761986423965D-05,   0.000000000000000D+00,  -1.250133986722482D-06, &
   0.000000000000000D+00,  -1.921481982842994D-06,   0.000000000000000D+00, &
  -3.655155137703740D-06,   0.000000000000000D+00,  -8.369671088131815D-07, &
   0.000000000000000D+00,   2.906822882765879D-05,   0.000000000000000D+00, &
   1.431251620019389D-04,   0.000000000000000D+00,   9.357902668049758D-04, &
  -2.165064330400595D-06,   0.000000000000000D+00,   4.082394122738607D-06, &
   0.000000000000000D+00,   1.064110289567965D-05,   0.000000000000000D+00, &
   1.652925180149349D-05,   0.000000000000000D+00,  -1.451057270756054D-05, &
   0.000000000000000D+00,  -2.103585322601944D-04,   0.000000000000000D+00, &
   1.457062788083543D-06,   0.000000000000000D+00,  -2.332586879562018D-07, &
   0.000000000000000D+00,  -1.293591142406640D-06,   0.000000000000000D+00, &
  -1.750590336714499D-05,   0.000000000000000D+00,  -6.807301281213363D-05, &
   0.000000000000000D+00,  -2.779494461735579D-04,   0.000000000000000D+00, &
  -1.817290325451267D-03,  -1.598942116937906D-08,   0.000000000000000D+00, &
  -1.686808677862445D-06,   0.000000000000000D+00,  -3.966845013332533D-06, &
   0.000000000000000D+00,  -2.392558799625183D-06,   0.000000000000000D+00, &
   2.085064401235581D-05,   0.000000000000000D+00,   1.707660223332044D-04, &
   0.000000000000000D+00,   6.022452792847012D-04,   0.000000000000000D+00, &
  -2.827991423355122D-07,   0.000000000000000D+00,   1.012306380826977D-06, &
   0.000000000000000D+00,   5.961183358581864D-07,   0.000000000000000D+00, &
   5.303681662892724D-06,   0.000000000000000D+00,   1.802266743838743D-05, &
   0.000000000000000D+00,   4.489962558753014D-05,   0.000000000000000D+00, &
   3.615067367513968D-04,   0.000000000000000D+00,   2.342774734613907D-03, &
   1.286992416275436D-07,   0.000000000000000D+00,  -9.913604122250558D-09, &
   0.000000000000000D+00,   1.169561013923862D-06,   0.000000000000000D+00, &
  -1.583526566944987D-06,   0.000000000000000D+00,  -1.076737936410185D-05, &
   0.000000000000000D+00,  -6.387773181459505D-05,   0.000000000000000D+00, &
  -3.760303487933162D-04,   0.000000000000000D+00,  -8.466105164362035D-04, &
   0.000000000000000D+00,  -2.861337775546544D-08,   0.000000000000000D+00, &
  -2.452916023774199D-07,   0.000000000000000D+00,   7.239588033394552D-08, &
   0.000000000000000D+00,   5.921381525070996D-07,   0.000000000000000D+00, &
  -1.744510434676694D-07,   0.000000000000000D+00,   8.750023798300625D-06, &
   0.000000000000000D+00,   4.911876503336968D-05,   0.000000000000000D+00, &
  -3.944014792212523D-04,   0.000000000000000D+00,  -2.028504787682577D-03, &
   3.585417319054295D-09,   0.000000000000000D+00,   1.224820711263768D-07, &
   0.000000000000000D+00,  -5.869278730632321D-07,   0.000000000000000D+00, &
   6.631357906034956D-07,   0.000000000000000D+00,   2.280743365029107D-06, &
   0.000000000000000D+00,   8.802141441659859D-06,   0.000000000000000D+00, &
   7.053464895918325D-05,   0.000000000000000D+00,   4.895846108825701D-04, &
   0.000000000000000D+00,   6.683844251438009D-04,   0.000000000000000D+00, &
  -2.430681700103368D-10,   0.000000000000000D+00,  -2.528773917561538D-08, &
   0.000000000000000D+00,   9.309991654078226D-08,   0.000000000000000D+00, &
  -1.203338333234223D-06,   0.000000000000000D+00,  -7.970051918426977D-07, &
   0.000000000000000D+00,  -3.775167115471762D-06,   0.000000000000000D+00, &
  -3.478381353946023D-05,   0.000000000000000D+00,  -1.097397673106215D-04, &
   0.000000000000000D+00,   4.087963866425604D-04,   0.000000000000000D+00, &
   1.187305065632463D-03,   6.965754332389217D-12,   0.000000000000000D+00, &
   8.108321091898298D-10,   0.000000000000000D+00,   1.715441669507778D-07, &
   0.000000000000000D+00,   7.306164169778654D-08,   0.000000000000000D+00, &
  -8.332479769459496D-07,   0.000000000000000D+00,   8.602092609352881D-08, &
   0.000000000000000D+00,   3.739817157775428D-06,   0.000000000000000D+00, &
  -6.490753236138527D-05,   0.000000000000000D+00,  -4.217581421963197D-04, &
   0.000000000000000D+00,  -2.671274758223080D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.943146347678805D-10, &
   0.000000000000000D+00,  -9.518750310047185D-08,   0.000000000000000D+00, &
   5.857339444219810D-07,   0.000000000000000D+00,   1.903626408682982D-07, &
   0.000000000000000D+00,  -2.407888753087149D-07,   0.000000000000000D+00, &
   8.877272505917253D-07,   0.000000000000000D+00,   4.471261368777936D-05, &
   0.000000000000000D+00,   5.307583432855719D-05,   0.000000000000000D+00, &
  -3.532327852065328D-04,   0.000000000000000D+00,  -4.776578586272621D-04 /)
lem(:,190) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.221383955680836D-03, &
  -5.281554044499223D-03,  -1.133555692945714D-02,   1.113253359011765D-02, &
   1.282164568002385D-02,  -6.191706121813095D-03,  -6.285073432119700D-03, &
  -1.052207323993738D-03,  -6.108999919379277D-04,   2.985702643312487D-04, &
   1.198945783519054D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.221383955680867D-03,  -5.281554044499243D-03,  -1.133555692945718D-02, &
   1.113253359011770D-02,   1.282164568002387D-02,  -6.191706121813146D-03, &
  -6.285073432119702D-03,  -1.052207323993668D-03,  -6.108999919379978D-04, &
   2.985702643312196D-04,   1.198945783519128D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.226381432939439D-04,   0.000000000000000D+00, &
   5.161687749810187D-06,   0.000000000000000D+00,   2.802245802759323D-07, &
   0.000000000000000D+00,   2.274550164557623D-08,   0.000000000000000D+00, &
   2.193316229750130D-09,   0.000000000000000D+00,   2.623856077945198D-10, &
  -2.655285260595511D-03,   0.000000000000000D+00,  -5.452696483030152D-05, &
   0.000000000000000D+00,  -4.198031093089579D-06,   0.000000000000000D+00, &
  -4.470855803495616D-07,   0.000000000000000D+00,  -5.450797001506366D-08, &
   0.000000000000000D+00,  -7.049426675364037D-09,   0.000000000000000D+00, &
   5.991963831269982D-03,   0.000000000000000D+00,   3.484429531804840D-05, &
   0.000000000000000D+00,   1.540782138261809D-05,   0.000000000000000D+00, &
   2.779378991774040D-06,   0.000000000000000D+00,   4.784703355404652D-07, &
   0.000000000000000D+00,   8.085820534789738D-08,   0.000000000000000D+00, &
   1.627612382820021D-08,  -3.798571551630510D-03,   0.000000000000000D+00, &
   7.298729841893743D-04,   0.000000000000000D+00,   2.370231051393785D-05, &
   0.000000000000000D+00,  -3.366828145899136D-06,   0.000000000000000D+00, &
  -1.537618656757391D-06,   0.000000000000000D+00,  -4.126096967157585D-07, &
   0.000000000000000D+00,  -9.623747535880172D-08,   0.000000000000000D+00, &
  -7.036856907083090D-03,   0.000000000000000D+00,  -2.245577164093712D-03, &
   0.000000000000000D+00,  -1.690381096231455D-04,   0.000000000000000D+00, &
  -2.163807856446797D-05,   0.000000000000000D+00,  -9.461811626143145D-07, &
   0.000000000000000D+00,   6.190935871265753D-07,   0.000000000000000D+00, &
   3.008768879110485D-07,   0.000000000000000D+00,   1.286992416275436D-07, &
   1.446998187821324D-02,   0.000000000000000D+00,   1.766962664457464D-03, &
   0.000000000000000D+00,   2.774812012383530D-05,   0.000000000000000D+00, &
   4.747822609012742D-05,   0.000000000000000D+00,   1.382438260908587D-05, &
   0.000000000000000D+00,   2.189596597197420D-06,   0.000000000000000D+00, &
  -3.566748042277531D-08,   0.000000000000000D+00,  -1.936258978043083D-07, &
   0.000000000000000D+00,  -7.480007924946745D-03,   0.000000000000000D+00, &
   2.396646714286282D-03,   0.000000000000000D+00,   7.827539166771905D-04, &
   0.000000000000000D+00,   6.735713343613001D-05,   0.000000000000000D+00, &
  -8.730512201292909D-06,   0.000000000000000D+00,  -7.333545430878997D-06, &
   0.000000000000000D+00,  -2.125936736814540D-06,   0.000000000000000D+00, &
  -2.639880113844331D-07,   0.000000000000000D+00,   1.681872207093659D-07, &
  -4.316657931785842D-03,   0.000000000000000D+00,  -5.390244086017165D-03, &
   0.000000000000000D+00,  -1.000900825706165D-03,   0.000000000000000D+00, &
  -2.003185492646766D-04,   0.000000000000000D+00,  -4.932200792671876D-05, &
   0.000000000000000D+00,  -3.656509891881802D-06,   0.000000000000000D+00, &
   3.014621269139756D-06,   0.000000000000000D+00,   1.586412753131874D-06, &
   0.000000000000000D+00,   3.594241926410059D-07,   0.000000000000000D+00, &
   7.249416148192576D-03,   0.000000000000000D+00,   2.548965785895708D-03, &
   0.000000000000000D+00,  -6.697126504763009D-04,   0.000000000000000D+00, &
  -1.020181221140951D-04,   0.000000000000000D+00,   4.242629859562863D-05, &
   0.000000000000000D+00,   2.704464675647275D-05,   0.000000000000000D+00, &
   6.289550549819865D-06,   0.000000000000000D+00,  -5.457810346198074D-07, &
   0.000000000000000D+00,  -9.958706835377638D-07,   0.000000000000000D+00, &
  -4.867321738175958D-07,  -3.432138916009880D-03,   0.000000000000000D+00, &
   2.006002195101194D-03,   0.000000000000000D+00,   2.040317648541763D-03, &
   0.000000000000000D+00,   4.481042150988476D-04,   0.000000000000000D+00, &
   1.008427085528609D-04,   0.000000000000000D+00,   2.493630725806013D-06, &
   0.000000000000000D+00,  -1.201923868716558D-05,   0.000000000000000D+00, &
  -5.404267169140477D-06,   0.000000000000000D+00,  -6.026653617737659D-07, &
   0.000000000000000D+00,   5.273642065053712D-07,   0.000000000000000D+00, &
   5.855175013318151D-04,   0.000000000000000D+00,  -2.718173139237192D-03, &
   0.000000000000000D+00,  -7.132938761341720D-04,   0.000000000000000D+00, &
   5.490039025748043D-05,   0.000000000000000D+00,  -7.390124308420272D-05, &
   0.000000000000000D+00,  -5.199902315923790D-05,   0.000000000000000D+00, &
  -1.218788777589759D-05,   0.000000000000000D+00,   3.556843907866907D-06, &
   0.000000000000000D+00,   3.600929893161572D-06,   0.000000000000000D+00, &
   9.482273759172680D-07,   0.000000000000000D+00,  -5.934934101903269D-07 /)
lem(:,191) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.690407763512195D-04,   2.687007761592501D-03,  -2.706984359366032D-03, &
  -4.500821451562435D-03,   8.201851979139103D-03,   2.000685045948750D-03, &
  -7.632472956814787D-03,  -2.794321699926283D-04,   2.643887941648254D-03, &
   7.069290153206859D-04,   3.377683226678028D-04,  -2.958068828728758D-04, &
  -6.573604289697544D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.690407763512234D-04,  -2.687007761592519D-03, &
   2.706984359366033D-03,   4.500821451562429D-03,  -8.201851979139080D-03, &
  -2.000685045948749D-03,   7.632472956814789D-03,   2.794321699926312D-04, &
  -2.643887941648231D-03,  -7.069290153206947D-04,  -3.377683226678148D-04, &
   2.958068828728993D-04,   6.573604289697478D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.111665043078483D-09, &
   0.000000000000000D+00,   2.558855736413937D-07,   0.000000000000000D+00, &
   1.763757222909296D-06,   0.000000000000000D+00,   2.607998563812215D-06, &
   0.000000000000000D+00,  -3.243067135072303D-06,   0.000000000000000D+00, &
  -6.049691015019519D-06,   0.000000000000000D+00,   4.170442453156060D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.950021377340298D-06,   0.000000000000000D+00, &
   1.289345007577024D-07,   0.000000000000000D+00,   8.790988688104114D-09, &
   0.000000000000000D+00,   5.372270865766900D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.234487724440900D-05,   0.000000000000000D+00, &
  -2.042322492004711D-06,   0.000000000000000D+00,  -1.921319416597831D-07, &
   0.000000000000000D+00,  -1.912528427914396D-08,   0.000000000000000D+00, &
  -5.708037794144892D-10,   0.000000000000000D+00,   1.498446214358508D-04, &
   0.000000000000000D+00,   9.871399667363590D-06,   0.000000000000000D+00, &
   1.397933568495767D-06,   0.000000000000000D+00,   2.053478330170487D-07, &
   0.000000000000000D+00,   2.576307153077543D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.706132399256132D-04,   0.000000000000000D+00, &
  -9.838219334165560D-06,   0.000000000000000D+00,  -3.463893422127666D-06, &
   0.000000000000000D+00,  -8.885938192799528D-07,   0.000000000000000D+00, &
  -1.866874956747439D-07,   0.000000000000000D+00,  -3.020932969894109D-08, &
   0.000000000000000D+00,   4.718669940727841D-04,   0.000000000000000D+00, &
  -7.282046511520466D-05,   0.000000000000000D+00,  -3.721954628055658D-06, &
   0.000000000000000D+00,   8.792995281093221D-07,   0.000000000000000D+00, &
   5.095254286555481D-07,   0.000000000000000D+00,   1.586867147995444D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.723221439443970D-05, &
   0.000000000000000D+00,   3.039650007388793D-04,   0.000000000000000D+00, &
   3.321294727099873D-05,   0.000000000000000D+00,   4.870606478555527D-06, &
   0.000000000000000D+00,   1.882445286510723D-07,   0.000000000000000D+00, &
  -2.586148718088494D-07,   0.000000000000000D+00,  -1.990239894851026D-07, &
   0.000000000000000D+00,  -7.058581060289880D-04,   0.000000000000000D+00, &
  -4.848820621580653D-04,   0.000000000000000D+00,  -3.883596126012328D-05, &
   0.000000000000000D+00,  -1.371829641345200D-05,   0.000000000000000D+00, &
  -3.722910785926876D-06,   0.000000000000000D+00,  -5.495676781021201D-07, &
   0.000000000000000D+00,   1.276866433155341D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.165991194847734D-03,   0.000000000000000D+00, &
   1.791414277255398D-04,   0.000000000000000D+00,  -8.698471584875189D-05, &
   0.000000000000000D+00,  -2.514999639708307D-06,   0.000000000000000D+00, &
   4.661027417697516D-06,   0.000000000000000D+00,   2.370040088348677D-06, &
   0.000000000000000D+00,   5.911208508309537D-07,   0.000000000000000D+00, &
  -1.559824227914752D-07,   0.000000000000000D+00,  -7.515125228849002D-04, &
   0.000000000000000D+00,   6.073435120902055D-04,   0.000000000000000D+00, &
   2.811545506297200D-04,   0.000000000000000D+00,   4.975584441531584D-05, &
   0.000000000000000D+00,   8.558494888797101D-06,   0.000000000000000D+00, &
  -6.746154660200734D-07,   0.000000000000000D+00,  -1.325388918447228D-06, &
   0.000000000000000D+00,  -5.004020994506079D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.990160527794985D-05,   0.000000000000000D+00, &
  -1.084337857201425D-03,   0.000000000000000D+00,  -2.012366284912053D-04, &
   0.000000000000000D+00,  -3.894427850896253D-05,   0.000000000000000D+00, &
  -2.085742510554973D-05,   0.000000000000000D+00,  -7.084838325676051D-06, &
   0.000000000000000D+00,  -8.312012277060443D-07,   0.000000000000000D+00, &
   6.655087423164066D-07,   0.000000000000000D+00,   7.723275860901201D-07, &
   0.000000000000000D+00,   5.745811604172751D-04,   0.000000000000000D+00, &
   6.473339757907150D-04,   0.000000000000000D+00,  -2.577442599836275D-04, &
   0.000000000000000D+00,  -8.026949882432194D-05,   0.000000000000000D+00, &
  -5.576499916443155D-06,   0.000000000000000D+00,   6.716113024449652D-06, &
   0.000000000000000D+00,   4.504571134361482D-06,   0.000000000000000D+00, &
   1.176078429022701D-06,   0.000000000000000D+00,  -3.001258431282351D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.850507631375258D-04, &
   0.000000000000000D+00,   2.036931122388398D-04,   0.000000000000000D+00, &
   5.658204311392581D-04,   0.000000000000000D+00,   1.270358897207956D-04, &
   0.000000000000000D+00,   4.052680795952040D-05,   0.000000000000000D+00, &
   1.063115984348157D-05,   0.000000000000000D+00,  -6.918581252361188D-07, &
   0.000000000000000D+00,  -2.413228568056092D-06,   0.000000000000000D+00, &
  -1.106342378851464D-06,   0.000000000000000D+00,   4.860851088596298D-07, &
   0.000000000000000D+00,   2.123552518300068D-04,   0.000000000000000D+00, &
  -5.908879220807053D-04,   0.000000000000000D+00,  -2.753168781571717D-04, &
   0.000000000000000D+00,   4.434899858404484D-05,   0.000000000000000D+00, &
  -5.893939850718507D-06,   0.000000000000000D+00,  -1.412598264755616D-05, &
   0.000000000000000D+00,  -7.758223853120284D-06,   0.000000000000000D+00, &
  -1.380259460411247D-06,   0.000000000000000D+00,   1.080182022843400D-06, &
   0.000000000000000D+00,   7.394427431859491D-07,   0.000000000000000D+00 /)
lem(:,192) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.452473077233087D-04,  -1.018280022249123D-03, &
   1.697837758167171D-04,   1.336032456309669D-03,   3.429277596411012D-03, &
  -4.723313127106385D-04,  -9.807920333831578D-03,  -8.148453622646178D-04, &
   8.899429427072093D-03,   2.113799427806252D-03,  -2.130165616565436D-03, &
  -1.469110675394522D-03,  -1.059406903400075D-03,  -1.231563735552900D-04, &
   4.758960232719497D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.452473077232889D-04, &
  -1.018280022249109D-03,   1.697837758167329D-04,   1.336032456309650D-03, &
   3.429277596411001D-03,  -4.723313127106277D-04,  -9.807920333831578D-03, &
  -8.148453622646082D-04,   8.899429427072076D-03,   2.113799427806244D-03, &
  -2.130165616565434D-03,  -1.469110675394510D-03,  -1.059406903400070D-03, &
  -1.231563735552979D-04,   4.758960232719397D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.865900864163237D-07,   0.000000000000000D+00, &
  -2.555361916024350D-06,   0.000000000000000D+00,  -8.419625219200676D-06, &
   0.000000000000000D+00,  -4.178433417394286D-06,   0.000000000000000D+00, &
   1.539703541341532D-05,   0.000000000000000D+00,   1.127744957971038D-05, &
   0.000000000000000D+00,  -1.411979023466381D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.277530997142283D-06,   0.000000000000000D+00,   9.436777507836212D-08, &
   0.000000000000000D+00,   4.329910777583593D-09,   0.000000000000000D+00, &
  -5.564137681514488D-10,  -4.583068925510649D-05,   0.000000000000000D+00, &
  -1.580037049144741D-06,   0.000000000000000D+00,  -1.193086007103830D-07, &
   0.000000000000000D+00,  -5.778142977026847D-09,   0.000000000000000D+00, &
   3.043263124445509D-05,   0.000000000000000D+00,   7.147539307156656D-06, &
   0.000000000000000D+00,   9.168625220185833D-07,   0.000000000000000D+00, &
   1.078830883163222D-07,   0.000000000000000D+00,  -6.694952731854858D-09, &
   6.217960237042021D-04,   0.000000000000000D+00,  -1.779738171604469D-06, &
   0.000000000000000D+00,  -1.976627526565405D-06,   0.000000000000000D+00, &
  -5.119497738706097D-07,   0.000000000000000D+00,  -7.695418263696588D-08, &
   0.000000000000000D+00,  -1.931637235903234D-03,   0.000000000000000D+00, &
   5.905062287686504D-06,   0.000000000000000D+00,  -3.075364405869293D-06, &
   0.000000000000000D+00,   4.028930475335123D-07,   0.000000000000000D+00, &
   2.723886389106113D-07,   0.000000000000000D+00,   6.326734006484835D-08, &
   1.528633027653042D-03,   0.000000000000000D+00,  -2.652569951504122D-04, &
   0.000000000000000D+00,   8.259482023470975D-06,   0.000000000000000D+00, &
   2.435075816722542D-06,   0.000000000000000D+00,   6.814912043362147D-08, &
   0.000000000000000D+00,  -1.662297809408989D-07,   0.000000000000000D+00, &
   2.093680223759923D-03,   0.000000000000000D+00,   8.299688744826711D-04, &
   0.000000000000000D+00,  -7.874501604801626D-07,   0.000000000000000D+00, &
  -2.929074916436719D-06,   0.000000000000000D+00,  -1.433220858866506D-06, &
   0.000000000000000D+00,  -1.333811696671857D-07,   0.000000000000000D+00, &
   3.154242515673320D-07,  -4.725334919004202D-03,   0.000000000000000D+00, &
  -6.679240571640717D-04,   0.000000000000000D+00,   1.081223155499791D-04, &
   0.000000000000000D+00,  -5.516119394493237D-07,   0.000000000000000D+00, &
   7.189912972348793D-07,   0.000000000000000D+00,   7.488066482597765D-07, &
   0.000000000000000D+00,   4.829161894732671D-08,   0.000000000000000D+00, &
   2.244844912180604D-03,   0.000000000000000D+00,  -9.363925748615912D-04, &
   0.000000000000000D+00,  -4.235046645220766D-04,   0.000000000000000D+00, &
  -1.796596422933300D-05,   0.000000000000000D+00,   4.249066421423012D-07, &
   0.000000000000000D+00,  -8.229816017984205D-08,   0.000000000000000D+00, &
  -4.201227690083197D-07,   0.000000000000000D+00,  -9.913604122250558D-09, &
   1.766962664457464D-03,   0.000000000000000D+00,   2.114293553409737D-03, &
   0.000000000000000D+00,   3.927097034197232D-04,   0.000000000000000D+00, &
   2.398487498643727D-06,   0.000000000000000D+00,   4.681067431948023D-06, &
   0.000000000000000D+00,  -3.667706302785146D-08,   0.000000000000000D+00, &
   4.172318332094177D-08,   0.000000000000000D+00,   2.773026674107132D-07, &
   0.000000000000000D+00,  -2.402930874275847D-03,   0.000000000000000D+00, &
  -1.003100963868457D-03,   0.000000000000000D+00,   4.724510448396389D-04, &
   0.000000000000000D+00,   1.659397683825673D-04,   0.000000000000000D+00, &
   1.062483004425175D-05,   0.000000000000000D+00,  -9.930762624844007D-07, &
   0.000000000000000D+00,  -1.069148805570843D-07,   0.000000000000000D+00, &
  -1.137469010036003D-07,   0.000000000000000D+00,  -9.868026555858872D-07, &
   8.379694141651865D-04,   0.000000000000000D+00,  -7.719717055210603D-04, &
   0.000000000000000D+00,  -1.136037476679387D-03,   0.000000000000000D+00, &
  -2.313373604426494D-04,   0.000000000000000D+00,  -4.020178743377834D-05, &
   0.000000000000000D+00,  -7.674834321735568D-06,   0.000000000000000D+00, &
  -3.631597740411696D-07,   0.000000000000000D+00,   1.671166029380796D-08, &
   0.000000000000000D+00,   2.561440870543772D-07,   0.000000000000000D+00, &
   4.172507620783626D-05,   0.000000000000000D+00,   1.030690127367594D-03, &
   0.000000000000000D+00,   5.058561951723382D-04,   0.000000000000000D+00, &
  -1.577089286189178D-04,   0.000000000000000D+00,  -2.490338759421234D-05, &
   0.000000000000000D+00,   8.694900341441175D-06,   0.000000000000000D+00, &
   4.434258917952984D-06,   0.000000000000000D+00,   8.305541777434235D-07, &
   0.000000000000000D+00,   2.633120537393027D-07,   0.000000000000000D+00, &
  -9.643799280999311D-08,  -7.194751354929285D-05,   0.000000000000000D+00, &
  -3.554909738592086D-04,   0.000000000000000D+00,   4.528483135648673D-04, &
   0.000000000000000D+00,   4.938029604474748D-04,   0.000000000000000D+00, &
   1.114612359391207D-04,   0.000000000000000D+00,   2.336915434316816D-05, &
   0.000000000000000D+00,   2.569128236049140D-07,   0.000000000000000D+00, &
  -2.139846107530199D-06,   0.000000000000000D+00,  -9.034548728315073D-07, &
   0.000000000000000D+00,  -1.626641653680363D-07,   0.000000000000000D+00, &
   7.361031805535412D-06,   0.000000000000000D+00,   8.817308836063491D-06, &
   0.000000000000000D+00,  -5.242191407015943D-04,   0.000000000000000D+00, &
  -1.690861711948870D-04,   0.000000000000000D+00,   1.553921497714943D-05, &
   0.000000000000000D+00,  -1.862270338538607D-05,   0.000000000000000D+00, &
  -1.226748145295808D-05,   0.000000000000000D+00,  -2.411737992769076D-06, &
   0.000000000000000D+00,   7.017671421470011D-07,   0.000000000000000D+00, &
   5.836896621135044D-07,   0.000000000000000D+00,   1.420027881361762D-06 /)
lem(:,193) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.359776627597792D-05, &
   3.656676467729653D-04,  -7.004934566313602D-04,  -1.333291387335572D-03, &
   3.358862291392439D-03,   3.489600291386719D-03,  -8.669021880008811D-03, &
  -5.012750372208420D-03,   1.224316375143186D-02,   4.021547068503580D-03, &
  -8.775711785171180D-03,  -2.654099695051735D-03,   2.475530534627560D-03, &
   1.763687110778521D-03,   5.387287708398877D-04,  -6.199977382441839D-04, &
  -7.083751256056590D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.359776627599899D-05,  -3.656676467729644D-04,   7.004934566313366D-04, &
   1.333291387335588D-03,  -3.358862291392430D-03,  -3.489600291386737D-03, &
   8.669021880008846D-03,   5.012750372208421D-03,  -1.224316375143187D-02, &
  -4.021547068503570D-03,   8.775711785171159D-03,   2.654099695051732D-03, &
  -2.475530534627553D-03,  -1.763687110778506D-03,  -5.387287708399101D-04, &
   6.199977382441826D-04,   7.083751256056750D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.734368866868577D-08,   0.000000000000000D+00, &
   2.337065253291493D-06,   0.000000000000000D+00,   1.247485407388296D-05, &
   0.000000000000000D+00,   1.790566659627029D-05,   0.000000000000000D+00, &
  -6.193067599583796D-06,   0.000000000000000D+00,  -2.775593988228994D-05, &
   0.000000000000000D+00,  -8.269656094680297D-06,   0.000000000000000D+00, &
   1.668900087604266D-05,   0.000000000000000D+00,   1.122480898239233D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.156980870163133D-08, &
   0.000000000000000D+00,   1.120609546881054D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.569600438896397D-06,   0.000000000000000D+00, &
  -7.780231998045811D-08,   0.000000000000000D+00,   1.218662882414851D-08, &
   0.000000000000000D+00,   8.020576064727137D-06,   0.000000000000000D+00, &
   6.953431649013828D-07,   0.000000000000000D+00,   3.979409013735545D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.736182380010297D-06, &
   0.000000000000000D+00,  -1.549224911338652D-06,   0.000000000000000D+00, &
  -3.259946556488173D-07,   0.000000000000000D+00,   2.856542683557664D-08, &
   0.000000000000000D+00,  -5.948574645340845D-05,   0.000000000000000D+00, &
  -3.574439519521421D-06,   0.000000000000000D+00,   2.337916783596043D-07, &
   0.000000000000000D+00,   1.476525443645217D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.563215409725594D-04,   0.000000000000000D+00, &
   1.297244566399429D-05,   0.000000000000000D+00,   2.422596028560011D-06, &
   0.000000000000000D+00,   5.864075159727675D-08,   0.000000000000000D+00, &
  -2.333192426085970D-07,   0.000000000000000D+00,  -4.150414591679771D-04, &
   0.000000000000000D+00,   4.042675998354024D-05,   0.000000000000000D+00, &
  -2.638226001634100D-06,   0.000000000000000D+00,  -1.213769810284105D-06, &
   0.000000000000000D+00,  -9.217261040598894D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.549164360572818D-04,   0.000000000000000D+00, &
  -2.506873645212392D-04,   0.000000000000000D+00,  -1.364928895622291D-05, &
   0.000000000000000D+00,  -2.782061759905692D-07,   0.000000000000000D+00, &
   4.851201033414626D-07,   0.000000000000000D+00,  -3.127179941669431D-07, &
   0.000000000000000D+00,   5.281837129021248D-04,   0.000000000000000D+00, &
   4.586832032050648D-04,   0.000000000000000D+00,   9.083445199000416D-06, &
   0.000000000000000D+00,   6.055723353769041D-06,   0.000000000000000D+00, &
   7.914705890549162D-07,   0.000000000000000D+00,  -7.772795281268802D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.476510316143676D-04, &
   0.000000000000000D+00,  -2.056563471647828D-04,   0.000000000000000D+00, &
   1.244180772147486D-04,   0.000000000000000D+00,   4.213619880241541D-06, &
   0.000000000000000D+00,  -2.168154485292682D-06,   0.000000000000000D+00, &
  -5.569947196788028D-07,   0.000000000000000D+00,   5.339038393769514D-07, &
   0.000000000000000D+00,   5.676056372656789D-04,   0.000000000000000D+00, &
  -5.605082929407533D-04,   0.000000000000000D+00,  -3.218310548684547D-04, &
   0.000000000000000D+00,  -3.326226984985677D-05,   0.000000000000000D+00, &
  -5.115922324654523D-06,   0.000000000000000D+00,   3.071709562168194D-07, &
   0.000000000000000D+00,   2.143989598001517D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.791414277255398D-04,   0.000000000000000D+00, &
   1.043358412745117D-03,   0.000000000000000D+00,   2.029719495467115D-04, &
   0.000000000000000D+00,  -6.506155678297233D-06,   0.000000000000000D+00, &
   9.678224465488051D-06,   0.000000000000000D+00,   3.606264139699075D-06, &
   0.000000000000000D+00,   2.869942394090603D-07,   0.000000000000000D+00, &
   4.164487920318765D-07,   0.000000000000000D+00,  -5.200880142081023D-04, &
   0.000000000000000D+00,  -6.021408257648072D-04,   0.000000000000000D+00, &
   3.514366357340658D-04,   0.000000000000000D+00,   1.456506571517006D-04, &
   0.000000000000000D+00,   1.742407276165966D-05,   0.000000000000000D+00, &
  -1.400125768107455D-06,   0.000000000000000D+00,  -1.936003545120932D-06, &
   0.000000000000000D+00,  -7.516651148064409D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.517717798319428D-04,   0.000000000000000D+00, &
  -2.377225736124120D-04,   0.000000000000000D+00,  -7.222193640495506D-04, &
   0.000000000000000D+00,  -1.521534766833258D-04,   0.000000000000000D+00, &
  -3.494154044203553D-05,   0.000000000000000D+00,  -1.153688923273701D-05, &
   0.000000000000000D+00,  -1.314904229796198D-06,   0.000000000000000D+00, &
   7.784722914534599D-07,   0.000000000000000D+00,  -1.331770801447411D-07, &
   0.000000000000000D+00,  -9.247433517326249D-05,   0.000000000000000D+00, &
   5.743683885952012D-04,   0.000000000000000D+00,   3.776530006396226D-04, &
   0.000000000000000D+00,  -1.272769320472504D-04,   0.000000000000000D+00, &
  -2.914171676441878D-05,   0.000000000000000D+00,   5.422479852182794D-06, &
   0.000000000000000D+00,   5.994104281369971D-06,   0.000000000000000D+00, &
   1.886348501040277D-06,   0.000000000000000D+00,  -1.082901508472183D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.234020685585182D-05, &
   0.000000000000000D+00,  -3.440548556915913D-04,   0.000000000000000D+00, &
   2.341105539063962D-04,   0.000000000000000D+00,   3.383724263994635D-04, &
   0.000000000000000D+00,   7.973390460431423D-05,   0.000000000000000D+00, &
   2.192135319731773D-05,   0.000000000000000D+00,   2.350811649328224D-06, &
   0.000000000000000D+00,  -2.491224169766631D-06,   0.000000000000000D+00, &
  -1.242816434193938D-06,   0.000000000000000D+00,  -3.739400892311191D-07, &
   0.000000000000000D+00,   1.447263817258957D-05,   0.000000000000000D+00, &
   6.386531157646577D-05,   0.000000000000000D+00,  -4.026563383271306D-04, &
   0.000000000000000D+00,  -1.370659240460748D-04,   0.000000000000000D+00, &
   1.666834233282375D-05,   0.000000000000000D+00,  -1.014408974279600D-05, &
   0.000000000000000D+00,  -1.027764227332076D-05,   0.000000000000000D+00, &
  -3.482085911835005D-06,   0.000000000000000D+00,   4.666188357638897D-07, &
   0.000000000000000D+00,   1.364778307036290D-06,   0.000000000000000D+00 /)
lem(:,194) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.056306207248804D-05,  -1.324351789833256D-04,   3.627806005020972D-04, &
   4.017692561129125D-04,  -1.105709098059747D-03,   6.179530526790729D-04, &
  -2.135992063036518D-03,  -3.402701172308988D-03,   1.064310024920904D-02, &
   5.798850598345773D-03,  -1.515454030927085D-02,  -5.109626741346606D-03, &
   9.845654399657264D-03,   2.748190894708729D-03,  -2.869039894136251D-03, &
  -1.274009419965869D-03,   3.360652616394729D-04,   7.243528465020342D-04, &
   5.565601189369409D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.056306207248413D-05,  -1.324351789833404D-04, &
   3.627806005021087D-04,   4.017692561129163D-04,  -1.105709098059759D-03, &
   6.179530526790790D-04,  -2.135992063036508D-03,  -3.402701172308985D-03, &
   1.064310024920904D-02,   5.798850598345780D-03,  -1.515454030927084D-02, &
  -5.109626741346607D-03,   9.845654399657237D-03,   2.748190894708736D-03, &
  -2.869039894136241D-03,  -1.274009419965878D-03,   3.360652616394702D-04, &
   7.243528465020499D-04,   5.565601189368674D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.533036365022320D-06,   0.000000000000000D+00,  -1.444242658415786D-05, &
   0.000000000000000D+00,  -3.287964280131317D-05,   0.000000000000000D+00, &
  -1.261982844040061D-05,   0.000000000000000D+00,   2.384088982630042D-05, &
   0.000000000000000D+00,   1.991211018670905D-05,   0.000000000000000D+00, &
   1.213126993322035D-05,   0.000000000000000D+00,  -6.424381654624102D-06, &
   0.000000000000000D+00,  -2.544712349799557D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.694921939909953D-07,   0.000000000000000D+00,  -3.852095318040439D-09, &
  -2.696466722580747D-06,   0.000000000000000D+00,  -4.890993752207801D-08, &
   0.000000000000000D+00,   1.049824377324716D-05,   0.000000000000000D+00, &
   5.956459971131639D-07,   0.000000000000000D+00,  -6.333583993769739D-08, &
   1.505583750838868D-05,   0.000000000000000D+00,  -1.513913481732565D-06, &
   0.000000000000000D+00,  -1.791905248796745D-07,   0.000000000000000D+00, &
  -1.177734486389767D-04,   0.000000000000000D+00,  -2.043976249791351D-06, &
   0.000000000000000D+00,   2.028225460046511D-07,   0.000000000000000D+00, &
   9.405970472821142D-08,   2.376616120054049D-05,   0.000000000000000D+00, &
   5.858518696720253D-06,   0.000000000000000D+00,   2.006367274156256D-06, &
   0.000000000000000D+00,  -3.845529246413880D-08,   0.000000000000000D+00, &
   5.608037706500861D-04,   0.000000000000000D+00,   7.565531305516386D-09, &
   0.000000000000000D+00,  -3.503464143741838D-06,   0.000000000000000D+00, &
  -9.510943391005848D-07,   0.000000000000000D+00,   5.219110411247988D-07, &
  -7.330192691033938D-04,   0.000000000000000D+00,   9.022009602214853D-05, &
   0.000000000000000D+00,  -5.200725570755341D-06,   0.000000000000000D+00, &
   5.014451830626365D-07,   0.000000000000000D+00,   3.027589686412020D-07, &
   0.000000000000000D+00,  -4.963302949880304D-04,   0.000000000000000D+00, &
  -3.571618643978381D-04,   0.000000000000000D+00,   3.331808284799745D-05, &
   0.000000000000000D+00,   4.427144890258750D-06,   0.000000000000000D+00, &
   1.733220991355529D-07,   0.000000000000000D+00,  -4.702812642185089D-07, &
   1.530771306131741D-03,   0.000000000000000D+00,   3.330553561877189D-04, &
   0.000000000000000D+00,  -1.364821502300053D-04,   0.000000000000000D+00, &
  -7.567107847206247D-06,   0.000000000000000D+00,  -2.399464139179381D-06, &
   0.000000000000000D+00,  -2.917319714108803D-07,   0.000000000000000D+00, &
  -5.394871433627488D-04,   0.000000000000000D+00,   4.064145781598479D-04, &
   0.000000000000000D+00,   3.174978954985336D-04,   0.000000000000000D+00, &
  -6.585192858776305D-06,   0.000000000000000D+00,   1.140559954965409D-06, &
   0.000000000000000D+00,   1.253760260489356D-06,   0.000000000000000D+00, &
  -5.516305511397086D-07,  -8.665725691264918D-04,   0.000000000000000D+00, &
  -9.803837701518551D-04,   0.000000000000000D+00,  -2.234165496947115D-04, &
   0.000000000000000D+00,   8.350034277978468D-05,   0.000000000000000D+00, &
   7.738798795285711D-06,   0.000000000000000D+00,   5.847329050301942D-07, &
   0.000000000000000D+00,  -3.669095853464858D-07,   0.000000000000000D+00, &
   7.272047200762058D-04,   0.000000000000000D+00,   4.392822744378019D-04, &
   0.000000000000000D+00,  -3.982393130350830D-04,   0.000000000000000D+00, &
  -2.243547912690518D-04,   0.000000000000000D+00,  -1.927648248867730D-05, &
   0.000000000000000D+00,  -4.767449865042469D-06,   0.000000000000000D+00, &
  -8.091981299116294D-07,   0.000000000000000D+00,   1.169561013923862D-06, &
   2.774812012383530D-05,   0.000000000000000D+00,   3.927097034197232D-04, &
   0.000000000000000D+00,   8.516398958750416D-04,   0.000000000000000D+00, &
   1.737839892459658D-04,   0.000000000000000D+00,  -6.318256940566725D-06, &
   0.000000000000000D+00,   4.036206617603433D-06,   0.000000000000000D+00, &
   2.455695324041560D-06,   0.000000000000000D+00,   1.064308949836084D-06, &
   0.000000000000000D+00,  -1.819055558908900D-04,   0.000000000000000D+00, &
  -4.569544212920669D-04,   0.000000000000000D+00,  -4.292029854605905D-04, &
   0.000000000000000D+00,   2.571744418177518D-04,   0.000000000000000D+00, &
   1.033592144533433D-04,   0.000000000000000D+00,   1.235869718143887D-05, &
   0.000000000000000D+00,   4.280826676808669D-07,   0.000000000000000D+00, &
  -1.539201328493296D-06,   0.000000000000000D+00,  -1.147567776831223D-06, &
   2.788106381471022D-05,   0.000000000000000D+00,   1.261818710356561D-04, &
   0.000000000000000D+00,  -2.615659733286196D-04,   0.000000000000000D+00, &
  -5.759216286366213D-04,   0.000000000000000D+00,  -1.234859093334540D-04, &
   0.000000000000000D+00,  -2.518337087469118D-05,   0.000000000000000D+00, &
  -7.771720880455951D-06,   0.000000000000000D+00,  -1.540425478084756D-06, &
   0.000000000000000D+00,   1.335458403997578D-07,   0.000000000000000D+00, &
   1.916383584420525D-05,   0.000000000000000D+00,  -1.762179849762342D-06, &
   0.000000000000000D+00,   4.014596010708953D-04,   0.000000000000000D+00, &
   2.770982514094167D-04,   0.000000000000000D+00,  -9.357907684213193D-05, &
   0.000000000000000D+00,  -1.857099738704503D-05,   0.000000000000000D+00, &
   4.314873578165982D-06,   0.000000000000000D+00,   4.094056491153388D-06, &
   0.000000000000000D+00,   1.614843450355483D-06,   0.000000000000000D+00, &
  -5.160138886713754D-07,  -5.696791860161142D-06,   0.000000000000000D+00, &
   1.262756430702105D-05,   0.000000000000000D+00,  -1.894520961209424D-04, &
   0.000000000000000D+00,   2.016665565964868D-04,   0.000000000000000D+00, &
   2.727292682450622D-04,   0.000000000000000D+00,   6.419366572315172D-05, &
   0.000000000000000D+00,   1.565529034755194D-05,   0.000000000000000D+00, &
   1.269299765648961D-06,   0.000000000000000D+00,  -1.466924872559225D-06, &
   0.000000000000000D+00,  -1.303167014724209D-06,   0.000000000000000D+00, &
   4.160583194422908D-07,   0.000000000000000D+00,  -5.332207018454229D-06, &
   0.000000000000000D+00,   5.555865693459801D-05,   0.000000000000000D+00, &
  -2.820992464617307D-04,   0.000000000000000D+00,  -1.010791275940242D-04, &
   0.000000000000000D+00,   1.033380050436361D-05,   0.000000000000000D+00, &
  -1.008050986844549D-05,   0.000000000000000D+00,  -7.879444596208980D-06, &
   0.000000000000000D+00,  -2.410035888209032D-06,   0.000000000000000D+00, &
   7.634186558166334D-07,   0.000000000000000D+00,   2.452398683050677D-06 /)
lem(:,195) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.150460633783014D-07,   2.742442849627292D-05, &
  -1.785115724661498D-04,   1.257798848636751D-04,   1.045253120827275D-03, &
  -9.545564225446372D-04,  -2.386538065102859D-03,  -6.408879412192719D-04, &
   7.077374997472257D-03,   4.857774385047901D-03,  -1.456223124638055D-02, &
  -6.335932216762480D-03,   1.687174083766183D-02,   3.413074951815139D-03, &
  -1.076533300100476D-02,  -8.631241581208830D-04,   3.761987138059861D-03, &
   3.973951817192651D-04,  -9.245685249378209D-04,  -3.110825882214870D-04, &
   3.639181052624070D-04,   0.000000000000000D+00,   8.150460633770808D-07, &
  -2.742442849626683D-05,   1.785115724661459D-04,  -1.257798848636871D-04, &
  -1.045253120827264D-03,   9.545564225446407D-04,   2.386538065102851D-03, &
   6.408879412192731D-04,  -7.077374997472235D-03,  -4.857774385047893D-03, &
   1.456223124638054D-02,   6.335932216762468D-03,  -1.687174083766183D-02, &
  -3.413074951815131D-03,   1.076533300100477D-02,   8.631241581208790D-04, &
  -3.761987138059842D-03,  -3.973951817192666D-04,   9.245685249378081D-04, &
   3.110825882214831D-04,  -3.639181052624045D-04,   0.000000000000000D+00, &
   4.075230316820916D-07,   0.000000000000000D+00,   1.296450086424739D-05, &
   0.000000000000000D+00,   4.718241678262883D-05,   0.000000000000000D+00, &
   4.000597686719507D-05,   0.000000000000000D+00,  -1.608593963480873D-05, &
   0.000000000000000D+00,  -1.287514659539121D-05,   0.000000000000000D+00, &
  -3.675160666351840D-07,   0.000000000000000D+00,  -3.613013476486561D-05, &
   0.000000000000000D+00,  -7.342860718890911D-06,   0.000000000000000D+00, &
   4.163730496727843D-05,   0.000000000000000D+00,   5.841848923485515D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.221752447775484D-08, &
   0.000000000000000D+00,   7.025599299122997D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.397637510034892D-06,   0.000000000000000D+00, &
  -4.097228656370468D-08,   0.000000000000000D+00,  -1.182376478032563D-06, &
   0.000000000000000D+00,   2.573032915031177D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.122845372273308D-05,   0.000000000000000D+00, &
   1.470930101249553D-06,   0.000000000000000D+00,  -6.031400734592363D-07, &
   0.000000000000000D+00,  -2.896790620575802D-05,   0.000000000000000D+00, &
  -1.921003804308579D-06,   0.000000000000000D+00,  -5.737952782473283D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.612079848181596D-05, &
   0.000000000000000D+00,  -1.109929737845447D-05,   0.000000000000000D+00, &
   7.908071535409735D-07,   0.000000000000000D+00,   1.355788720971706D-07, &
   0.000000000000000D+00,   1.987958233674384D-04,   0.000000000000000D+00, &
   8.917739410907439D-06,   0.000000000000000D+00,   1.840866012816570D-06, &
   0.000000000000000D+00,  -1.403544908314441D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.479226127679767D-04,   0.000000000000000D+00, &
   1.007272391955438D-04,   0.000000000000000D+00,   9.279404185521454D-07, &
   0.000000000000000D+00,  -1.046013707125796D-06,   0.000000000000000D+00, &
   1.534436845571678D-06,   0.000000000000000D+00,  -1.882643597836635D-04, &
   0.000000000000000D+00,  -2.691372861901551D-04,   0.000000000000000D+00, &
   8.616975666913328D-06,   0.000000000000000D+00,   6.773893908721552D-07, &
   0.000000000000000D+00,   4.278174142633074D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.209282519159283D-04,   0.000000000000000D+00, &
   1.726552771828699D-04,   0.000000000000000D+00,  -1.093236536591280D-04, &
   0.000000000000000D+00,  -3.646464585023811D-06,   0.000000000000000D+00, &
  -8.885581088009923D-07,   0.000000000000000D+00,  -1.289604004618827D-06, &
   0.000000000000000D+00,  -2.054684307792749D-04,   0.000000000000000D+00, &
   2.998900871970221D-04,   0.000000000000000D+00,   2.615501525849153D-04, &
   0.000000000000000D+00,   4.168433784289652D-06,   0.000000000000000D+00, &
   1.234763675448716D-06,   0.000000000000000D+00,   3.053750743019015D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.742525498007689D-04, &
   0.000000000000000D+00,  -6.206600512632393D-04,   0.000000000000000D+00, &
  -1.579803430386101D-04,   0.000000000000000D+00,   6.087726793778555D-05, &
   0.000000000000000D+00,   2.906157890866792D-06,   0.000000000000000D+00, &
   2.436900797261277D-07,   0.000000000000000D+00,  -1.543585958313504D-06, &
   0.000000000000000D+00,   2.604048957556261D-04,   0.000000000000000D+00, &
   3.254887454185666D-04,   0.000000000000000D+00,  -3.142273045499182D-04, &
   0.000000000000000D+00,  -1.861947056843112D-04,   0.000000000000000D+00, &
  -1.825470033957733D-05,   0.000000000000000D+00,  -3.040392469955626D-06, &
   0.000000000000000D+00,  -2.221714384589376D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.698471584875189D-05,   0.000000000000000D+00, &
   2.029719495467115D-04,   0.000000000000000D+00,   6.340543901090794D-04, &
   0.000000000000000D+00,   1.367072759655286D-04,   0.000000000000000D+00, &
  -1.779603948386356D-07,   0.000000000000000D+00,   5.418243790040906D-06, &
   0.000000000000000D+00,   2.584119377085560D-06,   0.000000000000000D+00, &
   2.763234753129178D-06,   0.000000000000000D+00,  -4.150375417016625D-05, &
   0.000000000000000D+00,  -3.485967174928456D-04,   0.000000000000000D+00, &
  -3.397882223246466D-04,   0.000000000000000D+00,   2.072156258011586D-04, &
   0.000000000000000D+00,   8.630914696820816D-05,   0.000000000000000D+00, &
   9.801684529495731D-06,   0.000000000000000D+00,  -1.020815018140637D-06, &
   0.000000000000000D+00,  -1.318781708939862D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.915096888921056D-05,   0.000000000000000D+00, &
   1.481151634912577D-04,   0.000000000000000D+00,  -1.845069893630866D-04, &
   0.000000000000000D+00,  -4.498811358951233D-04,   0.000000000000000D+00, &
  -1.004890209114343D-04,   0.000000000000000D+00,  -2.336597963320795D-05, &
   0.000000000000000D+00,  -7.302531381201347D-06,   0.000000000000000D+00, &
  -1.572405233632844D-06,   0.000000000000000D+00,   5.958918001625696D-07, &
   0.000000000000000D+00,  -5.290225765120466D-06,   0.000000000000000D+00, &
   9.941397323907691D-06,   0.000000000000000D+00,   3.343955794221301D-04, &
   0.000000000000000D+00,   2.239011141504473D-04,   0.000000000000000D+00, &
  -7.675153363896605D-05,   0.000000000000000D+00,  -1.581396251290346D-05, &
   0.000000000000000D+00,   4.740229118280038D-06,   0.000000000000000D+00, &
   4.208473171491968D-06,   0.000000000000000D+00,   1.015071633493533D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.670948237789472D-06, &
   0.000000000000000D+00,  -1.897079887209884D-05,   0.000000000000000D+00, &
  -1.565056889714319D-04,   0.000000000000000D+00,   1.582312958125735D-04, &
   0.000000000000000D+00,   2.171809173897918D-04,   0.000000000000000D+00, &
   5.288053270094156D-05,   0.000000000000000D+00,   1.398079290042101D-05, &
   0.000000000000000D+00,   1.219713417624751D-06,   0.000000000000000D+00, &
  -1.916438423632093D-06,   0.000000000000000D+00,  -3.270697866714804D-06, &
   0.000000000000000D+00,   2.196392115122669D-06,   0.000000000000000D+00, &
  -4.331966497751265D-06,   0.000000000000000D+00,   1.803174321883813D-05, &
   0.000000000000000D+00,  -2.359250712674516D-04,   0.000000000000000D+00, &
  -8.319303347659960D-05,   0.000000000000000D+00,   9.052416826313048D-06, &
   0.000000000000000D+00,  -8.211924138386904D-06,   0.000000000000000D+00, &
  -7.372465200679548D-06,   0.000000000000000D+00,  -1.866228547476931D-06, &
   0.000000000000000D+00,   9.962647222024809D-07,   0.000000000000000D+00 /)
lem(:,196) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -9.722512275860079D-06,   7.139124310078470D-05,  -4.603788035002149D-05, &
  -4.207956656630166D-04,   3.242214775794179D-04,   1.488583813451616D-04, &
   2.765076330204937D-04,   2.243587041138008D-03,   4.575076043414761D-04, &
  -9.237380627235368D-03,  -3.660273215968780D-03,   1.704677152800424D-02, &
   4.041391175278755D-03,  -1.705728118041509D-02,   7.345475223656749D-06, &
   9.617417103590334D-03,  -2.288716735134133D-03,  -3.232188437926192D-03, &
   1.323081343007383D-03,   9.323211162653016D-04,  -3.766026684686751D-04, &
  -3.685745125927411D-04,  -9.722512275872532D-06,   7.139124310079266D-05, &
  -4.603788035001572D-05,  -4.207956656630267D-04,   3.242214775794095D-04, &
   1.488583813451754D-04,   2.765076330205020D-04,   2.243587041137999D-03, &
   4.575076043414760D-04,  -9.237380627235340D-03,  -3.660273215968782D-03, &
   1.704677152800428D-02,   4.041391175278768D-03,  -1.705728118041508D-02, &
   7.345475223669850D-06,   9.617417103590299D-03,  -2.288716735134139D-03, &
  -3.232188437926183D-03,   1.323081343007381D-03,   9.323211162652873D-04, &
  -3.766026684686636D-04,  -3.685745125927357D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.568728967070233D-05,   0.000000000000000D+00, &
  -6.764369657022983D-05,   0.000000000000000D+00,   6.915906878269983D-06, &
   0.000000000000000D+00,   2.010447497147710D-05,   0.000000000000000D+00, &
  -3.622118690731094D-05,   0.000000000000000D+00,  -5.952947009646554D-06, &
   0.000000000000000D+00,   5.769684328504801D-05,   0.000000000000000D+00, &
   3.575930372282924D-05,   0.000000000000000D+00,  -5.019230063081975D-05, &
   0.000000000000000D+00,  -3.281849204432392D-05,   0.000000000000000D+00, &
   1.724174518292069D-06,  -2.430812221833787D-06,   0.000000000000000D+00, &
  -9.762913796094608D-06,   0.000000000000000D+00,   1.113722467538667D-06, &
   2.662239104728006D-05,   0.000000000000000D+00,   9.167986856768554D-08, &
   0.000000000000000D+00,   3.274533157424525D-05,   0.000000000000000D+00, &
  -1.068559568189365D-06,   0.000000000000000D+00,   2.911877643713502D-07, &
  -1.144288026775572D-04,   0.000000000000000D+00,   9.820391727445550D-07, &
   0.000000000000000D+00,   6.156830211002045D-07,   0.000000000000000D+00, &
  -5.364877740029180D-05,   0.000000000000000D+00,  -1.443265330985754D-05, &
   0.000000000000000D+00,   2.799596360830179D-06,   0.000000000000000D+00, &
  -2.239330691910922D-06,   2.665624746322010D-04,   0.000000000000000D+00, &
   3.329295945226669D-06,   0.000000000000000D+00,  -7.532500510294518D-06, &
   0.000000000000000D+00,  -2.171014656923762D-07,   0.000000000000000D+00, &
   3.086671473405033D-05,   0.000000000000000D+00,   1.149341691386843D-04, &
   0.000000000000000D+00,  -1.806570464814545D-06,   0.000000000000000D+00, &
  -1.084668712820032D-06,   0.000000000000000D+00,  -1.467989393376345D-07, &
  -3.746336570999131D-04,   0.000000000000000D+00,  -1.638232714809364D-04, &
   0.000000000000000D+00,   6.670994478461571D-05,   0.000000000000000D+00, &
   6.085105001216345D-06,   0.000000000000000D+00,   6.964842996584112D-07, &
   0.000000000000000D+00,   3.497979077060417D-05,   0.000000000000000D+00, &
  -1.130084385252220D-04,   0.000000000000000D+00,  -1.876920174504279D-04, &
   0.000000000000000D+00,   8.111840601581885D-06,   0.000000000000000D+00, &
   3.705154089221359D-07,   0.000000000000000D+00,   2.389274746397434D-06, &
   3.267595053219865D-04,   0.000000000000000D+00,   3.587411580736529D-04, &
   0.000000000000000D+00,   1.470507813089826D-04,   0.000000000000000D+00, &
  -8.645476052742768D-05,   0.000000000000000D+00,  -7.498784082909457D-06, &
   0.000000000000000D+00,  -1.393099902989086D-06,   0.000000000000000D+00, &
  -7.824909266943304D-05,   0.000000000000000D+00,  -1.240136533556764D-04, &
   0.000000000000000D+00,   2.184300020156041D-04,   0.000000000000000D+00, &
   2.021548173792724D-04,   0.000000000000000D+00,   6.247854483515524D-06, &
   0.000000000000000D+00,   2.055373944010563D-06,   0.000000000000000D+00, &
   2.967252813490544D-07,  -1.714258057538222D-04,   0.000000000000000D+00, &
  -1.972038333752597D-04,   0.000000000000000D+00,  -4.917811543978935D-04, &
   0.000000000000000D+00,  -1.315695643407012D-04,   0.000000000000000D+00, &
   4.935664578438932D-05,   0.000000000000000D+00,   3.738494858852575D-06, &
   0.000000000000000D+00,  -1.045014050151753D-06,   0.000000000000000D+00, &
   6.176176998501416D-05,   0.000000000000000D+00,   1.530035523553764D-04, &
   0.000000000000000D+00,   2.375815448387930D-04,   0.000000000000000D+00, &
  -2.458569693853017D-04,   0.000000000000000D+00,  -1.505536507286103D-04, &
   0.000000000000000D+00,  -1.748036678273068D-05,   0.000000000000000D+00, &
  -3.410015806162686D-06,   0.000000000000000D+00,  -1.583526566944987D-06, &
   4.747822609012742D-05,   0.000000000000000D+00,   2.398487498643727D-06, &
   0.000000000000000D+00,   1.737839892459658D-04,   0.000000000000000D+00, &
   5.131447826113334D-04,   0.000000000000000D+00,   1.153214701531801D-04, &
   0.000000000000000D+00,   9.019414063792555D-07,   0.000000000000000D+00, &
   5.765668054811901D-06,   0.000000000000000D+00,   3.605444817253983D-06, &
   0.000000000000000D+00,  -2.355178194662184D-05,   0.000000000000000D+00, &
  -1.907057388144436D-05,   0.000000000000000D+00,  -2.433407397838047D-04, &
   0.000000000000000D+00,  -2.661158466551857D-04,   0.000000000000000D+00, &
   1.666737861777551D-04,   0.000000000000000D+00,   7.070687944454380D-05, &
   0.000000000000000D+00,   8.568537074780326D-06,   0.000000000000000D+00, &
  -4.709368587238367D-07,   0.000000000000000D+00,  -2.576353366152242D-06, &
  -3.854589223726201D-06,   0.000000000000000D+00,  -1.083151686065353D-06, &
   0.000000000000000D+00,   1.004593628685650D-04,   0.000000000000000D+00, &
  -1.530732675125142D-04,   0.000000000000000D+00,  -3.713392926805651D-04, &
   0.000000000000000D+00,  -8.553375478635306D-05,   0.000000000000000D+00, &
  -2.047190374844721D-05,   0.000000000000000D+00,  -7.466736615114993D-06, &
   0.000000000000000D+00,  -9.975121808842161D-07,   0.000000000000000D+00, &
   3.550862748280768D-06,   0.000000000000000D+00,  -4.108598943091615D-06, &
   0.000000000000000D+00,  -1.240358294024619D-05,   0.000000000000000D+00, &
   2.576197817534103D-04,   0.000000000000000D+00,   1.801862839340043D-04, &
   0.000000000000000D+00,  -6.197434178259917D-05,   0.000000000000000D+00, &
  -1.206395159726409D-05,   0.000000000000000D+00,   4.316163223912096D-06, &
   0.000000000000000D+00,   3.811360636980585D-06,   0.000000000000000D+00, &
   3.583146094268207D-06,  -6.315733028990034D-07,   0.000000000000000D+00, &
   4.082243251814156D-06,   0.000000000000000D+00,  -8.818554673350908D-06, &
   0.000000000000000D+00,  -1.269374643034340D-04,   0.000000000000000D+00, &
   1.324163883596886D-04,   0.000000000000000D+00,   1.820046098464866D-04, &
   0.000000000000000D+00,   4.509149085057785D-05,   0.000000000000000D+00, &
   1.213431075381001D-05,   0.000000000000000D+00,   8.104934054195659D-07, &
   0.000000000000000D+00,  -2.479273685619581D-06,   0.000000000000000D+00, &
   3.584462338539511D-08,   0.000000000000000D+00,   4.439616655212060D-07, &
   0.000000000000000D+00,   8.035799574212992D-06,   0.000000000000000D+00, &
   2.281246077076592D-05,   0.000000000000000D+00,  -1.896920356816431D-04, &
   0.000000000000000D+00,  -6.724885379305015D-05,   0.000000000000000D+00, &
   6.612494492454363D-06,   0.000000000000000D+00,  -7.813791066773232D-06, &
   0.000000000000000D+00,  -6.152342882476614D-06,   0.000000000000000D+00, &
  -2.128464275910774D-06,   0.000000000000000D+00,   1.494540458556055D-06 /)
lem(:,197) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.558096391144162D-05,  -4.686212004163438D-04, &
   2.700623695272082D-04,   1.404625502311613D-03,   4.375092660923345D-04, &
  -1.976928881408161D-03,  -4.050573589067520D-03,   3.584133049028002D-04, &
   1.060916478927425D-02,   2.046285278877328D-03,  -1.599144920970777D-02, &
  -3.333893851102991D-04,   1.410225633774619D-02,  -3.764548304263401D-03, &
  -6.448973273453878D-03,   4.270634256177861D-03,   9.596957922306373D-04, &
  -2.095768053705992D-03,   2.448078730748949D-04,   7.188552766121525D-04, &
  -1.386460997967909D-04,   0.000000000000000D+00,   1.558096391144832D-05, &
   4.686212004163472D-04,  -2.700623695272117D-04,  -1.404625502311620D-03, &
  -4.375092660923264D-04,   1.976928881408169D-03,   4.050573589067510D-03, &
  -3.584133049028091D-04,  -1.060916478927424D-02,  -2.046285278877316D-03, &
   1.599144920970777D-02,   3.333893851102927D-04,  -1.410225633774617D-02, &
   3.764548304263407D-03,   6.448973273453857D-03,  -4.270634256177853D-03, &
  -9.596957922306245D-04,   2.095768053705997D-03,  -2.448078730749058D-04, &
  -7.188552766121495D-04,   1.386460997967992D-04,   0.000000000000000D+00, &
   7.790481955726327D-06,   0.000000000000000D+00,  -5.658580203023234D-05, &
   0.000000000000000D+00,  -8.614140816132017D-05,   0.000000000000000D+00, &
   3.690672071114317D-05,   0.000000000000000D+00,   7.551122950472073D-05, &
   0.000000000000000D+00,   3.250330898284701D-05,   0.000000000000000D+00, &
  -5.631224214862313D-05,   0.000000000000000D+00,  -1.008036976352387D-04, &
   0.000000000000000D+00,   4.271338537347143D-05,   0.000000000000000D+00, &
   8.016223709192293D-05,   0.000000000000000D+00,  -2.938162146519615D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.158916762631309D-07, &
   0.000000000000000D+00,  -7.366863816251422D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.310817560716653D-06,   0.000000000000000D+00, &
   5.480247605701108D-07,   0.000000000000000D+00,   2.456358228501179D-05, &
   0.000000000000000D+00,   3.798994767664921D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.602178324011682D-05,   0.000000000000000D+00, &
   1.977152123537849D-06,   0.000000000000000D+00,   1.581432499856176D-06, &
   0.000000000000000D+00,  -4.165442649760865D-05,   0.000000000000000D+00, &
  -2.316595261052789D-05,   0.000000000000000D+00,   7.669385787444099D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.578841119694538D-05, &
   0.000000000000000D+00,  -9.103785268140318D-08,   0.000000000000000D+00, &
  -6.068321025435817D-06,   0.000000000000000D+00,  -2.020049699014824D-06, &
   0.000000000000000D+00,   2.449385906457687D-05,   0.000000000000000D+00, &
   9.804147880233662D-05,   0.000000000000000D+00,   7.436333483346491D-06, &
   0.000000000000000D+00,  -3.046510866662228D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.136223075538826D-04,   0.000000000000000D+00, &
  -1.086098761948820D-04,   0.000000000000000D+00,   4.753860941041026D-05, &
   0.000000000000000D+00,   7.657075724525855D-06,   0.000000000000000D+00, &
   1.097465535354131D-06,   0.000000000000000D+00,   2.785309057952043D-05, &
   0.000000000000000D+00,  -8.910337100729531D-05,   0.000000000000000D+00, &
  -1.557504773501397D-04,   0.000000000000000D+00,   7.049783355227438D-07, &
   0.000000000000000D+00,   2.257630499169065D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.239158225615447D-05,   0.000000000000000D+00, &
   2.438699622730050D-04,   0.000000000000000D+00,   1.169975171533444D-04, &
   0.000000000000000D+00,  -6.658119396569137D-05,   0.000000000000000D+00, &
  -8.422652731974674D-06,   0.000000000000000D+00,  -1.784406451212481D-06, &
   0.000000000000000D+00,  -6.111836504641270D-05,   0.000000000000000D+00, &
  -9.843343070200173D-05,   0.000000000000000D+00,   1.752828703536348D-04, &
   0.000000000000000D+00,   1.710326377937255D-04,   0.000000000000000D+00, &
   9.199860760341536D-06,   0.000000000000000D+00,   3.010642327825726D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.411008151322504D-05, &
   0.000000000000000D+00,  -1.297720878952873D-04,   0.000000000000000D+00, &
  -3.839724566945816D-04,   0.000000000000000D+00,  -1.117548701367608D-04, &
   0.000000000000000D+00,   3.782454156390419D-05,   0.000000000000000D+00, &
   2.365677391793175D-06,   0.000000000000000D+00,  -1.425381210101359D-06, &
   0.000000000000000D+00,   4.557153833568638D-05,   0.000000000000000D+00, &
   1.334906468598344D-04,   0.000000000000000D+00,   1.909570075821797D-04, &
   0.000000000000000D+00,  -2.027933376166766D-04,   0.000000000000000D+00, &
  -1.296374993924570D-04,   0.000000000000000D+00,  -1.723081853058296D-05, &
   0.000000000000000D+00,  -4.733171583772289D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.514999639708307D-06,   0.000000000000000D+00, &
  -6.506155678297233D-06,   0.000000000000000D+00,   1.367072759655286D-04, &
   0.000000000000000D+00,   4.202892137703972D-04,   0.000000000000000D+00, &
   1.007939855441581D-04,   0.000000000000000D+00,   4.676926113452033D-06, &
   0.000000000000000D+00,   6.915090980305132D-06,   0.000000000000000D+00, &
   7.011930697992513D-06,   0.000000000000000D+00,  -1.487949189810461D-05, &
   0.000000000000000D+00,  -3.606307347411875D-05,   0.000000000000000D+00, &
  -2.024019943449647D-04,   0.000000000000000D+00,  -2.193039395594263D-04, &
   0.000000000000000D+00,   1.399311820285290D-04,   0.000000000000000D+00, &
   6.103964919596444D-05,   0.000000000000000D+00,   8.155115758356218D-06, &
   0.000000000000000D+00,  -1.242973446962883D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.050444485819691D-06,   0.000000000000000D+00, &
   5.148234333703136D-06,   0.000000000000000D+00,   7.438994133078915D-05, &
   0.000000000000000D+00,  -1.278496733597001D-04,   0.000000000000000D+00, &
  -3.116083052342537D-04,   0.000000000000000D+00,  -7.567297100655205D-05, &
   0.000000000000000D+00,  -2.050210225941692D-05,   0.000000000000000D+00, &
  -7.373758047645456D-06,   0.000000000000000D+00,  -3.369272757421360D-06, &
   0.000000000000000D+00,   9.710377112624663D-07,   0.000000000000000D+00, &
   8.174520748766482D-06,   0.000000000000000D+00,   1.631171598803669D-06, &
   0.000000000000000D+00,   2.157381838903726D-04,   0.000000000000000D+00, &
   1.509857365983602D-04,   0.000000000000000D+00,  -5.231710116300110D-05, &
   0.000000000000000D+00,  -1.006703542384805D-05,   0.000000000000000D+00, &
   4.445600595149978D-06,   0.000000000000000D+00,   4.546205072345180D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.291637172020151D-06, &
   0.000000000000000D+00,  -4.564521124029178D-07,   0.000000000000000D+00, &
  -1.756440047108388D-06,   0.000000000000000D+00,  -9.876410386104340D-05, &
   0.000000000000000D+00,   1.136062778153181D-04,   0.000000000000000D+00, &
   1.548650903514806D-04,   0.000000000000000D+00,   4.025821398976666D-05, &
   0.000000000000000D+00,   1.113389311643295D-05,   0.000000000000000D+00, &
   6.920792209721270D-07,   0.000000000000000D+00,  -5.398763858920268D-06, &
   0.000000000000000D+00,   3.645825171036840D-07,   0.000000000000000D+00, &
  -2.536969917141784D-06,   0.000000000000000D+00,  -2.218174423166447D-07, &
   0.000000000000000D+00,   1.470209445117988D-05,   0.000000000000000D+00, &
  -1.609615542040215D-04,   0.000000000000000D+00,  -5.682993993194825D-05, &
   0.000000000000000D+00,   5.371379837847345D-06,   0.000000000000000D+00, &
  -6.953469321860976D-06,   0.000000000000000D+00,  -6.425846651164009D-06, &
   0.000000000000000D+00,  -1.342223222125663D-06,   0.000000000000000D+00 /)
lem(:,198) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.233645919212021D-04,   5.913413891652684D-05,  -3.967794944737855D-04, &
  -4.029766673004104D-04,  -8.019251664438530D-04,  -7.749044546155490D-05, &
   2.232458433513773D-03,   3.932817863906550D-03,  -1.216754490445343D-03, &
  -9.660056215148155D-03,  -5.468497526364854D-04,   1.245354292665564D-02, &
  -1.952280911525050D-03,  -8.902587893314273D-03,   5.343069704768904D-03, &
   2.248710582801701D-03,  -4.097858684234824D-03,   1.230105772288011D-03, &
   1.176519399477716D-03,  -1.126986187085010D-03,  -4.364011461193080D-05, &
   4.748536897178367D-04,   2.233645919212024D-04,   5.913413891653049D-05, &
  -3.967794944737838D-04,  -4.029766673004134D-04,  -8.019251664438532D-04, &
  -7.749044546154826D-05,   2.232458433513774D-03,   3.932817863906548D-03, &
  -1.216754490445345D-03,  -9.660056215148146D-03,  -5.468497526364811D-04, &
   1.245354292665562D-02,  -1.952280911525060D-03,  -8.902587893314272D-03, &
   5.343069704768907D-03,   2.248710582801695D-03,  -4.097858684234835D-03, &
   1.230105772288012D-03,   1.176519399477718D-03,  -1.126986187085008D-03, &
  -4.364011461193319D-05,   4.748536897178341D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.914824971603270D-04,   0.000000000000000D+00, &
   2.023309059737866D-04,   0.000000000000000D+00,  -6.011674862214575D-06, &
   0.000000000000000D+00,  -8.272241673645235D-05,   0.000000000000000D+00, &
  -1.441808149228616D-04,   0.000000000000000D+00,   4.029657653432801D-06, &
   0.000000000000000D+00,   2.013208529492142D-04,   0.000000000000000D+00, &
   2.259346893872765D-05,   0.000000000000000D+00,  -1.524243082322084D-04, &
   0.000000000000000D+00,   4.940829752852388D-06,   0.000000000000000D+00, &
  -1.274323908805324D-05,  -1.547447972564076D-05,   0.000000000000000D+00, &
   2.217485809553218D-05,   0.000000000000000D+00,  -8.565773766707235D-06, &
   3.814144883455196D-05,   0.000000000000000D+00,   5.869689335404353D-06, &
   0.000000000000000D+00,  -3.587181761923502D-05,   0.000000000000000D+00, &
   7.382336836850024D-06,   0.000000000000000D+00,   8.369628064326481D-06, &
  -6.275532820050083D-05,   0.000000000000000D+00,  -2.482180131768104D-05, &
   0.000000000000000D+00,   4.374027847677353D-06,   0.000000000000000D+00, &
   2.091462234987620D-05,   0.000000000000000D+00,  -1.188133786760145D-05, &
   0.000000000000000D+00,  -1.896314196111657D-05,   0.000000000000000D+00, &
   3.071999269301056D-06,   7.827377842689436D-05,   0.000000000000000D+00, &
   6.367846110246040D-05,   0.000000000000000D+00,   1.612959127692354D-06, &
   0.000000000000000D+00,  -1.123764359488984D-05,   0.000000000000000D+00, &
   2.340479427376219D-05,   0.000000000000000D+00,   7.706689692845051D-06, &
   0.000000000000000D+00,   7.091814565215422D-05,   0.000000000000000D+00, &
   1.142531864432104D-05,   0.000000000000000D+00,  -8.230297663467074D-06, &
  -8.246666567926778D-05,   0.000000000000000D+00,  -9.389315494433712D-05, &
   0.000000000000000D+00,  -8.903295867909655D-05,   0.000000000000000D+00, &
   3.804465924695407D-05,   0.000000000000000D+00,   1.240510467275273D-05, &
   0.000000000000000D+00,  -4.946722120287540D-05,   0.000000000000000D+00, &
   9.063996074708437D-06,   0.000000000000000D+00,  -6.496643606240453D-05, &
   0.000000000000000D+00,  -1.289374588206117D-04,   0.000000000000000D+00, &
  -5.703524293570526D-06,   0.000000000000000D+00,   3.861592932462988D-06, &
   6.994323507771175D-05,   0.000000000000000D+00,   7.898784475170752D-05, &
   0.000000000000000D+00,   1.949633221475310D-04,   0.000000000000000D+00, &
   1.008841124751131D-04,   0.000000000000000D+00,  -5.477862310154228D-05, &
   0.000000000000000D+00,  -1.117854527884966D-05,   0.000000000000000D+00, &
   3.377292092407632D-05,   0.000000000000000D+00,  -1.846881704856494D-05, &
   0.000000000000000D+00,  -7.180460344408148D-05,   0.000000000000000D+00, &
   1.417768156137390D-04,   0.000000000000000D+00,   1.486029744436217D-04, &
   0.000000000000000D+00,   1.496834962015458D-05,   0.000000000000000D+00, &
   7.162619704812772D-06,  -4.103177322459683D-05,   0.000000000000000D+00, &
  -3.439842876744587D-05,   0.000000000000000D+00,  -1.052900452272084D-04, &
   0.000000000000000D+00,  -3.183634911835009D-04,   0.000000000000000D+00, &
  -1.004725447922547D-04,   0.000000000000000D+00,   2.929139309095698D-05, &
   0.000000000000000D+00,   2.195025812472930D-06,   0.000000000000000D+00, &
  -8.213077627226769D-06,   0.000000000000000D+00,   9.857453178522461D-06, &
   0.000000000000000D+00,   9.469929493674874D-05,   0.000000000000000D+00, &
   1.537473234836114D-04,   0.000000000000000D+00,  -1.704378482022359D-04, &
   0.000000000000000D+00,  -1.157844890036444D-04,   0.000000000000000D+00, &
  -2.026258996734342D-05,   0.000000000000000D+00,  -1.076737936410185D-05, &
   1.382438260908587D-05,   0.000000000000000D+00,   4.681067431948023D-06, &
   0.000000000000000D+00,  -6.318256940566725D-06,   0.000000000000000D+00, &
   1.153214701531801D-04,   0.000000000000000D+00,   3.584320753361778D-04, &
   0.000000000000000D+00,   9.318643739626760D-05,   0.000000000000000D+00, &
   8.003938897478693D-06,   0.000000000000000D+00,   8.983157866659187D-06, &
   0.000000000000000D+00,  -1.091801420461469D-06,   0.000000000000000D+00, &
   7.432271201207553D-07,   0.000000000000000D+00,  -2.030290672609924D-05, &
   0.000000000000000D+00,  -1.630246777852102D-04,   0.000000000000000D+00, &
  -1.832301472272907D-04,   0.000000000000000D+00,   1.195751246510811D-04, &
   0.000000000000000D+00,   5.542887308077155D-05,   0.000000000000000D+00, &
   8.384686285921724D-06,   0.000000000000000D+00,  -1.993533305036209D-06, &
  -1.819964334727542D-06,   0.000000000000000D+00,   7.762890480210459D-07, &
   0.000000000000000D+00,   9.682269168032223D-06,   0.000000000000000D+00, &
   6.168660606223382D-05,   0.000000000000000D+00,  -1.114615568087691D-04, &
   0.000000000000000D+00,  -2.711303393648741D-04,   0.000000000000000D+00, &
  -7.055712236691485D-05,   0.000000000000000D+00,  -2.071254036274397D-05, &
   0.000000000000000D+00,  -7.664976769963676D-06,   0.000000000000000D+00, &
   6.347412532100801D-07,   0.000000000000000D+00,  -2.268060205638390D-06, &
   0.000000000000000D+00,   4.038983872011482D-07,   0.000000000000000D+00, &
   2.110120852615451D-06,   0.000000000000000D+00,   1.815684650774082D-04, &
   0.000000000000000D+00,   1.284719113879243D-04,   0.000000000000000D+00, &
  -4.495738013380897D-05,   0.000000000000000D+00,  -8.614830962401707D-06, &
   0.000000000000000D+00,   5.122752464641401D-06,   0.000000000000000D+00, &
   1.253609524577096D-05,  -8.183999605525956D-08,   0.000000000000000D+00, &
   3.928288051624841D-07,   0.000000000000000D+00,  -3.860843486881278D-06, &
   0.000000000000000D+00,  -7.407698447952168D-06,   0.000000000000000D+00, &
  -8.242878371215190D-05,   0.000000000000000D+00,   1.008419527919584D-04, &
   0.000000000000000D+00,   1.369703617925872D-04,   0.000000000000000D+00, &
   3.698971811752411D-05,   0.000000000000000D+00,   1.072908509648443D-05, &
   0.000000000000000D+00,  -6.274437517738432D-07,   0.000000000000000D+00, &
   3.795313063377501D-09,   0.000000000000000D+00,   4.574592715689053D-07, &
   0.000000000000000D+00,   1.131112521114982D-06,   0.000000000000000D+00, &
  -3.993397184553721D-07,   0.000000000000000D+00,   9.076915486275873D-06, &
   0.000000000000000D+00,  -1.383779380916542D-04,   0.000000000000000D+00, &
  -4.844936342285760D-05,   0.000000000000000D+00,   4.518284127460294D-06, &
   0.000000000000000D+00,  -7.294374536404254D-06,   0.000000000000000D+00, &
  -7.369647652996471D-06,   0.000000000000000D+00,  -6.477564558683739D-06 /)
lem(:,199) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.418517530672113D-04,   2.234827226018523D-05, &
   8.751144653712434D-04,   1.331231506104650D-03,   3.359538302579648D-04, &
  -2.066866847891660D-03,  -3.619831637067400D-03,   6.660404009593702D-04, &
   7.099022163700382D-03,   2.122516791050293D-04,  -8.053917848055689D-03, &
   2.366316418933484D-03,   4.352163595990799D-03,  -4.240494513314303D-03, &
   6.864605164906502D-04,   2.109910056300178D-03,  -2.028584400124235D-03, &
   2.757906361284238D-04,   8.305659734699799D-04,  -6.159310436201861D-04, &
  -7.601519249939850D-05,   0.000000000000000D+00,   4.418517530672130D-04, &
  -2.234827226018010D-05,  -8.751144653712474D-04,  -1.331231506104653D-03, &
  -3.359538302579545D-04,   2.066866847891658D-03,   3.619831637067387D-03, &
  -6.660404009593723D-04,  -7.099022163700377D-03,  -2.122516791050281D-04, &
   8.053917848055705D-03,  -2.366316418933490D-03,  -4.352163595990802D-03, &
   4.240494513314302D-03,  -6.864605164906506D-04,  -2.109910056300182D-03, &
   2.028584400124242D-03,  -2.757906361284286D-04,  -8.305659734699839D-04, &
   6.159310436201987D-04,   7.601519249939804D-05,   0.000000000000000D+00, &
   2.209258765336081D-04,   0.000000000000000D+00,   1.979234565121732D-04, &
   0.000000000000000D+00,  -3.741333723280348D-04,   0.000000000000000D+00, &
  -2.221393255742278D-04,   0.000000000000000D+00,   3.207950111988033D-04, &
   0.000000000000000D+00,   2.757359058423789D-04,   0.000000000000000D+00, &
  -2.969089339902595D-04,   0.000000000000000D+00,  -2.374052351565895D-04, &
   0.000000000000000D+00,   2.338427943394171D-04,   0.000000000000000D+00, &
   8.153960579027803D-05,   0.000000000000000D+00,  -9.161327869510441D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.746572409788354D-05, &
   0.000000000000000D+00,   2.629171784677265D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.761781182484671D-05,   0.000000000000000D+00, &
   1.076491941971605D-05,   0.000000000000000D+00,  -6.006163673750178D-06, &
   0.000000000000000D+00,   8.087552183378067D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.100464114825061D-05,   0.000000000000000D+00, &
  -2.201196139883419D-05,   0.000000000000000D+00,   8.779144147348675D-06, &
   0.000000000000000D+00,   4.000345540887668D-06,   0.000000000000000D+00, &
  -1.284173575484446D-05,   0.000000000000000D+00,  -1.997572496669679D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.234889401104558D-05, &
   0.000000000000000D+00,   4.634350746951065D-05,   0.000000000000000D+00, &
   5.778377507307184D-06,   0.000000000000000D+00,  -2.328430570393959D-05, &
   0.000000000000000D+00,   4.143645459053564D-06,   0.000000000000000D+00, &
   7.826033066620664D-06,   0.000000000000000D+00,   6.003874345141958D-05, &
   0.000000000000000D+00,   1.970136616865214D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.696631946073358D-05,   0.000000000000000D+00, &
  -6.447046829908127D-05,   0.000000000000000D+00,  -7.516414684106749D-05, &
   0.000000000000000D+00,   2.783060220980730D-05,   0.000000000000000D+00, &
   2.870332567018804D-05,   0.000000000000000D+00,  -5.537787544269999D-06, &
   0.000000000000000D+00,   9.110223392120289D-06,   0.000000000000000D+00, &
  -5.301361444670677D-05,   0.000000000000000D+00,  -1.152818782428697D-04, &
   0.000000000000000D+00,  -1.804848870318966D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.811488472489452D-05,   0.000000000000000D+00, &
   5.657290722128428D-05,   0.000000000000000D+00,   1.567994736842224D-04, &
   0.000000000000000D+00,   9.273605064097729D-05,   0.000000000000000D+00, &
  -4.494813008399947D-05,   0.000000000000000D+00,  -2.701078374656345D-05, &
   0.000000000000000D+00,  -2.862080897808906D-06,   0.000000000000000D+00, &
  -2.075527253173003D-05,   0.000000000000000D+00,  -5.833797489541550D-05, &
   0.000000000000000D+00,   1.202246033850720D-04,   0.000000000000000D+00, &
   1.396442114964970D-04,   0.000000000000000D+00,   2.615487253521705D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.760108717510404D-05, &
   0.000000000000000D+00,  -2.994639012669677D-05,   0.000000000000000D+00, &
  -8.665206406552048D-05,   0.000000000000000D+00,  -2.715775311336025D-04, &
   0.000000000000000D+00,  -9.754637534987110D-05,   0.000000000000000D+00, &
   2.289199204628302D-05,   0.000000000000000D+00,   7.701488606968995D-06, &
   0.000000000000000D+00,   7.511931919573498D-06,   0.000000000000000D+00, &
   1.571734630122341D-05,   0.000000000000000D+00,   7.898178639000583D-05, &
   0.000000000000000D+00,   1.297938000679667D-04,   0.000000000000000D+00, &
  -1.493911885255945D-04,   0.000000000000000D+00,  -1.118897334456108D-04, &
   0.000000000000000D+00,  -2.573302629502149D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.661027417697516D-06,   0.000000000000000D+00, &
   9.678224465488051D-06,   0.000000000000000D+00,  -1.779603948386356D-07, &
   0.000000000000000D+00,   1.007939855441581D-04,   0.000000000000000D+00, &
   3.164208212033713D-04,   0.000000000000000D+00,   9.169720756063273D-05, &
   0.000000000000000D+00,   1.199038455677183D-05,   0.000000000000000D+00, &
   1.969922764373222D-05,   0.000000000000000D+00,  -4.158970546759652D-06, &
   0.000000000000000D+00,  -5.633909754072525D-06,   0.000000000000000D+00, &
  -1.903565104710568D-05,   0.000000000000000D+00,  -1.392306025414869D-04, &
   0.000000000000000D+00,  -1.589607937749792D-04,   0.000000000000000D+00, &
   1.072748116923373D-04,   0.000000000000000D+00,   5.349361651941626D-05, &
   0.000000000000000D+00,   6.787987023434719D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.762524742792917D-07,   0.000000000000000D+00, &
  -2.838927330313969D-06,   0.000000000000000D+00,   3.170758372057155D-06, &
   0.000000000000000D+00,   4.779361312832274D-05,   0.000000000000000D+00, &
  -1.012433419414817D-04,   0.000000000000000D+00,  -2.445222861410668D-04, &
   0.000000000000000D+00,  -6.903063431367236D-05,   0.000000000000000D+00, &
  -2.326879554465273D-05,   0.000000000000000D+00,  -2.254324595184748D-05, &
   0.000000000000000D+00,   6.492027390535483D-07,   0.000000000000000D+00, &
   1.346931805822303D-06,   0.000000000000000D+00,   3.185354791399892D-07, &
   0.000000000000000D+00,   4.023780662664385D-06,   0.000000000000000D+00, &
   1.591083911934501D-04,   0.000000000000000D+00,   1.127727838228724D-04, &
   0.000000000000000D+00,  -4.051566482866816D-05,   0.000000000000000D+00, &
  -7.039446530166921D-06,   0.000000000000000D+00,   7.679160944955619D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.084808471006274D-07, &
   0.000000000000000D+00,   1.022629363557703D-06,   0.000000000000000D+00, &
   6.296477219984623D-07,   0.000000000000000D+00,  -5.413894140092218D-06, &
   0.000000000000000D+00,  -6.763579641807812D-05,   0.000000000000000D+00, &
   9.346814735262552D-05,   0.000000000000000D+00,   1.248455363751998D-04, &
   0.000000000000000D+00,   3.600375888608327D-05,   0.000000000000000D+00, &
   1.281936618993906D-05,   0.000000000000000D+00,   1.766791407699267D-06, &
   0.000000000000000D+00,   6.196843851974008D-08,   0.000000000000000D+00, &
  -5.739163382396233D-07,   0.000000000000000D+00,   1.259176191226957D-06, &
   0.000000000000000D+00,   1.770810919210860D-06,   0.000000000000000D+00, &
   5.013815257614232D-06,   0.000000000000000D+00,  -1.232054110064129D-04, &
   0.000000000000000D+00,  -4.219416124876755D-05,   0.000000000000000D+00, &
   3.442933062564022D-06,   0.000000000000000D+00,  -8.438891727691801D-06, &
   0.000000000000000D+00,  -5.605478381047594D-06,   0.000000000000000D+00 /)
lem(:,200) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.974415095249315D-04,   6.189404619830340D-05,  -3.831899304570329D-04, &
  -7.976732953530759D-04,  -1.038855202648405D-03,  -4.037443736877909D-04, &
   1.458120891590642D-03,   2.976449698440771D-03,   3.364123682867097D-04, &
  -4.695331117160342D-03,  -8.504155226279594D-04,   4.484870390465530D-03, &
  -1.551708514291985D-03,  -1.580004085919714D-03,   2.340225560004911D-03, &
  -1.390813144190906D-03,  -1.989739166686104D-04,   1.259931815668552D-03, &
  -1.010072641211756D-03,   4.956156187507279D-05,   5.145411364494075D-04, &
  -3.618142678337893D-04,   1.974415095249381D-04,   6.189404619830405D-05, &
  -3.831899304570326D-04,  -7.976732953530748D-04,  -1.038855202648401D-03, &
  -4.037443736877910D-04,   1.458120891590637D-03,   2.976449698440770D-03, &
   3.364123682867110D-04,  -4.695331117160345D-03,  -8.504155226279530D-04, &
   4.484870390465524D-03,  -1.551708514291987D-03,  -1.580004085919715D-03, &
   2.340225560004920D-03,  -1.390813144190908D-03,  -1.989739166686129D-04, &
   1.259931815668557D-03,  -1.010072641211755D-03,   4.956156187506767D-05, &
   5.145411364494098D-04,  -3.618142678337897D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.209901995833740D-04,   0.000000000000000D+00, &
  -7.203156283567363D-04,   0.000000000000000D+00,   2.422992288306665D-04, &
   0.000000000000000D+00,   1.156153221233867D-03,   0.000000000000000D+00, &
  -1.306505152360607D-03,   0.000000000000000D+00,   5.713594942773833D-06, &
   0.000000000000000D+00,   7.854353487107504D-04,   0.000000000000000D+00, &
  -2.077489549059392D-04,   0.000000000000000D+00,  -3.083692509212223D-04, &
   0.000000000000000D+00,   1.354708080550339D-04,   0.000000000000000D+00, &
  -1.111214301157224D-05,  -2.167048020134302D-06,   0.000000000000000D+00, &
   1.410237723508557D-05,   0.000000000000000D+00,   1.158931621214146D-05, &
   1.133008114069452D-05,   0.000000000000000D+00,   6.989365770725115D-06, &
   0.000000000000000D+00,  -8.529282459204400D-06,   0.000000000000000D+00, &
  -9.217905255340864D-06,   0.000000000000000D+00,   1.353712780233445D-05, &
  -1.932128906417812D-05,   0.000000000000000D+00,  -1.432694855561531D-05, &
   0.000000000000000D+00,  -1.985055850401033D-05,   0.000000000000000D+00, &
  -1.057120114535510D-05,   0.000000000000000D+00,   4.789186520776777D-06, &
   0.000000000000000D+00,  -7.407803242058981D-06,   0.000000000000000D+00, &
  -4.589912767437336D-05,   1.416440472446688D-05,   0.000000000000000D+00, &
   1.985815004396117D-05,   0.000000000000000D+00,   3.824567232560691D-05, &
   0.000000000000000D+00,   1.553618420077929D-05,   0.000000000000000D+00, &
   2.854539917865650D-05,   0.000000000000000D+00,   6.623072633241841D-06, &
   0.000000000000000D+00,   3.804308282881093D-06,   0.000000000000000D+00, &
   5.405069434227579D-05,   0.000000000000000D+00,   4.902111769061729D-05, &
  -3.109218989151846D-06,   0.000000000000000D+00,  -2.089228083703511D-05, &
   0.000000000000000D+00,  -5.240420937581813D-05,   0.000000000000000D+00, &
  -7.138852836490812D-05,   0.000000000000000D+00,   1.731295007768328D-05, &
   0.000000000000000D+00,  -3.172112313191212D-05,   0.000000000000000D+00, &
  -1.254475633139205D-05,   0.000000000000000D+00,   5.061225751834713D-06, &
   0.000000000000000D+00,  -4.271901324738573D-05,   0.000000000000000D+00, &
  -1.141619893612672D-04,   0.000000000000000D+00,  -4.414750080745376D-05, &
   2.783506729007653D-07,   0.000000000000000D+00,   1.611803705885068D-05, &
   0.000000000000000D+00,   4.580970062357324D-05,   0.000000000000000D+00, &
   1.356422180908787D-04,   0.000000000000000D+00,   9.384322032602456D-05, &
   0.000000000000000D+00,  -3.835299724333374D-05,   0.000000000000000D+00, &
   2.013266240761629D-05,   0.000000000000000D+00,   7.075280154905629D-06, &
   0.000000000000000D+00,  -1.279568776237767D-05,   0.000000000000000D+00, &
  -4.818482794943218D-05,   0.000000000000000D+00,   1.071673989955909D-04, &
   0.000000000000000D+00,   1.441051210714556D-04,   0.000000000000000D+00, &
   5.992797053510559D-05,  -2.588069150894415D-06,   0.000000000000000D+00, &
  -6.784953762064741D-06,   0.000000000000000D+00,  -2.238674832294909D-05, &
   0.000000000000000D+00,  -7.542394732804663D-05,   0.000000000000000D+00, &
  -2.457323956933252D-04,   0.000000000000000D+00,  -1.017207370202181D-04, &
   0.000000000000000D+00,   1.384329840576913D-05,   0.000000000000000D+00, &
  -6.679193610931725D-06,   0.000000000000000D+00,  -8.937390645205385D-07, &
   0.000000000000000D+00,   1.122130008654056D-05,   0.000000000000000D+00, &
   6.526974204569158D-05,   0.000000000000000D+00,   1.130907064849989D-04, &
   0.000000000000000D+00,  -1.384961963169884D-04,   0.000000000000000D+00, &
  -1.172932108680935D-04,   0.000000000000000D+00,  -6.387773181459505D-05, &
   2.189596597197420D-06,   0.000000000000000D+00,  -3.667706302785146D-08, &
   0.000000000000000D+00,   4.036206617603433D-06,   0.000000000000000D+00, &
   9.019414063792555D-07,   0.000000000000000D+00,   9.318643739626760D-05, &
   0.000000000000000D+00,   2.938982709804387D-04,   0.000000000000000D+00, &
   9.695934498609986D-05,   0.000000000000000D+00,   2.342520746211112D-05, &
   0.000000000000000D+00,   6.616216910373164D-07,   0.000000000000000D+00, &
   5.366637234998707D-08,   0.000000000000000D+00,  -5.108404376211451D-06, &
   0.000000000000000D+00,  -1.784216594962091D-05,   0.000000000000000D+00, &
  -1.223973133559024D-04,   0.000000000000000D+00,  -1.418650160761575D-04, &
   0.000000000000000D+00,   1.006891004333143D-04,   0.000000000000000D+00, &
   5.668744131689976D-05,   0.000000000000000D+00,   2.468665347214100D-05, &
  -5.034161338132666D-07,   0.000000000000000D+00,   1.045492410363692D-06, &
   0.000000000000000D+00,   1.273515698322968D-06,   0.000000000000000D+00, &
   4.267808976936278D-06,   0.000000000000000D+00,   3.905782588543807D-05, &
   0.000000000000000D+00,  -9.707356121825145D-05,   0.000000000000000D+00, &
  -2.311041965836646D-04,   0.000000000000000D+00,  -7.362552642819425D-05, &
   0.000000000000000D+00,  -2.729337951556756D-05,   0.000000000000000D+00, &
   1.132243169586306D-07,   0.000000000000000D+00,  -4.760697831453564D-07, &
   0.000000000000000D+00,   1.505923425204227D-06,   0.000000000000000D+00, &
   2.422370842818462D-06,   0.000000000000000D+00,   7.235591562974224D-06, &
   0.000000000000000D+00,   1.438609038767604D-04,   0.000000000000000D+00, &
   1.021311275934576D-04,   0.000000000000000D+00,  -3.748305755490186D-05, &
   0.000000000000000D+00,  -8.449711003632670D-06,   0.000000000000000D+00, &
   1.945057958723276D-05,  -1.149442368910178D-08,   0.000000000000000D+00, &
  -1.505170576513560D-07,   0.000000000000000D+00,  -1.037233556897384D-06, &
   0.000000000000000D+00,  -6.454060141888031D-07,   0.000000000000000D+00, &
  -6.345321877593150D-06,   0.000000000000000D+00,  -5.697790546083242D-05, &
   0.000000000000000D+00,   9.028083549276156D-05,   0.000000000000000D+00, &
   1.197652615780521D-04,   0.000000000000000D+00,   3.838097309855116D-05, &
   0.000000000000000D+00,   8.586159227166802D-06,   0.000000000000000D+00, &
   4.500630866835223D-10,   0.000000000000000D+00,   1.657882080465162D-07, &
   0.000000000000000D+00,  -3.293777414432115D-07,   0.000000000000000D+00, &
  -7.750462378506711D-07,   0.000000000000000D+00,   4.953803800254770D-07, &
   0.000000000000000D+00,   7.204234850545261D-08,   0.000000000000000D+00, &
  -1.129511153117156D-04,   0.000000000000000D+00,  -3.839383760922465D-05, &
   0.000000000000000D+00,   2.585568026060686D-06,   0.000000000000000D+00, &
  -9.568315135375556D-06,   0.000000000000000D+00,  -2.564767711562584D-05 /)
lem(:,201) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.156811400164473D-04,   7.447650414324941D-05, &
   2.999385670612456D-04,   9.530745395246461D-04,   7.917174396983856D-04, &
  -4.541038825968726D-04,  -2.110964980240399D-03,  -1.239085191032241D-03, &
   2.716117518026922D-03,   1.313567489025694D-03,  -2.388649723600898D-03, &
   5.397369329048398D-04,   4.783578433295332D-04,  -8.690620246932798D-04, &
   9.371117533176882D-04,  -4.928231540085020D-04,  -2.233490350556952D-04, &
   6.805521188327464D-04,  -5.257232449735384D-04,   2.075450095625350D-05, &
   2.957965610373468D-04,   0.000000000000000D+00,  -2.156811400164449D-04, &
  -7.447650414325250D-05,  -2.999385670612490D-04,  -9.530745395246404D-04, &
  -7.917174396983792D-04,   4.541038825968673D-04,   2.110964980240399D-03, &
   1.239085191032240D-03,  -2.716117518026920D-03,  -1.313567489025691D-03, &
   2.388649723600899D-03,  -5.397369329048396D-04,  -4.783578433295312D-04, &
   8.690620246932792D-04,  -9.371117533176873D-04,   4.928231540085030D-04, &
   2.233490350556902D-04,  -6.805521188327514D-04,   5.257232449735436D-04, &
  -2.075450095624752D-05,  -2.957965610373484D-04,   0.000000000000000D+00, &
  -1.078405700082216D-04,   0.000000000000000D+00,  -5.815388547485241D-04, &
   0.000000000000000D+00,   6.754496357801892D-04,   0.000000000000000D+00, &
   1.323743664689196D-03,   0.000000000000000D+00,  -3.530194919262532D-03, &
   0.000000000000000D+00,   3.692860020447049D-03,   0.000000000000000D+00, &
  -1.730781211032026D-03,   0.000000000000000D+00,  -4.546504832646129D-04, &
   0.000000000000000D+00,   8.011261519163142D-04,   0.000000000000000D+00, &
  -9.842159028282533D-05,   0.000000000000000D+00,  -1.259222958032246D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.525545634920661D-06, &
   0.000000000000000D+00,   2.516650961775233D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.056923759813258D-06,   0.000000000000000D+00, &
   1.817308552842437D-05,   0.000000000000000D+00,  -2.177083166175167D-07, &
   0.000000000000000D+00,  -6.849411414450066D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.388703572053885D-06,   0.000000000000000D+00, &
  -1.233199435293841D-05,   0.000000000000000D+00,  -5.374344522660166D-05, &
   0.000000000000000D+00,  -3.782024974574862D-06,   0.000000000000000D+00, &
   4.374296691829702D-06,   0.000000000000000D+00,  -6.782767429584940D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.496458501928976D-06, &
   0.000000000000000D+00,   1.279651692229301D-05,   0.000000000000000D+00, &
   3.975083355703616D-05,   0.000000000000000D+00,   3.552991880743293D-05, &
   0.000000000000000D+00,   7.937917155005295D-06,   0.000000000000000D+00, &
   3.678377981877865D-06,   0.000000000000000D+00,   4.782266032711278D-07, &
   0.000000000000000D+00,   4.904904608788696D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.598103956431379D-06,   0.000000000000000D+00, &
  -1.456160442621458D-05,   0.000000000000000D+00,  -4.507503528330680D-05, &
   0.000000000000000D+00,  -7.387003104790293D-05,   0.000000000000000D+00, &
   4.452394761033676D-05,   0.000000000000000D+00,  -9.475991755032815D-06, &
   0.000000000000000D+00,  -8.142608059482852D-06,   0.000000000000000D+00, &
   5.075437303649700D-06,   0.000000000000000D+00,  -3.865010931643694D-05, &
   0.000000000000000D+00,  -1.291823804800371D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.997193091161098D-06,   0.000000000000000D+00, &
   1.456643704368719D-05,   0.000000000000000D+00,   3.931043974709304D-05, &
   0.000000000000000D+00,   1.255347380262549D-04,   0.000000000000000D+00, &
   1.020742912710588D-04,   0.000000000000000D+00,  -9.637034759287811D-05, &
   0.000000000000000D+00,   5.735275404238465D-06,   0.000000000000000D+00, &
   5.197845992934111D-06,   0.000000000000000D+00,  -1.100532852145089D-05, &
   0.000000000000000D+00,  -4.187999851068140D-05,   0.000000000000000D+00, &
   1.044138018219134D-04,   0.000000000000000D+00,   1.755169364483818D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.475972917715155D-06, &
   0.000000000000000D+00,  -9.557266042594554D-06,   0.000000000000000D+00, &
  -2.106837605139942D-05,   0.000000000000000D+00,  -7.096979305170024D-05, &
   0.000000000000000D+00,  -2.366641165964755D-04,   0.000000000000000D+00, &
  -1.159173811272673D-04,   0.000000000000000D+00,   4.737019835575439D-05, &
   0.000000000000000D+00,  -8.222650657462417D-07,   0.000000000000000D+00, &
  -1.369316108065286D-06,   0.000000000000000D+00,   9.084673666498285D-06, &
   0.000000000000000D+00,   5.788436421529511D-05,   0.000000000000000D+00, &
   1.032273287523155D-04,   0.000000000000000D+00,  -1.387881635936623D-04, &
   0.000000000000000D+00,  -1.391325209535254D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.370040088348677D-06,   0.000000000000000D+00, &
   3.606264139699075D-06,   0.000000000000000D+00,   5.418243790040906D-06, &
   0.000000000000000D+00,   4.676926113452033D-06,   0.000000000000000D+00, &
   9.169720756063273D-05,   0.000000000000000D+00,   2.912872315892793D-04, &
   0.000000000000000D+00,   1.104709893287914D-04,   0.000000000000000D+00, &
   4.794092383244523D-05,   0.000000000000000D+00,  -6.902446051598407D-07, &
   0.000000000000000D+00,   4.205185804323019D-07,   0.000000000000000D+00, &
  -2.957776898108329D-06,   0.000000000000000D+00,  -1.734087725101093D-05, &
   0.000000000000000D+00,  -1.126154981060877D-04,   0.000000000000000D+00, &
  -1.340961764683705D-04,   0.000000000000000D+00,   1.025305506862585D-04, &
   0.000000000000000D+00,   5.456775852592888D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.514111225686297D-07,   0.000000000000000D+00, &
  -9.803851072368914D-07,   0.000000000000000D+00,   1.181142102116645D-07, &
   0.000000000000000D+00,   3.365322702005197D-06,   0.000000000000000D+00, &
   3.083350572323683D-05,   0.000000000000000D+00,  -9.816565708540655D-05, &
   0.000000000000000D+00,  -2.338542063449625D-04,   0.000000000000000D+00, &
  -8.336502073824149D-05,   0.000000000000000D+00,  -8.255593604734964D-05, &
   0.000000000000000D+00,   2.238397734849389D-07,   0.000000000000000D+00, &
  -2.569685971962488D-07,   0.000000000000000D+00,  -3.061317349285739D-07, &
   0.000000000000000D+00,   1.666976876086462D-06,   0.000000000000000D+00, &
   1.000869396983364D-05,   0.000000000000000D+00,   1.364921341797712D-04, &
   0.000000000000000D+00,   9.818689031579365D-05,   0.000000000000000D+00, &
  -3.702618111530187D-05,   0.000000000000000D+00,   1.893832798537338D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.678315268483304D-08, &
   0.000000000000000D+00,   3.748128175573917D-07,   0.000000000000000D+00, &
  -3.855743125080315D-07,   0.000000000000000D+00,  -1.153863573004831D-06, &
   0.000000000000000D+00,  -6.889938635423535D-06,   0.000000000000000D+00, &
  -4.892647901745505D-05,   0.000000000000000D+00,   9.262202026339569D-05, &
   0.000000000000000D+00,   1.226799174946055D-04,   0.000000000000000D+00, &
   4.767840723871030D-05,   0.000000000000000D+00,   4.355791357026058D-05, &
   0.000000000000000D+00,   1.033820769856170D-08,   0.000000000000000D+00, &
  -1.853429942042377D-08,   0.000000000000000D+00,   5.975794155157572D-07, &
   0.000000000000000D+00,   2.904767180654759D-07,   0.000000000000000D+00, &
   9.428749674985042D-07,   0.000000000000000D+00,  -4.114923387408449D-06, &
   0.000000000000000D+00,  -1.093347998697084D-04,   0.000000000000000D+00, &
  -3.662205816775151D-05,   0.000000000000000D+00,  -4.769477241893886D-07, &
   0.000000000000000D+00,  -8.474822310805642D-06,   0.000000000000000D+00 /)
lem(:,202) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.366494922185700D-05,   3.325084241920137D-04,   3.033544451946825D-04, &
  -4.997328546820364D-04,  -7.655318638993842D-04,  -6.781915852637902D-04, &
  -3.861681283352032D-04,   1.272508475143008D-03,   1.769953242902585D-03, &
  -1.236589295011250D-03,  -1.449614063623566D-03,   1.189468622215609D-03, &
   1.231417316542442D-05,  -2.737495529240890D-04,   2.200041466423209D-04, &
  -3.452897478157301D-04,   3.916082357513482D-04,  -1.882329354546688D-04, &
  -1.412000112566798D-04,   3.500759245641213D-04,  -2.763512656733791D-04, &
   2.309743504706487D-05,   4.366494922185753D-05,   3.325084241920257D-04, &
   3.033544451946834D-04,  -4.997328546820411D-04,  -7.655318638993821D-04, &
  -6.781915852637816D-04,  -3.861681283352115D-04,   1.272508475143004D-03, &
   1.769953242902594D-03,  -1.236589295011250D-03,  -1.449614063623570D-03, &
   1.189468622215614D-03,   1.231417316542254D-05,  -2.737495529240890D-04, &
   2.200041466423213D-04,  -3.452897478157317D-04,   3.916082357513521D-04, &
  -1.882329354546658D-04,  -1.412000112566837D-04,   3.500759245641203D-04, &
  -2.763512656733751D-04,   2.309743504706444D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.367430114004874D-04,   0.000000000000000D+00, &
   4.321386047659272D-04,   0.000000000000000D+00,   1.043036371456839D-03, &
   0.000000000000000D+00,  -3.825288678076212D-03,   0.000000000000000D+00, &
   5.215271387192906D-03,   0.000000000000000D+00,  -5.232536339396063D-03, &
   0.000000000000000D+00,   4.125008938156947D-03,   0.000000000000000D+00, &
  -1.494373841637266D-03,   0.000000000000000D+00,  -3.560613227227485D-04, &
   0.000000000000000D+00,   3.732521445395117D-04,   0.000000000000000D+00, &
   2.583498677248823D-06,  -7.609577922079429D-07,   0.000000000000000D+00, &
  -4.004422949735328D-06,   0.000000000000000D+00,   1.286523625390813D-05, &
   3.149519751082174D-06,   0.000000000000000D+00,   3.817119295634970D-06, &
   0.000000000000000D+00,   8.463193124996803D-06,   0.000000000000000D+00, &
  -4.585009132781483D-06,   0.000000000000000D+00,  -2.811573121591960D-05, &
  -1.094766583712924D-06,   0.000000000000000D+00,  -4.499366998045788D-06, &
   0.000000000000000D+00,  -6.860634089251167D-06,   0.000000000000000D+00, &
  -1.027975280035415D-05,   0.000000000000000D+00,  -1.654701207111199D-06, &
   0.000000000000000D+00,   8.259504684384209D-06,   0.000000000000000D+00, &
  -1.009387104759381D-05,  -6.464885629571742D-06,   0.000000000000000D+00, &
   2.017819551046838D-06,   0.000000000000000D+00,   1.032028784005301D-05, &
   0.000000000000000D+00,   3.034573042123182D-05,   0.000000000000000D+00, &
   7.897803724690032D-06,   0.000000000000000D+00,   7.288153159952761D-06, &
   0.000000000000000D+00,   2.800393570388281D-06,   0.000000000000000D+00, &
   4.547085538392118D-06,   0.000000000000000D+00,   1.483235811412106D-04, &
   1.115649160959139D-05,   0.000000000000000D+00,   5.362920449022367D-07, &
   0.000000000000000D+00,  -1.217319718386052D-05,   0.000000000000000D+00, &
  -4.193110513455487D-05,   0.000000000000000D+00,  -9.154660920704312D-05, &
   0.000000000000000D+00,  -5.518304872905855D-06,   0.000000000000000D+00, &
  -8.022978270748101D-06,   0.000000000000000D+00,  -7.339942190403773D-06, &
   0.000000000000000D+00,   1.566447577307092D-06,   0.000000000000000D+00, &
  -4.916368955588300D-05,   0.000000000000000D+00,  -3.319393842351644D-04, &
  -8.159534448925587D-06,   0.000000000000000D+00,  -2.055722055727776D-07, &
   0.000000000000000D+00,   1.110944446468067D-05,   0.000000000000000D+00, &
   3.782468359638852D-05,   0.000000000000000D+00,   1.302244697995353D-04, &
   0.000000000000000D+00,   1.640778451980881D-04,   0.000000000000000D+00, &
   3.962066486161473D-06,   0.000000000000000D+00,   3.960380018957528D-06, &
   0.000000000000000D+00,   4.167339492080359D-06,   0.000000000000000D+00, &
  -8.247078084733807D-06,   0.000000000000000D+00,  -3.805929385089781D-05, &
   0.000000000000000D+00,   1.172457630190536D-04,   0.000000000000000D+00, &
   4.348799021008724D-04,   2.603252525729074D-06,   0.000000000000000D+00, &
  -5.815353751745146D-07,   0.000000000000000D+00,  -6.850942187518968D-06, &
   0.000000000000000D+00,  -2.093774125638237D-05,   0.000000000000000D+00, &
  -7.169694232669711D-05,   0.000000000000000D+00,  -2.546830315455607D-04, &
   0.000000000000000D+00,  -2.041097920266436D-04,   0.000000000000000D+00, &
  -1.864354617314255D-06,   0.000000000000000D+00,  -2.439880924282419D-07, &
   0.000000000000000D+00,   1.413504040371420D-07,   0.000000000000000D+00, &
   8.712077792817478D-06,   0.000000000000000D+00,   5.342709978647242D-05, &
   0.000000000000000D+00,   1.029569308917073D-04,   0.000000000000000D+00, &
  -1.539237630364898D-04,   0.000000000000000D+00,  -3.760303487933162D-04, &
  -3.566748042277531D-08,   0.000000000000000D+00,   4.172318332094177D-08, &
   0.000000000000000D+00,   2.455695324041560D-06,   0.000000000000000D+00, &
   5.765668054811901D-06,   0.000000000000000D+00,   8.003938897478693D-06, &
   0.000000000000000D+00,   9.695934498609986D-05,   0.000000000000000D+00, &
   3.203751570942605D-04,   0.000000000000000D+00,   1.854649195015159D-04, &
   0.000000000000000D+00,   3.273416089437095D-07,   0.000000000000000D+00, &
  -4.488983474351306D-07,   0.000000000000000D+00,  -1.148703288457579D-06, &
   0.000000000000000D+00,  -4.080278733580044D-06,   0.000000000000000D+00, &
  -1.793948318034834D-05,   0.000000000000000D+00,  -1.109231568857014D-04, &
   0.000000000000000D+00,  -1.378320512732138D-04,   0.000000000000000D+00, &
   1.208800273984850D-04,   0.000000000000000D+00,   1.940089969106714D-04, &
  -1.191183957916526D-07,   0.000000000000000D+00,   3.147667608348006D-07, &
   0.000000000000000D+00,  -4.606453331005372D-07,   0.000000000000000D+00, &
  -3.293633935231737D-07,   0.000000000000000D+00,   2.599474585631009D-06, &
   0.000000000000000D+00,   2.543362615728205D-05,   0.000000000000000D+00, &
  -1.074049852406195D-04,   0.000000000000000D+00,  -2.637497769643573D-04, &
   0.000000000000000D+00,  -1.131753706718524D-04,   0.000000000000000D+00, &
   1.995784910159913D-08,   0.000000000000000D+00,   2.673015159418132D-08, &
   0.000000000000000D+00,   5.975454011641117D-07,   0.000000000000000D+00, &
   6.766051796175500D-07,   0.000000000000000D+00,   2.283857276114199D-06, &
   0.000000000000000D+00,   1.401267507411623D-05,   0.000000000000000D+00, &
   1.387865510566221D-04,   0.000000000000000D+00,   1.024775950369332D-04, &
   0.000000000000000D+00,  -5.544068160334510D-05,   0.000000000000000D+00, &
  -2.261328474474156D-05,  -1.668685188604350D-09,   0.000000000000000D+00, &
  -9.763469615411920D-08,   0.000000000000000D+00,   3.108121201226960D-08, &
   0.000000000000000D+00,   1.352134672107380D-07,   0.000000000000000D+00, &
  -6.056812719440989D-07,   0.000000000000000D+00,  -8.506027114565057D-06, &
   0.000000000000000D+00,  -4.339476314469161D-05,   0.000000000000000D+00, &
   1.024826521442064D-04,   0.000000000000000D+00,   1.414454881577971D-04, &
   0.000000000000000D+00,   3.742704831058226D-05,   0.000000000000000D+00, &
   5.638045199086480D-11,   0.000000000000000D+00,   4.364035003238920D-08, &
   0.000000000000000D+00,  -2.109735572490951D-07,   0.000000000000000D+00, &
   3.657929763922669D-08,   0.000000000000000D+00,   1.265989012129843D-07, &
   0.000000000000000D+00,   1.386033809608242D-06,   0.000000000000000D+00, &
  -1.013204138995144D-05,   0.000000000000000D+00,  -1.127233584175408D-04, &
   0.000000000000000D+00,  -3.705446292628277D-05,   0.000000000000000D+00, &
   8.706444330939097D-06,   0.000000000000000D+00,  -4.402184232633596D-05 /)
lem(:,203) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.555423303454015D-05,   1.016527388409428D-04, &
  -3.290274775020716D-04,  -3.857939756680879D-04,   9.367421103549182D-04, &
   1.182181920070981D-03,  -6.081892582960189D-04,  -1.723036295641050D-03, &
   2.681392036131636D-04,   1.346080500483719D-03,  -5.392411429818440D-04, &
  -2.584031727994496D-04,   2.388988871405969D-04,  -8.214452676871188D-05, &
   8.495072595916488D-05,  -1.554715260369362D-04,   1.579741137602082D-04, &
  -6.106758125021509D-05,  -8.442565545170811D-05,   1.771556995070461D-04, &
  -1.417421388024587D-04,   0.000000000000000D+00,  -5.555423303453658D-05, &
  -1.016527388409495D-04,   3.290274775020681D-04,   3.857939756680900D-04, &
  -9.367421103549116D-04,  -1.182181920070989D-03,   6.081892582960129D-04, &
   1.723036295641059D-03,  -2.681392036131576D-04,  -1.346080500483724D-03, &
   5.392411429818395D-04,   2.584031727994488D-04,  -2.388988871405942D-04, &
   8.214452676871250D-05,  -8.495072595916559D-05,   1.554715260369384D-04, &
  -1.579741137602114D-04,   6.106758125021078D-05,   8.442565545171179D-05, &
  -1.771556995070404D-04,   1.417421388024553D-04,   0.000000000000000D+00, &
  -2.777711651726771D-05,   0.000000000000000D+00,   1.319786779178017D-05, &
   0.000000000000000D+00,   6.146546767318575D-04,   0.000000000000000D+00, &
  -2.356561582350663D-03,   0.000000000000000D+00,   2.287188400391179D-03, &
   0.000000000000000D+00,   8.030372572687001D-04,   0.000000000000000D+00, &
  -1.176478519744455D-03,   0.000000000000000D+00,  -1.952104693491101D-03, &
   0.000000000000000D+00,   2.621203253097821D-03,   0.000000000000000D+00, &
  -8.624309869164701D-04,   0.000000000000000D+00,   4.038907044360951D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.195973875661310D-06, &
   0.000000000000000D+00,   1.470245610870626D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.312887055074556D-06,   0.000000000000000D+00, &
   7.662422213204214D-07,   0.000000000000000D+00,  -9.253622534872890D-07, &
   0.000000000000000D+00,   1.995752728174508D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.431740712990794D-06,   0.000000000000000D+00, &
  -1.417446485805805D-06,   0.000000000000000D+00,  -3.442195825008331D-05, &
   0.000000000000000D+00,  -9.937678059176303D-07,   0.000000000000000D+00, &
  -4.130853490632541D-06,   0.000000000000000D+00,  -4.276074565866015D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.973543926241999D-06, &
   0.000000000000000D+00,   1.428478203438640D-06,   0.000000000000000D+00, &
   1.860010236869765D-05,   0.000000000000000D+00,   1.320340878681072D-04, &
   0.000000000000000D+00,   3.069317448398294D-06,   0.000000000000000D+00, &
   6.009853796535530D-06,   0.000000000000000D+00,   7.276130071426015D-06, &
   0.000000000000000D+00,  -4.480679696369301D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.996554039424922D-06,   0.000000000000000D+00, &
  -5.919895899732100D-07,   0.000000000000000D+00,  -1.506720250677385D-05, &
   0.000000000000000D+00,  -6.367067066289998D-05,   0.000000000000000D+00, &
  -2.595394910318938D-04,   0.000000000000000D+00,  -4.137735110765244D-06, &
   0.000000000000000D+00,  -6.302111622812761D-06,   0.000000000000000D+00, &
  -8.123001764921347D-06,   0.000000000000000D+00,   5.854727113496462D-06, &
   0.000000000000000D+00,  -1.359363261259290D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.946664349402569D-06,   0.000000000000000D+00, &
   4.838600631681223D-07,   0.000000000000000D+00,   1.227822649087481D-05, &
   0.000000000000000D+00,   4.165561136999599D-05,   0.000000000000000D+00, &
   1.562058946814602D-04,   0.000000000000000D+00,   3.380238945787757D-04, &
   0.000000000000000D+00,   3.299174552981335D-06,   0.000000000000000D+00, &
   4.314589734191615D-06,   0.000000000000000D+00,   4.660356190529302D-06, &
   0.000000000000000D+00,  -8.025590429428634D-06,   0.000000000000000D+00, &
  -5.066258155920595D-05,   0.000000000000000D+00,   5.015606072697769D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.888138450875643D-07, &
   0.000000000000000D+00,  -7.034179377209223D-07,   0.000000000000000D+00, &
  -7.719402444627299D-06,   0.000000000000000D+00,  -2.236973260237829D-05, &
   0.000000000000000D+00,  -7.783641413582491D-05,   0.000000000000000D+00, &
  -2.838072594189467D-04,   0.000000000000000D+00,  -3.693360433409147D-04, &
   0.000000000000000D+00,  -1.280054600234169D-06,   0.000000000000000D+00, &
  -1.653088349774700D-06,   0.000000000000000D+00,  -3.674237260125896D-07, &
   0.000000000000000D+00,   7.698739881631029D-06,   0.000000000000000D+00, &
   5.553292911719418D-05,   0.000000000000000D+00,   1.188853572004685D-04, &
   0.000000000000000D+00,  -8.406613432069192D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.911208508309537D-07,   0.000000000000000D+00, &
   2.869942394090603D-07,   0.000000000000000D+00,   2.584119377085560D-06, &
   0.000000000000000D+00,   6.915090980305132D-06,   0.000000000000000D+00, &
   1.199038455677183D-05,   0.000000000000000D+00,   1.104709893287914D-04, &
   0.000000000000000D+00,   3.553557198180564D-04,   0.000000000000000D+00, &
   3.849451302432146D-04,   0.000000000000000D+00,   6.782877137328790D-08, &
   0.000000000000000D+00,   3.984748893061747D-07,   0.000000000000000D+00, &
  -8.401759749326090D-07,   0.000000000000000D+00,  -3.616091898481279D-06, &
   0.000000000000000D+00,  -2.053736329845054D-05,   0.000000000000000D+00, &
  -1.196141065231320D-04,   0.000000000000000D+00,  -1.561291192570896D-04, &
   0.000000000000000D+00,   4.423867966421799D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.475892499949442D-07,   0.000000000000000D+00, &
   2.814003338643780D-08,   0.000000000000000D+00,  -1.009426145360502D-08, &
   0.000000000000000D+00,  -6.669793675172399D-07,   0.000000000000000D+00, &
   2.913544090586997D-06,   0.000000000000000D+00,   2.051424809275798D-05, &
   0.000000000000000D+00,  -1.270457297444497D-04,   0.000000000000000D+00, &
  -3.013566491304118D-04,   0.000000000000000D+00,  -3.340346888334408D-04, &
   0.000000000000000D+00,   5.978079926236121D-08,   0.000000000000000D+00, &
  -1.644721108547597D-07,   0.000000000000000D+00,   3.099375347426240D-07, &
   0.000000000000000D+00,   6.388860619018453D-07,   0.000000000000000D+00, &
   2.469849995809344D-06,   0.000000000000000D+00,   2.072914322002014D-05, &
   0.000000000000000D+00,   1.536470804819249D-04,   0.000000000000000D+00, &
   1.235021584733392D-04,   0.000000000000000D+00,   8.847461879758385D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.270340492735163D-08, &
   0.000000000000000D+00,   2.104012237459434D-08,   0.000000000000000D+00, &
  -2.766155114833786D-07,   0.000000000000000D+00,   3.198666810411677D-08, &
   0.000000000000000D+00,  -1.222162056029019D-06,   0.000000000000000D+00, &
  -1.117836753240267D-05,   0.000000000000000D+00,  -3.969105184734989D-05, &
   0.000000000000000D+00,   1.202981155273743D-04,   0.000000000000000D+00, &
   1.820205242625642D-04,   0.000000000000000D+00,   1.985855752734513D-04, &
   0.000000000000000D+00,   1.557981058354067D-09,   0.000000000000000D+00, &
   4.168761042655881D-08,   0.000000000000000D+00,  -6.647335842722019D-09, &
   0.000000000000000D+00,  -1.371588125644419D-07,   0.000000000000000D+00, &
   5.524863600525038D-07,   0.000000000000000D+00,   1.452358528498901D-06, &
   0.000000000000000D+00,  -1.825330563975812D-05,   0.000000000000000D+00, &
  -1.260917854185485D-04,   0.000000000000000D+00,  -5.742425357003462D-05, &
   0.000000000000000D+00,  -1.038278837496025D-05,   0.000000000000000D+00 /)
lem(:,204) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.524021533174731D-05,   3.259726320702949D-04,   5.557891507203426D-04, &
   3.687207250283492D-04,  -4.483033183676518D-04,  -6.215856342410371D-04, &
  -3.020430585765628D-04,  -3.856743889662903D-04,   1.049841972058652D-03, &
   6.530842031713331D-04,  -1.167621798125984D-03,   1.401915007271368D-04, &
   3.575792215170947D-04,  -2.097773164922552D-04,   6.437054435011117D-05, &
  -3.654387954575773D-05,   5.766384267729560D-05,  -7.799766599276707D-05, &
   6.399062732003159D-05,  -1.025570895737010D-05,  -6.144648536525737D-05, &
   1.011835925196092D-04,   9.524021533174486D-05,   3.259726320703109D-04, &
   5.557891507203408D-04,   3.687207250283441D-04,  -4.483033183676499D-04, &
  -6.215856342410334D-04,  -3.020430585765740D-04,  -3.856743889662912D-04, &
   1.049841972058661D-03,   6.530842031713369D-04,  -1.167621798125990D-03, &
   1.401915007271392D-04,   3.575792215170954D-04,  -2.097773164922545D-04, &
   6.437054435010930D-05,  -3.654387954575934D-05,   5.766384267730075D-05, &
  -7.799766599276897D-05,   6.399062732002712D-05,  -1.025570895736521D-05, &
  -6.144648536525528D-05,   1.011835925196051D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.547680341490214D-05,   0.000000000000000D+00, &
   1.561528684245177D-04,   0.000000000000000D+00,  -5.433514330719546D-04, &
   0.000000000000000D+00,  -3.840097776294531D-04,   0.000000000000000D+00, &
   7.613642719313233D-03,   0.000000000000000D+00,  -1.850898906768039D-02, &
   0.000000000000000D+00,   1.942003665550443D-02,   0.000000000000000D+00, &
  -8.972508613725930D-03,   0.000000000000000D+00,   7.630611179105693D-04, &
   0.000000000000000D+00,   6.405056733994684D-04,   0.000000000000000D+00, &
  -6.781684027779110D-07,  -1.893587098665129D-06,   0.000000000000000D+00, &
  -3.426071541305978D-06,   0.000000000000000D+00,   3.628641323953797D-06, &
  -2.457259537337562D-06,   0.000000000000000D+00,   3.184749278499335D-06, &
   0.000000000000000D+00,   3.972129216269742D-06,   0.000000000000000D+00, &
   1.356336805555518D-06,   0.000000000000000D+00,  -1.297107162585666D-05, &
   6.908374828296848D-06,   0.000000000000000D+00,  -1.881527760434013D-06, &
   0.000000000000000D+00,  -5.935667251585299D-06,   0.000000000000000D+00, &
  -6.781684027778102D-07,   0.000000000000000D+00,  -2.456582046425758D-06, &
   0.000000000000000D+00,   2.924897639253166D-06,   0.000000000000000D+00, &
   3.039071795175311D-05,  -6.690222754675954D-06,   0.000000000000000D+00, &
  -1.096180295399136D-06,   0.000000000000000D+00,   4.066131080291212D-06, &
   0.000000000000000D+00,   4.382180590676685D-06,   0.000000000000000D+00, &
  -2.510724364993570D-06,   0.000000000000000D+00,   8.330933553020293D-07, &
   0.000000000000000D+00,   2.161936436059725D-06,   0.000000000000000D+00, &
  -4.893088972127564D-06,   0.000000000000000D+00,  -1.250133986722514D-06, &
   4.419832952222165D-06,   0.000000000000000D+00,   3.038089836106778D-06, &
   0.000000000000000D+00,  -4.482330965934682D-07,   0.000000000000000D+00, &
  -6.099700345909412D-06,   0.000000000000000D+00,  -2.258390053495511D-05, &
   0.000000000000000D+00,   2.664808639891046D-06,   0.000000000000000D+00, &
  -3.062523956393866D-07,   0.000000000000000D+00,  -4.718098349448357D-06, &
   0.000000000000000D+00,  -6.621686742311856D-06,   0.000000000000000D+00, &
  -5.400913022095283D-06,   0.000000000000000D+00,  -2.103585322601944D-04, &
  -2.738475799682732D-06,   0.000000000000000D+00,  -2.334667401906459D-06, &
   0.000000000000000D+00,   2.587483332135405D-07,   0.000000000000000D+00, &
   8.183160187927377D-06,   0.000000000000000D+00,   3.220106730517380D-05, &
   0.000000000000000D+00,   1.495190230181749D-04,   0.000000000000000D+00, &
  -8.457304079234184D-07,   0.000000000000000D+00,   7.842532601617625D-07, &
   0.000000000000000D+00,   3.799777795108989D-06,   0.000000000000000D+00, &
   6.664549104456052D-06,   0.000000000000000D+00,   3.587237946032867D-07, &
   0.000000000000000D+00,   1.339666859539219D-06,   0.000000000000000D+00, &
   6.022452792847012D-04,   1.228895973634681D-06,   0.000000000000000D+00, &
   4.117643514432205D-07,   0.000000000000000D+00,  -1.510664431251173D-06, &
   0.000000000000000D+00,  -7.553613584037194D-06,   0.000000000000000D+00, &
  -2.498202531190891D-05,   0.000000000000000D+00,  -1.019245144327372D-04, &
   0.000000000000000D+00,  -3.981561688412321D-04,   0.000000000000000D+00, &
  -1.710368254629840D-07,   0.000000000000000D+00,  -3.881064275027709D-07, &
   0.000000000000000D+00,  -1.159916088792988D-06,   0.000000000000000D+00, &
  -1.733889189062117D-06,   0.000000000000000D+00,   6.425467514154582D-06, &
   0.000000000000000D+00,   5.298219717500372D-05,   0.000000000000000D+00, &
   4.706404452599148D-05,   0.000000000000000D+00,  -8.466105164362035D-04, &
  -1.936258978043083D-07,   0.000000000000000D+00,   2.773026674107132D-07, &
   0.000000000000000D+00,   1.064308949836084D-06,   0.000000000000000D+00, &
   3.605444817253983D-06,   0.000000000000000D+00,   8.983157866659187D-06, &
   0.000000000000000D+00,   2.342520746211112D-05,   0.000000000000000D+00, &
   1.854649195015159D-04,   0.000000000000000D+00,   5.541884124910150D-04, &
   0.000000000000000D+00,   1.082404626489526D-07,   0.000000000000000D+00, &
  -1.072781641213210D-07,   0.000000000000000D+00,  -3.759566246435057D-08, &
   0.000000000000000D+00,  -3.431538462179658D-07,   0.000000000000000D+00, &
  -4.394630126714863D-06,   0.000000000000000D+00,  -2.704078068997872D-05, &
   0.000000000000000D+00,  -1.448812605711903D-04,   0.000000000000000D+00, &
  -8.165321199935728D-05,   0.000000000000000D+00,   6.683844251438009D-04, &
  -2.621129230997358D-08,   0.000000000000000D+00,  -4.636839205165295D-08, &
   0.000000000000000D+00,  -2.162423392117538D-07,   0.000000000000000D+00, &
  -7.384911526786885D-07,   0.000000000000000D+00,  -5.541194488243226D-07, &
   0.000000000000000D+00,   2.766573778660442D-06,   0.000000000000000D+00, &
   1.142310559360769D-05,   0.000000000000000D+00,  -2.308383838664743D-04, &
   0.000000000000000D+00,  -4.327085031647380D-04,   0.000000000000000D+00, &
   3.459047035972346D-09,   0.000000000000000D+00,   6.532744795195437D-08, &
   0.000000000000000D+00,   2.272209639247286D-09,   0.000000000000000D+00, &
   1.441523998516801D-07,   0.000000000000000D+00,   4.398227592847342D-07, &
   0.000000000000000D+00,   4.197762553449259D-06,   0.000000000000000D+00, &
   3.395741668124653D-05,   0.000000000000000D+00,   1.987925034922540D-04, &
   0.000000000000000D+00,   4.196911902257469D-05,   0.000000000000000D+00, &
  -2.671274758223079D-04,  -2.430681696146325D-10,   0.000000000000000D+00, &
  -2.912462340030347D-08,   0.000000000000000D+00,   4.741471192290411D-08, &
   0.000000000000000D+00,   2.876997635821153D-08,   0.000000000000000D+00, &
  -2.726127386717087D-07,   0.000000000000000D+00,  -1.966619428870632D-06, &
   0.000000000000000D+00,  -1.610930809330957D-05,   0.000000000000000D+00, &
  -3.569513247689791D-05,   0.000000000000000D+00,   2.076798322492251D-04, &
   0.000000000000000D+00,   1.775197774400469D-04,   0.000000000000000D+00, &
   7.153534324373709D-12,   0.000000000000000D+00,   7.032544091260512D-09, &
   0.000000000000000D+00,   1.422948719880278D-08,   0.000000000000000D+00, &
   1.518665772117676D-08,   0.000000000000000D+00,   3.766586480335723D-07, &
   0.000000000000000D+00,   4.358499319357029D-07,   0.000000000000000D+00, &
   1.815345429344322D-06,   0.000000000000000D+00,  -3.351278344217126D-05, &
   0.000000000000000D+00,  -1.571723857683240D-04,   0.000000000000000D+00, &
   1.099677165101271D-05,   0.000000000000000D+00,   5.416087796904291D-06 /)
lem(:,205) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.146936295846311D-05,  -3.096871449938922D-04, &
  -9.241123113207545D-04,  -1.584879321514225D-03,  -9.088461291057721D-04, &
   2.125351891462577D-03,   2.458504315438751D-03,  -2.083462069179736D-03, &
  -1.534886990971190D-03,   1.682399197513091D-03,  -6.419412757249218D-05, &
  -5.060088795166263D-04,   2.269347473775514D-04,  -5.291881759358013D-05, &
   4.226364980067277D-05,  -1.916467611332851D-05,   3.766588741845729D-05, &
  -6.258299457069353D-05,   3.644233190382202D-05,   5.737903656990473D-06, &
  -4.389106062616842D-05,   0.000000000000000D+00,   5.146936295847397D-05, &
   3.096871449938897D-04,   9.241123113207421D-04,   1.584879321514232D-03, &
   9.088461291057811D-04,  -2.125351891462590D-03,  -2.458504315438747D-03, &
   2.083462069179747D-03,   1.534886990971200D-03,  -1.682399197513103D-03, &
   6.419412757248002D-05,   5.060088795166284D-04,  -2.269347473775409D-04, &
   5.291881759358337D-05,  -4.226364980068049D-05,   1.916467611333390D-05, &
  -3.766588741845866D-05,   6.258299457068256D-05,  -3.644233190380972D-05, &
  -5.737903656983557D-06,   4.389106062615365D-05,   0.000000000000000D+00, &
   2.573468147923535D-05,   0.000000000000000D+00,  -6.434521046841710D-05, &
   0.000000000000000D+00,   2.027641889024152D-04,   0.000000000000000D+00, &
  -8.304171399722752D-04,   0.000000000000000D+00,   5.222230969803728D-03, &
   0.000000000000000D+00,  -1.937749520631668D-02,   0.000000000000000D+00, &
   3.702432415953643D-02,   0.000000000000000D+00,  -3.831589825911436D-02, &
   0.000000000000000D+00,   2.150619327436531D-02,   0.000000000000000D+00, &
  -6.180815324432303D-03,   0.000000000000000D+00,   9.780229877983672D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.034505208333256D-06, &
   0.000000000000000D+00,  -1.911201862373840D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.294685132575639D-06,   0.000000000000000D+00, &
   6.141680946368469D-06,   0.000000000000000D+00,   3.787174197328991D-07, &
   0.000000000000000D+00,   5.266814348845657D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.897628630050546D-06,   0.000000000000000D+00, &
   2.017116274928637D-06,   0.000000000000000D+00,  -2.114375741084733D-05, &
   0.000000000000000D+00,   3.134750449203680D-06,   0.000000000000000D+00, &
  -4.412272478678575D-06,   0.000000000000000D+00,  -1.520069421680747D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.571054596445402D-06, &
   0.000000000000000D+00,  -7.274897411616231D-06,   0.000000000000000D+00, &
   3.151744179576239D-06,   0.000000000000000D+00,   8.679217511004627D-05, &
   0.000000000000000D+00,  -5.285106603465882D-06,   0.000000000000000D+00, &
   2.469680204055102D-06,   0.000000000000000D+00,   1.946877404641066D-05, &
   0.000000000000000D+00,   3.298229006236808D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.518186891233797D-06,   0.000000000000000D+00, &
   7.918062117281035D-06,   0.000000000000000D+00,   2.517048110309856D-06, &
   0.000000000000000D+00,  -4.345457195798993D-05,   0.000000000000000D+00, &
  -3.452891395205541D-04,   0.000000000000000D+00,   3.514145359848483D-06, &
   0.000000000000000D+00,  -2.279929680223774D-06,   0.000000000000000D+00, &
  -1.732099898508383D-05,   0.000000000000000D+00,  -3.332913058303436D-05, &
   0.000000000000000D+00,  -1.878438176187453D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.068749592877568D-06,   0.000000000000000D+00, &
  -4.002907849821682D-06,   0.000000000000000D+00,   1.752515866393965D-07, &
   0.000000000000000D+00,   3.176167531047004D-05,   0.000000000000000D+00, &
   1.538131326630015D-04,   0.000000000000000D+00,   9.976476255897542D-04, &
   0.000000000000000D+00,  -3.634368656564556D-07,   0.000000000000000D+00, &
   1.706140349783170D-06,   0.000000000000000D+00,   1.010579086174458D-05, &
   0.000000000000000D+00,   1.664158145063625D-05,   0.000000000000000D+00, &
  -1.671857842976333D-05,   0.000000000000000D+00,  -2.241046312973308D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.771560184858361D-06, &
   0.000000000000000D+00,   2.075969138771546D-07,   0.000000000000000D+00, &
  -3.319503162110522D-06,   0.000000000000000D+00,  -1.958395033260835D-05, &
   0.000000000000000D+00,  -7.273294412638319D-05,   0.000000000000000D+00, &
  -2.962790880174826D-04,   0.000000000000000D+00,  -1.922084345855228D-03, &
   0.000000000000000D+00,  -6.231034888635977D-07,   0.000000000000000D+00, &
  -3.628978836308478D-08,   0.000000000000000D+00,  -3.078971829889313D-06, &
   0.000000000000000D+00,  -2.141273604675251D-06,   0.000000000000000D+00, &
   2.304273560588361D-05,   0.000000000000000D+00,   1.823987935868073D-04, &
   0.000000000000000D+00,   6.348197227895361D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.559824227914752D-07,   0.000000000000000D+00, &
   4.164487920318765D-07,   0.000000000000000D+00,   2.763234753129178D-06, &
   0.000000000000000D+00,   7.011930697992513D-06,   0.000000000000000D+00, &
   1.969922764373222D-05,   0.000000000000000D+00,   4.794092383244523D-05, &
   0.000000000000000D+00,   3.849451302432146D-04,   0.000000000000000D+00, &
   2.469423230590010D-03,   0.000000000000000D+00,   1.879332659735006D-07, &
   0.000000000000000D+00,  -5.551460652964590D-07,   0.000000000000000D+00, &
   2.282248959411772D-07,   0.000000000000000D+00,  -9.870119910563955D-07, &
   0.000000000000000D+00,  -1.144648700214586D-05,   0.000000000000000D+00, &
  -6.855352139018906D-05,   0.000000000000000D+00,  -3.988019020829496D-04, &
   0.000000000000000D+00,  -8.895313652906659D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.733129527978334D-08,   0.000000000000000D+00, &
   9.903629841475407D-08,   0.000000000000000D+00,  -1.297841107889962D-06, &
   0.000000000000000D+00,  -8.908500305509171D-07,   0.000000000000000D+00, &
  -7.673002452755913D-07,   0.000000000000000D+00,   9.033345531893765D-06, &
   0.000000000000000D+00,   5.103901439563160D-05,   0.000000000000000D+00, &
  -4.194049114957014D-04,   0.000000000000000D+00,  -2.136643166602093D-03, &
   0.000000000000000D+00,   1.318128501525148D-08,   0.000000000000000D+00, &
   1.580254639051781D-07,   0.000000000000000D+00,   1.889980919283848D-07, &
   0.000000000000000D+00,  -5.053073678246792D-07,   0.000000000000000D+00, &
   1.475329747773212D-06,   0.000000000000000D+00,   9.227473933086708D-06, &
   0.000000000000000D+00,   7.587041797722263D-05,   0.000000000000000D+00, &
   5.176552930243696D-04,   0.000000000000000D+00,   7.023834772159586D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.511568725786969D-09, &
   0.000000000000000D+00,  -1.033801610832745D-07,   0.000000000000000D+00, &
   5.321562754087514D-07,   0.000000000000000D+00,   1.126694768787483D-07, &
   0.000000000000000D+00,  -7.828317146815452D-07,   0.000000000000000D+00, &
  -3.881326164748553D-06,   0.000000000000000D+00,  -3.703265719918010D-05, &
   0.000000000000000D+00,  -1.146531782116288D-04,   0.000000000000000D+00, &
   4.326565251996633D-04,   0.000000000000000D+00,   1.251756148063562D-03, &
   0.000000000000000D+00,   9.568746955445732D-11,   0.000000000000000D+00, &
   2.066852187741092D-08,   0.000000000000000D+00,  -2.232758747372727D-07, &
   0.000000000000000D+00,   8.884487379924595D-07,   0.000000000000000D+00, &
   5.952070113046300D-07,   0.000000000000000D+00,   7.396358831213971D-07, &
   0.000000000000000D+00,   4.204670455561099D-06,   0.000000000000000D+00, &
  -6.995080334335014D-05,   0.000000000000000D+00,  -4.450887696793046D-04, &
   0.000000000000000D+00,  -2.816241538500622D-04,   0.000000000000000D+00 /)
lem(:,206) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.913803454748216D-03,  -4.579024767873751D-03,  -1.073952454152942D-02, &
   9.599588157308945D-03,   1.203013224196182D-02,  -5.166498729007805D-03, &
  -5.511669175338172D-03,  -8.499523443907315D-04,  -6.887685745110614D-04, &
  -9.683900729870828D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.913803454748202D-03,  -4.579024767873661D-03, &
  -1.073952454152942D-02,   9.599588157308849D-03,   1.203013224196182D-02, &
  -5.166498729007772D-03,  -5.511669175338237D-03,  -8.499523443906553D-04, &
  -6.887685745110302D-04,  -9.683900729886415D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.552171617026273D-04,   0.000000000000000D+00,   5.602610072870282D-06, &
   0.000000000000000D+00,   3.075942785105414D-07,   0.000000000000000D+00, &
   2.540879311114882D-08,   0.000000000000000D+00,   2.520713601997269D-09, &
   0.000000000000000D+00,   2.713968311393437D-10,   0.000000000000000D+00, &
  -2.849256375396529D-03,   0.000000000000000D+00,  -5.887993023457569D-05, &
   0.000000000000000D+00,  -4.573362197581298D-06,   0.000000000000000D+00, &
  -4.934864037071268D-07,   0.000000000000000D+00,  -6.141285444363537D-08, &
   0.000000000000000D+00,  -8.103739756374472D-09,   0.000000000000000D+00, &
  -1.280123725576727D-09,   6.410564675303267D-03,   0.000000000000000D+00, &
   3.711098879643723D-05,   0.000000000000000D+00,   1.665656489610241D-05, &
   0.000000000000000D+00,   3.036359959391394D-06,   0.000000000000000D+00, &
   5.306074562546438D-07,   0.000000000000000D+00,   9.151196973852793D-08, &
   0.000000000000000D+00,   1.560605989376217D-08,   0.000000000000000D+00, &
  -4.057525303354003D-03,   0.000000000000000D+00,   7.839195888719506D-04, &
   0.000000000000000D+00,   2.579885217513818D-05,   0.000000000000000D+00, &
  -3.608257255857885D-06,   0.000000000000000D+00,  -1.678046134561173D-06, &
   0.000000000000000D+00,  -4.575785167887719D-07,   0.000000000000000D+00, &
  -1.065635417057295D-07,   0.000000000000000D+00,  -2.861337775546544D-08, &
  -7.480007924946745D-03,   0.000000000000000D+00,  -2.402930874275847D-03, &
   0.000000000000000D+00,  -1.819055558908900D-04,   0.000000000000000D+00, &
  -2.355178194662184D-05,   0.000000000000000D+00,  -1.091801420461469D-06, &
   0.000000000000000D+00,   6.616216910373164D-07,   0.000000000000000D+00, &
   3.273416089437095D-07,   0.000000000000000D+00,   1.082404626489526D-07, &
   0.000000000000000D+00,   1.536068377002750D-02,   0.000000000000000D+00, &
   1.886411063274246D-03,   0.000000000000000D+00,   2.897065188500167D-05, &
   0.000000000000000D+00,   5.109886274688685D-05,   0.000000000000000D+00, &
   1.507352292890799D-05,   0.000000000000000D+00,   2.456338105155462D-06, &
   0.000000000000000D+00,  -5.928668361689146D-09,   0.000000000000000D+00, &
  -1.894560871767062D-07,   0.000000000000000D+00,  -1.332519964196669D-07, &
  -7.924642073696581D-03,   0.000000000000000D+00,   2.548965785895708D-03, &
   0.000000000000000D+00,   8.383296766524495D-04,   0.000000000000000D+00, &
   7.292005763249218D-05,   0.000000000000000D+00,  -9.268690048917046D-06, &
   0.000000000000000D+00,  -7.992202032739052D-06,   0.000000000000000D+00, &
  -2.383199052386146D-06,   0.000000000000000D+00,  -3.371862721218134D-07, &
   0.000000000000000D+00,   8.263429129356054D-08,   0.000000000000000D+00, &
  -4.575938446906185D-03,   0.000000000000000D+00,  -5.724186116288449D-03, &
   0.000000000000000D+00,  -1.068260772763359D-03,   0.000000000000000D+00, &
  -2.148121126619998D-04,   0.000000000000000D+00,  -5.344238314889999D-05, &
   0.000000000000000D+00,  -4.154866444954186D-06,   0.000000000000000D+00, &
   3.244688952129152D-06,   0.000000000000000D+00,   1.769620009344265D-06, &
   0.000000000000000D+00,   4.516231661316888D-07,   0.000000000000000D+00, &
  -3.833351892257533D-08,   7.669478755237995D-03,   0.000000000000000D+00, &
   2.701710584792058D-03,   0.000000000000000D+00,  -7.132938761341721D-04, &
   0.000000000000000D+00,  -1.098443391704188D-04,   0.000000000000000D+00, &
   4.537830091041734D-05,   0.000000000000000D+00,   2.934313094183860D-05, &
   0.000000000000000D+00,   7.013741060405253D-06,   0.000000000000000D+00, &
  -5.030326014422852D-07,   0.000000000000000D+00,  -1.074437383021449D-06, &
   0.000000000000000D+00,  -4.265647345223706D-07,   0.000000000000000D+00, &
  -3.626663382760936D-03,   0.000000000000000D+00,   2.125711365720590D-03, &
   0.000000000000000D+00,   2.168442287488296D-03,   0.000000000000000D+00, &
   4.785478661335768D-04,   0.000000000000000D+00,   1.085150334952084D-04, &
   0.000000000000000D+00,   2.958725980955986D-06,   0.000000000000000D+00, &
  -1.301717640726414D-05,   0.000000000000000D+00,  -6.025457196319709D-06, &
   0.000000000000000D+00,  -7.907392134319581D-07,   0.000000000000000D+00, &
   5.190444273105669D-07,   0.000000000000000D+00,   5.414175167647428D-07 /)
lem(:,207) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.582644711828798D-04,   2.519242862768596D-03, &
  -2.254620248689380D-03,  -4.548910495471846D-03,   7.249098931208244D-03, &
   2.466281088061150D-03,  -6.757929924458944D-03,  -5.742737275129336D-04, &
   2.154322236638027D-03,   5.366039376297406D-04,   4.231521445470512D-04, &
  -2.009216456808017D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.582644711828965D-04, &
  -2.519242862768614D-03,   2.254620248689380D-03,   4.548910495471861D-03, &
  -7.249098931208258D-03,  -2.466281088061166D-03,   6.757929924458957D-03, &
   5.742737275129377D-04,  -2.154322236638054D-03,  -5.366039376297156D-04, &
  -4.231521445470482D-04,   2.009216456805269D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.144590180108382D-08,   0.000000000000000D+00, &
  -4.038926047396397D-07,   0.000000000000000D+00,  -1.724701286371560D-06, &
   0.000000000000000D+00,  -1.173973501160155D-06,   0.000000000000000D+00, &
   4.002665656464316D-06,   0.000000000000000D+00,   3.080378525803535D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.262696454201266D-06,   0.000000000000000D+00, &
   1.464030460218951D-07,   0.000000000000000D+00,   1.058699712978010D-08, &
   0.000000000000000D+00,   8.318354887923945D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.534021385027985D-05, &
   0.000000000000000D+00,  -2.271425830094999D-06,   0.000000000000000D+00, &
  -2.209373963500312D-07,   0.000000000000000D+00,  -2.400191983188210D-08, &
   0.000000000000000D+00,  -2.191149991795691D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.623461317553152D-04,   0.000000000000000D+00, &
   1.079416467784902D-05,   0.000000000000000D+00,   1.560788051842870D-06, &
   0.000000000000000D+00,   2.390778693334483D-07,   0.000000000000000D+00, &
   3.436697600608361D-08,   0.000000000000000D+00,   2.571378961093124D-09, &
   0.000000000000000D+00,  -3.991164125613035D-04,   0.000000000000000D+00, &
  -1.043679230739189D-05,   0.000000000000000D+00,  -3.760920449098271D-06, &
   0.000000000000000D+00,  -9.896471270016868D-07,   0.000000000000000D+00, &
  -2.182576611669044D-07,   0.000000000000000D+00,  -3.921631203386745D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.059692220950299D-04, &
   0.000000000000000D+00,  -7.940676875286581D-05,   0.000000000000000D+00, &
  -4.260642468079284D-06,   0.000000000000000D+00,   9.031078781426744D-07, &
   0.000000000000000D+00,   5.533798389478748D-07,   0.000000000000000D+00, &
   1.786363905850931D-07,   0.000000000000000D+00,   4.937010500194511D-08, &
   0.000000000000000D+00,  -9.356273933662664D-05,   0.000000000000000D+00, &
   3.278529234587712D-04,   0.000000000000000D+00,   3.618481855394546D-05, &
   0.000000000000000D+00,   5.466574519337734D-06,   0.000000000000000D+00, &
   2.873597752166643D-07,   0.000000000000000D+00,  -2.504244234434647D-07, &
   0.000000000000000D+00,  -1.426094934909911D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.515125228849002D-04,   0.000000000000000D+00, &
  -5.200880142081023D-04,   0.000000000000000D+00,  -4.150375417016625D-05, &
   0.000000000000000D+00,  -1.487949189810461D-05,   0.000000000000000D+00, &
  -4.158970546759652D-06,   0.000000000000000D+00,  -6.902446051598407D-07, &
   0.000000000000000D+00,   6.782877137328790D-08,   0.000000000000000D+00, &
   1.879332659735006D-07,   0.000000000000000D+00,   1.239274320475655D-03, &
   0.000000000000000D+00,   1.913849864471328D-04,   0.000000000000000D+00, &
  -9.452457939664436D-05,   0.000000000000000D+00,  -3.134223972752195D-06, &
   0.000000000000000D+00,   4.945208457336394D-06,   0.000000000000000D+00, &
   2.615571806290833D-06,   0.000000000000000D+00,   7.206759475786006D-07, &
   0.000000000000000D+00,   1.271430110553839D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.974553024608631D-04,   0.000000000000000D+00, &
   6.473339757907153D-04,   0.000000000000000D+00,   3.019714986298608D-04, &
   0.000000000000000D+00,   5.393758167204765D-05,   0.000000000000000D+00, &
   9.577257522444321D-06,   0.000000000000000D+00,  -5.584902872993456D-07, &
   0.000000000000000D+00,  -1.400848112806544D-06,   0.000000000000000D+00, &
  -6.099359611986326D-07,   0.000000000000000D+00,  -3.954008192928283D-08, &
   0.000000000000000D+00,  -1.062472697622862D-04,   0.000000000000000D+00, &
  -1.153350481710494D-03,   0.000000000000000D+00,  -2.148884341818009D-04, &
   0.000000000000000D+00,  -4.145057301347666D-05,   0.000000000000000D+00, &
  -2.251518238981013D-05,   0.000000000000000D+00,  -7.866933166071558D-06, &
   0.000000000000000D+00,  -1.092686179318778D-06,   0.000000000000000D+00, &
   6.065617152424367D-07,   0.000000000000000D+00,   4.625961949487693D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.090418092914094D-04, &
   0.000000000000000D+00,   6.875295563563542D-04,   0.000000000000000D+00, &
  -2.753168781571717D-04,   0.000000000000000D+00,  -8.661882209724809D-05, &
   0.000000000000000D+00,  -6.396720864330120D-06,   0.000000000000000D+00, &
   7.112693131262178D-06,   0.000000000000000D+00,   4.953608652584163D-06, &
   0.000000000000000D+00,   1.445581719608790D-06,   0.000000000000000D+00, &
  -1.621753673477553D-07,   0.000000000000000D+00,  -7.446214861887902D-07, &
   0.000000000000000D+00,  -5.137947621465896D-04,   0.000000000000000D+00, &
   2.160539495867563D-04,   0.000000000000000D+00,   6.025664101253570D-04, &
   0.000000000000000D+00,   1.358501383325397D-04,   0.000000000000000D+00, &
   4.366314868963289D-05,   0.000000000000000D+00,   1.172720963763831D-05, &
   0.000000000000000D+00,  -5.480136615364320D-07,   0.000000000000000D+00, &
  -2.580364749145952D-06,   0.000000000000000D+00,  -1.262669451498467D-06, &
   0.000000000000000D+00,  -2.304368801715395D-08,   0.000000000000000D+00 /)
lem(:,208) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.437939075172622D-04, &
  -9.899253085662693D-04,   2.051307548895130D-04,   1.187732347088552D-03, &
   2.891140915836612D-03,   2.089251010094764D-04,  -8.536965680783848D-03, &
  -1.735078340822483D-03,   7.481931358914708D-03,   2.311152076670815D-03, &
  -1.348142982650836D-03,  -1.001890973928294D-03,  -1.068658229690862D-03, &
  -4.634966690833982D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.437939075172762D-04,  -9.899253085662845D-04,   2.051307548894988D-04, &
   1.187732347088548D-03,   2.891140915836627D-03,   2.089251010094904D-04, &
  -8.536965680783859D-03,  -1.735078340822496D-03,   7.481931358914693D-03, &
   2.311152076670810D-03,  -1.348142982650811D-03,  -1.001890973928291D-03, &
  -1.068658229690875D-03,  -4.634966690833998D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.506497731089851D-09,   0.000000000000000D+00, &
   3.859004796419856D-07,   0.000000000000000D+00,   2.998124792770684D-06, &
   0.000000000000000D+00,   6.506341123620217D-06,   0.000000000000000D+00, &
  -9.465602096885839D-07,   0.000000000000000D+00,  -1.387027195935052D-05, &
   0.000000000000000D+00,  -2.807622787080599D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.720034203744618D-06,   0.000000000000000D+00, &
   1.113994086551253D-07,   0.000000000000000D+00,   6.251369733832454D-09, &
   0.000000000000000D+00,   1.653006419853997D-10,   0.000000000000000D+00, &
  -5.017670289136765D-05,   0.000000000000000D+00,  -1.795541857554435D-06, &
   0.000000000000000D+00,  -1.469478877639954D-07,   0.000000000000000D+00, &
  -1.182106216076087D-08,   0.000000000000000D+00,   1.485768661144444D-09, &
   3.261440844744009D-05,   0.000000000000000D+00,   7.984678341325602D-06, &
   0.000000000000000D+00,   1.062502366981418D-06,   0.000000000000000D+00, &
   1.417220222879507D-07,   0.000000000000000D+00,   1.161888964419345D-08, &
   0.000000000000000D+00,   6.758310350847696D-04,   0.000000000000000D+00, &
  -2.046322432369110D-06,   0.000000000000000D+00,  -2.204155613709190D-06, &
   0.000000000000000D+00,  -6.002643469503041D-07,   0.000000000000000D+00, &
  -1.133991606659040D-07,   0.000000000000000D+00,   7.613785895243098D-10, &
  -2.088458760948072D-03,   0.000000000000000D+00,   6.494911551271577D-06, &
   0.000000000000000D+00,  -3.551459560194285D-06,   0.000000000000000D+00, &
   4.023621121322034D-07,   0.000000000000000D+00,   3.089108344246352D-07, &
   0.000000000000000D+00,   8.070497353113320D-08,   0.000000000000000D+00, &
   1.647689123056277D-03,   0.000000000000000D+00,  -2.877587868230466D-04, &
   0.000000000000000D+00,   9.389499858767541D-06,   0.000000000000000D+00, &
   2.870340808410668D-06,   0.000000000000000D+00,   1.650824938521712D-07, &
   0.000000000000000D+00,  -1.520541826449241D-07,   0.000000000000000D+00, &
  -9.951031176410340D-08,   2.244844912180604D-03,   0.000000000000000D+00, &
   8.967154831373431D-04,   0.000000000000000D+00,  -1.133203679694510D-06, &
   0.000000000000000D+00,  -3.331858289556859D-06,   0.000000000000000D+00, &
  -1.725916069520274D-06,   0.000000000000000D+00,  -2.654726609514886D-07, &
   0.000000000000000D+00,   1.025944069898876D-07,   0.000000000000000D+00, &
  -5.057225281911330D-03,   0.000000000000000D+00,  -7.198987917058426D-04, &
   0.000000000000000D+00,   1.171374753498165D-04,   0.000000000000000D+00, &
  -8.269890654143804D-07,   0.000000000000000D+00,   7.449192701361720D-07, &
   0.000000000000000D+00,   9.013891748671817D-07,   0.000000000000000D+00, &
   2.091166630667318D-07,   0.000000000000000D+00,  -2.452916023774199D-07, &
   2.396646714286282D-03,   0.000000000000000D+00,  -1.003100963868457D-03, &
   0.000000000000000D+00,  -4.569544212920669D-04,   0.000000000000000D+00, &
  -1.907057388144436D-05,   0.000000000000000D+00,   7.432271201207553D-07, &
   0.000000000000000D+00,   5.366637234998707D-08,   0.000000000000000D+00, &
  -4.488983474351306D-07,   0.000000000000000D+00,  -1.072781641213210D-07, &
   0.000000000000000D+00,   1.886411063274246D-03,   0.000000000000000D+00, &
   2.261483794402819D-03,   0.000000000000000D+00,   4.227321301394397D-04, &
   0.000000000000000D+00,   2.283533000081228D-06,   0.000000000000000D+00, &
   4.919694608494048D-06,   0.000000000000000D+00,  -2.536412436036162D-07, &
   0.000000000000000D+00,  -1.501485425853169D-07,   0.000000000000000D+00, &
   2.762615418649931D-07,   0.000000000000000D+00,   2.499715869135115D-07, &
  -2.560474346162148D-03,   0.000000000000000D+00,  -1.070152144880910D-03, &
   0.000000000000000D+00,   5.058561951723381D-04,   0.000000000000000D+00, &
   1.789501125204395D-04,   0.000000000000000D+00,   1.141134091215218D-05, &
   0.000000000000000D+00,  -1.093492335793109D-06,   0.000000000000000D+00, &
   2.111053442806960D-10,   0.000000000000000D+00,   4.896099495158990D-08, &
   0.000000000000000D+00,  -2.314121278412485D-07,   0.000000000000000D+00, &
   8.920215260593442D-04,   0.000000000000000D+00,  -8.241235628187304D-04, &
   0.000000000000000D+00,  -1.214570409035943D-03,   0.000000000000000D+00, &
  -2.485804735978500D-04,   0.000000000000000D+00,  -4.318730555389897D-05, &
   0.000000000000000D+00,  -8.168995819305680D-06,   0.000000000000000D+00, &
  -2.911536494324118D-07,   0.000000000000000D+00,   1.576469582759871D-10, &
   0.000000000000000D+00,  -5.999449873543499D-09,   0.000000000000000D+00, &
   7.797383997259923D-07,   4.406234770596265D-05,   0.000000000000000D+00, &
   1.097856152432717D-03,   0.000000000000000D+00,   5.394503378759455D-04, &
   0.000000000000000D+00,  -1.690861711948870D-04,   0.000000000000000D+00, &
  -2.703567840646927D-05,   0.000000000000000D+00,   9.289064729801245D-06, &
   0.000000000000000D+00,   4.698852446959872D-06,   0.000000000000000D+00, &
   7.721627390596786D-07,   0.000000000000000D+00,   1.410916235009792D-07, &
   0.000000000000000D+00,  -1.322576401036110D-07,   0.000000000000000D+00, &
  -7.630498148417766D-05,   0.000000000000000D+00,  -3.781195944846609D-04, &
   0.000000000000000D+00,   4.830398905853312D-04,   0.000000000000000D+00, &
   5.281364032406298D-04,   0.000000000000000D+00,   1.197579959714180D-04, &
   0.000000000000000D+00,   2.524152882823122D-05,   0.000000000000000D+00, &
   3.101756074753165D-07,   0.000000000000000D+00,  -2.266855170916724D-06, &
   0.000000000000000D+00,  -8.109511916611176D-07,   0.000000000000000D+00, &
  -3.383566817946565D-07,   0.000000000000000D+00,  -4.040790615962174D-07 /)
lem(:,209) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.616521676472339D-05,   3.736552054007622D-04,  -6.645189355024239D-04, &
  -1.302671702599338D-03,   2.849377396710793D-03,   3.708928277994170D-03, &
  -7.189865710296249D-03,  -5.896433610952738D-03,   1.013651588830682D-02, &
   5.126704453517555D-03,  -6.931178848249555D-03,  -3.059442870837203D-03, &
   1.370886009257905D-03,   1.417159833331168D-03,   9.268688426562250D-04, &
  -1.378719901413425D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.616521676470734D-05,  -3.736552054007498D-04, &
   6.645189355024271D-04,   1.302671702599329D-03,  -2.849377396710789D-03, &
  -3.708928277994166D-03,   7.189865710296247D-03,   5.896433610952742D-03, &
  -1.013651588830681D-02,  -5.126704453517550D-03,   6.931178848249569D-03, &
   3.059442870837208D-03,  -1.370886009257933D-03,  -1.417159833331161D-03, &
  -9.268688426562108D-04,   1.378719901413147D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.271095340989752D-07,   0.000000000000000D+00,  -3.178062442196349D-06, &
   0.000000000000000D+00,  -1.115148835168313D-05,   0.000000000000000D+00, &
  -9.501425170169350D-06,   0.000000000000000D+00,   1.203125060493881D-05, &
   0.000000000000000D+00,   1.956272092770086D-05,   0.000000000000000D+00, &
  -1.272080044244692D-06,   0.000000000000000D+00,  -1.438158961433138D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.107881636139945D-07,   0.000000000000000D+00, &
   3.699790250342613D-09,   0.000000000000000D+00,  -8.093291171869555D-10, &
   0.000000000000000D+00,  -1.809012443014073D-06,   0.000000000000000D+00, &
  -1.086813386103679D-07,   0.000000000000000D+00,  -2.521371481015584D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.946924640780297D-06, &
   0.000000000000000D+00,   8.457741775958603D-07,   0.000000000000000D+00, &
   8.355797179666274D-08,   0.000000000000000D+00,  -1.703777331439395D-08, &
   0.000000000000000D+00,  -9.267713678665996D-06,   0.000000000000000D+00, &
  -1.772670774857296D-06,   0.000000000000000D+00,  -4.204556001800075D-07, &
   0.000000000000000D+00,  -4.386518197142380D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.618544896497200D-05,   0.000000000000000D+00, &
  -4.026074114868480D-06,   0.000000000000000D+00,   2.296407458045937D-07, &
   0.000000000000000D+00,   1.959274905849223D-07,   0.000000000000000D+00, &
   1.026808536275895D-08,   0.000000000000000D+00,   2.801169348222646D-04, &
   0.000000000000000D+00,   1.408723133771413D-05,   0.000000000000000D+00, &
   2.853980085612503D-06,   0.000000000000000D+00,   1.633112302464384D-07, &
   0.000000000000000D+00,  -1.030593650802877D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.498500667629974D-04,   0.000000000000000D+00, &
   4.521754504483098D-05,   0.000000000000000D+00,  -2.984975006588866D-06, &
   0.000000000000000D+00,  -1.505387763952037D-06,   0.000000000000000D+00, &
  -1.717723172415013D-07,   0.000000000000000D+00,   2.553502255714325D-07, &
   0.000000000000000D+00,   1.669781490858917D-04,   0.000000000000000D+00, &
  -2.739643499769489D-04,   0.000000000000000D+00,  -1.496246308663884D-05, &
   0.000000000000000D+00,  -3.243252853655058D-07,   0.000000000000000D+00, &
   6.316546321622480D-07,   0.000000000000000D+00,   4.581158854771778D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.676056372656788D-04, &
   0.000000000000000D+00,   4.970060905655846D-04,   0.000000000000000D+00, &
   9.313510509613099D-06,   0.000000000000000D+00,   6.763632317583393D-06, &
   0.000000000000000D+00,   9.893105407002325D-07,   0.000000000000000D+00, &
  -2.044563466882637D-07,   0.000000000000000D+00,   8.677358546239187D-08, &
   0.000000000000000D+00,  -1.015738326251009D-03,   0.000000000000000D+00, &
  -2.216203237384162D-04,   0.000000000000000D+00,   1.362662019472518D-04, &
   0.000000000000000D+00,   4.682030758260895D-06,   0.000000000000000D+00, &
  -2.508254206591940D-06,   0.000000000000000D+00,  -7.735351343886224D-07, &
   0.000000000000000D+00,   2.343023472915809D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.073435120902055D-04,   0.000000000000000D+00, &
  -6.021408257648072D-04,   0.000000000000000D+00,  -3.485967174928456D-04, &
   0.000000000000000D+00,  -3.606307347411875D-05,   0.000000000000000D+00, &
  -5.633909754072525D-06,   0.000000000000000D+00,   4.205185804323019D-07, &
   0.000000000000000D+00,   3.984748893061747D-07,   0.000000000000000D+00, &
  -5.551460652964590D-07,   0.000000000000000D+00,   1.913849864471328D-04, &
   0.000000000000000D+00,   1.117983495962564D-03,   0.000000000000000D+00, &
   2.184635519489865D-04,   0.000000000000000D+00,  -7.813143621033873D-06, &
   0.000000000000000D+00,   1.043751607224640D-05,   0.000000000000000D+00, &
   3.999106719155770D-06,   0.000000000000000D+00,   4.050142170898372D-07, &
   0.000000000000000D+00,  -1.895409457584857D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.554321500541940D-04,   0.000000000000000D+00, &
  -6.441224187142505D-04,   0.000000000000000D+00,   3.776530006396227D-04, &
   0.000000000000000D+00,   1.579088791112338D-04,   0.000000000000000D+00, &
   1.913506460803873D-05,   0.000000000000000D+00,  -1.454643755887772D-06, &
   0.000000000000000D+00,  -2.206076367312732D-06,   0.000000000000000D+00, &
  -4.756237528439882D-07,   0.000000000000000D+00,  -1.466585218044261D-07, &
   0.000000000000000D+00,   3.757151533908318D-04,   0.000000000000000D+00, &
  -2.539341389334899D-04,   0.000000000000000D+00,  -7.738430053493895D-04, &
   0.000000000000000D+00,  -1.636163454536218D-04,   0.000000000000000D+00, &
  -3.756874274299882D-05,   0.000000000000000D+00,  -1.257755086256498D-05, &
   0.000000000000000D+00,  -1.492831581660526D-06,   0.000000000000000D+00, &
   8.401894752308251D-07,   0.000000000000000D+00,   6.363544524835706D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.896933824663139D-05, &
   0.000000000000000D+00,   6.133095229143742D-04,   0.000000000000000D+00, &
   4.040802108319558D-04,   0.000000000000000D+00,  -1.370659240460748D-04, &
   0.000000000000000D+00,  -3.190913152792674D-05,   0.000000000000000D+00, &
   5.637342399934976D-06,   0.000000000000000D+00,   6.487951545575399D-06, &
   0.000000000000000D+00,   2.017634052639688D-06,   0.000000000000000D+00, &
  -1.682666225136007D-07,   0.000000000000000D+00,   2.101031504927788D-07, &
   0.000000000000000D+00,  -1.299861822393635D-05,   0.000000000000000D+00, &
  -3.674749347867609D-04,   0.000000000000000D+00,   2.498804151916682D-04, &
   0.000000000000000D+00,   3.627551563211225D-04,   0.000000000000000D+00, &
   8.580710924816366D-05,   0.000000000000000D+00,   2.379905159230350D-05, &
   0.000000000000000D+00,   2.711298489717076D-06,   0.000000000000000D+00, &
  -2.595754469575207D-06,   0.000000000000000D+00,  -1.684472609723360D-06, &
   0.000000000000000D+00,  -3.872636477350219D-07,   0.000000000000000D+00 /)
lem(:,210) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.118319367554957D-05,  -1.378634179311442D-04, &
   3.787162689589288D-04,   3.648275321615918D-04,  -1.129687259349322D-03, &
   7.017961987835426D-04,  -1.395052718474322D-03,  -4.209958518450967D-03, &
   8.539167575295990D-03,   7.710087216243781D-03,  -1.231182387721301D-02, &
  -7.080606712498764D-03,   7.558799317800794D-03,   3.678849148422719D-03, &
  -1.602424778155615D-03,  -1.359103283397701D-03,  -2.646957332728238D-04, &
   5.614186227576518D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.118319367556252D-05, &
  -1.378634179311459D-04,   3.787162689589178D-04,   3.648275321616027D-04, &
  -1.129687259349314D-03,   7.017961987835350D-04,  -1.395052718474340D-03, &
  -4.209958518450969D-03,   8.539167575295986D-03,   7.710087216243786D-03, &
  -1.231182387721301D-02,  -7.080606712498771D-03,   7.558799317800791D-03, &
   3.678849148422716D-03,  -1.602424778155606D-03,  -1.359103283397683D-03, &
  -2.646957332728284D-04,   5.614186227576353D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.358574552454562D-08,   0.000000000000000D+00,   2.635072043887836D-06, &
   0.000000000000000D+00,   1.436966021749580D-05,   0.000000000000000D+00, &
   2.189431481635558D-05,   0.000000000000000D+00,  -1.432450028665164D-06, &
   0.000000000000000D+00,  -2.122776463286134D-05,   0.000000000000000D+00, &
  -1.181034913372882D-05,   0.000000000000000D+00,  -5.262997753916541D-06, &
   0.000000000000000D+00,   1.139457299951627D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.972272802800671D-07,   0.000000000000000D+00,   1.792975275183929D-09, &
   0.000000000000000D+00,  -3.070376774778129D-06,   0.000000000000000D+00, &
  -8.816395611379936D-08,   0.000000000000000D+00,   1.368661139503161D-08, &
   1.170317918870853D-05,   0.000000000000000D+00,   7.633027482843837D-07, &
   0.000000000000000D+00,   3.227653805902903D-08,   0.000000000000000D+00, &
   1.727546619187504D-05,   0.000000000000000D+00,  -1.746877237634243D-06, &
   0.000000000000000D+00,  -3.005269878895945D-07,   0.000000000000000D+00, &
   5.393565733521179D-08,  -1.305422922571003D-04,   0.000000000000000D+00, &
  -2.576624927117547D-06,   0.000000000000000D+00,   1.953027617628895D-07, &
   0.000000000000000D+00,   1.093366840653112D-07,   0.000000000000000D+00, &
   2.456958844090387D-05,   0.000000000000000D+00,   7.088423579191362D-06, &
   0.000000000000000D+00,   2.447964727628545D-06,   0.000000000000000D+00, &
   7.988573820182934D-08,   0.000000000000000D+00,  -1.905163693947497D-07, &
   6.162174190933194D-04,   0.000000000000000D+00,  -4.112297368041471D-07, &
   0.000000000000000D+00,  -3.750658260266595D-06,   0.000000000000000D+00, &
  -1.243094989786318D-06,   0.000000000000000D+00,  -1.540381659026695D-08, &
   0.000000000000000D+00,  -7.996733717146503D-04,   0.000000000000000D+00, &
   9.914542786249221D-05,   0.000000000000000D+00,  -6.777530891676200D-06, &
   0.000000000000000D+00,   4.057592546768529D-07,   0.000000000000000D+00, &
   5.386354731886684D-07,   0.000000000000000D+00,  -3.249894737223463D-07, &
  -5.394871433627489D-04,   0.000000000000000D+00,  -3.902228128640278D-04, &
   0.000000000000000D+00,   3.789851521003255D-05,   0.000000000000000D+00, &
   5.183617731247308D-06,   0.000000000000000D+00,   4.205426914246940D-07, &
   0.000000000000000D+00,  -1.514591555793368D-07,   0.000000000000000D+00, &
   1.657573528987656D-03,   0.000000000000000D+00,   3.628059152056788D-04, &
   0.000000000000000D+00,  -1.495812611427084D-04,   0.000000000000000D+00, &
  -7.787831228089667D-06,   0.000000000000000D+00,  -2.774799133052872D-06, &
   0.000000000000000D+00,  -4.300717602197040D-07,   0.000000000000000D+00, &
   6.165623783546453D-07,  -5.827892683430647D-04,   0.000000000000000D+00, &
   4.392822744378018D-04,   0.000000000000000D+00,   3.451749155042537D-04, &
   0.000000000000000D+00,  -8.643341367855953D-06,   0.000000000000000D+00, &
   8.465744871828305D-07,   0.000000000000000D+00,   1.254469345520339D-06, &
   0.000000000000000D+00,   2.923097261097772D-07,   0.000000000000000D+00, &
  -9.336616385702186D-04,   0.000000000000000D+00,  -1.057937631327633D-03, &
   0.000000000000000D+00,  -2.424020885477146D-04,   0.000000000000000D+00, &
   9.196547693987144D-05,   0.000000000000000D+00,   8.713167423094190D-06, &
   0.000000000000000D+00,   9.573003802246408D-07,   0.000000000000000D+00, &
  -6.021055482860772D-07,   0.000000000000000D+00,   7.239588033394552D-08, &
   7.827539166771905D-04,   0.000000000000000D+00,   4.724510448396389D-04, &
   0.000000000000000D+00,  -4.292029854605905D-04,   0.000000000000000D+00, &
  -2.433407397838047D-04,   0.000000000000000D+00,  -2.030290672609924D-05, &
   0.000000000000000D+00,  -5.108404376211451D-06,   0.000000000000000D+00, &
  -1.148703288457579D-06,   0.000000000000000D+00,  -3.759566246435057D-08, &
   0.000000000000000D+00,   2.897065188500167D-05,   0.000000000000000D+00, &
   4.227321301394397D-04,   0.000000000000000D+00,   9.164020716964762D-04, &
   0.000000000000000D+00,   1.876198103959929D-04,   0.000000000000000D+00, &
  -7.878795778994625D-06,   0.000000000000000D+00,   3.888123032117673D-06, &
   0.000000000000000D+00,   2.553123586661460D-06,   0.000000000000000D+00, &
   9.866560474603202D-07,   0.000000000000000D+00,  -7.438272383284002D-07, &
  -1.947836756561436D-04,   0.000000000000000D+00,  -4.905348829299328D-04, &
   0.000000000000000D+00,  -4.605531236365895D-04,   0.000000000000000D+00, &
   2.770982514094166D-04,   0.000000000000000D+00,   1.121864033993638D-04, &
   0.000000000000000D+00,   1.352047695520183D-05,   0.000000000000000D+00, &
   7.037340206148654D-07,   0.000000000000000D+00,  -1.004734586925729D-06, &
   0.000000000000000D+00,  -8.007234075830249D-07,   0.000000000000000D+00, &
   2.995190546576231D-05,   0.000000000000000D+00,   1.352491284223209D-04, &
   0.000000000000000D+00,  -2.808679823595268D-04,   0.000000000000000D+00, &
  -6.190422225728379D-04,   0.000000000000000D+00,  -1.329741775603131D-04, &
   0.000000000000000D+00,  -2.680485727774145D-05,   0.000000000000000D+00, &
  -8.255409978652355D-06,   0.000000000000000D+00,  -1.780835701083723D-06, &
   0.000000000000000D+00,   5.514791827886822D-07,   0.000000000000000D+00, &
   1.202110319421490D-06,   2.042078621805989D-05,   0.000000000000000D+00, &
  -2.054484344526569D-06,   0.000000000000000D+00,   4.299301727088516D-04, &
   0.000000000000000D+00,   2.972765339258929D-04,   0.000000000000000D+00, &
  -1.010791275940241D-04,   0.000000000000000D+00,  -2.045641724916413D-05, &
   0.000000000000000D+00,   4.371950350220682D-06,   0.000000000000000D+00, &
   4.108179154594096D-06,   0.000000000000000D+00,   1.609697407201632D-06, &
   0.000000000000000D+00,   4.158034413906159D-07,   0.000000000000000D+00, &
  -6.070550917128262D-06,   0.000000000000000D+00,   1.367409393382627D-05, &
   0.000000000000000D+00,  -2.028641725038841D-04,   0.000000000000000D+00, &
   2.158766811774879D-04,   0.000000000000000D+00,   2.932036186530994D-04, &
   0.000000000000000D+00,   6.916785563500212D-05,   0.000000000000000D+00, &
   1.692289532305607D-05,   0.000000000000000D+00,   1.536958528306863D-06, &
   0.000000000000000D+00,  -1.682510520076938D-06,   0.000000000000000D+00, &
  -1.462368955523958D-06,   0.000000000000000D+00,  -4.557033766682066D-07 /)
lem(:,211) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.729586508127394D-06, &
   3.780446074840797D-05,  -2.025975776494699D-04,   1.002634706568157D-04, &
   1.054455302485463D-03,  -7.388121420556486D-04,  -2.170346655522550D-03, &
  -1.339385966868805D-03,   5.785429926762608D-03,   6.830985362156513D-03, &
  -1.199145891145735D-02,  -9.680659992580550D-03,   1.414135898037039D-02, &
   6.727637808893711D-03,  -8.925275117486387D-03,  -2.830237103770420D-03, &
   2.796661265734211D-03,   1.149713928101647D-03,  -4.311009603458609D-04, &
  -5.615835115823264D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.729586508127006D-06,  -3.780446074842740D-05,   2.025975776494779D-04, &
  -1.002634706568026D-04,  -1.054455302485475D-03,   7.388121420556430D-04, &
   2.170346655522562D-03,   1.339385966868803D-03,  -5.785429926762623D-03, &
  -6.830985362156505D-03,   1.199145891145737D-02,   9.680659992580555D-03, &
  -1.414135898037042D-02,  -6.727637808893704D-03,   8.925275117486399D-03, &
   2.830237103770412D-03,  -2.796661265734213D-03,  -1.149713928101644D-03, &
   4.311009603458675D-04,   5.615835115823255D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.560019013003318D-06, &
   0.000000000000000D+00,  -1.540818494907709D-05,   0.000000000000000D+00, &
  -3.584602673221236D-05,   0.000000000000000D+00,  -1.524892824349720D-05, &
   0.000000000000000D+00,   2.024324216812099D-05,   0.000000000000000D+00, &
   6.360430292384956D-06,   0.000000000000000D+00,   6.061332389970149D-06, &
   0.000000000000000D+00,   2.496674647712782D-05,   0.000000000000000D+00, &
  -9.502516331454614D-06,   0.000000000000000D+00,  -2.873368902811457D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.104952569758486D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.874107003060498D-07,   0.000000000000000D+00, &
  -5.158694921798788D-08,   0.000000000000000D+00,  -2.626428625890030D-06, &
   0.000000000000000D+00,  -2.073828043002058D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.915121715902582D-06,   0.000000000000000D+00, &
   2.458495057479905D-07,   0.000000000000000D+00,   7.561274008654276D-08, &
   0.000000000000000D+00,   2.424483579416956D-05,   0.000000000000000D+00, &
   1.965632524205920D-06,   0.000000000000000D+00,  -2.952806156478050D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.125147532433046D-05, &
   0.000000000000000D+00,  -2.284239812893494D-06,   0.000000000000000D+00, &
  -9.515002145670005D-07,   0.000000000000000D+00,   5.287319114864235D-07, &
   0.000000000000000D+00,  -6.399025627027735D-05,   0.000000000000000D+00, &
  -1.235728445556195D-05,   0.000000000000000D+00,   6.094034113887326D-07, &
   0.000000000000000D+00,   3.026779850793826D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.191919177592675D-04,   0.000000000000000D+00, &
   8.938521253820577D-06,   0.000000000000000D+00,   2.846933045531750D-06, &
   0.000000000000000D+00,  -2.383984463434444D-08,   0.000000000000000D+00, &
  -4.879897957258858D-07,   0.000000000000000D+00,  -1.609820624297830D-04, &
   0.000000000000000D+00,   1.125713758475013D-04,   0.000000000000000D+00, &
   9.794166784057983D-08,   0.000000000000000D+00,  -1.463627857221856D-06, &
   0.000000000000000D+00,  -6.270922487464291D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.054684307792750D-04,   0.000000000000000D+00, &
  -2.953152150003184D-04,   0.000000000000000D+00,   1.066640620973215D-05, &
   0.000000000000000D+00,   4.307044469776154D-07,   0.000000000000000D+00, &
   7.975200413803356D-07,   0.000000000000000D+00,  -8.854662313060814D-07, &
   0.000000000000000D+00,   4.567286012941521D-04,   0.000000000000000D+00, &
   1.876809488927795D-04,   0.000000000000000D+00,  -1.209708333383862D-04, &
   0.000000000000000D+00,  -3.006744345120450D-06,   0.000000000000000D+00, &
  -2.610452430597069D-07,   0.000000000000000D+00,  -2.620471598309825D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.228015812720763D-04, &
   0.000000000000000D+00,   3.254887454185666D-04,   0.000000000000000D+00, &
   2.854018854748742D-04,   0.000000000000000D+00,   3.170509307191083D-06, &
   0.000000000000000D+00,   7.913947403377679D-07,   0.000000000000000D+00, &
   3.716787512833762D-07,   0.000000000000000D+00,   1.496599289194139D-06, &
   0.000000000000000D+00,  -1.876827093026544D-04,   0.000000000000000D+00, &
  -6.711095231939786D-04,   0.000000000000000D+00,  -1.710955035722903D-04, &
   0.000000000000000D+00,   6.761301436470499D-05,   0.000000000000000D+00, &
   3.028124471899635D-06,   0.000000000000000D+00,  -2.597023307919505D-08, &
   0.000000000000000D+00,  -6.838393646044647D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.811545506297200D-04,   0.000000000000000D+00, &
   3.514366357340658D-04,   0.000000000000000D+00,  -3.397882223246466D-04, &
   0.000000000000000D+00,  -2.024019943449647D-04,   0.000000000000000D+00, &
  -1.903565104710568D-05,   0.000000000000000D+00,  -2.957776898108329D-06, &
   0.000000000000000D+00,  -8.401759749326090D-07,   0.000000000000000D+00, &
   2.282248959411772D-07,   0.000000000000000D+00,  -9.452457939664436D-05, &
   0.000000000000000D+00,   2.184635519489865D-04,   0.000000000000000D+00, &
   6.835403666418104D-04,   0.000000000000000D+00,   1.473096123214649D-04, &
   0.000000000000000D+00,  -1.319300027889045D-06,   0.000000000000000D+00, &
   5.547708881753694D-06,   0.000000000000000D+00,   2.108392963052461D-06, &
   0.000000000000000D+00,   3.918965198840519D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.421409771366374D-05,   0.000000000000000D+00, &
  -3.755524753305419D-04,   0.000000000000000D+00,  -3.658152630448167D-04, &
   0.000000000000000D+00,   2.239011141504473D-04,   0.000000000000000D+00, &
   9.388342773355986D-05,   0.000000000000000D+00,   1.063069404432594D-05, &
   0.000000000000000D+00,  -6.080824198529455D-07,   0.000000000000000D+00, &
  -1.239291384045499D-06,   0.000000000000000D+00,  -2.221899834808546D-06, &
   0.000000000000000D+00,   4.211253315266683D-05,   0.000000000000000D+00, &
   1.599646342080253D-04,   0.000000000000000D+00,  -1.981425821222246D-04, &
   0.000000000000000D+00,  -4.842563707199877D-04,   0.000000000000000D+00, &
  -1.080070911692979D-04,   0.000000000000000D+00,  -2.475135183080399D-05, &
   0.000000000000000D+00,  -7.692612820747820D-06,   0.000000000000000D+00, &
  -8.131976543326982D-07,   0.000000000000000D+00,   6.102262472667241D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.788313611192223D-06, &
   0.000000000000000D+00,   1.031293161550156D-05,   0.000000000000000D+00, &
   3.593321785210492D-04,   0.000000000000000D+00,   2.409068791879921D-04, &
   0.000000000000000D+00,  -8.319303347659957D-05,   0.000000000000000D+00, &
  -1.758142991938069D-05,   0.000000000000000D+00,   4.698366761476861D-06, &
   0.000000000000000D+00,   4.245360803655866D-06,   0.000000000000000D+00, &
   1.909286996435886D-06,   0.000000000000000D+00,   8.172751289530428D-07, &
   0.000000000000000D+00,  -4.977976817194909D-06,   0.000000000000000D+00, &
  -2.032029502348146D-05,   0.000000000000000D+00,  -1.684713688123103D-04, &
   0.000000000000000D+00,   1.693943629703174D-04,   0.000000000000000D+00, &
   2.337323650811333D-04,   0.000000000000000D+00,   5.689867329550888D-05, &
   0.000000000000000D+00,   1.513880832726342D-05,   0.000000000000000D+00, &
   1.123934188992853D-06,   0.000000000000000D+00,  -2.073897812296907D-06, &
   0.000000000000000D+00,  -1.076067231574940D-06,   0.000000000000000D+00 /)
lem(:,212) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.787209687917779D-07,  -1.456894838276560D-05,   8.118639760358376D-05, &
  -6.583552619140819D-05,  -3.970075803822169D-04,   4.026090159074689D-04, &
   3.085435794900483D-05,  -1.050974216954416D-04,   2.139168721703806D-03, &
   1.944638524645289D-03,  -8.172412861792168D-03,  -7.102685948737521D-03, &
   1.516954133475962D-02,   8.801551494821954D-03,  -1.568124551000366D-02, &
  -4.149652282245987D-03,   9.385787834004802D-03,   1.434396311866492D-04, &
  -3.480196872199560D-03,   2.288973056667493D-04,   1.086233339753633D-03, &
   9.882016632899706D-05,   5.787209688043265D-07,  -1.456894838276322D-05, &
   8.118639760358128D-05,  -6.583552619141610D-05,  -3.970075803822090D-04, &
   4.026090159074714D-04,   3.085435794899607D-05,  -1.050974216954426D-04, &
   2.139168721703805D-03,   1.944638524645285D-03,  -8.172412861792166D-03, &
  -7.102685948737530D-03,   1.516954133475964D-02,   8.801551494821962D-03, &
  -1.568124551000366D-02,  -4.149652282246001D-03,   9.385787834004839D-03, &
   1.434396311866385D-04,  -3.480196872199572D-03,   2.288973056667601D-04, &
   1.086233339753631D-03,   9.882016632898659D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.171401771412183D-05,   0.000000000000000D+00, &
   4.653746511115256D-05,   0.000000000000000D+00,   3.893513126828883D-05, &
   0.000000000000000D+00,  -1.845490935817142D-05,   0.000000000000000D+00, &
  -8.259737992709898D-06,   0.000000000000000D+00,   2.155903445905824D-05, &
   0.000000000000000D+00,  -1.346194890789629D-05,   0.000000000000000D+00, &
  -4.135378511905947D-05,   0.000000000000000D+00,  -4.133408731843569D-06, &
   0.000000000000000D+00,   4.260267268490523D-05,   0.000000000000000D+00, &
  -2.657945769400609D-07,   1.578308508918299D-06,   0.000000000000000D+00, &
  -2.398454621162823D-06,   0.000000000000000D+00,  -1.983829088756160D-07, &
  -1.182852403575287D-05,   0.000000000000000D+00,   4.999669532244341D-07, &
   0.000000000000000D+00,   2.986942724794427D-05,   0.000000000000000D+00, &
   8.742715533670003D-07,   0.000000000000000D+00,  -6.191644981563931D-07, &
   3.878572520045139D-05,   0.000000000000000D+00,  -1.593878088984075D-06, &
   0.000000000000000D+00,  -5.620455120513823D-07,   0.000000000000000D+00, &
  -1.285123993429288D-04,   0.000000000000000D+00,   4.859943529346152D-07, &
   0.000000000000000D+00,   6.638633951150272D-07,   0.000000000000000D+00, &
   1.310920565277620D-07,  -6.199447471071650D-05,   0.000000000000000D+00, &
  -1.510144025235158D-05,   0.000000000000000D+00,   3.625491979785719D-06, &
   0.000000000000000D+00,  -1.925590600027391D-07,   0.000000000000000D+00, &
   2.966000363011029D-04,   0.000000000000000D+00,   2.987799202163225D-06, &
   0.000000000000000D+00,  -7.460907856328605D-06,   0.000000000000000D+00, &
  -8.547173696035187D-07,   0.000000000000000D+00,   1.671580795124341D-06, &
   3.497979077060419D-05,   0.000000000000000D+00,   1.274574100468043D-04, &
   0.000000000000000D+00,  -4.337834968357184D-06,   0.000000000000000D+00, &
  -1.268393265360576D-06,   0.000000000000000D+00,   1.299496269251333D-07, &
   0.000000000000000D+00,  -4.128658874285371D-04,   0.000000000000000D+00, &
  -1.803890975786334D-04,   0.000000000000000D+00,   7.507521522235254D-05, &
   0.000000000000000D+00,   5.632980174291352D-06,   0.000000000000000D+00, &
   1.381539638182079D-07,   0.000000000000000D+00,  -8.564397893859244D-07, &
   3.903431014888694D-05,   0.000000000000000D+00,  -1.240136533556765D-04, &
   0.000000000000000D+00,  -2.057537419804512D-04,   0.000000000000000D+00, &
   1.146834041315696D-05,   0.000000000000000D+00,   3.085207256339321D-07, &
   0.000000000000000D+00,  -2.058546011764625D-07,   0.000000000000000D+00, &
   3.570084220156669D-04,   0.000000000000000D+00,   3.920363440738272D-04, &
   0.000000000000000D+00,   1.602272022883594D-04,   0.000000000000000D+00, &
  -9.636811770735413D-05,   0.000000000000000D+00,  -7.095055076554724D-06, &
   0.000000000000000D+00,  -6.051198765448920D-07,   0.000000000000000D+00, &
  -1.535831332286322D-06,  -8.623287160514105D-05,   0.000000000000000D+00, &
  -1.351527930107809D-04,   0.000000000000000D+00,   2.375815448387929D-04, &
   0.000000000000000D+00,   2.199740209250965D-04,   0.000000000000000D+00, &
   4.509963320888085D-06,   0.000000000000000D+00,   2.506393378461505D-06, &
   0.000000000000000D+00,   1.124127929315009D-06,   0.000000000000000D+00, &
  -1.858537219598393D-04,   0.000000000000000D+00,  -2.142014354299925D-04, &
   0.000000000000000D+00,  -5.334347479941962D-04,   0.000000000000000D+00, &
  -1.420935563817551D-04,   0.000000000000000D+00,   5.550571810417385D-05, &
   0.000000000000000D+00,   3.777299959487106D-06,   0.000000000000000D+00, &
  -6.574297802491540D-08,   0.000000000000000D+00,   5.921381525070996D-07, &
   6.735713343613001D-05,   0.000000000000000D+00,   1.659397683825673D-04, &
   0.000000000000000D+00,   2.571744418177518D-04,   0.000000000000000D+00, &
  -2.661158466551857D-04,   0.000000000000000D+00,  -1.630246777852102D-04, &
   0.000000000000000D+00,  -1.784216594962091D-05,   0.000000000000000D+00, &
  -4.080278733580044D-06,   0.000000000000000D+00,  -3.431538462179658D-07, &
   0.000000000000000D+00,   5.109886274688685D-05,   0.000000000000000D+00, &
   2.283533000081228D-06,   0.000000000000000D+00,   1.876198103959929D-04, &
   0.000000000000000D+00,   5.539237815666150D-04,   0.000000000000000D+00, &
   1.236249262799109D-04,   0.000000000000000D+00,  -5.173454620707978D-07, &
   0.000000000000000D+00,   5.178724083693248D-06,   0.000000000000000D+00, &
   2.115687321597687D-06,   0.000000000000000D+00,   1.479476843703843D-06, &
  -2.545766189994548D-05,   0.000000000000000D+00,  -2.021814607298990D-05, &
   0.000000000000000D+00,  -2.626115330995250D-04,   0.000000000000000D+00, &
  -2.869838216749015D-04,   0.000000000000000D+00,   1.801862839340043D-04, &
   0.000000000000000D+00,   7.661800304396016D-05,   0.000000000000000D+00, &
   9.639672819885853D-06,   0.000000000000000D+00,  -5.084166315791956D-07, &
   0.000000000000000D+00,  -2.117170538322880D-06,   0.000000000000000D+00, &
  -4.104547732033803D-06,   0.000000000000000D+00,  -1.324817327179123D-06, &
   0.000000000000000D+00,   1.089746952637812D-04,   0.000000000000000D+00, &
  -1.643882898518547D-04,   0.000000000000000D+00,  -3.998324983011005D-04, &
   0.000000000000000D+00,  -9.142154656976807D-05,   0.000000000000000D+00, &
  -2.122865354391061D-05,   0.000000000000000D+00,  -6.812761638199791D-06, &
   0.000000000000000D+00,  -1.187950701750647D-06,   0.000000000000000D+00, &
   7.075593815392832D-07,   3.808113073946486D-06,   0.000000000000000D+00, &
  -4.487284770419450D-06,   0.000000000000000D+00,  -1.416596640419216D-05, &
   0.000000000000000D+00,   2.771008454995598D-04,   0.000000000000000D+00, &
   1.941285965492779D-04,   0.000000000000000D+00,  -6.724885379305015D-05, &
   0.000000000000000D+00,  -1.360437365825148D-05,   0.000000000000000D+00, &
   4.126022402795049D-06,   0.000000000000000D+00,   4.488019216725941D-06, &
   0.000000000000000D+00,   1.785597686793358D-06,   0.000000000000000D+00, &
  -6.773020811835843D-07,   0.000000000000000D+00,   4.467584078592236D-06, &
   0.000000000000000D+00,  -9.232812747115577D-06,   0.000000000000000D+00, &
  -1.372928262234713D-04,   0.000000000000000D+00,   1.415639255945486D-04, &
   0.000000000000000D+00,   1.958210878282004D-04,   0.000000000000000D+00, &
   4.831193714576229D-05,   0.000000000000000D+00,   1.259241413157673D-05, &
   0.000000000000000D+00,   8.763920570955526D-07,   0.000000000000000D+00, &
  -1.613211541747941D-06,   0.000000000000000D+00,  -3.121747230011919D-06 /)
lem(:,213) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.095396306175414D-05,   2.360121246119579D-05, &
  -5.746463750309471D-04,   1.083158220376790D-04,   1.355377635040863D-03, &
   8.834583397084594D-04,  -1.333687681766748D-03,  -4.374637514712722D-03, &
  -1.788567457398021D-03,   1.042983486914287D-02,   6.103921938853482D-03, &
  -1.584566410000184D-02,  -5.154509451397162D-03,   1.491740672934998D-02, &
  -1.851764476670635D-04,  -8.022333430292603D-03,   2.700146111399280D-03, &
   2.296692763606372D-03,  -1.709740303740492D-03,  -4.879904669751816D-04, &
   6.556702715769701D-04,   0.000000000000000D+00,  -5.095396306175622D-05, &
  -2.360121246120120D-05,   5.746463750309392D-04,  -1.083158220376708D-04, &
  -1.355377635040856D-03,  -8.834583397084706D-04,   1.333687681766746D-03, &
   4.374637514712730D-03,   1.788567457398019D-03,  -1.042983486914290D-02, &
  -6.103921938853475D-03,   1.584566410000186D-02,   5.154509451397165D-03, &
  -1.491740672934996D-02,   1.851764476670627D-04,   8.022333430292627D-03, &
  -2.700146111399289D-03,  -2.296692763606377D-03,   1.709740303740499D-03, &
   4.879904669751826D-04,  -6.556702715769770D-04,   0.000000000000000D+00, &
  -2.547698153088149D-05,   0.000000000000000D+00,  -4.788901794142275D-05, &
   0.000000000000000D+00,   2.584181831088450D-05,   0.000000000000000D+00, &
   3.117973691192439D-05,   0.000000000000000D+00,  -3.997133195127447D-05, &
   0.000000000000000D+00,  -3.310102483005611D-05,   0.000000000000000D+00, &
   5.007376663089358D-06,   0.000000000000000D+00,   5.464671961476784D-05, &
   0.000000000000000D+00,   4.177912867892635D-05,   0.000000000000000D+00, &
  -5.536901011101991D-05,   0.000000000000000D+00,  -3.456637497316476D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.014132238590368D-06, &
   0.000000000000000D+00,   1.268366585031084D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.369435824342350D-06,   0.000000000000000D+00, &
   4.992665722576519D-07,   0.000000000000000D+00,   1.757465960252415D-06, &
   0.000000000000000D+00,  -4.356136153821630D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.879424907961721D-05,   0.000000000000000D+00, &
   3.971358523619988D-06,   0.000000000000000D+00,  -2.144005340410584D-06, &
   0.000000000000000D+00,  -2.770100108234454D-05,   0.000000000000000D+00, &
   2.805438219020053D-06,   0.000000000000000D+00,  -4.598826913614632D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.819874600058017D-05, &
   0.000000000000000D+00,  -2.541947224598155D-05,   0.000000000000000D+00, &
   1.298369132562246D-06,   0.000000000000000D+00,  -2.784725350879196D-07, &
   0.000000000000000D+00,   8.399386293450081D-05,   0.000000000000000D+00, &
  -1.991974573069791D-06,   0.000000000000000D+00,  -4.615683456645783D-06, &
   0.000000000000000D+00,   9.323627749212825D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.785309057952044D-05,   0.000000000000000D+00, &
   1.096208912474865D-04,   0.000000000000000D+00,   5.429009664625525D-06, &
   0.000000000000000D+00,  -2.324271577590999D-06,   0.000000000000000D+00, &
   1.878427934055564D-06,   0.000000000000000D+00,  -1.255618362424210D-04, &
   0.000000000000000D+00,  -1.189842100736601D-04,   0.000000000000000D+00, &
   5.420776223777919D-05,   0.000000000000000D+00,   4.563545220264987D-06, &
   0.000000000000000D+00,   1.703768536307951D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.118418848785516D-05,   0.000000000000000D+00, &
  -9.843343070200176D-05,   0.000000000000000D+00,  -1.705560269291402D-04, &
   0.000000000000000D+00,   4.396566450625547D-06,   0.000000000000000D+00, &
   6.756316228345414D-07,   0.000000000000000D+00,  -7.822488541629141D-07, &
   0.000000000000000D+00,   1.009252567074195D-04,   0.000000000000000D+00, &
   2.668506216283533D-04,   0.000000000000000D+00,   1.265406834269395D-04, &
   0.000000000000000D+00,  -7.413242801174466D-05,   0.000000000000000D+00, &
  -6.667123032080223D-06,   0.000000000000000D+00,  -2.048256948683538D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.752405717711948D-05, &
   0.000000000000000D+00,  -1.079344468353159D-04,   0.000000000000000D+00, &
   1.909570075821798D-04,   0.000000000000000D+00,   1.848200975469974D-04, &
   0.000000000000000D+00,   7.379833397070354D-06,   0.000000000000000D+00, &
   2.922165303959606D-06,   0.000000000000000D+00,   1.372467856969039D-06, &
   0.000000000000000D+00,  -3.649195230708372D-05,   0.000000000000000D+00, &
  -1.408544965544367D-04,   0.000000000000000D+00,  -4.163540085443533D-04, &
   0.000000000000000D+00,  -1.194299789561205D-04,   0.000000000000000D+00, &
   4.255676658233858D-05,   0.000000000000000D+00,   3.441140760426020D-06, &
   0.000000000000000D+00,   1.119461766874832D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.975584441531584D-05,   0.000000000000000D+00, &
   1.456506571517006D-04,   0.000000000000000D+00,   2.072156258011586D-04, &
   0.000000000000000D+00,  -2.193039395594263D-04,   0.000000000000000D+00, &
  -1.392306025414869D-04,   0.000000000000000D+00,  -1.734087725101093D-05, &
   0.000000000000000D+00,  -3.616091898481279D-06,   0.000000000000000D+00, &
  -9.870119910563955D-07,   0.000000000000000D+00,  -3.134223972752195D-06, &
   0.000000000000000D+00,  -7.813143621033873D-06,   0.000000000000000D+00, &
   1.473096123214649D-04,   0.000000000000000D+00,   4.528215357095772D-04, &
   0.000000000000000D+00,   1.068732058561280D-04,   0.000000000000000D+00, &
   2.711479773285582D-06,   0.000000000000000D+00,   5.138853249964861D-06, &
   0.000000000000000D+00,   3.114490599528014D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.604921438569874D-05,   0.000000000000000D+00, &
  -3.891170172610181D-05,   0.000000000000000D+00,  -2.190138425404549D-04, &
   0.000000000000000D+00,  -2.366534266814817D-04,   0.000000000000000D+00, &
   1.509857365983602D-04,   0.000000000000000D+00,   6.599321603347863D-05, &
   0.000000000000000D+00,   8.178388822536488D-06,   0.000000000000000D+00, &
  -1.252415927060622D-06,   0.000000000000000D+00,  -3.627751345133022D-06, &
   0.000000000000000D+00,   6.625340525135288D-06,   0.000000000000000D+00, &
   5.612780247672808D-06,   0.000000000000000D+00,   8.119637107568812D-05, &
   0.000000000000000D+00,  -1.367850115810443D-04,   0.000000000000000D+00, &
  -3.344969659852710D-04,   0.000000000000000D+00,  -7.994497350209884D-05, &
   0.000000000000000D+00,  -1.993458721756003D-05,   0.000000000000000D+00, &
  -6.903085092516467D-06,   0.000000000000000D+00,  -1.128222687083501D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.011366841053364D-06, &
   0.000000000000000D+00,   8.877760952472178D-06,   0.000000000000000D+00, &
   1.005285707136693D-06,   0.000000000000000D+00,   2.321651939645110D-04, &
   0.000000000000000D+00,   1.626717578866876D-04,   0.000000000000000D+00, &
  -5.682993993194827D-05,   0.000000000000000D+00,  -1.136620880455156D-05, &
   0.000000000000000D+00,   4.939835790847869D-06,   0.000000000000000D+00, &
   4.591776034522651D-06,   0.000000000000000D+00,   4.489544595449543D-06, &
   0.000000000000000D+00,  -1.391789269241292D-06,   0.000000000000000D+00, &
  -4.992528454829840D-07,   0.000000000000000D+00,  -1.463022329491763D-06, &
   0.000000000000000D+00,  -1.072335732891505D-04,   0.000000000000000D+00, &
   1.208785461997040D-04,   0.000000000000000D+00,   1.660147854768364D-04, &
   0.000000000000000D+00,   4.230516302690644D-05,   0.000000000000000D+00, &
   1.150001908651575D-05,   0.000000000000000D+00,   1.002390225892461D-06, &
   0.000000000000000D+00,  -2.241542715737267D-06,   0.000000000000000D+00 /)
lem(:,214) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.257475653143608D-05,   2.695348841505320D-04,   6.525289855156509D-05, &
  -2.492739430694097D-04,  -2.068948886887463D-04,  -9.101090641088653D-04, &
  -7.421668586544402D-04,   1.679652225370969D-03,   4.801971259075806D-03, &
   8.467618422676120D-04,  -1.049962997457186D-02,  -3.939801647342860D-03, &
   1.375115244847151D-02,   1.325152904394976D-03,  -1.111248077208057D-02, &
   3.694159611092551D-03,   4.524097918028975D-03,  -4.117877947786314D-03, &
  -7.240045750980291D-05,   1.767360040155996D-03,  -6.962402497408928D-04, &
  -4.919838793683121D-04,  -3.257475653143651D-05,   2.695348841505373D-04, &
   6.525289855157099D-05,  -2.492739430694188D-04,  -2.068948886887445D-04, &
  -9.101090641088577D-04,  -7.421668586544487D-04,   1.679652225370963D-03, &
   4.801971259075809D-03,   8.467618422676174D-04,  -1.049962997457183D-02, &
  -3.939801647342867D-03,   1.375115244847151D-02,   1.325152904394976D-03, &
  -1.111248077208061D-02,   3.694159611092552D-03,   4.524097918028976D-03, &
  -4.117877947786312D-03,  -7.240045750980933D-05,   1.767360040155996D-03, &
  -6.962402497408895D-04,  -4.919838793683231D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.262991238774594D-04,   0.000000000000000D+00, &
  -1.497078997080426D-04,   0.000000000000000D+00,  -3.720662207880085D-06, &
   0.000000000000000D+00,   5.693067997052100D-05,   0.000000000000000D+00, &
   5.350054432300477D-05,   0.000000000000000D+00,   4.486391322066664D-05, &
   0.000000000000000D+00,  -6.260495782843133D-05,   0.000000000000000D+00, &
  -1.067921179017668D-04,   0.000000000000000D+00,   4.582723851614364D-05, &
   0.000000000000000D+00,   8.469732522255384D-05,   0.000000000000000D+00, &
   6.256509013861874D-06,  -6.706550477162236D-06,   0.000000000000000D+00, &
  -1.957309455269297D-05,   0.000000000000000D+00,   2.197203798140615D-06, &
   2.349347312393303D-05,   0.000000000000000D+00,  -2.185049270110810D-06, &
   0.000000000000000D+00,   4.627535821274050D-05,   0.000000000000000D+00, &
   4.018570766860834D-06,   0.000000000000000D+00,  -9.977798701939426D-07, &
  -4.020263889210732D-05,   0.000000000000000D+00,   7.247575562256651D-06, &
   0.000000000000000D+00,   2.329579083921264D-06,   0.000000000000000D+00, &
  -7.262678645519565D-05,   0.000000000000000D+00,  -2.653712623063101D-05, &
   0.000000000000000D+00,   4.579680978495966D-06,   0.000000000000000D+00, &
  -2.477696909299154D-06,   2.340479427376214D-05,   0.000000000000000D+00, &
  -1.412964325396883D-05,   0.000000000000000D+00,  -1.731144347911928D-05, &
   0.000000000000000D+00,   6.941770019847493D-07,   0.000000000000000D+00, &
   8.846043623256441D-05,   0.000000000000000D+00,   7.119660789184331D-05, &
   0.000000000000000D+00,  -7.891749841256093D-07,   0.000000000000000D+00, &
  -7.209397371890980D-06,   0.000000000000000D+00,   1.615425633481753D-06, &
   2.594351023523854D-05,   0.000000000000000D+00,   9.063996074708432D-06, &
   0.000000000000000D+00,   7.817888578916733D-05,   0.000000000000000D+00, &
   8.266603341730587D-06,   0.000000000000000D+00,  -1.967625314998764D-06, &
   0.000000000000000D+00,  -9.229082030473591D-05,   0.000000000000000D+00, &
  -1.046217432297973D-04,   0.000000000000000D+00,  -9.678779039761526D-05, &
   0.000000000000000D+00,   4.308874643146214D-05,   0.000000000000000D+00, &
   8.194907276697311D-06,   0.000000000000000D+00,   1.643245902081265D-06, &
  -5.425479219807692D-05,   0.000000000000000D+00,   1.038692012267267D-05, &
   0.000000000000000D+00,  -7.180460344408148D-05,   0.000000000000000D+00, &
  -1.388955009740331D-04,   0.000000000000000D+00,  -2.573183221190986D-06, &
   0.000000000000000D+00,   8.409130232809264D-07,   0.000000000000000D+00, &
   7.771190489024729D-05,   0.000000000000000D+00,   8.695313494149865D-05, &
   0.000000000000000D+00,   2.134902067134996D-04,   0.000000000000000D+00, &
   1.076899368698198D-04,   0.000000000000000D+00,  -6.077405222581878D-05, &
   0.000000000000000D+00,  -9.566556391863495D-06,   0.000000000000000D+00, &
  -3.974778176041928D-06,   3.661328548813057D-05,   0.000000000000000D+00, &
  -2.062186227086453D-05,   0.000000000000000D+00,  -7.890020747787717D-05, &
   0.000000000000000D+00,   1.537473234836114D-04,   0.000000000000000D+00, &
   1.583128580887407D-04,   0.000000000000000D+00,   1.197162020149809D-05, &
   0.000000000000000D+00,   3.557151239793047D-06,   0.000000000000000D+00, &
  -4.518661444739779D-05,   0.000000000000000D+00,  -3.726641646135915D-05, &
   0.000000000000000D+00,  -1.142301393943694D-04,   0.000000000000000D+00, &
  -3.440212554461401D-04,   0.000000000000000D+00,  -1.055770284147542D-04, &
   0.000000000000000D+00,   3.474549507112611D-05,   0.000000000000000D+00, &
   3.553417873926685D-06,   0.000000000000000D+00,  -1.744510434676694D-07, &
  -8.730512201292909D-06,   0.000000000000000D+00,   1.062483004425175D-05, &
   0.000000000000000D+00,   1.033592144533433D-04,   0.000000000000000D+00, &
   1.666737861777551D-04,   0.000000000000000D+00,  -1.832301472272907D-04, &
   0.000000000000000D+00,  -1.223973133559024D-04,   0.000000000000000D+00, &
  -1.793948318034834D-05,   0.000000000000000D+00,  -4.394630126714863D-06, &
   0.000000000000000D+00,   1.507352292890799D-05,   0.000000000000000D+00, &
   4.919694608494048D-06,   0.000000000000000D+00,  -7.878795778994625D-06, &
   0.000000000000000D+00,   1.236249262799109D-04,   0.000000000000000D+00, &
   3.839807074465816D-04,   0.000000000000000D+00,   9.625939889622009D-05, &
   0.000000000000000D+00,   4.644124048983328D-06,   0.000000000000000D+00, &
   6.543431326640990D-06,   0.000000000000000D+00,   6.932589264324128D-06, &
  -1.231749710625376D-06,   0.000000000000000D+00,   1.034908991575972D-06, &
   0.000000000000000D+00,  -2.143988698455021D-05,   0.000000000000000D+00, &
  -1.761323965525997D-04,   0.000000000000000D+00,  -1.971775277608934D-04, &
   0.000000000000000D+00,   1.284719113879242D-04,   0.000000000000000D+00, &
   5.802353269977631D-05,   0.000000000000000D+00,   7.472944393058228D-06, &
   0.000000000000000D+00,  -1.899058947242322D-06,   0.000000000000000D+00, &
  -1.962774023063498D-06,   0.000000000000000D+00,   8.114712180501485D-07, &
   0.000000000000000D+00,   1.064703336679599D-05,   0.000000000000000D+00, &
   6.790267717088054D-05,   0.000000000000000D+00,  -1.182709712860696D-04, &
   0.000000000000000D+00,  -2.888577196739639D-04,   0.000000000000000D+00, &
  -7.218069777993009D-05,   0.000000000000000D+00,  -1.975694475447415D-05, &
   0.000000000000000D+00,  -7.645845313737331D-06,   0.000000000000000D+00, &
  -3.546733380360204D-06,   6.866118655259874D-07,   0.000000000000000D+00, &
  -2.525912209173275D-06,   0.000000000000000D+00,   2.268952343144815D-07, &
   0.000000000000000D+00,   1.156847109546282D-06,   0.000000000000000D+00, &
   1.947694651041991D-04,   0.000000000000000D+00,   1.378130624497006D-04, &
   0.000000000000000D+00,  -4.844936342285761D-05,   0.000000000000000D+00, &
  -8.860357833588340D-06,   0.000000000000000D+00,   5.140129623405627D-06, &
   0.000000000000000D+00,   5.077198121370472D-06,   0.000000000000000D+00, &
  -8.855236302127505D-08,   0.000000000000000D+00,   4.582657704107381D-07, &
   0.000000000000000D+00,  -4.218512049495033D-06,   0.000000000000000D+00, &
  -7.665486913469934D-06,   0.000000000000000D+00,  -8.994085021769675D-05, &
   0.000000000000000D+00,   1.063216201810778D-04,   0.000000000000000D+00, &
   1.452030815058162D-04,   0.000000000000000D+00,   3.826785287664572D-05, &
   0.000000000000000D+00,   1.115571044327457D-05,   0.000000000000000D+00, &
   1.057470944427196D-06,   0.000000000000000D+00,  -5.651786229106209D-06 /)
lem(:,215) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.173360688980697D-04,  -6.609413383283372D-04, &
  -2.992332568778610D-04,   6.839264855136833D-04,   1.629137520444256D-03, &
   1.041552544853057D-03,  -1.711568457366441D-03,  -4.603518018015632D-03, &
  -7.404497546995001D-04,   8.255024191009616D-03,   2.219259199448546D-03, &
  -9.689907063801698D-03,   9.624844334119561D-04,   6.416727241372954D-03, &
  -4.355528877968218D-03,  -8.063511983519508D-04,   3.172560902480730D-03, &
  -1.686768608038608D-03,  -5.732347633808498D-04,   1.096139797748001D-03, &
  -2.877328041039578D-04,   0.000000000000000D+00,   1.173360688980753D-04, &
   6.609413383283319D-04,   2.992332568778570D-04,  -6.839264855136785D-04, &
  -1.629137520444252D-03,  -1.041552544853065D-03,   1.711568457366444D-03, &
   4.603518018015633D-03,   7.404497546994994D-04,  -8.255024191009588D-03, &
  -2.219259199448540D-03,   9.689907063801691D-03,  -9.624844334119541D-04, &
  -6.416727241372964D-03,   4.355528877968203D-03,   8.063511983519532D-04, &
  -3.172560902480728D-03,   1.686768608038605D-03,   5.732347633808513D-04, &
  -1.096139797748002D-03,   2.877328041039595D-04,   0.000000000000000D+00, &
   5.866803444903738D-05,   0.000000000000000D+00,   1.493416786990422D-04, &
   0.000000000000000D+00,   1.321012362013364D-04,   0.000000000000000D+00, &
   4.360015806510472D-05,   0.000000000000000D+00,  -9.230654216940686D-05, &
   0.000000000000000D+00,  -1.530622547050845D-04,   0.000000000000000D+00, &
   5.075362680775138D-06,   0.000000000000000D+00,   2.107621811320952D-04, &
   0.000000000000000D+00,   2.528902105679230D-05,   0.000000000000000D+00, &
  -1.617846802349981D-04,   0.000000000000000D+00,   5.895457075906551D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.638115367604468D-06, &
   0.000000000000000D+00,  -9.651939404640038D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.056733353543195D-06,   0.000000000000000D+00, &
  -3.429486172105923D-06,   0.000000000000000D+00,   1.884707964240037D-05, &
   0.000000000000000D+00,   8.589458316404736D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.951346217322461D-06,   0.000000000000000D+00, &
   8.232842713057483D-06,   0.000000000000000D+00,   6.171240914177157D-06, &
   0.000000000000000D+00,  -2.373941870972095D-05,   0.000000000000000D+00, &
  -2.414513991952286D-05,   0.000000000000000D+00,   1.366235924481011D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.143645459053561D-06, &
   0.000000000000000D+00,  -1.472246244238813D-05,   0.000000000000000D+00, &
  -1.851578303289465D-05,   0.000000000000000D+00,   5.069017900853474D-07, &
   0.000000000000000D+00,   2.520416295388321D-05,   0.000000000000000D+00, &
   5.124665001387524D-05,   0.000000000000000D+00,   5.385401295663196D-06, &
   0.000000000000000D+00,  -1.076643368055745D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.393427082728741D-06,   0.000000000000000D+00, &
   9.110223392120265D-06,   0.000000000000000D+00,   6.558511972732672D-05, &
   0.000000000000000D+00,   1.454049514363367D-05,   0.000000000000000D+00, &
  -5.865335201792831D-06,   0.000000000000000D+00,  -3.022446717816985D-05, &
   0.000000000000000D+00,  -7.150676800602072D-05,   0.000000000000000D+00, &
  -8.090361171637975D-05,   0.000000000000000D+00,   3.242836598611326D-05, &
   0.000000000000000D+00,   1.458312918828840D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.498817141160859D-06,   0.000000000000000D+00, &
   1.066529053966638D-05,   0.000000000000000D+00,  -5.833797489541550D-05, &
   0.000000000000000D+00,  -1.223099796146540D-04,   0.000000000000000D+00, &
  -9.748187904112846D-06,   0.000000000000000D+00,   2.837164422956429D-06, &
   0.000000000000000D+00,   3.118834215924510D-05,   0.000000000000000D+00, &
   6.210723508397070D-05,   0.000000000000000D+00,   1.706589221904144D-04, &
   0.000000000000000D+00,   9.626497960193815D-05,   0.000000000000000D+00, &
  -5.093820055768604D-05,   0.000000000000000D+00,  -1.341094035385173D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.823264703106282D-06, &
   0.000000000000000D+00,  -2.356837253144727D-05,   0.000000000000000D+00, &
  -6.431671072738877D-05,   0.000000000000000D+00,   1.297938000679667D-04, &
   0.000000000000000D+00,   1.437979507008840D-04,   0.000000000000000D+00, &
   1.704998491137096D-05,   0.000000000000000D+00,   8.006472115757415D-06, &
   0.000000000000000D+00,  -1.921710113325977D-05,   0.000000000000000D+00, &
  -3.239099120009514D-05,   0.000000000000000D+00,  -9.331730369472067D-05, &
   0.000000000000000D+00,  -2.905492879264727D-04,   0.000000000000000D+00, &
  -9.773265426183524D-05,   0.000000000000000D+00,   2.792076819872266D-05, &
   0.000000000000000D+00,   2.958272690541992D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.558494888797101D-06,   0.000000000000000D+00, &
   1.742407276165966D-05,   0.000000000000000D+00,   8.630914696820816D-05, &
   0.000000000000000D+00,   1.399311820285290D-04,   0.000000000000000D+00, &
  -1.589607937749792D-04,   0.000000000000000D+00,  -1.126154981060877D-04, &
   0.000000000000000D+00,  -2.053736329845054D-05,   0.000000000000000D+00, &
  -1.144648700214586D-05,   0.000000000000000D+00,   4.945208457336394D-06, &
   0.000000000000000D+00,   1.043751607224640D-05,   0.000000000000000D+00, &
  -1.319300027889045D-06,   0.000000000000000D+00,   1.068732058561280D-04, &
   0.000000000000000D+00,   3.338746168978982D-04,   0.000000000000000D+00, &
   9.064795466190599D-05,   0.000000000000000D+00,   8.486109857450329D-06, &
   0.000000000000000D+00,   9.232926668137844D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.589401038575375D-06,   0.000000000000000D+00, &
  -6.159202151342601D-06,   0.000000000000000D+00,  -1.998936291333156D-05, &
   0.000000000000000D+00,  -1.497684398357018D-04,   0.000000000000000D+00, &
  -1.693310256391204D-04,   0.000000000000000D+00,   1.127727838228724D-04, &
   0.000000000000000D+00,   5.334722153403152D-05,   0.000000000000000D+00, &
   7.989761081391988D-06,   0.000000000000000D+00,  -2.208453303646228D-06, &
   0.000000000000000D+00,   4.583643317914154D-07,   0.000000000000000D+00, &
  -3.163321513287557D-06,   0.000000000000000D+00,   3.339189859353182D-06, &
   0.000000000000000D+00,   5.298053971740659D-05,   0.000000000000000D+00, &
  -1.058164222268487D-04,   0.000000000000000D+00,  -2.556627380570721D-04, &
   0.000000000000000D+00,  -6.877454177865064D-05,   0.000000000000000D+00, &
  -2.120688543818597D-05,   0.000000000000000D+00,  -7.963258965795666D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.989867081898749D-07, &
   0.000000000000000D+00,   1.515910438742415D-06,   0.000000000000000D+00, &
   1.330304345992742D-07,   0.000000000000000D+00,   2.826557100744237D-06, &
   0.000000000000000D+00,   1.693125079148792D-04,   0.000000000000000D+00, &
   1.199951378303614D-04,   0.000000000000000D+00,  -4.219416124876756D-05, &
   0.000000000000000D+00,  -7.596346842355203D-06,   0.000000000000000D+00, &
   5.611126285023065D-06,   0.000000000000000D+00,   1.352622264849873D-05, &
   0.000000000000000D+00,  -3.375708056928978D-07,   0.000000000000000D+00, &
   1.142658301507520D-06,   0.000000000000000D+00,   8.587459453778578D-07, &
   0.000000000000000D+00,  -5.233609340894733D-06,   0.000000000000000D+00, &
  -7.395539384780862D-05,   0.000000000000000D+00,   9.659672784063251D-05, &
   0.000000000000000D+00,   1.302865599857362D-04,   0.000000000000000D+00, &
   3.660945751811957D-05,   0.000000000000000D+00,   1.116648893115864D-05, &
   0.000000000000000D+00,  -6.298261314281319D-07,   0.000000000000000D+00 /)
lem(:,216) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.424581332885196D-04,   3.158590035194950D-04,   3.403048329880564D-04, &
   2.257628562859129D-05,  -7.171611115421591D-04,  -1.392236719632856D-03, &
  -9.500043392147006D-04,   1.253121394619028D-03,   3.722511135899826D-03, &
   1.152178483667406D-03,  -5.494104869364397D-03,  -1.758780187506689D-03, &
   5.602267681838459D-03,  -1.334317587708677D-03,  -2.746306232689293D-03, &
   3.103216357703009D-03,  -1.051628279576236D-03,  -1.123326251335282D-03, &
   1.751467450035328D-03,  -7.080690444965380D-04,  -4.799360274376915D-04, &
   6.624040138796491D-04,   3.424581332885247D-04,   3.158590035195030D-04, &
   3.403048329880558D-04,   2.257628562859365D-05,  -7.171611115421565D-04, &
  -1.392236719632854D-03,  -9.500043392147020D-04,   1.253121394619028D-03, &
   3.722511135899834D-03,   1.152178483667412D-03,  -5.494104869364407D-03, &
  -1.758780187506691D-03,   5.602267681838474D-03,  -1.334317587708678D-03, &
  -2.746306232689291D-03,   3.103216357703008D-03,  -1.051628279576232D-03, &
  -1.123326251335287D-03,   1.751467450035326D-03,  -7.080690444965260D-04, &
  -4.799360274376977D-04,   6.624040138796404D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.258067823253165D-04,   0.000000000000000D+00, &
   1.755903634512060D-04,   0.000000000000000D+00,  -3.909471482138662D-04, &
   0.000000000000000D+00,  -2.317178303546575D-04,   0.000000000000000D+00, &
   3.271664643079432D-04,   0.000000000000000D+00,   2.953532946867656D-04, &
   0.000000000000000D+00,  -3.123330083361087D-04,   0.000000000000000D+00, &
  -2.488654932657664D-04,   0.000000000000000D+00,   2.447546006549075D-04, &
   0.000000000000000D+00,   8.598214348237853D-05,   0.000000000000000D+00, &
  -1.597702276760915D-05,  -1.701575048398938D-05,   0.000000000000000D+00, &
   3.204547151568527D-06,   0.000000000000000D+00,  -1.424084401136962D-05, &
   2.114624169162339D-05,   0.000000000000000D+00,   7.871332497323119D-06, &
   0.000000000000000D+00,   1.116118514600174D-05,   0.000000000000000D+00, &
   1.286202457551652D-05,   0.000000000000000D+00,   1.689569715498388D-05, &
  -1.057120114535514D-05,   0.000000000000000D+00,  -1.172663827410010D-05, &
   0.000000000000000D+00,   7.112146232439420D-06,   0.000000000000000D+00, &
  -2.198321241382021D-05,   0.000000000000000D+00,  -1.837615632057181D-05, &
   0.000000000000000D+00,  -2.254240913236837D-05,   0.000000000000000D+00, &
   5.555474358055264D-06,  -1.225562914440495D-05,   0.000000000000000D+00, &
   6.623072633241856D-06,   0.000000000000000D+00,  -9.253317431204150D-06, &
   0.000000000000000D+00,  -2.296812070071748D-05,   0.000000000000000D+00, &
   1.654545274525837D-05,   0.000000000000000D+00,   2.280171919988297D-05, &
   0.000000000000000D+00,   4.311852743007091D-05,   0.000000000000000D+00, &
   8.188755649439833D-06,   0.000000000000000D+00,  -2.516624274203225D-05, &
   3.253273958765910D-05,   0.000000000000000D+00,   6.955396264001509D-06, &
   0.000000000000000D+00,   5.061225751834691D-06,   0.000000000000000D+00, &
   5.768666766168369D-05,   0.000000000000000D+00,   2.412636819682806D-05, &
   0.000000000000000D+00,  -4.191015896549517D-06,   0.000000000000000D+00, &
  -2.351144805455855D-05,   0.000000000000000D+00,  -5.879932830718967D-05, &
   0.000000000000000D+00,  -7.259342262414246D-05,   0.000000000000000D+00, &
   2.605799741837081D-05,   0.000000000000000D+00,   3.242392302065262D-05, &
  -3.578807803438344D-05,   0.000000000000000D+00,  -1.364401620739649D-05, &
   0.000000000000000D+00,   6.308988330746172D-06,   0.000000000000000D+00, &
  -4.818482794943219D-05,   0.000000000000000D+00,  -1.135979174364575D-04, &
   0.000000000000000D+00,  -2.180318481995559D-05,   0.000000000000000D+00, &
   9.117473031361262D-07,   0.000000000000000D+00,   1.787194398963524D-05, &
   0.000000000000000D+00,   5.015390938027009D-05,   0.000000000000000D+00, &
   1.454222719463021D-04,   0.000000000000000D+00,   9.079881150681305D-05, &
   0.000000000000000D+00,  -4.393485030416834D-05,   0.000000000000000D+00, &
  -3.004665106305449D-05,   2.243947602784934D-05,   0.000000000000000D+00, &
   7.578573998236979D-06,   0.000000000000000D+00,  -1.479781010773593D-05, &
   0.000000000000000D+00,  -5.195962116696766D-05,   0.000000000000000D+00, &
   1.130907064849989D-04,   0.000000000000000D+00,   1.382394019477321D-04, &
   0.000000000000000D+00,   2.875726274368140D-05,   0.000000000000000D+00, &
  -3.137187734750344D-06,   0.000000000000000D+00,  -7.209450732629709D-06, &
   0.000000000000000D+00,  -2.368585808469808D-05,   0.000000000000000D+00, &
  -8.029276775973197D-05,   0.000000000000000D+00,  -2.557661449492186D-04, &
   0.000000000000000D+00,  -9.650494599702847D-05,   0.000000000000000D+00, &
   2.206557633069135D-05,   0.000000000000000D+00,   8.750023798300625D-06, &
  -7.333545430878997D-06,   0.000000000000000D+00,  -9.930762624844007D-07, &
   0.000000000000000D+00,   1.235869718143887D-05,   0.000000000000000D+00, &
   7.070687944454380D-05,   0.000000000000000D+00,   1.195751246510811D-04, &
   0.000000000000000D+00,  -1.418650160761575D-04,   0.000000000000000D+00, &
  -1.109231568857014D-04,   0.000000000000000D+00,  -2.704078068997872D-05, &
   0.000000000000000D+00,   2.456338105155462D-06,   0.000000000000000D+00, &
  -2.536412436036162D-07,   0.000000000000000D+00,   3.888123032117673D-06, &
   0.000000000000000D+00,  -5.173454620707978D-07,   0.000000000000000D+00, &
   9.625939889622009D-05,   0.000000000000000D+00,   3.014567154522143D-04, &
   0.000000000000000D+00,   9.154909674363392D-05,   0.000000000000000D+00, &
   1.285755257664805D-05,   0.000000000000000D+00,   2.064575554666510D-05, &
   7.074576633694918D-07,   0.000000000000000D+00,   1.711193445453078D-07, &
   0.000000000000000D+00,  -5.488832023553196D-06,   0.000000000000000D+00, &
  -1.843787817598085D-05,   0.000000000000000D+00,  -1.296204395388156D-04, &
   0.000000000000000D+00,  -1.492249206558482D-04,   0.000000000000000D+00, &
   1.021311275934576D-04,   0.000000000000000D+00,   5.289710649932278D-05, &
   0.000000000000000D+00,   6.819106914820439D-06,   0.000000000000000D+00, &
  -5.495574918594667D-07,   0.000000000000000D+00,   1.183930908868187D-06, &
   0.000000000000000D+00,   1.566990787815751D-06,   0.000000000000000D+00, &
   4.671082792089535D-06,   0.000000000000000D+00,   4.370817558500269D-05, &
   0.000000000000000D+00,  -9.816427589679462D-05,   0.000000000000000D+00, &
  -2.356339431931522D-04,   0.000000000000000D+00,  -6.934025407019054D-05, &
   0.000000000000000D+00,  -2.403986152790932D-05,   0.000000000000000D+00, &
  -2.390899177510547D-05,   1.240195714877552D-07,   0.000000000000000D+00, &
  -5.676366728300423D-07,   0.000000000000000D+00,   1.637349874191551D-06, &
   0.000000000000000D+00,   2.500194509299865D-06,   0.000000000000000D+00, &
   5.724424173416880D-06,   0.000000000000000D+00,   1.508206533927301D-04, &
   0.000000000000000D+00,   1.075681723197740D-04,   0.000000000000000D+00, &
  -3.839383760922462D-05,   0.000000000000000D+00,  -6.554100492808529D-06, &
   0.000000000000000D+00,   8.211144341113336D-06,   0.000000000000000D+00, &
  -1.260364167026555D-08,   0.000000000000000D+00,  -1.550879591867370D-07, &
   0.000000000000000D+00,  -1.209520689257602D-06,   0.000000000000000D+00, &
  -6.303753208871602D-07,   0.000000000000000D+00,  -6.018933892275437D-06, &
   0.000000000000000D+00,  -6.275766476266933D-05,   0.000000000000000D+00, &
   9.109704055820318D-05,   0.000000000000000D+00,   1.219871320559690D-04, &
   0.000000000000000D+00,   3.638809044638110D-05,   0.000000000000000D+00, &
   1.332036273777871D-05,   0.000000000000000D+00,   2.010091160529620D-06 /)
lem(:,217) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.295797912724219D-04,  -6.461141809150894D-04, &
  -2.566977299766853D-04,   3.810730891905964D-04,   1.090394567498388D-03, &
   1.168489395598972D-03,  -2.458539116041422D-04,  -2.500864909347941D-03, &
  -1.692280717114779D-03,   3.076023940659708D-03,   1.699881386372377D-03, &
  -2.908517241598819D-03,   6.852007996960281D-04,   8.382952305789160D-04, &
  -1.463149839845815D-03,   1.208639821426812D-03,  -2.076966966511556D-04, &
  -7.513649469643529D-04,   9.392006425669407D-04,  -3.391183764305329D-04, &
  -3.058337598691269D-04,   0.000000000000000D+00,   4.295797912724229D-04, &
   6.461141809150947D-04,   2.566977299766828D-04,  -3.810730891905993D-04, &
  -1.090394567498387D-03,  -1.168489395598971D-03,   2.458539116041454D-04, &
   2.500864909347944D-03,   1.692280717114781D-03,  -3.076023940659712D-03, &
  -1.699881386372376D-03,   2.908517241598826D-03,  -6.852007996960327D-04, &
  -8.382952305789139D-04,   1.463149839845819D-03,  -1.208639821426814D-03, &
   2.076966966511542D-04,   7.513649469643494D-04,  -9.392006425669384D-04, &
   3.391183764305354D-04,   3.058337598691267D-04,   0.000000000000000D+00, &
   2.147898956362085D-04,   0.000000000000000D+00,  -9.164405978498039D-06, &
   0.000000000000000D+00,  -7.854454522154327D-04,   0.000000000000000D+00, &
   2.423881049176467D-04,   0.000000000000000D+00,   1.214048040208130D-03, &
   0.000000000000000D+00,  -1.359379367990366D-03,   0.000000000000000D+00, &
  -5.392827692268255D-07,   0.000000000000000D+00,   8.217263041463989D-04, &
   0.000000000000000D+00,  -2.162547507299201D-04,   0.000000000000000D+00, &
  -3.223853516703373D-04,   0.000000000000000D+00,   1.416607735831439D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.698063221500720D-05, &
   0.000000000000000D+00,  -4.252499468933845D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.790093825565050D-06,   0.000000000000000D+00, &
   1.489316182079550D-05,   0.000000000000000D+00,   5.427511399409951D-06, &
   0.000000000000000D+00,   1.054680964861937D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.782024974574861D-06,   0.000000000000000D+00, &
  -8.233410410741274D-06,   0.000000000000000D+00,   1.652449272553068D-05, &
   0.000000000000000D+00,  -5.894715114653533D-06,   0.000000000000000D+00, &
  -1.475819223150208D-05,   0.000000000000000D+00,  -2.242114376126916D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.519131021512736D-06, &
   0.000000000000000D+00,   3.678377981877853D-06,   0.000000000000000D+00, &
  -9.289457057230361D-06,   0.000000000000000D+00,  -5.167700436190973D-05, &
   0.000000000000000D+00,   5.382310492317744D-06,   0.000000000000000D+00, &
   1.618024696935191D-05,   0.000000000000000D+00,   3.809950946690496D-05, &
   0.000000000000000D+00,   1.747694842112098D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.410618463298417D-06,   0.000000000000000D+00, &
   4.100113818477097D-06,   0.000000000000000D+00,   5.075437303649706D-06, &
   0.000000000000000D+00,   5.533644669235109D-05,   0.000000000000000D+00, &
   5.534760220403144D-05,   0.000000000000000D+00,  -5.591929834278601D-06, &
   0.000000000000000D+00,  -1.737798595869266D-05,   0.000000000000000D+00, &
  -4.860688186348680D-05,   0.000000000000000D+00,  -7.027994851623639D-05, &
   0.000000000000000D+00,   1.590335713566152D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.030913676424800D-05,   0.000000000000000D+00, &
  -8.176396201858152D-06,   0.000000000000000D+00,   4.458250584561136D-06, &
   0.000000000000000D+00,  -4.187999851068141D-05,   0.000000000000000D+00, &
  -1.156082781559342D-04,   0.000000000000000D+00,  -4.998245383112602D-05, &
   0.000000000000000D+00,   7.210032253175086D-06,   0.000000000000000D+00, &
   1.646225698333917D-05,   0.000000000000000D+00,   4.244061300994341D-05, &
   0.000000000000000D+00,   1.281272878793174D-04,   0.000000000000000D+00, &
   9.415281496123268D-05,   0.000000000000000D+00,  -3.761798973589609D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.012843331671461D-06, &
   0.000000000000000D+00,   5.122778952825531D-06,   0.000000000000000D+00, &
  -1.215900843371171D-05,   0.000000000000000D+00,  -4.408537227763391D-05, &
   0.000000000000000D+00,   1.032273287523155D-04,   0.000000000000000D+00, &
   1.465248327318462D-04,   0.000000000000000D+00,   6.551257975941344D-05, &
   0.000000000000000D+00,  -6.289906059341867D-06,   0.000000000000000D+00, &
  -1.051928810151977D-05,   0.000000000000000D+00,  -2.228094418100337D-05, &
   0.000000000000000D+00,  -7.200405318426007D-05,   0.000000000000000D+00, &
  -2.359248073035503D-04,   0.000000000000000D+00,  -1.038343605486562D-04, &
   0.000000000000000D+00,   1.280909088306866D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.746154660200734D-07,   0.000000000000000D+00, &
  -1.400125768107455D-06,   0.000000000000000D+00,   9.801684529495731D-06, &
   0.000000000000000D+00,   6.103964919596444D-05,   0.000000000000000D+00, &
   1.072748116923373D-04,   0.000000000000000D+00,  -1.340961764683705D-04, &
   0.000000000000000D+00,  -1.196141065231320D-04,   0.000000000000000D+00, &
  -6.855352139018906D-05,   0.000000000000000D+00,   2.615571806290833D-06, &
   0.000000000000000D+00,   3.999106719155770D-06,   0.000000000000000D+00, &
   5.547708881753694D-06,   0.000000000000000D+00,   2.711479773285582D-06, &
   0.000000000000000D+00,   9.064795466190599D-05,   0.000000000000000D+00, &
   2.862422964490931D-04,   0.000000000000000D+00,   9.936167698424896D-05, &
   0.000000000000000D+00,   2.507480111740242D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.195629089752364D-07,   0.000000000000000D+00, &
   5.195401226504020D-07,   0.000000000000000D+00,  -2.923145974306833D-06, &
   0.000000000000000D+00,  -1.690007036967001D-05,   0.000000000000000D+00, &
  -1.171643056128585D-04,   0.000000000000000D+00,  -1.370636879900869D-04, &
   0.000000000000000D+00,   9.818689031579365D-05,   0.000000000000000D+00, &
   5.784862683384989D-05,   0.000000000000000D+00,   2.663879005768372D-05, &
   0.000000000000000D+00,  -2.571057723196095D-07,   0.000000000000000D+00, &
  -1.169376055618446D-06,   0.000000000000000D+00,   1.549282135367454D-07, &
   0.000000000000000D+00,   3.579830355404934D-06,   0.000000000000000D+00, &
   3.560670187369491D-05,   0.000000000000000D+00,  -9.617576364475661D-05, &
   0.000000000000000D+00,  -2.277401053755338D-04,   0.000000000000000D+00, &
  -7.534943546337623D-05,   0.000000000000000D+00,  -2.847339111973922D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.468617768959769D-07, &
   0.000000000000000D+00,  -2.853951078342324D-07,   0.000000000000000D+00, &
  -5.227861652247807D-07,   0.000000000000000D+00,   1.347805039494620D-06, &
   0.000000000000000D+00,   7.966704685944969D-06,   0.000000000000000D+00, &
   1.401607530482283D-04,   0.000000000000000D+00,   9.965063603376556D-05, &
   0.000000000000000D+00,  -3.662205816775150D-05,   0.000000000000000D+00, &
  -8.561941623439513D-06,   0.000000000000000D+00,   2.032453319181632D-05, &
   0.000000000000000D+00,  -7.506531979905188D-08,   0.000000000000000D+00, &
   4.524905174146304D-07,   0.000000000000000D+00,  -3.825531569709320D-07, &
   0.000000000000000D+00,  -1.140163236766463D-06,   0.000000000000000D+00, &
  -6.628812149978505D-06,   0.000000000000000D+00,  -5.349697568058300D-05, &
   0.000000000000000D+00,   9.047594370235093D-05,   0.000000000000000D+00, &
   1.192636133848721D-04,   0.000000000000000D+00,   3.929491369132923D-05, &
   0.000000000000000D+00,   8.754515179310236D-06,   0.000000000000000D+00 /)
lem(:,218) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.575599178738652D-04,   4.761473101396069D-04,   2.653472616643543D-04, &
   2.958708862773525D-04,   1.016415041515241D-04,  -7.461830739188064D-04, &
  -1.193688078883620D-03,  -5.463023124651737D-04,   1.404837616180511D-03, &
   2.025842427991905D-03,  -1.295452477990926D-03,  -1.594988586623455D-03, &
   1.365329933243775D-03,  -1.280696987715623D-04,  -2.726108041888042D-04, &
   4.277199091813351D-04,  -6.383538457838219D-04,   5.079160379908040D-04, &
  -2.315693979349946D-05,  -4.288620481385368D-04,   5.030442269089693D-04, &
  -1.882722782977093D-04,   2.575599178738762D-04,   4.761473101396064D-04, &
   2.653472616643547D-04,   2.958708862773566D-04,   1.016415041515261D-04, &
  -7.461830739188163D-04,  -1.193688078883621D-03,  -5.463023124651638D-04, &
   1.404837616180509D-03,   2.025842427991901D-03,  -1.295452477990929D-03, &
  -1.594988586623452D-03,   1.365329933243779D-03,  -1.280696987715649D-04, &
  -2.726108041888038D-04,   4.277199091813396D-04,  -6.383538457838257D-04, &
   5.079160379908019D-04,  -2.315693979349490D-05,  -4.288620481385338D-04, &
   5.030442269089660D-04,  -1.882722782977145D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.372702786864754D-05,   0.000000000000000D+00, &
  -6.601452253661014D-04,   0.000000000000000D+00,   6.846333556825368D-04, &
   0.000000000000000D+00,   1.410034728914574D-03,   0.000000000000000D+00, &
  -3.688635381472540D-03,   0.000000000000000D+00,   3.844449023820919D-03, &
   0.000000000000000D+00,  -1.795835060413011D-03,   0.000000000000000D+00, &
  -4.764293969421697D-04,   0.000000000000000D+00,   8.338003447535647D-04, &
   0.000000000000000D+00,  -1.021873994958485D-04,   0.000000000000000D+00, &
  -8.243709415584547D-06,  -1.148482593795079D-05,   0.000000000000000D+00, &
  -2.811316287878914D-06,   0.000000000000000D+00,   3.605742131132795D-06, &
   8.463193124996781D-06,   0.000000000000000D+00,   3.896197085439298D-06, &
   0.000000000000000D+00,   7.226982781108779D-06,   0.000000000000000D+00, &
   3.331286468060782D-06,   0.000000000000000D+00,   2.383137572122601D-05, &
   2.327514856933918D-06,   0.000000000000000D+00,  -1.654701207111207D-06, &
   0.000000000000000D+00,  -5.880095697544164D-06,   0.000000000000000D+00, &
  -2.375890389187176D-06,   0.000000000000000D+00,  -7.163766486231442D-06, &
   0.000000000000000D+00,  -1.467737521093410D-05,   0.000000000000000D+00, &
  -5.930753807795969D-05,  -9.687344821097332D-06,   0.000000000000000D+00, &
  -3.827105495344458D-06,   0.000000000000000D+00,   2.800393570388285D-06, &
   0.000000000000000D+00,  -1.117123432387653D-06,   0.000000000000000D+00, &
  -7.611565212819528D-06,   0.000000000000000D+00,   4.946200842053047D-06, &
   0.000000000000000D+00,   1.411595782217473D-05,   0.000000000000000D+00, &
   4.031971149941456D-05,   0.000000000000000D+00,   3.905095629555748D-05, &
   9.430169373391873D-06,   0.000000000000000D+00,   8.563572731788466D-06, &
   0.000000000000000D+00,   5.245448267263974D-06,   0.000000000000000D+00, &
   1.566447577307091D-06,   0.000000000000000D+00,   5.053722015778618D-05, &
   0.000000000000000D+00,   1.280244359106400D-05,   0.000000000000000D+00, &
  -8.626109610814325D-07,   0.000000000000000D+00,  -1.423293910335406D-05, &
   0.000000000000000D+00,  -4.325227575227905D-05,   0.000000000000000D+00, &
  -7.517676487908392D-05,   0.000000000000000D+00,   4.564921275036527D-05, &
  -6.986789569097378D-06,   0.000000000000000D+00,  -8.551898450182302D-06, &
   0.000000000000000D+00,  -8.845153365873782D-06,   0.000000000000000D+00, &
   2.846423405964968D-06,   0.000000000000000D+00,  -3.805929385089781D-05, &
   0.000000000000000D+00,  -1.341586635520609D-04,   0.000000000000000D+00, &
  -9.090830329679717D-06,   0.000000000000000D+00,   5.923051482652621D-08, &
   0.000000000000000D+00,   1.245937352806301D-05,   0.000000000000000D+00, &
   3.715043158846705D-05,   0.000000000000000D+00,   1.222227725848170D-04, &
   0.000000000000000D+00,   1.057144386116536D-04,   0.000000000000000D+00, &
  -1.004698852520413D-04,   4.776649320619277D-06,   0.000000000000000D+00, &
   3.989182485669879D-06,   0.000000000000000D+00,   4.522064506163154D-06, &
   0.000000000000000D+00,  -8.802071815476981D-06,   0.000000000000000D+00, &
  -3.963893556576631D-05,   0.000000000000000D+00,   1.029569308917073D-04, &
   0.000000000000000D+00,   1.828975391752859D-04,   0.000000000000000D+00, &
   2.797334482781059D-06,   0.000000000000000D+00,  -4.498794304493893D-07, &
   0.000000000000000D+00,  -7.465139723148578D-06,   0.000000000000000D+00, &
  -1.979698309666543D-05,   0.000000000000000D+00,  -6.868408465981140D-05, &
   0.000000000000000D+00,  -2.338814845151818D-04,   0.000000000000000D+00, &
  -1.210865768925074D-04,   0.000000000000000D+00,   4.911876503336968D-05, &
  -2.125936736814540D-06,   0.000000000000000D+00,  -1.069148805570843D-07, &
   0.000000000000000D+00,   4.280826676808669D-07,   0.000000000000000D+00, &
   8.568537074780326D-06,   0.000000000000000D+00,   5.542887308077155D-05, &
   0.000000000000000D+00,   1.006891004333143D-04,   0.000000000000000D+00, &
  -1.378320512732138D-04,   0.000000000000000D+00,  -1.448812605711903D-04, &
   0.000000000000000D+00,  -5.928668361689146D-09,   0.000000000000000D+00, &
  -1.501485425853169D-07,   0.000000000000000D+00,   2.553123586661460D-06, &
   0.000000000000000D+00,   5.178724083693248D-06,   0.000000000000000D+00, &
   4.644124048983328D-06,   0.000000000000000D+00,   9.154909674363392D-05, &
   0.000000000000000D+00,   2.905367138356208D-04,   0.000000000000000D+00, &
   1.151034386818255D-04,   0.000000000000000D+00,   5.105429334853263D-05, &
   3.572997592623339D-07,   0.000000000000000D+00,  -4.856798840810594D-07, &
   0.000000000000000D+00,  -1.430804688442488D-06,   0.000000000000000D+00, &
  -3.784962370237126D-06,   0.000000000000000D+00,  -1.742960423879594D-05, &
   0.000000000000000D+00,  -1.111284759532496D-04,   0.000000000000000D+00, &
  -1.319572193508490D-04,   0.000000000000000D+00,   1.024775950369332D-04, &
   0.000000000000000D+00,   5.648448546612810D-05,   0.000000000000000D+00, &
  -1.321722646669599D-07,   0.000000000000000D+00,   4.068416703052880D-07, &
   0.000000000000000D+00,  -4.402203820332510D-07,   0.000000000000000D+00, &
  -3.848219658398914D-07,   0.000000000000000D+00,   3.484770849200537D-06, &
   0.000000000000000D+00,   2.912154447631317D-05,   0.000000000000000D+00, &
  -9.986485094852350D-05,   0.000000000000000D+00,  -2.350450985050900D-04, &
   0.000000000000000D+00,  -8.645864300312188D-05,   0.000000000000000D+00, &
  -8.753325303837436D-05,   2.227879380342745D-08,   0.000000000000000D+00, &
   4.254148311379264D-09,   0.000000000000000D+00,   7.503467720365970D-07, &
   0.000000000000000D+00,   6.391164720872304D-07,   0.000000000000000D+00, &
   2.028574151176716D-06,   0.000000000000000D+00,   1.190239448271889D-05, &
   0.000000000000000D+00,   1.359331190887288D-04,   0.000000000000000D+00, &
   9.723925216335020D-05,   0.000000000000000D+00,  -3.705446292628276D-05, &
   0.000000000000000D+00,   2.279971453290220D-06,   0.000000000000000D+00, &
  -1.868483414831970D-09,   0.000000000000000D+00,  -1.082585703057941D-07, &
   0.000000000000000D+00,  -5.150240348523744D-09,   0.000000000000000D+00, &
   3.128197242727889D-07,   0.000000000000000D+00,  -7.422880442949688D-07, &
   0.000000000000000D+00,  -7.586674877137429D-06,   0.000000000000000D+00, &
  -4.632892119077701D-05,   0.000000000000000D+00,   9.452541204268298D-05, &
   0.000000000000000D+00,   1.242155540887591D-04,   0.000000000000000D+00, &
   4.952177873356385D-05,   0.000000000000000D+00,   4.645335977140615D-05 /)
lem(:,219) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.011122457834909D-05,  -5.316611800196500D-04, &
  -8.706183488895273D-04,  -2.036156291728076D-04,   3.708367418388800D-04, &
   7.266929142138643D-04,   1.052539254482994D-03,  -3.450384149018058D-04, &
  -1.866492468952925D-03,   2.030539473871566D-04,   1.378991441827527D-03, &
  -5.539747209351105D-04,  -1.685842798405071D-04,   1.637143282303899D-04, &
  -9.742827904233544D-05,   1.839386569297003D-04,  -2.848956754358986D-04, &
   2.161119494951279D-04,   1.140787772711300D-05,  -2.270733120823215D-04, &
   2.655618759446344D-04,   0.000000000000000D+00,   7.011122457834902D-05, &
   5.316611800196525D-04,   8.706183488895245D-04,   2.036156291728103D-04, &
  -3.708367418388880D-04,  -7.266929142138649D-04,  -1.052539254482983D-03, &
   3.450384149018061D-04,   1.866492468952921D-03,  -2.030539473871565D-04, &
  -1.378991441827527D-03,   5.539747209351097D-04,   1.685842798405040D-04, &
  -1.637143282303911D-04,   9.742827904234221D-05,  -1.839386569297006D-04, &
   2.848956754358926D-04,  -2.161119494951262D-04,  -1.140787772711052D-05, &
   2.270733120823176D-04,  -2.655618759446369D-04,   0.000000000000000D+00, &
   3.505561228917395D-05,   0.000000000000000D+00,  -2.739267531369639D-04, &
   0.000000000000000D+00,   4.395095301716382D-04,   0.000000000000000D+00, &
   1.128806209751430D-03,   0.000000000000000D+00,  -3.997106649960503D-03, &
   0.000000000000000D+00,   5.421393377320578D-03,   0.000000000000000D+00, &
  -5.437795132584633D-03,   0.000000000000000D+00,   4.280811080318264D-03, &
   0.000000000000000D+00,  -1.547616454622814D-03,   0.000000000000000D+00, &
  -3.702417415946628D-04,   0.000000000000000D+00,   3.873619851756580D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.771389490139397D-06, &
   0.000000000000000D+00,  -3.638035864598410D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.253622534872865D-07,   0.000000000000000D+00, &
   1.579163566468239D-05,   0.000000000000000D+00,   4.607480193417823D-06, &
   0.000000000000000D+00,   5.186734923063125D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.829978181929694D-06,   0.000000000000000D+00, &
  -4.130853490632537D-06,   0.000000000000000D+00,  -2.980268212733850D-05, &
   0.000000000000000D+00,  -3.045845639587245D-06,   0.000000000000000D+00, &
  -5.369053536595672D-06,   0.000000000000000D+00,  -6.939978698669339D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.251459586304588D-06, &
   0.000000000000000D+00,  -2.552650301334421D-06,   0.000000000000000D+00, &
   7.276130071426000D-06,   0.000000000000000D+00,  -1.243132798342323D-05, &
   0.000000000000000D+00,  -4.290252386290326D-07,   0.000000000000000D+00, &
   3.136155241346859D-06,   0.000000000000000D+00,   9.633046145875052D-06, &
   0.000000000000000D+00,   3.135105585545839D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.819121013800688D-06,   0.000000000000000D+00, &
   7.159750631222013D-06,   0.000000000000000D+00,   4.387275408524585D-06, &
   0.000000000000000D+00,   5.854727113496472D-06,   0.000000000000000D+00, &
   1.595295769868729D-04,   0.000000000000000D+00,   2.554337781859674D-06, &
   0.000000000000000D+00,  -1.425960470867532D-06,   0.000000000000000D+00, &
  -1.113751839126320D-05,   0.000000000000000D+00,  -4.160850671271564D-05, &
   0.000000000000000D+00,  -9.663354155777738D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.588333918818385D-06,   0.000000000000000D+00, &
  -8.056392784023287D-06,   0.000000000000000D+00,  -8.543498330741411D-06, &
   0.000000000000000D+00,   6.127497863439564D-07,   0.000000000000000D+00, &
  -5.066258155920596D-05,   0.000000000000000D+00,  -3.538347777356960D-04, &
   0.000000000000000D+00,  -1.381584321888951D-06,   0.000000000000000D+00, &
   1.377551440485464D-06,   0.000000000000000D+00,   1.066276328840265D-05, &
   0.000000000000000D+00,   3.651262591746097D-05,   0.000000000000000D+00, &
   1.315525984352056D-04,   0.000000000000000D+00,   1.746281240868026D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.615633128941117D-06, &
   0.000000000000000D+00,   5.225537945958704D-06,   0.000000000000000D+00, &
   5.390332984603721D-06,   0.000000000000000D+00,  -7.603648051882244D-06, &
   0.000000000000000D+00,  -3.738631819508166D-05,   0.000000000000000D+00, &
   1.188853572004684D-04,   0.000000000000000D+00,   4.620646364603822D-04, &
   0.000000000000000D+00,  -5.320773829621393D-07,   0.000000000000000D+00, &
  -1.233405466918691D-06,   0.000000000000000D+00,  -6.809991668197071D-06, &
   0.000000000000000D+00,  -1.994339794718193D-05,   0.000000000000000D+00, &
  -7.195711342923363D-05,   0.000000000000000D+00,  -2.583263069655135D-04, &
   0.000000000000000D+00,  -2.173425304243358D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.325388918447228D-06,   0.000000000000000D+00, &
  -1.936003545120932D-06,   0.000000000000000D+00,  -1.020815018140637D-06, &
   0.000000000000000D+00,   8.155115758356218D-06,   0.000000000000000D+00, &
   5.349361651941626D-05,   0.000000000000000D+00,   1.025305506862585D-04, &
   0.000000000000000D+00,  -1.561291192570896D-04,   0.000000000000000D+00, &
  -3.988019020829496D-04,   0.000000000000000D+00,   7.206759475786006D-07, &
   0.000000000000000D+00,   4.050142170898372D-07,   0.000000000000000D+00, &
   2.108392963052461D-06,   0.000000000000000D+00,   5.138853249964861D-06, &
   0.000000000000000D+00,   8.486109857450329D-06,   0.000000000000000D+00, &
   9.936167698424896D-05,   0.000000000000000D+00,   3.261714231306241D-04, &
   0.000000000000000D+00,   1.964533892832111D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.181072815100847D-08,   0.000000000000000D+00, &
   5.459837468701710D-07,   0.000000000000000D+00,  -3.786189244285041D-07, &
   0.000000000000000D+00,  -3.606962639042259D-06,   0.000000000000000D+00, &
  -1.904061006244512D-05,   0.000000000000000D+00,  -1.119313801397589D-04, &
   0.000000000000000D+00,  -1.380973606878615D-04,   0.000000000000000D+00, &
   1.235021584733392D-04,   0.000000000000000D+00,   2.058760307370229D-04, &
   0.000000000000000D+00,  -1.580398242889859D-07,   0.000000000000000D+00, &
  -1.618453546375659D-08,   0.000000000000000D+00,   1.364847612724659D-07, &
   0.000000000000000D+00,   2.436022171738322D-07,   0.000000000000000D+00, &
   3.089567080076807D-06,   0.000000000000000D+00,   2.372468647778984D-05, &
   0.000000000000000D+00,  -1.110752169386838D-04,   0.000000000000000D+00, &
  -2.699074676953315D-04,   0.000000000000000D+00,  -1.189989977198700D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.876623112621351D-08, &
   0.000000000000000D+00,  -2.254304211414578D-07,   0.000000000000000D+00, &
   1.279448524879614D-07,   0.000000000000000D+00,   5.291113546456525D-07, &
   0.000000000000000D+00,   2.790746112579231D-06,   0.000000000000000D+00, &
   1.609642750156279D-05,   0.000000000000000D+00,   1.406826776983664D-04, &
   0.000000000000000D+00,   1.032570962822611D-04,   0.000000000000000D+00, &
  -5.742425357003462D-05,   0.000000000000000D+00,  -2.457665625720454D-05, &
   0.000000000000000D+00,  -1.518911445295235D-08,   0.000000000000000D+00, &
   5.743350691423328D-08,   0.000000000000000D+00,  -3.390279793660200D-07, &
   0.000000000000000D+00,  -3.481020275740987D-07,   0.000000000000000D+00, &
  -1.211113241053706D-06,   0.000000000000000D+00,  -8.812671923485613D-06, &
   0.000000000000000D+00,  -4.166321304667463D-05,   0.000000000000000D+00, &
   1.059155181251776D-04,   0.000000000000000D+00,   1.455826036351378D-04, &
   0.000000000000000D+00,   3.890845121527766D-05,   0.000000000000000D+00 /)
lem(:,220) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.309132804528704D-05,   1.930286786855197D-04,   1.902720894568217D-04, &
   5.238791235586382D-06,   3.726944850391351D-04,   7.978247182715570D-04, &
  -3.948995640717555D-04,  -1.450506552839283D-03,  -3.545635660012778D-04, &
   1.422521026030059D-03,   4.595067816217942D-04,  -1.041851356163160D-03, &
   1.590880636270083D-04,   2.641434884957382D-04,  -1.552892634687103D-04, &
   4.318729326958226D-05,  -3.868359903300404D-05,   9.280813544700205D-05, &
  -1.270578446748547D-04,   8.239978609166422D-05,   2.295291429175393D-05, &
  -1.197928565071845D-04,   7.309132804528349D-05,   1.930286786855165D-04, &
   1.902720894568231D-04,   5.238791235585845D-06,   3.726944850391390D-04, &
   7.978247182715519D-04,  -3.948995640717603D-04,  -1.450506552839272D-03, &
  -3.545635660012754D-04,   1.422521026030051D-03,   4.595067816217896D-04, &
  -1.041851356163155D-03,   1.590880636270118D-04,   2.641434884957357D-04, &
  -1.552892634687119D-04,   4.318729326958497D-05,  -3.868359903300872D-05, &
   9.280813544699953D-05,  -1.270578446748493D-04,   8.239978609166544D-05, &
   2.295291429175169D-05,  -1.197928565071870D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.911343911640875D-05,   0.000000000000000D+00, &
   5.783439571278515D-06,   0.000000000000000D+00,   6.679553973481400D-04, &
   0.000000000000000D+00,  -2.467192271526806D-03,   0.000000000000000D+00, &
   2.351823510667721D-03,   0.000000000000000D+00,   8.477767339889263D-04, &
   0.000000000000000D+00,  -1.207424349958248D-03,   0.000000000000000D+00, &
  -2.030035319326738D-03,   0.000000000000000D+00,   2.713246882354356D-03, &
   0.000000000000000D+00,  -8.921929044540271D-04,   0.000000000000000D+00, &
  -1.423860066438076D-06,  -3.426071541305959D-06,   0.000000000000000D+00, &
  -4.354956747534771D-06,   0.000000000000000D+00,   1.586503051346842D-06, &
   3.552310681215345D-07,   0.000000000000000D+00,   1.356336805555515D-06, &
   0.000000000000000D+00,   4.550706455003589D-06,   0.000000000000000D+00, &
   5.764995999371311D-07,   0.000000000000000D+00,   2.543131510416589D-06, &
   1.493415800056319D-06,   0.000000000000000D+00,   1.071338961964004D-06, &
   0.000000000000000D+00,   2.924897639253172D-06,   0.000000000000000D+00, &
   3.972355046572864D-07,   0.000000000000000D+00,   3.647611069485121D-07, &
   0.000000000000000D+00,  -2.759183362210760D-06,   0.000000000000000D+00, &
  -3.936978729459198D-05,   1.323652162822481D-06,   0.000000000000000D+00, &
  -3.603131161999463D-06,   0.000000000000000D+00,  -5.996607400458928D-06, &
   0.000000000000000D+00,  -4.893088972127556D-06,   0.000000000000000D+00, &
  -4.296295700654839D-06,   0.000000000000000D+00,  -2.800691740477831D-06, &
   0.000000000000000D+00,   1.888563392227240D-06,   0.000000000000000D+00, &
   2.070389340826287D-05,   0.000000000000000D+00,   1.431251620019389D-04, &
  -4.078931515352120D-06,   0.000000000000000D+00,   3.574744532148957D-06, &
   0.000000000000000D+00,   8.045959725861842D-06,   0.000000000000000D+00, &
   8.201613837450138D-06,   0.000000000000000D+00,  -5.400913022095290D-06, &
   0.000000000000000D+00,   4.911725386832484D-06,   0.000000000000000D+00, &
   3.377347744514201D-06,   0.000000000000000D+00,  -1.038278771384379D-08, &
   0.000000000000000D+00,  -1.535913025669427D-05,   0.000000000000000D+00, &
  -6.729508500318408D-05,   0.000000000000000D+00,  -2.779494461735579D-04, &
   3.148636966590079D-06,   0.000000000000000D+00,  -2.346524894800168D-06, &
   0.000000000000000D+00,  -7.487063015896519D-06,   0.000000000000000D+00, &
  -8.852266046091312D-06,   0.000000000000000D+00,   6.121271204251233D-06, &
   0.000000000000000D+00,   1.339666859539218D-06,   0.000000000000000D+00, &
  -3.463824567043381D-06,   0.000000000000000D+00,  -1.642436646690619D-06, &
   0.000000000000000D+00,   6.310928977388946D-08,   0.000000000000000D+00, &
   1.121198044591219D-05,   0.000000000000000D+00,   4.251096049027574D-05, &
   0.000000000000000D+00,   1.618822761100893D-04,   0.000000000000000D+00, &
   3.615067367513969D-04,  -7.209176631955364D-07,   0.000000000000000D+00, &
   1.146428434869874D-06,   0.000000000000000D+00,   4.615963692712065D-06, &
   0.000000000000000D+00,   4.904135573562567D-06,   0.000000000000000D+00, &
  -8.064143530421671D-06,   0.000000000000000D+00,  -5.190884977364252D-05, &
   0.000000000000000D+00,   4.706404452599146D-05,   0.000000000000000D+00, &
   1.410588049582722D-06,   0.000000000000000D+00,  -5.217432822383241D-08, &
   0.000000000000000D+00,  -1.102787635713732D-06,   0.000000000000000D+00, &
  -6.590440893456512D-06,   0.000000000000000D+00,  -2.223952061956891D-05, &
   0.000000000000000D+00,  -8.021696174304483D-05,   0.000000000000000D+00, &
  -2.920656254110370D-04,   0.000000000000000D+00,  -3.944014792212523D-04, &
  -2.639880113844331D-07,   0.000000000000000D+00,  -1.137469010036003D-07, &
   0.000000000000000D+00,  -1.539201328493296D-06,   0.000000000000000D+00, &
  -4.709368587238367D-07,   0.000000000000000D+00,   8.384686285921724D-06, &
   0.000000000000000D+00,   5.668744131689976D-05,   0.000000000000000D+00, &
   1.208800273984850D-04,   0.000000000000000D+00,  -8.165321199935728D-05, &
   0.000000000000000D+00,  -1.894560871767062D-07,   0.000000000000000D+00, &
   2.762615418649931D-07,   0.000000000000000D+00,   9.866560474603202D-07, &
   0.000000000000000D+00,   2.115687321597687D-06,   0.000000000000000D+00, &
   6.543431326640990D-06,   0.000000000000000D+00,   1.285755257664805D-05, &
   0.000000000000000D+00,   1.151034386818255D-04,   0.000000000000000D+00, &
   3.656664041392472D-04,   0.000000000000000D+00,   4.087963866425604D-04, &
   1.167335312841707D-07,   0.000000000000000D+00,  -2.212066815515296D-07, &
   0.000000000000000D+00,   1.545180475469804D-07,   0.000000000000000D+00, &
  -5.494048924510122D-07,   0.000000000000000D+00,  -4.512812736814442D-06, &
   0.000000000000000D+00,  -2.203391316758353D-05,   0.000000000000000D+00, &
  -1.227722763904277D-04,   0.000000000000000D+00,  -1.588788015554859D-04, &
   0.000000000000000D+00,   4.196911902257470D-05,   0.000000000000000D+00, &
  -2.916121465791158D-08,   0.000000000000000D+00,   1.453986082379752D-08, &
   0.000000000000000D+00,  -3.794832451947079D-07,   0.000000000000000D+00, &
  -3.751421842796957D-08,   0.000000000000000D+00,  -2.501287163921293D-07, &
   0.000000000000000D+00,   2.861522410940299D-06,   0.000000000000000D+00, &
   1.953540246714421D-05,   0.000000000000000D+00,  -1.329701152282114D-04, &
   0.000000000000000D+00,  -3.113511444415528D-04,   0.000000000000000D+00, &
  -3.532327852065328D-04,   3.914518556706426D-09,   0.000000000000000D+00, &
   6.171789961321437D-08,   0.000000000000000D+00,   3.888968221012932D-08, &
   0.000000000000000D+00,  -3.584050443182005D-08,   0.000000000000000D+00, &
   1.053677594376740D-06,   0.000000000000000D+00,   3.383207181358802D-06, &
   0.000000000000000D+00,   2.287497684781333D-05,   0.000000000000000D+00, &
   1.582159589348211D-04,   0.000000000000000D+00,   1.264683909397834D-04, &
   0.000000000000000D+00,   1.099677165101271D-05,   0.000000000000000D+00, &
  -2.781830836403745D-10,   0.000000000000000D+00,  -3.724466067985994D-08, &
   0.000000000000000D+00,   1.127221407897013D-07,   0.000000000000000D+00, &
  -1.022315621811394D-07,   0.000000000000000D+00,  -3.360770456626977D-07, &
   0.000000000000000D+00,  -1.013507793677085D-06,   0.000000000000000D+00, &
  -1.160562555480516D-05,   0.000000000000000D+00,  -3.906296812897033D-05, &
   0.000000000000000D+00,   1.255279601615982D-04,   0.000000000000000D+00, &
   1.893955470519623D-04,   0.000000000000000D+00,   2.099664665947903D-04 /)
lem(:,221) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.835000762766309D-05,  -2.623978602834577D-04, &
  -6.608026424662702D-04,  -8.201640242019322D-04,  -9.266283705637097D-05, &
   4.127363104097850D-04,   4.055489364721190D-04,   8.584977532540390D-04, &
  -5.297642182952582D-04,  -9.734623084744893D-04,   7.425159480060628D-04, &
   8.375860685068175D-05,  -2.877721144186711D-04,   1.428050686430141D-04, &
  -3.464070273547708D-05,   1.864401757542988D-05,  -3.562473495757343D-05, &
   5.706275578948790D-05,  -5.837136771134544D-05,   2.610938111519604D-05, &
   2.748040760006317D-05,   0.000000000000000D+00,   4.835000762765718D-05, &
   2.623978602834551D-04,   6.608026424662723D-04,   8.201640242019326D-04, &
   9.266283705636564D-05,  -4.127363104097862D-04,  -4.055489364721107D-04, &
  -8.584977532540375D-04,   5.297642182952497D-04,   9.734623084744882D-04, &
  -7.425159480060609D-04,  -8.375860685068143D-05,   2.877721144186666D-04, &
  -1.428050686430138D-04,   3.464070273548485D-05,  -1.864401757543383D-05, &
   3.562473495757053D-05,  -5.706275578948079D-05,   5.837136771134320D-05, &
  -2.610938111520022D-05,  -2.748040760006267D-05,   0.000000000000000D+00, &
   2.417500381382550D-05,   0.000000000000000D+00,  -5.806762895929701D-05, &
   0.000000000000000D+00,   1.702569733376820D-04,   0.000000000000000D+00, &
  -5.735758461388542D-04,   0.000000000000000D+00,  -4.224262255827252D-04, &
   0.000000000000000D+00,   7.944545619605030D-03,   0.000000000000000D+00, &
  -1.917253253487636D-02,   0.000000000000000D+00,   2.005205406980712D-02, &
   0.000000000000000D+00,  -9.249881417200206D-03,   0.000000000000000D+00, &
   7.840567529058775D-04,   0.000000000000000D+00,   6.615890890712495D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.911201862373843D-06, &
   0.000000000000000D+00,  -1.776155340608358D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.100922731782234D-06,   0.000000000000000D+00, &
   5.266814348845667D-06,   0.000000000000000D+00,   6.673285481880158D-07, &
   0.000000000000000D+00,   3.929447288822205D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.813168096371214D-06,   0.000000000000000D+00, &
   6.165167297980917D-07,   0.000000000000000D+00,  -1.520069421680747D-05, &
   0.000000000000000D+00,   1.049433422480229D-06,   0.000000000000000D+00, &
  -3.040966350118232D-06,   0.000000000000000D+00,  -6.382811253416725D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.483004191988487D-06, &
   0.000000000000000D+00,  -2.802102411477392D-06,   0.000000000000000D+00, &
   3.617970842091877D-06,   0.000000000000000D+00,   3.298229006236808D-05, &
   0.000000000000000D+00,  -1.757185595076252D-06,   0.000000000000000D+00, &
  -3.563602196414057D-07,   0.000000000000000D+00,   4.237536280993395D-06, &
   0.000000000000000D+00,   4.164322096451036D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.219981304088249D-07,   0.000000000000000D+00, &
   2.441247380080189D-06,   0.000000000000000D+00,   2.405150373985143D-06, &
   0.000000000000000D+00,  -5.162006559558046D-06,   0.000000000000000D+00, &
  -1.878438176187450D-07,   0.000000000000000D+00,   2.022134247931225D-06, &
   0.000000000000000D+00,   3.188602144484650D-06,   0.000000000000000D+00, &
   1.179800188833615D-07,   0.000000000000000D+00,  -5.813038998031579D-06, &
   0.000000000000000D+00,  -2.328508066919901D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.529491256350214D-08,   0.000000000000000D+00, &
  -2.128197431963528D-06,   0.000000000000000D+00,  -5.374406747432938D-06, &
   0.000000000000000D+00,  -7.266396344600874D-06,   0.000000000000000D+00, &
  -6.601623108193971D-06,   0.000000000000000D+00,  -2.241046312973308D-04, &
   0.000000000000000D+00,  -1.781337740339019D-06,   0.000000000000000D+00, &
  -2.914261159206306D-06,   0.000000000000000D+00,  -9.669677219617993D-07, &
   0.000000000000000D+00,   7.817729678676349D-06,   0.000000000000000D+00, &
   3.333571943556243D-05,   0.000000000000000D+00,   1.561774697574059D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.932452733447065D-07, &
   0.000000000000000D+00,   1.829455634431061D-06,   0.000000000000000D+00, &
   4.259318061527114D-06,   0.000000000000000D+00,   6.994697524979545D-06, &
   0.000000000000000D+00,   1.059174169504457D-06,   0.000000000000000D+00, &
   4.094693794982140D-06,   0.000000000000000D+00,   6.348197227895361D-04, &
   0.000000000000000D+00,   7.307739722207702D-07,   0.000000000000000D+00, &
   9.270090967052188D-07,   0.000000000000000D+00,  -2.784542346713600D-07, &
   0.000000000000000D+00,  -7.094090614889026D-06,   0.000000000000000D+00, &
  -2.583565225278135D-05,   0.000000000000000D+00,  -1.074301567017706D-04, &
   0.000000000000000D+00,  -4.151462748079044D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.004020994506079D-07,   0.000000000000000D+00, &
  -7.516651148064409D-07,   0.000000000000000D+00,  -1.318781708939862D-06, &
   0.000000000000000D+00,  -1.242973446962883D-06,   0.000000000000000D+00, &
   6.787987023434719D-06,   0.000000000000000D+00,   5.456775852592888D-05, &
   0.000000000000000D+00,   4.423867966421799D-05,   0.000000000000000D+00, &
  -8.895313652906659D-04,   0.000000000000000D+00,   1.271430110553839D-08, &
   0.000000000000000D+00,  -1.895409457584857D-08,   0.000000000000000D+00, &
   3.918965198840519D-07,   0.000000000000000D+00,   3.114490599528014D-06, &
   0.000000000000000D+00,   9.232926668137844D-06,   0.000000000000000D+00, &
   2.507480111740242D-05,   0.000000000000000D+00,   1.964533892832111D-04, &
   0.000000000000000D+00,   5.774011076878162D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.086335031965273D-07,   0.000000000000000D+00, &
   2.548016969330405D-08,   0.000000000000000D+00,  -4.681418339178125D-08, &
   0.000000000000000D+00,  -1.109177747211360D-06,   0.000000000000000D+00, &
  -5.237944719411556D-06,   0.000000000000000D+00,  -2.877726528156695D-05, &
   0.000000000000000D+00,  -1.508433630852731D-04,   0.000000000000000D+00, &
  -7.961595200774834D-05,   0.000000000000000D+00,   7.023834772159586D-04, &
   0.000000000000000D+00,  -5.730916323837423D-08,   0.000000000000000D+00, &
   4.856771960945720D-08,   0.000000000000000D+00,  -3.428547815846163D-08, &
   0.000000000000000D+00,  -3.539696900446440D-07,   0.000000000000000D+00, &
  -7.165542442550829D-07,   0.000000000000000D+00,   2.870470377872249D-06, &
   0.000000000000000D+00,   1.052767734417701D-05,   0.000000000000000D+00, &
  -2.444297536187434D-04,   0.000000000000000D+00,  -4.510793159436025D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.537993585431198D-08, &
   0.000000000000000D+00,   7.599450945580198D-09,   0.000000000000000D+00, &
   7.636857727088578D-08,   0.000000000000000D+00,   5.504802724461387D-07, &
   0.000000000000000D+00,   1.180701864570030D-06,   0.000000000000000D+00, &
   4.852288008790119D-06,   0.000000000000000D+00,   3.670980705483865D-05, &
   0.000000000000000D+00,   2.073404326364890D-04,   0.000000000000000D+00, &
   3.989149360546534D-05,   0.000000000000000D+00,  -2.816241538500622D-04, &
   0.000000000000000D+00,  -2.492914472969253D-09,   0.000000000000000D+00, &
  -3.347043228863592D-08,   0.000000000000000D+00,   1.801189445698958D-08, &
   0.000000000000000D+00,  -1.247823996086333D-07,   0.000000000000000D+00, &
   1.598168907742895D-08,   0.000000000000000D+00,  -1.736694260449707D-06, &
   0.000000000000000D+00,  -1.713089583157887D-05,   0.000000000000000D+00, &
  -3.565249749461526D-05,   0.000000000000000D+00,   2.189808504919438D-04, &
   0.000000000000000D+00,   1.851897491480064D-04,   0.000000000000000D+00 /)
lem(:,222) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.481849603394979D-05,  -2.064623309584095D-04,  -4.288907443797990D-04, &
  -5.226903999762667D-04,  -5.246664086069652D-05,   1.206987343473593D-03, &
   1.898260126393861D-03,  -7.607752452988592D-04,  -2.904850012849724D-03, &
   7.569994937402523D-04,   1.845676061180520D-03,  -9.673737843276095D-04, &
  -1.981329949027824D-04,   3.526119926455642D-04,  -1.820072730357857D-04, &
   2.375179299593979D-05,  -3.899997064050096D-06,   3.160505072115290D-05, &
  -2.479003707541539D-05,   3.543175363533393D-05,  -4.454167326763433D-05, &
   1.349911202305192D-05,  -5.481849603397747D-05,  -2.064623309584046D-04, &
  -4.288907443798040D-04,  -5.226903999762745D-04,  -5.246664086068805D-05, &
   1.206987343473592D-03,   1.898260126393852D-03,  -7.607752452988407D-04, &
  -2.904850012849715D-03,   7.569994937402311D-04,   1.845676061180513D-03, &
  -9.673737843275970D-04,  -1.981329949027740D-04,   3.526119926455609D-04, &
  -1.820072730357905D-04,   2.375179299594668D-05,  -3.899997064059823D-06, &
   3.160505072114414D-05,  -2.479003707540056D-05,   3.543175363533983D-05, &
  -4.454167326764391D-05,   1.349911202304798D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.076114938506552D-05,   0.000000000000000D+00, &
  -6.477452319831951D-05,   0.000000000000000D+00,   2.114721751435253D-04, &
   0.000000000000000D+00,  -8.682522784504177D-04,   0.000000000000000D+00, &
   5.449217082247894D-03,   0.000000000000000D+00,  -2.009739456094397D-02, &
   0.000000000000000D+00,   3.824751055368562D-02,   0.000000000000000D+00, &
  -3.949730718822409D-02,   0.000000000000000D+00,   2.214828305363794D-02, &
   0.000000000000000D+00,  -6.364670102153277D-03,   0.000000000000000D+00, &
   9.247750946951839D-08,   2.774325284091319D-07,   0.000000000000000D+00, &
   6.011038115532074D-07,   0.000000000000000D+00,  -1.676154859138262D-06, &
   8.168846669821569D-07,   0.000000000000000D+00,  -1.575970890546049D-06, &
   0.000000000000000D+00,   2.881030102710757D-07,   0.000000000000000D+00, &
  -1.016109338252954D-06,   0.000000000000000D+00,   6.480193514800479D-06, &
  -1.573303270080554D-06,   0.000000000000000D+00,   4.844483451661788D-07, &
   0.000000000000000D+00,   5.418776971816492D-06,   0.000000000000000D+00, &
  -2.959280303030723D-06,   0.000000000000000D+00,   2.446800771385686D-06, &
   0.000000000000000D+00,   1.643098948396831D-06,   0.000000000000000D+00, &
  -2.324092876422877D-05,   1.731666770729618D-06,   0.000000000000000D+00, &
   1.774814473178722D-06,   0.000000000000000D+00,  -5.396624430437901D-06, &
   0.000000000000000D+00,  -1.637613565599716D-05,   0.000000000000000D+00, &
   4.865739729020847D-06,   0.000000000000000D+00,  -3.312041352202362D-06, &
   0.000000000000000D+00,  -7.202688171926459D-06,   0.000000000000000D+00, &
   4.397905789888870D-06,   0.000000000000000D+00,   9.491262616930138D-05, &
  -2.483681682900434D-06,   0.000000000000000D+00,  -2.525389717162316D-06, &
   0.000000000000000D+00,   4.347685011747555D-06,   0.000000000000000D+00, &
   2.154180272886029D-05,   0.000000000000000D+00,   3.493732505745005D-05, &
   0.000000000000000D+00,  -3.576897955560075D-06,   0.000000000000000D+00, &
   4.040111194957385D-06,   0.000000000000000D+00,   8.642628061725914D-06, &
   0.000000000000000D+00,   1.698922258805447D-06,   0.000000000000000D+00, &
  -4.803752585565731D-05,   0.000000000000000D+00,  -3.714167097688231D-04, &
   2.667373811368493D-06,   0.000000000000000D+00,   6.022597782226881D-07, &
   0.000000000000000D+00,  -4.209611311653803D-06,   0.000000000000000D+00, &
  -1.945159699794049D-05,   0.000000000000000D+00,  -3.542614448503281D-05, &
   0.000000000000000D+00,   1.252019608432232D-06,   0.000000000000000D+00, &
   8.330890902985170D-07,   0.000000000000000D+00,  -3.062967812413605D-06, &
   0.000000000000000D+00,  -5.581349601808763D-06,   0.000000000000000D+00, &
   1.670767416618773D-07,   0.000000000000000D+00,   3.480851342643082D-05, &
   0.000000000000000D+00,   1.650302938793897D-04,   0.000000000000000D+00, &
   1.060067700284547D-03,  -1.336607143562493D-06,   0.000000000000000D+00, &
   1.265086845963140D-06,   0.000000000000000D+00,   3.149347768350607D-06, &
   0.000000000000000D+00,   1.106441872426132D-05,   0.000000000000000D+00, &
   1.778161998553770D-05,   0.000000000000000D+00,  -1.866258663531467D-05, &
   0.000000000000000D+00,  -2.383978466058864D-04,   0.000000000000000D+00, &
   3.112553892580085D-07,   0.000000000000000D+00,   8.217241648930325D-07, &
   0.000000000000000D+00,   2.013274794746220D-06,   0.000000000000000D+00, &
  -2.477030279803564D-06,   0.000000000000000D+00,  -2.068712852946639D-05, &
   0.000000000000000D+00,  -7.762755356542751D-05,   0.000000000000000D+00, &
  -3.151586807952425D-04,   0.000000000000000D+00,  -2.028504787682577D-03, &
   1.681872207093659D-07,   0.000000000000000D+00,  -9.868026555858872D-07, &
   0.000000000000000D+00,  -1.147567776831223D-06,   0.000000000000000D+00, &
  -2.576353366152242D-06,   0.000000000000000D+00,  -1.993533305036209D-06, &
   0.000000000000000D+00,   2.468665347214100D-05,   0.000000000000000D+00, &
   1.940089969106714D-04,   0.000000000000000D+00,   6.683844251438009D-04, &
   0.000000000000000D+00,  -1.332519964196669D-07,   0.000000000000000D+00, &
   2.499715869135115D-07,   0.000000000000000D+00,  -7.438272383284002D-07, &
   0.000000000000000D+00,   1.479476843703843D-06,   0.000000000000000D+00, &
   6.932589264324128D-06,   0.000000000000000D+00,   2.064575554666510D-05, &
   0.000000000000000D+00,   5.105429334853263D-05,   0.000000000000000D+00, &
   4.087963866425604D-04,   0.000000000000000D+00,   2.597937022767604D-03, &
   4.184290537316993D-08,   0.000000000000000D+00,   1.418351635955778D-07, &
   0.000000000000000D+00,   2.547428850453311D-07,   0.000000000000000D+00, &
  -8.876705420844142D-07,   0.000000000000000D+00,  -2.031964685805803D-06, &
   0.000000000000000D+00,  -1.249750163080626D-05,   0.000000000000000D+00, &
  -7.296267095599671D-05,   0.000000000000000D+00,  -4.217581421963197D-04, &
   0.000000000000000D+00,  -9.335253644238789D-04,   0.000000000000000D+00, &
  -7.496182954292764D-09,   0.000000000000000D+00,  -1.264161685616119D-07, &
   0.000000000000000D+00,   4.384496530536542D-07,   0.000000000000000D+00, &
   8.273024579300141D-10,   0.000000000000000D+00,  -8.629257919197803D-07, &
   0.000000000000000D+00,  -6.353698711943508D-07,   0.000000000000000D+00, &
   9.842830448432575D-06,   0.000000000000000D+00,   5.307583432855719D-05, &
   0.000000000000000D+00,  -4.446871746125545D-04,   0.000000000000000D+00, &
  -2.245920133862107D-03,   8.008477200357684D-10,   0.000000000000000D+00, &
   4.046917264169744D-08,   0.000000000000000D+00,  -2.216580334003818D-07, &
   0.000000000000000D+00,   9.295539319382801D-07,   0.000000000000000D+00, &
   9.167052724586657D-07,   0.000000000000000D+00,   2.223088943142765D-06, &
   0.000000000000000D+00,   1.008046080115512D-05,   0.000000000000000D+00, &
   8.103316290704374D-05,   0.000000000000000D+00,   5.459524197891498D-04, &
   0.000000000000000D+00,   7.370153174111758D-04,   0.000000000000000D+00, &
  -4.773028456778899D-11,   0.000000000000000D+00,  -5.348944281358945D-09, &
   0.000000000000000D+00,  -8.833489987710955D-08,   0.000000000000000D+00, &
  -3.987508981432228D-07,   0.000000000000000D+00,   4.173923938473921D-07, &
   0.000000000000000D+00,  -3.504841907599989D-07,   0.000000000000000D+00, &
  -4.145343474422820D-06,   0.000000000000000D+00,  -3.963077364307036D-05, &
   0.000000000000000D+00,  -1.197477290547169D-04,   0.000000000000000D+00, &
   4.567683650986554D-04,   0.000000000000000D+00,   1.316518699882476D-03 /)
lem(:,223) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.638414066402645D-03,  -3.996108389831299D-03, &
  -1.015337488528637D-02,   8.334978107915381D-03,   1.126654447273055D-02, &
  -4.375180263200058D-03,  -4.882226992860420D-03,  -6.483045315013837D-04, &
  -6.514087843691127D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.638414066402676D-03, &
  -3.996108389831359D-03,  -1.015337488528637D-02,   8.334978107915381D-03, &
   1.126654447273051D-02,  -4.375180263200012D-03,  -4.882226992860420D-03, &
  -6.483045315014486D-04,  -6.514087843690619D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.877947688477916D-04,   0.000000000000000D+00,   6.042915196633645D-06, &
   0.000000000000000D+00,   3.348434926362646D-07,   0.000000000000000D+00, &
   2.804671989572912D-08,   0.000000000000000D+00,   2.845649339891154D-09, &
   0.000000000000000D+00,   3.151353384654175D-10,   0.000000000000000D+00, &
   4.145162566281613D-11,  -3.043411497249097D-03,   0.000000000000000D+00, &
  -6.323468624576155D-05,   0.000000000000000D+00,  -4.948242724515414D-06, &
   0.000000000000000D+00,  -5.396889234255248D-07,   0.000000000000000D+00, &
  -6.827549722587220D-08,   0.000000000000000D+00,  -9.237790800758476D-09, &
   0.000000000000000D+00,  -1.280123725232840D-09,   0.000000000000000D+00, &
   6.829750841615885D-03,   0.000000000000000D+00,   3.940651516863447D-05, &
   0.000000000000000D+00,   1.790888370397503D-05,   0.000000000000000D+00, &
   3.293582936805048D-06,   0.000000000000000D+00,   5.827317280356852D-07, &
   0.000000000000000D+00,   1.023717186182376D-07,   0.000000000000000D+00, &
   1.766230635155172D-08,   0.000000000000000D+00,   3.585417319054295D-09, &
  -4.316657931785842D-03,   0.000000000000000D+00,   8.379694141651865D-04, &
   0.000000000000000D+00,   2.788106381471022D-05,   0.000000000000000D+00, &
  -3.854589223726201D-06,   0.000000000000000D+00,  -1.819964334727542D-06, &
   0.000000000000000D+00,  -5.034161338132666D-07,   0.000000000000000D+00, &
  -1.191183957916526D-07,   0.000000000000000D+00,  -2.621129230997358D-08, &
   0.000000000000000D+00,  -7.924642073696581D-03,   0.000000000000000D+00, &
  -2.560474346162148D-03,   0.000000000000000D+00,  -1.947836756561436D-04, &
   0.000000000000000D+00,  -2.545766189994548D-05,   0.000000000000000D+00, &
  -1.231749710625376D-06,   0.000000000000000D+00,   7.074576633694918D-07, &
   0.000000000000000D+00,   3.572997592623339D-07,   0.000000000000000D+00, &
   1.167335312841707D-07,   0.000000000000000D+00,   4.184290537316993D-08, &
   1.625375883440400D-02,   0.000000000000000D+00,   2.006002195101194D-03, &
   0.000000000000000D+00,   3.024722906717499D-05,   0.000000000000000D+00, &
   5.473352792424587D-05,   0.000000000000000D+00,   1.631999140712507D-05, &
   0.000000000000000D+00,   2.717622145128055D-06,   0.000000000000000D+00, &
   1.827067200627219D-08,   0.000000000000000D+00,  -1.968698297587715D-07, &
   0.000000000000000D+00,  -1.033177663213987D-07,   0.000000000000000D+00, &
  -8.370511478300615D-03,   0.000000000000000D+00,   2.701710584792058D-03, &
   0.000000000000000D+00,   8.939267615397676D-04,   0.000000000000000D+00, &
   7.845733290466745D-05,   0.000000000000000D+00,  -9.822364754227937D-06, &
   0.000000000000000D+00,  -8.653336524050898D-06,   0.000000000000000D+00, &
  -2.636935014954742D-06,   0.000000000000000D+00,  -4.032331814776471D-07, &
   0.000000000000000D+00,   6.147901701524751D-08,   0.000000000000000D+00, &
   1.177049730412304D-07,  -4.835342640220692D-03,   0.000000000000000D+00, &
  -6.058893075932913D-03,   0.000000000000000D+00,  -1.135719322593626D-03, &
   0.000000000000000D+00,  -2.293285076853964D-04,   0.000000000000000D+00, &
  -5.755091459944866D-05,   0.000000000000000D+00,  -4.637824921922318D-06, &
   0.000000000000000D+00,   3.482654143076275D-06,   0.000000000000000D+00, &
   1.954384034819455D-06,   0.000000000000000D+00,   5.307391091249542D-07, &
   0.000000000000000D+00,   2.564415408458496D-08,   0.000000000000000D+00, &
   8.089874539108791D-03,   0.000000000000000D+00,   2.854809977514230D-03, &
   0.000000000000000D+00,  -7.569359840893331D-04,   0.000000000000000D+00, &
  -1.176375227757986D-04,   0.000000000000000D+00,   4.835190882873903D-05, &
   0.000000000000000D+00,   3.163928185161514D-05,   0.000000000000000D+00, &
   7.724463349518118D-06,   0.000000000000000D+00,  -4.736965703182943D-07, &
   0.000000000000000D+00,  -1.166524577192382D-06,   0.000000000000000D+00, &
  -4.994836153418678D-07,   0.000000000000000D+00,  -8.466372747589597D-08 /)
lem(:,224) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.442755550237446D-04, &
   2.358363738711021D-03,  -1.875092698849757D-03,  -4.512064934940326D-03, &
   6.409438908279238D-03,   2.749898822471807D-03,  -5.992881965116243D-03, &
  -7.306778010650692D-04,   1.789782466937172D-03,   3.636419664194281D-04, &
   4.042610263600931D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.442755550237505D-04,  -2.358363738711016D-03,   1.875092698849738D-03, &
   4.512064934940327D-03,  -6.409438908279204D-03,  -2.749898822471829D-03, &
   5.992881965116217D-03,   7.306778010651058D-04,  -1.789782466937189D-03, &
  -3.636419664194561D-04,  -4.042610263600565D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.427152147572568D-10,   0.000000000000000D+00, &
   5.321566929954437D-08,   0.000000000000000D+00,   5.425647640954340D-07, &
   0.000000000000000D+00,   1.493840237366546D-06,   0.000000000000000D+00, &
  -6.376360369805256D-08,   0.000000000000000D+00,  -3.846443031299473D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.571663921455250D-06, &
   0.000000000000000D+00,   1.634094604587638D-07,   0.000000000000000D+00, &
   1.227129212752654D-08,   0.000000000000000D+00,   1.071303280922049D-09, &
   0.000000000000000D+00,   7.588398235567626D-11,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.832185289843566D-05,   0.000000000000000D+00, &
  -2.497487026968925D-06,   0.000000000000000D+00,  -2.487944325473709D-07, &
   0.000000000000000D+00,  -2.845255478596222D-08,   0.000000000000000D+00, &
  -3.160344678852327D-09,   0.000000000000000D+00,  -8.905208525554365D-11, &
   0.000000000000000D+00,   1.748482924669026D-04,   0.000000000000000D+00, &
   1.171545784880627D-05,   0.000000000000000D+00,   1.721633946568100D-06, &
   0.000000000000000D+00,   2.715312692402316D-07,   0.000000000000000D+00, &
   4.190326261886884D-08,   0.000000000000000D+00,   5.329555272372994D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.277007487660230D-04, &
   0.000000000000000D+00,  -1.105631744350698D-05,   0.000000000000000D+00, &
  -4.061810379639237D-06,   0.000000000000000D+00,  -1.090567834685028D-06, &
   0.000000000000000D+00,  -2.489647120027733D-07,   0.000000000000000D+00, &
  -4.909823110608486D-08,   0.000000000000000D+00,  -6.717601643732351D-09, &
   0.000000000000000D+00,   5.402165803255590D-04,   0.000000000000000D+00, &
  -8.595541998716730D-05,   0.000000000000000D+00,  -4.781982611104446D-06, &
   0.000000000000000D+00,   9.344699976263584D-07,   0.000000000000000D+00, &
   6.004344058651668D-07,   0.000000000000000D+00,   2.009280247257791D-07, &
   0.000000000000000D+00,   5.070399951087409D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.990160527794985D-05,   0.000000000000000D+00, &
   3.517717798319428D-04,   0.000000000000000D+00,   3.915096888921056D-05, &
   0.000000000000000D+00,   6.050444485819691D-06,   0.000000000000000D+00, &
   3.762524742792917D-07,   0.000000000000000D+00,  -2.514111225686297D-07, &
   0.000000000000000D+00,  -1.475892499949442D-07,   0.000000000000000D+00, &
  -6.733129527978334D-08,   0.000000000000000D+00,  -7.974553024608631D-04, &
   0.000000000000000D+00,  -5.554321500541940D-04,   0.000000000000000D+00, &
  -4.421409771366374D-05,   0.000000000000000D+00,  -1.604921438569874D-05, &
   0.000000000000000D+00,  -4.589401038575375D-06,   0.000000000000000D+00, &
  -8.195629089752364D-07,   0.000000000000000D+00,   3.181072815100847D-08, &
   0.000000000000000D+00,   1.086335031965273D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.312975439783002D-03,   0.000000000000000D+00, &
   2.036931122388397D-04,   0.000000000000000D+00,  -1.020243987130660D-04, &
   0.000000000000000D+00,  -3.721532319718036D-06,   0.000000000000000D+00, &
   5.245757336512866D-06,   0.000000000000000D+00,   2.863802809376545D-06, &
   0.000000000000000D+00,   8.426197788674532D-07,   0.000000000000000D+00, &
   8.297002515845458D-08,   0.000000000000000D+00,  -1.472307075674457D-07, &
   0.000000000000000D+00,  -8.436354161957146D-04,   0.000000000000000D+00, &
   6.875295563563543D-04,   0.000000000000000D+00,   3.228419134528766D-04, &
   0.000000000000000D+00,   5.811720774260601D-05,   0.000000000000000D+00, &
   1.057559894040674D-05,   0.000000000000000D+00,  -4.629971581224267D-07, &
   0.000000000000000D+00,  -1.490935648418050D-06,   0.000000000000000D+00, &
  -6.915897341924751D-07,   0.000000000000000D+00,  -1.174913428961993D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.125985488013751D-04, &
   0.000000000000000D+00,  -1.222701092439568D-03,   0.000000000000000D+00, &
  -2.286190080687797D-04,   0.000000000000000D+00,  -4.399915936545186D-05, &
   0.000000000000000D+00,  -2.418589866637483D-05,   0.000000000000000D+00, &
  -8.639732403029280D-06,   0.000000000000000D+00,  -1.332206282849700D-06, &
   0.000000000000000D+00,   5.760394463218963D-07,   0.000000000000000D+00, &
   5.101170072968352D-07,   0.000000000000000D+00,   1.998319934422818D-07, &
   0.000000000000000D+00,   6.435821944892648D-04,   0.000000000000000D+00, &
   7.278939819388261D-04,   0.000000000000000D+00,  -2.929401128454495D-04, &
   0.000000000000000D+00,  -9.295676598484422D-05,   0.000000000000000D+00, &
  -7.190839784086323D-06,   0.000000000000000D+00,   7.528271707888774D-06, &
   0.000000000000000D+00,   5.404346333185709D-06,   0.000000000000000D+00, &
   1.694061238661683D-06,   0.000000000000000D+00,  -5.866256519556398D-08, &
   0.000000000000000D+00,  -3.521944028294727D-07,   0.000000000000000D+00 /)
lem(:,225) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.418448713027486D-04,  -9.592310480549429D-04,   2.224584967860555D-04, &
   1.093843562477128D-03,   2.423860158807245D-03,   6.429932828422757D-04, &
  -7.389739411970418D-03,  -2.274509130724119D-03,   6.273079174315280D-03, &
   2.264265007241312D-03,  -8.330203500415842D-04,  -5.579131158713940D-04, &
  -9.028849544162442D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.418448713027518D-04,  -9.592310480549422D-04, &
   2.224584967860540D-04,   1.093843562477136D-03,   2.423860158807253D-03, &
   6.429932828422565D-04,  -7.389739411970421D-03,  -2.274509130724086D-03, &
   6.273079174315266D-03,   2.264265007241293D-03,  -8.330203500415736D-04, &
  -5.579131158713758D-04,  -9.028849544162566D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.170898295341131D-08,   0.000000000000000D+00,  -6.314027406320748D-07, &
   0.000000000000000D+00,  -3.167871064239224D-06,   0.000000000000000D+00, &
  -4.321381509608080D-06,   0.000000000000000D+00,   4.221966845851634D-06, &
   0.000000000000000D+00,   1.058145627026917D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.161687749810156D-06,   0.000000000000000D+00,   1.276451557501405D-07, &
   0.000000000000000D+00,   7.912677542022607D-09,   0.000000000000000D+00, &
   4.769056826194756D-10,   0.000000000000000D+00,  -8.529563136546220D-11, &
  -5.452696483030158D-05,   0.000000000000000D+00,  -2.006048523012538D-06, &
   0.000000000000000D+00,  -1.726507641286418D-07,   0.000000000000000D+00, &
  -1.646174437777956D-08,   0.000000000000000D+00,  -8.017789329885840D-10, &
   0.000000000000000D+00,   3.484429531804844D-05,   0.000000000000000D+00, &
   8.810670779946010D-06,   0.000000000000000D+00,   1.203115333859049D-06, &
   0.000000000000000D+00,   1.719625063055478D-07,   0.000000000000000D+00, &
   2.078355443093395D-08,   0.000000000000000D+00,  -2.115907675279073D-09, &
   7.298729841893741D-04,   0.000000000000000D+00,  -2.304124887046316D-06, &
   0.000000000000000D+00,  -2.430056025065957D-06,   0.000000000000000D+00, &
  -6.850480642114199D-07,   0.000000000000000D+00,  -1.432668992090305D-07, &
   0.000000000000000D+00,  -1.831585458652096D-08,   0.000000000000000D+00, &
  -2.245577164093711D-03,   0.000000000000000D+00,   7.080054248163696D-06, &
   0.000000000000000D+00,  -4.011597539617730D-06,   0.000000000000000D+00, &
   4.078945432415283D-07,   0.000000000000000D+00,   3.459513258842746D-07, &
   0.000000000000000D+00,   1.056277228423076D-07,   0.000000000000000D+00, &
   1.051171079090721D-08,   1.766962664457464D-03,   0.000000000000000D+00, &
  -3.103044425133791D-04,   0.000000000000000D+00,   1.048964154376532D-05, &
   0.000000000000000D+00,   3.286579611373572D-06,   0.000000000000000D+00, &
   2.484527330795823D-07,   0.000000000000000D+00,  -1.473286177342732D-07, &
   0.000000000000000D+00,  -7.656201345748434D-08,   0.000000000000000D+00, &
   2.396646714286282D-03,   0.000000000000000D+00,   9.636344417739803D-04, &
   0.000000000000000D+00,  -1.456337706856421D-06,   0.000000000000000D+00, &
  -3.724085687538793D-06,   0.000000000000000D+00,  -2.002444115022358D-06, &
   0.000000000000000D+00,  -3.768654507789690D-07,   0.000000000000000D+00, &
   5.629185957273016D-08,   0.000000000000000D+00,   1.224820711263768D-07, &
  -5.390244086017165D-03,   0.000000000000000D+00,  -7.719717055210603D-04, &
   0.000000000000000D+00,   1.261818710356561D-04,   0.000000000000000D+00, &
  -1.083151686065353D-06,   0.000000000000000D+00,   7.762890480210459D-07, &
   0.000000000000000D+00,   1.045492410363692D-06,   0.000000000000000D+00, &
   3.147667608348006D-07,   0.000000000000000D+00,  -4.636839205165295D-08, &
   0.000000000000000D+00,   2.548965785895708D-03,   0.000000000000000D+00, &
  -1.070152144880910D-03,   0.000000000000000D+00,  -4.905348829299328D-04, &
   0.000000000000000D+00,  -2.021814607298990D-05,   0.000000000000000D+00, &
   1.034908991575972D-06,   0.000000000000000D+00,   1.711193445453078D-07, &
   0.000000000000000D+00,  -4.856798840810594D-07,   0.000000000000000D+00, &
  -2.212066815515296D-07,   0.000000000000000D+00,   1.418351635955778D-07, &
   2.006002195101194D-03,   0.000000000000000D+00,   2.409260402078505D-03, &
   0.000000000000000D+00,   4.528483135648673D-04,   0.000000000000000D+00, &
   2.193134860992385D-06,   0.000000000000000D+00,   5.176977926097199D-06, &
   0.000000000000000D+00,  -4.470291288082495D-07,   0.000000000000000D+00, &
  -3.098432144022509D-07,   0.000000000000000D+00,   2.300041465227281D-07, &
   0.000000000000000D+00,   1.337295742935327D-07,   0.000000000000000D+00, &
  -2.718173139237192D-03,   0.000000000000000D+00,  -1.137481650765530D-03, &
   0.000000000000000D+00,   5.394503378759455D-04,   0.000000000000000D+00, &
   1.920182330635061D-04,   0.000000000000000D+00,   1.221220910626693D-05, &
   0.000000000000000D+00,  -1.194336419201727D-06,   0.000000000000000D+00, &
   9.097418470056828D-08,   0.000000000000000D+00,   2.019484532163208D-07, &
   0.000000000000000D+00,  -1.582233741234562D-07,   0.000000000000000D+00, &
  -3.943828854975936D-07,   9.460752983366179D-04,   0.000000000000000D+00, &
  -8.763398539107509D-04,   0.000000000000000D+00,  -1.293452648497122D-03, &
   0.000000000000000D+00,  -2.659083501690629D-04,   0.000000000000000D+00, &
  -4.621119289361296D-05,   0.000000000000000D+00,  -8.684065724468918D-06, &
   0.000000000000000D+00,  -2.306683584813176D-07,   0.000000000000000D+00, &
  -1.347224144111791D-08,   0.000000000000000D+00,  -8.685812304694845D-08, &
   0.000000000000000D+00,   1.766009793584626D-07,   0.000000000000000D+00, &
   4.641503496032222D-05,   0.000000000000000D+00,   1.165112541589966D-03, &
   0.000000000000000D+00,   5.731979646554537D-04,   0.000000000000000D+00, &
  -1.805118797033451D-04,   0.000000000000000D+00,  -2.916396560105545D-05, &
   0.000000000000000D+00,   9.899168121865801D-06,   0.000000000000000D+00, &
   4.981161114562020D-06,   0.000000000000000D+00,   7.308543429482645D-07, &
   0.000000000000000D+00,   9.361418373577898D-08,   0.000000000000000D+00, &
   9.102278564802990D-08,   0.000000000000000D+00,  -4.828262506823343D-07 /)
lem(:,226) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.786588125467149D-05,   3.760202093338931D-04, &
  -6.293688037585722D-04,  -1.255122740972130D-03,   2.424214121115794D-03, &
   3.735738399293885D-03,  -5.883507101267844D-03,  -6.283444046781997D-03, &
   8.220984909973429D-03,   5.600896580713692D-03,  -5.319974233718400D-03, &
  -3.003375841617150D-03,   5.818942423018463D-04,   9.085690976479074D-04, &
   1.009457945747327D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.786588125467740D-05, &
  -3.760202093339082D-04,   6.293688037585678D-04,   1.255122740972144D-03, &
  -2.424214121115791D-03,  -3.735738399293887D-03,   5.883507101267832D-03, &
   6.283444046782004D-03,  -8.220984909973413D-03,  -5.600896580713687D-03, &
   5.319974233718391D-03,   3.003375841617157D-03,  -5.818942423018426D-04, &
  -9.085690976479209D-04,  -1.009457945747331D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.136590449043141D-09,   0.000000000000000D+00,   4.756517147310603D-07, &
   0.000000000000000D+00,   3.795125392296143D-06,   0.000000000000000D+00, &
   9.128646742357214D-06,   0.000000000000000D+00,   2.990063882836013D-06, &
   0.000000000000000D+00,  -1.355455480811528D-05,   0.000000000000000D+00, &
  -1.167994397953411D-05,   0.000000000000000D+00,   6.194871216383805D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.289345007576916D-07,   0.000000000000000D+00, &
   5.612588897946475D-09,   0.000000000000000D+00,   3.756833011390984D-12, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.042322492004714D-06, &
   0.000000000000000D+00,  -1.357719364043831D-07,   0.000000000000000D+00, &
  -8.768447691510303D-09,   0.000000000000000D+00,   2.551828660969299D-09, &
   0.000000000000000D+00,   9.871399667363579D-06,   0.000000000000000D+00, &
   9.872990997864227D-07,   0.000000000000000D+00,   1.170459967160040D-07, &
   0.000000000000000D+00,   4.927025587127951D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.838219334165535D-06,   0.000000000000000D+00, &
  -1.991513728280008D-06,   0.000000000000000D+00,  -5.057986959904422D-07, &
   0.000000000000000D+00,  -8.281968750309408D-08,   0.000000000000000D+00, &
   1.620355238472149D-08,   0.000000000000000D+00,  -7.282046511520464D-05, &
   0.000000000000000D+00,  -4.473152128576444D-06,   0.000000000000000D+00, &
   2.304772968589756D-07,   0.000000000000000D+00,   2.344149733278557D-07, &
   0.000000000000000D+00,   4.667472941464160D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.039650007388793D-04,   0.000000000000000D+00, &
   1.522865356104047D-05,   0.000000000000000D+00,   3.268645211920764D-06, &
   0.000000000000000D+00,   2.514063402354939D-07,   0.000000000000000D+00, &
  -9.953845895327885D-08,   0.000000000000000D+00,  -5.202444022096961D-08, &
   0.000000000000000D+00,  -4.848820621580653D-04,   0.000000000000000D+00, &
   4.994271592788823D-05,   0.000000000000000D+00,  -3.326536907328955D-06, &
   0.000000000000000D+00,  -1.778196285545568D-06,   0.000000000000000D+00, &
  -2.913365068703193D-07,   0.000000000000000D+00,   6.485394852323057D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.791414277255398D-04, &
   0.000000000000000D+00,  -2.972921725007927D-04,   0.000000000000000D+00, &
  -1.628966668315447D-05,   0.000000000000000D+00,  -3.691645301087503D-07, &
   0.000000000000000D+00,   7.692964612056563D-07,   0.000000000000000D+00, &
   1.920528805278518D-07,   0.000000000000000D+00,  -2.254788859210767D-07, &
   0.000000000000000D+00,   6.073435120902055D-04,   0.000000000000000D+00, &
   5.355868198764417D-04,   0.000000000000000D+00,   9.604706399547819D-06, &
   0.000000000000000D+00,   7.469871221745377D-06,   0.000000000000000D+00, &
   1.172287811390652D-06,   0.000000000000000D+00,  -2.290290181167915D-07, &
   0.000000000000000D+00,  -5.487136205416378D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.084337857201425D-03,   0.000000000000000D+00, &
  -2.377225736124120D-04,   0.000000000000000D+00,   1.481151634912577D-04, &
   0.000000000000000D+00,   5.148234333703136D-06,   0.000000000000000D+00, &
  -2.838927330313969D-06,   0.000000000000000D+00,  -9.803851072368914D-07, &
   0.000000000000000D+00,   2.814003338643780D-08,   0.000000000000000D+00, &
   9.903629841475407D-08,   0.000000000000000D+00,   6.473339757907153D-04, &
   0.000000000000000D+00,  -6.441224187142505D-04,   0.000000000000000D+00, &
  -3.755524753305419D-04,   0.000000000000000D+00,  -3.891170172610181D-05, &
   0.000000000000000D+00,  -6.159202151342601D-06,   0.000000000000000D+00, &
   5.195401226504020D-07,   0.000000000000000D+00,   5.459837468701710D-07, &
   0.000000000000000D+00,   2.548016969330405D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.036931122388397D-04,   0.000000000000000D+00, &
   1.193194834361660D-03,   0.000000000000000D+00,   2.341105539063962D-04, &
   0.000000000000000D+00,  -9.063983706333045D-06,   0.000000000000000D+00, &
   1.122354204221141D-05,   0.000000000000000D+00,   4.401502535354291D-06, &
   0.000000000000000D+00,   4.357177230679232D-07,   0.000000000000000D+00, &
  -2.133741145859296D-07,   0.000000000000000D+00,   4.650284686545723D-07, &
   0.000000000000000D+00,  -5.908879220807054D-04,   0.000000000000000D+00, &
  -6.863830814201899D-04,   0.000000000000000D+00,   4.040802108319559D-04, &
   0.000000000000000D+00,   1.702418998334970D-04,   0.000000000000000D+00, &
   2.084917994497706D-05,   0.000000000000000D+00,  -1.521903412390190D-06, &
   0.000000000000000D+00,  -2.450566030207751D-06,   0.000000000000000D+00, &
  -6.355954706594202D-07,   0.000000000000000D+00,  -1.241397185995711D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.996783049388927D-04, &
   0.000000000000000D+00,  -2.702336328431006D-04,   0.000000000000000D+00, &
  -8.258742581684043D-04,   0.000000000000000D+00,  -1.752062499395887D-04, &
   0.000000000000000D+00,  -4.025306082889557D-05,   0.000000000000000D+00, &
  -1.363581536479836D-05,   0.000000000000000D+00,  -1.668147698526579D-06, &
   0.000000000000000D+00,   9.919488549163183D-07,   0.000000000000000D+00, &
   4.541022375817646D-07,   0.000000000000000D+00,  -5.333191708167225D-08, &
   0.000000000000000D+00,  -1.054448140042075D-04,   0.000000000000000D+00, &
   6.523795824663346D-04,   0.000000000000000D+00,   4.306760360418309D-04, &
   0.000000000000000D+00,  -1.469109739818385D-04,   0.000000000000000D+00, &
  -3.466766239593811D-05,   0.000000000000000D+00,   5.881014310823306D-06, &
   0.000000000000000D+00,   6.999528263572369D-06,   0.000000000000000D+00, &
   2.224634304669314D-06,   0.000000000000000D+00,  -6.907870706564532D-08, &
   0.000000000000000D+00,  -4.137276604247899D-07,   0.000000000000000D+00 /)
lem(:,227) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.169706246618103D-05, &
  -1.420275905056559D-04,   3.885696546885108D-04,   3.417316182967864D-04, &
  -1.144804038671542D-03,   6.915570464422982D-04,  -7.282833322886310D-04, &
  -4.544601451180142D-03,   6.501389135883465D-03,   8.670632881674150D-03, &
  -9.479063665567828D-03,  -7.985114497759658D-03,   5.332698856113390D-03, &
   3.838143083603960D-03,  -5.108148287710710D-04,  -1.036458093152822D-03, &
  -6.474622125746496D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.169706246617372D-05,  -1.420275905056439D-04,   3.885696546885180D-04, &
   3.417316182967740D-04,  -1.144804038671547D-03,   6.915570464423050D-04, &
  -7.282833322886256D-04,  -4.544601451180157D-03,   6.501389135883456D-03, &
   8.670632881674167D-03,  -9.479063665567793D-03,  -7.985114497759661D-03, &
   5.332698856113392D-03,   3.838143083603974D-03,  -5.108148287710804D-04, &
  -1.036458093152832D-03,  -6.474622125746304D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.559987125808790D-07, &
   0.000000000000000D+00,  -3.640572548997478D-06,   0.000000000000000D+00, &
  -1.307227488128896D-05,   0.000000000000000D+00,  -1.262064405468898D-05, &
   0.000000000000000D+00,   8.498801540419126D-06,   0.000000000000000D+00, &
   1.610412838677471D-05,   0.000000000000000D+00,   6.197574336805414D-06, &
   0.000000000000000D+00,   6.380960879713946D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.250384573408557D-07,   0.000000000000000D+00, &
   4.329910777581229D-09,   0.000000000000000D+00,  -8.852037221665739D-10, &
  -3.446204314423369D-06,   0.000000000000000D+00,  -1.193086007103854D-07, &
   0.000000000000000D+00,  -1.566129662055289D-09,   0.000000000000000D+00, &
   1.292783260932360D-05,   0.000000000000000D+00,   9.168625220185844D-07, &
   0.000000000000000D+00,   7.659005777015582D-08,   0.000000000000000D+00, &
  -2.229661801831149D-08,   1.944798891526730D-05,   0.000000000000000D+00, &
  -1.976627526565409D-06,   0.000000000000000D+00,  -3.951924949469125D-07, &
   0.000000000000000D+00,  -2.781959825527839D-08,   0.000000000000000D+00, &
  -1.433501939336112D-04,   0.000000000000000D+00,  -3.075364405869295D-06, &
   0.000000000000000D+00,   1.932396235083975D-07,   0.000000000000000D+00, &
   1.628152721626476D-07,   0.000000000000000D+00,  -1.882724084196798D-08, &
   2.552493535055889D-05,   0.000000000000000D+00,   8.259482023470951D-06, &
   0.000000000000000D+00,   2.867448474989186D-06,   0.000000000000000D+00, &
   1.774762363875138D-07,   0.000000000000000D+00,  -7.630004117257147D-08, &
   0.000000000000000D+00,   6.716892231970755D-04,   0.000000000000000D+00, &
  -7.874501604801202D-07,   0.000000000000000D+00,  -4.025302375402282D-06, &
   0.000000000000000D+00,  -1.516501586004559D-06,   0.000000000000000D+00, &
  -1.660636380429425D-07,   0.000000000000000D+00,   2.376974360881995D-07, &
  -8.665725691264919D-04,   0.000000000000000D+00,   1.081223155499791D-04, &
   0.000000000000000D+00,  -8.247401466742629D-06,   0.000000000000000D+00, &
   3.377453351561830D-07,   0.000000000000000D+00,   6.688596963686461D-07, &
   0.000000000000000D+00,   3.929300334437422D-08,   0.000000000000000D+00, &
  -5.827892683430648D-04,   0.000000000000000D+00,  -4.235046645220767D-04, &
   0.000000000000000D+00,   4.239249424197367D-05,   0.000000000000000D+00, &
   5.916269507328871D-06,   0.000000000000000D+00,   6.257589999600995D-07, &
   0.000000000000000D+00,  -2.621003502405827D-07,   0.000000000000000D+00, &
   1.005552388056796D-07,   1.784822972777737D-03,   0.000000000000000D+00, &
   3.927097034197233D-04,   0.000000000000000D+00,  -1.627937305506170D-04, &
   0.000000000000000D+00,  -8.084060452667724D-06,   0.000000000000000D+00, &
  -3.139749243396192D-06,   0.000000000000000D+00,  -6.497019707037894D-07, &
   0.000000000000000D+00,   5.657905388140023D-08,   0.000000000000000D+00, &
  -6.262054004842212D-04,   0.000000000000000D+00,   4.724510448396389D-04, &
   0.000000000000000D+00,   3.731889742624513D-04,   0.000000000000000D+00, &
  -1.057066802714807D-05,   0.000000000000000D+00,   6.067438892652721D-07, &
   0.000000000000000D+00,   1.378348032605796D-06,   0.000000000000000D+00, &
   4.554380557661598D-07,   0.000000000000000D+00,  -5.869278730632321D-07, &
  -1.000900825706165D-03,   0.000000000000000D+00,  -1.136037476679387D-03, &
   0.000000000000000D+00,  -2.615659733286196D-04,   0.000000000000000D+00, &
   1.004593628685650D-04,   0.000000000000000D+00,   9.682269168032223D-06, &
   0.000000000000000D+00,   1.273515698322968D-06,   0.000000000000000D+00, &
  -4.606453331005372D-07,   0.000000000000000D+00,  -2.162423392117538D-07, &
   0.000000000000000D+00,   8.383296766524495D-04,   0.000000000000000D+00, &
   5.058561951723381D-04,   0.000000000000000D+00,  -4.605531236365895D-04, &
   0.000000000000000D+00,  -2.626115330995250D-04,   0.000000000000000D+00, &
  -2.143988698455021D-05,   0.000000000000000D+00,  -5.488832023553196D-06, &
   0.000000000000000D+00,  -1.430804688442488D-06,   0.000000000000000D+00, &
   1.545180475469804D-07,   0.000000000000000D+00,   2.547428850453311D-07, &
   3.024722906717499D-05,   0.000000000000000D+00,   4.528483135648673D-04, &
   0.000000000000000D+00,   9.818233614907448D-04,   0.000000000000000D+00, &
   2.016665565964868D-04,   0.000000000000000D+00,  -9.359012530631673D-06, &
   0.000000000000000D+00,   3.812243026677166D-06,   0.000000000000000D+00, &
   2.606182153950875D-06,   0.000000000000000D+00,   1.083349611217446D-06, &
   0.000000000000000D+00,   2.553683287822053D-07,   0.000000000000000D+00, &
  -2.076701249341536D-04,   0.000000000000000D+00,  -5.242191407015943D-04, &
   0.000000000000000D+00,  -4.922067156140453D-04,   0.000000000000000D+00, &
   2.972765339258929D-04,   0.000000000000000D+00,   1.211377181300459D-04, &
   0.000000000000000D+00,   1.470217643801394D-05,   0.000000000000000D+00, &
   9.772757481590671D-07,   0.000000000000000D+00,  -9.723412670880292D-07, &
   0.000000000000000D+00,  -8.646335729646066D-07,   0.000000000000000D+00, &
   3.027748916335630D-07,   3.201356547893330D-05,   0.000000000000000D+00, &
   1.443346154330549D-04,   0.000000000000000D+00,  -3.002791603279786D-04, &
   0.000000000000000D+00,  -6.626559497331646D-04,   0.000000000000000D+00, &
  -1.426384334216252D-04,   0.000000000000000D+00,  -2.851765515259088D-05, &
   0.000000000000000D+00,  -8.786441088291755D-06,   0.000000000000000D+00, &
  -1.949196469688150D-06,   0.000000000000000D+00,   9.896351575754642D-08, &
   0.000000000000000D+00,   4.526290112960162D-07,   0.000000000000000D+00, &
   2.167929075276315D-05,   0.000000000000000D+00,  -2.335978695242678D-06, &
   0.000000000000000D+00,   4.585596588505354D-04,   0.000000000000000D+00, &
   3.176547135837128D-04,   0.000000000000000D+00,  -1.086532363175542D-04, &
   0.000000000000000D+00,  -2.233962607543736D-05,   0.000000000000000D+00, &
   4.467507185461708D-06,   0.000000000000000D+00,   4.249793346375971D-06, &
   0.000000000000000D+00,   1.802603661765543D-06,   0.000000000000000D+00, &
   7.989937425807684D-08,   0.000000000000000D+00,  -9.551687684559049D-07 /)
lem(:,228) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.473566147000109D-06,   4.523936656907345D-05,  -2.169827185614590D-04, &
   7.972498807430811D-05,   1.032580091132900D-03,  -5.740356131534213D-04, &
  -1.902182073602562D-03,  -1.736399993318016D-03,   4.375373089008650D-03, &
   7.943140058370083D-03,  -9.041376341180027D-03,  -1.170104379657267D-02, &
   1.077627432478414D-02,   8.742204683810472D-03,  -6.500076943536734D-03, &
  -3.879311522411955D-03,   1.517325994226151D-03,   1.368159847474710D-03, &
   1.626697594356613D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.473566147020009D-06,  -4.523936656906596D-05, &
   2.169827185614403D-04,  -7.972498807430141D-05,  -1.032580091132885D-03, &
   5.740356131534053D-04,   1.902182073602559D-03,   1.736399993318031D-03, &
  -4.375373089008650D-03,  -7.943140058370084D-03,   9.041376341180025D-03, &
   1.170104379657268D-02,  -1.077627432478413D-02,  -8.742204683810468D-03, &
   6.500076943536717D-03,   3.879311522411955D-03,  -1.517325994226143D-03, &
  -1.368159847474707D-03,  -1.626697594356651D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.488091262487475D-08, &
   0.000000000000000D+00,   2.825458877169845D-06,   0.000000000000000D+00, &
   1.570755064547652D-05,   0.000000000000000D+00,   2.450653989354149D-05, &
   0.000000000000000D+00,   3.623481949602663D-07,   0.000000000000000D+00, &
  -1.752278987178328D-05,   0.000000000000000D+00,  -2.216951682783282D-06, &
   0.000000000000000D+00,  -8.765258707094589D-06,   0.000000000000000D+00, &
  -1.399430128425244D-05,   0.000000000000000D+00,   1.625242533574456D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.370377233420050D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.290942198196240D-07,   0.000000000000000D+00, &
   1.218662882414605D-08,   0.000000000000000D+00,   1.062786094418747D-06, &
   0.000000000000000D+00,   3.979409013737290D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.891016991165004D-06,   0.000000000000000D+00, &
  -3.259946556488213D-07,   0.000000000000000D+00,   6.174908794704933D-08, &
   0.000000000000000D+00,  -2.563074900744991D-06,   0.000000000000000D+00, &
   2.337916783596095D-07,   0.000000000000000D+00,   9.793115779792266D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.724540117028070D-05, &
   0.000000000000000D+00,   2.422596028559995D-06,   0.000000000000000D+00, &
   8.455254228754003D-08,   0.000000000000000D+00,  -1.698978507557404D-07, &
   0.000000000000000D+00,  -3.368580990236548D-05,   0.000000000000000D+00, &
  -2.638226001634109D-06,   0.000000000000000D+00,  -1.236934145788302D-06, &
   0.000000000000000D+00,  -1.159456701954315D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.173652426428016D-05,   0.000000000000000D+00, &
  -1.364928895622291D-05,   0.000000000000000D+00,   5.035275226475004D-07, &
   0.000000000000000D+00,   5.296248362789739D-07,   0.000000000000000D+00, &
  -3.537516516022247D-07,   0.000000000000000D+00,   2.397416291655131D-04, &
   0.000000000000000D+00,   9.083445199000418D-06,   0.000000000000000D+00, &
   3.733198920217930D-06,   0.000000000000000D+00,   2.438832718963117D-07, &
   0.000000000000000D+00,  -1.625450736784428D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.742525498007689D-04,   0.000000000000000D+00, &
   1.244180772147488D-04,   0.000000000000000D+00,  -6.031496386672049D-07, &
   0.000000000000000D+00,  -1.902699573028097D-06,   0.000000000000000D+00, &
  -2.320728066692198D-07,   0.000000000000000D+00,   6.725056907645831D-07, &
   0.000000000000000D+00,  -2.228015812720763D-04,   0.000000000000000D+00, &
  -3.218310548684547D-04,   0.000000000000000D+00,   1.260248130981536D-05, &
   0.000000000000000D+00,   2.315669846884841D-07,   0.000000000000000D+00, &
   8.447071007038026D-07,   0.000000000000000D+00,   1.088977436891489D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.928557563638327D-04, &
   0.000000000000000D+00,   2.029719495467116D-04,   0.000000000000000D+00, &
  -1.327370033742221D-04,   0.000000000000000D+00,  -2.504913487141662D-06, &
   0.000000000000000D+00,   9.379648830493821D-08,   0.000000000000000D+00, &
  -5.273797250266416D-07,   0.000000000000000D+00,   2.802224713148970D-07, &
   0.000000000000000D+00,  -2.402339580671915D-04,   0.000000000000000D+00, &
   3.514366357340658D-04,   0.000000000000000D+00,   3.097244396655799D-04, &
   0.000000000000000D+00,   2.341906484866961D-06,   0.000000000000000D+00, &
   5.340587432010526D-07,   0.000000000000000D+00,  -2.140355396054670D-08, &
   0.000000000000000D+00,   1.249360507599833D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.012366284912053D-04,   0.000000000000000D+00, &
  -7.222193640495506D-04,   0.000000000000000D+00,  -1.845069893630866D-04, &
   0.000000000000000D+00,   7.438994133078915D-05,   0.000000000000000D+00, &
   3.170758372057155D-06,   0.000000000000000D+00,   1.181142102116645D-07, &
   0.000000000000000D+00,  -1.009426145360502D-08,   0.000000000000000D+00, &
  -1.297841107889962D-06,   0.000000000000000D+00,   3.019714986298608D-04, &
   0.000000000000000D+00,   3.776530006396227D-04,   0.000000000000000D+00, &
  -3.658152630448167D-04,   0.000000000000000D+00,  -2.190138425404549D-04, &
   0.000000000000000D+00,  -1.998936291333156D-05,   0.000000000000000D+00, &
  -2.923145974306833D-06,   0.000000000000000D+00,  -3.786189244285041D-07, &
   0.000000000000000D+00,  -4.681418339178125D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.020243987130660D-04,   0.000000000000000D+00, &
   2.341105539063962D-04,   0.000000000000000D+00,   7.338573556386840D-04, &
   0.000000000000000D+00,   1.582312958125735D-04,   0.000000000000000D+00, &
  -2.345867751112298D-06,   0.000000000000000D+00,   5.666746810816224D-06, &
   0.000000000000000D+00,   1.961553110367963D-06,   0.000000000000000D+00, &
   8.999108420420001D-07,   0.000000000000000D+00,   5.948047675913967D-07, &
   0.000000000000000D+00,  -4.695783300010958D-05,   0.000000000000000D+00, &
  -4.026563383271306D-04,   0.000000000000000D+00,  -3.922006074368048D-04, &
   0.000000000000000D+00,   2.409068791879921D-04,   0.000000000000000D+00, &
   1.016378966750809D-04,   0.000000000000000D+00,   1.152966729676214D-05, &
   0.000000000000000D+00,  -6.574940901215429D-07,   0.000000000000000D+00, &
  -1.177994083665852D-06,   0.000000000000000D+00,  -3.690237351154326D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.507036990401916D-05, &
   0.000000000000000D+00,   1.718144215863910D-04,   0.000000000000000D+00, &
  -2.119517826069314D-04,   0.000000000000000D+00,  -5.192767294737469D-04, &
   0.000000000000000D+00,  -1.157818738927336D-04,   0.000000000000000D+00, &
  -2.627002090610527D-05,   0.000000000000000D+00,  -8.094091769972646D-06, &
   0.000000000000000D+00,  -1.125537221032533D-06,   0.000000000000000D+00, &
   2.585522177173884D-07,   0.000000000000000D+00,   1.374558057728394D-06, &
   0.000000000000000D+00,  -6.278353444861142D-06,   0.000000000000000D+00, &
   1.071178341214324D-05,   0.000000000000000D+00,   3.844757068796200D-04, &
   0.000000000000000D+00,   2.581590194348227D-04,   0.000000000000000D+00, &
  -8.973240143509224D-05,   0.000000000000000D+00,  -1.935088495831335D-05, &
   0.000000000000000D+00,   4.793428336254489D-06,   0.000000000000000D+00, &
   4.503708109075570D-06,   0.000000000000000D+00,   1.303698375512066D-06, &
   0.000000000000000D+00,  -1.415464519824769D-07,   0.000000000000000D+00 /)
lem(:,229) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.933022977161586D-07,  -1.673134726262293D-05, &
   9.011659316873932D-05,  -8.677583710206977D-05,  -3.862121167704638D-04, &
   4.781084461437378D-04,   4.740099396903772D-06,  -4.359217383210175D-04, &
   1.785756509657203D-03,   3.040376511413362D-03,  -6.460748611877895D-03, &
  -9.623561697913605D-03,   1.208251586594655D-02,   1.243905736544586D-02, &
  -1.279910286278401D-02,  -7.495472645400192D-03,   7.850959634952807D-03, &
   2.200770771948888D-03,  -2.888545133346831D-03,  -7.181578016770314D-04, &
   7.868970402041475D-04,   0.000000000000000D+00,   8.933022977153103D-07, &
  -1.673134726263536D-05,   9.011659316875621D-05,  -8.677583710206747D-05, &
  -3.862121167704790D-04,   4.781084461437399D-04,   4.740099396917618D-06, &
  -4.359217383210260D-04,   1.785756509657197D-03,   3.040376511413367D-03, &
  -6.460748611877889D-03,  -9.623561697913634D-03,   1.208251586594655D-02, &
   1.243905736544586D-02,  -1.279910286278397D-02,  -7.495472645400205D-03, &
   7.850959634952807D-03,   2.200770771948900D-03,  -2.888545133346831D-03, &
  -7.181578016770404D-04,   7.868970402041469D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.441322090103769D-06,   0.000000000000000D+00, &
  -1.564432949777588D-05,   0.000000000000000D+00,  -3.715149668173642D-05, &
   0.000000000000000D+00,  -1.558006678183403D-05,   0.000000000000000D+00, &
   2.055627004941015D-05,   0.000000000000000D+00,   3.541134883600588D-07, &
   0.000000000000000D+00,  -1.033737084275428D-05,   0.000000000000000D+00, &
   1.936368430588518D-05,   0.000000000000000D+00,   2.463659493133539D-05, &
   0.000000000000000D+00,  -1.135228255335206D-05,   0.000000000000000D+00, &
   1.412434949912706D-08,  -2.657945769399360D-07,   0.000000000000000D+00, &
   1.777767669879424D-06,   0.000000000000000D+00,  -1.168468913117993D-08, &
  -2.520040757017048D-06,   0.000000000000000D+00,  -3.076423397125972D-07, &
   0.000000000000000D+00,  -1.384340334737458D-05,   0.000000000000000D+00, &
   4.421271583743143D-07,   0.000000000000000D+00,   9.405970472822168D-08, &
   3.327070134513918D-05,   0.000000000000000D+00,   1.462068614444328D-06, &
   0.000000000000000D+00,  -3.845529246414030D-08,   0.000000000000000D+00, &
   4.466287327952552D-05,   0.000000000000000D+00,  -2.070657544487873D-06, &
   0.000000000000000D+00,  -9.510943391005755D-07,   0.000000000000000D+00, &
   5.320248995214150D-07,  -1.427330426917325D-04,   0.000000000000000D+00, &
   8.941707451979671D-08,   0.000000000000000D+00,   5.014451830626345D-07, &
   0.000000000000000D+00,   3.001792854134021D-07,   0.000000000000000D+00, &
  -7.017950960470539D-05,   0.000000000000000D+00,  -1.595194632310964D-05, &
   0.000000000000000D+00,   4.427144890258774D-06,   0.000000000000000D+00, &
  -6.701651510968322D-08,   0.000000000000000D+00,  -4.841519236615819D-07, &
   3.267595053219864D-04,   0.000000000000000D+00,   2.736254705700860D-06, &
   0.000000000000000D+00,  -7.567107847206264D-06,   0.000000000000000D+00, &
  -1.296509036924955D-06,   0.000000000000000D+00,   1.154461264971164D-08, &
   0.000000000000000D+00,   3.903431014888693D-05,   0.000000000000000D+00, &
   1.401646999228513D-04,   0.000000000000000D+00,  -6.585192858776349D-06, &
   0.000000000000000D+00,  -1.250985247087169D-06,   0.000000000000000D+00, &
   5.630359244537728D-07,   0.000000000000000D+00,  -1.036054626223595D-06, &
  -4.511968830521604D-04,   0.000000000000000D+00,  -1.972038333752597D-04, &
   0.000000000000000D+00,   8.350034277978467D-05,   0.000000000000000D+00, &
   5.339783832154179D-06,   0.000000000000000D+00,   6.022764563864340D-07, &
   0.000000000000000D+00,  -3.450840380594280D-08,   0.000000000000000D+00, &
   4.304486273903047D-05,   0.000000000000000D+00,  -1.351527930107809D-04, &
   0.000000000000000D+00,  -2.243547912690518D-04,   0.000000000000000D+00, &
   1.453076605546590D-05,   0.000000000000000D+00,   1.772748936531770D-08, &
   0.000000000000000D+00,   5.341204836205378D-09,   0.000000000000000D+00, &
   1.236559917456650D-06,   3.873249594157238D-04,   0.000000000000000D+00, &
   4.256994224203359D-04,   0.000000000000000D+00,   1.737839892459658D-04, &
   0.000000000000000D+00,  -1.064475855371683D-04,   0.000000000000000D+00, &
  -7.080958064506108D-06,   0.000000000000000D+00,  -8.295341303561286D-07, &
   0.000000000000000D+00,  -1.782464356249972D-08,   0.000000000000000D+00, &
  -9.415110646724745D-05,   0.000000000000000D+00,  -1.463922844156403D-04, &
   0.000000000000000D+00,   2.571744418177518D-04,   0.000000000000000D+00, &
   2.385317400267561D-04,   0.000000000000000D+00,   3.207454500370189D-06, &
   0.000000000000000D+00,   2.409936541014415D-06,   0.000000000000000D+00, &
   2.985582621793514D-07,   0.000000000000000D+00,   6.631357906034956D-07, &
  -2.003185492646766D-04,   0.000000000000000D+00,  -2.313373604426494D-04, &
   0.000000000000000D+00,  -5.759216286366213D-04,   0.000000000000000D+00, &
  -1.530732675125142D-04,   0.000000000000000D+00,   6.168660606223382D-05, &
   0.000000000000000D+00,   4.267808976936278D-06,   0.000000000000000D+00, &
  -3.293633935231737D-07,   0.000000000000000D+00,  -7.384911526786885D-07, &
   0.000000000000000D+00,   7.292005763249218D-05,   0.000000000000000D+00, &
   1.789501125204395D-04,   0.000000000000000D+00,   2.770982514094166D-04, &
   0.000000000000000D+00,  -2.869838216749015D-04,   0.000000000000000D+00, &
  -1.761323965525997D-04,   0.000000000000000D+00,  -1.843787817598085D-05, &
   0.000000000000000D+00,  -3.784962370237126D-06,   0.000000000000000D+00, &
  -5.494048924510122D-07,   0.000000000000000D+00,  -8.876705420844142D-07, &
   5.473352792424587D-05,   0.000000000000000D+00,   2.193134860992385D-06, &
   0.000000000000000D+00,   2.016665565964868D-04,   0.000000000000000D+00, &
   5.957936595138912D-04,   0.000000000000000D+00,   1.324163883596886D-04, &
   0.000000000000000D+00,  -1.855911602752209D-06,   0.000000000000000D+00, &
   5.015032475376303D-06,   0.000000000000000D+00,   2.754488253042815D-06, &
   0.000000000000000D+00,   8.653233203480088D-07,   0.000000000000000D+00, &
  -2.735732340158949D-05,   0.000000000000000D+00,  -2.139868401775447D-05, &
   0.000000000000000D+00,  -2.820992464617308D-04,   0.000000000000000D+00, &
  -3.083080682221210D-04,   0.000000000000000D+00,   1.941285965492779D-04, &
   0.000000000000000D+00,   8.284567026545772D-05,   0.000000000000000D+00, &
   1.034050005752572D-05,   0.000000000000000D+00,  -3.435595031822973D-07, &
   0.000000000000000D+00,  -9.366473920608391D-07,   0.000000000000000D+00, &
  -1.015773903317888D-06,  -4.357272907548071D-06,   0.000000000000000D+00, &
  -1.555254161950876D-06,   0.000000000000000D+00,   1.175064755858271D-04, &
   0.000000000000000D+00,  -1.759569476614578D-04,   0.000000000000000D+00, &
  -4.291975230230603D-04,   0.000000000000000D+00,  -9.769131191524808D-05, &
   0.000000000000000D+00,  -2.214182624560110D-05,   0.000000000000000D+00, &
  -7.353591213040619D-06,   0.000000000000000D+00,  -1.416146393362972D-06, &
   0.000000000000000D+00,   9.024730924008295D-07,   0.000000000000000D+00, &
   4.065294810050979D-06,   0.000000000000000D+00,  -4.861449969023181D-06, &
   0.000000000000000D+00,  -1.587417233194586D-05,   0.000000000000000D+00, &
   2.968813811726063D-04,   0.000000000000000D+00,   2.083940000685976D-04, &
   0.000000000000000D+00,  -7.266479464635369D-05,   0.000000000000000D+00, &
  -1.507209541656229D-05,   0.000000000000000D+00,   4.155159385553522D-06, &
   0.000000000000000D+00,   3.858207924669375D-06,   0.000000000000000D+00, &
   1.420791556653482D-06,   0.000000000000000D+00,   3.739520199132799D-07 /)
lem(:,230) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.620937623856301D-05,   8.259976678042103D-05, &
   6.510651692554574D-05,  -6.179489264001202D-04,  -2.901525444177292D-05, &
   1.204448085691919D-03,   1.150796706788563D-03,  -6.086419486410867D-04, &
  -4.191335264534872D-03,  -3.782577080935007D-03,   9.111622574641381D-03, &
   9.840262118955251D-03,  -1.395526927773110D-02,  -9.856369559401180D-03, &
   1.391562281043947D-02,   3.706393951808622D-03,  -8.292908931716711D-03, &
   5.730909959208972D-04,   2.970083015258819D-03,  -8.385328818196057D-04, &
  -9.449534754372494D-04,   0.000000000000000D+00,   1.620937623856396D-05, &
  -8.259976678042441D-05,  -6.510651692553703D-05,   6.179489264001244D-04, &
   2.901525444176552D-05,  -1.204448085691913D-03,  -1.150796706788559D-03, &
   6.086419486410871D-04,   4.191335264534865D-03,   3.782577080935002D-03, &
  -9.111622574641390D-03,  -9.840262118955233D-03,   1.395526927773107D-02, &
   9.856369559401185D-03,  -1.391562281043946D-02,  -3.706393951808637D-03, &
   8.292908931716683D-03,  -5.730909959208939D-04,  -2.970083015258812D-03, &
   8.385328818196045D-04,   9.449534754372515D-04,   0.000000000000000D+00, &
   8.104688119285616D-06,   0.000000000000000D+00,   4.198881803171351D-05, &
   0.000000000000000D+00,   3.341032209104755D-05,   0.000000000000000D+00, &
  -2.505277198528167D-05,   0.000000000000000D+00,  -9.681413403987852D-06, &
   0.000000000000000D+00,   2.925325759990340D-05,   0.000000000000000D+00, &
   9.123805872321174D-06,   0.000000000000000D+00,  -1.682156397995440D-05, &
   0.000000000000000D+00,  -4.044168330856233D-05,   0.000000000000000D+00, &
  -7.088640588986829D-06,   0.000000000000000D+00,   4.729540258459305D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.407318545559246D-07, &
   0.000000000000000D+00,   1.055730923478826D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.197267853816607D-06,   0.000000000000000D+00, &
  -6.837819379885558D-07,   0.000000000000000D+00,  -9.289440698446000D-06, &
   0.000000000000000D+00,  -4.573993544580735D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.045221585671061D-07,   0.000000000000000D+00, &
  -1.623856617800082D-07,   0.000000000000000D+00,   1.355788720971713D-07, &
   0.000000000000000D+00,   3.299682319112247D-05,   0.000000000000000D+00, &
   4.630130420805188D-06,   0.000000000000000D+00,  -1.403544908314541D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.966676833884113D-05, &
   0.000000000000000D+00,   3.268611557104688D-06,   0.000000000000000D+00, &
  -1.046013707125784D-06,   0.000000000000000D+00,   1.710667693483836D-06, &
   0.000000000000000D+00,  -5.468067433090107D-05,   0.000000000000000D+00, &
  -2.789304030294761D-05,   0.000000000000000D+00,   6.773893908721617D-07, &
   0.000000000000000D+00,   6.246348471354790D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.239158225615444D-05,   0.000000000000000D+00, &
  -3.635688015320572D-06,   0.000000000000000D+00,  -3.646464585023818D-06, &
   0.000000000000000D+00,   3.059032487806984D-07,   0.000000000000000D+00, &
  -7.230960023145388D-07,   0.000000000000000D+00,   3.118418848785516D-05, &
   0.000000000000000D+00,   1.214632515765755D-04,   0.000000000000000D+00, &
   4.168433784289642D-06,   0.000000000000000D+00,  -1.876906404051227D-06, &
   0.000000000000000D+00,  -4.784348997478632D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.376289495134076D-04,   0.000000000000000D+00, &
  -1.297720878952872D-04,   0.000000000000000D+00,   6.087726793778554D-05, &
   0.000000000000000D+00,   3.281362303974192D-06,   0.000000000000000D+00, &
   5.490190387853214D-07,   0.000000000000000D+00,  -1.165662061381185D-06, &
   0.000000000000000D+00,   3.449496598549945D-05,   0.000000000000000D+00, &
  -1.079344468353159D-04,   0.000000000000000D+00,  -1.861947056843112D-04, &
   0.000000000000000D+00,   7.264983871413899D-06,   0.000000000000000D+00, &
   1.412765026244339D-06,   0.000000000000000D+00,   5.258000894271972D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.095588377038783D-04, &
   0.000000000000000D+00,   2.903429485859620D-04,   0.000000000000000D+00, &
   1.367072759655286D-04,   0.000000000000000D+00,  -8.215895198345082D-05, &
   0.000000000000000D+00,  -6.257182036659461D-06,   0.000000000000000D+00, &
  -9.189726202562671D-07,   0.000000000000000D+00,   7.025023068997220D-07, &
   0.000000000000000D+00,  -7.390590125185929D-05,   0.000000000000000D+00, &
  -1.175592949741628D-04,   0.000000000000000D+00,   2.072156258011585D-04, &
   0.000000000000000D+00,   1.998825280730075D-04,   0.000000000000000D+00, &
   5.489763718957564D-06,   0.000000000000000D+00,   1.647495806521627D-06, &
   0.000000000000000D+00,   1.107786710957110D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.894427850896253D-05,   0.000000000000000D+00, &
  -1.521534766833258D-04,   0.000000000000000D+00,  -4.498811358951233D-04, &
   0.000000000000000D+00,  -1.278496733597001D-04,   0.000000000000000D+00, &
   4.779361312832274D-05,   0.000000000000000D+00,   3.365322702005197D-06, &
   0.000000000000000D+00,  -6.669793675172399D-07,   0.000000000000000D+00, &
  -8.908500305509171D-07,   0.000000000000000D+00,   5.393758167204765D-05, &
   0.000000000000000D+00,   1.579088791112338D-04,   0.000000000000000D+00, &
   2.239011141504473D-04,   0.000000000000000D+00,  -2.366534266814817D-04, &
   0.000000000000000D+00,  -1.497684398357018D-04,   0.000000000000000D+00, &
  -1.690007036967001D-05,   0.000000000000000D+00,  -3.606962639042259D-06, &
   0.000000000000000D+00,  -1.109177747211360D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.721532319718036D-06,   0.000000000000000D+00, &
  -9.063983706333045D-06,   0.000000000000000D+00,   1.582312958125735D-04, &
   0.000000000000000D+00,   4.868733541509405D-04,   0.000000000000000D+00, &
   1.136062778153181D-04,   0.000000000000000D+00,   1.311683244667325D-06, &
   0.000000000000000D+00,   5.484577930766048D-06,   0.000000000000000D+00, &
   2.624996591655047D-06,   0.000000000000000D+00,   1.710170853401269D-06, &
   0.000000000000000D+00,  -1.722517708669060D-05,   0.000000000000000D+00, &
  -4.179582982912928D-05,   0.000000000000000D+00,  -2.359250712674516D-04, &
   0.000000000000000D+00,  -2.546085645481721D-04,   0.000000000000000D+00, &
   1.626717578866875D-04,   0.000000000000000D+00,   7.096435142243197D-05, &
   0.000000000000000D+00,   8.660889050841522D-06,   0.000000000000000D+00, &
  -2.050353769483022D-07,   0.000000000000000D+00,  -1.495063980450833D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.193373878584103D-06, &
   0.000000000000000D+00,   6.076056578381753D-06,   0.000000000000000D+00, &
   8.802491574747587D-05,   0.000000000000000D+00,  -1.461013652562690D-04, &
   0.000000000000000D+00,  -3.586172585534414D-04,   0.000000000000000D+00, &
  -8.474602224743460D-05,   0.000000000000000D+00,  -2.077743483479517D-05, &
   0.000000000000000D+00,  -6.979560919618401D-06,   0.000000000000000D+00, &
  -9.262525185124195D-07,   0.000000000000000D+00,   1.120485607487242D-06, &
   0.000000000000000D+00,   1.054514542052064D-06,   0.000000000000000D+00, &
   9.579788684084681D-06,   0.000000000000000D+00,   4.426187792148556D-07, &
   0.000000000000000D+00,   2.490162673218992D-04,   0.000000000000000D+00, &
   1.747923268600898D-04,   0.000000000000000D+00,  -6.146231147463362D-05, &
   0.000000000000000D+00,  -1.257265938204223D-05,   0.000000000000000D+00, &
   4.187087459316430D-06,   0.000000000000000D+00,   3.945207956230306D-06, &
   0.000000000000000D+00,   1.643240724021970D-06,   0.000000000000000D+00 /)
lem(:,231) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.849477370138747D-05,  -4.905894405109629D-05,   2.550836220744938D-04, &
   1.196451939107515D-05,  -1.584186705073535D-04,   3.074070018700853D-05, &
  -8.233870943540086D-04,  -1.294456838734316D-03,   8.271593749480471D-04, &
   5.149647116045555D-03,   3.234051029544556D-03,  -1.023995728720538D-02, &
  -7.833941822533989D-03,   1.366981083074327D-02,   5.476872595844049D-03, &
  -1.223412374063357D-02,   9.130583895070671D-04,   6.366217011028160D-03, &
  -3.162803787065178D-03,  -1.465590656764934D-03,   1.790811770683746D-03, &
  -1.531090504102720D-05,  -2.849477370137976D-05,  -4.905894405110352D-05, &
   2.550836220744868D-04,   1.196451939109041D-05,  -1.584186705073502D-04, &
   3.074070018699106D-05,  -8.233870943540033D-04,  -1.294456838734306D-03, &
   8.271593749480388D-04,   5.149647116045554D-03,   3.234051029544560D-03, &
  -1.023995728720536D-02,  -7.833941822533993D-03,   1.366981083074325D-02, &
   5.476872595844048D-03,  -1.223412374063358D-02,   9.130583895070671D-04, &
   6.366217011028167D-03,  -3.162803787065163D-03,  -1.465590656764932D-03, &
   1.790811770683738D-03,  -1.531090504101956D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.395199783221439D-05,   0.000000000000000D+00, &
   5.510882473162914D-05,   0.000000000000000D+00,   5.114015600314457D-05, &
   0.000000000000000D+00,  -3.218870186208761D-05,   0.000000000000000D+00, &
  -3.578392758044761D-05,   0.000000000000000D+00,  -2.298878149684640D-05, &
   0.000000000000000D+00,  -2.525442529182631D-06,   0.000000000000000D+00, &
   6.181272427128369D-05,   0.000000000000000D+00,   4.287565311643191D-05, &
   0.000000000000000D+00,  -5.846162982276075D-05,   0.000000000000000D+00, &
   9.692513835903736D-07,   6.339971079084783D-06,   0.000000000000000D+00, &
  -6.591982742247055D-06,   0.000000000000000D+00,   8.978409779430284D-07, &
  -2.401948646711561D-05,   0.000000000000000D+00,   4.753113544989233D-07, &
   0.000000000000000D+00,   2.579566038119520D-05,   0.000000000000000D+00, &
  -1.507048393353319D-06,   0.000000000000000D+00,  -1.700646459966986D-06, &
   5.424644068913959D-05,   0.000000000000000D+00,   4.181829570246137D-06, &
   0.000000000000000D+00,  -1.807326883806144D-06,   0.000000000000000D+00, &
  -4.476880933558708D-05,   0.000000000000000D+00,   7.758461733163283D-06, &
   0.000000000000000D+00,   2.372841784801264D-06,   0.000000000000000D+00, &
  -1.467989393376351D-07,  -8.246666567926761D-05,   0.000000000000000D+00, &
  -2.898404862351587D-05,   0.000000000000000D+00,   5.690169752225577D-06, &
   0.000000000000000D+00,   6.964842996584105D-07,   0.000000000000000D+00, &
   2.594351023523844D-05,   0.000000000000000D+00,  -1.630391291650340D-05, &
   0.000000000000000D+00,  -1.780083350546279D-05,   0.000000000000000D+00, &
   3.705154089221315D-07,   0.000000000000000D+00,   1.775120614083177D-06, &
   9.867212238364244D-05,   0.000000000000000D+00,   7.898784475170752D-05, &
   0.000000000000000D+00,  -2.800233706103405D-06,   0.000000000000000D+00, &
  -7.498784082909457D-06,   0.000000000000000D+00,   4.409509387617953D-07, &
   0.000000000000000D+00,   2.851792966013809D-05,   0.000000000000000D+00, &
   1.038692012267270D-05,   0.000000000000000D+00,   8.627051996030386D-05, &
   0.000000000000000D+00,   6.247854483515517D-06,   0.000000000000000D+00, &
  -2.368181098803722D-06,   0.000000000000000D+00,  -1.397829164400967D-06, &
  -1.021041191578757D-04,   0.000000000000000D+00,  -1.155234426687471D-04, &
   0.000000000000000D+00,  -1.052900452272084D-04,   0.000000000000000D+00, &
   4.935664578438932D-05,   0.000000000000000D+00,   6.730948398231338D-06, &
   0.000000000000000D+00,  -3.189720773104371D-07,   0.000000000000000D+00, &
  -5.910602841857929D-05,   0.000000000000000D+00,   1.169142254492287D-05, &
   0.000000000000000D+00,  -7.890020747787720D-05,   0.000000000000000D+00, &
  -1.505536507286103D-04,   0.000000000000000D+00,   1.532396437671412D-06, &
   0.000000000000000D+00,   1.430968461028531D-06,   0.000000000000000D+00, &
   1.725876560712009D-07,   8.544619064097900D-05,   0.000000000000000D+00, &
   9.504031585117143D-05,   0.000000000000000D+00,   2.327577392235544D-04, &
   0.000000000000000D+00,   1.153214701531801D-04,   0.000000000000000D+00, &
  -6.798148477487473D-05,   0.000000000000000D+00,  -7.275755709281035D-06, &
   0.000000000000000D+00,  -9.829050261331650D-07,   0.000000000000000D+00, &
   3.950183357911352D-05,   0.000000000000000D+00,  -2.276597316614146D-05, &
   0.000000000000000D+00,  -8.617532767602914D-05,   0.000000000000000D+00, &
   1.666737861777551D-04,   0.000000000000000D+00,   1.692736013537761D-04, &
   0.000000000000000D+00,   8.714616585722223D-06,   0.000000000000000D+00, &
   3.021461403823303D-06,   0.000000000000000D+00,   2.280743365029107D-06, &
  -4.932200792671876D-05,   0.000000000000000D+00,  -4.020178743377834D-05, &
   0.000000000000000D+00,  -1.234859093334540D-04,   0.000000000000000D+00, &
  -3.713392926805651D-04,   0.000000000000000D+00,  -1.114615568087691D-04, &
   0.000000000000000D+00,   3.905782588543807D-05,   0.000000000000000D+00, &
   2.599474585631009D-06,   0.000000000000000D+00,  -5.541194488243226D-07, &
   0.000000000000000D+00,  -9.268690048917046D-06,   0.000000000000000D+00, &
   1.141134091215218D-05,   0.000000000000000D+00,   1.121864033993638D-04, &
   0.000000000000000D+00,   1.801862839340043D-04,   0.000000000000000D+00, &
  -1.971775277608934D-04,   0.000000000000000D+00,  -1.296204395388156D-04, &
   0.000000000000000D+00,  -1.742960423879594D-05,   0.000000000000000D+00, &
  -4.512812736814442D-06,   0.000000000000000D+00,  -2.031964685805803D-06, &
   1.631999140712507D-05,   0.000000000000000D+00,   5.176977926097199D-06, &
   0.000000000000000D+00,  -9.359012530631673D-06,   0.000000000000000D+00, &
   1.324163883596886D-04,   0.000000000000000D+00,   4.116212781041549D-04, &
   0.000000000000000D+00,   1.008419527919583D-04,   0.000000000000000D+00, &
   3.591189130017973D-06,   0.000000000000000D+00,   5.855928477339118D-06, &
   0.000000000000000D+00,   3.177688955982108D-06,   0.000000000000000D+00, &
  -1.367525035888935D-06,   0.000000000000000D+00,   1.307151476556868D-06, &
   0.000000000000000D+00,  -2.265641728208764D-05,   0.000000000000000D+00, &
  -1.896920356816431D-04,   0.000000000000000D+00,  -2.119415864416367D-04, &
   0.000000000000000D+00,   1.378130624497005D-04,   0.000000000000000D+00, &
   6.173662527589239D-05,   0.000000000000000D+00,   8.572259471746749D-06, &
   0.000000000000000D+00,  -5.816427055562743D-07,   0.000000000000000D+00, &
  -3.228157473252620D-06,  -2.106148299592645D-06,   0.000000000000000D+00, &
   8.494460301255502D-07,   0.000000000000000D+00,   1.160846986048235D-05, &
   0.000000000000000D+00,   7.414247977107101D-05,   0.000000000000000D+00, &
  -1.256597958578279D-04,   0.000000000000000D+00,  -3.083335357310107D-04, &
   0.000000000000000D+00,  -7.579773909290732D-05,   0.000000000000000D+00, &
  -1.973280502626783D-05,   0.000000000000000D+00,  -6.750826230289048D-06, &
   0.000000000000000D+00,  -8.475644088123569D-07,   0.000000000000000D+00, &
   7.383463375200068D-07,   0.000000000000000D+00,  -2.778013725201097D-06, &
   0.000000000000000D+00,   6.869296386490642D-08,   0.000000000000000D+00, &
   3.078068035736711D-07,   0.000000000000000D+00,   2.085949992565282D-04, &
   0.000000000000000D+00,   1.478552583046517D-04,   0.000000000000000D+00, &
  -5.214136201633949D-05,   0.000000000000000D+00,  -1.056258461863147D-05, &
   0.000000000000000D+00,   4.287589862553717D-06,   0.000000000000000D+00, &
   4.477020215169327D-06,   0.000000000000000D+00,   4.531200473689555D-06 /)
lem(:,232) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.989903956533517D-04,   3.532139947062909D-05, &
  -6.982845583213128D-04,  -5.802147166133445D-04,   3.526426453661653D-04, &
   1.701470559793998D-03,   1.774679646361575D-03,  -9.794463641125907D-04, &
  -5.312993834345387D-03,  -2.717424373469373D-03,   8.848507030166607D-03, &
   5.060844187524603D-03,  -1.081383421342154D-02,  -1.483754255750894D-03, &
   8.413125523782239D-03,  -3.504739516693463D-03,  -2.743079441040552D-03, &
   3.704805376797532D-03,  -7.459318193934739D-04,  -1.357195706148762D-03, &
   9.504062640178180D-04,   0.000000000000000D+00,  -1.989903956533578D-04, &
  -3.532139947063227D-05,   6.982845583213226D-04,   5.802147166133485D-04, &
  -3.526426453661757D-04,  -1.701470559793999D-03,  -1.774679646361569D-03, &
   9.794463641125907D-04,   5.312993834345379D-03,   2.717424373469367D-03, &
  -8.848507030166607D-03,  -5.060844187524587D-03,   1.081383421342154D-02, &
   1.483754255750891D-03,  -8.413125523782242D-03,   3.504739516693461D-03, &
   2.743079441040541D-03,  -3.704805376797519D-03,   7.459318193934752D-04, &
   1.357195706148754D-03,  -9.504062640178167D-04,   0.000000000000000D+00, &
  -9.949519782668157D-05,   0.000000000000000D+00,  -2.019486398555056D-04, &
   0.000000000000000D+00,  -5.888441261331548D-05,   0.000000000000000D+00, &
   1.352858446659848D-05,   0.000000000000000D+00,   3.038183322306608D-05, &
   0.000000000000000D+00,   6.668973198277617D-05,   0.000000000000000D+00, &
   4.467444185591678D-05,   0.000000000000000D+00,  -6.533783068958914D-05, &
   0.000000000000000D+00,  -1.124757155199695D-04,   0.000000000000000D+00, &
   4.787998778669208D-05,   0.000000000000000D+00,   8.987412061534371D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.865786034533449D-06, &
   0.000000000000000D+00,   1.058704624334300D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.215311559486416D-05,   0.000000000000000D+00, &
   2.670112456823198D-07,   0.000000000000000D+00,   8.217540702145458D-07, &
   0.000000000000000D+00,  -4.172056337503886D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.211429848792713D-05,   0.000000000000000D+00, &
   7.245102341967292D-06,   0.000000000000000D+00,  -5.172987143847838D-06, &
   0.000000000000000D+00,  -5.690525519257561D-06,   0.000000000000000D+00, &
   1.093830001948445D-05,   0.000000000000000D+00,   2.432511908578289D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.696631946073357D-05, &
   0.000000000000000D+00,  -2.567991567585530D-05,   0.000000000000000D+00, &
   3.614146374641646D-06,   0.000000000000000D+00,   1.097465535354115D-06, &
   0.000000000000000D+00,   4.393427082728745D-06,   0.000000000000000D+00, &
  -1.791493401000041D-05,   0.000000000000000D+00,  -1.770114199142554D-05, &
   0.000000000000000D+00,   2.257630499169069D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.813976843765110D-05,   0.000000000000000D+00, &
   5.657290722128429D-05,   0.000000000000000D+00,   2.474189932679552D-06, &
   0.000000000000000D+00,  -8.422652731974667D-06,   0.000000000000000D+00, &
   2.238731086576394D-06,   0.000000000000000D+00,   4.670010634386017D-06, &
   0.000000000000000D+00,   1.066529053966639D-05,   0.000000000000000D+00, &
   7.213335708563619D-05,   0.000000000000000D+00,   9.199860760341528D-06, &
   0.000000000000000D+00,  -3.851088959231865D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.351444101242739D-05,   0.000000000000000D+00, &
  -7.893491167628769D-05,   0.000000000000000D+00,  -8.665206406552048D-05, &
   0.000000000000000D+00,   3.782454156390418D-05,   0.000000000000000D+00, &
   8.834115799657551D-06,   0.000000000000000D+00,  -2.436311156580258D-07, &
   0.000000000000000D+00,  -5.532425822184609D-06,   0.000000000000000D+00, &
   1.219054792504617D-05,   0.000000000000000D+00,  -6.431671072738878D-05, &
   0.000000000000000D+00,  -1.296374993924570D-04,   0.000000000000000D+00, &
  -3.704578360347210D-06,   0.000000000000000D+00,   1.159495793382121D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.428570776313759D-05, &
   0.000000000000000D+00,   6.785185663490729D-05,   0.000000000000000D+00, &
   1.854672819814874D-04,   0.000000000000000D+00,   1.007939855441580D-04, &
   0.000000000000000D+00,  -5.534814081138659D-05,   0.000000000000000D+00, &
  -9.029978252539768D-06,   0.000000000000000D+00,  -2.938324993400566D-06, &
   0.000000000000000D+00,  -4.722425517998103D-06,   0.000000000000000D+00, &
  -2.636271651988766D-05,   0.000000000000000D+00,  -7.044139784614882D-05, &
   0.000000000000000D+00,   1.399311820285290D-04,   0.000000000000000D+00, &
   1.498478936289847D-04,   0.000000000000000D+00,   1.314455698077464D-05, &
   0.000000000000000D+00,   4.467189464488960D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.085742510554973D-05,   0.000000000000000D+00, &
  -3.494154044203553D-05,   0.000000000000000D+00,  -1.004890209114343D-04, &
   0.000000000000000D+00,  -3.116083052342537D-04,   0.000000000000000D+00, &
  -1.012433419414817D-04,   0.000000000000000D+00,   3.083350572323683D-05, &
   0.000000000000000D+00,   2.913544090586997D-06,   0.000000000000000D+00, &
  -7.673002452755913D-07,   0.000000000000000D+00,   9.577257522444321D-06, &
   0.000000000000000D+00,   1.913506460803873D-05,   0.000000000000000D+00, &
   9.388342773355986D-05,   0.000000000000000D+00,   1.509857365983602D-04, &
   0.000000000000000D+00,  -1.693310256391204D-04,   0.000000000000000D+00, &
  -1.171643056128585D-04,   0.000000000000000D+00,  -1.904061006244512D-05, &
   0.000000000000000D+00,  -5.237944719411556D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.245757336512866D-06,   0.000000000000000D+00, &
   1.122354204221141D-05,   0.000000000000000D+00,  -2.345867751112298D-06, &
   0.000000000000000D+00,   1.136062778153181D-04,   0.000000000000000D+00, &
   3.549659730247821D-04,   0.000000000000000D+00,   9.346814735262551D-05, &
   0.000000000000000D+00,   6.224036623948053D-06,   0.000000000000000D+00, &
   6.764709594030992D-06,   0.000000000000000D+00,   7.580078466142035D-06, &
   0.000000000000000D+00,  -5.015613324136078D-06,   0.000000000000000D+00, &
  -6.689137153446103D-06,   0.000000000000000D+00,  -2.105893551667286D-05, &
   0.000000000000000D+00,  -1.609615542040215D-04,   0.000000000000000D+00, &
  -1.813856345678557D-04,   0.000000000000000D+00,   1.199951378303614D-04, &
   0.000000000000000D+00,   5.634922082604448D-05,   0.000000000000000D+00, &
   8.078201219808588D-06,   0.000000000000000D+00,  -1.649791601436801D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.357056924262491D-07, &
   0.000000000000000D+00,  -3.483138447413903D-06,   0.000000000000000D+00, &
   3.523712105048840D-06,   0.000000000000000D+00,   5.820823998436113D-05, &
   0.000000000000000D+00,  -1.112361297353906D-04,   0.000000000000000D+00, &
  -2.706294732650726D-04,   0.000000000000000D+00,  -7.036444412053373D-05, &
   0.000000000000000D+00,  -1.964605599843557D-05,   0.000000000000000D+00, &
  -7.575402621526247D-06,   0.000000000000000D+00,  -3.880082062428633D-06, &
   0.000000000000000D+00,   7.494185914662516D-07,   0.000000000000000D+00, &
   1.680793232655886D-06,   0.000000000000000D+00,  -3.108452209191699D-08, &
   0.000000000000000D+00,   1.790395351492428D-06,   0.000000000000000D+00, &
   1.805463769806270D-04,   0.000000000000000D+00,   1.282363664764740D-04, &
   0.000000000000000D+00,  -4.577704056266730D-05,   0.000000000000000D+00, &
  -8.959497870035050D-06,   0.000000000000000D+00,   5.091518221692529D-06, &
   0.000000000000000D+00,   5.135784593481678D-06,   0.000000000000000D+00 /)
lem(:,233) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.146844109862631D-04,   2.401784328569584D-04,   2.919196058368769D-04, &
   4.178283871749863D-04,   2.981398148095012D-04,  -4.442094959284898D-04, &
  -1.570101988577953D-03,  -1.610787885397292D-03,   7.551224680272587D-04, &
   4.413866551032861D-03,   2.424008685068478D-03,  -6.195773850114181D-03, &
  -3.259064818080003D-03,   6.798766428212812D-03,  -5.132109235314366D-04, &
  -4.322516394918099D-03,   3.538520682769807D-03,  -1.107435210681133D-04, &
  -2.177392847669075D-03,   1.828346396323810D-03,  -3.968317547479909D-05, &
  -9.308784576905958D-04,  -1.146844109862630D-04,   2.401784328569526D-04, &
   2.919196058368838D-04,   4.178283871749903D-04,   2.981398148094986D-04, &
  -4.442094959284938D-04,  -1.570101988577953D-03,  -1.610787885397293D-03, &
   7.551224680272641D-04,   4.413866551032862D-03,   2.424008685068470D-03, &
  -6.195773850114171D-03,  -3.259064818080002D-03,   6.798766428212804D-03, &
  -5.132109235314404D-04,  -4.322516394918101D-03,   3.538520682769804D-03, &
  -1.107435210681132D-04,  -2.177392847669079D-03,   1.828346396323803D-03, &
  -3.968317547479658D-05,  -9.308784576905897D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.391274182362171D-04,   0.000000000000000D+00, &
   5.365673907432917D-05,   0.000000000000000D+00,   1.859498187051659D-04, &
   0.000000000000000D+00,   4.159429198037780D-05,   0.000000000000000D+00, &
  -9.511283040478254D-05,   0.000000000000000D+00,  -1.590841447488417D-04, &
   0.000000000000000D+00,   2.350867060881136D-06,   0.000000000000000D+00, &
   2.225386679690506D-04,   0.000000000000000D+00,   2.599598904187262D-05, &
   0.000000000000000D+00,  -1.693738249588746D-04,   0.000000000000000D+00, &
   1.082051167258837D-05,  -2.467296476889505D-06,   0.000000000000000D+00, &
  -1.918436423495766D-05,   0.000000000000000D+00,   9.277775367067025D-07, &
  -1.046918234050215D-06,   0.000000000000000D+00,  -5.128889188349949D-06, &
   0.000000000000000D+00,   2.467299438917215D-05,   0.000000000000000D+00, &
   5.215304633652444D-06,   0.000000000000000D+00,  -7.165085488414927D-06, &
   1.416440472446691D-05,   0.000000000000000D+00,   1.267338046398573D-05, &
   0.000000000000000D+00,   5.685716904755840D-06,   0.000000000000000D+00, &
  -1.225562914440491D-05,   0.000000000000000D+00,  -1.125153266690854D-05, &
   0.000000000000000D+00,   8.130047213091291D-06,   0.000000000000000D+00, &
   4.115078843951780D-06,  -2.526503892269882D-05,   0.000000000000000D+00, &
  -2.089228083703509D-05,   0.000000000000000D+00,  -2.008486570530813D-05, &
   0.000000000000000D+00,   2.802080502769115D-06,   0.000000000000000D+00, &
  -1.386292400708614D-05,   0.000000000000000D+00,   6.955396264001499D-06, &
   0.000000000000000D+00,  -1.106130309276023D-05,   0.000000000000000D+00, &
  -1.771644645551902D-05,   0.000000000000000D+00,   3.861592932462969D-06, &
   1.892391684200441D-05,   0.000000000000000D+00,   2.582140127132252D-05, &
   0.000000000000000D+00,   4.580970062357324D-05,   0.000000000000000D+00, &
   4.862572647822637D-06,   0.000000000000000D+00,  -1.117854527884966D-05, &
   0.000000000000000D+00,   3.652069281774566D-05,   0.000000000000000D+00, &
   7.651438223117655D-06,   0.000000000000000D+00,   6.308988330746158D-06, &
   0.000000000000000D+00,   5.976969387622577D-05,   0.000000000000000D+00, &
   1.496834962015458D-05,   0.000000000000000D+00,  -7.900418679193386D-06, &
  -5.198560764020832D-06,   0.000000000000000D+00,  -2.628111377632024D-05, &
   0.000000000000000D+00,  -6.455958639164686D-05,   0.000000000000000D+00, &
  -7.542394732804667D-05,   0.000000000000000D+00,   2.929139309095697D-05, &
   0.000000000000000D+00,   1.494575943574779D-05,   0.000000000000000D+00, &
  -3.984850035864523D-05,   0.000000000000000D+00,  -1.482209317175121D-05, &
   0.000000000000000D+00,   7.256034637345701D-06,   0.000000000000000D+00, &
  -5.195962116696768D-05,   0.000000000000000D+00,  -1.157844890036444D-04, &
   0.000000000000000D+00,  -1.194646067913975D-05,   0.000000000000000D+00, &
   3.293559447706156D-06,   1.480362207488238D-06,   0.000000000000000D+00, &
   1.967709233667479D-05,   0.000000000000000D+00,   5.464018909126182D-05, &
   0.000000000000000D+00,   1.562845070225248D-04,   0.000000000000000D+00, &
   9.318643739626760D-05,   0.000000000000000D+00,  -4.687908158671962D-05, &
   0.000000000000000D+00,  -1.374470033904040D-05,   0.000000000000000D+00, &
   2.474350600058438D-05,   0.000000000000000D+00,   8.122055211943368D-06, &
   0.000000000000000D+00,  -1.668350523440920D-05,   0.000000000000000D+00, &
  -5.685990454145791D-05,   0.000000000000000D+00,   1.195751246510810D-04, &
   0.000000000000000D+00,   1.395466931358522D-04,   0.000000000000000D+00, &
   1.946541153186402D-05,   0.000000000000000D+00,   8.802141441659859D-06, &
  -3.656509891881802D-06,   0.000000000000000D+00,  -7.674834321735568D-06, &
   0.000000000000000D+00,  -2.518337087469118D-05,   0.000000000000000D+00, &
  -8.553375478635306D-05,   0.000000000000000D+00,  -2.711303393648741D-04, &
   0.000000000000000D+00,  -9.707356121825145D-05,   0.000000000000000D+00, &
   2.543362615728205D-05,   0.000000000000000D+00,   2.766573778660442D-06, &
   0.000000000000000D+00,  -7.992202032739052D-06,   0.000000000000000D+00, &
  -1.093492335793109D-06,   0.000000000000000D+00,   1.352047695520183D-05, &
   0.000000000000000D+00,   7.661800304396016D-05,   0.000000000000000D+00, &
   1.284719113879242D-04,   0.000000000000000D+00,  -1.492249206558482D-04, &
   0.000000000000000D+00,  -1.111284759532496D-04,   0.000000000000000D+00, &
  -2.203391316758353D-05,   0.000000000000000D+00,  -1.249750163080626D-05, &
   2.717622145128055D-06,   0.000000000000000D+00,  -4.470291288082495D-07, &
   0.000000000000000D+00,   3.812243026677166D-06,   0.000000000000000D+00, &
  -1.855911602752209D-06,   0.000000000000000D+00,   1.008419527919583D-04, &
   0.000000000000000D+00,   3.162354223682567D-04,   0.000000000000000D+00, &
   9.028083549276155D-05,   0.000000000000000D+00,   9.365640963295851D-06, &
   0.000000000000000D+00,   9.792871165858155D-06,   0.000000000000000D+00, &
   7.553330780832099D-07,   0.000000000000000D+00,   2.763479858960385D-07, &
   0.000000000000000D+00,  -5.892031802574869D-06,   0.000000000000000D+00, &
  -1.918955065049504D-05,   0.000000000000000D+00,  -1.383779380916542D-04, &
   0.000000000000000D+00,  -1.581815656723966D-04,   0.000000000000000D+00, &
   1.075681723197740D-04,   0.000000000000000D+00,   5.310941712986958D-05, &
   0.000000000000000D+00,   8.536090945585502D-06,   0.000000000000000D+00, &
  -2.031169300788143D-06,  -5.957863669549901D-07,   0.000000000000000D+00, &
   1.318461731788463D-06,   0.000000000000000D+00,   1.843580166393435D-06, &
   0.000000000000000D+00,   5.086099327876450D-06,   0.000000000000000D+00, &
   4.847205200282096D-05,   0.000000000000000D+00,  -1.017647794012816D-04, &
   0.000000000000000D+00,  -2.449051358667865D-04,   0.000000000000000D+00, &
  -6.781207454526163D-05,   0.000000000000000D+00,  -2.137412590529461D-05, &
   0.000000000000000D+00,  -8.266829069890626D-06,   0.000000000000000D+00, &
   1.347494843811492D-07,   0.000000000000000D+00,  -6.540539772017890D-07, &
   0.000000000000000D+00,   1.771561011526714D-06,   0.000000000000000D+00, &
   2.597857656062465D-06,   0.000000000000000D+00,   4.553402160804710D-06, &
   0.000000000000000D+00,   1.593825366789872D-04,   0.000000000000000D+00, &
   1.130119176191189D-04,   0.000000000000000D+00,  -4.082486833544025D-05, &
   0.000000000000000D+00,  -7.496641483356039D-06,   0.000000000000000D+00, &
   5.622026519723173D-06,   0.000000000000000D+00,   1.419191310154802D-05 /)
lem(:,234) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.185487267672309D-04,  -3.185606193617972D-04, &
  -6.214198789673220D-04,  -5.515803285148335D-04,   1.640252606037143D-04, &
   1.207034098161395D-03,   1.591323086850937D-03,   9.932583264857649D-05, &
  -2.899096075811537D-03,  -2.382530869084187D-03,   3.486925939951582D-03, &
   2.344541529933424D-03,  -3.616803872218934D-03,   6.603812716468078D-04, &
   1.559002001436316D-03,  -2.167796599134706D-03,   1.192624770343383D-03, &
   4.337760615674344D-04,  -1.316022768776292D-03,   9.163115523765572D-04, &
   8.972270665341530D-05,   0.000000000000000D+00,   1.185487267672249D-04, &
   3.185606193617934D-04,   6.214198789673259D-04,   5.515803285148310D-04, &
  -1.640252606037220D-04,  -1.207034098161391D-03,  -1.591323086850930D-03, &
  -9.932583264857710D-05,   2.899096075811535D-03,   2.382530869084181D-03, &
  -3.486925939951593D-03,  -2.344541529933420D-03,   3.616803872218943D-03, &
  -6.603812716468116D-04,  -1.559002001436314D-03,   2.167796599134704D-03, &
  -1.192624770343385D-03,  -4.337760615674296D-04,   1.316022768776287D-03, &
  -9.163115523765660D-04,  -8.972270665341020D-05,   0.000000000000000D+00, &
   5.927436338361391D-05,   0.000000000000000D+00,   3.546214174957490D-04, &
   0.000000000000000D+00,   1.958016608062649D-04,   0.000000000000000D+00, &
  -3.924315610979108D-04,   0.000000000000000D+00,  -2.544258394786315D-04, &
   0.000000000000000D+00,   3.424776264198547D-04,   0.000000000000000D+00, &
   3.089564007954139D-04,   0.000000000000000D+00,  -3.240512208221560D-04, &
   0.000000000000000D+00,  -2.615300388924301D-04,   0.000000000000000D+00, &
   2.560317753143088D-04,   0.000000000000000D+00,   8.971850052716267D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.686049878717594D-06, &
   0.000000000000000D+00,  -5.176496144903330D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.765498916717544D-07,   0.000000000000000D+00, &
  -1.069840312062509D-05,   0.000000000000000D+00,   5.712148150407415D-06, &
   0.000000000000000D+00,   6.933294466998926D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.496458501928970D-06,   0.000000000000000D+00, &
   1.086222571123381D-05,   0.000000000000000D+00,   1.415698337952303D-05, &
   0.000000000000000D+00,  -2.519131021512728D-06,   0.000000000000000D+00, &
  -8.433308288596034D-06,   0.000000000000000D+00,   5.005675668220130D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.556554652297497D-06, &
   0.000000000000000D+00,  -1.456160442621457D-05,   0.000000000000000D+00, &
  -2.123922242831889D-05,   0.000000000000000D+00,   6.616743135223169D-06, &
   0.000000000000000D+00,  -3.154373621802899D-06,   0.000000000000000D+00, &
   4.100113818477098D-06,   0.000000000000000D+00,  -7.525210330508650D-06, &
   0.000000000000000D+00,  -2.237205335608971D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.296355646194695D-06,   0.000000000000000D+00, &
   1.693420485700574D-05,   0.000000000000000D+00,   3.931043974709304D-05, &
   0.000000000000000D+00,   1.011526826590353D-05,   0.000000000000000D+00, &
  -2.701078374656346D-05,   0.000000000000000D+00,   9.275081226669540D-06, &
   0.000000000000000D+00,   4.503230546167889D-06,   0.000000000000000D+00, &
   4.458250584561109D-06,   0.000000000000000D+00,   5.435353421845499D-05, &
   0.000000000000000D+00,   2.615487253521705D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.641613684849831D-06,   0.000000000000000D+00, &
  -1.913834648394661D-05,   0.000000000000000D+00,  -5.217238138723947D-05, &
   0.000000000000000D+00,  -7.096979305170024D-05,   0.000000000000000D+00, &
   2.289199204628304D-05,   0.000000000000000D+00,   3.495369216324462D-05, &
   0.000000000000000D+00,  -1.120897621350297D-05,   0.000000000000000D+00, &
  -8.639093250299687D-06,   0.000000000000000D+00,   6.137757375244322D-06, &
   0.000000000000000D+00,  -4.408537227763391D-05,   0.000000000000000D+00, &
  -1.118897334456108D-04,   0.000000000000000D+00,  -2.487842926762715D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.380305875676559D-06, &
   0.000000000000000D+00,   1.824834975540136D-05,   0.000000000000000D+00, &
   4.534185147470755D-05,   0.000000000000000D+00,   1.360339689929573D-04, &
   0.000000000000000D+00,   9.169720756063272D-05,   0.000000000000000D+00, &
  -4.154939029690186D-05,   0.000000000000000D+00,  -3.201898158050232D-05, &
   0.000000000000000D+00,   6.345701570372750D-06,   0.000000000000000D+00, &
   5.246594270622900D-06,   0.000000000000000D+00,  -1.401480983130200D-05, &
   0.000000000000000D+00,  -4.853173148260311D-05,   0.000000000000000D+00, &
   1.072748116923373D-04,   0.000000000000000D+00,   1.385635740867772D-04, &
   0.000000000000000D+00,   3.155134099353444D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.084838325676051D-06,   0.000000000000000D+00, &
  -1.153688923273701D-05,   0.000000000000000D+00,  -2.336597963320795D-05, &
   0.000000000000000D+00,  -7.567297100655205D-05,   0.000000000000000D+00, &
  -2.445222861410668D-04,   0.000000000000000D+00,  -9.816565708540655D-05, &
   0.000000000000000D+00,   2.051424809275798D-05,   0.000000000000000D+00, &
   9.033345531893765D-06,   0.000000000000000D+00,  -5.584902872993456D-07, &
   0.000000000000000D+00,  -1.454643755887772D-06,   0.000000000000000D+00, &
   1.063069404432594D-05,   0.000000000000000D+00,   6.599321603347863D-05, &
   0.000000000000000D+00,   1.127727838228724D-04,   0.000000000000000D+00, &
  -1.370636879900869D-04,   0.000000000000000D+00,  -1.119313801397589D-04, &
   0.000000000000000D+00,  -2.877726528156695D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.863802809376545D-06,   0.000000000000000D+00, &
   4.401502535354291D-06,   0.000000000000000D+00,   5.666746810816224D-06, &
   0.000000000000000D+00,   1.311683244667325D-06,   0.000000000000000D+00, &
   9.346814735262551D-05,   0.000000000000000D+00,   2.913079010679365D-04, &
   0.000000000000000D+00,   9.262202026339571D-05,   0.000000000000000D+00, &
   1.406214145060302D-05,   0.000000000000000D+00,   2.212671703853364D-05, &
   0.000000000000000D+00,  -9.416816164687176D-07,   0.000000000000000D+00, &
   6.115714896787794D-07,   0.000000000000000D+00,  -2.914896219742205D-06, &
   0.000000000000000D+00,  -1.722084298308033D-05,   0.000000000000000D+00, &
  -1.232054110064129D-04,   0.000000000000000D+00,  -1.416911035642557D-04, &
   0.000000000000000D+00,   9.965063603376556D-05,   0.000000000000000D+00, &
   5.352724990859746D-05,   0.000000000000000D+00,   7.215740763864568D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.655479006559587D-07, &
   0.000000000000000D+00,  -1.348328933546672D-06,   0.000000000000000D+00, &
   1.888411626883917D-07,   0.000000000000000D+00,   3.884438609862693D-06, &
   0.000000000000000D+00,   3.941841728083481D-05,   0.000000000000000D+00, &
  -9.687627556947392D-05,   0.000000000000000D+00,  -2.292063613300664D-04, &
   0.000000000000000D+00,  -6.974071543768337D-05,   0.000000000000000D+00, &
  -2.492239135046754D-05,   0.000000000000000D+00,  -2.561272050025838D-05, &
   0.000000000000000D+00,   2.698431407741391D-07,   0.000000000000000D+00, &
  -3.136783918782093D-07,   0.000000000000000D+00,  -7.188408854951918D-07, &
   0.000000000000000D+00,   1.162251706925325D-06,   0.000000000000000D+00, &
   6.485187030527794D-06,   0.000000000000000D+00,   1.448827839739571D-04, &
   0.000000000000000D+00,   1.027726172881070D-04,   0.000000000000000D+00, &
  -3.760093387442135D-05,   0.000000000000000D+00,  -6.604725368103485D-06, &
   0.000000000000000D+00,   8.493977075866153D-06,   0.000000000000000D+00 /)
lem(:,235) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.378983969496708D-04,   6.914570983309882D-05,   3.409222393200400D-04, &
   6.300712800446767D-04,   4.725239694563815D-04,   5.397962926731126D-05, &
  -7.149458863640556D-04,  -1.396712853915747D-03,  -8.187684978544555D-04, &
   1.520373995418107D-03,   2.371593158263596D-03,  -1.414644885469618D-03, &
  -1.830771232340610D-03,   1.672008204830899D-03,  -2.713847047750129D-04, &
  -4.311582314150004D-04,   8.300648647703142D-04,  -9.553645451304868D-04, &
   4.321168995811320D-04,   3.601134892962729D-04,  -7.469141173625074D-04, &
   4.885389527096583D-04,   1.378983969496694D-04,   6.914570983309323D-05, &
   3.409222393200429D-04,   6.300712800446811D-04,   4.725239694563767D-04, &
   5.397962926730994D-05,  -7.149458863640469D-04,  -1.396712853915748D-03, &
  -8.187684978544601D-04,   1.520373995418109D-03,   2.371593158263598D-03, &
  -1.414644885469621D-03,  -1.830771232340608D-03,   1.672008204830904D-03, &
  -2.713847047750154D-04,  -4.311582314150013D-04,   8.300648647703141D-04, &
  -9.553645451304908D-04,   4.321168995811331D-04,   3.601134892962754D-04, &
  -7.469141173625129D-04,   4.885389527096610D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.137240351520953D-04,   0.000000000000000D+00, &
   7.732162100189771D-06,   0.000000000000000D+00,  -8.133407655562778D-04, &
   0.000000000000000D+00,   2.294807870647616D-04,   0.000000000000000D+00, &
   1.275719912654162D-03,   0.000000000000000D+00,  -1.413383826833486D-03, &
   0.000000000000000D+00,  -5.063169041524536D-06,   0.000000000000000D+00, &
   8.557736824749128D-04,   0.000000000000000D+00,  -2.239398471700363D-04, &
   0.000000000000000D+00,  -3.359648099526909D-04,   0.000000000000000D+00, &
  -7.351228054353110D-06,  -5.707183441558477D-06,   0.000000000000000D+00, &
  -1.823163382671448D-06,   0.000000000000000D+00,  -1.336217774861021D-05, &
  -1.094766583712921D-06,   0.000000000000000D+00,   1.278278387418141D-06, &
   0.000000000000000D+00,   2.327514856933929D-06,   0.000000000000000D+00, &
   8.394927555132406D-06,   0.000000000000000D+00,   1.598899915446957D-05, &
   7.727166852432858D-06,   0.000000000000000D+00,   2.017819551046823D-06, &
   0.000000000000000D+00,   7.781491300240842D-06,   0.000000000000000D+00, &
   5.583043543280519D-06,   0.000000000000000D+00,  -3.827105495344455D-06, &
   0.000000000000000D+00,  -1.009534323885802D-05,   0.000000000000000D+00, &
   1.674707890967960D-05,  -3.193023752324895D-06,   0.000000000000000D+00, &
  -5.791558564189709D-06,   0.000000000000000D+00,  -1.217319718386050D-05, &
   0.000000000000000D+00,  -2.301775229903317D-05,   0.000000000000000D+00, &
  -1.188465224420958D-05,   0.000000000000000D+00,  -3.930314540584710D-06, &
   0.000000000000000D+00,   5.245448267263976D-06,   0.000000000000000D+00, &
  -7.337893629578542D-06,   0.000000000000000D+00,  -5.483676842919187D-05, &
  -8.375093779761063D-06,   0.000000000000000D+00,   4.546357268700668D-06, &
   0.000000000000000D+00,   1.451984806030163D-05,   0.000000000000000D+00, &
   3.782468359638852D-05,   0.000000000000000D+00,   1.983420605506433D-05, &
   0.000000000000000D+00,   1.125044454999954D-05,   0.000000000000000D+00, &
   9.825117296179254D-06,   0.000000000000000D+00,   3.942967856716201D-06, &
   0.000000000000000D+00,   2.846423405964955D-06,   0.000000000000000D+00, &
   5.482594733957814D-05,   0.000000000000000D+00,   5.992797053510557D-05, &
   1.434068144405011D-05,   0.000000000000000D+00,  -9.595585048613920D-07, &
   0.000000000000000D+00,  -1.517840789503506D-05,   0.000000000000000D+00, &
  -4.683564140979155D-05,   0.000000000000000D+00,  -7.169694232669713D-05, &
   0.000000000000000D+00,   1.384329840576914D-05,   0.000000000000000D+00, &
  -8.355286063250623D-06,   0.000000000000000D+00,  -9.503669450444258D-06, &
   0.000000000000000D+00,  -8.495743304961809D-06,   0.000000000000000D+00, &
   4.576568785312897D-06,   0.000000000000000D+00,  -3.963893556576631D-05, &
   0.000000000000000D+00,  -1.172932108680935D-04,   0.000000000000000D+00, &
  -5.473796917941583D-05,  -1.005049985553653D-05,   0.000000000000000D+00, &
   1.417775552407557D-07,   0.000000000000000D+00,   1.317836339365469D-05, &
   0.000000000000000D+00,   3.999486513400433D-05,   0.000000000000000D+00, &
   1.250499705151469D-04,   0.000000000000000D+00,   9.695934498609987D-05, &
   0.000000000000000D+00,  -3.646398742605504D-05,   0.000000000000000D+00, &
   5.546714867278306D-06,   0.000000000000000D+00,   4.194456982920293D-06, &
   0.000000000000000D+00,   4.348165751845187D-06,   0.000000000000000D+00, &
  -1.097487800702617D-05,   0.000000000000000D+00,  -4.157274217652752D-05, &
   0.000000000000000D+00,   1.006891004333143D-04,   0.000000000000000D+00, &
   1.498909484190325D-04,   0.000000000000000D+00,   7.053464895918325D-05, &
   3.014621269139756D-06,   0.000000000000000D+00,  -3.631597740411696D-07, &
   0.000000000000000D+00,  -7.771720880455951D-06,   0.000000000000000D+00, &
  -2.047190374844721D-05,   0.000000000000000D+00,  -7.055712236691485D-05, &
   0.000000000000000D+00,  -2.311041965836646D-04,   0.000000000000000D+00, &
  -1.074049852406195D-04,   0.000000000000000D+00,   1.142310559360769D-05, &
   0.000000000000000D+00,  -2.383199052386146D-06,   0.000000000000000D+00, &
   2.111053442806960D-10,   0.000000000000000D+00,   7.037340206148654D-07, &
   0.000000000000000D+00,   9.639672819885853D-06,   0.000000000000000D+00, &
   5.802353269977631D-05,   0.000000000000000D+00,   1.021311275934576D-04, &
   0.000000000000000D+00,  -1.319572193508490D-04,   0.000000000000000D+00, &
  -1.227722763904277D-04,   0.000000000000000D+00,  -7.296267095599671D-05, &
   1.827067200627219D-08,   0.000000000000000D+00,  -3.098432144022509D-07, &
   0.000000000000000D+00,   2.606182153950875D-06,   0.000000000000000D+00, &
   5.015032475376303D-06,   0.000000000000000D+00,   3.591189130017973D-06, &
   0.000000000000000D+00,   9.028083549276155D-05,   0.000000000000000D+00, &
   2.819491509402355D-04,   0.000000000000000D+00,   1.024826521442064D-04, &
   0.000000000000000D+00,   2.706574058124700D-05,   0.000000000000000D+00, &
   3.886876641381362D-07,   0.000000000000000D+00,  -5.257257457334870D-07, &
   0.000000000000000D+00,  -1.668686000508449D-06,   0.000000000000000D+00, &
  -4.001669695943166D-06,   0.000000000000000D+00,  -1.718316973190151D-05, &
   0.000000000000000D+00,  -1.129511153117156D-04,   0.000000000000000D+00, &
  -1.321795139069115D-04,   0.000000000000000D+00,   9.723925216335021D-05, &
   0.000000000000000D+00,   5.950997473991586D-05,   0.000000000000000D+00, &
   2.842066084329331D-05,  -1.453302700322681D-07,   0.000000000000000D+00, &
   4.904245836879994D-07,   0.000000000000000D+00,  -4.323715722690545D-07, &
   0.000000000000000D+00,  -3.767689162647804D-07,   0.000000000000000D+00, &
   3.375673008537472D-06,   0.000000000000000D+00,   3.268533547065315D-05, &
   0.000000000000000D+00,  -9.618234571028476D-05,   0.000000000000000D+00, &
  -2.257894643903737D-04,   0.000000000000000D+00,  -7.737286396967558D-05, &
   0.000000000000000D+00,  -2.991308411233357D-05,   0.000000000000000D+00, &
   2.458524223656090D-08,   0.000000000000000D+00,  -1.418397475902034D-08, &
   0.000000000000000D+00,   8.892304893331483D-07,   0.000000000000000D+00, &
   6.821789222816936D-07,   0.000000000000000D+00,   1.985055883125411D-06, &
   0.000000000000000D+00,   9.198069850168681D-06,   0.000000000000000D+00, &
   1.366703211299455D-04,   0.000000000000000D+00,   9.708119130064546D-05, &
   0.000000000000000D+00,  -3.643589364269878D-05,   0.000000000000000D+00, &
  -8.919460601262779D-06,   0.000000000000000D+00,   2.131792182067722D-05 /)
lem(:,236) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.002329040924052D-04,  -4.505938348017801D-04, &
  -2.591370700339375D-04,  -3.510399526443746D-04,  -5.001060462611574D-04, &
   1.133249644485418D-04,   9.345594980078846D-04,   1.165879372673940D-03, &
  -3.452145576711752D-04,  -2.000076477592027D-03,   2.017268959540615D-04, &
   1.455665543040401D-03,  -6.467092486781277D-04,  -5.969727120885923D-05, &
   1.277799709059544D-04,  -2.008212202814706D-04,   3.899892863016681D-04, &
  -4.399219770023847D-04,   1.712597394822222D-04,   2.205172998197043D-04, &
  -4.107246350380445D-04,   0.000000000000000D+00,   2.002329040924013D-04, &
   4.505938348017861D-04,   2.591370700339417D-04,   3.510399526443641D-04, &
   5.001060462611574D-04,  -1.133249644485313D-04,  -9.345594980078868D-04, &
  -1.165879372673943D-03,   3.452145576711781D-04,   2.000076477592027D-03, &
  -2.017268959540642D-04,  -1.455665543040399D-03,   6.467092486781268D-04, &
   5.969727120886102D-05,  -1.277799709059586D-04,   2.008212202814712D-04, &
  -3.899892863016621D-04,   4.399219770023868D-04,  -1.712597394822283D-04, &
  -2.205172998197107D-04,   4.107246350380491D-04,   0.000000000000000D+00, &
   1.001164520462002D-04,   0.000000000000000D+00,  -1.442349084831599D-05, &
   0.000000000000000D+00,  -7.079899062352527D-04,   0.000000000000000D+00, &
   6.908476520894726D-04,   0.000000000000000D+00,   1.491893559204493D-03, &
   0.000000000000000D+00,  -3.840742928368895D-03,   0.000000000000000D+00, &
   3.990164163591345D-03,   0.000000000000000D+00,  -1.858120704770775D-03, &
   0.000000000000000D+00,  -4.976935463551129D-04,   0.000000000000000D+00, &
   8.653313879379439D-04,   0.000000000000000D+00,  -1.058988895726153D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.914900493025535D-06, &
   0.000000000000000D+00,  -4.004422949735453D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.431740712990790D-06,   0.000000000000000D+00, &
   4.768858528624199D-06,   0.000000000000000D+00,   2.829978181929700D-06, &
   0.000000000000000D+00,   3.793706293527953D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.088968241631945D-07,   0.000000000000000D+00, &
   1.428478203438647D-06,   0.000000000000000D+00,   2.551148259738402D-05, &
   0.000000000000000D+00,   6.956684512202091D-07,   0.000000000000000D+00, &
  -2.552650301334413D-06,   0.000000000000000D+00,  -7.109985276162786D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.002895869787296D-08, &
   0.000000000000000D+00,  -4.317457981452423D-06,   0.000000000000000D+00, &
  -1.506720250677384D-05,   0.000000000000000D+00,  -6.316777652593629D-05, &
   0.000000000000000D+00,  -3.382592243942506D-06,   0.000000000000000D+00, &
  -1.536498703459089D-06,   0.000000000000000D+00,   4.387275408524590D-06, &
   0.000000000000000D+00,   2.033943703056921D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.609180519410166D-06,   0.000000000000000D+00, &
   3.332965081997447D-06,   0.000000000000000D+00,   1.385683850462090D-05, &
   0.000000000000000D+00,   4.165561136999599D-05,   0.000000000000000D+00, &
   4.176004316552767D-05,   0.000000000000000D+00,   4.837314277104507D-06, &
   0.000000000000000D+00,   6.172096971217263D-06,   0.000000000000000D+00, &
   3.201064223635796D-06,   0.000000000000000D+00,   6.127497863439598D-07, &
   0.000000000000000D+00,   5.131266344095265D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.631142949946538D-06,   0.000000000000000D+00, &
  -1.583013552534162D-06,   0.000000000000000D+00,  -1.383792632320035D-05, &
   0.000000000000000D+00,  -4.345089029750039D-05,   0.000000000000000D+00, &
  -7.783641413582491D-05,   0.000000000000000D+00,   4.737019835575434D-05, &
   0.000000000000000D+00,  -5.480653014386482D-06,   0.000000000000000D+00, &
  -7.989333712251705D-06,   0.000000000000000D+00,  -7.178533905221297D-06, &
   0.000000000000000D+00,   2.934689306473592D-06,   0.000000000000000D+00, &
  -3.738631819508166D-05,   0.000000000000000D+00,  -1.391325209535254D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.077609480805548D-06, &
   0.000000000000000D+00,   1.473037030042468D-06,   0.000000000000000D+00, &
   1.231272551480222D-05,   0.000000000000000D+00,   3.738595956100998D-05, &
   0.000000000000000D+00,   1.219068770101146D-04,   0.000000000000000D+00, &
   1.104709893287914D-04,   0.000000000000000D+00,  -1.050065097794215D-04, &
   0.000000000000000D+00,   4.058205978929093D-06,   0.000000000000000D+00, &
   5.563309496557967D-06,   0.000000000000000D+00,   4.433491271351753D-06, &
   0.000000000000000D+00,  -8.456431114092623D-06,   0.000000000000000D+00, &
  -3.787868184592714D-05,   0.000000000000000D+00,   1.025305506862585D-04, &
   0.000000000000000D+00,   1.905405625380020D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.312012277060443D-07,   0.000000000000000D+00, &
  -1.314904229796198D-06,   0.000000000000000D+00,  -7.302531381201347D-06, &
   0.000000000000000D+00,  -2.050210225941692D-05,   0.000000000000000D+00, &
  -6.903063431367236D-05,   0.000000000000000D+00,  -2.338542063449625D-04, &
   0.000000000000000D+00,  -1.270457297444497D-04,   0.000000000000000D+00, &
   5.103901439563160D-05,   0.000000000000000D+00,  -1.400848112806544D-06, &
   0.000000000000000D+00,  -2.206076367312732D-06,   0.000000000000000D+00, &
  -6.080824198529455D-07,   0.000000000000000D+00,   8.178388822536488D-06, &
   0.000000000000000D+00,   5.334722153403152D-05,   0.000000000000000D+00, &
   9.818689031579365D-05,   0.000000000000000D+00,  -1.380973606878615D-04, &
   0.000000000000000D+00,  -1.508433630852731D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.426197788674532D-07,   0.000000000000000D+00, &
   4.357177230679232D-07,   0.000000000000000D+00,   1.961553110367963D-06, &
   0.000000000000000D+00,   5.484577930766048D-06,   0.000000000000000D+00, &
   6.224036623948053D-06,   0.000000000000000D+00,   9.262202026339571D-05, &
   0.000000000000000D+00,   2.918743029959147D-04,   0.000000000000000D+00, &
   1.202981155273742D-04,   0.000000000000000D+00,   5.420352392877237D-05, &
   0.000000000000000D+00,   4.845679010078038D-09,   0.000000000000000D+00, &
   7.019845808472722D-07,   0.000000000000000D+00,  -3.823635665483722D-07, &
   0.000000000000000D+00,  -3.492557844954443D-06,   0.000000000000000D+00, &
  -1.702918359258072D-05,   0.000000000000000D+00,  -1.093347998697084D-04, &
   0.000000000000000D+00,  -1.300464494477475D-04,   0.000000000000000D+00, &
   1.032570962822610D-04,   0.000000000000000D+00,   5.847108482625844D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.705076022232827D-07, &
   0.000000000000000D+00,  -5.262231977228530D-08,   0.000000000000000D+00, &
   3.229366599603510D-07,   0.000000000000000D+00,   6.908190796180910D-08, &
   0.000000000000000D+00,   3.087396195832210D-06,   0.000000000000000D+00, &
   2.738429611879531D-05,   0.000000000000000D+00,  -1.017098047831503D-04, &
   0.000000000000000D+00,  -2.373951463869292D-04,   0.000000000000000D+00, &
  -8.992454278079721D-05,   0.000000000000000D+00,  -9.258143212135926D-05, &
   0.000000000000000D+00,   7.755569252890826D-08,   0.000000000000000D+00, &
  -2.803475813892230D-07,   0.000000000000000D+00,   4.209744727239519D-08, &
   0.000000000000000D+00,   5.964259706634066D-07,   0.000000000000000D+00, &
   1.888264261638642D-06,   0.000000000000000D+00,   1.248713199774328D-05, &
   0.000000000000000D+00,   1.354737406918140D-04,   0.000000000000000D+00, &
   9.647325355584607D-05,   0.000000000000000D+00,  -3.747617516986098D-05, &
   0.000000000000000D+00,   2.676548150715798D-06,   0.000000000000000D+00 /)
lem(:,237) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.182232814900278D-05,  -5.651974777978774D-05,  -2.122635291758168D-04, &
   9.322955872164833D-05,   7.080964831475015D-04,   5.006099681836830D-04, &
   1.292204858753893D-04,  -2.755225349223114D-04,  -1.171527899488467D-03, &
  -5.359554217753594D-04,   1.425080913077551D-03,   4.665895990675903D-04, &
  -1.048858225679758D-03,   1.882797668312693D-04,   2.187744717426349D-04, &
  -9.601010554906479D-05,   2.606084259074267D-05,  -9.473079560002069D-05, &
   1.888286043416057D-04,  -1.975312190991627D-04,   6.661730099037937D-05, &
   1.205012434934586D-04,   1.182232814900208D-05,  -5.651974777979886D-05, &
  -2.122635291758178D-04,   9.322955872165721D-05,   7.080964831474928D-04, &
   5.006099681836799D-04,   1.292204858754039D-04,  -2.755225349223104D-04, &
  -1.171527899488478D-03,  -5.359554217753582D-04,   1.425080913077553D-03, &
   4.665895990675901D-04,  -1.048858225679757D-03,   1.882797668312723D-04, &
   2.187744717426373D-04,  -9.601010554906711D-05,   2.606084259074048D-05, &
  -9.473079560001887D-05,   1.888286043416062D-04,  -1.975312190991635D-04, &
   6.661730099037707D-05,   1.205012434934604D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.679497162982350D-05,   0.000000000000000D+00, &
  -3.047427774096723D-04,   0.000000000000000D+00,   4.505413508624836D-04, &
   0.000000000000000D+00,   1.203803620697113D-03,   0.000000000000000D+00, &
  -4.159199570352192D-03,   0.000000000000000D+00,   5.618751461711570D-03, &
   0.000000000000000D+00,  -5.634747913246654D-03,   0.000000000000000D+00, &
   4.430387182831808D-03,   0.000000000000000D+00,  -1.598362518858990D-03, &
   0.000000000000000D+00,  -3.842671721461833D-04,   0.000000000000000D+00, &
  -1.617622467231901D-06,  -2.457259537337573D-06,   0.000000000000000D+00, &
   3.552310681215217D-07,   0.000000000000000D+00,  -3.035610945767253D-06, &
   2.209297863595004D-06,   0.000000000000000D+00,  -1.881527760434014D-06, &
   0.000000000000000D+00,  -2.667281720016261D-06,   0.000000000000000D+00, &
   1.071338961964005D-06,   0.000000000000000D+00,   1.750585704443136D-05, &
  -3.253988849692021D-06,   0.000000000000000D+00,   2.819678461419452D-06, &
   0.000000000000000D+00,   4.066131080291206D-06,   0.000000000000000D+00, &
   3.965815530072586D-06,   0.000000000000000D+00,   7.205169113168067D-07, &
   0.000000000000000D+00,  -5.996607400458939D-06,   0.000000000000000D+00, &
  -3.154761986423964D-05,   5.810301211193466D-06,   0.000000000000000D+00, &
  -9.888181133304385D-07,   0.000000000000000D+00,  -4.916734406731237D-06, &
   0.000000000000000D+00,  -6.099700345909401D-06,   0.000000000000000D+00, &
  -3.860671290809350D-07,   0.000000000000000D+00,  -2.018805218740978D-06, &
   0.000000000000000D+00,  -6.534802057322068D-07,   0.000000000000000D+00, &
   8.201613837450140D-06,   0.000000000000000D+00,  -1.451057270756055D-05, &
  -7.059422236891090D-06,   0.000000000000000D+00,  -2.726642024120173D-06, &
   0.000000000000000D+00,   3.216347435852936D-06,   0.000000000000000D+00, &
   9.680146759398490D-06,   0.000000000000000D+00,   3.220106730517380D-05, &
   0.000000000000000D+00,  -3.510947529577148D-06,   0.000000000000000D+00, &
   2.151862612583055D-06,   0.000000000000000D+00,   5.439194917684101D-06, &
   0.000000000000000D+00,   3.922552855701593D-06,   0.000000000000000D+00, &
   6.121271204251234D-06,   0.000000000000000D+00,   1.707660223332044D-04, &
   6.294871131542212D-06,   0.000000000000000D+00,   4.065607926138019D-06, &
   0.000000000000000D+00,  -1.164264103784871D-06,   0.000000000000000D+00, &
  -1.194613355474645D-05,   0.000000000000000D+00,  -4.259180063636747D-05, &
   0.000000000000000D+00,  -1.019245144327372D-04,   0.000000000000000D+00, &
   3.060480717875587D-06,   0.000000000000000D+00,  -1.798895990439537D-06, &
   0.000000000000000D+00,  -6.277504188111470D-06,   0.000000000000000D+00, &
  -8.334489943865829D-06,   0.000000000000000D+00,   2.659663060761089D-07, &
   0.000000000000000D+00,  -5.190884977364252D-05,   0.000000000000000D+00, &
  -3.760303487933162D-04,  -4.115152487700175D-06,   0.000000000000000D+00, &
  -2.052951920121643D-06,   0.000000000000000D+00,   1.059102638665090D-06, &
   0.000000000000000D+00,   1.142559991709096D-05,   0.000000000000000D+00, &
   3.738766289003607D-05,   0.000000000000000D+00,   1.344520637429826D-04, &
   0.000000000000000D+00,   1.854649195015159D-04,   0.000000000000000D+00, &
  -6.116026459833094D-07,   0.000000000000000D+00,   9.530956334072193D-07, &
   0.000000000000000D+00,   3.732167139882350D-06,   0.000000000000000D+00, &
   4.981744513938477D-06,   0.000000000000000D+00,  -6.717416062683984D-06, &
   0.000000000000000D+00,  -3.691705615772343D-05,   0.000000000000000D+00, &
   1.208800273984850D-04,   0.000000000000000D+00,   4.895846108825701D-04, &
   1.586412753131874D-06,   0.000000000000000D+00,   1.671166029380796D-08, &
   0.000000000000000D+00,  -1.540425478084756D-06,   0.000000000000000D+00, &
  -7.466736615114993D-06,   0.000000000000000D+00,  -2.071254036274397D-05, &
   0.000000000000000D+00,  -7.362552642819425D-05,   0.000000000000000D+00, &
  -2.637497769643573D-04,   0.000000000000000D+00,  -2.308383838664743D-04, &
   0.000000000000000D+00,  -3.371862721218134D-07,   0.000000000000000D+00, &
   4.896099495158990D-08,   0.000000000000000D+00,  -1.004734586925729D-06, &
   0.000000000000000D+00,  -5.084166315791956D-07,   0.000000000000000D+00, &
   7.472944393058228D-06,   0.000000000000000D+00,   5.289710649932278D-05, &
   0.000000000000000D+00,   1.024775950369332D-04,   0.000000000000000D+00, &
  -1.588788015554859D-04,   0.000000000000000D+00,  -4.217581421963197D-04, &
  -1.968698297587715D-07,   0.000000000000000D+00,   2.300041465227281D-07, &
   0.000000000000000D+00,   1.083349611217446D-06,   0.000000000000000D+00, &
   2.754488253042815D-06,   0.000000000000000D+00,   5.855928477339118D-06, &
   0.000000000000000D+00,   9.365640963295851D-06,   0.000000000000000D+00, &
   1.024826521442064D-04,   0.000000000000000D+00,   3.335648461176220D-04, &
   0.000000000000000D+00,   2.076798322492251D-04,   0.000000000000000D+00, &
   1.285006372353759D-07,   0.000000000000000D+00,  -2.868067392171013D-07, &
   0.000000000000000D+00,  -2.384668934201227D-08,   0.000000000000000D+00, &
  -5.579004097414035D-07,   0.000000000000000D+00,  -3.331083602579456D-06, &
   0.000000000000000D+00,  -1.901377504716518D-05,   0.000000000000000D+00, &
  -1.127233584175407D-04,   0.000000000000000D+00,  -1.387940498962089D-04, &
   0.000000000000000D+00,   1.264683909397834D-04,   0.000000000000000D+00, &
   2.177304086653227D-04,  -3.263148181479900D-08,   0.000000000000000D+00, &
   5.434652550918348D-08,   0.000000000000000D+00,  -4.371339338543453D-07, &
   0.000000000000000D+00,  -4.034752844412494D-07,   0.000000000000000D+00, &
  -3.426759201290022D-07,   0.000000000000000D+00,   3.294610939561013D-06, &
   0.000000000000000D+00,   2.273089858996605D-05,   0.000000000000000D+00, &
  -1.150816397529898D-04,   0.000000000000000D+00,  -2.770799181187242D-04, &
   0.000000000000000D+00,  -1.250041624737237D-04,   0.000000000000000D+00, &
   4.417060884089854D-09,   0.000000000000000D+00,   6.147003872111689D-08, &
   0.000000000000000D+00,   1.077522310172104D-07,   0.000000000000000D+00, &
  -2.582483759443241D-09,   0.000000000000000D+00,   3.340838713714681D-07, &
   0.000000000000000D+00,   2.396072841497272D-06,   0.000000000000000D+00, &
   1.738149590031631D-05,   0.000000000000000D+00,   1.427281414439409D-04, &
   0.000000000000000D+00,   1.042463611771478D-04,   0.000000000000000D+00, &
  -5.950269077599141D-05,   0.000000000000000D+00,  -2.642881452373784D-05 /)
lem(:,238) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.275323097356671D-05,  -1.208595482059155D-04, &
  -9.756407928746942D-05,   2.447785188547742D-04,  -9.356619910423914D-06, &
  -9.245956597090673D-04,  -3.729831808206081D-04,   1.048830437749382D-03, &
   9.307887486999888D-04,  -8.245758358942446D-04,  -7.681789030270269D-04, &
   7.023814326642265D-04,   5.010687667334219D-05,  -2.330909218720559D-04, &
   9.690351927347912D-05,  -1.591593922322716D-05,   2.020789021627878D-05, &
  -5.369998975186937D-05,   8.954946098923487D-05,  -8.423998048586385D-05, &
   1.727869150494759D-05,   0.000000000000000D+00,   3.275323097356638D-05, &
   1.208595482059206D-04,   9.756407928747419D-05,  -2.447785188547797D-04, &
   9.356619910420249D-06,   9.245956597090777D-04,   3.729831808206110D-04, &
  -1.048830437749391D-03,  -9.307887486999897D-04,   8.245758358942496D-04, &
   7.681789030270280D-04,  -7.023814326642292D-04,  -5.010687667334289D-05, &
   2.330909218720556D-04,  -9.690351927348278D-05,   1.591593922322769D-05, &
  -2.020789021627347D-05,   5.369998975186991D-05,  -8.954946098923839D-05, &
   8.423998048586046D-05,  -1.727869150494328D-05,   0.000000000000000D+00, &
   1.637661548678674D-05,   0.000000000000000D+00,  -2.667972141373388D-05, &
   0.000000000000000D+00,   7.720895514046114D-06,   0.000000000000000D+00, &
   7.106306230282202D-04,   0.000000000000000D+00,  -2.568543072199442D-03, &
   0.000000000000000D+00,   2.413720959165769D-03,   0.000000000000000D+00, &
   8.914345467957424D-04,   0.000000000000000D+00,  -1.238397336028898D-03, &
   0.000000000000000D+00,  -2.104797194221069D-03,   0.000000000000000D+00, &
   2.801757046579681D-03,   0.000000000000000D+00,  -9.205804045644003D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.294685132575634D-06, &
   0.000000000000000D+00,  -1.100922731782231D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.296694191224728D-07,   0.000000000000000D+00, &
   2.017116274928641D-06,   0.000000000000000D+00,   1.297395096223219D-06, &
   0.000000000000000D+00,   6.165167297980949D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.770937829531303D-07,   0.000000000000000D+00, &
  -6.672962867159931D-07,   0.000000000000000D+00,   3.151744179576224D-06, &
   0.000000000000000D+00,  -1.409858587592948D-06,   0.000000000000000D+00, &
   4.919874479026158D-07,   0.000000000000000D+00,   3.617970842091880D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.201191386738294D-06, &
   0.000000000000000D+00,   8.805446165937523D-07,   0.000000000000000D+00, &
  -2.880499401495464D-06,   0.000000000000000D+00,  -4.345457195798992D-05, &
   0.000000000000000D+00,   1.455668292912018D-06,   0.000000000000000D+00, &
  -1.881038594202415D-06,   0.000000000000000D+00,  -6.025162857614545D-06, &
   0.000000000000000D+00,  -5.162006559558050D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.850137078208944D-06,   0.000000000000000D+00, &
  -1.954136373323255D-06,   0.000000000000000D+00,   2.552430117615665D-06, &
   0.000000000000000D+00,   2.243636424728915D-05,   0.000000000000000D+00, &
   1.538131326630015D-04,   0.000000000000000D+00,  -9.586086806553373D-07, &
   0.000000000000000D+00,   2.247300467522461D-06,   0.000000000000000D+00, &
   7.585758815987931D-06,   0.000000000000000D+00,   8.149628622007439D-06, &
   0.000000000000000D+00,  -6.601623108193964D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.325083513738286D-06,   0.000000000000000D+00, &
   2.678440485121714D-06,   0.000000000000000D+00,  -1.406061264618724D-06, &
   0.000000000000000D+00,  -1.654922238768624D-05,   0.000000000000000D+00, &
  -7.105916802771200D-05,   0.000000000000000D+00,  -2.962790880174827D-04, &
   0.000000000000000D+00,  -1.607020876565753D-07,   0.000000000000000D+00, &
  -1.967414457913507D-06,   0.000000000000000D+00,  -7.266819591200746D-06, &
   0.000000000000000D+00,  -8.919066991481026D-06,   0.000000000000000D+00, &
   6.930303954483818D-06,   0.000000000000000D+00,   4.094693794982148D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.263763838452371D-06, &
   0.000000000000000D+00,  -2.030969816697011D-06,   0.000000000000000D+00, &
   1.293744915365217D-06,   0.000000000000000D+00,   1.224698674979757D-05, &
   0.000000000000000D+00,   4.427793442232718D-05,   0.000000000000000D+00, &
   1.681816032247189D-04,   0.000000000000000D+00,   3.849451302432145D-04, &
   0.000000000000000D+00,   9.071398746118963D-07,   0.000000000000000D+00, &
   1.330093689132353D-06,   0.000000000000000D+00,   4.526814058202821D-06, &
   0.000000000000000D+00,   5.565143593915274D-06,   0.000000000000000D+00, &
  -8.097793172604240D-06,   0.000000000000000D+00,  -5.365326071353536D-05, &
   0.000000000000000D+00,   4.423867966421799D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.655087423164066D-07,   0.000000000000000D+00, &
   7.784722914534599D-07,   0.000000000000000D+00,  -1.572405233632844D-06, &
   0.000000000000000D+00,  -7.373758047645456D-06,   0.000000000000000D+00, &
  -2.326879554465273D-05,   0.000000000000000D+00,  -8.336502073824149D-05, &
   0.000000000000000D+00,  -3.013566491304118D-04,   0.000000000000000D+00, &
  -4.194049114957014D-04,   0.000000000000000D+00,  -6.099359611986326D-07, &
   0.000000000000000D+00,  -4.756237528439882D-07,   0.000000000000000D+00, &
  -1.239291384045499D-06,   0.000000000000000D+00,  -1.252415927060622D-06, &
   0.000000000000000D+00,   7.989761081391988D-06,   0.000000000000000D+00, &
   5.784862683384989D-05,   0.000000000000000D+00,   1.235021584733392D-04, &
   0.000000000000000D+00,  -7.961595200774834D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.297002515845458D-08,   0.000000000000000D+00, &
  -2.133741145859296D-07,   0.000000000000000D+00,   8.999108420420001D-07, &
   0.000000000000000D+00,   2.624996591655047D-06,   0.000000000000000D+00, &
   6.764709594030992D-06,   0.000000000000000D+00,   1.406214145060302D-05, &
   0.000000000000000D+00,   1.202981155273742D-04,   0.000000000000000D+00, &
   3.770417013754257D-04,   0.000000000000000D+00,   4.326565251996633D-04, &
   0.000000000000000D+00,   9.926892720789963D-08,   0.000000000000000D+00, &
   6.557369336268932D-08,   0.000000000000000D+00,  -2.140353645654603D-07, &
   0.000000000000000D+00,  -2.138581353429215D-07,   0.000000000000000D+00, &
  -3.971348161460608D-06,   0.000000000000000D+00,  -2.270020348544091D-05, &
   0.000000000000000D+00,  -1.260917854185485D-04,   0.000000000000000D+00, &
  -1.622283091163491D-04,   0.000000000000000D+00,   3.989149360546537D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.351429496736497D-08, &
   0.000000000000000D+00,   1.157390014478303D-07,   0.000000000000000D+00, &
  -1.400660527969095D-07,   0.000000000000000D+00,  -3.915241990034036D-07, &
   0.000000000000000D+00,   1.405007503285515D-08,   0.000000000000000D+00, &
   3.210923617676067D-06,   0.000000000000000D+00,   1.870595389281316D-05, &
   0.000000000000000D+00,  -1.392844665699516D-04,   0.000000000000000D+00, &
  -3.220477029024913D-04,   0.000000000000000D+00,  -3.724539703565499D-04, &
   0.000000000000000D+00,   1.872345489803555D-08,   0.000000000000000D+00, &
  -4.927054340717054D-08,   0.000000000000000D+00,   2.292297124220746D-07, &
   0.000000000000000D+00,   2.593694541208614D-08,   0.000000000000000D+00, &
   5.334928577315244D-07,   0.000000000000000D+00,   3.378893806387829D-06, &
   0.000000000000000D+00,   2.452119699882019D-05,   0.000000000000000D+00, &
   1.629948209896985D-04,   0.000000000000000D+00,   1.297789296424235D-04, &
   0.000000000000000D+00,   1.314908975431461D-05,   0.000000000000000D+00 /)
lem(:,239) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.300631434689353D-05,  -1.877461989735025D-04,  -3.538430340100840D-04, &
  -3.568249260401622D-04,   5.235171353587857D-05,   6.441396741261801D-04, &
   4.067700332966224D-04,   1.371276641389936D-05,  -4.275417781630836D-05, &
  -9.177699643629581D-04,  -4.513285105569371D-06,   9.617372145288152D-04, &
  -4.149004404785281D-04,  -1.990487068108462D-04,   2.182707165500841D-04, &
  -8.622144939806428D-05,   2.145941395845339D-05,  -1.013401826919712D-05, &
   2.172364002005732D-05,  -3.835654431233665D-05,   4.877129030844694D-05, &
  -3.444461665484385D-05,  -5.300631434689217D-05,  -1.877461989735162D-04, &
  -3.538430340100843D-04,  -3.568249260401545D-04,   5.235171353587234D-05, &
   6.441396741261781D-04,   4.067700332966349D-04,   1.371276641389873D-05, &
  -4.275417781631871D-05,  -9.177699643629543D-04,  -4.513285105569931D-06, &
   9.617372145288141D-04,  -4.149004404785261D-04,  -1.990487068108451D-04, &
   2.182707165500875D-04,  -8.622144939806916D-05,   2.145941395845214D-05, &
  -1.013401826919156D-05,   2.172364002005389D-05,  -3.835654431233893D-05, &
   4.877129030844796D-05,  -3.444461665484023D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.922328646765324D-05,   0.000000000000000D+00, &
  -5.807166502853050D-05,   0.000000000000000D+00,   1.770113092719973D-04, &
   0.000000000000000D+00,  -5.981514664817229D-04,   0.000000000000000D+00, &
  -4.595501807711615D-04,   0.000000000000000D+00,   8.261669692848944D-03, &
   0.000000000000000D+00,  -1.981432388877080D-02,   0.000000000000000D+00, &
   2.066520881746686D-02,   0.000000000000000D+00,  -9.518950425607105D-03, &
   0.000000000000000D+00,   8.045657846967595D-04,   0.000000000000000D+00, &
   2.774325284091350D-07,   7.911964699074549D-07,   0.000000000000000D+00, &
   1.556704742739812D-06,   0.000000000000000D+00,  -1.575970890546045D-06, &
   1.695929125128415D-06,   0.000000000000000D+00,  -1.517734900913655D-06, &
   0.000000000000000D+00,  -2.580675798448778D-07,   0.000000000000000D+00, &
  -1.049729824754199D-06,   0.000000000000000D+00,   5.418776971816487D-06, &
  -3.934020352477601D-06,   0.000000000000000D+00,   2.159562771832318D-07, &
   0.000000000000000D+00,   3.755325067275796D-06,   0.000000000000000D+00, &
  -2.925405757436886D-06,   0.000000000000000D+00,   1.652414448434929D-06, &
   0.000000000000000D+00,   3.227961907158869D-08,   0.000000000000000D+00, &
  -1.637613565599714D-05,   3.103811697561687D-06,   0.000000000000000D+00, &
   1.107739984082847D-06,   0.000000000000000D+00,  -3.011553969369944D-06, &
   0.000000000000000D+00,  -6.112945098769632D-06,   0.000000000000000D+00, &
   4.215348453629796D-06,   0.000000000000000D+00,  -1.429929255857445D-07, &
   0.000000000000000D+00,  -1.720601085835444D-06,   0.000000000000000D+00, &
   4.553235401547781D-06,   0.000000000000000D+00,   3.493732505745005D-05, &
  -6.930732028390599D-07,   0.000000000000000D+00,  -1.088967840066522D-06, &
   0.000000000000000D+00,   1.717157173699538D-07,   0.000000000000000D+00, &
   4.053215351204212D-06,   0.000000000000000D+00,   3.446832756296213D-06, &
   0.000000000000000D+00,  -3.561832802367134D-06,   0.000000000000000D+00, &
  -1.448852723118351D-06,   0.000000000000000D+00,   1.421146787839603D-06, &
   0.000000000000000D+00,   1.685543023738087D-06,   0.000000000000000D+00, &
  -5.755918757993510D-06,   0.000000000000000D+00,   1.252019608432203D-06, &
  -3.294853144634423D-07,   0.000000000000000D+00,   1.009640133325007D-06, &
   0.000000000000000D+00,   1.979962626142478D-06,   0.000000000000000D+00, &
  -1.377244507827445D-07,   0.000000000000000D+00,  -5.507022150707788D-06, &
   0.000000000000000D+00,  -2.368148368226603D-05,   0.000000000000000D+00, &
   2.317537947928233D-06,   0.000000000000000D+00,   1.205038345501960D-06, &
   0.000000000000000D+00,  -1.616441002770354D-06,   0.000000000000000D+00, &
  -5.359679215840410D-06,   0.000000000000000D+00,  -7.360654648076430D-06, &
   0.000000000000000D+00,  -7.484177833051193D-06,   0.000000000000000D+00, &
  -2.383978466058864D-04,  -1.494080458749326D-07,   0.000000000000000D+00, &
  -9.107662323656564D-07,   0.000000000000000D+00,  -1.591718413441436D-06, &
   0.000000000000000D+00,  -2.912059330990458D-07,   0.000000000000000D+00, &
   8.259991923563529D-06,   0.000000000000000D+00,   3.461794709674705D-05, &
   0.000000000000000D+00,   1.629126309325736D-04,   0.000000000000000D+00, &
  -9.177974980769405D-07,   0.000000000000000D+00,  -5.534359681200354D-08, &
   0.000000000000000D+00,   1.711530758311665D-06,   0.000000000000000D+00, &
   4.940306325503907D-06,   0.000000000000000D+00,   7.497784505885667D-06, &
   0.000000000000000D+00,   1.413504089516071D-06,   0.000000000000000D+00, &
   6.458958833614568D-06,   0.000000000000000D+00,   6.683844251438009D-04, &
   3.594241926410059D-07,   0.000000000000000D+00,   2.561440870543772D-07, &
   0.000000000000000D+00,   1.335458403997578D-07,   0.000000000000000D+00, &
  -9.975121808842161D-07,   0.000000000000000D+00,  -7.664976769963676D-06, &
   0.000000000000000D+00,  -2.729337951556756D-05,   0.000000000000000D+00, &
  -1.131753706718524D-04,   0.000000000000000D+00,  -4.327085031647380D-04, &
   0.000000000000000D+00,   8.263429129356054D-08,   0.000000000000000D+00, &
  -2.314121278412485D-07,   0.000000000000000D+00,  -8.007234075830249D-07, &
   0.000000000000000D+00,  -2.117170538322880D-06,   0.000000000000000D+00, &
  -1.899058947242322D-06,   0.000000000000000D+00,   6.819106914820439D-06, &
   0.000000000000000D+00,   5.648448546612810D-05,   0.000000000000000D+00, &
   4.196911902257470D-05,   0.000000000000000D+00,  -9.335253644238789D-04, &
  -1.033177663213987D-07,   0.000000000000000D+00,   1.337295742935327D-07, &
   0.000000000000000D+00,   2.553683287822053D-07,   0.000000000000000D+00, &
   8.653233203480088D-07,   0.000000000000000D+00,   3.177688955982108D-06, &
   0.000000000000000D+00,   9.792871165858155D-06,   0.000000000000000D+00, &
   2.706574058124700D-05,   0.000000000000000D+00,   2.076798322492251D-04, &
   0.000000000000000D+00,   6.013351909759593D-04,   0.000000000000000D+00, &
   3.644216339922261D-08,   0.000000000000000D+00,   1.360271462285522D-10, &
   0.000000000000000D+00,   1.143841145055106D-07,   0.000000000000000D+00, &
   3.738220939464025D-07,   0.000000000000000D+00,  -4.615257703730436D-07, &
   0.000000000000000D+00,  -5.177003652814152D-06,   0.000000000000000D+00, &
  -3.020858659984367D-05,   0.000000000000000D+00,  -1.571723857683240D-04, &
   0.000000000000000D+00,  -7.814934046932618D-05,   0.000000000000000D+00, &
   7.370153174111758D-04,  -6.960263973810314D-09,   0.000000000000000D+00, &
  -4.875504517113272D-08,   0.000000000000000D+00,  -3.298942347980315D-08, &
   0.000000000000000D+00,  -2.011849093066686D-07,   0.000000000000000D+00, &
  -7.754655789933091D-08,   0.000000000000000D+00,  -4.097178384050260D-07, &
   0.000000000000000D+00,   3.016667910501330D-06,   0.000000000000000D+00, &
   9.535964993652134D-06,   0.000000000000000D+00,  -2.582198288497048D-04, &
   0.000000000000000D+00,  -4.698667578793968D-04,   0.000000000000000D+00, &
   7.761584596531596D-10,   0.000000000000000D+00,   3.097891383540878D-08, &
   0.000000000000000D+00,  -3.859410221116349D-08,   0.000000000000000D+00, &
  -9.848378743892824D-09,   0.000000000000000D+00,   7.285670305160601D-08, &
   0.000000000000000D+00,   1.071469691550162D-06,   0.000000000000000D+00, &
   5.022774632823846D-06,   0.000000000000000D+00,   3.918345427522751D-05, &
   0.000000000000000D+00,   2.162162735610206D-04,   0.000000000000000D+00, &
   3.817458379518016D-05,   0.000000000000000D+00,  -2.963054203481746D-04 /)
lem(:,240) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.573468147922215D-05,   1.565736682790405D-04, &
   4.873463082504794D-04,   9.543439930052776D-04,   1.016004861126284D-03, &
  -2.418909806570762D-04,  -2.230697891530804D-03,  -8.462538956231172D-04, &
   2.421016994519559D-03,   3.196018834102631D-04,  -1.603843584499964D-03, &
   5.139279585511036D-04,   3.307330273829422D-04,  -2.615856925499649D-04, &
   8.595724253192032D-05,  -3.890161407911733D-05,  -2.565174908461146D-06, &
  -4.882660616947671D-06,   3.011149921514943D-05,  -2.715090529832726D-05, &
   2.898981096417160D-05,   0.000000000000000D+00,  -2.573468147924220D-05, &
  -1.565736682790374D-04,  -4.873463082504681D-04,  -9.543439930052871D-04, &
  -1.016004861126286D-03,   2.418909806570909D-04,   2.230697891530802D-03, &
   8.462538956230953D-04,  -2.421016994519554D-03,  -3.196018834102524D-04, &
   1.603843584499971D-03,  -5.139279585511054D-04,  -3.307330273829513D-04, &
   2.615856925499689D-04,  -8.595724253192151D-05,   3.890161407911257D-05, &
   2.565174908470707D-06,   4.882660616947334D-06,  -3.011149921516229D-05, &
   2.715090529832597D-05,  -2.898981096415690D-05,   0.000000000000000D+00, &
  -1.286734073961736D-05,   0.000000000000000D+00,   2.736633189129712D-05, &
   0.000000000000000D+00,  -7.104907960870594D-05,   0.000000000000000D+00, &
   2.228965649144419D-04,   0.000000000000000D+00,  -9.057859171700323D-04, &
   0.000000000000000D+00,   5.665993478030948D-03,   0.000000000000000D+00, &
  -2.079067972989783D-02,   0.000000000000000D+00,   3.943289824754460D-02, &
   0.000000000000000D+00,  -4.064464963563603D-02,   0.000000000000000D+00, &
   2.277171294829318D-02,   0.000000000000000D+00,  -6.543078801379863D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.017252604166524D-06, &
   0.000000000000000D+00,   9.710138494319293D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.398200757574823D-07,   0.000000000000000D+00, &
  -2.700125914847626D-06,   0.000000000000000D+00,   9.366311856555393D-08, &
   0.000000000000000D+00,  -2.445022357742118D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.027923086028591D-06,   0.000000000000000D+00, &
  -1.450179378558906D-06,   0.000000000000000D+00,   7.840722295373466D-06, &
   0.000000000000000D+00,  -1.862083771263552D-06,   0.000000000000000D+00, &
   8.377437249680484D-07,   0.000000000000000D+00,   6.515133697140412D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.466358836474937D-07, &
   0.000000000000000D+00,   3.498411843225668D-06,   0.000000000000000D+00, &
   1.974404707253754D-06,   0.000000000000000D+00,  -2.602760132018892D-05, &
   0.000000000000000D+00,   3.426749032217800D-06,   0.000000000000000D+00, &
   2.319234878439028D-06,   0.000000000000000D+00,  -6.235950983875625D-06, &
   0.000000000000000D+00,  -1.822814598426460D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.634108079420494D-06,   0.000000000000000D+00, &
  -3.799439605714163D-06,   0.000000000000000D+00,  -7.981520432692344D-06, &
   0.000000000000000D+00,   5.048919510055489D-06,   0.000000000000000D+00, &
   1.034756719847738D-04,   0.000000000000000D+00,  -4.291312122122657D-06, &
   0.000000000000000D+00,  -3.447031416391785D-06,   0.000000000000000D+00, &
   4.711169471758985D-06,   0.000000000000000D+00,   2.342674229118307D-05, &
   0.000000000000000D+00,   3.743146980678167D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.539481026785161D-07,   0.000000000000000D+00, &
   3.268366020495210D-06,   0.000000000000000D+00,   8.643971212366635D-06, &
   0.000000000000000D+00,   1.324795151347240D-06,   0.000000000000000D+00, &
  -5.218610460266790D-05,   0.000000000000000D+00,  -3.980035937967754D-04, &
   0.000000000000000D+00,   3.306070963541612D-06,   0.000000000000000D+00, &
   1.590176217809779D-06,   0.000000000000000D+00,  -4.658879021500317D-06, &
   0.000000000000000D+00,  -2.069641460048096D-05,   0.000000000000000D+00, &
  -3.741243449238278D-05,   0.000000000000000D+00,   2.262369880455808D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.176315963165973D-07, &
   0.000000000000000D+00,  -1.631628447667663D-06,   0.000000000000000D+00, &
  -4.596632214483711D-06,   0.000000000000000D+00,   1.047981303188809D-06, &
   0.000000000000000D+00,   3.770498347521970D-05,   0.000000000000000D+00, &
   1.760574211548985D-04,   0.000000000000000D+00,   1.123477656280079D-03, &
   0.000000000000000D+00,  -1.136103274872931D-06,   0.000000000000000D+00, &
   2.022294077524740D-07,   0.000000000000000D+00,   4.070249897720250D-06, &
   0.000000000000000D+00,   1.195110697020075D-05,   0.000000000000000D+00, &
   1.844279183611540D-05,   0.000000000000000D+00,  -2.072559251874392D-05, &
   0.000000000000000D+00,  -2.526103244902797D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.723275860901201D-07,   0.000000000000000D+00, &
  -1.331770801447411D-07,   0.000000000000000D+00,   5.958918001625696D-07, &
   0.000000000000000D+00,  -3.369272757421360D-06,   0.000000000000000D+00, &
  -2.254324595184748D-05,   0.000000000000000D+00,  -8.255593604734964D-05, &
   0.000000000000000D+00,  -3.340346888334408D-04,   0.000000000000000D+00, &
  -2.136643166602093D-03,   0.000000000000000D+00,  -3.954008192928283D-08, &
   0.000000000000000D+00,  -1.466585218044261D-07,   0.000000000000000D+00, &
  -2.221899834808546D-06,   0.000000000000000D+00,  -3.627751345133022D-06, &
   0.000000000000000D+00,  -2.208453303646228D-06,   0.000000000000000D+00, &
   2.663879005768372D-05,   0.000000000000000D+00,   2.058760307370229D-04, &
   0.000000000000000D+00,   7.023834772159586D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.472307075674457D-07,   0.000000000000000D+00, &
   4.650284686545723D-07,   0.000000000000000D+00,   5.948047675913967D-07, &
   0.000000000000000D+00,   1.710170853401269D-06,   0.000000000000000D+00, &
   7.580078466142035D-06,   0.000000000000000D+00,   2.212671703853364D-05, &
   0.000000000000000D+00,   5.420352392877237D-05,   0.000000000000000D+00, &
   4.326565251996633D-04,   0.000000000000000D+00,   2.728280783134165D-03, &
   0.000000000000000D+00,   7.922678840814772D-08,   0.000000000000000D+00, &
  -2.212922789576891D-07,   0.000000000000000D+00,   8.847004691568142D-07, &
   0.000000000000000D+00,   4.553133933045281D-07,   0.000000000000000D+00, &
  -1.564593815851678D-06,   0.000000000000000D+00,  -1.313908543988176D-05, &
   0.000000000000000D+00,  -7.757973945462446D-05,   0.000000000000000D+00, &
  -4.450887696793047D-04,   0.000000000000000D+00,  -9.781347043911115D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.134323273574300D-08, &
   0.000000000000000D+00,  -7.177713103519579D-08,   0.000000000000000D+00, &
  -4.732095263888418D-07,   0.000000000000000D+00,   2.520672854098924D-07, &
   0.000000000000000D+00,  -4.480065684461049D-07,   0.000000000000000D+00, &
  -6.623250917563934D-07,   0.000000000000000D+00,   1.031734153380720D-05, &
   0.000000000000000D+00,   5.517441504134404D-05,   0.000000000000000D+00, &
  -4.699555042734211D-04,   0.000000000000000D+00,  -2.356359515104339D-03, &
   0.000000000000000D+00,   3.463208940456888D-09,   0.000000000000000D+00, &
   1.019175996796216D-07,   0.000000000000000D+00,  -3.147947993434675D-07, &
   0.000000000000000D+00,  -4.008818027123064D-07,   0.000000000000000D+00, &
   4.079527794332191D-07,   0.000000000000000D+00,   2.042096651823464D-06, &
   0.000000000000000D+00,   1.057716294964572D-05,   0.000000000000000D+00, &
   8.629045621995765D-05,   0.000000000000000D+00,   5.746151382288841D-04, &
   0.000000000000000D+00,   7.719733130104782D-04,   0.000000000000000D+00 /)
lem(:,241) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.391394905551710D-03, &
  -3.508965075278462D-03,  -9.590683274529554D-03,   7.285142882974730D-03, &
   1.054901935414486D-02,  -3.754824554489107D-03,  -4.371711136746921D-03, &
  -4.737708224391704D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.391394905551785D-03,  -3.508965075278416D-03,  -9.590683274529653D-03, &
   7.285142882974773D-03,   1.054901935414488D-02,  -3.754824554489242D-03, &
  -4.371711136746870D-03,  -4.737708224390195D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.203707094697056D-04,   0.000000000000000D+00, &
   6.482701651245800D-06,   0.000000000000000D+00,   3.619929650116304D-07, &
   0.000000000000000D+00,   3.066293350707587D-08,   0.000000000000000D+00, &
   3.166135507219896D-09,   0.000000000000000D+00,   3.607272553519103D-10, &
   0.000000000000000D+00,   4.270374144553907D-11,   0.000000000000000D+00, &
  -3.237714694889330D-03,   0.000000000000000D+00,  -6.759066825807615D-05, &
   0.000000000000000D+00,  -5.322704336100407D-06,   0.000000000000000D+00, &
  -5.857099408164771D-07,   0.000000000000000D+00,  -7.508334003490698D-08, &
   0.000000000000000D+00,  -1.037216111654230D-08,   0.000000000000000D+00, &
  -1.467941105894157D-09,   0.000000000000000D+00,  -2.430681700103368D-10, &
   7.249416148192576D-03,   0.000000000000000D+00,   4.172507620783626D-05, &
   0.000000000000000D+00,   1.916383584420525D-05,   0.000000000000000D+00, &
   3.550862748280768D-06,   0.000000000000000D+00,   6.347412532100801D-07, &
   0.000000000000000D+00,   1.132243169586306D-07,   0.000000000000000D+00, &
   1.995784910159913D-08,   0.000000000000000D+00,   3.459047035972346D-09, &
   0.000000000000000D+00,  -4.575938446906185D-03,   0.000000000000000D+00, &
   8.920215260593442D-04,   0.000000000000000D+00,   2.995190546576231D-05, &
   0.000000000000000D+00,  -4.104547732033803D-06,   0.000000000000000D+00, &
  -1.962774023063498D-06,   0.000000000000000D+00,  -5.495574918594667D-07, &
   0.000000000000000D+00,  -1.321722646669599D-07,   0.000000000000000D+00, &
  -2.916121465791158D-08,   0.000000000000000D+00,  -7.496182954292764D-09, &
  -8.370511478300615D-03,   0.000000000000000D+00,  -2.718173139237192D-03, &
   0.000000000000000D+00,  -2.076701249341536D-04,   0.000000000000000D+00, &
  -2.735732340158949D-05,   0.000000000000000D+00,  -1.367525035888935D-06, &
   0.000000000000000D+00,   7.553330780832099D-07,   0.000000000000000D+00, &
   3.886876641381362D-07,   0.000000000000000D+00,   1.285006372353759D-07, &
   0.000000000000000D+00,   3.644216339922261D-08,   0.000000000000000D+00, &
   1.714882202687083D-02,   0.000000000000000D+00,   2.125711365720589D-03, &
   0.000000000000000D+00,   3.156766036931115D-05,   0.000000000000000D+00, &
   5.837909698479359D-05,   0.000000000000000D+00,   1.756419111790384D-05, &
   0.000000000000000D+00,   2.974888630436758D-06,   0.000000000000000D+00, &
   3.913466009280519D-08,   0.000000000000000D+00,  -2.082673623735184D-07, &
   0.000000000000000D+00,  -1.074381185280568D-07,   0.000000000000000D+00, &
  -5.287177092997294D-08,  -8.817420615492080D-03,   0.000000000000000D+00, &
   2.854809977514229D-03,   0.000000000000000D+00,   9.495412354642717D-04, &
   0.000000000000000D+00,   8.397396336753180D-05,   0.000000000000000D+00, &
  -1.038795510715184D-05,   0.000000000000000D+00,  -9.315977617543752D-06, &
   0.000000000000000D+00,  -2.887889392038571D-06,   0.000000000000000D+00, &
  -4.649771086187222D-07,   0.000000000000000D+00,   5.127555045916170D-08, &
   0.000000000000000D+00,   8.133203935234029D-08,   0.000000000000000D+00, &
  -5.094851322871740D-03,   0.000000000000000D+00,  -6.394240902090415D-03, &
   0.000000000000000D+00,  -1.203259435108714D-03,   0.000000000000000D+00, &
  -2.438633578399640D-04,   0.000000000000000D+00,  -6.164993823762254D-05, &
   0.000000000000000D+00,  -5.109096851299239D-06,   0.000000000000000D+00, &
   3.725870722959832D-06,   0.000000000000000D+00,   2.139231960409271D-06, &
   0.000000000000000D+00,   6.067332943594085D-07,   0.000000000000000D+00, &
   5.809421318147445D-08,   0.000000000000000D+00,  -8.308804177147084D-08 /)
lem(:,242) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -5.283151506597398D-04,   2.206816285498488D-03,  -1.557556940288941D-03, &
  -4.419789446182310D-03,   5.677604561119694D-03,   2.906556492344311D-03, &
  -5.335297442613297D-03,  -8.037309384465645D-04,   1.525317901391646D-03, &
   2.189948874320647D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.283151506597303D-04,  -2.206816285498501D-03, &
   1.557556940288962D-03,   4.419789446182302D-03,  -5.677604561119697D-03, &
  -2.906556492344292D-03,   5.335297442613277D-03,   8.037309384465608D-04, &
  -1.525317901391620D-03,  -2.189948874320758D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.820067984358133D-09,   0.000000000000000D+00,  -1.007827159868525D-07, &
   0.000000000000000D+00,  -6.428551712400412D-07,   0.000000000000000D+00, &
  -1.111372324901234D-06,   0.000000000000000D+00,   9.899351641856822D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.877806543294190D-06,   0.000000000000000D+00, &
   1.800838951992935D-07,   0.000000000000000D+00,   1.388409052145502D-08, &
   0.000000000000000D+00,   1.285563937211140D-09,   0.000000000000000D+00, &
   1.170781443084380D-10,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.129276422162263D-05,   0.000000000000000D+00, &
  -2.721267749679925D-06,   0.000000000000000D+00,  -2.759922121125328D-07, &
   0.000000000000000D+00,  -3.265332400426220D-08,   0.000000000000000D+00, &
  -3.967648222743598D-09,   0.000000000000000D+00,  -3.867404844288634D-10, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.873506705606788D-04, &
   0.000000000000000D+00,   1.263550809175754D-05,   0.000000000000000D+00, &
   1.880969648359458D-06,   0.000000000000000D+00,   3.031174457344950D-07, &
   0.000000000000000D+00,   4.889206882004981D-08,   0.000000000000000D+00, &
   7.224683193173467D-09,   0.000000000000000D+00,   4.876661811401078D-10, &
   0.000000000000000D+00,  -4.563496413700349D-04,   0.000000000000000D+00, &
  -1.169193607997233D-05,   0.000000000000000D+00,  -4.365397485900213D-06, &
   0.000000000000000D+00,  -1.191236221004669D-06,   0.000000000000000D+00, &
  -2.789856516696358D-07,   0.000000000000000D+00,  -5.817889139345375D-08, &
   0.000000000000000D+00,  -9.907333872694177D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.745811604172751D-04,   0.000000000000000D+00, &
  -9.247433517326249D-05,   0.000000000000000D+00,  -5.290225765120466D-06, &
   0.000000000000000D+00,   9.710377112624663D-07,   0.000000000000000D+00, &
   6.492027390535483D-07,   0.000000000000000D+00,   2.238397734849389D-07, &
   0.000000000000000D+00,   5.978079926236121D-08,   0.000000000000000D+00, &
   1.318128501525148D-08,   0.000000000000000D+00,  -1.062472697622862D-04, &
   0.000000000000000D+00,   3.757151533908318D-04,   0.000000000000000D+00, &
   4.211253315266683D-05,   0.000000000000000D+00,   6.625340525135288D-06, &
   0.000000000000000D+00,   4.583643317914154D-07,   0.000000000000000D+00, &
  -2.571057723196095D-07,   0.000000000000000D+00,  -1.580398242889859D-07, &
   0.000000000000000D+00,  -5.730916323837423D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.436354161957146D-04,   0.000000000000000D+00, &
  -5.908879220807054D-04,   0.000000000000000D+00,  -4.695783300010958D-05, &
   0.000000000000000D+00,  -1.722517708669060D-05,   0.000000000000000D+00, &
  -5.015613324136078D-06,   0.000000000000000D+00,  -9.416816164687176D-07, &
   0.000000000000000D+00,   4.845679010078038D-09,   0.000000000000000D+00, &
   9.926892720789963D-08,   0.000000000000000D+00,   7.922678840814772D-08, &
   0.000000000000000D+00,   1.387022851367670D-03,   0.000000000000000D+00, &
   2.160539495867563D-04,   0.000000000000000D+00,  -1.094921855587940D-04, &
   0.000000000000000D+00,  -4.284196134875921D-06,   0.000000000000000D+00, &
   5.558105019486322D-06,   0.000000000000000D+00,   3.113379817031362D-06, &
   0.000000000000000D+00,   9.590624626985826D-07,   0.000000000000000D+00, &
   1.307353362503985D-07,   0.000000000000000D+00,  -6.261436081074758D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.900132121622945D-04, &
   0.000000000000000D+00,   7.278939819388261D-04,   0.000000000000000D+00, &
   3.437555305237547D-04,   0.000000000000000D+00,   6.229508705012527D-05, &
   0.000000000000000D+00,   1.155844683315836D-05,   0.000000000000000D+00, &
  -3.818963984882561D-07,   0.000000000000000D+00,  -1.589072100162013D-06, &
   0.000000000000000D+00,  -7.749978215128054D-07,   0.000000000000000D+00, &
  -1.832331774904505D-07,   0.000000000000000D+00,   8.329818030391308D-08, &
   0.000000000000000D+00,  -1.189545185300219D-04,   0.000000000000000D+00, &
  -1.292331749713833D-03,   0.000000000000000D+00,  -2.424139141286396D-04, &
   0.000000000000000D+00,  -4.658152632275726D-05,   0.000000000000000D+00, &
  -2.586653146365229D-05,   0.000000000000000D+00,  -9.405540262541705D-06, &
   0.000000000000000D+00,  -1.557143313746564D-06,   0.000000000000000D+00, &
   5.625571187804162D-07,   0.000000000000000D+00,   5.429327534901412D-07, &
   0.000000000000000D+00,   1.791287123716920D-07,   0.000000000000000D+00 /)
lem(:,243) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.395005741317509D-04,  -9.270672228641981D-04, &
   2.268017136174672D-04,   1.034136323148311D-03,   2.029044688807138D-03, &
   9.041494799728739D-04,  -6.389790605427370D-03,  -2.555588182601501D-03, &
   5.281371280814997D-03,   2.096571600306272D-03,  -5.320636500511058D-04, &
  -2.060679857998451D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.395005741317587D-04, &
  -9.270672228641742D-04,   2.268017136174644D-04,   1.034136323148282D-03, &
   2.029044688807142D-03,   9.041494799728905D-04,  -6.389790605427384D-03, &
  -2.555588182601495D-03,   5.281371280815002D-03,   2.096571600306241D-03, &
  -5.320636500510985D-04,  -2.060679857997844D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.098146592279610D-09,   0.000000000000000D+00,   8.009936535904505D-08, &
   0.000000000000000D+00,   8.866617718059080D-07,   0.000000000000000D+00, &
   3.035852021564385D-06,   0.000000000000000D+00,   2.134567330030225D-06, &
   0.000000000000000D+00,  -5.841430228493608D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.602610072870281D-06,   0.000000000000000D+00,   1.433609048058881D-07, &
   0.000000000000000D+00,   9.432051988265465D-09,   0.000000000000000D+00, &
   7.077388190019304D-10,   0.000000000000000D+00,   1.890535205595603D-11, &
   0.000000000000000D+00,  -5.887993023457575D-05,   0.000000000000000D+00, &
  -2.212979484206307D-06,   0.000000000000000D+00,  -1.971387090522987D-07, &
   0.000000000000000D+00,  -2.049958218260547D-08,   0.000000000000000D+00, &
  -1.827320431096988D-09,   0.000000000000000D+00,   3.021410032510299D-10, &
   3.711098879643724D-05,   0.000000000000000D+00,   9.628404676611331D-06, &
   0.000000000000000D+00,   1.340289769804167D-06,   0.000000000000000D+00, &
   2.001869326545069D-07,   0.000000000000000D+00,   2.792649648646232D-08, &
   0.000000000000000D+00,   2.156991040424754D-09,   0.000000000000000D+00, &
   7.839195888719511D-04,   0.000000000000000D+00,  -2.555390472879539D-06, &
   0.000000000000000D+00,  -2.654723965015250D-06,   0.000000000000000D+00, &
  -7.675202083265739D-07,   0.000000000000000D+00,  -1.703463789555830D-07, &
   0.000000000000000D+00,  -2.916556225999307D-08,   0.000000000000000D+00, &
   1.587750973340133D-09,  -2.402930874275847D-03,   0.000000000000000D+00, &
   7.661669075443939D-06,   0.000000000000000D+00,  -4.460021540421970D-06, &
   0.000000000000000D+00,   4.174027578159681D-07,   0.000000000000000D+00, &
   3.831459473688856D-07,   0.000000000000000D+00,   1.268977283090137D-07, &
   0.000000000000000D+00,   2.457472877220651D-08,   0.000000000000000D+00, &
   1.886411063274246D-03,   0.000000000000000D+00,  -3.328841862478262D-04, &
   0.000000000000000D+00,   1.156729562664227D-05,   0.000000000000000D+00, &
   3.689366112200712D-06,   0.000000000000000D+00,   3.234391412712453D-07, &
   0.000000000000000D+00,  -1.475662332877719D-07,   0.000000000000000D+00, &
  -8.718216120794930D-08,   0.000000000000000D+00,  -2.528773917561538D-08, &
   2.548965785895708D-03,   0.000000000000000D+00,   1.030690127367594D-03, &
   0.000000000000000D+00,  -1.762179849762342D-06,   0.000000000000000D+00, &
  -4.108598943091615D-06,   0.000000000000000D+00,  -2.268060205638390D-06, &
   0.000000000000000D+00,  -4.760697831453564D-07,   0.000000000000000D+00, &
   2.673015159418132D-08,   0.000000000000000D+00,   6.532744795195437D-08, &
   0.000000000000000D+00,  -5.724186116288449D-03,   0.000000000000000D+00, &
  -8.241235628187304D-04,   0.000000000000000D+00,   1.352491284223209D-04, &
   0.000000000000000D+00,  -1.324817327179123D-06,   0.000000000000000D+00, &
   8.114712180501485D-07,   0.000000000000000D+00,   1.183930908868187D-06, &
   0.000000000000000D+00,   4.068416703052880D-07,   0.000000000000000D+00, &
   1.453986082379752D-08,   0.000000000000000D+00,  -1.264161685616119D-07, &
   2.701710584792058D-03,   0.000000000000000D+00,  -1.137481650765530D-03, &
   0.000000000000000D+00,  -5.242191407015943D-04,   0.000000000000000D+00, &
  -2.139868401775447D-05,   0.000000000000000D+00,   1.307151476556868D-06, &
   0.000000000000000D+00,   2.763479858960385D-07,   0.000000000000000D+00, &
  -5.257257457334870D-07,   0.000000000000000D+00,  -2.868067392171013D-07, &
   0.000000000000000D+00,   1.360271462285522D-10,   0.000000000000000D+00, &
   2.125711365720589D-03,   0.000000000000000D+00,   2.557516746656444D-03, &
   0.000000000000000D+00,   4.830398905853312D-04,   0.000000000000000D+00, &
   2.121917905712801D-06,   0.000000000000000D+00,   5.448177955209335D-06, &
   0.000000000000000D+00,  -6.239724600842188D-07,   0.000000000000000D+00, &
  -4.510611699468166D-07,   0.000000000000000D+00,   1.986142873503335D-07, &
   0.000000000000000D+00,   1.971269274408078D-07,   0.000000000000000D+00, &
  -2.602173426168478D-08,  -2.876000712547205D-03,   0.000000000000000D+00, &
  -1.205040236634688D-03,   0.000000000000000D+00,   5.731979646554538D-04, &
   0.000000000000000D+00,   2.051322797407905D-04,   0.000000000000000D+00, &
   1.302413844714418D-05,   0.000000000000000D+00,  -1.295367179334134D-06, &
   0.000000000000000D+00,   1.714630189815976D-07,   0.000000000000000D+00, &
   3.302662552197573D-07,   0.000000000000000D+00,  -8.368709300954763D-08, &
   0.000000000000000D+00,  -1.369234337647969D-07,   0.000000000000000D+00, &
   1.000130300214467D-03,   0.000000000000000D+00,  -9.286094257263453D-04, &
   0.000000000000000D+00,  -1.372620714326772D-03,   0.000000000000000D+00, &
  -2.833044242780526D-04,   0.000000000000000D+00,  -4.926511962561687D-05, &
   0.000000000000000D+00,  -9.214826826855018D-06,   0.000000000000000D+00, &
  -1.784885756658673D-07,   0.000000000000000D+00,  -2.407223221744199D-08, &
   0.000000000000000D+00,  -1.822626621029663D-07,   0.000000000000000D+00, &
   6.533533824921976D-08,   0.000000000000000D+00,   4.399452641729985D-07 /)
lem(:,244) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.892692545775752D-05, &
   3.745452721661506D-04,  -5.960036405441328D-04,  -1.200064847199536D-03, &
   2.076815360972200D-03,   3.641273417607956D-03,  -4.777185875798953D-03, &
  -6.330427439382061D-03,   6.581305850204596D-03,   5.651786608465147D-03, &
  -4.025125064558048D-03,  -2.697344346033448D-03,   1.151227775243558D-04, &
   4.100660125615509D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.892692545776392D-05,  -3.745452721661460D-04,   5.960036405441262D-04, &
   1.200064847199542D-03,  -2.076815360972194D-03,  -3.641273417607968D-03, &
   4.777185875798972D-03,   6.330427439382063D-03,  -6.581305850204610D-03, &
  -5.651786608465137D-03,   4.025125064558079D-03,   2.697344346033436D-03, &
  -1.151227775243699D-04,  -4.100660125615199D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.895583443415034D-08, &
   0.000000000000000D+00,  -7.867049442568774D-07,   0.000000000000000D+00, &
  -4.100703709364830D-06,   0.000000000000000D+00,  -6.709920164131742D-06, &
   0.000000000000000D+00,   1.637163257806285D-06,   0.000000000000000D+00, &
   1.241753975194047D-05,   0.000000000000000D+00,   5.193924158745646D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.464030460218923D-07, &
   0.000000000000000D+00,   7.247051606809718D-09,   0.000000000000000D+00, &
   3.272080156935302D-10,   0.000000000000000D+00,  -1.299742953297385D-10, &
   0.000000000000000D+00,  -2.271425830095020D-06,   0.000000000000000D+00, &
  -1.607900189070785D-07,   0.000000000000000D+00,  -1.337008308516611D-08, &
   0.000000000000000D+00,  -1.169768654350691D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.079416467784902D-05,   0.000000000000000D+00, &
   1.123234182376185D-06,   0.000000000000000D+00,   1.462352389882939D-07, &
   0.000000000000000D+00,   1.454403273260091D-08,   0.000000000000000D+00, &
  -4.629308582534725D-09,   0.000000000000000D+00,  -1.043679230739189D-05, &
   0.000000000000000D+00,  -2.207207950311490D-06,   0.000000000000000D+00, &
  -5.861465742692682D-07,   0.000000000000000D+00,  -1.125432272915156D-07, &
   0.000000000000000D+00,  -8.342158892148238D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.940676875286580D-05,   0.000000000000000D+00, &
  -4.916726902533147D-06,   0.000000000000000D+00,   2.344655069519024D-07, &
   0.000000000000000D+00,   2.705187725111171D-07,   0.000000000000000D+00, &
   7.478862895396795D-08,   0.000000000000000D+00,  -8.621508443358333D-09, &
   0.000000000000000D+00,   3.278529234587710D-04,   0.000000000000000D+00, &
   1.638894300501731D-05,   0.000000000000000D+00,   3.671805441246062D-06, &
   0.000000000000000D+00,   3.294861711958345D-07,   0.000000000000000D+00, &
  -9.802559473087801D-08,   0.000000000000000D+00,  -4.613651275099898D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.200880142081024D-04, &
   0.000000000000000D+00,   5.461874001598685D-05,   0.000000000000000D+00, &
  -3.664308476192714D-06,   0.000000000000000D+00,  -2.038745637233536D-06, &
   0.000000000000000D+00,  -3.944863662876364D-07,   0.000000000000000D+00, &
   2.799351217897311D-08,   0.000000000000000D+00,   8.562133967975432D-08, &
   0.000000000000000D+00,   1.913849864471329D-04,   0.000000000000000D+00, &
  -3.206585378080420D-04,   0.000000000000000D+00,  -1.762666737027510D-05, &
   0.000000000000000D+00,  -4.131366408568506D-07,   0.000000000000000D+00, &
   8.987449689629845D-07,   0.000000000000000D+00,   2.893514424057034D-07, &
   0.000000000000000D+00,  -3.259653910561398D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.473339757907150D-04,   0.000000000000000D+00, &
   5.743683885952012D-04,   0.000000000000000D+00,   9.941397323907691D-06, &
   0.000000000000000D+00,   8.174520748766482D-06,   0.000000000000000D+00, &
   1.346931805822303D-06,   0.000000000000000D+00,  -2.569685971962488D-07, &
   0.000000000000000D+00,  -1.644721108547597D-07,   0.000000000000000D+00, &
   1.580254639051781D-07,   0.000000000000000D+00,  -1.153350481710494D-03, &
   0.000000000000000D+00,  -2.539341389334899D-04,   0.000000000000000D+00, &
   1.599646342080253D-04,   0.000000000000000D+00,   5.612780247672808D-06, &
   0.000000000000000D+00,  -3.163321513287557D-06,   0.000000000000000D+00, &
  -1.169376055618446D-06,   0.000000000000000D+00,  -1.618453546375659D-08, &
   0.000000000000000D+00,   4.856771960945720D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.875295563563543D-04,   0.000000000000000D+00, &
  -6.863830814201899D-04,   0.000000000000000D+00,  -4.026563383271306D-04, &
   0.000000000000000D+00,  -4.179582982912928D-05,   0.000000000000000D+00, &
  -6.689137153446103D-06,   0.000000000000000D+00,   6.115714896787794D-07, &
   0.000000000000000D+00,   7.019845808472722D-07,   0.000000000000000D+00, &
   6.557369336268932D-08,   0.000000000000000D+00,  -2.212922789576891D-07, &
   0.000000000000000D+00,   2.160539495867563D-04,   0.000000000000000D+00, &
   1.268878980618115D-03,   0.000000000000000D+00,   2.498804151916681D-04, &
   0.000000000000000D+00,  -1.027187671264012D-05,   0.000000000000000D+00, &
   1.202864995002898D-05,   0.000000000000000D+00,   4.807877398628847D-06, &
   0.000000000000000D+00,   4.697603497266871D-07,   0.000000000000000D+00, &
  -2.784686344137440D-07,   0.000000000000000D+00,  -1.930048473687811D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.264352024782882D-04, &
   0.000000000000000D+00,  -7.288703751486533D-04,   0.000000000000000D+00, &
   4.306760360418309D-04,   0.000000000000000D+00,   1.826332318541640D-04, &
   0.000000000000000D+00,   2.256558845505964D-05,   0.000000000000000D+00, &
  -1.597316517777375D-06,   0.000000000000000D+00,  -2.699824703387650D-06, &
   0.000000000000000D+00,  -7.313205152740576D-07,   0.000000000000000D+00, &
   5.519575848522850D-08,   0.000000000000000D+00,  -3.174024801219159D-07, &
   0.000000000000000D+00,   4.236576539765877D-04,   0.000000000000000D+00, &
  -2.866049144267170D-04,   0.000000000000000D+00,  -8.782343053886232D-04, &
   0.000000000000000D+00,  -1.868966345632514D-04,   0.000000000000000D+00, &
  -4.298110684655552D-05,   0.000000000000000D+00,  -1.470686482843044D-05, &
   0.000000000000000D+00,  -1.840728943846536D-06,   0.000000000000000D+00, &
   1.104722426213944D-06,   0.000000000000000D+00,   5.872827333129580D-07, &
   0.000000000000000D+00,   2.015267035043524D-07,   0.000000000000000D+00 /)
lem(:,245) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.211727239560483D-05,  -1.449633874757389D-04,   3.930365952786190D-04, &
   3.277595109804113D-04,  -1.146358496683441D-03,   6.235007386738999D-04, &
  -1.774819711561894D-04,  -4.542320910708007D-03,   4.700191012213930D-03, &
   8.910576524251205D-03,  -6.949955459282474D-03,  -8.087745487576743D-03, &
   3.455347930809192D-03,   3.478053208496349D-03,   2.263190606460913D-04, &
  -5.215838768457250D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.211727239561184D-05,  -1.449633874757549D-04, &
   3.930365952786124D-04,   3.277595109804252D-04,  -1.146358496683440D-03, &
   6.235007386739007D-04,  -1.774819711561964D-04,  -4.542320910708019D-03, &
   4.700191012213927D-03,   8.910576524251207D-03,  -6.949955459282444D-03, &
  -8.087745487576722D-03,   3.455347930809169D-03,   3.478053208496343D-03, &
   2.263190606461071D-04,  -5.215838768457169D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.029879877826760D-08, &
   0.000000000000000D+00,   5.437047465909832D-07,   0.000000000000000D+00, &
   4.396901390873822D-06,   0.000000000000000D+00,   1.092451148540841D-05, &
   0.000000000000000D+00,   5.353116104992933D-06,   0.000000000000000D+00, &
  -1.098716665289833D-05,   0.000000000000000D+00,  -1.080129150151648D-05, &
   0.000000000000000D+00,  -2.630040243142562D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.527116214857326D-07,   0.000000000000000D+00,   6.251369733854208D-09, &
   0.000000000000000D+00,  -4.382971578302438D-11,   0.000000000000000D+00, &
  -3.822263274967502D-06,   0.000000000000000D+00,  -1.469478877640186D-07, &
   0.000000000000000D+00,  -7.883870108046609D-09,   0.000000000000000D+00, &
   3.100678448615469D-09,   1.416411344468132D-05,   0.000000000000000D+00, &
   1.062502366981426D-06,   0.000000000000000D+00,   1.100321868602302D-07, &
   0.000000000000000D+00,   1.606829946251353D-09,   0.000000000000000D+00, &
   2.158732521142971D-05,   0.000000000000000D+00,  -2.204155613709174D-06, &
   0.000000000000000D+00,  -4.798582723292204D-07,   0.000000000000000D+00, &
  -6.822153441328371D-08,   0.000000000000000D+00,   2.517615635445244D-08, &
  -1.561846767760491D-04,   0.000000000000000D+00,  -3.551459560194294D-06, &
   0.000000000000000D+00,   1.949647826667213D-07,   0.000000000000000D+00, &
   2.025124152393867D-07,   0.000000000000000D+00,   2.832973532589152D-08, &
   0.000000000000000D+00,   2.659329335774389D-05,   0.000000000000000D+00, &
   9.389499858767548D-06,   0.000000000000000D+00,   3.272232464089589D-06, &
   0.000000000000000D+00,   2.608444597796019D-07,   0.000000000000000D+00, &
  -7.819566411270720D-08,   0.000000000000000D+00,  -2.644250990202216D-08, &
   7.272047200762057D-04,   0.000000000000000D+00,  -1.133203679694523D-06, &
   0.000000000000000D+00,  -4.317143959061764D-06,   0.000000000000000D+00, &
  -1.774834162934793D-06,   0.000000000000000D+00,  -2.789602772858333D-07, &
   0.000000000000000D+00,   5.314856053423058D-08,   0.000000000000000D+00, &
  -9.336616385702186D-04,   0.000000000000000D+00,   1.171374753498166D-04, &
   0.000000000000000D+00,  -9.643258955346020D-06,   0.000000000000000D+00, &
   2.851098739653252D-07,   0.000000000000000D+00,   7.960596493383984D-07, &
   0.000000000000000D+00,   1.772905533249782D-07,   0.000000000000000D+00, &
  -2.278085102316019D-07,  -6.262054004842212D-04,   0.000000000000000D+00, &
  -4.569544212920669D-04,   0.000000000000000D+00,   4.682416926131373D-05, &
   0.000000000000000D+00,   6.633090982621226D-06,   0.000000000000000D+00, &
   8.064690275619453D-07,   0.000000000000000D+00,  -2.787350035878322D-07, &
   0.000000000000000D+00,  -4.864888179532930D-08,   0.000000000000000D+00, &
   1.912427213546813D-03,   0.000000000000000D+00,   4.227321301394397D-04, &
   0.000000000000000D+00,  -1.760906098022989D-04,   0.000000000000000D+00, &
  -8.432469217346670D-06,   0.000000000000000D+00,  -3.501412278623731D-06, &
   0.000000000000000D+00,  -8.536264966165166D-07,   0.000000000000000D+00, &
   7.782798550787888D-08,   0.000000000000000D+00,   9.309991654078226D-08, &
  -6.697126504763009D-04,   0.000000000000000D+00,   5.058561951723382D-04, &
   0.000000000000000D+00,   4.014596010708953D-04,   0.000000000000000D+00, &
  -1.240358294024619D-05,   0.000000000000000D+00,   4.038983872011482D-07, &
   0.000000000000000D+00,   1.505923425204227D-06,   0.000000000000000D+00, &
   5.975454011641117D-07,   0.000000000000000D+00,   2.272209639247286D-09, &
   0.000000000000000D+00,  -1.068260772763359D-03,   0.000000000000000D+00, &
  -1.214570409035943D-03,   0.000000000000000D+00,  -2.808679823595268D-04, &
   0.000000000000000D+00,   1.089746952637812D-04,   0.000000000000000D+00, &
   1.064703336679599D-05,   0.000000000000000D+00,   1.566990787815751D-06, &
   0.000000000000000D+00,  -4.402203820332510D-07,   0.000000000000000D+00, &
  -3.794832451947079D-07,   0.000000000000000D+00,   4.384496530536542D-07, &
   8.939267615397676D-04,   0.000000000000000D+00,   5.394503378759455D-04, &
   0.000000000000000D+00,  -4.922067156140453D-04,   0.000000000000000D+00, &
  -2.820992464617308D-04,   0.000000000000000D+00,  -2.265641728208764D-05, &
   0.000000000000000D+00,  -5.892031802574869D-06,   0.000000000000000D+00, &
  -1.668686000508449D-06,   0.000000000000000D+00,  -2.384668934201227D-08, &
   0.000000000000000D+00,   1.143841145055106D-07,   0.000000000000000D+00, &
   3.156766036931115D-05,   0.000000000000000D+00,   4.830398905853312D-04, &
   0.000000000000000D+00,   1.047767482555652D-03,   0.000000000000000D+00, &
   2.158766811774879D-04,   0.000000000000000D+00,  -1.077957034315127D-05, &
   0.000000000000000D+00,   3.784414252226515D-06,   0.000000000000000D+00, &
   2.692942414903026D-06,   0.000000000000000D+00,   1.263317756172308D-06, &
   0.000000000000000D+00,   1.255047262705803D-07,   0.000000000000000D+00, &
  -4.098152967165714D-07,  -2.205632078428272D-04,   0.000000000000000D+00, &
  -5.579872331461355D-04,   0.000000000000000D+00,  -5.241029542386782D-04, &
   0.000000000000000D+00,   3.176547135837127D-04,   0.000000000000000D+00, &
   1.301836715768905D-04,   0.000000000000000D+00,   1.589889500461351D-05, &
   0.000000000000000D+00,   1.226658840383346D-06,   0.000000000000000D+00, &
  -9.082326152611204D-07,   0.000000000000000D+00,  -7.853310307163568D-07, &
   0.000000000000000D+00,  -3.291354463671607D-07,   0.000000000000000D+00, &
   3.406769890232546D-05,   0.000000000000000D+00,   1.534348811930898D-04, &
   0.000000000000000D+00,  -3.197781293773501D-04,   0.000000000000000D+00, &
  -7.066609367702738D-04,   0.000000000000000D+00,  -1.524391142640039D-04, &
   0.000000000000000D+00,  -3.029868559585604D-05,   0.000000000000000D+00, &
  -9.349673953878187D-06,   0.000000000000000D+00,  -2.158567480979843D-06, &
   0.000000000000000D+00,   1.361948039917754D-08,   0.000000000000000D+00, &
   6.192228605714879D-07,   0.000000000000000D+00,   6.002438218579532D-08 /)
lem(:,246) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.042234014581550D-06,   5.049712245617803D-05, &
  -2.254005426519246D-04,   6.469331080919713D-05,   9.958548812282802D-04, &
  -4.571728464734997D-04,  -1.634654086389581D-03,  -1.885680151005556D-03, &
   3.041081479033108D-03,   8.310174361911867D-03,  -6.151430340726283D-03, &
  -1.249953393386382D-02,   7.380553405718520D-03,   9.508216911981343D-03, &
  -4.058922837480583D-03,  -4.048301049093238D-03,   3.320112931366319D-04, &
   1.108499464827591D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.042234014567793D-06, &
  -5.049712245617162D-05,   2.254005426519320D-04,  -6.469331080920682D-05, &
  -9.958548812282785D-04,   4.571728464735002D-04,   1.634654086389577D-03, &
   1.885680151005564D-03,  -3.041081479033127D-03,  -8.310174361911862D-03, &
   6.151430340726302D-03,   1.249953393386382D-02,  -7.380553405718518D-03, &
  -9.508216911981332D-03,   4.058922837480584D-03,   4.048301049093228D-03, &
  -3.320112931366401D-04,  -1.108499464827593D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.749867287250219D-07,   0.000000000000000D+00, &
  -3.984695511015491D-06,   0.000000000000000D+00,  -1.450749692093803D-05, &
   0.000000000000000D+00,  -1.461759960320827D-05,   0.000000000000000D+00, &
   7.280159442666784D-06,   0.000000000000000D+00,   1.244813719572357D-05, &
   0.000000000000000D+00,   4.986468141569819D-07,   0.000000000000000D+00, &
   9.245981688240721D-06,   0.000000000000000D+00,   4.929904163231219D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.798364585259131D-09,   0.000000000000000D+00, &
  -8.093291171838460D-10,   0.000000000000000D+00,  -1.618658234622322D-07, &
   0.000000000000000D+00,  -2.521371481029844D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.232322964903595D-06,   0.000000000000000D+00, &
   8.355797179667149D-08,   0.000000000000000D+00,  -2.393158828683223D-08, &
   0.000000000000000D+00,  -3.172706888573538D-06,   0.000000000000000D+00, &
  -4.204556001800255D-07,   0.000000000000000D+00,  -2.292334829133762D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.159506541616414D-06, &
   0.000000000000000D+00,   2.296407458046004D-07,   0.000000000000000D+00, &
   1.530317350727926D-07,   0.000000000000000D+00,  -3.414749895737850D-08, &
   0.000000000000000D+00,   3.023383172723505D-05,   0.000000000000000D+00, &
   2.853980085612513D-06,   0.000000000000000D+00,   1.800463789950010D-07, &
   0.000000000000000D+00,  -6.334478606887594D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.622366004979266D-05,   0.000000000000000D+00, &
  -2.984975006588870D-06,   0.000000000000000D+00,  -1.505347895064731D-06, &
   0.000000000000000D+00,  -1.571600267417641D-07,   0.000000000000000D+00, &
   2.310416455269243D-07,   0.000000000000000D+00,  -7.939384358809687D-05, &
   0.000000000000000D+00,  -1.496246308663882D-05,   0.000000000000000D+00, &
   4.295083089298563D-07,   0.000000000000000D+00,   6.539326981029514D-07, &
   0.000000000000000D+00,   2.363041532780011D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.604048957556262D-04,   0.000000000000000D+00, &
   9.313510509613111D-06,   0.000000000000000D+00,   4.551871467890827D-06, &
   0.000000000000000D+00,   4.580370314599603D-07,   0.000000000000000D+00, &
  -2.685479130163423D-07,   0.000000000000000D+00,   1.319230529847389D-07, &
   0.000000000000000D+00,  -1.876827093026544D-04,   0.000000000000000D+00, &
   1.362662019472518D-04,   0.000000000000000D+00,  -1.222828952000394D-06, &
   0.000000000000000D+00,  -2.299635590798432D-06,   0.000000000000000D+00, &
  -4.655087638180170D-07,   0.000000000000000D+00,   8.839011227334916D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.402339580671914D-04, &
   0.000000000000000D+00,  -3.485967174928456D-04,   0.000000000000000D+00, &
   1.446207077537008D-05,   0.000000000000000D+00,   6.150061325869208D-08, &
   0.000000000000000D+00,   9.901165458067445D-07,   0.000000000000000D+00, &
   2.923694756442284D-07,   0.000000000000000D+00,  -6.729845023794154D-07, &
   0.000000000000000D+00,   5.292372125909348D-04,   0.000000000000000D+00, &
   2.184635519489865D-04,   0.000000000000000D+00,  -1.445892333935517D-04, &
   0.000000000000000D+00,  -2.097688995402541D-06,   0.000000000000000D+00, &
   3.776471997643987D-07,   0.000000000000000D+00,  -3.981316035468768D-07, &
   0.000000000000000D+00,  -1.013896679272653D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.577442599836275D-04,   0.000000000000000D+00, &
   3.776530006396226D-04,   0.000000000000000D+00,   3.343955794221301D-04, &
   0.000000000000000D+00,   1.631171598803669D-06,   0.000000000000000D+00, &
   3.185354791399892D-07,   0.000000000000000D+00,  -3.061317349285739D-07, &
   0.000000000000000D+00,   3.099375347426240D-07,   0.000000000000000D+00, &
   1.889980919283848D-07,   0.000000000000000D+00,  -2.148884341818009D-04, &
   0.000000000000000D+00,  -7.738430053493895D-04,   0.000000000000000D+00, &
  -1.981425821222246D-04,   0.000000000000000D+00,   8.119637107568812D-05, &
   0.000000000000000D+00,   3.339189859353182D-06,   0.000000000000000D+00, &
   1.549282135367454D-07,   0.000000000000000D+00,   1.364847612724659D-07, &
   0.000000000000000D+00,  -3.428547815846163D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.228419134528766D-04,   0.000000000000000D+00, &
   4.040802108319559D-04,   0.000000000000000D+00,  -3.922006074368048D-04, &
   0.000000000000000D+00,  -2.359250712674516D-04,   0.000000000000000D+00, &
  -2.105893551667286D-05,   0.000000000000000D+00,  -2.914896219742205D-06, &
   0.000000000000000D+00,  -3.823635665483722D-07,   0.000000000000000D+00, &
  -2.140353645654603D-07,   0.000000000000000D+00,   8.847004691568142D-07, &
   0.000000000000000D+00,  -1.094921855587940D-04,   0.000000000000000D+00, &
   2.498804151916681D-04,   0.000000000000000D+00,   7.848208906833642D-04, &
   0.000000000000000D+00,   1.693943629703175D-04,   0.000000000000000D+00, &
  -3.292278552367135D-06,   0.000000000000000D+00,   5.846874698570475D-06, &
   0.000000000000000D+00,   1.948336969013476D-06,   0.000000000000000D+00, &
   4.664006668749683D-07,   0.000000000000000D+00,   9.107450391179900D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.972828278299928D-05, &
   0.000000000000000D+00,  -4.298778949392931D-04,   0.000000000000000D+00, &
  -4.188673491184603D-04,   0.000000000000000D+00,   2.581590194348226D-04, &
   0.000000000000000D+00,   1.095266038480532D-04,   0.000000000000000D+00, &
   1.245330859071476D-05,   0.000000000000000D+00,  -6.480408638284343D-07, &
   0.000000000000000D+00,  -1.021241028291210D-06,   0.000000000000000D+00, &
  -7.129461945609274D-07,   0.000000000000000D+00,  -9.706425325231555D-07, &
   0.000000000000000D+00,   4.802513874593013D-05,   0.000000000000000D+00, &
   1.836644083723721D-04,   0.000000000000000D+00,  -2.258982231420233D-04, &
   0.000000000000000D+00,  -5.547991677495476D-04,   0.000000000000000D+00, &
  -1.237511188366805D-04,   0.000000000000000D+00,  -2.788426014625294D-05, &
   0.000000000000000D+00,  -8.587593266230653D-06,   0.000000000000000D+00, &
  -1.117986156819211D-06,   0.000000000000000D+00,   4.750585553027463D-07, &
   0.000000000000000D+00,   2.780589276850175D-07,   0.000000000000000D+00 /)
lem(:,247) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.037751904855668D-06, &
  -1.843006461960731D-05,   9.860896154459257D-05,  -1.041006128514723D-04, &
  -3.857820336056624D-04,   5.354933539247871D-04,   4.041441685866115D-05, &
  -6.733772628505854D-04,   1.302054506902045D-03,   3.678944735562160D-03, &
  -4.461920891339691D-03,  -1.104682274775786D-02,   8.400798687569110D-03, &
   1.459282196290696D-02,  -9.079085016547842D-03,  -9.557651620277819D-03, &
   5.530273304107125D-03,   3.451240938055428D-03,  -1.780245255024669D-03, &
  -1.215061064646121D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.037751904877679D-06,  -1.843006461961445D-05,   9.860896154458220D-05, &
  -1.041006128514593D-04,  -3.857820336056578D-04,   5.354933539247723D-04, &
   4.041441685866131D-05,  -6.733772628505771D-04,   1.302054506902050D-03, &
   3.678944735562164D-03,  -4.461920891339699D-03,  -1.104682274775787D-02, &
   8.400798687569121D-03,   1.459282196290695D-02,  -9.079085016547833D-03, &
  -9.557651620277829D-03,   5.530273304107110D-03,   3.451240938055411D-03, &
  -1.780245255024656D-03,  -1.215061064646112D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.918346032741719D-08,   0.000000000000000D+00, &
   2.899891588073856D-06,   0.000000000000000D+00,   1.653645340511471D-05, &
   0.000000000000000D+00,   2.606803784106977D-05,   0.000000000000000D+00, &
   6.959278129233227D-07,   0.000000000000000D+00,  -1.696198933648919D-05, &
   0.000000000000000D+00,   3.928739799470039D-06,   0.000000000000000D+00, &
   2.166198383647698D-06,   0.000000000000000D+00,  -1.859626300136525D-05, &
   0.000000000000000D+00,  -1.120005141598454D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.479204602059287D-08,   0.000000000000000D+00, &
  -3.078594578119279D-07,   0.000000000000000D+00,   7.164897291451128D-09, &
   2.017482134814934D-06,   0.000000000000000D+00,   6.574631776550527D-08, &
   0.000000000000000D+00,  -2.699805205189676D-06,   0.000000000000000D+00, &
  -4.175437864360622D-07,   0.000000000000000D+00,   5.393565733521165D-08, &
  -1.582276853164971D-05,   0.000000000000000D+00,   4.134373893187016D-07, &
   0.000000000000000D+00,   1.093366840653054D-07,   0.000000000000000D+00, &
   3.675817237447758D-05,   0.000000000000000D+00,   1.971398748822764D-06, &
   0.000000000000000D+00,   7.988573820183067D-08,   0.000000000000000D+00, &
  -1.636293210174190D-07,   5.043168749189936D-05,   0.000000000000000D+00, &
  -2.511222297283430D-06,   0.000000000000000D+00,  -1.243094989786310D-06, &
   0.000000000000000D+00,  -9.814871217590054D-09,   0.000000000000000D+00, &
  -1.570464753676091D-04,   0.000000000000000D+00,  -2.499155536815297D-07, &
   0.000000000000000D+00,   4.057592546768423D-07,   0.000000000000000D+00, &
   5.248191518024793D-07,   0.000000000000000D+00,  -3.727577869998259D-07, &
  -7.824909266943306D-05,   0.000000000000000D+00,  -1.691894950413843D-05, &
   0.000000000000000D+00,   5.183617731247332D-06,   0.000000000000000D+00, &
   2.051432758603117D-07,   0.000000000000000D+00,  -1.608962070417818D-07, &
   0.000000000000000D+00,   3.570084220156669D-04,   0.000000000000000D+00, &
   2.545637416870728D-06,   0.000000000000000D+00,  -7.787831228089671D-06, &
   0.000000000000000D+00,  -1.747969351482100D-06,   0.000000000000000D+00, &
  -1.650086617860320D-07,   0.000000000000000D+00,   6.873918627410565D-07, &
   4.304486273903045D-05,   0.000000000000000D+00,   1.530035523553764D-04, &
   0.000000000000000D+00,  -8.643341367856004D-06,   0.000000000000000D+00, &
  -1.273974061940714D-06,   0.000000000000000D+00,   6.399524663972951D-07, &
   0.000000000000000D+00,   2.986689645436527D-08,   0.000000000000000D+00, &
  -4.896030683354715D-04,   0.000000000000000D+00,  -2.142014354299925D-04, &
   0.000000000000000D+00,   9.196547693987149D-05,   0.000000000000000D+00, &
   5.195291344452029D-06,   0.000000000000000D+00,   8.613822402369028D-07, &
   0.000000000000000D+00,  -3.513329389762476D-07,   0.000000000000000D+00, &
   4.233838006338194D-07,   4.702155960827982D-05,   0.000000000000000D+00, &
  -1.463922844156403D-04,   0.000000000000000D+00,  -2.433407397838047D-04, &
   0.000000000000000D+00,   1.740573531054120D-05,   0.000000000000000D+00, &
  -1.348542351445901D-07,   0.000000000000000D+00,  -3.303192064079006D-07, &
   0.000000000000000D+00,  -1.921573593869110D-08,   0.000000000000000D+00, &
   4.176938797670679D-04,   0.000000000000000D+00,   4.596423102619445D-04, &
   0.000000000000000D+00,   1.876198103959929D-04,   0.000000000000000D+00, &
  -1.166483141777014D-04,   0.000000000000000D+00,  -7.205413842529361D-06, &
   0.000000000000000D+00,  -6.743730012261097D-07,   0.000000000000000D+00, &
  -7.210138907325513D-10,   0.000000000000000D+00,  -1.203338333234223D-06, &
  -1.020181221140951D-04,   0.000000000000000D+00,  -1.577089286189178D-04, &
   0.000000000000000D+00,   2.770982514094167D-04,   0.000000000000000D+00, &
   2.576197817534103D-04,   0.000000000000000D+00,   2.110120852615451D-06, &
   0.000000000000000D+00,   2.422370842818462D-06,   0.000000000000000D+00, &
   6.766051796175500D-07,   0.000000000000000D+00,   1.441523998516801D-07, &
   0.000000000000000D+00,  -2.148121126619998D-04,   0.000000000000000D+00, &
  -2.485804735978500D-04,   0.000000000000000D+00,  -6.190422225728379D-04, &
   0.000000000000000D+00,  -1.643882898518547D-04,   0.000000000000000D+00, &
   6.790267717088054D-05,   0.000000000000000D+00,   4.671082792089535D-06, &
   0.000000000000000D+00,  -3.848219658398914D-07,   0.000000000000000D+00, &
  -3.751421842796957D-08,   0.000000000000000D+00,   8.273024579300141D-10, &
   7.845733290466745D-05,   0.000000000000000D+00,   1.920182330635061D-04, &
   0.000000000000000D+00,   2.972765339258929D-04,   0.000000000000000D+00, &
  -3.083080682221210D-04,   0.000000000000000D+00,  -1.896920356816431D-04, &
   0.000000000000000D+00,  -1.918955065049504D-05,   0.000000000000000D+00, &
  -4.001669695943166D-06,   0.000000000000000D+00,  -5.579004097414035D-07, &
   0.000000000000000D+00,   3.738220939464025D-07,   0.000000000000000D+00, &
   5.837909698479359D-05,   0.000000000000000D+00,   2.121917905712801D-06, &
   0.000000000000000D+00,   2.158766811774879D-04,   0.000000000000000D+00, &
   6.384929451475414D-04,   0.000000000000000D+00,   1.415639255945486D-04, &
   0.000000000000000D+00,  -3.066089814007063D-06,   0.000000000000000D+00, &
   5.024503611814026D-06,   0.000000000000000D+00,   2.455321178974458D-06, &
   0.000000000000000D+00,   6.012566642687948D-07,   0.000000000000000D+00, &
   8.355458085723141D-07,  -2.925195938871648D-05,   0.000000000000000D+00, &
  -2.260510646745369D-05,   0.000000000000000D+00,  -3.017560728707810D-04, &
   0.000000000000000D+00,  -3.299831099866026D-04,   0.000000000000000D+00, &
   2.083940000685976D-04,   0.000000000000000D+00,   8.927551818483453D-05, &
   0.000000000000000D+00,   1.116381916467259D-05,   0.000000000000000D+00, &
  -9.044643294045664D-08,   0.000000000000000D+00,  -1.431956228890467D-06, &
   0.000000000000000D+00,  -9.348927146355129D-07,   0.000000000000000D+00, &
  -4.612162190532008D-06,   0.000000000000000D+00,  -1.776804151587500D-06, &
   0.000000000000000D+00,   1.260511574776833D-04,   0.000000000000000D+00, &
  -1.877220966311915D-04,   0.000000000000000D+00,  -4.592259312367593D-04, &
   0.000000000000000D+00,  -1.042415563718348D-04,   0.000000000000000D+00, &
  -2.324746261650476D-05,   0.000000000000000D+00,  -7.636788535685975D-06, &
   0.000000000000000D+00,  -1.289640348134355D-06,   0.000000000000000D+00, &
   1.583573112413638D-07,   0.000000000000000D+00,   5.250226006794714D-07 /)
lem(:,248) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.111441794591643D-06,  -3.072850652032822D-05, &
   9.970865978801879D-05,   9.922108380217018D-05,  -6.216587001902133D-04, &
  -1.235099061142710D-04,   1.011856446153872D-03,   1.238811510978399D-03, &
   5.858548936097138D-05,  -3.590984488417328D-03,  -5.323379630905996D-03, &
   6.916200475288919D-03,   1.266627923515766D-02,  -1.063610813992622D-02, &
  -1.361593823809748D-02,   1.126696447166283D-02,   7.090115270113159D-03, &
  -7.215363071888555D-03,  -1.496779890000800D-03,   2.811411541088615D-03, &
   1.460433223021520D-04,   0.000000000000000D+00,  -2.111441794604824D-06, &
   3.072850652033922D-05,  -9.970865978801069D-05,  -9.922108380218031D-05, &
   6.216587001902044D-04,   1.235099061142813D-04,  -1.011856446153869D-03, &
  -1.238811510978407D-03,  -5.858548936097283D-05,   3.590984488417340D-03, &
   5.323379630905985D-03,  -6.916200475288942D-03,  -1.266627923515766D-02, &
   1.063610813992622D-02,   1.361593823809747D-02,  -1.126696447166285D-02, &
  -7.090115270113152D-03,   7.215363071888544D-03,   1.496779890000801D-03, &
  -2.811411541088613D-03,  -1.460433223021540D-04,   0.000000000000000D+00, &
  -1.055720897301632D-06,   0.000000000000000D+00,  -1.492644020429135D-05, &
   0.000000000000000D+00,  -3.665016573128330D-05,   0.000000000000000D+00, &
  -1.382069798658583D-05,   0.000000000000000D+00,   2.308515595220694D-05, &
   0.000000000000000D+00,  -1.235233905097663D-06,   0.000000000000000D+00, &
  -1.813862050676312D-05,   0.000000000000000D+00,   2.638631440678765D-06, &
   0.000000000000000D+00,   1.827758680973248D-05,   0.000000000000000D+00, &
   2.541391117956528D-05,   0.000000000000000D+00,  -1.076452619536673D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.346206522256908D-08, &
   0.000000000000000D+00,  -5.975883869946131D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.208562260789209D-07,   0.000000000000000D+00, &
   1.557530540234983D-07,   0.000000000000000D+00,   2.942059199630860D-06, &
   0.000000000000000D+00,  -7.445049678204855D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.033710516527148D-05,   0.000000000000000D+00, &
  -8.956731831265110D-07,   0.000000000000000D+00,   5.287319114864217D-07, &
   0.000000000000000D+00,  -3.287558203986110D-07,   0.000000000000000D+00, &
  -2.287039036659231D-07,   0.000000000000000D+00,   3.026779850793803D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.719281005473127D-05, &
   0.000000000000000D+00,   5.346171628261917D-06,   0.000000000000000D+00, &
  -2.383984463434631D-08,   0.000000000000000D+00,  -4.818925994589027D-07, &
   0.000000000000000D+00,  -3.182276834896121D-05,   0.000000000000000D+00, &
   3.742300310804671D-06,   0.000000000000000D+00,  -1.463627857221856D-06, &
   0.000000000000000D+00,   3.249124703203885D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.111836504641266D-05,   0.000000000000000D+00, &
  -3.052836081705153D-05,   0.000000000000000D+00,   4.307044469776215D-07, &
   0.000000000000000D+00,   5.096025555234926D-07,   0.000000000000000D+00, &
  -1.105575422703780D-06,   0.000000000000000D+00,   1.009252567074195D-04, &
   0.000000000000000D+00,  -5.123634076511546D-06,   0.000000000000000D+00, &
  -3.006744345120459D-06,   0.000000000000000D+00,   7.390052404078627D-07, &
   0.000000000000000D+00,   3.454574347483732D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.449496598549944D-05,   0.000000000000000D+00, &
   1.334906468598345D-04,   0.000000000000000D+00,   3.170509307191117D-06, &
   0.000000000000000D+00,  -1.857176962740605D-06,   0.000000000000000D+00, &
  -1.839027589836648D-07,   0.000000000000000D+00,   1.142939515835250D-06, &
   0.000000000000000D+00,  -1.497905928130990D-04,   0.000000000000000D+00, &
  -1.408544965544367D-04,   0.000000000000000D+00,   6.761301436470501D-05, &
   0.000000000000000D+00,   2.172498850711521D-06,   0.000000000000000D+00, &
   1.368069074859417D-07,   0.000000000000000D+00,   1.914405821499281D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.779052035309390D-05, &
   0.000000000000000D+00,  -1.175592949741629D-04,   0.000000000000000D+00, &
  -2.024019943449647D-04,   0.000000000000000D+00,   9.926997561993735D-06, &
   0.000000000000000D+00,   1.483757380289102D-06,   0.000000000000000D+00, &
  -1.511823469826868D-07,   0.000000000000000D+00,   4.263429340986675D-07, &
   0.000000000000000D+00,   1.182680425585441D-04,   0.000000000000000D+00, &
   3.142137506440675D-04,   0.000000000000000D+00,   1.473096123214649D-04, &
   0.000000000000000D+00,  -9.039441275682023D-05,   0.000000000000000D+00, &
  -5.700815069918569D-06,   0.000000000000000D+00,  -1.125816074037707D-06, &
   0.000000000000000D+00,  -6.441238512775862D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.026949882432194D-05,   0.000000000000000D+00, &
  -1.272769320472504D-04,   0.000000000000000D+00,   2.239011141504473D-04, &
   0.000000000000000D+00,   2.157381838903726D-04,   0.000000000000000D+00, &
   4.023780662664385D-06,   0.000000000000000D+00,   1.666976876086462D-06, &
   0.000000000000000D+00,   6.388860619018453D-07,   0.000000000000000D+00, &
  -5.053073678246792D-07,   0.000000000000000D+00,  -4.145057301347666D-05, &
   0.000000000000000D+00,  -1.636163454536218D-04,   0.000000000000000D+00, &
  -4.842563707199877D-04,   0.000000000000000D+00,  -1.367850115810443D-04, &
   0.000000000000000D+00,   5.298053971740659D-05,   0.000000000000000D+00, &
   3.579830355404934D-06,   0.000000000000000D+00,   2.436022171738322D-07, &
   0.000000000000000D+00,  -3.539696900446440D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.811720774260601D-05,   0.000000000000000D+00, &
   1.702418998334970D-04,   0.000000000000000D+00,   2.409068791879921D-04, &
   0.000000000000000D+00,  -2.546085645481721D-04,   0.000000000000000D+00, &
  -1.609615542040215D-04,   0.000000000000000D+00,  -1.722084298308033D-05, &
   0.000000000000000D+00,  -3.492557844954443D-06,   0.000000000000000D+00, &
  -2.138581353429215D-07,   0.000000000000000D+00,   4.553133933045281D-07, &
   0.000000000000000D+00,  -4.284196134875921D-06,   0.000000000000000D+00, &
  -1.027187671264012D-05,   0.000000000000000D+00,   1.693943629703175D-04, &
   0.000000000000000D+00,   5.220467688857051D-04,   0.000000000000000D+00, &
   1.208785461997040D-04,   0.000000000000000D+00,   1.697152648277293D-07, &
   0.000000000000000D+00,   4.983358205106899D-06,   0.000000000000000D+00, &
   2.206152622135807D-06,   0.000000000000000D+00,   1.200938686130211D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.840585674008354D-05, &
   0.000000000000000D+00,  -4.470709467719569D-05,   0.000000000000000D+00, &
  -2.530643266597164D-04,   0.000000000000000D+00,  -2.730169826450611D-04, &
   0.000000000000000D+00,   1.747923268600898D-04,   0.000000000000000D+00, &
   7.632925828671482D-05,   0.000000000000000D+00,   9.341650256765796D-06, &
   0.000000000000000D+00,  -6.474446233140101D-07,   0.000000000000000D+00, &
  -1.505014685043270D-06,   0.000000000000000D+00,  -1.647214602101262D-06, &
   0.000000000000000D+00,   7.756028958376641D-06,   0.000000000000000D+00, &
   6.538339521503504D-06,   0.000000000000000D+00,   9.487049989444262D-05, &
   0.000000000000000D+00,  -1.557060704771083D-04,   0.000000000000000D+00, &
  -3.836488595851126D-04,   0.000000000000000D+00,  -9.000663053962181D-05, &
   0.000000000000000D+00,  -2.150463205384777D-05,   0.000000000000000D+00, &
  -6.886799635396870D-06,   0.000000000000000D+00,  -1.435872109159456D-06, &
   0.000000000000000D+00,   3.702951256779039D-07,   0.000000000000000D+00 /)
lem(:,249) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.126461494437288D-05,  -3.911445924501062D-05,  -4.193256373268675D-05, &
   2.537566267505169D-04,  -5.923479021266602D-05,  -1.388413149327845D-04, &
   2.446320982037756D-04,  -5.849543609272916D-04,  -1.632705124482196D-03, &
  -1.650503152130347D-04,   4.882781646492284D-03,   5.553063466662941D-03, &
  -8.763117973501015D-03,  -1.158606925477711D-02,   1.192520136794275D-02, &
   9.810413845054825D-03,  -1.177662604593452D-02,  -2.504435499471269D-03, &
   7.201927184204227D-03,  -1.431761816898866D-03,  -2.468888368961101D-03, &
   1.230274044278815D-03,   1.126461494436556D-05,  -3.911445924501524D-05, &
  -4.193256373267728D-05,   2.537566267505183D-04,  -5.923479021267262D-05, &
  -1.388413149327808D-04,   2.446320982037786D-04,  -5.849543609272972D-04, &
  -1.632705124482201D-03,  -1.650503152130351D-04,   4.882781646492280D-03, &
   5.553063466662951D-03,  -8.763117973501025D-03,  -1.158606925477710D-02, &
   1.192520136794274D-02,   9.810413845054825D-03,  -1.177662604593453D-02, &
  -2.504435499471255D-03,   7.201927184204228D-03,  -1.431761816898876D-03, &
  -2.468888368961090D-03,   1.230274044278818D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.173791242410757D-05,   0.000000000000000D+00, &
   2.254161340764737D-05,   0.000000000000000D+00,  -3.582488630134230D-05, &
   0.000000000000000D+00,  -1.535016781475988D-05,   0.000000000000000D+00, &
   3.010806899556804D-05,   0.000000000000000D+00,   1.605666152778508D-05, &
   0.000000000000000D+00,   7.202873144067887D-06,   0.000000000000000D+00, &
  -1.371537425726934D-05,   0.000000000000000D+00,  -4.595071762393019D-05, &
   0.000000000000000D+00,  -5.785691745177876D-06,   0.000000000000000D+00, &
  -1.308267872366115D-06,   9.062054235509180D-08,   0.000000000000000D+00, &
   7.557425804329910D-06,   0.000000000000000D+00,  -9.462431392641745D-07, &
  -6.626518333174421D-06,   0.000000000000000D+00,  -2.310476557830246D-07, &
   0.000000000000000D+00,  -2.836821317395789D-05,   0.000000000000000D+00, &
   7.328361831722516D-07,   0.000000000000000D+00,   1.356017803911420D-07, &
   2.832326921745558D-05,   0.000000000000000D+00,  -8.636552923819757D-07, &
   0.000000000000000D+00,   1.013013772579662D-07,   0.000000000000000D+00, &
   6.212772740227632D-05,   0.000000000000000D+00,   4.179761786007957D-06, &
   0.000000000000000D+00,  -2.109224615305689D-06,   0.000000000000000D+00, &
   1.671580795124329D-06,  -4.946722120287545D-05,   0.000000000000000D+00, &
   8.412664397637088D-06,   0.000000000000000D+00,   1.525998207762389D-06, &
   0.000000000000000D+00,   1.299496269251311D-07,   0.000000000000000D+00, &
  -9.229082030473599D-05,   0.000000000000000D+00,  -3.162406105775583D-05, &
   0.000000000000000D+00,   6.764071832690060D-06,   0.000000000000000D+00, &
   1.381539638182048D-07,   0.000000000000000D+00,  -6.860997656179171D-07, &
   2.851792966013797D-05,   0.000000000000000D+00,  -1.846881704856498D-05, &
   0.000000000000000D+00,  -1.841302069917897D-05,   0.000000000000000D+00, &
   3.085207256339404D-07,   0.000000000000000D+00,  -5.349185375123171D-07, &
   0.000000000000000D+00,   1.089012834350040D-04,   0.000000000000000D+00, &
   8.695313494149864D-05,   0.000000000000000D+00,  -4.645094181567548D-06, &
   0.000000000000000D+00,  -7.095055076554729D-06,   0.000000000000000D+00, &
   7.439730579878235D-07,   0.000000000000000D+00,  -1.079822134194396D-06, &
   3.111860017823468D-05,   0.000000000000000D+00,   1.169142254492289D-05, &
   0.000000000000000D+00,   9.469929493674874D-05,   0.000000000000000D+00, &
   4.509963320888090D-06,   0.000000000000000D+00,  -1.045570812210382D-06, &
   0.000000000000000D+00,   2.013143602082236D-07,   0.000000000000000D+00, &
  -1.119094176384783D-04,   0.000000000000000D+00,  -1.265506469155292D-04, &
   0.000000000000000D+00,  -1.142301393943694D-04,   0.000000000000000D+00, &
   5.550571810417385D-05,   0.000000000000000D+00,   5.188819823863466D-06, &
   0.000000000000000D+00,   3.631130943761550D-07,   0.000000000000000D+00, &
   1.184406076746002D-06,  -6.400443560714663D-05,   0.000000000000000D+00, &
   1.298295145204354D-05,   0.000000000000000D+00,  -8.617532767602917D-05, &
   0.000000000000000D+00,  -1.630246777852102D-04,   0.000000000000000D+00, &
   4.314301876188870D-06,   0.000000000000000D+00,   4.739344556224158D-07, &
   0.000000000000000D+00,   4.691467394563776D-07,   0.000000000000000D+00, &
   9.315450159950361D-05,   0.000000000000000D+00,   1.032169398662661D-04, &
   0.000000000000000D+00,   2.525497092025625D-04,   0.000000000000000D+00, &
   1.236249262799109D-04,   0.000000000000000D+00,  -7.505414444311641D-05, &
   0.000000000000000D+00,  -6.786673892517795D-06,   0.000000000000000D+00, &
  -1.598546479444323D-06,   0.000000000000000D+00,  -7.970051918426977D-07, &
   4.242629859562863D-05,   0.000000000000000D+00,  -2.490338759421234D-05, &
   0.000000000000000D+00,  -9.357907684213193D-05,   0.000000000000000D+00, &
   1.801862839340043D-04,   0.000000000000000D+00,   1.815684650774082D-04, &
   0.000000000000000D+00,   7.235591562974224D-06,   0.000000000000000D+00, &
   2.283857276114199D-06,   0.000000000000000D+00,   4.398227592847342D-07, &
   0.000000000000000D+00,  -5.344238314889999D-05,   0.000000000000000D+00, &
  -4.318730555389897D-05,   0.000000000000000D+00,  -1.329741775603131D-04, &
   0.000000000000000D+00,  -3.998324983011005D-04,   0.000000000000000D+00, &
  -1.182709712860696D-04,   0.000000000000000D+00,   4.370817558500269D-05, &
   0.000000000000000D+00,   3.484770849200537D-06,   0.000000000000000D+00, &
  -2.501287163921293D-07,   0.000000000000000D+00,  -8.629257919197803D-07, &
  -9.822364754227937D-06,   0.000000000000000D+00,   1.221220910626693D-05, &
   0.000000000000000D+00,   1.211377181300459D-04,   0.000000000000000D+00, &
   1.941285965492779D-04,   0.000000000000000D+00,  -2.119415864416367D-04, &
   0.000000000000000D+00,  -1.383779380916542D-04,   0.000000000000000D+00, &
  -1.718316973190151D-05,   0.000000000000000D+00,  -3.331083602579456D-06, &
   0.000000000000000D+00,  -4.615257703730436D-07,   0.000000000000000D+00, &
   1.756419111790384D-05,   0.000000000000000D+00,   5.448177955209335D-06, &
   0.000000000000000D+00,  -1.077957034315127D-05,   0.000000000000000D+00, &
   1.415639255945486D-04,   0.000000000000000D+00,   4.408343221883796D-04, &
   0.000000000000000D+00,   1.063216201810778D-04,   0.000000000000000D+00, &
   1.828034506408980D-06,   0.000000000000000D+00,   5.282358908539528D-06, &
   0.000000000000000D+00,   3.014062173880095D-06,   0.000000000000000D+00, &
   2.289117555010530D-06,  -1.500127292193003D-06,   0.000000000000000D+00, &
   1.564721883734063D-06,   0.000000000000000D+00,  -2.393232139153818D-05, &
   0.000000000000000D+00,  -2.035857734677348D-04,   0.000000000000000D+00, &
  -2.273237063374402D-04,   0.000000000000000D+00,   1.478552583046517D-04, &
   0.000000000000000D+00,   6.605901024301085D-05,   0.000000000000000D+00, &
   8.406193606266798D-06,   0.000000000000000D+00,  -7.890373462729547D-07, &
   0.000000000000000D+00,  -1.769853392256878D-06,   0.000000000000000D+00, &
  -2.249896623162722D-06,   0.000000000000000D+00,   8.895358740381842D-07, &
   0.000000000000000D+00,   1.256732477510695D-05,   0.000000000000000D+00, &
   8.040003899217336D-05,   0.000000000000000D+00,  -1.334741022682386D-04, &
   0.000000000000000D+00,  -3.291448407404200D-04,   0.000000000000000D+00, &
  -7.970094218768506D-05,   0.000000000000000D+00,  -1.978687024213040D-05, &
   0.000000000000000D+00,  -7.151824296983374D-06,   0.000000000000000D+00, &
  -1.520581058025159D-06,   0.000000000000000D+00,   6.987047132072353D-07 /)
lem(:,250) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.727750488371213D-05,   2.854096994562599D-04, &
   1.655899917281754D-04,  -6.393522331592311D-04,  -7.601246855187153D-04, &
  -3.153204794305543D-05,   1.531051683825308D-03,   2.382905477360945D-03, &
   7.537610793873762D-05,  -5.516663096962954D-03,  -5.060668970466414D-03, &
   8.494799385780710D-03,   8.475142556238126D-03,  -1.087609130418435D-02, &
  -4.850550551577378D-03,   9.741410910322695D-03,  -1.563904825153295D-03, &
  -4.643466707286407D-03,   3.395660789193124D-03,   5.506606938262868D-04, &
  -1.744683212311534D-03,   0.000000000000000D+00,   3.727750488371599D-05, &
  -2.854096994562476D-04,  -1.655899917281771D-04,   6.393522331592195D-04, &
   7.601246855187180D-04,   3.153204794305980D-05,  -1.531051683825309D-03, &
  -2.382905477360938D-03,  -7.537610793874113D-05,   5.516663096962950D-03, &
   5.060668970466429D-03,  -8.494799385780698D-03,  -8.475142556238151D-03, &
   1.087609130418435D-02,   4.850550551577381D-03,  -9.741410910322724D-03, &
   1.563904825153305D-03,   4.643466707286413D-03,  -3.395660789193123D-03, &
  -5.506606938262785D-04,   1.744683212311528D-03,   0.000000000000000D+00, &
   1.863875244186248D-05,   0.000000000000000D+00,   9.339137393785246D-05, &
   0.000000000000000D+00,   7.876896634748522D-05,   0.000000000000000D+00, &
  -1.544856553112459D-05,   0.000000000000000D+00,  -2.602073794718939D-05, &
   0.000000000000000D+00,  -2.445723832161944D-05,   0.000000000000000D+00, &
  -3.250090382939770D-05,   0.000000000000000D+00,   1.078909898069225D-06, &
   0.000000000000000D+00,   6.362029949156743D-05,   0.000000000000000D+00, &
   4.580553861293856D-05,   0.000000000000000D+00,  -6.184314672217862D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.473522761507311D-06, &
   0.000000000000000D+00,   3.944545605603609D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.667917443417206D-06,   0.000000000000000D+00, &
  -8.942059682998175D-07,   0.000000000000000D+00,  -1.304671610651562D-05, &
   0.000000000000000D+00,  -1.156257235947032D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.350605800541119D-07,   0.000000000000000D+00, &
  -3.418924351093847D-06,   0.000000000000000D+00,   4.310093926392329D-07, &
   0.000000000000000D+00,   2.507854144964910D-05,   0.000000000000000D+00, &
   7.675240731376844D-06,   0.000000000000000D+00,  -6.915492643794380D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.537787544270006D-06, &
   0.000000000000000D+00,   1.231947143753683D-05,   0.000000000000000D+00, &
   1.544917230930554D-06,   0.000000000000000D+00,   1.878427934055580D-06, &
   0.000000000000000D+00,  -3.022446717816984D-05,   0.000000000000000D+00, &
  -2.764866945635656D-05,   0.000000000000000D+00,   2.663068377727483D-06, &
   0.000000000000000D+00,   1.703768536307899D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.670010634386026D-06,   0.000000000000000D+00, &
  -2.075527253173002D-05,   0.000000000000000D+00,  -1.762300970295542D-05, &
   0.000000000000000D+00,   6.756316228345471D-07,   0.000000000000000D+00, &
  -1.528656601662424D-06,   0.000000000000000D+00,   3.111933391883783D-05, &
   0.000000000000000D+00,   6.210723508397070D-05,   0.000000000000000D+00, &
   1.113143233626295D-06,   0.000000000000000D+00,  -6.667123032080213D-06, &
   0.000000000000000D+00,  -1.747128032576280D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.971333903035599D-06,   0.000000000000000D+00, &
   1.219054792504620D-05,   0.000000000000000D+00,   7.898178639000583D-05, &
   0.000000000000000D+00,   7.379833397070349D-06,   0.000000000000000D+00, &
  -1.933999716830372D-06,   0.000000000000000D+00,  -5.164054023605024D-07, &
   0.000000000000000D+00,  -3.682852130450034D-05,   0.000000000000000D+00, &
  -8.656163671874140D-05,   0.000000000000000D+00,  -9.331730369472067D-05, &
   0.000000000000000D+00,   4.255676658233859D-05,   0.000000000000000D+00, &
   7.501542171233522D-06,   0.000000000000000D+00,   8.996343594847989D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.617674056874094D-06, &
   0.000000000000000D+00,   1.369513142753737D-05,   0.000000000000000D+00, &
  -7.044139784614878D-05,   0.000000000000000D+00,  -1.392306025414869D-04, &
   0.000000000000000D+00,  -7.684273888715403D-07,   0.000000000000000D+00, &
   1.381952853415264D-06,   0.000000000000000D+00,   1.012301917328786D-06, &
   0.000000000000000D+00,   3.740000193259881D-05,   0.000000000000000D+00, &
   7.374027779329453D-05,   0.000000000000000D+00,   2.010431488649749D-04, &
   0.000000000000000D+00,   1.068732058561280D-04,   0.000000000000000D+00, &
  -6.146827829037027D-05,   0.000000000000000D+00,  -8.523667781050933D-06, &
   0.000000000000000D+00,  -1.712970740331702D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.576499916443155D-06,   0.000000000000000D+00, &
  -2.914171676441878D-05,   0.000000000000000D+00,  -7.675153363896605D-05, &
   0.000000000000000D+00,   1.509857365983602D-04,   0.000000000000000D+00, &
   1.591083911934501D-04,   0.000000000000000D+00,   1.000869396983364D-05, &
   0.000000000000000D+00,   2.469849995809344D-06,   0.000000000000000D+00, &
   1.475329747773212D-06,   0.000000000000000D+00,  -2.251518238981013D-05, &
   0.000000000000000D+00,  -3.756874274299882D-05,   0.000000000000000D+00, &
  -1.080070911692979D-04,   0.000000000000000D+00,  -3.344969659852710D-04, &
   0.000000000000000D+00,  -1.058164222268487D-04,   0.000000000000000D+00, &
   3.560670187369491D-05,   0.000000000000000D+00,   3.089567080076807D-06, &
   0.000000000000000D+00,  -7.165542442550829D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.057559894040674D-05,   0.000000000000000D+00, &
   2.084917994497706D-05,   0.000000000000000D+00,   1.016378966750809D-04, &
   0.000000000000000D+00,   1.626717578866875D-04,   0.000000000000000D+00, &
  -1.813856345678557D-04,   0.000000000000000D+00,  -1.232054110064129D-04, &
   0.000000000000000D+00,  -1.702918359258072D-05,   0.000000000000000D+00, &
  -3.971348161460608D-06,   0.000000000000000D+00,  -1.564593815851678D-06, &
   0.000000000000000D+00,   5.558105019486322D-06,   0.000000000000000D+00, &
   1.202864995002898D-05,   0.000000000000000D+00,  -3.292278552367135D-06, &
   0.000000000000000D+00,   1.208785461997040D-04,   0.000000000000000D+00, &
   3.784175209372886D-04,   0.000000000000000D+00,   9.659672784063251D-05, &
   0.000000000000000D+00,   4.318192392331468D-06,   0.000000000000000D+00, &
   6.277882088274952D-06,   0.000000000000000D+00,   3.696658806148235D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.438580573824009D-06, &
   0.000000000000000D+00,  -7.222257752835420D-06,   0.000000000000000D+00, &
  -2.221138866874117D-05,   0.000000000000000D+00,  -1.726479584958697D-04, &
   0.000000000000000D+00,  -1.942269886900826D-04,   0.000000000000000D+00, &
   1.282363664764739D-04,   0.000000000000000D+00,   5.877417971430589D-05, &
   0.000000000000000D+00,   7.756542601462762D-06,   0.000000000000000D+00, &
  -8.356283655811720D-07,   0.000000000000000D+00,  -3.726737268053251D-06, &
   0.000000000000000D+00,   6.095415105603179D-07,   0.000000000000000D+00, &
  -3.799489399865534D-06,   0.000000000000000D+00,   3.719797435424997D-06, &
   0.000000000000000D+00,   6.346631523324063D-05,   0.000000000000000D+00, &
  -1.173402751315254D-04,   0.000000000000000D+00,  -2.871478004009014D-04, &
   0.000000000000000D+00,  -7.266791186583122D-05,   0.000000000000000D+00, &
  -1.982975446513090D-05,   0.000000000000000D+00,  -7.350033920337306D-06, &
   0.000000000000000D+00,  -1.188011830721333D-06,   0.000000000000000D+00 /)
lem(:,251) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.171131519216280D-04,  -1.590905409677209D-04,   1.053369080134939D-04, &
   1.443094869873736D-04,   4.232727791782635D-04,   5.652803031655260D-04, &
  -5.484186756913856D-05,  -1.549072363381057D-03,  -2.285047741226910D-03, &
  -7.479848535931274D-05,   4.865892446930311D-03,   4.168410525611549D-03, &
  -6.497996377555422D-03,  -5.442285669275054D-03,   7.737172734505927D-03, &
   1.142578403500102D-03,  -6.061523819545971D-03,   3.292091009041387D-03, &
   1.384861246288000D-03,  -3.005684841038439D-03,   1.332077077523483D-03, &
   7.876023124133118D-04,  -1.171131519216341D-04,  -1.590905409677194D-04, &
   1.053369080134907D-04,   1.443094869873784D-04,   4.232727791782651D-04, &
   5.652803031655240D-04,  -5.484186756913788D-05,  -1.549072363381054D-03, &
  -2.285047741226914D-03,  -7.479848535931208D-05,   4.865892446930310D-03, &
   4.168410525611554D-03,  -6.497996377555428D-03,  -5.442285669275061D-03, &
   7.737172734505941D-03,   1.142578403500097D-03,  -6.061523819545972D-03, &
   3.292091009041397D-03,   1.384861246287996D-03,  -3.005684841038438D-03, &
   1.332077077523486D-03,   7.876023124133205D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.274723273014850D-04,   0.000000000000000D+00, &
  -9.183154191549803D-05,   0.000000000000000D+00,  -4.215819594414485D-05, &
   0.000000000000000D+00,  -1.787555563632152D-05,   0.000000000000000D+00, &
   4.155982701950546D-05,   0.000000000000000D+00,   6.649183423115579D-05, &
   0.000000000000000D+00,   4.792405250441660D-05,   0.000000000000000D+00, &
  -6.788260790325924D-05,   0.000000000000000D+00,  -1.187698644684998D-04, &
   0.000000000000000D+00,   5.070818090418334D-05,   0.000000000000000D+00, &
   4.656244291194703D-06,   1.104283535386522D-05,   0.000000000000000D+00, &
  -2.948697108882384D-06,   0.000000000000000D+00,   4.864112984790838D-06, &
  -2.348966241918460D-05,   0.000000000000000D+00,  -1.648726533100699D-06, &
   0.000000000000000D+00,  -1.899148506563503D-06,   0.000000000000000D+00, &
  -6.234692438568276D-06,   0.000000000000000D+00,  -3.293939115057447D-06, &
   2.854539917865655D-05,   0.000000000000000D+00,   6.605595525840017D-06, &
   0.000000000000000D+00,  -4.455443986007150D-06,   0.000000000000000D+00, &
   1.654545274525839D-05,   0.000000000000000D+00,   1.433700925147396D-05, &
   0.000000000000000D+00,   6.183114204489696D-06,   0.000000000000000D+00, &
  -1.935863218990556D-06,  -1.386292400708617D-05,   0.000000000000000D+00, &
  -1.254475633139206D-05,   0.000000000000000D+00,   9.052484606345401D-06, &
   0.000000000000000D+00,   4.242411299612529D-06,   0.000000000000000D+00, &
  -2.853740542093732D-05,   0.000000000000000D+00,  -2.351144805455855D-05, &
   0.000000000000000D+00,  -2.155506095350630D-05,   0.000000000000000D+00, &
   2.919795739225533D-06,   0.000000000000000D+00,   1.643245902081270D-06, &
  -1.549562017766975D-05,   0.000000000000000D+00,   7.651438223117667D-06, &
   0.000000000000000D+00,  -1.279568776237766D-05,   0.000000000000000D+00, &
  -1.829365795301037D-05,   0.000000000000000D+00,   8.409130232809211D-07, &
   0.000000000000000D+00,   2.128841777068574D-05,   0.000000000000000D+00, &
   2.880974012223605D-05,   0.000000000000000D+00,   5.015390938027009D-05, &
   0.000000000000000D+00,   3.035825373699615D-06,   0.000000000000000D+00, &
  -9.566556391863500D-06,   0.000000000000000D+00,   3.992245413785356D-07, &
   4.051603870480506D-05,   0.000000000000000D+00,   8.375026813910210D-06, &
   0.000000000000000D+00,   7.256034637345693D-06,   0.000000000000000D+00, &
   6.526974204569154D-05,   0.000000000000000D+00,   1.197162020149809D-05, &
   0.000000000000000D+00,  -4.014377619822075D-06,   0.000000000000000D+00, &
  -6.156309815272161D-06,   0.000000000000000D+00,  -2.908011272539711D-05, &
   0.000000000000000D+00,  -7.083932515537062D-05,   0.000000000000000D+00, &
  -8.029276775973195D-05,   0.000000000000000D+00,   3.474549507112610D-05, &
   0.000000000000000D+00,   1.117269538626056D-05,   0.000000000000000D+00, &
   1.147990403813625D-06,  -4.390310486350425D-05,   0.000000000000000D+00, &
  -1.605487528084676D-05,   0.000000000000000D+00,   8.297182942004439D-06, &
   0.000000000000000D+00,  -5.685990454145788D-05,   0.000000000000000D+00, &
  -1.223973133559024D-04,   0.000000000000000D+00,  -6.199926517862620D-06, &
   0.000000000000000D+00,   1.863489238603883D-06,   0.000000000000000D+00, &
   2.003563492483021D-06,   0.000000000000000D+00,   2.151250595388388D-05, &
   0.000000000000000D+00,   5.933613208453420D-05,   0.000000000000000D+00, &
   1.687830405024037D-04,   0.000000000000000D+00,   9.625939889622011D-05, &
   0.000000000000000D+00,  -5.256487073727487D-05,   0.000000000000000D+00, &
  -1.037889507616424D-05,   0.000000000000000D+00,  -3.775167115471762D-06, &
   2.704464675647275D-05,   0.000000000000000D+00,   8.694900341441175D-06, &
   0.000000000000000D+00,  -1.857099738704503D-05,   0.000000000000000D+00, &
  -6.197434178259917D-05,   0.000000000000000D+00,   1.284719113879243D-04, &
   0.000000000000000D+00,   1.438609038767604D-04,   0.000000000000000D+00, &
   1.401267507411623D-05,   0.000000000000000D+00,   4.197762553449259D-06, &
   0.000000000000000D+00,  -4.154866444954186D-06,   0.000000000000000D+00, &
  -8.168995819305680D-06,   0.000000000000000D+00,  -2.680485727774145D-05, &
   0.000000000000000D+00,  -9.142154656976807D-05,   0.000000000000000D+00, &
  -2.888577196739639D-04,   0.000000000000000D+00,  -9.816427589679462D-05, &
   0.000000000000000D+00,   2.912154447631317D-05,   0.000000000000000D+00, &
   2.861522410940299D-06,   0.000000000000000D+00,  -6.353698711943508D-07, &
  -8.653336524050898D-06,   0.000000000000000D+00,  -1.194336419201727D-06, &
   0.000000000000000D+00,   1.470217643801394D-05,   0.000000000000000D+00, &
   8.284567026545772D-05,   0.000000000000000D+00,   1.378130624497005D-04, &
   0.000000000000000D+00,  -1.581815656723966D-04,   0.000000000000000D+00, &
  -1.129511153117156D-04,   0.000000000000000D+00,  -1.901377504716518D-05, &
   0.000000000000000D+00,  -5.177003652814152D-06,   0.000000000000000D+00, &
   2.974888630436758D-06,   0.000000000000000D+00,  -6.239724600842188D-07, &
   0.000000000000000D+00,   3.784414252226515D-06,   0.000000000000000D+00, &
  -3.066089814007063D-06,   0.000000000000000D+00,   1.063216201810778D-04, &
   0.000000000000000D+00,   3.334971674583149D-04,   0.000000000000000D+00, &
   9.109704055820317D-05,   0.000000000000000D+00,   7.318371257444545D-06, &
   0.000000000000000D+00,   7.390271930474232D-06,   0.000000000000000D+00, &
   8.074880236134254D-06,   8.045605130954156D-07,   0.000000000000000D+00, &
   3.729777218278155D-07,   0.000000000000000D+00,  -6.310270697300318D-06, &
   0.000000000000000D+00,  -2.007231506870128D-05,   0.000000000000000D+00, &
  -1.478045425977890D-04,   0.000000000000000D+00,  -1.683221137887486D-04, &
   0.000000000000000D+00,   1.130119176191189D-04,   0.000000000000000D+00, &
   5.414740356639941D-05,   0.000000000000000D+00,   7.962597437621406D-06, &
   0.000000000000000D+00,  -1.891703051147778D-06,   0.000000000000000D+00, &
  -6.420138609133115D-07,   0.000000000000000D+00,   1.450170900396455D-06, &
   0.000000000000000D+00,   2.108107176348652D-06,   0.000000000000000D+00, &
   5.509705378853024D-06,   0.000000000000000D+00,   5.323260327685528D-05, &
   0.000000000000000D+00,  -1.060060019582623D-04,   0.000000000000000D+00, &
  -2.568033371733419D-04,   0.000000000000000D+00,  -6.903443240191183D-05, &
   0.000000000000000D+00,  -2.032824517978607D-05,   0.000000000000000D+00, &
  -8.107165224194627D-06,   0.000000000000000D+00,  -4.277561160953880D-06 /)
lem(:,252) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.266110133003947D-04,   2.686800247277972D-04, &
  -1.536004613529017D-04,  -6.813157346445992D-04,  -8.389428376124613D-04, &
  -1.777367889588203D-04,   1.217619156396308D-03,   2.092158978043325D-03, &
   6.541638835576578D-04,  -3.251726727551734D-03,  -3.398032970713841D-03, &
   3.844684655494238D-03,   3.396144928957942D-03,  -4.449362869673815D-03, &
   2.584758204740255D-04,   2.677912006478384D-03,  -2.767434019513882D-03, &
   7.023287969285580D-04,   1.340255860116522D-03,  -1.666370274790044D-03, &
   5.184464791974446D-04,   0.000000000000000D+00,  -3.266110133003913D-04, &
  -2.686800247277957D-04,   1.536004613528974D-04,   6.813157346445975D-04, &
   8.389428376124643D-04,   1.777367889588176D-04,  -1.217619156396309D-03, &
  -2.092158978043318D-03,  -6.541638835576551D-04,   3.251726727551736D-03, &
   3.398032970713844D-03,  -3.844684655494242D-03,  -3.396144928957937D-03, &
   4.449362869673823D-03,  -2.584758204740260D-04,  -2.677912006478387D-03, &
   2.767434019513887D-03,  -7.023287969285625D-04,  -1.340255860116522D-03, &
   1.666370274790047D-03,  -5.184464791974482D-04,   0.000000000000000D+00, &
  -1.633055066501957D-04,   0.000000000000000D+00,  -2.458295403187362D-04, &
   0.000000000000000D+00,   9.495866677884910D-05,   0.000000000000000D+00, &
   1.847619564641879D-04,   0.000000000000000D+00,   4.631669789208862D-05, &
   0.000000000000000D+00,  -9.532425352321505D-05,   0.000000000000000D+00, &
  -1.686897643731036D-04,   0.000000000000000D+00,   2.660918767523557D-06, &
   0.000000000000000D+00,   2.317164512394880D-04,   0.000000000000000D+00, &
   2.809161284726717D-05,   0.000000000000000D+00,  -1.774692877464355D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.291043388655999D-05, &
   0.000000000000000D+00,   2.415861062875754D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.443115440042624D-06,   0.000000000000000D+00, &
  -2.315614581786451D-06,   0.000000000000000D+00,  -3.460404898445454D-06, &
   0.000000000000000D+00,  -8.299363992635330D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.937917155005299D-06,   0.000000000000000D+00, &
   5.273919698912036D-06,   0.000000000000000D+00,  -1.032063198836697D-05, &
   0.000000000000000D+00,   5.382310492317756D-06,   0.000000000000000D+00, &
   1.394705947286475D-05,   0.000000000000000D+00,   7.585027615089838D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.154373621802900D-06, &
   0.000000000000000D+00,  -8.142608059482856D-06,   0.000000000000000D+00, &
   8.964762870593276D-06,   0.000000000000000D+00,   7.797471344084744D-06, &
   0.000000000000000D+00,  -7.243712977303180D-06,   0.000000000000000D+00, &
  -1.737798595869266D-05,   0.000000000000000D+00,  -2.117896831576186D-05, &
   0.000000000000000D+00,   3.367862484588420D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.464737136987723D-06,   0.000000000000000D+00, &
   4.503230546167890D-06,   0.000000000000000D+00,  -1.100532852145090D-05, &
   0.000000000000000D+00,  -1.974522369611858D-05,   0.000000000000000D+00, &
   2.837164422956440D-06,   0.000000000000000D+00,   7.108556574079045D-06, &
   0.000000000000000D+00,   1.900057579026964D-05,   0.000000000000000D+00, &
   4.244061300994342D-05,   0.000000000000000D+00,   5.910208229451751D-06, &
   0.000000000000000D+00,  -1.341094035385173D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.017471747436565D-05,   0.000000000000000D+00, &
   4.673266208813706D-06,   0.000000000000000D+00,   6.137757375244304D-06, &
   0.000000000000000D+00,   5.788436421529511D-05,   0.000000000000000D+00, &
   1.704998491137098D-05,   0.000000000000000D+00,  -8.454295522488836D-06, &
   0.000000000000000D+00,  -7.651443967479317D-06,   0.000000000000000D+00, &
  -2.116350932137319D-05,   0.000000000000000D+00,  -5.711199831171848D-05, &
   0.000000000000000D+00,  -7.200405318426007D-05,   0.000000000000000D+00, &
   2.792076819872264D-05,   0.000000000000000D+00,   1.721798391309308D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.215437395834039D-05, &
   0.000000000000000D+00,  -9.080130308931080D-06,   0.000000000000000D+00, &
   7.076994903198347D-06,   0.000000000000000D+00,  -4.853173148260313D-05, &
   0.000000000000000D+00,  -1.126154981060877D-04,   0.000000000000000D+00, &
  -1.367262189834554D-05,   0.000000000000000D+00,   3.706088495917185D-06, &
   0.000000000000000D+00,   9.517604540089325D-06,   0.000000000000000D+00, &
   2.006709486919097D-05,   0.000000000000000D+00,   4.903998561187980D-05, &
   0.000000000000000D+00,   1.446656268055321D-04,   0.000000000000000D+00, &
   9.064795466190599D-05,   0.000000000000000D+00,  -4.508953332719159D-05, &
   0.000000000000000D+00,  -1.506058590288592D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.716113024449652D-06,   0.000000000000000D+00, &
   5.422479852182794D-06,   0.000000000000000D+00,  -1.581396251290346D-05, &
   0.000000000000000D+00,  -5.231710116300110D-05,   0.000000000000000D+00, &
   1.127727838228724D-04,   0.000000000000000D+00,   1.364921341797712D-04, &
   0.000000000000000D+00,   2.072914322002014D-05,   0.000000000000000D+00, &
   9.227473933086708D-06,   0.000000000000000D+00,  -7.866933166071558D-06, &
   0.000000000000000D+00,  -1.257755086256498D-05,   0.000000000000000D+00, &
  -2.475135183080399D-05,   0.000000000000000D+00,  -7.994497350209884D-05, &
   0.000000000000000D+00,  -2.556627380570721D-04,   0.000000000000000D+00, &
  -9.617576364475661D-05,   0.000000000000000D+00,   2.372468647778984D-05, &
   0.000000000000000D+00,   2.870470377872249D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.629971581224267D-07,   0.000000000000000D+00, &
  -1.521903412390190D-06,   0.000000000000000D+00,   1.152966729676214D-05, &
   0.000000000000000D+00,   7.096435142243197D-05,   0.000000000000000D+00, &
   1.199951378303614D-04,   0.000000000000000D+00,  -1.416911035642557D-04, &
   0.000000000000000D+00,  -1.093347998697084D-04,   0.000000000000000D+00, &
  -2.270020348544091D-05,   0.000000000000000D+00,  -1.313908543988176D-05, &
   0.000000000000000D+00,   3.113379817031362D-06,   0.000000000000000D+00, &
   4.807877398628847D-06,   0.000000000000000D+00,   5.846874698570475D-06, &
   0.000000000000000D+00,   1.697152648277293D-07,   0.000000000000000D+00, &
   9.659672784063251D-05,   0.000000000000000D+00,   3.019167348295940D-04, &
   0.000000000000000D+00,   9.047594370235094D-05,   0.000000000000000D+00, &
   1.066541566386660D-05,   0.000000000000000D+00,   1.039401727109985D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.058895595343981D-06, &
   0.000000000000000D+00,   6.993648268904404D-07,   0.000000000000000D+00, &
  -2.946445837712679D-06,   0.000000000000000D+00,  -1.767342317502922D-05, &
   0.000000000000000D+00,  -1.303272515453218D-04,   0.000000000000000D+00, &
  -1.497633927187970D-04,   0.000000000000000D+00,   1.027726172881071D-04, &
   0.000000000000000D+00,   5.245914519737911D-05,   0.000000000000000D+00, &
   8.620929203523179D-06,   0.000000000000000D+00,  -2.209746210003732D-06, &
   0.000000000000000D+00,  -2.757493241745044D-07,   0.000000000000000D+00, &
  -1.520495196855275D-06,   0.000000000000000D+00,   2.216009561166740D-07, &
   0.000000000000000D+00,   4.165661062701285D-06,   0.000000000000000D+00, &
   4.364070134567229D-05,   0.000000000000000D+00,  -9.886617697098886D-05, &
   0.000000000000000D+00,  -2.364543333594610D-04,   0.000000000000000D+00, &
  -6.817518664528140D-05,   0.000000000000000D+00,  -2.222346737011970D-05, &
   0.000000000000000D+00,  -8.717654167073194D-06,   0.000000000000000D+00 /)
lem(:,253) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.259186503217917D-04,  -1.790658500618580D-04,  -1.594222992427175D-05, &
   1.660308540453100D-04,   5.296191011565833D-04,   7.291652091528369D-04, &
   3.360006623284897D-04,  -6.798283608130348D-04,  -1.680618832363974D-03, &
  -1.205227731572926D-03,   1.636390735973023D-03,   2.882299188353317D-03, &
  -1.567297917793418D-03,  -2.237708677200072D-03,   2.121117536093115D-03, &
  -3.372044080482774D-04,  -8.394591440223850D-04,   1.396175359142639D-03, &
  -1.140096224338246D-03,   5.414659909056549D-05,   8.877318070761592D-04, &
  -9.159365557115590D-04,  -2.259186503218026D-04,  -1.790658500618585D-04, &
  -1.594222992427380D-05,   1.660308540453010D-04,   5.296191011565884D-04, &
   7.291652091528435D-04,   3.360006623284863D-04,  -6.798283608130390D-04, &
  -1.680618832363972D-03,  -1.205227731572930D-03,   1.636390735973027D-03, &
   2.882299188353319D-03,  -1.567297917793418D-03,  -2.237708677200081D-03, &
   2.121117536093120D-03,  -3.372044080482796D-04,  -8.394591440223849D-04, &
   1.396175359142649D-03,  -1.140096224338253D-03,   5.414659909055774D-05, &
   8.877318070761708D-04,  -9.159365557115541D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.007506077468731D-04,   0.000000000000000D+00, &
   3.479364329460513D-04,   0.000000000000000D+00,   2.279923989263479D-04, &
   0.000000000000000D+00,  -4.093951311803401D-04,   0.000000000000000D+00, &
  -2.662140366101959D-04,   0.000000000000000D+00,   3.518231963464998D-04, &
   0.000000000000000D+00,   3.250914132433966D-04,   0.000000000000000D+00, &
  -3.369581874910330D-04,   0.000000000000000D+00,  -2.727339288221702D-04, &
   0.000000000000000D+00,   2.662024093236509D-04,   0.000000000000000D+00, &
   1.100335572991829D-05,   7.481927902403087D-06,   0.000000000000000D+00, &
  -5.409188363289174D-06,   0.000000000000000D+00,   3.638242756650482D-07, &
  -1.027975280035413D-05,   0.000000000000000D+00,  -5.259692908004478D-06, &
   0.000000000000000D+00,  -2.375890389187158D-06,   0.000000000000000D+00, &
  -1.406506253004953D-06,   0.000000000000000D+00,  -1.565918857161386D-05, &
   5.583043543280505D-06,   0.000000000000000D+00,   7.288153159952722D-06, &
   0.000000000000000D+00,   5.036816162872035D-06,   0.000000000000000D+00, &
   8.814051915120619D-06,   0.000000000000000D+00,   4.946200842053047D-06, &
   0.000000000000000D+00,   1.157793492771812D-05,   0.000000000000000D+00, &
   1.576780341270431D-05,   5.911837300413248D-06,   0.000000000000000D+00, &
  -3.930314540584695D-06,   0.000000000000000D+00,  -7.339942190403786D-06, &
   0.000000000000000D+00,   6.816150018845667D-06,   0.000000000000000D+00, &
  -3.594569363423258D-06,   0.000000000000000D+00,  -7.801652196999713D-06, &
   0.000000000000000D+00,  -1.423293910335406D-05,   0.000000000000000D+00, &
  -2.071085374288215D-05,   0.000000000000000D+00,   7.725085972709376D-06, &
  -1.364442534505104D-05,   0.000000000000000D+00,  -4.186765650264331D-06, &
   0.000000000000000D+00,   3.942967856716197D-06,   0.000000000000000D+00, &
  -8.247078084733821D-06,   0.000000000000000D+00,  -2.430707338057382D-05, &
   0.000000000000000D+00,  -9.330510552078064D-06,   0.000000000000000D+00, &
   5.435617065507261D-06,   0.000000000000000D+00,   1.648178254002429D-05, &
   0.000000000000000D+00,   3.715043158846705D-05,   0.000000000000000D+00, &
   1.008520226971179D-05,   0.000000000000000D+00,  -3.004665106305447D-05, &
   1.302009508975328D-05,   0.000000000000000D+00,   1.080936328802085D-05, &
   0.000000000000000D+00,   4.850261067695400D-06,   0.000000000000000D+00, &
   4.576568785312908D-06,   0.000000000000000D+00,   5.342709978647242D-05, &
   0.000000000000000D+00,   2.875726274368142D-05,   0.000000000000000D+00, &
   1.591234219066087D-05,   0.000000000000000D+00,  -1.215937843774778D-06, &
   0.000000000000000D+00,  -1.706254574857973D-05,   0.000000000000000D+00, &
  -4.902652579407962D-05,   0.000000000000000D+00,  -6.868408465981141D-05, &
   0.000000000000000D+00,   2.206557633069135D-05,   0.000000000000000D+00, &
   3.855927712565527D-05,  -9.668457623716237D-06,   0.000000000000000D+00, &
  -1.039161279303662D-05,   0.000000000000000D+00,  -8.988955705252805D-06, &
   0.000000000000000D+00,   4.688992400504394D-06,   0.000000000000000D+00, &
  -4.157274217652753D-05,   0.000000000000000D+00,  -1.109231568857014D-04, &
   0.000000000000000D+00,  -2.767352572558118D-05,   0.000000000000000D+00, &
  -1.102898925933108D-05,   0.000000000000000D+00,   2.003709226524959D-07, &
   0.000000000000000D+00,   1.443557398837055D-05,   0.000000000000000D+00, &
   4.220800440264299D-05,   0.000000000000000D+00,   1.290757108828395D-04, &
   0.000000000000000D+00,   9.154909674363392D-05,   0.000000000000000D+00, &
  -4.062084568560877D-05,   0.000000000000000D+00,  -3.478381353946023D-05, &
   6.289550549819865D-06,   0.000000000000000D+00,   4.434258917952984D-06, &
   0.000000000000000D+00,   4.314873578165982D-06,   0.000000000000000D+00, &
  -1.206395159726409D-05,   0.000000000000000D+00,  -4.495738013380897D-05, &
   0.000000000000000D+00,   1.021311275934576D-04,   0.000000000000000D+00, &
   1.387865510566221D-04,   0.000000000000000D+00,   3.395741668124653D-05, &
   0.000000000000000D+00,   3.244688952129152D-06,   0.000000000000000D+00, &
  -2.911536494324118D-07,   0.000000000000000D+00,  -8.255409978652355D-06, &
   0.000000000000000D+00,  -2.122865354391061D-05,   0.000000000000000D+00, &
  -7.218069777993009D-05,   0.000000000000000D+00,  -2.356339431931522D-04, &
   0.000000000000000D+00,  -9.986485094852350D-05,   0.000000000000000D+00, &
   1.953540246714421D-05,   0.000000000000000D+00,   9.842830448432575D-06, &
  -2.636935014954742D-06,   0.000000000000000D+00,   9.097418470056828D-08, &
   0.000000000000000D+00,   9.772757481590671D-07,   0.000000000000000D+00, &
   1.034050005752572D-05,   0.000000000000000D+00,   6.173662527589239D-05, &
   0.000000000000000D+00,   1.075681723197740D-04,   0.000000000000000D+00, &
  -1.321795139069115D-04,   0.000000000000000D+00,  -1.127233584175407D-04, &
   0.000000000000000D+00,  -3.020858659984367D-05,   0.000000000000000D+00, &
   3.913466009280519D-08,   0.000000000000000D+00,  -4.510611699468166D-07, &
   0.000000000000000D+00,   2.692942414903026D-06,   0.000000000000000D+00, &
   5.024503611814026D-06,   0.000000000000000D+00,   1.828034506408980D-06, &
   0.000000000000000D+00,   9.109704055820317D-05,   0.000000000000000D+00, &
   2.841403911287421D-04,   0.000000000000000D+00,   9.452541204268299D-05, &
   0.000000000000000D+00,   1.522537546080586D-05,   0.000000000000000D+00, &
   2.328948437625208D-05,   4.207526992456868D-07,   0.000000000000000D+00, &
  -5.672987901574428D-07,   0.000000000000000D+00,  -1.900197023347111D-06, &
   0.000000000000000D+00,  -4.188804169018388D-06,   0.000000000000000D+00, &
  -1.717684263941916D-05,   0.000000000000000D+00,  -1.181787200049683D-04, &
   0.000000000000000D+00,  -1.366367535939890D-04,   0.000000000000000D+00, &
   9.708119130064545D-05,   0.000000000000000D+00,   5.394890881527811D-05, &
   0.000000000000000D+00,   7.362664388893385D-06,   0.000000000000000D+00, &
  -1.584778044150050D-07,   0.000000000000000D+00,   5.685833592392109D-07, &
   0.000000000000000D+00,  -4.310400674018314D-07,   0.000000000000000D+00, &
  -4.113567515544922D-07,   0.000000000000000D+00,   3.645149122902624D-06, &
   0.000000000000000D+00,   3.664484591215111D-05,   0.000000000000000D+00, &
  -9.618314482840672D-05,   0.000000000000000D+00,  -2.256092064059145D-04, &
   0.000000000000000D+00,  -7.100672252785514D-05,   0.000000000000000D+00, &
  -2.594616426360134D-05,   0.000000000000000D+00,  -2.710820226473692D-05 /)
lem(:,254) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.346943534614288D-05,   2.938416645769784D-04, &
   1.562008296473229D-04,  -3.366465556690712D-04,  -5.915706465809004D-04, &
  -5.907937486335814D-04,  -3.239216084633777D-05,   1.011752049748130D-03, &
   1.422409885114364D-03,  -3.084821370582758D-04,  -2.236627547655164D-03, &
   1.996618172750542D-04,   1.595828665672563D-03,  -8.107413644683344D-04, &
   7.119242650302149D-05,   1.743464784748592D-04,  -4.525176543753730D-04, &
   6.749126510488932D-04,  -5.018214902940955D-04,  -6.256543356944548D-05, &
   5.273602747128979D-04,   0.000000000000000D+00,  -2.346943534613818D-05, &
  -2.938416645769814D-04,  -1.562008296473288D-04,   3.366465556690807D-04, &
   5.915706465809054D-04,   5.907937486335745D-04,   3.239216084633636D-05, &
  -1.011752049748130D-03,  -1.422409885114363D-03,   3.084821370582791D-04, &
   2.236627547655161D-03,  -1.996618172750537D-04,  -1.595828665672560D-03, &
   8.107413644683321D-04,  -7.119242650302614D-05,  -1.743464784748564D-04, &
   4.525176543753768D-04,  -6.749126510488947D-04,   5.018214902940952D-04, &
   6.256543356944979D-05,  -5.273602747128994D-04,   0.000000000000000D+00, &
  -1.173471767306745D-05,   0.000000000000000D+00,   3.136548761160612D-04, &
   0.000000000000000D+00,   4.964483781987226D-05,   0.000000000000000D+00, &
  -8.551345260875043D-04,   0.000000000000000D+00,   2.258349546255620D-04, &
   0.000000000000000D+00,   1.331163571199924D-03,   0.000000000000000D+00, &
  -1.463826639838616D-03,   0.000000000000000D+00,  -1.018214779668051D-05, &
   0.000000000000000D+00,   8.888490960049011D-04,   0.000000000000000D+00, &
  -2.313772649202183D-04,   0.000000000000000D+00,  -3.489667908422632D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.277108886483404D-07, &
   0.000000000000000D+00,   1.553712491212549D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.937678059176388D-07,   0.000000000000000D+00, &
  -1.398597929891694D-05,   0.000000000000000D+00,  -3.045845639587246D-06, &
   0.000000000000000D+00,  -9.791655862415653D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.956684512202077D-07,   0.000000000000000D+00, &
   6.009853796535539D-06,   0.000000000000000D+00,   1.491703446999449D-05, &
   0.000000000000000D+00,   3.231547061743286D-06,   0.000000000000000D+00, &
   3.136155241346879D-06,   0.000000000000000D+00,   7.785756300830923D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.167009984495274D-06, &
   0.000000000000000D+00,  -1.536498703459090D-06,   0.000000000000000D+00, &
  -8.123001764921341D-06,   0.000000000000000D+00,   1.963522084423208D-05, &
   0.000000000000000D+00,  -1.428889074893431D-06,   0.000000000000000D+00, &
  -5.034695349928813D-06,   0.000000000000000D+00,  -1.113751839126321D-05, &
   0.000000000000000D+00,  -2.347164263254870D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.521124279803820D-06,   0.000000000000000D+00, &
  -3.185140417885023D-06,   0.000000000000000D+00,   3.201064223635787D-06, &
   0.000000000000000D+00,  -8.025590429428634D-06,   0.000000000000000D+00, &
  -5.981664679365698D-05,   0.000000000000000D+00,  -1.366481872088513D-06, &
   0.000000000000000D+00,   4.219805007630945D-06,   0.000000000000000D+00, &
   1.274982242527812D-05,   0.000000000000000D+00,   3.651262591746097D-05, &
   0.000000000000000D+00,   2.124187972113179D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.433926047774003D-06,   0.000000000000000D+00, &
   7.398483850310372D-06,   0.000000000000000D+00,   4.371059773374256D-06, &
   0.000000000000000D+00,   2.934689306473595D-06,   0.000000000000000D+00, &
   5.553292911719418D-05,   0.000000000000000D+00,   6.551257975941347D-05, &
   0.000000000000000D+00,   2.616521748276330D-06,   0.000000000000000D+00, &
  -2.428771586521507D-06,   0.000000000000000D+00,  -1.397871707361348D-05, &
   0.000000000000000D+00,  -4.376007965930402D-05,   0.000000000000000D+00, &
  -7.195711342923363D-05,   0.000000000000000D+00,   1.280909088306866D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.243515460891264D-06, &
   0.000000000000000D+00,  -8.739018904391984D-06,   0.000000000000000D+00, &
  -7.980289262914372D-06,   0.000000000000000D+00,   4.145797725124652D-06, &
   0.000000000000000D+00,  -3.787868184592714D-05,   0.000000000000000D+00, &
  -1.196141065231320D-04,   0.000000000000000D+00,  -6.000202963193547D-05, &
   0.000000000000000D+00,  -8.643524791175222D-07,   0.000000000000000D+00, &
   1.975547493503456D-06,   0.000000000000000D+00,   1.296356315851790D-05, &
   0.000000000000000D+00,   3.760643035360819D-05,   0.000000000000000D+00, &
   1.214483748404275D-04,   0.000000000000000D+00,   9.936167698424896D-05, &
   0.000000000000000D+00,  -3.613728631376689D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.504571134361482D-06,   0.000000000000000D+00, &
   5.994104281369971D-06,   0.000000000000000D+00,   4.740229118280038D-06, &
   0.000000000000000D+00,  -1.006703542384805D-05,   0.000000000000000D+00, &
  -4.051566482866816D-05,   0.000000000000000D+00,   9.818689031579365D-05, &
   0.000000000000000D+00,   1.536470804819249D-04,   0.000000000000000D+00, &
   7.587041797722263D-05,   0.000000000000000D+00,  -1.092686179318778D-06, &
   0.000000000000000D+00,  -1.492831581660526D-06,   0.000000000000000D+00, &
  -7.692612820747820D-06,   0.000000000000000D+00,  -1.993458721756003D-05, &
   0.000000000000000D+00,  -6.877454177865064D-05,   0.000000000000000D+00, &
  -2.277401053755338D-04,   0.000000000000000D+00,  -1.110752169386838D-04, &
   0.000000000000000D+00,   1.052767734417701D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.490935648418050D-06,   0.000000000000000D+00, &
  -2.450566030207751D-06,   0.000000000000000D+00,  -6.574940901215429D-07, &
   0.000000000000000D+00,   8.660889050841522D-06,   0.000000000000000D+00, &
   5.634922082604448D-05,   0.000000000000000D+00,   9.965063603376556D-05, &
   0.000000000000000D+00,  -1.300464494477475D-04,   0.000000000000000D+00, &
  -1.260917854185485D-04,   0.000000000000000D+00,  -7.757973945462446D-05, &
   0.000000000000000D+00,   9.590624626985826D-07,   0.000000000000000D+00, &
   4.697603497266871D-07,   0.000000000000000D+00,   1.948336969013476D-06, &
   0.000000000000000D+00,   4.983358205106899D-06,   0.000000000000000D+00, &
   4.318192392331468D-06,   0.000000000000000D+00,   9.047594370235094D-05, &
   0.000000000000000D+00,   2.799915006918462D-04,   0.000000000000000D+00, &
   1.059155181251776D-04,   0.000000000000000D+00,   2.884305272901526D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.716699970050111D-08, &
   0.000000000000000D+00,   8.427712383116169D-07,   0.000000000000000D+00, &
  -3.338238196532878D-07,   0.000000000000000D+00,  -3.355201996887104D-06, &
   0.000000000000000D+00,  -1.708687360300404D-05,   0.000000000000000D+00, &
  -1.110396775294163D-04,   0.000000000000000D+00,  -1.293879425691747D-04, &
   0.000000000000000D+00,   9.647325355584610D-05,   0.000000000000000D+00, &
   6.113000640373311D-05,   0.000000000000000D+00,   3.029583951940378D-05, &
   0.000000000000000D+00,  -1.838814399160541D-07,   0.000000000000000D+00, &
  -8.315172794979937D-08,   0.000000000000000D+00,   4.501809086849846D-07, &
   0.000000000000000D+00,   2.176800683616010D-07,   0.000000000000000D+00, &
   3.464469554686875D-06,   0.000000000000000D+00,   3.014715196130160D-05, &
   0.000000000000000D+00,  -9.759051595904120D-05,   0.000000000000000D+00, &
  -2.256601368777985D-04,   0.000000000000000D+00,  -7.971524306434967D-05, &
   0.000000000000000D+00,  -3.126077143674168D-05,   0.000000000000000D+00 /)
lem(:,255) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -8.960231674972779D-05,  -3.019090338040091D-04,  -2.612444630954037D-04, &
   5.485314697470765D-05,   1.444747692901812D-05,   2.200783402743987D-04, &
   7.061551817475408D-04,   5.041612787836761D-04,  -3.403770416372419D-04, &
  -1.295460609708184D-03,  -6.249273897261325D-04,   1.477457893563655D-03, &
   4.997309221527915D-04,  -1.078245319071810D-03,   2.348628216060099D-04, &
   1.410511222601645D-04,  -4.290615722202998D-05,   6.996483820687986D-05, &
  -2.308086927107570D-04,   3.316275620395610D-04,  -2.288700952127732D-04, &
  -5.602610068409430D-05,  -8.960231674973573D-05,  -3.019090338040061D-04, &
  -2.612444630954042D-04,   5.485314697469567D-05,   1.444747692902409D-05, &
   2.200783402744124D-04,   7.061551817475348D-04,   5.041612787836671D-04, &
  -3.403770416372405D-04,  -1.295460609708184D-03,  -6.249273897261298D-04, &
   1.477457893563651D-03,   4.997309221527874D-04,  -1.078245319071807D-03, &
   2.348628216060073D-04,   1.410511222601627D-04,  -4.290615722202402D-05, &
   6.996483820688096D-05,  -2.308086927107618D-04,   3.316275620395579D-04, &
  -2.288700952127678D-04,  -5.602610068409103D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.008398684417483D-04,   0.000000000000000D+00, &
   8.274767386595761D-06,   0.000000000000000D+00,  -7.549236522066601D-04, &
   0.000000000000000D+00,   7.008050586432084D-04,   0.000000000000000D+00, &
   1.568308745250299D-03,   0.000000000000000D+00,  -3.987017595672918D-03, &
   0.000000000000000D+00,   4.130920697096041D-03,   0.000000000000000D+00, &
  -1.918382877745401D-03,   0.000000000000000D+00,  -5.180690714849957D-04, &
   0.000000000000000D+00,   8.955008014834865D-04,   0.000000000000000D+00, &
   7.133979301948402D-07,   3.972129216269722D-06,   0.000000000000000D+00, &
   4.550706455003570D-06,   0.000000000000000D+00,  -5.795257260101034D-06, &
  -2.667281720016274D-06,   0.000000000000000D+00,  -2.456582046425760D-06, &
   0.000000000000000D+00,  -6.603503917957010D-06,   0.000000000000000D+00, &
   3.647611069485093D-07,   0.000000000000000D+00,   2.640351456269439D-06, &
  -1.659719774266918D-06,   0.000000000000000D+00,   7.205169113168067D-07, &
   0.000000000000000D+00,   2.161936436059707D-06,   0.000000000000000D+00, &
   1.956160101549159D-06,   0.000000000000000D+00,   2.098639956999245D-06, &
   0.000000000000000D+00,   1.888563392227241D-06,   0.000000000000000D+00, &
   2.906822882765880D-05,   3.518816999173640D-06,   0.000000000000000D+00, &
   2.791896892528657D-06,   0.000000000000000D+00,  -6.534802057321898D-07, &
   0.000000000000000D+00,  -6.621686742311859D-06,   0.000000000000000D+00, &
   4.955058656778154D-06,   0.000000000000000D+00,  -9.668185752763948D-07, &
   0.000000000000000D+00,  -4.039009294632970D-06,   0.000000000000000D+00, &
  -1.535913025669427D-05,   0.000000000000000D+00,  -6.807301281213363D-05, &
   2.694878867329592D-09,   0.000000000000000D+00,  -4.264550698415521D-06, &
   0.000000000000000D+00,  -3.310040802136255D-06,   0.000000000000000D+00, &
   3.922552855701597D-06,   0.000000000000000D+00,   3.587237946032855D-07, &
   0.000000000000000D+00,  -8.222631001698584D-06,   0.000000000000000D+00, &
  -2.508963049122620D-06,   0.000000000000000D+00,   2.622318048324806D-06, &
   0.000000000000000D+00,   1.268990140051789D-05,   0.000000000000000D+00, &
   4.251096049027573D-05,   0.000000000000000D+00,   4.489962558753013D-05, &
  -3.854813862573230D-06,   0.000000000000000D+00,   3.277458658322216D-06, &
   0.000000000000000D+00,   7.013030765891129D-06,   0.000000000000000D+00, &
   3.525118946399681D-06,   0.000000000000000D+00,   2.659663060761089D-07, &
   0.000000000000000D+00,   5.298219717500372D-05,   0.000000000000000D+00, &
   7.524671901603998D-06,   0.000000000000000D+00,   3.976536078772789D-06, &
   0.000000000000000D+00,  -6.997406811012245D-07,   0.000000000000000D+00, &
  -1.208174871206213D-05,   0.000000000000000D+00,  -4.257670354636689D-05, &
   0.000000000000000D+00,  -8.021696174304484D-05,   0.000000000000000D+00, &
   4.911876503336971D-05,   3.190891550191664D-06,   0.000000000000000D+00, &
  -2.017993613179798D-06,   0.000000000000000D+00,  -7.277003678068085D-06, &
   0.000000000000000D+00,  -7.342009783728861D-06,   0.000000000000000D+00, &
   3.141361068189126D-06,   0.000000000000000D+00,  -3.691705615772343D-05, &
   0.000000000000000D+00,  -1.448812605711903D-04,   0.000000000000000D+00, &
  -4.770165616118815D-06,   0.000000000000000D+00,  -2.030813792563988D-06, &
   0.000000000000000D+00,   1.059433156608613D-06,   0.000000000000000D+00, &
   1.076480609259697D-05,   0.000000000000000D+00,   3.621592290411080D-05, &
   0.000000000000000D+00,   1.217705597131962D-04,   0.000000000000000D+00, &
   1.151034386818255D-04,   0.000000000000000D+00,  -1.097397673106215D-04, &
  -5.457810346198074D-07,   0.000000000000000D+00,   8.305541777434235D-07, &
   0.000000000000000D+00,   4.094056491153388D-06,   0.000000000000000D+00, &
   4.316163223912096D-06,   0.000000000000000D+00,  -8.614830962401707D-06, &
   0.000000000000000D+00,  -3.748305755490186D-05,   0.000000000000000D+00, &
   1.024775950369332D-04,   0.000000000000000D+00,   1.987925034922540D-04, &
   0.000000000000000D+00,   1.769620009344265D-06,   0.000000000000000D+00, &
   1.576469582759871D-10,   0.000000000000000D+00,  -1.780835701083723D-06, &
   0.000000000000000D+00,  -6.812761638199791D-06,   0.000000000000000D+00, &
  -1.975694475447415D-05,   0.000000000000000D+00,  -6.934025407019054D-05, &
   0.000000000000000D+00,  -2.350450985050900D-04,   0.000000000000000D+00, &
  -1.329701152282114D-04,   0.000000000000000D+00,   5.307583432855719D-05, &
  -4.032331814776471D-07,   0.000000000000000D+00,   2.019484532163208D-07, &
   0.000000000000000D+00,  -9.723412670880292D-07,   0.000000000000000D+00, &
  -3.435595031822973D-07,   0.000000000000000D+00,   8.572259471746749D-06, &
   0.000000000000000D+00,   5.310941712986958D-05,   0.000000000000000D+00, &
   9.723925216335021D-05,   0.000000000000000D+00,  -1.387940498962089D-04, &
   0.000000000000000D+00,  -1.571723857683240D-04,   0.000000000000000D+00, &
  -2.082673623735184D-07,   0.000000000000000D+00,   1.986142873503335D-07, &
   0.000000000000000D+00,   1.263317756172308D-06,   0.000000000000000D+00, &
   2.455321178974458D-06,   0.000000000000000D+00,   5.282358908539528D-06, &
   0.000000000000000D+00,   7.318371257444545D-06,   0.000000000000000D+00, &
   9.452541204268299D-05,   0.000000000000000D+00,   2.946732379019233D-04, &
   0.000000000000000D+00,   1.255279601615982D-04,   0.000000000000000D+00, &
   5.734922491333556D-05,   1.411520815506827D-07,   0.000000000000000D+00, &
  -3.465707071589597D-07,   0.000000000000000D+00,  -1.127927850294588D-07, &
   0.000000000000000D+00,  -4.408062200661905D-07,   0.000000000000000D+00, &
  -3.771535906944601D-06,   0.000000000000000D+00,  -1.803317654784666D-05, &
   0.000000000000000D+00,  -1.091420100661669D-04,   0.000000000000000D+00, &
  -1.293679986183278D-04,   0.000000000000000D+00,   1.042463611771479D-04, &
   0.000000000000000D+00,   6.060016794723726D-05,   0.000000000000000D+00, &
  -3.619321971465199D-08,   0.000000000000000D+00,   8.701697454002154D-08, &
   0.000000000000000D+00,  -5.255708968553366D-07,   0.000000000000000D+00, &
  -4.320157372040117D-07,   0.000000000000000D+00,  -1.049485608857222D-07, &
   0.000000000000000D+00,   2.805237790183865D-06,   0.000000000000000D+00, &
   2.529296676654063D-05,   0.000000000000000D+00,  -1.044738651946238D-04, &
   0.000000000000000D+00,  -2.407929943200917D-04,   0.000000000000000D+00, &
  -9.343664376767805D-05,   0.000000000000000D+00,  -9.763519935865877D-05 /)
lem(:,256) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   9.580877087506947D-06,   1.412911555372504D-04, &
   5.209961499988049D-04,   4.788854795006237D-04,  -2.657722729105922D-04, &
  -5.202539488618533D-04,  -5.200312780028876D-04,  -3.568413244316251D-04, &
   8.163000053271349D-04,   1.074716494929592D-03,  -7.823418817853367D-04, &
  -7.824602365272270D-04,   6.734520193604302D-04,   2.369990551276532D-05, &
  -1.874682072231190D-04,   6.483030244396329D-05,   5.806150315204216D-06, &
   3.863179085343456D-05,  -1.195477946817101D-04,   1.580182165278479D-04, &
  -1.004996308204829D-04,   0.000000000000000D+00,  -9.580877087505975D-06, &
  -1.412911555372492D-04,  -5.209961499988107D-04,  -4.788854795006187D-04, &
   2.657722729106009D-04,   5.202539488618471D-04,   5.200312780028770D-04, &
   3.568413244316280D-04,  -8.163000053271325D-04,  -1.074716494929590D-03, &
   7.823418817853351D-04,   7.824602365272260D-04,  -6.734520193604268D-04, &
  -2.369990551276887D-05,   1.874682072231129D-04,  -6.483030244395969D-05, &
  -5.806150315202604D-06,  -3.863179085343737D-05,   1.195477946817083D-04, &
  -1.580182165278422D-04,   1.004996308204846D-04,   0.000000000000000D+00, &
  -4.790438543754831D-06,   0.000000000000000D+00,   5.196250848065800D-05, &
   0.000000000000000D+00,  -3.269992951156264D-04,   0.000000000000000D+00, &
   4.591618962009505D-04,   0.000000000000000D+00,   1.274479039875717D-03, &
   0.000000000000000D+00,  -4.315779903975792D-03,   0.000000000000000D+00, &
   5.810264628507311D-03,   0.000000000000000D+00,  -5.825644392669009D-03, &
   0.000000000000000D+00,   4.575366663407179D-03,   0.000000000000000D+00, &
  -1.647599463259163D-03,   0.000000000000000D+00,  -3.976330045186157D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.787174197328887D-07, &
   0.000000000000000D+00,   6.673285481880150D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.297395096223227D-06,   0.000000000000000D+00, &
  -4.412272478678561D-06,   0.000000000000000D+00,  -4.766632487057150D-07, &
   0.000000000000000D+00,  -3.040966350118237D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.846139320748673D-06,   0.000000000000000D+00, &
   4.919874479026264D-07,   0.000000000000000D+00,   1.946877404641066D-05, &
   0.000000000000000D+00,  -4.810185474243505D-08,   0.000000000000000D+00, &
   2.919017985982324D-06,   0.000000000000000D+00,   4.237536280993383D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.855845713422518D-06, &
   0.000000000000000D+00,   1.744529490526621D-06,   0.000000000000000D+00, &
  -6.025162857614552D-06,   0.000000000000000D+00,  -3.332913058303437D-05, &
   0.000000000000000D+00,   9.307657524632789D-07,   0.000000000000000D+00, &
  -5.730970160811662D-07,   0.000000000000000D+00,  -4.154162290471595D-06, &
   0.000000000000000D+00,  -5.813038998031571D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.279506513696729D-06,   0.000000000000000D+00, &
  -2.930379375338365D-06,   0.000000000000000D+00,  -9.357932502578143D-07, &
   0.000000000000000D+00,   8.149628622007446D-06,   0.000000000000000D+00, &
  -1.671857842976333D-05,   0.000000000000000D+00,  -2.311288870766782D-06, &
   0.000000000000000D+00,  -2.321049370874294D-06,   0.000000000000000D+00, &
   1.883472012827750D-06,   0.000000000000000D+00,   9.002825544332542D-06, &
   0.000000000000000D+00,   3.333571943556245D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.627106140893852D-07,   0.000000000000000D+00, &
   3.304440293305896D-06,   0.000000000000000D+00,   5.463854821436069D-06, &
   0.000000000000000D+00,   4.050832712304216D-06,   0.000000000000000D+00, &
   6.930303954483806D-06,   0.000000000000000D+00,   1.823987935868071D-04, &
   0.000000000000000D+00,   3.159057485302791D-06,   0.000000000000000D+00, &
   3.462658531620759D-06,   0.000000000000000D+00,  -9.642195223544744D-08, &
   0.000000000000000D+00,  -1.115288743586457D-05,   0.000000000000000D+00, &
  -4.317140282013910D-05,   0.000000000000000D+00,  -1.074301567017706D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -6.344285081967431D-07, &
   0.000000000000000D+00,  -3.007207885278287D-06,   0.000000000000000D+00, &
  -6.529947514682244D-06,   0.000000000000000D+00,  -7.860420211038035D-06, &
   0.000000000000000D+00,   1.018701268717363D-07,   0.000000000000000D+00, &
  -5.365326071353536D-05,   0.000000000000000D+00,  -3.988019020829497D-04, &
   0.000000000000000D+00,  -2.270175244758515D-06,   0.000000000000000D+00, &
  -2.341610144373988D-06,   0.000000000000000D+00,   2.572961290598971D-07, &
   0.000000000000000D+00,   1.081016202728983D-05,   0.000000000000000D+00, &
   3.762022095529544D-05,   0.000000000000000D+00,   1.375394327857082D-04, &
   0.000000000000000D+00,   1.964533892832111D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.176078429022701D-06,   0.000000000000000D+00, &
   1.886348501040277D-06,   0.000000000000000D+00,   4.208473171491968D-06, &
   0.000000000000000D+00,   4.445600595149978D-06,   0.000000000000000D+00, &
  -7.039446530166921D-06,   0.000000000000000D+00,  -3.702618111530187D-05, &
   0.000000000000000D+00,   1.235021584733392D-04,   0.000000000000000D+00, &
   5.176552930243696D-04,   0.000000000000000D+00,   6.065617152424367D-07, &
   0.000000000000000D+00,   8.401894752308251D-07,   0.000000000000000D+00, &
  -8.131976543326982D-07,   0.000000000000000D+00,  -6.903085092516467D-06, &
   0.000000000000000D+00,  -2.120688543818597D-05,   0.000000000000000D+00, &
  -7.534943546337623D-05,   0.000000000000000D+00,  -2.699074676953315D-04, &
   0.000000000000000D+00,  -2.444297536187434D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.915897341924751D-07,   0.000000000000000D+00, &
  -6.355954706594202D-07,   0.000000000000000D+00,  -1.177994083665852D-06, &
   0.000000000000000D+00,  -2.050353769483022D-07,   0.000000000000000D+00, &
   8.078201219808588D-06,   0.000000000000000D+00,   5.352724990859746D-05, &
   0.000000000000000D+00,   1.032570962822610D-04,   0.000000000000000D+00, &
  -1.622283091163491D-04,   0.000000000000000D+00,  -4.450887696793047D-04, &
   0.000000000000000D+00,   1.307353362503985D-07,   0.000000000000000D+00, &
  -2.784686344137440D-07,   0.000000000000000D+00,   4.664006668749683D-07, &
   0.000000000000000D+00,   2.206152622135807D-06,   0.000000000000000D+00, &
   6.277882088274952D-06,   0.000000000000000D+00,   1.066541566386660D-05, &
   0.000000000000000D+00,   1.059155181251776D-04,   0.000000000000000D+00, &
   3.418229942089438D-04,   0.000000000000000D+00,   2.189808504919438D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.784957933393629D-08, &
   0.000000000000000D+00,   1.351832000490542D-07,   0.000000000000000D+00, &
  -1.708815301438210D-07,   0.000000000000000D+00,  -7.994558924681393D-07, &
   0.000000000000000D+00,  -3.977207603033184D-06,   0.000000000000000D+00, &
  -1.984394075055498D-05,   0.000000000000000D+00,  -1.145206898691573D-04, &
   0.000000000000000D+00,  -1.402998250870352D-04,   0.000000000000000D+00, &
   1.297789296424235D-04,   0.000000000000000D+00,   2.297498895215344D-04, &
   0.000000000000000D+00,  -7.067527557556711D-08,   0.000000000000000D+00, &
   1.402611795802449D-07,   0.000000000000000D+00,   1.775965722217278D-08, &
   0.000000000000000D+00,  -9.056885498254973D-08,   0.000000000000000D+00, &
  -4.100922506188099D-07,   0.000000000000000D+00,   2.794650108933255D-06, &
   0.000000000000000D+00,   2.135731251530711D-05,   0.000000000000000D+00, &
  -1.194401536636237D-04,   0.000000000000000D+00,  -2.848216065684504D-04, &
   0.000000000000000D+00,  -1.310503867167950D-04,   0.000000000000000D+00 /)
lem(:,257) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.394540591152921D-05,  -1.321826821433429D-04,  -1.866484999576725D-04, &
  -9.477362388508538D-05,   6.694053253776195D-06,  -3.625698543195719D-04, &
  -3.587149872992695D-04,   7.636545334159067D-04,   9.391274607024978D-04, &
  -3.970370807480503D-04,  -1.131984693842424D-03,   1.915070416749457D-04, &
   7.996280555243166D-04,  -3.814970846021432D-04,  -1.365082328366315D-04, &
   1.785018429640101D-04,  -6.682374327438416D-05,   3.085376087133469D-06, &
  -4.532775455091117D-06,   3.346857581045707D-05,  -6.078270470332086D-05, &
   7.060840675663854D-05,  -4.394540591152938D-05,  -1.321826821433401D-04, &
  -1.866484999576720D-04,  -9.477362388508852D-05,   6.694053253776184D-06, &
  -3.625698543195654D-04,  -3.587149872992696D-04,   7.636545334158987D-04, &
   9.391274607024967D-04,  -3.970370807480458D-04,  -1.131984693842419D-03, &
   1.915070416749406D-04,   7.996280555243133D-04,  -3.814970846021380D-04, &
  -1.365082328366340D-04,   1.785018429640095D-04,  -6.682374327437733D-05, &
   3.085376087132875D-06,  -4.532775455097334D-06,   3.346857581045655D-05, &
  -6.078270470331619D-05,   7.060840675663923D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.153397188059042D-05,   0.000000000000000D+00, &
  -2.455737417958404D-05,   0.000000000000000D+00,   4.706979914314547D-06, &
   0.000000000000000D+00,   7.523525933618484D-04,   0.000000000000000D+00, &
  -2.666740100537530D-03,   0.000000000000000D+00,   2.475909232219228D-03, &
   0.000000000000000D+00,   9.322666447999501D-04,   0.000000000000000D+00, &
  -1.268068366414194D-03,   0.000000000000000D+00,  -2.177577776561889D-03, &
   0.000000000000000D+00,   2.887712591578258D-03,   0.000000000000000D+00, &
   6.011038115532376D-07,   1.556704742739801D-06,   0.000000000000000D+00, &
   2.514168773934400D-06,   0.000000000000000D+00,  -1.016109338252965D-06, &
   1.503352333430585D-06,   0.000000000000000D+00,  -1.049729824754196D-06, &
   0.000000000000000D+00,  -2.243511154644223D-06,   0.000000000000000D+00, &
  -7.701801372455798D-07,   0.000000000000000D+00,   1.643098948396848D-06, &
  -3.249585158764701D-06,   0.000000000000000D+00,  -6.809388566856068D-08, &
   0.000000000000000D+00,   3.227961907158944D-08,   0.000000000000000D+00, &
   9.095315491810670D-08,   0.000000000000000D+00,  -1.990129553524768D-09, &
   0.000000000000000D+00,  -5.588947163043682D-07,   0.000000000000000D+00, &
   4.397905789888834D-06,   4.693318291949331D-07,   0.000000000000000D+00, &
   7.697567054255956D-07,   0.000000000000000D+00,   1.681138248390284D-06, &
   0.000000000000000D+00,   4.553235401547791D-06,   0.000000000000000D+00, &
   1.862253143990891D-07,   0.000000000000000D+00,   2.227209029577145D-06, &
   0.000000000000000D+00,   1.502572513162069D-06,   0.000000000000000D+00, &
  -3.417037623428010D-06,   0.000000000000000D+00,  -4.803752585565729D-05, &
   3.134835150226061D-06,   0.000000000000000D+00,   1.386961410166760D-08, &
   0.000000000000000D+00,  -2.925690387467039D-06,   0.000000000000000D+00, &
  -7.100054019893413D-06,   0.000000000000000D+00,  -5.755918757993512D-06, &
   0.000000000000000D+00,   6.870236248981195D-07,   0.000000000000000D+00, &
  -3.047977859300876D-06,   0.000000000000000D+00,  -3.241389536715152D-06, &
   0.000000000000000D+00,   2.434950622542183D-06,   0.000000000000000D+00, &
   2.428868511917115D-05,   0.000000000000000D+00,   1.650302938793897D-04, &
  -4.147136647323197D-06,   0.000000000000000D+00,  -7.208674248084534D-07, &
   0.000000000000000D+00,   2.537919130178118D-06,   0.000000000000000D+00, &
   7.998816676818831D-06,   0.000000000000000D+00,   8.697012826207119D-06, &
   0.000000000000000D+00,  -7.484177833051179D-06,   0.000000000000000D+00, &
  -4.450754073874750D-07,   0.000000000000000D+00,   2.224521303017490D-06, &
   0.000000000000000D+00,   3.734441122482583D-06,   0.000000000000000D+00, &
  -7.890006509359904D-07,   0.000000000000000D+00,  -1.714145405379966D-05, &
   0.000000000000000D+00,  -7.496843227807915D-05,   0.000000000000000D+00, &
  -3.151586807952425D-04,   2.832078829671312D-06,   0.000000000000000D+00, &
   3.078525092577053D-07,   0.000000000000000D+00,  -1.989883080139868D-06, &
   0.000000000000000D+00,  -6.779060749722217D-06,   0.000000000000000D+00, &
  -9.002618010120568D-06,   0.000000000000000D+00,   7.456437637109816D-06, &
   0.000000000000000D+00,   6.458958833614565D-06,   0.000000000000000D+00, &
  -3.874719007214490D-07,   0.000000000000000D+00,  -9.390969521505893D-07, &
   0.000000000000000D+00,  -2.328767603356499D-06,   0.000000000000000D+00, &
   7.083835317861039D-07,   0.000000000000000D+00,   1.232101673452541D-05, &
   0.000000000000000D+00,   4.576248102849166D-05,   0.000000000000000D+00, &
   1.748797330349027D-04,   0.000000000000000D+00,   4.087963866425604D-04, &
  -9.958706835377638D-07,   0.000000000000000D+00,   2.633120537393027D-07, &
   0.000000000000000D+00,   1.614843450355483D-06,   0.000000000000000D+00, &
   3.811360636980585D-06,   0.000000000000000D+00,   5.122752464641401D-06, &
   0.000000000000000D+00,  -8.449711003632670D-06,   0.000000000000000D+00, &
  -5.544068160334510D-05,   0.000000000000000D+00,   4.196911902257469D-05, &
   0.000000000000000D+00,   4.516231661316888D-07,   0.000000000000000D+00, &
  -5.999449873543499D-09,   0.000000000000000D+00,   5.514791827886822D-07, &
   0.000000000000000D+00,  -1.187950701750647D-06,   0.000000000000000D+00, &
  -7.645845313737331D-06,   0.000000000000000D+00,  -2.403986152790932D-05, &
   0.000000000000000D+00,  -8.645864300312188D-05,   0.000000000000000D+00, &
  -3.113511444415528D-04,   0.000000000000000D+00,  -4.446871746125545D-04, &
   6.147901701524751D-08,   0.000000000000000D+00,  -1.582233741234562D-07, &
   0.000000000000000D+00,  -8.646335729646066D-07,   0.000000000000000D+00, &
  -9.366473920608391D-07,   0.000000000000000D+00,  -5.816427055562743D-07, &
   0.000000000000000D+00,   8.536090945585502D-06,   0.000000000000000D+00, &
   5.950997473991586D-05,   0.000000000000000D+00,   1.264683909397834D-04, &
   0.000000000000000D+00,  -7.814934046932618D-05,   0.000000000000000D+00, &
  -1.074381185280568D-07,   0.000000000000000D+00,   1.971269274408078D-07, &
   0.000000000000000D+00,   1.255047262705803D-07,   0.000000000000000D+00, &
   6.012566642687948D-07,   0.000000000000000D+00,   3.014062173880095D-06, &
   0.000000000000000D+00,   7.390271930474232D-06,   0.000000000000000D+00, &
   1.522537546080586D-05,   0.000000000000000D+00,   1.255279601615982D-04, &
   0.000000000000000D+00,   3.891603144573532D-04,   0.000000000000000D+00, &
   4.567683650986554D-04,   3.975438674675684D-08,   0.000000000000000D+00, &
  -6.123717625004956D-08,   0.000000000000000D+00,   2.621519626664873D-07, &
   0.000000000000000D+00,  -1.155953064648076D-07,   0.000000000000000D+00, &
  -7.859729571789726D-07,   0.000000000000000D+00,  -4.366190110179097D-06, &
   0.000000000000000D+00,  -2.390955495877763D-05,   0.000000000000000D+00, &
  -1.299857629856502D-04,   0.000000000000000D+00,  -1.660059530526962D-04, &
   0.000000000000000D+00,   3.817458379518014D-05,   0.000000000000000D+00, &
  -7.765715551410966D-09,   0.000000000000000D+00,  -3.328801651212232D-08, &
   0.000000000000000D+00,  -1.028730528411170D-07,   0.000000000000000D+00, &
   7.382175302162306D-08,   0.000000000000000D+00,  -5.228603000625393D-07, &
   0.000000000000000D+00,  -5.643655401015740D-07,   0.000000000000000D+00, &
   3.000756266817583D-06,   0.000000000000000D+00,   1.793498805395955D-05, &
   0.000000000000000D+00,  -1.456610156143652D-04,   0.000000000000000D+00, &
  -3.332104621389328D-04,   0.000000000000000D+00,  -3.918773457928461D-04 /)
lem(:,258) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.456492323019380D-05,   1.388401865126172D-04, &
   3.859953919335351D-04,   6.268511662820535D-04,   4.441666296663061D-04, &
  -2.977902642769358D-04,  -4.748666089071653D-04,  -3.071247717774360D-04, &
  -3.847648993282499D-04,   7.382488902763875D-04,   4.449473867087243D-04, &
  -7.702876064337351D-04,   1.565234139382218D-04,   2.100901986023727D-04, &
  -1.653481438814362D-04,   5.178012379072432D-05,  -1.041223384969494D-05, &
   5.669337551676962D-06,  -1.381769313259437D-05,   2.634030441180136D-05, &
  -3.623788492408603D-05,   0.000000000000000D+00,  -2.456492323018073D-05, &
  -1.388401865126165D-04,  -3.859953919335410D-04,  -6.268511662820483D-04, &
  -4.441666296663007D-04,   2.977902642769313D-04,   4.748666089071605D-04, &
   3.071247717774379D-04,   3.847648993282521D-04,  -7.382488902763814D-04, &
  -4.449473867087267D-04,   7.702876064337327D-04,  -1.565234139382164D-04, &
  -2.100901986023788D-04,   1.653481438814329D-04,  -5.178012379071706D-05, &
   1.041223384969276D-05,  -5.669337551681890D-06,   1.381769313259658D-05, &
  -2.634030441179739D-05,   3.623788492408553D-05,   0.000000000000000D+00, &
  -1.228246161508855D-05,   0.000000000000000D+00,   2.551022005085561D-05, &
   0.000000000000000D+00,  -6.383618644809655D-05,   0.000000000000000D+00, &
   1.866074795940537D-04,   0.000000000000000D+00,  -6.233349333473479D-04, &
   0.000000000000000D+00,  -4.924758512898433D-04,   0.000000000000000D+00, &
   8.564738507059971D-03,   0.000000000000000D+00,  -2.043503986830637D-02, &
   0.000000000000000D+00,   2.126015578448134D-02,   0.000000000000000D+00, &
  -9.779795904545897D-03,   0.000000000000000D+00,   8.241580955341009D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.710138494319489D-07, &
   0.000000000000000D+00,   9.216134704414782D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.746115754905952D-07,   0.000000000000000D+00, &
  -2.445022357742130D-06,   0.000000000000000D+00,  -4.452083135184844D-09, &
   0.000000000000000D+00,  -2.052501060679718D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.095615719639163D-06,   0.000000000000000D+00, &
  -9.319189944016490D-07,   0.000000000000000D+00,   6.515133697140422D-06, &
   0.000000000000000D+00,  -1.545864888150005D-06,   0.000000000000000D+00, &
   9.633013412834183D-07,   0.000000000000000D+00,   4.348225593037649D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.942296202061538D-07, &
   0.000000000000000D+00,   2.191737541120207D-06,   0.000000000000000D+00, &
  -2.246729236474618D-07,   0.000000000000000D+00,  -1.822814598426464D-05, &
   0.000000000000000D+00,   2.041675320483899D-06,   0.000000000000000D+00, &
   8.979718604962199D-07,   0.000000000000000D+00,  -3.456920364094685D-06, &
   0.000000000000000D+00,  -6.472092558365839D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.787028238688755D-07,   0.000000000000000D+00, &
  -9.702093289739447D-07,   0.000000000000000D+00,  -1.774401627154302D-06, &
   0.000000000000000D+00,   5.147541015232004D-06,   0.000000000000000D+00, &
   3.743146980678164D-05,   0.000000000000000D+00,  -1.279949705242674D-06, &
   0.000000000000000D+00,  -1.950892368552321D-06,   0.000000000000000D+00, &
  -3.657330075137049D-08,   0.000000000000000D+00,   4.050972657024017D-06, &
   0.000000000000000D+00,   3.215808134812599D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.578190656877378D-07,   0.000000000000000D+00, &
  -8.639287903579482D-07,   0.000000000000000D+00,   1.327483711689442D-06, &
   0.000000000000000D+00,   1.608829854354249D-06,   0.000000000000000D+00, &
  -6.089093667652733D-06,   0.000000000000000D+00,   2.262369880455776D-06, &
   0.000000000000000D+00,   7.504327375475041D-07,   0.000000000000000D+00, &
   2.331150066465516D-06,   0.000000000000000D+00,   2.725832568037650D-06, &
   0.000000000000000D+00,   2.963457273658171D-07,   0.000000000000000D+00, &
  -5.264011783803277D-06,   0.000000000000000D+00,  -2.447340411143772D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.902216649735651D-07, &
   0.000000000000000D+00,   1.002467826584194D-06,   0.000000000000000D+00, &
  -1.092037719436197D-06,   0.000000000000000D+00,  -5.298858356000283D-06, &
   0.000000000000000D+00,  -7.791270366499046D-06,   0.000000000000000D+00, &
  -8.542696347850618D-06,   0.000000000000000D+00,  -2.526103244902797D-04, &
   0.000000000000000D+00,  -7.632386693951583D-07,   0.000000000000000D+00, &
  -1.809479694353513D-06,   0.000000000000000D+00,  -2.357736202758533D-06, &
   0.000000000000000D+00,  -6.193035056775716D-07,   0.000000000000000D+00, &
   8.237430035394004D-06,   0.000000000000000D+00,   3.594910023527131D-05, &
   0.000000000000000D+00,   1.701722244959680D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.001258431282351D-07,   0.000000000000000D+00, &
  -1.082901508472183D-07,   0.000000000000000D+00,   1.015071633493533D-06, &
   0.000000000000000D+00,   4.546205072345180D-06,   0.000000000000000D+00, &
   7.679160944955619D-06,   0.000000000000000D+00,   1.893832798537338D-06, &
   0.000000000000000D+00,   8.847461879758385D-06,   0.000000000000000D+00, &
   7.023834772159586D-04,   0.000000000000000D+00,   4.625961949487693D-07, &
   0.000000000000000D+00,   6.363544524835706D-07,   0.000000000000000D+00, &
   6.102262472667241D-07,   0.000000000000000D+00,  -1.128222687083501D-06, &
   0.000000000000000D+00,  -7.963258965795666D-06,   0.000000000000000D+00, &
  -2.847339111973922D-05,   0.000000000000000D+00,  -1.189989977198700D-04, &
   0.000000000000000D+00,  -4.510793159436025D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.174913428961993D-07,   0.000000000000000D+00, &
  -1.241397185995711D-07,   0.000000000000000D+00,  -3.690237351154326D-07, &
   0.000000000000000D+00,  -1.495063980450833D-06,   0.000000000000000D+00, &
  -1.649791601436801D-06,   0.000000000000000D+00,   7.215740763864568D-06, &
   0.000000000000000D+00,   5.847108482625844D-05,   0.000000000000000D+00, &
   3.989149360546537D-05,   0.000000000000000D+00,  -9.781347043911115D-04, &
   0.000000000000000D+00,  -6.261436081074758D-08,   0.000000000000000D+00, &
  -1.930048473687811D-08,   0.000000000000000D+00,   9.107450391179900D-08, &
   0.000000000000000D+00,   1.200938686130211D-06,   0.000000000000000D+00, &
   3.696658806148235D-06,   0.000000000000000D+00,   1.039401727109985D-05, &
   0.000000000000000D+00,   2.884305272901526D-05,   0.000000000000000D+00, &
   2.189808504919438D-04,   0.000000000000000D+00,   6.261204747238963D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.690920312096260D-08, &
   0.000000000000000D+00,  -3.633412823754385D-08,   0.000000000000000D+00, &
  -1.783904860129548D-08,   0.000000000000000D+00,  -1.345791959967847D-07, &
   0.000000000000000D+00,  -6.828989312729481D-07,   0.000000000000000D+00, &
  -5.612179673682647D-06,   0.000000000000000D+00,  -3.194636816253304D-05, &
   0.000000000000000D+00,  -1.637154897141845D-04,   0.000000000000000D+00, &
  -7.695317621150137D-05,   0.000000000000000D+00,   7.719733130104784D-04, &
   0.000000000000000D+00,  -2.076146614719427D-08,   0.000000000000000D+00, &
   9.370895390406930D-09,   0.000000000000000D+00,  -4.781878854178976D-08, &
   0.000000000000000D+00,  -3.723997872203808D-07,   0.000000000000000D+00, &
  -6.533256229321111D-07,   0.000000000000000D+00,  -7.470055113554638D-07, &
   0.000000000000000D+00,   3.030790039610256D-06,   0.000000000000000D+00, &
   8.772326448377643D-06,   0.000000000000000D+00,  -2.720643027637076D-04, &
   0.000000000000000D+00,  -4.891925431505729D-04,   0.000000000000000D+00 /)
lem(:,259) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.239274765643018D-05,   1.227758762238410D-04,   2.631836121178345D-04, &
   3.636092914731818D-04,   2.039406232570550D-04,  -4.440660316289193D-04, &
  -1.275244370915803D-03,  -8.685397125439126D-04,   1.616791555433292D-03, &
   1.823905793049122D-03,  -1.608769000745357D-03,  -1.025220559052228D-03, &
   1.185446029226710D-03,  -1.420186143679377D-04,  -2.848023075435894D-04, &
   2.159642619828938D-04,  -4.841936928652733D-05,   4.457977816983538D-06, &
  -1.660323057798937D-05,   2.318470794505790D-07,  -8.740993400193724D-06, &
   2.882541558223410D-05,   3.239274765644613D-05,   1.227758762238443D-04, &
   2.631836121178332D-04,   3.636092914731845D-04,   2.039406232570521D-04, &
  -4.440660316289143D-04,  -1.275244370915803D-03,  -8.685397125439277D-04, &
   1.616791555433290D-03,   1.823905793049139D-03,  -1.608769000745342D-03, &
  -1.025220559052241D-03,   1.185446029226703D-03,  -1.420186143679291D-04, &
  -2.848023075435950D-04,   2.159642619828883D-04,  -4.841936928650651D-05, &
   4.457977816983598D-06,  -1.660323057801058D-05,   2.318470794541390D-07, &
  -8.740993400178296D-06,   2.882541558223362D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.515526801481516D-06,   0.000000000000000D+00, &
   2.650928503488332D-05,   0.000000000000000D+00,  -7.292794754848671D-05, &
   0.000000000000000D+00,   2.311785145556011D-04,   0.000000000000000D+00, &
  -9.397485431899716D-04,   0.000000000000000D+00,   5.871917148705073D-03, &
   0.000000000000000D+00,  -2.145889700615451D-02,   0.000000000000000D+00, &
   4.058284812047366D-02,   0.000000000000000D+00,  -4.176002821276301D-02, &
   0.000000000000000D+00,   2.337757503152102D-02,   0.000000000000000D+00, &
  -4.238552517356801D-08,  -1.271565755206994D-07,   0.000000000000000D+00, &
  -2.806929534227790D-07,   0.000000000000000D+00,   7.721279236170333D-07, &
  -4.176308039830182D-07,   0.000000000000000D+00,   7.478229371540282D-07, &
   0.000000000000000D+00,  -3.061835489815674D-07,   0.000000000000000D+00, &
   5.715482705229628D-07,   0.000000000000000D+00,  -2.727975378503061D-06, &
   3.666496128653669D-07,   0.000000000000000D+00,   4.005665016396929D-08, &
   0.000000000000000D+00,  -2.430757998308766D-06,   0.000000000000000D+00, &
   1.402279158016768D-06,   0.000000000000000D+00,  -8.561791398704661D-07, &
   0.000000000000000D+00,  -1.309817527239995D-06,   0.000000000000000D+00, &
   8.470594770490771D-06,   1.051423552033578D-06,   0.000000000000000D+00, &
  -1.344819460053798D-06,   0.000000000000000D+00,   1.173281936908050D-06, &
   0.000000000000000D+00,   6.948611436984310D-06,   0.000000000000000D+00, &
  -2.365734749462589D-06,   0.000000000000000D+00,   1.002898265470456D-07, &
   0.000000000000000D+00,   3.771689824965868D-06,   0.000000000000000D+00, &
   1.831034309843713D-06,   0.000000000000000D+00,  -2.828233902522370D-05, &
  -2.244336846692952D-06,   0.000000000000000D+00,   2.494923797719321D-06, &
   0.000000000000000D+00,   1.691395758206347D-06,   0.000000000000000D+00, &
  -7.196185405967047D-06,   0.000000000000000D+00,  -1.960676206853288D-05, &
   0.000000000000000D+00,   3.045469849974317D-06,   0.000000000000000D+00, &
   1.164860936589200D-07,   0.000000000000000D+00,  -4.890327885513695D-06, &
   0.000000000000000D+00,  -8.444380371164896D-06,   0.000000000000000D+00, &
   6.029054891363018D-06,   0.000000000000000D+00,   1.116360909554543D-04, &
   1.413860018291211D-06,   0.000000000000000D+00,  -2.925278727891019D-06, &
   0.000000000000000D+00,  -3.070402667514558D-06,   0.000000000000000D+00, &
   5.879344047740801D-06,   0.000000000000000D+00,   2.549188966758989D-05, &
   0.000000000000000D+00,   3.963639284228230D-05,   0.000000000000000D+00, &
  -2.550195764612346D-06,   0.000000000000000D+00,   9.700822994280427D-07, &
   0.000000000000000D+00,   4.847269072611760D-06,   0.000000000000000D+00, &
   9.574273700939097D-06,   0.000000000000000D+00,   8.720856938024546D-07, &
   0.000000000000000D+00,  -5.658128997543570D-05,   0.000000000000000D+00, &
  -4.244206537290533D-04,   1.512033149951538D-07,   0.000000000000000D+00, &
   1.781498019477693D-06,   0.000000000000000D+00,   1.903276050083151D-06, &
   0.000000000000000D+00,  -5.212809560292718D-06,   0.000000000000000D+00, &
  -2.247920506748467D-05,   0.000000000000000D+00,  -3.962462175303654D-05, &
   0.000000000000000D+00,   3.440303400815927D-06,   0.000000000000000D+00, &
   1.010022829285178D-06,   0.000000000000000D+00,  -1.574587510882383D-06, &
   0.000000000000000D+00,  -3.207143222034068D-06,   0.000000000000000D+00, &
  -5.330612484135430D-06,   0.000000000000000D+00,   1.178787562274071D-06, &
   0.000000000000000D+00,   4.063167544690201D-05,   0.000000000000000D+00, &
   1.873675878922278D-04,   0.000000000000000D+00,   1.187305065632463D-03, &
  -4.867321738175958D-07,   0.000000000000000D+00,  -9.643799280999311D-08, &
   0.000000000000000D+00,  -5.160138886713754D-07,   0.000000000000000D+00, &
   3.583146094268207D-06,   0.000000000000000D+00,   1.253609524577096D-05, &
   0.000000000000000D+00,   1.945057958723276D-05,   0.000000000000000D+00, &
  -2.261328474474156D-05,   0.000000000000000D+00,  -2.671274758223079D-04, &
   0.000000000000000D+00,  -3.833351892257533D-08,   0.000000000000000D+00, &
   7.797383997259923D-07,   0.000000000000000D+00,   1.202110319421490D-06, &
   0.000000000000000D+00,   7.075593815392832D-07,   0.000000000000000D+00, &
  -3.546733380360204D-06,   0.000000000000000D+00,  -2.390899177510547D-05, &
   0.000000000000000D+00,  -8.753325303837436D-05,   0.000000000000000D+00, &
  -3.532327852065328D-04,   0.000000000000000D+00,  -2.245920133862107D-03, &
   1.177049730412304D-07,   0.000000000000000D+00,  -3.943828854975936D-07, &
   0.000000000000000D+00,   3.027748916335630D-07,   0.000000000000000D+00, &
  -1.015773903317888D-06,   0.000000000000000D+00,  -3.228157473252620D-06, &
   0.000000000000000D+00,  -2.031169300788143D-06,   0.000000000000000D+00, &
   2.842066084329331D-05,   0.000000000000000D+00,   2.177304086653227D-04, &
   0.000000000000000D+00,   7.370153174111758D-04,   0.000000000000000D+00, &
  -5.287177092997294D-08,   0.000000000000000D+00,  -2.602173426168478D-08, &
   0.000000000000000D+00,  -4.098152967165714D-07,   0.000000000000000D+00, &
   8.355458085723141D-07,   0.000000000000000D+00,   2.289117555010530D-06, &
   0.000000000000000D+00,   8.074880236134254D-06,   0.000000000000000D+00, &
   2.328948437625208D-05,   0.000000000000000D+00,   5.734922491333556D-05, &
   0.000000000000000D+00,   4.567683650986554D-04,   0.000000000000000D+00, &
   2.859948249627605D-03,   1.289104521792068D-08,   0.000000000000000D+00, &
   1.096118729937687D-07,   0.000000000000000D+00,  -2.382089697672896D-07, &
   0.000000000000000D+00,  -4.038259839922591D-07,   0.000000000000000D+00, &
  -8.902081664313933D-08,   0.000000000000000D+00,  -2.010483586232932D-06, &
   0.000000000000000D+00,  -1.413966995725609D-05,   0.000000000000000D+00, &
  -8.213186798411549D-05,   0.000000000000000D+00,  -4.685518769826647D-04, &
   0.000000000000000D+00,  -1.023463714836473D-03,   0.000000000000000D+00, &
  -1.940883883094095D-09,   0.000000000000000D+00,  -5.242484366604355D-08, &
   0.000000000000000D+00,   2.739214636467384D-07,   0.000000000000000D+00, &
  -5.580115795902997D-07,   0.000000000000000D+00,  -7.588510775675014D-07, &
   0.000000000000000D+00,  -9.520746844993043D-07,   0.000000000000000D+00, &
  -7.684074108400887D-07,   0.000000000000000D+00,   1.101581327267430D-05, &
   0.000000000000000D+00,   5.737805549767128D-05,   0.000000000000000D+00, &
  -4.953969989504616D-04,   0.000000000000000D+00,  -2.467623654120907D-03 /)
lem(:,260) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.169297324278317D-03,  -3.098998831776758D-03,  -9.058411611988773D-03, &
   6.407621853397221D-03,   9.884598123959538D-03,  -3.260157764906037D-03, &
  -3.955370019932633D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.169297324278273D-03,  -3.098998831776646D-03, &
  -9.058411611988823D-03,   6.407621853397094D-03,   9.884598123959693D-03, &
  -3.260157764905942D-03,  -3.955370019932776D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.529449355923201D-04,   0.000000000000000D+00,   6.922051415933808D-06, &
   0.000000000000000D+00,   3.890599318688151D-07,   0.000000000000000D+00, &
   3.326117486812377D-08,   0.000000000000000D+00,   3.482473436081629D-09, &
   0.000000000000000D+00,   4.057860464630538D-10,   0.000000000000000D+00, &
   4.941041901778782D-11,   0.000000000000000D+00,   6.965754332389217D-12, &
  -3.432138916009880D-03,   0.000000000000000D+00,  -7.194751354929285D-05, &
   0.000000000000000D+00,  -5.696791860161142D-06,   0.000000000000000D+00, &
  -6.315733028990034D-07,   0.000000000000000D+00,  -8.183999605525956D-08, &
   0.000000000000000D+00,  -1.149442368910178D-08,   0.000000000000000D+00, &
  -1.668685188604350D-09,   0.000000000000000D+00,  -2.430681696146325D-10, &
   0.000000000000000D+00,   7.669478755237995D-03,   0.000000000000000D+00, &
   4.406234770596265D-05,   0.000000000000000D+00,   2.042078621805989D-05, &
   0.000000000000000D+00,   3.808113073946486D-06,   0.000000000000000D+00, &
   6.866118655259874D-07,   0.000000000000000D+00,   1.240195714877552D-07, &
   0.000000000000000D+00,   2.227879380342745D-08,   0.000000000000000D+00, &
   3.914518556706426D-09,   0.000000000000000D+00,   8.008477200357684D-10, &
  -4.835342640220692D-03,   0.000000000000000D+00,   9.460752983366179D-04, &
   0.000000000000000D+00,   3.201356547893330D-05,   0.000000000000000D+00, &
  -4.357272907548071D-06,   0.000000000000000D+00,  -2.106148299592645D-06, &
   0.000000000000000D+00,  -5.957863669549901D-07,   0.000000000000000D+00, &
  -1.453302700322681D-07,   0.000000000000000D+00,  -3.263148181479900D-08, &
   0.000000000000000D+00,  -6.960263973810314D-09,   0.000000000000000D+00, &
  -8.817420615492080D-03,   0.000000000000000D+00,  -2.876000712547205D-03, &
   0.000000000000000D+00,  -2.205632078428272D-04,   0.000000000000000D+00, &
  -2.925195938871648D-05,   0.000000000000000D+00,  -1.500127292193003D-06, &
   0.000000000000000D+00,   8.045605130954156D-07,   0.000000000000000D+00, &
   4.207526992456868D-07,   0.000000000000000D+00,   1.411520815506827D-07, &
   0.000000000000000D+00,   3.975438674675684D-08,   0.000000000000000D+00, &
   1.289104521792068D-08,   1.804556733654862D-02,   0.000000000000000D+00, &
   2.245519276652059D-03,   0.000000000000000D+00,   3.292418136036498D-05, &
   0.000000000000000D+00,   6.203332343229487D-05,   0.000000000000000D+00, &
   1.880647376243266D-05,   0.000000000000000D+00,   3.229099118738510D-06, &
   0.000000000000000D+00,   5.780253891576613D-08,   0.000000000000000D+00, &
  -2.215153121920104D-07,   0.000000000000000D+00,  -1.157539058902820D-07, &
   0.000000000000000D+00,  -4.346073242766863D-08,   0.000000000000000D+00, &
  -9.265213178996501D-03,   0.000000000000000D+00,   3.008207849433394D-03, &
   0.000000000000000D+00,   1.005170081656508D-03,   0.000000000000000D+00, &
   8.947372184893196D-05,   0.000000000000000D+00,  -1.096292428584015D-05, &
   0.000000000000000D+00,  -9.979574044525985D-06,   0.000000000000000D+00, &
  -3.136642664976336D-06,   0.000000000000000D+00,  -5.239169510893736D-07, &
   0.000000000000000D+00,   4.495323687405276D-08,   0.000000000000000D+00, &
   7.945212939240710D-08,   0.000000000000000D+00,   5.846759402607434D-08 /)
lem(:,261) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.112570257273658D-04,   2.065648086465403D-03, &
  -1.292052873700877D-03,  -4.292909388269924D-03,   5.043374202929146D-03, &
   2.976310857129599D-03,  -4.774344544202585D-03,  -8.293657864431577D-04, &
   1.334390071371089D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.112570257274021D-04, &
  -2.065648086465433D-03,   1.292052873700856D-03,   4.292909388269948D-03, &
  -5.043374202929164D-03,  -2.976310857129620D-03,   4.774344544202614D-03, &
   8.293657864431540D-04,  -1.334390071371113D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.161507142971809D-10,   0.000000000000000D+00,   1.129216920495516D-08, &
   0.000000000000000D+00,   1.598266002258679D-07,   0.000000000000000D+00, &
   6.813280348903170D-07,   0.000000000000000D+00,   6.388450583970957D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.181742731822966D-06,   0.000000000000000D+00, &
   1.965104667209384D-07,   0.000000000000000D+00,   1.544761422901139D-08, &
   0.000000000000000D+00,   1.485347521953355D-09,   0.000000000000000D+00, &
   1.503030230718491D-10,   0.000000000000000D+00,   1.175900125595566D-11, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.425510988358650D-05, &
   0.000000000000000D+00,  -2.943290101554190D-06,   0.000000000000000D+00, &
  -3.027093577745238D-07,   0.000000000000000D+00,  -3.669068966627076D-08, &
   0.000000000000000D+00,  -4.697628694423271D-09,   0.000000000000000D+00, &
  -5.607186895976360D-10,   0.000000000000000D+00,  -1.482768802845533D-11, &
   0.000000000000000D+00,   1.998530318886273D-04,   0.000000000000000D+00, &
   1.355451328160797D-05,   0.000000000000000D+00,   2.039140173808006D-06, &
   0.000000000000000D+00,   3.340818332301639D-07,   0.000000000000000D+00, &
   5.554392913984698D-08,   0.000000000000000D+00,   8.849678772857640D-09, &
   0.000000000000000D+00,   1.135390283089498D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.850507631375258D-04,   0.000000000000000D+00, &
  -1.234020685585182D-05,   0.000000000000000D+00,  -4.670948237789472D-06, &
   0.000000000000000D+00,  -1.291637172020151D-06,   0.000000000000000D+00, &
  -3.084808471006274D-07,   0.000000000000000D+00,  -6.678315268483304D-08, &
   0.000000000000000D+00,  -1.270340492735163D-08,   0.000000000000000D+00, &
  -1.511568725786969D-09,   0.000000000000000D+00,   6.090418092914094D-04, &
   0.000000000000000D+00,  -9.896933824663139D-05,   0.000000000000000D+00, &
  -5.788313611192223D-06,   0.000000000000000D+00,   1.011366841053364D-06, &
   0.000000000000000D+00,   6.989867081898749D-07,   0.000000000000000D+00, &
   2.468617768959769D-07,   0.000000000000000D+00,   6.876623112621351D-08, &
   0.000000000000000D+00,   1.537993585431198D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.125985488013751D-04,   0.000000000000000D+00, &
   3.996783049388927D-04,   0.000000000000000D+00,   4.507036990401916D-05, &
   0.000000000000000D+00,   7.193373878584103D-06,   0.000000000000000D+00, &
   5.357056924262491D-07,   0.000000000000000D+00,  -2.655479006559587D-07, &
   0.000000000000000D+00,  -1.705076022232827D-07,   0.000000000000000D+00, &
  -6.351429496736497D-08,   0.000000000000000D+00,  -2.134323273574300D-08, &
   0.000000000000000D+00,  -8.900132121622945D-04,   0.000000000000000D+00, &
  -6.264352024782882D-04,   0.000000000000000D+00,  -4.972828278299928D-05, &
   0.000000000000000D+00,  -1.840585674008354D-05,   0.000000000000000D+00, &
  -5.438580573824009D-06,   0.000000000000000D+00,  -1.058895595343981D-06, &
   0.000000000000000D+00,  -1.716699970050111D-08,   0.000000000000000D+00, &
   9.784957933393629D-08,   0.000000000000000D+00,   5.690920312096260D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.461360369014112D-03, &
   0.000000000000000D+00,   2.284583831608360D-04,   0.000000000000000D+00, &
  -1.169339357114261D-04,   0.000000000000000D+00,  -4.827402751427659D-06, &
   0.000000000000000D+00,   5.879274662189761D-06,   0.000000000000000D+00, &
   3.363669284465252D-06,   0.000000000000000D+00,   1.071740761268238D-06, &
   0.000000000000000D+00,   1.713984697133120D-07,   0.000000000000000D+00, &
  -4.191866455488466D-08,   0.000000000000000D+00,  -8.091879090584786D-08, &
   0.000000000000000D+00,  -9.365574101270223D-04,   0.000000000000000D+00, &
   7.683990603413082D-04,   0.000000000000000D+00,   3.647045652884812D-04, &
   0.000000000000000D+00,   6.647150991188986D-05,   0.000000000000000D+00, &
   1.252923347372544D-05,   0.000000000000000D+00,  -3.113008686790378D-07, &
   0.000000000000000D+00,  -1.692239763636800D-06,   0.000000000000000D+00, &
  -8.581109408038838D-07,   0.000000000000000D+00,  -2.295373494900302D-07, &
   0.000000000000000D+00,   1.229422848062403D-08,   0.000000000000000D+00 /)
lem(:,262) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.368492323476940D-04, &
  -8.941724385532417D-04,   2.221951854390676D-04,   9.952002532857245D-04, &
   1.700884175431589D-03,   1.047734097660547D-03,  -5.535796497980130D-03, &
  -2.669486351791023D-03,   4.485062141333799D-03,   1.887454729561612D-03, &
  -3.809658194447766D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.368492323476916D-04,  -8.941724385532574D-04,   2.221951854390654D-04, &
   9.952002532857248D-04,   1.700884175431602D-03,   1.047734097660558D-03, &
  -5.535796497980133D-03,  -2.669486351791019D-03,   4.485062141333792D-03, &
   1.887454729561632D-03,  -3.809658194447933D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.670751532053100D-09, &
   0.000000000000000D+00,  -1.556730546428769D-07,   0.000000000000000D+00, &
  -1.108601070008453D-06,   0.000000000000000D+00,  -2.618342844913530D-06, &
   0.000000000000000D+00,  -1.814558794691070D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.042915196633671D-06,   0.000000000000000D+00, &
   1.586989326444745D-07,   0.000000000000000D+00,   1.086301526900967D-08, &
   0.000000000000000D+00,   9.044271928314429D-10,   0.000000000000000D+00, &
   6.284341571017151D-11,   0.000000000000000D+00,  -1.390302955118615D-11, &
  -6.323468624576151D-05,   0.000000000000000D+00,  -2.417251118270301D-06, &
   0.000000000000000D+00,  -2.208017279427432D-07,   0.000000000000000D+00, &
  -2.420444947476824D-08,   0.000000000000000D+00,  -2.598388716822263D-09, &
   0.000000000000000D+00,  -1.231281382730736D-10,   0.000000000000000D+00, &
   3.940651516863459D-05,   0.000000000000000D+00,   1.043983114359501D-05, &
   0.000000000000000D+00,   1.474987785873834D-06,   0.000000000000000D+00, &
   2.271313294494896D-07,   0.000000000000000D+00,   3.419263868760322D-08, &
   0.000000000000000D+00,   4.184329135053695D-09,   0.000000000000000D+00, &
  -5.768350254088788D-10,   8.379694141651865D-04,   0.000000000000000D+00, &
  -2.801643769178326D-06,   0.000000000000000D+00,  -2.878435371578486D-06, &
   0.000000000000000D+00,  -8.483717340239765D-07,   0.000000000000000D+00, &
  -1.958313172042737D-07,   0.000000000000000D+00,  -3.776831374016271D-08, &
   0.000000000000000D+00,  -4.291904358472736D-09,   0.000000000000000D+00, &
  -2.560474346162147D-03,   0.000000000000000D+00,   8.240562848051663D-06, &
   0.000000000000000D+00,  -4.899566823779754D-06,   0.000000000000000D+00, &
   4.296764951876112D-07,   0.000000000000000D+00,   4.203340100053728D-07, &
   0.000000000000000D+00,   1.467633274873211D-07,   0.000000000000000D+00, &
   3.509490665961318D-08,   0.000000000000000D+00,   8.108321091898298D-10, &
   2.006002195101194D-03,   0.000000000000000D+00,  -3.554909738592086D-04, &
   0.000000000000000D+00,   1.262756430702105D-05,   0.000000000000000D+00, &
   4.082243251814156D-06,   0.000000000000000D+00,   3.928288051624841D-07, &
   0.000000000000000D+00,  -1.505170576513560D-07,   0.000000000000000D+00, &
  -9.763469615411920D-08,   0.000000000000000D+00,  -2.912462340030347D-08, &
   0.000000000000000D+00,   2.701710584792058D-03,   0.000000000000000D+00, &
   1.097856152432717D-03,   0.000000000000000D+00,  -2.054484344526569D-06, &
   0.000000000000000D+00,  -4.487284770419450D-06,   0.000000000000000D+00, &
  -2.525912209173275D-06,   0.000000000000000D+00,  -5.676366728300423D-07, &
   0.000000000000000D+00,   4.254148311379264D-09,   0.000000000000000D+00, &
   6.171789961321437D-08,   0.000000000000000D+00,   4.046917264169744D-08, &
  -6.058893075932913D-03,   0.000000000000000D+00,  -8.763398539107509D-04, &
   0.000000000000000D+00,   1.443346154330549D-04,   0.000000000000000D+00, &
  -1.555254161950876D-06,   0.000000000000000D+00,   8.494460301255502D-07, &
   0.000000000000000D+00,   1.318461731788463D-06,   0.000000000000000D+00, &
   4.904245836879994D-07,   0.000000000000000D+00,   5.434652550918348D-08, &
   0.000000000000000D+00,  -4.875504517113272D-08,   0.000000000000000D+00, &
   2.854809977514229D-03,   0.000000000000000D+00,  -1.205040236634688D-03, &
   0.000000000000000D+00,  -5.579872331461355D-04,   0.000000000000000D+00, &
  -2.260510646745369D-05,   0.000000000000000D+00,   1.564721883734063D-06, &
   0.000000000000000D+00,   3.729777218278155D-07,   0.000000000000000D+00, &
  -5.672987901574428D-07,   0.000000000000000D+00,  -3.465707071589597D-07, &
   0.000000000000000D+00,  -6.123717625004956D-08,   0.000000000000000D+00, &
   1.096118729937687D-07,   2.245519276652059D-03,   0.000000000000000D+00, &
   2.706170576906811D-03,   0.000000000000000D+00,   5.132930092295205D-04, &
   0.000000000000000D+00,   2.065983041709690D-06,   0.000000000000000D+00, &
   5.730063744462857D-06,   0.000000000000000D+00,  -7.889081869957405D-07, &
   0.000000000000000D+00,  -5.800771023223820D-07,   0.000000000000000D+00, &
   1.769965314196430D-07,   0.000000000000000D+00,   2.233617284215253D-07, &
   0.000000000000000D+00,   3.444429845309630D-08,   0.000000000000000D+00, &
  -3.033936273236948D-03,   0.000000000000000D+00,  -1.272789634535695D-03, &
   0.000000000000000D+00,   6.070719419217208D-04,   0.000000000000000D+00, &
   2.182834490005913D-04,   0.000000000000000D+00,   1.384477896747589D-05, &
   0.000000000000000D+00,  -1.396475028787605D-06,   0.000000000000000D+00, &
   2.448329679690422D-07,   0.000000000000000D+00,   4.445234491410603D-07, &
   0.000000000000000D+00,  -2.198368097513856D-08,   0.000000000000000D+00, &
  -1.545000094334480D-07,   0.000000000000000D+00,  -8.141594408188327D-08 /)
lem(:,263) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.951497677715947D-05,   3.704419092323246D-04,  -5.646735695988578D-04, &
  -1.143284304231506D-03,   1.796160007563428D-03,   3.477205449611476D-03, &
  -3.865749045438841D-03,  -6.162941701406177D-03,   5.232509360689294D-03, &
   5.450044271771400D-03,  -3.043937424423276D-03,  -2.300939341037783D-03, &
  -9.305401343971010D-05,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.951497677716107D-05,  -3.704419092323232D-04, &
   5.646735695988468D-04,   1.143284304231506D-03,  -1.796160007563404D-03, &
  -3.477205449611489D-03,   3.865749045438817D-03,   6.162941701406186D-03, &
  -5.232509360689269D-03,  -5.450044271771422D-03,   3.043937424423283D-03, &
   2.300939341037821D-03,   9.305401343967011D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.349119807501207D-09, &
   0.000000000000000D+00,   9.915594089767809D-08,   0.000000000000000D+00, &
   1.117792545375140D-06,   0.000000000000000D+00,   4.055410514042104D-06, &
   0.000000000000000D+00,   4.186843670077697D-06,   0.000000000000000D+00, &
  -4.517099011779443D-06,   0.000000000000000D+00,  -9.848013072054867D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.634094604587889D-07,   0.000000000000000D+00, &
   8.730835295146190D-09,   0.000000000000000D+00,   5.574655080898666D-10, &
   0.000000000000000D+00,  -9.452676137235116D-12,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.497487026968932D-06,   0.000000000000000D+00, &
  -1.845162356885817D-07,   0.000000000000000D+00,  -1.728253246469771D-08, &
   0.000000000000000D+00,  -1.191037177083429D-09,   0.000000000000000D+00, &
   5.189125266498008D-10,   0.000000000000000D+00,   1.171545784880626D-05, &
   0.000000000000000D+00,   1.255375047983942D-06,   0.000000000000000D+00, &
   1.730723722079702D-07,   0.000000000000000D+00,   2.163333198808223D-08, &
   0.000000000000000D+00,   5.395902555751229D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.105631744350697D-05,   0.000000000000000D+00, &
  -2.420629697587887D-06,   0.000000000000000D+00,  -6.633482152720837D-07, &
   0.000000000000000D+00,  -1.385547799834611D-07,   0.000000000000000D+00, &
  -1.993549677152591D-08,   0.000000000000000D+00,   6.014196184717017D-09, &
   0.000000000000000D+00,  -8.595541998716722D-05,   0.000000000000000D+00, &
  -5.357578844664797D-06,   0.000000000000000D+00,   2.405645934874660D-07, &
   0.000000000000000D+00,   3.053789390070406D-07,   0.000000000000000D+00, &
   9.629764731165878D-08,   0.000000000000000D+00,   1.223155635045261D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.517717798319429D-04, &
   0.000000000000000D+00,   1.756314406530065D-05,   0.000000000000000D+00, &
   4.066685882480669D-06,   0.000000000000000D+00,   4.012317937685847D-07, &
   0.000000000000000D+00,  -9.906621879252237D-08,   0.000000000000000D+00, &
  -6.037708667779733D-08,   0.000000000000000D+00,  -4.603234230457598D-09, &
   0.000000000000000D+00,  -5.554321500541940D-04,   0.000000000000000D+00, &
   5.925692396929402D-05,   0.000000000000000D+00,  -3.999233505975092D-06, &
   0.000000000000000D+00,  -2.290882283573376D-06,   0.000000000000000D+00, &
  -4.874469206110132D-07,   0.000000000000000D+00,   2.076474070465781D-09, &
   0.000000000000000D+00,   4.156500639699954D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.036931122388398D-04,   0.000000000000000D+00, &
  -3.440548556915913D-04,   0.000000000000000D+00,  -1.897079887209884D-05, &
   0.000000000000000D+00,  -4.564521124029178D-07,   0.000000000000000D+00, &
   1.022629363557703D-06,   0.000000000000000D+00,   3.748128175573917D-07, &
   0.000000000000000D+00,   2.104012237459434D-08,   0.000000000000000D+00, &
  -1.033801610832745D-07,   0.000000000000000D+00,   6.875295563563542D-04, &
   0.000000000000000D+00,   6.133095229143742D-04,   0.000000000000000D+00, &
   1.031293161550156D-05,   0.000000000000000D+00,   8.877760952472178D-06, &
   0.000000000000000D+00,   1.515910438742415D-06,   0.000000000000000D+00, &
  -2.853951078342324D-07,   0.000000000000000D+00,  -2.254304211414578D-07, &
   0.000000000000000D+00,   7.599450945580198D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.222701092439568D-03,   0.000000000000000D+00, &
  -2.702336328431006D-04,   0.000000000000000D+00,   1.718144215863910D-04, &
   0.000000000000000D+00,   6.076056578381753D-06,   0.000000000000000D+00, &
  -3.483138447413903D-06,   0.000000000000000D+00,  -1.348328933546672D-06, &
   0.000000000000000D+00,  -5.262231977228530D-08,   0.000000000000000D+00, &
   1.157390014478303D-07,   0.000000000000000D+00,  -7.177713103519579D-08, &
   0.000000000000000D+00,   7.278939819388261D-04,   0.000000000000000D+00, &
  -7.288703751486533D-04,   0.000000000000000D+00,  -4.298778949392931D-04, &
   0.000000000000000D+00,  -4.470709467719569D-05,   0.000000000000000D+00, &
  -7.222257752835420D-06,   0.000000000000000D+00,   6.993648268904404D-07, &
   0.000000000000000D+00,   8.427712383116169D-07,   0.000000000000000D+00, &
   1.351832000490542D-07,   0.000000000000000D+00,  -3.633412823754385D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.284583831608360D-04, &
   0.000000000000000D+00,   1.344949986717178D-03,   0.000000000000000D+00, &
   2.657491057112158D-04,   0.000000000000000D+00,  -1.144614811994248D-05, &
   0.000000000000000D+00,   1.284792534459096D-05,   0.000000000000000D+00, &
   5.216537896193591D-06,   0.000000000000000D+00,   5.067835631802803D-07, &
   0.000000000000000D+00,  -3.678197920914980D-07,   0.000000000000000D+00, &
  -9.775184213226504D-08,   0.000000000000000D+00,   2.745901733285585D-07, &
   0.000000000000000D+00,  -6.620584463084396D-04,   0.000000000000000D+00, &
  -7.715442318197078D-04,   0.000000000000000D+00,   4.574088218423572D-04, &
   0.000000000000000D+00,   1.950709221004012D-04,   0.000000000000000D+00, &
   2.428372260767170D-05,   0.000000000000000D+00,  -1.678517027198210D-06, &
   0.000000000000000D+00,  -2.950249818166323D-06,   0.000000000000000D+00, &
  -8.196108825634154D-07,   0.000000000000000D+00,   5.306243705042164D-08, &
   0.000000000000000D+00,  -6.995699555970522D-09,   0.000000000000000D+00 /)
lem(:,264) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.244875317140258D-05,  -1.467748546763090D-04, &
   3.931338172640457D-04,   3.192603265156830D-04,  -1.134505561566791D-03, &
   5.260326130782800D-04,   2.487693367786800D-04,  -4.325536434836449D-03, &
   3.206327409185232D-03,   8.665614436956988D-03,  -4.858931075800627D-03, &
  -7.676962820105453D-03,   2.039049604346664D-03,   2.863944293230582D-03, &
   5.840312264381938D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.244875317140710D-05, &
  -1.467748546762982D-04,   3.931338172640420D-04,   3.192603265156905D-04, &
  -1.134505561566793D-03,   5.260326130782756D-04,   2.487693367786834D-04, &
  -4.325536434836455D-03,   3.206327409185230D-03,   8.665614436957004D-03, &
  -4.858931075800628D-03,  -7.676962820105454D-03,   2.039049604346671D-03, &
   2.863944293230579D-03,   5.840312264381830D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.449415992853366D-08,   0.000000000000000D+00, &
  -9.072292345637440D-07,   0.000000000000000D+00,  -4.801139984800442D-06, &
   0.000000000000000D+00,  -8.289894367128633D-06,   0.000000000000000D+00, &
  -1.203266963062804D-07,   0.000000000000000D+00,   1.068672731480960D-05, &
   0.000000000000000D+00,   6.354801842958855D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.802245802759748D-07,   0.000000000000000D+00,   7.912677542033784D-09, &
   0.000000000000000D+00,   2.836560226648907D-10,   0.000000000000000D+00, &
  -1.530947225665813D-10,  -4.198031093089650D-06,   0.000000000000000D+00, &
  -1.726507641286646D-07,   0.000000000000000D+00,  -1.248174866920080D-08, &
   0.000000000000000D+00,   2.265801893819882D-10,   0.000000000000000D+00, &
   1.540782138261812D-05,   0.000000000000000D+00,   1.203115333859035D-06, &
   0.000000000000000D+00,   1.389635756286641D-07,   0.000000000000000D+00, &
   1.149964618159621D-08,   0.000000000000000D+00,  -6.119315357500470D-09, &
   2.370231051393786D-05,   0.000000000000000D+00,  -2.430056025065901D-06, &
   0.000000000000000D+00,  -5.591010272632454D-07,   0.000000000000000D+00, &
  -9.803572900012858D-08,   0.000000000000000D+00,  -2.696634303061219D-09, &
   0.000000000000000D+00,  -1.690381096231454D-04,   0.000000000000000D+00, &
  -4.011597539617705D-06,   0.000000000000000D+00,   1.991470014977699D-07, &
   0.000000000000000D+00,   2.384869169906713D-07,   0.000000000000000D+00, &
   5.845532571458012D-08,   0.000000000000000D+00,  -2.021705010106777D-08, &
   2.774812012383504D-05,   0.000000000000000D+00,   1.048964154376533D-05, &
   0.000000000000000D+00,   3.666812678299063D-06,   0.000000000000000D+00, &
   3.353179038795566D-07,   0.000000000000000D+00,  -7.852930267474912D-08, &
   0.000000000000000D+00,  -2.989706912166565D-08,   0.000000000000000D+00, &
   7.827539166771909D-04,   0.000000000000000D+00,  -1.456337706856368D-06, &
   0.000000000000000D+00,  -4.620611482872691D-06,   0.000000000000000D+00, &
  -2.023069296913810D-06,   0.000000000000000D+00,  -3.769290907215429D-07, &
   0.000000000000000D+00,   2.095572728773918D-08,   0.000000000000000D+00, &
   6.795690127786417D-08,  -1.000900825706165D-03,   0.000000000000000D+00, &
   1.261818710356560D-04,   0.000000000000000D+00,  -1.098548073433846D-05, &
   0.000000000000000D+00,   2.423935843760142D-07,   0.000000000000000D+00, &
   9.177151433022941D-07,   0.000000000000000D+00,   2.694276080095424D-07, &
   0.000000000000000D+00,  -3.269079501938751D-08,   0.000000000000000D+00, &
  -6.697126504763011D-04,   0.000000000000000D+00,  -4.905348829299326D-04, &
   0.000000000000000D+00,   5.120929819892323D-05,   0.000000000000000D+00, &
   7.338633395447437D-06,   0.000000000000000D+00,   9.733649352844797D-07, &
   0.000000000000000D+00,  -3.029217103923405D-07,   0.000000000000000D+00, &
  -1.534141447225670D-07,   0.000000000000000D+00,   1.715441669507778D-07, &
   2.040317648541763D-03,   0.000000000000000D+00,   4.528483135648673D-04, &
   0.000000000000000D+00,  -1.894520961209424D-04,   0.000000000000000D+00, &
  -8.818554673350908D-06,   0.000000000000000D+00,  -3.860843486881278D-06, &
   0.000000000000000D+00,  -1.037233556897384D-06,   0.000000000000000D+00, &
   3.108121201226960D-08,   0.000000000000000D+00,   4.741471192290411D-08, &
   0.000000000000000D+00,  -7.132938761341721D-04,   0.000000000000000D+00, &
   5.394503378759455D-04,   0.000000000000000D+00,   4.299301727088516D-04, &
   0.000000000000000D+00,  -1.416596640419216D-05,   0.000000000000000D+00, &
   2.268952343144815D-07,   0.000000000000000D+00,   1.637349874191551D-06, &
   0.000000000000000D+00,   7.503467720365970D-07,   0.000000000000000D+00, &
   3.888968221012932D-08,   0.000000000000000D+00,  -2.216580334003818D-07, &
  -1.135719322593626D-03,   0.000000000000000D+00,  -1.293452648497122D-03, &
   0.000000000000000D+00,  -3.002791603279786D-04,   0.000000000000000D+00, &
   1.175064755858271D-04,   0.000000000000000D+00,   1.160846986048235D-05, &
   0.000000000000000D+00,   1.843580166393435D-06,   0.000000000000000D+00, &
  -4.323715722690545D-07,   0.000000000000000D+00,  -4.371339338543453D-07, &
   0.000000000000000D+00,  -3.298942347980315D-08,   0.000000000000000D+00, &
   9.495412354642717D-04,   0.000000000000000D+00,   5.731979646554538D-04, &
   0.000000000000000D+00,  -5.241029542386782D-04,   0.000000000000000D+00, &
  -3.017560728707810D-04,   0.000000000000000D+00,  -2.393232139153818D-05, &
   0.000000000000000D+00,  -6.310270697300318D-06,   0.000000000000000D+00, &
  -1.900197023347111D-06,   0.000000000000000D+00,  -1.127927850294588D-07, &
   0.000000000000000D+00,   2.621519626664873D-07,   0.000000000000000D+00, &
  -2.382089697672896D-07,   3.292418136036498D-05,   0.000000000000000D+00, &
   5.132930092295205D-04,   0.000000000000000D+00,   1.114133303491444D-03, &
   0.000000000000000D+00,   2.302159623422686D-04,   0.000000000000000D+00, &
  -1.215450809206318D-05,   0.000000000000000D+00,   3.791190381447498D-06, &
   0.000000000000000D+00,   2.802342816257761D-06,   0.000000000000000D+00, &
   1.404557071530829D-06,   0.000000000000000D+00,   2.727723129737987D-07, &
   0.000000000000000D+00,  -1.599890714080151D-08,   0.000000000000000D+00, &
  -2.334616654933485D-04,   0.000000000000000D+00,  -5.918240310336983D-04, &
   0.000000000000000D+00,  -5.561962712419447D-04,   0.000000000000000D+00, &
   3.381927832479124D-04,   0.000000000000000D+00,   1.393034605957514D-04, &
   0.000000000000000D+00,   1.710691158267549D-05,   0.000000000000000D+00, &
   1.460962080831040D-06,   0.000000000000000D+00,  -8.619244301903450D-07, &
   0.000000000000000D+00,  -8.770744296494286D-07,   0.000000000000000D+00, &
  -2.723513224702747D-07,   0.000000000000000D+00,   4.139275961628620D-07 /)
lem(:,265) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.469984832119781D-06, &
   5.419727226967899D-05,  -2.300396740786239D-04,   5.403720389800469D-05, &
   9.541288011106494D-04,  -3.780490147057850D-04,  -1.397643290130740D-03, &
  -1.857952732513885D-03,   1.894790648207347D-03,   8.119947848128743D-03, &
  -3.606200419582606D-03,  -1.234458853520320D-02,   4.372737673453900D-03, &
   9.300681308796308D-03,  -1.998408959599460D-03,  -3.582372333494362D-03, &
  -4.953018261627752D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.469984832130228D-06,  -5.419727226969298D-05,   2.300396740786217D-04, &
  -5.403720389799393D-05,  -9.541288011106453D-04,   3.780490147057759D-04, &
   1.397643290130730D-03,   1.857952732513897D-03,  -1.894790648207339D-03, &
  -8.119947848128740D-03,   3.606200419582603D-03,   1.234458853520319D-02, &
  -4.372737673453914D-03,  -9.300681308796294D-03,   1.998408959599454D-03, &
   3.582372333494362D-03,   4.953018261627783D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.106268448978790D-08,   0.000000000000000D+00, &
   5.954625462923718D-07,   0.000000000000000D+00,   4.868336415493039D-06, &
   0.000000000000000D+00,   1.228492887453045D-05,   0.000000000000000D+00, &
   6.738206901131966D-06,   0.000000000000000D+00,  -1.003167084884444D-05, &
   0.000000000000000D+00,  -7.434405522630568D-06,   0.000000000000000D+00, &
  -4.429807165872065D-07,   0.000000000000000D+00,  -8.066646871470008D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.790988688095131D-09, &
   0.000000000000000D+00,   3.756833010527194D-12,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.921319416598022D-07,   0.000000000000000D+00, &
  -8.768447691517327D-09,   0.000000000000000D+00,   3.272107424389454D-09, &
   0.000000000000000D+00,   1.397933568495762D-06,   0.000000000000000D+00, &
   1.170459967160283D-07,   0.000000000000000D+00,   5.856356943952926D-10, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.463893422127647D-06, &
   0.000000000000000D+00,  -5.057986959904506D-07,   0.000000000000000D+00, &
  -6.379868260078590D-08,   0.000000000000000D+00,   3.000172492324074D-08, &
   0.000000000000000D+00,  -3.721954628055633D-06,   0.000000000000000D+00, &
   2.304772968589854D-07,   0.000000000000000D+00,   1.931347112646564D-07, &
   0.000000000000000D+00,   1.877454379634362D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.321294727099874D-05,   0.000000000000000D+00, &
   3.268645211920744D-06,   0.000000000000000D+00,   2.619663906282169D-07, &
   0.000000000000000D+00,  -6.805415721235018D-08,   0.000000000000000D+00, &
  -1.123060948951494D-08,   0.000000000000000D+00,  -3.883596126012327D-05, &
   0.000000000000000D+00,  -3.326536907328953D-06,   0.000000000000000D+00, &
  -1.759407916300222D-06,   0.000000000000000D+00,  -2.669824271933597D-07, &
   0.000000000000000D+00,   4.821773391518108D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.698471584875190D-05,   0.000000000000000D+00, &
  -1.628966668315446D-05,   0.000000000000000D+00,   3.732832362176235D-07, &
   0.000000000000000D+00,   7.765553358529259D-07,   0.000000000000000D+00, &
   1.586486569222069D-07,   0.000000000000000D+00,  -2.359309949231650D-07, &
   0.000000000000000D+00,   2.811545506297201D-04,   0.000000000000000D+00, &
   9.604706399547836D-06,   0.000000000000000D+00,   5.326458117678828D-06, &
   0.000000000000000D+00,   6.422325375468891D-07,   0.000000000000000D+00, &
  -2.816124483818635D-07,   0.000000000000000D+00,  -3.169632047829754D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.012366284912053D-04, &
   0.000000000000000D+00,   1.481151634912577D-04,   0.000000000000000D+00, &
  -1.786797417654831D-06,   0.000000000000000D+00,  -2.674845178769629D-06, &
   0.000000000000000D+00,  -6.734369267143954D-07,   0.000000000000000D+00, &
   1.070872790304099D-07,   0.000000000000000D+00,   7.469656824970808D-08, &
   0.000000000000000D+00,  -2.577442599836276D-04,   0.000000000000000D+00, &
  -3.755524753305420D-04,   0.000000000000000D+00,   1.626712402970330D-05, &
   0.000000000000000D+00,  -8.759580553551124D-08,   0.000000000000000D+00, &
   1.126377979965907D-06,   0.000000000000000D+00,   4.431782844375087D-07, &
   0.000000000000000D+00,  -4.519972924474051D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.658204311392581D-04,   0.000000000000000D+00, &
   2.341105539063962D-04,   0.000000000000000D+00,  -1.565056889714319D-04, &
   0.000000000000000D+00,  -1.756440047108388D-06,   0.000000000000000D+00, &
   6.296477219984623D-07,   0.000000000000000D+00,  -3.855743125080315D-07, &
   0.000000000000000D+00,  -2.766155114833786D-07,   0.000000000000000D+00, &
   5.321562754087514D-07,   0.000000000000000D+00,  -2.753168781571717D-04, &
   0.000000000000000D+00,   4.040802108319558D-04,   0.000000000000000D+00, &
   3.593321785210492D-04,   0.000000000000000D+00,   1.005285707136693D-06, &
   0.000000000000000D+00,   1.330304345992742D-07,   0.000000000000000D+00, &
  -5.227861652247807D-07,   0.000000000000000D+00,   1.279448524879614D-07, &
   0.000000000000000D+00,   7.636857727088578D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.286190080687797D-04,   0.000000000000000D+00, &
  -8.258742581684043D-04,   0.000000000000000D+00,  -2.119517826069314D-04, &
   0.000000000000000D+00,   8.802491574747587D-05,   0.000000000000000D+00, &
   3.523712105048840D-06,   0.000000000000000D+00,   1.888411626883917D-07, &
   0.000000000000000D+00,   3.229366599603510D-07,   0.000000000000000D+00, &
  -1.400660527969095D-07,   0.000000000000000D+00,  -4.732095263888418D-07, &
   0.000000000000000D+00,   3.437555305237547D-04,   0.000000000000000D+00, &
   4.306760360418309D-04,   0.000000000000000D+00,  -4.188673491184603D-04, &
   0.000000000000000D+00,  -2.530643266597164D-04,   0.000000000000000D+00, &
  -2.221138866874117D-05,   0.000000000000000D+00,  -2.946445837712679D-06, &
   0.000000000000000D+00,  -3.338238196532878D-07,   0.000000000000000D+00, &
  -1.708815301438210D-07,   0.000000000000000D+00,  -1.783904860129548D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.169339357114261D-04, &
   0.000000000000000D+00,   2.657491057112158D-04,   0.000000000000000D+00, &
   8.362973476672112D-04,   0.000000000000000D+00,   1.807444078044576D-04, &
   0.000000000000000D+00,  -4.179356067140862D-06,   0.000000000000000D+00, &
   6.069593893930727D-06,   0.000000000000000D+00,   1.932747618926402D-06, &
   0.000000000000000D+00,   4.087641703369801D-07,   0.000000000000000D+00, &
   3.265343537783065D-07,   0.000000000000000D+00,  -4.128034197513196D-07, &
   0.000000000000000D+00,  -5.252045692262341D-05,   0.000000000000000D+00, &
  -4.571945625922477D-04,   0.000000000000000D+00,  -4.457591390083918D-04, &
   0.000000000000000D+00,   2.756047391035378D-04,   0.000000000000000D+00, &
   1.175177276492078D-04,   0.000000000000000D+00,   1.339676159432827D-05, &
   0.000000000000000D+00,  -6.434385883156395D-07,   0.000000000000000D+00, &
  -1.045143180265104D-06,   0.000000000000000D+00,  -4.160990103718939D-07, &
   0.000000000000000D+00,  -9.786708176573565D-08,   0.000000000000000D+00 /)
lem(:,266) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.146810421401124D-06,  -1.996036342805378D-05,   1.060444935852683D-04, &
  -1.164811275172997D-04,  -3.906563966711545D-04,   5.701731743360650D-04, &
   1.072482990894837D-04,  -8.102300395026363D-04,   7.936964710822747D-04, &
   3.894391434564525D-03,  -2.489652163375317D-03,  -1.143388732434786D-02, &
   4.708601992699185D-03,   1.526650607467589D-02,  -5.227829989902291D-03, &
  -1.025343022114836D-02,   3.038554424784628D-03,   3.785771414007467D-03, &
  -5.857904901755370D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.146810421398908D-06,  -1.996036342804227D-05, &
   1.060444935852698D-04,  -1.164811275173076D-04,  -3.906563966711559D-04, &
   5.701731743360679D-04,   1.072482990894911D-04,  -8.102300395026330D-04, &
   7.936964710822570D-04,   3.894391434564536D-03,  -2.489652163375304D-03, &
  -1.143388732434788D-02,   4.708601992699174D-03,   1.526650607467589D-02, &
  -5.227829989902296D-03,  -1.025343022114837D-02,   3.038554424784627D-03, &
   3.785771414007481D-03,  -5.857904901755480D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.833051202523317D-07,   0.000000000000000D+00,  -4.220133684125492D-06, &
   0.000000000000000D+00,  -1.554686359535579D-05,   0.000000000000000D+00, &
  -1.589548347401999D-05,   0.000000000000000D+00,   7.155563597939448D-06, &
   0.000000000000000D+00,   1.133277248053872D-05,   0.000000000000000D+00, &
  -5.170475083981780D-06,   0.000000000000000D+00,   3.228937965601645D-06, &
   0.000000000000000D+00,   1.476452148981905D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.736010956604256D-08,   0.000000000000000D+00,  -5.564137681634481D-10, &
  -3.538791565445274D-07,   0.000000000000000D+00,  -5.778142977017632D-09, &
   0.000000000000000D+00,   2.268287879256125D-06,   0.000000000000000D+00, &
   1.078830883163391D-07,   0.000000000000000D+00,  -2.229661801830731D-08, &
  -2.907956496468300D-06,   0.000000000000000D+00,  -5.119497738706249D-07, &
   0.000000000000000D+00,  -2.781959825528572D-08,   0.000000000000000D+00, &
  -1.777743177580640D-05,   0.000000000000000D+00,   4.028930475335314D-07, &
   0.000000000000000D+00,   1.628152721626416D-07,   0.000000000000000D+00, &
  -3.897077157907449D-08,   4.029923206278637D-05,   0.000000000000000D+00, &
   2.435075816722526D-06,   0.000000000000000D+00,   1.774762363875203D-07, &
   0.000000000000000D+00,  -5.939506622669672D-08,   0.000000000000000D+00, &
   5.612450043020509D-05,   0.000000000000000D+00,  -2.929074916436743D-06, &
   0.000000000000000D+00,  -1.516501586004570D-06,   0.000000000000000D+00, &
  -1.538894144332519D-07,   0.000000000000000D+00,   2.283109657046516D-07, &
  -1.714258057538221D-04,   0.000000000000000D+00,  -5.516119394493206D-07, &
   0.000000000000000D+00,   3.377453351561931D-07,   0.000000000000000D+00, &
   6.474517958416172D-07,   0.000000000000000D+00,   1.340496449440399D-08, &
   0.000000000000000D+00,  -8.623287160514115D-05,   0.000000000000000D+00, &
  -1.796596422933300D-05,   0.000000000000000D+00,   5.916269507328889D-06, &
   0.000000000000000D+00,   4.211441139051364D-07,   0.000000000000000D+00, &
  -2.659568519534415D-07,   0.000000000000000D+00,   1.567501362212697D-07, &
   3.873249594157237D-04,   0.000000000000000D+00,   2.398487498643718D-06, &
   0.000000000000000D+00,  -8.084060452667733D-06,   0.000000000000000D+00, &
  -2.151414840409400D-06,   0.000000000000000D+00,  -4.023105438265990D-07, &
   0.000000000000000D+00,   9.635763707633788D-08,   0.000000000000000D+00, &
   4.702155960827984D-05,   0.000000000000000D+00,   1.659397683825674D-04, &
   0.000000000000000D+00,  -1.057066802714808D-05,   0.000000000000000D+00, &
  -1.319483390770139D-06,   0.000000000000000D+00,   8.013474764512626D-07, &
   0.000000000000000D+00,   2.200433210789462D-07,   0.000000000000000D+00, &
  -7.105008719092168D-07,  -5.280678074221608D-04,   0.000000000000000D+00, &
  -2.313373604426494D-04,   0.000000000000000D+00,   1.004593628685650D-04, &
   0.000000000000000D+00,   5.140953852883357D-06,   0.000000000000000D+00, &
   1.084860137326866D-06,   0.000000000000000D+00,  -2.443721286750149D-07, &
   0.000000000000000D+00,  -2.709191845738822D-08,   0.000000000000000D+00, &
   5.097163229124517D-05,   0.000000000000000D+00,  -1.577089286189178D-04, &
   0.000000000000000D+00,  -2.626115330995250D-04,   0.000000000000000D+00, &
   2.015434880733202D-05,   0.000000000000000D+00,  -2.733004005921906D-07, &
   0.000000000000000D+00,  -5.803274369709333D-07,   0.000000000000000D+00, &
   2.043087929028076D-07,   0.000000000000000D+00,   7.306164169778654D-08, &
   4.481042150988476D-04,   0.000000000000000D+00,   4.938029604474748D-04, &
   0.000000000000000D+00,   2.016665565964868D-04,   0.000000000000000D+00, &
  -1.269374643034340D-04,   0.000000000000000D+00,  -7.407698447952168D-06, &
   0.000000000000000D+00,  -6.454060141888031D-07,   0.000000000000000D+00, &
   1.352134672107380D-07,   0.000000000000000D+00,   2.876997635821153D-08, &
   0.000000000000000D+00,  -1.098443391704188D-04,   0.000000000000000D+00, &
  -1.690861711948870D-04,   0.000000000000000D+00,   2.972765339258929D-04, &
   0.000000000000000D+00,   2.771008454995598D-04,   0.000000000000000D+00, &
   1.156847109546282D-06,   0.000000000000000D+00,   2.500194509299865D-06, &
   0.000000000000000D+00,   6.391164720872304D-07,   0.000000000000000D+00, &
  -3.584050443182005D-08,   0.000000000000000D+00,   9.295539319382801D-07, &
  -2.293285076853964D-04,   0.000000000000000D+00,  -2.659083501690629D-04, &
   0.000000000000000D+00,  -6.626559497331646D-04,   0.000000000000000D+00, &
  -1.759569476614578D-04,   0.000000000000000D+00,   7.414247977107101D-05, &
   0.000000000000000D+00,   5.086099327876450D-06,   0.000000000000000D+00, &
  -3.767689162647804D-07,   0.000000000000000D+00,  -4.034752844412494D-07, &
   0.000000000000000D+00,  -2.011849093066686D-07,   0.000000000000000D+00, &
   8.397396336753180D-05,   0.000000000000000D+00,   2.051322797407905D-04, &
   0.000000000000000D+00,   3.176547135837127D-04,   0.000000000000000D+00, &
  -3.299831099866026D-04,   0.000000000000000D+00,  -2.035857734677348D-04, &
   0.000000000000000D+00,  -2.007231506870128D-05,   0.000000000000000D+00, &
  -4.188804169018388D-06,   0.000000000000000D+00,  -4.408062200661905D-07, &
   0.000000000000000D+00,  -1.155953064648076D-07,   0.000000000000000D+00, &
  -4.038259839922591D-07,   6.203332343229487D-05,   0.000000000000000D+00, &
   2.065983041709690D-06,   0.000000000000000D+00,   2.302159623422686D-04, &
   0.000000000000000D+00,   6.818375981277722D-04,   0.000000000000000D+00, &
   1.509804829904389D-04,   0.000000000000000D+00,  -4.183917478400995D-06, &
   0.000000000000000D+00,   5.063379978245968D-06,   0.000000000000000D+00, &
   2.552577575022065D-06,   0.000000000000000D+00,   8.187282880046705D-07, &
   0.000000000000000D+00,  -7.411983182931629D-08,   0.000000000000000D+00, &
  -3.114247516553934D-05,   0.000000000000000D+00,  -2.383224697154986D-05, &
   0.000000000000000D+00,  -3.215473556424558D-04,   0.000000000000000D+00, &
  -3.519337023407202D-04,   0.000000000000000D+00,   2.229080484713501D-04, &
   0.000000000000000D+00,   9.585734028572447D-05,   0.000000000000000D+00, &
   1.202368666576861D-05,   0.000000000000000D+00,  -8.810241510817076D-09, &
   0.000000000000000D+00,  -1.224049311366901D-06,   0.000000000000000D+00, &
  -4.405258281047706D-07,   0.000000000000000D+00,  -6.219407894362764D-07 /)
lem(:,267) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.013492061256229D-07,   4.309553049407641D-06, &
  -4.109904591763472D-05,   1.085764765456415D-04,   1.223784230601403D-04, &
  -6.053830512488282D-04,  -1.749002740370315D-04,   8.243558247520502D-04, &
   1.184958587200663D-03,   5.743686678112534D-04,  -2.732019704921007D-03, &
  -6.247140046372003D-03,   4.249220919027500D-03,   1.424649731140172D-02, &
  -6.491454708249489D-03,  -1.588714696378768D-02,   7.542335466371721D-03, &
   9.333839248384176D-03,  -5.151785906717265D-03,  -2.964272921639788D-03, &
   1.980422262107422D-03,   0.000000000000000D+00,   1.013492061461253D-07, &
  -4.309553049408524D-06,   4.109904591761991D-05,  -1.085764765456426D-04, &
  -1.223784230601279D-04,   6.053830512488208D-04,   1.749002740370255D-04, &
  -8.243558247520448D-04,  -1.184958587200656D-03,  -5.743686678112614D-04, &
   2.732019704920998D-03,   6.247140046372006D-03,  -4.249220919027501D-03, &
  -1.424649731140173D-02,   6.491454708249487D-03,   1.588714696378767D-02, &
  -7.542335466371715D-03,  -9.333839248384211D-03,   5.151785906717264D-03, &
   2.964272921639797D-03,  -1.980422262107425D-03,   0.000000000000000D+00, &
   5.067460307226333D-08,   0.000000000000000D+00,   2.822264238104903D-06, &
   0.000000000000000D+00,   1.681113756065406D-05,   0.000000000000000D+00, &
   2.662389604718496D-05,   0.000000000000000D+00,  -8.834647157009698D-08, &
   0.000000000000000D+00,  -1.783117378593785D-05,   0.000000000000000D+00, &
   6.449096606499967D-06,   0.000000000000000D+00,   9.143815590925439D-06, &
   0.000000000000000D+00,  -7.475050367727129D-06,   0.000000000000000D+00, &
  -1.580863745887231D-05,   0.000000000000000D+00,  -1.287704526330208D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.006179130539050D-09, &
   0.000000000000000D+00,   1.299782562400809D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.890285696315752D-07,   0.000000000000000D+00, &
   2.856542683559307D-08,   0.000000000000000D+00,   8.748603385109318D-07, &
   0.000000000000000D+00,   1.476525443645141D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.619707607206806D-06,   0.000000000000000D+00, &
   5.864075159727163D-08,   0.000000000000000D+00,  -1.698978507557429D-07, &
   0.000000000000000D+00,  -1.143842076240589D-05,   0.000000000000000D+00, &
  -1.213769810284104D-06,   0.000000000000000D+00,  -1.159456701954639D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.133125141114453D-06, &
   0.000000000000000D+00,  -2.782061759905625D-07,   0.000000000000000D+00, &
   5.296248362789797D-07,   0.000000000000000D+00,  -3.792127906236556D-07, &
   0.000000000000000D+00,   4.138409301572558D-05,   0.000000000000000D+00, &
   6.055723353769038D-06,   0.000000000000000D+00,   2.438832718963131D-07, &
   0.000000000000000D+00,  -1.597658790041490D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.411008151322505D-05,   0.000000000000000D+00, &
   4.213619880241559D-06,   0.000000000000000D+00,  -1.902699573028107D-06, &
   0.000000000000000D+00,  -1.457953553816794D-07,   0.000000000000000D+00, &
   6.933573184650414D-07,   0.000000000000000D+00,  -6.752405717711948D-05, &
   0.000000000000000D+00,  -3.326226984985678D-05,   0.000000000000000D+00, &
   2.315669846884759D-07,   0.000000000000000D+00,   5.932512228391311D-07, &
   0.000000000000000D+00,  -5.794041246490118D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.095588377038783D-04,   0.000000000000000D+00, &
  -6.506155678297237D-06,   0.000000000000000D+00,  -2.504913487141664D-06, &
   0.000000000000000D+00,   9.788004229152853D-07,   0.000000000000000D+00, &
  -2.990240302005754D-07,   0.000000000000000D+00,   4.929758145816674D-07, &
   0.000000000000000D+00,   3.779052035309391D-05,   0.000000000000000D+00, &
   1.456506571517006D-04,   0.000000000000000D+00,   2.341906484866978D-06, &
   0.000000000000000D+00,  -1.800055428157534D-06,   0.000000000000000D+00, &
  -4.843276329354191D-07,   0.000000000000000D+00,  -7.056977565545099D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.620241832851591D-04, &
   0.000000000000000D+00,  -1.521534766833258D-04,   0.000000000000000D+00, &
   7.438994133078912D-05,   0.000000000000000D+00,   1.255945014856946D-06, &
   0.000000000000000D+00,   1.832633842858908D-07,   0.000000000000000D+00, &
   1.444733015562726D-07,   0.000000000000000D+00,  -1.141049212601456D-06, &
   0.000000000000000D+00,   4.107436271826625D-05,   0.000000000000000D+00, &
  -1.272769320472504D-04,   0.000000000000000D+00,  -2.190138425404549D-04, &
   0.000000000000000D+00,   1.238312325300816D-05,   0.000000000000000D+00, &
   1.610349158811429D-06,   0.000000000000000D+00,   2.931516228751032D-07, &
   0.000000000000000D+00,   7.422027263696349D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.270358897207956D-04,   0.000000000000000D+00, &
   3.383724263994635D-04,   0.000000000000000D+00,   1.582312958125735D-04, &
   0.000000000000000D+00,  -9.876410386104340D-05,   0.000000000000000D+00, &
  -5.413894140092218D-06,   0.000000000000000D+00,  -1.153863573004831D-06, &
   0.000000000000000D+00,   3.198666810411677D-08,   0.000000000000000D+00, &
   1.126694768787483D-07,   0.000000000000000D+00,  -8.661882209724809D-05, &
   0.000000000000000D+00,  -1.370659240460748D-04,   0.000000000000000D+00, &
   2.409068791879921D-04,   0.000000000000000D+00,   2.321651939645110D-04, &
   0.000000000000000D+00,   2.826557100744237D-06,   0.000000000000000D+00, &
   1.347805039494620D-06,   0.000000000000000D+00,   5.291113546456525D-07, &
   0.000000000000000D+00,   5.504802724461387D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.399915936545186D-05,   0.000000000000000D+00, &
  -1.752062499395887D-04,   0.000000000000000D+00,  -5.192767294737469D-04, &
   0.000000000000000D+00,  -1.461013652562690D-04,   0.000000000000000D+00, &
   5.820823998436113D-05,   0.000000000000000D+00,   3.884438609862693D-06, &
   0.000000000000000D+00,   6.908190796180910D-08,   0.000000000000000D+00, &
  -3.915241990034036D-07,   0.000000000000000D+00,   2.520672854098924D-07, &
   0.000000000000000D+00,   6.229508705012527D-05,   0.000000000000000D+00, &
   1.826332318541640D-04,   0.000000000000000D+00,   2.581590194348226D-04, &
   0.000000000000000D+00,  -2.730169826450611D-04,   0.000000000000000D+00, &
  -1.726479584958697D-04,   0.000000000000000D+00,  -1.767342317502922D-05, &
   0.000000000000000D+00,  -3.355201996887104D-06,   0.000000000000000D+00, &
  -7.994558924681393D-07,   0.000000000000000D+00,  -1.345791959967847D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.827402751427659D-06, &
   0.000000000000000D+00,  -1.144614811994248D-05,   0.000000000000000D+00, &
   1.807444078044576D-04,   0.000000000000000D+00,   5.580729107773259D-04, &
   0.000000000000000D+00,   1.285468891246864D-04,   0.000000000000000D+00, &
  -8.794031090607202D-07,   0.000000000000000D+00,   4.947184283086292D-06, &
   0.000000000000000D+00,   2.282942447494380D-06,   0.000000000000000D+00, &
   5.162807821686528D-07,   0.000000000000000D+00,  -8.304391831676980D-08, &
   0.000000000000000D+00,  -1.959019658286450D-05,   0.000000000000000D+00, &
  -4.763960312141699D-05,   0.000000000000000D+00,  -2.703813820643770D-04, &
   0.000000000000000D+00,  -2.917739169259423D-04,   0.000000000000000D+00, &
   1.872410332605680D-04,   0.000000000000000D+00,   8.192495745103279D-05, &
   0.000000000000000D+00,   9.932281804911033D-06,   0.000000000000000D+00, &
  -3.531348933717112D-07,   0.000000000000000D+00,  -9.484373108211013D-07, &
   0.000000000000000D+00,  -9.271606312051729D-07,   0.000000000000000D+00 /)
lem(:,268) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.533296045052201D-06,   1.652203204275829D-05,  -4.633159886185783D-05, &
  -3.016166202904593D-05,   2.632841916257333D-04,  -1.235959428876170D-04, &
  -1.725469355966516D-04,   3.931750016724983D-04,  -2.686849039827911D-04, &
  -1.720181728096405D-03,  -1.112532830590020D-03,   4.049959459425862D-03, &
   7.418984293315779D-03,  -6.235607357074073D-03,  -1.453464255144274D-02, &
   8.647431978617386D-03,   1.349645731329116D-02,  -9.664554887353584D-03, &
  -5.805447009086020D-03,   6.754593454705534D-03,   5.760381118028628D-04, &
  -2.746583724539179D-03,  -1.533296045057006D-06,   1.652203204276936D-05, &
  -4.633159886185679D-05,  -3.016166202906082D-05,   2.632841916257358D-04, &
  -1.235959428876005D-04,  -1.725469355966599D-04,   3.931750016724883D-04, &
  -2.686849039827865D-04,  -1.720181728096395D-03,  -1.112532830590018D-03, &
   4.049959459425843D-03,   7.418984293315790D-03,  -6.235607357074058D-03, &
  -1.453464255144273D-02,   8.647431978617377D-03,   1.349645731329118D-02, &
  -9.664554887353584D-03,  -5.805447009086025D-03,   6.754593454705532D-03, &
   5.760381118028642D-04,  -2.746583724539169D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.274128727652546D-05,   0.000000000000000D+00, &
  -3.390049305064106D-05,   0.000000000000000D+00,  -9.810726864994649D-06, &
   0.000000000000000D+00,   2.751930487717295D-05,   0.000000000000000D+00, &
  -4.128850622155995D-07,   0.000000000000000D+00,  -2.094753471290551D-05, &
   0.000000000000000D+00,  -4.663946109595923D-06,   0.000000000000000D+00, &
   1.425625803807910D-07,   0.000000000000000D+00,   1.842504713885620D-05, &
   0.000000000000000D+00,   2.860658072958324D-05,   0.000000000000000D+00, &
   3.551310875225913D-07,  -1.067897124524747D-06,   0.000000000000000D+00, &
  -2.187733334253266D-07,   0.000000000000000D+00,   2.946291154348404D-07, &
   8.799598141198264D-06,   0.000000000000000D+00,  -2.243604766761965D-07, &
   0.000000000000000D+00,  -6.896161986548835D-06,   0.000000000000000D+00, &
  -7.555583329428351D-07,   0.000000000000000D+00,   5.219110411247925D-07, &
  -3.263502547638058D-05,   0.000000000000000D+00,   6.987764554929510D-07, &
   0.000000000000000D+00,   3.027589686411978D-07,   0.000000000000000D+00, &
   3.099949764968391D-05,   0.000000000000000D+00,  -3.463483869684371D-07, &
   0.000000000000000D+00,   1.733220991355446D-07,   0.000000000000000D+00, &
  -4.841519236615756D-07,   6.994323507771172D-05,   0.000000000000000D+00, &
   4.293690818018080D-06,   0.000000000000000D+00,  -2.399464139179371D-06, &
   0.000000000000000D+00,   1.154461264971188D-08,   0.000000000000000D+00, &
  -5.425479219807682D-05,   0.000000000000000D+00,   9.116389086806821D-06, &
   0.000000000000000D+00,   1.140559954965399D-06,   0.000000000000000D+00, &
   5.630359244537759D-07,   0.000000000000000D+00,  -1.126947700896971D-06, &
  -1.021041191578757D-04,   0.000000000000000D+00,  -3.439842876744588D-05, &
   0.000000000000000D+00,   7.738798795285721D-06,   0.000000000000000D+00, &
   6.022764563864215D-07,   0.000000000000000D+00,   5.358416857404380D-08, &
   0.000000000000000D+00,   3.111860017823468D-05,   0.000000000000000D+00, &
  -2.062186227086454D-05,   0.000000000000000D+00,  -1.927648248867731D-05, &
   0.000000000000000D+00,   1.772748936531506D-08,   0.000000000000000D+00, &
  -2.209764668434328D-07,   0.000000000000000D+00,   1.108508143079305D-06, &
   1.191433287144560D-04,   0.000000000000000D+00,   9.504031585117140D-05, &
   0.000000000000000D+00,  -6.318256940566707D-06,   0.000000000000000D+00, &
  -7.080958064506110D-06,   0.000000000000000D+00,   2.970638127888545D-07, &
   0.000000000000000D+00,   2.451145337146138D-07,   0.000000000000000D+00, &
   3.373904569911270D-05,   0.000000000000000D+00,   1.298295145204358D-05, &
   0.000000000000000D+00,   1.033592144533433D-04,   0.000000000000000D+00, &
   3.207454500370191D-06,   0.000000000000000D+00,  -6.181665623878282D-07, &
   0.000000000000000D+00,  -3.910214088409301D-07,   0.000000000000000D+00, &
   3.539287952408322D-07,  -1.217086214794309D-04,   0.000000000000000D+00, &
  -1.376709495352444D-04,   0.000000000000000D+00,  -1.234859093334541D-04, &
   0.000000000000000D+00,   6.168660606223377D-05,   0.000000000000000D+00, &
   4.431087397893792D-06,   0.000000000000000D+00,  -6.419097110114519D-08, &
   0.000000000000000D+00,  -3.890935367977604D-07,   0.000000000000000D+00, &
  -6.893880658423893D-05,   0.000000000000000D+00,   1.426480301681312D-05, &
   0.000000000000000D+00,  -9.357907684213191D-05,   0.000000000000000D+00, &
  -1.761323965525997D-04,   0.000000000000000D+00,   6.793863186018509D-06, &
   0.000000000000000D+00,   6.843406033226455D-07,   0.000000000000000D+00, &
   1.566503423195811D-07,   0.000000000000000D+00,  -8.332479769459496D-07, &
   1.008427085528609D-04,   0.000000000000000D+00,   1.114612359391207D-04, &
   0.000000000000000D+00,   2.727292682450622D-04,   0.000000000000000D+00, &
   1.324163883596886D-04,   0.000000000000000D+00,  -8.242878371215190D-05, &
   0.000000000000000D+00,  -6.345321877593150D-06,   0.000000000000000D+00, &
  -6.056812719440989D-07,   0.000000000000000D+00,  -2.726127386717087D-07, &
   0.000000000000000D+00,   4.537830091041734D-05,   0.000000000000000D+00, &
  -2.703567840646927D-05,   0.000000000000000D+00,  -1.010791275940241D-04, &
   0.000000000000000D+00,   1.941285965492779D-04,   0.000000000000000D+00, &
   1.947694651041991D-04,   0.000000000000000D+00,   5.724424173416880D-06, &
   0.000000000000000D+00,   2.028574151176716D-06,   0.000000000000000D+00, &
   1.053677594376740D-06,   0.000000000000000D+00,   9.167052724586657D-07, &
  -5.755091459944866D-05,   0.000000000000000D+00,  -4.621119289361296D-05, &
   0.000000000000000D+00,  -1.426384334216252D-04,   0.000000000000000D+00, &
  -4.291975230230603D-04,   0.000000000000000D+00,  -1.256597958578279D-04, &
   0.000000000000000D+00,   4.847205200282096D-05,   0.000000000000000D+00, &
   3.375673008537472D-06,   0.000000000000000D+00,  -3.426759201290022D-07, &
   0.000000000000000D+00,  -7.754655789933091D-08,   0.000000000000000D+00, &
  -1.038795510715184D-05,   0.000000000000000D+00,   1.302413844714418D-05, &
   0.000000000000000D+00,   1.301836715768905D-04,   0.000000000000000D+00, &
   2.083940000685976D-04,   0.000000000000000D+00,  -2.273237063374402D-04, &
   0.000000000000000D+00,  -1.478045425977890D-04,   0.000000000000000D+00, &
  -1.717684263941916D-05,   0.000000000000000D+00,  -3.771535906944601D-06, &
   0.000000000000000D+00,  -7.859729571789726D-07,   0.000000000000000D+00, &
  -8.902081664313933D-08,   1.880647376243266D-05,   0.000000000000000D+00, &
   5.730063744462857D-06,   0.000000000000000D+00,  -1.215450809206318D-05, &
   0.000000000000000D+00,   1.509804829904389D-04,   0.000000000000000D+00, &
   4.712106885289491D-04,   0.000000000000000D+00,   1.123522945082391D-04, &
   0.000000000000000D+00,   6.319851026023241D-07,   0.000000000000000D+00, &
   5.250368725906707D-06,   0.000000000000000D+00,   2.227153524735248D-06, &
   0.000000000000000D+00,   6.984583934693264D-07,   0.000000000000000D+00, &
  -1.630252806885567D-06,   0.000000000000000D+00,   1.810918164852495D-06, &
   0.000000000000000D+00,  -2.525381129024642D-05,   0.000000000000000D+00, &
  -2.177340237026636D-04,   0.000000000000000D+00,  -2.431670864303329D-04, &
   0.000000000000000D+00,   1.583548726191047D-04,   0.000000000000000D+00, &
   7.058262833795528D-05,   0.000000000000000D+00,   9.114220604508244D-06, &
   0.000000000000000D+00,  -2.639227455790342D-07,   0.000000000000000D+00, &
  -1.489369245321459D-06,   0.000000000000000D+00,  -1.775872287288028D-06 /)
lem(:,269) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.736428565805347D-05,  -9.232748978689187D-05, &
   3.123744182020544D-04,   2.659392436184982D-04,  -5.219963801956276D-04, &
  -8.234354953737063D-04,  -3.916947146540276D-04,   1.153574116190469D-03, &
   2.739850240805591D-03,   1.302870404170175D-03,  -5.068737935425852D-03, &
  -7.413477639712860D-03,   6.975907656888894D-03,   1.193368924451464D-02, &
  -9.478619900516579D-03,  -8.651719224486324D-03,   9.832036309854641D-03, &
   1.224920685113235D-03,  -5.920439882057077D-03,   2.211431514163259D-03, &
   1.771113812544137D-03,   0.000000000000000D+00,   2.736428565805581D-05, &
   9.232748978688943D-05,  -3.123744182020612D-04,  -2.659392436184973D-04, &
   5.219963801956310D-04,   8.234354953737072D-04,   3.916947146540278D-04, &
  -1.153574116190473D-03,  -2.739850240805592D-03,  -1.302870404170170D-03, &
   5.068737935425839D-03,   7.413477639712854D-03,  -6.975907656888879D-03, &
  -1.193368924451464D-02,   9.478619900516606D-03,   8.651719224486317D-03, &
  -9.832036309854644D-03,  -1.224920685113236D-03,   5.920439882057091D-03, &
  -2.211431514163267D-03,  -1.771113812544134D-03,   0.000000000000000D+00, &
   1.368214282902316D-05,   0.000000000000000D+00,   5.015888274868128D-06, &
   0.000000000000000D+00,  -5.104123664180167D-05,   0.000000000000000D+00, &
  -2.477576934666115D-05,   0.000000000000000D+00,   2.629920722979211D-05, &
   0.000000000000000D+00,   1.555118412413053D-05,   0.000000000000000D+00, &
   1.422861129014091D-05,   0.000000000000000D+00,   1.205305730631416D-05, &
   0.000000000000000D+00,  -1.755000432141532D-05,   0.000000000000000D+00, &
  -4.675037911154501D-05,   0.000000000000000D+00,  -6.967872365134377D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.081668365286650D-06, &
   0.000000000000000D+00,  -1.793196738330070D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.991262961598747D-06,   0.000000000000000D+00, &
  -1.622962659310291D-08,   0.000000000000000D+00,   4.031542816922381D-06, &
   0.000000000000000D+00,   5.745661216318299D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.449342693393904D-05,   0.000000000000000D+00, &
  -1.515813000354034D-06,   0.000000000000000D+00,   1.534436845571682D-06, &
   0.000000000000000D+00,  -1.812285818153171D-06,   0.000000000000000D+00, &
  -4.132350453082401D-06,   0.000000000000000D+00,   4.278174142633099D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.811488472489454D-05, &
   0.000000000000000D+00,   8.308536150604466D-06,   0.000000000000000D+00, &
  -8.885581088009923D-07,   0.000000000000000D+00,  -7.230960023145379D-07, &
   0.000000000000000D+00,  -5.498817141160832D-06,   0.000000000000000D+00, &
   1.401462033320932D-05,   0.000000000000000D+00,   1.234763675448716D-06, &
   0.000000000000000D+00,  -4.784348997478507D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.351444101242739D-05,   0.000000000000000D+00, &
  -2.994639012669679D-05,   0.000000000000000D+00,   2.906157890866794D-06, &
   0.000000000000000D+00,   5.490190387853140D-07,   0.000000000000000D+00, &
  -1.058910771277853D-06,   0.000000000000000D+00,   4.971333903035614D-06, &
   0.000000000000000D+00,  -2.356837253144726D-05,   0.000000000000000D+00, &
  -1.825470033957734D-05,   0.000000000000000D+00,   1.412765026244330D-06, &
   0.000000000000000D+00,   1.311222112806028D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.412877781885074D-05,   0.000000000000000D+00, &
   6.785185663490725D-05,   0.000000000000000D+00,  -1.779603948386414D-07, &
   0.000000000000000D+00,  -6.257182036659439D-06,   0.000000000000000D+00, &
   5.845617468058177D-07,   0.000000000000000D+00,   1.323234051913271D-06, &
   0.000000000000000D+00,   5.291159682054524D-06,   0.000000000000000D+00, &
   1.369513142753734D-05,   0.000000000000000D+00,   8.630914696820814D-05, &
   0.000000000000000D+00,   5.489763718957549D-06,   0.000000000000000D+00, &
  -2.232394889233928D-06,   0.000000000000000D+00,   1.195755250189854D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.016048972262819D-05, &
   0.000000000000000D+00,  -9.433599401151117D-05,   0.000000000000000D+00, &
  -1.004890209114343D-04,   0.000000000000000D+00,   4.779361312832277D-05, &
   0.000000000000000D+00,   5.777610368685016D-06,   0.000000000000000D+00, &
  -2.245943466066908D-07,   0.000000000000000D+00,  -2.862561525928940D-07, &
   0.000000000000000D+00,  -5.741201724533210D-06,   0.000000000000000D+00, &
   1.518657769008571D-05,   0.000000000000000D+00,  -7.675153363896602D-05, &
   0.000000000000000D+00,  -1.497684398357018D-04,   0.000000000000000D+00, &
   2.544758240418010D-06,   0.000000000000000D+00,   9.871582589486363D-07, &
   0.000000000000000D+00,  -2.488173656619604D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.052680795952040D-05,   0.000000000000000D+00, &
   7.973390460431423D-05,   0.000000000000000D+00,   2.171809173897918D-04, &
   0.000000000000000D+00,   1.136062778153181D-04,   0.000000000000000D+00, &
  -6.763579641807812D-05,   0.000000000000000D+00,  -6.889938635423535D-06, &
   0.000000000000000D+00,  -1.222162056029019D-06,   0.000000000000000D+00, &
  -7.828317146815452D-07,   0.000000000000000D+00,  -6.396720864330120D-06, &
   0.000000000000000D+00,  -3.190913152792674D-05,   0.000000000000000D+00, &
  -8.319303347659957D-05,   0.000000000000000D+00,   1.626717578866876D-04, &
   0.000000000000000D+00,   1.693125079148792D-04,   0.000000000000000D+00, &
   7.966704685944969D-06,   0.000000000000000D+00,   2.790746112579231D-06, &
   0.000000000000000D+00,   1.180701864570030D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.418589866637483D-05,   0.000000000000000D+00, &
  -4.025306082889557D-05,   0.000000000000000D+00,  -1.157818738927336D-04, &
   0.000000000000000D+00,  -3.586172585534414D-04,   0.000000000000000D+00, &
  -1.112361297353906D-04,   0.000000000000000D+00,   3.941841728083481D-05, &
   0.000000000000000D+00,   3.087396195832210D-06,   0.000000000000000D+00, &
   1.405007503285515D-08,   0.000000000000000D+00,  -4.480065684461049D-07, &
   0.000000000000000D+00,   1.155844683315836D-05,   0.000000000000000D+00, &
   2.256558845505964D-05,   0.000000000000000D+00,   1.095266038480532D-04, &
   0.000000000000000D+00,   1.747923268600898D-04,   0.000000000000000D+00, &
  -1.942269886900826D-04,   0.000000000000000D+00,  -1.303272515453218D-04, &
   0.000000000000000D+00,  -1.708687360300404D-05,   0.000000000000000D+00, &
  -3.977207603033184D-06,   0.000000000000000D+00,  -6.828989312729481D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.879274662189761D-06, &
   0.000000000000000D+00,   1.284792534459096D-05,   0.000000000000000D+00, &
  -4.179356067140862D-06,   0.000000000000000D+00,   1.285468891246864D-04, &
   0.000000000000000D+00,   4.034719442027376D-04,   0.000000000000000D+00, &
   1.009947715747169D-04,   0.000000000000000D+00,   3.084556203926777D-06, &
   0.000000000000000D+00,   5.192469322021370D-06,   0.000000000000000D+00, &
   2.569444866376371D-06,   0.000000000000000D+00,   1.984878000179688D-06, &
   0.000000000000000D+00,  -5.858995500855205D-06,   0.000000000000000D+00, &
  -7.757672176580976D-06,   0.000000000000000D+00,  -2.342564740381300D-05, &
   0.000000000000000D+00,  -1.846978126254659D-04,   0.000000000000000D+00, &
  -2.076926546641004D-04,   0.000000000000000D+00,   1.369768210097738D-04, &
   0.000000000000000D+00,   6.253373703375017D-05,   0.000000000000000D+00, &
   8.395527721708592D-06,   0.000000000000000D+00,  -7.513211999102635D-07, &
   0.000000000000000D+00,  -1.960809740361586D-06,   0.000000000000000D+00 /)
lem(:,270) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.307657589576818D-05,  -1.277989735717054D-04,  -1.277366747810731D-04, &
   7.326130005941249D-05,  -3.811784620575297D-06,   3.093098723660578D-04, &
   7.335316360245919D-04,   4.030979724166309D-04,  -1.279244016110396D-03, &
  -2.837458235821611D-03,  -1.213096331466039D-03,   4.863217555920816D-03, &
   6.260823656204524D-03,  -6.052350206358142D-03,  -8.191882508696294D-03, &
   7.949633888778378D-03,   3.676847166577760D-03,  -7.501624250447475D-03, &
   2.106524753669267D-03,   3.138578484505892D-03,  -3.232642497754558D-03, &
   3.209876192735984D-04,   5.307657589577294D-05,  -1.277989735717073D-04, &
  -1.277366747810722D-04,   7.326130005940267D-05,  -3.811784620573260D-06, &
   3.093098723660673D-04,   7.335316360245921D-04,   4.030979724166212D-04, &
  -1.279244016110389D-03,  -2.837458235821599D-03,  -1.213096331466045D-03, &
   4.863217555920804D-03,   6.260823656204525D-03,  -6.052350206358140D-03, &
  -8.191882508696305D-03,   7.949633888778385D-03,   3.676847166577759D-03, &
  -7.501624250447482D-03,   2.106524753669260D-03,   3.138578484505901D-03, &
  -3.232642497754558D-03,   3.209876192735972D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.149292218123464D-04,   0.000000000000000D+00, &
   1.091121705527848D-04,   0.000000000000000D+00,   7.803240105774279D-06, &
   0.000000000000000D+00,  -7.093488401575535D-06,   0.000000000000000D+00, &
  -1.276516776166843D-05,   0.000000000000000D+00,  -3.395927144438671D-05, &
   0.000000000000000D+00,  -3.010358009880306D-05,   0.000000000000000D+00, &
  -8.073145769940777D-07,   0.000000000000000D+00,   6.729569509706442D-05, &
   0.000000000000000D+00,   4.809892252695678D-05,   0.000000000000000D+00, &
  -5.270942401560007D-06,   3.090391620034274D-07,   0.000000000000000D+00, &
   1.331259905791526D-05,   0.000000000000000D+00,  -2.892764684865404D-06, &
  -1.326485608154377D-06,   0.000000000000000D+00,   1.584379109586482D-06, &
   0.000000000000000D+00,  -2.765599544166694D-05,   0.000000000000000D+00, &
  -3.548592652417539D-07,   0.000000000000000D+00,   1.468884742573553D-06, &
  -3.109218989151876D-06,   0.000000000000000D+00,  -6.246714947469227D-06, &
   0.000000000000000D+00,   2.212144155338389D-07,   0.000000000000000D+00, &
   3.253273958765912D-05,   0.000000000000000D+00,   6.635893527969846D-06, &
   0.000000000000000D+00,  -4.392305283795383D-06,   0.000000000000000D+00, &
   2.389274746397436D-06,   1.892391684200440D-05,   0.000000000000000D+00, &
   1.611803705885071D-05,   0.000000000000000D+00,   4.519211538229660D-06, &
   0.000000000000000D+00,  -1.393099902989085D-06,   0.000000000000000D+00, &
  -1.549562017766971D-05,   0.000000000000000D+00,  -1.364401620739650D-05, &
   0.000000000000000D+00,   1.023904644287294D-05,   0.000000000000000D+00, &
   2.055373944010564D-06,   0.000000000000000D+00,  -1.397829164400969D-06, &
  -3.181603125337392D-05,   0.000000000000000D+00,  -2.628111377632027D-05, &
   0.000000000000000D+00,  -2.238674832294910D-05,   0.000000000000000D+00, &
   3.738494858852573D-06,   0.000000000000000D+00,  -3.189720773104326D-07, &
   0.000000000000000D+00,  -1.713847620745557D-05,   0.000000000000000D+00, &
   8.375026813910205D-06,   0.000000000000000D+00,  -1.479781010773592D-05, &
   0.000000000000000D+00,  -1.748036678273068D-05,   0.000000000000000D+00, &
   1.430968461028530D-06,   0.000000000000000D+00,  -6.642303920364193D-07, &
   2.364090823537926D-05,   0.000000000000000D+00,   3.185030433950663D-05, &
   0.000000000000000D+00,   5.464018909126179D-05,   0.000000000000000D+00, &
   9.019414063792612D-07,   0.000000000000000D+00,  -7.275755709281037D-06, &
   0.000000000000000D+00,   1.227559323867717D-06,   0.000000000000000D+00, &
   4.451711953030083D-05,   0.000000000000000D+00,   9.117611101902769D-06, &
   0.000000000000000D+00,   8.297182942004424D-06,   0.000000000000000D+00, &
   7.070687944454378D-05,   0.000000000000000D+00,   8.714616585722216D-06, &
   0.000000000000000D+00,  -2.318005435852889D-06,   0.000000000000000D+00, &
   2.805040964020591D-07,  -7.078161213520282D-06,   0.000000000000000D+00, &
  -3.190518858708481D-05,   0.000000000000000D+00,  -7.735945574087309D-05, &
   0.000000000000000D+00,  -8.553375478635304D-05,   0.000000000000000D+00, &
   3.905782588543808D-05,   0.000000000000000D+00,   7.648170256031733D-06, &
   0.000000000000000D+00,   3.415989773524664D-07,   0.000000000000000D+00, &
  -4.795297523572742D-05,   0.000000000000000D+00,  -1.732501795717659D-05, &
   0.000000000000000D+00,   9.320310309623323D-06,   0.000000000000000D+00, &
  -6.197434178259917D-05,   0.000000000000000D+00,  -1.296204395388156D-04, &
   0.000000000000000D+00,  -2.837859104857746D-06,   0.000000000000000D+00, &
   6.843129546973277D-07,   0.000000000000000D+00,   8.602092609352881D-08, &
   2.493630725806013D-06,   0.000000000000000D+00,   2.336915434316816D-05, &
   0.000000000000000D+00,   6.419366572315172D-05,   0.000000000000000D+00, &
   1.820046098464866D-04,   0.000000000000000D+00,   1.008419527919584D-04, &
   0.000000000000000D+00,  -5.697790546083242D-05,   0.000000000000000D+00, &
  -8.506027114565057D-06,   0.000000000000000D+00,  -1.966619428870632D-06, &
   0.000000000000000D+00,   2.934313094183860D-05,   0.000000000000000D+00, &
   9.289064729801245D-06,   0.000000000000000D+00,  -2.045641724916413D-05, &
   0.000000000000000D+00,  -6.724885379305015D-05,   0.000000000000000D+00, &
   1.378130624497006D-04,   0.000000000000000D+00,   1.508206533927301D-04, &
   0.000000000000000D+00,   1.190239448271889D-05,   0.000000000000000D+00, &
   3.383207181358802D-06,   0.000000000000000D+00,   2.223088943142765D-06, &
  -4.637824921922318D-06,   0.000000000000000D+00,  -8.684065724468918D-06, &
   0.000000000000000D+00,  -2.851765515259088D-05,   0.000000000000000D+00, &
  -9.769131191524808D-05,   0.000000000000000D+00,  -3.083335357310107D-04, &
   0.000000000000000D+00,  -1.017647794012816D-04,   0.000000000000000D+00, &
   3.268533547065315D-05,   0.000000000000000D+00,   3.294610939561013D-06, &
   0.000000000000000D+00,  -4.097178384050260D-07,   0.000000000000000D+00, &
  -9.315977617543752D-06,   0.000000000000000D+00,  -1.295367179334134D-06, &
   0.000000000000000D+00,   1.589889500461351D-05,   0.000000000000000D+00, &
   8.927551818483453D-05,   0.000000000000000D+00,   1.478552583046517D-04, &
   0.000000000000000D+00,  -1.683221137887486D-04,   0.000000000000000D+00, &
  -1.181787200049683D-04,   0.000000000000000D+00,  -1.803317654784666D-05, &
   0.000000000000000D+00,  -4.366190110179097D-06,   0.000000000000000D+00, &
  -2.010483586232932D-06,   3.229099118738510D-06,   0.000000000000000D+00, &
  -7.889081869957405D-07,   0.000000000000000D+00,   3.791190381447498D-06, &
   0.000000000000000D+00,  -4.183917478400995D-06,   0.000000000000000D+00, &
   1.123522945082391D-04,   0.000000000000000D+00,   3.535281970796041D-04, &
   0.000000000000000D+00,   9.369100125572141D-05,   0.000000000000000D+00, &
   4.901071050098324D-06,   0.000000000000000D+00,   6.006314981883220D-06, &
   0.000000000000000D+00,   3.589816521535874D-06,   0.000000000000000D+00, &
   8.547338461242150D-07,   0.000000000000000D+00,   4.632889098278034D-07, &
   0.000000000000000D+00,  -6.739255939648318D-06,   0.000000000000000D+00, &
  -2.104765297285690D-05,   0.000000000000000D+00,  -1.577614266851603D-04, &
   0.000000000000000D+00,  -1.795026896038119D-04,   0.000000000000000D+00, &
   1.202014661556055D-04,   0.000000000000000D+00,   5.687238131440269D-05, &
   0.000000000000000D+00,   7.819287321965664D-06,   0.000000000000000D+00, &
  -9.420917338988244D-07,   0.000000000000000D+00,  -3.788815865392385D-06 /)
lem(:,271) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.789668518997873D-05,   3.249104375337216D-04, &
   4.165159684368001D-04,   5.717752747720806D-05,  -6.177915385981686D-04, &
  -1.063417228035267D-03,  -6.092682866594083D-04,   1.043098715603481D-03, &
   2.578922599657523D-03,   1.472092802924334D-03,  -3.408045197653572D-03, &
  -4.777170772303774D-03,   3.934283753381495D-03,   4.968453101144872D-03, &
  -5.198259001541403D-03,  -7.525694646553572D-04,   4.090713799611313D-03, &
  -2.955509921155468D-03,  -3.422715023539647D-04,   2.264997231391467D-03, &
  -1.571239084087195D-03,   0.000000000000000D+00,   3.789668518997875D-05, &
  -3.249104375337184D-04,  -4.165159684368038D-04,  -5.717752747720849D-05, &
   6.177915385981744D-04,   1.063417228035267D-03,   6.092682866594016D-04, &
  -1.043098715603483D-03,  -2.578922599657520D-03,  -1.472092802924333D-03, &
   3.408045197653567D-03,   4.777170772303768D-03,  -3.934283753381494D-03, &
  -4.968453101144864D-03,   5.198259001541404D-03,   7.525694646553582D-04, &
  -4.090713799611330D-03,   2.955509921155465D-03,   3.422715023539697D-04, &
  -2.264997231391460D-03,   1.571239084087192D-03,   0.000000000000000D+00, &
   1.894834259498961D-05,   0.000000000000000D+00,   1.351188330699623D-05, &
   0.000000000000000D+00,  -6.964537360709226D-05,   0.000000000000000D+00, &
  -7.627842615174438D-05,   0.000000000000000D+00,  -8.982434785297268D-06, &
   0.000000000000000D+00,   3.872947466117457D-05,   0.000000000000000D+00, &
   6.992249267280667D-05,   0.000000000000000D+00,   5.106300780427987D-05, &
   0.000000000000000D+00,  -7.118600235728083D-05,   0.000000000000000D+00, &
  -1.238162236200652D-04,   0.000000000000000D+00,   5.254701507478027D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.497998012133864D-06, &
   0.000000000000000D+00,   4.720582629211378D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.212908999932379D-06,   0.000000000000000D+00, &
   2.192384496162285D-06,   0.000000000000000D+00,  -8.443207133725538D-06, &
   0.000000000000000D+00,  -1.949887918668203D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.598103956431375D-06,   0.000000000000000D+00, &
  -7.138865134835070D-06,   0.000000000000000D+00,   1.771773797070784D-07, &
   0.000000000000000D+00,   8.410618463298417D-06,   0.000000000000000D+00, &
   5.126558491789411D-06,   0.000000000000000D+00,  -3.006448165904639D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.296355646194702D-06, &
   0.000000000000000D+00,   1.456643704368719D-05,   0.000000000000000D+00, &
   5.777238076705587D-06,   0.000000000000000D+00,  -1.784406451212471D-06, &
   0.000000000000000D+00,  -3.464737136987737D-06,   0.000000000000000D+00, &
  -8.176396201858162D-06,   0.000000000000000D+00,   7.782337883029635D-06, &
   0.000000000000000D+00,   3.010642327825725D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.059035515038112D-06,   0.000000000000000D+00, &
  -1.913834648394661D-05,   0.000000000000000D+00,  -2.106837605139940D-05, &
   0.000000000000000D+00,   2.365677391793167D-06,   0.000000000000000D+00, &
  -2.436311156580268D-07,   0.000000000000000D+00,  -3.791925726578202D-06, &
   0.000000000000000D+00,   4.673266208813699D-06,   0.000000000000000D+00, &
  -1.215900843371170D-05,   0.000000000000000D+00,  -1.723081853058295D-05, &
   0.000000000000000D+00,   1.159495793382125D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.931721427297478D-06,   0.000000000000000D+00, &
   2.096050348725599D-05,   0.000000000000000D+00,   4.534185147470755D-05, &
   0.000000000000000D+00,   4.676926113452024D-06,   0.000000000000000D+00, &
  -9.029978252539761D-06,   0.000000000000000D+00,   1.854929093627056D-06, &
   0.000000000000000D+00,   1.109715660968724D-05,   0.000000000000000D+00, &
   4.992709015375541D-06,   0.000000000000000D+00,   7.076994903198330D-06, &
   0.000000000000000D+00,   6.103964919596442D-05,   0.000000000000000D+00, &
   1.314455698077465D-05,   0.000000000000000D+00,  -3.786047330192701D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.638679172999461D-06, &
   0.000000000000000D+00,  -2.325476609176459D-05,   0.000000000000000D+00, &
  -6.199423243983808D-05,   0.000000000000000D+00,  -7.567297100655206D-05, &
   0.000000000000000D+00,   3.083350572323679D-05,   0.000000000000000D+00, &
   1.163516344742597D-05,   0.000000000000000D+00,   6.502103240738651D-07, &
   0.000000000000000D+00,  -1.312951462632104D-05,   0.000000000000000D+00, &
  -9.589302209008892D-06,   0.000000000000000D+00,   8.055490605335464D-06, &
   0.000000000000000D+00,  -5.231710116300109D-05,   0.000000000000000D+00, &
  -1.171643056128585D-04,   0.000000000000000D+00,  -8.838763922786746D-06, &
   0.000000000000000D+00,   1.376465021237273D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.063115984348157D-05,   0.000000000000000D+00, &
   2.192135319731773D-05,   0.000000000000000D+00,   5.288053270094156D-05, &
   0.000000000000000D+00,   1.548650903514806D-04,   0.000000000000000D+00, &
   9.346814735262552D-05,   0.000000000000000D+00,  -4.892647901745505D-05, &
   0.000000000000000D+00,  -1.117836753240267D-05,   0.000000000000000D+00, &
  -3.881326164748553D-06,   0.000000000000000D+00,   7.112693131262178D-06, &
   0.000000000000000D+00,   5.637342399934976D-06,   0.000000000000000D+00, &
  -1.758142991938069D-05,   0.000000000000000D+00,  -5.682993993194827D-05, &
   0.000000000000000D+00,   1.199951378303614D-04,   0.000000000000000D+00, &
   1.401607530482283D-04,   0.000000000000000D+00,   1.609642750156279D-05, &
   0.000000000000000D+00,   4.852288008790119D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.639732403029280D-06,   0.000000000000000D+00, &
  -1.363581536479836D-05,   0.000000000000000D+00,  -2.627002090610527D-05, &
   0.000000000000000D+00,  -8.474602224743460D-05,   0.000000000000000D+00, &
  -2.706294732650726D-04,   0.000000000000000D+00,  -9.687627556947392D-05, &
   0.000000000000000D+00,   2.738429611879531D-05,   0.000000000000000D+00, &
   3.210923617676067D-06,   0.000000000000000D+00,  -6.623250917563934D-07, &
   0.000000000000000D+00,  -3.818963984882561D-07,   0.000000000000000D+00, &
  -1.597316517777375D-06,   0.000000000000000D+00,   1.245330859071476D-05, &
   0.000000000000000D+00,   7.632925828671482D-05,   0.000000000000000D+00, &
   1.282363664764739D-04,   0.000000000000000D+00,  -1.497633927187970D-04, &
   0.000000000000000D+00,  -1.110396775294163D-04,   0.000000000000000D+00, &
  -1.984394075055498D-05,   0.000000000000000D+00,  -5.612179673682647D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.363669284465252D-06, &
   0.000000000000000D+00,   5.216537896193591D-06,   0.000000000000000D+00, &
   6.069593893930727D-06,   0.000000000000000D+00,  -8.794031090607202D-07, &
   0.000000000000000D+00,   1.009947715747169D-04,   0.000000000000000D+00, &
   3.165423994475130D-04,   0.000000000000000D+00,   8.981124887994176D-05, &
   0.000000000000000D+00,   7.713456163677985D-06,   0.000000000000000D+00, &
   7.459457572684182D-06,   0.000000000000000D+00,   8.500689241421090D-06, &
   0.000000000000000D+00,  -1.172586994916132D-06,   0.000000000000000D+00, &
   7.841833648284720D-07,   0.000000000000000D+00,  -3.003872029175900D-06, &
   0.000000000000000D+00,  -1.826533039927326D-05,   0.000000000000000D+00, &
  -1.384039661920313D-04,   0.000000000000000D+00,  -1.583778562627284D-04, &
   0.000000000000000D+00,   1.081526806774998D-04,   0.000000000000000D+00, &
   5.312710832106343D-05,   0.000000000000000D+00,   7.961704215386985D-06, &
   0.000000000000000D+00,  -1.939707519595813D-06,   0.000000000000000D+00 /)
lem(:,272) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.373231011759990D-04,  -1.956970692386510D-04,  -2.975299651300409D-04, &
  -2.384179672821541D-04,   4.634631907081945D-05,   5.184525806242690D-04, &
   9.040159473406158D-04,   6.563827575986254D-04,  -5.398016344593464D-04, &
  -1.969037871702357D-03,  -1.733156244869202D-03,   1.690639172802014D-03, &
   3.600938476913295D-03,  -1.690513572368749D-03,  -2.910613139730545D-03, &
   2.697716238155838D-03,  -1.902799253574251D-04,  -1.576771908953254D-03, &
   2.007721675227284D-03,  -1.004129281050174D-03,  -6.389255475214026D-04, &
   1.373869770692903D-03,  -1.373231011759932D-04,  -1.956970692386510D-04, &
  -2.975299651300503D-04,  -2.384179672821507D-04,   4.634631907082609D-05, &
   5.184525806242695D-04,   9.040159473406146D-04,   6.563827575986294D-04, &
  -5.398016344593489D-04,  -1.969037871702358D-03,  -1.733156244869201D-03, &
   1.690639172802008D-03,   3.600938476913298D-03,  -1.690513572368750D-03, &
  -2.910613139730544D-03,   2.697716238155833D-03,  -1.902799253574262D-04, &
  -1.576771908953259D-03,   2.007721675227286D-03,  -1.004129281050168D-03, &
  -6.389255475214063D-04,   1.373869770692899D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.083792820333847D-04,   0.000000000000000D+00, &
  -2.066330333479213D-04,   0.000000000000000D+00,   7.974674293061727D-05, &
   0.000000000000000D+00,   1.911895770008764D-04,   0.000000000000000D+00, &
   5.326460298599542D-05,   0.000000000000000D+00,  -9.961951261173978D-05, &
   0.000000000000000D+00,  -1.747896230829385D-04,   0.000000000000000D+00, &
   7.540211221348963D-07,   0.000000000000000D+00,   2.419213116264741D-04, &
   0.000000000000000D+00,   2.924048974484226D-05,   0.000000000000000D+00, &
   5.315045843375522D-06,   8.350981880209791D-06,   0.000000000000000D+00, &
   3.696212897747694D-06,   0.000000000000000D+00,   8.448498771177069D-06, &
  -6.464885629571788D-06,   0.000000000000000D+00,  -1.959494720490262D-06, &
   0.000000000000000D+00,  -9.687344821097315D-06,   0.000000000000000D+00, &
  -7.699902588065971D-06,   0.000000000000000D+00,  -4.184713718941853D-06, &
  -3.193023752324904D-06,   0.000000000000000D+00,   5.362920449022579D-07, &
   0.000000000000000D+00,  -6.542123825412111D-06,   0.000000000000000D+00, &
   5.911837300413260D-06,   0.000000000000000D+00,   8.563572731788455D-06, &
   0.000000000000000D+00,   6.110276050676163D-06,   0.000000000000000D+00, &
  -8.116074532411982D-06,   1.030027891395081D-05,   0.000000000000000D+00, &
   4.546357268700671D-06,   0.000000000000000D+00,   1.110944446468065D-05, &
   0.000000000000000D+00,   8.521349055656016D-06,   0.000000000000000D+00, &
   6.701809076796192D-06,   0.000000000000000D+00,  -4.186765650264330D-06, &
   0.000000000000000D+00,  -8.845153365873787D-06,   0.000000000000000D+00, &
   7.451442911610366D-06,   0.000000000000000D+00,   7.162619704812788D-06, &
  -4.156116494190726D-06,   0.000000000000000D+00,  -8.559586546576961D-06, &
   0.000000000000000D+00,  -1.517840789503505D-05,   0.000000000000000D+00, &
  -2.093774125638238D-05,   0.000000000000000D+00,   2.195025812472940D-06, &
   0.000000000000000D+00,  -1.539463375881575D-05,   0.000000000000000D+00, &
  -4.785258581947925D-06,   0.000000000000000D+00,   4.850261067695398D-06, &
   0.000000000000000D+00,  -8.802071815477001D-06,   0.000000000000000D+00, &
  -2.026258996734342D-05,   0.000000000000000D+00,   3.293559447706148D-06, &
  -1.029406809095440D-05,   0.000000000000000D+00,   6.140135095960162D-06, &
   0.000000000000000D+00,   1.779464788804446D-05,   0.000000000000000D+00, &
   3.999486513400433D-05,   0.000000000000000D+00,   8.003938897478706D-06, &
   0.000000000000000D+00,  -1.374470033904040D-05,   0.000000000000000D+00, &
   1.474887336643993D-05,   0.000000000000000D+00,   1.194285479743490D-05, &
   0.000000000000000D+00,   5.095682958064302D-06,   0.000000000000000D+00, &
   4.688992400504394D-06,   0.000000000000000D+00,   5.542887308077155D-05, &
   0.000000000000000D+00,   1.946541153186401D-05,   0.000000000000000D+00, &
  -9.050921062375390D-06,   1.749972000869274D-05,   0.000000000000000D+00, &
  -1.472757720157571D-06,   0.000000000000000D+00,  -1.860269656548919D-05, &
   0.000000000000000D+00,  -5.311985243432893D-05,   0.000000000000000D+00, &
  -7.055712236691487D-05,   0.000000000000000D+00,   2.543362615728204D-05, &
   0.000000000000000D+00,   1.851658589307912D-05,   0.000000000000000D+00, &
  -1.094226489936661D-05,   0.000000000000000D+00,  -1.131778054296302D-05, &
   0.000000000000000D+00,  -9.519218156999232D-06,   0.000000000000000D+00, &
   5.879238946655952D-06,   0.000000000000000D+00,  -4.495738013380897D-05, &
   0.000000000000000D+00,  -1.111284759532496D-04,   0.000000000000000D+00, &
  -1.644479835576124D-05,   0.000000000000000D+00,   3.739817157775428D-06, &
  -1.201923868716558D-05,   0.000000000000000D+00,   2.569128236049140D-07, &
   0.000000000000000D+00,   1.565529034755194D-05,   0.000000000000000D+00, &
   4.509149085057785D-05,   0.000000000000000D+00,   1.369703617925872D-04, &
   0.000000000000000D+00,   9.028083549276156D-05,   0.000000000000000D+00, &
  -4.339476314469161D-05,   0.000000000000000D+00,  -1.610930809330957D-05, &
   0.000000000000000D+00,   7.013741060405253D-06,   0.000000000000000D+00, &
   4.698852446959872D-06,   0.000000000000000D+00,   4.371950350220682D-06, &
   0.000000000000000D+00,  -1.360437365825148D-05,   0.000000000000000D+00, &
  -4.844936342285761D-05,   0.000000000000000D+00,   1.075681723197740D-04, &
   0.000000000000000D+00,   1.359331190887288D-04,   0.000000000000000D+00, &
   2.287497684781333D-05,   0.000000000000000D+00,   1.008046080115512D-05, &
   3.482654143076275D-06,   0.000000000000000D+00,  -2.306683584813176D-07, &
   0.000000000000000D+00,  -8.786441088291755D-06,   0.000000000000000D+00, &
  -2.214182624560110D-05,   0.000000000000000D+00,  -7.579773909290732D-05, &
   0.000000000000000D+00,  -2.449051358667865D-04,   0.000000000000000D+00, &
  -9.618234571028476D-05,   0.000000000000000D+00,   2.273089858996605D-05, &
   0.000000000000000D+00,   3.016667910501330D-06,   0.000000000000000D+00, &
  -2.887889392038571D-06,   0.000000000000000D+00,   1.714630189815976D-07, &
   0.000000000000000D+00,   1.226658840383346D-06,   0.000000000000000D+00, &
   1.116381916467259D-05,   0.000000000000000D+00,   6.605901024301085D-05, &
   0.000000000000000D+00,   1.130119176191189D-04,   0.000000000000000D+00, &
  -1.366367535939890D-04,   0.000000000000000D+00,  -1.091420100661669D-04, &
   0.000000000000000D+00,  -2.390955495877763D-05,   0.000000000000000D+00, &
  -1.413966995725609D-05,   5.780253891576613D-08,   0.000000000000000D+00, &
  -5.800771023223820D-07,   0.000000000000000D+00,   2.802342816257761D-06, &
   0.000000000000000D+00,   5.063379978245968D-06,   0.000000000000000D+00, &
   6.319851026023241D-07,   0.000000000000000D+00,   9.369100125572141D-05, &
   0.000000000000000D+00,   2.913835067056667D-04,   0.000000000000000D+00, &
   9.040987871367662D-05,   0.000000000000000D+00,   1.135359023558300D-05, &
   0.000000000000000D+00,   1.087122216339661D-05,   0.000000000000000D+00, &
   4.531684667337895D-07,   0.000000000000000D+00,  -6.097895578671173D-07, &
   0.000000000000000D+00,  -2.125662277343869D-06,   0.000000000000000D+00, &
  -4.400471211265544D-06,   0.000000000000000D+00,  -1.754303577645773D-05, &
   0.000000000000000D+00,  -1.238979783083233D-04,   0.000000000000000D+00, &
  -1.424318287906998D-04,   0.000000000000000D+00,   9.947317630357246D-05, &
   0.000000000000000D+00,   5.224367642348283D-05,   0.000000000000000D+00, &
   8.948426040676203D-06,   0.000000000000000D+00,  -2.129273875834180D-06 /)
lem(:,273) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.703441286707562D-04,   4.037819635794525D-04, &
   3.221326874396685D-04,   1.799131700087649D-04,  -2.226630539081674D-04, &
  -6.968573330971123D-04,  -7.950134552282765D-04,  -1.628616543413172D-04, &
   1.101856846491073D-03,   1.725377386460359D-03,  -2.512541034060450D-04, &
  -2.561905126092624D-03,   2.127766144834170D-04,   1.830745390220171D-03, &
  -1.082340696205420D-03,   1.834325372907926D-04,   3.854724143546503D-04, &
  -8.668481991151166D-04,   9.377822094127944D-04,  -3.484770419208628D-04, &
  -4.869282056619387D-04,   0.000000000000000D+00,  -2.703441286707538D-04, &
  -4.037819635794530D-04,  -3.221326874396678D-04,  -1.799131700087625D-04, &
   2.226630539081704D-04,   6.968573330971104D-04,   7.950134552282716D-04, &
   1.628616543413163D-04,  -1.101856846491073D-03,  -1.725377386460360D-03, &
   2.512541034060481D-04,   2.561905126092625D-03,  -2.127766144834227D-04, &
  -1.830745390220174D-03,   1.082340696205429D-03,  -1.834325372907937D-04, &
  -3.854724143546570D-04,   8.668481991151156D-04,  -9.377822094127848D-04, &
   3.484770419208645D-04,   4.869282056619345D-04,   0.000000000000000D+00, &
  -1.351720643353774D-04,   0.000000000000000D+00,  -1.515398853019700D-04, &
   0.000000000000000D+00,   3.610235199445334D-04,   0.000000000000000D+00, &
   2.412329148326283D-04,   0.000000000000000D+00,  -4.161419880951692D-04, &
   0.000000000000000D+00,  -2.829545573219867D-04,   0.000000000000000D+00, &
   3.641194118646957D-04,   0.000000000000000D+00,   3.387986146014003D-04, &
   0.000000000000000D+00,  -3.482902836319618D-04,   0.000000000000000D+00, &
  -2.841491664550914D-04,   0.000000000000000D+00,   2.762766165101593D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.068628998316513D-05, &
   0.000000000000000D+00,   3.573093457725190D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.973543926241989D-06,   0.000000000000000D+00, &
  -2.241016804371679D-06,   0.000000000000000D+00,  -3.251459586304583D-06, &
   0.000000000000000D+00,  -6.795364323794171D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.002895869787150D-08,   0.000000000000000D+00, &
  -5.919895899732180D-07,   0.000000000000000D+00,  -1.649247605911069D-05, &
   0.000000000000000D+00,   1.167009984495272D-06,   0.000000000000000D+00, &
   7.159750631222012D-06,   0.000000000000000D+00,   7.064629331992151D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.180552495272101D-06, &
   0.000000000000000D+00,   3.332965081997457D-06,   0.000000000000000D+00, &
   1.227822649087481D-05,   0.000000000000000D+00,   1.741908898247055D-05, &
   0.000000000000000D+00,   1.559715892822082D-06,   0.000000000000000D+00, &
  -3.185140417885023D-06,   0.000000000000000D+00,  -8.543498330741399D-06, &
   0.000000000000000D+00,   6.274026100430684D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.397768736765282D-06,   0.000000000000000D+00, &
  -5.150205752212546D-06,   0.000000000000000D+00,  -1.383792632320036D-05, &
   0.000000000000000D+00,  -2.236973260237828D-05,   0.000000000000000D+00, &
   7.701488606968997D-06,   0.000000000000000D+00,  -4.070627563263875D-06, &
   0.000000000000000D+00,  -2.733278390431320D-06,   0.000000000000000D+00, &
   4.371059773374262D-06,   0.000000000000000D+00,  -7.603648051882227D-06, &
   0.000000000000000D+00,  -2.573302629502150D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.349800828656869D-06,   0.000000000000000D+00, &
   4.389575952151072D-06,   0.000000000000000D+00,   1.474985565684116D-05, &
   0.000000000000000D+00,   3.738595956100997D-05,   0.000000000000000D+00, &
   1.199038455677183D-05,   0.000000000000000D+00,  -3.201898158050235D-05, &
   0.000000000000000D+00,   6.152360836215132D-06,   0.000000000000000D+00, &
   7.956137871810849D-06,   0.000000000000000D+00,   3.748073937748418D-06, &
   0.000000000000000D+00,   4.145797725124651D-06,   0.000000000000000D+00, &
   5.349361651941627D-05,   0.000000000000000D+00,   3.155134099353443D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.631314928200424D-06, &
   0.000000000000000D+00,  -2.799643423149977D-06,   0.000000000000000D+00, &
  -1.550506022079550D-05,   0.000000000000000D+00,  -4.706049035490309D-05, &
   0.000000000000000D+00,  -6.903063431367237D-05,   0.000000000000000D+00, &
   2.051424809275798D-05,   0.000000000000000D+00,   4.137502218310073D-05, &
   0.000000000000000D+00,  -7.003307630859236D-06,   0.000000000000000D+00, &
  -9.504020766117847D-06,   0.000000000000000D+00,  -7.824664634111138D-06, &
   0.000000000000000D+00,   4.596504221212211D-06,   0.000000000000000D+00, &
  -4.051566482866815D-05,   0.000000000000000D+00,  -1.119313801397589D-04, &
   0.000000000000000D+00,  -3.079304450028801D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.918581252361188D-07,   0.000000000000000D+00, &
   2.350811649328224D-06,   0.000000000000000D+00,   1.398079290042101D-05, &
   0.000000000000000D+00,   4.025821398976666D-05,   0.000000000000000D+00, &
   1.248455363751998D-04,   0.000000000000000D+00,   9.262202026339569D-05, &
   0.000000000000000D+00,  -3.969105184734989D-05,   0.000000000000000D+00, &
  -3.703265719918010D-05,   0.000000000000000D+00,   4.953608652584163D-06, &
   0.000000000000000D+00,   6.487951545575399D-06,   0.000000000000000D+00, &
   4.698366761476861D-06,   0.000000000000000D+00,  -1.136620880455156D-05, &
   0.000000000000000D+00,  -4.219416124876756D-05,   0.000000000000000D+00, &
   9.965063603376556D-05,   0.000000000000000D+00,   1.406826776983664D-04, &
   0.000000000000000D+00,   3.670980705483865D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.332206282849700D-06,   0.000000000000000D+00, &
  -1.668147698526579D-06,   0.000000000000000D+00,  -8.094091769972646D-06, &
   0.000000000000000D+00,  -2.077743483479517D-05,   0.000000000000000D+00, &
  -7.036444412053373D-05,   0.000000000000000D+00,  -2.292063613300664D-04, &
   0.000000000000000D+00,  -1.017098047831503D-04,   0.000000000000000D+00, &
   1.870595389281316D-05,   0.000000000000000D+00,   1.031734153380720D-05, &
   0.000000000000000D+00,  -1.589072100162013D-06,   0.000000000000000D+00, &
  -2.699824703387650D-06,   0.000000000000000D+00,  -6.480408638284343D-07, &
   0.000000000000000D+00,   9.341650256765796D-06,   0.000000000000000D+00, &
   5.877417971430589D-05,   0.000000000000000D+00,   1.027726172881071D-04, &
   0.000000000000000D+00,  -1.293879425691747D-04,   0.000000000000000D+00, &
  -1.145206898691573D-04,   0.000000000000000D+00,  -3.194636816253304D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.071740761268238D-06, &
   0.000000000000000D+00,   5.067835631802803D-07,   0.000000000000000D+00, &
   1.932747618926402D-06,   0.000000000000000D+00,   4.947184283086292D-06, &
   0.000000000000000D+00,   3.084556203926777D-06,   0.000000000000000D+00, &
   8.981124887994176D-05,   0.000000000000000D+00,   2.783722810644874D-04, &
   0.000000000000000D+00,   9.628490354696910D-05,   0.000000000000000D+00, &
   1.629716091240068D-05,   0.000000000000000D+00,   2.468410752470286D-05, &
   0.000000000000000D+00,  -3.616552737625497D-08,   0.000000000000000D+00, &
   9.750782105695845D-07,   0.000000000000000D+00,  -2.919188994911613D-07, &
   0.000000000000000D+00,  -3.424353894107841D-06,   0.000000000000000D+00, &
  -1.674795148564198D-05,   0.000000000000000D+00,  -1.141033171326573D-04, &
   0.000000000000000D+00,  -1.323300687053442D-04,   0.000000000000000D+00, &
   9.538662807553911D-05,   0.000000000000000D+00,   5.480166095033302D-05, &
   0.000000000000000D+00,   7.720540967828400D-06,   0.000000000000000D+00 /)
lem(:,274) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.455785955282222D-04,  -2.115962699403794D-04,  -1.516987695777749D-04, &
  -3.676208069006813D-04,  -3.622694644786036D-04,   5.236780950382938D-05, &
   4.149593368839221D-04,   7.326486343104130D-04,   6.223662091216745D-04, &
  -3.072501215878688D-04,  -1.443393670391292D-03,  -7.195109727012477D-04, &
   1.591252428286601D-03,   5.254085796731316D-04,  -1.157151281354344D-03, &
   3.272934966341982D-04,   4.200183475246609D-05,  -2.766607922145348D-05, &
   2.146959447957576D-04,  -4.557410072378389D-04,   4.558299195403553D-04, &
  -1.273601954732730D-04,  -1.455785955282166D-04,  -2.115962699403795D-04, &
  -1.516987695777831D-04,  -3.676208069006799D-04,  -3.622694644785966D-04, &
   5.236780950382891D-05,   4.149593368839133D-04,   7.326486343104149D-04, &
   6.223662091216797D-04,  -3.072501215878683D-04,  -1.443393670391297D-03, &
  -7.195109727012460D-04,   1.591252428286603D-03,   5.254085796731269D-04, &
  -1.157151281354342D-03,   3.272934966342009D-04,   4.200183475246545D-05, &
  -2.766607922145329D-05,   2.146959447957561D-04,  -4.557410072378375D-04, &
   4.558299195403555D-04,  -1.273601954732743D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.457790169824690D-05,   0.000000000000000D+00, &
   3.454937823888373D-04,   0.000000000000000D+00,   6.775922131324473D-05, &
   0.000000000000000D+00,  -8.871701475232480D-04,   0.000000000000000D+00, &
   2.199549424168102D-04,   0.000000000000000D+00,   1.385906682700681D-03, &
   0.000000000000000D+00,  -1.513336244440110D-03,   0.000000000000000D+00, &
  -1.475495898503477D-05,   0.000000000000000D+00,   9.205750964586542D-04, &
   0.000000000000000D+00,  -2.384088474186373D-04,   0.000000000000000D+00, &
   5.569201125841725D-06,   6.908374828296874D-06,   0.000000000000000D+00, &
   1.493415800056313D-06,   0.000000000000000D+00,   2.468325222231499D-06, &
  -3.253988849692027D-06,   0.000000000000000D+00,  -1.096180295399148D-06, &
   0.000000000000000D+00,  -1.659719774266934D-06,   0.000000000000000D+00, &
  -3.603131161999475D-06,   0.000000000000000D+00,  -1.676539968300051D-05, &
  -2.028715653976686D-06,   0.000000000000000D+00,  -9.888181133304323D-07, &
   0.000000000000000D+00,  -4.482330965934612D-07,   0.000000000000000D+00, &
  -3.169545135413808D-06,   0.000000000000000D+00,   2.791896892528648D-06, &
   0.000000000000000D+00,   8.045959725861847D-06,   0.000000000000000D+00, &
   1.652925180149349D-05,   3.290622929385451D-07,   0.000000000000000D+00, &
   2.638005796431304D-06,   0.000000000000000D+00,   3.216347435852935D-06, &
   0.000000000000000D+00,   8.183160187927377D-06,   0.000000000000000D+00, &
   4.923138005825862D-06,   0.000000000000000D+00,   1.265520868577601D-06, &
   0.000000000000000D+00,  -3.310040802136249D-06,   0.000000000000000D+00, &
  -8.852266046091325D-06,   0.000000000000000D+00,   2.085064401235581D-05, &
   5.952127976632868D-06,   0.000000000000000D+00,  -7.093200227700261D-07, &
   0.000000000000000D+00,  -5.412767873998047D-06,   0.000000000000000D+00, &
  -1.194613355474646D-05,   0.000000000000000D+00,  -2.498202531190891D-05, &
   0.000000000000000D+00,  -4.123260121201170D-07,   0.000000000000000D+00, &
  -4.047426605654686D-06,   0.000000000000000D+00,  -2.315369927837800D-06, &
   0.000000000000000D+00,   3.525118946399669D-06,   0.000000000000000D+00, &
  -8.064143530421678D-06,   0.000000000000000D+00,  -6.387773181459507D-05, &
  -9.498536891936983D-06,   0.000000000000000D+00,  -3.142007926016599D-06, &
   0.000000000000000D+00,   3.802826111912216D-06,   0.000000000000000D+00, &
   1.358860401206082D-05,   0.000000000000000D+00,   3.738766289003608D-05, &
   0.000000000000000D+00,   2.342520746211112D-05,   0.000000000000000D+00, &
  -4.086817401973734D-06,   0.000000000000000D+00,   3.641989750277392D-06, &
   0.000000000000000D+00,   6.785301345751179D-06,   0.000000000000000D+00, &
   4.231725108692808D-06,   0.000000000000000D+00,   3.141361068189115D-06, &
   0.000000000000000D+00,   5.668744131689977D-05,   0.000000000000000D+00, &
   7.053464895918323D-05,   8.677939223551047D-06,   0.000000000000000D+00, &
   4.418763384734989D-06,   0.000000000000000D+00,  -1.250426371822419D-06, &
   0.000000000000000D+00,  -1.392265423135514D-05,   0.000000000000000D+00, &
  -4.347328868975921D-05,   0.000000000000000D+00,  -7.362552642819425D-05, &
   0.000000000000000D+00,   1.142310559360769D-05,   0.000000000000000D+00, &
   3.361489466822727D-06,   0.000000000000000D+00,  -2.225895486357716D-06, &
   0.000000000000000D+00,  -7.318818808289693D-06,   0.000000000000000D+00, &
  -7.660543862135279D-06,   0.000000000000000D+00,   2.988343660728292D-06, &
   0.000000000000000D+00,  -3.748305755490186D-05,   0.000000000000000D+00, &
  -1.227722763904277D-04,   0.000000000000000D+00,  -6.490753236138527D-05, &
  -5.404267169140477D-06,   0.000000000000000D+00,  -2.139846107530199D-06, &
   0.000000000000000D+00,   1.269299765648961D-06,   0.000000000000000D+00, &
   1.213431075381001D-05,   0.000000000000000D+00,   3.698971811752411D-05, &
   0.000000000000000D+00,   1.197652615780521D-04,   0.000000000000000D+00, &
   1.024826521442064D-04,   0.000000000000000D+00,  -3.569513247689791D-05, &
   0.000000000000000D+00,  -5.030326014422852D-07,   0.000000000000000D+00, &
   7.721627390596786D-07,   0.000000000000000D+00,   4.108179154594096D-06, &
   0.000000000000000D+00,   4.126022402795049D-06,   0.000000000000000D+00, &
  -8.860357833588340D-06,   0.000000000000000D+00,  -3.839383760922462D-05, &
   0.000000000000000D+00,   9.723925216335020D-05,   0.000000000000000D+00, &
   1.582159589348211D-04,   0.000000000000000D+00,   8.103316290704374D-05, &
   1.954384034819455D-06,   0.000000000000000D+00,  -1.347224144111791D-08, &
   0.000000000000000D+00,  -1.949196469688150D-06,   0.000000000000000D+00, &
  -7.353591213040619D-06,   0.000000000000000D+00,  -1.973280502626783D-05, &
   0.000000000000000D+00,  -6.781207454526163D-05,   0.000000000000000D+00, &
  -2.257894643903737D-04,   0.000000000000000D+00,  -1.150816397529898D-04, &
   0.000000000000000D+00,   9.535964993652134D-06,   0.000000000000000D+00, &
  -4.649771086187222D-07,   0.000000000000000D+00,   3.302662552197573D-07, &
   0.000000000000000D+00,  -9.082326152611204D-07,   0.000000000000000D+00, &
  -9.044643294045664D-08,   0.000000000000000D+00,   8.406193606266798D-06, &
   0.000000000000000D+00,   5.414740356639941D-05,   0.000000000000000D+00, &
   9.708119130064545D-05,   0.000000000000000D+00,  -1.293679986183278D-04, &
   0.000000000000000D+00,  -1.299857629856502D-04,   0.000000000000000D+00, &
  -8.213186798411549D-05,  -2.215153121920104D-07,   0.000000000000000D+00, &
   1.769965314196430D-07,   0.000000000000000D+00,   1.404557071530829D-06, &
   0.000000000000000D+00,   2.552577575022065D-06,   0.000000000000000D+00, &
   5.250368725906707D-06,   0.000000000000000D+00,   4.901071050098324D-06, &
   0.000000000000000D+00,   9.040987871367662D-05,   0.000000000000000D+00, &
   2.791819802947033D-04,   0.000000000000000D+00,   1.095189939536081D-04, &
   0.000000000000000D+00,   3.075147992461853D-05,   0.000000000000000D+00, &
   1.540910873102190D-07,   0.000000000000000D+00,  -4.022310879059607D-07, &
   0.000000000000000D+00,  -1.833099862989341D-07,   0.000000000000000D+00, &
  -4.832712872895759D-07,   0.000000000000000D+00,  -3.541814406613975D-06, &
   0.000000000000000D+00,  -1.696953734597303D-05,   0.000000000000000D+00, &
  -1.093201367924644D-04,   0.000000000000000D+00,  -1.273113347596496D-04, &
   0.000000000000000D+00,   9.637853339132117D-05,   0.000000000000000D+00, &
   6.305580637129234D-05,   0.000000000000000D+00,   3.212185655106162D-05 /)
lem(:,275) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.330485027418766D-05,   3.499039502434624D-04, &
   4.419994168751149D-04,   1.398090906360704D-04,   1.739101487319150D-04, &
   9.906956403532987D-05,  -4.801752601209401D-04,  -7.498215959022820D-04, &
  -3.337913594140006D-04,   8.680930855230768D-04,   1.126275836149585D-03, &
  -8.042078416309590D-04,  -8.059511163757506D-04,   6.872100218476441D-04, &
  -3.594554073169567D-06,  -1.378119286883153D-04,   1.644128391638637D-05, &
   2.727929596686715D-06,   1.221830668625199D-04,  -2.372789074531203D-04, &
   2.217952327940194D-04,   0.000000000000000D+00,  -7.330485027418624D-05, &
  -3.499039502434711D-04,  -4.419994168751148D-04,  -1.398090906360609D-04, &
  -1.739101487319170D-04,  -9.906956403533796D-05,   4.801752601209428D-04, &
   7.498215959022847D-04,   3.337913594140014D-04,  -8.680930855230760D-04, &
  -1.126275836149584D-03,   8.042078416309625D-04,   8.059511163757474D-04, &
  -6.872100218476441D-04,   3.594554073176340D-06,   1.378119286883117D-04, &
  -1.644128391639253D-05,  -2.727929596686739D-06,  -1.221830668625128D-04, &
   2.372789074531219D-04,  -2.217952327940238D-04,   0.000000000000000D+00, &
  -3.665242513708752D-05,   0.000000000000000D+00,   1.179460987385351D-04, &
   0.000000000000000D+00,   2.019178267543493D-05,   0.000000000000000D+00, &
  -7.950858665979933D-04,   0.000000000000000D+00,   7.115086666386720D-04, &
   0.000000000000000D+00,   1.640482202957864D-03,   0.000000000000000D+00, &
  -4.127807568499465D-03,   0.000000000000000D+00,   4.266772584033841D-03, &
   0.000000000000000D+00,  -1.976415192650958D-03,   0.000000000000000D+00, &
  -5.379960281468562D-04,   0.000000000000000D+00,   9.247526976815786D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.897628630050546D-06, &
   0.000000000000000D+00,   2.813168096371207D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.770937829531348D-07,   0.000000000000000D+00, &
  -7.274897411616240D-06,   0.000000000000000D+00,  -2.846139320748676D-06, &
   0.000000000000000D+00,  -2.802102411477386D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.113117568195664D-06,   0.000000000000000D+00, &
   8.805446165937527D-07,   0.000000000000000D+00,   2.517048110309865D-06, &
   0.000000000000000D+00,   1.210500604308407D-06,   0.000000000000000D+00, &
   1.744529490526621D-06,   0.000000000000000D+00,   2.405150373985149D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.447762632438351D-08, &
   0.000000000000000D+00,   1.848477371560479D-06,   0.000000000000000D+00, &
   2.552430117615667D-06,   0.000000000000000D+00,   3.176167531047001D-05, &
   0.000000000000000D+00,   4.513128253316377D-07,   0.000000000000000D+00, &
   7.562379746401799D-07,   0.000000000000000D+00,  -9.357932502578219D-07, &
   0.000000000000000D+00,  -7.266396344600874D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.664391322579362D-06,   0.000000000000000D+00, &
  -1.291127516255197D-06,   0.000000000000000D+00,  -4.963125672685721D-06, &
   0.000000000000000D+00,  -1.654922238768622D-05,   0.000000000000000D+00, &
  -7.273294412638321D-05,   0.000000000000000D+00,  -6.869111274043034D-07, &
   0.000000000000000D+00,  -2.404793014946834D-06,   0.000000000000000D+00, &
  -2.770402371971401D-06,   0.000000000000000D+00,   4.050832712304216D-06, &
   0.000000000000000D+00,   1.059174169504446D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.068576234213195D-06,   0.000000000000000D+00, &
  -1.464331179458972D-06,   0.000000000000000D+00,   3.714492120229656D-06, &
   0.000000000000000D+00,   1.328676636586768D-05,   0.000000000000000D+00, &
   4.427793442232718D-05,   0.000000000000000D+00,   4.794092383244523D-05, &
   0.000000000000000D+00,   4.369381427226944D-07,   0.000000000000000D+00, &
   3.029285569845636D-06,   0.000000000000000D+00,   6.323206314942813D-06, &
   0.000000000000000D+00,   3.917343298557616D-06,   0.000000000000000D+00, &
   1.018701268717415D-07,   0.000000000000000D+00,   5.456775852592886D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.604306114473789D-06, &
   0.000000000000000D+00,   3.263910305953127D-06,   0.000000000000000D+00, &
  -1.673586988758205D-06,   0.000000000000000D+00,  -1.244698685761121D-05, &
   0.000000000000000D+00,  -4.292964107673270D-05,   0.000000000000000D+00, &
  -8.336502073824147D-05,   0.000000000000000D+00,   5.103901439563163D-05, &
   0.000000000000000D+00,  -1.046533543656542D-06,   0.000000000000000D+00, &
  -3.018230524485836D-06,   0.000000000000000D+00,  -7.002125323612567D-06, &
   0.000000000000000D+00,  -7.925630517438175D-06,   0.000000000000000D+00, &
   2.456445217266440D-06,   0.000000000000000D+00,  -3.702618111530187D-05, &
   0.000000000000000D+00,  -1.508433630852731D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.413228568056092D-06,   0.000000000000000D+00, &
  -2.491224169766631D-06,   0.000000000000000D+00,   1.219713417624751D-06, &
   0.000000000000000D+00,   1.113389311643295D-05,   0.000000000000000D+00, &
   3.600375888608327D-05,   0.000000000000000D+00,   1.226799174946055D-04, &
   0.000000000000000D+00,   1.202981155273743D-04,   0.000000000000000D+00, &
  -1.146531782116288D-04,   0.000000000000000D+00,   1.445581719608790D-06, &
   0.000000000000000D+00,   2.017634052639688D-06,   0.000000000000000D+00, &
   4.245360803655866D-06,   0.000000000000000D+00,   4.939835790847869D-06, &
   0.000000000000000D+00,  -7.596346842355203D-06,   0.000000000000000D+00, &
  -3.662205816775150D-05,   0.000000000000000D+00,   1.032570962822611D-04, &
   0.000000000000000D+00,   2.073404326364890D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.760394463218963D-07,   0.000000000000000D+00, &
   9.919488549163183D-07,   0.000000000000000D+00,  -1.125537221032533D-06, &
   0.000000000000000D+00,  -6.979560919618401D-06,   0.000000000000000D+00, &
  -1.964605599843557D-05,   0.000000000000000D+00,  -6.974071543768337D-05, &
   0.000000000000000D+00,  -2.373951463869292D-04,   0.000000000000000D+00, &
  -1.392844665699516D-04,   0.000000000000000D+00,   5.517441504134404D-05, &
   0.000000000000000D+00,  -7.749978215128054D-07,   0.000000000000000D+00, &
  -7.313205152740576D-07,   0.000000000000000D+00,  -1.021241028291210D-06, &
   0.000000000000000D+00,  -6.474446233140101D-07,   0.000000000000000D+00, &
   7.756542601462762D-06,   0.000000000000000D+00,   5.245914519737911D-05, &
   0.000000000000000D+00,   9.647325355584610D-05,   0.000000000000000D+00, &
  -1.402998250870352D-04,   0.000000000000000D+00,  -1.637154897141845D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.713984697133120D-07, &
   0.000000000000000D+00,  -3.678197920914980D-07,   0.000000000000000D+00, &
   4.087641703369801D-07,   0.000000000000000D+00,   2.282942447494380D-06, &
   0.000000000000000D+00,   5.192469322021370D-06,   0.000000000000000D+00, &
   7.713456163677985D-06,   0.000000000000000D+00,   9.628490354696910D-05, &
   0.000000000000000D+00,   2.985365572095028D-04,   0.000000000000000D+00, &
   1.310433610559568D-04,   0.000000000000000D+00,   6.054540937019465D-05, &
   0.000000000000000D+00,   9.943171801944118D-08,   0.000000000000000D+00, &
   1.989776777121191D-07,   0.000000000000000D+00,  -2.703435741324592D-07, &
   0.000000000000000D+00,  -6.337870078690225D-07,   0.000000000000000D+00, &
  -3.399491856418174D-06,   0.000000000000000D+00,  -1.832319074097136D-05, &
   0.000000000000000D+00,  -1.094534084148293D-04,   0.000000000000000D+00, &
  -1.291577477879766D-04,   0.000000000000000D+00,   1.057550794881416D-04, &
   0.000000000000000D+00,   6.281820705151257D-05,   0.000000000000000D+00 /)
lem(:,276) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -1.827283201131882D-05,  -1.910605140125611D-05,   5.136641702326943D-05, &
   7.659479409828247D-05,  -3.228972563853270D-04,  -6.217081161685184D-04, &
  -1.252912188610508D-04,   2.205972319495875D-04,   4.675234724185957D-04, &
   7.602604372772551D-04,  -2.108235800573346D-04,  -1.162294196455502D-03, &
   1.834273176060731D-04,   7.980859942067257D-04,  -3.783899565135351D-04, &
  -1.295052353066428D-04,   1.463629600138583D-04,  -3.552980732570401D-05, &
  -1.134733391093444D-05,  -8.290762553844317D-06,   6.887721223875745D-05, &
  -1.171325795375154D-04,  -1.827283201131717D-05,  -1.910605140124735D-05, &
   5.136641702326857D-05,   7.659479409827846D-05,  -3.228972563853192D-04, &
  -6.217081161685164D-04,  -1.252912188610630D-04,   2.205972319495903D-04, &
   4.675234724186021D-04,   7.602604372772559D-04,  -2.108235800573405D-04, &
  -1.162294196455503D-03,   1.834273176060765D-04,   7.980859942067204D-04, &
  -3.783899565135364D-04,  -1.295052353066369D-04,   1.463629600138556D-04, &
  -3.552980732570719D-05,  -1.134733391093175D-05,  -8.290762553842003D-06, &
   6.887721223875698D-05,  -1.171325795375171D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.999033005732695D-06,   0.000000000000000D+00, &
   5.895295105843810D-05,   0.000000000000000D+00,  -3.488250217520470D-04, &
   0.000000000000000D+00,   4.698924763915544D-04,   0.000000000000000D+00, &
   1.338942741334868D-03,   0.000000000000000D+00,  -4.465479683023489D-03, &
   0.000000000000000D+00,   5.995031260027777D-03,   0.000000000000000D+00, &
  -6.010031080823063D-03,   0.000000000000000D+00,   4.715456337784409D-03, &
   0.000000000000000D+00,  -1.694959340249962D-03,   0.000000000000000D+00, &
   8.168846669821257D-07,   1.695929125128421D-06,   0.000000000000000D+00, &
   1.503352333430595D-06,   0.000000000000000D+00,   4.844483451661665D-07, &
  -1.115053256515652D-06,   0.000000000000000D+00,   2.159562771832410D-07, &
   0.000000000000000D+00,  -1.375645296543523D-06,   0.000000000000000D+00, &
  -6.809388566855455D-08,   0.000000000000000D+00,  -5.396624430437912D-06, &
   2.049071262938549D-06,   0.000000000000000D+00,  -9.379559223482426D-07, &
   0.000000000000000D+00,  -3.011553969369938D-06,   0.000000000000000D+00, &
   6.644492118123145D-07,   0.000000000000000D+00,  -9.910905667219181D-07, &
   0.000000000000000D+00,   1.681138248390282D-06,   0.000000000000000D+00, &
   2.154180272886025D-05,  -3.757195224480349D-06,   0.000000000000000D+00, &
   6.122037742607822D-07,   0.000000000000000D+00,   3.569751031342451D-06, &
   0.000000000000000D+00,   4.053215351204230D-06,   0.000000000000000D+00, &
  -2.136343036278069D-06,   0.000000000000000D+00,   1.062423905028842D-06, &
   0.000000000000000D+00,   8.066980040168380D-07,   0.000000000000000D+00, &
  -7.100054019893400D-06,   0.000000000000000D+00,  -3.542614448503280D-05, &
   2.769159860520600D-06,   0.000000000000000D+00,   1.003575168875261D-06, &
   0.000000000000000D+00,  -1.661834343878844D-06,   0.000000000000000D+00, &
  -4.385394181883556D-06,   0.000000000000000D+00,  -5.507022150707782D-06, &
   0.000000000000000D+00,   4.373970947021826D-06,   0.000000000000000D+00, &
   5.670651344628880D-07,   0.000000000000000D+00,  -2.416917396700609D-06, &
   0.000000000000000D+00,  -7.071086856528417D-07,   0.000000000000000D+00, &
   8.697012826207119D-06,   0.000000000000000D+00,  -1.866258663531472D-05, &
  -4.644317965280365D-07,   0.000000000000000D+00,  -2.214854063933862D-06, &
   0.000000000000000D+00,  -1.314691877672810D-06,   0.000000000000000D+00, &
   2.386182054109009D-06,   0.000000000000000D+00,   9.146896579687648D-06, &
   0.000000000000000D+00,   3.461794709674706D-05,   0.000000000000000D+00, &
  -4.839400902034631D-06,   0.000000000000000D+00,  -1.545254230723247D-06, &
   0.000000000000000D+00,   2.599813353919404D-06,   0.000000000000000D+00, &
   5.940032160547950D-06,   0.000000000000000D+00,   4.347000539569270D-06, &
   0.000000000000000D+00,   7.456437637109811D-06,   0.000000000000000D+00, &
   1.940089969106714D-04,  -1.476050477243915D-08,   0.000000000000000D+00, &
   2.051876604498146D-06,   0.000000000000000D+00,   2.542055740721383D-06, &
   0.000000000000000D+00,  -5.657476653077574D-07,   0.000000000000000D+00, &
  -1.141554109391776D-05,   0.000000000000000D+00,  -4.445355268698689D-05, &
   0.000000000000000D+00,  -1.131753706718524D-04,   0.000000000000000D+00, &
   3.192069762457796D-06,   0.000000000000000D+00,   7.338672930018135D-07, &
   0.000000000000000D+00,  -2.223353064387005D-06,   0.000000000000000D+00, &
  -7.020317243879304D-06,   0.000000000000000D+00,  -8.563350476682885D-06, &
   0.000000000000000D+00,  -3.434114636003633D-07,   0.000000000000000D+00, &
  -5.544068160334509D-05,   0.000000000000000D+00,  -4.217581421963197D-04, &
  -6.026653617737659D-07,   0.000000000000000D+00,  -9.034548728315073D-07, &
   0.000000000000000D+00,  -1.466924872559225D-06,   0.000000000000000D+00, &
   8.104934054195659D-07,   0.000000000000000D+00,   1.072908509648443D-05, &
   0.000000000000000D+00,   3.838097309855116D-05,   0.000000000000000D+00, &
   1.414454881577971D-04,   0.000000000000000D+00,   2.076798322492251D-04, &
   0.000000000000000D+00,  -1.074437383021449D-06,   0.000000000000000D+00, &
   1.410916235009792D-07,   0.000000000000000D+00,   1.609697407201632D-06, &
   0.000000000000000D+00,   4.488019216725941D-06,   0.000000000000000D+00, &
   5.140129623405627D-06,   0.000000000000000D+00,  -6.554100492808529D-06, &
   0.000000000000000D+00,  -3.705446292628276D-05,   0.000000000000000D+00, &
   1.264683909397834D-04,   0.000000000000000D+00,   5.459524197891498D-04, &
   5.307391091249542D-07,   0.000000000000000D+00,  -8.685812304694845D-08, &
   0.000000000000000D+00,   9.896351575754642D-08,   0.000000000000000D+00, &
  -1.416146393362972D-06,   0.000000000000000D+00,  -6.750826230289048D-06, &
   0.000000000000000D+00,  -2.137412590529461D-05,   0.000000000000000D+00, &
  -7.737286396967558D-05,   0.000000000000000D+00,  -2.770799181187242D-04, &
   0.000000000000000D+00,  -2.582198288497048D-04,   0.000000000000000D+00, &
   5.127555045916170D-08,   0.000000000000000D+00,  -8.368709300954763D-08, &
   0.000000000000000D+00,  -7.853310307163568D-07,   0.000000000000000D+00, &
  -1.431956228890467D-06,   0.000000000000000D+00,  -7.890373462729547D-07, &
   0.000000000000000D+00,   7.962597437621406D-06,   0.000000000000000D+00, &
   5.394890881527811D-05,   0.000000000000000D+00,   1.042463611771479D-04, &
   0.000000000000000D+00,  -1.660059530526962D-04,   0.000000000000000D+00, &
  -4.685518769826647D-04,  -1.157539058902820D-07,   0.000000000000000D+00, &
   2.233617284215253D-07,   0.000000000000000D+00,   2.727723129737987D-07, &
   0.000000000000000D+00,   8.187282880046705D-07,   0.000000000000000D+00, &
   2.227153524735248D-06,   0.000000000000000D+00,   6.006314981883220D-06, &
   0.000000000000000D+00,   1.135359023558300D-05,   0.000000000000000D+00, &
   1.095189939536081D-04,   0.000000000000000D+00,   3.510488112908878D-04, &
   0.000000000000000D+00,   2.304534353663988D-04,   0.000000000000000D+00, &
   4.398496416188756D-08,   0.000000000000000D+00,  -9.979148381218784D-08, &
   0.000000000000000D+00,   2.528023444542473D-07,   0.000000000000000D+00, &
   1.391490416819402D-07,   0.000000000000000D+00,  -3.267161001867995D-07, &
   0.000000000000000D+00,  -4.021706118338959D-06,   0.000000000000000D+00, &
  -2.066195032149522D-05,   0.000000000000000D+00,  -1.165273463508776D-04, &
   0.000000000000000D+00,  -1.421194053563970D-04,   0.000000000000000D+00, &
   1.333613256171873D-04,   0.000000000000000D+00,   2.417675622014313D-04 /)
lem(:,277) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.871613198489092D-05,   8.450852474049048D-05, &
   1.517723537492806D-04,   5.436413336289007D-05,  -1.665348613763635D-04, &
   2.293435235886289D-04,   6.842877770554068D-04,  -2.196649722061246D-04, &
  -1.006989257279724D-03,  -2.189916296347253D-04,   9.436532319160378D-04, &
   2.337110136084642D-04,  -6.854548955859771D-04,   1.516221203015977D-04, &
   1.693758191707653D-04,  -1.248532260501766D-04,   3.928766826263231D-05, &
  -4.776857171450607D-06,  -2.615881909903460D-06,  -1.571914095110402D-05, &
   4.241829515410917D-05,   0.000000000000000D+00,  -1.871613198489513D-05, &
  -8.450852474049551D-05,  -1.517723537492810D-04,  -5.436413336288332D-05, &
   1.665348613763627D-04,  -2.293435235886391D-04,  -6.842877770554030D-04, &
   2.196649722061325D-04,   1.006989257279726D-03,   2.189916296347164D-04, &
  -9.436532319160382D-04,  -2.337110136084611D-04,   6.854548955859736D-04, &
  -1.516221203015951D-04,  -1.693758191707603D-04,   1.248532260501715D-04, &
  -3.928766826263646D-05,   4.776857171453648D-06,   2.615881909907928D-06, &
   1.571914095110428D-05,  -4.241829515411326D-05,   0.000000000000000D+00, &
  -9.358065992450859D-06,   0.000000000000000D+00,   1.622966084868147D-05, &
   0.000000000000000D+00,  -2.777172064513909D-05,   0.000000000000000D+00, &
   5.162052992195188D-06,   0.000000000000000D+00,   7.889199857660623D-04, &
   0.000000000000000D+00,  -2.759343153030523D-03,   0.000000000000000D+00, &
   2.536202881383451D-03,   0.000000000000000D+00,   9.719524403269025D-04, &
   0.000000000000000D+00,  -1.297471180312458D-03,   0.000000000000000D+00, &
  -2.247885918476053D-03,   0.000000000000000D+00,   2.970876026388674D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.398200757574787D-07, &
   0.000000000000000D+00,   6.746115754905967D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.659902690314799D-07,   0.000000000000000D+00, &
  -1.450179378558912D-06,   0.000000000000000D+00,  -3.498514676667224D-07, &
   0.000000000000000D+00,  -9.319189944016516D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.288642362721600D-07,   0.000000000000000D+00, &
   1.346714821538766D-08,   0.000000000000000D+00,   1.974404707253768D-06, &
   0.000000000000000D+00,   2.222170190920347D-07,   0.000000000000000D+00, &
   5.990491590554280D-07,   0.000000000000000D+00,  -2.246729236474696D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.267082852043329D-07, &
   0.000000000000000D+00,  -8.287770520961663D-08,   0.000000000000000D+00, &
  -9.101047952104763D-07,   0.000000000000000D+00,   5.048919510055488D-06, &
   0.000000000000000D+00,  -1.024027513285316D-06,   0.000000000000000D+00, &
  -5.290276504355945D-08,   0.000000000000000D+00,   1.734167034808735D-06, &
   0.000000000000000D+00,   5.147541015231997D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.319708956025234D-07,   0.000000000000000D+00, &
   1.111468556265136D-06,   0.000000000000000D+00,   1.351692470141221D-06, &
   0.000000000000000D+00,  -3.817408786007081D-06,   0.000000000000000D+00, &
  -5.218610460266790D-05,   0.000000000000000D+00,   1.994019807478267D-06, &
   0.000000000000000D+00,   1.369551031529902D-07,   0.000000000000000D+00, &
  -3.119425045727307D-06,   0.000000000000000D+00,  -7.603024508401008D-06, &
   0.000000000000000D+00,  -6.089093667652741D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.054917265531495D-07,   0.000000000000000D+00, &
  -1.841022404198240D-06,   0.000000000000000D+00,  -2.523030078699364D-06, &
   0.000000000000000D+00,   2.976599818105447D-06,   0.000000000000000D+00, &
   2.616923825910540D-05,   0.000000000000000D+00,   1.760574211548985D-04, &
   0.000000000000000D+00,  -2.116841218752819D-06,   0.000000000000000D+00, &
  -6.021713594752867D-07,   0.000000000000000D+00,   3.247923057703529D-06, &
   0.000000000000000D+00,   8.492046327553492D-06,   0.000000000000000D+00, &
   8.994252686387397D-06,   0.000000000000000D+00,  -8.542696347850615D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.096997150231724D-06, &
   0.000000000000000D+00,   1.849464110973787D-06,   0.000000000000000D+00, &
   3.053973783879094D-06,   0.000000000000000D+00,  -1.231738014246962D-06, &
   0.000000000000000D+00,  -1.824399254458148D-05,   0.000000000000000D+00, &
  -7.906992506986133D-05,   0.000000000000000D+00,  -3.340346888334408D-04, &
   0.000000000000000D+00,   1.224041320874281D-06,   0.000000000000000D+00, &
   5.672157543931072D-07,   0.000000000000000D+00,  -2.773364182296379D-06, &
   0.000000000000000D+00,  -7.505843494125838D-06,   0.000000000000000D+00, &
  -9.401794779658960D-06,   0.000000000000000D+00,   8.112272778580795D-06, &
   0.000000000000000D+00,   8.847461879758397D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.106342378851464D-06,   0.000000000000000D+00, &
  -1.242816434193938D-06,   0.000000000000000D+00,  -1.916438423632093D-06, &
   0.000000000000000D+00,   6.920792209721270D-07,   0.000000000000000D+00, &
   1.281936618993906D-05,   0.000000000000000D+00,   4.767840723871030D-05, &
   0.000000000000000D+00,   1.820205242625642D-04,   0.000000000000000D+00, &
   4.326565251996633D-04,   0.000000000000000D+00,  -1.621753673477553D-07, &
   0.000000000000000D+00,  -1.682666225136007D-07,   0.000000000000000D+00, &
   1.909286996435886D-06,   0.000000000000000D+00,   4.591776034522651D-06, &
   0.000000000000000D+00,   5.611126285023065D-06,   0.000000000000000D+00, &
  -8.561941623439513D-06,   0.000000000000000D+00,  -5.742425357003462D-05, &
   0.000000000000000D+00,   3.989149360546534D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.101170072968352D-07,   0.000000000000000D+00, &
   4.541022375817646D-07,   0.000000000000000D+00,   2.585522177173884D-07, &
   0.000000000000000D+00,  -9.262525185124195D-07,   0.000000000000000D+00, &
  -7.575402621526247D-06,   0.000000000000000D+00,  -2.492239135046754D-05, &
   0.000000000000000D+00,  -8.992454278079721D-05,   0.000000000000000D+00, &
  -3.220477029024913D-04,   0.000000000000000D+00,  -4.699555042734211D-04, &
   0.000000000000000D+00,  -1.832331774904505D-07,   0.000000000000000D+00, &
   5.519575848522850D-08,   0.000000000000000D+00,  -7.129461945609274D-07, &
   0.000000000000000D+00,  -1.505014685043270D-06,   0.000000000000000D+00, &
  -8.356283655811720D-07,   0.000000000000000D+00,   8.620929203523179D-06, &
   0.000000000000000D+00,   6.113000640373311D-05,   0.000000000000000D+00, &
   1.297789296424235D-04,   0.000000000000000D+00,  -7.695317621150137D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.191866455488466D-08, &
   0.000000000000000D+00,  -9.775184213226504D-08,   0.000000000000000D+00, &
   3.265343537783065D-07,   0.000000000000000D+00,   5.162807821686528D-07, &
   0.000000000000000D+00,   2.569444866376371D-06,   0.000000000000000D+00, &
   7.459457572684182D-06,   0.000000000000000D+00,   1.629716091240068D-05, &
   0.000000000000000D+00,   1.310433610559568D-04,   0.000000000000000D+00, &
   4.020103681491829D-04,   0.000000000000000D+00,   4.809065193713811D-04, &
   0.000000000000000D+00,   5.867065568247820D-08,   0.000000000000000D+00, &
  -6.346197880444880D-08,   0.000000000000000D+00,   9.058833258333245D-09, &
   0.000000000000000D+00,   2.027475479318410D-07,   0.000000000000000D+00, &
  -7.491890892121195D-07,   0.000000000000000D+00,  -4.625457883308485D-06, &
   0.000000000000000D+00,  -2.494564367461223D-05,   0.000000000000000D+00, &
  -1.340075086325075D-04,   0.000000000000000D+00,  -1.701512519925588D-04, &
   0.000000000000000D+00,   3.660296389516593D-05,   0.000000000000000D+00 /)
lem(:,278) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   3.156216438318735D-05,   1.141976490641087D-04,   2.281613719238571D-04, &
   2.790379950687630D-04,   1.052009048265464D-04,  -3.298898346755977D-04, &
  -5.910339245396022D-04,  -1.056007549734680D-04,   1.934576055500781D-04, &
   2.119337326355441D-04,   6.273043918209745D-04,  -3.879131481597461D-04, &
  -6.387070565460220D-04,   5.574287593167621D-04,   2.108318464031015D-05, &
  -1.906112936510688D-04,   1.108129384919682D-04,  -3.447924244462461D-05, &
   5.340797738312944D-06,  -2.543433763853505D-06,   7.535175041801204D-06, &
  -1.822883189842097D-05,   3.156216438318786D-05,   1.141976490641130D-04, &
   2.281613719238590D-04,   2.790379950687579D-04,   1.052009048265502D-04, &
  -3.298898346755991D-04,  -5.910339245396117D-04,  -1.056007549734655D-04, &
   1.934576055500843D-04,   2.119337326355446D-04,   6.273043918209706D-04, &
  -3.879131481597422D-04,  -6.387070565460181D-04,   5.574287593167540D-04, &
   2.108318464031293D-05,  -1.906112936510604D-04,   1.108129384919603D-04, &
  -3.447924244462805D-05,   5.340797738320515D-06,  -2.543433763852632D-06, &
   7.535175041797971D-06,  -1.822883189842262D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -8.983189637771483D-06,   0.000000000000000D+00, &
   2.457405717834903D-05,   0.000000000000000D+00,  -6.535459617609575D-05, &
   0.000000000000000D+00,   1.932597598617521D-04,   0.000000000000000D+00, &
  -6.457305561118393D-04,   0.000000000000000D+00,  -5.240881732240409D-04, &
   0.000000000000000D+00,   8.856941025781889D-03,   0.000000000000000D+00, &
  -2.103762874746390D-02,   0.000000000000000D+00,   2.183911723260118D-02, &
   0.000000000000000D+00,  -1.003364703566952D-02,   0.000000000000000D+00, &
  -1.271565755206647D-07,  -3.672424174134842D-07,   0.000000000000000D+00, &
  -7.567150053719636D-07,   0.000000000000000D+00,   7.478229371540181D-07, &
  -9.812185562919915D-07,   0.000000000000000D+00,   7.572231235561210D-07, &
   0.000000000000000D+00,  -4.205948267227394D-07,   0.000000000000000D+00, &
   6.399748526212970D-07,   0.000000000000000D+00,  -2.430757998308745D-06, &
   1.316745930393565D-06,   0.000000000000000D+00,   1.844106065995726D-07, &
   0.000000000000000D+00,  -1.963130482188662D-06,   0.000000000000000D+00, &
   2.737359686212096D-06,   0.000000000000000D+00,  -6.158392388860976D-07, &
   0.000000000000000D+00,  -7.204321913031795D-07,   0.000000000000000D+00, &
   6.948611436984311D-06,   2.884840989088797D-07,   0.000000000000000D+00, &
  -1.075686195315176D-06,   0.000000000000000D+00,   1.136481174987799D-06, &
   0.000000000000000D+00,   4.428688224617128D-06,   0.000000000000000D+00, &
  -3.322902378383442D-06,   0.000000000000000D+00,  -2.504387498894055D-07, &
   0.000000000000000D+00,   1.810718845058727D-06,   0.000000000000000D+00, &
  -6.939092160382761D-07,   0.000000000000000D+00,  -1.960676206853289D-05, &
  -1.789020610736581D-06,   0.000000000000000D+00,   5.437873331765982D-07, &
   0.000000000000000D+00,   2.240871244055363D-07,   0.000000000000000D+00, &
  -3.757098286614837D-06,   0.000000000000000D+00,  -6.474766364581160D-06, &
   0.000000000000000D+00,   1.982105520618386D-06,   0.000000000000000D+00, &
   8.044482253538260D-07,   0.000000000000000D+00,  -8.248827025013797D-07, &
   0.000000000000000D+00,  -1.320123966708196D-06,   0.000000000000000D+00, &
   5.959008684499491D-06,   0.000000000000000D+00,   3.963639284228228D-05, &
   2.128824646219715D-06,   0.000000000000000D+00,   4.825802636075879D-07, &
   0.000000000000000D+00,  -9.938949945094412D-07,   0.000000000000000D+00, &
   4.024746691159046D-07,   0.000000000000000D+00,   4.084222610521048D-06, &
   0.000000000000000D+00,   2.744704381287080D-06,   0.000000000000000D+00, &
  -8.902290381917169D-07,   0.000000000000000D+00,  -1.041191994769106D-06, &
   0.000000000000000D+00,  -4.182778647721666D-07,   0.000000000000000D+00, &
   1.323287252660904D-06,   0.000000000000000D+00,   1.267635777395108D-06, &
   0.000000000000000D+00,  -6.628138316261676D-06,   0.000000000000000D+00, &
   3.440303400815905D-06,  -1.529071964448533D-06,   0.000000000000000D+00, &
  -4.368857506596637D-07,   0.000000000000000D+00,   1.500348856849344D-06, &
   0.000000000000000D+00,   2.473163224616074D-06,   0.000000000000000D+00, &
   2.163861465082152D-07,   0.000000000000000D+00,  -5.124259934830275D-06, &
   0.000000000000000D+00,  -2.512375515334318D-05,   0.000000000000000D+00, &
   6.834413276598173D-07,   0.000000000000000D+00,   7.908989636079195D-07, &
   0.000000000000000D+00,   2.486628964156903D-07,   0.000000000000000D+00, &
  -1.656597098103416D-06,   0.000000000000000D+00,  -5.598637559575195D-06, &
   0.000000000000000D+00,  -8.088484829916405D-06,   0.000000000000000D+00, &
  -9.411939897724173D-06,   0.000000000000000D+00,  -2.671274758223080D-04, &
   5.273642065053712D-07,   0.000000000000000D+00,  -1.626641653680363D-07, &
   0.000000000000000D+00,  -1.303167014724209D-06,   0.000000000000000D+00, &
  -2.479273685619581D-06,   0.000000000000000D+00,  -6.274437517738432D-07, &
   0.000000000000000D+00,   8.586159227166802D-06,   0.000000000000000D+00, &
   3.742704831058226D-05,   0.000000000000000D+00,   1.775197774400469D-04, &
   0.000000000000000D+00,  -4.265647345223706D-07,   0.000000000000000D+00, &
  -1.322576401036110D-07,   0.000000000000000D+00,   4.158034413906159D-07, &
   0.000000000000000D+00,   1.785597686793358D-06,   0.000000000000000D+00, &
   5.077198121370472D-06,   0.000000000000000D+00,   8.211144341113336D-06, &
   0.000000000000000D+00,   2.279971453290220D-06,   0.000000000000000D+00, &
   1.099677165101271D-05,   0.000000000000000D+00,   7.370153174111758D-04, &
   2.564415408458496D-08,   0.000000000000000D+00,   1.766009793584626D-07, &
   0.000000000000000D+00,   4.526290112960162D-07,   0.000000000000000D+00, &
   9.024730924008295D-07,   0.000000000000000D+00,  -8.475644088123569D-07, &
   0.000000000000000D+00,  -8.266829069890626D-06,   0.000000000000000D+00, &
  -2.991308411233357D-05,   0.000000000000000D+00,  -1.250041624737237D-04, &
   0.000000000000000D+00,  -4.698667578793968D-04,   0.000000000000000D+00, &
   8.133203935234029D-08,   0.000000000000000D+00,  -1.369234337647969D-07, &
   0.000000000000000D+00,  -3.291354463671607D-07,   0.000000000000000D+00, &
  -9.348927146355129D-07,   0.000000000000000D+00,  -1.769853392256878D-06, &
   0.000000000000000D+00,  -1.891703051147778D-06,   0.000000000000000D+00, &
   7.362664388893385D-06,   0.000000000000000D+00,   6.060016794723726D-05, &
   0.000000000000000D+00,   3.817458379518014D-05,   0.000000000000000D+00, &
  -1.023463714836473D-03,  -4.346073242766863D-08,   0.000000000000000D+00, &
   3.444429845309630D-08,   0.000000000000000D+00,  -1.599890714080151D-08, &
   0.000000000000000D+00,  -7.411983182931629D-08,   0.000000000000000D+00, &
   6.984583934693264D-07,   0.000000000000000D+00,   3.589816521535874D-06, &
   0.000000000000000D+00,   1.087122216339661D-05,   0.000000000000000D+00, &
   3.075147992461853D-05,   0.000000000000000D+00,   2.304534353663988D-04, &
   0.000000000000000D+00,   6.514290116598832D-04,   0.000000000000000D+00, &
   1.148496089646976D-08,   0.000000000000000D+00,   2.917301338884701D-08, &
   0.000000000000000D+00,   4.046791462934332D-08,   0.000000000000000D+00, &
   1.995618817599720D-07,   0.000000000000000D+00,  -1.043985903303732D-07, &
   0.000000000000000D+00,  -8.364462665601973D-07,   0.000000000000000D+00, &
  -5.860715590631283D-06,   0.000000000000000D+00,  -3.351415702446555D-05, &
   0.000000000000000D+00,  -1.704637158145791D-04,   0.000000000000000D+00, &
  -7.614072702757303D-05,   0.000000000000000D+00,   8.073607663216225D-04 /)
lem(:,279) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.393961913458059D-05,  -8.542625183002900D-05, &
  -2.729767084353653D-04,  -5.761278987921154D-04,  -7.779001945958573D-04, &
  -3.662950385590100D-04,   9.261869738232498D-04,   1.723732600825333D-03, &
  -4.107992135525899D-04,  -2.086067844801860D-03,   6.491323422168717D-04, &
   1.206561094710638D-03,  -8.209419472946875D-04,  -9.390907031818552D-05, &
   2.523642358192572D-04,  -1.217568984791668D-04,   5.175834027410020D-05, &
   1.972138598204497D-06,  -2.837073521006182D-06,  -1.316127886133005D-05, &
   4.465230838690002D-06,   0.000000000000000D+00,   1.393961913460180D-05, &
   8.542625183002560D-05,   2.729767084353517D-04,   5.761278987921300D-04, &
   7.779001945958542D-04,   3.662950385589943D-04,  -9.261869738232373D-04, &
  -1.723732600825318D-03,   4.107992135525824D-04,   2.086067844801846D-03, &
  -6.491323422168711D-04,  -1.206561094710630D-03,   8.209419472946853D-04, &
   9.390907031818367D-05,  -2.523642358192469D-04,   1.217568984791569D-04, &
  -5.175834027411042D-05,  -1.972138598191549D-06,   2.837073521014060D-06, &
   1.316127886132660D-05,  -4.465230838701345D-06,   0.000000000000000D+00, &
   6.969809567288024D-06,   0.000000000000000D+00,  -1.311350560437832D-05, &
   0.000000000000000D+00,   2.946776068221608D-05,   0.000000000000000D+00, &
  -7.693668414207122D-05,   0.000000000000000D+00,   2.404067698732010D-04, &
   0.000000000000000D+00,  -9.730043060993357D-04,   0.000000000000000D+00, &
   6.070006607814336D-03,   0.000000000000000D+00,  -2.210574414873325D-02, &
   0.000000000000000D+00,   4.170110626946237D-02,   0.000000000000000D+00, &
  -4.284649204702546D-02,   0.000000000000000D+00,   2.396768538640084D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.510118272568938D-07, &
   0.000000000000000D+00,  -5.314492771768268D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.336365267761846D-07,   0.000000000000000D+00, &
   1.327331725891293D-06,   0.000000000000000D+00,  -1.554418210912397D-07, &
   0.000000000000000D+00,   1.238284014162913D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.648711992217899D-07,   0.000000000000000D+00, &
   8.794255467838961D-07,   0.000000000000000D+00,  -3.384210013009699D-06, &
   0.000000000000000D+00,   9.250714969709468D-07,   0.000000000000000D+00, &
  -7.452400030444693D-09,   0.000000000000000D+00,  -3.001184174508699D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   8.082890010379816D-07, &
   0.000000000000000D+00,  -1.345369921419750D-06,   0.000000000000000D+00, &
  -1.624469712619628D-06,   0.000000000000000D+00,   9.550909604887881D-06, &
   0.000000000000000D+00,  -7.516338235333200D-07,   0.000000000000000D+00, &
  -1.960722213713046D-06,   0.000000000000000D+00,   1.340077023670622D-06, &
   0.000000000000000D+00,   7.814022713753345D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.339820722083717D-06,   0.000000000000000D+00, &
   1.002860903841435D-07,   0.000000000000000D+00,   4.364328424936825D-06, &
   0.000000000000000D+00,   2.057722597713006D-06,   0.000000000000000D+00, &
  -3.084752967942762D-05,   0.000000000000000D+00,   4.189646142159611D-07, &
   0.000000000000000D+00,   3.100705285494694D-06,   0.000000000000000D+00, &
   1.927518153725799D-06,   0.000000000000000D+00,  -7.952968596315961D-06, &
   0.000000000000000D+00,  -2.130062861756963D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.111609900245198D-06,   0.000000000000000D+00, &
   7.388710903256771D-07,   0.000000000000000D+00,  -5.161986173825774D-06, &
   0.000000000000000D+00,  -9.065843158235920D-06,   0.000000000000000D+00, &
   6.702579068254104D-06,   0.000000000000000D+00,   1.199988736992787D-04, &
   0.000000000000000D+00,  -9.764972914997265D-07,   0.000000000000000D+00, &
  -2.819273817162588D-06,   0.000000000000000D+00,  -2.948161014574018D-06, &
   0.000000000000000D+00,   6.502993502112490D-06,   0.000000000000000D+00, &
   2.739503165316098D-05,   0.000000000000000D+00,   4.210869653175477D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.162670963024975D-06, &
   0.000000000000000D+00,  -1.082964241067019D-07,   0.000000000000000D+00, &
   4.831392444086751D-06,   0.000000000000000D+00,   9.861980149449433D-06, &
   0.000000000000000D+00,   5.037493179966933D-07,   0.000000000000000D+00, &
  -6.074631115892424D-05,   0.000000000000000D+00,  -4.510791977207168D-04, &
   0.000000000000000D+00,   1.379547926483990D-06,   0.000000000000000D+00, &
   1.212178915478443D-06,   0.000000000000000D+00,   1.028270984330977D-06, &
   0.000000000000000D+00,  -6.038621560156226D-06,   0.000000000000000D+00, &
  -2.397721816722097D-05,   0.000000000000000D+00,  -4.175160283602444D-05, &
   0.000000000000000D+00,   4.388504612744750D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.860851088596298D-07,   0.000000000000000D+00, &
  -3.739400892311191D-07,   0.000000000000000D+00,  -3.270697866714804D-06, &
   0.000000000000000D+00,  -5.398763858920268D-06,   0.000000000000000D+00, &
   1.766791407699267D-06,   0.000000000000000D+00,   4.355791357026058D-05, &
   0.000000000000000D+00,   1.985855752734513D-04,   0.000000000000000D+00, &
   1.251756148063562D-03,   0.000000000000000D+00,  -7.446214861887902D-07, &
   0.000000000000000D+00,   2.101031504927788D-07,   0.000000000000000D+00, &
   8.172751289530428D-07,   0.000000000000000D+00,   4.489544595449543D-06, &
   0.000000000000000D+00,   1.352622264849873D-05,   0.000000000000000D+00, &
   2.032453319181632D-05,   0.000000000000000D+00,  -2.457665625720454D-05, &
   0.000000000000000D+00,  -2.816241538500622D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.998319934422818D-07,   0.000000000000000D+00, &
  -5.333191708167225D-08,   0.000000000000000D+00,   1.374558057728394D-06, &
   0.000000000000000D+00,   1.120485607487242D-06,   0.000000000000000D+00, &
  -3.880082062428633D-06,   0.000000000000000D+00,  -2.561272050025838D-05, &
   0.000000000000000D+00,  -9.258143212135926D-05,   0.000000000000000D+00, &
  -3.724539703565499D-04,   0.000000000000000D+00,  -2.356359515104339D-03, &
   0.000000000000000D+00,   8.329818030391308D-08,   0.000000000000000D+00, &
  -3.174024801219159D-07,   0.000000000000000D+00,  -9.706425325231555D-07, &
   0.000000000000000D+00,  -1.647214602101262D-06,   0.000000000000000D+00, &
  -3.726737268053251D-06,   0.000000000000000D+00,  -2.209746210003732D-06, &
   0.000000000000000D+00,   3.029583951940378D-05,   0.000000000000000D+00, &
   2.297498895215344D-04,   0.000000000000000D+00,   7.719733130104784D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.091879090584786D-08, &
   0.000000000000000D+00,   2.745901733285585D-07,   0.000000000000000D+00, &
  -4.128034197513196D-07,   0.000000000000000D+00,  -8.304391831676980D-08, &
   0.000000000000000D+00,   1.984878000179688D-06,   0.000000000000000D+00, &
   8.500689241421090D-06,   0.000000000000000D+00,   2.468410752470286D-05, &
   0.000000000000000D+00,   6.054540937019465D-05,   0.000000000000000D+00, &
   4.809065193713811D-04,   0.000000000000000D+00,   2.992894364094969D-03, &
   0.000000000000000D+00,   2.977985681338603D-08,   0.000000000000000D+00, &
  -1.451676304197259D-08,   0.000000000000000D+00,   5.705327277045459D-07, &
   0.000000000000000D+00,  -3.534986366361005D-08,   0.000000000000000D+00, &
  -1.331575244813013D-07,   0.000000000000000D+00,  -2.007435046017054D-06, &
   0.000000000000000D+00,  -1.487310131909106D-05,   0.000000000000000D+00, &
  -8.675731296072317D-05,   0.000000000000000D+00,  -4.922663293958269D-04, &
   0.000000000000000D+00,  -1.069247761013141D-03,   0.000000000000000D+00 /)
lem(:,280) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.969072307435329D-03,  -2.751646711814412D-03, &
  -8.559490324614054D-03,   5.668977329566651D-03,   9.274211909771998D-03, &
  -2.858970401667879D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   2.969072307435254D-03, &
  -2.751646711814439D-03,  -8.559490324613965D-03,   5.668977329566658D-03, &
   9.274211909771941D-03,  -2.858970401667891D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   5.855175013318151D-04,   0.000000000000000D+00,   7.361031805535412D-06, &
   0.000000000000000D+00,   4.160583194422908D-07,   0.000000000000000D+00, &
   3.584462338539511D-08,   0.000000000000000D+00,   3.795313063377501D-09, &
   0.000000000000000D+00,   4.500630866835223D-10,   0.000000000000000D+00, &
   5.638045199086480D-11,   0.000000000000000D+00,   7.153534324373709D-12, &
   0.000000000000000D+00,  -3.626663382760936D-03,   0.000000000000000D+00, &
  -7.630498148417766D-05,   0.000000000000000D+00,  -6.070550917128262D-06, &
   0.000000000000000D+00,  -6.773020811835843D-07,   0.000000000000000D+00, &
  -8.855236302127505D-08,   0.000000000000000D+00,  -1.260364167026555D-08, &
   0.000000000000000D+00,  -1.868483414831970D-09,   0.000000000000000D+00, &
  -2.781830836403745D-10,   0.000000000000000D+00,  -4.773028456778899D-11, &
   8.089874539108791D-03,   0.000000000000000D+00,   4.641503496032222D-05, &
   0.000000000000000D+00,   2.167929075276315D-05,   0.000000000000000D+00, &
   4.065294810050979D-06,   0.000000000000000D+00,   7.383463375200068D-07, &
   0.000000000000000D+00,   1.347494843811492D-07,   0.000000000000000D+00, &
   2.458524223656090D-08,   0.000000000000000D+00,   4.417060884089854D-09, &
   0.000000000000000D+00,   7.761584596531596D-10,   0.000000000000000D+00, &
  -5.094851322871740D-03,   0.000000000000000D+00,   1.000130300214467D-03, &
   0.000000000000000D+00,   3.406769890232546D-05,   0.000000000000000D+00, &
  -4.612162190532008D-06,   0.000000000000000D+00,  -2.249896623162722D-06, &
   0.000000000000000D+00,  -6.420138609133115D-07,   0.000000000000000D+00, &
  -1.584778044150050D-07,   0.000000000000000D+00,  -3.619321971465199D-08, &
   0.000000000000000D+00,  -7.765715551410966D-09,   0.000000000000000D+00, &
  -1.940883883094095D-09,  -9.265213178996501D-03,   0.000000000000000D+00, &
  -3.033936273236948D-03,   0.000000000000000D+00,  -2.334616654933485D-04, &
   0.000000000000000D+00,  -3.114247516553934D-05,   0.000000000000000D+00, &
  -1.630252806885567D-06,   0.000000000000000D+00,   8.547338461242150D-07, &
   0.000000000000000D+00,   4.531684667337895D-07,   0.000000000000000D+00, &
   1.540910873102190D-07,   0.000000000000000D+00,   4.398496416188756D-08, &
   0.000000000000000D+00,   1.148496089646976D-08,   0.000000000000000D+00, &
   1.894374849564869D-02,   0.000000000000000D+00,   2.365410625478693D-03, &
   0.000000000000000D+00,   3.431076885416128D-05,   0.000000000000000D+00, &
   6.569454744484824D-05,   0.000000000000000D+00,   2.004713428574508D-05, &
   0.000000000000000D+00,   3.480923972972222D-06,   0.000000000000000D+00, &
   7.493183085350146D-08,   0.000000000000000D+00,  -2.357651373989403D-07, &
   0.000000000000000D+00,  -1.253500931276657D-07,   0.000000000000000D+00, &
  -4.627848161172777D-08,   0.000000000000000D+00,  -1.909176252189519D-08 /)
lem(:,281) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.937114797933228D-04, &
   1.935065036032065D-03,  -1.069879464703071D-03,  -4.145872643030815D-03, &
   4.494901201427747D-03,   2.987561756475944D-03,  -4.296321117973460D-03, &
  -8.295764116363314D-04,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   4.937114797933231D-04,  -1.935065036032034D-03,   1.069879464703033D-03, &
   4.145872643030809D-03,  -4.494901201427705D-03,  -2.987561756475967D-03, &
   4.296321117973434D-03,   8.295764116363570D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -7.087773263730010D-10, &
   0.000000000000000D+00,  -2.498771740148417D-08,   0.000000000000000D+00, &
  -2.217436211385796D-07,   0.000000000000000D+00,  -6.449725163172609D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.483919945544193D-06, &
   0.000000000000000D+00,   2.127463742275859D-07,   0.000000000000000D+00, &
   1.697540025155638D-08,   0.000000000000000D+00,   1.675776691303188D-09, &
   0.000000000000000D+00,   1.798497713320260D-10,   0.000000000000000D+00, &
   1.809077117758828D-11,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.721050327279449D-05,   0.000000000000000D+00, &
  -3.163925396357155D-06,   0.000000000000000D+00,  -3.290634184863721D-07, &
   0.000000000000000D+00,  -4.061364510152360D-08,   0.000000000000000D+00, &
  -5.382004402739814D-09,   0.000000000000000D+00,  -7.040325086601305D-10, &
   0.000000000000000D+00,  -7.185725685466161D-11,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.123552518300068D-04,   0.000000000000000D+00, &
   1.447263817258957D-05,   0.000000000000000D+00,   2.196392115122669D-06, &
   0.000000000000000D+00,   3.645825171036840D-07,   0.000000000000000D+00, &
   6.196843851974008D-08,   0.000000000000000D+00,   1.033820769856170D-08, &
   0.000000000000000D+00,   1.557981058354067D-09,   0.000000000000000D+00, &
   9.568746955445732D-11,   0.000000000000000D+00,  -5.137947621465896D-04, &
   0.000000000000000D+00,  -1.299861822393635D-05,   0.000000000000000D+00, &
  -4.977976817194909D-06,   0.000000000000000D+00,  -1.391789269241292D-06, &
   0.000000000000000D+00,  -3.375708056928978D-07,   0.000000000000000D+00, &
  -7.506531979905188D-08,   0.000000000000000D+00,  -1.518911445295235D-08, &
   0.000000000000000D+00,  -2.492914472969253D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   6.435821944892648D-04,   0.000000000000000D+00, &
  -1.054448140042075D-04,   0.000000000000000D+00,  -6.278353444861142D-06, &
   0.000000000000000D+00,   1.054514542052064D-06,   0.000000000000000D+00, &
   7.494185914662516D-07,   0.000000000000000D+00,   2.698431407741391D-07, &
   0.000000000000000D+00,   7.755569252890826D-08,   0.000000000000000D+00, &
   1.872345489803555D-08,   0.000000000000000D+00,   3.463208940456888D-09, &
   0.000000000000000D+00,  -1.189545185300219D-04,   0.000000000000000D+00, &
   4.236576539765877D-04,   0.000000000000000D+00,   4.802513874593013D-05, &
   0.000000000000000D+00,   7.756028958376641D-06,   0.000000000000000D+00, &
   6.095415105603179D-07,   0.000000000000000D+00,  -2.757493241745044D-07, &
   0.000000000000000D+00,  -1.838814399160541D-07,   0.000000000000000D+00, &
  -7.067527557556711D-08,   0.000000000000000D+00,  -2.076146614719427D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.365574101270223D-04, &
   0.000000000000000D+00,  -6.620584463084396D-04,   0.000000000000000D+00, &
  -5.252045692262341D-05,   0.000000000000000D+00,  -1.959019658286450D-05, &
   0.000000000000000D+00,  -5.858995500855205D-06,   0.000000000000000D+00, &
  -1.172586994916132D-06,   0.000000000000000D+00,  -3.616552737625497D-08, &
   0.000000000000000D+00,   9.943171801944118D-08,   0.000000000000000D+00, &
   5.867065568247820D-08,   0.000000000000000D+00,   2.977985681338603D-08, &
   0.000000000000000D+00,   1.535943339686959D-03,   0.000000000000000D+00, &
   2.408992881653844D-04,   0.000000000000000D+00,  -1.243542286814037D-04, &
   0.000000000000000D+00,  -5.354960346073285D-06,   0.000000000000000D+00, &
   6.207233938148690D-06,   0.000000000000000D+00,   3.614347132710731D-06, &
   0.000000000000000D+00,   1.181731084509279D-06,   0.000000000000000D+00, &
   2.078851474572805D-07,   0.000000000000000D+00,  -3.076061880710644D-08, &
   0.000000000000000D+00,  -4.870024534986803D-08,   0.000000000000000D+00 /)
lem(:,282) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.339679928267094D-04,  -8.611429582507440D-04,   2.117167495052182D-04, &
   9.683882110346160D-04,   1.430659678476454D-03,   1.113356575773613D-03, &
  -4.814385165702922D-03,  -2.679962233128472D-03,   3.851103132716341D-03, &
   1.681305046729956D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.339679928267160D-04,  -8.611429582507224D-04, &
   2.117167495052041D-04,   9.683882110346116D-04,   1.430659678476472D-03, &
   1.113356575773599D-03,  -4.814385165702909D-03,  -2.679962233128420D-03, &
   3.851103132716310D-03,   1.681305046729889D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.724214983561654D-10, &
   0.000000000000000D+00,   1.697817966690733D-08,   0.000000000000000D+00, &
   2.550997431553349D-07,   0.000000000000000D+00,   1.255491717231636D-06, &
   0.000000000000000D+00,   1.972222146917436D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   6.482701651245815D-06,   0.000000000000000D+00,   1.737566232055574D-07, &
   0.000000000000000D+00,   1.223420074268873D-08,   0.000000000000000D+00, &
   1.082439489598731D-09,   0.000000000000000D+00,   9.492822508738392D-11, &
   0.000000000000000D+00,   2.475579260931331D-12,   0.000000000000000D+00, &
  -6.759066825807605D-05,   0.000000000000000D+00,  -2.619484521390580D-06, &
   0.000000000000000D+00,  -2.438734022647428D-07,   0.000000000000000D+00, &
  -2.769992721955712D-08,   0.000000000000000D+00,  -3.260191230558117D-09, &
   0.000000000000000D+00,  -3.086737820454430D-10,   0.000000000000000D+00, &
   6.183145868357614D-11,   4.172507620783621D-05,   0.000000000000000D+00, &
   1.124632149442350D-05,   0.000000000000000D+00,   1.607833162663034D-06, &
   0.000000000000000D+00,   2.531981316568740D-07,   0.000000000000000D+00, &
   3.996804457869507D-08,   0.000000000000000D+00,   5.730647191224161D-09, &
   0.000000000000000D+00,   4.159057365864043D-10,   0.000000000000000D+00, &
   8.920215260593442D-04,   0.000000000000000D+00,  -3.043960871498729D-06, &
   0.000000000000000D+00,  -3.101389343676275D-06,   0.000000000000000D+00, &
  -9.280332549827649D-07,   0.000000000000000D+00,  -2.202849866786760D-07, &
   0.000000000000000D+00,  -4.540121429227469D-08,   0.000000000000000D+00, &
  -7.344217807247117D-09,   0.000000000000000D+00,   6.943146347678805D-10, &
  -2.718173139237192D-03,   0.000000000000000D+00,   8.817308836063491D-06, &
   0.000000000000000D+00,  -5.332207018454229D-06,   0.000000000000000D+00, &
   4.439616655212060D-07,   0.000000000000000D+00,   4.574592715689053D-07, &
   0.000000000000000D+00,   1.657882080465162D-07,   0.000000000000000D+00, &
   4.364035003238920D-08,   0.000000000000000D+00,   7.032544091260512D-09, &
   0.000000000000000D+00,   2.125711365720590D-03,   0.000000000000000D+00, &
  -3.781195944846609D-04,   0.000000000000000D+00,   1.367409393382627D-05, &
   0.000000000000000D+00,   4.467584078592236D-06,   0.000000000000000D+00, &
   4.582657704107381D-07,   0.000000000000000D+00,  -1.550879591867370D-07, &
   0.000000000000000D+00,  -1.082585703057941D-07,   0.000000000000000D+00, &
  -3.724466067985994D-08,   0.000000000000000D+00,  -5.348944281358945D-09, &
   2.854809977514230D-03,   0.000000000000000D+00,   1.165112541589966D-03, &
   0.000000000000000D+00,  -2.335978695242678D-06,   0.000000000000000D+00, &
  -4.861449969023181D-06,   0.000000000000000D+00,  -2.778013725201097D-06, &
   0.000000000000000D+00,  -6.540539772017890D-07,   0.000000000000000D+00, &
  -1.418397475902034D-08,   0.000000000000000D+00,   6.147003872111689D-08, &
   0.000000000000000D+00,   3.097891383540878D-08,   0.000000000000000D+00, &
  -6.394240902090415D-03,   0.000000000000000D+00,  -9.286094257263453D-04, &
   0.000000000000000D+00,   1.534348811930898D-04,   0.000000000000000D+00, &
  -1.776804151587500D-06,   0.000000000000000D+00,   8.895358740381842D-07, &
   0.000000000000000D+00,   1.450170900396455D-06,   0.000000000000000D+00, &
   5.685833592392109D-07,   0.000000000000000D+00,   8.701697454002154D-08, &
   0.000000000000000D+00,  -3.328801651212232D-08,   0.000000000000000D+00, &
  -5.242484366604355D-08,   3.008207849433394D-03,   0.000000000000000D+00, &
  -1.272789634535695D-03,   0.000000000000000D+00,  -5.918240310336983D-04, &
   0.000000000000000D+00,  -2.383224697154986D-05,   0.000000000000000D+00, &
   1.810918164852495D-06,   0.000000000000000D+00,   4.632889098278034D-07, &
   0.000000000000000D+00,  -6.097895578671173D-07,   0.000000000000000D+00, &
  -4.022310879059607D-07,   0.000000000000000D+00,  -9.979148381218784D-08, &
   0.000000000000000D+00,   2.917301338884701D-08,   0.000000000000000D+00, &
   2.365410625478693D-03,   0.000000000000000D+00,   2.855157429072669D-03, &
   0.000000000000000D+00,   5.435970211855232D-04,   0.000000000000000D+00, &
   2.022429382292918D-06,   0.000000000000000D+00,   6.020348653846282D-06, &
   0.000000000000000D+00,  -9.447705155900994D-07,   0.000000000000000D+00, &
  -7.004923773276639D-07,   0.000000000000000D+00,   1.608771464988933D-07, &
   0.000000000000000D+00,   2.499733651470588D-07,   0.000000000000000D+00, &
   8.604684653992556D-08,   0.000000000000000D+00,  -7.456678039006511D-08 /)
lem(:,283) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.975180768986226D-05,   3.645433539783505D-04, &
  -5.353375310706649D-04,  -1.088053474838930D-03,   1.570379498293325D-03, &
   3.278716925102082D-03,  -3.128071501528734D-03,  -5.872921425368955D-03, &
   4.151244818022622D-03,   5.119118099148623D-03,  -2.329915810931448D-03, &
  -1.913430949028056D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.975180768986013D-05, &
  -3.645433539783590D-04,   5.353375310706730D-04,   1.088053474838933D-03, &
  -1.570379498293336D-03,  -3.278716925102079D-03,   3.128071501528731D-03, &
   5.872921425368931D-03,  -4.151244818022616D-03,  -5.119118099148611D-03, &
   2.329915810931415D-03,   1.913430949028060D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.002743446785481D-09,   0.000000000000000D+00, &
  -1.941405440826680D-07,   0.000000000000000D+00,  -1.418238153580233D-06, &
   0.000000000000000D+00,  -3.668812445943319D-06,   0.000000000000000D+00, &
  -1.817106859081354D-06,   0.000000000000000D+00,   5.887378801107770D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.800838951993302D-07,   0.000000000000000D+00, &
   1.012126528306924D-08,   0.000000000000000D+00,   7.491631336111697D-10, &
   0.000000000000000D+00,   3.673039821697282D-11,   0.000000000000000D+00, &
  -2.168113773032231D-11,   0.000000000000000D+00,  -2.721267749679910D-06, &
   0.000000000000000D+00,  -2.073693338660516D-07,   0.000000000000000D+00, &
  -2.082074180776383D-08,   0.000000000000000D+00,  -1.958954570913974D-09, &
   0.000000000000000D+00,   1.530433256443777D-11,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.263550809175753D-05,   0.000000000000000D+00, &
   1.384798490247918D-06,   0.000000000000000D+00,   1.984397940456065D-07, &
   0.000000000000000D+00,   2.766773562181054D-08,   0.000000000000000D+00, &
   2.683793958774408D-09,   0.000000000000000D+00,  -1.174680342119233D-09, &
   0.000000000000000D+00,  -1.169193607997234D-05,   0.000000000000000D+00, &
  -2.632347844049752D-06,   0.000000000000000D+00,  -7.384022485977567D-07, &
   0.000000000000000D+00,  -1.625303050353998D-07,   0.000000000000000D+00, &
  -2.850672797699199D-08,   0.000000000000000D+00,  -1.440789552333344D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,  -9.247433517326246D-05, &
   0.000000000000000D+00,  -5.796282324523070D-06,   0.000000000000000D+00, &
   2.481506551958836D-07,   0.000000000000000D+00,   3.394358519428652D-07, &
   0.000000000000000D+00,   1.154001378595954D-07,   0.000000000000000D+00, &
   2.379653455007376D-08,   0.000000000000000D+00,  -5.657608473356593D-09, &
   0.000000000000000D+00,   3.757151533908318D-04,   0.000000000000000D+00, &
   1.874792966113659D-05,   0.000000000000000D+00,   4.455390625466726D-06, &
   0.000000000000000D+00,   4.686935834051351D-07,   0.000000000000000D+00, &
  -1.016698758991308D-07,   0.000000000000000D+00,  -7.149630139944640D-08, &
   0.000000000000000D+00,  -1.577497866672470D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.908879220807053D-04,   0.000000000000000D+00, &
   6.386531157646577D-05,   0.000000000000000D+00,  -4.331966497751265D-06, &
   0.000000000000000D+00,  -2.536969917141784D-06,   0.000000000000000D+00, &
  -5.739163382396233D-07,   0.000000000000000D+00,  -1.853429942042377D-08, &
   0.000000000000000D+00,   4.168761042655881D-08,   0.000000000000000D+00, &
   2.066852187741092D-08,   0.000000000000000D+00,   2.160539495867563D-04, &
   0.000000000000000D+00,  -3.674749347867609D-04,   0.000000000000000D+00, &
  -2.032029502348146D-05,   0.000000000000000D+00,  -4.992528454829840D-07, &
   0.000000000000000D+00,   1.142658301507520D-06,   0.000000000000000D+00, &
   4.524905174146304D-07,   0.000000000000000D+00,   5.743350691423328D-08, &
   0.000000000000000D+00,  -3.347043228863592D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.278939819388261D-04,   0.000000000000000D+00, &
   6.523795824663346D-04,   0.000000000000000D+00,   1.071178341214324D-05, &
   0.000000000000000D+00,   9.579788684084681D-06,   0.000000000000000D+00, &
   1.680793232655886D-06,   0.000000000000000D+00,  -3.136783918782093D-07, &
   0.000000000000000D+00,  -2.803475813892230D-07,   0.000000000000000D+00, &
  -4.927054340717054D-08,   0.000000000000000D+00,   1.019175996796216D-07, &
   0.000000000000000D+00,  -1.292331749713833D-03,   0.000000000000000D+00, &
  -2.866049144267170D-04,   0.000000000000000D+00,   1.836644083723721D-04, &
   0.000000000000000D+00,   6.538339521503504D-06,   0.000000000000000D+00, &
  -3.799489399865534D-06,   0.000000000000000D+00,  -1.520495196855275D-06, &
   0.000000000000000D+00,  -8.315172794979937D-08,   0.000000000000000D+00, &
   1.402611795802449D-07,   0.000000000000000D+00,   9.370895390406930D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,   7.683990603413082D-04, &
   0.000000000000000D+00,  -7.715442318197078D-04,   0.000000000000000D+00, &
  -4.571945625922477D-04,   0.000000000000000D+00,  -4.763960312141699D-05, &
   0.000000000000000D+00,  -7.757672176580976D-06,   0.000000000000000D+00, &
   7.841833648284720D-07,   0.000000000000000D+00,   9.750782105695845D-07, &
   0.000000000000000D+00,   1.989776777121191D-07,   0.000000000000000D+00, &
  -6.346197880444880D-08,   0.000000000000000D+00,  -1.451676304197259D-08, &
   0.000000000000000D+00,   2.408992881653844D-04,   0.000000000000000D+00, &
   1.421341546930961D-03,   0.000000000000000D+00,   2.816984895731512D-04, &
   0.000000000000000D+00,  -1.259358424240024D-05,   0.000000000000000D+00, &
   1.367804015007348D-05,   0.000000000000000D+00,   5.626654975095936D-06, &
   0.000000000000000D+00,   5.450659426261992D-07,   0.000000000000000D+00, &
  -4.474174254606454D-07,   0.000000000000000D+00,  -1.625262307735531D-07, &
   0.000000000000000D+00,   2.514544002121169D-08,   0.000000000000000D+00 /)
lem(:,284) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.269735923019527D-05, &
  -1.476059051883546D-04,   3.898794688840761D-04,   3.137467114806685D-04, &
  -1.111644455572382D-03,   4.188373294671743D-04,   5.605666151994539D-04, &
  -3.990220621817281D-03,   2.024658353270220D-03,   8.134848726640380D-03, &
  -3.226507285941499D-03,  -6.998454141848748D-03,   1.070312177940606D-03, &
   2.204069154752255D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.269735923019370D-05,  -1.476059051883478D-04,   3.898794688840691D-04, &
   3.137467114806618D-04,  -1.111644455572373D-03,   4.188373294671692D-04, &
   5.605666151994541D-04,  -3.990220621817287D-03,   2.024658353270215D-03, &
   8.134848726640359D-03,  -3.226507285941476D-03,  -6.998454141848738D-03, &
   1.070312177940583D-03,   2.204069154752237D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.540923287498840D-09,   0.000000000000000D+00, &
   1.141157337548749D-07,   0.000000000000000D+00,   1.298280232945372D-06, &
   0.000000000000000D+00,   4.805904539436045D-06,   0.000000000000000D+00, &
   5.509218724131010D-06,   0.000000000000000D+00,  -3.229315414398279D-06, &
   0.000000000000000D+00,  -8.902616227951762D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.075942785105441D-07,   0.000000000000000D+00, &
   9.432051988252860D-09,   0.000000000000000D+00,   5.141315622231244D-10, &
   0.000000000000000D+00,  -2.375287806226161D-11,   0.000000000000000D+00, &
  -4.573362197581300D-06,   0.000000000000000D+00,  -1.971387090522940D-07, &
   0.000000000000000D+00,  -1.636384362950219D-08,   0.000000000000000D+00, &
  -8.773719377248601D-10,   0.000000000000000D+00,   6.487637400714108D-10, &
   1.665656489610244D-05,   0.000000000000000D+00,   1.340289769804136D-06, &
   0.000000000000000D+00,   1.654406602196844D-07,   0.000000000000000D+00, &
   1.859872670786714D-08,   0.000000000000000D+00,  -3.918675765687707D-10, &
   0.000000000000000D+00,   2.579885217513817D-05,   0.000000000000000D+00, &
  -2.654723965015260D-06,   0.000000000000000D+00,  -6.349531220327396D-07, &
   0.000000000000000D+00,  -1.236631231625098D-07,   0.000000000000000D+00, &
  -1.480850315581006D-08,   0.000000000000000D+00,   8.792547312964153D-09, &
  -1.819055558908901D-04,   0.000000000000000D+00,  -4.460021540421971D-06, &
   0.000000000000000D+00,   2.049950208667283D-07,   0.000000000000000D+00, &
   2.725725368784435D-07,   0.000000000000000D+00,   8.028668979115336D-08, &
   0.000000000000000D+00,   4.931918549626985D-09,   0.000000000000000D+00, &
   2.897065188500171D-05,   0.000000000000000D+00,   1.156729562664224D-05, &
   0.000000000000000D+00,   4.054009720035296D-06,   0.000000000000000D+00, &
   4.041036745748266D-07,   0.000000000000000D+00,  -8.029859567928667D-08, &
   0.000000000000000D+00,  -4.637308584985502D-08,   0.000000000000000D+00, &
   7.846269848622232D-09,   8.383296766524495D-04,   0.000000000000000D+00, &
  -1.762179849762337D-06,   0.000000000000000D+00,  -4.932364754042550D-06, &
   0.000000000000000D+00,  -2.264282385873360D-06,   0.000000000000000D+00, &
  -4.653729762904867D-07,   0.000000000000000D+00,  -2.909440242144060D-09, &
   0.000000000000000D+00,   3.005994164775927D-08,   0.000000000000000D+00, &
  -1.068260772763359D-03,   0.000000000000000D+00,   1.352491284223208D-04, &
   0.000000000000000D+00,  -1.228745231311116D-05,   0.000000000000000D+00, &
   2.065510025886052D-07,   0.000000000000000D+00,   1.035315185060441D-06, &
   0.000000000000000D+00,   3.506737758352459D-07,   0.000000000000000D+00, &
   1.739135177812572D-08,   0.000000000000000D+00,  -9.518750310047185D-08, &
  -7.132938761341720D-04,   0.000000000000000D+00,  -5.242191407015943D-04, &
   0.000000000000000D+00,   5.555865693459801D-05,   0.000000000000000D+00, &
   8.035799574212992D-06,   0.000000000000000D+00,   1.131112521114982D-06, &
   0.000000000000000D+00,  -3.293777414432115D-07,   0.000000000000000D+00, &
  -2.109735572490951D-07,   0.000000000000000D+00,   1.422948719880278D-08, &
   0.000000000000000D+00,   2.168442287488296D-03,   0.000000000000000D+00, &
   4.830398905853312D-04,   0.000000000000000D+00,  -2.028641725038841D-04, &
   0.000000000000000D+00,  -9.232812747115577D-06,   0.000000000000000D+00, &
  -4.218512049495033D-06,   0.000000000000000D+00,  -1.209520689257602D-06, &
   0.000000000000000D+00,  -5.150240348523744D-09,   0.000000000000000D+00, &
   1.127221407897013D-07,   0.000000000000000D+00,  -8.833489987710955D-08, &
  -7.569359840893331D-04,   0.000000000000000D+00,   5.731979646554537D-04, &
   0.000000000000000D+00,   4.585596588505354D-04,   0.000000000000000D+00, &
  -1.587417233194586D-05,   0.000000000000000D+00,   6.869296386490642D-08, &
   0.000000000000000D+00,   1.771561011526714D-06,   0.000000000000000D+00, &
   8.892304893331483D-07,   0.000000000000000D+00,   1.077522310172104D-07, &
   0.000000000000000D+00,  -3.859410221116349D-08,   0.000000000000000D+00, &
  -1.203259435108714D-03,   0.000000000000000D+00,  -1.372620714326772D-03, &
   0.000000000000000D+00,  -3.197781293773501D-04,   0.000000000000000D+00, &
   1.260511574776833D-04,   0.000000000000000D+00,   1.256732477510695D-05, &
   0.000000000000000D+00,   2.108107176348652D-06,   0.000000000000000D+00, &
  -4.310400674018314D-07,   0.000000000000000D+00,  -5.255708968553366D-07, &
   0.000000000000000D+00,  -1.028730528411170D-07,   0.000000000000000D+00, &
   2.739214636467384D-07,   1.005170081656508D-03,   0.000000000000000D+00, &
   6.070719419217208D-04,   0.000000000000000D+00,  -5.561962712419447D-04, &
   0.000000000000000D+00,  -3.215473556424558D-04,   0.000000000000000D+00, &
  -2.525381129024642D-05,   0.000000000000000D+00,  -6.739255939648318D-06, &
   0.000000000000000D+00,  -2.125662277343869D-06,   0.000000000000000D+00, &
  -1.833099862989341D-07,   0.000000000000000D+00,   2.528023444542473D-07, &
   0.000000000000000D+00,   4.046791462934332D-08,   0.000000000000000D+00, &
   3.431076885416128D-05,   0.000000000000000D+00,   5.435970211855232D-04, &
   0.000000000000000D+00,   1.180844195777839D-03,   0.000000000000000D+00, &
   2.446591312841147D-04,   0.000000000000000D+00,  -1.349373662333213D-05, &
   0.000000000000000D+00,   3.823879013882323D-06,   0.000000000000000D+00, &
   2.926688618455078D-06,   0.000000000000000D+00,   1.547689129493714D-06, &
   0.000000000000000D+00,   3.796817145400408D-07,   0.000000000000000D+00, &
  -1.369653542203758D-07,   0.000000000000000D+00,   4.458671694458606D-08 /)
lem(:,285) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -3.790432550820729D-06,   5.677786588025998D-05,  -2.321674097481004D-04, &
   4.638164818304661D-05,   9.125820823390491D-04,  -3.254847649028467D-04, &
  -1.203145273399015D-03,  -1.720494249822064D-03,   9.805665661694934D-04, &
   7.568826519317641D-03,  -1.540519102552758D-03,  -1.155448317898824D-02, &
   1.963507127428543D-03,   8.474523327029118D-03,  -5.009123050499085D-04, &
  -2.794404960415972D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.790432550829609D-06,  -5.677786588025031D-05, &
   2.321674097480965D-04,  -4.638164818304852D-05,  -9.125820823390559D-04, &
   3.254847649028502D-04,   1.203145273399012D-03,   1.720494249822065D-03, &
  -9.805665661694973D-04,  -7.568826519317635D-03,   1.540519102552765D-03, &
   1.155448317898822D-02,  -1.963507127428547D-03,  -8.474523327029112D-03, &
   5.009123050499075D-04,   2.794404960415950D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -4.874076277391457D-08,   0.000000000000000D+00,  -1.003625313340645D-06, &
   0.000000000000000D+00,  -5.361025885081817D-06,   0.000000000000000D+00, &
  -9.469824202974018D-06,   0.000000000000000D+00,  -1.009110409834493D-06, &
   0.000000000000000D+00,   9.775319625484858D-06,   0.000000000000000D+00, &
   3.564260233828023D-06,   0.000000000000000D+00,   1.164051211534770D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.058699712978415D-08,   0.000000000000000D+00, &
   3.272080156959073D-10,   0.000000000000000D+00,  -1.603319280005432D-10, &
   0.000000000000000D+00,  -2.209373963500383D-07,   0.000000000000000D+00, &
  -1.337008308517458D-08,   0.000000000000000D+00,   3.326887501488398D-10, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.560788051842864D-06, &
   0.000000000000000D+00,   1.462352389883015D-07,   0.000000000000000D+00, &
   1.057035083498611D-08,   0.000000000000000D+00,  -6.927945776196511D-09, &
   0.000000000000000D+00,  -3.760920449098265D-06,   0.000000000000000D+00, &
  -5.861465742692736D-07,   0.000000000000000D+00,  -9.366265577293958D-08, &
   0.000000000000000D+00,   2.948485060975339D-10,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.260642468079286D-06,   0.000000000000000D+00, &
   2.344655069519129D-07,   0.000000000000000D+00,   2.291084539223320D-07, &
   0.000000000000000D+00,   5.003683017731743D-08,   0.000000000000000D+00, &
  -2.741463701113055D-08,   0.000000000000000D+00,   3.618481855394543D-05, &
   0.000000000000000D+00,   3.671805441246071D-06,   0.000000000000000D+00, &
   3.352981115815769D-07,   0.000000000000000D+00,  -6.938194734377430D-08, &
   0.000000000000000D+00,  -2.030494870522078D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.150375417016624D-05,   0.000000000000000D+00, &
  -3.664308476192686D-06,   0.000000000000000D+00,  -2.003784631307719D-06, &
   0.000000000000000D+00,  -3.623903988832829D-07,   0.000000000000000D+00, &
   1.859775639873156D-08,   0.000000000000000D+00,   5.754689294316654D-08, &
   0.000000000000000D+00,  -9.452457939664434D-05,   0.000000000000000D+00, &
  -1.762666737027512D-05,   0.000000000000000D+00,   3.285255868268361D-07, &
   0.000000000000000D+00,   8.941497882027376D-07,   0.000000000000000D+00, &
   2.483882407801735D-07,   0.000000000000000D+00,  -3.626082694628764D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.019714986298608D-04, &
   0.000000000000000D+00,   9.941397323907679D-06,   0.000000000000000D+00, &
   6.070301974402063D-06,   0.000000000000000D+00,   8.094899555285661D-07, &
   0.000000000000000D+00,  -3.035989639674472D-07,   0.000000000000000D+00, &
  -1.355443964002634D-07,   0.000000000000000D+00,   1.890533255898805D-07, &
   0.000000000000000D+00,  -2.148884341818009D-04,   0.000000000000000D+00, &
   1.599646342080253D-04,   0.000000000000000D+00,  -2.310684650970229D-06, &
   0.000000000000000D+00,  -3.035962484862383D-06,   0.000000000000000D+00, &
  -8.563638813271654D-07,   0.000000000000000D+00,   5.817837258365896D-08, &
   0.000000000000000D+00,   3.745958351127317D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.753168781571717D-04,   0.000000000000000D+00, &
  -4.026563383271306D-04,   0.000000000000000D+00,   1.803174321883813D-05, &
   0.000000000000000D+00,  -2.218174423166447D-07,   0.000000000000000D+00, &
   1.259176191226957D-06,   0.000000000000000D+00,   5.975794155157572D-07, &
   0.000000000000000D+00,  -6.647335842722019D-09,   0.000000000000000D+00, &
  -2.232758747372727D-07,   0.000000000000000D+00,   6.025664101253570D-04, &
   0.000000000000000D+00,   2.498804151916682D-04,   0.000000000000000D+00, &
  -1.684713688123103D-04,   0.000000000000000D+00,  -1.463022329491763D-06, &
   0.000000000000000D+00,   8.587459453778578D-07,   0.000000000000000D+00, &
  -3.825531569709320D-07,   0.000000000000000D+00,  -3.390279793660200D-07, &
   0.000000000000000D+00,   1.801189445698958D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.929401128454495D-04,   0.000000000000000D+00, &
   4.306760360418309D-04,   0.000000000000000D+00,   3.844757068796200D-04, &
   0.000000000000000D+00,   4.426187792148556D-07,   0.000000000000000D+00, &
  -3.108452209191699D-08,   0.000000000000000D+00,  -7.188408854951918D-07, &
   0.000000000000000D+00,   4.209744727239519D-08,   0.000000000000000D+00, &
   2.292297124220746D-07,   0.000000000000000D+00,  -3.147947993434675D-07, &
   0.000000000000000D+00,  -2.424139141286396D-04,   0.000000000000000D+00, &
  -8.782343053886232D-04,   0.000000000000000D+00,  -2.258982231420233D-04, &
   0.000000000000000D+00,   9.487049989444262D-05,   0.000000000000000D+00, &
   3.719797435424997D-06,   0.000000000000000D+00,   2.216009561166740D-07, &
   0.000000000000000D+00,   4.501809086849846D-07,   0.000000000000000D+00, &
   1.775965722217278D-08,   0.000000000000000D+00,  -4.781878854178976D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.647045652884812D-04, &
   0.000000000000000D+00,   4.574088218423572D-04,   0.000000000000000D+00, &
  -4.457591390083918D-04,   0.000000000000000D+00,  -2.703813820643770D-04, &
   0.000000000000000D+00,  -2.342564740381300D-05,   0.000000000000000D+00, &
  -3.003872029175900D-06,   0.000000000000000D+00,  -2.919188994911613D-07, &
   0.000000000000000D+00,  -2.703435741324592D-07,   0.000000000000000D+00, &
   9.058833258333245D-09,   0.000000000000000D+00,   5.705327277045459D-07, &
   0.000000000000000D+00,  -1.243542286814037D-04,   0.000000000000000D+00, &
   2.816984895731512D-04,   0.000000000000000D+00,   8.881875518674175D-04, &
   0.000000000000000D+00,   1.922422718375659D-04,   0.000000000000000D+00, &
  -5.021119707244868D-06,   0.000000000000000D+00,   6.322274356500995D-06, &
   0.000000000000000D+00,   1.943982354973991D-06,   0.000000000000000D+00, &
   3.466650516522377D-07,   0.000000000000000D+00,   1.713564140388741D-07, &
   0.000000000000000D+00,   4.236021137063723D-08,   0.000000000000000D+00 /)
lem(:,286) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.243530376981887D-06,  -2.132122244227582D-05, &
   1.120937808858848D-04,  -1.243390400671266D-04,  -3.967381614680134D-04, &
   5.842515966189254D-04,   1.819976181676072D-04,  -8.613834469990378D-04, &
   3.342972475248404D-04,   3.779387927108123D-03,  -7.604566835230654D-04, &
  -1.099768122379781D-02,   1.436345864909935D-03,   1.472743921232727D-02, &
  -1.799917090191172D-03,  -9.813571821961273D-03,   8.781567298416907D-04, &
   3.376674557314919D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.243530376984503D-06, &
  -2.132122244228464D-05,   1.120937808858833D-04,  -1.243390400671177D-04, &
  -3.967381614680074D-04,   5.842515966189160D-04,   1.819976181675991D-04, &
  -8.613834469990212D-04,   3.342972475248436D-04,   3.779387927108114D-03, &
  -7.604566835230608D-04,  -1.099768122379779D-02,   1.436345864909927D-03, &
   1.472743921232727D-02,  -1.799917090191148D-03,  -9.813571821961235D-03, &
   8.781567298416714D-04,   3.376674557314911D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.139733241904403D-08,   0.000000000000000D+00,   6.322587182111900D-07, &
   0.000000000000000D+00,   5.231721784841231D-06,   0.000000000000000D+00, &
   1.332344696351763D-05,   0.000000000000000D+00,   7.600661446436790D-06, &
   0.000000000000000D+00,  -1.001761981832768D-05,   0.000000000000000D+00, &
  -5.910391891397722D-06,   0.000000000000000D+00,   4.339061090294847D-06, &
   0.000000000000000D+00,  -6.166440221865851D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.005647789597624D-08,   0.000000000000000D+00,   1.653006419842522D-10, &
   0.000000000000000D+00,  -4.005027929794071D-07,   0.000000000000000D+00, &
  -1.182106216077176D-08,   0.000000000000000D+00,   3.100678448611911D-09, &
   2.523044264604196D-06,   0.000000000000000D+00,   1.417220222879505D-07, &
   0.000000000000000D+00,   1.606829946258047D-09,   0.000000000000000D+00, &
  -3.132292129842922D-06,   0.000000000000000D+00,  -6.002643469503280D-07, &
   0.000000000000000D+00,  -6.822153441327561D-08,   0.000000000000000D+00, &
   3.153383858567032D-08,  -1.971433991549795D-05,   0.000000000000000D+00, &
   4.023621121322074D-07,   0.000000000000000D+00,   2.025124152393751D-07, &
   0.000000000000000D+00,   1.578134333401319D-08,   0.000000000000000D+00, &
   4.387610906754082D-05,   0.000000000000000D+00,   2.870340808410676D-06, &
   0.000000000000000D+00,   2.608444597796202D-07,   0.000000000000000D+00, &
  -6.498666583649367D-08,   0.000000000000000D+00,  -2.979999711157653D-09, &
   6.176176998501419D-05,   0.000000000000000D+00,  -3.331858289556837D-06, &
   0.000000000000000D+00,  -1.774834162934798D-06,   0.000000000000000D+00, &
  -2.627924419534741D-07,   0.000000000000000D+00,   4.597379744062087D-08, &
   0.000000000000000D+00,  -1.858537219598393D-04,   0.000000000000000D+00, &
  -8.269890654143825D-07,   0.000000000000000D+00,   2.851098739653239D-07, &
   0.000000000000000D+00,   7.687196453576160D-07,   0.000000000000000D+00, &
   1.471699098429258D-07,   0.000000000000000D+00,  -2.427385990480124D-07, &
  -9.415110646724764D-05,   0.000000000000000D+00,  -1.907057388144435D-05, &
   0.000000000000000D+00,   6.633090982621242D-06,   0.000000000000000D+00, &
   6.059952546997325D-07,   0.000000000000000D+00,  -2.776344762325870D-07, &
   0.000000000000000D+00,  -1.829599565305326D-08,   0.000000000000000D+00, &
   4.176938797670679D-04,   0.000000000000000D+00,   2.283533000081281D-06, &
   0.000000000000000D+00,  -8.432469217346658D-06,   0.000000000000000D+00, &
  -2.529740725476107D-06,   0.000000000000000D+00,  -6.108834377726716D-07, &
   0.000000000000000D+00,   1.158196746397801D-07,   0.000000000000000D+00, &
   5.586818676649600D-08,   5.097163229124507D-05,   0.000000000000000D+00, &
   1.789501125204395D-04,   0.000000000000000D+00,  -1.240358294024620D-05, &
   0.000000000000000D+00,  -1.377337133702468D-06,   0.000000000000000D+00, &
   9.463565375029841D-07,   0.000000000000000D+00,   3.744413075370888D-07, &
   0.000000000000000D+00,  -6.431517924570710D-08,   0.000000000000000D+00, &
  -5.665790301456866D-04,   0.000000000000000D+00,  -2.485804735978500D-04, &
   0.000000000000000D+00,   1.089746952637811D-04,   0.000000000000000D+00, &
   5.147693781273942D-06,   0.000000000000000D+00,   1.297063423142190D-06, &
   0.000000000000000D+00,  -2.446107872754382D-07,   0.000000000000000D+00, &
  -2.104720292062854D-07,   0.000000000000000D+00,   5.857339444219810D-07, &
   5.490039025748043D-05,   0.000000000000000D+00,  -1.690861711948870D-04, &
   0.000000000000000D+00,  -2.820992464617307D-04,   0.000000000000000D+00, &
   2.281246077076592D-05,   0.000000000000000D+00,  -3.993397184553721D-07, &
   0.000000000000000D+00,  -7.750462378506711D-07,   0.000000000000000D+00, &
   3.657929763922669D-08,   0.000000000000000D+00,   1.518665772117676D-08, &
   0.000000000000000D+00,   4.785478661335768D-04,   0.000000000000000D+00, &
   5.281364032406298D-04,   0.000000000000000D+00,   2.158766811774879D-04, &
   0.000000000000000D+00,  -1.372928262234713D-04,   0.000000000000000D+00, &
  -7.665486913469934D-06,   0.000000000000000D+00,  -6.303753208871602D-07, &
   0.000000000000000D+00,   3.128197242727889D-07,   0.000000000000000D+00, &
  -1.022315621811394D-07,   0.000000000000000D+00,  -3.987508981432228D-07, &
  -1.176375227757986D-04,   0.000000000000000D+00,  -1.805118797033451D-04, &
   0.000000000000000D+00,   3.176547135837128D-04,   0.000000000000000D+00, &
   2.968813811726063D-04,   0.000000000000000D+00,   3.078068035736711D-07, &
   0.000000000000000D+00,   2.597857656062465D-06,   0.000000000000000D+00, &
   6.821789222816936D-07,   0.000000000000000D+00,  -2.582483759443241D-09, &
   0.000000000000000D+00,  -9.848378743892824D-09,   0.000000000000000D+00, &
  -2.438633578399640D-04,   0.000000000000000D+00,  -2.833044242780526D-04, &
   0.000000000000000D+00,  -7.066609367702738D-04,   0.000000000000000D+00, &
  -1.877220966311915D-04,   0.000000000000000D+00,   8.040003899217336D-05, &
   0.000000000000000D+00,   5.509705378853024D-06,   0.000000000000000D+00, &
  -4.113567515544922D-07,   0.000000000000000D+00,  -4.320157372040117D-07, &
   0.000000000000000D+00,   7.382175302162306D-08,   0.000000000000000D+00, &
  -5.580115795902997D-07,   8.947372184893196D-05,   0.000000000000000D+00, &
   2.182834490005913D-04,   0.000000000000000D+00,   3.381927832479124D-04, &
   0.000000000000000D+00,  -3.519337023407202D-04,   0.000000000000000D+00, &
  -2.177340237026636D-04,   0.000000000000000D+00,  -2.104765297285690D-05, &
   0.000000000000000D+00,  -4.400471211265544D-06,   0.000000000000000D+00, &
  -4.832712872895759D-07,   0.000000000000000D+00,   1.391490416819402D-07, &
   0.000000000000000D+00,   1.995618817599720D-07,   0.000000000000000D+00, &
   6.569454744484824D-05,   0.000000000000000D+00,   2.022429382292918D-06, &
   0.000000000000000D+00,   2.446591312841147D-04,   0.000000000000000D+00, &
   7.256942864982596D-04,   0.000000000000000D+00,   1.606054129713848D-04, &
   0.000000000000000D+00,  -5.233960483672621D-06,   0.000000000000000D+00, &
   5.152837384580692D-06,   0.000000000000000D+00,   2.649685411223737D-06, &
   0.000000000000000D+00,   6.827833132901565D-07,   0.000000000000000D+00, &
   1.748451466049415D-07,   0.000000000000000D+00,   5.711932548426334D-07 /)
lem(:,287) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.165186073198546D-07, &
   6.034800661291727D-06,  -4.832397472197545D-05,   1.135122910187729D-04, &
   1.361104215237279D-04,  -5.824692622257551D-04,  -1.933359680362663D-04, &
   6.692803723871540D-04,   1.042891292881920D-03,   9.011782415879827D-04, &
  -1.786195728602802D-03,  -6.536499284947641D-03,   1.532444413203296D-03, &
   1.453818091170681D-02,  -2.203338841924259D-03,  -1.651504563384562D-02, &
   3.487258920712218D-03,   1.017532113239370D-02,  -2.694366264536625D-03, &
  -3.553576573633010D-03,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.165186073113118D-07,  -6.034800661291832D-06,   4.832397472198530D-05, &
  -1.135122910187795D-04,  -1.361104215237319D-04,   5.824692622257620D-04, &
   1.933359680362719D-04,  -6.692803723871550D-04,  -1.042891292881929D-03, &
  -9.011782415879767D-04,   1.786195728602808D-03,   6.536499284947632D-03, &
  -1.532444413203303D-03,  -1.453818091170680D-02,   2.203338841924260D-03, &
   1.651504563384563D-02,  -3.487258920712217D-03,  -1.017532113239372D-02, &
   2.694366264536618D-03,   3.553576573633006D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.776140501356878D-07, &
   0.000000000000000D+00,  -4.337757607265784D-06,   0.000000000000000D+00, &
  -1.620463030064949D-05,   0.000000000000000D+00,  -1.655601276157301D-05, &
   0.000000000000000D+00,   7.697898241169488D-06,   0.000000000000000D+00, &
   1.135642144882996D-05,   0.000000000000000D+00,  -7.892542956971879D-06, &
   0.000000000000000D+00,  -2.673731161777537D-06,   0.000000000000000D+00, &
   8.618714838289651D-06,   0.000000000000000D+00,   1.208130274647383D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.335393043563064D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.695795671872977D-07,   0.000000000000000D+00, &
  -1.703777331440426D-08,   0.000000000000000D+00,  -7.878563134604132D-07, &
   0.000000000000000D+00,  -4.386518197141286D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.671781816125761D-07,   0.000000000000000D+00, &
   1.959274905849259D-07,   0.000000000000000D+00,  -3.414749895737554D-08, &
   0.000000000000000D+00,   4.257678236419093D-06,   0.000000000000000D+00, &
   1.633112302464342D-07,   0.000000000000000D+00,  -6.334478606887480D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.256924046105279D-05, &
   0.000000000000000D+00,  -1.505387763952033D-06,   0.000000000000000D+00, &
  -1.571600267417627D-07,   0.000000000000000D+00,   2.275576111678755D-07, &
   0.000000000000000D+00,  -1.853303474099336D-06,   0.000000000000000D+00, &
  -3.243252853655003D-07,   0.000000000000000D+00,   6.539326981029621D-07, &
   0.000000000000000D+00,   9.954014262140822D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.557153833568637D-05,   0.000000000000000D+00, &
   6.763632317583372D-06,   0.000000000000000D+00,   4.580370314599601D-07, &
   0.000000000000000D+00,  -2.646681462283288D-07,   0.000000000000000D+00, &
   1.717028529964347D-07,   0.000000000000000D+00,  -3.649195230708370D-05, &
   0.000000000000000D+00,   4.682030758260896D-06,   0.000000000000000D+00, &
  -2.299635590798435D-06,   0.000000000000000D+00,  -3.841334256853866D-07, &
   0.000000000000000D+00,   9.926224986618239D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.390590125185929D-05,   0.000000000000000D+00, &
  -3.606307347411875D-05,   0.000000000000000D+00,   6.150061325869984D-08, &
   0.000000000000000D+00,   7.582985867820806D-07,   0.000000000000000D+00, &
   1.864155457270626D-07,   0.000000000000000D+00,  -7.315827886631377D-07, &
   0.000000000000000D+00,   1.182680425585440D-04,   0.000000000000000D+00, &
  -7.813143621033870D-06,   0.000000000000000D+00,  -2.097688995402538D-06, &
   0.000000000000000D+00,   1.189308500340803D-06,   0.000000000000000D+00, &
  -1.992309504049048D-07,   0.000000000000000D+00,   7.779247111660590D-09, &
   0.000000000000000D+00,   0.000000000000000D+00,   4.107436271826624D-05, &
   0.000000000000000D+00,   1.579088791112338D-04,   0.000000000000000D+00, &
   1.631171598803688D-06,   0.000000000000000D+00,  -1.788099484354864D-06, &
   0.000000000000000D+00,  -7.116650957987761D-07,   0.000000000000000D+00, &
   1.702826596923537D-07,   0.000000000000000D+00,   1.778064945470548D-08, &
   0.000000000000000D+00,  -1.743137893140169D-04,   0.000000000000000D+00, &
  -1.636163454536218D-04,   0.000000000000000D+00,   8.119637107568813D-05, &
   0.000000000000000D+00,   4.748933576281288D-07,   0.000000000000000D+00, &
   1.427913560211364D-07,   0.000000000000000D+00,   2.557914099705574D-07, &
   0.000000000000000D+00,   6.762173464247981D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.434899858404484D-05,   0.000000000000000D+00, &
  -1.370659240460748D-04,   0.000000000000000D+00,  -2.359250712674516D-04, &
   0.000000000000000D+00,   1.470209445117988D-05,   0.000000000000000D+00, &
   1.770810919210860D-06,   0.000000000000000D+00,   2.904767180654759D-07, &
   0.000000000000000D+00,  -1.371588125644419D-07,   0.000000000000000D+00, &
   8.884487379924595D-07,   0.000000000000000D+00,   1.358501383325397D-04, &
   0.000000000000000D+00,   3.627551563211225D-04,   0.000000000000000D+00, &
   1.693943629703174D-04,   0.000000000000000D+00,  -1.072335732891505D-04, &
   0.000000000000000D+00,  -5.233609340894733D-06,   0.000000000000000D+00, &
  -1.140163236766463D-06,   0.000000000000000D+00,  -3.481020275740987D-07, &
   0.000000000000000D+00,  -1.247823996086333D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.295676598484422D-05,   0.000000000000000D+00, &
  -1.469109739818385D-04,   0.000000000000000D+00,   2.581590194348227D-04, &
   0.000000000000000D+00,   2.490162673218992D-04,   0.000000000000000D+00, &
   1.790395351492428D-06,   0.000000000000000D+00,   1.162251706925325D-06, &
   0.000000000000000D+00,   5.964259706634066D-07,   0.000000000000000D+00, &
   2.593694541208614D-08,   0.000000000000000D+00,  -4.008818027123064D-07, &
   0.000000000000000D+00,  -4.658152632275726D-05,   0.000000000000000D+00, &
  -1.868966345632514D-04,   0.000000000000000D+00,  -5.547991677495476D-04, &
   0.000000000000000D+00,  -1.557060704771083D-04,   0.000000000000000D+00, &
   6.346631523324063D-05,   0.000000000000000D+00,   4.165661062701285D-06, &
   0.000000000000000D+00,   2.176800683616010D-07,   0.000000000000000D+00, &
  -9.056885498254973D-08,   0.000000000000000D+00,  -3.723997872203808D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.647150991188986D-05, &
   0.000000000000000D+00,   1.950709221004012D-04,   0.000000000000000D+00, &
   2.756047391035378D-04,   0.000000000000000D+00,  -2.917739169259423D-04, &
   0.000000000000000D+00,  -1.846978126254659D-04,   0.000000000000000D+00, &
  -1.826533039927326D-05,   0.000000000000000D+00,  -3.424353894107841D-06, &
   0.000000000000000D+00,  -6.337870078690225D-07,   0.000000000000000D+00, &
   2.027475479318410D-07,   0.000000000000000D+00,  -3.534986366361005D-08, &
   0.000000000000000D+00,  -5.354960346073285D-06,   0.000000000000000D+00, &
  -1.259358424240024D-05,   0.000000000000000D+00,   1.922422718375659D-04, &
   0.000000000000000D+00,   5.947626138724019D-04,   0.000000000000000D+00, &
   1.365138580388074D-04,   0.000000000000000D+00,  -1.830397359992888D-06, &
   0.000000000000000D+00,   4.951492393424260D-06,   0.000000000000000D+00, &
   2.068737302012766D-06,   0.000000000000000D+00,   7.983954766882115D-07, &
   0.000000000000000D+00,   3.905933339953730D-07,   0.000000000000000D+00 /)
lem(:,288) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.359821014401073D-08,  -2.327864578552854D-06,   1.917100510213443D-05, &
  -5.452823290563587D-05,  -1.922252313409928D-05,   2.804327615848416D-04, &
  -1.694490741598188D-04,  -2.347211557549217D-04,   4.636829201086275D-04, &
   5.106267130805294D-05,  -1.583858092735750D-03,  -1.862947394395914D-03, &
   2.817336625974726D-03,   8.560852689277004D-03,  -3.049419713825020D-03, &
  -1.620061001192015D-02,   4.340109191015350D-03,   1.584894791125475D-02, &
  -6.278109389345597D-03,  -8.258282347732709D-03,   5.164056748820684D-03, &
   2.271381211664083D-03,   7.359821016274069D-08,  -2.327864578553271D-06, &
   1.917100510212105D-05,  -5.452823290562584D-05,  -1.922252313409374D-05, &
   2.804327615848269D-04,  -1.694490741598136D-04,  -2.347211557549112D-04, &
   4.636829201086244D-04,   5.106267130804670D-05,  -1.583858092735754D-03, &
  -1.862947394395911D-03,   2.817336625974739D-03,   8.560852689277001D-03, &
  -3.049419713825027D-03,  -1.620061001192017D-02,   4.340109191015351D-03, &
   1.584894791125475D-02,  -6.278109389345592D-03,  -8.258282347732722D-03, &
   5.164056748820685D-03,   2.271381211664069D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.506619261590031D-06,   0.000000000000000D+00, &
   1.639213582607971D-05,   0.000000000000000D+00,   2.607503749809489D-05, &
   0.000000000000000D+00,  -1.949869340844512D-06,   0.000000000000000D+00, &
  -1.972038129011333D-05,   0.000000000000000D+00,   7.314431318084812D-06, &
   0.000000000000000D+00,   1.208603807927189D-05,   0.000000000000000D+00, &
  -6.982375842867377D-07,   0.000000000000000D+00,  -3.570665105505552D-06, &
   0.000000000000000D+00,  -1.780988222460352D-05,   0.000000000000000D+00, &
  -4.131372228519565D-08,   3.329858195504313D-07,   0.000000000000000D+00, &
  -1.147937245073747D-06,   0.000000000000000D+00,  -2.268081622433245D-08, &
  -4.411590270409586D-07,   0.000000000000000D+00,   2.346587435700158D-07, &
   0.000000000000000D+00,   1.005805259130412D-05,   0.000000000000000D+00, &
  -5.269395198612485D-08,   0.000000000000000D+00,  -1.905163693947422D-07, &
  -7.277720524527080D-06,   0.000000000000000D+00,  -1.114906195645419D-06, &
   0.000000000000000D+00,  -1.540381659026662D-08,   0.000000000000000D+00, &
  -3.685110288856343D-05,   0.000000000000000D+00,   7.013762913260534D-07, &
   0.000000000000000D+00,   5.386354731886648D-07,   0.000000000000000D+00, &
  -3.727577869998249D-07,   3.377292092407631D-05,   0.000000000000000D+00, &
   6.842699758503759D-08,   0.000000000000000D+00,   4.205426914246842D-07, &
   0.000000000000000D+00,  -1.608962070417857D-07,   0.000000000000000D+00, &
   7.771190489024742D-05,   0.000000000000000D+00,   4.468392063962734D-06, &
   0.000000000000000D+00,  -2.774799133052859D-06,   0.000000000000000D+00, &
  -1.650086617860391D-07,   0.000000000000000D+00,   6.950544403084417D-07, &
  -5.910602841857933D-05,   0.000000000000000D+00,   9.857453178522480D-06, &
   0.000000000000000D+00,   8.465744871828291D-07,   0.000000000000000D+00, &
   6.399524663972971D-07,   0.000000000000000D+00,  -1.663791157252839D-08, &
   0.000000000000000D+00,  -1.119094176384784D-04,   0.000000000000000D+00, &
  -3.726641646135911D-05,   0.000000000000000D+00,   8.713167423094192D-06, &
   0.000000000000000D+00,   8.613822402368922D-07,   0.000000000000000D+00, &
  -2.847073080621138D-07,   0.000000000000000D+00,   5.297198706320437D-07, &
   3.373904569911279D-05,   0.000000000000000D+00,  -2.276597316614149D-05, &
   0.000000000000000D+00,  -2.030290672609923D-05,   0.000000000000000D+00, &
  -1.348542351445838D-07,   0.000000000000000D+00,  -5.135163771924727D-07, &
   0.000000000000000D+00,  -8.985184625886952D-08,   0.000000000000000D+00, &
   1.293951763618676D-04,   0.000000000000000D+00,   1.032169398662661D-04, &
   0.000000000000000D+00,  -7.878795778994631D-06,   0.000000000000000D+00, &
  -7.205413842529359D-06,   0.000000000000000D+00,   3.229479921033057D-07, &
   0.000000000000000D+00,   1.763574818556499D-07,   0.000000000000000D+00, &
  -1.129140776448501D-06,   3.637473975176332D-05,   0.000000000000000D+00, &
   1.426480301681308D-05,   0.000000000000000D+00,   1.121864033993638D-04, &
   0.000000000000000D+00,   2.110120852615431D-06,   0.000000000000000D+00, &
  -2.493488132243861D-07,   0.000000000000000D+00,   9.369132291168010D-08, &
   0.000000000000000D+00,  -4.729201585735729D-08,   0.000000000000000D+00, &
  -1.315030703557183D-04,   0.000000000000000D+00,  -1.488621318338421D-04, &
   0.000000000000000D+00,  -1.329741775603131D-04,   0.000000000000000D+00, &
   6.790267717088057D-05,   0.000000000000000D+00,   3.778567309572550D-06, &
   0.000000000000000D+00,  -2.187864208907325D-07,   0.000000000000000D+00, &
   2.288956986742698D-07,   0.000000000000000D+00,   1.903626408682982D-07, &
  -7.390124308420272D-05,   0.000000000000000D+00,   1.553921497714943D-05, &
   0.000000000000000D+00,  -1.010791275940242D-04,   0.000000000000000D+00, &
  -1.896920356816431D-04,   0.000000000000000D+00,   9.076915486275873D-06, &
   0.000000000000000D+00,   4.953803800254770D-07,   0.000000000000000D+00, &
   1.265989012129843D-07,   0.000000000000000D+00,   3.766586480335723D-07, &
   0.000000000000000D+00,   1.085150334952084D-04,   0.000000000000000D+00, &
   1.197579959714180D-04,   0.000000000000000D+00,   2.932036186530994D-04, &
   0.000000000000000D+00,   1.415639255945486D-04,   0.000000000000000D+00, &
  -8.994085021769675D-05,   0.000000000000000D+00,  -6.018933892275437D-06, &
   0.000000000000000D+00,  -7.422880442949688D-07,   0.000000000000000D+00, &
  -3.360770456626977D-07,   0.000000000000000D+00,   4.173923938473921D-07, &
   4.835190882873903D-05,   0.000000000000000D+00,  -2.916396560105545D-05, &
   0.000000000000000D+00,  -1.086532363175542D-04,   0.000000000000000D+00, &
   2.083940000685976D-04,   0.000000000000000D+00,   2.085949992565282D-04, &
   0.000000000000000D+00,   4.553402160804710D-06,   0.000000000000000D+00, &
   1.985055883125411D-06,   0.000000000000000D+00,   3.340838713714681D-07, &
   0.000000000000000D+00,   7.285670305160601D-08,   0.000000000000000D+00, &
  -6.164993823762254D-05,   0.000000000000000D+00,  -4.926511962561687D-05, &
   0.000000000000000D+00,  -1.524391142640039D-04,   0.000000000000000D+00, &
  -4.592259312367593D-04,   0.000000000000000D+00,  -1.334741022682386D-04, &
   0.000000000000000D+00,   5.323260327685528D-05,   0.000000000000000D+00, &
   3.645149122902624D-06,   0.000000000000000D+00,  -1.049485608857222D-07, &
   0.000000000000000D+00,  -5.228603000625393D-07,   0.000000000000000D+00, &
  -7.588510775675014D-07,  -1.096292428584015D-05,   0.000000000000000D+00, &
   1.384477896747589D-05,   0.000000000000000D+00,   1.393034605957514D-04, &
   0.000000000000000D+00,   2.229080484713501D-04,   0.000000000000000D+00, &
  -2.431670864303329D-04,   0.000000000000000D+00,  -1.577614266851603D-04, &
   0.000000000000000D+00,  -1.754303577645773D-05,   0.000000000000000D+00, &
  -3.541814406613975D-06,   0.000000000000000D+00,  -3.267161001867995D-07, &
   0.000000000000000D+00,  -1.043985903303732D-07,   0.000000000000000D+00, &
   2.004713428574508D-05,   0.000000000000000D+00,   6.020348653846282D-06, &
   0.000000000000000D+00,  -1.349373662333213D-05,   0.000000000000000D+00, &
   1.606054129713848D-04,   0.000000000000000D+00,   5.024709509100963D-04, &
   0.000000000000000D+00,   1.188199876353736D-04,   0.000000000000000D+00, &
  -4.544002858284752D-07,   0.000000000000000D+00,   4.988751144480321D-06, &
   0.000000000000000D+00,   2.456100334922962D-06,   0.000000000000000D+00, &
   1.025919284926382D-06,   0.000000000000000D+00,   6.146501992430868D-07 /)
lem(:,289) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.593082834028412D-05,  -3.668975462993597D-05, &
  -1.330999730520214D-04,   3.086018478635570D-04,   3.233308880611200D-04, &
  -3.912899194612646D-04,  -7.819579892236716D-04,  -6.607116687916085D-04, &
   6.500296743094968D-04,   2.775986281787053D-03,   2.497357616422591D-03, &
  -3.963374126028072D-03,  -9.361119558737483D-03,   4.338928916816620D-03, &
   1.478374971519912D-02,  -6.561864426595789D-03,  -1.214921996516293D-02, &
   8.396984277343334D-03,   4.275722873409716D-03,  -6.123981884057040D-03, &
   4.581641724068651D-04,   0.000000000000000D+00,  -1.593082834028032D-05, &
   3.668975462992290D-05,   1.330999730520224D-04,  -3.086018478635475D-04, &
  -3.233308880611198D-04,   3.912899194612588D-04,   7.819579892236659D-04, &
   6.607116687916146D-04,  -6.500296743094954D-04,  -2.775986281787061D-03, &
  -2.497357616422580D-03,   3.963374126028077D-03,   9.361119558737472D-03, &
  -4.338928916816623D-03,  -1.478374971519910D-02,   6.561864426595785D-03, &
   1.214921996516293D-02,  -8.396984277343353D-03,  -4.275722873409720D-03, &
   6.123981884057045D-03,  -4.581641724068643D-04,   0.000000000000000D+00, &
  -7.965414170140399D-06,   0.000000000000000D+00,  -2.806858637647980D-05, &
   0.000000000000000D+00,  -3.145313103900539D-06,   0.000000000000000D+00, &
   3.386511927828273D-05,   0.000000000000000D+00,   2.341630189506742D-06, &
   0.000000000000000D+00,  -2.103148612403700D-05,   0.000000000000000D+00, &
  -6.563888227537639D-06,   0.000000000000000D+00,  -7.739445604191605D-06, &
   0.000000000000000D+00,  -9.427681571708486D-07,   0.000000000000000D+00, &
   2.115716614846337D-05,   0.000000000000000D+00,   2.884850590281819D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.297212821056065D-07, &
   0.000000000000000D+00,  -2.948992971204457D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -2.357989969918697D-06,   0.000000000000000D+00, &
   5.339037625905814D-07,   0.000000000000000D+00,   4.018356769951131D-06, &
   0.000000000000000D+00,   2.176235530782622D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.280208048269495D-06,   0.000000000000000D+00, &
   5.724354303179536D-07,   0.000000000000000D+00,  -4.879897957258918D-07, &
   0.000000000000000D+00,  -1.602049654544037D-05,   0.000000000000000D+00, &
  -1.807898631347235D-06,   0.000000000000000D+00,  -6.270922487464588D-08, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.862080897808897D-06, &
   0.000000000000000D+00,  -4.612272670857870D-06,   0.000000000000000D+00, &
   7.975200413803352D-07,   0.000000000000000D+00,  -1.105575422703793D-06, &
   0.000000000000000D+00,   3.118834215924511D-05,   0.000000000000000D+00, &
   8.958780874747969D-06,   0.000000000000000D+00,  -2.610452430597083D-07, &
   0.000000000000000D+00,   3.454574347483774D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.532425822184630D-06,   0.000000000000000D+00, &
   1.571734630122343D-05,   0.000000000000000D+00,   7.913947403377642D-07, &
   0.000000000000000D+00,  -1.839027589836634D-07,   0.000000000000000D+00, &
   1.098826846704146D-06,   0.000000000000000D+00,  -3.682852130450033D-05, &
   0.000000000000000D+00,  -3.239099120009515D-05,   0.000000000000000D+00, &
   3.028124471899634D-06,   0.000000000000000D+00,   1.368069074859399D-07, &
   0.000000000000000D+00,   2.588814768376132D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.291159682054547D-06,   0.000000000000000D+00, &
  -2.636271651988766D-05,   0.000000000000000D+00,  -1.903565104710569D-05, &
   0.000000000000000D+00,   1.483757380289089D-06,   0.000000000000000D+00, &
  -4.425025217886840D-07,   0.000000000000000D+00,   3.331767548181195D-07, &
   0.000000000000000D+00,   3.716016798612354D-05,   0.000000000000000D+00, &
   7.374027779329453D-05,   0.000000000000000D+00,  -1.319300027889034D-06, &
   0.000000000000000D+00,  -5.700815069918568D-06,   0.000000000000000D+00, &
   1.161717145466750D-07,   0.000000000000000D+00,  -3.147281226316350D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.624959288410944D-06, &
   0.000000000000000D+00,   1.518657769008570D-05,   0.000000000000000D+00, &
   9.388342773355987D-05,   0.000000000000000D+00,   4.023780662664383D-06, &
   0.000000000000000D+00,  -1.617557944030206D-06,   0.000000000000000D+00, &
  -1.040665112506709D-07,   0.000000000000000D+00,  -9.611702506927345D-07, &
   0.000000000000000D+00,  -4.350602910272271D-05,   0.000000000000000D+00, &
  -1.022202003553103D-04,   0.000000000000000D+00,  -1.080070911692979D-04, &
   0.000000000000000D+00,   5.298053971740661D-05,   0.000000000000000D+00, &
   4.720127850363343D-06,   0.000000000000000D+00,   5.248997018587881D-07, &
   0.000000000000000D+00,  -9.005056477830635D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.893939850718507D-06,   0.000000000000000D+00, &
   1.666834233282375D-05,   0.000000000000000D+00,  -8.319303347659960D-05, &
   0.000000000000000D+00,  -1.609615542040215D-04,   0.000000000000000D+00, &
   5.013815257614232D-06,   0.000000000000000D+00,   9.428749674985042D-07, &
   0.000000000000000D+00,   5.524863600525038D-07,   0.000000000000000D+00, &
   5.952070113046300D-07,   0.000000000000000D+00,   4.366314868963289D-05, &
   0.000000000000000D+00,   8.580710924816366D-05,   0.000000000000000D+00, &
   2.337323650811333D-04,   0.000000000000000D+00,   1.208785461997040D-04, &
   0.000000000000000D+00,  -7.395539384780862D-05,   0.000000000000000D+00, &
  -6.628812149978505D-06,   0.000000000000000D+00,  -1.211113241053706D-06, &
   0.000000000000000D+00,   1.598168907742895D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.190839784086323D-06,   0.000000000000000D+00, &
  -3.466766239593811D-05,   0.000000000000000D+00,  -8.973240143509224D-05, &
   0.000000000000000D+00,   1.747923268600898D-04,   0.000000000000000D+00, &
   1.805463769806270D-04,   0.000000000000000D+00,   6.485187030527794D-06, &
   0.000000000000000D+00,   1.888264261638642D-06,   0.000000000000000D+00, &
   5.334928577315244D-07,   0.000000000000000D+00,   4.079527794332191D-07, &
   0.000000000000000D+00,  -2.586653146365229D-05,   0.000000000000000D+00, &
  -4.298110684655552D-05,   0.000000000000000D+00,  -1.237511188366805D-04, &
   0.000000000000000D+00,  -3.836488595851126D-04,   0.000000000000000D+00, &
  -1.173402751315254D-04,   0.000000000000000D+00,   4.364070134567229D-05, &
   0.000000000000000D+00,   3.464469554686875D-06,   0.000000000000000D+00, &
  -4.100922506188099D-07,   0.000000000000000D+00,  -6.533256229321111D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.252923347372544D-05, &
   0.000000000000000D+00,   2.428372260767170D-05,   0.000000000000000D+00, &
   1.175177276492078D-04,   0.000000000000000D+00,   1.872410332605680D-04, &
   0.000000000000000D+00,  -2.076926546641004D-04,   0.000000000000000D+00, &
  -1.384039661920313D-04,   0.000000000000000D+00,  -1.674795148564198D-05, &
   0.000000000000000D+00,  -3.399491856418174D-06,   0.000000000000000D+00, &
  -7.491890892121195D-07,   0.000000000000000D+00,  -1.331575244813013D-07, &
   0.000000000000000D+00,   6.207233938148690D-06,   0.000000000000000D+00, &
   1.367804015007348D-05,   0.000000000000000D+00,  -5.021119707244868D-06, &
   0.000000000000000D+00,   1.365138580388074D-04,   0.000000000000000D+00, &
   4.297448651550749D-04,   0.000000000000000D+00,   1.060021405795111D-04, &
   0.000000000000000D+00,   1.729548654358046D-06,   0.000000000000000D+00, &
   5.222906815656704D-06,   0.000000000000000D+00,   2.735222303752310D-06, &
   0.000000000000000D+00,   1.065001712492768D-06,   0.000000000000000D+00 /)
lem(:,290) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   9.998316850629006D-06,   7.054142284610504D-05,  -1.228668191128645D-04, &
  -8.066314063569990D-05,   8.399927115327128D-05,  -1.219864869647827D-04, &
   1.271487839254407D-04,   7.685721527978945D-04,   8.389189396076943D-04, &
  -7.776022729332082D-04,  -3.126629625690773D-03,  -2.533894772584211D-03, &
   4.233455532993571D-03,   8.414318832066781D-03,  -4.591292178810030D-03, &
  -1.113210612470908D-02,   6.995791913346683D-03,   6.820421345634381D-03, &
  -8.085500129180944D-03,  -4.823059732867782D-06,   4.649379229753042D-03, &
  -2.639126206668560D-03,   9.998316850625555D-06,   7.054142284610859D-05, &
  -1.228668191128710D-04,  -8.066314063569689D-05,   8.399927115327271D-05, &
  -1.219864869647875D-04,   1.271487839254423D-04,   7.685721527979014D-04, &
   8.389189396076952D-04,  -7.776022729332106D-04,  -3.126629625690776D-03, &
  -2.533894772584212D-03,   4.233455532993575D-03,   8.414318832066791D-03, &
  -4.591292178810024D-03,  -1.113210612470909D-02,   6.995791913346682D-03, &
   6.820421345634379D-03,  -8.085500129180934D-03,  -4.823059732861458D-06, &
   4.649379229753045D-03,  -2.639126206668565D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.861140512784935D-05,   0.000000000000000D+00, &
  -7.054034932530094D-05,   0.000000000000000D+00,  -3.754789259802773D-05, &
   0.000000000000000D+00,   1.862252028449445D-05,   0.000000000000000D+00, &
   1.009501760942771D-05,   0.000000000000000D+00,   1.326956598654126D-05, &
   0.000000000000000D+00,   1.952183386650109D-05,   0.000000000000000D+00, &
   9.473261591523329D-06,   0.000000000000000D+00,  -1.648078088249850D-05, &
   0.000000000000000D+00,  -5.004441143457031D-05,   0.000000000000000D+00, &
   5.030884636460702D-07,  -4.270167324599732D-06,   0.000000000000000D+00, &
  -5.135907735203558D-07,   0.000000000000000D+00,   1.393955592662060D-07, &
   1.551118364136671D-05,   0.000000000000000D+00,  -5.668679145348312D-07, &
   0.000000000000000D+00,  -4.509401944155013D-07,   0.000000000000000D+00, &
   8.735921714181566D-07,   0.000000000000000D+00,   1.117143760986505D-06, &
  -3.172112313191211D-05,   0.000000000000000D+00,  -7.093031526431901D-07, &
   0.000000000000000D+00,   9.925132034820833D-07,   0.000000000000000D+00, &
  -4.191015896549489D-06,   0.000000000000000D+00,  -6.416487766246295D-06, &
   0.000000000000000D+00,  -2.876198707155626D-08,   0.000000000000000D+00, &
  -8.564397893859128D-07,   3.652069281774567D-05,   0.000000000000000D+00, &
   7.075280154905624D-06,   0.000000000000000D+00,  -4.377687974023206D-06, &
   0.000000000000000D+00,  -2.058546011764565D-07,   0.000000000000000D+00, &
   2.128841777068580D-05,   0.000000000000000D+00,   1.787194398963523D-05, &
   0.000000000000000D+00,   4.276392920537134D-06,   0.000000000000000D+00, &
  -6.051198765448907D-07,   0.000000000000000D+00,  -1.079822134194387D-06, &
  -1.713847620745555D-05,   0.000000000000000D+00,  -1.482209317175120D-05, &
   0.000000000000000D+00,   1.122130008654053D-05,   0.000000000000000D+00, &
   2.506393378461507D-06,   0.000000000000000D+00,   2.013143602082153D-07, &
   0.000000000000000D+00,  -3.510123880716526D-05,   0.000000000000000D+00, &
  -2.908011272539710D-05,   0.000000000000000D+00,  -2.368585808469806D-05, &
   0.000000000000000D+00,   3.777299959487109D-06,   0.000000000000000D+00, &
   3.631130943761542D-07,   0.000000000000000D+00,   1.360911026960773D-06, &
  -1.878818167575000D-05,   0.000000000000000D+00,   9.117611101902720D-06, &
   0.000000000000000D+00,  -1.668350523440921D-05,   0.000000000000000D+00, &
  -1.784216594962090D-05,   0.000000000000000D+00,   4.739344556224156D-07, &
   0.000000000000000D+00,   4.440094689684855D-08,   0.000000000000000D+00, &
   2.598429638269277D-05,   0.000000000000000D+00,   3.492285034800830D-05, &
   0.000000000000000D+00,   5.933613208453422D-05,   0.000000000000000D+00, &
  -5.173454620708117D-07,   0.000000000000000D+00,  -6.786673892517790D-06, &
   0.000000000000000D+00,   1.097038488508894D-08,   0.000000000000000D+00, &
  -1.433764192730908D-07,   4.852248110037242D-05,   0.000000000000000D+00, &
   9.875973415858750D-06,   0.000000000000000D+00,   9.320310309623323D-06, &
   0.000000000000000D+00,   7.661800304396016D-05,   0.000000000000000D+00, &
   7.235591562974228D-06,   0.000000000000000D+00,  -1.983369122816320D-06, &
   0.000000000000000D+00,  -6.287367462752404D-07,   0.000000000000000D+00, &
  -7.973165557962296D-06,   0.000000000000000D+00,  -3.474996131851343D-05, &
   0.000000000000000D+00,  -8.403874042497026D-05,   0.000000000000000D+00, &
  -9.142154656976810D-05,   0.000000000000000D+00,   4.370817558500269D-05, &
   0.000000000000000D+00,   6.812160455592401D-06,   0.000000000000000D+00, &
   2.702448487615787D-07,   0.000000000000000D+00,  -2.407888753087149D-07, &
  -5.199902315923790D-05,   0.000000000000000D+00,  -1.862270338538607D-05, &
   0.000000000000000D+00,   1.033380050436361D-05,   0.000000000000000D+00, &
  -6.724885379305015D-05,   0.000000000000000D+00,  -1.383779380916542D-04, &
   0.000000000000000D+00,   7.204234850545261D-08,   0.000000000000000D+00, &
   1.386033809608242D-06,   0.000000000000000D+00,   4.358499319357029D-07, &
   0.000000000000000D+00,   2.958725980955986D-06,   0.000000000000000D+00, &
   2.524152882823122D-05,   0.000000000000000D+00,   6.916785563500212D-05, &
   0.000000000000000D+00,   1.958210878282004D-04,   0.000000000000000D+00, &
   1.063216201810778D-04,   0.000000000000000D+00,  -6.275766476266933D-05, &
   0.000000000000000D+00,  -7.586674877137429D-06,   0.000000000000000D+00, &
  -1.013507793677085D-06,   0.000000000000000D+00,  -3.504841907599989D-07, &
   3.163928185161514D-05,   0.000000000000000D+00,   9.899168121865801D-06, &
   0.000000000000000D+00,  -2.233962607543736D-05,   0.000000000000000D+00, &
  -7.266479464635369D-05,   0.000000000000000D+00,   1.478552583046517D-04, &
   0.000000000000000D+00,   1.593825366789872D-04,   0.000000000000000D+00, &
   9.198069850168681D-06,   0.000000000000000D+00,   2.396072841497272D-06, &
   0.000000000000000D+00,   1.071469691550162D-06,   0.000000000000000D+00, &
  -5.109096851299239D-06,   0.000000000000000D+00,  -9.214826826855018D-06, &
   0.000000000000000D+00,  -3.029868559585604D-05,   0.000000000000000D+00, &
  -1.042415563718348D-04,   0.000000000000000D+00,  -3.291448407404200D-04, &
   0.000000000000000D+00,  -1.060060019582623D-04,   0.000000000000000D+00, &
   3.664484591215111D-05,   0.000000000000000D+00,   2.805237790183865D-06, &
   0.000000000000000D+00,  -5.643655401015740D-07,   0.000000000000000D+00, &
  -9.520746844993043D-07,  -9.979574044525985D-06,   0.000000000000000D+00, &
  -1.396475028787605D-06,   0.000000000000000D+00,   1.710691158267549D-05, &
   0.000000000000000D+00,   9.585734028572447D-05,   0.000000000000000D+00, &
   1.583548726191047D-04,   0.000000000000000D+00,  -1.795026896038119D-04, &
   0.000000000000000D+00,  -1.238979783083233D-04,   0.000000000000000D+00, &
  -1.696953734597303D-05,   0.000000000000000D+00,  -4.021706118338959D-06, &
   0.000000000000000D+00,  -8.364462665601973D-07,   0.000000000000000D+00, &
   3.480923972972222D-06,   0.000000000000000D+00,  -9.447705155900994D-07, &
   0.000000000000000D+00,   3.823879013882323D-06,   0.000000000000000D+00, &
  -5.233960483672621D-06,   0.000000000000000D+00,   1.188199876353736D-04, &
   0.000000000000000D+00,   3.752461992017545D-04,   0.000000000000000D+00, &
   9.693032613593990D-05,   0.000000000000000D+00,   3.761929946750341D-06, &
   0.000000000000000D+00,   5.888772466859235D-06,   0.000000000000000D+00, &
   3.015089013038335D-06,   0.000000000000000D+00,   2.447183432879596D-06 /)
lem(:,291) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.221780515842493D-04,  -1.735241144988332D-04, &
   2.653084119526107D-04,   5.027424501465387D-04,   2.794153366049593D-04, &
  -4.402776785653302D-04,  -1.186475375503929D-03,  -1.090327315258988D-03, &
   6.481305633928539D-04,   2.942242685024866D-03,   2.541262170799822D-03, &
  -3.188226776480588D-03,  -6.452607971329137D-03,   3.474185697505299D-03, &
   7.044795265987561D-03,  -5.511158456041849D-03,  -2.518044751714025D-03, &
   5.493398179029520D-03,  -2.421240523310541D-03,  -1.821432612483580D-03, &
   2.856156184319791D-03,   0.000000000000000D+00,   1.221780515842468D-04, &
   1.735241144988202D-04,  -2.653084119526055D-04,  -5.027424501465328D-04, &
  -2.794153366049608D-04,   4.402776785653287D-04,   1.186475375503933D-03, &
   1.090327315258987D-03,  -6.481305633928534D-04,  -2.942242685024869D-03, &
  -2.541262170799832D-03,   3.188226776480590D-03,   6.452607971329136D-03, &
  -3.474185697505300D-03,  -7.044795265987552D-03,   5.511158456041857D-03, &
   2.518044751714024D-03,  -5.493398179029506D-03,   2.421240523310541D-03, &
   1.821432612483573D-03,  -2.856156184319782D-03,   0.000000000000000D+00, &
   6.108902579212175D-05,   0.000000000000000D+00,   1.103666341601158D-04, &
   0.000000000000000D+00,   3.067037682854872D-05,   0.000000000000000D+00, &
   1.772236349708432D-05,   0.000000000000000D+00,   8.127568605456295D-06, &
   0.000000000000000D+00,  -2.156628139862103D-05,   0.000000000000000D+00, &
  -3.091432465664181D-05,   0.000000000000000D+00,  -3.342028590137117D-05, &
   0.000000000000000D+00,  -5.537807086355188D-07,   0.000000000000000D+00, &
   7.023752254839831D-05,   0.000000000000000D+00,   5.025840400531404D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -4.829511538597134D-06, &
   0.000000000000000D+00,  -2.634612237025542D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.827266103000359D-06,   0.000000000000000D+00, &
  -5.133735642948517D-07,   0.000000000000000D+00,   4.602803640183311D-06, &
   0.000000000000000D+00,   3.332330578890793D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.475991755032815D-06,   0.000000000000000D+00, &
  -1.522744833439025D-06,   0.000000000000000D+00,   3.296692078200505D-06, &
   0.000000000000000D+00,  -5.591929834278602D-06,   0.000000000000000D+00, &
  -8.837997515699185D-06,   0.000000000000000D+00,  -4.502467716077008D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.275081226669544D-06, &
   0.000000000000000D+00,   5.197845992934126D-06,   0.000000000000000D+00, &
  -3.957907812096804D-06,   0.000000000000000D+00,  -7.822488541629099D-07, &
   0.000000000000000D+00,   7.108556574079050D-06,   0.000000000000000D+00, &
   1.646225698333916D-05,   0.000000000000000D+00,   5.872183069518668D-06, &
   0.000000000000000D+00,  -2.048256948683540D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.791925726578216D-06,   0.000000000000000D+00, &
  -8.639093250299704D-06,   0.000000000000000D+00,   9.084673666498279D-06, &
   0.000000000000000D+00,   2.922165303959604D-06,   0.000000000000000D+00, &
  -5.164054023605020D-07,   0.000000000000000D+00,  -8.880857847336853D-06, &
   0.000000000000000D+00,  -2.116350932137317D-05,   0.000000000000000D+00, &
  -2.228094418100337D-05,   0.000000000000000D+00,   3.441140760426020D-06, &
   0.000000000000000D+00,   8.996343594848017D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.131887812385644D-06,   0.000000000000000D+00, &
   4.992709015375541D-06,   0.000000000000000D+00,  -1.401480983130202D-05, &
   0.000000000000000D+00,  -1.734087725101092D-05,   0.000000000000000D+00, &
   1.381952853415268D-06,   0.000000000000000D+00,   1.257238537609422D-07, &
   0.000000000000000D+00,   8.757039499723927D-06,   0.000000000000000D+00, &
   2.296366483139619D-05,   0.000000000000000D+00,   4.903998561187983D-05, &
   0.000000000000000D+00,   2.711479773285586D-06,   0.000000000000000D+00, &
  -8.523667781050933D-06,   0.000000000000000D+00,   7.016408814252885D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.203789572992409D-05, &
   0.000000000000000D+00,   5.331490575482450D-06,   0.000000000000000D+00, &
   8.055490605335467D-06,   0.000000000000000D+00,   6.599321603347865D-05, &
   0.000000000000000D+00,   1.000869396983364D-05,   0.000000000000000D+00, &
  -3.351243293532568D-06,   0.000000000000000D+00,  -7.189145229462911D-07, &
   0.000000000000000D+00,  -9.610029986986167D-06,   0.000000000000000D+00, &
  -2.537949001718391D-05,   0.000000000000000D+00,  -6.723916750081515D-05, &
   0.000000000000000D+00,  -7.994497350209886D-05,   0.000000000000000D+00, &
   3.560670187369493D-05,   0.000000000000000D+00,   9.072628159972067D-06, &
   0.000000000000000D+00,   2.468066400742059D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.412598264755616D-05,   0.000000000000000D+00, &
  -1.014408974279600D-05,   0.000000000000000D+00,   9.052416826313048D-06, &
   0.000000000000000D+00,  -5.682993993194825D-05,   0.000000000000000D+00, &
  -1.232054110064129D-04,   0.000000000000000D+00,  -4.114923387408449D-06, &
   0.000000000000000D+00,   1.452358528498901D-06,   0.000000000000000D+00, &
   7.396358831213971D-07,   0.000000000000000D+00,   1.172720963763831D-05, &
   0.000000000000000D+00,   2.379905159230350D-05,   0.000000000000000D+00, &
   5.689867329550888D-05,   0.000000000000000D+00,   1.660147854768364D-04, &
   0.000000000000000D+00,   9.659672784063251D-05,   0.000000000000000D+00, &
  -5.349697568058300D-05,   0.000000000000000D+00,  -8.812671923485613D-06, &
   0.000000000000000D+00,  -1.736694260449707D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.528271707888774D-06,   0.000000000000000D+00, &
   5.881014310823306D-06,   0.000000000000000D+00,  -1.935088495831335D-05, &
   0.000000000000000D+00,  -6.146231147463362D-05,   0.000000000000000D+00, &
   1.282363664764740D-04,   0.000000000000000D+00,   1.448827839739571D-04, &
   0.000000000000000D+00,   1.248713199774328D-05,   0.000000000000000D+00, &
   3.378893806387829D-06,   0.000000000000000D+00,   2.042096651823464D-06, &
   0.000000000000000D+00,  -9.405540262541705D-06,   0.000000000000000D+00, &
  -1.470686482843044D-05,   0.000000000000000D+00,  -2.788426014625294D-05, &
   0.000000000000000D+00,  -9.000663053962181D-05,   0.000000000000000D+00, &
  -2.871478004009014D-04,   0.000000000000000D+00,  -9.886617697098886D-05, &
   0.000000000000000D+00,   3.014715196130160D-05,   0.000000000000000D+00, &
   2.794650108933255D-06,   0.000000000000000D+00,  -7.470055113554638D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -3.113008686790378D-07, &
   0.000000000000000D+00,  -1.678517027198210D-06,   0.000000000000000D+00, &
   1.339676159432827D-05,   0.000000000000000D+00,   8.192495745103279D-05, &
   0.000000000000000D+00,   1.369768210097738D-04,   0.000000000000000D+00, &
  -1.583778562627284D-04,   0.000000000000000D+00,  -1.141033171326573D-04, &
   0.000000000000000D+00,  -1.832319074097136D-05,   0.000000000000000D+00, &
  -4.625457883308485D-06,   0.000000000000000D+00,  -2.007435046017054D-06, &
   0.000000000000000D+00,   3.614347132710731D-06,   0.000000000000000D+00, &
   5.626654975095936D-06,   0.000000000000000D+00,   6.322274356500995D-06, &
   0.000000000000000D+00,  -1.830397359992888D-06,   0.000000000000000D+00, &
   1.060021405795111D-04,   0.000000000000000D+00,   3.333598102077213D-04, &
   0.000000000000000D+00,   9.163257262967101D-05,   0.000000000000000D+00, &
   6.264336089769132D-06,   0.000000000000000D+00,   6.539606596468565D-06, &
   0.000000000000000D+00,   3.966852875967855D-06,   0.000000000000000D+00 /)
lem(:,292) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   1.354852584631371D-04,  -4.927759839892809D-05,  -1.115199971124875D-04, &
  -2.260868276474280D-04,  -3.371899396320261D-04,  -1.798589439918819D-04, &
   3.860340905785964D-04,   1.037521925498522D-03,   1.050527034138627D-03, &
  -2.520801733842427D-04,  -2.223385855801588D-03,  -2.442774638420883D-03, &
   1.591280280515430D-03,   4.571392694303493D-03,  -1.651346878426540D-03, &
  -3.959607601020093D-03,   3.294143397783479D-03,   3.654538472259741D-04, &
  -2.641671072324951D-03,   2.437930981184262D-03,  -3.945486209434046D-04, &
  -1.518205136304230D-03,   1.354852584631377D-04,  -4.927759839893150D-05, &
  -1.115199971124821D-04,  -2.260868276474246D-04,  -3.371899396320308D-04, &
  -1.798589439918825D-04,   3.860340905786014D-04,   1.037521925498522D-03, &
   1.050527034138630D-03,  -2.520801733842446D-04,  -2.223385855801595D-03, &
  -2.442774638420883D-03,   1.591280280515429D-03,   4.571392694303482D-03, &
  -1.651346878426541D-03,  -3.959607601020092D-03,   3.294143397783477D-03, &
   3.654538472259654D-04,  -2.641671072324942D-03,   2.437930981184259D-03, &
  -3.945486209434132D-04,  -1.518205136304232D-03,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.676676895409998D-04,   0.000000000000000D+00, &
   4.523640774280160D-05,   0.000000000000000D+00,  -1.014479161496219D-04, &
   0.000000000000000D+00,  -6.778704844548786D-05,   0.000000000000000D+00, &
  -1.468492364891919D-05,   0.000000000000000D+00,   3.988170151282608D-05, &
   0.000000000000000D+00,   7.321619138387159D-05,   0.000000000000000D+00, &
   5.314799212660586D-05,   0.000000000000000D+00,  -7.329176417530845D-05, &
   0.000000000000000D+00,  -1.295266002951160D-04,   0.000000000000000D+00, &
  -9.075891015413231D-06,  -3.337024109319626D-06,   0.000000000000000D+00, &
   1.009218044679802D-05,   0.000000000000000D+00,  -3.957449977790220D-06, &
   7.897803724690038D-06,   0.000000000000000D+00,   3.790236721662274D-06, &
   0.000000000000000D+00,  -7.611565212819532D-06,   0.000000000000000D+00, &
   7.459020539686772D-07,   0.000000000000000D+00,   6.059751250466208D-06, &
  -1.188465224420958D-05,   0.000000000000000D+00,  -8.022978270748111D-06, &
   0.000000000000000D+00,  -2.717976446222407D-07,   0.000000000000000D+00, &
  -3.594569363423295D-06,   0.000000000000000D+00,  -8.626109610814151D-07, &
   0.000000000000000D+00,  -7.302165532880538D-06,   0.000000000000000D+00, &
   4.656927441083872D-07,   6.701809076796206D-06,   0.000000000000000D+00, &
   9.825117296179248D-06,   0.000000000000000D+00,   4.167339492080376D-06, &
   0.000000000000000D+00,  -4.776052597330332D-06,   0.000000000000000D+00, &
   1.161991785079578D-05,   0.000000000000000D+00,   5.435617065507250D-06, &
   0.000000000000000D+00,   1.245937352806299D-05,   0.000000000000000D+00, &
   5.476989025035600D-06,   0.000000000000000D+00,  -3.974778176041934D-06, &
   7.500113686605931D-06,   0.000000000000000D+00,  -4.785258581947926D-06, &
   0.000000000000000D+00,  -8.495743304961798D-06,   0.000000000000000D+00, &
   8.712077792817492D-06,   0.000000000000000D+00,   3.557151239793045D-06, &
   0.000000000000000D+00,  -4.699259338320084D-06,   0.000000000000000D+00, &
  -9.532215347638338D-06,   0.000000000000000D+00,  -1.706254574857974D-05, &
   0.000000000000000D+00,  -1.979698309666543D-05,   0.000000000000000D+00, &
   3.553417873926694D-06,   0.000000000000000D+00,   1.147990403813613D-06, &
  -1.714447384238347D-05,   0.000000000000000D+00,  -5.278562409326530D-06, &
   0.000000000000000D+00,   5.095682958064287D-06,   0.000000000000000D+00, &
  -1.097487800702617D-05,   0.000000000000000D+00,  -1.793948318034836D-05, &
   0.000000000000000D+00,   1.863489238603884D-06,   0.000000000000000D+00, &
  -1.126976565928369D-05,   0.000000000000000D+00,   6.832823925825707D-06, &
   0.000000000000000D+00,   1.963669257507024D-05,   0.000000000000000D+00, &
   4.220800440264300D-05,   0.000000000000000D+00,   4.644124048983329D-06, &
   0.000000000000000D+00,  -1.037889507616426D-05,   0.000000000000000D+00, &
   1.415648243323581D-06,   1.645114485548197D-05,   0.000000000000000D+00, &
   1.309271861642871D-05,   0.000000000000000D+00,   5.431444128102040D-06, &
   0.000000000000000D+00,   5.879238946655946D-06,   0.000000000000000D+00, &
   5.802353269977634D-05,   0.000000000000000D+00,   1.401267507411623D-05, &
   0.000000000000000D+00,  -4.753286199609019D-06,   0.000000000000000D+00, &
   1.909701748781807D-05,   0.000000000000000D+00,  -1.715730144682914D-06, &
   0.000000000000000D+00,  -2.031179966916721D-05,   0.000000000000000D+00, &
  -5.735464865051421D-05,   0.000000000000000D+00,  -7.218069777993009D-05, &
   0.000000000000000D+00,   2.912154447631317D-05,   0.000000000000000D+00, &
   1.265593256213949D-05,   0.000000000000000D+00,   8.877272505917253D-07, &
  -1.218788777589759D-05,   0.000000000000000D+00,  -1.226748145295808D-05, &
   0.000000000000000D+00,  -1.008050986844549D-05,   0.000000000000000D+00, &
   6.612494492454363D-06,   0.000000000000000D+00,  -4.844936342285760D-05, &
   0.000000000000000D+00,  -1.129511153117156D-04,   0.000000000000000D+00, &
  -1.013204138995144D-05,   0.000000000000000D+00,   1.815345429344322D-06, &
   0.000000000000000D+00,  -1.301717640726414D-05,   0.000000000000000D+00, &
   3.101756074753165D-07,   0.000000000000000D+00,   1.692289532305607D-05, &
   0.000000000000000D+00,   4.831193714576229D-05,   0.000000000000000D+00, &
   1.452030815058162D-04,   0.000000000000000D+00,   9.109704055820318D-05, &
   0.000000000000000D+00,  -4.632892119077701D-05,   0.000000000000000D+00, &
  -1.160562555480516D-05,   0.000000000000000D+00,  -4.145343474422820D-06, &
   7.724463349518118D-06,   0.000000000000000D+00,   4.981161114562020D-06, &
   0.000000000000000D+00,   4.467507185461708D-06,   0.000000000000000D+00, &
  -1.507209541656229D-05,   0.000000000000000D+00,  -5.214136201633949D-05, &
   0.000000000000000D+00,   1.130119176191189D-04,   0.000000000000000D+00, &
   1.366703211299455D-04,   0.000000000000000D+00,   1.738149590031631D-05, &
   0.000000000000000D+00,   5.022774632823846D-06,   0.000000000000000D+00, &
   3.725870722959832D-06,   0.000000000000000D+00,  -1.784885756658673D-07, &
   0.000000000000000D+00,  -9.349673953878187D-06,   0.000000000000000D+00, &
  -2.324746261650476D-05,   0.000000000000000D+00,  -7.970094218768506D-05, &
   0.000000000000000D+00,  -2.568033371733419D-04,   0.000000000000000D+00, &
  -9.618314482840672D-05,   0.000000000000000D+00,   2.529296676654063D-05, &
   0.000000000000000D+00,   3.000756266817583D-06,   0.000000000000000D+00, &
  -7.684074108400887D-07,  -3.136642664976336D-06,   0.000000000000000D+00, &
   2.448329679690422D-07,   0.000000000000000D+00,   1.460962080831040D-06, &
   0.000000000000000D+00,   1.202368666576861D-05,   0.000000000000000D+00, &
   7.058262833795528D-05,   0.000000000000000D+00,   1.202014661556055D-04, &
   0.000000000000000D+00,  -1.424318287906998D-04,   0.000000000000000D+00, &
  -1.093201367924644D-04,   0.000000000000000D+00,  -2.066195032149522D-05, &
   0.000000000000000D+00,  -5.860715590631283D-06,   0.000000000000000D+00, &
   7.493183085350146D-08,   0.000000000000000D+00,  -7.004923773276639D-07, &
   0.000000000000000D+00,   2.926688618455078D-06,   0.000000000000000D+00, &
   5.152837384580692D-06,   0.000000000000000D+00,  -4.544002858284752D-07, &
   0.000000000000000D+00,   9.693032613593990D-05,   0.000000000000000D+00, &
   3.034317683410059D-04,   0.000000000000000D+00,   8.979014604809294D-05, &
   0.000000000000000D+00,   8.847728858850675D-06,   0.000000000000000D+00, &
   7.983910107694168D-06,   0.000000000000000D+00,   9.079582407070105D-06 /)
lem(:,293) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.030753639578801D-05,  -1.345200570145134D-05, &
   2.648180341573946D-04,   4.575584678572749D-04,   3.549504776496688D-04, &
  -9.350499686725300D-05,  -7.394129847378032D-04,  -1.046475468820291D-03, &
  -3.902506416323501D-04,   1.180982896768484D-03,   2.130111881663168D-03, &
  -1.326685014631848D-04,  -3.018474897818764D-03,   2.062074994716629D-04, &
   2.230304381715701D-03,  -1.468279145086675D-03,   1.900087974063368D-04, &
   8.436757540301659D-04,  -1.398056521599469D-03,   1.027244136437427D-03, &
   9.822529127108641D-05,   0.000000000000000D+00,   6.030753639577651D-05, &
   1.345200570145727D-05,  -2.648180341573883D-04,  -4.575584678572794D-04, &
  -3.549504776496712D-04,   9.350499686725777D-05,   7.394129847378009D-04, &
   1.046475468820290D-03,   3.902506416323489D-04,  -1.180982896768484D-03, &
  -2.130111881663167D-03,   1.326685014631819D-04,   3.018474897818763D-03, &
  -2.062074994716599D-04,  -2.230304381715704D-03,   1.468279145086674D-03, &
  -1.900087974063395D-04,  -8.436757540301640D-04,   1.398056521599468D-03, &
  -1.027244136437428D-03,  -9.822529127108276D-05,   0.000000000000000D+00, &
   3.015376819788663D-05,   0.000000000000000D+00,  -1.252789650757792D-04, &
   0.000000000000000D+00,  -2.292640591419017D-04,   0.000000000000000D+00, &
   7.468465195590954D-05,   0.000000000000000D+00,   2.001347751720831D-04, &
   0.000000000000000D+00,   5.557238740596404D-05,   0.000000000000000D+00, &
  -1.005391762554571D-04,   0.000000000000000D+00,  -1.829247302476908D-04, &
   0.000000000000000D+00,   3.694583461297502D-07,   0.000000000000000D+00, &
   2.507695887441527D-04,   0.000000000000000D+00,   3.101095751452163D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.383864688552264D-06, &
   0.000000000000000D+00,   1.027892754729740D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   3.069317448398303D-06,   0.000000000000000D+00, &
   7.364105820281045D-06,   0.000000000000000D+00,  -4.290252386290322D-07, &
   0.000000000000000D+00,  -8.591396041273769D-08,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.382592243942504D-06,   0.000000000000000D+00, &
  -6.302111622812775D-06,   0.000000000000000D+00,  -1.221544117626674D-06, &
   0.000000000000000D+00,  -1.428889074893432D-06,   0.000000000000000D+00, &
  -1.425960470867529D-06,   0.000000000000000D+00,  -6.170391584454801D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.559715892822081D-06, &
   0.000000000000000D+00,   6.172096971217271D-06,   0.000000000000000D+00, &
   4.660356190529319D-06,   0.000000000000000D+00,  -1.016510603502517D-05, &
   0.000000000000000D+00,   2.691719428455621D-06,   0.000000000000000D+00, &
   4.219805007630948D-06,   0.000000000000000D+00,   1.066276328840265D-05, &
   0.000000000000000D+00,   7.780649070644781D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.576853778886021D-06,   0.000000000000000D+00, &
  -2.733278390431324D-06,   0.000000000000000D+00,  -7.178533905221292D-06, &
   0.000000000000000D+00,   7.698739881631031D-06,   0.000000000000000D+00, &
   8.006472115757420D-06,   0.000000000000000D+00,  -1.576499000476968D-06, &
   0.000000000000000D+00,  -6.047842023183511D-06,   0.000000000000000D+00, &
  -1.397871707361348D-05,   0.000000000000000D+00,  -1.994339794718193D-05, &
   0.000000000000000D+00,   2.958272690541997D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.500308540576926D-06,   0.000000000000000D+00, &
  -3.055354216780081D-06,   0.000000000000000D+00,   3.748073937748421D-06, &
   0.000000000000000D+00,  -8.456431114092630D-06,   0.000000000000000D+00, &
  -2.053736329845053D-05,   0.000000000000000D+00,   3.706088495917181D-06, &
   0.000000000000000D+00,  -1.370623384514412D-06,   0.000000000000000D+00, &
   5.021635266031639D-06,   0.000000000000000D+00,   1.568974396179137D-05, &
   0.000000000000000D+00,   3.760643035360820D-05,   0.000000000000000D+00, &
   8.486109857450336D-06,   0.000000000000000D+00,  -1.506058590288591D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.862325049889300D-06, &
   0.000000000000000D+00,   8.608745967199175D-06,   0.000000000000000D+00, &
   4.294278989785562D-06,   0.000000000000000D+00,   4.596504221212215D-06, &
   0.000000000000000D+00,   5.334722153403152D-05,   0.000000000000000D+00, &
   2.072914322002014D-05,   0.000000000000000D+00,  -1.002937457112427D-05, &
   0.000000000000000D+00,   2.680849504046547D-06,   0.000000000000000D+00, &
  -3.259847779207474D-06,   0.000000000000000D+00,  -1.687837254385695D-05, &
   0.000000000000000D+00,  -4.931806657883116D-05,   0.000000000000000D+00, &
  -6.877454177865064D-05,   0.000000000000000D+00,   2.372468647778980D-05, &
   0.000000000000000D+00,   2.008650906732161D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -7.758223853120284D-06,   0.000000000000000D+00, &
  -1.027764227332076D-05,   0.000000000000000D+00,  -8.211924138386904D-06, &
   0.000000000000000D+00,   5.371379837847345D-06,   0.000000000000000D+00, &
  -4.219416124876755D-05,   0.000000000000000D+00,  -1.093347998697084D-04, &
   0.000000000000000D+00,  -1.825330563975812D-05,   0.000000000000000D+00, &
   4.204670455561099D-06,   0.000000000000000D+00,  -5.480136615364320D-07, &
   0.000000000000000D+00,   2.711298489717076D-06,   0.000000000000000D+00, &
   1.513880832726342D-05,   0.000000000000000D+00,   4.230516302690644D-05, &
   0.000000000000000D+00,   1.302865599857362D-04,   0.000000000000000D+00, &
   9.047594370235093D-05,   0.000000000000000D+00,  -4.166321304667463D-05, &
   0.000000000000000D+00,  -1.713089583157887D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   5.404346333185709D-06,   0.000000000000000D+00, &
   6.999528263572369D-06,   0.000000000000000D+00,   4.793428336254489D-06, &
   0.000000000000000D+00,  -1.257265938204223D-05,   0.000000000000000D+00, &
  -4.577704056266730D-05,   0.000000000000000D+00,   1.027726172881070D-04, &
   0.000000000000000D+00,   1.354737406918140D-04,   0.000000000000000D+00, &
   2.452119699882019D-05,   0.000000000000000D+00,   1.057716294964572D-05, &
   0.000000000000000D+00,  -1.557143313746564D-06,   0.000000000000000D+00, &
  -1.840728943846536D-06,   0.000000000000000D+00,  -8.587593266230653D-06, &
   0.000000000000000D+00,  -2.150463205384777D-05,   0.000000000000000D+00, &
  -7.266791186583122D-05,   0.000000000000000D+00,  -2.364543333594610D-04, &
   0.000000000000000D+00,  -9.759051595904120D-05,   0.000000000000000D+00, &
   2.135731251530711D-05,   0.000000000000000D+00,   3.030790039610256D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.692239763636800D-06, &
   0.000000000000000D+00,  -2.950249818166323D-06,   0.000000000000000D+00, &
  -6.434385883156395D-07,   0.000000000000000D+00,   9.932281804911033D-06, &
   0.000000000000000D+00,   6.253373703375017D-05,   0.000000000000000D+00, &
   1.081526806774998D-04,   0.000000000000000D+00,  -1.323300687053442D-04, &
   0.000000000000000D+00,  -1.094534084148293D-04,   0.000000000000000D+00, &
  -2.494564367461223D-05,   0.000000000000000D+00,  -1.487310131909106D-05, &
   0.000000000000000D+00,   1.181731084509279D-06,   0.000000000000000D+00, &
   5.450659426261992D-07,   0.000000000000000D+00,   1.943982354973991D-06, &
   0.000000000000000D+00,   4.951492393424260D-06,   0.000000000000000D+00, &
   1.729548654358046D-06,   0.000000000000000D+00,   9.163257262967101D-05, &
   0.000000000000000D+00,   2.839497366483619D-04,   0.000000000000000D+00, &
   9.165022246654915D-05,   0.000000000000000D+00,   1.248195312836092D-05, &
   0.000000000000000D+00,   1.151241284388782D-05,   0.000000000000000D+00 /)
lem(:,294) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   7.852787310642701D-06,   8.810179222805745D-05,  -1.045717209553425D-04, &
  -2.807809293579484D-04,  -3.076600152374521D-04,  -3.544671129332269D-04, &
  -1.393009605509444D-04,   3.736531528128047D-04,   8.605584230842621D-04, &
   8.037137048836808D-04,  -2.485543039035215D-04,  -1.632705017150638D-03, &
  -8.584969686114031D-04,   1.758821301943478D-03,   5.599110439178402D-04, &
  -1.289230323274360D-03,   4.849307660302913D-04,  -7.160424447380326D-05, &
  -1.076264580997800D-04,   5.000269956941321D-04,  -7.209501313162549D-04, &
   4.541317233588111D-04,   7.852787310649266D-06,   8.810179222805146D-05, &
  -1.045717209553378D-04,  -2.807809293579378D-04,  -3.076600152374597D-04, &
  -3.544671129332370D-04,  -1.393009605509377D-04,   3.736531528128086D-04, &
   8.605584230842587D-04,   8.037137048836799D-04,  -2.485543039035188D-04, &
  -1.632705017150634D-03,  -8.584969686114029D-04,   1.758821301943480D-03, &
   5.599110439178402D-04,  -1.289230323274361D-03,   4.849307660302905D-04, &
  -7.160424447380839D-05,  -1.076264580997729D-04,   5.000269956941354D-04, &
  -7.209501313162622D-04,   4.541317233588057D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.423114684343750D-04,   0.000000000000000D+00, &
  -1.711813815096109D-04,   0.000000000000000D+00,   3.589524138331944D-04, &
   0.000000000000000D+00,   2.631367871959921D-04,   0.000000000000000D+00, &
  -4.289849456110488D-04,   0.000000000000000D+00,  -2.955026225118109D-04, &
   0.000000000000000D+00,   3.741509227072016D-04,   0.000000000000000D+00, &
   3.530983091672206D-04,   0.000000000000000D+00,  -3.598103423642412D-04, &
   0.000000000000000D+00,  -2.948289746267931D-04,   0.000000000000000D+00, &
   9.121285976756568D-07,  -6.781684027778140D-07,   0.000000000000000D+00, &
   3.972355046572592D-07,   0.000000000000000D+00,   6.908600658600529D-06, &
   3.965815530072582D-06,   0.000000000000000D+00,   8.330933553020353D-07, &
   0.000000000000000D+00,   1.956160101549133D-06,   0.000000000000000D+00, &
  -2.800691740477826D-06,   0.000000000000000D+00,  -8.369671088131758D-07, &
  -3.169545135413804D-06,   0.000000000000000D+00,  -2.018805218740980D-06, &
   0.000000000000000D+00,  -4.718098349448366D-06,   0.000000000000000D+00, &
  -4.458654564487705D-06,   0.000000000000000D+00,  -9.668185752763974D-07, &
   0.000000000000000D+00,  -1.038278771383879D-08,   0.000000000000000D+00, &
  -1.750590336714500D-05,  -2.892847338709675D-06,   0.000000000000000D+00, &
   1.265520868577612D-06,   0.000000000000000D+00,   5.439194917684107D-06, &
   0.000000000000000D+00,   6.664549104456058D-06,   0.000000000000000D+00, &
   7.067717652252458D-07,   0.000000000000000D+00,   3.441168657089653D-06, &
   0.000000000000000D+00,   2.622318048324802D-06,   0.000000000000000D+00, &
   1.121198044591221D-05,   0.000000000000000D+00,   1.802266743838744D-05, &
   5.453505514422518D-06,   0.000000000000000D+00,   2.224032166834955D-06, &
   0.000000000000000D+00,  -2.315369927837792D-06,   0.000000000000000D+00, &
  -8.334489943865835D-06,   0.000000000000000D+00,   6.425467514154587D-06, &
   0.000000000000000D+00,   6.763103289680305D-06,   0.000000000000000D+00, &
  -1.243919532234205D-06,   0.000000000000000D+00,  -5.085682206033279D-06, &
   0.000000000000000D+00,  -1.208174871206214D-05,   0.000000000000000D+00, &
  -2.223952061956892D-05,   0.000000000000000D+00,   8.750023798300623D-06, &
  -6.173233574717278D-07,   0.000000000000000D+00,  -4.635105502362163D-06, &
   0.000000000000000D+00,  -3.270226895393352D-06,   0.000000000000000D+00, &
   4.231725108692807D-06,   0.000000000000000D+00,  -6.717416062683983D-06, &
   0.000000000000000D+00,  -2.704078068997871D-05,   0.000000000000000D+00, &
  -1.078480984074275D-05,   0.000000000000000D+00,  -3.310367072637556D-06, &
   0.000000000000000D+00,   3.923597365381508D-06,   0.000000000000000D+00, &
   1.338917824819949D-05,   0.000000000000000D+00,   3.621592290411081D-05, &
   0.000000000000000D+00,   1.285755257664804D-05,   0.000000000000000D+00, &
  -3.478381353946024D-05,  -4.355610217782858D-06,   0.000000000000000D+00, &
   4.049733345692801D-06,   0.000000000000000D+00,   7.720402279393862D-06, &
   0.000000000000000D+00,   3.998879626170358D-06,   0.000000000000000D+00, &
   2.988343660728294D-06,   0.000000000000000D+00,   5.289710649932277D-05, &
   0.000000000000000D+00,   3.395741668124653D-05,   0.000000000000000D+00, &
   9.805610993544880D-06,   0.000000000000000D+00,   4.781265751600553D-06, &
   0.000000000000000D+00,  -1.538190058105250D-06,   0.000000000000000D+00, &
  -1.415758490152761D-05,   0.000000000000000D+00,  -4.497333341779854D-05, &
   0.000000000000000D+00,  -6.934025407019054D-05,   0.000000000000000D+00, &
   1.953540246714422D-05,   0.000000000000000D+00,   4.471261368777936D-05, &
   3.556843907866907D-06,   0.000000000000000D+00,  -2.411737992769076D-06, &
   0.000000000000000D+00,  -7.879444596208980D-06,   0.000000000000000D+00, &
  -7.813791066773232D-06,   0.000000000000000D+00,   4.518284127460294D-06, &
   0.000000000000000D+00,  -3.839383760922465D-05,   0.000000000000000D+00, &
  -1.127233584175408D-04,   0.000000000000000D+00,  -3.351278344217126D-05, &
   0.000000000000000D+00,  -6.025457196319709D-06,   0.000000000000000D+00, &
  -2.266855170916724D-06,   0.000000000000000D+00,   1.536958528306863D-06, &
   0.000000000000000D+00,   1.259241413157673D-05,   0.000000000000000D+00, &
   3.826785287664572D-05,   0.000000000000000D+00,   1.219871320559690D-04, &
   0.000000000000000D+00,   9.452541204268298D-05,   0.000000000000000D+00, &
  -3.906296812897033D-05,   0.000000000000000D+00,  -3.963077364307036D-05, &
  -4.736965703182943D-07,   0.000000000000000D+00,   7.308543429482645D-07, &
   0.000000000000000D+00,   4.249793346375971D-06,   0.000000000000000D+00, &
   4.155159385553522D-06,   0.000000000000000D+00,  -1.056258461863147D-05, &
   0.000000000000000D+00,  -4.082486833544025D-05,   0.000000000000000D+00, &
   9.708119130064546D-05,   0.000000000000000D+00,   1.427281414439409D-04, &
   0.000000000000000D+00,   3.918345427522751D-05,   0.000000000000000D+00, &
   2.139231960409271D-06,   0.000000000000000D+00,  -2.407223221744199D-08, &
   0.000000000000000D+00,  -2.158567480979843D-06,   0.000000000000000D+00, &
  -7.636788535685975D-06,   0.000000000000000D+00,  -1.978687024213040D-05, &
   0.000000000000000D+00,  -6.903443240191183D-05,   0.000000000000000D+00, &
  -2.256092064059145D-04,   0.000000000000000D+00,  -1.044738651946238D-04, &
   0.000000000000000D+00,   1.793498805395955D-05,   0.000000000000000D+00, &
   1.101581327267430D-05,  -5.239169510893736D-07,   0.000000000000000D+00, &
   4.445234491410603D-07,   0.000000000000000D+00,  -8.619244301903450D-07, &
   0.000000000000000D+00,  -8.810241510817076D-09,   0.000000000000000D+00, &
   9.114220604508244D-06,   0.000000000000000D+00,   5.687238131440269D-05, &
   0.000000000000000D+00,   9.947317630357246D-05,   0.000000000000000D+00, &
  -1.273113347596496D-04,   0.000000000000000D+00,  -1.165273463508776D-04, &
   0.000000000000000D+00,  -3.351415702446555D-05,   0.000000000000000D+00, &
  -2.357651373989403D-07,   0.000000000000000D+00,   1.608771464988933D-07, &
   0.000000000000000D+00,   1.547689129493714D-06,   0.000000000000000D+00, &
   2.649685411223737D-06,   0.000000000000000D+00,   4.988751144480321D-06, &
   0.000000000000000D+00,   3.761929946750341D-06,   0.000000000000000D+00, &
   8.979014604809294D-05,   0.000000000000000D+00,   2.753587482386020D-04, &
   0.000000000000000D+00,   9.876874960726808D-05,   0.000000000000000D+00, &
   1.747622106817088D-05,   0.000000000000000D+00,   2.594957756043337D-05 /)
lem(:,295) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.930361052575366D-05,   7.581368439973742D-05, &
  -9.253396589875990D-05,   1.347485810111982D-04,   4.175380839589966D-04, &
   3.167358341500645D-04,   6.928620118098011D-05,  -4.298356772943440D-04, &
  -8.483461733221018D-04,  -4.455460911760186D-04,   9.220758440335885D-04, &
   1.235520443612283D-03,  -8.358804167513448D-04,  -8.398834494432101D-04, &
   7.167996599123155D-04,  -5.987493322247880D-05,  -6.405150447347882D-05, &
  -1.976576926866615D-05,  -6.585507048529777D-05,   2.856437091995172D-04, &
  -3.757089529857508D-04,   0.000000000000000D+00,  -7.930361052576030D-05, &
  -7.581368439972723D-05,   9.253396589876593D-05,  -1.347485810112114D-04, &
  -4.175380839590031D-04,  -3.167358341500561D-04,  -6.928620118097794D-05, &
   4.298356772943412D-04,   8.483461733221012D-04,   4.455460911760178D-04, &
  -9.220758440335884D-04,  -1.235520443612285D-03,   8.358804167513446D-04, &
   8.398834494432097D-04,  -7.167996599123144D-04,   5.987493322247768D-05, &
   6.405150447347675D-05,   1.976576926866889D-05,   6.585507048529742D-05, &
  -2.856437091995214D-04,   3.757089529857533D-04,   0.000000000000000D+00, &
  -3.965180526288310D-05,   0.000000000000000D+00,  -8.245361866777655D-05, &
   0.000000000000000D+00,   3.558002834313628D-04,   0.000000000000000D+00, &
   8.986614003352484D-05,   0.000000000000000D+00,  -9.213442833341482D-04, &
   0.000000000000000D+00,   2.168886015444206D-04,   0.000000000000000D+00, &
   1.437706416749725D-03,   0.000000000000000D+00,  -1.560886015184944D-03, &
   0.000000000000000D+00,  -1.933132163947558D-05,   0.000000000000000D+00, &
   9.512961046519366D-04,   0.000000000000000D+00,  -2.451509370343702D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   3.134750449203660D-06, &
   0.000000000000000D+00,   1.049433422480228D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.409858587592951D-06,   0.000000000000000D+00, &
   2.469680204055095D-06,   0.000000000000000D+00,  -4.810185474243258D-08, &
   0.000000000000000D+00,  -3.563602196413962D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.210500604308408D-06,   0.000000000000000D+00, &
  -1.881038594202418D-06,   0.000000000000000D+00,  -1.732099898508381D-05, &
   0.000000000000000D+00,  -1.150063642902420D-06,   0.000000000000000D+00, &
  -5.730970160811700D-07,   0.000000000000000D+00,   1.179800188833643D-07, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.443473205087069D-06, &
   0.000000000000000D+00,   7.562379746401870D-07,   0.000000000000000D+00, &
   7.585758815987931D-06,   0.000000000000000D+00,   1.664158145063624D-05, &
   0.000000000000000D+00,   6.166788865473878D-07,   0.000000000000000D+00, &
   1.381271965672737D-06,   0.000000000000000D+00,   1.883472012827763D-06, &
   0.000000000000000D+00,   7.817729678676351D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.497863339640506D-06,   0.000000000000000D+00, &
   1.651865704418265D-06,   0.000000000000000D+00,  -2.770402371971401D-06, &
   0.000000000000000D+00,  -8.919066991481029D-06,   0.000000000000000D+00, &
   2.304273560588360D-05,   0.000000000000000D+00,   1.312933681587883D-06, &
   0.000000000000000D+00,  -5.582148735288890D-07,   0.000000000000000D+00, &
  -3.645595914459904D-06,   0.000000000000000D+00,  -1.115288743586457D-05, &
   0.000000000000000D+00,  -2.583565225278134D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -9.700610184167437D-07,   0.000000000000000D+00, &
  -3.385528643403104D-06,   0.000000000000000D+00,  -2.176624548447287D-06, &
   0.000000000000000D+00,   3.917343298557625D-06,   0.000000000000000D+00, &
  -8.097793172604237D-06,   0.000000000000000D+00,  -6.855352139018906D-05, &
   0.000000000000000D+00,  -3.268760429481094D-06,   0.000000000000000D+00, &
  -1.797994312435369D-06,   0.000000000000000D+00,   2.798465931190664D-06, &
   0.000000000000000D+00,   1.247377053742862D-05,   0.000000000000000D+00, &
   3.762022095529544D-05,   0.000000000000000D+00,   2.507480111740242D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,   6.835687851405842D-07, &
   0.000000000000000D+00,   3.828890410589409D-06,   0.000000000000000D+00, &
   6.367398863172071D-06,   0.000000000000000D+00,   3.333837602713509D-06, &
   0.000000000000000D+00,   2.456445217266435D-06,   0.000000000000000D+00, &
   5.784862683384986D-05,   0.000000000000000D+00,   7.587041797722258D-05, &
   0.000000000000000D+00,   3.960673922590106D-06,   0.000000000000000D+00, &
   3.376776893047392D-06,   0.000000000000000D+00,  -1.238273132275045D-06, &
   0.000000000000000D+00,  -1.292071576528403D-05,   0.000000000000000D+00, &
  -4.291374440092752D-05,   0.000000000000000D+00,  -7.534943546337624D-05, &
   0.000000000000000D+00,   1.052767734417702D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.380259460411247D-06,   0.000000000000000D+00, &
  -3.482085911835005D-06,   0.000000000000000D+00,  -7.372465200679548D-06, &
   0.000000000000000D+00,  -6.953469321860976D-06,   0.000000000000000D+00, &
   3.442933062564022D-06,   0.000000000000000D+00,  -3.662205816775151D-05, &
   0.000000000000000D+00,  -1.260917854185485D-04,   0.000000000000000D+00, &
  -6.995080334335014D-05,   0.000000000000000D+00,  -2.580364749145952D-06, &
   0.000000000000000D+00,  -2.595754469575207D-06,   0.000000000000000D+00, &
   1.123934188992853D-06,   0.000000000000000D+00,   1.150001908651575D-05, &
   0.000000000000000D+00,   3.660945751811957D-05,   0.000000000000000D+00, &
   1.192636133848721D-04,   0.000000000000000D+00,   1.059155181251776D-04, &
   0.000000000000000D+00,  -3.565249749461526D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.694061238661683D-06,   0.000000000000000D+00, &
   2.224634304669314D-06,   0.000000000000000D+00,   4.503708109075570D-06, &
   0.000000000000000D+00,   4.187087459316430D-06,   0.000000000000000D+00, &
  -8.959497870035050D-06,   0.000000000000000D+00,  -3.760093387442135D-05, &
   0.000000000000000D+00,   9.647325355584607D-05,   0.000000000000000D+00, &
   1.629948209896985D-04,   0.000000000000000D+00,   8.629045621995765D-05, &
   0.000000000000000D+00,   5.625571187804162D-07,   0.000000000000000D+00, &
   1.104722426213944D-06,   0.000000000000000D+00,  -1.117986156819211D-06, &
   0.000000000000000D+00,  -6.886799635396870D-06,   0.000000000000000D+00, &
  -1.982975446513090D-05,   0.000000000000000D+00,  -6.817518664528140D-05, &
   0.000000000000000D+00,  -2.256601368777985D-04,   0.000000000000000D+00, &
  -1.194401536636237D-04,   0.000000000000000D+00,   8.772326448377643D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.581109408038838D-07, &
   0.000000000000000D+00,  -8.196108825634154D-07,   0.000000000000000D+00, &
  -1.045143180265104D-06,   0.000000000000000D+00,  -3.531348933717112D-07, &
   0.000000000000000D+00,   8.395527721708592D-06,   0.000000000000000D+00, &
   5.312710832106343D-05,   0.000000000000000D+00,   9.538662807553911D-05, &
   0.000000000000000D+00,  -1.291577477879766D-04,   0.000000000000000D+00, &
  -1.340075086325075D-04,   0.000000000000000D+00,  -8.675731296072317D-05, &
   0.000000000000000D+00,   2.078851474572805D-07,   0.000000000000000D+00, &
  -4.474174254606454D-07,   0.000000000000000D+00,   3.466650516522377D-07, &
   0.000000000000000D+00,   2.068737302012766D-06,   0.000000000000000D+00, &
   5.222906815656704D-06,   0.000000000000000D+00,   6.264336089769132D-06, &
   0.000000000000000D+00,   9.165022246654915D-05,   0.000000000000000D+00, &
   2.800813109864269D-04,   0.000000000000000D+00,   1.134516233195660D-04, &
   0.000000000000000D+00,   3.258014902213019D-05,   0.000000000000000D+00 /)
lem(:,296) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.899490699316102D-05,   1.305180261735334D-04,   2.228045443746431D-04, &
   9.734526935502892D-06,  -2.375737804721874D-04,  -9.884643170226467D-05, &
  -2.572884958742408D-04,  -4.461484500269600D-04,   1.187146753410161D-05, &
   6.185723458326034D-04,   8.153579867166565D-04,  -1.849532523896441D-04, &
  -1.187907477758979D-03,   1.687023957874680D-04,   7.998577772720732D-04, &
  -3.808566022464827D-04,  -1.092526579413166D-04,   1.107617028746795D-04, &
  -4.059025176667268D-07,  -2.394934499070037D-05,  -4.833507602462124D-05, &
   1.568654129392146D-04,   2.899490699317341D-05,   1.305180261735287D-04, &
   2.228045443746465D-04,   9.734526935514074D-06,  -2.375737804721958D-04, &
  -9.884643170227637D-05,  -2.572884958742327D-04,  -4.461484500269577D-04, &
   1.187146753410262D-05,   6.185723458326018D-04,   8.153579867166554D-04, &
  -1.849532523896402D-04,  -1.187907477758981D-03,   1.687023957874688D-04, &
   7.998577772720770D-04,  -3.808566022464855D-04,  -1.092526579413204D-04, &
   1.107617028746805D-04,  -4.059025176614194D-07,  -2.394934499069879D-05, &
  -4.833507602462703D-05,   1.568654129392125D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.720445288662916D-05,   0.000000000000000D+00, &
   1.250409293974893D-04,   0.000000000000000D+00,   3.162490947920653D-05, &
   0.000000000000000D+00,  -8.323116106009148D-04,   0.000000000000000D+00, &
   7.229040795111082D-04,   0.000000000000000D+00,   1.709375366491624D-03, &
   0.000000000000000D+00,  -4.264130294994804D-03,   0.000000000000000D+00, &
   4.398621608632327D-03,   0.000000000000000D+00,  -2.032819754853376D-03, &
   0.000000000000000D+00,  -5.571935189834851D-04,   0.000000000000000D+00, &
   2.881030102710867D-07,  -2.580675798448881D-07,   0.000000000000000D+00, &
  -2.243511154644221D-06,   0.000000000000000D+00,   2.446800771385682D-06, &
  -1.375645296543531D-06,   0.000000000000000D+00,   1.652414448434933D-06, &
   0.000000000000000D+00,   3.498563068875601D-06,   0.000000000000000D+00, &
  -1.990129553519770D-09,   0.000000000000000D+00,  -7.202688171926449D-06, &
   3.444363795926181D-06,   0.000000000000000D+00,  -9.910905667219099D-07, &
   0.000000000000000D+00,  -1.720601085835431D-06,   0.000000000000000D+00, &
  -1.304508750797863D-06,   0.000000000000000D+00,  -6.550490254104087D-07, &
   0.000000000000000D+00,   1.502572513162066D-06,   0.000000000000000D+00, &
   1.698922258805442D-06,  -2.549670610034272D-06,   0.000000000000000D+00, &
  -1.066603657054818D-06,   0.000000000000000D+00,   8.066980040168238D-07, &
   0.000000000000000D+00,   1.685543023738079D-06,   0.000000000000000D+00, &
  -2.191617874973684D-06,   0.000000000000000D+00,  -5.988110882373960D-07, &
   0.000000000000000D+00,   7.492921958744871D-07,   0.000000000000000D+00, &
   2.434950622542176D-06,   0.000000000000000D+00,   3.480851342643084D-05, &
  -1.518832238846140D-06,   0.000000000000000D+00,   1.882637255182735D-06, &
   0.000000000000000D+00,   1.334206623338427D-06,   0.000000000000000D+00, &
  -7.071086856528381D-07,   0.000000000000000D+00,  -7.360654648076421D-06, &
   0.000000000000000D+00,   2.523823554261967D-06,   0.000000000000000D+00, &
   2.288473705492514D-06,   0.000000000000000D+00,   3.733424272237198D-08, &
   0.000000000000000D+00,  -4.399283439126640D-06,   0.000000000000000D+00, &
  -1.714145405379967D-05,   0.000000000000000D+00,  -7.762755356542751D-05, &
   5.088959815145169D-06,   0.000000000000000D+00,  -2.582170067781213D-07, &
   0.000000000000000D+00,  -2.607053412970906D-06,   0.000000000000000D+00, &
  -2.319143429171435D-06,   0.000000000000000D+00,   4.347000539569268D-06, &
   0.000000000000000D+00,   1.413504089516081D-06,   0.000000000000000D+00, &
  -6.871069817905259D-07,   0.000000000000000D+00,  -2.826989183095552D-06, &
   0.000000000000000D+00,  -2.471481808718031D-06,   0.000000000000000D+00, &
   2.944163428456075D-06,   0.000000000000000D+00,   1.317149383568732D-05, &
   0.000000000000000D+00,   4.576248102849166D-05,   0.000000000000000D+00, &
   5.105429334853264D-05,  -5.633897612306407D-06,   0.000000000000000D+00, &
  -1.200737762375810D-06,   0.000000000000000D+00,   2.656345011872785D-06, &
   0.000000000000000D+00,   5.490655058269340D-06,   0.000000000000000D+00, &
   3.422867280519547D-06,   0.000000000000000D+00,  -3.434114636003579D-07, &
   0.000000000000000D+00,   5.648448546612812D-05,   0.000000000000000D+00, &
   2.433601621259907D-07,   0.000000000000000D+00,   2.146354103447444D-06, &
   0.000000000000000D+00,   3.283649657050082D-06,   0.000000000000000D+00, &
  -9.089899000337996D-07,   0.000000000000000D+00,  -1.234215335577404D-05, &
   0.000000000000000D+00,  -4.331161516657076D-05,   0.000000000000000D+00, &
  -8.645864300312188D-05,   0.000000000000000D+00,   5.307583432855719D-05, &
   3.600929893161572D-06,   0.000000000000000D+00,   7.017671421470011D-07, &
   0.000000000000000D+00,  -2.410035888209032D-06,   0.000000000000000D+00, &
  -6.152342882476614D-06,   0.000000000000000D+00,  -7.294374536404254D-06, &
   0.000000000000000D+00,   2.585568026060686D-06,   0.000000000000000D+00, &
  -3.705446292628277D-05,   0.000000000000000D+00,  -1.571723857683240D-04, &
   0.000000000000000D+00,  -7.907392134319581D-07,   0.000000000000000D+00, &
  -8.109511916611176D-07,   0.000000000000000D+00,  -1.682510520076938D-06, &
   0.000000000000000D+00,   8.763920570955526D-07,   0.000000000000000D+00, &
   1.115571044327457D-05,   0.000000000000000D+00,   3.638809044638110D-05, &
   0.000000000000000D+00,   1.242155540887591D-04,   0.000000000000000D+00, &
   1.255279601615982D-04,   0.000000000000000D+00,  -1.197477290547169D-04, &
  -1.166524577192382D-06,   0.000000000000000D+00,   9.361418373577898D-08, &
   0.000000000000000D+00,   1.802603661765543D-06,   0.000000000000000D+00, &
   3.858207924669375D-06,   0.000000000000000D+00,   4.287589862553717D-06, &
   0.000000000000000D+00,  -7.496641483356039D-06,   0.000000000000000D+00, &
  -3.643589364269878D-05,   0.000000000000000D+00,   1.042463611771478D-04, &
   0.000000000000000D+00,   2.162162735610206D-04,   0.000000000000000D+00, &
   6.067332943594085D-07,   0.000000000000000D+00,  -1.822626621029663D-07, &
   0.000000000000000D+00,   1.361948039917754D-08,   0.000000000000000D+00, &
  -1.289640348134355D-06,   0.000000000000000D+00,  -7.151824296983374D-06, &
   0.000000000000000D+00,  -2.032824517978607D-05,   0.000000000000000D+00, &
  -7.100672252785514D-05,   0.000000000000000D+00,  -2.407929943200917D-04, &
   0.000000000000000D+00,  -1.456610156143652D-04,   0.000000000000000D+00, &
   5.737805549767128D-05,   4.495323687405276D-08,   0.000000000000000D+00, &
  -2.198368097513856D-08,   0.000000000000000D+00,  -8.770744296494286D-07, &
   0.000000000000000D+00,  -1.224049311366901D-06,   0.000000000000000D+00, &
  -2.639227455790342D-07,   0.000000000000000D+00,   7.819287321965664D-06, &
   0.000000000000000D+00,   5.224367642348283D-05,   0.000000000000000D+00, &
   9.637853339132117D-05,   0.000000000000000D+00,  -1.421194053563970D-04, &
   0.000000000000000D+00,  -1.704637158145791D-04,   0.000000000000000D+00, &
  -1.253500931276657D-07,   0.000000000000000D+00,   2.499733651470588D-07, &
   0.000000000000000D+00,   3.796817145400408D-07,   0.000000000000000D+00, &
   6.827833132901565D-07,   0.000000000000000D+00,   2.456100334922962D-06, &
   0.000000000000000D+00,   5.888772466859235D-06,   0.000000000000000D+00, &
   8.847728858850675D-06,   0.000000000000000D+00,   9.876874960726808D-05, &
   0.000000000000000D+00,   3.034808280381746D-04,   0.000000000000000D+00, &
   1.366404058338821D-04,   0.000000000000000D+00,   6.372577149190007D-05 /)
lem(:,297) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,   2.369510299370034D-06,  -2.596252708891148D-05, &
  -1.843092829103558D-04,  -4.131252691253750D-04,  -1.155154478450638D-04, &
   4.844767119868537D-04,   4.037868267733688D-04,   2.158650070463725D-04, &
  -8.451151179482500D-05,  -8.216436034492833D-04,  -3.809525578196853D-04, &
   9.225501192601438D-04,   2.510539769364918D-04,  -6.525717565017621D-04, &
   1.565406779670816D-04,   1.480669140435896D-04,  -1.135999201336056D-04, &
   1.891535984050937D-05,   1.436820127552847D-05,  -3.490679392752584D-06, &
  -3.538019237558386D-05,   0.000000000000000D+00,  -2.369510299372307D-06, &
   2.596252708891188D-05,   1.843092829103625D-04,   4.131252691253661D-04, &
   1.155154478450554D-04,  -4.844767119868429D-04,  -4.037868267733660D-04, &
  -2.158650070463749D-04,   8.451151179482393D-05,   8.216436034492840D-04, &
   3.809525578196847D-04,  -9.225501192601451D-04,  -2.510539769364913D-04, &
   6.525717565017661D-04,  -1.565406779670827D-04,  -1.480669140435918D-04, &
   1.135999201336066D-04,  -1.891535984050726D-05,  -1.436820127552741D-05, &
   3.490679392747690D-06,   3.538019237558161D-05,   0.000000000000000D+00, &
  -1.184755149683300D-06,   0.000000000000000D+00,  -8.637063982230535D-06, &
   0.000000000000000D+00,   6.206573501590571D-05,   0.000000000000000D+00, &
  -3.664331426723476D-04,   0.000000000000000D+00,   4.793101724606559D-04, &
   0.000000000000000D+00,   1.400429975248260D-03,   0.000000000000000D+00, &
  -4.610585399863202D-03,   0.000000000000000D+00,   6.174728334599008D-03, &
   0.000000000000000D+00,  -6.189253803310088D-03,   0.000000000000000D+00, &
   4.851648968525008D-03,   0.000000000000000D+00,  -1.741031287438332D-03, &
   0.000000000000000D+00,   0.000000000000000D+00,   9.366311856554352D-08, &
   0.000000000000000D+00,  -4.452083135183610D-09,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.498514676667250D-07,   0.000000000000000D+00, &
   8.377437249680638D-07,   0.000000000000000D+00,  -5.788675919814441D-07, &
   0.000000000000000D+00,   9.633013412834228D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.773240369501953D-07,   0.000000000000000D+00, &
   5.990491590554325D-07,   0.000000000000000D+00,  -6.235950983875629D-06, &
   0.000000000000000D+00,   1.495343423007971D-06,   0.000000000000000D+00, &
  -8.232179972030957D-07,   0.000000000000000D+00,  -3.456920364094678D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.016575113254742D-06, &
   0.000000000000000D+00,  -1.459394061497187D-06,   0.000000000000000D+00, &
   1.734167034808732D-06,   0.000000000000000D+00,   2.342674229118308D-05, &
   0.000000000000000D+00,  -1.824373431536798D-06,   0.000000000000000D+00, &
  -2.021075658965018D-07,   0.000000000000000D+00,   3.358733748304811D-06, &
   0.000000000000000D+00,   4.050972657024009D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.384050857441865D-07,   0.000000000000000D+00, &
   1.133128106788217D-06,   0.000000000000000D+00,   5.198577646265371D-07, &
   0.000000000000000D+00,  -7.603024508401014D-06,   0.000000000000000D+00, &
  -3.741243449238283D-05,   0.000000000000000D+00,   1.362021322055115D-06, &
   0.000000000000000D+00,   1.686728061438089D-06,   0.000000000000000D+00, &
  -1.009148949458107D-06,   0.000000000000000D+00,  -3.974873565077538D-06, &
   0.000000000000000D+00,  -5.264011783803267D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   8.781263860408546D-07,   0.000000000000000D+00, &
   1.877536816355626D-08,   0.000000000000000D+00,  -1.673549960569969D-06, &
   0.000000000000000D+00,  -3.726844778658303D-07,   0.000000000000000D+00, &
   8.994252686387387D-06,   0.000000000000000D+00,  -2.072559251874392D-05, &
   0.000000000000000D+00,  -1.077871372354266D-06,   0.000000000000000D+00, &
  -2.768033651409225D-06,   0.000000000000000D+00,  -1.835746577303632D-06, &
   0.000000000000000D+00,   1.995914755419505D-06,   0.000000000000000D+00, &
   8.896561812510026D-06,   0.000000000000000D+00,   3.594910023527132D-05, &
   0.000000000000000D+00,   0.000000000000000D+00,  -1.671384990659081D-06, &
   0.000000000000000D+00,  -7.227221575688694D-07,   0.000000000000000D+00, &
   1.955595596492756D-06,   0.000000000000000D+00,   5.382195852507861D-06, &
   0.000000000000000D+00,   4.353819508711697D-06,   0.000000000000000D+00, &
   8.112272778580795D-06,   0.000000000000000D+00,   2.058760307370230D-04, &
   0.000000000000000D+00,   1.350235294510407D-06,   0.000000000000000D+00, &
   2.740278956348551D-06,   0.000000000000000D+00,   3.070456253104014D-06, &
   0.000000000000000D+00,  -6.273306131061563D-07,   0.000000000000000D+00, &
  -1.145482303683267D-05,   0.000000000000000D+00,  -4.561006860508925D-05, &
   0.000000000000000D+00,  -1.189989977198700D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.080182022843400D-06,   0.000000000000000D+00, &
   4.666188357638897D-07,   0.000000000000000D+00,  -1.866228547476931D-06, &
   0.000000000000000D+00,  -6.425846651164009D-06,   0.000000000000000D+00, &
  -8.438891727691801D-06,   0.000000000000000D+00,  -4.769477241893886D-07, &
   0.000000000000000D+00,  -5.742425357003462D-05,   0.000000000000000D+00, &
  -4.450887696793046D-04,   0.000000000000000D+00,  -1.262669451498467D-06, &
   0.000000000000000D+00,  -1.684472609723360D-06,   0.000000000000000D+00, &
  -2.073897812296907D-06,   0.000000000000000D+00,   1.002390225892461D-06, &
   0.000000000000000D+00,   1.116648893115864D-05,   0.000000000000000D+00, &
   3.929491369132923D-05,   0.000000000000000D+00,   1.455826036351378D-04, &
   0.000000000000000D+00,   2.189808504919438D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -5.866256519556398D-08,   0.000000000000000D+00, &
  -6.907870706564532D-08,   0.000000000000000D+00,   1.303698375512066D-06, &
   0.000000000000000D+00,   3.945207956230306D-06,   0.000000000000000D+00, &
   5.091518221692529D-06,   0.000000000000000D+00,  -6.604725368103485D-06, &
   0.000000000000000D+00,  -3.747617516986098D-05,   0.000000000000000D+00, &
   1.297789296424235D-04,   0.000000000000000D+00,   5.746151382288841D-04, &
   0.000000000000000D+00,   5.429327534901412D-07,   0.000000000000000D+00, &
   5.872827333129580D-07,   0.000000000000000D+00,   4.750585553027463D-07, &
   0.000000000000000D+00,  -1.435872109159456D-06,   0.000000000000000D+00, &
  -7.350033920337306D-06,   0.000000000000000D+00,  -2.222346737011970D-05, &
   0.000000000000000D+00,  -7.971524306434967D-05,   0.000000000000000D+00, &
  -2.848216065684504D-04,   0.000000000000000D+00,  -2.720643027637076D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,  -2.295373494900302D-07, &
   0.000000000000000D+00,   5.306243705042164D-08,   0.000000000000000D+00, &
  -4.160990103718939D-07,   0.000000000000000D+00,  -9.484373108211013D-07, &
   0.000000000000000D+00,  -7.513211999102635D-07,   0.000000000000000D+00, &
   7.961704215386985D-06,   0.000000000000000D+00,   5.480166095033302D-05, &
   0.000000000000000D+00,   1.057550794881416D-04,   0.000000000000000D+00, &
  -1.701512519925588D-04,   0.000000000000000D+00,  -4.922663293958269D-04, &
   0.000000000000000D+00,  -3.076061880710644D-08,   0.000000000000000D+00, &
  -1.625262307735531D-07,   0.000000000000000D+00,   1.713564140388741D-07, &
   0.000000000000000D+00,   7.983954766882115D-07,   0.000000000000000D+00, &
   2.735222303752310D-06,   0.000000000000000D+00,   6.539606596468565D-06, &
   0.000000000000000D+00,   1.248195312836092D-05,   0.000000000000000D+00, &
   1.134516233195660D-04,   0.000000000000000D+00,   3.609207379687146D-04, &
   0.000000000000000D+00,   2.419741271286598D-04,   0.000000000000000D+00 /)
lem(:,298) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   2.740924801699078D-05,   8.840297998234821D-05,   1.466084274777781D-04, &
   1.253064121545688D-04,   6.843505117698049D-06,   4.895437008469072D-06, &
   2.966949253626519D-04,   8.968065581526254D-06,  -8.117420077919048D-04, &
  -3.727729279681497D-04,   7.371644110534922D-04,   6.608542543442218D-04, &
  -5.477189700812512D-04,  -4.631228631073716D-04,   4.832550051986215D-04, &
  -1.463403464006981D-05,  -1.563912575892813D-04,   9.274662874304234D-05, &
  -2.003718464303283D-05,   2.585887514580404D-07,  -4.084251633307572D-07, &
   5.742430087183135D-06,   2.740924801699067D-05,   8.840297998234794D-05, &
   1.466084274777779D-04,   1.253064121545741D-04,   6.843505117693005D-06, &
   4.895437008462574D-06,   2.966949253626565D-04,   8.968065581531424D-06, &
  -8.117420077919028D-04,  -3.727729279681546D-04,   7.371644110534881D-04, &
   6.608542543442243D-04,  -5.477189700812501D-04,  -4.631228631073705D-04, &
   4.832550051986260D-04,  -1.463403464007418D-05,  -1.563912575892870D-04, &
   9.274662874304510D-05,  -2.003718464302683D-05,   2.585887514586145D-07, &
  -4.084251633376299D-07,   5.742430087180668D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -6.321503819169671D-06,   0.000000000000000D+00, &
   1.489791789562906D-05,   0.000000000000000D+00,  -2.748783931410407D-05, &
   0.000000000000000D+00,   3.665986392492393D-06,   0.000000000000000D+00, &
   8.243593792787864D-04,   0.000000000000000D+00,  -2.849161814878216D-03, &
   0.000000000000000D+00,   2.596060415746171D-03,   0.000000000000000D+00, &
   1.009675942996165D-03,   0.000000000000000D+00,  -1.325933067517771D-03, &
   0.000000000000000D+00,  -2.316412006419817D-03,   0.000000000000000D+00, &
  -2.806929534227890D-07,  -7.567150053719656D-07,   0.000000000000000D+00, &
  -1.368065866967344D-06,   0.000000000000000D+00,   5.715482705229559D-07, &
  -1.313951280381987D-06,   0.000000000000000D+00,   6.399748526212923D-07, &
   0.000000000000000D+00,   2.557951624114927D-07,   0.000000000000000D+00, &
   6.057192183901321D-07,   0.000000000000000D+00,  -1.309817527239991D-06, &
   2.442820512278462D-06,   0.000000000000000D+00,   2.515971383682502D-07, &
   0.000000000000000D+00,  -7.204321913031743D-07,   0.000000000000000D+00, &
   1.598158717909563D-06,   0.000000000000000D+00,  -2.872773182222276D-07, &
   0.000000000000000D+00,   1.459251909386494D-07,   0.000000000000000D+00, &
   1.831034309843729D-06,  -1.043801779275257D-06,   0.000000000000000D+00, &
  -3.748853617627662D-07,   0.000000000000000D+00,   1.677047090472612D-07, &
   0.000000000000000D+00,  -6.939092160382779D-07,   0.000000000000000D+00, &
  -6.466214868062944D-07,   0.000000000000000D+00,  -7.889580930710724D-07, &
   0.000000000000000D+00,  -8.103041940191536D-07,   0.000000000000000D+00, &
  -1.093272283971703D-06,   0.000000000000000D+00,   6.029054891363028D-06, &
  -4.984427668146395D-07,   0.000000000000000D+00,  -9.476404129725543D-07, &
   0.000000000000000D+00,   7.373893903028928D-08,   0.000000000000000D+00, &
   2.330232824619859D-06,   0.000000000000000D+00,   5.959008684499495D-06, &
   0.000000000000000D+00,  -1.787242197092850D-06,   0.000000000000000D+00, &
   1.041164048114889D-06,   0.000000000000000D+00,   2.135883659408610D-06, &
   0.000000000000000D+00,   1.845635624488169D-06,   0.000000000000000D+00, &
  -4.203012677273291D-06,   0.000000000000000D+00,  -5.658128997543573D-05, &
   9.025379484950011D-07,   0.000000000000000D+00,   1.868953804576533D-06, &
   0.000000000000000D+00,   6.775517765606700D-07,   0.000000000000000D+00, &
  -3.309378693699374D-06,   0.000000000000000D+00,  -8.388835624126092D-06, &
   0.000000000000000D+00,  -6.628138316261674D-06,   0.000000000000000D+00, &
   2.753110591513201D-06,   0.000000000000000D+00,  -6.058820523629014D-07, &
   0.000000000000000D+00,  -2.527031541791730D-06,   0.000000000000000D+00, &
  -2.927622901813304D-06,   0.000000000000000D+00,   3.057605538381726D-06, &
   0.000000000000000D+00,   2.800190375634929D-05,   0.000000000000000D+00, &
   1.873675878922278D-04,  -9.850992537798034D-07,   0.000000000000000D+00, &
  -1.520803040436786D-06,   0.000000000000000D+00,  -1.280674206227470D-06, &
   0.000000000000000D+00,   2.822123384024537D-06,   0.000000000000000D+00, &
   8.762629860204927D-06,   0.000000000000000D+00,   9.452709798816852D-06, &
   0.000000000000000D+00,  -9.411939897724168D-06,   0.000000000000000D+00, &
  -1.846509058561484D-06,   0.000000000000000D+00,   4.054007708041892D-07, &
   0.000000000000000D+00,   2.161217760526612D-06,   0.000000000000000D+00, &
   2.988859720824322D-06,   0.000000000000000D+00,  -1.294104001636407D-06, &
   0.000000000000000D+00,  -1.909087219745931D-05,   0.000000000000000D+00, &
  -8.320432199401562D-05,   0.000000000000000D+00,  -3.532327852065328D-04, &
   9.482273759172680D-07,   0.000000000000000D+00,   5.836896621135044D-07, &
   0.000000000000000D+00,   7.634186558166334D-07,   0.000000000000000D+00, &
  -2.128464275910774D-06,   0.000000000000000D+00,  -7.369647652996471D-06, &
   0.000000000000000D+00,  -9.568315135375556D-06,   0.000000000000000D+00, &
   8.706444330939097D-06,   0.000000000000000D+00,   1.099677165101271D-05, &
   0.000000000000000D+00,   5.190444273105669D-07,   0.000000000000000D+00, &
  -3.383566817946565D-07,   0.000000000000000D+00,  -1.462368955523958D-06, &
   0.000000000000000D+00,  -1.613211541747941D-06,   0.000000000000000D+00, &
   1.057470944427196D-06,   0.000000000000000D+00,   1.332036273777871D-05, &
   0.000000000000000D+00,   4.952177873356385D-05,   0.000000000000000D+00, &
   1.893955470519623D-04,   0.000000000000000D+00,   4.567683650986554D-04, &
  -4.994836153418678D-07,   0.000000000000000D+00,   9.102278564802990D-08, &
   0.000000000000000D+00,   7.989937425807684D-08,   0.000000000000000D+00, &
   1.420791556653482D-06,   0.000000000000000D+00,   4.477020215169327D-06, &
   0.000000000000000D+00,   5.622026519723173D-06,   0.000000000000000D+00, &
  -8.919460601262779D-06,   0.000000000000000D+00,  -5.950269077599141D-05, &
   0.000000000000000D+00,   3.817458379518016D-05,   0.000000000000000D+00, &
   5.809421318147445D-08,   0.000000000000000D+00,   6.533533824921976D-08, &
   0.000000000000000D+00,   6.192228605714879D-07,   0.000000000000000D+00, &
   1.583573112413638D-07,   0.000000000000000D+00,  -1.520581058025159D-06, &
   0.000000000000000D+00,  -8.107165224194627D-06,   0.000000000000000D+00, &
  -2.594616426360134D-05,   0.000000000000000D+00,  -9.343664376767805D-05, &
   0.000000000000000D+00,  -3.332104621389328D-04,   0.000000000000000D+00, &
  -4.953969989504616D-04,   7.945212939240710D-08,   0.000000000000000D+00, &
  -1.545000094334480D-07,   0.000000000000000D+00,  -2.723513224702747D-07, &
   0.000000000000000D+00,  -4.405258281047706D-07,   0.000000000000000D+00, &
  -1.489369245321459D-06,   0.000000000000000D+00,  -9.420917338988244D-07, &
   0.000000000000000D+00,   8.948426040676203D-06,   0.000000000000000D+00, &
   6.305580637129234D-05,   0.000000000000000D+00,   1.333613256171873D-04, &
   0.000000000000000D+00,  -7.614072702757303D-05,   0.000000000000000D+00, &
  -4.627848161172777D-08,   0.000000000000000D+00,   8.604684653992556D-08, &
   0.000000000000000D+00,  -1.369653542203758D-07,   0.000000000000000D+00, &
   1.748451466049415D-07,   0.000000000000000D+00,   1.025919284926382D-06, &
   0.000000000000000D+00,   3.015089013038335D-06,   0.000000000000000D+00, &
   7.983910107694168D-06,   0.000000000000000D+00,   1.747622106817088D-05, &
   0.000000000000000D+00,   1.366404058338821D-04,   0.000000000000000D+00, &
   4.153743257962606D-04,   0.000000000000000D+00,   5.052080112267632D-04 /)
lem(:,299) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.344472141383848D-05,  -7.792362492885448D-05, &
  -2.294222141308357D-04,  -4.270361095941767D-04,  -4.626635443805377D-04, &
  -7.547152856048195D-05,   4.932079182075902D-04,   3.629412530875298D-04, &
   1.262686317224810D-04,   5.531312900634196D-05,  -6.963333367634845D-04, &
  -1.649431206322900D-05,   6.465643671951960D-04,  -3.332658375856092D-04, &
  -9.741101158659648D-05,   1.650107033680689D-04,  -7.269121020787796D-05, &
   1.910598438380383D-05,  -3.835638011862515D-06,   1.667962755647525D-06, &
  -4.406080309384300D-06,   0.000000000000000D+00,   1.344472141382615D-05, &
   7.792362492885428D-05,   2.294222141308455D-04,   4.270361095941670D-04, &
   4.626635443805329D-04,   7.547152856049230D-05,  -4.932079182075896D-04, &
  -3.629412530875335D-04,  -1.262686317224814D-04,  -5.531312900634303D-05, &
   6.963333367634823D-04,   1.649431206322582D-05,  -6.465643671951959D-04, &
   3.332658375856162D-04,   9.741101158659161D-05,  -1.650107033680717D-04, &
   7.269121020788289D-05,  -1.910598438380267D-05,   3.835638011863093D-06, &
  -1.667962755651127D-06,   4.406080309382820D-06,   0.000000000000000D+00, &
   6.722360706919343D-06,   0.000000000000000D+00,  -1.245707580714603D-05, &
   0.000000000000000D+00,   2.738532116226510D-05,   0.000000000000000D+00, &
  -6.899731059758047D-05,   0.000000000000000D+00,   2.009505273521665D-04, &
   0.000000000000000D+00,  -6.680269871112318D-04,   0.000000000000000D+00, &
  -5.531286477696872D-04,   0.000000000000000D+00,   9.138457487147996D-03, &
   0.000000000000000D+00,  -2.162290796105441D-02,   0.000000000000000D+00, &
   2.240281357861489D-02,   0.000000000000000D+00,  -1.028069393391832D-02, &
   0.000000000000000D+00,   0.000000000000000D+00,  -5.314492771768276D-07, &
   0.000000000000000D+00,  -5.108704907288865D-07,   0.000000000000000D+00, &
   0.000000000000000D+00,  -4.079765584892631D-07,   0.000000000000000D+00, &
   1.238284014162916D-06,   0.000000000000000D+00,  -1.179257618466572D-07, &
   0.000000000000000D+00,   1.102828174971711D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.047584766579141D-07,   0.000000000000000D+00, &
   6.846892528045315D-07,   0.000000000000000D+00,  -3.001184174508703D-06, &
   0.000000000000000D+00,   8.840832968030072D-07,   0.000000000000000D+00, &
  -1.506207473702158D-07,   0.000000000000000D+00,  -2.358885739775644D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,   5.239968771462760D-07, &
   0.000000000000000D+00,  -1.111868444326970D-06,   0.000000000000000D+00, &
  -7.626377579533910D-07,   0.000000000000000D+00,   7.814022713753335D-06, &
   0.000000000000000D+00,  -1.030680374408774D-06,   0.000000000000000D+00, &
  -1.173110384751532D-06,   0.000000000000000D+00,   1.449768593354781D-06, &
   0.000000000000000D+00,   4.897988919119686D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,  -1.196711727161186D-06,   0.000000000000000D+00, &
   1.804379977310323D-07,   0.000000000000000D+00,   2.036331413730174D-06, &
   0.000000000000000D+00,  -8.977770351569670D-07,   0.000000000000000D+00, &
  -2.130062861756966D-05,   0.000000000000000D+00,   5.668643673620718D-07, &
   0.000000000000000D+00,   1.180414168515859D-06,   0.000000000000000D+00, &
   2.319368031140716D-07,   0.000000000000000D+00,  -4.070896680885770D-06, &
   0.000000000000000D+00,  -6.793103851962748D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   1.044966216780128D-06,   0.000000000000000D+00, &
   1.051007592540323D-06,   0.000000000000000D+00,  -7.250739069554818D-07, &
   0.000000000000000D+00,  -1.216474317641375D-06,   0.000000000000000D+00, &
   6.550627105671959D-06,   0.000000000000000D+00,   4.210869653175476D-05, &
   0.000000000000000D+00,   1.455546880962050D-07,   0.000000000000000D+00, &
  -3.388651442462359D-07,   0.000000000000000D+00,  -1.162616876306239D-06, &
   0.000000000000000D+00,   3.429847679729593D-07,   0.000000000000000D+00, &
   4.099732225133031D-06,   0.000000000000000D+00,   2.527331852262579D-06, &
   0.000000000000000D+00,   0.000000000000000D+00,  -8.334506727445517D-07, &
   0.000000000000000D+00,  -1.755795271741424D-06,   0.000000000000000D+00, &
  -9.961282604063157D-07,   0.000000000000000D+00,   1.054255063498632D-06, &
   0.000000000000000D+00,   1.112969546651137D-06,   0.000000000000000D+00, &
  -7.010130689297564D-06,   0.000000000000000D+00,   4.388504612744742D-06, &
   0.000000000000000D+00,  -3.470359757059061D-07,   0.000000000000000D+00, &
   1.548338689854494D-07,   0.000000000000000D+00,   1.477966531305627D-06, &
   0.000000000000000D+00,   2.582966090654734D-06,   0.000000000000000D+00, &
   4.543236987993985D-07,   0.000000000000000D+00,  -4.963033396294079D-06, &
   0.000000000000000D+00,  -2.599187542950752D-05,   0.000000000000000D+00, &
   0.000000000000000D+00,   7.394427431859491D-07,   0.000000000000000D+00, &
   1.364778307036290D-06,   0.000000000000000D+00,   9.962647222024809D-07, &
   0.000000000000000D+00,  -1.342223222125663D-06,   0.000000000000000D+00, &
  -5.605478381047594D-06,   0.000000000000000D+00,  -8.474822310805642D-06, &
   0.000000000000000D+00,  -1.038278837496025D-05,   0.000000000000000D+00, &
  -2.816241538500622D-04,   0.000000000000000D+00,  -2.304368801715395D-08, &
   0.000000000000000D+00,  -3.872636477350219D-07,   0.000000000000000D+00, &
  -1.076067231574940D-06,   0.000000000000000D+00,  -2.241542715737267D-06, &
   0.000000000000000D+00,  -6.298261314281319D-07,   0.000000000000000D+00, &
   8.754515179310236D-06,   0.000000000000000D+00,   3.890845121527766D-05, &
   0.000000000000000D+00,   1.851897491480064D-04,   0.000000000000000D+00, &
   0.000000000000000D+00,  -3.521944028294727D-07,   0.000000000000000D+00, &
  -4.137276604247899D-07,   0.000000000000000D+00,  -1.415464519824769D-07, &
   0.000000000000000D+00,   1.643240724021970D-06,   0.000000000000000D+00, &
   5.135784593481678D-06,   0.000000000000000D+00,   8.493977075866153D-06, &
   0.000000000000000D+00,   2.676548150715798D-06,   0.000000000000000D+00, &
   1.314908975431461D-05,   0.000000000000000D+00,   7.719733130104782D-04, &
   0.000000000000000D+00,   1.791287123716920D-07,   0.000000000000000D+00, &
   2.015267035043524D-07,   0.000000000000000D+00,   2.780589276850175D-07, &
   0.000000000000000D+00,   3.702951256779039D-07,   0.000000000000000D+00, &
  -1.188011830721333D-06,   0.000000000000000D+00,  -8.717654167073194D-06, &
   0.000000000000000D+00,  -3.126077143674168D-05,   0.000000000000000D+00, &
  -1.310503867167950D-04,   0.000000000000000D+00,  -4.891925431505729D-04, &
   0.000000000000000D+00,   0.000000000000000D+00,   1.229422848062403D-08, &
   0.000000000000000D+00,  -6.995699555970522D-09,   0.000000000000000D+00, &
  -9.786708176573565D-08,   0.000000000000000D+00,  -9.271606312051729D-07, &
   0.000000000000000D+00,  -1.960809740361586D-06,   0.000000000000000D+00, &
  -1.939707519595813D-06,   0.000000000000000D+00,   7.720540967828400D-06, &
   0.000000000000000D+00,   6.281820705151257D-05,   0.000000000000000D+00, &
   3.660296389516593D-05,   0.000000000000000D+00,  -1.069247761013141D-03, &
   0.000000000000000D+00,  -4.870024534986803D-08,   0.000000000000000D+00, &
   2.514544002121169D-08,   0.000000000000000D+00,   4.236021137063723D-08, &
   0.000000000000000D+00,   3.905933339953730D-07,   0.000000000000000D+00, &
   1.065001712492768D-06,   0.000000000000000D+00,   3.966852875967855D-06, &
   0.000000000000000D+00,   1.151241284388782D-05,   0.000000000000000D+00, &
   3.258014902213019D-05,   0.000000000000000D+00,   2.419741271286598D-04, &
   0.000000000000000D+00,   6.773253901489880D-04,   0.000000000000000D+00 /)
lem(:,300) = (/ 0.000000000000000D+00,   0.000000000000000D+00,   0.000000000000000D+00, &
  -2.024546728527013D-05,  -7.707000963832286D-05,  -1.687023605847237D-04, &
  -2.513300870401680D-04,  -2.087326408705703D-04,   1.157005934204661D-04, &
   7.014996723505375D-04,   9.690195975733130D-04,  -5.743456406661187D-05, &
  -1.789278913119071D-03,  -6.624771650578121D-04,   1.861743074990350D-03, &
   1.766099774378761D-04,  -1.121955112211936D-03,   4.427269701623289D-04, &
   1.525493620321472D-04,  -2.263979354895397D-04,   7.880298467483956D-05, &
  -1.483766951530606D-05,   1.524180589357915D-05,   4.964322455036438D-06, &
  -1.423199452195497D-06,  -2.024546728527375D-05,  -7.707000963833351D-05, &
  -1.687023605847193D-04,  -2.513300870401632D-04,  -2.087326408705785D-04, &
   1.157005934204637D-04,   7.014996723505441D-04,   9.690195975733211D-04, &
  -5.743456406660963D-05,  -1.789278913119076D-03,  -6.624771650578302D-04, &
   1.861743074990365D-03,   1.766099774378851D-04,  -1.121955112211948D-03, &
   4.427269701623386D-04,   1.525493620321422D-04,  -2.263979354895589D-04, &
   7.880298467484695D-05,  -1.483766951528733D-05,   1.524180589357600D-05, &
   4.964322455016449D-06,  -1.423199452194552D-06,   0.000000000000000D+00, &
   0.000000000000000D+00,   4.757763400729286D-06,   0.000000000000000D+00, &
  -1.220361359534430D-05,   0.000000000000000D+00,   2.971361829867939D-05, &
   0.000000000000000D+00,  -7.911943288481939D-05,   0.000000000000000D+00, &
   2.481212306238817D-04,   0.000000000000000D+00,  -1.004114624384731D-03, &
   0.000000000000000D+00,   6.260171143938123D-03,   0.000000000000000D+00, &
  -2.273254435447944D-02,   0.000000000000000D+00,   4.278965703332629D-02, &
   0.000000000000000D+00,  -4.390580171783246D-02,   0.000000000000000D+00, &
   2.119276258673093D-08,   6.357828776031219D-08,   0.000000000000000D+00, &
   1.420613755821542D-07,   0.000000000000000D+00,  -3.873501359613433D-07, &
   2.231062259136308D-07,   0.000000000000000D+00,  -3.823430546911281D-07, &
   0.000000000000000D+00,   2.129015190540212D-07,   0.000000000000000D+00, &
  -3.213963956915607D-07,   0.000000000000000D+00,   1.281172364622298D-06, &
  -3.793949106443673D-08,   0.000000000000000D+00,  -1.157927240537006D-07, &
   0.000000000000000D+00,   1.182883253424776D-06,   0.000000000000000D+00, &
  -5.657578403857180D-07,   0.000000000000000D+00,   2.828974453348401D-07, &
   0.000000000000000D+00,   7.943812505905118D-07,   0.000000000000000D+00, &
  -3.580799979022173D-06,  -8.827600723656708D-07,   0.000000000000000D+00, &
   6.819448853216357D-07,   0.000000000000000D+00,  -1.431034148263159D-07, &
   0.000000000000000D+00,  -3.145183577682122D-06,   0.000000000000000D+00, &
   9.473041582912668D-08,   0.000000000000000D+00,   4.798106770239020D-07, &
   0.000000000000000D+00,  -1.480566136192586D-06,   0.000000000000000D+00, &
  -1.607243383390752D-06,   0.000000000000000D+00,   1.028783732850875D-05, &
   2.024491045792465D-06,   0.000000000000000D+00,  -7.384753372477850D-07, &
   0.000000000000000D+00,  -1.852281752734435D-06,   0.000000000000000D+00, &
   1.645191864597767D-06,   0.000000000000000D+00,   8.360165568857964D-06, &
   0.000000000000000D+00,   4.448151268220333D-07,   0.000000000000000D+00, &
  -1.394129652217666D-06,   0.000000000000000D+00,   7.076486756884091D-07, &
   0.000000000000000D+00,   4.804607840005181D-06,   0.000000000000000D+00, &
   2.032222140240467D-06,   0.000000000000000D+00,  -3.315102847995812D-05, &
  -2.646067785838592D-06,   0.000000000000000D+00,   9.103401171478346D-07, &
   0.000000000000000D+00,   3.477724852601323D-06,   0.000000000000000D+00, &
   1.719138296895295D-06,   0.000000000000000D+00,  -8.836786411486896D-06, &
   0.000000000000000D+00,  -2.275960160571696D-05,   0.000000000000000D+00, &
   7.165172112691862D-08,   0.000000000000000D+00,   1.625969159417252D-06, &
   0.000000000000000D+00,  -3.922200689119510D-08,   0.000000000000000D+00, &
  -6.043534688275249D-06,   0.000000000000000D+00,  -9.659653955060580D-06, &
   0.000000000000000D+00,   7.547535571339693D-06,   0.000000000000000D+00, &
   1.281514507595194D-04,   1.970926920573156D-06,   0.000000000000000D+00, &
  -1.507073240547787D-06,   0.000000000000000D+00,  -3.762471067791737D-06, &
   0.000000000000000D+00,  -3.128121039980161D-06,   0.000000000000000D+00, &
   7.382523596173003D-06,   0.000000000000000D+00,   2.942325457436106D-05, &
   0.000000000000000D+00,   4.443720810943157D-05,   0.000000000000000D+00, &
  -7.594073260275182D-07,   0.000000000000000D+00,  -6.603078721838386D-07, &
   0.000000000000000D+00,   3.260155309080622D-07,   0.000000000000000D+00, &
   5.488669231007688D-06,   0.000000000000000D+00,   1.063996492366813D-05, &
   0.000000000000000D+00,   1.750372395555585D-07,   0.000000000000000D+00, &
  -6.503969559800011D-05,   0.000000000000000D+00,  -4.776578586272621D-04, &
  -5.934934101903269D-07,   0.000000000000000D+00,   1.420027881361762D-06, &
   0.000000000000000D+00,   2.452398683050677D-06,   0.000000000000000D+00, &
   1.494540458556055D-06,   0.000000000000000D+00,  -6.477564558683739D-06, &
   0.000000000000000D+00,  -2.564767711562584D-05,   0.000000000000000D+00, &
  -4.402184232633596D-05,   0.000000000000000D+00,   5.416087796904291D-06, &
   0.000000000000000D+00,   5.414175167647428D-07,   0.000000000000000D+00, &
  -4.040790615962174D-07,   0.000000000000000D+00,  -4.557033766682066D-07, &
   0.000000000000000D+00,  -3.121747230011919D-06,   0.000000000000000D+00, &
  -5.651786229106209D-06,   0.000000000000000D+00,   2.010091160529620D-06, &
   0.000000000000000D+00,   4.645335977140615D-05,   0.000000000000000D+00, &
   2.099664665947903D-04,   0.000000000000000D+00,   1.316518699882476D-03, &
  -8.466372747589597D-08,   0.000000000000000D+00,  -4.828262506823343D-07, &
   0.000000000000000D+00,  -9.551687684559049D-07,   0.000000000000000D+00, &
   3.739520199132799D-07,   0.000000000000000D+00,   4.531200473689555D-06, &
   0.000000000000000D+00,   1.419191310154802D-05,   0.000000000000000D+00, &
   2.131792182067722D-05,   0.000000000000000D+00,  -2.642881452373784D-05, &
   0.000000000000000D+00,  -2.963054203481746D-04,   0.000000000000000D+00, &
  -8.308804177147084D-08,   0.000000000000000D+00,   4.399452641729985D-07, &
   0.000000000000000D+00,   6.002438218579532D-08,   0.000000000000000D+00, &
   5.250226006794714D-07,   0.000000000000000D+00,   6.987047132072353D-07, &
   0.000000000000000D+00,  -4.277561160953880D-06,   0.000000000000000D+00, &
  -2.710820226473692D-05,   0.000000000000000D+00,  -9.763519935865877D-05, &
   0.000000000000000D+00,  -3.918773457928461D-04,   0.000000000000000D+00, &
  -2.467623654120907D-03,   5.846759402607434D-08,   0.000000000000000D+00, &
  -8.141594408188327D-08,   0.000000000000000D+00,   4.139275961628620D-07, &
   0.000000000000000D+00,  -6.219407894362764D-07,   0.000000000000000D+00, &
  -1.775872287288028D-06,   0.000000000000000D+00,  -3.788815865392385D-06, &
   0.000000000000000D+00,  -2.129273875834180D-06,   0.000000000000000D+00, &
   3.212185655106162D-05,   0.000000000000000D+00,   2.417675622014313D-04, &
   0.000000000000000D+00,   8.073607663216225D-04,   0.000000000000000D+00, &
  -1.909176252189519D-08,   0.000000000000000D+00,  -7.456678039006511D-08, &
   0.000000000000000D+00,   4.458671694458606D-08,   0.000000000000000D+00, &
   5.711932548426334D-07,   0.000000000000000D+00,   6.146501992430868D-07, &
   0.000000000000000D+00,   2.447183432879596D-06,   0.000000000000000D+00, &
   9.079582407070105D-06,   0.000000000000000D+00,   2.594957756043337D-05, &
   0.000000000000000D+00,   6.372577149190007D-05,   0.000000000000000D+00, &
   5.052080112267632D-04,   0.000000000000000D+00,   3.126811269506710D-03 /)
end subroutine init_lem

end module laplace_elem_mat
